package com.focustech.baseandroid;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import android.app.Application;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.database.sqlite.SQLiteDatabase;
import android.net.ConnectivityManager;
import android.os.Bundle;
import android.os.SystemClock;
import android.view.ViewGroup;

import com.focustech.baseandroid.activity.BaseActivity;
import com.focustech.baseandroid.bean.IntentBean;
import com.focustech.baseandroid.db.bean.DBean;
import com.focustech.baseandroid.general.BaseAnnotation;
import com.focustech.baseandroid.general.BaseAnnotation.isCatchException;
import com.focustech.baseandroid.listener.OnLifeCycleChangeListener.LifeCycle;
import com.focustech.baseandroid.manager.ACache;
import com.focustech.baseandroid.manager.NetStateManager;
import com.focustech.baseandroid.manager.UnCeManager;
import com.focustech.baseandroid.manager.NetStateManager.NetState;
import com.focustech.baseandroid.net.util.NetUtils;

/**
 * 整个App 可以对全局进行操作，是所有类中权利最大的，使用它需要在AndroidManifest.xml进行声明 否者会报错
 * 
 * @author zhangzheng
 * 
 */
public class BaseApp extends Application {

	@Override
	public void onCreate() {
		super.onCreate();
		init(this);
		initDatasAndListener();
		catchException();
	}

	// 捕获异常
	protected void catchException() {
		isCatchException annotation = BaseAnnotation.getAnnotation(this, isCatchException.class);
		if (annotation != null) {
			UnCeManager catchExcep = new UnCeManager(this,annotation);
			Thread.setDefaultUncaughtExceptionHandler(catchExcep);
		}
	}

	private void initDatasAndListener() {
		initBroadCast();
	}

	private void initBroadCast() {
		IntentFilter mFilter = new IntentFilter();
		mFilter.addAction(ConnectivityManager.CONNECTIVITY_ACTION);
		registerReceiver(new NetReceiver(), mFilter);
	}

	private List<BaseActivity> recodeAcs = new ArrayList<>();

	private BaseActivity currenActivity;

	/**
	 * 获得当前正在显示的Activity
	 * 
	 * @return
	 */
	public BaseActivity getCurrenActivity() {
		if (currenActivity != null && !currenActivity.isRunning) {
			return null;
		}
		return currenActivity;
	}

	public boolean currenActivityIsRunning() {
		if (currenActivity == null) {
			return false;
		}
		return currenActivity.isRunning;
	}

	public LifeCycle currenActivityLife() {
		if (currenActivity == null) {
			return null;
		}
		return currenActivity.currenLifeCycle;
	}

	/**
	 * 记录Activity不建议手动调用
	 * 
	 * @param mActivity
	 */
	public void recodeActivity(BaseActivity mActivity) {
		if (recodeAcs.contains(mActivity)) {
			recodeAcs.remove(mActivity);
		}
		recodeAcs.add(mActivity);
		currenActivity = mActivity;
	}

	public BaseActivity getFirstActivity() {
		return recodeAcs.get(0);
	}

	/**
	 * 取消记录 不建议手动调用
	 * 
	 * @param mActivity
	 */
	public void unRecodeActivity(BaseActivity mActivity) {
		recodeAcs.remove(mActivity);
		if (recodeAcs.size() == 0) {
			currenActivity = null;
			onDestory();
		} else {
			currenActivity = recodeAcs.get(recodeAcs.size() - 1);
		}
		clearAIntentRecoder(mActivity);
	}

	/**
	 * 所有Activity结束，一般情况下Application不会立马结束，所以可以在 该回调中处理些回收资源的事情
	 */
	protected void onDestory() {
		DBean.forcedClosed();
		android.os.Process.killProcess(android.os.Process.myPid());
	}

	/**
	 * 远程结束Activity
	 * 
	 * @param mActivitys
	 */
	public void remoteFinish(BaseActivity... mActivitys) {
		if (mActivitys == null) {
			return;
		}
		if (mActivitys.length > 1) {
			for (BaseActivity baseActivity : mActivitys) {
				remoteFinish(baseActivity);
			}
			return;
		}
		BaseActivity mActivity = mActivitys[0];
		if (recodeAcs.contains(mActivity)) {
			mActivity.finish();
			unRecodeActivity(mActivity);
		}
	}

	/**
	 * 删除所有Activity退出应用
	 */
	public void remoteAllFinish() {
		remoteFinish(recodeAcs.toArray(new BaseActivity[recodeAcs.size()]));
	}

	private static BaseApp application;

	public static <T extends BaseApp> void init(T application) {
		BaseApp.application = application;
	}

	@SuppressWarnings("unchecked")
	public static <T extends BaseApp> T getAPP() {
		return (T) application;
	}
	
	private static Context context;
	
	public static void init(Context context){
		BaseApp.context=context;
	}
	
	public static Context getContext() {
		return context;
	}

	/**
	 * 数据库升级
	 * 
	 * @param tableName
	 *            表名
	 * @param db
	 *            数据库
	 * @param oldVersion
	 *            老数据库版本
	 * @param newVersion
	 *            新数据库版本
	 * @return true表示已经被处理了，不会继续向下传递
	 */
	public boolean onDbUpgrade(String tableName, SQLiteDatabase db,
			int oldVersion, int newVersion) {
		return false;
	}

	/**
	 * 数据库降级
	 * 
	 * @param tableName
	 *            表名
	 * @param db
	 *            数据库
	 * @param oldVersion
	 *            老数据库版本
	 * @param newVersion
	 *            新数据库版本
	 * @return true表示已经被处理了，不会继续向下传递
	 */
	public boolean onDbDowngrade(String tableName, SQLiteDatabase db,
			int oldVersion, int newVersion) {
		return false;
	}

	/**
	 * 数据库升级成功
	 */
	public void onDbUpdateCompute() {

	}

	private boolean isconnection = true;

	/**
	 * 获得当前的网络状态
	 * 
	 * @return
	 */
	public NetState getCurrenNetState() {
		boolean isConnected = NetUtils.isNetworkConnected(this);
		if (!isConnected) {
			return NetState.DISCONNECTION;
		}

		if (NetUtils.isWifiConnected(this)) {
			return NetState.WIFI;
		}

		if (NetUtils.isMobileConnected(this)) {
			return NetState.MOBILE;
		}

		return NetState.DISCONNECTION;
	}

	public class NetReceiver extends BroadcastReceiver {

		@Override
		public void onReceive(Context context, Intent intent) {
			String action = intent.getAction();
			if (ConnectivityManager.CONNECTIVITY_ACTION.equals(action)) {
				boolean isConnected = NetUtils.isNetworkConnected(context);
				if (!isConnected) {
					NetStateManager.getInstance().post(NetState.DISCONNECTION);
					isconnection = false;
					return;
				}

				if (!isconnection) {
					NetStateManager.getInstance().post(NetState.RECONNECTION);
				}

				isconnection = true;

				if (NetUtils.isWifiConnected(context)) {
					NetStateManager.getInstance().post(NetState.WIFI);
				}

				if (NetUtils.isMobileConnected(context)) {
					NetStateManager.getInstance().post(NetState.MOBILE);
				}

			}

		}

	}

	private LinkedList<IntentBean> intentBeans = new LinkedList<>();

	public void recodeIntent(Class<? extends BaseActivity> cla,
			Integer requestCode, Bundle bundler) {
		IntentBean ib = new IntentBean();
		ib.setBundler(bundler);
		ib.setClaName(cla.getName());
		ib.setRequestCode(requestCode);
		intentBeans.add(ib);
	}

	private void clearAIntentRecoder(BaseActivity mActivity) {
		IntentBean clearIntentBean = null;
		for (IntentBean bean : intentBeans) {
			if (bean.getClaName().equals(mActivity.getClass().getName())) {
				clearIntentBean = bean;
				break;
			}
		}
		intentBeans.remove(clearIntentBean);
	}

	public void clearLocalIntentRecode(String saveItbeanKeys) {
		saveItbeanKeys = "itBean" + saveItbeanKeys;
		ACache.get(this).remove(saveItbeanKeys);
	}

	public void saveIntentRecodeLocal(String saveItbeanKeys) {
		saveItbeanKeys = "itBean" + saveItbeanKeys;
		IntentBean[] itbeans = new IntentBean[intentBeans.size()];
		for (int i = 0; i < itbeans.length; i++) {
			itbeans[i] = intentBeans.get(i);
		}

		ACache.get(this).put(saveItbeanKeys, itbeans);
	}

	public void recoveryIntentRecode(String saveItbeanKeys) {
		saveItbeanKeys = "itBean" + saveItbeanKeys;
		IntentBean[] itbeans = (IntentBean[]) ACache.get(this).getAsObject(
				saveItbeanKeys);
		intentBeans.clear();
		for (IntentBean intentBean : itbeans) {
			intentBeans.add(intentBean);
		}
	}

	public IntentBean getIntentRecode() {
		return intentBeans.removeLast();
	}

}