package com.focustech.baseandroid.activity;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;

import android.content.Intent;
import android.os.Bundle;
import android.support.v4.app.FragmentActivity;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;

import com.focustech.baseandroid.bean.IntentBean;
import com.focustech.baseandroid.general.BaseAnnotation.isClickMethod;
import com.focustech.baseandroid.listener.OnBaseClickListener;
import com.focustech.baseandroid.listener.OnLifeCycleChangeListener;
import com.focustech.baseandroid.listener.OnLifeCycleChangeListener.LifeCycle;
import com.focustech.baseandroid.manager.ControlManager;
import com.focustech.baseandroid.manager.EventBus;
import com.focustech.baseandroid.manager.LifeCycleManager;
import com.focustech.baseandroid.manager.NetStateManager;
import com.focustech.baseandroid.net.Response;
import com.focustech.baseandroid.net.params.Url.onRequestListener;
import com.focustech.baseandroid.util.BaseViewUtils;

import static com.focustech.baseandroid.general.BaseAnnotation.getAnnotation;
import static com.focustech.baseandroid.general.BaseAnnotation.getAnnotationMethods;
import static com.focustech.baseandroid.general.BaseAnnotation.setContentView;

/**
 * 基础的Activity类 理论上所有的Activity都应该继承该Activity 有自己的生命周期回调接口，可以在其他的类中监听的到生命周期的回调
 * 支持注解 设置ContentView 注解指定点击事件类 注解指定网络访问 回调类，注解指定EventBus等 带有Holoer的fv()找控件的方法
 */
public abstract class BaseActivity extends FragmentActivity implements
		onRequestListener<Response> {

	protected ViewGroup rootView;

	public LifeCycleManager lifeCycleManager;

	protected BaseActivity mActivity;
	
	public LifeCycle currenLifeCycle;

	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		currenLifeCycle=LifeCycle.ONCREATE;
		mActivity = this;
		isRunning = true;
		EventBus.getInstance().register(this);
		NetStateManager.getInstance().register(this);
		rootView = (ViewGroup) LayoutInflater.from(this).inflate(
				getAnnotation(this, setContentView.class).value(), null);
		setContentView(rootView);

		initViews(rootView);
		initListener();
		initDatas(savedInstanceState);

		lifeCycleManager = new LifeCycleManager();

		setListenerByAnnotation();

		Log.i("Activity", getClass().getSimpleName());
		
	}


	private void setListenerByAnnotation() {
		List<Method> methods = getAnnotationMethods(mActivity,
				isClickMethod.class);
		for (Method method : methods) {
			isClickMethod annotation = method
					.getAnnotation(isClickMethod.class);
			int[] value = annotation.value();
			if (value == null || value.length == 0) {
				continue;
			}
			setClickListenerView(value);
		}
	}

	public void registerLifeCycle(OnLifeCycleChangeListener lis) {
		mActivity.registerLifeCycle(lis);
	}

	/**
	 * 初始化视图
	 * 
	 * @param rootView
	 *            根View
	 */
	protected abstract void initViews(ViewGroup rootView);

	/**
	 * 初始化数据
	 * 
	 * @param savedInstanceState
	 * 
	 */
	protected abstract void initDatas(Bundle savedInstanceState);

	/**
	 * 初始化监听
	 */
	protected abstract void initListener();

	public void setClickListenerView(int... ids) {
		for (int id : ids) {
			setClickListenerView(id);
		}
	}

	public void setClickListenerView(View... views) {
		for (View view : views) {
			setClickListenerView(view);
		}
	}

	public void setClickListenerView(int id) {
		setClickListenerView(fv(id));
	}

	public void setClickListenerView(View view) {
		List<Method> annotationMethods = getAnnotationMethods(this,
				isClickMethod.class);
		if (annotationMethods != null && annotationMethods.size() > 0) {
			view.setOnClickListener(new ViewListener(view, annotationMethods));
		}
	}

	private class ViewListener extends OnBaseClickListener{

		private List<Method> annotationMethods;
		private View view;

		public ViewListener(View view, List<Method> annotationMethods) {

			this.annotationMethods = annotationMethods;
			this.view = view;
		}

		@Override
		public void onViewClick(View v) {
			for (Method annotationMethod : annotationMethods) {

				try {
					isClickMethod annotation = annotationMethod
							.getAnnotation(isClickMethod.class);
					int[] ids = annotation.value();
					List<Integer> idList = new ArrayList<>();
					for (int id : ids) {
						idList.add(id);
					}
					if (ids == null || ids.length == 0
							|| idList.contains(v.getId())) {
						annotationMethod.invoke(BaseActivity.this, view);
					}
				} catch (IllegalAccessException e) {
					e.printStackTrace();
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
				} catch (InvocationTargetException e) {
					throw new RuntimeException(e);
				} 

			}
		}
	}

	public boolean isRunning = false;

	@Override
	protected void onStart() {
		super.onStart();
		lifeCycleManager.notifyLifeCycleChange(LifeCycle.START);
		currenLifeCycle=LifeCycle.START;
	}

	protected void onResume() {
		super.onResume();
		isRunning = true;
		lifeCycleManager.notifyLifeCycleChange(LifeCycle.ONRESUM);
		currenLifeCycle=LifeCycle.ONRESUM;
	}

	protected void onPause() {
		super.onPause();
		lifeCycleManager.notifyLifeCycleChange(LifeCycle.PAURSE);
		currenLifeCycle=LifeCycle.PAURSE;
		
	}

	@Override
	protected void onStop() {
		super.onStop();
		isRunning = false;
		lifeCycleManager.notifyLifeCycleChange(LifeCycle.STOP);
		currenLifeCycle=LifeCycle.STOP;
	}

	@Override
	protected void onDestroy() {
		super.onDestroy();
		isRunning = false;
		lifeCycleManager.notifyLifeCycleChange(LifeCycle.DESTROY);
		releaseMemory();
		EventBus.getInstance().unRegister(this);
		NetStateManager.getInstance().unRegister(this);
		currenLifeCycle=LifeCycle.DESTROY;
	}

	protected void releaseMemory() {
		HashMap<Integer, View> viewsHashMap = (HashMap) rootView.getTag();
		if (viewsHashMap != null) {
			viewsHashMap.clear();
			rootView.setTag(null);
		}
	}

	public <T extends View> T fv(int id) {
		return ControlManager.fv(rootView, id);
	}

	public <T extends View> T fv(int id, Class<T> cla) {
		return fv(id);
	}

	public void showProgressDialog() {

	}

	public void dismissProgressDialog() {

	}

	public void showToast(String content) {
		BaseViewUtils.showToast(this, content);
	}

	public void showToast(int contentId) {
		BaseViewUtils.showToast(this, contentId);
	}

	/**
	 * 跳转到指定的页面
	 * 
	 * @param cla
	 * @param requestCode
	 * @param bundler
	 */
	public void turnTo(Class<? extends BaseActivity> cla, Integer requestCode,
			Bundle bundler) {
		ControlManager.turnTo(cla, requestCode, bundler, mActivity);
	}

	public void turnTo(Class<? extends BaseActivity> cla) {
		turnTo(cla, null, null);
	}

	public void turnTo(Class<? extends BaseActivity> cla, Bundle bundler) {
		turnTo(cla, null, bundler);
	}

	@Override
	public void onCompute(Response resp, Class cla) {
		ControlManager.onCompute(this, resp, cla);
	}

	/**
	 * EventBus进行推送
	 * 
	 * @param mClass
	 *            推送的类
	 * @param type
	 *            推送的类型 用于识别相同类，相同参数的情况
	 * @param parms
	 *            推送方法的参数
	 */
	public void eventPost(final Class mClass, final Integer type,
			final Object... parms) {
		EventBus.getInstance().post(mClass, type, parms);

	}

	public void eventPost(Class mClass, Object... parms) {
		EventBus.getInstance().post(mClass, parms);
	}

	/**
	 * 强制EventBus进行推送提交 忽略类型只要类和参数满足就全部提交
	 */
	public void forcedEventPost(Class mClass, Object... parms) {
		EventBus.getInstance().forcedPost(mClass, parms);
	}
}
