package com.focustech.baseandroid.bean;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;

import android.content.ContentValues;
import android.database.Cursor;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.focustech.baseandroid.BaseApp;
import com.focustech.baseandroid.bean.interfaces.BaseBeanInterFace;
import com.focustech.baseandroid.db.SqliteHelp;
import com.focustech.baseandroid.db.interfaces.DbAble;
import com.focustech.baseandroid.general.BaseAnnotation.isPrimaryKey;
import com.focustech.baseandroid.manager.ACache;
import com.focustech.baseandroid.util.BaseStringUtils;
import com.focustech.baseandroid.util.PropertyDescriptor;

/**
 * 所有的数据模型类的基类，设计这个基类将会更容易的对某些具有高级功能的数据模型，进行拓展和修改
 * 数据库操作的数据模型，和网络访问的数据模型都属于高级功能数据模型类
 * @author zhangzheng
 *
 */
public class BaseBean implements BaseBeanInterFace{
	
	/**
	 * 转换字符串的null值为空值("")
	 */
	@Override
	public void conversionStringNull(){
		Field[] declaredFields = getClass().getDeclaredFields();
		for (Field field : declaredFields) {
			Method getMethod = PropertyDescriptor.getGetMethod(getClass(), field.getName());
			if(getMethod!=null){
				field.setAccessible(true);
				try {
					Object object = field.get(this);
					if (field.getGenericType().toString().equals(  
	                        "class java.lang.String")&&object==null){
						field.set(this, "");
					}
				} catch (IllegalAccessException e) {
					e.printStackTrace();
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
				}
			}
		}
	}

	/**
	 * 用自己本身复制出来一个对象，当然这个方法是比较费性能的，但是如果在具体的子类中
	 * 复写该方法，自己去创建对象，将会为系统节省大量资源
	 * @param bean 复制的数据模型对象
	 * @return
	 */
	@Override
	public <T extends BaseBeanInterFace> T copy() {

		try {
			Constructor<? extends BaseBean> constructor = this.getClass()
					.getConstructor();
			@SuppressWarnings("unchecked")
			T copyBean = (T) constructor.newInstance();
//			String jsonString = this.toJSONString();
//			copyBean.parseJsonToBean(jsonString);
			return copyBean;
		} catch (NoSuchMethodException e) {
			return null;
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (InvocationTargetException e) {
			e.printStackTrace();
		}

		return null;
	}
	
	/**
	 * 根据其他对象填充自己，你可能需要根据其他的数据，为自己填充数据
	 * 它会将变量名相同的值，赋值与你，但是你自己需要保证这是有意义的
	 * @param bean BaseBean的子类
	 */
	@Override
	@SuppressWarnings("unchecked")
	public <T extends BaseBean> T fillData(T bean){
		parseJsonToBean(bean.toJSONString());
		return (T) this;
	}

	/**
	 * 将bean转换成JSON
	 * 
	 * @return
	 */
	@Override
	public String toJSONString() {
		return JSON.toJSONString(this);
	}

	/**
	 * 将Json字符串用于填充自己
	 * 
	 * @param json
	 */
	@Override
	public boolean parseJsonToBean(String json) {
		BaseBean bean = JSON.parseObject(json, getClass());
		if(bean==null){
			return false;
		}
		Field[] declaredFields = getClass().getDeclaredFields();
		Class<? extends BaseBean> cla = bean.getClass();
		for (Field field : declaredFields) {
			Method setMethod = PropertyDescriptor.getSetMethod(getClass(),
					field.getName());
			Method getMethod = PropertyDescriptor.getGetMethod(cla,
					field.getName());
			if (setMethod != null ) {
				if (getMethod != null ) {
					try {
						setMethod.invoke(this, getMethod.invoke(bean));
					} catch (IllegalAccessException e) {
						e.printStackTrace();
					} catch (IllegalArgumentException e) {
						e.printStackTrace();
					} catch (InvocationTargetException e) {
						e.printStackTrace();
					}
				}
			}
		}
		
		return true;
	}
	
	/**
	 * 根据游标填充自己，操作数据库的时候可以排上大用场，当然如果你复写这个方法
	 * 可以让性能更上一个台阶
	 */
	public void  fillData(Cursor query){
		String[] columnNames = query.getColumnNames();
		for (String columnName : columnNames) {
			String fieldvalue = query.getString(query
					.getColumnIndex(columnName));
			PropertyDescriptor.invokeSet(this, columnName, fieldvalue);
		}
	}
	
	
	/**
	 * 根据自己的实例构建成ContentValues，用于数据库操作，这是默认实现
	 * 利用反射填充数据，如果你对性能要求较高，建议复写该方法，但是如果仅仅操作
	 * 几条数据，我觉得没必要为十几毫秒较真
	 */
	public ContentValues getContentValues() {
		
		ContentValues values = new ContentValues();
		
		if(!(this instanceof DbAble)){
			return values;
		}

		List<Field> usedFields = SqliteHelp.getUsedFields((DbAble)this);

		for (Field field : usedFields) {
			String fieldName = field.getName();
			Object value = PropertyDescriptor.invokeGet(this, fieldName);
			if (value == null || value.equals(0)) {
				continue;
			}
			boolean isPrimary = field.getAnnotation(isPrimaryKey.class) != null;
			if (value != null && !isPrimary) {
				values.put(fieldName, String.valueOf(value));
			}
		}

		return values;
	}
	
	/**
	 * 保存实例到本地，将实例转换成JSon数据，所以不需要序列化
	 */
	@Override
	public void saveLocation(String key){
		ACache.get(BaseApp.getAPP()).put(key, toJSONString());
	}
	
	/**
	 * 从本地读取缓存填充自己
	 */
	@Override
	public boolean fillDataFromLocation(String key){
		String json = ACache.get(BaseApp.getAPP()).getAsString(key);
		if(BaseStringUtils.isNull(json)){
			return false;
		}
		return parseJsonToBean(json);
	}
}
