package com.focustech.baseandroid.db;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import com.focustech.baseandroid.db.interfaces.DbAbilityInterface;
import com.focustech.baseandroid.db.interfaces.DbAble;
import com.focustech.baseandroid.general.BaseAnnotation;
import com.focustech.baseandroid.general.BaseAnnotation.isIgnore;
import com.focustech.baseandroid.general.BaseAnnotation.isPrimaryKey;
import com.focustech.baseandroid.util.PropertyDescriptor;

import android.content.ContentValues;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;

/**
 * 数据库的帮助类
 * 
 * @author zhangzheng
 * 
 * @param <T>
 */
public class SqliteHelp<T extends DbAble> implements DbAbilityInterface<T> {

	private final SQLiteDatabase db;
	private final T bean;

	public SqliteHelp(T bean) {
		this.db = bean.getDb();
		this.bean = bean;
	}

	/**
	 * 插入数据
	 */
	@Override
	public void insert() {
		db.insert(bean.getTabName(), null, bean.getContentValues());
	}

	@Override
	public void update(T updateBean) {
		ContentValues values = updateBean.getContentValues();
		String whereClause = getSqlCondition(bean);
		db.update(bean.getTabName(), values, whereClause, null);
	}

	@Override
	public void update(String fieldName, String FieldValue) {
		ContentValues values = bean.getContentValues();
		db.update(bean.getTabName(), values, fieldName + " = ?",
				new String[] { FieldValue });
	}

	@Override
	public void delete() {

		bean.getDb().delete(bean.getTabName(), getSqlCondition(bean), null);
	}

	private String getSqlCondition(T bean) {
		StringBuilder whereClause = new StringBuilder();

		List<Field> usedFields = getUsedFields(bean);
		for (Field field : usedFields) {
			Method getMethod = PropertyDescriptor.getGetMethod(bean.getClass(),
					field.getName());
			if (getMethod != null) {
				try {
					Object value = PropertyDescriptor.invokeGet(bean,
							field.getName());
					if (value == null || value.equals(0)) {
						continue;
					}

					if ((value + "").contains("||")) {
						String[] split = (value + "").split("\\|\\|");
						String name = field.getName();
						StringBuilder result = new StringBuilder();
						for (int i = 0; i < split.length; i++) {
							String sv = split[i];
							sv = "'" + sv + "'";
							if (i != 0) {
								result.append(" OR " + name + " = " + sv);
							} else {
								result.append(sv);
							}
						}
						value = result.toString();
					} else {
						value = "'" + value + "'";
					}

					whereClause.append(field.getName() + " = " + value
							+ " AND ");
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
				}
			}
		}

		try {
			whereClause.delete(whereClause.length() - 5,
					whereClause.length() - 1);
		} catch (Exception e) {
			return null;
		}
		return whereClause.toString();
	}

	/**
	 * 插入多条数据
	 * 
	 * @param beans
	 */
	public static <T extends DbAble> void inserts(List<T> beans,
			SQLiteDatabase db) {
		if (beans != null && beans.size() > 0) {
			db.beginTransaction();
			for (T t : beans) {
				db.insert(t.getTabName(), null, t.getContentValues());
			}
			db.setTransactionSuccessful();
			db.endTransaction();
		}
	}

	public static <T extends DbAble> void inserts(List<T> beans) {
		SQLiteDatabase db = beans.get(0).getDb();
		inserts(beans, db);
		db.close();
	}

	/**
	 * 删除所有数据
	 * 
	 * @param bean
	 * @return 删除了多少数据
	 */
	public static <T extends DbAble> int deleteAll(T bean) {
		if (bean == null) {
			return -1;
		}

		int delete = 0;
		try {
			delete = bean.getDb().delete(bean.getTabName(), null, null);
		} catch (Exception e) {
			return -1;
		}
		return delete;
	}

	/**
	 * 查找所有的对象从数据库
	 * 
	 * @param bean
	 * @return
	 */
	public static <T extends DbAble> List<T> selectAll(T bean, SQLiteDatabase db) {
		List<T> beans = new ArrayList<>();

		if (bean == null) {
			return beans;
		}

		Cursor query = null;
		try {
			query = db.query(bean.getTabName(), null, null, null, null, null,
					null);
		} catch (Exception e) {
			return beans;
		}

		while (query.moveToNext()) {
			T copyBean = bean.copy();
			String[] columnNames = query.getColumnNames();
			for (String columnName : columnNames) {
				String fieldvalue = query.getString(query
						.getColumnIndex(columnName));
				PropertyDescriptor.invokeSet(copyBean, columnName, fieldvalue);
			}
			beans.add(copyBean);
		}
		query.close();

		return beans;
	}

	public static <T extends DbAble> List<T> selectAll(T bean) {
		SQLiteDatabase db = bean.getDb();
		List<T> selectAll = selectAll(bean, db);
		db.close();
		return selectAll;
	}

	/**
	 * 按条件查找
	 */
	@Override
	public T select(String fieldName, String value) {
		Cursor query = db.query(bean.getTabName(), null, fieldName + "=?",
				new String[] { value }, null, null, null);

		if (query.moveToFirst()) {
			String[] columnNames = query.getColumnNames();
			for (String columnName : columnNames) {
				String fieldvalue = query.getString(query
						.getColumnIndex(columnName));
				PropertyDescriptor.invokeSet(bean, columnName, fieldvalue);
			}
		}

		return bean;
	}

	/**
	 * 查找数据库的第几个到第几个
	 * 
	 * @param bean
	 * @param start
	 * @param end
	 * @return
	 */
	public static <T extends DbAble> List<T> selectLimit(T bean, int start,
			int count) {
		List<T> beans = new ArrayList<>();
		Cursor query = bean.getDb().rawQuery(
				"select * from " + bean.getTabName() + " LIMIT " + start + ","
						+ count, null);
		while (query.moveToNext()) {
			T copyBean = bean.copy();
			String[] columnNames = query.getColumnNames();
			for (String columnName : columnNames) {
				String fieldvalue = query.getString(query
						.getColumnIndex(columnName));
				PropertyDescriptor.invokeSet(copyBean, columnName, fieldvalue);
			}
			beans.add(copyBean);
		}
		return beans;
	}

	/**
	 * 根据本身变量查找
	 */
	@Override
	public List<T> select() {

		Cursor query = getCursor();

		return parseCursor(query);

	}

	/**
	 * 得到当前查找数据库获得的Cursor
	 * @return
	 */
	public Cursor getCursor() {
		Cursor query = bean.getDb().query(bean.getTabName(), null,
				getSqlCondition(bean), null, null, null, null);
		return query;
	}

	/**
	 * 根据Sql语句查找
	 * @param sql
	 * @return
	 */
	public List<T> select(String sql) {
		Cursor query = bean.getDb().query(bean.getTabName(), null, sql, null,
				null, null, null);
		return parseCursor(query);
	}

	/**
	 * 根据Cursor解析成数据
	 * @param query
	 * @return
	 */
	private List<T> parseCursor(Cursor query) {
		List<T> beans = new ArrayList<>();

		while (query.moveToNext()) {
			beans.add(parseCursorToBean(query));
		}
		return beans;
	}

	private T parseCursorToBean(Cursor query) {
		T copyBean = bean.copy();
		copyBean.fillData(query);
		return copyBean;
	}

	/**
	 * 得到有用的全局变量，没有被忽略
	 * @param bean
	 * @return
	 */
	public static List<Field> getUsedFields(DbAble bean) {
		List<Field> userdFields = new ArrayList<>();
		List<Field> ignoreFields = BaseAnnotation.getFieldsHasAnnotation(bean,
				isIgnore.class);

		Field[] allFields = bean.getClass().getDeclaredFields();

		for (Field field : allFields) {
			isPrimaryKey annotation = field.getAnnotation(isPrimaryKey.class);
			if (!ignoreFields.contains(field) && annotation == null) {
				userdFields.add(field);
			}
		}
		return userdFields;
	}

	/**
	 * 用于连表查询，因为这种情况比较复杂，所以还是需要自己拼Sql语句
	 * @param sql 完整的Sql语句
	 * @param clas 查询出来的类Class
	 * @return
	 */
	public static <T extends DbAble> HashMap<Class<T>, List<T>> selectSqlMap(
			String sql, Class<T>... clas) {
		
		HashMap<Class<T>, List<T>> maps=new HashMap<>();
		
		for (Class<T> cla : clas) {
			try {
				T instance = cla.newInstance();
				List<T> dbbeans=new ArrayList<>();
				Cursor rawQuery = instance.getDb().rawQuery(sql, null);
				while(rawQuery.moveToNext()){
					instance.fillData(rawQuery);
					dbbeans.add(instance);
					instance=instance.copy();
				}
				
				maps.put(cla, dbbeans);
				rawQuery.close();
			} catch (InstantiationException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			}
		}
		
		return maps;
	}
	
	/**
	 * 自定义Sql语句进行查询
	 * @param sql
	 * @param cla
	 * @return
	 */
	public static <T extends DbAble> List<T> selectSql(String sql, Class<T> cla){
		HashMap<Class<T>, List<T>> selectSqlMap = selectSqlMap(sql, cla);
		return selectSqlMap.get(cla);
	}
			

}
