package com.focustech.baseandroid.dialog;

import static com.focustech.baseandroid.general.BaseAnnotation.getAnnotation;
import static com.focustech.baseandroid.general.BaseAnnotation.getAnnotationMethods;

import java.lang.reflect.Method;
import java.util.List;

import com.focustech.baseandroid.activity.BaseActivity;
import com.focustech.baseandroid.general.BaseAnnotation.isClickMethod;
import com.focustech.baseandroid.general.BaseAnnotation.setContentView;
import com.focustech.baseandroid.listener.OnBaseClickListener;
import com.focustech.baseandroid.listener.OnLifeCycleChangeListener;
import com.focustech.baseandroid.manager.ControlManager;
import com.focustech.baseandroid.manager.NetStateManager;
import com.focustech.baseandroid.net.Response;
import com.focustech.baseandroid.net.params.Url.onRequestListener;
import com.focustech.baseandroid.util.BaseViewUtils;

import android.app.Dialog;
import android.content.Context;
import android.os.Bundle;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;

/**
 * dialog的基础类，对其进行二次封装，代码风格会和Activity完全一致
 * @author zhangzheng
 *
 * @param <T>
 */
public abstract class BaseDialog<T extends BaseActivity> extends Dialog implements
		onRequestListener<Response> {

	protected T mActivity;

	protected ViewGroup rootView;

	public BaseDialog(Context context, int theme) {
		super(context, theme);

		this.mActivity = (T) context;
	}

	public BaseDialog(Context context) {
		super(context, 0);
		this.mActivity = (T) context;
	}

	@Override
	protected void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		NetStateManager.getInstance().register(this);
		rootView = (ViewGroup) LayoutInflater.from(mActivity).inflate(
				getAnnotation(this, setContentView.class).value(), null);
		setContentView(rootView);

		initViews(rootView);

		initDatas(savedInstanceState);

		initListener();
	}
	
	public void registerLifeCycle(OnLifeCycleChangeListener lis){
		mActivity.registerLifeCycle(lis);
	}

	// 初始化视图
	protected abstract void initViews(ViewGroup rootView);

	// 初始化数据
	protected abstract void initDatas(Bundle savedInstanceState);

	// 初始化监听
	protected abstract void initListener();

	public void setClickListenerView(int... ids) {
		for (int id : ids) {
			setClickListenerView(id);
		}
	}

	public void setClickListenerView(View... views) {
		for (View view : views) {
			setClickListenerView(view);
		}
	}

	public void setClickListenerView(int id) {
		setClickListenerView(fv(id));
	}

	public void setClickListenerView(View view) {
		List<Method> annotationMethods = getAnnotationMethods(this,
				isClickMethod.class);
		if (annotationMethods != null && annotationMethods.size() > 0) {
			view.setOnClickListener(new ViewListener(view, annotationMethods));
		}
	}

	private class ViewListener extends OnBaseClickListener {

		private List<Method> annotationMethods;
		private View view;

		public ViewListener(View view, List<Method> annotationMethods) {

			this.annotationMethods = annotationMethods;
			this.view = view;
		}

		@Override
		public void onViewClick(View v) {
			for (Method annotationMethod : annotationMethods) {
				try {
					annotationMethod.invoke(BaseDialog.this, view);
				} catch (Exception e) {
					Log.e("error", e.toString());
				}
			}
		}
	}

	public <T extends View> T fv(int id) {
		return ControlManager.fv(rootView,id);
	}

	public <T extends View> T fv(int id, Class<T> cla) {
		return fv(id);
	}

	public void showToast(String content) {
		BaseViewUtils.showToast(mActivity, content);
	}

	public void showToast(int contentId) {
		BaseViewUtils.showToast(mActivity, contentId);
	}
	
	@Override
	public void onCompute(Response resp, Class cla) {
		ControlManager.onCompute(this, resp, cla);
	}

	/**
	 * 跳转到指定的页面
	 * 
	 * @param cla
	 * @param requestCode
	 * @param bundler
	 */
	public void turnTo(Class<? extends BaseActivity> cla, Integer requestCode,
			Bundle bundler) {
		ControlManager.turnTo(cla, requestCode, bundler, mActivity);
	}

	public void turnTo(Class<? extends BaseActivity> cla) {
		turnTo(cla, null, null);
	}

	public void turnTo(Class<? extends BaseActivity> cla, Bundle bundler) {
		turnTo(cla, null, bundler);
	}
}
