package com.focustech.baseandroid.fragment;

import android.os.Bundle;
import android.support.v4.app.Fragment;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;

import java.lang.reflect.Method;
import java.util.HashMap;

import com.focustech.baseandroid.activity.BaseActivity;
import com.focustech.baseandroid.general.BaseAnnotation;
import com.focustech.baseandroid.listener.OnLifeCycleChangeListener;
import com.focustech.baseandroid.listener.OnLifeCycleChangeListener.LifeCycle;
import com.focustech.baseandroid.manager.ControlManager;
import com.focustech.baseandroid.manager.EventBus;
import com.focustech.baseandroid.manager.LifeCycleManager;
import com.focustech.baseandroid.manager.NetStateManager;
import com.focustech.baseandroid.net.Response;
import com.focustech.baseandroid.net.params.Url.onRequestListener;

import static com.focustech.baseandroid.general.BaseAnnotation.getAnnotation;
import static com.focustech.baseandroid.general.BaseAnnotation.getAnnotationMethod;
import static com.focustech.baseandroid.general.BaseAnnotation.setContentView;

/**
 * 基础的Fragment类，进行封装，使其代码风格和Activity一致
 * Created by zhangzheng on 2015/10/19.
 */
public abstract class BaseFragment<T extends BaseActivity> extends Fragment
		implements onRequestListener<Response> {

	protected T mActivity;
	protected ViewGroup rootView;
	private LifeCycleManager lifeManager;

	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		NetStateManager.getInstance().register(this);
		mActivity = (T) getActivity();
		lifeManager = mActivity.lifeCycleManager;
		EventBus.getInstance().register(this);
	}

	public void registerLifeCycle(OnLifeCycleChangeListener lis) {
		lifeManager.registerLifeCycle(lis);
	}

	@Override
	public View onCreateView(LayoutInflater inflater, ViewGroup container,
			Bundle savedInstanceState) {
		rootView = (ViewGroup) LayoutInflater.from(mActivity).inflate(
				getAnnotation(this, setContentView.class).value(), null);

		return rootView;
	}

	@Override
	public void onActivityCreated(Bundle savedInstanceState) {
		super.onActivityCreated(savedInstanceState);

		initViews(rootView);

		initDatas(savedInstanceState);

		initListener();

	}

	// 初始化视图
	protected abstract void initViews(ViewGroup rootView);

	// 初始化数据
	protected abstract void initDatas(Bundle savedInstanceState);

	// 初始化监听
	protected abstract void initListener();

	public void setClickListenerView(int... ids) {
		for (int id : ids) {
			setClickListenerView(id);
		}
	}

	public void setClickListenerView(View... views) {
		for (View view : views) {
			setClickListenerView(view);
		}
	}

	public void setClickListenerView(int id) {
		setClickListenerView(fv(id));
	}

	public void setClickListenerView(View view) {
		Method annotationMethod = getAnnotationMethod(this,
				BaseAnnotation.isClickMethod.class);
		if (annotationMethod != null) {
			view.setOnClickListener(new ViewListener(annotationMethod, view));
		}
	}

	private class ViewListener implements View.OnClickListener {

		private Method annotationMethod;
		private View view;

		public ViewListener(Method annotationMethod, View view) {

			this.annotationMethod = annotationMethod;
			this.view = view;
		}

		@Override
		public void onClick(View v) {
			try {
				annotationMethod.invoke(BaseFragment.this, view);
			} catch (Exception e) {
				Log.e("error", e.toString());
			}
		}
	}

	@Override
	public void onHiddenChanged(boolean hidden) {
		super.onHiddenChanged(hidden);
		lifeManager.notifyLifeCycleChange(hidden ? LifeCycle.HIDE
				: LifeCycle.SHOW);
	}

	@Override
	public void onPause() {
		super.onPause();
		lifeManager.notifyLifeCycleChange(LifeCycle.PAURSE);
	}

	@Override
	public void onResume() {
		super.onResume();
		lifeManager.notifyLifeCycleChange(LifeCycle.ONRESUM);
	}

	@Override
	public void onStart() {
		super.onStart();
		lifeManager.notifyLifeCycleChange(LifeCycle.START);
	}

	@Override
	public void onStop() {
		super.onStop();
		lifeManager.notifyLifeCycleChange(LifeCycle.STOP);
	}

	@Override
	public void onDestroy() {
		super.onDestroy();
		lifeManager.notifyLifeCycleChange(LifeCycle.DESTROY);
		releaseMemory();
		EventBus.getInstance().unRegister(this);
	}

	protected void releaseMemory() {
		HashMap<Integer, View> viewsHashMap = (HashMap) rootView.getTag();
		if (viewsHashMap != null) {
			viewsHashMap.clear();
			rootView.setTag(null);
		}
	}

	public <T extends View> T fv(int id) {
		return ControlManager.fv(rootView, id);
	}

	public <T extends View> T fv(int id, Class<T> cla) {
		return fv(id);
	}

	public void showToast(String content) {
		mActivity.showToast(content);
	}

	public void showToast(int contentId) {
		mActivity.showToast(contentId);
	}

	@Override
	public void onCompute(Response resp, Class cla) {
		ControlManager.onCompute(this, resp, cla);
	}

	/**
	 * 跳转到指定的页面
	 * 
	 * @param cla
	 * @param requestCode
	 * @param bundler
	 */
	public void turnTo(Class<? extends BaseActivity> cla, Integer requestCode,
			Bundle bundler) {
		ControlManager.turnTo(cla, requestCode, bundler, mActivity);
	}

	public void turnTo(Class<? extends BaseActivity> cla) {
		turnTo(cla, null, null);
	}

	public void turnTo(Class<? extends BaseActivity> cla, Bundle bundler) {
		turnTo(cla, null, bundler);
	}

}
