package com.focustech.baseandroid.manager;


import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import android.os.Handler;
import android.os.Looper;

import com.focustech.baseandroid.general.BaseAnnotation;

/**
 * 类与类之前不需要有任何引用，就可以在一个类中操作另外的一个类
 * 这样设计的两个类，耦合度会很低，但实际中却会对跟踪代码带来障碍
 * 因为你可能并不能知道，这个方法具体被哪个类调用了，所以详细的注释可能会帮助到我们
 * 并且默认情况下，我仅仅将该功能集成到了Activity中，毕竟Activity之间的走向
 * 清晰明了，我们和容易可以跟踪的到代码，而且Activity和Activity之间交互太过于
 * 复杂，所以我目前主要是为了解决这个问题，而去设计了这个类
 * Created by zhangzheng on 2015/10/19.
 * 用于解耦代码
 */
public class EventBus {

    public static class MethodField {
        public Method method;
        public int type;
        public boolean isMainThread;
        public Object instance;
    }

    private static EventBus evenBus = new EventBus();

    private static HashMap<String, List<MethodField>> methodMap = new HashMap<>();

    public static EventBus getInstance() {
        return evenBus;
    }

    /**
     * 注册EvenBus到对象中
     * @param position 实体类
     */
    public void register(Object position) {
        Class<?> mClass = position.getClass();
        Method[] methods = mClass.getMethods();
        for (Method method : methods) {
            BaseAnnotation.isEventBusMethod annotation = method.getAnnotation(BaseAnnotation.isEventBusMethod.class);
            if (annotation != null) {
                MethodField mf = new MethodField();
                mf.isMainThread = annotation.isMainThread();
                mf.type = annotation.type();
                mf.method = method;
                mf.instance=position;
                List<MethodField> methodsList = methodMap.get(mClass.getName());
                if(methodsList==null){
                    methodsList=new ArrayList<>();
                }
                methodsList.add(mf);
                methodMap.put(mClass.getName(),methodsList);
            }
        }
    }

    /**
     * 解除对 对象的注册
     * @param position 实体类
     */
    public void unRegister(Object position){
        Class<?> mClass = position.getClass();
        methodMap.remove(mClass.getName());
    }

    /**
	 * EventBus进行推送
	 * @param mClass 推送的类
	 * @param type 推送的类型 用于识别相同类，相同参数的情况
	 * @param parms 推送方法的参数
	 */
    public void post(final Class mClass,final Integer type,final Object... parms){
        List<MethodField> methodFields = methodMap.get(mClass.getName());
        for (MethodField mf:methodFields){
            if(type==null||mf.type==type){
            	
            	if(mf.isMainThread&&Looper.myLooper()!=Looper.getMainLooper()){
            		new Handler(Looper.getMainLooper()).post(new Runnable() {
						
						@Override
						public void run() {
							post(mClass,type,parms);
						}
					});
            		return;
            	}
            	
                try {
                    mf.method.invoke(mf.instance,parms);
                } catch (IllegalAccessException e) {
                    e.printStackTrace();
                } catch (InvocationTargetException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    public void post(Class mClass,Object... parms){
        post(mClass,-1,parms);
    }

    /**
	 * 强制EventBus进行推送提交 忽略类型只要类和参数满足就全部提交
	 */
    public void forcedPost(Class mClass,Object... parms){
        post(mClass,null,parms);
    }

}
