package com.nuance.dragon.toolkit.audio;

import junit.framework.Assert;

/**
 * Represents a 'chunk' of audio, with binary audio data and a type that indicates
 * how the audio data is encoded.
 * <p>
 * Either {@link AbstractAudioChunk#audioShorts} or {@link AbstractAudioChunk#audioBytes} will
 * be non-null; the other will be null. *
 */
public abstract class AbstractAudioChunk
{
    //predefined parameters
    public static final int PCM_8K_SAMPLE_RATE_IN_HZ      = 8000;
    public static final int PCM_8K_SAMPLE_SIZE_IN_BYTES   = 2;
    public static final int PCM_11K_SAMPLE_RATE_IN_HZ     = 11025;
    public static final int PCM_11K_SAMPLE_SIZE_IN_BYTES  = 2;
    public static final int PCM_16K_SAMPLE_RATE_IN_HZ     = 16000;
    public static final int PCM_16K_SAMPLE_SIZE_IN_BYTES  = 2;
    public static final int PCM_22K_SAMPLE_RATE_IN_HZ     = 22050;
    public static final int PCM_22K_SAMPLE_SIZE_IN_BYTES  = 2;
    public static final int PCM_44K_SAMPLE_RATE_IN_HZ     = 44100;
    public static final int PCM_44K_SAMPLE_SIZE_IN_BYTES  = 2;
    public static final int PCM_48K_SAMPLE_RATE_IN_HZ     = 48000;
    public static final int PCM_48K_SAMPLE_SIZE_IN_BYTES  = 2;
    
    public static final int ULAW_SAMPLE_RATE_IN_HZ        = 8000;
    public static final int ULAW_SAMPLE_SIZE_IN_BYTES     = 1;
    
    public static final int GSM_FR_DOUBLE_FRAME_LENGTH_IN_MILLISECONDS = 40; // 20 * 2
    public static final int GSM_FR_DOUBLE_FRAME_SIZE_IN_BYTES          = 65; // 32.5 * 2
    
    public static final int AMR_FRAME_LENGTH_IN_MILLISECONDS  = 20;
    public static final int AMR0_FRAME_SIZE_IN_BYTES          = 13;
    public static final int AMR1_FRAME_SIZE_IN_BYTES          = 14;
    public static final int AMR2_FRAME_SIZE_IN_BYTES          = 16;
    public static final int AMR3_FRAME_SIZE_IN_BYTES          = 18;
    public static final int AMR4_FRAME_SIZE_IN_BYTES          = 20;
    public static final int AMR5_FRAME_SIZE_IN_BYTES          = 21;
    public static final int AMR6_FRAME_SIZE_IN_BYTES          = 27;
    public static final int AMR7_FRAME_SIZE_IN_BYTES          = 32;
    
    public static final int MP3_128KBPS_BIT_RATE              = 128000;

    /**
     * Valid speech status values for an audio chunk.
     */
    public enum SpeechStatus
    {
        /**
         * Indicates that the speech status is unknown
         */
        UNKNOWN,

        /**
         * Indicates that this chunk does not contain speech data
         */
        NO_SPEECH,

        /**
         * Indicates that this chunk contains speech data
         */
        SPEECH
    }

    /**
     * Base constructor of AbstractAudioChunk
     *
     * @param type The {@link AudioType} of the chunk. Must not be null.
     * @param shorts Audio data in shorts. May be null if <code>bytes</code> is Not null.
     * @param bytes Audio data in bytes. May be null if <code>shorts</code> is Not null.
     * @param timestamp TODO
     * @param duration Duration of the audio data in milliseconds.
     * @param speechStatus The {@link SpeechStatus} of this chunk
     * @param ttsMarkers The array of {@link TtsMarker} for this chunk. May be null.
     */
    protected AbstractAudioChunk(AudioType type, short[] shorts, byte[] bytes, long timestamp, int duration, SpeechStatus speechStatus, TtsMarker[] ttsMarkers)
    {
        Assert.assertNotNull(type);
        Assert.assertNotNull(speechStatus);
        Assert.assertTrue(duration >= 0);
        Assert.assertTrue(shorts != null || bytes != null);
        audioType = type;
        audioShorts = shorts;
        audioBytes = bytes;
        audioTimestamp = timestamp;
        audioEndTimestamp = timestamp + duration;
        audioDuration = duration;
        audioSpeechStatus = speechStatus;
        audioTtsMarkers = ttsMarkers;
    }

    /** The type of audio in this chunk. */
    // TODO Never null?
    public final AudioType audioType;

    /** The audio data. May be null if {@link #audioShorts} is non-null. */
    public final byte[] audioBytes;

    /** The audio data. May be null if {@link #audioBytes} is non-null. */
    public final short[] audioShorts;

    /** The duration of this chunk in milliseconds, or 0 if unknown. */
    public final int audioDuration;

    /** The timestamp at the start of this audio */
    public final long audioTimestamp;

    /** The timestamp at the end of this audio (based on {@link #audioTimestamp}
     * + {@link #audioDuration}) */
    public final long audioEndTimestamp;

    /**
     * Speech status, from {@link SpeechStatus}.
     */
    public final SpeechStatus audioSpeechStatus;

    /**
     * TTS markers, from {@link TtsMarker}.
     */
    public final TtsMarker[] audioTtsMarkers;

    @Override
    public String toString()
    {
        return "[" + audioType + "], " + (audioBytes == null ? -1 : audioBytes.length) + ", " + (audioShorts == null ? -1 : audioShorts.length) + ", " + audioDuration + ", " + audioTimestamp + ", " + audioEndTimestamp + ", " + audioSpeechStatus;
    }
}
