package com.nuance.dragon.toolkit.audio;

import com.nuance.dragon.toolkit.oem.api.internal.SystemClock;

/**
 * An audio chunk that represents a chunk of audio from a single audio channel.
 */
public class AudioChunk extends AbstractAudioChunk
{
    /**
     * AudioChunk constructor
     * @param type The audio type
     * @param data The 8-bit audio data. The data is NOT copied, so don't change
     * the contents!
     * @param duration Duration in ms (or 0 if unknown). Must be >= 0.
     */
    public AudioChunk(AudioType type, byte [] data, int duration)
    {
        this(type, null, data, 0, duration, SpeechStatus.UNKNOWN, null);
    }

    /**
     * AudioChunk constructor
     * @param type The audio type
     * @param data The 8-bit audio data. The data is NOT copied, so don't change
     * the contents!
     * @param duration Duration in ms (or 0 if unknown). Must be >= 0.
     * @param speechStatus The speech status
     */
    public AudioChunk(AudioType type, byte [] data, int duration, SpeechStatus speechStatus)
    {
        this(type, null, data, SystemClock.uptimeMillis(), duration, speechStatus, null);
    }

    /**
     * AudioChunk constructor
     * @param type The audio type
     * @param data The 8-bit audio data. The data is NOT copied, so don't change
     * the contents!
     * @param duration Duration in ms (or 0 if unknown). Must be >= 0.
     * @param ttsMarkers The TTS markers
     */
    public AudioChunk(AudioType type, byte [] data, int duration, TtsMarker[] ttsMarkers)
    {
        this(type, null, data, SystemClock.uptimeMillis(), duration, SpeechStatus.UNKNOWN, ttsMarkers);
    }

    /**
     * @param type The audio type
     * @param data The 16-bit audio data. The data is NOT copied, so don't change
     * the contents!
     */
    public AudioChunk(AudioType type, short [] data)
    {
        this(type, data, null, SystemClock.uptimeMillis(), type.getDuration(data), SpeechStatus.UNKNOWN, null);
    }

    /**
     * @param type The audio type
     * @param data The 16-bit audio data. The data is NOT copied, so don't change
     * @param speechStatus The speech status
     * the contents!
     */
    public AudioChunk(AudioType type, short [] data, SpeechStatus speechStatus)
    {
        this(type, data, null, SystemClock.uptimeMillis(), type.getDuration(data), speechStatus, null);
    }

    /**
     * @param type The audio type
     * @param data The 16-bit audio data. The data is NOT copied, so don't change
     * @param ttsMarkers The TTS markers. May be null.
     * the contents!
     */
    public AudioChunk(AudioType type, short [] data, TtsMarker[] ttsMarkers)
    {
        this(type, data, null, SystemClock.uptimeMillis(), type.getDuration(data), SpeechStatus.UNKNOWN, ttsMarkers);
    }

    /**
     * @param type The audio type
     * @param data The 16-bit audio data. The data is NOT copied, so don't change
     * the contents!
     * @param timestamp
     */
    public AudioChunk(AudioType type, short [] data, long timestamp)
    {
        this(type, data, null, timestamp, type.getDuration(data), SpeechStatus.UNKNOWN, null);
    }

    /**
     * @param type The audio type
     * @param data The 16-bit audio data. The data is NOT copied, so don't change
     * the contents!
     * @param timestamp
     * @param ttsMarkers
     */
    public AudioChunk(AudioType type, short [] data, long timestamp, TtsMarker[] ttsMarkers)
    {
        this(type, data, null, timestamp, type.getDuration(data), SpeechStatus.UNKNOWN, ttsMarkers);
    }

    private AudioChunk(AudioType type, short[] shorts, byte[] bytes, long timestamp, int duration, SpeechStatus speechStatus, TtsMarker[] ttsMarkers)
    {
        super(type, shorts, bytes, timestamp, duration, speechStatus, ttsMarkers);
    }

    /**
     * Returns a numeric representation of the audio energy of the current audio chunk.
     * The audio level is within the range from 0 to 90, the value is proportional to
     * the loudness of the audio. Based on the experiments, only the audio with the
     * energy value greater than 40 will be heard by normal human ears.
     *
     * @return audio energy level
     */
    public float getAudioEnergyLevel()
    {
        long acc = 0;

        if (audioShorts != null)
        {
            acc = accumulate(audioShorts);
        }
        else
        {
            acc = accumulate(audioBytes);
        }

        return (90 + computeAverageLevel(acc));
    }

    /**
     * A group of helper functions for generating the audio energy
     * level values.
     *
     * @param shorts
     * @return audio energy level
     */
    private long accumulate(short[] shorts)
    {
        long acc = 0, val;
        for (int i = 0; i < shorts.length; i++)
        {
            val = shorts[i];
            acc += (val * val) >> 9;
        }
        return acc;
    }

    private long accumulate(byte[] audioBytes)
    {
        long acc = 0, val;
        for (int i = 0; i < audioBytes.length; i++)
        {
            val = audioBytes[i];
            acc += (val * val) >> 9;
        }
        return acc;
    }

    private float computeAverageLevel(long acc)
    {
        float average;//, peak = 0;
        double pwr, dbpwr;


        pwr = acc / (32768.0 * 32768.0);
        if (pwr < 1e-9)
            dbpwr = -90;
        else
        {
            dbpwr = 10 * Math.log10(pwr);
            if (dbpwr > 0)
                dbpwr = 0;
        }

        average = (float)dbpwr;
        //peak = (float)dbpwr;

        return average;
    }
}
