package com.nuance.dragon.toolkit.audio;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Base class for audio modules that monitor an existing audio source via an
 * {@link AudioSink}, and also expose corresponding audio as an {@link AudioSource}.
 * The exposed audio may be unmodified (e.g. if this module is simply monitoring
 * audio or duplicating it for multiple sinks), or it may be modified (e.g. if
 * this module encodes it to a different format).
 */
public abstract class AudioPipe<InputType extends AbstractAudioChunk, OutputType extends AbstractAudioChunk> extends AudioSource<OutputType>
{

    private final AudioSink<InputType> _internalSink;
    private boolean _sourceConnected; // Don't report closed until an initial connection has been made

    protected AudioPipe(NMTHandler mainThreadHandler)
    {
        super(mainThreadHandler);
        _internalSink = new AudioSink<InputType>(mainThreadHandler)
        {
            @Override
            public void chunksAvailable(AudioSource<InputType> source)
            {
                AudioPipe.this.chunksAvailable(source, this);
            }

            @Override
            public void framesDropped(AudioSource<InputType> source)
            {
                AudioPipe.this.framesDropped(source, this);
            }

            @Override
            public void sourceClosed(AudioSource<InputType> source)
            {
                AudioPipe.this.sourceClosed(source, this);
            }

            @Override
            public boolean isAudioSourceTypeSupported(AudioType type)
            {
                return AudioPipe.this.isAudioSourceTypeSupported(type);
            }
        };
    }

    /**
     * Connect the audio source that this module will obtain its audio from.
     * This should only be called once during the lifetime of this audio pipe.
     * If you need to create a second connection, you should use a new audio
     * pipe instance.
     * @param source The audio source this pipe connects to
     */
    public void connectAudioSource(final AudioSource<InputType> source)
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                Checker.checkState(this, !_sourceConnected, "Source was already connected to this pipe");

                _sourceConnected = true;
                _internalSink.connectAudioSource(source);
            }
        });
    }

    /**
     * Disconnect the audio source that was previously connected via
     * {@link #connectAudioSource}. It is safe to call this even if there is no
     * connected source.
     * @return The disconnected source, or null if there was no source to
     * disconnect or this method is invoked asynchronously because not called
     * from the communication thread.
     */
    public AudioSource<InputType> disconnectAudioSource()
    {
        final List<AudioSource<InputType>> ret = new ArrayList<AudioSource<InputType>>();
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                AudioSource<InputType> connectedSource = _internalSink.getConnectedSource();
                if (connectedSource != null)
                {
                    boolean active = connectedSource.isActive();
                    _internalSink.disconnectAudioSource();

                    if (active)
                        _internalSink.sourceClosed(connectedSource);

                    _sourceConnected = false;
                    ret.add(connectedSource);
                }
            }
        });
        return ret.isEmpty() ? null : ret.get(0);
    }

    /**
     * Determine whether a source with the given audio type can be connected to
     * this pipe. The default AudioPipe implementation always returns true.
     * @param type The audio type
     * @return true if the audio type is supported.
     */
    protected boolean isAudioSourceTypeSupported(AudioType type)
    {
        return true;
    }

    /**
     * Internal method for obtaining the audio source that has been connected
     * via {@link #connectAudioSource}.
     * @return The connected audio source, or null if none is connected.
     */
    protected final AudioSource<InputType> getConnectedSource()
    {
        return _internalSink.getConnectedSource();
    }

    /**
     * Internal method for determining whether the connected audio source (or
     * the to-be connected audio source) is active.
     * @return true if a source is connected and active. Also returns true if
     * no source has been connected yet, since it is expected that a source will
     * be connected eventually.
     */
    protected final boolean isSourceActive()
    {
        if (!_sourceConnected) return true;

        AudioSource<InputType> source = _internalSink.getConnectedSource();
        if (source == null) return false;

        return source.isActive();
    }

    /**
     * Internal method for getting the available chunks for the connected audio
     * source.
     * @return The number of chunks available from the connected source. Always
     * >= 0.
     */
    protected final int getChunksAvailableFromSource()
    {
        AudioSource<InputType> source = _internalSink.getConnectedSource();
        if (source == null) return 0;

        return source.getChunksAvailableForSink(_internalSink);
    }

    /**
     * Internal method for getting an audio chunk from the connected audio source.
     * @return The next chunk from the connected source, or null if there are
     * no available chunks.
     */
    protected final InputType getAudioChunkFromSource()
    {
        AudioSource<InputType> source = _internalSink.getConnectedSource();
        if (source == null) return null;

        return source.getAudioChunkForSink(_internalSink);
    }

    /**
     * Internal method for getting the audio type of the connected audio source.
     * @return The audio type of the connected source.
     */
    protected final AudioType getAudioTypeFromSource()
    {
        AudioSource<InputType> source = _internalSink.getConnectedSource();
        if (source == null) return AudioType.UNKNOWN;

        return source.getAudioType();
    }

    /**
     * The connected source has called {@link AudioSink#chunksAvailable} on the
     * internal sink.
     * @param fromSource The source that called {@link AudioSink#chunksAvailable}.
     * @param internalSink The internal sink, which should be used when calling
     * methods on the source.
     */
    protected abstract void chunksAvailable(AudioSource<InputType> fromSource, AudioSink<InputType> internalSink);

    /**
     * The connected source has called {@link AudioSink#framesDropped} on the
     * internal sink.
     * @param fromSource The source that called {@link AudioSink#chunksAvailable}.
     * @param internalSink The internal sink, which should be used when calling
     * methods on the source.
     */
    protected abstract void framesDropped(AudioSource<InputType> fromSource, AudioSink<InputType> internalSink);

    /**
     * The connected source has called {@link AudioSink#sourceClosed} on the
     * internal sink.
     * @param fromSource The source that called {@link AudioSink#sourceClosed}.
     * @param internalSink The internal sink, which should be used when calling
     * methods on the source.
     */
    protected abstract void sourceClosed(AudioSource<InputType> fromSource, AudioSink<InputType> internalSink);
}
