package com.nuance.dragon.toolkit.audio;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.impl.HandlerOem;

/**
 * An audio sink accepts audio from an {@link AudioSource}.
 */
public abstract class AudioSink<AudioChunkType extends AbstractAudioChunk>
{
    protected final NMTHandler _mainThreadHandler;

    private AudioSource<AudioChunkType> _audioSource;

    protected AudioSink(NMTHandler mainThreadHandler)
    {
        _mainThreadHandler = mainThreadHandler == null ? new HandlerOem() : mainThreadHandler;
    }

    public NMTHandler getMainThreadHandler()
    {
        return _mainThreadHandler;
    }

    /**
     * Connect an AudioSource to this AudioSink. Set up all the necessary
     * information so that the source will generate audio and the sink will read
     * it automatically.
     *
     * @param source The audio source this sink connects to
     */
    public final void connectAudioSource(final AudioSource<AudioChunkType> source)
    {
        Checker.checkArgForNull("source", source);
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                AudioType sourceType = source.getAudioType();
                Checker.checkArgForCondition("source", "a supported audio type", sourceType != null && isAudioSourceTypeSupported(sourceType));

                AudioSource<AudioChunkType> oldSource = _audioSource;
                _audioSource = source;
                if (oldSource != null)
                {
                    oldSource.audioSinkDisconnected(AudioSink.this);
                }
                source.audioSinkConnected(AudioSink.this);
            }
        });
    }

    /**
     * Disconnect the current AudioSource from this AudioSink. It is safe to
     * call this even if there is no connected source.
     * @return The disconnected source, or null if there was no source to
     * disconnect or this method is invoked asynchronously because not called
     * from the communication thread.
     */
    public final AudioSource<AudioChunkType> disconnectAudioSource()
    {
        final List<AudioSource<AudioChunkType>> ret = new ArrayList<AudioSource<AudioChunkType>>();
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                AudioSource<AudioChunkType> source = _audioSource;
                _audioSource = null;
                if (source != null)
                {
                    source.audioSinkDisconnected(AudioSink.this);
                    audioSourceDisconnected(source);
                    ret.add(source);
                }
            }
        });
        return ret.isEmpty() ? null : ret.get(0);
    }

    /**
     * Called when the connected audio source has been disconnected.
     * @param source The source that was just disconnected.
     */
    protected void audioSourceDisconnected(AudioSource<AudioChunkType> source)
    {
        // Nothing by default
    }

    /**
     * Get the connected source.
     * @return The connected source, or null if no source is connected.
     */
    protected AudioSource<AudioChunkType> getConnectedSource()
    {
        return _audioSource;
    }

    /**
     * Determine whether a source with the given audio type can be connected to
     * this sink. The default AudioSink implementation always returns true.
     * @param type The audio type
     * @return true if the audio type is supported.
     */
    protected boolean isAudioSourceTypeSupported(AudioType type)
    {
        return true;
    }

    /**
     * Called by the AudioSource associated with this AudioSink when new data
     * has been added to the source. This tells this AudioSink that it can read
     * some audio from the source if the sink was waiting on new data.
     *
     * @param source The AudioSource that is notifying this AudioSink of data
     */
    public abstract void chunksAvailable(AudioSource<AudioChunkType> source);

    /**
     * Called by the AudioSource associated with this AudioSink when data was
     * not read fast enough and the source's internal buffer overflowed. The
     * entire utterance is probably invalid and it is up to this sink to notify
     * the system appropriately.
     *
     * @param source The source that overflowed
     */
    public abstract void framesDropped(AudioSource<AudioChunkType> source);

    /**
     * Called by the AudioSource associated with this AudioSink when the source
     * will no longer generate any new audio chunks other than what is already
     * available inside it.
     *
     * @param source The source that just closed
     */
    public abstract void sourceClosed(AudioSource<AudioChunkType> source);
}
