package com.nuance.dragon.toolkit.audio;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.impl.HandlerOem;
import com.nuance.dragon.toolkit.util.internal.ReadOnlyList;

/**
 * Audio sources provide audio to an {@link AudioSink}, or multiple audio sinks
 * depending on the implementation. Audio sources should notify attached audio
 * sinks when audio is available, and provide the audio upon request.
 */
public abstract class AudioSource<AudioChunkType extends AbstractAudioChunk>
{
    protected final NMTHandler _mainThreadHandler;
    
    // As an optimization, keep this around to return whenever there are no chunks
    private final ReadOnlyList<AudioChunkType> _emptyChunkList = new ReadOnlyList<AudioChunkType>();

    /**
     * The type of audio generated by this source. Never returns null.
     *
     * @return The audio type.
     */
    public abstract AudioType getAudioType();

    /**
     * How many AudioChunks are currently available for reading by a given sink
     * from this source.
     *
     * @param sink The sink making the request
     *
     * @return The number of available chunks. Always >= 0.
     */
    public abstract int getChunksAvailableForSink(AudioSink<AudioChunkType> sink);

    /**
     * Whether or not the source is going to generate any more audio than is
     * currently available.
     *
     * @return true if this source is active.
     */
    public abstract boolean isActive();

    protected AudioSource(NMTHandler mainThreadHandler)
    {
        _mainThreadHandler = mainThreadHandler == null ? new HandlerOem() : mainThreadHandler;
    }
    
    public NMTHandler getMainThreadHandler()
    {
        return _mainThreadHandler;
    }
    
    /**
     * Equivalent to checking if {@link #getChunksAvailableForSink} == 0, but
     * possibly more efficient.
     * <p>
     * Subclasses: override if you can improve efficiency.
     * @param sink The sink that is the object of this call
     * @return true if there are no available chunks for the sink.
     */
    public boolean isEmptyForSink(AudioSink<AudioChunkType> sink)
    {
        return getChunksAvailableForSink(sink) == 0;
    }

    /**
     * Equivalent to checking if {@link #isEmpty} == 0, but possibly more
     * efficient.
     * <p>
     * Subclasses: override if you can improve efficiency
     * @return true if there are no available chunks.
     */
    public boolean isEmpty()
    {
        return getChunksAvailable() == 0;
    }

    /**
     * How many audio chunks would be available to a newly connected sink?
     * @return The number of available chunks. Always >= 0.
     */
    public abstract int getChunksAvailable();

    /**
     * Get a chunk of audio from this audio source, intended for deposit into
     * the specified AudioSink. This will remove the audio from the source.
     * Called by the AudioSink associated with this AudioSource.
     * @param sink  The intended AudioSink
     * @return The next audio chunk, or null if there are no more available.
     */
    public abstract AudioChunkType getAudioChunkForSink(AudioSink<AudioChunkType> sink);

    /**
     * Read all of the available audio chunks into a chunk list. This will
     * remove the audio from the source. Called by the AudioSink associated with
     * this AudioSource.
     * @param sink The intended AudioSink
     * @param chunkList The list of chunks to read into.
     * @return The number of chunks that were read.
     */
    public final int readAllAudioChunksForSink(AudioSink<AudioChunkType> sink, List<AudioChunkType> chunkList)
    {
        AudioChunkType chunk;
        int count = 0;
        do {
            chunk = getAudioChunkForSink(sink);
            if (chunk != null)
            {
                count++;
                chunkList.add(chunk);
            }
            else
            {
                break;
            }
        } while (true);
        return count;
    }

    /**
     * Get all of the available audio chunk. This will remove the audio from the
     * source. Called by the AudioSink associated with this AudioSource.
     * @param sink The intended AudioSink
     * @return The chunks that were read.
     */
    public final List<AudioChunkType> getAllAudioChunksForSink(AudioSink<AudioChunkType> sink)
    {
        int available = this.getChunksAvailableForSink(sink);
        if (available == 0)
            return _emptyChunkList;

        ArrayList<AudioChunkType> list = new ArrayList<AudioChunkType>(available);
        getAllAudioChunksForSink(sink, list);
        return list;
    }

    /**
     * Get all of the available audio chunk. This will remove the audio from the
     * source. Called by the AudioSink associated with this AudioSource.
     * @param sink The intended AudioSink
     * @param chunksList The list to add the new chunks to
     */
    public void getAllAudioChunksForSink(AudioSink<AudioChunkType> sink, List<AudioChunkType> chunksList)
    {
        List<AudioChunkType> list = chunksList;
        do {
            AudioChunkType chunk = getAudioChunkForSink(sink);
            if (chunk == null)
                break;

            list.add(chunk);
        } while (true);
    }

    /**
     * Called by an AudioSink when it connects to this source. This is protected
     * to make sure no other modules are tempted to call this directly.
     *
     * @param sink  The connected AudioSink
     */
    protected abstract void audioSinkConnected(AudioSink<AudioChunkType> sink);

    /**
     * Called by an AudioSink when it disconnects from this source. This is
     * protected to make sure no other modules are tempted to call this directly.
     *
     * @param sink The disconnected AudioSink
     */
    protected abstract void audioSinkDisconnected(AudioSink<AudioChunkType> sink);
}
