package com.nuance.dragon.toolkit.audio;

import java.util.Arrays;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * Valid audio encoding types.
 */
public class AudioType implements JSONCompliant
{
    /**
     * Encoding types
     */
    public enum Encoding
    {
        PCM_16,
        SPEEX,
        OPUS,
        ULAW,
        GSM_FR,
        AMR0,
        AMR1,
        AMR2,
        AMR3,
        AMR4,
        AMR5,
        AMR6,
        AMR7,
        MP3_8KBPS,
        MP3_16KBPS,
        MP3_24KBPS,
        MP3_32KBPS,
        MP3_40KBPS,
        MP3_48KBPS,
        MP3_56KBPS,
        MP3_64KBPS,
        MP3_80KBPS,
        MP3_96KBPS,
        MP3_112KBPS,
        MP3_128KBPS,
        MP3_144KBPS,
        MP3_160KBPS,
        UNKNOWN;
    }

    /**
     * Sampling rates
     */
    public static class Frequency
    {
        public static final int FREQ_8KHZ  = 8000;
        public static final int FREQ_11KHZ = 11025;
        public static final int FREQ_16KHZ = 16000;
        public static final int FREQ_22KHZ = 22050;
        public static final int FREQ_44KHZ = 44100;
        public static final int FREQ_48KHZ = 48000;
    }

    /** {@link Encoding#PCM_16} and {@link Frequency#FREQ_48KHZ} */
    public static final AudioType PCM_48k = new AudioType(Encoding.PCM_16, Frequency.FREQ_48KHZ);

    /** {@link Encoding#PCM_16} and {@link Frequency#FREQ_44KHZ} */
    public static final AudioType PCM_44k = new AudioType(Encoding.PCM_16, Frequency.FREQ_44KHZ);

    /** {@link Encoding#PCM_16} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType PCM_22k = new AudioType(Encoding.PCM_16, Frequency.FREQ_22KHZ);

    /** {@link Encoding#PCM_16} and {@link Frequency#FREQ_16KHZ} */
    public static final AudioType PCM_16k = new AudioType(Encoding.PCM_16, Frequency.FREQ_16KHZ);

    /** {@link Encoding#PCM_16} and {@link Frequency#FREQ_11KHZ} */
    public static final AudioType PCM_11k = new AudioType(Encoding.PCM_16, Frequency.FREQ_11KHZ);

    /** {@link Encoding#PCM_16} and {@link Frequency#FREQ_8KHZ} */
    public static final AudioType PCM_8k = new AudioType(Encoding.PCM_16, Frequency.FREQ_8KHZ);

    /** {@link Encoding#SPEEX} and {@link Frequency#FREQ_16KHZ} */
    public static final AudioType SPEEX_WB = new AudioType(Encoding.SPEEX, Frequency.FREQ_16KHZ);

    /** {@link Encoding#SPEEX} and {@link Frequency#FREQ_8KHZ} */
    public static final AudioType SPEEX_NB = new AudioType(Encoding.SPEEX, Frequency.FREQ_8KHZ);

    /** {@link Encoding#OPUS} and {@link Frequency#FREQ_16KHZ} */
    public static final AudioType OPUS_WB = new AudioType(Encoding.OPUS, Frequency.FREQ_16KHZ);

    /** {@link Encoding#OPUS} and {@link Frequency#FREQ_8KHZ} */
    public static final AudioType OPUS_NB = new AudioType(Encoding.OPUS, Frequency.FREQ_8KHZ);

    /** {@link Encoding#ULAW} and {@link Frequency#FREQ_8KHZ} */
    public static final AudioType ULAW = new AudioType(Encoding.ULAW, Frequency.FREQ_8KHZ);

    /** {@link Encoding#GSM_FR} and {@link Frequency#FREQ_8KHZ} */
    public static final AudioType GSM_FR = new AudioType(Encoding.GSM_FR, Frequency.FREQ_8KHZ);

    /** {@link Encoding#AMR0} and {@link Frequency#FREQ_8KHZ} */
    public static final AudioType AMR0 = new AudioType(Encoding.AMR0, Frequency.FREQ_8KHZ);

    /** {@link Encoding#AMR1} and {@link Frequency#FREQ_8KHZ} */
    public static final AudioType AMR1 = new AudioType(Encoding.AMR1, Frequency.FREQ_8KHZ);

    /** {@link Encoding#AMR2} and {@link Frequency#FREQ_8KHZ} */
    public static final AudioType AMR2 = new AudioType(Encoding.AMR2, Frequency.FREQ_8KHZ);

    /** {@link Encoding#AMR3} and {@link Frequency#FREQ_8KHZ} */
    public static final AudioType AMR3 = new AudioType(Encoding.AMR3, Frequency.FREQ_8KHZ);

    /** {@link Encoding#AMR4} and {@link Frequency#FREQ_8KHZ} */
    public static final AudioType AMR4 = new AudioType(Encoding.AMR4, Frequency.FREQ_8KHZ);

    /** {@link Encoding#AMR5} and {@link Frequency#FREQ_8KHZ} */
    public static final AudioType AMR5 = new AudioType(Encoding.AMR5, Frequency.FREQ_8KHZ);

    /** {@link Encoding#AMR6} and {@link Frequency#FREQ_8KHZ} */
    public static final AudioType AMR6 = new AudioType(Encoding.AMR6, Frequency.FREQ_8KHZ);

    /** {@link Encoding#AMR7} and {@link Frequency#FREQ_8KHZ} */
    public static final AudioType AMR7 = new AudioType(Encoding.AMR7, Frequency.FREQ_8KHZ);

    /** {@link Encoding#MP3_160KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_160KBPS = new AudioType(Encoding.MP3_160KBPS, Frequency.FREQ_22KHZ);

    /** {@link Encoding#MP3_144KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_144KBPS = new AudioType(Encoding.MP3_144KBPS, Frequency.FREQ_22KHZ);

    /** {@link Encoding#MP3_128KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_128KBPS = new AudioType(Encoding.MP3_128KBPS, Frequency.FREQ_22KHZ);

    /** {@link Encoding#MP3_112KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_112KBPS = new AudioType(Encoding.MP3_112KBPS, Frequency.FREQ_22KHZ);

    /** {@link Encoding#MP3_96KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_96KBPS = new AudioType(Encoding.MP3_96KBPS, Frequency.FREQ_22KHZ);

    /** {@link Encoding#MP3_80KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_80KBPS = new AudioType(Encoding.MP3_80KBPS, Frequency.FREQ_22KHZ);

    /** {@link Encoding#MP3_64KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_64KBPS = new AudioType(Encoding.MP3_64KBPS, Frequency.FREQ_22KHZ);

    /** {@link Encoding#MP3_56KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_56KBPS = new AudioType(Encoding.MP3_56KBPS, Frequency.FREQ_22KHZ);

    /** {@link Encoding#MP3_48KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_48KBPS = new AudioType(Encoding.MP3_48KBPS, Frequency.FREQ_22KHZ);

    /** {@link Encoding#MP3_40KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_40KBPS = new AudioType(Encoding.MP3_40KBPS, Frequency.FREQ_22KHZ);

    /** {@link Encoding#MP3_32KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_32KBPS = new AudioType(Encoding.MP3_32KBPS, Frequency.FREQ_22KHZ);

    /** {@link Encoding#MP3_24KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_24KBPS = new AudioType(Encoding.MP3_24KBPS, Frequency.FREQ_22KHZ);

    /** {@link Encoding#MP3_16KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_16KBPS = new AudioType(Encoding.MP3_16KBPS, Frequency.FREQ_22KHZ);

    /** {@link Encoding#MP3_8KBPS} and {@link Frequency#FREQ_22KHZ} */
    public static final AudioType MP3_8KBPS = new AudioType(Encoding.MP3_8KBPS, Frequency.FREQ_22KHZ);

    /** Unknown audio type. */
    public static final AudioType UNKNOWN = new AudioType(Encoding.UNKNOWN, 0);

    private static final String KEY_ENCODING = "enc";
    private static final String KEY_FREQUENCY = "freq";

    /** The sampling rate, as samples / second. See {@link Frequency}. */
    public final int frequency;

    /** The audio encoding. */
    public final Encoding encoding;

    /** The SSE configuration (optional). */
    public final byte[] sse;

    public AudioType(Encoding encoding, int frequency)
    {
        this(encoding, frequency, null);
    }

    public AudioType(Encoding encoding, int frequency, byte[] sse)
    {
        this.frequency = frequency;
        this.encoding = encoding;
        this.sse = sse;
    }

    public int getSampleCount(int durationMs)
    {
        return (frequency * durationMs) / 1000;
    }

    public int getDuration(short[] shorts)
    {
        return getDuration(shorts.length);
    }

    public int getDuration(int sampleCount)
    {
        long ret = 0;
        if (encoding == Encoding.PCM_16)
        {
            ret = (long)sampleCount * 1000 /frequency;
            return (int)ret;
        }

        Logger.error(this, "Unable to detect duration for encoding " + encoding.name());
        return 0;
    }

    /** @return Number of bytes per audio sample. */
    public int getBytesPerSample() {
        int bytesPerSample;

        switch (encoding) {
            case PCM_16:
                bytesPerSample = 2;
                break;
            default:
                bytesPerSample = 1;
        }

        return bytesPerSample;
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result
                + ((encoding == null) ? 0 : encoding.hashCode());
        result = prime * result + frequency;
        result = prime * result + Arrays.hashCode(sse);
        return result;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        AudioType other = (AudioType) obj;
        if (encoding != other.encoding)
            return false;
        if (frequency != other.frequency)
            return false;
        if (!Arrays.equals(sse, other.sse))
            return false;
        return true;
    }

    @Override
    public String toString()
    {
        return "AudioType [frequency=" + frequency + ", encoding=" + encoding
                + ", sse=" + Arrays.toString(sse) + "]";
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        o.tryPut(KEY_FREQUENCY, frequency);
        o.tryPut(KEY_ENCODING, encoding.name());
        return o;
    }

    public static AudioType fromJSON(org.json.JSONObject json) throws JSONException
    {
        int frequency = json.getInt(KEY_FREQUENCY);
        String encoding = json.getString(KEY_ENCODING);
        Encoding[] types = Encoding.values();
        for (Encoding t : types)
        {
            if (t.name().equals(encoding))
                return new AudioType(t, frequency);
        }
        throw new JSONException("Encoding '" + encoding + "' not valid");
    }
}
