package com.nuance.dragon.toolkit.audio;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.internal.SystemClock;


/**
 * An audio chunk that represents multiple audio channels.
 */
public class MultiChannelAudioChunk extends AbstractAudioChunk
{
    /**
     * @param type The audio type
     * @param shorts The 16-bit audio data. The data is NOT copied, so don't change
     * the contents!
     * @param channelCount The number of channels
     * @param speechStatus The speech status
     */
    public MultiChannelAudioChunk(AudioType type, short[] shorts, int channelCount, SpeechStatus speechStatus)
    {
        this(type, shorts, null, SystemClock.uptimeMillis(), type.getDuration(shorts) / channelCount, channelCount, new int[channelCount], speechStatus);
    }

    /**
     * @param type The audio type
     * @param bytes The 8-bit audio data. The data is NOT copied, so don't change
     * the contents!
     * @param timestamp
     * @param duration Duration in ms (or 0 if unknown). Must be >= 0.
     * @param channelCount The number of channels. Must be > 0.
     * @param speechStatus The speech status
     */
    public MultiChannelAudioChunk(AudioType type, byte[] bytes, long timestamp, int duration, int channelCount, SpeechStatus speechStatus)
    {
        this(type, null, bytes, timestamp, duration, channelCount, new int[channelCount], speechStatus);
    }

    /**
     * Create a multi-channel chunk from multiple single-channel chunks
     * @param audioChunks The single-channel chunks
     */
    public MultiChannelAudioChunk(AudioChunk ... audioChunks)
    {
        this(audioChunks[0].audioType, concatShorts(audioChunks), concatBytes(audioChunks), audioChunks[0].audioTimestamp, audioChunks[0].audioDuration, (byte)audioChunks.length, getChannelOffsets(audioChunks), audioChunks[0].audioSpeechStatus);
    }

    private MultiChannelAudioChunk(AudioType type, short[] shorts, byte[] bytes, long timestamp, int duration, int channelCount, int [] channelOffsetsMs, SpeechStatus speechStatus)
    {
        // TODO: add TtsMarker support?
        super(type, shorts, bytes, timestamp, duration, speechStatus, null);
        Assert.assertTrue(channelCount > 0);
        this.channelCount = channelCount;
        this.channelOffsetsMs = channelOffsetsMs;
    }

    public AudioChunk getSingleChannelChunk(int channel)
    {
        Checker.checkArgForCondition("channel", "valid", channel >= 0 && channel < channelCount);

        if (audioShorts != null)
        {
            int channelLength = audioShorts.length / channelCount;
            int offset = channelLength * channel;

            short [] channelShorts = new short[channelLength];
            System.arraycopy(audioShorts, offset, channelShorts, 0, channelLength);
            return new AudioChunk(audioType, channelShorts, audioTimestamp);
        }
        else
        {
            int channelLength = audioBytes.length / channelCount;
            int offset = channelLength * channel;

            byte [] channelBytes = new byte[channelLength];
            System.arraycopy(audioBytes, offset, channelBytes, 0, channelLength);
            return new AudioChunk(audioType, channelBytes, audioDuration, audioSpeechStatus);
        }
    }

    private static int[] getChannelOffsets(AudioChunk ... chunks)
    {
        int [] offsets = new int[chunks.length];
        long baseTimestamp = chunks[0].audioTimestamp;
        offsets[0] = 0;

        for (int i = 1; i < chunks.length; i++)
        {
            AudioChunk chunk = chunks[i];
            offsets[i] = (int)(chunk.audioTimestamp - baseTimestamp);
        }
        return offsets;
    }

    private static byte[] concatBytes(AudioChunk ... chunks)
    {
        int length = 0;
        for (AudioChunk chunk : chunks)
        {
            if (chunk.audioBytes == null)
                return null;

            length += chunk.audioBytes.length;
        }

        byte[] bytes = new byte[length];
        int offset = 0;
        for (AudioChunk chunk : chunks)
        {
            System.arraycopy(chunk.audioBytes, 0, bytes, offset, chunk.audioBytes.length);
            offset += chunk.audioBytes.length;
        }
        return bytes;
    }

    private static short[] concatShorts(AudioChunk ... chunks)
    {
        int length = 0;
        for (AudioChunk chunk : chunks)
        {
            if (chunk.audioShorts == null)
                return null;

            length += chunk.audioShorts.length;
        }

        short[] shorts = new short[length];
        int offset = 0;
        for (AudioChunk chunk : chunks)
        {
            System.arraycopy(chunk.audioShorts, 0, shorts, offset, chunk.audioShorts.length);
            offset += chunk.audioShorts.length;
        }
        return shorts;
    }

    /**
     * The number of channels represented by this audio chunk.
     */
    public final int channelCount;

    /**
     * The offsets, in ms, of each channel from channel 0. By definition, the
     * first element of this array is 0. A positive value means that the
     * channel's audio comes LATER than channel 0.
     */
    public final int[] channelOffsetsMs;
}
