/*------------------------------------------------------------------------

  Copyright 2012-2012 Nuance Communications, Inc.

------------------------------------------------------------------------*/
package com.nuance.dragon.toolkit.audio.bluetooth;

import android.bluetooth.BluetoothAdapter;
import android.content.Context;
import android.os.Handler;
import android.util.Log;

import com.nuance.dragon.toolkit.Build;
import com.nuance.dragon.toolkit.audio.bluetooth.impl.Bluetooth;

/**
 * This Bluetooth Manager provides simplified API to operate Bluetooth for the
 * purpose of audio recording, playback and speech recognition. <br>
 * It is implemented as a singleton, since there is a one-to-one relationship
 * between Bluetooth SCO connection and the recognizer.
 * <br>
 * Permissions: <br>
 * It is necessary to define the following permissions
 * for this Bluetooth module to work correctly: <br>
 * BLUETOOTH_ADMIN, BLUETOOTH_PERM, BLUETOOTH and MODIFY_AUDIO_SETTINGS
 */
public class BluetoothManager implements Bluetooth.HeadsetStateListener {
    private static final String TAG = "DMT-BluetoothManager";
    private static final boolean DEBUG = Build.DEBUG;

    private int mConnectAudioTimeoutSec;

    private Bluetooth mBluetoothImpl;
    private BluetoothListener mListener;

    private boolean mIsInitialized;
    private int mAudioState = AudioState.DISCONNECTED;
    private Handler mHandler;

    private class AudioState {
        private static final int DISCONNECTED   = 0;
        private static final int CONNECTING     = 1;
        private static final int CONNECTED      = 2;
        private static final int DISCONNECTING  = 3;
    }

    private Runnable mConnectAudioTimeoutRunnable = new Runnable() {
        public void run() {
            if (mIsInitialized) {
                // timeout reached, stop the SCO
                // since BT implementation is still trying
                // establish audio connection
                mBluetoothImpl.stopBluetoothSco();
                mAudioState = AudioState.DISCONNECTED;
                notifyAudioConnected(Status.TIMEOUT);
            }
        }
    };

    private Runnable mDisconnectAudioTimeoutRunnable = new Runnable() {
        public void run() {
            mAudioState = AudioState.DISCONNECTED;
            notifyAudioDisconnected(Status.TIMEOUT);
        }
    };

    // Singleton
    private BluetoothManager() {};
    private static final BluetoothManager sInstance = new BluetoothManager();

    /**
     * Get the instance of the Bluetooth manager
     * @return The instance of the Bluetooth manager,
     * or null if this module is not supported
     */
    public static BluetoothManager getInstance() {
        return sInstance;
    }

    /**
     * Initialize the bluetooth manager for a new context.
     *
     * The Bluetooth manager requires a context to operate
     * on the underlying Bluetooth headset object, and therefore
     * is coupled/dependent on a given context. The user should call
     * release() before the context becomes invalid, e.g. at onDestroy().
     *
     * If application uses multiple activities, we recommend either:
     * 1) Using the application context, which can be obtained by calling
     *    a getApplicationContext() on any context. The application context
     *    has a life time of a process hosting the application.
     * or
     * 2) Wrap BT into a service, and use a service context
     * @param context - the context used to initialize Bluetooth manager
     * @param listener - the BluetoothListener for notifications
     */
    public void initialize(Context context, BluetoothListener listener) {
        if (DEBUG) Log.d(TAG, "initialize(), context = " + context);

        // validate state and parameters
        if (context == null) {
            throw new IllegalArgumentException("The context argument is null");
        }
        if (listener == null) {
            throw new IllegalArgumentException("The listener argument is null");
        }
        if (mIsInitialized) {
            throw new IllegalStateException("Attempt at double-initialization of Bluetooth manager, it has been already initialized");
        }

        registerListener(listener);
        mHandler = new Handler();
        mBluetoothImpl = Bluetooth.createInstance(context, this);

        mAudioState = AudioState.DISCONNECTED;

        if (DEBUG) Log.d(TAG, "init(), returning mBluetoothImpl = " + mBluetoothImpl);
    }

    /**
     * Release the bluetooth manager.
     *
     * This method releases the underlying bluetooth headset object
     * and cleans up the resources
     */
    public void release() {
        if (DEBUG) Log.d(TAG, "release()");
        if (mIsInitialized) {
            mIsInitialized = false;

            if ( (mAudioState == AudioState.CONNECTED) || (mAudioState == AudioState.CONNECTING)) {
                Log.w(TAG, "release(), disconnecting audio: recommended to disconnect BT audio before calling release()");
                mBluetoothImpl.stopBluetoothSco();
            }

            mAudioState = AudioState.DISCONNECTED;

            mBluetoothImpl.close();
            mBluetoothImpl = null;

            // for some reason, Android BluetoothProfile.ServiceListener.onServiceDisconnected(int profile) does not get called
            // after mBluetoothImpl.close(); it is possible that Android did not intend to call it if the user called release/close
            // but only if the service was shut down by the system. It is unclear to me from the doc. Therefore, to keep
            // the API asynch and symmetrical with initialize(), the callback is 'emulated' by a post.
            if (mHandler != null)
                mHandler.post(mOnBluetoothReleasedRunnable);
        }
        else {
            throw new IllegalStateException("release() - BluetoothManager is not active.");
        }
    }

    private void finish() {
        if (DEBUG) Log.d(TAG, "finish()");

        mHandler = null;

        // Call the listener only after all other teardown has completed.
        if (mListener != null) {
            BluetoothListener listener = mListener;
            mListener = null;
            listener.onBluetoothReleased(Status.SUCCESS);
        }

    }

    private Runnable mOnBluetoothReleasedRunnable = new Runnable() {
        public void run() {
            finish();
        }
    };

    /**
     * Convenience method to get the state of the Bluetooth adapter.
     * @return True if the adapter is enabled, else false.
     */
    public static boolean isAdapterEnabled()
    {
        final BluetoothAdapter bluetoothAdapter = BluetoothAdapter.getDefaultAdapter();
        if (bluetoothAdapter != null)
        {
            return bluetoothAdapter.isEnabled();
        }
        return false;
    }

    /**
     * Call this method to find out if an audio-capable
     * Bluetooth device is connected to the host device.
     * Please note that this state is not the same a "bluetooth audio
     * connected" state, see isAudioConnected()
     * @return True if there is one
     */
    public boolean isDeviceConnected() {
        checkInitialized();
        boolean result = (mBluetoothImpl.getHeadsetState() == Bluetooth.HEADSET_STATE_CONNECTED);
        if (DEBUG) Log.d(TAG, "isDeviceConnected() returning " + result);
        return result;
    }

    /**
     * Call this method to find out if there is an already existing audio connection.
     * If there is, no need to call connectAudio(), the Bluetooth audio is ready to use.
     * @return True if Bluetooth audio (SCO) link is connected and ready to use
     */
    public boolean isAudioConnected() {
        if (DEBUG) Log.d(TAG, "isAudioConnected(), mBluetoothImpl = " + mBluetoothImpl);
        checkInitialized();

        boolean result = (mBluetoothImpl.getAudioState() == Bluetooth.HEADSET_AUDIO_STATE_CONNECTED);

        if (DEBUG) Log.d(TAG, "isAudioConnected() returning " + result);
        return result;
    }

    /**
     * Connect Bluetooth audio; in more detail, establish SCO link
     * This step is needed before running recording and speech recognition using BT audio.
     * For most non-media profile BT headsets, this step is often needed for playback
     * of the audio via BT device
     * @param connectionTimeoutSec - connection timeout in seconds, must be a positive integer
     *          connection attempt
     *          will time out if unsuccessful after the specified number of seconds;
     *          the client will receive a notification via
     *          BluetoothListener.onAudioConnected() with status
     *          BluetoothManager.Status.TIMEOUT
     *          In case of success, onAudioConnected() will be called with
     *          BluetoothManager.Status.TIMEOUT
     *          In case the BT device is not connected, onAudioConnected()
     *          will be called with BluetoothManager.Status.DEVICE_NOT_CONNECTED
     */
    public void connectAudio(int connectionTimeoutSec) {
        if (DEBUG) Log.d(TAG, "connectAudio()");
        checkInitialized();

        // Check parameters
        if (connectionTimeoutSec < 1) {
            throw new IllegalArgumentException("Connection timeout cannot be 0 or negative");
        }
        mConnectAudioTimeoutSec = connectionTimeoutSec;

        // Check device connectivity
        boolean deviceConnected = isDeviceConnected();
        boolean audioConnected = isAudioConnected();
        if (DEBUG) Log.d(TAG, "connectAudio(), deviceConnected = <" + deviceConnected + ">, audioConnected = <" + audioConnected + ">");

        if (!deviceConnected) {
            notifyAudioConnected(Status.DEVICE_NOT_CONNECTED);
            return;
        }

        // State machine
        switch (mAudioState) {
            case AudioState.DISCONNECTED:
                mAudioState = AudioState.CONNECTING;
                if (mBluetoothImpl.startBluetoothSco())
                    mHandler.postDelayed(mConnectAudioTimeoutRunnable, mConnectAudioTimeoutSec*1000);
                else
                    mHandler.post(mConnectAudioTimeoutRunnable);
                break;
            case AudioState.CONNECTING:
                throw new IllegalStateException("connectAudio() called twice");
            case AudioState.CONNECTED:
                throw new IllegalStateException("connectAudio() called twice, already connected");
             case AudioState.DISCONNECTING:
                throw new IllegalStateException("connectAudio() should not be called until the disconnectAudio() has been acknowledged");

             default:
                throw new IllegalStateException("connectAudio() - illegal audio state, mAudioState = " + mAudioState);
        }
   }

    /**
     * Connect Bluetooth audio assuming 5 second connection attempt timeout
     */
    public void connectAudio() {
        connectAudio(5);
    }

    /**
     * Disconnect Bluetooth audio.
     * Call this method when done with active BT audio / speech interaction,
     * to save BT device power and allow other clients to use BT
     * The caller will receive a notification via BluetoothListener.onAudioDisconnected()
     */
    public void disconnectAudio() {
        if (DEBUG) Log.d(TAG, "disconnectAudio()");
        checkInitialized();

        boolean initiateDisconnect = false;

        switch (mAudioState) {
            case AudioState.DISCONNECTED:
                throw new IllegalStateException("disconnectAudio() - audio already disconnected");
            case AudioState.CONNECTING:
                notifyAudioConnected(Status.ABORTED_BY_CALLER); // acknowledge the connect callback and proceed with disconnect
                mAudioState = AudioState.DISCONNECTING;
                initiateDisconnect = true;
                break;
            case AudioState.CONNECTED:
                mAudioState = AudioState.DISCONNECTING;
                initiateDisconnect = true;
                break;
             case AudioState.DISCONNECTING:
                throw new IllegalStateException("calling disconnectAudio() twice");

             default:
                throw new IllegalStateException("disconnectAudio() - illegal audio state, mAudioState = " + mAudioState);
        }

        if (initiateDisconnect) {
            mBluetoothImpl.stopBluetoothSco();
            mHandler.postDelayed(mDisconnectAudioTimeoutRunnable, 1*1000);
        }
    }

    /**
     * Get the recommended playback stream for BT.
     * This method provides a recommended playback streams for a given
     * Android version and some specific device models, based on the field data.
     * When using Bluetooth in connection with SpeakerPlayerSink, the client
     * should specify the audio stream.
     * The client has an option to specify their own value, or use this method as a hint.
     * See also: public SpeakerPlayerSink(int streamType, AudioType audioType).
     * @return Android's audio stream type, e.g. AudioManager.STREAM_BLUETOOTH_SCO
     */
    public int getPlaybackStream() {
        if (DEBUG) Log.d(TAG, "getPlaybackStream(), mBluetoothImpl = " + mBluetoothImpl);
        checkInitialized();
        return mBluetoothImpl.getPlaybackStream();
    }

    /**
     * Status codes for Bluetooth Manager
     */
    public class Status {
        public static final int SUCCESS=0;
        public static final int TIMEOUT=1;
        public static final int UNSPECIFIED_REASON=2;
        public static final int DEVICE_NOT_CONNECTED=3;
        public static final int ABORTED_BY_CALLER=4;
    }


    // =========== Internal methods =============
    private void registerListener(BluetoothListener l) {
        mListener = l;
    }

    private void checkInitialized() {
        if (!mIsInitialized) {
            throw new IllegalStateException("Attempt at using uninitialized BluetoothManager");
        }
    }

    private void notifyAudioDisconnected(int status) {
        if (mListener != null)
            mListener.onAudioDisconnected(status);
    }

    private void notifyAudioConnected(int status) {
        if (mListener != null)
            mListener.onAudioConnected(status);
    }

    private void notifyDeviceDisconnected(int status) {
        if (mListener != null)
            mListener.onDeviceDisconnected(status);
    }

    private void notifyDeviceConnected(int status) {
        if (mListener != null)
            mListener.onDeviceConnected(status);
    }

    private void onAudioConnected() {
        switch (mAudioState) {
            case AudioState.DISCONNECTED:
            case AudioState.CONNECTED:
            case AudioState.DISCONNECTING:
                Log.w(TAG, "onAudioStateChanged(), possible state error, mAudioState = " + mAudioState);
                break;

            case AudioState.CONNECTING:
                mAudioState = AudioState.CONNECTED;
                notifyAudioConnected(Status.SUCCESS);
                break;

             default:
                throw new IllegalStateException("onAudioConnected() - illegal audio state, mAudioState = " + mAudioState);
        }
    }

    private void onAudioDisconnected() {
        switch (mAudioState) {
            case AudioState.DISCONNECTED:
                break;  // do nothing, already in disconnected state
            case AudioState.CONNECTING:
            case AudioState.CONNECTED:
                mBluetoothImpl.stopBluetoothSco();
                mAudioState = AudioState.DISCONNECTED;
                notifyAudioDisconnected(Status.UNSPECIFIED_REASON); // unsolicited disconnect, i.e. BT power down or out of range
                break;
            case AudioState.DISCONNECTING:
                mAudioState = AudioState.DISCONNECTED;
                notifyAudioDisconnected(Status.SUCCESS);
                break;

             default:
                throw new IllegalStateException("onAudioDisconnected() - illegal audio state, mAudioState = " + mAudioState);
        }
    }


    // ==== from HeadsetStateListener =====
    /** Not for public use
    */
    public void onConnectionStateChanged(int connectionState) {
        if (DEBUG) Log.d(TAG, "onConnectionStateChanged(), state = " + connectionState);

        if (connectionState == Bluetooth.HEADSET_STATE_CONNECTED) {
            notifyDeviceConnected(Status.SUCCESS);
        }

        if (connectionState == Bluetooth.HEADSET_STATE_DISCONNECTED) {
            notifyDeviceDisconnected(Status.UNSPECIFIED_REASON);
        }
    }

    /** Not for public use
    */
    public void onAudioStateChanged(int audioState) {
        if (DEBUG) Log.d(TAG, "onAudioStateChanged(), state = " + audioState + ", mListener = " + mListener);

        if (audioState == Bluetooth.HEADSET_AUDIO_STATE_CONNECTED) {
            mHandler.removeCallbacks(mConnectAudioTimeoutRunnable);
            onAudioConnected();
        }

        if (audioState == Bluetooth.HEADSET_AUDIO_STATE_DISCONNECTED) {
            mHandler.removeCallbacks(mDisconnectAudioTimeoutRunnable);
            onAudioDisconnected();
        }
    }

    /** Not for public use
    */
    public void onServiceConnected(boolean statusSuccess) {
        if (statusSuccess) {
            mIsInitialized = true;
            if (mListener != null) {
                mListener.onBluetoothInitialized(Status.SUCCESS);
            }
        } else {
            mIsInitialized = false;
            if (mListener != null) {
                mListener.onBluetoothInitialized(Status.UNSPECIFIED_REASON);
            }
        }
    }

    /** Not for public use
    */
    public void onServiceDisconnected() {
        mIsInitialized = false;
        if (mListener != null) {
            mListener.onBluetoothReleased(Status.SUCCESS);
        }
    }

}
