/*------------------------------------------------------------------------

  Copyright 2006-2012 Nuance Communications, Inc.

------------------------------------------------------------------------*/
package com.nuance.dragon.toolkit.audio.bluetooth.impl;


import android.bluetooth.BluetoothAdapter;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.IntentFilter;
import android.util.Log;

/**
 * Abstract base class for the Bluetooth functionality.
 * Derived classes are using platform specific Bluetooth APIs
 * for implementing functionality defined by this class.
 */
public abstract class Bluetooth {

    private static final String TAG = "DMT-Bluetooth";
    static final boolean DEBUG = true;  // one flag for the entire bluetooth impl package

    public static final int HEADSET_STATE_DISCONNECTED = 0;
    public static final int HEADSET_STATE_CONNECTED = 1;

    public static final int HEADSET_AUDIO_STATE_DISCONNECTED = 0;
    public static final int HEADSET_AUDIO_STATE_CONNECTED = 1;

    protected Context mContext;
    protected HeadsetStateListener mListener;
    protected boolean mServiceConnected;
    protected boolean mStateReceiverRegistered;
    protected boolean mHeadsetConnected;
    protected boolean mAudioConnected;

    public interface HeadsetStateListener {
        public void onServiceConnected(boolean statusSuccess);
        public void onServiceDisconnected();
        public void onConnectionStateChanged(int connectionState);
        public void onAudioStateChanged(int audioState);
    }

    public static Bluetooth createInstance(Context context, HeadsetStateListener listener) {
        int sdk = AndroidVersion.SDK;
        Bluetooth result = null;

        if ( (sdk < AndroidVersion.FROYO) || (AndroidVersion.IS_BROKEN_HTC)) {
            result = new Bluetooth_2_0(context, listener);
        } else if (sdk < AndroidVersion.GINGERBREAD){
            result = new Bluetooth_2_2(context, listener);
        } else if (sdk < AndroidVersion.HONEYCOMB) {
            result = new BluetoothImpl_2_3(context, listener);
        } else if (sdk < AndroidVersion.JELLY_BEAN_MR2) {
            result = new BluetoothImpl(context, listener);
        } else {
            result = new BluetoothImpl_4_3(context, listener);
        }

        Log.i(TAG, "createInstance() returns " + result + "> for SDK version " + sdk);
        return result;
    }

    protected Bluetooth(Context context, HeadsetStateListener listener) {
        mContext = context;
        mListener = listener;

        createHeadset();
        IntentFilter filter = createFilter();
        mContext.registerReceiver(getStateReceiver(), filter);
        mStateReceiverRegistered = true;


        if(mServiceConnected) {
            // Note: I believe that mServiceConnected will never be true here, but I'm leaving this in for now. (dbreslau)
            mListener.onConnectionStateChanged(getHeadsetState());
        }
    }

    protected abstract BroadcastReceiver getStateReceiver();

    protected abstract IntentFilter createFilter();

    public boolean isEnabled() {
        boolean enabled = false;
        BluetoothAdapter adapter = getBluetoothAdapter();

        if(adapter != null) {
            enabled = adapter.isEnabled();
        }

        return enabled;
    }

    protected BluetoothAdapter getBluetoothAdapter()
    {
        /*
        TODO(LXU): do it with reflection
        int sdk = AndroidVersion.SDK;
        if (sdk <= AndroidVersion.JELLY_BEAN_4_2)
        {
            return BluetoothAdapter.getDefaultAdapter();
        }
        else
        {
            return ((BluetoothManager)mContext.getSystemService(Context.BLUETOOTH_SERVICE)).getAdapter();
        }*/
        return BluetoothAdapter.getDefaultAdapter();
    }

    protected abstract void createHeadset();

    public abstract int getHeadsetState();

    public abstract int getAudioState();

    public abstract boolean startBluetoothSco();

    public abstract void stopBluetoothSco();

    public abstract int getPlaybackStream();

    public abstract int getRecordingSource();

    public abstract String[] getBondedDevices();

    public abstract void close();

    static void LOGD(String msg)
    {
        Log.d(TAG, msg);
    }
}
