/*------------------------------------------------------------------------

  Copyright 2006-2011 Nuance Communications, Inc.

------------------------------------------------------------------------*/
package com.nuance.dragon.toolkit.audio.bluetooth.impl;

import android.bluetooth.BluetoothDevice;
import android.content.Context;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;

/**
 * Reflection-based wrapper for the BluetoothHeadsetLegacy API.
 * This wrapper is using the bluetooth API signatures from Android API level 10 and below,
 * which was changed after, so we call this legacy.
 */
public class BluetoothHeadsetLegacy {

    public interface ServiceListener {
        public void onServiceConnected();
        public void onServiceDisconnected();
    }

    private static final Constructor<?> sConstructor;
    private static final Method sGetState;
    private static final Method sStartVoiceRecognition;
    private static final Method sStopVoiceRecognition;
    private static final Method sGetCurrentHeadset;
    private static final Method sDisconnectHeadset;
    private static final Method sConnectHeadset;
    private static final Method sClose;
    private static final Class<?> sServiceListenerClass;
    private static final boolean sGetStateRequiresDevice;

    private static final String ON_SERVICE_CONNECTED = "onServiceConnected";
    private static final String ON_SERVICE_DISCONNECTED = "onServiceDisconnected";

    public static final String ACTION_STATE_CHANGED;
    public static final String ACTION_AUDIO_STATE_CHANGED;
    public static final String EXTRA_STATE;
    public static final String EXTRA_AUDIO_STATE;

    /** There was an error trying to obtain the state */
    public static final int STATE_ERROR;
    /** No headset currently connected */
    public static final int STATE_DISCONNECTED;
    /** Connection attempt in progress */
    public static final int STATE_CONNECTING;
    /** A headset is currently connected */
    public static final int STATE_CONNECTED;

    /** A SCO audio channel is not established */
    public static final int AUDIO_STATE_DISCONNECTED;
    /** A SCO audio channel is established */
    public static final int AUDIO_STATE_CONNECTED;

    android.os.Handler mHandler = new android.os.Handler();

    static {
        Reflection r = new Reflection();
        Class<?> cls = r.getClassForName("android.bluetooth.BluetoothHeadset");
        Class<?> clsListener = r.getClassForName("android.bluetooth.BluetoothHeadset$ServiceListener");

        Method tmpGetState;
        Method tmpDisconnect;
        boolean tmpGetStateRequiresDevice = false;

        sConstructor = r.getConstructor(cls, Context.class, clsListener);

        // since sGetState is 'final', below finds valid getState method in local
        //  and then assigns it to sGetState.
        tmpGetState = r.getMethodOrNull(cls, "getState");
        if(tmpGetState == null) {
            tmpGetState = r.getMethod(cls, "getState", BluetoothDevice.class);
            tmpGetStateRequiresDevice = true;
        }
        sGetState = tmpGetState;
        sGetStateRequiresDevice = tmpGetStateRequiresDevice;

        sGetCurrentHeadset = r.getMethod(cls, "getCurrentHeadset");
        sConnectHeadset = r.getMethod(cls, "connectHeadset", BluetoothDevice.class);

        tmpDisconnect = r.getMethodOrNull(cls, "disconnectHeadset");
        if(tmpDisconnect == null) {
            tmpDisconnect = r.getMethod(cls, "disconnectHeadset", BluetoothDevice.class);
        }
        sDisconnectHeadset = tmpDisconnect;

        sStartVoiceRecognition = r.getMethod(cls, "startVoiceRecognition");
        sStopVoiceRecognition = r.getMethod(cls, "stopVoiceRecognition");
        sClose = r.getMethod(cls, "close");

        sServiceListenerClass = clsListener;

        ACTION_STATE_CHANGED = (String) r.getFieldValue(cls, "ACTION_STATE_CHANGED");
        ACTION_AUDIO_STATE_CHANGED = (String) r.getFieldValue(cls, "ACTION_AUDIO_STATE_CHANGED");
        EXTRA_STATE = (String) r.getFieldValue(cls, "EXTRA_STATE");
        EXTRA_AUDIO_STATE = (String) r.getFieldValue(cls, "EXTRA_AUDIO_STATE");

        STATE_ERROR = (Integer) r.getFieldValue(cls, "STATE_ERROR");
        STATE_DISCONNECTED = (Integer) r.getFieldValue(cls, "STATE_DISCONNECTED");
        STATE_CONNECTING = (Integer) r.getFieldValue(cls, "STATE_CONNECTING");
        STATE_CONNECTED = (Integer) r.getFieldValue(cls, "STATE_CONNECTED");

        AUDIO_STATE_DISCONNECTED = (Integer) r.getFieldValue(cls, "AUDIO_STATE_DISCONNECTED");
        AUDIO_STATE_CONNECTED = (Integer) r.getFieldValue(cls, "AUDIO_STATE_CONNECTED");
    }

    private Object mBluetoothHeadset;
    private ServiceListener mServiceListener;

    private Runnable mOnServiceConnected = new Runnable() {
        public void run() {
            mServiceListener.onServiceConnected();
        }
    };

    private Runnable mOnServiceDisconnected = new Runnable() {
        public void run() {
            mServiceListener.onServiceDisconnected();
        }
    };

    public BluetoothHeadsetLegacy(Context context, ServiceListener listener) {
        mServiceListener = listener;
        Object serviceListenerProxy = Proxy.newProxyInstance(this.getClass().getClassLoader(),
                new Class<?>[] {
                    sServiceListenerClass
                }, new InvocationHandler() {
                    @Override
                    public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
                        if (method.getName().equals(ON_SERVICE_CONNECTED)) {
                            mHandler.post(mOnServiceConnected);
                        }
                        else if (method.getName().equals(ON_SERVICE_DISCONNECTED)) {
                            mHandler.post(mOnServiceDisconnected);
                        }
                        return null;
                    }
                });

        try {
            mBluetoothHeadset = sConstructor.newInstance(context, serviceListenerProxy);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public boolean connectHeadset(BluetoothDevice device) {
        boolean status = false;

        try {
            if(sConnectHeadset != null) {
                status = (Boolean)sConnectHeadset.invoke(mBluetoothHeadset, device);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }

        return status;
    }

    public void disconnectHeadset() {
        try {
            if(sDisconnectHeadset != null) {
                sDisconnectHeadset.invoke(mBluetoothHeadset);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }
    }

    public void disconnectHeadset(BluetoothDevice device) {
        try {
            if(sDisconnectHeadset != null) {
                sDisconnectHeadset.invoke(mBluetoothHeadset, device);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }
    }

    public BluetoothDevice getConnectedDevice() {
        BluetoothDevice device = null;

        try {
            if(sGetCurrentHeadset != null) {
                device = (BluetoothDevice)sGetCurrentHeadset.invoke(mBluetoothHeadset);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }

        return device;
    }

    public int getState() {
        try {
            if(sGetStateRequiresDevice) {
                Object current = sGetCurrentHeadset.invoke(mBluetoothHeadset);
                return(Integer) sGetState.invoke(mBluetoothHeadset, current);
            } else {
                return(Integer) sGetState.invoke(mBluetoothHeadset);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }

        return 0;
    }

    public boolean startVoiceRecognition() {
        try {
            return(Boolean) sStartVoiceRecognition.invoke(mBluetoothHeadset);
        } catch(Exception e) {
            e.printStackTrace();
        }
        return false;
    }

    public boolean stopVoiceRecognition() {
        try {
            return(Boolean) sStopVoiceRecognition.invoke(mBluetoothHeadset);
        } catch(Exception e) {
            e.printStackTrace();
        }
        return false;
    }

    public void close() {
        try {
            sClose.invoke(mBluetoothHeadset);
        } catch(Exception e) {
            e.printStackTrace();
        }
    }

}
