/*------------------------------------------------------------------------

  Copyright 2006-2013 Nuance Communications, Inc.

------------------------------------------------------------------------*/
package com.nuance.dragon.toolkit.audio.bluetooth.impl;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.content.Context;
import android.util.Log;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.List;

/**
 * Reflection-based wrapper for the BluetoothHeadset_4_3 API.
 * A bluetooth issue appeared on Android 4.3:DMTADRSDK-476,
 * an specific implementation is made.
 */
public class BluetoothHeadset_4_3 {

    private static final String TAG = "BluetoothHeadset_4_3";

    public interface ServiceListener {
        public void onServiceConnected(boolean success);
        public void onServiceDisconnected();
    }

    private static final Method sGetConnectionState;
    private static final Method sStartVoiceRecognition;
    private static final Method sStopVoiceRecognition;
    private static final Method sGetCurrentHeadset;
    private static final Method sDisconnectHeadset;
    private static final Method sConnectHeadset;
    private static final Method sGetProfileProxy;
    private static final Method sCloseProfileProxy;
    private static final Class<?> sServiceListenerClass;

    private static final int   sProfile;

    private static final String ON_SERVICE_CONNECTED = "onServiceConnected";
    private static final String ON_SERVICE_DISCONNECTED = "onServiceDisconnected";
    protected static final String HEADSET_CLASSNAME = "android.bluetooth.BluetoothHeadset";
    protected static final String DEVICE_CLASSNAME = "android.bluetooth.BluetoothDevice";
    protected static final String ADAPTER_CLASSNAME = "android.bluetooth.BluetoothAdapter";
    protected static final String PROFILE_CLASSNAME = "android.bluetooth.BluetoothProfile";
    protected static final String LISTENER_CLASSNAME = "android.bluetooth.BluetoothProfile$ServiceListener";

    public static final String ACTION_STATE_CHANGED;
    public static final String ACTION_AUDIO_STATE_CHANGED;
    public static final String EXTRA_STATE;
    public static final String EXTRA_AUDIO_STATE;

    /** No headset currently connected */
    public static final int STATE_DISCONNECTED;
    /** Connection attempt in progress */
    public static final int STATE_CONNECTING;
    /** A headset is currently connected */
    public static final int STATE_CONNECTED;

    /** A SCO audio channel is not established */
    public static final int AUDIO_STATE_DISCONNECTED;
    /** A SCO audio channel is established */
    public static final int AUDIO_STATE_CONNECTED;
    /** A SCO audio channel is connecting */
    public static final int AUDIO_STATE_CONNECTING;

    android.os.Handler mHandler = new android.os.Handler();

    static {
        Reflection r = new Reflection();
        Class<?> cls = r.getClassForName(HEADSET_CLASSNAME);
        Class<?> clsListener = r.getClassForName(LISTENER_CLASSNAME);
        Class<?> profileClass = r.getClassForName(PROFILE_CLASSNAME);

        sGetConnectionState = r.getMethod(cls, "getConnectionState", BluetoothDevice.class);

        sGetCurrentHeadset = r.getMethod(cls, "getConnectedDevices");
        sConnectHeadset = r.getMethod(cls, "connect", BluetoothDevice.class);

        sDisconnectHeadset = r.getMethod(cls, "disconnect", BluetoothDevice.class);

        sStartVoiceRecognition = ReflectionFactory.getMethodFromStrings(HEADSET_CLASSNAME,
                                                                        "startScoUsingVirtualVoiceCall",
                                                                        DEVICE_CLASSNAME);
        sStopVoiceRecognition = ReflectionFactory.getMethodFromStrings(HEADSET_CLASSNAME,
                                                                       "stopScoUsingVirtualVoiceCall",
                                                                       DEVICE_CLASSNAME);
        sGetProfileProxy = ReflectionFactory.getMethodFromStrings(ADAPTER_CLASSNAME,
                                                                  "getProfileProxy",
                                                                  "android.content.Context",
                                                                  LISTENER_CLASSNAME,
                                                                  "int");
        sCloseProfileProxy = ReflectionFactory.getMethod(ADAPTER_CLASSNAME,
                                                         "closeProfileProxy",
                                                         int.class,
                                                         profileClass);

        sServiceListenerClass = clsListener;

        sProfile = (Integer)r.getFieldValue(profileClass, "HEADSET");

        ACTION_STATE_CHANGED = (String) r.getFieldValue(cls, "ACTION_CONNECTION_STATE_CHANGED");
        ACTION_AUDIO_STATE_CHANGED = (String) r.getFieldValue(cls, "ACTION_AUDIO_STATE_CHANGED");
        EXTRA_STATE = (String) r.getFieldValue(cls, "EXTRA_STATE");
        EXTRA_AUDIO_STATE = (String) r.getFieldValue(cls, "EXTRA_STATE");

        STATE_DISCONNECTED = (Integer) r.getFieldValue(profileClass, "STATE_DISCONNECTED");
        STATE_CONNECTING = (Integer) r.getFieldValue(profileClass, "STATE_CONNECTING");
        STATE_CONNECTED = (Integer) r.getFieldValue(profileClass, "STATE_CONNECTED");

        AUDIO_STATE_DISCONNECTED = (Integer) r.getFieldValue(cls, "STATE_AUDIO_DISCONNECTED");
        AUDIO_STATE_CONNECTED = (Integer) r.getFieldValue(cls, "STATE_AUDIO_CONNECTED");
        AUDIO_STATE_CONNECTING = (Integer) r.getFieldValue(cls, "STATE_AUDIO_CONNECTING");

    }

    private BluetoothAdapter mAdapter;

    private Object mBluetoothHeadset;
    private ServiceListener mServiceListener;

    private Runnable mOnServiceConnectedSuccess = new Runnable() {
        public void run() {
            mServiceListener.onServiceConnected(true);
        }
    };

    private Runnable mOnServiceConnectedFailure = new Runnable() {
        public void run() {
            mServiceListener.onServiceConnected(false);
        }
    };

    private Runnable mOnServiceDisconnected = new Runnable() {
        public void run() {
            mServiceListener.onServiceDisconnected();
        }
    };

    public BluetoothHeadset_4_3(Context context, ServiceListener listener) {
        mAdapter = BluetoothAdapter.getDefaultAdapter();
        mServiceListener = listener;
        Object serviceListenerProxy = Proxy.newProxyInstance(this.getClass().getClassLoader(),
                new Class<?>[] {
                    sServiceListenerClass
                }, new InvocationHandler() {
                    @Override
                    public Object invoke(Object proxy, Method method, Object[] args)
                            throws Throwable {

                        Log.i(TAG, "onService_Callback: method.getName() is <" + method.getName() + ">");
                        if (method.getName().equals(ON_SERVICE_CONNECTED)) {
                            mBluetoothHeadset = args[1];
                            Log.i(TAG, "onServiceConnected: mBluetoothHeadset<"+mBluetoothHeadset+">");

                            if (mBluetoothHeadset != null) {
                                mHandler.post(mOnServiceConnectedSuccess);
                            }
                            else {
                                mHandler.post(mOnServiceConnectedFailure);
                            }
                        }
                        else if (method.getName().equals(ON_SERVICE_DISCONNECTED)) {
                            mBluetoothHeadset = null;
                            Log.i(TAG, "onServiceDisconnected: mBluetoothHeadset<"+mBluetoothHeadset+">");
                            mHandler.post(mOnServiceDisconnected);
                        }
                        return null;
                    }
                });

        try {
            if (mAdapter == null ||
                    ((Boolean)sGetProfileProxy.invoke(mAdapter, context, serviceListenerProxy, sProfile)) == false) {
                throw new RuntimeException("Failed to get BT profile proxy");
            }
        } catch(IllegalAccessException iae) {
            throw new RuntimeException("IllegalAccessException calling sGetProfileProxy");
        } catch(InvocationTargetException ite) {
            throw new RuntimeException("InvocationTargetException calling sGetProfileProxy");
        }

    }

    public boolean connectHeadset(BluetoothDevice device) {
        boolean status = false;

        try {
            if(sConnectHeadset != null) {
                status = (Boolean)sConnectHeadset.invoke(mBluetoothHeadset, device);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }

        return status;
    }

    public void disconnectHeadset() {
        try {
            if(sDisconnectHeadset != null) {
                sDisconnectHeadset.invoke(mBluetoothHeadset);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }
    }

    public void disconnectHeadset(BluetoothDevice device) {
        try {
            if(sDisconnectHeadset != null) {
                sDisconnectHeadset.invoke(mBluetoothHeadset, device);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }
    }

    @SuppressWarnings("unchecked")
    public BluetoothDevice getConnectedDevice() {
        List<BluetoothDevice> connectedDevices = null;
        BluetoothDevice device = null;

        try {
            if((mBluetoothHeadset != null) && (sGetCurrentHeadset != null)) {
                connectedDevices = (List<BluetoothDevice>)sGetCurrentHeadset.invoke(mBluetoothHeadset);
                if (connectedDevices != null && connectedDevices.size() > 0) {
                    // only one headset can be connected at a time
                    device = connectedDevices.get(0);
                }
            }
        } catch(Exception e) {
            e.printStackTrace();
        }

        return device;
    }

    public int getState() {
        int state = STATE_DISCONNECTED;
        BluetoothDevice device = null;
        try {
            device = getConnectedDevice();
            if ((mBluetoothHeadset != null) && (device != null)) {
                state =(Integer)sGetConnectionState.invoke(mBluetoothHeadset, device);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }
        Log.i(TAG, "getState: state = <"+state+">, device = <" + ((device != null) ? device.getName() : null) + ">, mBluetoothHeadset = <" + mBluetoothHeadset + ">");
        return state;
    }

    public boolean startVoiceRecognition() {
        try {
            return(Boolean) sStartVoiceRecognition.invoke(mBluetoothHeadset, getConnectedDevice());
        } catch(Exception e) {
            e.printStackTrace();
        }
        return false;
    }

    public boolean stopVoiceRecognition() {
        try {
            return(Boolean) sStopVoiceRecognition.invoke(mBluetoothHeadset, getConnectedDevice());
        } catch(Exception e) {
            e.printStackTrace();
        }
        return false;
    }

    public void close() {
        Log.d(TAG, "close()");
        try {
            if (mBluetoothHeadset != null && mAdapter != null) {
                Log.d(TAG, "close(), calling closeProfileProxy()");
                sCloseProfileProxy.invoke(mAdapter, sProfile, mBluetoothHeadset);
            }
        } catch(Exception e) {
            Log.d(TAG, "close(), exception()");
            e.printStackTrace();
        }
    }

}
