/*------------------------------------------------------------------------

  Copyright 2006-2011 Nuance Communications, Inc.

------------------------------------------------------------------------*/
package com.nuance.dragon.toolkit.audio.bluetooth.impl;

import java.lang.reflect.Field;
import java.util.Set;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;

/**
 * Bluetooth functionality for Android >= 3.0, from which, the bluetooth APIs
 * are quite stable. If there is an exception on a special version of Android
 * a specific implementation should be made.
 */
class BluetoothImpl extends Bluetooth {
    private BluetoothHeadset mHeadset;

    public BluetoothImpl(Context context, HeadsetStateListener listener) {
        super(context, listener);
    }

    protected void createHeadset() {
        mHeadset = new BluetoothHeadset(mContext, getServiceListener());
    }

    @Override
    public int getHeadsetState() {
        mHeadsetConnected = (mHeadset.getState() == BluetoothHeadset.STATE_CONNECTED);
        return mHeadsetConnected ? HEADSET_STATE_CONNECTED : HEADSET_STATE_DISCONNECTED;
    }

    @Override
    public int getAudioState() {
        return mAudioConnected ? HEADSET_AUDIO_STATE_CONNECTED : HEADSET_AUDIO_STATE_DISCONNECTED;
    }

    @Override
    public int getPlaybackStream() {
        int streamIdForBt = Audio.VOICE_CALL_STREAM;
        try {
            Field field = ReflectionFactory.getField("android.media.AudioManager", "STREAM_BLUETOOTH_SCO");
            streamIdForBt = field.getInt(null);
        } catch (Exception e) {
            if (DEBUG) LOGD("getPlaybackStream, problem finding field STREAM_BLUETOOTH_SCO in class AudioManager");
        }
        if (DEBUG) LOGD("getPlaybackStream returns = "+streamIdForBt);
        return streamIdForBt;
    }

    @Override
    public int getRecordingSource() {
        return Audio.DEFAULT_AUDIO_SOURCE;
    }

    @Override
    public boolean startBluetoothSco() {
        boolean value = mHeadset.startVoiceRecognition();
        if(DEBUG) LOGD("mHeadset.startVoiceRecognition() returned "+ value);
        return value;
    }

    @Override
    public void stopBluetoothSco() {
        boolean value = mHeadset.stopVoiceRecognition();
        if(DEBUG) LOGD("mHeadset.stopVoiceRecognition() returned "+ value);
    }

    @Override
    public void close() {
        if(DEBUG) LOGD("close()");
        if(mStateReceiverRegistered) {
            mStateReceiverRegistered = false;
            mContext.unregisterReceiver(mStateReceiver);
        }
        if(mHeadset != null) {
            try {
                mHeadset.close();
            } catch(Throwable t) {
            }
            mHeadset = null;
        }
    }

    @Override
    public String[] getBondedDevices() {
        BluetoothAdapter adapter = getBluetoothAdapter();
        if(adapter == null) return null;
        Set<BluetoothDevice> devices = adapter.getBondedDevices();
        if(devices == null) return null;

        String[] names = new String[devices.size()];
        int i = 0;
        for(BluetoothDevice bd: devices) {
            String name = bd.getName();
            names[i++] = (name != null) ? name : bd.getAddress();
        }
        return names;
    }

    // Bluetooth headset service listener
    private BluetoothHeadset.ServiceListener mServiceListener;

    private BluetoothHeadset.ServiceListener getServiceListener() {
        if (mServiceListener == null) {
            mServiceListener = new BluetoothHeadset.ServiceListener() {
                public void onServiceConnected(boolean statusSuccess) {
                    if(DEBUG) LOGD("onServiceConnected(), BluetoothHeadset service connected");
                    if(DEBUG) LOGD("onServiceConnected(), mHeadset = " + mHeadset + ", mListener = " + mListener);
                    if( (mHeadset == null) || (!statusSuccess)) {
                        if (mListener != null) {
                            mListener.onServiceConnected(false);
                        }
                        return;
                    }
                    mServiceConnected = true;
                    mHeadsetConnected = (mHeadset.getState() == BluetoothHeadset.STATE_CONNECTED);
                    if(mListener != null) {
                        mListener.onServiceConnected(true);
                        mListener.onConnectionStateChanged(getHeadsetState());
                    }
                }

                public void onServiceDisconnected() {
                    if(DEBUG) LOGD("BluetoothHeadset service disconnected");
                    if (mListener != null) {
                        mListener.onServiceDisconnected();
                    }
                    mServiceConnected = false;
                }
            };
        }
        return mServiceListener;
    }

    // Bluetooth headset state receiver
    private BroadcastReceiver mStateReceiver;

    @Override
    protected BroadcastReceiver getStateReceiver() {
        if (mStateReceiver == null) {
            mStateReceiver = new BroadcastReceiver() {
                public void onReceive(Context context, Intent intent) {
                    String action = intent.getAction();
                    if(action.equals(BluetoothHeadset.ACTION_STATE_CHANGED)) {
                        int state = intent.getIntExtra(BluetoothHeadset.EXTRA_STATE, -1);
                        if(DEBUG) LOGD("Bluetooth state changed: " + state);
                        boolean connected = (state == BluetoothHeadset.STATE_CONNECTED);
                        if(mHeadsetConnected != connected) {
                            mHeadsetConnected = connected;
                            if(mListener != null) {
                                mListener.onConnectionStateChanged(getHeadsetState());
                            }
                        }
                    } else if(action.equals(BluetoothHeadset.ACTION_AUDIO_STATE_CHANGED)) {
                        int state = intent.getIntExtra(BluetoothHeadset.EXTRA_STATE, -1);
                        if(DEBUG) LOGD("Bluetooth audio state changed: " + state);
                        if(state == BluetoothHeadset.AUDIO_STATE_DISCONNECTED
                                || state == BluetoothHeadset.AUDIO_STATE_CONNECTED
                                || state == BluetoothHeadset.AUDIO_STATE_CONNECTING) {
                            boolean connected = (state == BluetoothHeadset.AUDIO_STATE_CONNECTED);
                            if(mAudioConnected != connected) {
                                mAudioConnected = connected;
                                if(mListener != null) {
                                    mListener.onAudioStateChanged(getAudioState());
                                }
                            }
                        }
                    }
                }
            };
        }
        return mStateReceiver;
    }

    @Override
    protected IntentFilter createFilter() {
        IntentFilter filter = new IntentFilter(BluetoothHeadset.ACTION_STATE_CHANGED);
        filter.addAction(BluetoothHeadset.ACTION_AUDIO_STATE_CHANGED);
        return filter;
    }
}
