/*------------------------------------------------------------------------

  Copyright 2006-2012 Nuance Communications, Inc.

------------------------------------------------------------------------*/
package com.nuance.dragon.toolkit.audio.bluetooth.impl;

import java.util.Set;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;

/**
 * Bluetooth functionality for Android 2.0 - 2.1.
 */
class Bluetooth_2_0 extends Bluetooth {
    private BluetoothHeadsetLegacy mHeadset;
    public Bluetooth_2_0(Context context, HeadsetStateListener listener) {
        super(context, listener);
    }

    @Override
    public int getHeadsetState() {
        return mHeadsetConnected ? HEADSET_STATE_CONNECTED : HEADSET_STATE_DISCONNECTED;
    }

    @Override
    public int getAudioState() {
        return mAudioConnected ? HEADSET_AUDIO_STATE_CONNECTED : HEADSET_AUDIO_STATE_DISCONNECTED;
    }

    @Override
    public int getPlaybackStream() {
        if(AndroidVersion.USE_MUSIC_STREAM_FOR_BLUETOOTH) {
            return Audio.MUSIC_STREAM;
        } else {
            return Audio.BLUETOOTH_STREAM;
        }
    }

    @Override
    public int getRecordingSource() {
        return Audio.DEFAULT_AUDIO_SOURCE;
    }

    @Override
    public boolean startBluetoothSco() {
        if(!AndroidVersion.USE_MUSIC_STREAM_FOR_BLUETOOTH) {
            mHeadset.startVoiceRecognition();
        }
        return true;
    }

    @Override
    public void stopBluetoothSco() {
        if(!AndroidVersion.USE_MUSIC_STREAM_FOR_BLUETOOTH && mHeadset != null) {
            mHeadset.stopVoiceRecognition();
        }
    }

    @Override
    public void close() {
        if(mStateReceiverRegistered) {
            mStateReceiverRegistered = false;
            mContext.unregisterReceiver(mStateReceiver);
        }
        if(mHeadset != null) {
            try {
                mHeadset.close();
            } catch(Throwable t) {
            }
            mHeadset = null;
        }
    }

    @Override
    public String[] getBondedDevices() {
        BluetoothAdapter adapter = getBluetoothAdapter();
        if(adapter == null) return null;
        Set<BluetoothDevice> devices = adapter.getBondedDevices();
        if(devices == null) return null;

        String[] names = new String[devices.size()];
        int i = 0;
        for(BluetoothDevice bd: devices) {
            String name = bd.getName();
            names[i++] = (name != null) ? name : bd.getAddress();
        }
        return names;
    }


    // Bluetooth headset service listener
    private BluetoothHeadsetLegacy.ServiceListener mServiceListener;

    private BluetoothHeadsetLegacy.ServiceListener getServiceListener() {
        if (mServiceListener == null) {
            mServiceListener = new BluetoothHeadsetLegacy.ServiceListener() {
                public void onServiceConnected() {
                    if(DEBUG) LOGD("BluetoothHeadset service connected");
                    if(mHeadset == null) {
                        if(mListener != null) {
                            mListener.onServiceConnected(false);
                        }
                        return;
                    }
                    mServiceConnected = true;
                    mHeadsetConnected = (mHeadset.getState() == BluetoothHeadsetLegacy.STATE_CONNECTED);
                    if(mListener != null) {
                        mListener.onServiceConnected(true);
                        mListener.onConnectionStateChanged(getHeadsetState());
                    }
                }

                public void onServiceDisconnected() {
                    if(DEBUG) LOGD("BluetoothHeadset service disconnected");
                    if(mListener != null) {
                        mListener.onServiceDisconnected();
                    }
                    mServiceConnected = false;
                }
            };
        }
        return mServiceListener;
    }

    // Bluetooth headset state receiver
    private BroadcastReceiver mStateReceiver;

    @Override
    protected BroadcastReceiver getStateReceiver() {
        if (mStateReceiver == null) {
            mStateReceiver = new BroadcastReceiver() {
                public void onReceive(Context context, Intent intent) {
                    String action = intent.getAction();
                    if(action.equals(BluetoothHeadsetLegacy.ACTION_STATE_CHANGED)) {
                        int state = intent.getIntExtra(BluetoothHeadsetLegacy.EXTRA_STATE, -1);
                        if(DEBUG) LOGD("Bluetooth state changed: " + state);
                        boolean connected = (state == BluetoothHeadsetLegacy.STATE_CONNECTED);
                        if(mHeadsetConnected != connected) {
                            mHeadsetConnected = connected;
                            if(mListener != null) {
                                mListener.onConnectionStateChanged(getHeadsetState());
                            }
                        }
                    } else if(action.equals(BluetoothHeadsetLegacy.ACTION_AUDIO_STATE_CHANGED)) {
                        int state = intent.getIntExtra(BluetoothHeadsetLegacy.EXTRA_AUDIO_STATE, -1);
                        if(DEBUG) LOGD("Bluetooth audio state changed: " + state);
                        // Ignore all states other than connected and disconnected.
                        // E.g., Espresso sends a state (2) that probably means connecting.
                        if(state == BluetoothHeadsetLegacy.AUDIO_STATE_CONNECTED ||
                                state == BluetoothHeadsetLegacy.AUDIO_STATE_DISCONNECTED) {
                            boolean connected = (state == BluetoothHeadsetLegacy.AUDIO_STATE_CONNECTED);
                            if(mAudioConnected != connected) {
                                mAudioConnected = connected;
                                if(mListener != null) {
                                    mListener.onAudioStateChanged(getAudioState());
                                }
                            }
                        }
                    }
                }
            };
        }
        return mStateReceiver;
    }

    @Override
    protected IntentFilter createFilter() {
        IntentFilter filter = new IntentFilter(BluetoothHeadsetLegacy.ACTION_STATE_CHANGED);
        filter.addAction(BluetoothHeadsetLegacy.ACTION_AUDIO_STATE_CHANGED);
        return filter;
    }

    @Override
    protected void createHeadset() {
        mHeadset = new BluetoothHeadsetLegacy(mContext, getServiceListener());
    }
}
