/*------------------------------------------------------------------------

  Copyright 2006-2012 Nuance Communications, Inc.

------------------------------------------------------------------------*/
package com.nuance.dragon.toolkit.audio.bluetooth.impl;

import java.util.Set;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.media.AudioManager;
/**
 * Bluetooth functionality for Android 2.2.
 */
class Bluetooth_2_2 extends Bluetooth {
    private BluetoothHeadsetLegacy mHeadset;
    private AudioManager mAudioManager;

    public Bluetooth_2_2(Context context, HeadsetStateListener listener) {
        super(context, listener);
        mAudioManager = (AudioManager) mContext.getSystemService(Context.AUDIO_SERVICE);
        LOGD("Bluetooth 2.2 created");
    }

    @Override
    public int getHeadsetState() {
        return mHeadsetConnected ? HEADSET_STATE_CONNECTED : HEADSET_STATE_DISCONNECTED;
    }

    @Override
    public int getAudioState() {
        return mAudioConnected ? HEADSET_AUDIO_STATE_CONNECTED : HEADSET_AUDIO_STATE_DISCONNECTED;
    }

    @Override
    public int getPlaybackStream() {
        if (AndroidVersion.USE_MUSIC_STREAM_FOR_BLUETOOTH) {
            LOGD("Bluetooth 2.2 playback stream = MUSIC_STREAM");
            return Audio.MUSIC_STREAM;
        } else {
            LOGD("Bluetooth 2.2 playback stream = VOICE_CALL_STREAM");
            return Audio.VOICE_CALL_STREAM;
        }
    }

    @Override
    public int getRecordingSource() {
        return Audio.DEFAULT_AUDIO_SOURCE;
    }

    @Override
    public boolean startBluetoothSco() {
        if (!AndroidVersion.USE_MUSIC_STREAM_FOR_BLUETOOTH) {
            mAudioManager.startBluetoothSco();
        }
        return true;
    }

    @Override
    public void stopBluetoothSco() {
        if (!AndroidVersion.USE_MUSIC_STREAM_FOR_BLUETOOTH) {
            mAudioManager.stopBluetoothSco();
        }
    }

    @Override
    public void close() {
        if (mStateReceiverRegistered) {
            mStateReceiverRegistered = false;
            mContext.unregisterReceiver(mStateReceiver);
        }
        if (mHeadset != null) {
            try {
                mHeadset.close();
            } catch (Throwable t) {}
            mHeadset = null;
        }
    }

    @Override
    public String[] getBondedDevices() {
        BluetoothAdapter adapter = getBluetoothAdapter();
        if(adapter == null) return null;
        Set<BluetoothDevice> devices = adapter.getBondedDevices();
        if (devices == null) return null;

        String[] names = new String[devices.size()];
        int i = 0;
        for (BluetoothDevice bd: devices) {
            String name = bd.getName();
            names[i++] = (name != null) ? name : bd.getAddress();
        }
        return names;
    }

    // Bluetooth headset service listener
    private BluetoothHeadsetLegacy.ServiceListener mServiceListener;

    private BluetoothHeadsetLegacy.ServiceListener getServiceListener() {
        if (mServiceListener == null) {
            mServiceListener = new BluetoothHeadsetLegacy.ServiceListener() {
                public void onServiceConnected() {
                    if (DEBUG) LOGD("BluetoothHeadset service connected mHeadset = " + mHeadset);
                    if (mHeadset == null) {
                        if(mListener != null) {
                            mListener.onServiceConnected(false);
                        }
                        return;
                    }
                    mServiceConnected = true;
                    mHeadsetConnected = (mHeadset.getState() == BluetoothHeadsetLegacy.STATE_CONNECTED);
                    if (mListener != null) {
                        mListener.onServiceConnected(true);
                        mListener.onConnectionStateChanged(getHeadsetState());
                    }
                }

                public void onServiceDisconnected() {
                    if (DEBUG) LOGD("BluetoothHeadset service disconnected");
                    mListener.onServiceDisconnected();
                    mServiceConnected = false;
                }
            };
        }

        return mServiceListener;
    }

    // Bluetooth headset state receiver
    private BroadcastReceiver mStateReceiver;

    @Override
    protected BroadcastReceiver getStateReceiver() {
        if (mStateReceiver == null) {
            mStateReceiver = new BroadcastReceiver() {
                public void onReceive(Context context, Intent intent) {
                    String action = intent.getAction();
                    if (action.equals(BluetoothHeadsetLegacy.ACTION_STATE_CHANGED)) {
                        int state = intent.getIntExtra(BluetoothHeadsetLegacy.EXTRA_STATE, -1);
                        if (DEBUG) LOGD("Bluetooth state changed: " + state);
                        boolean connected = (state == BluetoothHeadsetLegacy.STATE_CONNECTED);
                        if (mHeadsetConnected != connected) {
                            mHeadsetConnected = connected;
                            if (mListener != null) {
                                mListener.onConnectionStateChanged(getHeadsetState());
                            }
                        }
                    } else if (action.equals(AudioManager.ACTION_SCO_AUDIO_STATE_CHANGED)) {
                        int state = intent.getIntExtra(AudioManager.EXTRA_SCO_AUDIO_STATE, -1);
                        if (DEBUG) LOGD("Bluetooth audio state changed: " + state);
                        if (state == AudioManager.SCO_AUDIO_STATE_CONNECTED
                                || state == AudioManager.SCO_AUDIO_STATE_DISCONNECTED
                                || state == AudioManager.SCO_AUDIO_STATE_ERROR) {
                            boolean connected = (state == AudioManager.SCO_AUDIO_STATE_CONNECTED);
                            if (mAudioConnected != connected) {
                                mAudioConnected = connected;
                                if (mListener != null) {
                                    mListener.onAudioStateChanged(getAudioState());
                                }
                            }
                        }
                    }
                }
            };
        }
        return mStateReceiver;
    }

    @Override
    protected IntentFilter createFilter() {
        IntentFilter filter = new IntentFilter(BluetoothHeadsetLegacy.ACTION_STATE_CHANGED);
        filter.addAction(AudioManager.ACTION_SCO_AUDIO_STATE_CHANGED);
        return filter;
    }

    @Override
    protected void createHeadset() {
        mHeadset = new BluetoothHeadsetLegacy(mContext, getServiceListener());
    }
}
