/*------------------------------------------------------------------------

  Copyright 2006-2011 Nuance Communications, Inc.

------------------------------------------------------------------------*/
package com.nuance.dragon.toolkit.audio.bluetooth.impl;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

// TODO: should be converged with the other Reflection class that provides the
// same functionality.
/**
 * Reflection utilities. Simplifies dynamic object creation, field access,
 * and method invocation.
 */
public class ReflectionFactory {

//    private static final Logger LOGGER = new Logger("ReflectionFactory");

    private ReflectionFactory() {
    }

    public static Object newObject(String classname) {
        Object result = null;
        if (classname != null) {
            try {
                Class<?> clazz = Class.forName(classname);
                result = clazz.newInstance();
            } catch (ClassNotFoundException cnfe) {
                handleException(cnfe);
            } catch (IllegalAccessException iae) {
                handleException(iae);
            } catch (InstantiationException ie) {
                handleException(ie);
            }
        }
        return result;
    }

    @SuppressWarnings("rawtypes")
    public static Object newObject(Constructor construct, Object...args) {
        Object result = null;
        if (construct != null) {
            try {
                result = (Object)construct.newInstance(args);
            }  catch (IllegalAccessException iae) {
                handleException(iae);
            } catch (InstantiationException ie) {
                handleException(ie);
            } catch (InvocationTargetException ite) {
                handleException(ite);
            }
        }
        return result;
    }

    private static void handleException(Exception e) {
//        LOGGER.logException("Exception received", e);
    }

    public static Method getMethod(String className, String methodName, final Class<?>... parameterTypes) {
//        LOGGER.logD("getMethod " + methodName);
        Method result = null;

        if ((className != null) && (methodName != null)) {
            try {
                Class<?> clazz = Class.forName(className);
                result = clazz.getMethod(methodName, parameterTypes);
            } catch (ClassNotFoundException cnfe) {
                handleException(cnfe);
            } catch (IllegalArgumentException iare) {
                handleException(iare);
            } catch (NoSuchMethodException nsme) {
                handleException(nsme);
            }
        }
        return result;
    }

    public static Class<?> getClassForName(String name) {
        Class<?> clazz = null;
        try {
            clazz = Class.forName(name);
        } catch(ClassNotFoundException cnfe) {
            clazz = getPrimitiveClassForName(name);
            if (clazz == null) {
                handleException(cnfe);
            }
        }
        return clazz;
    }

    private static Class<?> getPrimitiveClassForName(String name) {
        Class<?> clazz = null;
        if ("int".equals(name)) {
            clazz = int.class;
        } else if ("boolean".equals(name)) {
            clazz = boolean.class;
        } else if ("float".equals(name)) {
            clazz = float.class;
        } else if ("double".equals(name)) {
            clazz = double.class;
        } else if ("char".equals(name)) {
            clazz = char.class;
        } else if ("byte".equals(name)) {
            clazz = byte.class;
        } else if ("void".equals(name)) {
            clazz = void.class;
        } else if ("short".equals(name)) {
            clazz = short.class;
        } else if ("long".equals(name)) {
            clazz = long.class;
        }
        return clazz;
    }

    public static Method getMethodFromStrings(String className, String methodName, String... parameterTypeNames) {
//        LOGGER.logD("getMethodFromStrings(String, String, String[]) " + methodName);
        Method result = null;

        Class<?>[] parameterTypes = new Class[parameterTypeNames.length];

        for(int i = 0; i < parameterTypes.length ; i++) {
            parameterTypes[i] = getClassForName(parameterTypeNames[i]);
        }

        if((className != null) && (methodName != null)) {
            try {
                Class<?> clazz = Class.forName(className);
                result = clazz.getMethod(methodName, parameterTypes);
            } catch(ClassNotFoundException cnfe) {
                handleException(cnfe);
            } catch(IllegalArgumentException iare) {
                handleException(iare);
            } catch(NoSuchMethodException nsme) {
                handleException(nsme);
            }
        }
        return result;
    }

    @SuppressWarnings("rawtypes")
    public static Constructor getConstructorFromStrings(String className, String... parameterTypeNames) {
        //LOGGER.logD("getMethodFromStrings(String, String, String[]) ");
        Constructor result = null;

        Class<?>[] parameterTypes = new Class[parameterTypeNames.length];

        for(int i = 0; i < parameterTypes.length ; i++) {
            parameterTypes[i] = getClassForName(parameterTypeNames[i]);
        }

        if((className != null)) {
            try {
                Class<?> clazz = Class.forName(className);
                result = clazz.getConstructor(parameterTypes);
            } catch(ClassNotFoundException cnfe) {
                handleException(cnfe);
            } catch(IllegalArgumentException iare) {
                handleException(iare);
            } catch(NoSuchMethodException nsme) {
                handleException(nsme);
            }
        }
        return result;
    }

    public static Field getField(String className, String fieldName) {
//        LOGGER.logD("getField class " + className + " field name " + fieldName);
        Field result = null;

        if ((className != null) && (fieldName != null)) {
            try {
                Class<?> clazz = Class.forName(className);
                result = clazz.getField(fieldName);
            } catch (ClassNotFoundException cnfe) {
                handleException(cnfe);
            } catch (NoSuchFieldException nsfe) {
                handleException(nsfe);
            }
        }
        return result;
    }
}
