package com.nuance.dragon.toolkit.audio.nat;

import java.util.LinkedList;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.Logger;

class NativeAudioImpl implements NativeAudio
{
    static {
        System.loadLibrary("dmt_audio");
    }

    native static long audioCreate();
    native static void audioDestroy(long audio);

    native static long recorderCreate(long audio, int sampleRate, int bufferSizeMs);
    native static int recorderStart(long recorder);
    native static int recorderStop(long recorder);
    native static int recorderRead(long recorder, short[] buffer);
    native static int getRecorderBufferSize(long recorder);
    native static int getRecorderBufferDuration(long recorder);
    native static void recorderDestroy(long recorder);

    native static long playerCreate(long audio, int sampleRate, int bufferSizeMs);
    native static boolean playerStart(long player);
    native static boolean playerStop(long player);
    native static boolean playerStalled(long player);
    native static boolean playerHasStartTime(long player);
    native static int getPlayerStartTime(long player);
    native static int playerWrite(long player, short[] buffer);
    native static void playerDestroy(long player);

    private static final class RecorderImpl implements Recorder
    {
        private long _recorder;
        private final short[] _buffer;
        private final int _bufferSizeMs;

        private AudioChunk _lastChunk;

        private RecorderImpl(long recorder)
        {
            Assert.assertTrue(recorder != 0);
            _recorder = recorder;

            // Constant buffer size
            int bufferSize = getRecorderBufferSize(_recorder);
            _buffer = new short[bufferSize];
            _bufferSizeMs = getRecorderBufferDuration(_recorder);
        }

        @Override
        public boolean start()
        {
            Assert.assertTrue(_recorder != 0);
            return 0 == recorderStart(_recorder);
        }

        @Override
        public boolean stop()
        {
            Assert.assertTrue(_recorder != 0);
            return 0 == recorderStop(_recorder);
        }

        @Override
        public int getBufferSizeMs()
        {
            return _bufferSizeMs;
        }

        @Override
        public AudioChunk read()
        {
            Assert.assertTrue(_recorder != 0);

            // Should always give getRecorderBufferSize() shorts
            // TODO: checking for 0 is not safe, since 0 could be a valid timestamp
            int timestamp = recorderRead(_recorder, _buffer);
            if (timestamp == 0)
                return null;

            // Since we might be polling recorderRead() frequently, don't
            // allocate the new buffer until we actually have valid audio

            short[] buffer = new short[_buffer.length];
            System.arraycopy(_buffer, 0, buffer, 0, buffer.length);

            AudioChunk chunk = new AudioChunk(AudioType.PCM_16k, buffer, timestamp);

            if (_lastChunk != null && _lastChunk.audioEndTimestamp != timestamp)
            {
                Logger.warn(this, "Audio timestamp mismatch: " + _lastChunk.audioTimestamp + " != " + timestamp);
            }
            _lastChunk = chunk;

            return chunk;
        }

        @Override
        public void release()
        {
            Assert.assertTrue(_recorder != 0);
            recorderDestroy(_recorder);
            _recorder = 0;
        }
    }

    private static final class PlayerImpl implements Player
    {
        private long _player;
        private boolean _startTimeDetermined;
        private int _startTime;
        private final LinkedList<short[]> _playingBuffers;
        private final LinkedList<AudioChunk> _playedBuffers;
        private int _playedDuration;

        private PlayerImpl(long player)
        {
            Assert.assertTrue(player != 0);
            _player = player;
            _playingBuffers = new LinkedList<short[]>();
            _playedBuffers = new LinkedList<AudioChunk>();
        }

        @Override
        public boolean start()
        {
            Assert.assertTrue(_player != 0);
            if (playerStart(_player))
            {
                return true;
            }
            return false;
        }

        @Override
        public boolean stop()
        {
            Assert.assertTrue(_player != 0);
            return playerStop(_player);
        }

        @Override
        public int write(short[] audio)
        {
            Assert.assertTrue(_player != 0);

            int ret = playerWrite(_player, audio);
            int length = audio.length;

            Assert.assertTrue(ret <= length);

            if (ret > 0)
            {
                if (ret < audio.length)
                {
                    short[] playingAudio = new short[ret];
                    System.arraycopy(audio, 0, playingAudio, 0, ret);
                    audio = playingAudio;
                }
                _playingBuffers.add(audio);
            }

            if (!_startTimeDetermined)
            {
                if (playerHasStartTime(_player))
                {
                    _startTimeDetermined = true;
                    _startTime = getPlayerStartTime(_player);
                    _playedDuration = 0;
                }
            }
            else
            {
                if (!playerHasStartTime(_player))
                {
                    _startTimeDetermined = false;
                }
            }

            if (_startTimeDetermined)
            {
                for (short[] buffer : _playingBuffers)
                {
                    AudioChunk chunk = new AudioChunk(AudioType.PCM_16k, buffer, _startTime + _playedDuration);
                    _playedDuration += chunk.audioDuration;
                    _playedBuffers.add(chunk);
                }
                _playingBuffers.clear();
            }

            return ret;
        }

        @Override
        public AudioChunk getPlayedAudio()
        {
            if (!_playedBuffers.isEmpty())
            {
                return _playedBuffers.remove();
            }
            return null;
        }

        @Override
        public boolean isStalled()
        {
            Assert.assertTrue(_player != 0);

            return playerStalled(_player);
        }

        @Override
        public void release()
        {
            Assert.assertTrue(_player != 0);
            playerDestroy(_player);
            _player = 0;
        }
    }

    private long _audio;

    @Override
    public boolean initialize()
    {
        if (_audio != 0)
            audioDestroy(_audio);

        _audio = audioCreate();
        return _audio != 0;
    }

    @Override
    public void release()
    {
        if (_audio != 0)
        {
            audioDestroy(_audio);
            _audio = 0;
        }
    }

    @Override
    public Recorder createRecorder(int sampleRate)
    {
        if (_audio != 0)
        {
            long recorder = recorderCreate(_audio, sampleRate, 400);
            RecorderImpl rec = new RecorderImpl(recorder);
            return rec;
        }
        return null;
    }

    @Override
    public Player createPlayer(int sampleRate)
    {
        if (_audio != 0)
        {
            long player = playerCreate(_audio, sampleRate, 400);
            if (player != 0)
            {
                PlayerImpl p = new PlayerImpl(player);
                return p;
            }
        }
        return null;
    }
}
