package com.nuance.dragon.toolkit.audio.nat;

import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.sinks.PlayerSink;
import com.nuance.dragon.toolkit.audio.sources.RecorderSource;
import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.WorkerThread;

/**
 * Manages the native audio system (based on OpenSLES). This manager is required
 * in order to create a {@link NativePlayerSink} or a {@link NativeRecorderSource}.
 */
public class NativeAudioManager
{
    private final WorkerThread _workerThread;
    private final NMTHandler _workerThreadHandler;
    private final NativeAudio _nativeAudio;
    private final List<WeakReference<RecorderSource<AudioChunk>>> _recorders;
    private final ArrayList<WeakReference<PlayerSink>> _players;

    /**
     * Create the audio manager. Be sure to call {@link #release} when finished.
     */
    public NativeAudioManager()
    {
        _workerThread = Factory.createWorkerThread("com.nuance.dragon.toolkit.audio.nat.NativeAudioManager");
        _nativeAudio = new NativeAudioImpl();

        _recorders = new ArrayList<WeakReference<RecorderSource<AudioChunk>>>();
        _players = new ArrayList<WeakReference<PlayerSink>>();

        _workerThread.start();
        _workerThreadHandler = _workerThread.getHandler();
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _nativeAudio.initialize();
            }
        });
    }

    /**
     * Releases the native audio manager
     */
    public void release()
    {
        // Make sure to free any recorders/players prior to freeing the audio
        for (WeakReference<RecorderSource<AudioChunk>> recorderRef : _recorders)
        {
            RecorderSource<AudioChunk> recorder = recorderRef.get();
            if (recorder != null)
                recorder.stopRecording();
        }

        for (WeakReference<PlayerSink> playerRef : _players)
        {
            PlayerSink player = playerRef.get();
            if (player != null)
                player.stopPlaying();
        }

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _nativeAudio.release();
            }
        });
        _workerThread.stop();
    }

    /**
     * @param sampleRate
     * @return A recorder source that uses the native recorder
     */
    public RecorderSource<AudioChunk> createRecorderSource(AudioType audioType)
    {
        NativeRecorderSource ret = new NativeRecorderSource(AudioType.PCM_16k, _nativeAudio, _workerThreadHandler);

        // Take this opportunity to clean up the recorder list
        for (int i = 0; i < _recorders.size(); i++)
        {
            WeakReference<RecorderSource<AudioChunk>> recorderRef = _recorders.get(i);
            RecorderSource<AudioChunk> recorder = recorderRef.get();
            if (recorder == null)
                _recorders.remove(i--);
        }
        _recorders.add(new WeakReference<RecorderSource<AudioChunk>>(ret));
        return ret;
    }

    /**
     * @param audioType
     * @return A player sink that uses the native player
     */
    public PlayerSink createPlayerSink(AudioType audioType)
    {
        NativePlayerSink ret = new NativePlayerSink(audioType, _nativeAudio, _workerThreadHandler);

        // Take this opportunity to clean up the player list
        for (int i = 0; i < _players.size(); i++)
        {
            WeakReference<PlayerSink> playerRef = _players.get(i);
            PlayerSink player = playerRef.get();
            if (player == null)
                _players.remove(i--);
        }
        _players.add(new WeakReference<PlayerSink>(ret));
        return ret;
    }
}
