package com.nuance.dragon.toolkit.audio.nat;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import android.os.Handler;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.sinks.PlayerSink;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;

/**
 * A player implementation that uses Android's native audio API (OpenSLES), for
 * slightly improved latency.
 * <p>
 * TODO: This is still in progress.
 */
class NativePlayerSink extends PlayerSink
{
    private static final int REQUEST_SIZE_MS = 400;

    private final NativeAudio _audio;
    private NativeAudio.Player _player;
    private boolean _playing;
    private boolean _started;
    private final LinkedList<short[]> _buffers;
    private boolean _noMore;

    NativePlayerSink(AudioType audioType, NativeAudio audio, NMTHandler workerThreadHandler)
    {
        super(audioType, workerThreadHandler);
        _audio = audio;
        _buffers = new LinkedList<short[]>();
    }

    @Override
    protected boolean isCodecSupported(AudioType audioType)
    {
        if (audioType.encoding ==  AudioType.Encoding.PCM_16 &&
                (audioType.frequency ==  AudioType.Frequency.FREQ_16KHZ))
            return true;

        return false;
    }

    @Override
    protected boolean startPlayingInternal(AudioType audioType)
    {
        if (_player != null)
        {
            return false;
        }

        _player = _audio.createPlayer(audioType.frequency);
        if (_player == null)
            return false;

        if (!_player.start())
        {
            _player.release();
            _player = null;
            return false;
        }

        _playing = true;

        final Handler handler = new Handler();
        Runnable runnable = new Runnable()
        {
            @Override
            public void run()
            {
                if (_player == null || !_playing)
                    return;

                boolean wrote = tryWrite();

                if (!_playing) // Is this necessary? TODO
                    return;

                if (!wrote && _noMore && _buffers.isEmpty())
                {
                    handleNoMore();
                    if (_playing) // If we're still playing, try again later
                    {
                        handler.postDelayed(this, REQUEST_SIZE_MS);
                    }
                } else
                {
                    handler.postDelayed(this, REQUEST_SIZE_MS / 2);
                }
            }
        };

        handler.post(runnable);

        return true;
    }

    @Override
    protected void stopPlayingInternal()
    {
        if (!_playing)
            return;

        _player.stop();
        _player.release();
        _player = null;
        _playing = false;
        handleStopped();
    }

    @Override
    protected void newChunksInternal(List<AudioChunk> chunks)
    {
        for (AudioChunk chunk : chunks)
        {
            if (chunk.audioShorts.length > 0)
                _buffers.add(chunk.audioShorts);
            else
                Logger.warn(this, "Skipping empty audio chunk");
        }

        // The SpeakerPlayerSink has this commented out, and allows the timer
        // task to do all the writing. But this results in slightly delayed
        // audio start... is it ok to also try to write here? TODO
        tryWrite();
    }

    @Override
    protected void noNewChunksInternal()
    {
        if (!_playing)
            return;

        handleNoMore();
    }

    private boolean tryWrite()
    {
        if (!_playing)
            return false;

        int written = 0;
        while (!_buffers.isEmpty())
        {
            short [] buffer = _buffers.remove();
            int ret = _player.write(buffer);

            if (ret < 0)
            {
                Logger.error(this, "Error writing audio buffer");
                break;
            }
            if (ret == 0)
            {
                if (buffer.length > 0)
                    _buffers.addFirst(buffer);
                break;
            }
            else
            {
                written += ret;
                if (ret < buffer.length)
                {
                    short [] leftover = new short[buffer.length - ret];
                    System.arraycopy(buffer, ret, leftover, 0, leftover.length);
                    _buffers.addFirst(leftover);
                    break;
                }
            }
        }

        if (_buffers.isEmpty())
        {
            handleAudioNeeded(REQUEST_SIZE_MS);
        }

        if (!_started && written > 0)
        {
            _started = true;
            handleStarted();
        }

        if (written > 0)
        {
            ArrayList<AudioChunk> playedChunks = new ArrayList<AudioChunk>();
            AudioChunk playedChunk;
            do
            {
                playedChunk = _player.getPlayedAudio();
                if (playedChunk != null)
                    playedChunks.add(playedChunk);
            } while (playedChunk != null);

            if (!playedChunks.isEmpty())
            {
                handleChunksPlayed(playedChunks);
            }
        }

        return written > 0;
    }

    private void handleNoMore()
    {
        _noMore = true;

        if (_player.isStalled() && _buffers.isEmpty())
        {
            stopPlayingInternal();
        }
    }
}
