package com.nuance.dragon.toolkit.audio.pipes;

import java.util.LinkedList;
import java.util.List;

import junit.framework.Assert;
import android.os.Handler;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.MultiChannelAudioChunk;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.AsyncTaskHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * A pipe that aligns a secondary audio stream with a primary audio stream for
 * the purposes of SSE echo cancellation. This pipe uses a basic cross-
 * correlation algorithm to determine the alignment of the two streams.
 */
public class AlignedMergerPipe extends SingleSinkPipe<AudioChunk, MultiChannelAudioChunk>
{
    /** Not for public use. */
    public static boolean LOAD_NATIVE_LIBRARY_SUCCESS = true;
    static {
        try {
            System.loadLibrary("dmt_native_utils");
        }
        catch (UnsatisfiedLinkError e) {
            Logger.error(AlignedMergerPipe.class, "Failed to load native library.", e);
            LOAD_NATIVE_LIBRARY_SUCCESS = false;
        }
    }

    /**
     * Guess the alignment of two audio buffers based on a simple cross-
     * correlation.
     * <p>
     * In a typical scenario, 'microphone' would be the raw recorded audio from
     * the microphone, and 'reference' would be the TTS audio that is being
     * played.
     * @param mic The 'microphone' input. This input is assumed to have 'ref' as
     * background noise.
     * @param ref The 'reference' input. This input is assumed to be present in
     * the background of the microphone input.
     * @param sampleRate The sampling rate of both the mic and ref audio.
     * @param minOffsetMs The minimum reference offset, in ms. This corresponds
     * to the offset of the reference audio in relation to the microphone audio.
     * @param maxOffsetMs The maximum reference offset, in ms.
     * @return The detected offset of the reference audio in relation to the
     * microphone audio. This value will be >= minOffsetMs and <= maxOffsetMs,
     * or it will be < minOffsetMs if an error occurs.
     * <p>
     * If the reference audio is determined to come before the microphone audio,
     * the return value will be negative. If the audio appears to be already
     * perfectly aligned, it will return 0.
     */
    native static int guessAudioAlignment(short[] mic, short[] ref, int sampleRate, int minOffsetMs, int maxOffsetMs);

    /**
     * Represents a duration, with both ms and sample count
     */
    private static final class Duration
    {
        final int ms;
        final int samples;

        Duration(AudioType type, int ms)
        {
            this.ms = ms;
            samples = type.getSampleCount(ms);
        }
    }

    private final AsyncTaskHandler _workerHandler;
    private final Handler _mainHandler;

    private AudioSource<AudioChunk> _secondarySource;
    private final AudioSink<AudioChunk> _secondarySink;
    private final LinkedList<AudioChunk> _primaryAudio;
    private final LinkedList<AudioChunk> _secondaryAudio;
    private final LinkedList<AudioChunk> _secondaryAudioExtra; // Secondary audio that isn't continuous
    private final LinkedList<MultiChannelAudioChunk> _mergedAudio;
    private final AudioType _audioType;
    private final Duration _frameSize;
    private long _nextTimestamp;
    private int _primarySampleCount;
    private int _secondarySampleCount;

    private final Duration _correlationWindow; // The window to use for the correlation

    private final Duration _minSecondaryAdjustment;
    private final Duration _maxSecondaryAdjustment;
    private Duration _secondaryAdjustment;

    private boolean _correlationDetermined;  // Indicates that the correlation window has been determined.
                                             // Prior to this, primary/secondary audio will not be merged.
    private Object _correlationDetermining;

    /**
     * Create an aligned merger pipe with a fixed alignment between the primary
     * and secondary audio. No cross-correlation will occur.
     * @param audioType The type of the input and output audio (must be PCM).
     * @param frameSizeMs The size of a frame. All merged multi-channel output
     * chunks will be this size.
     * @param secondaryAdjustmentMs The adjustment to make to the secondary
     * audio timestamps in order to make the primary and secondary audio line up
     * exactly.
     */
    public AlignedMergerPipe(AudioType audioType, int frameSizeMs, int secondaryAdjustmentMs)
    {
        this(audioType, frameSizeMs, secondaryAdjustmentMs, secondaryAdjustmentMs, 0);
    }

    /**
     * Create an aligned merger pipe with an unknown alignment. Cross-correlation
     * will be used to determine the actual alignment, using a search window
     * defined by the minimum and maximum secondary adjustment parameters.
     * <p>
     * The timestamps of the input audio chunks must be approximately
     * correct, so that they are within the bounds of search window. Note
     * that there will be an initial delay before output audio
     * becomes available, based on the values of the secondary
     * adjustment parameters and the correlation window size, due to the fact
     * that the audio must be correlated before the proper alignment can be
     * determined.
     * @param audioType The type of the input and output audio (must be PCM).
     * @param frameSizeMs The size of a frame. All merged multi-channel output
     * chunks will be a multiple of this size. IMPORTANT: Please make sure that
     * the frameSizeMs corresponds to an integer number of samples.
     * @param minSecondaryAdjustmentMs The minimum offset used in the search
     * for the aligning offset..
     * @param maxSecondaryAdjustmentMs The maximum offset used in the search
     * for the aligning offset..
     * @param correlationWindowMs The correlation window to use. A bigger value
     * may yield better results, but will be more CPU-intensive. If 0, the
     * cross-correlation will be disabled, and the delay will be assumed to be
     * maxSecondaryDelayMs.
     */
    public AlignedMergerPipe(AudioType audioType, int frameSizeMs, int minSecondaryAdjustmentMs, int maxSecondaryAdjustmentMs, int correlationWindowMs)
    {
        super(null);
        
        Checker.checkArgForNull("audioType", audioType);
        Checker.checkArgForCondition("correlationWindowMs", "greater than or equal to 0", correlationWindowMs >= 0);
        Checker.checkArgForCondition("frameSizeMs", "greater than 0", frameSizeMs > 0);
        Checker.checkArgForCondition("maxSecondaryOffsetMs", "greater than or equal to minSecondaryOffsetMs", maxSecondaryAdjustmentMs >= minSecondaryAdjustmentMs);

        _workerHandler = new AsyncTaskHandler();
        _mainHandler = new Handler();

        _correlationWindow = new Duration(audioType, correlationWindowMs);
        _minSecondaryAdjustment = new Duration(audioType, minSecondaryAdjustmentMs);
        _maxSecondaryAdjustment = new Duration(audioType, maxSecondaryAdjustmentMs);
        _secondaryAdjustment = _maxSecondaryAdjustment;

        _audioType = audioType;
        _frameSize = new Duration(audioType, frameSizeMs);
        _primaryAudio = new LinkedList<AudioChunk>();
        _secondaryAudio = new LinkedList<AudioChunk>();
        _secondaryAudioExtra = new LinkedList<AudioChunk>();
        _mergedAudio = new LinkedList<MultiChannelAudioChunk>();

        // Special value for the correlation window to disable the correlation
        if (correlationWindowMs == 0)
        {
            _correlationDetermined = true;
        }

        _secondarySink = new AudioSink<AudioChunk>(_mainThreadHandler)
        {
            @Override
            public void chunksAvailable(AudioSource<AudioChunk> source)
            {
                List<AudioChunk> chunks = source.getAllAudioChunksForSink(this);
                boolean added = false;

                for (AudioChunk chunk : chunks)
                {
                    int missing = 0;
                    if (!_secondaryAudio.isEmpty())
                    {
                        missing =  (int)(chunk.audioTimestamp - _secondaryAudio.getLast().audioEndTimestamp);
                    }

                    if (missing <= 1)
                    {
                        _secondarySampleCount += chunk.audioShorts.length;
                        _secondaryAudio.add(chunk);
                    } else
                    {
                        if (_secondaryAudioExtra.isEmpty())
                        {
                           Logger.warn(AlignedMergerPipe.this, "Got delayed secondary audio (" + missing + " missing)");
                        }
                        _secondaryAudioExtra.add(chunk);
                    }

                    added = true;
                }

                if (added)
                    checkAudioAndNotify();
            }

            @Override
            public void framesDropped(AudioSource<AudioChunk> source)
            {
                // Do we care?
            }

            @Override
            public void sourceClosed(AudioSource<AudioChunk> source)
            {
                // Do we care?
                // TODO: properly handle cases like a new secondary source
                // getting connected while we still have old audio
            }
        };
    }

    @Override
    public void connectAudioSource(AudioSource<AudioChunk> source)
    {
        _primaryAudio.clear();
        _primarySampleCount = 0;
        resetCorrelation();

        super.connectAudioSource(source);
    }

    /**
     * @return True if there is an active secondary source attached
     */
    private boolean isSecondaryActive()
    {
        return _secondarySource != null && _secondarySource.isActive();
    }

    /**
     * @return True if there is an active primary source attached
     */
    private boolean isPrimaryActive()
    {
        return isSourceActive();
    }

    /**
     * Trigger a check of the current available audio. Clean things up and merge
     * if possible.
     * @return true if new audio is available to the sink
     */
    private boolean checkAudio()
    {
        // Keep trying to trim as long as we keep loading more extra audio
        boolean trim;
        do
        {
            trim = false;
            if (_secondarySampleCount > 0)
            {
                _secondarySampleCount -= trimSecondary();
                Assert.assertTrue(_secondarySampleCount >= 0);
                if (_secondarySampleCount == 0)
                {
                    loadExtraSecondaryAudio();
                    trim = true;
                }
            }
        } while (trim);

        correlateIfNecessary();

        boolean merged = merge();

        boolean flushed = false;
        if (!isSourceActive()) // Flush all primary audio
            flushed = flushPrimary();

        boolean notify = merged || flushed;

        cleanupSecondaryExtra();

        if (merged && _secondarySampleCount == 0)
        {
            // We just emptied the secondary audio, so give this a chance to
            // retry with what's in the merged secondary audio
            notify = notify || checkAudio();
        }
        return notify;
    }

    /**
     * Trigger a check of the current available audio. Clean things up and merge
     * if possible. Notify the sink if new chunks are available.
     */
    private void checkAudioAndNotify()
    {
        if (checkAudio())
        {
            notifyChunksAvailable();
        }
    }

    /**
     * Reset the correlation state.
     */
    private void resetCorrelation()
    {
        if (_correlationWindow.ms != 0)
        {
            _correlationDetermined = false;
            _correlationDetermining = null;
            _secondaryAdjustment = _maxSecondaryAdjustment;
        }
    }

    /**
     * Loads audio from the extra buffer into the secondary buffer
     */
    private void loadExtraSecondaryAudio()
    {
        Assert.assertTrue(_secondaryAudio.isEmpty());

        while (!_secondaryAudioExtra.isEmpty())
        {
            AudioChunk chunk = _secondaryAudioExtra.peek();

            if (_secondaryAudio.isEmpty() || chunk.audioTimestamp == _secondaryAudio.getLast().audioEndTimestamp)
            {
                _secondaryAudio.add(_secondaryAudioExtra.removeFirst());
                _secondarySampleCount += chunk.audioShorts.length;
            } else
            {
                break;
            }
        }
    }

    /**
     * Clean up the secondary extra audio, i.e. get rid of stuff that we don't
     * need to hold on to:
     * <br> a) Audio that comes after the last primary audio (when the primary
     * source isn't active)
     * <br> b) Extra audio in excess of the maximum reference delay
     */
    private void cleanupSecondaryExtra()
    {
        if (isPrimaryActive())
            return;

        if (_secondaryAudioExtra.isEmpty())
            return;

        for (int i = 0; i < _secondaryAudioExtra.size(); i++)
        {
            AudioChunk chunk = _secondaryAudioExtra.get(i);
            int toSecondaryEnd = (int)(_secondaryAudioExtra.getLast().audioEndTimestamp - chunk.audioEndTimestamp);
            if (toSecondaryEnd < -_secondaryAdjustment.ms) // TODO: check math
            {
                // Keep the audio if it is within the maximum delay from the last buffer
                continue;
            }

            if (!_primaryAudio.isEmpty())
            {
                int diff = (int)(chunk.audioTimestamp - _primaryAudio.getLast().audioEndTimestamp);
                if (diff < _maxSecondaryAdjustment.ms) // TODO: check math
                {
                    // Keep the audio if it can potentially be merged with primary audio
                    continue;
                }
            }

            // We don't need this chunk
            _secondaryAudioExtra.remove(i--);
        }
    }

    /**
     * Flush all remaining primary audio to the merged audio (unless a
     * correlation is in progress)
     * @return true if audio was flushed
     */
    private boolean flushPrimary()
    {
        if (!_primaryAudio.isEmpty())
        {
            if (_correlationDetermining == null)
            {
                // Only flush if there is no correlation in progress
                _mergedAudio.add(new MultiChannelAudioChunk(removeAudio(_primaryAudio, _primarySampleCount)));
                _primarySampleCount = 0;
                return true;
            }
        }
        return false;
    }

    /**
     * Trim excess secondary audio based on the start of the primary audio. Also
     * will add padding to the start of the secondary audio if it comes later
     * than the primary audio.
     * @return the number of samples that were trimmed (negative if samples were
     * added)
     */
    private int trimSecondary()
    {
        if (_secondaryAudio.isEmpty())
            return 0;

        if (!_primaryAudio.isEmpty())
        {
            long primaryTime = _primaryAudio.peek().audioTimestamp;
            long secondaryTime = _secondaryAudio.peek().audioTimestamp;
            long effectiveSecondaryTime;

            if (_correlationDetermined)
                effectiveSecondaryTime = secondaryTime + _secondaryAdjustment.ms;
            else
                effectiveSecondaryTime = secondaryTime + _maxSecondaryAdjustment.ms;

            int timeDiff = (int)(primaryTime - effectiveSecondaryTime);
            if (timeDiff > 0)
            {
                // Need to chop off some secondary from the beginning
                //long endTime = _secondaryAudio.getLast().audioEndTimestamp;

                int toChop = timeDiff;
                int toChopSamples = _audioType.getSampleCount(toChop);
                if (toChopSamples >= _secondarySampleCount)
                    toChopSamples = _secondarySampleCount;

                int chopped = removeAndDiscardAudio(_secondaryAudio, toChopSamples);

                if (!_secondaryAudio.isEmpty()) // TODO: remove
                {
                    //Assert.assertTrue(endTime == _secondaryAudio.getLast().audioEndTimestamp);
                }

                Logger.warn(this, "Removing " + toChop + "ms from secondary audio");
                return chopped;
            }
        } else if (!isPrimaryActive())
        {
            final int maxMsToKeep = _maxSecondaryAdjustment.ms - _minSecondaryAdjustment.ms; // TODO: check this math
            final long endTime;
            if (_secondaryAudioExtra.isEmpty())
                endTime = _secondaryAudio.getLast().audioEndTimestamp;
            else
                endTime = _secondaryAudioExtra.getLast().audioEndTimestamp;

            final long keepAfter = endTime - maxMsToKeep;
            final long startTime = _secondaryAudio.getFirst().audioTimestamp;
            final int toDelete = (int)(keepAfter - startTime);

            if (toDelete > 0)
            {
                int chopped = removeAndDiscardAudio(_secondaryAudio, _audioType.getSampleCount(toDelete));
                return chopped;
            }
        }
        return 0;
    }

    /**
     * Performs asynchronous correlation of the secondary + primary sources if:
     * <br> a) it hasn't happened already, and
     * <br> b) there is sufficient audio buffered for the correlation
     */
    private void correlateIfNecessary()
    {
        if (_correlationDetermined || _correlationDetermining != null)
            return;

        if (_secondaryAudio.isEmpty() || _primaryAudio.isEmpty())
            return;

        final int correlationWindowMs = _correlationWindow.samples;
        final int curSecondaryAdjustmentMs = (int)(_primaryAudio.peek().audioTimestamp - _secondaryAudio.peek().audioTimestamp);

        final int minOffsetMs = (_minSecondaryAdjustment.ms - curSecondaryAdjustmentMs);
        final int maxOffsetMs = (_maxSecondaryAdjustment.ms - curSecondaryAdjustmentMs);

        int neededPrimarySamples = correlationWindowMs;
        int neededSecondarySamples = correlationWindowMs;

        if (minOffsetMs < 0)
            neededSecondarySamples += _audioType.getSampleCount(-minOffsetMs);
        if (maxOffsetMs > 0)
            neededPrimarySamples += _audioType.getSampleCount(maxOffsetMs);

        // If the secondary source is closed, we want to try the cross correlation
        // no matter how many samples we have
        if (!isSecondaryActive() && _secondarySampleCount > 0 && _secondarySampleCount < neededSecondarySamples)
            neededSecondarySamples = _secondarySampleCount;

        if (_primarySampleCount < neededPrimarySamples || _secondarySampleCount < neededSecondarySamples)
            return;

        // Assume secondaryAudio is ahead by maxRefDelay
        AudioChunk primaryFrame = peekAudio(_primaryAudio, neededPrimarySamples);
        AudioChunk secondaryFrame = peekAudio(_secondaryAudio, neededSecondarySamples);

        final short[] primaryAudio = primaryFrame.audioShorts;
        final short[] secondaryAudio = secondaryFrame.audioShorts;

        _correlationDetermined = false;
        final Object correlation = new Object();
        _correlationDetermining = correlation;

        _workerHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final int offsetMs = guessAudioAlignment(primaryAudio, secondaryAudio, _audioType.frequency, minOffsetMs, maxOffsetMs);
                _mainHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (offsetMs >= minOffsetMs && offsetMs <= maxOffsetMs)
                        {
                            if (correlation != _correlationDetermining) // TODO: would this ever happen?
                                return;

                            int newSecondaryAdjustment = curSecondaryAdjustmentMs + offsetMs;
                            Assert.assertTrue(_minSecondaryAdjustment.ms <= newSecondaryAdjustment && newSecondaryAdjustment <= _maxSecondaryAdjustment.ms);
                            Logger.debug(AlignedMergerPipe.this, "New secondary adjustment: " + newSecondaryAdjustment + "ms");

                            _secondaryAdjustment = new Duration(_audioType, newSecondaryAdjustment);
                        } else
                        {
                            Logger.error(AlignedMergerPipe.this, "Got invalid alignment " + offsetMs + " (expected between " + minOffsetMs + " and " + maxOffsetMs + ")");

                            // TODO: this is not at all the right thing to do. But what should we do instead?
                            // Maybe disable merging completely and simply pass all primary audio through?
                            _secondaryAdjustment = new Duration(_audioType, (minOffsetMs + maxOffsetMs) / 2);
                        }
                        _correlationDetermined = true;
                        _correlationDetermining = null;

                        checkAudioAndNotify();

                        // The source may have closed while in the middle of the
                        // correlation. If so, notify now.
                        if (!isSourceActive())
                        {
                            notifySourceClosed();
                        }
                    }
                });
            }
        });
    }

    /**
     * Merges the audio from the secondary + primary sources
     * @return True if any primary audio was transferred to the merged audio
     */
    private boolean merge()
    {
        if (_primaryAudio.isEmpty())
            return false;

        if (_secondaryAudio.isEmpty() && isSecondaryActive())
            return false; // Wait for more secondary

        boolean merged = false;

        if (_secondaryAudio.isEmpty())
        {
            if (_primarySampleCount > 0)
            {
                // Make sure we're always merging full frames
                int samplesToMerge = _primarySampleCount - (_primarySampleCount % _frameSize.samples);
                if (samplesToMerge > 0)
                {
                    AudioChunk chunk = removeAudio(_primaryAudio, samplesToMerge);
                    _primarySampleCount -= chunk.audioShorts.length;
                    _mergedAudio.add(new MultiChannelAudioChunk(chunk));
                    merged = true;

                    Logger.warn(this, "Passing primary audio through due to empty secondary audio");
                }
            }
        }
        else
        {
            long primaryTime = _primaryAudio.peek().audioTimestamp;
            long secondaryTime = _secondaryAudio.peek().audioTimestamp;
            long adjustedSecondaryTime;
            if (_correlationDetermined)
                adjustedSecondaryTime = secondaryTime + _secondaryAdjustment.ms;
            else
                adjustedSecondaryTime = secondaryTime + _minSecondaryAdjustment.ms;

            int excessPrimaryMs = (int)(adjustedSecondaryTime - primaryTime);
            if (excessPrimaryMs > 0)
            {
                // Move excess single-channel chunks to merged audio
                // But, make sure we're always merging full frames!
                int excessPrimarySamples = _audioType.getSampleCount(excessPrimaryMs);
                int remainder = (excessPrimarySamples % _frameSize.samples);

                // The whole samples
                //int samplesToMerge = excessPrimarySamples - remainder;

                if (remainder > 0)
                {
                    // We don't have full primary frames before the secondary audio starts,
                    // so chop off a little bit of secondary so that the single-channel
                    // primary can be big enough for a full frame.
                    int secondarySamplesToDiscard = _frameSize.samples - remainder;
                    if (secondarySamplesToDiscard <= _secondarySampleCount)
                    {
                        int removed = removeAndDiscardAudio(_secondaryAudio, secondarySamplesToDiscard);
                        Assert.assertEquals(removed, secondarySamplesToDiscard);
                        _secondarySampleCount -= removed;
                        excessPrimarySamples += removed;
                    }

                    // Else, it should be ok to leave things alone until later,
                    // because there's not enough secondary to merge anyway
                }

                // Ensure that we're not trying to pass through more audio than
                // is available
                if (excessPrimarySamples > _primarySampleCount)
                    excessPrimarySamples = _primarySampleCount - (_primarySampleCount % _frameSize.samples);

                if (excessPrimarySamples > 0)
                {
                    AudioChunk chunk = removeAudio(_primaryAudio, excessPrimarySamples);
                    _primarySampleCount -= chunk.audioShorts.length;
                    Assert.assertTrue(_primarySampleCount >= 0);

                    _mergedAudio.add(new MultiChannelAudioChunk(chunk));
                    merged = true;

                    Logger.warn(this, "Passing " + chunk.audioDuration + "ms excess primary audio through");
                }
            }

            if (_correlationDetermined)
            {
                if (_secondarySampleCount < _frameSize.samples && !isSecondaryActive())
                {
                    // Pad the secondary audio (TODO: maybe we should look at the extra secondary audio first?)
                    int padding = _frameSize.samples - _secondarySampleCount;
                    _secondaryAudio.add(new AudioChunk(_audioType, new short[padding], _secondaryAudio.getLast().audioEndTimestamp));
                    _secondarySampleCount = _frameSize.samples;
                }

                int samplesToRead;
                if (_primarySampleCount < _secondarySampleCount)
                    samplesToRead = _primarySampleCount;
                else
                    samplesToRead = _secondarySampleCount;

                if (samplesToRead >= _frameSize.samples)
                {
                    // Make sure it's a multiple of the frame size
                    samplesToRead -= (samplesToRead % _frameSize.samples);

                    AudioChunk primary = removeAudio(_primaryAudio, samplesToRead);
                    AudioChunk secondary = removeAudio(_secondaryAudio, primary.audioType, samplesToRead, primary.audioTimestamp - _frameSize.ms);

                    Assert.assertEquals(primary.audioShorts.length, samplesToRead);
                    Assert.assertEquals(secondary.audioShorts.length, samplesToRead);

                    _primarySampleCount -= samplesToRead;
                    _secondarySampleCount -= samplesToRead;

                    _mergedAudio.add(new MultiChannelAudioChunk(primary, secondary));
                    merged = true;
                }

                Assert.assertTrue(_secondarySampleCount >= 0);
                Assert.assertTrue(_primarySampleCount >= 0);
            }
        }
        return merged;
    }

    /**
     * Connect a secondary audio source. The secondary audio will be merged with
     * the primary audio. Secondary audio chunks will be discarded if there is
     * no primary audio to merge them with.
     * <p>
     * If no secondary audio source is connected, the primary audio will be
     * converted directly to a multi-channel audio chunk with 1 channel.
     * @param source The secondary source
     */
    public void connectSecondarySource(AudioSource<AudioChunk> source)
    {
        Checker.checkArgForCondition("source", "the correct audio type", source.getAudioType().equals(_audioType));

        _secondarySource = source;
        _secondarySink.connectAudioSource(_secondarySource);
    }

    /**
     * Disconnect the secondary audio source. Merged audio production will
     * continue as long as there is a primary source connected.
     * @return The disconnected source, or null if there was no source to
     * disconnect.
     */
    public AudioSource<AudioChunk> disconnectSecondarySource()
    {
        _secondarySource = null;
        return _secondarySink.disconnectAudioSource();
    }

    @Override
    public AudioType getAudioType()
    {
        return _audioType;
    }

    @Override
    public int getChunksAvailable()
    {
        return _mergedAudio.size();
    }

    @Override
    public boolean isActive()
    {
        return isSourceActive();
    }

    @Override
    protected MultiChannelAudioChunk getAudioChunk()
    {
        if (!_mergedAudio.isEmpty())
            return _mergedAudio.remove();
        return null;
    }

    @Override
    protected void chunksAvailable(AudioSource<AudioChunk> fromSource, AudioSink<AudioChunk> internalSink)
    {
        List<AudioChunk> chunks = fromSource.getAllAudioChunksForSink(internalSink);
        for (AudioChunk chunk : chunks)
        {
            if (_primaryAudio.isEmpty())
            {
                _primaryAudio.add(chunk);
                _primarySampleCount += chunk.audioShorts.length;
            } else
            {
                long prevEnd = _primaryAudio.getLast().audioEndTimestamp;
                int missingMs = (int)(chunk.audioTimestamp - prevEnd);
                if (missingMs == 0)
                {
                    _primaryAudio.add(chunk);
                    _primarySampleCount += chunk.audioShorts.length;
                }
                else
                {
                    // TODO: is there a better way to handle cases where we have a gap in the input audio?
                    if (missingMs > 0)
                    {
                        Logger.warn(this, "Inserting silence to preserve time alignment");

                        int missingSamples = _audioType.getSampleCount(missingMs);
                        _primaryAudio.add(new AudioChunk(_audioType, new short[missingSamples], prevEnd));
                        _primarySampleCount += missingSamples;

                        _primaryAudio.add(chunk);
                        _primarySampleCount += chunk.audioShorts.length;
                    } else
                    {
                        Logger.error(this, "Discarding audio buffer because it overlaps previous buffer");
                        notifyFramesDropped();
                    }
                }
            }
        }

        checkAudioAndNotify();
    }

    @Override
    protected void framesDropped(AudioSource<AudioChunk> fromSource, AudioSink<AudioChunk> internalSink)
    {
        // notifyFramesDropped();
        // We'll see the dropped frame when we see the timings mismatch
    }

    @Override
    protected void sourceClosed(AudioSource<AudioChunk> fromSource, AudioSink<AudioChunk> internalSink)
    {
        boolean checkAudio = false;

        int silencePadding = _primarySampleCount % _frameSize.samples;
        if (silencePadding > 0)
        {
            AudioChunk newChunk = new AudioChunk(_audioType, new short[silencePadding], _nextTimestamp);
            _nextTimestamp += newChunk.audioDuration;
            _primaryAudio.add(newChunk);
            _primarySampleCount += silencePadding;

            checkAudio = true;
        }

        // Push the secondary audio that comes after the final primary audio into
        // the extra buffers
        if (!_primaryAudio.isEmpty())
        {
            long lastTime = _primaryAudio.getLast().audioEndTimestamp;
            int startIndex;
            for (startIndex = 0; startIndex < _secondaryAudio.size(); startIndex++)
            {
                AudioChunk secondaryChunk = _secondaryAudio.get(startIndex);
                if (secondaryChunk.audioTimestamp - lastTime > 0)
                {
                    break;
                }
            }

            for (int i = _secondaryAudio.size() - 1; i >= startIndex; i--)
            {
                AudioChunk chunk = _secondaryAudio.remove(i);
                _secondarySampleCount -= chunk.audioShorts.length;
                _secondaryAudioExtra.addFirst(chunk);

                checkAudio = true;
            }

            Assert.assertTrue(_secondarySampleCount >= 0);
        }

        // If the primary source closed, and it has audio, but we never had enough
        // for the correlation, just pass the primary audio through
        if (!_primaryAudio.isEmpty())
        {
            checkAudio = true;
        }

        if (checkAudio)
            checkAudioAndNotify();

        // Only notify of the source closing if we're not currently determining
        // the correlation. If the correlation is in progress, the notification
        // will occur when it finishes.
        if (_correlationDetermining == null)
            notifySourceClosed();
    }

    /**
     * Remove a frame, and optionally return the removed audio
     * @param chunks
     * @param audioType The audio type of the audio
     * @param numSamples
     * @param newTimestamp The timestamp for the new frame
     * @return The audio
     */
    private static AudioChunk removeAudio(LinkedList<AudioChunk> chunks, AudioType audioType, final int numSamples, long newTimestamp)
    {
        short [] newAudio = new short[numSamples];
        int offset = 0;
        while (!chunks.isEmpty() && offset < numSamples)
        {
            int reqLen = numSamples - offset;

            AudioChunk removed = chunks.remove();

            short [] audio = removed.audioShorts;

            if (audio.length <= reqLen)
            {
                System.arraycopy(audio, 0, newAudio, offset, audio.length);
                offset += audio.length;
            }
            else
            {
                System.arraycopy(audio, 0, newAudio, offset, reqLen);
                offset = numSamples;

                short [] leftover = new short[audio.length - reqLen];
                System.arraycopy(audio, reqLen, leftover, 0, audio.length - reqLen);

                chunks.addFirst(new AudioChunk(removed.audioType, leftover, removed.audioTimestamp + removed.audioType.getDuration(reqLen)));
            }
        }

        if (offset < numSamples)
        {
            // Need to re-alloc smaller buffer
            short [] fixedAudio = new short[offset];
            System.arraycopy(newAudio, 0, fixedAudio, 0, fixedAudio.length);
            newAudio = fixedAudio;
        }

        return new AudioChunk(audioType, newAudio, newTimestamp);
    }

    /**
     * Remove a frame from chunks, and return the resulting AudioChunk
     * @param chunks
     * @param numSamples
     * @return The new audio frame
     */
    private static AudioChunk removeAudio(LinkedList<AudioChunk> chunks, int numSamples)
    {
        return removeAudio(chunks, chunks.peek().audioType, numSamples, chunks.peek().audioTimestamp);
    }

    /**
     * Return a frame from chunks, but don't remove it from chunks
     * @param chunks
     * @param sampleCount
     * @return The new audio frame
     */
    private static AudioChunk peekAudio(List<AudioChunk> chunks, int sampleCount)
    {
        short [] newAudio = new short[sampleCount];
        int offset = 0;
        AudioChunk firstChunk = chunks.get(0);
        for (AudioChunk chunk : chunks)
        {
            int toCopy = sampleCount - offset;
            if (toCopy > chunk.audioShorts.length)
                toCopy = chunk.audioShorts.length;

            System.arraycopy(chunk.audioShorts, 0, newAudio, offset, toCopy);
            offset += toCopy;

            if (offset >= sampleCount)
                break;
        }

        return new AudioChunk(firstChunk.audioType, newAudio, firstChunk.audioTimestamp);
    }

    /**
     * Removes a frame and discards it
     * @param chunks
     * @param numSamples
     * @return the number of samples removed
     */
    private static int removeAndDiscardAudio(LinkedList<AudioChunk> chunks, final int numSamples)
    {
        int removedSamples = 0;
        while (!chunks.isEmpty() && removedSamples < numSamples)
        {
            int reqLen = numSamples - removedSamples;

            AudioChunk removed = chunks.remove();

            short [] audio = removed.audioShorts;

            if (audio.length <= reqLen)
            {
                removedSamples += audio.length;
            }
            else
            {
                removedSamples = numSamples;

                short [] leftover = new short[audio.length - reqLen];
                System.arraycopy(audio, reqLen, leftover, 0, audio.length - reqLen);

                chunks.addFirst(new AudioChunk(removed.audioType, leftover, removed.audioTimestamp + removed.audioType.getDuration(reqLen)));
            }
        }
        return removedSamples;
    }
}
