package com.nuance.dragon.toolkit.audio.pipes;

import java.util.LinkedList;

import android.os.Handler;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk.SpeechStatus;
import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioEnergyListener;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * A simple pipe that provides audio energy notifications, based on the
 * {@link AudioChunk#getAudioEnergyLevel} and {@link AudioChunk#audioDuration},
 * this class will deliver the audio energy notifications via {@link AudioEnergyListener}
 * one at a time.
 */
public class AudioEnergyCalculatorPipe extends BufferingPipe<AudioChunk>
{
    private final AudioEnergyListener _listener;
    private final LinkedList<AudioChunk> _buffer;
    private static final int ENERGY_AUDIO_CHUNK_DURATION = 20; //millisecond
    private final Handler _mainThreadHandler;
    private boolean _newChunk;

    /**
     * AudioEnergyCalculatorPipe constructor
     *
     * @param listener A AudioEnergyListener that receives notifications
     */
    public AudioEnergyCalculatorPipe(AudioEnergyListener listener)
    {
        Checker.checkArgForNull("AudioEnergyListener", listener);
        _listener = listener;
        _buffer = new LinkedList<AudioChunk>();
        _mainThreadHandler = new Handler();
    }

    @Override
    protected void sourceClosed(AudioSource<AudioChunk> fromSource, AudioSink<AudioChunk> internalSink)
    {
        _buffer.clear();
        super.sourceClosed(fromSource, internalSink);
    }

    @Override
    protected void onChunkBuffered(AudioChunk chunk)
    {
        _newChunk = true;

        if (chunk != null)
        {
            if (chunk.audioDuration > ENERGY_AUDIO_CHUNK_DURATION)
            {
                AudioChunk[] smallChunks = splitAudioChunks(chunk);
                for (AudioChunk smallChunk : smallChunks)
                {
                    _buffer.add(smallChunk);
                }
            }
            else
            {
                _buffer.add(chunk);
            }
        }

        if (_buffer.size() > 0)
        {
            _mainThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    if (_buffer.size() <= 0)
                        return;

                    AudioChunk smallChunk = _buffer.remove();
                    float energyLevel = smallChunk.getAudioEnergyLevel();
                    _listener.onEnergyLevelAvailable(energyLevel, smallChunk.audioSpeechStatus == SpeechStatus.SPEECH);

                    if (_newChunk)
                    {
                        _newChunk = false;
                        _mainThreadHandler.post(this);
                    }
                    else
                    {
                        _mainThreadHandler.postDelayed(this, smallChunk.audioDuration);
                    }
                }
            });
        }

        super.onChunkBuffered(chunk);
    }

    /**
     * Assuming the bigChunk's audioDuration is multiplier of ENERGY_AUDIO_CHUNK_DURATION.
     *
     * @param bigChunk
     * @return an array of smaller AudioChunks
     */
    private AudioChunk[] splitAudioChunks(AudioChunk bigChunk)
    {
        int numOfChunks = bigChunk.audioDuration / ENERGY_AUDIO_CHUNK_DURATION;
        if (bigChunk.audioDuration % ENERGY_AUDIO_CHUNK_DURATION != 0)
        {
            Logger.warn(this, "the given big Chunk's duration is not the multipler of " + ENERGY_AUDIO_CHUNK_DURATION
                        + "ms we may miss some audio in the end for calculating the audio energy level.");
        }
        AudioChunk[] chunks = new AudioChunk[numOfChunks];
        int audioShortsBufLength = bigChunk.audioShorts.length / numOfChunks;
        for (int i = 0; i < numOfChunks; i++)
        {
            short[] buffer = new short[audioShortsBufLength];
            System.arraycopy(bigChunk.audioShorts, i * audioShortsBufLength, buffer, 0, audioShortsBufLength);
            long timeStamp = 0;
            if (bigChunk.audioTimestamp > 0)
            {
                timeStamp = bigChunk.audioTimestamp + (long) bigChunk.audioDuration / (long) numOfChunks;
            }
            chunks[i] = new AudioChunk(bigChunk.audioType, buffer, timeStamp);
        }

        return chunks;
    }
}
