package com.nuance.dragon.toolkit.audio.pipes;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CopyOnWriteArrayList;

import android.util.Pair;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk;
import com.nuance.dragon.toolkit.audio.AudioPipe;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.util.MultiIteratorQueue;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Similar than {@link DuplicatorPipe} but with built-in buffering abilities
 * so sinks can be connected at different times but still get the same audio stream.
 */
public class BufferingDuplicatorPipe<AudioChunkType extends AbstractAudioChunk> extends AudioPipe<AudioChunkType, AudioChunkType>
{
    private final List<AudioSink<AudioChunkType>> _currentAudioSinks;
    private final MultiIteratorQueue<AudioChunkType> _buffers;
    private final List<MultiIteratorQueue<AudioChunkType>.Iterator> _iterators;
    private final Map<AudioSink<AudioChunkType>, Pair<Integer, Integer>> _cachedAudioInfoForSink;
    private final List<AudioChunkType> _cache;
    private final int _expectedSinkCount;

    /**
     * BufferingDuplicatorPipe constructor
     * @param expectedSinkCount The number of sinks to be expected to connect to this duplicator. 
     */
    public BufferingDuplicatorPipe(int expectedSinkCount)
    {
        super(null);
        
        Checker.checkArgForCondition("expectedSinkCount", "> 0", expectedSinkCount > 0);

        _currentAudioSinks = new CopyOnWriteArrayList<AudioSink<AudioChunkType>>();
        _buffers = new MultiIteratorQueue<AudioChunkType>();
        _iterators = new ArrayList<MultiIteratorQueue<AudioChunkType>.Iterator>();
        _cachedAudioInfoForSink = new HashMap<AudioSink<AudioChunkType>,  Pair<Integer, Integer>>(expectedSinkCount);
        _expectedSinkCount = expectedSinkCount; 
        _cache = new ArrayList<AudioChunkType>(); 
    }

    @Override
    public AudioType getAudioType()
    {
        return getAudioTypeFromSource();
    }

    @Override
    public int getChunksAvailableForSink(AudioSink<AudioChunkType> sink)
    {
        int index = _currentAudioSinks.indexOf(sink);
        if (index != -1)
        {
            int cachedSize = 0;
            Pair<Integer, Integer> cachedAudioInfo = _cachedAudioInfoForSink.get(sink);
            if (cachedAudioInfo != null)
            {
                cachedSize = _cachedAudioInfoForSink.get(sink).second - _cachedAudioInfoForSink.get(sink).first;
            }
            return cachedSize + _iterators.get(index).size();
        }

        return 0;
    }

    @Override
    public int getChunksAvailable()
    {
        final String msg = "Method is not supported. Call getChunksAvailableForSink() instead.";
        Logger.error(this, msg);
        throw new UnsupportedOperationException(msg);
    }

    @Override
    public boolean isActive()
    {
        return isSourceActive();
    }

    @Override
    public AudioChunkType getAudioChunkForSink(AudioSink<AudioChunkType> sink)
    {
        int index = _currentAudioSinks.indexOf(sink);
        if (index != -1)
        {
            Pair<Integer, Integer> cachedAudioInfo = _cachedAudioInfoForSink.get(sink);
            if(cachedAudioInfo != null && cachedAudioInfo.first < cachedAudioInfo.second)
            {
                int readIndex = cachedAudioInfo.first;
                AudioChunkType next = _cache.get(readIndex++);
                if (readIndex < cachedAudioInfo.second)
                {
                    _cachedAudioInfoForSink.put(sink, new Pair<Integer, Integer>(readIndex, cachedAudioInfo.second));
                }
                else
                {
                    _cachedAudioInfoForSink.remove(sink);
                }
                return next;
            }

            MultiIteratorQueue<AudioChunkType>.Iterator chunks = _iterators.get(index);
            if (chunks.size() > 0) {
                AudioChunkType next = chunks.next();
                if (next != null)
                {
                    return next;
                }
                else
                {
                    Logger.error(this, "getAudioChunkForSink(): Expected non-null chunk");
                }
            }
        }
        else
        {
            Logger.warn(this, "getAudioChunkForSink(): Can't find sink");
        }

        return null;
    }

    @Override
    public void getAllAudioChunksForSink(AudioSink<AudioChunkType> sink, List<AudioChunkType> chunksList)
    {
        int index = _currentAudioSinks.indexOf(sink);
        if (index != -1)
        {
            Pair<Integer, Integer> cachedAudioInfo = _cachedAudioInfoForSink.get(sink);
            if(cachedAudioInfo != null && cachedAudioInfo.first < cachedAudioInfo.second)
            {
                chunksList.addAll(_cache.subList(cachedAudioInfo.first, cachedAudioInfo.second));
                _cachedAudioInfoForSink.remove(sink);
            }

            MultiIteratorQueue<AudioChunkType>.Iterator chunks = _iterators.get(index);
            AudioChunkType next = chunks.next();
            while (next != null)
            {
                chunksList.add(next);
                next = chunks.next();
            }
        } else
        {
            Logger.warn(this, "getAudioChunkForSink(): Can't find sink");
        }
    }

    @Override
    protected void audioSinkConnected(AudioSink<AudioChunkType> sink)
    {
        Logger.debug(this, "Audio Sink connected!");
        _currentAudioSinks.add(sink);
        // "first" of the pair is what has been currently read in the cache for that sink.
        // "second" of the pair is the snapshot of the cache size at this point.
        _cachedAudioInfoForSink.put(sink, new Pair<Integer, Integer>(0, _cache.size()));
        _iterators.add(_buffers.createIterator());

        if (!isSourceActive())
            sink.sourceClosed(this);
    }

    @Override
    protected void audioSinkDisconnected(AudioSink<AudioChunkType> sink)
    {
        Logger.debug(this, "Audio Sink disconnected!");
        _cachedAudioInfoForSink.remove(sink);
        int index = _currentAudioSinks.indexOf(sink);
        if (index >= 0) {
            _currentAudioSinks.remove(index);
            MultiIteratorQueue<AudioChunkType>.Iterator removed = _iterators.remove(index);
            removed.release();
        } else {
            Logger.warn(this, "Couldn't find audio sink to remove");
        }
    }

    @Override
    public void chunksAvailable(AudioSource<AudioChunkType> fromSource, AudioSink<AudioChunkType> internalSink)
    {
        // Logger.debug(this,
        // "Chunks available! Notifying " +
        // _currentAudioSinks.size() +
        // " connected sinks.");

        int chunksAvail = fromSource.getChunksAvailableForSink(internalSink);
        if (chunksAvail > 0)
        {
            List<AudioChunkType> available = fromSource.getAllAudioChunksForSink(internalSink);
            if (_currentAudioSinks.size() < _expectedSinkCount)
            {
                _cache.addAll(available);
            }
            else
            {
                // Cache as all been dequeued in connected sink now,
                // wipe it so we release memory.
                if (_cachedAudioInfoForSink.isEmpty())
                {
                    _cache.clear();   
                }
            }

            for (AudioChunkType chunk : available)
            {
                _buffers.add(chunk);
            }

            for (AudioSink<AudioChunkType> sink : _currentAudioSinks)
            {
                sink.chunksAvailable(this);
            }
        }
    }

    @Override
    public void framesDropped(AudioSource<AudioChunkType> fromSource, AudioSink<AudioChunkType> internalSink)
    {
        for (AudioSink<AudioChunkType> sink : _currentAudioSinks)
        {
            sink.framesDropped(this);
        }
    }

    @Override
    public void sourceClosed(AudioSource<AudioChunkType> fromSource, AudioSink<AudioChunkType> internalSink)
    {
        Logger.debug(BufferingDuplicatorPipe.this, "Source closed.");

        // Iterate the list manually instead of using an Iterator
        // to avoid ConcurrentModificationException.
        for (AudioSink<AudioChunkType> sink : _currentAudioSinks)
        {
            sink.sourceClosed(this);
        }
    }
}
