package com.nuance.dragon.toolkit.audio.pipes;

import java.util.ArrayList;
import java.util.LinkedList;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;

/**
 * A basic audio pipe that consumes audio as fast as possible from its source,
 * and buffers it here.
 */
public class BufferingPipe<AudioChunkType extends AbstractAudioChunk> extends SingleSinkPipe<AudioChunkType, AudioChunkType>
{
    private final LinkedList<AudioChunkType> _buffer;
    private final ArrayList<AudioChunkType> _tempBuffer; // Re-usable temporary buffer
    private final int _maxDuration;
    private AudioType _audioType;
    private int _duration;

    /**
     * BufferingPipe constructor
     */
    public BufferingPipe()
    {
        this(AudioType.UNKNOWN, -1);
    }

    /**
     * BufferingPipe constructor
     * @param expectedAudioType The expected audio type of the chunks for this pipe.
     */
    public BufferingPipe(AudioType expectedAudioType)
    {
        this(expectedAudioType, -1);
    }

    /**
     * BufferingPipe constructor
     * @param expectedAudioType The expected audio type of the chunks for this pipe.
     * @param maxDuration The duration (in ms.) after when non consumed audio is considered too old and cleared, or -1 to never clear the pipe. Default is -1.
     */
    public BufferingPipe(AudioType expectedAudioType, int maxDuration)
    {
        super(null);
        
        _buffer = new LinkedList<AudioChunkType>();
        _tempBuffer = new ArrayList<AudioChunkType>();
        _maxDuration = maxDuration;
        _audioType = expectedAudioType;
    }

    /**
     * Get the duration of the currently buffered audio
     * @return The buffered duration, in ms
     */
    public int getBufferedDuration()
    {
        return _duration;
    }

    /**
     * Discard the audio chunks already buffered and set
     * the duration to 0, this is used to discard the unwanted
     * buffers at some point
     */
    public void clearBuffers()
    {
        _buffer.clear();
        _duration = 0;
    }

    @Override
    public void connectAudioSource(AudioSource<AudioChunkType> source)
    {
        _buffer.clear();

        // Keep track of the audio type, because we want it to remain valid
        // even after the source is disconnected (because valid audio may remain
        // buffered)
        Assert.assertTrue(_audioType == AudioType.UNKNOWN || _audioType == source.getAudioType());
        _audioType = source.getAudioType();
        super.connectAudioSource(source);
    }

    /**
     * Override to be notified when every chunk is buffered.
     * @param chunk Never null
     */
    protected void onChunkBuffered(AudioChunkType chunk)
    {
        // Nothing by default
    }

    @Override
    protected void chunksAvailable(AudioSource<AudioChunkType> fromSource, AudioSink<AudioChunkType> internalSink)
    {
        _tempBuffer.ensureCapacity(fromSource.getChunksAvailableForSink(internalSink));
        fromSource.getAllAudioChunksForSink(internalSink, _tempBuffer);
        for (AudioChunkType chunk : _tempBuffer)
        {
            _buffer.add(chunk);
            _duration += chunk.audioDuration;

            // Remove first chunk when the buffer goes above the limit duration.
            if (_maxDuration >= 0 && _duration > _maxDuration)
            {
                AudioChunkType first =_buffer.removeFirst();
                if (first != null)
                    _duration -= first.audioDuration;
            }

            onChunkBuffered(chunk);
        }
        _tempBuffer.clear();

        // Since onChunkBuffered() can be overridden, test buffered duration
        // because we have no idea what child classes did in that method.
        if (_duration > 0)
            notifyChunksAvailable();
    }

    @Override
    protected void framesDropped(AudioSource<AudioChunkType> fromSource, AudioSink<AudioChunkType> internalSink)
    {
        notifyFramesDropped();
    }

    @Override
    protected void sourceClosed(AudioSource<AudioChunkType> fromSource, AudioSink<AudioChunkType> internalSink)
    {
        notifySourceClosed();
    }

    @Override
    public AudioType getAudioType()
    {
        //return getAudioTypeFromSource();
        return _audioType; // Use the cached audio type because the source may have been disconnected
    }

    @Override
    public int getChunksAvailable()
    {
        return _buffer.size();
    }

    @Override
    public boolean isActive()
    {
        return isSourceActive();
    }

    @Override
    protected AudioChunkType getAudioChunk()
    {
        if (_buffer.isEmpty())
            return null;

        AudioChunkType chunk = _buffer.remove();
        _duration -= chunk.audioDuration;
        return chunk;
    }

    // For internal usage only (see PrecisionClearBufferingPipe)
    final AudioChunkType getLastAudioChunk()
    {
        if (_buffer.isEmpty())
            return null;

        AudioChunkType chunk = _buffer.removeLast();
        _duration -= chunk.audioDuration;
        return chunk;
    }

    final void addAudioChunk(AudioChunkType chunk, boolean last)
    {
        if (last)
            _buffer.addLast(chunk);
        else
            _buffer.addFirst(chunk);
        _duration += chunk.audioDuration;
    }
}
