package com.nuance.dragon.toolkit.audio.pipes;

import java.util.LinkedList;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.MultiChannelAudioChunk;

/**
 * A pipe that takes a multi-channel input, and outputs a single channel from
 * that input.
 */
public class ChannelFilterPipe extends SingleSinkPipe<MultiChannelAudioChunk, AudioChunk>
{
    private final int _channel;
    private final LinkedList<AudioChunk> _chunks;
    private final AudioType _audioType;
    private final boolean _useSilenceUponAudioUnavailable;

    /**
     * @param channel The channel to output (0-based channel index)
     * @param audioType The audio type
     */
    public ChannelFilterPipe(int channel, AudioType audioType)
    {
        this(channel, audioType, false);
    }

    /**
     * @param channel The channel to output (0-based channel index)
     * @param audioType The audio type
     * @param useSilenceUponAudioUnavailable Whether to use silence when audio is unavailable for the specified channel (only for PCM)
     */
    public ChannelFilterPipe(int channel, AudioType audioType, boolean useSilenceUponAudioUnavailable)
    {
        super(null);
        
        _channel = channel;
        _chunks = new LinkedList<AudioChunk>();
        _audioType = audioType;
        _useSilenceUponAudioUnavailable = useSilenceUponAudioUnavailable;
    }

    @Override
    protected AudioChunk getAudioChunk()
    {
        if (!_chunks.isEmpty())
            return _chunks.remove();
        return null;
    }

    @Override
    protected void chunksAvailable(AudioSource<MultiChannelAudioChunk> fromSource, AudioSink<MultiChannelAudioChunk> internalSink)
    {
        while (true)
        {
            MultiChannelAudioChunk chunk = fromSource.getAudioChunkForSink(internalSink);
            if (chunk == null)
                break;

            if (_channel < chunk.channelCount)
            {
                _chunks.add(chunk.getSingleChannelChunk(_channel));
                notifyChunksAvailable();
            }
            else
            {
                //Logger.warn(this, "Channel " + _channel + " not valid");
                if (_useSilenceUponAudioUnavailable)
                {
                    Assert.assertTrue(_audioType.encoding == AudioType.Encoding.PCM_16);

                    AudioChunk baseChunk = chunk.getSingleChannelChunk(0);
                    AudioChunk silenceChunk = new AudioChunk(baseChunk.audioType, new short[baseChunk.audioShorts.length], baseChunk.audioTimestamp, baseChunk.audioTtsMarkers);
                    _chunks.add(silenceChunk);
                    notifyChunksAvailable();
                }
            }
        }
    }

    @Override
    protected void framesDropped(AudioSource<MultiChannelAudioChunk> fromSource, AudioSink<MultiChannelAudioChunk> internalSink)
    {
        notifyFramesDropped();
    }

    @Override
    protected void sourceClosed(AudioSource<MultiChannelAudioChunk> fromSource, AudioSink<MultiChannelAudioChunk> internalSink)
    {
        notifySourceClosed();
    }

    @Override
    public AudioType getAudioType()
    {
        return _audioType;
    }

    @Override
    public boolean isActive()
    {
        return isSourceActive();
    }

    @Override
    public int getChunksAvailable()
    {
        // Fix for DMTADRSDK-419, was "return getChunksAvailableFromSource();"
        return _chunks.size();
    }
}
