package com.nuance.dragon.toolkit.audio.pipes;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.WorkerThread;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Base class for audio pipes that perform some kind of encoding/decoding/re-sampling.
 */
public abstract class ConverterPipe<InChunkType extends AbstractAudioChunk, OutChunkType extends AbstractAudioChunk> extends SingleSinkPipe<InChunkType, OutChunkType>
{
    private final LinkedList<InChunkType> _rawAudio;     // All audio that hasn't been consumed by the converter yet. Modified from both threads.
    private final LinkedList<InChunkType> _unconvertedAudio; // All audio that hasn't been moved to _convertedAudio yet
    private final LinkedList<OutChunkType> _convertedAudio;   // Audio that has been converted
    private boolean _initialized; // Touched from worker thread
    private boolean _isReleased; // Touched from main thread
    private WorkerThread _workerThread;
    private final NMTHandler _workerThreadHandler;
    private final NMTHandler _mainThreadHandler;
    private int _durationBeforeInit;
    private final Object _sync;
    private final ArrayList<InChunkType> _tempBuffer; // Re-usable temporary buffer

    /**
     * ConverterPipe constructor
     */
    public ConverterPipe()
    {
        this(0);
    }

    /**
     * ConverterPipe constructor
     * @param durationBeforeInit The duration of audio to buffer before calling
     * {@link #initialize} and {@link #convert}.
     */
    public ConverterPipe(int durationBeforeInit)
    {
        super(null);
        
        _rawAudio = new LinkedList<InChunkType>();
        _unconvertedAudio = new LinkedList<InChunkType>();
        _convertedAudio = new LinkedList<OutChunkType>();
        _workerThreadHandler = getWorkerThreadHandler();
        _mainThreadHandler = Factory.createNMTHandler();
        _durationBeforeInit = durationBeforeInit;
        _sync = new Object();
        _tempBuffer = new ArrayList<InChunkType>();
    }

    /**
     * Called on MAIN thread exactly once in the constructor to obtain the
     * worker thread handler.
     * @return The worker thread handler.
     */
    private NMTHandler getWorkerThreadHandler()
    {
        _workerThread = Factory.createWorkerThread("com.nuance.dragon.toolkit.audio.pipes.ConverterPipe");
        _workerThread.start();
        return _workerThread.getHandler();
    }

    /**
     * Release the resources used by the converter.
     */
    public final void release()
    {
        Checker.checkState(this, !_isReleased);

        _isReleased = true;
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (_initialized)
                {
                    _initialized = false;
                    cleanup();
                }
            }
        });

        if (_workerThread != null)
        {
            _workerThread.stop();
            _workerThread = null;
        }
    }

    @Override
    public int getChunksAvailable()
    {
        return _convertedAudio.size();
    }

    @Override
    public boolean isActive()
    {
        return isSourceActive() || _unconvertedAudio.size() > 0;
    }

    @Override
    protected OutChunkType getAudioChunk()
    {
        if (_convertedAudio.isEmpty())
        {
            return null;
        }
        return _convertedAudio.remove(0);
    }

    @Override
    protected void getAllAudioChunks(List<OutChunkType> chunksList)
    {
        chunksList.addAll(_convertedAudio);
        _convertedAudio.clear();
    }

    /**
     * Runs on worker thread. Removes all available audio chunks from _rawAudio
     * and converts them. Then switches to main thread, and moves converted chunks from
     * _unconvertedAudio to _convertedAudio.
     */
    private void convertNext()
    {
        final ArrayList<InChunkType> pending;
        synchronized (_sync)
        {
            pending = new ArrayList<InChunkType>(_rawAudio.size());
            pending.addAll(_rawAudio);
            _rawAudio.clear();
        }

        if (!pending.isEmpty())
        {
            if (!_initialized)
            {
                initialize(pending.get(0).audioType);
                _initialized = true;
            }
            final ArrayList<OutChunkType> converted = new ArrayList<OutChunkType>(pending.size());

            List<OutChunkType> convertedChunks = convert(pending);
            converted.addAll(convertedChunks);

            _mainThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    // This on longer holds since a PCM chunk can now be decoded
                    // into 2 chunks based on end pointer results.
                    //Assert.assertTrue(converted.size() <= _unconvertedAudio.size());

                    _convertedAudio.addAll(converted);
                    if (!_unconvertedAudio.isEmpty())
                    {
                        for (InChunkType chunk : pending)
                        {
                            InChunkType removed = _unconvertedAudio.remove(0);
                            Assert.assertEquals(chunk, removed);
                        }
                    }

                    if(_convertedAudio.size() > 0)
                        notifyChunksAvailable();

                    AudioSource<InChunkType> connectedSource = getConnectedSource();
                    if (_unconvertedAudio.isEmpty() && (connectedSource == null || !connectedSource.isActive()))
                    {
                        notifySourceClosed();
                    }
                }
            });
        }
    }

    /**
     * Called on worker thread to initialize the converter (triggered by constructor).
     * @param audioType The type of the source audio.
     */
    protected abstract void initialize(AudioType audioType);

    /**
     * Called on worker thread to cleanup the converter (triggered by {@link #release}.
     * TODO Is this safe to call more than once?
     */
    protected abstract void cleanup();

    /**
     * Called from worker thread to convert a raw audio chunk.
     * TODO: allow this method to consume any number of audio chunks, and return
     * any number of audio chunks?
     * @param raw The raw audio
     * @return The converted audio
     */
    protected List<OutChunkType> convert(List<InChunkType> raw)
    {
        ArrayList<OutChunkType> converted = new ArrayList<OutChunkType>(raw.size());
        for (InChunkType chunk : raw)
        {
            OutChunkType[] chunks = convert(chunk);
            if (chunks != null)
            {
                for (OutChunkType c : chunks)
                {
                    converted.add(c);
                }
            }
        }

        return converted;
    }

    /**
     * Called from worker thread to convert a raw audio chunk.
     * TODO: allow this method to consume any number of audio chunks, and return
     * any number of audio chunks?
     * @param raw The raw audio
     * @return The converted audio
     */
    protected abstract OutChunkType[] convert(InChunkType raw);

    @Override
    public void chunksAvailable(AudioSource<InChunkType> source, AudioSink<InChunkType> internalSink)
    {
        int chunksAvail = source.getChunksAvailableForSink(internalSink);
        if (chunksAvail > 0)
        {
            _tempBuffer.ensureCapacity(chunksAvail);
            source.getAllAudioChunksForSink(internalSink, _tempBuffer);

            synchronized (_sync)
            {
                for (InChunkType chunk : _tempBuffer)
                {
                    _rawAudio.add(chunk);
                    _unconvertedAudio.add(chunk);

                    if (_durationBeforeInit > 0)
                    {
                        _durationBeforeInit -= chunk.audioDuration;
                    }
                }
            }
            _tempBuffer.clear();

            if (_durationBeforeInit <= 0)
            {
                _workerThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        convertNext();
                    }
                });
            }
        }
    }

    @Override
    public void framesDropped(AudioSource<InChunkType> source, AudioSink<InChunkType> internalSink)
    {
        notifyFramesDropped();
    }

    @Override
    public void sourceClosed(AudioSource<InChunkType> source, AudioSink<InChunkType> internalSink)
    {
        if (_unconvertedAudio.isEmpty())
        {
            notifySourceClosed();
        }
    }
}
