package com.nuance.dragon.toolkit.audio.pipes;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk;
import com.nuance.dragon.toolkit.audio.AudioPipe;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.util.MultiIteratorQueue;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;

/**
 * An implementation of audio source that reads audio from a single audio source,
 * and allows multiple audio sinks to read the audio independently of each other.
 */
public class DuplicatorPipe<AudioChunkType extends AbstractAudioChunk> extends AudioPipe<AudioChunkType, AudioChunkType>
{
    private final List<AudioSink<AudioChunkType>> _currentAudioSinks;
    private final MultiIteratorQueue<AudioChunkType> _buffers;
    private final List<MultiIteratorQueue<AudioChunkType>.Iterator> _iterators;

    /**
     * DuplicatorPipe constructor
     */
    public DuplicatorPipe()
    {
        this(null);
    }
    
    /**
     * DuplicatorPipe constructor
     */
    public DuplicatorPipe(NMTHandler mainThreadHandler)
    {
        super(mainThreadHandler);
        _currentAudioSinks = new CopyOnWriteArrayList<AudioSink<AudioChunkType>>();
        _buffers = new MultiIteratorQueue<AudioChunkType>();
        _iterators = new ArrayList<MultiIteratorQueue<AudioChunkType>.Iterator>();
    }

    @Override
    public AudioType getAudioType()
    {
        return getAudioTypeFromSource();
    }

    @Override
    public int getChunksAvailableForSink(AudioSink<AudioChunkType> sink)
    {
        int index = _currentAudioSinks.indexOf(sink);
        if (index != -1)
        {
            return _iterators.get(index).size();
        }

        return 0;
    }

    @Override
    public int getChunksAvailable()
    {
        final String msg = "Method is not supported. Call getChunksAvailableForSink() instead.";
        Logger.error(this, msg);
        throw new UnsupportedOperationException(msg);
    }

    @Override
    public boolean isActive()
    {
        return isSourceActive();
    }

    @Override
    public AudioChunkType getAudioChunkForSink(AudioSink<AudioChunkType> sink)
    {
        int index = _currentAudioSinks.indexOf(sink);
        if (index != -1)
        {
            MultiIteratorQueue<AudioChunkType>.Iterator chunks = _iterators.get(index);
            if (chunks.size() > 0) {
                AudioChunkType next = chunks.next();
                if (next != null)
                {
                    return next;
                } else
                {
                    Logger.error(this, "getAudioChunkForSink(): Expected non-null chunk");
                }
            }
        } else
        {
            Logger.warn(this, "getAudioChunkForSink(): Can't find sink");
        }

        return null;
    }

    @Override
    public void getAllAudioChunksForSink(AudioSink<AudioChunkType> sink, List<AudioChunkType> chunksList)
    {
        int index = _currentAudioSinks.indexOf(sink);
        if (index != -1)
        {
            MultiIteratorQueue<AudioChunkType>.Iterator chunks = _iterators.get(index);
            AudioChunkType next = chunks.next();
            while (next != null)
            {
                chunksList.add(next);
                next = chunks.next();
            }
        } else
        {
            Logger.warn(this, "getAudioChunkForSink(): Can't find sink");
        }
    }

    @Override
    protected void audioSinkConnected(AudioSink<AudioChunkType> sink)
    {
        Logger.debug(this, "Audio Sink connected!");
        _currentAudioSinks.add(sink);
        _iterators.add(_buffers.createIterator());

        if (!isSourceActive())
            sink.sourceClosed(this);
    }

    @Override
    protected void audioSinkDisconnected(AudioSink<AudioChunkType> sink)
    {
        Logger.debug(this, "Audio Sink disconnected!");
        int index = _currentAudioSinks.indexOf(sink);
        if (index >= 0) {
            _currentAudioSinks.remove(index);
            MultiIteratorQueue<AudioChunkType>.Iterator removed = _iterators.remove(index);
            removed.release();
        } else {
            Logger.warn(this, "Couldn't find audio sink to remove");
        }
    }

    @Override
    public void chunksAvailable(AudioSource<AudioChunkType> fromSource, AudioSink<AudioChunkType> internalSink)
    {
        // Logger.debug(this,
        // "Chunks available! Notifying " +
        // _currentAudioSinks.size() +
        // " connected sinks.");

        int chunksAvail = fromSource.getChunksAvailableForSink(internalSink);
        if (chunksAvail > 0)
        {
            List<AudioChunkType> available = fromSource.getAllAudioChunksForSink(internalSink);
            for (AudioChunkType chunk : available)
            {
                _buffers.add(chunk);
            }

            for (AudioSink<AudioChunkType> sink : _currentAudioSinks)
            {
                sink.chunksAvailable(this);
            }
        }
    }

    @Override
    public void framesDropped(AudioSource<AudioChunkType> fromSource, AudioSink<AudioChunkType> internalSink)
    {
        for (AudioSink<AudioChunkType> sink : _currentAudioSinks)
        {
            sink.framesDropped(this);
        }
    }

    @Override
    public void sourceClosed(AudioSource<AudioChunkType> fromSource, AudioSink<AudioChunkType> internalSink)
    {
        Logger.debug(DuplicatorPipe.this, "Source closed.");

        // Iterate the list manually instead of using an Iterator
        // to avoid ConcurrentModificationException.
        for (AudioSink<AudioChunkType> sink : _currentAudioSinks)
        {
            sink.sourceClosed(this);
        }
    }
}
