package com.nuance.dragon.toolkit.audio.pipes;

import java.util.Hashtable;

import android.content.Context;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.sse.Sse;

/**
 * A pipe that performs noise-suppression using SSE.
 * <p>
 * The input to this pipe is a source of audio chunks (single channel).
 * <p>
 * The output audio will be the audio with noise suppression applied.
 * <p>
 */
public class NoiseSuppressionPipe extends ConverterPipe<AudioChunk, AudioChunk>
{
    private final Sse _sse;
    private boolean _initialized;
    private FileManager _configDataFileManager;
    private Hashtable<AudioType, String> _configDataFileNames;
    @SuppressWarnings({ "deprecation", "unused" })
    private SseTuningType _tuningType;
    private final boolean _internalSse;
    private final SseInitParam _initParam;

    /**
     * Create an Noise Suppression pipe.
     * @param context The Android context object.
     */
    public NoiseSuppressionPipe(Context context)
    {
        this(context,  new SseInitParam.Builder().build(), null, null);
    }

    /**
     * Create an Noise Suppression pipe.
     * @param context The Android context object.
     * @param configDataFileManager The file manager where SSE should look for the configData file.
     * @param configDataFileNames The configData file names for different AudioTypes.
     */
    public NoiseSuppressionPipe(Context context, FileManager configDataFileManager, Hashtable<AudioType, String> configDataFileNames)
    {
        this(context,  new SseInitParam.Builder().build(), configDataFileManager, configDataFileNames);
    }

    /**
     * @deprecated
     * Create an Noise Suppression pipe.
     * @param context The Android context object.
     * @param configDataFileManager The file manager where SSE should look for the configData file.
     * @param configDataFileNames The configData file names for different AudioTypes.
     */
    @Deprecated
    public NoiseSuppressionPipe(Context context, FileManager configDataFileManager, Hashtable<AudioType, String> configDataFileNames, SseTuningType tuningType)
    {
        super(0);
        Checker.checkArgForNull("context", context.getApplicationContext());
        _sse = new Sse(context);
        _configDataFileManager = configDataFileManager;
        _configDataFileNames = configDataFileNames;
        _tuningType = tuningType;
        _internalSse = true;
        _initParam = null;
    }

    /**
     * Create an Noise Suppression pipe.
     * @param context The Android context object.
     * @param initParam The parameters have to be set prior to SSE initialization
     * @param configDataFileManager The file manager where SSE should look for the configData file.
     * @param configDataFileNames The configData file names for different AudioTypes. It will override any parameters
     *                             set in initParam if applicable
     */
    public NoiseSuppressionPipe(Context context, SseInitParam initParam, FileManager configDataFileManager, Hashtable<AudioType, String> configDataFileNames)
    {
        super(0);
        Checker.checkArgForNull("context", context.getApplicationContext());
        _sse = new Sse(context);
        _configDataFileManager = configDataFileManager;
        _configDataFileNames = configDataFileNames;
        _internalSse = true;
        _initParam = initParam;
    }

    /**
     * Create an Noise Suppression pipe.
     * @param sseProxy An existing SSE proxy instance. The proxy must be have been initialized already.
     */
    public NoiseSuppressionPipe(SseProxy sseProxy)
    {
        super(0);
        Checker.checkArgForNull("sseProxy", sseProxy);
        _sse = sseProxy.getSse();
        _configDataFileManager = null;
        _configDataFileNames = null;
        _tuningType = null;
        _internalSse = false;
        _initParam = null;
    }

    /**
     * Get the required frame size (in ms) for the SSE engine. All audio
     * chunks input to this pipe from its connected source should have a
     * duration that is an exact multiple of this value.
     * @return The required frame size in ms
     */
    public static int getRequiredFrameSize()
    {
        return Sse.getRequiredFrameSize();
    }

    @Override
    protected void initialize(AudioType audioType)
    {
        if (!_initialized)
        {
            if (_internalSse)
            {
                boolean success = _sse.initialize(audioType, (_initParam != null ? _initParam : new SseInitParam.Builder().build()),
                                                _configDataFileManager, _configDataFileNames != null ? _configDataFileNames.get(audioType) : null);
                if (!success)
                {
                    Logger.error(this, "Failure initializing SSE");
                }
            }
            _initialized = true;
        }
    }

    @Override
    protected void cleanup()
    {
        if (_initialized)
        {
            if (_internalSse)
            {
                _sse.release();
            }
        }
        _initialized = false;
    }

    @Override
    protected AudioChunk[] convert(AudioChunk raw)
    {
        AudioChunk chunk = _sse.process(raw);
        if (chunk != null)
        {
            return new AudioChunk[] { chunk };
        }

        return new AudioChunk[] { raw };
    }

    @Override
    public AudioType getAudioType()
    {
        return getAudioTypeFromSource();
    }
}
