package com.nuance.dragon.toolkit.audio.pipes;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Implementation of audio pipe that reads from a single audio source, normalizes
 * the PCM audio, and provides the amplified audio data to a single
 * audio sink.
 */
public class NormalizerPipe extends ConverterPipe<AudioChunk, AudioChunk>
{
    // ~1/4th of the maximum PCM16 amplitude
    private static int DEFAULT_THRESHOLD = 8000;

    private static short MIN_AMPLITUDE = -32768;
    private static short MAX_AMPLITUDE = 32767;

    private final short _threshold;
    private boolean _active;

    /**
     * NormalizerPipe constructor
     *
     */
    public NormalizerPipe()
    {
        this(DEFAULT_THRESHOLD);
    }

    /**
     * NormalizerPipe constructor
     * @param threshold The threshold value. Only a signal that
     * has a peak value above this threshold will be normalized.
     * This is to avoid to amplify noise on chunks that
     * don't carry speech. Value must be comprised between
     * -32768 and 32767.
     *
     */
    public NormalizerPipe(int threshold)
    {
        Checker.checkArgForCondition("threshold", "between 32768 and 32767", threshold >= MIN_AMPLITUDE && threshold <= MAX_AMPLITUDE);
        _threshold = (short) Math.abs(threshold);
        _active = true;
    }

    /**
     * Decide whether to normalize or not the input signal. This might be useful
     * when normalization must happen in only certain conditions or for certain
     * sources (ex: wired headset).
     *@param active True to normalize audio. False to skip processing.
     */
    public void setActive(boolean active)
    {
        _active = active;
    }

    @Override
    public AudioType getAudioType()
    {
        return getAudioTypeFromSource();
    }

    @Override
    protected boolean isAudioSourceTypeSupported(AudioType type)
    {
        return type.encoding == AudioType.Encoding.PCM_16;
    }

    @Override
    protected AudioChunk[] convert(AudioChunk raw)
    {
        AudioChunk chunk = null;

        if (raw != null)
        {
            // Make a copy of the chunk to mess with
            short[] data = new short[raw.audioShorts.length];
            System.arraycopy(raw.audioShorts, 0, data, 0, data.length);
            chunk = new AudioChunk(raw.audioType, data, raw.audioTimestamp, raw.audioTtsMarkers);

            if (_active)
            {
                chunk = normalize(chunk, _threshold);
            }
        }

        return createAudioChunks(chunk);
    }

    @Override
    protected void initialize(AudioType audioType)
    {
    }

    @Override
    protected void cleanup()
    {
    }

    private static AudioChunk normalize(AudioChunk chunk, short threshold)
    {
        short frame = 0, peak = 0;
        short[] frames = chunk.audioShorts;

        // 1st pass to determine the peak of the chunk
        for (int idx = 0; idx < frames.length; idx++)
        {
            frame = (short) Math.abs(frames[idx]);
            if (frame >= peak)
                peak = frame;
        }

        // Amplify if necessary
        if (peak > 0 && peak >= threshold)
        {
            // 2nd pass to apply the amplification factor i.e. gain
            float gain = MAX_AMPLITUDE / peak;
            for (int idx = 0; idx < frames.length; idx++)
            {
                frame = (short) (gain * frames[idx]);

                // Make sure we don't clip
                if (frame >= MIN_AMPLITUDE && frame <= MAX_AMPLITUDE)
                    frames[idx] = frame;
                else if (frame > MAX_AMPLITUDE)
                    frames[idx] = MAX_AMPLITUDE;
                else
                    frames[idx] = MIN_AMPLITUDE;
            }
        }

        return new AudioChunk(chunk.audioType, frames, chunk.audioTimestamp, chunk.audioTtsMarkers);
    }

    private static AudioChunk[] createAudioChunks(AudioChunk chunk)
    {
        AudioChunk[] chunks = null;

        if (chunk != null)
        {
            chunks = new AudioChunk[1];
            chunks[0] = new AudioChunk(chunk.audioType, chunk.audioShorts, chunk.audioTimestamp, chunk.audioTtsMarkers);
        }
        return chunks;
    }
}
