package com.nuance.dragon.toolkit.audio.pipes;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.util.OpusEncoder;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;


/**
 * Implementation of audio pipe that reads from a single audio source, converts
 * the audio to Opus, and provides the Opus-encoded audio data to a single
 * audio sink.
 */
public class OpusEncoderPipe extends ConverterPipe<AudioChunk, AudioChunk>
{
    private final int _bitRate, _complexity;
    private final boolean _vbr;
    private final OpusEncoder _encoder;
    private final int _durationMs;
    
    // By request from research we have to force 20ms audio for opus
    private static final int DEFAULT_DURATION_MS = 20;
    // With 20 ms chunks we get 320 samples at 16 kHz
    private int _samplesPerDuration = 320;
    
    /**
     * OpusEncoderPipe constructor
     */
    public OpusEncoderPipe()
    {
        this(28000, 10, false, DEFAULT_DURATION_MS);
    }

    /**
     * OpusEncoderPipe constructor
     *
     * @param bitRate The bit rate of the encoded packets. Default is 28000.
     * @param complexity The complexity factor, between 0 and 10. Default is 10.
     * @param vbr Variable Bit Rate or Constant Bit Rate. Default is true (i.e. VBR).
     */
    public OpusEncoderPipe(int bitRate, int complexity, boolean vbr)
    {
        this(bitRate, complexity, vbr, DEFAULT_DURATION_MS);
    }

    /**
     * OpusEncoderPipe constructor
     *
     * @param bitRate The bit rate of the encoded packets. Default is 28000.
     * @param complexity The complexity factor, between 0 and 10. Default is 10.
     * @param vbr Variable Bit Rate or Constant Bit Rate. Default is true (i.e. VBR).
     * @param durationMs This is to change the minimum duration for a chunk. Default is 20 ms
     */
    public OpusEncoderPipe(int bitRate, int complexity, boolean vbr, int durationMs)
    {
        Checker.checkArgForCondition("bitRate", "stricly greater than 0", bitRate > 0);
        Checker.checkArgForCondition("complexity", "between 0 and 10", complexity >= 0 && complexity <= 10);

        _bitRate = bitRate;
        _complexity = complexity;
        _vbr = vbr;
        _durationMs = durationMs;

        _encoder = new OpusEncoder();
    }

    private AudioType getEncodedAudioType(AudioType audioType)
    {
        Assert.assertTrue(isAudioSourceTypeSupported(audioType));
        if (audioType.frequency == AudioType.Frequency.FREQ_8KHZ)
            return AudioType.OPUS_NB;
        else
            return AudioType.OPUS_WB;
    }

    @Override
    protected boolean isAudioSourceTypeSupported(AudioType type)
    {
        return type.encoding == AudioType.Encoding.PCM_16;
    }

    @Override
    protected void initialize(AudioType audioType)
    {
        _samplesPerDuration = audioType.getSampleCount(_durationMs);
        _encoder.initialize(getEncodedAudioType(audioType), _bitRate, _complexity, _vbr);
    }

    @Override
    protected AudioChunk[] convert(AudioChunk raw)
    {
        Logger.verbose(this, "[LATCHK] convert() start raw:" + raw);
        AudioChunk paddedChunk = raw;
        
        if(raw.audioShorts.length < _samplesPerDuration){
            short[] padBuffer = new short[_samplesPerDuration];
            System.arraycopy(raw.audioShorts, 0, padBuffer, 0, raw.audioShorts.length);
            paddedChunk = new AudioChunk(raw.audioType, padBuffer, 20);
        }
        
        AudioChunk[] chunks = _encoder.encode(paddedChunk);
        if (com.nuance.dragon.toolkit.Build.DEBUG)
        {
            StringBuilder log = new StringBuilder();
            for (AudioChunk ac : chunks)
            {
                log.append(ac);
                log.append("; ");
            }
            Logger.verbose(this, "[LATCHK] convert() end chunk:" + log);
        }
        return chunks;
    }

    @Override
    public AudioType getAudioType()
    {
        AudioSource<AudioChunk> connectedSource = getConnectedSource();
        if (connectedSource != null)
        {
            AudioType sourceType = connectedSource.getAudioType();
            if (sourceType != null)
            {
                return getEncodedAudioType(sourceType);
            }
        }
        return AudioType.UNKNOWN;
    }

    @Override
    protected void cleanup()
    {
        _encoder.release();
    }
}
