package com.nuance.dragon.toolkit.audio.pipes;

import java.util.Hashtable;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.MultiChannelAudioChunk;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;

import android.content.Context;

/**
 * A pipe that performs passenger interference cancellation, which removes
 * off-axis signal from passenger
 * <p>
 * The input to this pipe is a source of multi-channel audio chunks. Channel 0
 * is considered to be the desire audio input (i.e. from the passenger). An optional
 * channel 1 is considered to be the interfering audio that comes from other directions
 * <p>
 * The output audio will be the audio from channel 0, with PIC applied for
 * every chunk that has 2 channels.
 * <p>
 * Typically, {@link AlignedMergerPipe} would be used to align the input and
 * interference audio streams in order to generate the multi-channel chunks.
 */
public class PassengerInterCancelPipe extends SsePipe
{

    /**
     * Create an SSE pipe.
     * @param context The Android context object.
     */
    public PassengerInterCancelPipe(Context context)
    {
        super(context);
    }

    /**
     * Create an SSE pipe.
     * @param context The Android context object.
     * @param configDataFileManager The file manager where SSE should look for the configData file.
     * @param configDataFileNames The configData file names for different AudioTypes.
     */
    public PassengerInterCancelPipe(Context context, FileManager configDataFileManager, Hashtable<AudioType, String> configDataFileNames)
    {
        super(context, configDataFileManager, configDataFileNames);
    }

    /**
     * Create an SSE pipe.
     * @param context The Android context object.
     * @param initParam The parameters have to be set prior to SSE initialization
     * @param configDataFileManager The file manager where SSE should look for the configData file.
     * @param configDataFileNames The configData file names for different AudioTypes. It will override any parameters
     *                             set in initParam if applicable
     */
    public PassengerInterCancelPipe(Context context, SseInitParam initParam, FileManager configDataFileManager, Hashtable<AudioType, String> configDataFileNames)
    {
        super(context, initParam, configDataFileManager, configDataFileNames);
    }

    /**
     * Create an SSE pipe.
     * @param sseProxy An existing SSE proxy instance. The proxy must be have been initialized already.
     */
    public PassengerInterCancelPipe(SseProxy sseProxy)
    {
        super(sseProxy);
    }

    @Override
    protected void initialize(AudioType audioType)
    {
        if (!_initialized)
        {
            if (_internalSse)
            {
                // force AEC disabled for conflict between config data and sseSetData
                // TODO: confirm with SSE team
                SseInitParam rInitParam;
                if (_initParam != null)
                {
                    rInitParam = new SseInitParam.Builder().setMicrophoneNumber(2).setNrOnOff(_initParam.isNrOn()).
                            setBeamFormingOnOff(_initParam.isBeamFormingOn()).setVoconASR(_initParam.isForVoconASR()).setEchoCancelOnOff(false).setPicOnOff(true).build();
                }
                else
                {
                    rInitParam = new SseInitParam.Builder().setMicrophoneNumber(2).setEchoCancelOnOff(false).setPicOnOff(true).build();
                }
                boolean success = _sse.initialize(audioType, rInitParam, _configDataFileManager, _configDataFileNames != null ? _configDataFileNames.get(audioType) : null);
                if (success && rInitParam.isForVoconASR())
                {
                    _voconConfiguration = _sse.getVoconConfiguration();
                }
                else if (!success)
                {
                    Logger.error(this, "Failure initializing SSE");
                }
            }
            _initialized = true;
        }
    }

    @Override
    protected AudioChunk[] convert(MultiChannelAudioChunk raw)
    {
        AudioChunk chunk = _sse.processPic(raw);
        if (chunk != null)
        {
            return new AudioChunk[] { chunk };
        }

        return new AudioChunk[] { raw.getSingleChannelChunk(0) };
    }

}
