package com.nuance.dragon.toolkit.audio.pipes;

import java.util.Timer;
import java.util.TimerTask;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.oem.api.Logger;

/**
 * A pipe that buffers audio but reports it to the rest of the system at a certain rate.
 * Not particularly useful with live audio but handy for testing purposes (using canned audio).
 */
public class PeriodicUpdaterPipe<AudioChunkType extends AbstractAudioChunk> extends BufferingPipe<AudioChunkType>
{
    private int _updatePeriod = 1;
    private Timer _timer = null;
    private boolean _pendingClose = false;
    private boolean _paused = false;

    private class Sync {
        public int chunksAdded = 0;
        public int chunksRead  = 0;
        public int timerExpired = 0;
    }
    private Sync _sync = new Sync();

    private TimerTask _periodicUpdate = new TimerTask()
    {
        @Override
        public void run() {
            boolean notifySink = false;
            boolean notifySourceIsClosed = false;
            synchronized(_sync) {
                _sync.timerExpired++;
                // if we have more than one chunk available (likely if file is Burst), notify sink
                if ( (_sync.chunksAdded - _sync.chunksRead) > 1) {
                    notifySink = true;
                }
                else if (_pendingClose && ((_sync.chunksAdded - _sync.chunksRead) == 1)) {
                    notifySourceIsClosed = true;
                }
            }
            if (notifySink) {
                if (_paused == false) {
                    notifyChunksAvailable();
                }
            }
            else if (notifySourceIsClosed) {
                _timer.cancel();
                notifySourceClosed();
            }
        }
    };

    public PeriodicUpdaterPipe(int period)
    {
        _updatePeriod = period;
        _timer = new Timer();
        _timer.scheduleAtFixedRate(_periodicUpdate, _updatePeriod, _updatePeriod);
    }

    public PeriodicUpdaterPipe()
    {
        this(400);
    }

    public void pauseUpdater()
    {
        _paused = true;
    }

    public void resumeUpdater()
    {
        _paused = false;
    }

    @Override
    protected void chunksAvailable(AudioSource<AudioChunkType> fromSource, AudioSink<AudioChunkType> internalSink)
    {
        int chunksToAdd = fromSource.getChunksAvailableForSink(internalSink);
        int chunksAdded, chunksRead;
        synchronized (_sync) {
            _sync.chunksAdded += chunksToAdd;
            chunksAdded = _sync.chunksAdded;
            chunksRead = _sync.chunksRead;
        }
        Logger.debug(this, "chunksAvailable(): chunksAdded/chunksRead: "+ chunksAdded + " / " + chunksRead);
        if (_paused == false) {
            super.chunksAvailable(fromSource, internalSink);
        }
    }

    @Override
    public int getChunksAvailable()
    {
        int chunks = 0;
        boolean available = false;

        if (_paused == false) {
            synchronized (_sync) {
                if (_sync.timerExpired > 0) {
                    available = true;
                }
            }
        }

        if (available) {
            chunks = super.getChunksAvailable();
        }

        return chunks;
    }

    @Override
    protected AudioChunkType getAudioChunk()
    {
        int added = 0;
        int read  = 0;
        AudioChunkType chunk = null;

        if (_paused == false) {
            synchronized (_sync) {
                added = _sync.chunksAdded;
                read = _sync.chunksRead;
                if (_sync.timerExpired > 0) {
                    _sync.timerExpired--;
                    chunk = super.getAudioChunk();
                    if (chunk != null) {
                        _sync.chunksRead++;
                        read++;
                    }
                }
            }
        }

        if (chunk != null) {
            Logger.debug(this, "getAudioChunk(): AVAILABLE chunksAdded/chunksRead: " + added +" / " + read);
        }
        else {
            Logger.debug(this, "getAudioChunk(): NO CHUNK AVAILABLE");
        }

        return chunk;
    }

    @Override
    protected void sourceClosed(AudioSource<AudioChunkType> fromSource, AudioSink<AudioChunkType> internalSink)
    {
        boolean allDone = false;

        synchronized (_sync) {
            if (_sync.chunksAdded == _sync.chunksRead) {
                allDone = true;
            }
            else {
                _pendingClose = true;
            }
        }

        if (allDone) {
            _timer.cancel();
            super.sourceClosed(fromSource, internalSink);
        }
    }

}
