package com.nuance.dragon.toolkit.audio.pipes;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * A special implementation of a {@link BufferingPipe}. It can discards
 * old accumulated audio up to a specified duration. This pipe might be
 * useful when implementing seamless wake-up into the application
 * in order to "strip" speech that comes before the wake-up utterance.
 */
public class PrecisionClearBufferingPipe extends BufferingPipe<AudioChunk>
{
    /**
     * PrecisionClearBufferingPipe constructor
     */
    public PrecisionClearBufferingPipe()
    {
        this(10000);
    }

    /**
     * PrecisionClearBufferingPipe constructor
     * @param maxDuration The duration (in ms.) after when non consumed audio is considered too old and cleared. Default is 10000 ms.
     */
    public PrecisionClearBufferingPipe(int maxDuration)
    {
        super(AudioType.UNKNOWN, maxDuration);
    }

    /**
     * Clear the audio buffered up before a certain timestamp.
     * @param timestamp The timestamp value (in ms.).
     * You may want to substract an extra padding duration for time data returned by the recognizers.
     */
    public void clearBuffersBeforeTimestamp(long timestamp)
    {
        Checker.checkArgForCondition("timestamp", "timestamp > 0", timestamp > 0);

        boolean keepRemoving = true;
        while (keepRemoving)
        {
            if (getChunksAvailable() == 0)
                return;

            AudioChunk chunk = getAudioChunk();
            if (chunk.audioEndTimestamp <= timestamp)
            {
                // Removed the whole first chunk of the list
                Logger.debug(this, "Removed a whole chunk out of the buffering pipe.");
            }
            else if (chunk.audioTimestamp < timestamp)
            {
                // Insert left-over audio in place of the first chunk of the list
                Logger.debug(this, "Removed a partial chunk out of the buffering pipe.");
                int leftOverDuration = (int)(chunk.audioDuration - (timestamp - chunk.audioTimestamp));
                if (leftOverDuration > 0) {
                    short leftOverBuffer[] = new short[chunk.audioType.getSampleCount(leftOverDuration)];
                    System.arraycopy(chunk.audioShorts, chunk.audioType.getSampleCount((int)(timestamp - chunk.audioTimestamp)), leftOverBuffer, 0, leftOverBuffer.length);

                    AudioChunk leftOverChunk = new AudioChunk(chunk.audioType, leftOverBuffer, chunk.audioEndTimestamp - leftOverDuration);
                    addAudioChunk(leftOverChunk, false);
                }
                keepRemoving = false;
            }
            else
            {
                // Re-insert whole chunk
                addAudioChunk(chunk, false);
                keepRemoving = false;
            }
        }
    }

    /**
     * Clear the audio buffered up after a certain timestamp.
     * @param timestamp The timestamp value (in ms.).
     * You may want to add an extra padding duration for time data returned by the recognizers.
     */
    public void clearBuffersAfterTimestamp(long timestamp)
    {
        Checker.checkArgForCondition("timestamp", "timestamp > 0", timestamp > 0);

        boolean keepRemoving = true;
        while (keepRemoving)
        {
            if (getChunksAvailable() == 0)
                return;

            AudioChunk chunk = getLastAudioChunk();
            if (chunk.audioTimestamp >= timestamp)
            {
                // Removed the whole last chunk of the list
                Logger.debug(this, "Removed a whole chunk out of the buffering pipe.");
            }
            else if (chunk.audioEndTimestamp > timestamp)
            {
                // Insert left-over audio in place of the last chunk of the list
                Logger.debug(this, "Removed a partial chunk out of the buffering pipe.");
                int leftOverDuration = (int)(chunk.audioDuration - (chunk.audioEndTimestamp - timestamp));
                if (leftOverDuration > 0) {
                    short leftOverBuffer[] = new short[chunk.audioType.getSampleCount(leftOverDuration)];
                    System.arraycopy(chunk.audioShorts, 0, leftOverBuffer, 0, leftOverBuffer.length);

                    AudioChunk leftOverChunk = new AudioChunk(chunk.audioType, leftOverBuffer, chunk.audioTimestamp);
                    addAudioChunk(leftOverChunk, true);
                }
                keepRemoving = false;
            }
            else
            {
                // Re-insert whole chunk
                addAudioChunk(chunk, true);
                keepRemoving = false;
            }
        }
    }

    /**
     * @deprecated Clear the audio buffered up before and after certain timestamps.
     * Use {@link #clearBuffersBeforeAndAfterTimestamps(long,long)} instead.
     * @param startTimestamp Indicates the time stamp (in ms.) after which audio needs to be kept.
     * You may want to subtract an extra padding duration for time data returned by the recognizers
     * @param endTimestamp Indicates the time stamp (in ms.) before which audio needs to be kept.
     * You may want to add an extra padding duration for time data returned by the recognizers
     */
    @Deprecated
    public void clearBuffers(long startTimestamp, long endTimestamp)
    {
        Checker.checkArgForCondition("startTimestamp", "startTimestamp > 0", startTimestamp > 0);
        Checker.checkArgForCondition("endTimestamp", "endTimestamp > 0", endTimestamp > 0);
        Checker.checkArgForCondition("startTimestamp, endTimestamp", "startTimestamp < endTimestamp", startTimestamp < endTimestamp);

        clearBuffersBeforeTimestamp(startTimestamp);
        clearBuffersAfterTimestamp(endTimestamp);
    }
 
    /**
     * Clear the audio buffered up before and after certain timestamps.
     * @param beforeTimestamp Indicates the time stamp (in ms.) after which audio needs to be kept.
     * You may want to subtract an extra padding duration for time data returned by the recognizers
     * @param afterTimestamp Indicates the time stamp (in ms.) before which audio needs to be kept.
     * You may want to add an extra padding duration for time data returned by the recognizers
     */
    public void clearBuffersBeforeAndAfterTimestamps(long beforeTimestamp, long afterTimestamp)
    {
        Checker.checkArgForCondition("beforeTimestamp", "beforeTimestamp > 0", beforeTimestamp > 0);
        Checker.checkArgForCondition("afterTimestamp", "afterTimestamp > 0", afterTimestamp > 0);
        Checker.checkArgForCondition("beforeTimestamp, afterTimestamp", "beforeTimestamp < afterTimestamp", beforeTimestamp < afterTimestamp);

        clearBuffersBeforeTimestamp(beforeTimestamp);
        clearBuffersAfterTimestamp(afterTimestamp);
    }
}
