package com.nuance.dragon.toolkit.audio.pipes;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.util.SpeexResampler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Implementation of audio pipe that reads from a single audio source, re-sample
 * the PCM audio, and provides the re-sampled audio data to a single
 * audio sink.
 */
public class ResamplerPipe extends ConverterPipe<AudioChunk, AudioChunk>
{
    private final AudioType _resampledAudioType;
    private final int _quality;
    private final SpeexResampler _resampler;

    /**
     * ResamplerPipe constructor
     *
     * @param resampledAudioType The desired, re-sampled audio type.
     */
    public ResamplerPipe(AudioType resampledAudioType)
    {
        this(resampledAudioType, 3);
    }

    /**
     * ResamplerPipe constructor
     *
     * @param resampledAudioType The desired, re-sampled audio type.
     * @param quality The quality factor, between 0 and 10. Default is 3 as recommended per SPEEX doc.
     */
    public ResamplerPipe(AudioType resampledAudioType, int quality)
    {
        Checker.checkArgForNull("resampledAudioType", resampledAudioType);
        Checker.checkArgForCondition("quality", "between 0 and 10", quality >= 0 && quality <= 10);

        _resampledAudioType = resampledAudioType;
        _quality = quality;
        _resampler = new SpeexResampler();
    }

    @Override
    protected boolean isAudioSourceTypeSupported(AudioType type)
    {
        return type.encoding == AudioType.Encoding.PCM_16;
    }

    @Override
    protected AudioChunk[] convert(AudioChunk raw)
    {
        return _resampler.resample(raw, 0);
    }

    @Override
    public AudioType getAudioType()
    {
        return _resampledAudioType;
    }

    @Override
    protected void initialize(AudioType audioType)
    {
        _resampler.initialize(audioType, _resampledAudioType, _quality);
    }

    @Override
    protected void cleanup()
    {
        _resampler.release();
    }
}
