package com.nuance.dragon.toolkit.audio.pipes;

import java.util.List;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk;
import com.nuance.dragon.toolkit.audio.AudioPipe;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;

/**
 * Base class for pipes that only allow one sink to be connected
 *
 * @param <InputType>
 * @param <OutputType>
 */
public abstract class SingleSinkPipe<InputType extends AbstractAudioChunk, OutputType extends AbstractAudioChunk> extends AudioPipe<InputType, OutputType>
{
    private AudioSink<OutputType> _connectedSink;
    private boolean _closed;
    private boolean _framesDropped;

    protected SingleSinkPipe(NMTHandler mainThreadHandler)
    {
        super(mainThreadHandler);
    }

    @Override
    public void connectAudioSource(AudioSource<InputType> source)
    {
        if (_closed && source.isActive())
            _closed = false;

        super.connectAudioSource(source);
    }

    /**
     * Gets the next audio chunk
     * @return The audio chunk, or null if there are none
     */
    protected abstract OutputType getAudioChunk();

    /**
     * Get all audio chunks. Calls {@link #getAudioChunk} by default, override
     * for more efficiency.
     * @param chunks List that the new chunks will be added to
     */
    protected void getAllAudioChunks(List<OutputType> chunks)
    {
        while (true)
        {
            OutputType chunk = getAudioChunk();
            if (chunk == null)
                break;
            chunks.add(chunk);
        }
    }

    /**
     * Override to do something when a sink is connected.
     * @param sink The sink that was connected
     */
    protected void onSinkConnected(AudioSink<OutputType> sink)
    {
    }

    /**
     * Override to do something when a sink is disconnected.
     * @param sink The sink that was disconnected
     */
    protected void onSinkDisconnected(AudioSink<OutputType> sink)
    {
    }

    /**
     * Notify the connected sink (if any) that chunks are available.
     */
    protected void notifyChunksAvailable()
    {
        if (_connectedSink != null)
            _connectedSink.chunksAvailable(this);
    }

    /**
     * Notify the connected sink (if any) that frames were dropped.
     */
    protected void notifyFramesDropped()
    {
        if (_connectedSink != null)
            _connectedSink.framesDropped(this);
        else
            _framesDropped = true;
    }

    /**
     * Notify the connected sink (if any) that the source has closed.
     */
    protected void notifySourceClosed()
    {
        if (!_closed)
        {
            if (_connectedSink != null)
                _connectedSink.sourceClosed(this);
            else
                _closed = true;
        }
    }

    @Override
    public final int getChunksAvailableForSink(AudioSink<OutputType> sink)
    {
        if (sink == _connectedSink)
            return getChunksAvailable();

        return 0;
    }

    @Override
    public final boolean isEmptyForSink(AudioSink<OutputType> sink)
    {
        if (sink == _connectedSink)
            return isEmpty();

        return true;
    }

    @Override
    public final OutputType getAudioChunkForSink(AudioSink<OutputType> sink)
    {
        if (sink == _connectedSink)
            return getAudioChunk();

        return null;
    }

    @Override
    public final void getAllAudioChunksForSink(AudioSink<OutputType> sink, List<OutputType> chunksList)
    {
        if (sink == _connectedSink)
            getAllAudioChunks(chunksList);
    }

    @Override
    protected final void audioSinkConnected(final AudioSink<OutputType> sink)
    {
        _connectedSink = sink;

        onSinkConnected(sink);

        // Posting the job is necessary to avoid a race condition where the recognizers
        // are connected to a pipe with already chunks in it and start pushing the audio
        // to non set-up back-end engines.
        _mainThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (sink != _connectedSink)
                    return;

                if (_framesDropped)
                {
                    _framesDropped = false;
                    sink.framesDropped(SingleSinkPipe.this);
                }

                if (!isEmpty())
                {
                    sink.chunksAvailable(SingleSinkPipe.this);
                }

                if (_closed)
                {
                    sink.sourceClosed(SingleSinkPipe.this);
                }
            }
        });
    }

    @Override
    protected final void audioSinkDisconnected(AudioSink<OutputType> sink)
    {
        if (sink == _connectedSink)
        {
            _connectedSink = null;
            onSinkDisconnected(sink);
        } else
        {
            Logger.warn(this, "Wrong sink disconnected");
        }
    }
}
