package com.nuance.dragon.toolkit.audio.pipes;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.util.SpeexEncoder;
import com.nuance.dragon.toolkit.oem.api.Logger;


/**
 * Implementation of audio pipe that reads from a single audio source, converts
 * the audio to Speex, and provides the Speex-encoded audio data to a single
 * audio sink.
 */
public class SpeexEncoderPipe extends ConverterPipe<AudioChunk, AudioChunk>
{
    public static final int DEFAULT_STOP_ON_END_OF_SPEECH = 1;
    public static final int DEFAULT_VAD_LONG_UTTERANCE    = 0;
    public static final int DEFAULT_VAD_HISTORY_LEN       = 50;
    public static final int DEFAULT_VAD_BEGIN_LEN         = 15;
    public static final int DEFAULT_VAD_BEGIN_THRESHOLD   = 7;
    public static final int DEFAULT_VAD_BEGIN_DELAY       = 0;
    public static final int DEFAULT_VAD_END_LEN           = 50;
    public static final int DEFAULT_VAD_END_THRESHOLD     = 5;
    public static final int DEFAULT_VAD_INTERSPEECH_LEN   = 35;

    // With 20 ms chunks we get 320 samples at 16 kHz
    private int _samplesPerFrame;

    private final SpeexEncoder _encoder;
    private final int _endpointerStopOnEndOfSpeech;
    private final int _endpointerVadLongUtterance;
    private final int _endpointerVadHistoryLength;
    private final int _endpointerVadBeginLength;
    private final int _endpointerVadBeginThreshold;
    private final int _endpointerVadBeginDelay;
    private final int _endpointerVadEndLength;
    private final int _endpointerVadEndThreshold;
    private final int _endpointerVadInterSpeechLength;

    /**
     * SpeexDecoderPipe constructor
     */
    public SpeexEncoderPipe()
    {
        this(DEFAULT_STOP_ON_END_OF_SPEECH,
             DEFAULT_VAD_LONG_UTTERANCE,
             DEFAULT_VAD_HISTORY_LEN,
             DEFAULT_VAD_BEGIN_LEN,
             DEFAULT_VAD_BEGIN_THRESHOLD,
             DEFAULT_VAD_BEGIN_DELAY,
             DEFAULT_VAD_END_LEN,
             DEFAULT_VAD_END_THRESHOLD,
             DEFAULT_VAD_INTERSPEECH_LEN);
    }

    /**
     * SpeexDecoderPipe constructor
     *
     * Creates the SpeexDecoderPipe object and allows setting speech detection parameters.
     *
     * @param endpointerStopOnEndOfSpeech    If set to 1, speech detection algorithm will
     *                                       stop running after detecting first end-of-speech.
     *                                       Set to 0 otherwise.
     *
     * @param endpointerVadLongUtterance     If set to 1, this value indicates that the utterance
     *                                       is a long one, and that the endpointerVadInterSpeechLength
     *                                       parameter should be considered. Set to 0 otherwise.
     *
     * @param endpointerVadHistoryLength     History length in number of Speex frames over which the
     *                                       speech detection algorithm runs.
     *
     * @param endpointerVadBeginLength       Window length in number of Speex frames over which the
     *                                       speech-start detection statistic is evaluated
     *
     * @param endpointerVadBeginThreshold    Threshold used to detect start-of-speech. When speech-start
     *                                       detection statistic exceeds this number start-of-speech is
     *                                       detected.
     *
     * @param endpointerVadBeginDelay        Initial delay in Speex frames before detection algorithm
     *                                       starts running. This is useful if first few frames may
     *                                       contain noise or artifacts introduced by recorder.
     *
     * @param endpointerVadEndLength         Window length in number of Speex frames over which the speech
     *                                       end detection statistic is evaluated.
     *
     * @param endpointerVadEndThreshold      Threshold used to detect end-of-speech. When start detection
     *                                       statistic exceeds this number end-of-speech is detected.
     *
     * @param endpointerVadInterSpeechLength Only relevant if endpointerVadLongUtterance is 1, and
     *                                       endpointerStopOnEndOfSpeech is 0. After end-of-speech is
     *                                       detected, the speech detection algorithm waits for the number
     *                                       of Speex frames specified by endpointerVadInterSpeechLength
     *                                       before checking for start-of-speech again.
     */
    // TODO What do any of these mean?  What are their valid ranges?
    public SpeexEncoderPipe(int endpointerStopOnEndOfSpeech,
                            int endpointerVadLongUtterance,
                            int endpointerVadHistoryLength,
                            int endpointerVadBeginLength,
                            int endpointerVadBeginThreshold,
                            int endpointerVadBeginDelay,
                            int endpointerVadEndLength,
                            int endpointerVadEndThreshold,
                            int endpointerVadInterSpeechLength)
    {
        _endpointerStopOnEndOfSpeech    = endpointerStopOnEndOfSpeech;
        _endpointerVadLongUtterance     = endpointerVadLongUtterance;
        _endpointerVadHistoryLength     = endpointerVadHistoryLength;
        _endpointerVadBeginLength       = endpointerVadBeginLength;
        _endpointerVadBeginThreshold    = endpointerVadBeginThreshold;
        _endpointerVadBeginDelay        = endpointerVadBeginDelay;
        _endpointerVadEndLength         = endpointerVadEndLength;
        _endpointerVadEndThreshold      = endpointerVadEndThreshold;
        _endpointerVadInterSpeechLength = endpointerVadInterSpeechLength;
        _encoder = new SpeexEncoder();
    }

    private AudioType getEncodedAudioType(AudioType audioType)
    {
        Assert.assertTrue(isAudioSourceTypeSupported(audioType));
        if (audioType.frequency == AudioType.Frequency.FREQ_8KHZ)
            return AudioType.SPEEX_NB;
        else
            return AudioType.SPEEX_WB;
    }

    @Override
    protected boolean isAudioSourceTypeSupported(AudioType type)
    {
        return type.encoding == AudioType.Encoding.PCM_16;
    }

    @Override
    protected void initialize(AudioType audioType)
    {
        _samplesPerFrame = audioType.frequency == AudioType.Frequency.FREQ_8KHZ ? 160 : 320;
        _encoder.initialize(getEncodedAudioType(audioType),
                            _endpointerStopOnEndOfSpeech,
                            _endpointerVadLongUtterance,
                            _endpointerVadHistoryLength,
                            _endpointerVadBeginLength,
                            _endpointerVadBeginThreshold,
                            _endpointerVadBeginDelay,
                            _endpointerVadEndLength,
                            _endpointerVadEndThreshold,
                            _endpointerVadInterSpeechLength);
    }

    @Override
    protected AudioChunk[] convert(AudioChunk raw)
    {
        Logger.verbose(this, "[LATCHK] convert() start raw:" + raw);
        AudioChunk paddedChunk = raw;

        int differenceLength = raw.audioShorts.length % (_samplesPerFrame); // 2 bytes per sample (1 short)
        if (differenceLength != 0) {
            Logger.warn(this,  "Padding audio chunk for Speex since not aligned on 20ms.");
            int closestLength = raw.audioShorts.length + (_samplesPerFrame - differenceLength);
            short[] padBuffer = new short[closestLength];
            System.arraycopy(raw.audioShorts, 0, padBuffer, 0, raw.audioShorts.length);
            paddedChunk = new AudioChunk(raw.audioType, padBuffer, raw.audioTimestamp);
        }

        AudioChunk[] chunks = _encoder.encode(paddedChunk.audioShorts, 0, paddedChunk.audioShorts.length);

        if (com.nuance.dragon.toolkit.Build.DEBUG)
        {
            StringBuilder log = new StringBuilder();
            for (AudioChunk ac : chunks)
            {
                log.append(ac);
                log.append("; ");
            }
            Logger.verbose(this, "[LATCHK] convert() end chunk:" + log);
        }
        return chunks;
    }

    @Override
    public AudioType getAudioType()
    {
        AudioSource<AudioChunk> connectedSource = getConnectedSource();
        if (connectedSource != null)
        {
            AudioType sourceType = connectedSource.getAudioType();
            if (sourceType != null)
            {
                return getEncodedAudioType(sourceType);
            }
        }
        return AudioType.UNKNOWN;
    }

    @Override
    protected void cleanup()
    {
        _encoder.release();
    }
}
