package com.nuance.dragon.toolkit.audio.pipes;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioPipe;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.MultiChannelAudioChunk;
import com.nuance.dragon.toolkit.audio.util.MultiIteratorQueue;
import com.nuance.dragon.toolkit.oem.api.Logger;

/**
 * Pipe that converts a single multi-channel audio source into multiple
 * single-channel input audio sources.
 */
public class SplitterPipe extends AudioPipe<MultiChannelAudioChunk, AudioChunk>
{
    private final List<AudioSink<AudioChunk>> _currentAudioSinks;
    private final MultiIteratorQueue<MultiChannelAudioChunk> _buffers;
    private final List<MultiIteratorQueue<MultiChannelAudioChunk>.Iterator> _iterators;

    public SplitterPipe()
    {
        super(null);
        
        _currentAudioSinks = new ArrayList<AudioSink<AudioChunk>>();
        _buffers = new MultiIteratorQueue<MultiChannelAudioChunk>();
        _iterators = new ArrayList<MultiIteratorQueue<MultiChannelAudioChunk>.Iterator>();
    }

    @Override
    protected void chunksAvailable(AudioSource<MultiChannelAudioChunk> fromSource, AudioSink<MultiChannelAudioChunk> internalSink)
    {
        int chunksAvail = fromSource.getChunksAvailableForSink(internalSink);
        if (chunksAvail > 0)
        {
            while (chunksAvail-- > 0)
            {
                MultiChannelAudioChunk chunk = fromSource.getAudioChunkForSink(internalSink);
                _buffers.add(chunk);
            }

            for (AudioSink<AudioChunk> sink : _currentAudioSinks)
            {
                sink.chunksAvailable(SplitterPipe.this);
            }
        }
    }

    @Override
    protected void framesDropped(AudioSource<MultiChannelAudioChunk> fromSource, AudioSink<MultiChannelAudioChunk> internalSink)
    {
        for (AudioSink<AudioChunk> sink : _currentAudioSinks)
        {
            sink.framesDropped(SplitterPipe.this);
        }
    }

    @Override
    protected void sourceClosed(AudioSource<MultiChannelAudioChunk> fromSource, AudioSink<MultiChannelAudioChunk> internalSink)
    {
        Logger.debug(SplitterPipe.this, "Source closed.");

        // iterate the list manually instead of using an Interator
        // to avoid ConcurrentModificationException.
        for (int i = _currentAudioSinks.size() - 1; i >= 0; i--)
        {
            AudioSink<AudioChunk> sink = _currentAudioSinks.get(i);
            sink.sourceClosed(SplitterPipe.this);
        }
    }

    @Override
    public AudioType getAudioType()
    {
        return getAudioTypeFromSource();
    }

    @Override
    public int getChunksAvailableForSink(AudioSink<AudioChunk> sink)
    {
        int index = _currentAudioSinks.indexOf(sink);
        if (index != -1)
        {
            return _iterators.get(index).size();
        }

        return 0;
    }

    @Override
    public boolean isActive()
    {
        return isSourceActive();
    }

    @Override
    public int getChunksAvailable()
    {
        return 0;
    }

    @Override
    public AudioChunk getAudioChunkForSink(AudioSink<AudioChunk> sink)
    {
        int index = _currentAudioSinks.indexOf(sink);
        if (index != -1)
        {
            MultiIteratorQueue<MultiChannelAudioChunk>.Iterator chunks = _iterators.get(index);
            if (chunks.size() > 0) {
                MultiChannelAudioChunk next = chunks.next();
                if (next != null)
                {
                    return next.getSingleChannelChunk(index);
                } else
                {
                    Logger.error(this, "getAudioChunkForSink(): Expected non-null chunk");
                }
            }
        } else
        {
            Logger.warn(this, "getAudioChunkForSink(): Can't find sink");
        }

        return null;
    }

    @Override
    protected void audioSinkConnected(AudioSink<AudioChunk> sink)
    {
        Logger.debug(this, "Audio Sink connected!");
        _currentAudioSinks.add(sink);
        _iterators.add(_buffers.createIterator());

        if (!isSourceActive())
            sink.sourceClosed(this);
    }

    @Override
    protected void audioSinkDisconnected(AudioSink<AudioChunk> sink)
    {
        Logger.debug(this, "Audio Sink disconnected!");
        int index = _currentAudioSinks.indexOf(sink);
        if (index >= 0) {
            _currentAudioSinks.remove(index);
            MultiIteratorQueue<MultiChannelAudioChunk>.Iterator removed = _iterators.remove(index);
            removed.release();
        } else {
            Logger.warn(this, "Couldn't find audio sink to remove");
        }
    }
}
