package com.nuance.dragon.toolkit.audio.pipes;

import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * SSE pre-init parameters, which cannot be changed after SSE initialized
 */
public class SseInitParam
{
    public static class Builder {
        private int _micNum;
        private boolean _isEchoCancelOn;
        private boolean _isNrOn;
        private boolean _isBeamFormingOn;
        private boolean _isPicOn;
        private boolean _isForVoconASR;

        /**
         * Builder to help to set pre-init parameters before passing to SSE initializing.
         */
        public Builder()
        {
            this._micNum = 1;
            this._isEchoCancelOn = true;
            this._isNrOn = true;
            this._isBeamFormingOn = false;
            this._isPicOn = false;
            this._isForVoconASR = false;
        }

        /**
         * Set custom microphone number
         * @param number The microphone number up to 2, it is set to multiple microphones
         *              for passenger interference canceller.
         *
         */
        public Builder setMicrophoneNumber(int number)
        {
            this._micNum = number;
            return this;
        }

        /**
         * Enable or disable echo cancellation
         * @param on true to enable and false to disable. Echo Cancellation is turned on
         *          by default.
         *
         */
        public Builder setEchoCancelOnOff(boolean on)
        {
            this._isEchoCancelOn = on;
            return this;
        }

        /**
         * Enable or disable noise reduction
         * @param on true to enable and false to disable. Noise Reduction is turned on
         *          by default.
         *
         */
        public Builder setNrOnOff(boolean on)
        {
            this._isNrOn = on;
            return this;
        }

        /**
         * Enable or disable beam former
         * @param on true to enable and false to disable. Beam Former is turned off
         *          by default.
         *
         */
        public Builder setBeamFormingOnOff(boolean on)
        {
            this._isBeamFormingOn = on;
            return this;
        }

        /**
         * Enable or disable passenger interference canceller
         * @param on true to enable and false to disable. Passenger Interference Canceller is turned off
         *          by default.
         *
         */
        public Builder setPicOnOff(boolean on)
        {
            this._isPicOn = on;
            return this;
        }

        /**
         * Set SSE to prepare audio data for the VoCon ASR
         * @param on true to enable and false to disable. By default, SSE doesn't add additional information
         *          to the audio signal for ASR.
         *
         */
        public Builder setVoconASR(boolean on)
        {
            this._isForVoconASR = on;
            return this;
        }

        /**
         * Create instance of {@link SseInitParam}
         */
        public SseInitParam build()
        {
            Checker.checkArgForCondition("_micNum", "> 0 and <= 2", (_micNum > 0 &&_micNum <= 2));
            return new SseInitParam(this);
        }

    }

    private final int _micNum;
    private final boolean _isEchoCancelOn;
    private final boolean _isNrOn;
    private final boolean _isBeamFormingOn;
    private final boolean _isPicOn;
    private final boolean _isForVoconASR;

    private SseInitParam(Builder builder)
    {
        this._micNum = builder._micNum;
        this._isEchoCancelOn = builder._isEchoCancelOn;
        this._isNrOn = builder._isNrOn;
        this._isBeamFormingOn = builder._isBeamFormingOn;
        this._isPicOn = builder._isPicOn;
        this._isForVoconASR = builder._isForVoconASR;
    }

    public int getMicNum()
    {
        return _micNum;
    }

    public boolean isEchoCancelOn()
    {
        return _isEchoCancelOn;
    }

    public boolean isNrOn()
    {
        return _isNrOn;
    }

    public boolean isBeamFormingOn()
    {
        return _isBeamFormingOn;
    }

    public boolean isPicOn()
    {
        return _isPicOn;
    }

    public boolean isForVoconASR()
    {
        return _isForVoconASR;
    }

    @Override
    public String toString()
    {
        return "SseInitParam [_micNum=" + _micNum
                + ", _isEchoCancelOn=" + _isEchoCancelOn
                + ", _isNrOn=" + _isNrOn
                + ", _isBeamFormingOn=" + _isBeamFormingOn
                + ", _isPicOn=" + _isPicOn
                + ", _isForVoconASR=" + _isForVoconASR + "]";
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + _micNum;
        result = prime * result + (_isEchoCancelOn ? 1231 : 1237);
        result = prime * result + (_isNrOn ? 1151 : 1153);
        result = prime * result + (_isBeamFormingOn ? 1049 : 1051);
        result = prime * result + (_isPicOn ? 2143 : 2153);
        result = prime * result + (_isForVoconASR ? 2311 : 2333);
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        SseInitParam other = (SseInitParam) obj;
        if (_micNum != other._micNum)
            return false;
        if (_isEchoCancelOn != other._isEchoCancelOn)
            return false;
        if (_isNrOn != other._isNrOn)
            return false;
        if (_isBeamFormingOn != other._isBeamFormingOn)
            return false;
        if (_isPicOn != other._isPicOn)
            return false;
        if (_isForVoconASR != other._isForVoconASR)
            return false;
        return true;
    }

}
