package com.nuance.dragon.toolkit.audio.pipes;

import java.util.Hashtable;

import android.content.Context;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.MultiChannelAudioChunk;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.sse.Sse;

/**
 * A pipe that performs noise-suppression and/or echo cancellation.
 * <p>
 * The input to this pipe is a source of multi-channel audio chunks. Channel 0
 * is considered to be the audio input (i.e. from the microphone). An optional
 * channel 1 is considered to be the audio that was playing in the background
 * at the same time the audio was recorded.
 * <p>
 * The output audio will be the audio from channel 0, with noise suppression
 * applied, and echo-cancellation applied for every chunk that has 2 channels.
 * <p>
 * Typically, {@link AlignedMergerPipe} would be used to align the input and
 * background audio streams in order to generate the multi-channel chunks.
 */
public class SsePipe extends ConverterPipe<MultiChannelAudioChunk, AudioChunk>
{
    protected final Sse _sse;
    protected boolean _initialized;
    protected FileManager _configDataFileManager;
    protected Hashtable<AudioType, String> _configDataFileNames;
    @SuppressWarnings({ "deprecation", "unused" })
    private SseTuningType _tuningType;
    protected final boolean _internalSse;
    protected SseInitParam _initParam;
    protected byte[] _voconConfiguration;

    /**
     * Create an SSE pipe.
     * @param context The Android context object.
     */
    public SsePipe(Context context)
    {
        this(context, new SseInitParam.Builder().build(), null, null);
    }

    /**
     * Create an SSE pipe.
     * @param context The Android context object.
     * @param configDataFileManager The file manager where SSE should look for the configData file.
     * @param configDataFileNames The configData file names for different AudioTypes.
     */
    public SsePipe(Context context, FileManager configDataFileManager, Hashtable<AudioType, String> configDataFileNames)
    {
        this(context, new SseInitParam.Builder().build(), configDataFileManager, configDataFileNames);
    }


    /**
     * @deprecated
     * Create an SSE pipe.
     * @param context The Android context object.
     * @param configDataFileManager The file manager where SSE should look for the configData file.
     * @param configDataFileNames The configData file names for different AudioTypes.
     * @param tuningType The type of audio tuning to be done on the pipe (echo cancellation, noise suppression, beam forming)
     */
    @Deprecated
    public SsePipe(Context context, FileManager configDataFileManager, Hashtable<AudioType, String> configDataFileNames, SseTuningType tuningType)
    {
        super(0);
        Checker.checkArgForNull("context", context.getApplicationContext());
        _sse = new Sse(context);
        _configDataFileManager = configDataFileManager;
        _configDataFileNames = configDataFileNames;
        _tuningType = tuningType;
        _internalSse = true;
        _initParam = null;
        _voconConfiguration = null;
    }

    /**
     * Create an SSE pipe.
     * @param context The Android context object.
     * @param initParam The parameters have to be set prior to SSE initialization
     * @param configDataFileManager The file manager where SSE should look for the configData file.
     * @param configDataFileNames The configData file names for different AudioTypes. It will override any parameters
     *                             set in initParam if applicable
     */
    public SsePipe(Context context, SseInitParam initParam, FileManager configDataFileManager, Hashtable<AudioType, String> configDataFileNames)
    {
        super(0);
        Checker.checkArgForNull("context", context.getApplicationContext());
        _sse = new Sse(context);
        _configDataFileManager = configDataFileManager;
        _configDataFileNames = configDataFileNames;
        _internalSse = true;
        _initParam = initParam;
        _voconConfiguration = null;
    }

    /**
     * Create an SSE pipe.
     * @param sseProxy An existing SSE proxy instance. The proxy must be have been initialized already.
     */
    public SsePipe(SseProxy sseProxy)
    {
        super(0);
        Checker.checkArgForNull("sseProxy", sseProxy);
        _sse = sseProxy.getSse();
        _voconConfiguration = _sse.getVoconConfiguration();
        _configDataFileManager = null;
        _configDataFileNames = null;
        _internalSse = false;
        _initParam = null;
    }

    /**
     * Get the required frame size (in ms) for the SSE engine. All audio
     * chunks input to this pipe from its connected source should have a
     * duration that is an exact multiple of this value.
     * @return The required frame size in ms
     */
    public static int getRequiredFrameSize()
    {
        return Sse.getRequiredFrameSize();
    }

    @Override
    protected void initialize(AudioType audioType)
    {
        if (!_initialized)
        {
            if (_internalSse)
            {
                if (_initParam == null)
                    _initParam = new SseInitParam.Builder().build();
                boolean success = _sse.initialize(audioType, _initParam, _configDataFileManager, _configDataFileNames != null ? _configDataFileNames.get(audioType) : null);
                if (success && _initParam.isForVoconASR())
                {
                    _voconConfiguration = _sse.getVoconConfiguration();
                }
                else if (!success)
                {
                    Logger.error(this, "Failure initializing SSE");
                }
            }
            _initialized = true;
        }
    }

    @Override
    protected void cleanup()
    {
        if (_initialized)
        {
            if (_internalSse)
            {
                _sse.release();
            }
        }
        _initialized = false;
    }

    @Override
    protected AudioChunk[] convert(MultiChannelAudioChunk raw)
    {
        AudioChunk chunk = _sse.process(raw);
        if (chunk != null)
        {
            return new AudioChunk[] { chunk };
        }

        return new AudioChunk[] { raw.getSingleChannelChunk(0) };
    }

    @Override
    public AudioType getAudioType()
    {
        AudioType sourceAudioType = getAudioTypeFromSource();
        return new AudioType(sourceAudioType.encoding,
                             sourceAudioType.frequency,
                             _voconConfiguration);
    }


    /**
     * Get embedded SSE library version
     * @return SSE library version string
     */
    public String getSseVersion()
    {
        return _sse.getVersion();
    }
}
