package com.nuance.dragon.toolkit.audio.pipes;

import android.content.Context;

import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.WorkerThread;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.sse.Sse;

/**
 * Wraps native SSE and runs loading/releasing operations on a worker thread for convenience.
 * This is solely intended to be used with {@link SsePipe} and {@link NoiseSuppressionPipe}.
 */
public class SseProxy
{
    /**
     * Listens for the completion of {@link SseProxy#initialize}.
     */
    public interface InitializeListener
    {
        /**
         * The initialization completed.
         * @param success false if the initialization was unsuccessful
         */
        void onLoaded(boolean success);
    }

    /**
     * Listens for the completion of {@link SseProxy#release}.
     */
    public interface ReleaseListener
    {
        /**
         * The release completed.
         */
        void onReleased();
    }

    private final Context _context;
    private final WorkerThread _workerThread;
    private final NMTHandler _mainThreadHandler, _workerThreadHandler;
    private boolean _initialized;
    private Sse _sse;

    /**
     * Create the SSE proxy.
     * @param context The application context.
     */
    public SseProxy(Context context)
    {
        Checker.checkArgForNull("context", context.getApplicationContext());
        _context = context;

        _workerThread = Factory.createWorkerThread("com.nuance.dragon.toolkit.audio.pipes.SseProxy");
        _workerThread.start();

        _mainThreadHandler = Factory.createNMTHandler();
        _workerThreadHandler = _workerThread.getHandler();
    }

    /**
     * @deprecated
     * Initialize SSE, same SSE setup params than in {@link SsePipe} and {@link NoiseSuppressionPipe}.
     * @param useVocon True if VoCon is used as the application recognizer, else false (Elvis, cloud).
     */
    @Deprecated
    public void initialize(final AudioType audioType, final FileManager configDataFileManager, final String configDataFileName, final SseTuningType tuningType, final boolean useVocon)
    {
        initialize(audioType, configDataFileManager, configDataFileName, tuningType, useVocon, null);
    }

    /**
     * @deprecated
     * Initialize SSE, same SSE setup params than in {@link SsePipe} and {@link NoiseSuppressionPipe}.
     * @param useVocon True if VoCon is used as the application recognizer, else false (Elvis, cloud).
     * @param listener Notifies when SSE is loaded. Can be null.
     */
    @Deprecated
    public void initialize(final AudioType audioType, final FileManager configDataFileManager, final String configDataFileName, final SseTuningType tuningType, final boolean useVocon, final InitializeListener listener)
    {
        Checker.checkState(this, !_initialized);
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _sse = new Sse(_context);
                final boolean success = _sse.initialize(audioType, configDataFileManager, configDataFileName, tuningType, useVocon);
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        _initialized  = true;

                        if (listener != null)
                            listener.onLoaded(success);
                    }
                });
            }
        });
    }

    /**
     * Initialize SSE, same SSE setup parameters as in {@link SsePipe} and {@link NoiseSuppressionPipe}.
     * @param audioType It includes the sample rate to initialize SSE
     * @param initParam The parameters have to be set prior to SSE initialization
     * @param configDataFileManager Optional the file manager to find SSE initial configuration file
     * @param configDataFileName Optional the SSE initial configuration file. It will override any parameters
     *                           set in initParam if applicable.
     * @param listener Notifies when SSE is loaded. Can be null.
     */
    public void initialize(final AudioType audioType, final SseInitParam initParam, final FileManager configDataFileManager, final String configDataFileName, final InitializeListener listener)
    {
        Checker.checkState(this, !_initialized);
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _sse = new Sse(_context);
                final boolean success = _sse.initialize(audioType, initParam, configDataFileManager, configDataFileName);
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        _initialized  = true;

                        if (listener != null)
                            listener.onLoaded(success);
                    }
                });
            }
        });
    }

    /**
     * Releases SSE.
     */
    public void release()
    {
        release(null);
    }

    /**
     * Releases SSE.
     * @param listener Notifies when SSE is released. Can be null.
     */
    public void release(final ReleaseListener listener)
    {
        Checker.checkState(this, _initialized);
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _sse.release();
                _sse = null;

                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        _workerThread.stop();
                        _initialized = false;

                        if (listener != null)
                            listener.onReleased();
                    }
                });
            }
        });
    }

    Sse getSse()
    {
        Checker.checkState(this, _initialized, "SseProxy is not initialized yet.");
        return _sse;
    }
}