package com.nuance.dragon.toolkit.audio.pipes;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * A pipe that reports a notification as soon as a certain duration of audio has
 * been processed. This does not automatically close the stream--it just calls
 * {@link Listener#onTimeout}.
 */
public class TimeoutPipe<AudioChunkType extends AudioChunk> extends BufferingPipe<AudioChunkType>
{
    /**
     * Timeout listener.
     */
    public interface Listener
    {
        /**
         * Listener method that is called as soon as the audio duration has been
         * met or surpassed. This method will only be called once. The pipe will
         * continue passing audio along even after timing out.
         */
        void onTimeout();
    }

    private final int _timeoutMs;
    private final Listener _listener;
    private int _ms;
    private boolean _timedOut;

    /**
     * @param timeoutMs The timeout, in milliseconds. Must be > 0.
     * @param listener The listener to be called when the timeout is met.
     */
    public TimeoutPipe(int timeoutMs, Listener listener)
    {
        Checker.checkArgForCondition("timeoutMs", "greater than 0", timeoutMs > 0);
        Checker.checkArgForNull("listener", listener);
        _timeoutMs = timeoutMs;
        _listener = listener;
    }

    @Override
    protected void onChunkBuffered(AudioChunkType chunk)
    {
        super.onChunkBuffered(chunk);

        _ms += chunk.audioDuration;
        if (_ms >= _timeoutMs && !_timedOut)
        {
            _timedOut = true;
            _listener.onTimeout();
        }
    }

    @Override
    public void connectAudioSource(AudioSource<AudioChunkType> source)
    {
        _ms = 0;
        _timedOut = false;
        super.connectAudioSource(source);
    }
}
