package com.nuance.dragon.toolkit.audio.pipes;

import java.util.Hashtable;

import android.content.Context;

import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;

/**
 * A pipe that performs noise-suppression and/or echo cancellation and is meant to
 * be used with {@link com.nuance.dragon.toolkit.vocon.VoconRecognizer}.
 * <p>
 * The input to this pipe is a source of multi-channel audio chunks. Channel 0
 * is considered to be the audio input (i.e. from the microphone). An optional
 * channel 1 is considered to be the audio that was playing in the background
 * at the same time the audio was recorded.
 * <p>
 * The output audio will be the audio from channel 0, with noise suppression
 * applied, and echo-cancellation applied for every chunk that has 2 channels.
 * <p>
 * Typically, {@link AlignedMergerPipe} would be used to align the input and
 * background audio streams in order to generate the multi-channel chunks.
 */
public class VoconSsePipe extends SsePipe
{
    /**
     * Create a VoCon SSE pipe.
     * @param context The Android context object.
     */
    public VoconSsePipe(Context context)
    {
        super(context);
    }

    /**
     * Create a VoCon SSE pipe.
     * @param context The Android context object.
     * @param configDataFileManager The file manager where SSE should look for the configData file.
     * @param configDataFileNames The configData file names for different AudioTypes.
     */
    public VoconSsePipe(Context context, FileManager configDataFileManager, Hashtable<AudioType, String> configDataFileNames)
    {
        super(context, configDataFileManager, configDataFileNames);
    }


    /**
     * @deprecated
     * Create a VoCon SSE pipe.
     * @param context The Android context object.
     * @param configDataFileManager The file manager where SSE should look for the configData file.
     * @param configDataFileNames The configData file names for different AudioTypes.
     * @param tuningType The type of audio tuning to be done on the pipe (echo cancellation, noise suppression, beam forming)
     */
    @Deprecated
    public VoconSsePipe(Context context, FileManager configDataFileManager, Hashtable<AudioType, String> configDataFileNames, SseTuningType tuningType)
    {
        super(context, configDataFileManager, configDataFileNames, tuningType);
    }

    /**
     * Create a VoCon SSE pipe.
     * @param context The Android context object.
     * @param initParam The parameters have to be set prior to SSE initialization
     * @param configDataFileManager The file manager where SSE should look for the configData file.
     * @param configDataFileNames The configData file names for different AudioTypes. It will override any parameters
     *                             set in initParam if applicable
     */
    public VoconSsePipe(Context context, SseInitParam initParam, FileManager configDataFileManager, Hashtable<AudioType, String> configDataFileNames)
    {
        super(context, initParam, configDataFileManager, configDataFileNames);
    }

    /**
     * Create a VoCon SSE pipe.
     * @param sseProxy An existing SSE proxy instance. The proxy must be have been initialized already.
     */
    public VoconSsePipe(SseProxy sseProxy)
    {
        super(sseProxy);
    }

    @Override
    protected void initialize(AudioType audioType)
    {
        if (!_initialized)
        {
            if (_internalSse)
            {
                SseInitParam rInitParam;
                if (_initParam != null)
                {
                    rInitParam = new SseInitParam.Builder().setMicrophoneNumber(_initParam.getMicNum()).setEchoCancelOnOff(_initParam.isEchoCancelOn()).
                            setNrOnOff(_initParam.isNrOn()).setBeamFormingOnOff(_initParam.isBeamFormingOn()).setPicOnOff(_initParam.isPicOn()).setVoconASR(true).build();
                }
                else
                {
                    rInitParam = new SseInitParam.Builder().setVoconASR(true).build();
                }
                boolean success = _sse.initialize(audioType, rInitParam, _configDataFileManager, _configDataFileNames != null ? _configDataFileNames.get(audioType) : null);
                if (success)
                {
                    _voconConfiguration = _sse.getVoconConfiguration();
                }
                else
                {
                    Logger.error(this, "Failure initializing SSE");
                }
            }
            _initialized = true;
        }
    }
}
