package com.nuance.dragon.toolkit.audio.sinks;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.List;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;

/**
 * The file sink acts like a player, but saves all audio to a file instead of
 * playing it to the speaker.
 */
public class FilePlayerSink extends PlayerSink
{
    private static final int AUDIO_REQUEST_SIZE_MS = 1000;
    private String _fileName;
    private boolean _littleEndian;
    private AudioType _audioType;
    private OutputStream _outputStream = null;
    private boolean _done = false;
    private final FileManager _fileManager;

    /**
     * Create a file player sink.
     *
     * @param audioType The audio type to play. Can't be null.
     * @param fileName The file to write audio to
     * @param fileManager is used to access file
     * @param littleEndian True to save in little-endian format; false for big-
     * endian.
     * @param workerThreadHandler The worker thread to use, or null if we should
     * create a temporary one.
     * @param mainThreadHandler The main thread handler to use, or null if we should
     * create a temporary one.
     */
    public FilePlayerSink(AudioType audioType, String fileName, FileManager fileManager, boolean littleEndian, NMTHandler workerThreadHandler, NMTHandler mainThreadHandler)
    {
        super(audioType, workerThreadHandler, mainThreadHandler);
        _fileName = fileName;
        _littleEndian = littleEndian;
        _fileManager = fileManager;
    }

    /**
     * Create a file player sink.
     *
     * @param audioType The audio type to play. Can't be null.
     * @param fileName The file to write audio to
     * @param fileManager is used to access file
     * @param littleEndian True to save in little-endian format; false for big-
     * endian.
     * @param workerThreadHandler The worker thread to use, or null if we should
     * create a temporary one.
     */
    public FilePlayerSink(AudioType audioType, String fileName, FileManager fileManager, boolean littleEndian, NMTHandler workerThreadHandler)
    {
        this(audioType, fileName, fileManager, littleEndian, workerThreadHandler, null);
    }

    /**
     * Create a file player sink.
     * @param audioType The audio type to play. Can't be null.
     * @param fileName The file to write audio to
     * @param littleEndian True to save in little-endian format; false for big-
     * endian.
     */
    public FilePlayerSink(AudioType audioType, String fileName, boolean littleEndian)
    {
        this(audioType, fileName, null, littleEndian, null);
    }

    /**
     * Create a file player sink that saves the audio in little-endian format.
     * @param audioType The audio type to play. Can't be null.
     * @param fileName The file to write audio to
     */
    public FilePlayerSink(AudioType audioType, String fileName)
    {
        this(audioType, fileName, null, true, null);
    }

    /**
     * Create a stream player sink that saves the audio in little-endian format.
     * @param audioType The audio type to play. Can't be null.
     * @param outputStream The stream to write audio to.
     */
    public FilePlayerSink(AudioType audioType, OutputStream outputStream)
    {
        this(audioType, outputStream, true);
    }

    /**
     * Create a stream player sink.
     * @param audioType The audio type to play. Can't be null.
     * @param outputStream The stream to write audio to.
     * @param littleEndian True to save in little-endian format; false for big-
     * endian.
     */
    public FilePlayerSink(AudioType audioType, OutputStream outputStream, boolean littleEndian)
    {
        this(audioType, (String)null, littleEndian);
        _outputStream = outputStream;
    }

    @Override
    protected boolean startPlayingInternal(AudioType audioType)
    {
        _audioType = audioType;

        if (_outputStream == null) {
            boolean isAbsolute = new File(_fileName).isAbsolute();
            if (_fileManager != null && !isAbsolute)
            {
                _outputStream = _fileManager.openFileForWriting(_fileName);
                if (_outputStream == null)
                {
                    _done = true;
                    return false;
                }
            }
            else
            {
                try
                {
                    _outputStream = new FileOutputStream(new File(_fileName));
                }
                catch (FileNotFoundException e)
                {
                    _done = true;
                    return false;
                }
            }
        }
        handleStarted();
        handleAudioNeeded(AUDIO_REQUEST_SIZE_MS);
        return true;
    }

    @Override
    protected void stopPlayingInternal()
    {
        cleanup();
    }

    // Called from worker thread
    @Override
    protected void newChunksInternal(List<AudioChunk> chunks)
    {
        if (_done) return;

        // Always ask for 1000 more ms--simple way of constantly getting as much
        // audio as possible.
        handleAudioNeeded(AUDIO_REQUEST_SIZE_MS);

        for (AudioChunk chunk : chunks)
        {
            byte [] buf;
            if (_audioType.encoding == AudioType.Encoding.PCM_16)
            {
                short[] shorts = chunk.audioShorts;
                buf = new byte[shorts.length * 2];
                for (int i = 0; i < shorts.length; i++)
                {
                    short s = shorts[i];
                    if (_littleEndian)
                    {
                        buf[i * 2] = (byte)s;
                        buf[i * 2 + 1] = (byte)(s >> 8);
                    }
                    else
                    {
                        buf[i * 2 + 1] = (byte)s;
                        buf[i * 2] = (byte)(s >> 8);
                    }
                }
            }
            else
            {
                buf = chunk.audioBytes;
            }

            try
            {
                if (_audioType.encoding == AudioType.Encoding.SPEEX || _audioType.encoding == AudioType.Encoding.OPUS )
                {
                    // for speex add Nuance header for each buffer so audiomorphosis can playback it
                    if(buf.length < 0x80) {
                        _outputStream.write(buf.length & 0x7F);
                    }
                    else if(buf.length < 0x4000) {
                        _outputStream.write(((buf.length >> 7) & 0x7F) | 0x80);
                        _outputStream.write(buf.length & 0x7F);
                    }
                    else if(buf.length < 0x200000) {
                        _outputStream.write(((buf.length >> 14) & 0x7F) | 0x80);
                        _outputStream.write(((buf.length >> 7) & 0x7F) | 0x80);
                        _outputStream.write(buf.length & 0x7F);
                    }
                    else if(buf.length < 0x10000000) {
                        _outputStream.write(((buf.length >> 21) & 0x7F) | 0x80);
                        _outputStream.write(((buf.length >> 14) & 0x7F) | 0x80);
                        _outputStream.write(((buf.length >> 7) & 0x7F) | 0x80);
                        _outputStream.write(buf.length & 0x7F);
                    }
                    else {
                        Logger.error(FilePlayerSink.this, "chunksAvailableInternal() buffer size is too big!!!");
                        cleanup();
                        return;
                    }
                }
                _outputStream.write(buf);
            }
            catch (IOException e)
            {
                Logger.error(FilePlayerSink.this, "chunksAvailableInternal() exception thrown!!!");
                cleanup();
                return;
            }
        }
    }

    // Called from worker thread
    @Override
    protected void noNewChunksInternal()
    {
        cleanup();
    }

    private void cleanup()
    {
        if (_done) return;

        try
        {
            _outputStream.close();
        }
        catch (IOException e) {}
        _done = true;

        handleStopped();
    }

    // supports all codecs
    @Override
    protected boolean isCodecSupported(AudioType audioType) {
        return true;
    }
}
