package com.nuance.dragon.toolkit.audio.sinks;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.pipes.DuplicatorPipe;
import com.nuance.dragon.toolkit.audio.sources.SingleSinkSource;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.WorkerThread;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.impl.WorkerThreadOem;

/**
 * An audio sink that consumes audio from a source and plays it via the device
 * speaker.
 */
public abstract class PlayerSink extends AudioSink<AudioChunk>
{
    /**
     * Listens for when the player starts and stops.
     */
    public interface Listener
    {
        /**
         * The player started.
         * @param player The player. Never null.
         */
        void onStarted(PlayerSink player);

        /**
         * The player stopped.
         * @param player The player. Never null.
         */
        void onStopped(PlayerSink player);
    }

    private class RefSource extends SingleSinkSource<AudioChunk>
    {
        private final LinkedList<AudioChunk> _chunks;

        RefSource(NMTHandler mainThreadHandler)
        {
            super(mainThreadHandler);
            _chunks = new LinkedList<AudioChunk>();
        }

        void addChunks(List<AudioChunk> chunks)
        {
            _chunks.addAll(chunks);

            notifyChunksAvailable();
        }

        void close()
        {
            notifySourceClosed();
        }

        @Override
        public AudioType getAudioType()
        {
            return _audioType;
        }

        @Override
        public int getChunksAvailable()
        {
            return _chunks.size();
        }

        @Override
        public boolean isActive()
        {
            return this == _refSource;
        }

        @Override
        protected AudioChunk getAudioChunk()
        {
            if (!_chunks.isEmpty())
                return _chunks.remove();
            else
                return null;
        }
    }

    private final AudioType _audioType;
    private WorkerThread _workerThread;
    protected NMTHandler _workerThreadHandler;
    private Listener _listener;
    private boolean _useOwnWorkerThread; // create own thread, false if user specifies a worker thread
    private boolean _started; // keeps track of start, to avoid starting twice
    private boolean _stopped;

    private int _missingAudioMs; // Accessed from main thread
    private int _requestingAudioMs; // Accessed from worker thread

    private RefSource _refSource;
    private DuplicatorPipe<AudioChunk> _refPipe;

    /**
     * PlayerSink constructor
     * @param audioType The audio type to play. Can't be null.
     * @param workerThreadHandler The worker thread to use, or null if we should
     * create a temporary one.
     */
    public PlayerSink(AudioType audioType, NMTHandler workerThreadHandler) {
        this(audioType, workerThreadHandler, null);
    }
    
    /**
     * PlayerSink constructor
     * @param audioType The audio type to play. Can't be null.
     * @param workerThreadHandler The worker thread to use, or null if we should
     * create a temporary one.
     */
    public PlayerSink(AudioType audioType, NMTHandler workerThreadHandler, NMTHandler mainThreadHandler) {
        super(mainThreadHandler);
        
        Checker.checkArgForNull("audioType", audioType);
        Checker.checkArgForCondition("audioType", "a type supported by this player", isCodecSupported(audioType));

        _audioType = audioType;
        if (workerThreadHandler == null)
        {
            _useOwnWorkerThread = true;
            _workerThread = null;

        } else
        {
            _workerThreadHandler = workerThreadHandler;
            _workerThread = null;
        }

        _refPipe = new DuplicatorPipe<AudioChunk>(mainThreadHandler);

        // null this out to make it appear closed, until 'startPlaying' occurs
        _refSource = null;
    }

    /**
     * Get the reference source
     * @return The reference source
     */
    public AudioSource<AudioChunk> getReferenceSource()
    {
        return _refPipe;
    }

    @Override
    public boolean isAudioSourceTypeSupported(AudioType type)
    {
        return (_audioType.encoding == type.encoding && _audioType.frequency == type.frequency);
    }

    @Override
    public void chunksAvailable(AudioSource<AudioChunk> source)
    {
        if (!_started) return;

        if (_missingAudioMs > 0)
        {
            final ArrayList<AudioChunk> chunks = new ArrayList<AudioChunk>();
            int read = readOffChunks(source, _missingAudioMs, chunks);
            _missingAudioMs -= read;
            _workerThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    newChunksInternal(chunks);
                }
            });
        }
    }

    @Override
    public void framesDropped(AudioSource<AudioChunk> source)
    {
    }

    @Override
    public void sourceClosed(AudioSource<AudioChunk> source)
    {
        Logger.debug(this, "Source closed.");
        if (!_started) return;

        final ArrayList<AudioChunk> remainingChunks = new ArrayList<AudioChunk>();
        while (source.getChunksAvailableForSink(this) > 0)
        {
            remainingChunks.add(source.getAudioChunkForSink(this));
        }

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (!remainingChunks.isEmpty())
                    newChunksInternal(remainingChunks);
                noNewChunksInternal();
            }
        });
    }

    @Override
    protected void audioSourceDisconnected(AudioSource<AudioChunk> source)
    {
        Logger.debug(this, "Source disconnected.");
        if (!_started) return;

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                noNewChunksInternal();
            }
        });

        super.audioSourceDisconnected(source);
    }

    /**
     * Read off chunks until ms is reached
     * @param connectedSource
     * @param ms The milliseconds to read
     * @param chunks The empty array of chunks to add to
     * @return The ms that were read
     */
    private int readOffChunks(AudioSource<AudioChunk> connectedSource, int ms, List<AudioChunk> chunks)
    {
        int read = 0;
        do {
            AudioChunk chunk = connectedSource.getAudioChunkForSink(this);
            if (chunk == null)
                break;

            read += chunk.audioDuration;
            chunks.add(chunk);
        } while (read < ms);
        return read;
    }

    /**
     * Start playing audio. Make sure to only call this once.
     */
    public void startPlaying()
    {
        startPlaying(null);
    }

    /**
     * Start playing audio. Make sure to only call this once.
     * @param listener Listens for the player starting and stopping. May be
     * null.
     */
    public void startPlaying(final Listener listener)
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                Checker.checkState(this, !_started, "Already started.");
        
                if (_useOwnWorkerThread) {
                    _workerThread = new WorkerThreadOem("com.nuance.dragon.toolkit.audio.sinks.PlayerSink");
                    _workerThread.start();
                    _workerThreadHandler = _workerThread.getHandler();
                }
        
                Logger.debug(this, "Starting playing");
                _started = true;
                _listener = listener;
                _missingAudioMs = 0;
        
                _refSource = new RefSource(_mainThreadHandler);
                if (_refPipe == null)
                    _refPipe = new DuplicatorPipe<AudioChunk>(_mainThreadHandler);
                _refPipe.connectAudioSource(_refSource);
        
                _workerThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        _requestingAudioMs = 0;
                        boolean started = startPlayingInternal(_audioType);
                        if (!started)
                        {
                            handleStopped();
                        }
                    }
                });
            }
        });
    }

    /**
     * Stop playing. It is safe to call this even if the player has already
     * stopped playing or hasn't started yet.
     */
    public void stopPlaying()
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                Logger.debug(this, "Stopping playing");
                if (_started && !_stopped)
                {
                    _stopped = true;
                    _workerThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            stopPlayingInternal();
                        }
                    });
                }
            }
        });
    }

    /**
     * Checks if codec is supported.
     * @param audioType Can't be null.
     * @return false if not supported, true if supported
     */
    protected abstract boolean isCodecSupported(AudioType audioType);

    /**
     * Start playing. Called once by the base class. Called from the worker thread.
     * @param audioType Can't be null.
     * @return false on failure
     */
    protected abstract boolean startPlayingInternal(AudioType audioType);

    /**
     * Stop playing. Called once by the base class. Called from the worker thread.
     */
    protected abstract void stopPlayingInternal();

    /**
     * Called from the worker thread upon a new audio chunk available.
     * @param chunks Audio chunks to add. Pass as many as possible in one call,
     * to make sure the player doesn't pad with silence. Can't be null.
     */
    protected abstract void newChunksInternal(List<AudioChunk> chunks);

    /**
     * Called from the worker thread when no more chunks will be available.
     */
    protected abstract void noNewChunksInternal();

    protected AudioType getAudioType()
    {
        return _audioType;
    }

    protected void handleChunksPlayed(final List<AudioChunk> chunksPlayed)
    {
        _mainThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _refSource.addChunks(chunksPlayed);
            }
        });
    }

    /**
     * Call from audio thread by the implementing class when more audio is
     * needed. This will trigger {@link #newChunksInternal} to be called soon if
     * audio is available (possibly with more or less audio than was requested).
     * Audio will NOT be given to {@link #newChunksInternal} unless this method
     * is called regularly!
     * @param msNeeded The total ms of audio that are needed.
     */
    protected void handleAudioNeeded(int msNeeded)
    {
        // If there is already audio being requested, adjust msNeeded so we only
        // request the remaining ms
        if (_requestingAudioMs > 0) {
            msNeeded -= _requestingAudioMs;
        }

        if (msNeeded > 0)
        {
            final int fNeeded = msNeeded;
            _requestingAudioMs += fNeeded; // Keep track of the pending audio request

            _mainThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    if (!_started)
                        return;

                    final List<AudioChunk> chunks;
                    AudioSource<AudioChunk> source = getConnectedSource();
                    int read = 0;
                    if (source != null)
                    {
                        chunks = new ArrayList<AudioChunk>();
                        read = readOffChunks(source, fNeeded, chunks);
                    } else
                    {
                        chunks = null;
                    }

                    // Keep track of the missing audio for when new chunks become
                    // available
                    _missingAudioMs = fNeeded - read;

                    final boolean finished;
                    if (_missingAudioMs > 0 && (source == null || !source.isActive()))
                    {
                        finished = true;
                        _missingAudioMs = 0;
                    }
                    else
                    {
                        finished = false;
                    }

                    _workerThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            // Now that we finished this audio request, reset
                            // _requestingAudioMs
                            _requestingAudioMs -= fNeeded;
                            Assert.assertTrue(_requestingAudioMs >= 0);

                            if (chunks != null && !chunks.isEmpty())
                                newChunksInternal(chunks);

                            if (finished)
                                noNewChunksInternal();
                        }
                    });
                }
            });
        }
    }

    /**
     * Call from audio thread by the implementing class when playback has started.
     */
    protected void handleStarted()
    {
        _mainThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (_listener != null)
                {
                    _listener.onStarted(PlayerSink.this);
                }
            }
        });
    }

    /**
     * Call from audio thread by the implementing class when playback has stopped.
     */
    protected void handleStopped()
    {
        _mainThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                disconnectAudioSource();

                if (_started)
                {
                    _started = false;
                    if (_workerThread != null)
                    {
                        _workerThread.stop();
                        _workerThread = null;
                    }
                }

                if (_listener != null)
                {
                    _listener.onStopped(PlayerSink.this);
                    _listener = null;
                }

                if (_refSource != null)
                {
                    RefSource refSource = _refSource;
                    _refSource = null;
                    _refPipe = null;
                    refSource.close();
                }
            }
        });
    }
}
