package com.nuance.dragon.toolkit.audio.sinks;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import junit.framework.Assert;
import android.media.AudioFormat;
import android.media.AudioManager;
import android.media.AudioTrack;
import android.os.Handler;
import android.os.SystemClock;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk.SpeechStatus;
import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioEnergyListener;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.AudioType.Encoding;
import com.nuance.dragon.toolkit.audio.TtsMarker;
import com.nuance.dragon.toolkit.audio.TtsMarkerListener;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.util.internal.ObjectUtils;

/**
 * Implementation of {@link PlayerSink} that plays to the device speaker.
 */
public class SpeakerPlayerSink extends PlayerSink
{
    // initial parameters
    private static final int DEFAULT_STREAM_TYPE = AudioManager.STREAM_MUSIC;
    private static final int CHANNEL_CONFIG      = AudioFormat.CHANNEL_CONFIGURATION_MONO;
    private static final int MODE                = AudioTrack.MODE_STREAM;

    // default buffer length
    private static final int PCM_BUFFER_LENGTH_IN_MILLISECS     = 400;
    private static final int PCM_CHUNK_LENGTH_IN_MILLISECS      = 20;

    private static final int PCM_8K_AUDIO_FORMAT                = AudioFormat.ENCODING_PCM_16BIT;
    private static final int PCM_11K_AUDIO_FORMAT               = AudioFormat.ENCODING_PCM_16BIT;
    private static final int PCM_16K_AUDIO_FORMAT               = AudioFormat.ENCODING_PCM_16BIT;
    private static final int PCM_22K_AUDIO_FORMAT               = AudioFormat.ENCODING_PCM_16BIT;
    private static final int PCM_44K_AUDIO_FORMAT               = AudioFormat.ENCODING_PCM_16BIT;
    private static final int PCM_48K_AUDIO_FORMAT               = AudioFormat.ENCODING_PCM_16BIT;

    private AudioTrack _player;
    private int _streamType;
    private int _adjustedBufferSizeInBytes;
    private int _adjustedBufferSizeInShorts;
    private boolean _done = false;
    private final LinkedList<AudioChunk> _bufs = new LinkedList<AudioChunk>();
    private int _off;
    private boolean _needToWrite;
    private boolean _noNewChunks;
    private Handler _readHandler; // Accessed from worker thread
    private int _totalWritten;    // Total number of samples written to the player
    private long _startTimestamp;
    private final int _bufferLengthInMs;

    private int _stalledPosition;
    private boolean _headMoved;
    private final ArrayList<AudioChunk> _pendingPlayedAudio;
    private final LinkedList<AudioChunk> _audioInfoChunks;
    private final Handler _mainThreadHandler;
    private int _audioInfoSamples;
    private final int _audioInfoDurationMs;

    // Optional event listeners
    private TtsMarkerListener _ttsMarkerListener;
    private AudioEnergyListener _audioEnergyListener;

    /**
     * Create a player that uses its own dedicated worker thread.
     * @param audioType The audio type to play. Can't be null.
     */
    public SpeakerPlayerSink(AudioType audioType)
    {
        this(audioType, PCM_BUFFER_LENGTH_IN_MILLISECS, null, DEFAULT_STREAM_TYPE, PCM_CHUNK_LENGTH_IN_MILLISECS);
    }

    /**
     * Create a player that uses its own dedicated worker thread.
     * @param audioType The audio type to play. Can't be null.
     * @param bufferLengthMs The size of the internal buffer used by the
     * internal {@link AudioTrack}. Use a bigger buffer to avoid choppiness
     * under high CPU load.
     * @param audioInfoDurationMs The duration to calculate audio information such as energy level, etc.
     * Default value is PCM_CHUNK_LENGTH_IN_MILLISECS.
     */
    public SpeakerPlayerSink(AudioType audioType, int bufferLengthMs, int audioInfoDurationMs)
    {
        this(audioType, bufferLengthMs, null, DEFAULT_STREAM_TYPE, audioInfoDurationMs);
    }

    /**
     * Create a player that uses its own dedicated worker thread.
     * @param streamType Set the stream type. Default value is
     * AudioManager.STREAM_MUSIC.
     * @param audioType The audio type to play. Can't be null.
     */
    public SpeakerPlayerSink(int streamType, AudioType audioType)
    {
        this(audioType, PCM_BUFFER_LENGTH_IN_MILLISECS, null, streamType, PCM_CHUNK_LENGTH_IN_MILLISECS);
    }

    /**
     * Create a player that uses its own dedicated worker thread.
     * @param audioType The audio type to play. Can't be null.
     * @param bufferLengthMs The size of the internal buffer used by the
     * internal {@link AudioTrack}. Use a bigger buffer to avoid choppiness
     * under high CPU load.
     */
    public SpeakerPlayerSink(AudioType audioType, int bufferLengthMs)
    {
        this(audioType, bufferLengthMs, null, DEFAULT_STREAM_TYPE, PCM_CHUNK_LENGTH_IN_MILLISECS);
    }

    /**
     * Create a player that uses the given worker thread.
     * @param audioType The audio type to play. Can't be null.
     * @param workerThreadHandler The external worker thread handler. If null,
     * the player will use its own dedicated worker thread.
     */
    public SpeakerPlayerSink(AudioType audioType, NMTHandler workerThreadHandler)
    {
        this(audioType, PCM_BUFFER_LENGTH_IN_MILLISECS, workerThreadHandler, DEFAULT_STREAM_TYPE, PCM_CHUNK_LENGTH_IN_MILLISECS);
    }

    /**
     * Create a player that uses the given worker thread.
     * @param audioType The audio type to play. Can't be null.
     * @param bufferLengthMs The size of the internal buffer used by the
     * internal {@link AudioTrack}. Use a bigger buffer to avoid choppiness
     * under high CPU load.
     * @param workerThreadHandler The external worker thread handler. If null,
     * the player will use its own dedicated worker thread.
     */
    public SpeakerPlayerSink(AudioType audioType, int bufferLengthMs, NMTHandler workerThreadHandler)
    {
        this(audioType, bufferLengthMs, workerThreadHandler, DEFAULT_STREAM_TYPE, PCM_CHUNK_LENGTH_IN_MILLISECS);
    }

    /**
     * Create a player that uses the given worker thread.
     * @param audioType The audio type to play. Can't be null.
     * @param bufferLengthMs The size of the internal buffer used by the
     * internal {@link AudioTrack}. Use a bigger buffer to avoid choppiness
     * under high CPU load.
     * @param workerThreadHandler The external worker thread handler. If null,
     * the player will use its own dedicated worker thread.
     * @param streamType Set the stream type. Default value is
     * AudioManager.STREAM_MUSIC.
     */
    public SpeakerPlayerSink(AudioType audioType, int bufferLengthMs, NMTHandler workerThreadHandler, int streamType)
    {
        this(audioType, bufferLengthMs, workerThreadHandler, streamType, PCM_CHUNK_LENGTH_IN_MILLISECS);
    }

    /**
     * Create a player that uses the given worker thread.
     * @param audioType The audio type to play. Can't be null.
     * @param bufferLengthMs The size of the internal buffer used by the
     * internal {@link AudioTrack}. Use a bigger buffer to avoid choppiness
     * under high CPU load.
     * @param workerThreadHandler The external worker thread handler. If null,
     * the player will use its own dedicated worker thread.
     * @param streamType Set the stream type. Default value is
     * AudioManager.STREAM_MUSIC.
     * @param audioInfoDurationMs The duration to calculate audio information such as energy level, etc.
     * Default value is PCM_CHUNK_LENGTH_IN_MILLISECS.
     */
    public SpeakerPlayerSink(AudioType audioType, int bufferLengthMs, NMTHandler workerThreadHandler, int streamType, int audioInfoDurationMs)
    {
        super(audioType, workerThreadHandler);

        Checker.checkArgForCondition("bufferLengthMs", "greater than 0", bufferLengthMs > 0);

        _bufferLengthInMs = bufferLengthMs;
        _pendingPlayedAudio = new ArrayList<AudioChunk>();
        _audioInfoChunks = new LinkedList<AudioChunk>();
        _streamType = streamType;
        _mainThreadHandler = new Handler();
        _audioInfoDurationMs = audioInfoDurationMs;
    }

    /**
     * Register a listener to be notified with the TTS markers associated with the audio playback.
     * @param listener The notification listener. May be null.
     */
    public void setTtsMarkerListener(TtsMarkerListener listener)
    {
        _ttsMarkerListener = listener;
    }

    /**
     * Register a listener to be notified with the energy levels associated with the audio playback.
     * @param listener The notification listener. May be null.
     */
    public void setAudioEnergyListener(AudioEnergyListener listener)
    {
        _audioEnergyListener = listener;
    }

    @Override
    protected boolean startPlayingInternal(AudioType audioType)
    {
        Assert.assertTrue(audioType.encoding == AudioType.Encoding.PCM_16);
        _done = false;
        _noNewChunks = false;
        _totalWritten = 0;

        int bufferSizeInBytes;
        int sampleRateInHz;
        int audioFormat;
        int sampleSizeInBytes;
        final int readInterval = _bufferLengthInMs/2;
        if (audioType.frequency == AudioType.Frequency.FREQ_8KHZ)
        {
            sampleRateInHz = AudioChunk.PCM_8K_SAMPLE_RATE_IN_HZ;
            sampleSizeInBytes = AudioChunk.PCM_8K_SAMPLE_SIZE_IN_BYTES;
            audioFormat = PCM_8K_AUDIO_FORMAT;
        }
        else if (audioType.frequency == AudioType.Frequency.FREQ_11KHZ)
        {
            sampleRateInHz = AudioChunk.PCM_11K_SAMPLE_RATE_IN_HZ;
            sampleSizeInBytes = AudioChunk.PCM_11K_SAMPLE_SIZE_IN_BYTES;
            audioFormat = PCM_11K_AUDIO_FORMAT;
        }
        else if (audioType.frequency == AudioType.Frequency.FREQ_22KHZ)
        {
            sampleRateInHz = AudioChunk.PCM_22K_SAMPLE_RATE_IN_HZ;
            sampleSizeInBytes = AudioChunk.PCM_22K_SAMPLE_SIZE_IN_BYTES;
            audioFormat = PCM_22K_AUDIO_FORMAT;
        }
        else if (audioType.frequency == AudioType.Frequency.FREQ_44KHZ)
        {
            sampleRateInHz = AudioChunk.PCM_44K_SAMPLE_RATE_IN_HZ;
            sampleSizeInBytes = AudioChunk.PCM_44K_SAMPLE_SIZE_IN_BYTES;
            audioFormat = PCM_44K_AUDIO_FORMAT;
        }
        else if (audioType.frequency == AudioType.Frequency.FREQ_48KHZ)
        {
            sampleRateInHz = AudioChunk.PCM_48K_SAMPLE_RATE_IN_HZ;
            sampleSizeInBytes = AudioChunk.PCM_48K_SAMPLE_SIZE_IN_BYTES;
            audioFormat = PCM_48K_AUDIO_FORMAT;
        }
        else
        {
            // Default: 16 KHz audio
            sampleRateInHz = AudioChunk.PCM_16K_SAMPLE_RATE_IN_HZ;
            sampleSizeInBytes = AudioChunk.PCM_16K_SAMPLE_SIZE_IN_BYTES;
            audioFormat = PCM_16K_AUDIO_FORMAT;
        }

        _audioInfoSamples =  (sampleSizeInBytes/2) * sampleRateInHz * _audioInfoDurationMs / 1000;

        bufferSizeInBytes = getBufferSizeInBytes(sampleSizeInBytes, sampleRateInHz, _bufferLengthInMs);

        // adjust buffer size
        _adjustedBufferSizeInBytes  = getAdjustedBufferSizeInBytes(bufferSizeInBytes, sampleRateInHz, CHANNEL_CONFIG, audioFormat);
        _adjustedBufferSizeInShorts = _adjustedBufferSizeInBytes / sampleSizeInBytes;

        // create player
        _player = new AudioTrack(_streamType,
                                 sampleRateInHz,
                                 CHANNEL_CONFIG,
                                 audioFormat,
                                 _adjustedBufferSizeInBytes,
                                 MODE);
        if (_player.getState() != AudioTrack.STATE_INITIALIZED)
        {
            Logger.error(this, "startPlayingInternal() instantiate AudioTrack failed!!!");
            stopPlayingInternal();
            return false;
        }
        _stalledPosition = 0;
        _headMoved = false;

        _readHandler = new Handler();
        Runnable readRunnable = new Runnable()
        {
            private boolean _started = false;

            @Override
            public void run()
            {
                if (_done) return;

                _needToWrite = true;
                int written = writeBuffer();
                if (!_started && written > 0) {
                    handleStarted();
                    if (!_done)
                        _started = true;
                }

                if (!_done)
                    _readHandler.postDelayed(this, readInterval);
            }
        };
        _readHandler.post(readRunnable);

        // start player
        _player.play();

        // TODO: write off 1 buffer right after starting player otherwise OnPlaybackPositionUpdateListener won't be triggered
        _needToWrite = true;

        return true;
    }

    @Override
    protected void stopPlayingInternal()
    {
         if (_done) return;

        _done = true;

        if (_readHandler != null)
        {
            _readHandler.removeCallbacksAndMessages(null);
            _readHandler = null;
        }

        if (_player != null)
        {
            if (_player.getState() != AudioTrack.STATE_UNINITIALIZED)
            {
                Logger.debug(this, "stop _player.");
                _player.stop();
            }
            Logger.debug(this, "release _player.");
            _player.release();
            _player = null;
        }

        handleStopped();
    }

    // Called from worker thread
    @Override
    protected void newChunksInternal(List<AudioChunk> chunks)
    {
        if (chunks != null)
        {
                _bufs.addAll(chunks);
        }
        // writeBuffer();
    }

    // Called from worker thread
    @Override
    protected void noNewChunksInternal()
    {
        if (!_noNewChunks)
        {
            _noNewChunks = true;
            // Trigger writeBuffer(), to check if there's anything left to play, or
            // clean up if not.
           // writeBuffer();
        }
    }

    private static int getBufferSizeInBytes(int sampleSizeInBytes, int sampleRateInHz, int bufferLengthInMs)
    {
        // Note: We do multiplication before devision to avoid losing precision
        //       without casting to double.
        return sampleSizeInBytes * sampleRateInHz * bufferLengthInMs / 1000;
    }

    private static int getAdjustedBufferSizeInBytes(int defaultBufferSizeInBytes, int sampleRateInHz, int channelConfig, int audioFormat)
    {
        int minBufferSizeInBytes = AudioTrack.getMinBufferSize(sampleRateInHz, channelConfig, audioFormat);
        if ((minBufferSizeInBytes != AudioTrack.ERROR_BAD_VALUE) &&
            (minBufferSizeInBytes != AudioTrack.ERROR) &&
            (minBufferSizeInBytes > defaultBufferSizeInBytes))
        {
            return minBufferSizeInBytes;
        }
        else
        {
            return defaultBufferSizeInBytes;
        }
    }

    private int writeBuffer()
    {
        if (!_needToWrite)
            return 0;

        if (_headMoved)
        {
            boolean stalled = _player.getPlaybackHeadPosition() >= _totalWritten;
            if (stalled)
            {
                _headMoved = false;
                Logger.warn(this, "Player stalled");
            }
        }

        // If we're going to leave the player waiting for more data, make sure
        // we write _adjustedBufferSizeInShorts total bytes, otherwise it seems
        // like things can get screwed up.
        int toWrite = _adjustedBufferSizeInShorts;
        int written = 0;
        final LinkedList<AudioChunk> writtenChunks = new LinkedList<AudioChunk>();
        while (toWrite > 0 && !_bufs.isEmpty())
        {
            AudioChunk chunk = _bufs.getFirst();
            int len = chunk.audioShorts.length - _off;
            if (len > toWrite)
                len = toWrite;

            int count = _player.write(chunk.audioShorts, _off, len);
            if (count < 0)
            {
                // TODO: report error?
                toWrite = 0;
                _off = 0;
                _bufs.removeFirst();
            } else
            {
                if (count > 0)
                {
                    if (!_headMoved && _stalledPosition != _player.getPlaybackHeadPosition())
                    {
                        int pendingHeadPosition = _stalledPosition;

                        _headMoved = true;
                        _startTimestamp = SystemClock.uptimeMillis() - getAudioType().getDuration(_player.getPlaybackHeadPosition() - _stalledPosition);

                        for (AudioChunk pendingChunk : _pendingPlayedAudio)
                        {
                            long pendingTimestamp = _startTimestamp + getAudioType().getDuration(pendingHeadPosition);
                            AudioChunk writtenChunk = new AudioChunk(getAudioType(), pendingChunk.audioShorts, pendingTimestamp, pendingChunk.audioTtsMarkers);
                            writtenChunks.add(writtenChunk);
                            _audioInfoChunks.add(writtenChunk);
                            pendingHeadPosition += pendingChunk.audioShorts.length;
                        }
                        _pendingPlayedAudio.clear();

                        assert(pendingHeadPosition == _totalWritten);

                        Logger.warn(this, "Player unstalled");
                    }

                    if (_headMoved)
                    {
                        long timestamp = _startTimestamp + getAudioType().getDuration(_totalWritten);
                        AudioChunk writtenChunk = new AudioChunk(getAudioType(), ObjectUtils.createSubArray(chunk.audioShorts, _off, count), timestamp, chunk.audioTtsMarkers);
                        writtenChunks.add(writtenChunk);
                        _audioInfoChunks.add(writtenChunk);
                    } else
                    {
                        // We're stalled, so save the written audio to the pending buffers
                        AudioChunk writtenChunk = new AudioChunk(getAudioType(), ObjectUtils.createSubArray(chunk.audioShorts, _off, count), chunk.audioTtsMarkers);
                        _pendingPlayedAudio.add(writtenChunk);
                    }

                    written += count;
                    _totalWritten += count;
                }

                // If we wrote less than we wanted, than the buffer is full and we
                // don't need to write anymore
                if (count < len)
                    toWrite = 0;
                else
                    toWrite -= count;

                if (_off + count < chunk.audioShorts.length)
                {
                    _off += count;
                }
                else
                {
                    _bufs.removeFirst();
                    _off = 0;
                }

                AudioChunk choppedAudioChunk = null;
                do
                {
                    choppedAudioChunk = removeAudio(_audioInfoChunks, _audioInfoSamples);
                    if (choppedAudioChunk != null)
                    {
                        handleAudioInfoCallBacks(choppedAudioChunk);
                    }
                } while (!_audioInfoChunks.isEmpty() && choppedAudioChunk != null);
            }
        }

        if (!writtenChunks.isEmpty())
        {
            handleChunksPlayed(writtenChunks);
            _mainThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    if(_ttsMarkerListener != null)
                    {
                        for (AudioChunk chunk : writtenChunks)
                        {
                            if (chunk.audioTtsMarkers != null)
                            {
                                for (TtsMarker m : chunk.audioTtsMarkers)
                                    _ttsMarkerListener.onTtsMarkerAvailable(m);
                            }
                        }
                    }
                }
            });
        }

        // If we wrote all we needed to write, we don't need to write again until
        // we get the notification callback.
        if (toWrite == 0)
            _needToWrite = false;

        // If no new chunks are available, and both the player's buffer and the connected source's buffers are empty
        AudioSource<AudioChunk> audioSource = getConnectedSource();
        if (_noNewChunks && _bufs.isEmpty() && ((audioSource == null)? true : (audioSource.getChunksAvailableForSink(this) == 0)))
        {
            if (written > 0 && written < _adjustedBufferSizeInShorts)
            {
                // Write silence to make sure we get the playback notification again
                short [] filler = new short[_adjustedBufferSizeInShorts - written]; // Initialized with 0s
                _player.write(filler, 0, filler.length);
            }
            //else if (_player.getPlaybackHeadPosition() >= _totalWritten)
            else if ((_totalWritten < _adjustedBufferSizeInShorts) || (_player.getPlaybackHeadPosition() - _totalWritten >= 0)) // To handle overflow case correctly
            {
                Logger.debug(this, "writeBuffer() No more chunks. Cleanup player.");
                stopPlayingInternal();
                for (AudioChunk c : _audioInfoChunks)
                {
                    handleAudioInfoCallBacks(c);
                }
            }
        } else if (_bufs.isEmpty())
        {
            // When _bufs is empty, it's time to request more audio
            handleAudioNeeded(_bufferLengthInMs);
        }

        return written;
    }

    @Override
    protected boolean isCodecSupported(AudioType audioType) {
        if (audioType.encoding == Encoding.PCM_16)
            return true;
        return false;
    }

    /**
     * Remove a frame, and optionally return the removed audio
     * @param chunks
     * @param audioType The audio type of the audio
     * @param numSamples
     * @param newTimestamp The timestamp for the new frame
     * @return The audio
     */
    private static AudioChunk removeAudio(LinkedList<AudioChunk> chunks, AudioType audioType, final int numSamples, long newTimestamp)
    {
        short [] newAudio = new short[numSamples];
        int offset = 0;
        while (!chunks.isEmpty() && offset < numSamples)
        {
            int reqLen = numSamples - offset;

            AudioChunk removed = chunks.remove();

            short [] audio = removed.audioShorts;

            if (audio.length <= reqLen)
            {
                System.arraycopy(audio, 0, newAudio, offset, audio.length);
                offset += audio.length;
            }
            else
            {
                System.arraycopy(audio, 0, newAudio, offset, reqLen);
                offset = numSamples;

                short [] leftover = new short[audio.length - reqLen];
                System.arraycopy(audio, reqLen, leftover, 0, audio.length - reqLen);

                chunks.addFirst(new AudioChunk(removed.audioType, leftover, removed.audioTimestamp + removed.audioType.getDuration(reqLen)));
            }
        }

        if (offset < numSamples)
        {
            // Need to re-alloc smaller buffer
            short [] fixedAudio = new short[offset];
            System.arraycopy(newAudio, 0, fixedAudio, 0, fixedAudio.length);
            newAudio = fixedAudio;
            chunks.addFirst(new AudioChunk(audioType, newAudio, newTimestamp));
            return null;
        }

        return new AudioChunk(audioType, newAudio, newTimestamp);
    }

    /**
     * Remove a frame from chunks, and return the resulting AudioChunk
     * @param chunks
     * @param numSamples
     * @return The new audio frame
     */
    private static AudioChunk removeAudio(LinkedList<AudioChunk> chunks, int numSamples)
    {
        if (!chunks.isEmpty())
            return removeAudio(chunks, chunks.peek().audioType, numSamples, chunks.peek().audioTimestamp);
        else
            return null;
    }

    private void handleAudioInfoCallBacks(final AudioChunk chunk)
    {
        _mainThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (_audioEnergyListener != null)
                {
                    _audioEnergyListener.onEnergyLevelAvailable(chunk.getAudioEnergyLevel(), chunk.audioSpeechStatus == SpeechStatus.SPEECH);
                }
            }
        });
    }
}
