package com.nuance.dragon.toolkit.audio.sources;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.LinkedBlockingQueue;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.util.Mp3HeaderParser;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTAssetFileDescriptor;
import com.nuance.dragon.toolkit.oem.api.NMTContext;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * The file source which will load the audio from file in one shot,
 * create audio chunks with the length of 400ms, save the audio chunks
 * into a queue and provide them upon pulling.
 */
public class BurstFileRecorderSource extends RecorderSource<AudioChunk>
{
    // initial parameters
    private static final int PCM_8K_BUFFER_LENGTH_IN_MILLISECS  = 400;
    private static final int PCM_8K_BUFFER_SIZE_IN_BYTES        = AudioChunk.PCM_8K_SAMPLE_SIZE_IN_BYTES *
                                                                  AudioChunk.PCM_8K_SAMPLE_RATE_IN_HZ *
                                                                  PCM_8K_BUFFER_LENGTH_IN_MILLISECS /
                                                                  1000; // yes we do multiplication before devision to avoid loosing precision without casting to double

    private static final int PCM_11K_BUFFER_LENGTH_IN_MILLISECS = 400;
    private static final int PCM_11K_BUFFER_SIZE_IN_BYTES       = AudioChunk.PCM_11K_SAMPLE_SIZE_IN_BYTES *
                                                                  AudioChunk.PCM_11K_SAMPLE_RATE_IN_HZ *
                                                                  PCM_11K_BUFFER_LENGTH_IN_MILLISECS /
                                                                  1000; // yes we do multiplication before devision to avoid loosing precision without casting to double

    private static final int PCM_16K_BUFFER_LENGTH_IN_MILLISECS = 400;
    private static final int PCM_16K_BUFFER_SIZE_IN_BYTES       = AudioChunk.PCM_16K_SAMPLE_SIZE_IN_BYTES *
                                                                  AudioChunk.PCM_16K_SAMPLE_RATE_IN_HZ *
                                                                  PCM_16K_BUFFER_LENGTH_IN_MILLISECS /
                                                                  1000; // yes we do multiplication before devision to avoid loosing precision without casting to double

    private static final int PCM_22K_BUFFER_LENGTH_IN_MILLISECS = 400;
    private static final int PCM_22K_BUFFER_SIZE_IN_BYTES       = AudioChunk.PCM_22K_SAMPLE_SIZE_IN_BYTES *
                                                                  AudioChunk.PCM_22K_SAMPLE_RATE_IN_HZ *
                                                                  PCM_22K_BUFFER_LENGTH_IN_MILLISECS /
                                                                  1000; // yes we do multiplication before devision to avoid loosing precision without casting to double

    private static final int PCM_44K_BUFFER_LENGTH_IN_MILLISECS = 400;
    private static final int PCM_44K_BUFFER_SIZE_IN_BYTES       = AudioChunk.PCM_44K_SAMPLE_SIZE_IN_BYTES *
                                                                  AudioChunk.PCM_44K_SAMPLE_RATE_IN_HZ *
                                                                  PCM_44K_BUFFER_LENGTH_IN_MILLISECS /
                                                                  1000; // yes we do multiplication before devision to avoid loosing precision without casting to double
    private static final int PCM_48K_BUFFER_LENGTH_IN_MILLISECS = 400;
    private static final int PCM_48K_BUFFER_SIZE_IN_BYTES       = AudioChunk.PCM_48K_SAMPLE_SIZE_IN_BYTES *
                                                                  AudioChunk.PCM_48K_SAMPLE_RATE_IN_HZ *
                                                                  PCM_48K_BUFFER_LENGTH_IN_MILLISECS /
                                                                  1000; // yes we do multiplication before devision to avoid loosing precision without casting to double

    private static final int ULAW_BUFFER_LENGTH_IN_MILLISECS    = 400;
    private static final int ULAW_BUFFER_SIZE_IN_BYTES          = AudioChunk.ULAW_SAMPLE_SIZE_IN_BYTES *
                                                                  AudioChunk.ULAW_SAMPLE_RATE_IN_HZ *
                                                                  ULAW_BUFFER_LENGTH_IN_MILLISECS /
                                                                  1000; // yes we do multiplication before devision to avoid loosing precision without casting to double

    private static final int GSM_FR_BUFFER_LENGTH_IN_MILLISECS  = 400;
    private static final int GSM_FR_BUFFER_SIZE_IN_BYTES        = GSM_FR_BUFFER_LENGTH_IN_MILLISECS *
                                                                  AudioChunk.GSM_FR_DOUBLE_FRAME_SIZE_IN_BYTES /
                                                                  AudioChunk.GSM_FR_DOUBLE_FRAME_LENGTH_IN_MILLISECONDS; // yes we do multiplication before devision to avoid loosing precision without casting to double

    private static final int AMR0_BUFFER_LENGTH_IN_MILLISECS    = 400;
    private static final int AMR0_BUFFER_SIZE_IN_BYTES          = AMR0_BUFFER_LENGTH_IN_MILLISECS *
                                                                  AudioChunk.AMR0_FRAME_SIZE_IN_BYTES /
                                                                  AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS; // yes we do multiplication before devision to avoid loosing precision without casting to double
    private static final int AMR1_BUFFER_LENGTH_IN_MILLISECS    = 400;
    private static final int AMR1_BUFFER_SIZE_IN_BYTES          = AMR0_BUFFER_LENGTH_IN_MILLISECS *
                                                                  AudioChunk.AMR1_FRAME_SIZE_IN_BYTES /
                                                                  AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS; // yes we do multiplication before devision to avoid loosing precision without casting to double
    private static final int AMR2_BUFFER_LENGTH_IN_MILLISECS    = 400;
    private static final int AMR2_BUFFER_SIZE_IN_BYTES          = AMR0_BUFFER_LENGTH_IN_MILLISECS *
                                                                  AudioChunk.AMR2_FRAME_SIZE_IN_BYTES /
                                                                  AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS; // yes we do multiplication before devision to avoid loosing precision without casting to double
    private static final int AMR3_BUFFER_LENGTH_IN_MILLISECS    = 400;
    private static final int AMR3_BUFFER_SIZE_IN_BYTES          = AMR0_BUFFER_LENGTH_IN_MILLISECS *
                                                                  AudioChunk.AMR3_FRAME_SIZE_IN_BYTES /
                                                                  AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS; // yes we do multiplication before devision to avoid loosing precision without casting to double
    private static final int AMR4_BUFFER_LENGTH_IN_MILLISECS    = 400;
    private static final int AMR4_BUFFER_SIZE_IN_BYTES          = AMR0_BUFFER_LENGTH_IN_MILLISECS *
                                                                  AudioChunk.AMR4_FRAME_SIZE_IN_BYTES /
                                                                  AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS; // yes we do multiplication before devision to avoid loosing precision without casting to double
    private static final int AMR5_BUFFER_LENGTH_IN_MILLISECS    = 400;
    private static final int AMR5_BUFFER_SIZE_IN_BYTES          = AMR0_BUFFER_LENGTH_IN_MILLISECS *
                                                                  AudioChunk.AMR5_FRAME_SIZE_IN_BYTES /
                                                                  AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS; // yes we do multiplication before devision to avoid loosing precision without casting to double
    private static final int AMR6_BUFFER_LENGTH_IN_MILLISECS    = 400;
    private static final int AMR6_BUFFER_SIZE_IN_BYTES          = AMR0_BUFFER_LENGTH_IN_MILLISECS *
                                                                  AudioChunk.AMR6_FRAME_SIZE_IN_BYTES /
                                                                  AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS; // yes we do multiplication before devision to avoid loosing precision without casting to double
    private static final int AMR7_BUFFER_LENGTH_IN_MILLISECS    = 400;
    private static final int AMR7_BUFFER_SIZE_IN_BYTES          = AMR0_BUFFER_LENGTH_IN_MILLISECS *
                                                                  AudioChunk.AMR7_FRAME_SIZE_IN_BYTES /
                                                                  AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS; // yes we do multiplication before devision to avoid loosing precision without casting to double

    private static final int MP3_128KBPS_BUFFER_LENGTH_IN_MILLISECS = 400;
    private static final int MP3_128KBPS_BUFFER_SIZE_IN_BYTES       = AudioChunk.MP3_128KBPS_BIT_RATE *
                                                                      MP3_128KBPS_BUFFER_LENGTH_IN_MILLISECS /
                                                                      8 /
                                                                      1000; // yes we do multiplication before devision to avoid loosing precision without casting to double

    private String _fileName;
    private final int _resourceId;
    private NMTAssetFileDescriptor _assetFileDesc;
    private final NMTContext _context;
    private boolean _littleEndian;
    private InputStream _inputStream = null;
    private boolean _done;

    private LinkedBlockingQueue<AudioChunk> _audioChunkQueue;
    private AudioType _audioType;

    private final FileManager _fileManager;
    
    private Mp3HeaderParser _mp3Parser;

    /**
     * Create a file source.
     *
     * @param audioType The audio type of the file to load. Generally, it is possible
     * that only PCM 22khz, 16Khz, PCM8Khz will be supported.
     * @param fileName The file to read the audio from
     * @param fileManager is used to access file
     * @param littleEndian True to save in little-endian format; false for big-
     * endian.
     */
    public BurstFileRecorderSource(AudioType audioType, String fileName, FileManager fileManager, boolean littleEndian)
    {
        this(audioType, fileName, fileManager, littleEndian, null);
    }

    /**
     * Create a file source.
     *
     * @param audioType The audio type of the file to load. Generally, it is possible
     * that only PCM 22khz, 16Khz, PCM8Khz will be supported.
     * @param fileName The file to read the audio from
     * @param fileManager is used to access file
     * @param littleEndian True to read in little-endian format; false for big-
     * endian.
     */
    public BurstFileRecorderSource(AudioType audioType, String fileName, FileManager fileManager, boolean littleEndian, NMTHandler mainThreadHandler)
    {
        super(audioType, null, mainThreadHandler);
        Checker.checkArgForNull("audioType", audioType);
        //Checker.checkArgForNull("fileName", fileName);

        _audioType = audioType;
        _fileName = fileName;
        _littleEndian = littleEndian;
        _resourceId = -1;
        _context = null;
        _fileManager = fileManager;
        _audioChunkQueue = new LinkedBlockingQueue<AudioChunk>();
        _done = false;
    }

    /**
     * Create a file source.
     *
     * @param audioType The audio type of the file to load. Generally, it is possible
     * that only PCM 22khz, 16Khz, PCM8Khz will be supported.
     * @param fileName The file to read the audio from
     * @param littleEndian True to read in little-endian format; false for big-
     * endian.
     */
    public BurstFileRecorderSource(AudioType audioType, String fileName, boolean littleEndian)
    {
        this(audioType, fileName, null, littleEndian);
    }

    /**
     * Create a file source that reads the audio in little-endian format.
     * @param audioType The audio type of the file to load. Generally, it is possible
     * that only PCM 22khz, 16Khz, PCM8Khz will be supported.
     * @param fileName The file to read the audio from
     */
    public BurstFileRecorderSource(AudioType audioType, String fileName)
    {
        this(audioType, fileName, true);
    }

    /**
     * Create a stream source that reads the audio in little-endian format.
     * @param audioType The audio type of the file to load. Generally, it is possible
     * that only PCM 22khz, 16Khz, PCM8Khz will be supported.
     * @param inputStream The input stream to read the audio from.
     */
    public BurstFileRecorderSource(AudioType audioType, InputStream inputStream)
    {
        this(audioType, inputStream, true);
    }

    /**
     * Create a stream source.
     *
     * @param audioType The audio type of the file to load. Generally, it is possible
     * that only PCM 22khz, 16Khz, PCM8Khz will be supported.
     * @param inputStream The input stream to read the audio from.
     * @param littleEndian True to read in little-endian format; false for big-
     * endian.
     */
    public BurstFileRecorderSource(AudioType audioType, InputStream inputStream, boolean littleEndian)
    {
        this(audioType, (String)null, littleEndian);
        _inputStream = inputStream;
    }

    /**
     * Create a file source with a given resource file Id, the file should
     * be deployed in the "res/raw" folder of the application.
     *
     * @param audioType The audio type of the file to load. Generally, it is possible
     * that only PCM 22khz, 16Khz, PCM8Khz will be supported.
     * @param resourceId The resource id of the file, normally the file should be
     * in the res/raw folder of the application
     * @param context The application context object which will be used to load the file
     */
    public BurstFileRecorderSource(AudioType audioType, int resourceId, NMTContext context)
    {
        super(audioType, null, null);
        Checker.checkArgForNull("audioType", audioType);
        Checker.checkArgForCondition("audioType", "a type supported by this player", isCodecSupported(audioType));
        Checker.checkArgForNull("NMTContext", context);

        _audioType = audioType;
        _resourceId = resourceId;
        _context = context;
        _fileName = null;
        _littleEndian = true;
        _audioChunkQueue = new LinkedBlockingQueue<AudioChunk>();
        _done = false;
        _fileManager = null;
    }

    @Override
    protected AudioChunk getAudioChunk()
    {
        if (_audioChunkQueue.size() > 0)
            return _audioChunkQueue.remove();
        else
            return null;
    }

    @Override
    public AudioType getAudioType()
    {
        return _audioType;
    }

    @Override
    public boolean isActive()
    {
        return !_done || (_audioChunkQueue.size() > 0);
    }

    @Override
    public int getChunksAvailable()
    {
        return _audioChunkQueue.size();
    }

    private void prepareAudioChunks()
    {
        byte[] buf = null;
        int bufIdx = 0;

        if (_inputStream == null) {
            try
            {
                if (_fileName != null)
                {
                    boolean isAbsolute = new File(_fileName).isAbsolute();
                    if (_fileManager != null && !isAbsolute)
                    {
                        _inputStream = _fileManager.openFileForReading(_fileName);
                    }
                    else
                        _inputStream = new FileInputStream(new File(_fileName));
                }
                else if (_resourceId != -1 && _context != null)
                {
                    _assetFileDesc = _context.openRawResourceFd(_resourceId);
                    if (_assetFileDesc == null)
                    {
                        throw new IllegalArgumentException("resourceId must refer to an uncompressed resource");
                    }
                    _inputStream = _assetFileDesc.createInputStream();
                }
                else
                {
                    Logger.error(this,  "prepareAudioChunks() FileInputStream has not been created!!!");
                    return;
                }
            }
            catch (FileNotFoundException e)
            {
                Logger.error(this,  "prepareAudioChunks() FileNotFoundException!!!");
                return;
            }
        }
        // file stream returned by FileManager could be null
        if (_inputStream == null)
            return;

        final int bufferSizeInBytes;
        final int bufferLengthInMillisecs;
        if (_audioType.encoding == AudioType.Encoding.PCM_16)
        {
            if (_audioType.frequency == AudioType.Frequency.FREQ_8KHZ)
            {
                bufferSizeInBytes = PCM_8K_BUFFER_SIZE_IN_BYTES;
                bufferLengthInMillisecs = PCM_8K_BUFFER_LENGTH_IN_MILLISECS;
            }
            else if (_audioType.frequency == AudioType.Frequency.FREQ_11KHZ)
            {
                bufferSizeInBytes = PCM_11K_BUFFER_SIZE_IN_BYTES;
                bufferLengthInMillisecs = PCM_11K_BUFFER_LENGTH_IN_MILLISECS;
            }
            else if (_audioType.frequency == AudioType.Frequency.FREQ_16KHZ)
            {
                bufferSizeInBytes = PCM_16K_BUFFER_SIZE_IN_BYTES;
                bufferLengthInMillisecs = PCM_16K_BUFFER_LENGTH_IN_MILLISECS;
            }
            else if (_audioType.frequency == AudioType.Frequency.FREQ_22KHZ)
            {
                bufferSizeInBytes = PCM_22K_BUFFER_SIZE_IN_BYTES;
                bufferLengthInMillisecs = PCM_22K_BUFFER_LENGTH_IN_MILLISECS;
            }
            else if (_audioType.frequency == AudioType.Frequency.FREQ_44KHZ)
            {
                bufferSizeInBytes = PCM_44K_BUFFER_SIZE_IN_BYTES;
                bufferLengthInMillisecs = PCM_44K_BUFFER_LENGTH_IN_MILLISECS;
            }
            else if (_audioType.frequency == AudioType.Frequency.FREQ_48KHZ)
            {
                bufferSizeInBytes = PCM_48K_BUFFER_SIZE_IN_BYTES;
                bufferLengthInMillisecs = PCM_48K_BUFFER_LENGTH_IN_MILLISECS;
            }
            else
            {
                bufferSizeInBytes = -1;
                bufferLengthInMillisecs = 400; // we assume the buffer len is 400ms, so it will be enough to start playback
            }
        }
        else if (_audioType.encoding == AudioType.Encoding.ULAW)
        {
            bufferSizeInBytes = ULAW_BUFFER_SIZE_IN_BYTES;
            bufferLengthInMillisecs = ULAW_BUFFER_LENGTH_IN_MILLISECS;
        }
        else if (_audioType.encoding == AudioType.Encoding.GSM_FR)
        {
            bufferSizeInBytes = GSM_FR_BUFFER_SIZE_IN_BYTES;
            bufferLengthInMillisecs = GSM_FR_BUFFER_LENGTH_IN_MILLISECS;
        }
        else if (_audioType.encoding == AudioType.Encoding.AMR0)
        {
            bufferSizeInBytes = AMR0_BUFFER_SIZE_IN_BYTES;
            bufferLengthInMillisecs = AMR0_BUFFER_LENGTH_IN_MILLISECS;
        }
        else if (_audioType.encoding == AudioType.Encoding.AMR1)
        {
            bufferSizeInBytes = AMR1_BUFFER_SIZE_IN_BYTES;
            bufferLengthInMillisecs = AMR1_BUFFER_LENGTH_IN_MILLISECS;
        }
        else if (_audioType.encoding == AudioType.Encoding.AMR2)
        {
            bufferSizeInBytes = AMR2_BUFFER_SIZE_IN_BYTES;
            bufferLengthInMillisecs = AMR2_BUFFER_LENGTH_IN_MILLISECS;
        }
        else if (_audioType.encoding == AudioType.Encoding.AMR3)
        {
            bufferSizeInBytes = AMR3_BUFFER_SIZE_IN_BYTES;
            bufferLengthInMillisecs = AMR3_BUFFER_LENGTH_IN_MILLISECS;
        }
        else if (_audioType.encoding == AudioType.Encoding.AMR4)
        {
            bufferSizeInBytes = AMR4_BUFFER_SIZE_IN_BYTES;
            bufferLengthInMillisecs = AMR4_BUFFER_LENGTH_IN_MILLISECS;
        }
        else if (_audioType.encoding == AudioType.Encoding.AMR5)
        {
            bufferSizeInBytes = AMR5_BUFFER_SIZE_IN_BYTES;
            bufferLengthInMillisecs = AMR5_BUFFER_LENGTH_IN_MILLISECS;
        }
        else if (_audioType.encoding == AudioType.Encoding.AMR6)
        {
            bufferSizeInBytes = AMR6_BUFFER_SIZE_IN_BYTES;
            bufferLengthInMillisecs = AMR6_BUFFER_LENGTH_IN_MILLISECS;
        }
        else if (_audioType.encoding == AudioType.Encoding.AMR7)
        {
            bufferSizeInBytes = AMR7_BUFFER_SIZE_IN_BYTES;
            bufferLengthInMillisecs = AMR7_BUFFER_LENGTH_IN_MILLISECS;
        }
        else if (_audioType.encoding == AudioType.Encoding.MP3_128KBPS)
        {
			if (_mp3Parser == null) {
				_mp3Parser = new Mp3HeaderParser(_inputStream);
				_mp3Parser.stripID3v2Tag();
			}
			bufferSizeInBytes = _mp3Parser.getNextBufferSize();
			bufferLengthInMillisecs = 0;
        }
        else
        {
            bufferSizeInBytes = -1;
            bufferLengthInMillisecs = 400; // we assume the buffer len is 400ms, so it will be enough to start playback
        }


        if (bufferSizeInBytes > 0)
            buf = new byte[bufferSizeInBytes];


        boolean allDone = false;
        while (!allDone)
        {
            int bytesRead;
            if (_audioType.encoding == AudioType.Encoding.SPEEX || _audioType.encoding == AudioType.Encoding.OPUS)
            {
                // read Nuance speex header
                int speexBufferLen = 0;
                byte[] nextByte = new byte[1];
                do
                {
                    try
                    {
                        bytesRead = _inputStream.read(nextByte);
                    } catch (IOException ioe)
                    {
                        Logger.error(this,  "prepareAudioChunks() reading header _fis.read() threw " + ioe + "!!!");
                        bytesRead = 0;
                    }
                    if (bytesRead != nextByte.length)
                    {
                        Logger.debug(this, "allDone 1");
                        cleanup();
                        return;
                    }
                    speexBufferLen = (speexBufferLen << 7) | nextByte[0] & 0x7F;
                } while((nextByte[0] & 0x80) != 0);

                if (speexBufferLen <= 0) {
                    Logger.error(this,  "prepareAudioChunks() Read file length error: speexBufferLen =" + speexBufferLen);

                    cleanup();
                    return;
                }

                buf = new byte[speexBufferLen];
            }
            if(_audioType.encoding == AudioType.Encoding.MP3_128KBPS)
            {
            	int bufferSize = _mp3Parser.getNextBufferSize();
            	bytesRead = -1;
            	if(bufferSize > 0){
            		// Reallocate buffer only if needed
            		if(buf.length != bufferSize){
            			buf = new byte[bufferSize];
            		}
					bytesRead = _mp3Parser.fillBuffer(buf);
            	}
            }else{
	            try
	            {
	                bytesRead = _inputStream.read(buf, 0, buf.length);
	            } catch (IOException ioe)
	            {
	                Logger.error(this,  "prepareAudioChunks() _fis.read() threw " + ioe + "!!!");
	                bytesRead = 0;
	            }
            }

            if (bytesRead == -1)
            {
                allDone = true;
            }
            else if (bytesRead != buf.length)
            {
                 // this piece of code will do silence padding
                 // we will not do that for now
//                 if (audioType.encoding == AudioType.Encoding.PCM_16)
//                 {
//                     // pcm auto padding
//                     Arrays.fill(_buf, bytesRead, _buf.length -1, (byte)0);
//                     bytesRead = _buf.length;
//                 }
            }

            if (bytesRead > 0)
            {
                AudioChunk audioChunk;
                if (_audioType.encoding == AudioType.Encoding.PCM_16)
                {
                    short[] shorts = new short[bytesRead / 2];
                    for (int i = 0; i < bytesRead - 1; i += 2)
                    {
                        if (_littleEndian)
                        {
                            shorts[i / 2] = (short)((((buf[i + 1] << 8) & 0xFF00) | (buf[i] & 0xFF)));
                        }
                        else
                        {
                            shorts[i / 2] = (short)((((buf[i] << 8) & 0xFF00) | (buf[i + 1] & 0xFF)));
                        }
                    }
                    audioChunk = new AudioChunk(_audioType, shorts, bufferLengthInMillisecs * bufIdx++);
                }
                else
                {
                    byte[] bufCopy = new byte[bytesRead];
                    System.arraycopy(buf, 0, bufCopy, 0, bytesRead);
                    audioChunk = new AudioChunk(_audioType, bufCopy, 100);
                }

                _audioChunkQueue.add(audioChunk);
            }
            if (allDone)
            {
                cleanup();
            }
        }

    }

    private void cleanup()
    {
        if (_inputStream != null)
        {
            try
            {
                _inputStream.close();
            } catch (IOException ioe) {}
        }

        if (_assetFileDesc != null)
        {
            _assetFileDesc.close();
        }

        _done = true;
    }

    @Override
    protected boolean isCodecSupported(AudioType audioType) {
        // All are supported
        return true;
    }

    @Override
    protected boolean startRecordingInternal(AudioType audioType)
    {
        handleStarted();
        prepareAudioChunks();
        //use this to notify new chunks available on main thread
        handleNewAudio(null);
        handleSourceClosed();
        return true;
    }

    @Override
    protected void stopRecordingInternal()
    {
        //nothing to do, no real recording going on
    }

    @Override
    protected void onSinkConnected(AudioSink<AudioChunk> sink)
    {
        super.onSinkConnected(sink);
        if (!isEmpty())
        {
            sink.chunksAvailable(this);
        }
    }
}
