package com.nuance.dragon.toolkit.audio.sources;

import junit.framework.Assert;
import android.media.AudioFormat;
import android.media.AudioRecord;
import android.media.AudioRecord.OnRecordPositionUpdateListener;
import android.os.Handler;
import android.os.SystemClock;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk;
import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.AudioType.Encoding;
import com.nuance.dragon.toolkit.audio.MultiChannelAudioChunk;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.util.internal.ObjectUtils;

public abstract class MicrophoneRecorderSourceBase<AudioChunkType extends AbstractAudioChunk> extends RecorderSource<AudioChunkType>
{
    // initial parameters
    private static final int DEFAULT_AUDIO_SOURCE       = android.media.MediaRecorder.AudioSource.VOICE_RECOGNITION;

    // default buffer lengths
    private static final int PCM_BUFFER_LENGTH_IN_MILLISECS = 400;
    private static final int PCM_CHUNK_LENGTH_IN_MILLISECS = 20;

    private static final int PCM_8K_AUDIO_FORMAT  = AudioFormat.ENCODING_PCM_16BIT;
    private static final int PCM_11K_AUDIO_FORMAT = AudioFormat.ENCODING_PCM_16BIT;
    private static final int PCM_16K_AUDIO_FORMAT = AudioFormat.ENCODING_PCM_16BIT;
    private static final int PCM_22K_AUDIO_FORMAT = AudioFormat.ENCODING_PCM_16BIT;
    private static final int PCM_44K_AUDIO_FORMAT = AudioFormat.ENCODING_PCM_16BIT;
    private static final int PCM_48K_AUDIO_FORMAT = AudioFormat.ENCODING_PCM_16BIT;

    private AudioType _audioType;
    private int _sourceType;
    private AudioRecord _recorder;
    private boolean _done;
    private int _readBufferSize;
    private int _totalRead;
    private long _startTimestamp;
    private final int _bufferLengthInMs, _chunkLengthInMs;
    private Handler _fallbackReadHandler;
    private Runnable _fallBackReadRunnable;

    // override by subclass
    protected int channelConfig = AudioFormat.CHANNEL_IN_MONO;
    protected int channelCount = 1;

    /**
     * Create a recorder that uses its own dedicated worker thread.
     * This is a shortcut constructor that uses default values of:
     * sourceType = MediaRecorder.AudioSource.VOICE_RECOGNITION
     * defaultBufferLengthInMs = 400
     * workerThreadHandler = null
     * @param audioType The audio type to record.
     */
    protected MicrophoneRecorderSourceBase(AudioType audioType)
    {
        this(audioType, PCM_BUFFER_LENGTH_IN_MILLISECS, null);
    }

    /**
     * Create a recorder that uses its own dedicated worker thread.
     * This is a shortcut constructor that uses default values of:
     * defaultBufferLengthInMs = 400
     * workerThreadHandler = null
     * @param sourceType The source type to use. Default value is
     * MediaRecorder.AudioSource.VOICE_RECOGNITION.
     * @param audioType The audio type to record.
     */
    protected MicrophoneRecorderSourceBase(int sourceType, AudioType audioType)
    {
        this(sourceType, audioType, PCM_BUFFER_LENGTH_IN_MILLISECS, null);
    }

    /**
     * Create a recorder that uses the given worker thread.
     * This is a shortcut constructor that uses default values of:
     * sourceType = MediaRecorder.AudioSource.VOICE_RECOGNITION
     * defaultBufferLengthInMs = 400
     * @param audioType The audio type to record.
     * @param workerThreadHandler The external worker thread handler. If null,
     * the recorder will use its own dedicated worker thread. Default value is null.
     */
    protected MicrophoneRecorderSourceBase(AudioType audioType, NMTHandler workerThreadHandler)
    {
        this(audioType, PCM_BUFFER_LENGTH_IN_MILLISECS, workerThreadHandler);
    }

    /**
     * Create a recorder that uses the given worker thread.
     * This is a shortcut constructor that uses default values of:
     * sourceType = MediaRecorder.AudioSource.VOICE_RECOGNITION
     * @param audioType The audio type to record.
     * @param bufferLengthInMs The total size (in ms) of the buffer where
     * audio data is written to during the recording. Setting it to a higher value
     * can help record audio smoothly under load. Default value is 400.
     * @param workerThreadHandler The external worker thread handler. If null,
     * the recorder will use its own dedicated worker thread. Default value is null.
     */
    protected MicrophoneRecorderSourceBase(AudioType audioType,  int bufferLengthInMs, NMTHandler workerThreadHandler)
    {
        this(DEFAULT_AUDIO_SOURCE, audioType, bufferLengthInMs, workerThreadHandler);
    }

    /**
     * Create a recorder that uses the given worker thread.
     * Other constructors are supplied that leave out some of these
     * arguments and use a default value instead.
     * @param sourceType The source type to use. Default value is
     * MediaRecorder.AudioSource.VOICE_RECOGNITION.
     * @param audioType The audio type to record.
     * @param bufferLengthInMs The total size (in ms) of the buffer where
     * audio data is written to during the recording. Setting it to a higher value
     * can help record audio smoothly under load. Default value is 400.
     * @param workerThreadHandler The external worker thread handler. If null,
     * the recorder will use its own dedicated worker thread. Default value is null.
     */
    protected MicrophoneRecorderSourceBase(int sourceType, AudioType audioType,  int bufferLengthInMs, NMTHandler workerThreadHandler)
    {
        this(sourceType, audioType, bufferLengthInMs, PCM_CHUNK_LENGTH_IN_MILLISECS, workerThreadHandler);
    }

    /**
     * Create a recorder that uses the given worker thread.
     * Other constructors are supplied that leave out some of these
     * arguments and use a default value instead.
     * @param sourceType The source type to use. Default value is
     * MediaRecorder.AudioSource.VOICE_RECOGNITION.
     * @param audioType The audio type to record.
     * @param bufferLengthInMs The total size (in ms) of the buffer where
     * audio data is written to during the recording. Setting it to a higher value
     * can help record audio smoothly under load. Default value is 400.
     * @param chunkLengthInMs The size (in ms) of the chunks that
     * are propagated to the rest of the system during the recording. Default value is 20.
     * @param workerThreadHandler The external worker thread handler. If null,
     * the recorder will use its own dedicated worker thread. Default value is null.
     */
    protected MicrophoneRecorderSourceBase(int sourceType, AudioType audioType, int bufferLengthInMs, int chunkLengthInMs, NMTHandler workerThreadHandler)
    {
        super(audioType, workerThreadHandler);

        Checker.checkArgForCondition("bufferLengthInMs", "stricly greater than 0", bufferLengthInMs > 0);
        _sourceType = sourceType;
        _bufferLengthInMs = bufferLengthInMs;
        _chunkLengthInMs = chunkLengthInMs;
    }

    @Override
    protected boolean startRecordingInternal(final AudioType audioType)
    {
        Assert.assertTrue(audioType.encoding == AudioType.Encoding.PCM_16);

        _done = false;
        _audioType = audioType;
        _totalRead = 0;

        int bufferSizeInBytes;
        int sampleRateInHz;
        int audioFormat;
        int sampleSizeInBytes;
        if (audioType.frequency == AudioType.Frequency.FREQ_8KHZ)
        {
            sampleRateInHz = AudioChunk.PCM_8K_SAMPLE_RATE_IN_HZ;
            audioFormat = PCM_8K_AUDIO_FORMAT;
            sampleSizeInBytes = AudioChunk.PCM_8K_SAMPLE_SIZE_IN_BYTES;
        }
        else if (audioType.frequency == AudioType.Frequency.FREQ_11KHZ)
        {
            sampleRateInHz = AudioChunk.PCM_11K_SAMPLE_RATE_IN_HZ;
            audioFormat = PCM_11K_AUDIO_FORMAT;
            sampleSizeInBytes = AudioChunk.PCM_11K_SAMPLE_SIZE_IN_BYTES;
        }
        else if (audioType.frequency == AudioType.Frequency.FREQ_22KHZ)
        {
            sampleRateInHz = AudioChunk.PCM_22K_SAMPLE_RATE_IN_HZ;
            audioFormat = PCM_22K_AUDIO_FORMAT;
            sampleSizeInBytes = AudioChunk.PCM_22K_SAMPLE_SIZE_IN_BYTES;
        }
        else if (audioType.frequency == AudioType.Frequency.FREQ_44KHZ)
        {
            sampleRateInHz = AudioChunk.PCM_44K_SAMPLE_RATE_IN_HZ;
            audioFormat = PCM_44K_AUDIO_FORMAT;
            sampleSizeInBytes = AudioChunk.PCM_44K_SAMPLE_SIZE_IN_BYTES;
        }
        else if (audioType.frequency == AudioType.Frequency.FREQ_48KHZ)
        {
            sampleRateInHz = AudioChunk.PCM_48K_SAMPLE_RATE_IN_HZ;
            audioFormat = PCM_48K_AUDIO_FORMAT;
            sampleSizeInBytes = AudioChunk.PCM_48K_SAMPLE_SIZE_IN_BYTES;
        }
        else
        {
            // Default: 16 KHz audio
            sampleRateInHz = AudioChunk.PCM_16K_SAMPLE_RATE_IN_HZ;
            audioFormat = PCM_16K_AUDIO_FORMAT;
            sampleSizeInBytes = AudioChunk.PCM_16K_SAMPLE_SIZE_IN_BYTES;
        }

        bufferSizeInBytes = getBufferSizeInBytes(sampleSizeInBytes, sampleRateInHz, _bufferLengthInMs);

        // Compute buffer size
        _readBufferSize = sampleRateInHz * _chunkLengthInMs * channelCount / 1000;

        // create recorder
        _recorder = new AudioRecord(_sourceType,
                                    sampleRateInHz,
                                    channelConfig,
                                    audioFormat,
                                    getAdjustedBufferSizeInBytes(bufferSizeInBytes, sampleRateInHz, channelConfig, audioFormat));
        if (_recorder.getState() != AudioRecord.STATE_INITIALIZED)
        {
            Logger.error(this, "startRecordingInternal() instantiate AudioRecord failed!!!");
            cleanup();
            return false;
        }

        // Safety code in case Android audio system fails to notify us when audio buffer is filled.
        // This will trigger a read (see listener below, might happen on certain devices).
        // TODO: (LXU) investigate marker position doesn't impact onMarkerReached call's frequency
        final int notificationMarkerPosition = _readBufferSize/2;
        final int fallbackReadPeriod = _chunkLengthInMs / 2;
        _fallbackReadHandler = new Handler();
        _fallBackReadRunnable = new Runnable()
        {
            @Override
            public void run()
            {
                if (_done) return;

                Logger.verbose(this, "Android audio system notifications stopped, execute periodic read.");
                _recorder.setNotificationMarkerPosition(notificationMarkerPosition);
                AudioChunkType buf = readBuffer();
                if (buf != null) {
                    handleNewAudio(buf);
                }
                _fallbackReadHandler.postDelayed(this, fallbackReadPeriod);
            }
        };

        OnRecordPositionUpdateListener readListener = new OnRecordPositionUpdateListener()
        {
            private boolean _started = false;

            @Override
            public void onMarkerReached(AudioRecord recorder)
            {
                if (_done) return;

                Logger.verbose(this, "Android audio system onMarkerReached invoked");
                _fallbackReadHandler.removeCallbacks(_fallBackReadRunnable);
                recorder.setNotificationMarkerPosition(notificationMarkerPosition);
                AudioChunkType buf = readBuffer();
                if (buf != null) {

                    if (!_started) {
                        _started = true;
                        handleStarted();
                    }

                    handleNewAudio(buf);
                }
                _fallbackReadHandler.postDelayed(_fallBackReadRunnable, fallbackReadPeriod);
            }

            @Override
            public void onPeriodicNotification(AudioRecord recorder)
            {
                // nothing to do, we use onMarkerReached() instead
            }
        };
        _recorder.setRecordPositionUpdateListener(readListener);

        // start recorder and trigger first read
        _recorder.startRecording();
        readListener.onMarkerReached(_recorder);
        return true;
    }

    @Override
    protected void stopRecordingInternal()
    {
        if (_done) return;

        _fallbackReadHandler.removeCallbacks(_fallBackReadRunnable);

        // Read the last bit of audio
        AudioChunkType buf = readBuffer();
        if (buf != null)
            handleNewAudio(buf);

        cleanup();

        handleSourceClosed();
    }

    private int getBufferSizeInBytes(int sampleSizeInBytes, int sampleRateInHz, int bufferLengthInMs)
    {
        // Note: We do multiplication before devision to avoid losing precision
        //       without casting to double.
        return sampleSizeInBytes * sampleRateInHz * bufferLengthInMs * channelCount / 1000;
    }

    private int getAdjustedBufferSizeInBytes(int defaultBufferSizeInBytes, int sampleRateInHz, int channelConfig, int audioFormat) {
        int minBufferSizeInBytes = AudioRecord.getMinBufferSize(sampleRateInHz, channelConfig, audioFormat);
        if ((minBufferSizeInBytes != AudioRecord.ERROR_BAD_VALUE) &&
            (minBufferSizeInBytes != AudioRecord.ERROR) &&
            (minBufferSizeInBytes > defaultBufferSizeInBytes))
        {
            return minBufferSizeInBytes;
        }
        else
        {
            return defaultBufferSizeInBytes;
        }
    }

    private void cleanup()
    {
        _done = true;

        if (_recorder != null)
        {
            if (_recorder.getState() == AudioRecord.STATE_INITIALIZED)
            {
                Logger.debug(this, "cleanup() stop _recorder.");
                _recorder.stop();
            }
            Logger.debug(this, "cleanup() release _recorder.");
            _recorder.release();
        }
    }

    private AudioChunkType readBuffer()
    {
        if (_done)
            return null;

        int lenRead = 0;
        int toRead = _readBufferSize;
        short[] buf = new short[_readBufferSize];
        do
        {
            int shortsRead = _recorder.read(buf, lenRead, toRead);
            if (shortsRead < 0)
                break;

            lenRead += shortsRead;
            toRead -= shortsRead;
        } while (toRead > 0);

        if (lenRead <= 0)
        {
            Logger.error(this, "readBuffer() _recorder.read() didn't read any audio!!!");
            cleanup();
            handleSourceClosed(false);
            return null;
        }

        if (lenRead < _readBufferSize)
        {
            Logger.warn(this, "readBuffer() _recorder.read() didn't read expected len!!! " + lenRead);
            buf = ObjectUtils.createSubArray(buf, 0, lenRead);
        }

        if (_totalRead == 0)
        {
            _startTimestamp = SystemClock.uptimeMillis() - _audioType.getDuration(lenRead);
        }

        long timestamp = _startTimestamp + _audioType.getDuration(_totalRead);
        _totalRead += lenRead;

        return createNewAudioChunk(_audioType, buf, timestamp);
    }

    @Override
    protected boolean isCodecSupported(AudioType audioType) {
        if (audioType.encoding == Encoding.PCM_16)
            return true;
        return false;
    }

    /**
     * Create an audio chunk whose type could be {@link AudioChunk} or {@link MultiChannelAudioChunk}
     * or any other AudioChunkType which is a subclass of {@link AbstractAudioChunk}
     *
     * @param type The audio type to record.
     * @param data Audio data in shorts
     * @param timestamp
     */
    protected abstract AudioChunkType createNewAudioChunk(AudioType type, short [] data, long timestamp) ;

}
