package com.nuance.dragon.toolkit.audio.sources;

import java.util.ArrayList;
import java.util.LinkedList;

import android.util.Pair;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * The queuing audio source appears as a single audio source, but is backed by
 * multiple queued audio sources.
 * <p>
 * This audio source may be configured to remain open until explicitly stopped,
 * or to automatically close when it runs out of queued audio sources.
 */
public class QueuingAudioSource<AudioChunkType extends AbstractAudioChunk> extends AudioSource<AudioChunkType>
{
    /**
     * Optional listener for determining when a queued audio source has been
     * fully consumed.
     */
    public interface Listener
    {
        /**
         * The audio source has been consumed.
         * @param doneSource The audio source.
         * @param remaining How many are remaining in the queue.
         */
        void onSourceDone(AudioSource<? extends AbstractAudioChunk> doneSource, int remaining);
    }

    private final LinkedList<Pair<AudioSource<AudioChunkType>, Listener>> _queue;
    private AudioSource<AudioChunkType> _current;
    private Listener _currentListener;
    private AudioSink<AudioChunkType> _sink;
    private boolean _closed;
    private final AudioType _audioType;
    private final boolean _autoClose;
    private final AudioSink<AudioChunkType> _internalSink;

    /**
     * @param audioType The audio type of the sources that will be queued.
     * @param autoStop true if the queue should automatically stop when the
     * last queued audio source has closed. If false, the queue will not stop
     * until {@link #stop} is called.
     */
    public QueuingAudioSource(AudioType audioType, boolean autoStop)
    {
        super(null);
        
        Checker.checkArgForNull("audioType", audioType);

        _audioType = audioType;
        _autoClose = autoStop;
        _closed = false;
        _queue = new LinkedList<Pair<AudioSource<AudioChunkType>, Listener>>();
        _internalSink = new AudioSink<AudioChunkType>(_mainThreadHandler)
        {
            @Override
            public void chunksAvailable(AudioSource<AudioChunkType> source)
            {
                if (_sink != null && _current == source)
                {
                    _sink.chunksAvailable(QueuingAudioSource.this);
                }
            }

            @Override
            public void framesDropped(AudioSource<AudioChunkType> source)
            {
                if (_sink != null && _current == source)
                {
                    _sink.framesDropped(QueuingAudioSource.this);
                }
            }

            @Override
            public void sourceClosed(AudioSource<AudioChunkType> source)
            {
                if (_current == source)
                {
                    if (source.isEmptyForSink(this))
                    {
                        moveToNextSource();
                    }
                }
            }
        };
    }

    /**
     * Stop the queue. This will finally close the queuing audio source,
     * notifying any attached sink via {@link AudioSink#sourceClosed}. If there
     * is a queued source currently active, it will be disconnected. If this
     * queuing audio source was not configured to auto-close, it will never
     * close until this method is called.
     */
    public void stop()
    {
        if (!_closed)
        {
            Logger.debug(this, "Stopping queue");
            _closed = true;

            disconnectAll();

            if (_sink != null)
            {
                _sink.sourceClosed(this);
            }
        }
    }

    /**
     * Queue an audio source. This audio source will block the queue until it
     * closes. Must not be called after the source has been stopped.
     * @param audioSource The audio source to queue
     */
    public void queue(AudioSource<AudioChunkType> audioSource)
    {
        queue(audioSource, null);
    }

    /**
     * Queue an audio source. This audio source will block the queue until it
     * closes.  Must not be called after the source has been stopped.
     * @param audioSource The audio source to queue
     * @param listener Will be notified when the queued audio source has has
     * been consumed. If null, there will be no notifications.
     */
    public void queue(AudioSource<AudioChunkType> audioSource, Listener listener)
    {
        Checker.checkState(this, !_closed, "Source is closed");
        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForCondition("audioSource", "a supported audio type", _audioType.equals(audioSource.getAudioType()));

        Logger.debug(this, "Queuing audio source");
        _queue.add(new Pair<AudioSource<AudioChunkType>, Listener>(audioSource, listener));

        if (_current == null)
        {
            moveToNextSource();
        }
    }

    private void moveToNextSource()
    {
        if (_closed) {
            Logger.warn(this, "Can't begin next source because queuing source is closed");
            return;
        }

        if (_current != null)
        {
            Logger.debug(this, "Done with audio source");
            _internalSink.disconnectAudioSource();
            if (_currentListener != null)
            {
                _currentListener.onSourceDone(_current, _queue.size());
                _currentListener = null;
            }
            _current = null;
        }

        while (!_queue.isEmpty())
        {
            Pair<AudioSource<AudioChunkType>, Listener> removed = _queue.remove();

            // Used to be "if (removed.first.isEmptyForSink(this._internalSink))".
            // TODO: Either redefine what an active source means or change the TtsSource
            //        or FileSource so we can simplify the logic here.
            if (removed.first.isActive() || !removed.first.isEmpty())
            {
                Logger.debug(this, "Starting next queued audio source");
                _current = removed.first;
                _currentListener = removed.second;
                _internalSink.connectAudioSource(_current);
                break;
            } else
            {
                Logger.debug(this, "Discarding dead audio source");
            }
        }

        if (_current == null && _queue.isEmpty())
        {
            if (_autoClose)
            {
                Logger.debug(this, "Finished playing final queued audio source, stopping queue");
                stop();
            } else
            {
                Logger.debug(this, "Finished playing final queued audio source, going idle");
            }
        }
    }

    @Override
    public AudioType getAudioType()
    {
        return _audioType;
    }

    @Override
    public int getChunksAvailableForSink(AudioSink<AudioChunkType> sink)
    {
        if (_sink == sink && _current != null)
        {
            int available = _current.getChunksAvailableForSink(_internalSink);
            return available;
        }
        return 0;
    }

    @Override
    public int getChunksAvailable()
    {
        if (_current != null)
        {
            int available = _current.getChunksAvailable();
            return available;
        }
        return 0;
    }

    @Override
    public boolean isActive()
    {
        return !_closed;
    }

    @Override
    public AudioChunkType getAudioChunkForSink(AudioSink<AudioChunkType> sink)
    {
        if (_sink == sink && _current != null)
        {
            AudioChunkType chunk = _current.getAudioChunkForSink(_internalSink);
            if (!_current.isActive() && _current.isEmptyForSink(_internalSink))
            {
                final AudioSource<AudioChunkType> current = _current;
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (_current == current)
                            moveToNextSource();
                    }
                });
            }
            return chunk;
        }
        return null;
    }

    @Override
    protected void audioSinkConnected(AudioSink<AudioChunkType> sink)
    {
        _sink = sink;
    }

    @Override
    protected void audioSinkDisconnected(AudioSink<AudioChunkType> sink)
    {
        if (_sink == sink)
        {
            _sink = null;
            disconnectAll();
        }
    }

    private void disconnectAll()
    {
        int numQueued = _queue.size();
        ArrayList<Pair<AudioSource<AudioChunkType>, Listener>> remaining = new ArrayList<Pair<AudioSource<AudioChunkType>, Listener>>(numQueued);
        remaining.addAll(_queue);
        _queue.clear();

        if (_current != null)
        {
            AudioSource<AudioChunkType> current = _current;
            _current = null;
            if (_currentListener != null)
                _currentListener.onSourceDone(current, numQueued);
            _currentListener = null;
            _internalSink.disconnectAudioSource();

            for (Pair<AudioSource<AudioChunkType>, Listener> p : remaining)
            {
                numQueued--;

                _internalSink.disconnectAudioSource();
                if (p.second != null)
                    p.second.onSourceDone(p.first, numQueued);
            }
        }
    }
}
