package com.nuance.dragon.toolkit.audio.sources;

import java.util.LinkedList;
import java.util.List;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.WorkerThread;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.impl.WorkerThreadOem;

/**
 * An audio source that obtains its audio data from the device's recorder.
 */
public abstract class RecorderSource<AudioChunkType extends AbstractAudioChunk> extends SingleSinkSource<AudioChunkType>
{
    /**
     * Notifies when the recorder actually starts recording.
     */
    public interface Listener<AudioChunkType extends AbstractAudioChunk>
    {
        /**
         * The recorder has started recording.
         * @param recorder
         */
        void onStarted(RecorderSource<AudioChunkType> recorder);


        /**
         * The recorder has stopped recording.
         * @param recorder
         */
        void onStopped(RecorderSource<AudioChunkType> recorder);

        /**
         * The recorder has failed to capture audio.
         * @param recorder
         */
        void onError(RecorderSource<AudioChunkType> recorder);
    }

    private final AudioType _audioType;
    private final List<AudioChunkType> _buffer;
    private WorkerThread _workerThread;
    private NMTHandler _workerThreadHandler;
    private Listener<AudioChunkType> _listener;
    private boolean _started; // keeps track of start, to avoid starting twice
    private boolean _stopped;
    private boolean _closed;
    private boolean _useOwnWorkerThread; // create own thread, false if user specifies a worker thread

    /**
     * @param audioType The audio type to record.
     * @param workerThreadHandler The worker thread to use, or null if we should
     * create a temporary one.
     */
    public RecorderSource(AudioType audioType, NMTHandler workerThreadHandler)
    {
        this(audioType, workerThreadHandler, null);
    }

    /**
     * @param audioType The audio type to record.
     * @param workerThreadHandler The worker thread to use, or null if we should
     * create a temporary one.
     */
    public RecorderSource(AudioType audioType, NMTHandler workerThreadHandler, NMTHandler mainThreadHandler)
    {
        super(mainThreadHandler);
        Checker.checkArgForNull("audioType", audioType);
        Checker.checkArgForCondition("audioType", "a type supported by this player", isCodecSupported(audioType));
        _audioType = audioType;
        _buffer = new LinkedList<AudioChunkType>();
        if (workerThreadHandler == null)
        {
            _useOwnWorkerThread = true;
            _workerThread = null;

        } else
        {
            _workerThread = null;
            _workerThreadHandler = workerThreadHandler;
        }
        _closed = false;
    }

    @Override
    public AudioType getAudioType()
    {
        return _audioType;
    }

    @Override
    public int getChunksAvailable()
    {
        return _buffer.size();
    }

    @Override
    protected AudioChunkType getAudioChunk()
    {
        // Return null if size is 0.
        if (!_buffer.isEmpty())
            return _buffer.remove(0);
        else
            return null;
    }

    /**
     * Start recording audio. Audio will be DISCARDED whenever no audio sink is
     * connected.
     * <p>
     * The recorder may only be started ONCE.
     * @param listener Notifies when the recorder actually starts and stops recording,
     * since there may be a delay. If null, there will be no notification.
     */
    public void startRecording(final Listener<AudioChunkType> listener)
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                Checker.checkState(this, !_started, "Already started");

                if (_useOwnWorkerThread) {
                    _workerThread = new WorkerThreadOem("com.nuance.dragon.toolkit.audio.sources.RecorderSource");
                    _workerThread.start();
                    _workerThreadHandler = _workerThread.getHandler();
                }

                Logger.debug(this, "Starting recording");
                _listener = listener;

                // Reset Recorder's state
                _started = true;
                _stopped = false;
                _closed = false;

                _workerThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        boolean success = startRecordingInternal(_audioType);
                        if (!success)
                        {
                            handleSourceClosed(success);
                        }
                    }
                });
            }
        });
    }

    /**
     * Start recording audio. Audio will be DISCARDED whenever no audio sink is
     * connected.
     * <p>
     * The recorder may only be started ONCE.
     */
    public void startRecording()
    {
        startRecording(null);
    }

    /**
     * Stop recording. There may still be delayed audio chunks after calling
     * this function. This is always safe to call, even if the recorder hasn't
     * started or has already stopped.
     */
    public void stopRecording()
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                Logger.debug(this, "Stopping recording");
                if (_started && !_stopped)
                {
                    _stopped = true;

                    _workerThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            stopRecordingInternal();
                        }
                    });
                }
            }
        });
    }

    /**
     * Checks if codec is supported.
     * @param audioType Can't be null.
     * @return false if not supported, true if supported
     */
    protected abstract boolean isCodecSupported(AudioType audioType);

    /**
     * Start recording. Called once by the base class. Called from worker thread.
     * @param audioType
     * @return false on failure
     */
    protected abstract boolean startRecordingInternal(AudioType audioType);

    /**
     * Stop recording. Called once by the base class. Called from worker thread.
     */
    protected abstract void stopRecordingInternal();

    /**
     * Called from audio thread by the implementing class as soon as recording
     * has actually started.
     */
    protected void handleStarted()
    {
        _mainThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (_listener != null)
                {
                    _listener.onStarted(RecorderSource.this);
                }
            }
        });
    }

    /**
     * Call from audio thread by the implementing class every time new audio is
     * available.
     * @param audio
     */
    protected void handleNewAudio(final AudioChunkType audio)
    {
        Logger.verbose(this, "[LATCHK] handleNewAudio() audio:" + audio);
        _mainThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                // Buffer the audio
                _buffer.add(audio);

                // If there is a connected sink, notify it that audio is available
                notifyChunksAvailable();
            }
        });
    }

    /**
     * Call from audio thread by the implementing class when the recorder has
     * closed.
     */
    protected void handleSourceClosed()
    {
        handleSourceClosed(true);
    }

    /**
     * Call from audio thread by the implementing class when the recorder has
     * closed.
     * @param success
     */
    protected void handleSourceClosed(final boolean success)
    {
        Logger.debug(this, "handleSourceClosed()");
        _mainThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if( _started)
                    _started = false;

                if (!_closed)
                {
                    _closed = true;

                    notifySourceClosed();
                    if (_workerThread != null)
                    {
                        _workerThread.stop();
                        _workerThread = null;
                    }
                }

                if (_listener != null)
                {
                    if (success)
                    {
                        _listener.onStopped(RecorderSource.this);
                    }
                    else
                    {
                        _listener.onError(RecorderSource.this);
                    }
                    _listener = null;
                }
            }
        });
    }

    /**
     * Call from audio thread by the implementing class when the frames dropped occurs
     * in player.
     */
    protected void handleFramesDropped()
    {
        _mainThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                notifyFramesDropped();
            }
        });
    }

    @Override
    public boolean isActive()
    {
        return !_closed;
    }
}
