package com.nuance.dragon.toolkit.audio.sources;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;

/**
 * Base class for sources that only allow one sink to be connected
 *
 * @param <ChunkType>
 */
public abstract class SingleSinkSource<ChunkType extends AbstractAudioChunk> extends AudioSource<ChunkType>
{
    private AudioSink<ChunkType> _connectedSink;
    private boolean _closed;
    private boolean _framesDropped;

    /**
     * Gets the next audio chunk
     * @return The audio chunk, or null if there are none
     */
    protected abstract ChunkType getAudioChunk();

    protected SingleSinkSource(NMTHandler mainThreadHandler)
    {
        super(mainThreadHandler);
    }
    
    /**
     * Override to do something when a sink is connected.
     * @param sink The sink that was connected
     */
    protected void onSinkConnected(AudioSink<ChunkType> sink)
    {
    }

    /**
     * Override to do something when a sink is disconnected.
     * @param sink The sink that was disconnected
     */
    protected void onSinkDisconnected(AudioSink<ChunkType> sink)
    {
    }

    /**
     * Notify the connected sink (if any) that chunks are available.
     */
    protected void notifyChunksAvailable()
    {
        if (_connectedSink != null)
            _connectedSink.chunksAvailable(this);
    }

    /**
     * Notify the connected sink (if any) that frames were dropped.
     */
    protected void notifyFramesDropped()
    {
        if (_connectedSink != null)
            _connectedSink.framesDropped(this);
        else
            _framesDropped = true;
    }

    /**
     * Notify the connected sink (if any) that the source has closed.
     */
    protected void notifySourceClosed()
    {
        if (!_closed)
        {
            if (_connectedSink != null)
                _connectedSink.sourceClosed(this);
            else
                _closed = true;
        }
    }

    @Override
    public final int getChunksAvailableForSink(AudioSink<ChunkType> sink)
    {
        if (sink == _connectedSink)
            return getChunksAvailable();

        return 0;
    }

    @Override
    public final boolean isEmptyForSink(AudioSink<ChunkType> sink)
    {
        if (sink == _connectedSink)
            return isEmpty();

        return true;
    }

    @Override
    public final ChunkType getAudioChunkForSink(AudioSink<ChunkType> sink)
    {
        if (sink == _connectedSink)
            return getAudioChunk();

        return null;
    }

    @Override
    protected final void audioSinkConnected(final AudioSink<ChunkType> sink)
    {
        _connectedSink = sink;

        onSinkConnected(sink);
        
        // Posting the job is necessary to avoid a race condition where the recognizers
        // are connected to a pipe with already chunks in it and start pushing the audio
        // to non set-up back-end engines.
        _mainThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (sink != _connectedSink)
                    return;

                if (_framesDropped)
                {
                    _framesDropped = false;
                    sink.framesDropped(SingleSinkSource.this);
                }

                if (!isEmpty())
                {
                    sink.chunksAvailable(SingleSinkSource.this);
                }

                if (_closed)
                {
                    sink.sourceClosed(SingleSinkSource.this);
                }
            }
        });
    }

    @Override
    protected final void audioSinkDisconnected(AudioSink<ChunkType> sink)
    {
        if (sink == _connectedSink)
        {
            _connectedSink = null;
            onSinkDisconnected(sink);
        } else
        {
            Logger.warn(this, "Wrong sink disconnected");
        }
    }
}
