package com.nuance.dragon.toolkit.audio.sources;

import java.util.ArrayList;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk;
import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * An AudioSource that is used to stream audio pushed through the audio chain.
 */
public class StreamingBufferAudioSource<AudioChunkType extends AbstractAudioChunk>
        extends SingleSinkSource<AudioChunkType> {

    /**
     * Notifies when the recorder actually starts recording.
     */
    public interface Listener<AudioChunkType extends AbstractAudioChunk> {
        /**
         * The recorder has started recording.
         *
         * @param audioSource
         */
        public void onAudioSourceStarted(AudioSource<AudioChunkType> audioSource);

        /**
         * The recorder has stopped recording.
         *
         * @param audioSource
         */
        public void onAudioSourceStopped(AudioSource<AudioChunkType> audioSource);

        /**
         * The Audio Source encountered a fatal error. No more events will be
         * raised for that session.
         *
         * @param audioSource
         *            AudioSource instance.
         * @param code
         *            Error code, implementation specific.
         * @param message
         *            Error message.
         */
        public void onAudioSourceError(AudioSource<AudioChunkType> audioSource,
                int code, String message);
    }

    private AudioType _audioType;
    private ArrayList<AudioChunkType> _audioChunkQueue;
    private Listener<AudioChunkType> _listener;
    private boolean _isActive;
    private boolean _isClosedSent;
    private boolean _isStarted;

    /**
     * @param audioType
     *            The audio type that we will wish to stream.
     * @param workerThreadHandler
     *            The worker thread to use, or null if we should create a
     *            temporary one.
     */
    public StreamingBufferAudioSource(AudioType audioType, NMTHandler workerThreadHandler) {
        super(workerThreadHandler);
        Checker.checkArgForNull("audioType", audioType);
        _audioType = audioType;
        _audioChunkQueue = new ArrayList<AudioChunkType>();
        _listener = null;
        _isActive = false;
        _isClosedSent = false;
        _isStarted = false;
    }

    /**
     * Puts the Audio Source in ready state to start streaming.
     * @param listener IListener instance to send notifications.
     */
    public void start(final Listener<AudioChunkType> listener) {
        _mainThreadHandler.postToLooper(new Runnable() {

          @Override
          public void run() {
              Checker.checkState(this, !_isStarted, "Audio Source cannot be started; already active.");
              _isStarted = true;

              _mainThreadHandler.post(new Runnable() {

                @Override
                public void run() {
                    _listener = listener;
                    _isActive = true;
                    _isClosedSent = false;
                    if (_listener != null) {
                        listener.onAudioSourceStarted(StreamingBufferAudioSource.this);
                    }

                    if (!_audioChunkQueue.isEmpty()) {
                        notifyChunksAvailable();
                    }

                }
              });
          }
        });
    }

    /**
     * Indicates that the Audio Source is done streaming.
     *
     */
    public void stop() {
        _mainThreadHandler.postToLooper(new Runnable() {
            @Override
            public void run() {
                Checker.checkState(this, _isStarted, "Audio Source cannot be stopped; already inactive.");
                _isStarted = false;
                _mainThreadHandler.post(new Runnable() {

                    @Override
                    public void run() {
                        notifySourceClosed();
                        _isClosedSent = true;
                        if (_listener != null) {
                            _listener.onAudioSourceStopped(StreamingBufferAudioSource.this);
                        }
                        _isActive = false;
                    }

                });
            }
        });

    }

    /**
     * Push audio buffer for streaming through the audio chain.
     *
     * @param buffer
     *            Buffer containing audio data that matches the AudioType
     *            specified at construction.
     */
    public void pushAudio(byte[] buffer) {
         byte[] chunk = new byte[buffer.length];
         int durationMs = _audioType.getDuration(chunk.length / _audioType.getBytesPerSample());

         System.arraycopy(buffer, 0, chunk, 0, chunk.length);
         pushAudio(new AudioChunk(_audioType, chunk, durationMs));
    }

    /**
     * Push audio chunk for streaming through the audio chain.
     *
     * @param audioChunk
     *            Audio Chunk object that will be passed as-is through the audio
     *            chain. Note that its AudioType must match the one specified at
     *            construction.
     */
    public void pushAudio(final AudioChunk audioChunk) {
        _mainThreadHandler.postToLooper(new Runnable() {
             @Override
             public void run() {
                 Checker.checkState(this, audioChunk.audioType == _audioType, "Inconsistent Audio Type.");
                 _mainThreadHandler.post(new Runnable() {
                        @SuppressWarnings("unchecked") //For cast to AudioChunkType.
                        @Override
                        public void run() {
                            _audioChunkQueue.add((AudioChunkType) audioChunk);
                            if (_isActive) {
                                notifyChunksAvailable();
                            }
                        }

                    });

             }

        });

    }

    @Override
    protected AudioChunkType getAudioChunk() {
        AudioChunkType audioChunk = null;
        if (!_audioChunkQueue.isEmpty()) {
            audioChunk = _audioChunkQueue.get(0);
            _audioChunkQueue.remove(0);
        }
        return audioChunk;
    }

    @Override
    public AudioType getAudioType() {
        return _audioType;
    }

    @Override
    public boolean isActive() {
        return _isActive;
    }

    @Override
    public int getChunksAvailable() {
        return _audioChunkQueue.size();
    }

    @Override
    protected void onSinkConnected(AudioSink<AudioChunkType> sink) {
        // NOTE: Standard NMT Audio Design requirement:
        // On new sink connection, if the source has already sent its final SourceClosed notification
        // and it still has pending audio chunks to be consumed, the ChunksAvailable/SourceClosed
        // notifications shall be resent.

        if (!_isActive && _isClosedSent && _audioChunkQueue.size() > 0) {
            notifyChunksAvailable();
            notifySourceClosed();
        }
    }
}
