package com.nuance.dragon.toolkit.audio.sources;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.util.Mp3HeaderParser;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTAssetFileDescriptor;
import com.nuance.dragon.toolkit.oem.api.NMTContext;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.impl.HandlerOem;

/**
 * The file source acts like an audio recorder; it will generate audio
 * at regular intervals and not all at once. However, it acquires its
 * audio from a file rather than from the device's microphone.
 */
public class StreamingFileRecorderSource extends RecorderSource<AudioChunk>
{
    // default buffer length
    private static final int BUFFER_LENGTH_IN_MILLISECS = 400;

    private String _fileName;
    private final int _resourceId;
    private final int _bufferLengthInMs;
    private NMTAssetFileDescriptor _assetFileDesc;
    private final NMTContext _context;
    private boolean _littleEndian;
    private FileInputStream _fis = null;
    private Mp3HeaderParser _mp3Parser = null;
    private boolean _done = false;

    private Runnable _timerRunnable;
    private byte[] _buf;
    private int _bufIdx = 0;

    private NMTHandler _handler;
    private final FileManager _fileManager;

    /**
     * Create a file recorder source that reads the audio in little-endian format.
     * @param audioType The audio type to record.
     * @param fileName The file to read the audio from
     */
    public StreamingFileRecorderSource(AudioType audioType, String fileName)
    {
        this(audioType, BUFFER_LENGTH_IN_MILLISECS, fileName, true);
    }

    /**
     * Create a file recorder source.
     * @param audioType The audio type to record.
     * @param fileName The file to read the audio from
     * @param littleEndian True to save in little-endian format; false for big-
     * endian.
     */
    public StreamingFileRecorderSource(AudioType audioType, String fileName, boolean littleEndian)
    {
        this(audioType, BUFFER_LENGTH_IN_MILLISECS, fileName, littleEndian);
    }

    /**
     * Create a file recorder source.
     * @param audioType The audio type to record.
     * @param bufferLengthInMs The total size (in ms) of the buffer where
     * audio data is written to during the "fake" recording. Default value is 400.
     * @param fileName The file to read the audio from
     * @param littleEndian True to save in little-endian format; false for big-
     * endian.
     */
    public StreamingFileRecorderSource(AudioType audioType, int bufferLengthInMs, String fileName, boolean littleEndian)
    {
        this(audioType, bufferLengthInMs, fileName, null, littleEndian);
    }

    /**
     * Create a file recorder source.
     * @param audioType The audio type to record.
     * @param bufferLengthInMs The total size (in ms) of the buffer where
     * audio data is written to during the "fake" recording. Default value is 400.
     * @param fileName The file to read the audio from
     * @param fileManager is used to access file
     * @param littleEndian True to save in little-endian format; false for big-
     * endian.
     */
    public StreamingFileRecorderSource(AudioType audioType, int bufferLengthInMs, String fileName, FileManager fileManager, boolean littleEndian)
    {
        this(audioType, bufferLengthInMs, fileName, fileManager, littleEndian, null);
    }
    
    /**
     * Create a file recorder source.
     * @param audioType The audio type to record.
     * @param bufferLengthInMs The total size (in ms) of the buffer where
     * audio data is written to during the "fake" recording. Default value is 400.
     * @param fileName The file to read the audio from
     * @param fileManager is used to access file
     * @param littleEndian True to save in little-endian format; false for big-
     * endian.
     */
    public StreamingFileRecorderSource(AudioType audioType, int bufferLengthInMs, String fileName, FileManager fileManager, boolean littleEndian, NMTHandler mainThreadHandler)
    {
        super(audioType, null, mainThreadHandler);
        if (audioType == AudioType.GSM_FR)
            Checker.checkArgForCondition("bufferLengthInMs", "multiple of two", bufferLengthInMs % 2 == 0);
        _fileName = fileName;
        _littleEndian = littleEndian;
        _resourceId = -1;
        _context = null;
        _bufferLengthInMs = bufferLengthInMs;
        _fileManager = fileManager;
    }

    /**
     * Create a file recorder source with a given resource file Id, the file should
     * be deployed in the "res/raw" folder of the application.
     *
     * @param audioType The audio type to record.
     * @param bufferLengthInMs The total size (in ms) of the buffer where
     * audio data is written to during the "fake" recording. Default value is 400.
     * @param resourceId The resource id of the file, normally the file should be
     * in the res/raw folder of the application
     * @param context The application context object which will be used to load the file
     */
    public StreamingFileRecorderSource(AudioType audioType, int bufferLengthInMs, int resourceId, NMTContext context)
    {
        super(audioType, null, null);
        if (audioType == AudioType.GSM_FR)
            Checker.checkArgForCondition("bufferLengthInMs", "multiple of two", bufferLengthInMs % 2 == 0);
        _resourceId = resourceId;
        _context = context;
        _fileName = null;
        _littleEndian = true;
        _bufferLengthInMs = bufferLengthInMs;
        _fileManager = null;
    }

    @Override
    protected boolean startRecordingInternal(final AudioType audioType)
    {
        _done = false;
        _handler = new HandlerOem();

        try
        {
            if (null != _fileName)
            {
                boolean isAbsolute = new File(_fileName).isAbsolute();
                if (_fileManager != null && !isAbsolute)
                {
                    _fis = _fileManager.openFileForReading(_fileName);
                }
                else
                    _fis = new FileInputStream(new File(_fileName));
            }
            else if (_resourceId != -1 && _context != null)
            {
                _assetFileDesc = _context.openRawResourceFd(_resourceId);
                if (_assetFileDesc == null)
                {
                    throw new IllegalArgumentException("resourceId must refer to an uncompressed resource");
                }
                _fis = _assetFileDesc.createInputStream();
            }
            else
            {
                Logger.error(this,  "startRecordingInternal() FileInputStream has not been created!!!");
                return false;
            }
        }
        catch (FileNotFoundException e)
        {
            Logger.error(this,  "startRecordingInternal() FileNotFoundException!!!");
            _done = true;
            return false;
        }

        handleStarted();

        final int bufferSizeInBytes;
        if (audioType.encoding == AudioType.Encoding.PCM_16)
        {
            if (audioType.frequency == AudioType.Frequency.FREQ_8KHZ)
            {
                bufferSizeInBytes = AudioChunk.PCM_8K_SAMPLE_SIZE_IN_BYTES *
                                    AudioChunk.PCM_8K_SAMPLE_RATE_IN_HZ *
                                    _bufferLengthInMs /
                                    1000; // yes we do multiplication before devision to avoid loosing precision without casting to double
            }
            else if (audioType.frequency == AudioType.Frequency.FREQ_11KHZ)
            {
                bufferSizeInBytes = AudioChunk.PCM_11K_SAMPLE_SIZE_IN_BYTES *
                                    AudioChunk.PCM_11K_SAMPLE_RATE_IN_HZ *
                                    _bufferLengthInMs /
                                    1000; // yes we do multiplication before devision to avoid loosing precision without casting to double
            }
            else if (audioType.frequency == AudioType.Frequency.FREQ_16KHZ)
            {
                bufferSizeInBytes = AudioChunk.PCM_16K_SAMPLE_SIZE_IN_BYTES *
                                    AudioChunk.PCM_16K_SAMPLE_RATE_IN_HZ *
                                    _bufferLengthInMs /
                                    1000; // yes we do multiplication before devision to avoid loosing precision without casting to double
            }
            else if (audioType.frequency == AudioType.Frequency.FREQ_22KHZ)
            {
                bufferSizeInBytes = AudioChunk.PCM_22K_SAMPLE_SIZE_IN_BYTES *
                                    AudioChunk.PCM_22K_SAMPLE_RATE_IN_HZ *
                                    _bufferLengthInMs /
                                    1000; // yes we do multiplication before devision to avoid loosing precision without casting to double
            }
            else if (audioType.frequency == AudioType.Frequency.FREQ_44KHZ)
            {
                bufferSizeInBytes = AudioChunk.PCM_44K_SAMPLE_SIZE_IN_BYTES *
                                    AudioChunk.PCM_44K_SAMPLE_RATE_IN_HZ *
                                    _bufferLengthInMs /
                                    1000; // yes we do multiplication before devision to avoid loosing precision without casting to double
            }
            else if (audioType.frequency == AudioType.Frequency.FREQ_48KHZ)
            {
                bufferSizeInBytes = AudioChunk.PCM_48K_SAMPLE_SIZE_IN_BYTES *
                                    AudioChunk.PCM_48K_SAMPLE_RATE_IN_HZ *
                                    _bufferLengthInMs /
                                    1000; // yes we do multiplication before devision to avoid loosing precision without casting to double
            }
            else
            {
                bufferSizeInBytes = -1;
            }
        }
        else if (audioType.encoding == AudioType.Encoding.ULAW)
        {
            bufferSizeInBytes = AudioChunk.ULAW_SAMPLE_SIZE_IN_BYTES *
                    AudioChunk.ULAW_SAMPLE_RATE_IN_HZ *
                    _bufferLengthInMs /
                    1000; // yes we do multiplication before devision to avoid loosing precision without casting to double
        }
        else if (audioType.encoding == AudioType.Encoding.GSM_FR)
        {
            // yes we do multiplication before devision to avoid loosing precision without casting to double
            bufferSizeInBytes = _bufferLengthInMs * AudioChunk.GSM_FR_DOUBLE_FRAME_SIZE_IN_BYTES / AudioChunk.GSM_FR_DOUBLE_FRAME_LENGTH_IN_MILLISECONDS;
        }
        else if (audioType.encoding == AudioType.Encoding.AMR0)
        {
            // yes we do multiplication before devision to avoid loosing precision without casting to double
            bufferSizeInBytes = _bufferLengthInMs * AudioChunk.AMR0_FRAME_SIZE_IN_BYTES / AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS;
        }
        else if (audioType.encoding == AudioType.Encoding.AMR1)
        {
            // yes we do multiplication before devision to avoid loosing precision without casting to double
            bufferSizeInBytes = _bufferLengthInMs * AudioChunk.AMR1_FRAME_SIZE_IN_BYTES / AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS;
        }
        else if (audioType.encoding == AudioType.Encoding.AMR2)
        {
            // yes we do multiplication before devision to avoid loosing precision without casting to double
            bufferSizeInBytes = _bufferLengthInMs * AudioChunk.AMR2_FRAME_SIZE_IN_BYTES / AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS;
        }
        else if (audioType.encoding == AudioType.Encoding.AMR3)
        {
            // yes we do multiplication before devision to avoid loosing precision without casting to double
            bufferSizeInBytes = _bufferLengthInMs * AudioChunk.AMR3_FRAME_SIZE_IN_BYTES / AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS;
        }
        else if (audioType.encoding == AudioType.Encoding.AMR4)
        {
            // yes we do multiplication before devision to avoid loosing precision without casting to double
            bufferSizeInBytes = _bufferLengthInMs * AudioChunk.AMR4_FRAME_SIZE_IN_BYTES / AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS;
        }
        else if (audioType.encoding == AudioType.Encoding.AMR5)
        {
            // yes we do multiplication before devision to avoid loosing precision without casting to double
            bufferSizeInBytes = _bufferLengthInMs * AudioChunk.AMR5_FRAME_SIZE_IN_BYTES / AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS;
        }
        else if (audioType.encoding == AudioType.Encoding.AMR6)
        {
            // yes we do multiplication before devision to avoid loosing precision without casting to double
            bufferSizeInBytes = _bufferLengthInMs * AudioChunk.AMR6_FRAME_SIZE_IN_BYTES / AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS;
        }
        else if (audioType.encoding == AudioType.Encoding.AMR7)
        {
            // yes we do multiplication before devision to avoid loosing precision without casting to double
            bufferSizeInBytes = _bufferLengthInMs * AudioChunk.AMR7_FRAME_SIZE_IN_BYTES / AudioChunk.AMR_FRAME_LENGTH_IN_MILLISECONDS;
        }
        else if (audioType.encoding == AudioType.Encoding.MP3_128KBPS) {
			if (_mp3Parser == null) {
				_mp3Parser = new Mp3HeaderParser(_fis);
				_mp3Parser.stripID3v2Tag();
			}
			bufferSizeInBytes = _mp3Parser.getNextBufferSize();

		} else {
			bufferSizeInBytes = -1;
		}

		if (bufferSizeInBytes > 0)
			_buf = new byte[bufferSizeInBytes];
		_timerRunnable = new Runnable() {
			@Override
            public void run()
            {
                if (_done) return;
                boolean allDone = false;
                int bytesRead;
                if (audioType.encoding == AudioType.Encoding.SPEEX || audioType.encoding == AudioType.Encoding.OPUS)
                {
                    // read Nuance speex header
                    int speexBufferLen = 0;
                    byte[] nextByte = new byte[1];
                    do
                    {
                        try
                        {
                            bytesRead = _fis.read(nextByte);
                        } catch (IOException ioe)
                        {
                            Logger.error(this,  "TimerRunnable.run() reading header _fis.read() threw " + ioe + "!!!");
                            bytesRead = 0;
                        }
                        if (bytesRead != nextByte.length)
                        {
                            Logger.debug(this, "allDone 1");
                            _done = true;
                            cleanup();
                            handleSourceClosed();
                            return;
                        }
                        speexBufferLen = (speexBufferLen << 7) | nextByte[0] & 0x7F;
                    } while((nextByte[0] & 0x80) != 0);

                    if (speexBufferLen <= 0) {
                        Logger.error(this,  "Read file length error: speexBufferLen =" + speexBufferLen);
                        handleFramesDropped();

                        _done = true;
                        cleanup();
                        handleSourceClosed();
                        return;
                    }

                    _buf = new byte[speexBufferLen];
                }
                if(audioType.encoding == AudioType.Encoding.MP3_128KBPS)
                {
                	int bufferSize = _mp3Parser.getNextBufferSize();
                	bytesRead = -1;
                	if(bufferSize > 0){
                		// Reallocate buffer only if needed
                		if(_buf.length != bufferSize){
                			_buf = new byte[bufferSize];
                		}
						bytesRead = _mp3Parser.fillBuffer(_buf);
                	}

                }else
                {
	                try
	                {
	                    bytesRead = _fis.read(_buf);
	                } catch (IOException ioe)
	                {
	                    Logger.error(this,  "TimerRunnable.run() _fis.read() threw " + ioe + "!!!");
	                    bytesRead = 0;
	                }
                }

                if (bytesRead == -1)
                {
                    allDone = true;
                }
                else if (bytesRead != _buf.length)
                {
                    // NOTE: requested by test team to append the audio with silence
                    //       to provide the real audio from the file, we should get
                    //       rid of this part
//                    if (audioType.encoding == AudioType.Encoding.PCM_16)
//                    {
//                        // pcm auto padding
//                        Arrays.fill(_buf, bytesRead, _buf.length -1, (byte)0);
//                        bytesRead = _buf.length;
//                    }
                }

                if (bytesRead > 0)
                {
                    AudioChunk audioChunk;
                    if (audioType.encoding == AudioType.Encoding.PCM_16)
                    {
                        short[] shorts = new short[bytesRead / 2];
                        for (int i = 0; i < bytesRead - 1; i += 2)
                        {
                            if (_littleEndian)
                            {
                                shorts[i / 2] = (short)((((_buf[i + 1] << 8) & 0xFF00) | (_buf[i] & 0xFF)));
                            }
                            else
                            {
                                shorts[i / 2] = (short)((((_buf[i] << 8) & 0xFF00) | (_buf[i + 1] & 0xFF)));
                            }
                        }
                        audioChunk = new AudioChunk(audioType, shorts, _bufferLengthInMs * _bufIdx++);
                    }
                    else
                    {
                        byte[] buf = new byte[bytesRead];
                        System.arraycopy(_buf, 0, buf, 0, bytesRead);
                        audioChunk = new AudioChunk(audioType, buf, 100);
                    }
                    handleNewAudio(audioChunk);

                }
                if (!allDone)
                {
                    _handler.postDelayed(_timerRunnable, _bufferLengthInMs);
                }
                else
                {
                    Logger.debug(this, "allDone 2");
                    _done = true;
                    cleanup();
                    handleSourceClosed();
                }
            }
        };
        _handler.postDelayed(_timerRunnable, _bufferLengthInMs);
        return true;
    }

    @Override
    protected void stopRecordingInternal()
    {
        if (_done) return;

        _done = true;
        cleanup();
        handleSourceClosed();
    }

    private void cleanup() {
        if (_handler != null)
        {
            _handler.removeCallbacks(_timerRunnable);
        }

        if (_fis != null)
        {
            try
            {
                _fis.close();
            } catch (IOException ioe) {}
        }

        if (_assetFileDesc != null)
        {
            _assetFileDesc.close();
        }
        stopRecording();
    }

    // All codecs are supported
    @Override
    protected boolean isCodecSupported(AudioType audioType) {
        return true;
    }
}
