package com.nuance.dragon.toolkit.audio.util;

import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

public class Mp3HeaderParser {

    private InputStream _fis = null;
    private final int MP3_HEADER_SIZE = 4;
    private final int ID3_TAG_SIZE = 10;
    private final int[] MPEG2_BITRATE_TABLE = { 0, 8000, 16000, 24000, 32000, 40000, 48000, 56000, 64000, 80000, 96000, 112000, 128000, 144000, 160000, -1 };
    private final byte[] ID3_TAG_MARKER = { 'I', 'D', '3' };

    private byte[] _mp3Header;
    private int _currentBufferSize;
    public Mp3HeaderParser(InputStream fis)
    {
        Checker.checkArgForCondition("FileInputStream", "not null", fis != null);
        _fis = fis;
    }
    public boolean stripID3v2Tag()
    {
        // Try reading ID3 tag
        // We are going to cheat a little here since FileInputStream does not support rewinding
        // The strategy is to read the first 4 bytes which will indicate if an ID3 tag is present or not
        // If it is not then 4 bytes is conveniently the size of the mp3 header which we'll reuse later on
        byte[] id3Tag = new byte[MP3_HEADER_SIZE];
        _mp3Header = id3Tag;

        int bytesRead = 0;
        try {
            bytesRead = _fis.read(id3Tag);
        } catch (IOException e) {
            Logger.error(this,  "Mp3HeaderParser.checkID3Tag() _fis.read() threw " + e + "!!!");
            return false;
        }
        if(bytesRead != MP3_HEADER_SIZE)
        {
            Logger.error(this,  "Mp3HeaderParser.checkID3Tag() : + error while reading id3 header");
            return false;
        }

        // Check if the first 3 bytes of the file match 'ID3'
        byte[] toCheck = new byte[3];
        System.arraycopy(id3Tag, 0, toCheck, 0, 3);
        if(!(Arrays.equals(toCheck, ID3_TAG_MARKER)))
        {
            return false;
        }

        byte[] id3TagRemainder = new byte[ID3_TAG_SIZE - MP3_HEADER_SIZE];
        try {
            bytesRead = _fis.read(id3TagRemainder);
        } catch (IOException e) {
            Logger.error(this,  "Mp3HeaderParser.checkID3Tag() _fis.read() threw " + e + "!!!");
            return false;
        }

        if(bytesRead != ID3_TAG_SIZE - MP3_HEADER_SIZE)
        {
            Logger.error(this,  "Mp3HeaderParser.checkID3Tag() : + error while reading id3 header");
            return false;
        }

        // Try to read the length of the ID3 tag
        int tagLength = id3Tag[2] << 21 | id3Tag[3] << 14 | id3Tag[4] << 7 | id3Tag[5];
        try {
            // We already read 10 bytes from the header
            long skipLength = _fis.skip(tagLength - 10);
            if(skipLength != tagLength)
            {
                Logger.error(this,  "Mp3HeaderParser.checkID3Tag() : Unable to skip ID3 of length : " + tagLength);
            }
        } catch (IOException e) {
            Logger.error(this,  "Mp3HeaderParser.checkID3Tag() _fis.read() threw " + e + "!!!");
            return false;
        }

        // We really skipped an ID3 tag so the first 4 bytes were not a mp3 header
        _mp3Header = null;

        return true;
    }

    public int fillBuffer(byte[] buffer)
    {
        int bytesRead = 0;
        if(_mp3Header == null){
            if(getNextBufferSize() != 4)
            {
                return -1;
            }
        }
        //buffer = new byte[_currentBufferSize];
        try
        {
            bytesRead = _fis.read(buffer, 4, buffer.length - 4);
        } catch (IOException ioe)
        {
            Logger.error(this,  "TimerRunnable.run() _fis.read() threw " + ioe + "!!!");
            bytesRead = 0;
        }

        System.arraycopy(_mp3Header, 0, buffer, 0, _mp3Header.length);

        _mp3Header = null;
        return bytesRead + 4;
    }

    public int getNextBufferSize()
    {
        // If we already have a header in bank, use it. This will get flushed when the fillBuffer method is called
        // (when the mp3 frame is actually read)
        if(_mp3Header == null)
        {
            _mp3Header = new byte[4];
            int bytesRead = 0;
            try
            {
                bytesRead = _fis.read(_mp3Header);
            } catch (IOException ioe)
            {
                Logger.error(this,  "TimerRunnable.run() reading header _fis.read() threw " + ioe + "!!!");
                bytesRead = 0;
            }
            if (bytesRead != _mp3Header.length)
            {
                return -1;
            }
        }


        Logger.info(this,  "Bitrate : " + String.valueOf((_mp3Header[2] & 0xF0) >> 4) );
        Logger.info(this,  "Sample rate : " + String.valueOf((_mp3Header[2] & 0x0C) >> 2) );
        Logger.info(this,  "Padding : " + String.valueOf((_mp3Header[2] & 0x02) >> 1) );
        Logger.info(this,  "CRC : " + String.valueOf(_mp3Header[1] & 0x01));

        int chunkBitrate = 0;
        if((_mp3Header[2] & 0xF0) == 0xC0){
            chunkBitrate = 128000;
        }
        if((_mp3Header[2] & 0xF0) == 0xB0){
            chunkBitrate = 112000;
        }
        if((_mp3Header[2] & 0xF0) == 0xA0){
            chunkBitrate = 96000;
        }
        if((_mp3Header[2] & 0xF0) == 0x09){
            chunkBitrate = 80000;
        }

        int sampleRate = 1;
        if((_mp3Header[2] & 0x0C) == 0){
            sampleRate = 22050;
        }
        if((_mp3Header[2] & 0x0C) == 1){
            sampleRate = 24000;
        }
        if((_mp3Header[2] & 0x0C) == 2){
            sampleRate = 16000;
        }
        if((_mp3Header[1] & 0x01) == 0){
            byte[] crc = new byte[2];
            try {
                int readCrc = _fis.read(crc);
            } catch (IOException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }

        int frameSize = (int) (72 * (double)chunkBitrate / (double)sampleRate);
        int bufSize = (72 * chunkBitrate / sampleRate) + ((_mp3Header[2] & 0x02) >> 1);
        _currentBufferSize = bufSize;

        return bufSize;
    }
}
