package com.nuance.dragon.toolkit.audio.util;

import junit.framework.Assert;

/**
 * A linked-list backed queue class that supports multiple iterators accessing
 * the queue at different places. The queue will only retain as much data as is
 * required by all active queue iterators.
 * @param <T>
 */
public final class MultiIteratorQueue<T extends Object>
{
    private static abstract class ListItem<T extends Object>
    {
        protected static final int TYPE_VAL = 0;
        protected static final int TYPE_IT = 1;

        final int _type;
        ListItem<T> _prev;
        ListItem<T> _next;
        private final T _val;

        ListItem(int type, T value)
        {
            _type = type;
            _val = value;
        }

        /**
         * Unlink this from the linked list. Links the previous and next
         * together.
         */
        void unlink()
        {
            if (_prev != null)
                _prev._next = _next;

            if (_next != null)
                _next._prev = _prev;

            _prev = null;
            _next = null;
        }

        /**
         * Insert an item after this.
         * @param newItem
         */
        void insertAfter(ListItem<T> newItem)
        {
            newItem._prev = this;
            if (_next != null)
                _next._prev = newItem;
            newItem._next = _next;
            _next = newItem;
        }

        /**
         * Finds the next (after this) item in the linked list of the given
         * type.
         * @param type
         * @return The next item, or null if none was found.
         */
        ListItem<T> findNext(int type)
        {
            ListItem<T> next = _next;
            while (next != null)
            {
                if (next._type == type)
                    return next;

                next = next._next;
            }
            return null;
        }
    }

    private static final class ValueItem<T extends Object> extends ListItem<T>
    {
        ValueItem(T value)
        {
            super(TYPE_VAL, value);
        }
    }

    private static final class IteratorItem<T extends Object> extends ListItem<T>
    {
        IteratorItem()
        {
            super(TYPE_IT, null);
        }
    }

    /**
     * Iterates through the contents of the queue.
     * TODO How is this used?
     */
    public final class Iterator
    {
        private final IteratorItem<T> _item;
        private int _startId;

        Iterator(IteratorItem<T> item, int startId)
        {
            _item = item;
            _startId = startId;
        }

        /**
         * Advance to the next item in the queue, and return it.
         * @return The next item, or null if there are no more items.
         */
        public T next()
        {
            ListItem<T> valueItem = _item.findNext(ListItem.TYPE_VAL);
            if (valueItem != null)
            {
                Assert.assertTrue(size() > 0);

                T value = valueItem._val;

                // Move this iterator item forward in the linked list
                if (MultiIteratorQueue.this._first == _item)
                    MultiIteratorQueue.this._first = _item._next;

                _item.unlink();
                valueItem.insertAfter(_item);

                if (_item._next == null)
                    MultiIteratorQueue.this._last = _item;

                MultiIteratorQueue.this.cleanup();

                _startId++;

                return value;
            }
            Assert.assertEquals(0, size());
            return null;
        }

        /**
         * Get the number of remaining items represented by this iterator.
         * @return
         */
        public int size()
        {
            int size = MultiIteratorQueue.this._endId - _startId;
            return size;
        }

        /**
         * Release this iterator. If you don't call this, the queue will
         * continue to retain items that it thinks are needed by this iterator.
         * TODO What happens if called twice?
         */
        public void release()
        {
            if (MultiIteratorQueue.this._first == _item)
                MultiIteratorQueue.this._first = _item._next;

            if (MultiIteratorQueue.this._last == _item)
                MultiIteratorQueue.this._last = _item._prev;

            _item.unlink();
            MultiIteratorQueue.this.cleanup();
        }
    }

    private ListItem<T> _first;
    private ListItem<T> _last;

    // startId and endId are used to keep track of the length of different
    // iterators. Since the state of the linked list may change unexpectedly
    // from the perspective of iterators, we keep track of the apparent size
    // using these relative values.
    private int _startId;
    private int _endId;

    /**
     * Add an item to the end of the queue. Note: items can only be added if
     * there is at least one active iterator. Items added with no active
     * iterator will be discarded.
     * @param item The item to add.
     */
    public void add(T item)
    {
        if (_first == null)
        {
            return;
        }

        add(new ValueItem<T>(item));
        _endId++;
    }

    /**
     * Create an iterator, which starts at the current end of this queue. I.e.,
     * it will not have any items at first. The iterator will only reflect items
     * added after the iterator's creation.
     * <p>
     * Make sure to {@link Iterator#release} the iterator if you don't want the
     * queue to continue buffering items for this iterator.
     * @return
     */
    public Iterator createIterator()
    {
        IteratorItem<T> item = new IteratorItem<T>();
        add(item);
        Iterator it = new Iterator(item, _endId);
        return it;
    }

    /**
     * Get the total number of items in this queue. This value will likely
     * be larger than the {@link Iterator#size} of iterators.
     * @return
     */
    public int size()
    {
        return _endId - _startId;
    }

    private void add(ListItem<T> item)
    {
        if (_first == null)
        {
            _first = item;
        } else
        {
            _last.insertAfter(item);
        }
        _last = item;
    }

    private void cleanup()
    {
        // Iterators 'care' only about the value items that fall after them in
        // the linked list. Therefore, if there are value items at the beginning
        // of the linked list, no one cares about them anymore, and we can
        // discard them.
        while (_first != null && _first._type != ListItem.TYPE_IT)
        {
            ListItem<T> next = _first._next;
            _first.unlink();

            _first = next;
            _startId++;
        }
    }
}
