package com.nuance.dragon.toolkit.audio.util;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.TtsMarker;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Decodes Opus audio into PCM. TODO What is PCM?
 */
public class OpusDecoder
{
    /** Not for public use. */
    public static boolean LOAD_NATIVE_LIBRARY_SUCCESS = true;
    static {
        try {
            System.loadLibrary("dmt_opus");
        }
        catch (UnsatisfiedLinkError e) {
            Logger.error(OpusDecoder.class, "Failed to load native library.", e);
            LOAD_NATIVE_LIBRARY_SUCCESS = false;
        }
    }

    /* From Opus doc: 120ms. */
    private static final int MAXIMUM_PCM_CHUNK_SIZE_MS = 120;

    /* jni methods */
    private static native long initializeNative(int[] retError,
                                                int samplingRate);

    private static native int decodeNative(long decoder, byte[] inputOpusBuffer, int[] inputOpusBufferLen, short[] outputPcmBuffer, int[] outputPcmBufferLen);
    private static native void releaseNative(long decoder);

    private long _decoder;
    private AudioType _audioType;
    private short[] _tempOutBuffer;

    /**
     * Initialize the Opus decoder. This should be called exactly once, and
     * should be followed by a call to {@link #release} when finished.
     * @param audioType
     */
    public void initialize(AudioType audioType)
    {
        Assert.assertTrue(audioType.encoding == AudioType.Encoding.OPUS);
        Assert.assertTrue(audioType.frequency == AudioType.Frequency.FREQ_16KHZ ||
                          audioType.frequency == AudioType.Frequency.FREQ_8KHZ);

        int samplingRate;
        if (audioType.frequency == AudioType.Frequency.FREQ_8KHZ)
        {
            samplingRate = 8000;
        }
        else
        {
            samplingRate = 16000;
        }

        int [] error= new int[1];
        if (_decoder != 0)
            releaseNative(_decoder);
        _decoder = initializeNative(error,
                                    samplingRate);
        _audioType = new AudioType(AudioType.Encoding.PCM_16, audioType.frequency);
        // TODO: handle error?
    }

    /**
     * Release the Opus decoder. This should be called exactly once after
     * calling {@link #initialize}, otherwise memory will be leaked.
     */
    public void release()
    {
        if (_decoder != 0)
        {
            releaseNative(_decoder);
            _decoder = 0;
        }
    }

    /**
     * Decode some Opus data. This should probably be run on a background
     * thread of some sort.
     * @param inputOpusBuffer The Opus data to decode.
     * @return The decoded audio chunks.
     */
    public AudioChunk[] decode(AudioChunk inputOpusBuffer)
    {
        Checker.checkArgForNull("inputOpusBuffer", inputOpusBuffer);

        // Make the buffer large enough i.e. maximum indicated by the Opus spec.
        // Also, recycle the buffer so we go faster.
        final int expectedBufferSize = _audioType.getSampleCount(MAXIMUM_PCM_CHUNK_SIZE_MS);
        if (_tempOutBuffer == null || _tempOutBuffer.length < expectedBufferSize)
        {
            _tempOutBuffer = new short[expectedBufferSize];
        }

        int[] inBufferLen = new int[1];
        inBufferLen[0] = inputOpusBuffer.audioBytes.length;
        int[] outBufferLen = new int[1];
        outBufferLen[0] = _tempOutBuffer.length;

        decodeNative(_decoder, inputOpusBuffer.audioBytes, inBufferLen, _tempOutBuffer, outBufferLen);

        return createAudioChunks(outBufferLen[0], inputOpusBuffer.audioTimestamp, inputOpusBuffer.audioTtsMarkers);
    }

    private AudioChunk[] createAudioChunks(int length, long timestamp, TtsMarker[] markers)
    {
        short[] chunk = new short[length];
        System.arraycopy(_tempOutBuffer, 0, chunk, 0, chunk.length);

        AudioChunk[] retChunks = new AudioChunk[1];
        retChunks[0] = new AudioChunk(_audioType, chunk, timestamp, markers);
        return retChunks;
    }
}
