package com.nuance.dragon.toolkit.audio.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import junit.framework.Assert;
import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.TtsMarker;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Encodes PCM audio into Opus. TODO What is PCM?
 */
public class OpusEncoder
{
    /** Not for public use. */
    public static boolean LOAD_NATIVE_LIBRARY_SUCCESS = true;
    static {
        try {
            System.loadLibrary("dmt_opus");
        }
        catch (UnsatisfiedLinkError e) {
            Logger.error(OpusEncoder.class, "Failed to load native library.", e);
            LOAD_NATIVE_LIBRARY_SUCCESS = false;
        }
    }

    /* From Opus doc: Various Opus supported packet lengths.
       Passing in a duration of less than 10 ms will prevent
       the encoder from using the LPC or hybrid modes. */
    private static final int PACKET_LENGTH_10_MS = 10;
    private static final int PACKET_LENGTH_20_MS = 20;
    private static final int PACKET_LENGTH_40_MS = 40;
    private static final int PACKET_LENGTH_60_MS = 60;

    private static final int MINIMUM_PACKET_LENGTH_MS = PACKET_LENGTH_10_MS;

    /* jni methods */
    private static native long initializeNative(int[] retError,
                                                int samplingRate,
                                                int bitRate,
                                                int complexity,
                                                boolean vbr);

    private static native int encodeNative(long encoder, short[] inputPcmBuffer, int[] inputPcmBufferLen, byte[] outputOpusBuffer, int[] outputOpusBufferLen);
    private static native void releaseNative(long encoder);

    private long _encoder;
    private AudioType _audioType;
    private short[] _tempInBuffer;
    private byte[] _tempOutBuffer;

    /**
     * Initialize the Opus encoder. This should be called exactly once, and
     * should be followed by a call to {@link #release} when finished.
     * @param audioType
     * @param bitRate
     * @param complexity
     * @param vbr
     */
    public void initialize(AudioType audioType, int bitRate, int complexity, boolean vbr)
    {
        Assert.assertTrue(audioType.encoding == AudioType.Encoding.OPUS);

        int samplingRate;
        if (audioType.frequency == AudioType.Frequency.FREQ_8KHZ)
        {
            samplingRate = 8000;
        }
        else
        {
            samplingRate = 16000;
        }

        int [] error= new int[1];
        if (_encoder != 0)
            releaseNative(_encoder);
        _encoder = initializeNative(error,
                                    samplingRate,
                                    bitRate,
                                    complexity,
                                    vbr);
        _audioType = audioType;
        // TODO: handle error?
    }

    /**
     * Release the Opus encoder. This should be called exactly once after
     * calling {@link #initialize}, otherwise memory will be leaked.
     */
    public void release()
    {
        if (_encoder != 0)
        {
            releaseNative(_encoder);
            _encoder = 0;
        }
    }

    /**
     * Encode some PCM data. This should probably be run on a background
     * thread of some sort.
     * @param inputPcmChunk The PCM data to encode.
     * @return The encoded audio chunks.
     */
    public AudioChunk[] encode(AudioChunk inputPcmChunk)
    {
        Checker.checkArgForNull("inputPcmChunk", inputPcmChunk);

        final int targetedSampleCount = getTargetSampleCount(inputPcmChunk);
        if (_tempInBuffer == null || _tempInBuffer.length < targetedSampleCount)
        {
            _tempInBuffer = new short[targetedSampleCount];
        }

        // Make the buffer large enough, assuming the compression is 0% (no compression).
        // Also, recycle the buffer so we go faster.
        final int expectedBufferSize = 2 * targetedSampleCount; // data in bytes
        if (_tempOutBuffer == null || _tempOutBuffer.length < expectedBufferSize)
        {
            _tempOutBuffer = new byte[expectedBufferSize];
        }

        // Encode the PCM chunk into several Opus packets.
        // We throw away a piece of chunk if it is too small: the audio recorder shall give us a reasonably sized chunk (20 ms, 40 ms, etc.)
        final List<AudioChunk> chunks = new ArrayList<AudioChunk>();
        final int[] inBufferLen = new int[1];
        final int[] outBufferLen = new int[1];
        final int minimumSupportedPcmSampleCount = inputPcmChunk.audioType.getSampleCount(MINIMUM_PACKET_LENGTH_MS);
        int sampleOffset = 0;
        while((inputPcmChunk.audioShorts.length - sampleOffset) >= minimumSupportedPcmSampleCount)
        {
            int sizeToEncode = 0;
            if((inputPcmChunk.audioShorts.length - sampleOffset) < _tempInBuffer.length){
                sizeToEncode = inputPcmChunk.audioShorts.length - sampleOffset;
                // Reset the buffer to 0 so that the last buffer does not contain previous data
                Arrays.fill(_tempInBuffer, (short) 0);
            }else{
                sizeToEncode = _tempInBuffer.length;
            }
            System.arraycopy(inputPcmChunk.audioShorts, sampleOffset, _tempInBuffer, 0, sizeToEncode);
            sampleOffset = sampleOffset + _tempInBuffer.length;

            inBufferLen[0] = sizeToEncode;
            outBufferLen[0] = _tempOutBuffer.length;
            if (0 == encodeNative(_encoder, _tempInBuffer, inBufferLen, _tempOutBuffer, outBufferLen))
            {
                // Transmit non-empty packets, data size > 1 per OPUS doc.
                if (outBufferLen[0] > 1)
                    chunks.add(createAudioChunk(_audioType, _tempOutBuffer, outBufferLen[0], inputPcmChunk.audioTtsMarkers));
            }
        }

        return chunks.toArray(new AudioChunk[chunks.size()]);
    }

    private static AudioChunk createAudioChunk(AudioType audioType, byte[] buffer, int len, TtsMarker[] markers)
    {
        byte[] chunk = new byte[len];
        System.arraycopy(buffer, 0, chunk, 0, chunk.length);
        return new AudioChunk(audioType, chunk, 0, markers);
    }

    private static int getTargetSampleCount(AudioChunk inputAudioChunk)
    {
        final int inputDurationMs = inputAudioChunk.audioType.getDuration(inputAudioChunk.audioShorts);
        if (inputDurationMs < MINIMUM_PACKET_LENGTH_MS) {
            Logger.error(OpusEncoder.class, "Input audio chunk is too small for proper Opus encoding.");
            return 0;
        }
        // By research request, the packets must be of 20 ms
        return (inputAudioChunk.audioType.frequency * 20) / 1000;

        //// Find the best configuration for the input duration (biggest divider with smallest remaining).
        //final SparseArray<Pair<Float, Integer>> configs = new SparseArray<Pair<Float, Integer>>();
        //configs.put(PACKET_LENGTH_10_MS, new Pair <Float, Integer>((float)(inputDurationMs / PACKET_LENGTH_10_MS), inputDurationMs % PACKET_LENGTH_10_MS));
        //configs.put(PACKET_LENGTH_20_MS, new Pair <Float, Integer>((float)(inputDurationMs / PACKET_LENGTH_20_MS), inputDurationMs % PACKET_LENGTH_20_MS));
        //configs.put(PACKET_LENGTH_40_MS, new Pair <Float, Integer>((float)(inputDurationMs / PACKET_LENGTH_40_MS), inputDurationMs % PACKET_LENGTH_40_MS));
        //configs.put(PACKET_LENGTH_60_MS, new Pair <Float, Integer>((float)(inputDurationMs / PACKET_LENGTH_60_MS), inputDurationMs % PACKET_LENGTH_60_MS));
        //
        //Pair<Float, Integer> bestConfig = configs.get(PACKET_LENGTH_10_MS);
        //int bestConfigDurationMs = PACKET_LENGTH_10_MS;
        //for(int idx = 0; idx < configs.size(); idx++)
        //{
//            Pair<Float, Integer> currentConfig = configs.valueAt(idx);
//            if (currentConfig != null)
//            {
//                if (currentConfig.first >= 1 && currentConfig.first <= bestConfig.first && currentConfig.second <= bestConfig.second)
//                {
//                    bestConfig = currentConfig;
//                    bestConfigDurationMs = configs.keyAt(idx);
//                }
//            }
        //}
        //return (inputAudioChunk.audioType.frequency * 20) / 1000;
    }
}
