package com.nuance.dragon.toolkit.audio.util;

import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.ShortBuffer;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.audio.AbstractAudioChunk;
import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.Logger;
/**
 * Decodes Speex audio data into PCM. TODO What is PCM?
 */
public class SpeexDecoder
{
    /** Not for public use. */
    public static boolean LOAD_NATIVE_LIBRARY_SUCCESS = true;
    static {
        try {
            System.loadLibrary("dmt_speex");
        }
        catch (UnsatisfiedLinkError e) {
            Logger.error(SpeexDecoder.class, "Failed to load native library.", e);
            LOAD_NATIVE_LIBRARY_SUCCESS = false;
        }
    }

    /* jni methods */
    private static native long initializeNative(int[] retFrameSize, int[] retError, int mode, int samplingRate);
    private static native int decodeNative(long decoder, byte[] inputSpeexBuffer, int off, int len, OutputStream outputPcmBuffer);
    private static native void releaseNative(long decoder);

    private AudioType _audioType;
    private AudioType _outputAudioType;
    private int _pcmFrameSizeInSamples;
    private long _decoder;

    /**
     * Initialize the Speex decoder. This should be called exactly once, and
     * should be followed by a call to {@link #release} when finished.
     * @param audioType
     */
    public void initialize(AudioType audioType)
    {
        Assert.assertTrue(audioType.encoding == AudioType.Encoding.SPEEX);
        Assert.assertTrue(audioType.frequency == AudioType.Frequency.FREQ_16KHZ ||
                          audioType.frequency == AudioType.Frequency.FREQ_8KHZ);

        _audioType = audioType;
        _outputAudioType = new AudioType(AudioType.Encoding.PCM_16, _audioType.frequency);

        int mode = audioType.frequency == AudioType.Frequency.FREQ_16KHZ ? 1 : 0;
        int samplingRate = audioType.frequency;

        int [] error= new int[1];
        int [] frameSize= new int[1];
        if (_decoder != 0)
            releaseNative(_decoder);
        _decoder = initializeNative(frameSize, error, mode, samplingRate);
        _pcmFrameSizeInSamples = frameSize[0];

        // TODO: handle error?
    }

    /**
     * Release the Speex decoder. This should be called exactly once after
     * calling {@link #initialize}, otherwise memory will be leaked.
     */
    public void release()
    {
        if (_decoder != 0) {
            releaseNative(_decoder);
            _decoder = 0;
        }
    }

    /**
     * Decode some Speex data. This should probably be run on a background
     * thread of some sort.
     * @param inputSpeexBuffer The Speex data to decode. This method will begin
     * at index speexOffset of the byte array, and will decode as much as will
     * fit into pcmData.
     * @param off The offset into inputSpeexBuffer to decode from.
     * @param len The length of audio to decode.
     * @param speechStatus The speech status of the input audio
     * @return The decoded audio chunk.
     */
    public AudioChunk[] decode(byte[] inputSpeexBuffer, int off, int len, AbstractAudioChunk.SpeechStatus speechStatus)
    {
        if (_decoder == 0)
        {
            Logger.error(this, "Speex decoder isn't initialized");
            return new AudioChunk[0];
        }

        // TODO: If issue dealing with parallel decoders arises, maybe
        //       pass a copy of inputSpeexBuffer into the native layer
        //       (same logic than in SpeexEncoder.java).
        ByteArrayOutputStream outputPcmBuffer = new ByteArrayOutputStream();
        int ret = decodeNative(_decoder, inputSpeexBuffer, off, len, outputPcmBuffer);
        if (ret >= 0)
        {
            byte[] pcm_bytes = outputPcmBuffer.toByteArray();
            Assert.assertEquals(pcm_bytes.length % 2, 0);
            ByteBuffer byteBuffer = ByteBuffer.wrap(pcm_bytes);
            byteBuffer.order(ByteOrder.LITTLE_ENDIAN);
            ShortBuffer shortBuffer = byteBuffer.asShortBuffer();
            short[] pcm_buffer = new short[pcm_bytes.length / 2];
            shortBuffer.get(pcm_buffer);
            return new AudioChunk[] {new AudioChunk(_outputAudioType, pcm_buffer, speechStatus)};
        }
        return new AudioChunk[0];
    }
}
