package com.nuance.dragon.toolkit.audio.util;

import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.util.ArrayList;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.Logger;

/**
 * Encodes PCM audio into Speex. TODO What is PCM?
 */
public class SpeexEncoder
{
    /** Not for public use. */
    public static boolean LOAD_NATIVE_LIBRARY_SUCCESS = true;
    static {
        try {
            System.loadLibrary("dmt_speex");
        }
        catch (UnsatisfiedLinkError e) {
            Logger.error(SpeexDecoder.class, "Failed to load native library.", e);
            LOAD_NATIVE_LIBRARY_SUCCESS = false;
        }
    }

    /* jni methods */
    private static native long initializeNative(int[] retFrameSize,
                                                int[] retError,
                                                int mode,
                                                int samplingRate,
                                                int complexity,
                                                int quality,
                                                int endpointerStopOnEndOfSpeech,
                                                int endpointerVadLongUtterance,
                                                int endpointerVadHistoryLength,
                                                int endpointerVadBeginLength,
                                                int endpointerVadBeginThreshold,
                                                int endpointerVadBeginDelay,
                                                int endpointerVadEndLength,
                                                int endpointerVadEndThreshold,
                                                int endpointerVadInterSpeechLength);

    private static native int encodeNative(long encoder, short[] inputPcmBuffer, int off, int length, OutputStream outputSpeexBuffer, int[] consumed, int[] endpointer);
    private static native void releaseNative(long encoder);

    private AudioType _audioType;
    private AudioType _inputAudioType;
    private int _pcmFrameSizeInSamples;
    private AudioChunk.SpeechStatus _curSpeechStatus = AudioChunk.SpeechStatus.NO_SPEECH;
    private long _encoder;

    /**
     * Initialize the Speex encoder. This should be called exactly once, and
     * should be followed by a call to {@link #release} when finished.
     * @param audioType
     * TODO What do all of these mean>
     */
    public void initialize(AudioType audioType,
                           int endpointerStopOnEndOfSpeech,
                           int endpointerVadLongUtterance,
                           int endpointerVadHistoryLength,
                           int endpointerVadBeginLength,
                           int endpointerVadBeginThreshold,
                           int endpointerVadBeginDelay,
                           int endpointerVadEndLength,
                           int endpointerVadEndThreshold,
                           int endpointerVadInterSpeechLength)
    {
        Assert.assertTrue(audioType.encoding == AudioType.Encoding.SPEEX);

        _audioType = audioType;
        _inputAudioType = new AudioType(AudioType.Encoding.PCM_16, _audioType.frequency);

        int mode, samplingRate, complexity, quality;
        if (audioType.frequency == AudioType.Frequency.FREQ_8KHZ)
        {
            mode = 0;
            samplingRate = 8000;
            complexity = 3;
            quality = 6;
        }
        else
        {
            mode = 1;
            samplingRate = 16000;
            complexity = 3;
            quality = 8;
        }

        int [] error= new int[1];
        int [] frameSize= new int[1];
        if (_encoder != 0)
            releaseNative(_encoder);
        _encoder = initializeNative(frameSize, error,
                                    mode,
                                    samplingRate,
                                    complexity,
                                    quality,
                                    endpointerStopOnEndOfSpeech,
                                    endpointerVadLongUtterance,
                                    endpointerVadHistoryLength,
                                    endpointerVadBeginLength,
                                    endpointerVadBeginThreshold,
                                    endpointerVadBeginDelay,
                                    endpointerVadEndLength,
                                    endpointerVadEndThreshold,
                                    endpointerVadInterSpeechLength);
        _pcmFrameSizeInSamples = frameSize[0];
        // TODO: handle error?
    }

    /**
     * Release the Speex encoder. This should be called exactly once after
     * calling {@link #initialize}, otherwise memory will be leaked.
     */
    public void release()
    {
        if (_encoder != 0)
        {
            releaseNative(_encoder);
            _encoder = 0;
        }
    }

    /**
     * Decode some Speex data. This should probably be run on a background
     * thread of some sort.
     * @param inputPcmBuffer The PCM data to encode. This method will begin at
     * dataOffset of the short array, and will encode as much as will fit into
     * speexData.
     * @param off The offset into inputPcmBuffer to encode from. TODO Negative numbers valid?
     * @param len The length of audio to encode. TODO negative values allowed?
     * @return The encoded audio chunk.
     */
    public AudioChunk[] encode(short[] inputPcmBuffer, int off, int len)
    {
        if (len % _pcmFrameSizeInSamples != 0)
        {
            Logger.error(this, "the buffer length (" + len + ") is NOT divisible by Speex encoder frame size (" + _pcmFrameSizeInSamples + ")!!!");
            return new AudioChunk[0];
        }

        if (_encoder == 0)
        {
            Logger.error(this, "Speex encoder isn't initialized");
            return new AudioChunk[0];
        }

        short[] _tempInputBuffer = new short[inputPcmBuffer.length];

        // This copy is necessary because the Speex encoder seems to modify the
        // input buffer
        System.arraycopy(inputPcmBuffer, off, _tempInputBuffer, off, len);

        ArrayList<AudioChunk> encodedChunks = new ArrayList<AudioChunk>();
        int[] consumed = new int[1];
        int[] endpointer = new int[1];
        ByteArrayOutputStream outputSpeexBuffer = new ByteArrayOutputStream();

        do
        {
            outputSpeexBuffer.reset();
            int ret = encodeNative(_encoder, _tempInputBuffer, off, len, outputSpeexBuffer, consumed, endpointer);
            if (ret >= 0)
            {
                len -= consumed[0];
                off += consumed[0];
                AudioChunk chunk = new AudioChunk(_audioType, outputSpeexBuffer.toByteArray(), _inputAudioType.getDuration(consumed[0]), _curSpeechStatus);
                encodedChunks.add(chunk);
                updateCurSpeechStatus(endpointer[0]);
            } else
            {
                len = 0;
            }
        } while (len > 0);

        AudioChunk[] speexBuffers = new AudioChunk[encodedChunks.size()];
        return encodedChunks.toArray(speexBuffers);
    }

    private void updateCurSpeechStatus(int endPointerResult) {
        if (endPointerResult == 1) // SPEECH_START (1)
        {
            Assert.assertTrue(_curSpeechStatus == AudioChunk.SpeechStatus.NO_SPEECH);
            _curSpeechStatus = AudioChunk.SpeechStatus.SPEECH;
        }
        else if (endPointerResult == 2) // SPEECH_END (2)
        {
            Assert.assertTrue(_curSpeechStatus == AudioChunk.SpeechStatus.SPEECH);
            _curSpeechStatus = AudioChunk.SpeechStatus.NO_SPEECH;
        }
    }
}
