package com.nuance.dragon.toolkit.audio.util;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.TtsMarker;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Re-sample PCM audio. TODO What is PCM?
 */
public class SpeexResampler
{
    /** Not for public use. */
    public static boolean LOAD_NATIVE_LIBRARY_SUCCESS = true;
    static {
        try {
            System.loadLibrary("dmt_speex");
        }
        catch (UnsatisfiedLinkError e) {
            Logger.error(SpeexDecoder.class, "Failed to load native library.", e);
            LOAD_NATIVE_LIBRARY_SUCCESS = false;
        }
    }

    /* jni methods */
    private static native long initializeNative(int[] retError,
                                                int quality,
                                                int channels,
                                                int inSamplingRate,
                                                int outSamplingRate);

    private static native int resampleNative(long resampler, int channel,
                                             short[] inputPcmBuffer, int[] inputPcmLength,
                                             short[] outputPcmBuffer, int[] outputPcmLength);

    private static native void releaseNative(long resampler);

    private long _resampler;
    private short[] _tempResampledBuffer;
    private AudioType _resampledAudioType;

    /**
     * Initialize the Speex re-sampler. This should be called exactly once, and
     * should be followed by a call to {@link #release} when finished.
     * @param inputAudioType
     * @param outputAudioType
     * @param quality
     * TODO What do all of these mean>
     */
    public void initialize(AudioType inputAudioType,
                           AudioType outputAudioType,
                           int quality)
    {
        Checker.checkArgForCondition("inputAudioType", "PCM", inputAudioType.encoding == AudioType.Encoding.PCM_16);
        Checker.checkArgForCondition("outputAudioType", "PCM", outputAudioType.encoding == AudioType.Encoding.PCM_16);

        if (_resampler != 0)
            releaseNative(_resampler);

        int [] error= new int[1];

        // TODO: support multiple channels?
        _resampler = initializeNative(error, quality, 1,
                                      inputAudioType.frequency,
                                      outputAudioType.frequency);

        if (error[0] != 0)
            Logger.error(this, "Failed to initialize Speex resampler. Error code = " + error[0]);

        _resampledAudioType = outputAudioType;
    }

    /**
     * Release the Speex re-sampler. This should be called exactly once after
     * calling {@link #initialize}, otherwise memory will be leaked.
     */
    public void release()
    {
        if (_resampler != 0)
        {
            releaseNative(_resampler);
            _resampler = 0;
        }
    }

    /**
     * Re-sample PCM data. This should probably be run on a background
     * thread of some sort.
     * @param raw The PCM chunk to be re-sampled.
     * @param channel The channel to be re-sampled.
     * @return The re-sampled audio chunks.
     */
    public AudioChunk[] resample(AudioChunk raw, int channel)
    {
        Checker.checkArgForNull("raw", raw);

        // Make the buffer large enough, assuming we don't up-sample
        // more than twice the input signal per Nyquist theorem.
        // Also, recycle the buffer so we go faster.
        final int expectedBufferSize = 2 * raw.audioShorts.length;
        if (_tempResampledBuffer == null || _tempResampledBuffer.length < expectedBufferSize)
        {
            _tempResampledBuffer = new short[expectedBufferSize];
        }

        int[] rawLen =  new int[1];
        rawLen[0] = raw.audioShorts.length;
        int[] resampledLen = new int[1];
        resampledLen[0] = _tempResampledBuffer.length;

        resampleNative(_resampler, channel,
                       raw.audioShorts, rawLen,
                       _tempResampledBuffer, resampledLen);

        return createAudioChunks(resampledLen[0], raw.audioTimestamp, raw.audioTtsMarkers);
    }

    private AudioChunk[] createAudioChunks(int length, long timestamp, TtsMarker[] markers)
    {
        short[] chunk = new short[length];
        System.arraycopy(_tempResampledBuffer, 0, chunk, 0, chunk.length);

        AudioChunk[] retChunks = new AudioChunk[1];
        retChunks[0] = new AudioChunk(_resampledAudioType, chunk, timestamp, markers);
        return retChunks;
    }
}
