package com.nuance.dragon.toolkit.cloudservices;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.HttpUtil;
import com.nuance.dragon.toolkit.oem.api.HttpUtil.HttpConnectionStatus;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTContext;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.oem.impl.HandlerOem;
import com.nuance.dragon.toolkit.util.internal.Base64;

class ASRegistration implements JSONCompliant
{
    static interface RegistrationListener
    {
        /* on new registration succeeded */
        void onNewRegistrationSucceeded(ASRegistration reg);
        /* on new registration failed */
        void onNewRegistrationFailed(ASError error);
    }

    private static class RegistrationEntry
    {
        private static class RegistrationRequest
        {
            private final RegistrationListener _l;
            private final NMTHandler _h;
            
            private RegistrationRequest(RegistrationListener l, NMTHandler h)
            {
                _l = l;
                _h = h;
            }
        }
        private ASRegistration _reg;
        private List<RegistrationRequest> _requests = new LinkedList<RegistrationRequest>();
        private RegistrationEntry(ASRegistration r)
        {
            _reg = r;
        }
        private RegistrationEntry(RegistrationRequest r)
        {
            _reg = null;
            _requests.add(r);
        }
    }
    
    private static class SessionEntry
    {
        private static class SessionRequest
        {
            private final ASSession.SessionListener _l;
            private final NMTHandler _h;
            
            private SessionRequest(ASSession.SessionListener l, NMTHandler h)
            {
                _l = l;
                _h = h;
            }
        }
        private ASSession _sess;
        private List<SessionRequest> _requests = new LinkedList<SessionRequest>();
        private SessionEntry(ASSession s)
        {
            _sess = s;
        }
        private SessionEntry(SessionRequest r)
        {
            _sess = null;
            _requests.add(r);
        }
    }

    private static final String REGISTRATION_FILE_EXTENSION = ".reg";
    private static final String ANONYMOUS_KEY = "N_U_A_N_C_E_ANONY_KEY";

    private static final String KEY_KEY = "key";
    private static final String KEY_NMAID = "nmaid";
    private static final String KEY_CLIENT_ID = "clientId";
    private static final String KEY_CLIENT_SECRET = "clientSecret";
    private static final String KEY_DEVICE_ID = "deviceId";
    private static final String KEY_PLATFORM = "platform";
    private static final String KEY_AS_HOST = "asHost";
    private static final String KEY_AS_PORT = "asPort";

    private static final Hashtable<String, RegistrationEntry> Registrations = new Hashtable<String, RegistrationEntry>();
    private static final Object SyncObj = new Object();

    private final String _rKey;
    private final String _nmaid;
    private final String _deviceId;
    private final String _platform;
    private final String _clientId;
    private final String _clientSecret;
    private final String _asHost;
    private final int    _asPort;
    private final Object _syncObj = new Object();
    
    private HttpUtil _http;
    private volatile int _refCount = 0;

    private final Hashtable<String, SessionEntry> Sessions = new Hashtable<String, SessionEntry>();

    protected ASRegistration(String rKey, String nmaid, String deviceId, String platform, String clientId, String clientSecret, String asHost, int asPort, HttpUtil http)
    {
        _rKey = rKey;
        _nmaid = nmaid;
        _clientId = clientId;
        _clientSecret = clientSecret;
        _deviceId = deviceId;
        _platform = platform;
        _asHost = asHost;
        _asPort = asPort;
        
        _http = http;
    }
    
    // called from main thread
    static void newRegistration(final String nmaid, String password, final String deviceId, final String platform, final String asHost, final int asPort, final RegistrationListener listener, boolean reset, final FileManager fileMgr, boolean enableSelfSignedCert, String certSummary, String certData, NMTContext context)
    {
        synchronized(SyncObj)
        {
            Logger.debug(ASRegistration.class, "newRegistration(" + nmaid + ", " + password + ", " + deviceId + ", " + platform + ", " + asHost + ", " + asPort + ", " + reset + ")");
            NMTHandler mainHandler = new HandlerOem();

            final String rKey = "" + (nmaid + "_" + deviceId + "_" + platform + "_" + asHost + "_" + asPort).hashCode();
            final String fileName = rKey + REGISTRATION_FILE_EXTENSION;

            if (reset)
            {
                Logger.debug(ASRegistration.class, "newRegistration() resetting");
                Registrations.remove(rKey);
                fileMgr.delete(fileName);
                if (fileMgr.exists(fileName))
                    Logger.warn(ASRegistration.class, "newRegistration(): reset failed - deleting old cached registration failed!!! (" + fileName + ")");
            }

            final RegistrationEntry re = Registrations.get(rKey);
            Logger.debug(ASRegistration.class, "newRegistration() re:" + re + (re == null ? "" : ", re._requests.size():" + re._requests.size() + ", re._reg:" + re._reg));
            if (re == null || (re._requests.size() == 0 && re._reg == null))
            {
                if (fileMgr.exists(fileName))
                {
                    Logger.debug(ASRegistration.class, "newRegistration() " + fileName + " exists");
                    int fileSize = fileMgr.getSize(fileName);
                    FileInputStream fis = fileMgr.openFileForReading(fileName);
                    byte[] json = new byte[fileSize]; // TODO assuming file size is small
                    try {
                        fis.read(json); // TODO assuming whole file is read off in 1 shot
                        fis.close();
                        final ASRegistration cachedReg = ASRegistration.fromJSON(new org.json.JSONObject(new String(json, "UTF-8")));
                        Logger.debug(ASRegistration.class, "newRegistration() cachedReg: (" + cachedReg._nmaid + ", " + cachedReg._deviceId + ", " + cachedReg._platform + ", " + cachedReg._asHost + ", " + cachedReg._asPort + ")");
                        if (cachedReg._nmaid.equals(nmaid) && cachedReg._deviceId.equals(deviceId) && cachedReg._platform.equals(platform) && cachedReg._asHost.equals(asHost) && cachedReg._asPort == asPort)
                        {
                            Logger.debug(ASRegistration.class, "newRegistration() cachedReg is equal to the requested");
                            cachedReg._http = Factory.createHttpUtil(enableSelfSignedCert, certSummary, certData, context);
                            Registrations.put(rKey, new RegistrationEntry(cachedReg));
                            cachedReg._refCount++;
                            mainHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    Logger.debug(ASRegistration.class, "newRegistration() onNewRegistrationSucceeded(" + cachedReg + ") on cachedReg");
                                    listener.onNewRegistrationSucceeded(cachedReg);
                                }
                            });
                            return;
                        }
                    } catch (Exception e) {
                        Logger.warn(ASRegistration.class, "Exception thrown while accessing/parsing cached registration:" + e);
                        try {
                            fis.close();
                        } catch (IOException ee) {}
                        fileMgr.delete(fileName);
                        if (fileMgr.exists(fileName))
                            Logger.warn(ASRegistration.class, "newRegistration(): cleaning up inaccessible/unparsable cached registration failed!!! (" + fileName + ")");
                    }
                }

                /* Commented out for NJIISESDK-40
                String[] fileNames = fileMgr.findFiles(REGISTRATION_FILE_EXTENSION);
                if (fileNames != null)
                {
                    for (String fn : fileNames)
                    {
                        fileMgr.delete(fn);
                        if (fileMgr.exists(fn))
                            Logger.warn(ASRegistration.class, "newRegistration(): deleting old cached registration failed!!! (" + fn + ")");
                    }
                }*/

                // do a new registration
                String url = "https://" + asHost + ":" + asPort + "/oauth/client";
                Logger.debug(ASRegistration.class, "newRegistration() url" + url);
                Registrations.put(rKey, new RegistrationEntry(new RegistrationEntry.RegistrationRequest(listener, mainHandler)));
                final HttpUtil http = Factory.createHttpUtil(enableSelfSignedCert, certSummary, certData, context);
                http.request(HttpUtil.METHOD_POST, url, HttpUtil.PROTOCOL_CONTENT_TYPE_JSON, constructRegistrationRequest(nmaid, deviceId, platform), new HttpUtil.ResponseCallback()
                {
                    @Override
                    public void okCallback(String response) {
                        ASError err = null;
                        ASRegistration newReg = null;
                        try {
                            org.json.JSONObject jsonResponse = new org.json.JSONObject(response);
                            String clientId = jsonResponse.getString(KEY_CLIENT_ID);
                            String clientSecret = jsonResponse.getString(KEY_CLIENT_SECRET);
                            if (clientId.length() == 0 || clientSecret.length() == 0)
                            {
                                Logger.error(ASRegistration.class, "newRegistration() okCallback(): " + KEY_CLIENT_ID + "/" + KEY_CLIENT_SECRET + " is empty");
                                err = new ASError(ASError.INTERNAL_ERROR_CODE_SERVER_ERROR);
                            }

                            if (err == null)
                            {
                                newReg = new ASRegistration(rKey, nmaid, deviceId, platform, clientId, clientSecret, asHost, asPort, http);

                                // cache registration
                                FileOutputStream fos = fileMgr.openFileForWriting(fileName);
                                try {
                                    fos.write(newReg.toJSON().toString().getBytes()); // TODO assuming whole file is written off in 1 shot
                                    fos.close();
                                } catch (IOException e) {
                                    try {
                                        fos.close();
                                    } catch (IOException ee) {}
                                    fileMgr.delete(fileName);
                                    if (fileMgr.exists(fileName))
                                        Logger.warn(ASRegistration.class, "newRegistration(): cleaning up unsuccessfully-cached-registration failed!!! (" + fileName + ")");
                                    err = new ASError(ASError.INTERNAL_ERROR_CODE_REGISTRATION_CACHING_FAILURE);
                                }
                            }
                        } catch (JSONException e) {
                            Logger.error(ASRegistration.class, "newRegistration(): json parsing error!!! (" + response + ")");
                            err = new ASError(ASError.INTERNAL_ERROR_CODE_SERVER_ERROR);
                        }

                        if (err == null)
                        {
                            handleNewRegistrationSucceeded(rKey, newReg);
                        }
                        else
                        {
                            newReg.release();
                            handleNewRegistrationFailed(rKey, err);
                        }
                    }

                    @Override
                    public void errorCallback(HttpConnectionStatus status, int responseCode, String response) {
                        ASError error;
                        if (status == HttpConnectionStatus.HTTP_CONNECTION_ERROR)
                            error = new ASError(ASError.INTERNAL_ERROR_CODE_CONNECTION_FAILURE, response);
                        else
                            error = new ASError(responseCode, response);
                        http.release();
                        handleNewRegistrationFailed(rKey, error);
                    }
                }, constructExtraHeaders(nmaid, password));
            }
            else
            {
                if (re._requests.size() == 0)
                {
                    final ASRegistration reg = re._reg;
                    reg._refCount++;
                    mainHandler.post(new Runnable()
                    {
                        @Override
                        public void run() {
                            Logger.debug(ASRegistration.class, "newRegistration() onNewRegistrationSucceeded(" + reg + ") on reuse");
                            listener.onNewRegistrationSucceeded(reg);
                        }
                    });
                }
                else
                {
                    Logger.debug(ASRegistration.class, "newRegistration() add listener");
                    re._requests.add(new RegistrationEntry.RegistrationRequest(listener, mainHandler));
                }
            }
        }
    }

    private static void handleNewRegistrationSucceeded(String rKey, final ASRegistration reg)
    {
        List<RegistrationEntry.RegistrationRequest> requests;
        synchronized(SyncObj)
        {
            RegistrationEntry re = Registrations.get(rKey);
            re._reg = reg;
            requests = new LinkedList<RegistrationEntry.RegistrationRequest>(re._requests);
            reg._refCount += requests.size();
            re._requests.clear();
        }
        for (final RegistrationEntry.RegistrationRequest r : requests)
        {
            r._h.post(new Runnable()
            {
                @Override
                public void run() {
                    Logger.debug(ASRegistration.class, "handleNewRegistrationSucceeded() onNewRegistrationSucceeded(" + reg + ") on new");
                    r._l.onNewRegistrationSucceeded(reg);
                }
            });
        }
    }

    private static void handleNewRegistrationFailed(String rKey, final ASError error)
    {
        List<RegistrationEntry.RegistrationRequest> requests;
        synchronized(SyncObj)
        {
            RegistrationEntry re = Registrations.get(rKey);
            requests = new LinkedList<RegistrationEntry.RegistrationRequest>(re._requests);
            re._requests.clear();
        }
        for (final RegistrationEntry.RegistrationRequest r : requests)
        {
            r._h.post(new Runnable()
            {
                @Override
                public void run() {
                    Logger.debug(ASRegistration.class, "handleNewRegistrationFailed() onNewRegistrationFailed(" + error + ")");
                    r._l.onNewRegistrationFailed(error);
                }
            });
        }
    }

    void getSession(final ThirdPartyLogin login, final ASSession.SessionListener listener)
    {
        synchronized(_syncObj)
        {
            Logger.debug(this, "getSession(" + login + ")");
            NMTHandler mainHandler = Factory.createNMTHandler();
            boolean reuseSess = false, renewSess = false;
            
            final String sKey = login == null ? ANONYMOUS_KEY : login.ProviderId + "_" + login.UserId;
            SessionEntry se = Sessions.get(sKey);
            final ASSession sess = se == null ? null : se._sess;
            Logger.debug(this, "getSession() se:" + se + ", sess:" + sess);

            //Logger.debug(ASRegistration.this, "getSession() sess == null:" + (sess == null ? "true" : "false, sess.isExpired():" + sess.isExpired()) + ", " +
            //                                               "login == null:" + (login == null ? "true" + (sess == null ? "" : ", sess._login == null:" + (sess._login == null)) : "false,  login.equals(sess._login):" + login.equals(sess == null ? null : sess._login)));
            // existing session available
            Logger.debug(this, "getSession() (sess != null):" + (sess != null));
            if (sess != null)
            {
                // non-anonymous login
                Logger.debug(this, "getSession() (login != null):" + (login != null) + " 1");
                if (login != null)
                {
                    // matching previous login
                    Logger.debug(this, "getSession() login.equals(sess._login):" + login.equals(sess._login));
                    if (login.equals(sess._login))
                    {
                        Logger.debug(this, "getSession() !sess.isExpired():" + !sess.isExpired() + " 1");
                        if (!sess.isExpired())
                        {
                            reuseSess = true;
                        }
                        else
                        {
                            renewSess = true;
                        }
                    }
                }
                // anonymous login
                else
                {
                    // matching previous login
                    Logger.debug(this, "getSession() (sess._login == null):" + (sess._login == null));
                    if (sess._login == null)
                    {
                        Logger.debug(this, "getSession() !sess.isExpired():" + !sess.isExpired() + " 2");
                        if (!sess.isExpired())
                        {
                            reuseSess = true;
                        }
                        else
                        {
                            renewSess = true;
                        }
                    }
                }
            }
    
            Logger.debug(this, "getSession() reuseSess:" + reuseSess + ", renewSess:" + renewSess); 
            if (reuseSess)
            {
                mainHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        Logger.debug(ASRegistration.this, "getSession() onGetSessionSucceeded() on reuse");
                        listener.onGetSessionSucceeded(sess);
                    }
                });
                return;
            }
    
            if (se == null)
            {
                Logger.debug(this, "getSession() creating new SessionEntry");
                se = new SessionEntry(new SessionEntry.SessionRequest(listener, mainHandler));
                Sessions.put(sKey, se);
            }
            else
            {
                Logger.debug(this, "getSession() add listener");
                se._requests.add(new SessionEntry.SessionRequest(listener, mainHandler));
                if (se._requests.size() > 1)
                {
                    Logger.debug(this, "getSession() (se._requests.size() > 1): true so return");
                    return;
                }
            }
    
            String url = "https://" + _asHost + ":" + _asPort + "/oauth/token";
            Logger.debug(this, "getSession() url:" + url);
            final boolean renewSessFinal = renewSess;
            _http.request(HttpUtil.METHOD_POST, url, HttpUtil.PROTOCOL_CONTENT_TYPE_OAUTH, ASSession.constructSessionRequest((renewSess ? sess.RefreshToken : null), login, (sess == null ? null : sess.NuanceUserId)), new HttpUtil.ResponseCallback()
            {
                @Override
                public void okCallback(String response) {
                    List<SessionEntry.SessionRequest> requests;
                    ASSession sess = null;
                    ASError err = null;
                    synchronized(_syncObj)
                    {
                        Logger.debug(ASRegistration.this, "getSession() okCallback() response:" + response);
                        SessionEntry se = Sessions.get(sKey);
                        try {
                            org.json.JSONObject jsonResponse = new org.json.JSONObject(response);
                            se._sess = ASSession.fromJSON(jsonResponse);
                            se._sess._login = login;
                            sess = se._sess;
                        } catch (JSONException e) {
                            Logger.error(ASRegistration.this, "getSession(): json parsing error!!! (" + response + ")");
                            err = new ASError(ASError.INTERNAL_ERROR_CODE_SERVER_ERROR);
                        }
                        requests = new LinkedList<SessionEntry.SessionRequest>(se._requests);
                        se._requests.clear();
                    }
                    for (SessionEntry.SessionRequest r : requests)
                    {
                        final ASSession.SessionListener l = r._l;
                        NMTHandler h = r._h;
                        if (err == null)
                        {
                            final ASSession sessFinal = sess;
                            h.post(new Runnable()
                            {
                                @Override
                                public void run() {
                                    Logger.debug(ASRegistration.this, "getSession() okCallback() onGetSessionSucceeded() on " + (renewSessFinal ? "renew" : "new"));
                                    l.onGetSessionSucceeded(sessFinal);
                                }
                            });
                        }
                        else
                        {
                            final ASError errFinal = err;
                            h.post(new Runnable()
                            {
                                @Override
                                public void run() {
                                    Logger.debug(ASRegistration.this, "getSession() okCallback() onGetSessionFailed(" + errFinal + ") on " + (renewSessFinal ? "renew" : "new"));
                                    l.onGetSessionFailed(errFinal);
                                }
                            });
                        }
                    }
                }
    
                @Override
                public void errorCallback(HttpConnectionStatus status, int responseCode, String response) {
                    synchronized(_syncObj)
                    {
                        List<SessionEntry.SessionRequest> requests;
                        ASError error;
                        synchronized(_syncObj)
                        {
                            Logger.debug(ASRegistration.this, "getSession() errorCallback() status:" + status + ", responseCode:" + responseCode + ", response:" + response);
                            SessionEntry se = Sessions.get(sKey);
                            if (status == HttpConnectionStatus.HTTP_CONNECTION_ERROR)
                                error = new ASError(ASError.INTERNAL_ERROR_CODE_CONNECTION_FAILURE);
                            else if (status == HttpConnectionStatus.HTTP_CONNECTION_ERROR_SECURITY)
                                error = new ASError(ASError.INTERNAL_ERROR_CODE_CONNECTION_FAILURE_SECURITY);
                            else
                                error = new ASError(responseCode, response);
                            requests = new LinkedList<SessionEntry.SessionRequest>(se._requests);
                            se._requests.clear();
                        }
                        for (SessionEntry.SessionRequest r : requests)
                        {
                            ASSession.SessionListener l = r._l;
                            Logger.debug(ASRegistration.this, "getSession() errorCallback() onGetSessionFailed(" + error + ") on " + (renewSessFinal ? "renew" : "new"));
                            l.onGetSessionFailed(error);
                        }
                    }
                }
            }, constructExtraHeaders(_clientId, _clientSecret));
        }
    }

    public String getNuanceUserId(ThirdPartyLogin login)
    {
        synchronized(_syncObj)
        {
            final String sKey = login == null ? ANONYMOUS_KEY : login.ProviderId + "_" + login.UserId;
            SessionEntry se = Sessions.get(sKey);
            ASSession sess = se == null ? null : se._sess;
            return sess != null ? sess.NuanceUserId : null;
        }
    }
    
    public String getClientId()
    {
        return _clientId;
    }
    
    public void release()
    {
        synchronized(_syncObj)
        {
            synchronized(SyncObj)
            {
                Logger.debug(this, "release() _refCount:" + _refCount);
                _refCount--;
                if (_refCount == 0)
                {
                    Logger.debug(this, "release() releasing");
                    _http.release();
                    Registrations.remove(_rKey);
                }
            }
        }
    }

    private static Map<String, String> constructExtraHeaders(String username, String password)
    {
        HashMap<String, String> extraHeaders = new HashMap<String, String>();
        //String auth = "Basic " + Base64.encodeToString(String.format("%s:%s", username, password).getBytes(), Base64.NO_WRAP);
        String auth = "Basic " + Base64.encode(String.format("%s:%s", username, password).getBytes());
        extraHeaders.put("Authorization", auth);
        return extraHeaders;
    }

    private static String constructRegistrationRequest(String nmaid, String deviceId, String platform)
    {
        JSONObject o = new JSONObject();
        o.tryPut("nmaid", nmaid);
        o.tryPut("deviceId", deviceId);
        o.tryPut("platform", platform);
        return o.toString();
    }

    @Override
    public JSONObject toJSON() {
        JSONObject o = new JSONObject();
        o.tryPut(KEY_KEY, _rKey);
        o.tryPut(KEY_NMAID, _nmaid);
        o.tryPut(KEY_DEVICE_ID, _deviceId);
        o.tryPut(KEY_PLATFORM, _platform);
        o.tryPut(KEY_CLIENT_ID, _clientId);
        o.tryPut(KEY_CLIENT_SECRET, _clientSecret);
        o.tryPut(KEY_AS_HOST, _asHost);
        o.tryPut(KEY_AS_PORT, _asPort);
        return o;
    }

    public static ASRegistration fromJSON(org.json.JSONObject json) throws JSONException
    {
        String rKey = json.getString(KEY_KEY);
        String nmaid = json.getString(KEY_NMAID);
        String deviceId = json.getString(KEY_DEVICE_ID);
        String platform = json.getString(KEY_PLATFORM);
        String clientId = json.getString(KEY_CLIENT_ID);
        String clientSecret = json.getString(KEY_CLIENT_SECRET);
        String asHost = json.getString(KEY_AS_HOST);
        int asPort = json.getInt(KEY_AS_PORT);

        return new ASRegistration(rKey, nmaid, deviceId, platform, clientId, clientSecret, asHost, asPort, null);
    }
}
