package com.nuance.dragon.toolkit.cloudservices;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * An audio parameter to be sent via {@link Transaction#addParam}. Note: the
 * audio source must be started in order to begin collecting audio, and it must
 * close in order to finalize the transaction.
 */
public class AudioParam extends StreamParam
{
    private final AudioSource<AudioChunk> _source;
    private final AudioSink<AudioChunk> _sink;
    private StreamListener _streamListener;
    private CommandContext _commandContext;
    private boolean _doneRecording;

    private final List<AudioChunk> _audioBuffer;
    private final Boolean[] _recordingDoneWithError;

    /**
     * @param key    the name of the AudioParam specified by the Command on Nuance Mobile
     *               Application Server. (cannot be NULL or Empty).
     * @param source the audio source associated with the audio parameter,
     *               it will provide the input audio data, cloud services
     *               will pick the audio and stream them to the speech
     *               server for recognition. A possible audio source would
     *               be recorder, Speex encoder, or file source. (cannot be NULL).
     */
    public AudioParam(final String key, AudioSource<AudioChunk> source)
    {
        super(TYPE_AUDIO, key);
        Checker.checkArgForNull("source", source);
        _source = source;
        _doneRecording = false;

        _audioBuffer = new ArrayList<AudioChunk>();
        _recordingDoneWithError = new Boolean[1];
        _recordingDoneWithError[0] = Boolean.valueOf(false);

        _sink = new AudioSink<AudioChunk>(source.getMainThreadHandler())
        {
            @Override
            public void chunksAvailable(AudioSource<AudioChunk> source)
            {
                int available = _source.getChunksAvailableForSink(this);
                try {
                    while (available-- > 0)
                    {
                        if (_streamListener != null)
                        {
                            _commandContext.sendAudioForParam(key, _source.getAudioChunkForSink(_sink), false);
                        } else
                        {
                            _audioBuffer.add(_source.getAudioChunkForSink(_sink));
                        }
                    }
                }
                catch (Exception e) {
                    Logger.error(AudioParam.this, "Error sending audio for param " + key, e);

                    _doneRecording = true;
                    disconnectAudioSource();

                    if (_streamListener != null)
                    {
                        _streamListener.onStreamComplete(AudioParam.this, true);
                        _streamListener = null;
                    } else
                    {
                        _recordingDoneWithError[0] = Boolean.valueOf(true);
                    }
                    _commandContext = null;
                }
            }

            @Override
            public void framesDropped(AudioSource<AudioChunk> source)
            {
                // TODO
            }

            @Override
            public void sourceClosed(AudioSource<AudioChunk> source)
            {
                Logger.debug(AudioParam.this, "sourceClosed()");
                // try to get all the remaining chunks
                chunksAvailable(source);

                _doneRecording = true;
                disconnectAudioSource();

                boolean error = false;
                try {
                    if (_streamListener != null)
                    {
                        _commandContext.sendAudioForParam(key, null, true);
                    }
                }
                catch (Exception e) {
                    error = true;
                    _recordingDoneWithError[0] = Boolean.valueOf(true);
                    Logger.error(AudioParam.this, "Error sending audio for param " + key, e);
                }

                if (_streamListener != null)
                {
                    _streamListener.onStreamComplete(AudioParam.this, error);
                    _streamListener = null;
                }
                _commandContext = null;
            }
        };
        _sink.connectAudioSource(source);
    }

    @Override
    void send(CommandContext commandContext) throws Exception
    {
        commandContext.sendAudioParam(key);

        if (!_doneRecording)
        {
            _commandContext = commandContext;
        }
    }

    @Override
    void cancelStreaming()
    {
        Logger.debug(this, "cancelStreaming()");
        _sink.disconnectAudioSource();
        _streamListener = null;
    }

    @Override
    boolean stream(CommandContext commandContext, StreamListener streamListener) throws Exception {
        Logger.debug(this, "stream()");

        if (!_doneRecording)
        {
            _streamListener = streamListener;
        }

        // Stream the buffered chunks first, the buffer chunks are
        // received before the transaction gets started
        if (_audioBuffer.size() > 0)
        {
            for (AudioChunk chunk : _audioBuffer)
            {
                commandContext.sendAudioForParam(key, chunk, false);
            }
            _audioBuffer.clear();
        }

        // Send the last chunk, the recording could finish even before
        // the transaction gets started
        if (_doneRecording)
        {
            commandContext.sendAudioForParam(key, null, true);
            streamListener.onStreamComplete(AudioParam.this, _recordingDoneWithError[0].booleanValue());
            return _doneRecording;
        }

        if (_source != null)
        {
            int available = _source.getChunksAvailableForSink(_sink);
            while (available-- > 0)
            {
                commandContext.sendAudioForParam(key, _source.getAudioChunkForSink(_sink), _doneRecording && available == 0);
            }
        }

        // TODO: should we always call the listener, even when we're already done?
        // How else can we return an error code?
        return _doneRecording;
    }

    @Override
    boolean isStreaming() {
        return _streamListener != null;
    }
}
