package com.nuance.dragon.toolkit.cloudservices;

import java.util.Enumeration;
import java.util.Hashtable;

import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines;
import com.nuance.nmsp.client2.sdk.components.core.internal.calllog.CalllogImpl;

/**
 * Represents a configuration for {@link CloudConfig}. It contains the basic calllog
 * information including diableCalllog, calllog file size, retention days, root id parameters.
 */
public class CalllogConfig implements NMTConfig
{
    private Hashtable<String, Object> _parameters = new Hashtable<String, Object>();

    /**
     * Make a CalllogConfig object.
     *
     * @param disableCalllog Enable or Disable Call log feature
     *                       Disable Call log by setting disableCalllog = true.
     *                       Enable Call log by setting disableCalllog = false.
     */
    public CalllogConfig(boolean disableCalllog)
    {
        if (disableCalllog) {
            _parameters.put(NMSPDefines.NMSP_DEFINES_CALLLOG_DISABLE, "TRUE".getBytes());
        }
        else {
            _parameters.put(NMSPDefines.NMSP_DEFINES_CALLLOG_DISABLE, "FALSE".getBytes());
        }
    }

    /**
     * Make a CalllogConfig object.
     *
     * @param disableCalllog Enable or Disable Call log feature
     *                       Disable Call log by setting disableCalllog = true.
     *                       Enable Call log by setting disableCalllog = false.
     * @param chunkSize      The call log data callback will be called when internal buffer size
     *                       exceeding the chunk size. The chunk size is 100K by default, the minimum
     *                       chunk size is 20K and the maximum chunk size is 150K.
     * @param retentionDays  Specifies the suggested number of days for preserving the call log data in the cloud.
     *                       The effective data retention period is subject to a number of conditions on the server
     *                       side and may be automatically adjusted to a higher or lower value than the specified one.
     *
     *                       The value of �1 indicates no specific value for the data retention
     *                       and the server default value will be used. Other negative values or 0 is not allowed.
     *                       The default value is -1.
     * @param rootId         Input RootId.
     *                       If RootId is null, a new root id will be generated.
     *                       If RootId is not null, RootId will be used as root session event id.
     *
     */
    public CalllogConfig(boolean disableCalllog, int chunkSize, int retentionDays, String rootId)
    {
        Checker.checkArgForCondition("chunkSize",
                                     "equal or greater than " + CalllogImpl.CALLLOG_CHUNK_SIZE_MINIMUM + " but less than " + CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM,
                                     (chunkSize >= CalllogImpl.CALLLOG_CHUNK_SIZE_MINIMUM && chunkSize < CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM));

        Checker.checkArgForCondition("retentionDays",
                                     "greater than 0 or equal to -1",
                                     (retentionDays > 0 || retentionDays == -1));

        if (disableCalllog) {
            _parameters.put(NMSPDefines.NMSP_DEFINES_CALLLOG_DISABLE, "TRUE".getBytes());
        }
        else {
            _parameters.put(NMSPDefines.NMSP_DEFINES_CALLLOG_DISABLE, "FALSE".getBytes());
        }

        _parameters.put(NMSPDefines.NMSP_DEFINES_CALLLOG_CHUNK_SIZE, Integer.valueOf(chunkSize).toString().getBytes());
        _parameters.put(NMSPDefines.NMSP_DEFINES_CALLLOG_RETENTION_DAYS, Integer.valueOf(retentionDays).toString().getBytes());

        if (rootId != null) {
            _parameters.put(NMSPDefines.NMSP_DEFINES_CALLLOG_ROOT_ID, rootId.getBytes());
        }
    }

    @Override
    public int size() {
        return _parameters.size();
    }

    @Override
    public Enumeration<String> keys() {
        return _parameters.keys();
    }

    @Override
    public Object getValue(String key) {
        return _parameters.get(key);
    }
}
