package com.nuance.dragon.toolkit.cloudservices;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Represents a configuration for {@link CloudServices}. It contains the basic connection
 * information including the server IP and port, credentials, input and output
 * codec parameters.
 */
public class CloudConfig
{
    /**
     * Builder for creating instances of {@code CloudConfig}.
     */
    public static class Builder
    {
        private String _appName = null;
        private String _gwHost;
        private int    _gwPort;
        private String _asHost;
        private int    _asPort;
        private String _nmaid;
        private byte[] _appKey;
        private String _password;
        private String _deviceId = null;
        private AudioType _recorderCodec;
        private AudioType _playerCodec;
        private List<NMTConfig> _configs = null;
        private boolean _enableCalllog = false;
        private boolean _enableSSL = false;
        private boolean _enableSelfSignedCert = false;
        private String  _certSummary = null;
        private String  _certData = null;
        private boolean _enableMdsMessages = false;

        private Builder()
        {
        }

        /**
         * Create a {@link CloudConfig} builder object with the given parameters.
         *
         * @param gwHost            NMSP Gateway IP address or host name.
         * @param gwPort            NMSP Gateway port.
         * @param appId             Application ID (NMAID) provided by Nuance Communications, Inc.
         * @param appKey            Application Key provided by Nuance Communications, Inc.
         * @param recorderCodec     Recorder codec (input codec for recognition).
         * @param playerCodec       Player codec (output codec for TTS).
         */
        public Builder(String gwHost, int gwPort, String appId, byte[] appKey, AudioType recorderCodec, AudioType playerCodec)
        {
            this(gwHost, gwPort, appId, appKey, recorderCodec, playerCodec, false);
        }
        
        /**
         * Create a {@link CloudConfig} builder object with the given parameters.
         *
         * @param gwHost            NMSP Gateway IP address or host name.
         * @param gwPort            NMSP Gateway port.
         * @param appId             Application ID (NMAID) provided by Nuance Communications, Inc.
         * @param appKey            Application Key provided by Nuance Communications, Inc.
         * @param recorderCodec     Recorder codec (input codec for recognition).
         * @param playerCodec       Player codec (output codec for TTS).
         * @param enableSSL         Enable SSL connection
         */
        public Builder(String gwHost, int gwPort, String appId, byte[] appKey, AudioType recorderCodec, AudioType playerCodec, boolean enableSSL)
        {
            this();
            _gwHost = gwHost;
            _gwPort = gwPort;
            _nmaid = appId;
            _appKey = appKey;
            _password = null;
            _recorderCodec = recorderCodec;
            _playerCodec = playerCodec;
            _enableSSL = enableSSL;
        }
        
        /**
         * Create a {@link CloudConfig} builder object with the given parameters.
         *
         * @param gwHost            NMSP Gateway IP address or host name.
         * @param gwPort            NMSP Gateway port.
         * @param asHost            NMSP Authorization Server IP address or host name.
         * @param asPort            NMSP Authorization Server port.
         * @param nmaid             NMAID provided by Nuance Communications, Inc.
         * @param password          Password provided by Nuance Communications, Inc.
         * @param recorderCodec     Recorder codec (input codec for recognition).
         * @param playerCodec       Player codec (output codec for TTS).
         */
        public Builder(String gwHost, int gwPort, String asHost, int asPort, String nmaid, String password, AudioType recorderCodec, AudioType playerCodec)
        {
            this();
            _gwHost = gwHost;
            _gwPort = gwPort;
            _asHost = asHost;
            _asPort = asPort;
            _nmaid = nmaid;
            _appKey = null;
            _password = password;
            _recorderCodec = recorderCodec;
            _playerCodec = playerCodec;
            _enableSSL = true;
        }

        /**
         * Set app name.
         * @param appName The NMT client / application name.
         * @return builder object itself.
         */
        public Builder setAppName(String appName)
        {
            _appName = appName;
            return this;
        }

        /**
         * Set device id.
         * @param deviceId The device ID, or null to generate a random one for the current user.
         * @return builder object itself.
         */
        public Builder setDeviceId(String deviceId)
        {
            _deviceId = deviceId;
            return this;
        }

        /**
         * Enable calllog. Default is disabled.
         * @return builder object itself.
         */
        public Builder enableCalllog()
        {
            _enableCalllog = true;
            return this;
        }

        /**
         * Set SSL settings
         * @param enableSelfSignedCert Allow self signed server certificate
         * @param certSummary Certificate summary (common name) information.
         * @param certData Certificate Data information encoded in base64 string
         * @return builder object itself.
         */
        public Builder setSslSettings(boolean enableSelfSignedCert, String certSummary, String certData)
        {
            _enableSelfSignedCert = enableSelfSignedCert;
            _certSummary = certSummary;
            _certData = certData;
            return this;
        }

        /**
         * Allow self signed server certificate. Same effective as setSslSettings(true, null, null).
         * @return builder object itself.
         */
        public Builder allowSslSelfSigned()
        {
            return setSslSettings(true, null, null);
        }

        /**
         * Enable to send and receive MDS messages.
         * @return builder object itself.
         */
        public Builder enableMdsMessages()
        {
            _enableMdsMessages = true;
            return this;
        }

        /**
         * Set NMT configuration list.
         * @param configs NMT configuration list.
         * @return builder object itself.
         */
        public Builder setConfigs(List<NMTConfig> configs)
        {
            _configs = configs;
            return this;
        }

        /**
         * Build the CloudConfig object.
         * @return the built CloudConfig object
         */
        public CloudConfig build()
        {
            ArrayList<NMTConfig> configs = null;
            if (_enableCalllog || _enableSSL || _enableMdsMessages)
            {
                configs = new ArrayList<NMTConfig>();
                if (_configs != null)
                    configs.addAll(_configs);

                // TODO there are already existing config objects inside, shall we remove them first?

                if (_enableCalllog)
                    configs.add(new CalllogConfig(false));
                if (_enableSSL)
                    configs.add(new SSLConfig(_enableSelfSignedCert, _certSummary, _certData));
                if (_enableMdsMessages)
                    configs.add(new MdsMessageConfig(true));
            }
            if (configs != null)
                _configs = configs;

            return new CloudConfig(this);
        }
    }

    public final String appName;
    public final String gwHost;
    public final int gwPort;
    public final String asHost;
    public final int asPort;
    public final String nmaid;
    public final byte[] appKey;
    public final String password;
    public final String deviceId;
    public final AudioType recorderCodec;
    public final AudioType playerCodec;
    public final List<NMTConfig> configs;
    public final boolean newCop;
    public final boolean enableSelfSignedCert;
    public final String  certSummary;
    public final String  certData; 

    /**
     * Create a {@link CloudConfig} object from the builder.
     * @param builder The builder object.
     */
    protected CloudConfig(Builder builder)
    {
        this(builder._appName, builder._gwHost, builder._gwPort, builder._asHost, builder._asPort, builder._nmaid, builder._appKey, builder._password, builder._configs, builder._deviceId, builder._recorderCodec, builder._playerCodec, builder._enableSelfSignedCert, builder._certSummary, builder._certData);
    }

    /**
     * Create a {@link CloudConfig} object with the given parameters {@link Builder}
     * 
     * @deprecated  Use {@link #CloudConfig(Builder)} instead.
     */
    @Deprecated
    protected CloudConfig(String appName, String gwHost, int gwPort, String nmaid, byte[] appKey, List<NMTConfig> configs, String deviceId, AudioType recorderCodec, AudioType playerCodec)
    {
        this(appName, gwHost, gwPort, null, -1, nmaid, appKey, null, configs, deviceId, recorderCodec, playerCodec, false, null, null);
    }
    
    /**
     * Create a {@link CloudConfig} object with the given parameters {@link Builder}
     * 
     * @deprecated  Use {@link #CloudConfig(Builder)} instead.
     */
    @Deprecated
    protected CloudConfig(String appName, String gwHost, int gwPort, String asHost, int asPort, String nmaid, String password, List<NMTConfig> configs, String deviceId, AudioType recorderCodec, AudioType playerCodec)
    {
        this(appName, gwHost, gwPort, asHost, asPort, nmaid, null, password, configs, deviceId, recorderCodec, playerCodec, false, null, null);
    }
    
    /**
     * Create a {@link CloudConfig} object with the given parameters {@link Builder}
     */
    private CloudConfig(String appName, String gwHost, int gwPort, String asHost, int asPort, String nmaid, byte[] appKey, String password, List<NMTConfig> configs, String deviceId, AudioType recorderCodec, AudioType playerCodec, boolean enableSelfSignedCert, String certSummary, String certData)
    {
        Checker.checkStringArgForNullOrEmpty("nmaid", nmaid);
        Checker.checkStringArgForNullOrEmpty("gwHost", gwHost);
        Checker.checkArgForCondition("gwPort", "greater than 0", (gwPort > 0));
        if (appKey != null)
        {
            Checker.checkArgForNull("appKey", appKey);
        } else
        {
            Checker.checkStringArgForNullOrEmpty("asHost", asHost);
            Checker.checkArgForCondition("asPort", "greater than 0", (asPort > 0));
            Checker.checkStringArgForNullOrEmpty("password", password);
        }
        Checker.checkArgForNull("recorderCodec", recorderCodec);
        Checker.checkArgForNull("playerCodec", playerCodec);

        this.appName = appName;
        this.gwHost = gwHost;
        this.gwPort = gwPort;
        this.asHost = asHost;
        this.asPort = asPort;
        this.configs = configs;
        this.nmaid = nmaid;
        this.appKey = appKey;
        this.password = password;
        this.deviceId = deviceId;
        this.recorderCodec = recorderCodec;
        this.playerCodec = playerCodec;
        this.newCop = password != null;
        this.enableSelfSignedCert = enableSelfSignedCert;
        this.certSummary = certSummary;
        this.certData = certData;
    }
}
