package com.nuance.dragon.toolkit.cloudservices;

import com.nuance.dragon.toolkit.core.calllog.CalllogManager;
import com.nuance.dragon.toolkit.core.data.Data.Dictionary;
import com.nuance.dragon.toolkit.oem.api.NMTContext;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.nmsp.client2.sdk.components.general.ResourceUtil;

/**
 * Module that represents the connection(s) to the NMAS. This module manages a
 * queue of asynchronous transactions (NMAS commands) that are executed one at
 * a time. When executing a transaction, this module will send the contents of
 * the transaction to Nuance's speech servers via the NMAS components of the
 * NMSP SDK.
 * <p>
 * To create a CloudServices, call {@link #createCloudServices}.
 */
public abstract class CloudServices
{
    /**
     * Possible connection states.
     */
    public enum ConnectionState
    {
        /** Currently connected to the server */
        CONNECTED,
        /** Not connected to the server */
        DISCONNECTED
    }

    /**
     * Retrieves third party login info
     */
    public interface ThirdPartyLoginRetriever
    {
        /*
         * @return The newest ThirdPartyLogin object,
         *         could be null to reuse the previously-returned object,
         *         but cannot be null for the first invocation.
         */
        ThirdPartyLogin fetchThirdPartyLogin();
    }

    /**
     * MDS message listener.
     */
    public interface MdsMessageListener
    {
        /**
         * Upon receiving a new MDS message.
         * @param id the message id
         * @param msg the message payload in Dictionary format
         */
        void onMessageReceived(int id, Dictionary msg);
    }

    /**
     * CloudServices listener callbacks
     */
    public interface CloudServicesListener
    {
        /**
         * It is called to indicate that release operation is completed and
         * all the internal services have been stopped properly.
         */
        void onReleaseCompleted();
    }

    /**
     * Set customized CloudServices listener.
     * @param listener
     */
    public abstract void setCloudServicesListener(CloudServicesListener listener);

    /**
     * Disconnects from the gateway. Does all cleanup. Cancels the active
     * transaction. Delegate called when cleanup is complete and we can
     * guarantee that no further callbacks will be called.
     */
    public abstract void release();

    /**
     * Same as release() but runs synchronously if called from a thread other than the main thread.
     */
    public abstract void releaseAndWait();

    /**
     * Gets the most recent session ID.
     *
     * @return The session ID, or NULL if no session has been established yet.
     */
    public abstract String getSessionID();

    /**
     * Gets the unique ID.
     *
     * @return The unique ID, or NULL in case of an error.
     */
    public abstract String getUniqueID();

    /**
     * Bit field mask indicating all 4 IDs (IMEI, Serial, MAC and Android ID) are to be included
     * to generate the NuanceID
     */
    public static final int    BM_ALL         = ResourceUtil.BM_ALL;
    /**
     * Bit field mask indicating IMEI hash to be included to generate the NuanceID
     */
    public static final int    BM_IMEI        = ResourceUtil.BM_IMEI;
    /**
     * Bit field mask indicating serial hash to be included to generate the NuanceID
     */
    public static final int    BM_SERIAL      = ResourceUtil.BM_SERIAL;
    /**
     * Bit field mask indicating MAC hash to be included to generate the NuanceID
     */
    public static final int    BM_MAC         = ResourceUtil.BM_MAC;
    /**
     * Bit field mask indicating Android ID hash to be included to generate the NuanceID
     */
    public static final int    BM_ANDROIDID   = ResourceUtil.BM_ANDROIDID;

    /**
     * Gets the Nuance ID
     * @param context a valid Android Context.
     * @param idMasks Bit field indicating which IDs are used to generate NuanceID.
     *
     * @return Nuance ID string
     **/
    public static final String getNuanceID(NMTContext context, int idMasks)
    {
        return ResourceUtil.getNuanceID(context.getNativeContext(), idMasks);
    }
    /**
     * Gets the Nuance ID including all ID masks
     * @param context a valid Android Context.
     *
     * @return Nuance ID string
     **/
    public static final String getNuanceID(NMTContext context)
    {
        return ResourceUtil.getNuanceID(context.getNativeContext());
    }

    /**
     * Gets the server-generated user ID.
     *
     * @return The user ID, or NULL if not yet connected.
     */
    public abstract String getUserID();

    /**
     * Gets the server-generated client ID.
     *
     * @return The client ID, or NULL if not yet registered.
     */
    public abstract String getClientID();

    /**
     * Retrieve the current connection state. This may be used for diagnostic
     * purposes. This is not especially useful information, because
     * CloudServices automatically connects when necessary.
     *
     * @return The current state which will never be NULL.
     */
    public abstract ConnectionState getConnectionState();

    /**
     * Explicitly reinitialize the CloudServices instance to hide the overhead
     * when processing the transactions issued later.
     */
    public abstract void warmUp();

    /**
     * Queue a transaction to be executed. The transaction will start executing
     * immediately if there is no transaction currently in progress. Otherwise,
     * the transaction will be queued. The transaction's position in the queue
     * depends on its priority, and the priority of the existing queued
     * transactions. This transaction will be placed after all existing
     * transactions with greater or equal priority.
     *
     * @param transaction (cannot be NULL)
     * @param priority (should be greater than 0)
     */
    public abstract void addTransaction(Transaction transaction, int priority);

    /**
     * Get the number of total transactions in the transaction queue, including
     * the transaction that is currently executing (if any).
     * @return The number of total transactions
     */
    public abstract int getTransactionCount();

    /**
     * Get the currently executing transaction.
     *
     * @return The current transaction, or NULL if there is no transaction.
     */
    public abstract Transaction currentTransaction();

    /**
     * Connect to the server explicitly if not connected yet.
     */
    public abstract void connect();

    /**
     * Get the context object.
     */
    public abstract NMTContext getContext();

    /**
     * @return main thread handler
     */
    public abstract NMTHandler getMainThreadHandler();

    /**
     * Activate the CalllogManager instance.
     *
     * @param calllogManager CalllogManager instance
     */
    public abstract void activateCalllogManager(CalllogManager calllogManager);

    /**
     * Create a CloudServices instance
     *
     * @param appContext      The application context object.
     * @param config          The cloud configuration.
     * @param loginRetriever  The third party login retriever, could be null for anonymous login with limited features.
     * @param resetRegistration Whether to reset the cached registration received previously.
     * @param mdsMsgListener  The MDA message listener.
     */
    public static final CloudServices createCloudServices(NMTContext appContext, CloudConfig config, ThirdPartyLoginRetriever loginRetriever, boolean resetRegistration, MdsMessageListener mdsMsgListener, NMTHandler mainHandler)
    {
        return new CloudServicesImpl(appContext, config, loginRetriever, resetRegistration, mdsMsgListener, mainHandler);
    }

    /**
     * Create a CloudServices instance
     *
     * @param appContext      The application context object.
     * @param config          The cloud configuration.
     * @param loginRetriever  The third party login retriever, could be null for anonymous login with limited features.
     * @param resetRegistration Whether to reset the cached registration received previously.
     * @param mdsMsgListener  The MDA message listener.
     */
    public static final CloudServices createCloudServices(NMTContext appContext, CloudConfig config, ThirdPartyLoginRetriever loginRetriever, boolean resetRegistration, MdsMessageListener mdsMsgListener)
    {
        return CloudServices.createCloudServices(appContext, config, loginRetriever, resetRegistration, mdsMsgListener, null);
    }

    /**
     * Create a CloudServices instance
     *
     * @param appContext      The application context object.
     * @param config          The cloud configuration.
     * @param loginRetriever  The third party login retriever, could be null for anonymous login with limited features.
     * @param resetRegistration Whether to reset the cached registration received previously.
     */
    public static final CloudServices createCloudServices(NMTContext appContext, CloudConfig config, ThirdPartyLoginRetriever loginRetriever, boolean resetRegistration)
    {
        return CloudServices.createCloudServices(appContext, config, loginRetriever, resetRegistration, null);
    }

    /**
     * Create a CloudServices instance
     *
     * @param appContext      The application context object.
     * @param config          The cloud configuration.
     * @param loginRetriever  The third party login retriever, could be null for anonymous login with limited features.
     * @param mdsMsgListener  The MDA message listener.
     */
    public static final CloudServices createCloudServices(NMTContext appContext, CloudConfig config, ThirdPartyLoginRetriever loginRetriever, MdsMessageListener mdsMsgListener)
    {
        return CloudServices.createCloudServices(appContext, config, loginRetriever, false, mdsMsgListener);
    }

    /**
     * Create a CloudServices instance
     *
     * @param appContext      The application context object.
     * @param config          The cloud configuration.
     * @param loginRetriever  The third party login retriever, could be null for anonymous login with limited features.
     */
    public static final CloudServices createCloudServices(NMTContext appContext, CloudConfig config, ThirdPartyLoginRetriever loginRetriever)
    {
        return CloudServices.createCloudServices(appContext, config, loginRetriever, false);
    }

    /**
     * Create a CloudServices instance anonymously
     *
     * @param appContext      The application context object.
     * @param config          The cloud configuration.
     */
    public static final CloudServices createCloudServices(NMTContext appContext, CloudConfig config)
    {
        return CloudServices.createCloudServices(appContext, config, null);
    }

    /**
     * @return A cloud services 'impersonator'. The impersonator doesn't do
     * any real connecting--it is a very light 'fake' cloud module that may be
     * useful for testing other modules.
     */
    public static final CloudServices createCloudImpersonator()
    {
        return new CloudServicesImpersonator();
    }

    /**
     * Get the account manager object.
     * @return account manager object.
     */
    abstract AccountManager getAccountManager();

    /**
     * acquire the singleton nmsp resource object and register a listener with it
     * @param session a valid session object
     * @param resourcelistener the resource listener to register
     * @return
     */
    abstract Object acquireNmspResource(ASSession session, Object resourcelistener);

    /**
     * release the acquired singleton nmsp resource object
     */
    abstract void releaseNmspResource();
}
