package com.nuance.dragon.toolkit.cloudservices;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Queue;

import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.impl.HandlerOem;

/**
 * Base class for cloud services implementations.
 */
abstract class CloudServicesBase extends CloudServices
{
    protected final NMTHandler         _mainThreadHandler;
    private final Queue<Transaction>   _transactionQueue;
    private Transaction                _currentTransaction;
    private boolean                    _isReleased;

    CloudServicesBase(NMTHandler mainThreadHandler)
    {
        Logger.debug(this, "CloudServicesBase() mainThreadHandler:" + mainThreadHandler);
        _mainThreadHandler = mainThreadHandler == null ? new HandlerOem() : mainThreadHandler;
//        _transactionQueue = new PriorityQueue<Transaction>(10, new TransactionComparator());
        _transactionQueue = new LinkedList<Transaction>();
    }

    @Override
    public void release()
    {
        Logger.debug(this, "release()");
        Checker.checkState(this, !_isReleased);

        _isReleased = true;

        ArrayList<Transaction> transactions = new ArrayList<Transaction>(_transactionQueue.size());
        transactions.addAll(_transactionQueue);
        _transactionQueue.clear();

        if (_currentTransaction != null)
        {
            _currentTransaction.cancel();
            _currentTransaction = null;
        }

        for (Transaction t : transactions)
        {
            t.cancel();
        }
    }

    @Override
    public void addTransaction(final Transaction transaction, final int priority)
    {
        Checker.checkArgForNull("transaction", transaction);
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                Checker.checkState(this, !_isReleased);
                
                transaction.transactionAdded();
                transaction.setPriority(priority);
                addTransactionWithPriority(transaction);
        //        Object[] objs = _transactionQueue.toArray();
        //        for (int i = 0; i < objs.length; i++)
        //        {
        //            Logger.debug(CloudServicesBase.this, "transaction queue index [" + i + "] elem [" + objs[i] + "]");
        //        }
                transaction.monitorTransaction(new Transaction.InternalListener()
                {
                    @Override
                    public void onReadyToStart(Transaction t)
                    {
                        if (!_transactionQueue.isEmpty() && _transactionQueue.peek() == t)
                            startNextTransaction();
                    }
        
                    @Override
                    public void onFinished(Transaction t)
                    {
                        t.monitorTransaction(null);
        //                _transactionQueue.remove(t); WLWL
                        if (_currentTransaction == t)
                            _currentTransaction = null;
                        else
                            _transactionQueue.remove(t);
        
        //                Logger.debug(CloudServicesBase.this, "onFinished Transaction [" + t + "]");
                        startNextTransaction();
                    }
                });
        
                startNextTransaction();
            }
        });
    }

    private void addTransactionWithPriority(Transaction tran)
    {
        LinkedList<Transaction> queue = (LinkedList<Transaction>) _transactionQueue;
        for (int i = 0; i < queue.size(); i++)
        {
            Transaction curTran = queue.get(i);
            if (tran.getPriority() > curTran.getPriority())
            {
                queue.add(i, tran);
                return;
            }
        }
        queue.add(tran);
    }

    @Override
    public int getTransactionCount()
    {
        int count = _transactionQueue.size();
        if (_currentTransaction != null)
            count++;
        return count;
    }

    @Override
    public Transaction currentTransaction()
    {
        return _currentTransaction;
    }

    /**
     * Call this whenever the connection state changes. It may trigger the next
     * transaction to start, if it was waiting for {@link #canStartTransaction}
     * to return true.
     */
    protected void connectionStateChanged()
    {
        startNextTransaction();
    }

    /**
     * Determines whether a transaction can be started. If this returns false,
     * transactions will remain in the queue until
     * {@link #connectionStateChanged} is called.
     *
     * @return true if a transaction can be started.
     */
    protected abstract boolean canStartTransaction();

    /**
     * Return a command context, which will be used by the transaction to send
     * things to the NMSP SDK (or an impersonator).
     *
     * @param t The transaction that this context is for
     * @return The command context.
     */
    protected abstract CommandContext createCommandContext(Transaction t);
    
    /** Starts the next transaction, if possible. Otherwise does nothing. */
    private void startNextTransaction()
    {
        if (_currentTransaction == null && canStartTransaction() && !_transactionQueue.isEmpty())
        {
            if (_transactionQueue.peek().isReadyToStart())
            {
                _currentTransaction = _transactionQueue.remove();
                CommandContext commandContext = createCommandContext(_currentTransaction);
                _currentTransaction.getListener().onTransactionStarted(_currentTransaction);
                _currentTransaction.start(commandContext);
    //            Logger.debug(CloudServicesBase.this, "startNextTransaction ... _currentTransaction [" + _currentTransaction + "]");
            }
        }
    }
    
    @Override
    public NMTHandler getMainThreadHandler()
    {
        return _mainThreadHandler;
    }
}
