package com.nuance.dragon.toolkit.cloudservices;

import java.util.List;
import java.util.Locale;

import com.nuance.dragon.toolkit.Build;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.language.GrammarLanguage;
import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.NMTContext;
import com.nuance.nmsp.client2.sdk.components.general.ResourceUtil;

/**
 * A Configure class contains all the cloud service settings including
 * server ip and port, user id and key, playback and record codec ...
 */
final class CloudServicesConfig
{
    private static final String UNKNOWN = "unknown";
    private final String        _appName;
    private final String        _gwHost;
    private final short         _gwPort;
    private final String        _asHost;
    private final short         _asPort;
    private final String        _nmaid;
    private final byte[]        _appKey;
    private final String        _password;
    private final NMTContext    _appContext;
    private final String        _deviceId;
    private final AudioType     _recorderCodec;
    private final AudioType     _playerCodec;
    private final String        _platform;
    private final List<NMTConfig> _configs;
    private final boolean       _newCop;

    /**
     * @param appContext     The Android Application Context object, used internally as the main application context.
     * @param appName        The application name.
     * @param host           NMSP Gateway IP address or hostname
     * @param port           NMSP Gateway port
     * @param appId          application ID provided by Nuance Communications, Inc.
     * @param appKey         application private key provided by Nuance Communications, Inc.
     * @param deviceId       device UID provided by client application i.e. IMEI, phone #, Google Play ID, etc.
     * @param recorderCodec  input codec
     * @param playerCodec    output codec
     * @param newCop         whether it's new COP version
     */
    public CloudServicesConfig(NMTContext appContext, String appName, String gwHost, int gwPort, String asHost, int asPort, String nmaid, byte[] appKey, String password, String deviceId,
            AudioType recorderCodec, AudioType playerCodec, List<NMTConfig> configs, boolean newCop)
    {
        _appContext = appContext;
        _appName = (appName != null && appName.length() > 0) ? appName : appName(appContext);
        _gwHost = gwHost;
        _gwPort = (short) gwPort;
        _asHost = asHost;
        _asPort = (short) asPort;
        _nmaid = nmaid;
        _appKey = appKey;
        _password = password;
        _deviceId = (deviceId != null && deviceId.length() > 0) ? deviceId : uid(appContext);
        _recorderCodec = recorderCodec;
        _playerCodec = playerCodec;
        _platform = platformName(appContext);
        _configs = configs;
        _newCop = newCop;
    }

    /**
     * @return the host application's name
     */
    public String applicationName()
    {
        return _appName;
    }

    /**
     * @return the application ID provided by Nuance Communications, Inc.
     */
    public String nmaid()
    {
        return _nmaid;
    }

    /**
     * @return application private key provided by Nuance Communications, Inc.
     */
    public String password()
    {
        return _password;
    }
    
    /**
     * @return application private key provided by Nuance Communications, Inc.
     */
    public byte[] appKey()
    {
        return _appKey;
    }

    /**
     * @return NMSP Gateway IP address or hostname
     */
    public String gwHost()
    {
        return _gwHost;
    }

    /**
     * @return NMSP Gateway port
     */
    public short gwPort()
    {
        return _gwPort;
    }

    /**
     * @return Authorization Server IP address or hostname
     */
    public String asHost()
    {
        return _asHost;
    }

    /**
     * @return Authorization Server port
     */
    public short asPort()
    {
        return _asPort;
    }

    /**
     * @return the version of the client, used for server logging purpose
     */
    public String clientVersion()
    {
        return Build.VERSION;
    }

    /**
     * @return the name of the wireless carrier name, "carrier" will be returned
     *         if there is no actual active carrier
     */
    public String carrier()
    {
        String carrier = Factory.getDeviceInfo(_appContext).getCarrierName();
        if (carrier == null || carrier.length() == 0)
        {
            carrier = "carrier";
        }
        return carrier;
    }

    /**
     * @return the phone's local language
     */
    public String defaultLanguage()
    {
        String lang = GrammarLanguage.getGrammarLanguage(Locale.getDefault());
        return lang == null ? "xxx-XXX" : lang;
    }

    /**
     * @return the device model number
     */
    public String model()
    {
        String model = ResourceUtil.getDeviceModel();
        if (model != null && model.length() > 0)
        {
            return model;
        }
        return UNKNOWN;
    }

    /**
     * @return the device OS type along with its version number
     */
    public String os()
    {
        return ResourceUtil.getOSType() + " " + ResourceUtil.getOSVersion();
    }
    
    /**
     * @return the device OS type
     */
    public String osType()
    {
        return ResourceUtil.getOSType();
    }
    
    /**
     * @return the device OS version number
     */
    public String osVersion()
    {
        return ResourceUtil.getOSVersion();
    }

    /**
     * @return the device locale string
     */
    public String locale()
    {
        Locale l = Locale.getDefault();
        if (l != null)
        {
            String c = l.getCountry();
            if (c != null && c.length() > 0)
            {
                return c;
            }
        }
        return "xxx-XXX";
    }

    /**
     * @return the type of the network, for example "WIFI" or "MOBILE".
     */
    public String networkType()
    {
        String name = Factory.getDeviceInfo(_appContext).getNetworkType();
        if (name != null && name.length() > 0)
        {
            return name;
        }
        return UNKNOWN;
    }

    /**
     * @return the uid of the device, could be IMEI number for example
     */
    public String uid()
    {
        return _deviceId == null ? UNKNOWN : _deviceId;
    }
    
    /**
     * @return whether it's new cop
     */
    public boolean newCop()
    {
        return _newCop;
    }

    private static String appName(NMTContext context)
    {
        String appName = Factory.getDeviceInfo(context).getAppName();

        return appName == null ? UNKNOWN : appName;
    }

    private static String uid(NMTContext context)
    {
        String uid = Factory.getDeviceInfo(context).getUid();
        if (uid != null && uid.length() == 0)
            uid = null;
        return uid;
    }

    private static String platformName(NMTContext context)
    {
        return Factory.getDeviceInfo(context).getPlatform();
    }

    /**
     * @return the host application's package name
     */
    public byte[] packageName()
    {
        return Factory.getDeviceInfo(_appContext).getPackageName();
    }

    /**
     * @return the config parameters
     */
    public List<NMTConfig> configs()
    {
        return _configs;
    }

    /**
     * @return the corresponding NMSP SDK's recorder codec
     */
    public Object nmspRecorderCodec()
    {
        return getNMSPCodec(_recorderCodec, _newCop);
    }

    /**
     * @return the corresponding NMSP SDK's player codec
     */
    public Object nmspPlayerCodec()
    {
        return getNMSPCodec(_playerCodec, _newCop);
    }
    
    /**
     * @return the recorder audio type
     */
    public AudioType recorderCodec()
    {
        return _recorderCodec;
    }

    /**
     * @return the player audio type
     */
    public AudioType playerCodec()
    {
        return _playerCodec;
    }

    /**
     * @return the platform name
     */
    public String platform()
    {
        return _platform;
    }

    private static Object getNMSPCodec(AudioType audioType, boolean newCop)
    {
        if (audioType.encoding == AudioType.Encoding.PCM_16)
        {
            if (audioType.frequency == AudioType.Frequency.FREQ_8KHZ)
            {
                return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.PCM_16_8K : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.PCM_16_8K;
            } else if (audioType.frequency == AudioType.Frequency.FREQ_16KHZ)
            {
                return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.PCM_16_16K : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.PCM_16_16K;
            }
        } else if (audioType.encoding == AudioType.Encoding.SPEEX)
        {
            if (audioType.frequency == AudioType.Frequency.FREQ_8KHZ)
            {
                return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.SPEEX_8K : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.SPEEX_8K;
            } else if (audioType.frequency == AudioType.Frequency.FREQ_16KHZ)
            {
                return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.SPEEX_16K : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.SPEEX_16K;
            }
        } else if (audioType.encoding == AudioType.Encoding.OPUS)
        {
            if (audioType.frequency == AudioType.Frequency.FREQ_8KHZ)
            {
                return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.OPUS_8K : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.OPUS_8K;
            } else if (audioType.frequency == AudioType.Frequency.FREQ_16KHZ)
            {
                return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.OPUS_16K : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.OPUS_16K;
            }
        } else if (audioType.encoding == AudioType.Encoding.ULAW)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.ULAW : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.ULAW;
        } else if (audioType.encoding == AudioType.Encoding.GSM_FR)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MS_GSM_FR : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MS_GSM_FR;
        } else if (audioType.encoding == AudioType.Encoding.AMR0)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.AMR_00 : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.AMR_00;
        } else if (audioType.encoding == AudioType.Encoding.AMR1)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.AMR_01 : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.AMR_01;
        } else if (audioType.encoding == AudioType.Encoding.AMR2)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.AMR_02 : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.AMR_02;
        } else if (audioType.encoding == AudioType.Encoding.AMR3)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.AMR_03 : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.AMR_03;
        } else if (audioType.encoding == AudioType.Encoding.AMR4)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.AMR_04 : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.AMR_04;
        } else if (audioType.encoding == AudioType.Encoding.AMR5)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.AMR_05 : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.AMR_05;
        } else if (audioType.encoding == AudioType.Encoding.AMR6)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.AMR_06 : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.AMR_06;
        } else if (audioType.encoding == AudioType.Encoding.AMR7)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.AMR_07 : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.AMR_07;
        } else if (audioType.encoding == AudioType.Encoding.MP3_8KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_8KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_8KBPS;
        } else if (audioType.encoding == AudioType.Encoding.MP3_16KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_16KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_16KBPS;
        } else if (audioType.encoding == AudioType.Encoding.MP3_24KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_24KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_24KBPS;
        } else if (audioType.encoding == AudioType.Encoding.MP3_32KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_32KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_32KBPS;
        } else if (audioType.encoding == AudioType.Encoding.MP3_40KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_40KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_40KBPS;
        } else if (audioType.encoding == AudioType.Encoding.MP3_48KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_48KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_48KBPS;
        } else if (audioType.encoding == AudioType.Encoding.MP3_56KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_56KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_56KBPS;
        } else if (audioType.encoding == AudioType.Encoding.MP3_64KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_64KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_64KBPS;
        } else if (audioType.encoding == AudioType.Encoding.MP3_80KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_80KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_80KBPS;
        } else if (audioType.encoding == AudioType.Encoding.MP3_96KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_96KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_96KBPS;
        } else if (audioType.encoding == AudioType.Encoding.MP3_112KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_112KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_112KBPS;
        } else if (audioType.encoding == AudioType.Encoding.MP3_128KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_128KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_128KBPS;
        } else if (audioType.encoding == AudioType.Encoding.MP3_144KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_144KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_144KBPS;
        } else if (audioType.encoding == AudioType.Encoding.MP3_160KBPS)
        {
            return !newCop ? com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec.MP3_160KBPS : com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec.MP3_160KBPS;
        }
        return null;
    }
}
