package com.nuance.dragon.toolkit.cloudservices;


import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.core.calllog.CalllogManager;
import com.nuance.dragon.toolkit.core.calllog.CalllogManager.CalllogDataListener;
import com.nuance.dragon.toolkit.core.calllog.CalllogSender;
import com.nuance.dragon.toolkit.core.calllog.CalllogSender.SenderListener;
import com.nuance.dragon.toolkit.core.calllog.SessionEventBuilder;
import com.nuance.dragon.toolkit.core.calllog.internal.Calllog;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTContext;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.impl.HandlerOem;

/**
 * A Dummy CloudServices implementation to exercise the basic event flow
 */
public class CloudServicesImpersonator extends CloudServicesBase implements Calllog
{
    private enum Action
    {
        IDLE,
        CONNECT,
        DISCONNECT
    }

    private ConnectionState _state;
    private Action _pendingAction;
    private boolean _stopping = true;

    private final Runnable _fakeNetworkThread;
    private final NMTHandler _messageHandler;

    public CloudServicesImpersonator()
    {
        super(null);
        _pendingAction = Action.CONNECT;
        _stopping = false;
        _state = ConnectionState.DISCONNECTED;
        _messageHandler = new HandlerOem();

        _fakeNetworkThread = new Runnable()
        {
            @Override
            public void run()
            {
                if (_stopping == true)
                {
                    _stopping = false;
                    _pendingAction = Action.IDLE;
                    _state = ConnectionState.DISCONNECTED;
                    connectionStateChanged();
                    return;
                }

                if (_pendingAction == Action.CONNECT &&
                    _state != ConnectionState.CONNECTED)
                {
                    Logger.debug(this, "Connected!");
                    _state = ConnectionState.CONNECTED;
                    _pendingAction = Action.IDLE;
                    connectionStateChanged();
                }
                else if (_pendingAction == Action.DISCONNECT &&
                         _state != ConnectionState.DISCONNECTED)
                {
                    Logger.debug(this, "Disconnected!");
                    _state = ConnectionState.DISCONNECTED;
                    _pendingAction = Action.IDLE;
                    connectionStateChanged();
                }
                else if (_state == ConnectionState.CONNECTED)
                {
                    // Sit at connected, allowing the base class to drive
                    // the processing of transactions.
                }
                _messageHandler.postDelayed(_fakeNetworkThread, 200);
            }
        };
        _messageHandler.postDelayed(_fakeNetworkThread, 200);
    }

    @Override
    public void release()
    {
        super.release();
        _stopping = true;
    }

    @Override
    public void releaseAndWait()
    {
        // TODO
    }

    @Override
    protected boolean canStartTransaction()
    {
        return true;
    }

    @Override
    protected CommandContext createCommandContext(final Transaction t)
    {
        return new CommandContext()
        {
            private boolean _isEnded = false;
            private int _expectingAudio = 0;

            @Override
            public void initializeCommand(String command, Data.Dictionary customKeys, int timeout) throws Exception
            {
                Logger.debug(CloudServicesImpersonator.this, "Initializing command " + command);
            }

            @Override
            public void endCommand() throws Exception
            {
                Logger.debug(CloudServicesImpersonator.this, "Ending command");
                _isEnded = true;
                if (_expectingAudio == 0)
                    completeTransaction();
            }

            @Override
            public void sendTextParam(String name, String text) throws Exception
            {
                Logger.debug(CloudServicesImpersonator.this, "Sending text param " + name);
            }

            @Override
            public void sendDictParam(String name, Data.Dictionary dict) throws Exception
            {
                Logger.debug(CloudServicesImpersonator.this, "Sending dict param " + name);
            }

            @Override
            public void sendAudioParam(String name) throws Exception
            {
                Logger.debug(CloudServicesImpersonator.this, "Sending audio param " + name);
                _expectingAudio++;
            }

            @Override
            public void sendAudioForParam(String audioParamName, AudioChunk audio, boolean isFinal) throws Exception
            {
                Logger.debug(CloudServicesImpersonator.this, "Sending" + (isFinal ? " final" : "") + " audio for param " + audioParamName);
                if (isFinal)
                    _expectingAudio--;

                if (_expectingAudio == 0 && _isEnded)
                    completeTransaction();
            }

            private void completeTransaction()
            {
                t.handleResult(null, true);
            }

            @Override
            public void sendChoiceParam(String name, String choicename) throws Exception
            {
                Logger.debug(CloudServicesImpersonator.this, "Sending choice param " + name);
            }

            @Override
            public void sendDataParam(String name, byte[] data) throws Exception
            {
                Logger.debug(CloudServicesImpersonator.this, "Sending data param " + name);
            }

            @Override
            public void sendSeqChunkParam(String name, Data.Dictionary dict) throws Exception
            {
                Logger.debug(CloudServicesImpersonator.this, "Sending seq chunk param " + name);
            }

            @Override
            public void sendSeqEndParam(String name, Data.Dictionary dict) throws Exception
            {
                Logger.debug(CloudServicesImpersonator.this, "Sending seq end param " + name);
            }

            @Override
            public void sendSeqStartParam(String name, Data.Dictionary dict) throws Exception
            {
                Logger.debug(CloudServicesImpersonator.this, "Sending seq start param " + name);
            }

            @Override
            public void sendTTSParam(String name, Data.Dictionary dict, NMTAudioSink audioSink) throws Exception
            {
                Logger.debug(CloudServicesImpersonator.this, "Sending tts param " + name);
            }

            @Override
            public void freeResource()
            {
                Logger.debug(CloudServicesImpersonator.this, "free resource");
            }

            @Override
            public void getSession() {
                // TODO
            }

            @Override
            public boolean newCop() {
                // TODO Auto-generated method stub
                return false;
            }
        };
    }

    @Override
    public String getSessionID()
    {
        return "SESSIONID";
    }

    @Override
    public String getUniqueID()
    {
        return "UNIQUEID";
    }

    @Override
    public ConnectionState getConnectionState()
    {
        return _state;
    }

    @Override
    public void registerCalllogDataListener(CalllogDataListener calllogDataListener) {
    }

    @Override
    public SessionEventBuilder logAppEvent(String appName, String schemaVersion) {
        return null;
    }

    @Override
    public String getCalllogRootParentId() {
        return null;
    }

    @Override
    public void flushCallLogData()
    {
    }

    @Override
    protected AccountManager getAccountManager() {
        // TODO
        return null;
    }

    @Override
    public String getUserID() {
        // TODO
        return null;
    }

    @Override
    public String getClientID() {
        // TODO
        return null;
    }

    @Override
    public void warmUp() {
        // TODO
    }

    @Override
    public CalllogSender createCalllogSender(SenderListener senderListener) {
        return null;
    }

    @Override
    public void connect() {
        // TODO
    }

    @Override
    public NMTContext getContext() {
        // TODO
        return null;
    }

    @Override
    public void activateCalllogManager(CalllogManager calllogManager)
    {
        calllogManager.attachCalllogImpl(this);
    }

    @Override
    Object acquireNmspResource(ASSession session, Object resourcelistener) {
        // TODO
        return null;
    }

    @Override
    void releaseNmspResource() {
        // TODO
    }

    @Override
    public void setCloudServicesListener(CloudServicesListener listener) {
        // TODO Auto-generated method stub

    }
}
