package com.nuance.dragon.toolkit.cloudservices;

import java.util.Enumeration;
import java.util.List;
import java.util.Vector;

import com.nuance.dragon.toolkit.cloudservices.util.PdxConverter;
import com.nuance.dragon.toolkit.core.calllog.CalllogManager;
import com.nuance.dragon.toolkit.core.calllog.CalllogManager.CalllogDataListener;
import com.nuance.dragon.toolkit.core.calllog.CalllogSender;
import com.nuance.dragon.toolkit.core.calllog.CalllogSender.SenderListener;
import com.nuance.dragon.toolkit.core.calllog.SessionEventBuilder;
import com.nuance.dragon.toolkit.core.calllog.internal.Calllog;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTContext;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.impl.FileManagerOem;
import com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines;

/**
 * The actual implementation of the CloudServices Interface, it have the NMSP Manager and Nmas
 * resource embedded inside. It manages a transaction queue which can do different NMAS transactions
 * sequentially. The result will be posted via the transaction listener to each transaction object.
 * <p>
 * Don't create this directly. Use {@link com.nuance.dragon.toolkit.cloudservices.CloudServices#createCloudServices}.
 */
class CloudServicesImpl extends CloudServicesBase implements Calllog
{
    // SDK1
    private final com.nuance.nmsp.client.sdk.components.resource.common.Manager             _mgr1;
    private final com.nuance.nmsp.client.sdk.components.resource.common.ManagerListener     _mgrListener1;
    private final com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResourceListener  _rscListener1;
    private com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResource                _rsc1;
    private com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResourceListener        _registeredRscListener1;
    // SDK2
    private final com.nuance.nmsp.client2.sdk.components.resource.common.Manager            _mgr2;
    private final com.nuance.nmsp.client2.sdk.components.resource.common.ManagerListener    _mgrListener2;
    private final com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResourceListener _rscListener2;
    private com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource               _rsc2;
    private com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResourceListener       _registeredRscListener2;
    private final MdsMessageListener                                                        _mdsListener;

    private ConnectionState     _state;
    private String              _sessionId;
    private final AccountManager _accManager;
    private final CloudServicesConfig _config;
    private boolean             _isValid;
    private CalllogDataListener _calllogDataListener;
    private boolean             _isCalllogDisabled = true;
    private NMTContext          _context;
    private CloudServicesListener _csListener = null;


    /**
     * Create a CloudServices object
     *
     * @param appContext      The application context object (cannot be NULL)
     * @param config          The cloud configuration (cannot be NULL)
     */
    CloudServicesImpl(NMTContext appContext, CloudConfig config, ThirdPartyLoginRetriever loginRetriever, boolean resetRegistation, MdsMessageListener mdsListener, NMTHandler mainThreadHandler)
    {
        super(mainThreadHandler);
        Checker.checkArgForNull("appContext", appContext);
        Checker.checkArgForNull("config", config);

        _config = new CloudServicesConfig(appContext.getAppContext(), config.appName, config.gwHost, config.gwPort, config.asHost, config.asPort, config.nmaid, config.appKey, config.password, config.deviceId, config.recorderCodec, config.playerCodec, config.configs, config.newCop);
        if (!_config.newCop())
        {
            _mgrListener1 = (com.nuance.nmsp.client.sdk.components.resource.common.ManagerListener)createNmspManagerListener();
            _mgrListener2 = null;
            _mgr1 = (com.nuance.nmsp.client.sdk.components.resource.common.Manager)createNmspManager();
            _mgr2 = null;
            _accManager = null;
            _rscListener1 = (com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResourceListener)createNmasResourceListener();
            _rscListener2 = null;
        }
        else
        {
            _mgrListener1 = null;
            _mgrListener2 = (com.nuance.nmsp.client2.sdk.components.resource.common.ManagerListener)createNmspManagerListener();
            _accManager = new AccountManager(_config.nmaid(), _config.password(), _config.asHost(), _config.asPort(), _config.uid(), _config.platform(), loginRetriever, resetRegistation, new FileManagerOem(appContext.getAppContext(), "as"), config.enableSelfSignedCert, config.certSummary, config.certData, appContext.getAppContext(), _mainThreadHandler);
            _mgr1 = null;
            _mgr2 = (com.nuance.nmsp.client2.sdk.components.resource.common.Manager)createNmspManager();
            _rscListener1 = null;
            _rscListener2 = (com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResourceListener)createNmasResourceListener();
        }
        _registeredRscListener1 = null;
        _registeredRscListener2 = null;
        _mdsListener = mdsListener;

        _context = appContext;

        _state = ConnectionState.DISCONNECTED;

        if ((!_config.newCop() ? _mgr1 : _mgr2) == null)
        {
            _isValid = false;
            return;
        }
        _isValid = true;
    }

    @Override
    public void release()
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                CloudServicesImpl.super.release();
                if (_accManager != null)
                    _accManager.release();
                if (_isValid)
                {
                    if (!_config.newCop())
                    {
                        if (_rsc1 != null)
                        {
                            try
                            {
                                _rsc1.freeResource(0);
                            } catch (com.nuance.nmsp.client.sdk.components.resource.common.ResourceException ex)
                            {
                                // Don't log this because it happens way too often, whenever the
                                // resource was already freed by the SDK.
                                // Logger.error(this, "Error freeing NMAS resource", ex);
                                // this can be simply ignored
                            }
                            _rsc1 = null;
                        }
                        _mgr1.shutdown();
                    }
                    else
                    {
                        if (_rsc2 != null)
                        {
                            try
                            {
                                _rsc2.freeResource(0);
                            } catch (com.nuance.nmsp.client2.sdk.components.resource.common.ResourceException ex)
                            {
                                // Don't log this because it happens way too often, whenever the
                                // resource was already freed by the SDK.
                                // Logger.error(this, "Error freeing NMAS resource", ex);
                                // this can be simply ignored
                            }
                            _rsc2 = null;
                        }
                        _mgr2.shutdown();
                    }
                    _isValid = false;
                }
            }
        });
    }

    @Override
    public void releaseAndWait()
    {
        final boolean [] done = new boolean[1];
        done[0] = false;
        final Object syncObj = new Object();
        synchronized(syncObj)
        {
            _mainThreadHandler.postToLooper(new Runnable()
            {
                @Override
                public void run() {
                    release();
                    synchronized(syncObj)
                    {
                        done[0] = true;
                        syncObj.notify();
                    }
                }
            });
            if (!done[0])
            {
                try {
                    syncObj.wait(3000);
                } catch (InterruptedException e) {
                    Logger.error(CloudServicesImpl.this, "releaseAndWait() e:" + e);
                }
            }
        }
    }

    @Override
    protected boolean canStartTransaction()
    {
        // if (_isValid)
        // {
        // return _state == ConnectionState.CONNECTED;
        // }
        //
        // return false;

        // TODO: WLWL, not sure about the logic above yet
        // as long as we have the nmsp manager
        // we can start the transaction since
        // the SDK will reconnect when create the command
        // or resource.
        return _isValid;
    }

    @Override
    public ConnectionState getConnectionState()
    {
        if (_isValid)
        {
            return _state;
        }

        return ConnectionState.DISCONNECTED;
    }

    @Override
    public String getSessionID()
    {
        if (_isValid)
        {
            return _sessionId;
        }

        return null;
    }

    @Override
    public String getUniqueID()
    {
        if (_config != null)
        {
            return _config.uid();
        }

        return null;
    }

    @Override
    public String getUserID() {
        if(_accManager != null){
            return _accManager.getNuanceUserId();
        }
        Logger.error(this, "getUserID() Trying to use a null account manager");
        return null;
    }

    @Override
    public String getClientID() {
        if(_accManager != null){
            return _accManager.getClientId();
        }
        Logger.error(this, "getClientID() Trying to use a null account manager");
        return null;
    }

    @Override
    public void warmUp() {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                _accManager.getSession(new ASSession.SessionListener()
                {
                    @Override
                    public void onGetSessionSucceeded(ASSession session) {
                        Logger.info(this, "warmUp() succeeded");
                    }
                    @Override
                    public void onGetSessionFailed(ASError error) {
                        Logger.info(this, "warmUp() failed");
                    }
                });
            }
        });
    }

    @Override
    public void activateCalllogManager(CalllogManager calllogManager)
    {
        if (!_isCalllogDisabled)
            calllogManager.attachCalllogImpl(this);
        else
            Logger.warn(this, "activateCalllogManager() calllog is disabled!");
    }

    @Override
    public void registerCalllogDataListener(CalllogDataListener calllogDataListener) {
        _calllogDataListener = calllogDataListener;
    }

    @Override
    public SessionEventBuilder logAppEvent(String appEventName, String schemaVersion)
    {
        if ((!_config.newCop() ? _mgr1 : _mgr2) == null)
            return null;

        if (_isCalllogDisabled)
            return null;

        CalllogSessionEventBuilderImpl sessionEventBuilder;
        if (!_config.newCop())
        {
            com.nuance.nmsp.client.sdk.components.core.calllog.SessionEventBuilder  nmspBuilder1 =  _mgr1.logAppEvent(appEventName, schemaVersion);
            sessionEventBuilder = new CalllogSessionEventBuilderImpl(nmspBuilder1, _config.newCop());
        }
        else
        {
            com.nuance.nmsp.client2.sdk.components.core.calllog.SessionEventBuilder  nmspBuilder2 =  _mgr2.logAppEvent(appEventName, schemaVersion);
            sessionEventBuilder = new CalllogSessionEventBuilderImpl(nmspBuilder2, _config.newCop());
        }

        return sessionEventBuilder;
    }

    @Override
    public String getCalllogRootParentId() {
        if ((!_config.newCop() ? _mgr1 : _mgr2) == null)
            return null;

        if (_isCalllogDisabled)
            return null;

        return !_config.newCop() ? _mgr1.getCalllogRootParentId() : _mgr2.getCalllogRootParentId();
    }

    @Override
    public void flushCallLogData()
    {
        if ((!_config.newCop() ? _mgr1 : _mgr2) == null || _isCalllogDisabled)
            return;

        if (!_config.newCop())
            _mgr1.flushCallLogData();
        else
            _mgr2.flushCallLogData();
    }

    @Override
    public CalllogSender createCalllogSender(SenderListener senderListener) {
        return new CalllogSenderImpl(senderListener, _config.newCop(), _accManager, _mainThreadHandler);
    }

    @Override
    public void connect() {
        if (_config.newCop())
        {
            _mainThreadHandler.postToLooper(new Runnable()
            {
                @Override
                public void run() {
                    _accManager.getSession(new ASSession.SessionListener()
                    {
                        @Override
                        public void onGetSessionSucceeded(ASSession session) {
                            if (_rsc2 == null)
                            {
                                createNmasResource(session);
                            }
                            else
                            {
                                _rsc2.loadResource();
                            }
                        }
                        @Override
                        public void onGetSessionFailed(ASError error) {
                        }
                    });
                }
            });
        }
    }

    @Override
    public NMTContext getContext()
    {
        return _context;
    }

    @Override
    protected CommandContext createCommandContext(Transaction t)
    {
        if (_isValid)
        {
            return new CommandContextImpl(this, !_config.newCop() ? _mgr1 : _mgr2, _config, t);
        }

        return null;
    }

    @Override
    AccountManager getAccountManager() {
        return _accManager;
    }

    // called from mainThread
    @Override
    Object acquireNmspResource(ASSession session, Object resourcelistener)
    {
        if ((!_config.newCop() ? _rsc1 : _rsc2) == null)
        {
            createNmasResource(session);
        }
        if (!_config.newCop())
        {
            Checker.checkState(this, _registeredRscListener1 == null, "the singleton nmsp resource is not yet released by the previous acquisition!!!");
            _registeredRscListener1 = (com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResourceListener)resourcelistener;
            return _rsc1;
        }
        else
        {
            Checker.checkState(this, _registeredRscListener2 == null, "the singleton nmsp resource is not yet released by the previous acquisition!!!");
            _registeredRscListener2 = (com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResourceListener)resourcelistener;
            return _rsc2;
        }
    }

    // called from mainThread
    @Override
    void releaseNmspResource()
    {
        if (!_config.newCop())
            _registeredRscListener1 = null;
        else
            _registeredRscListener2 = null;
    }

    private Object createNmspManagerListener()
    {
        class ConnectedRunnable implements Runnable
        {
            private String _sessId;
            private ConnectedRunnable(String sessId)
            {
                _sessId = sessId;
            }
            @Override
            public void run() {
                Logger.info(CloudServicesImpl.this, "Connected with session ID " + _sessId);
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        _sessionId = _sessId;
                        _state = ConnectionState.CONNECTED;
                    }
                });
            }
        }
        class ConnectionFailedRunnable implements Runnable
        {
            private short _reasonCode;
            private ConnectionFailedRunnable(short reasonCode)
            {
                _reasonCode = reasonCode;
            }
            @Override
            public void run() {
                Logger.warn(CloudServicesImpl.this, "Connection failed reasonCode [" + _reasonCode + "]");
            }
        }
        class DisconnectedRunnable implements Runnable
        {
            private short _reasonCode;
            private DisconnectedRunnable(short reasonCode)
            {
                _reasonCode = reasonCode;
            }
            @Override
            public void run() {
                Logger.debug(CloudServicesImpl.this, "Disconnected reasonCode [" + _reasonCode + "]");
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        _state = ConnectionState.DISCONNECTED;

                        // TODO: is a notification necessary?
                    }
                });
            }
        }
        class ShutdownCompletedRunnable implements Runnable
        {
            @Override
            public void run() {
                Logger.debug(CloudServicesImpl.this, "Shutdown completed");
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (_csListener != null)
                        {
                            _csListener.onReleaseCompleted();
                        }
                    }
                });
            }
        }
        class CallLogDataGeneratedRunnable implements Runnable
        {
            private byte[] _calllogData;
            @SuppressWarnings({ "rawtypes" })
            private List _sessionEventIds;
            @SuppressWarnings({ "rawtypes" })
            private CallLogDataGeneratedRunnable(byte[] callLogData, List sessionEventIds)
            {
                _calllogData = callLogData;
                _sessionEventIds = sessionEventIds;
            }
            @SuppressWarnings({ "unchecked", "deprecation" })
            @Override
            public void run() {
                Logger.debug(CloudServicesImpl.this, "callLogDataGenerated data [" + _calllogData + "] length [" + ((_calllogData != null) ? _calllogData.length : 0) + "]");
                Logger.debug(CloudServicesImpl.this, "_calllogDataListener [" + _calllogDataListener + "]");
                boolean isFlushed = false;
                if (_sessionEventIds != null && _sessionEventIds.size() > 0 )
                {
                    if (((String)_sessionEventIds.get(0)).startsWith("###flushed###"))
                    {
                        isFlushed = true;
                    }
                }
                final boolean isFlushedFinal = isFlushed;
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (_calllogDataListener != null)
                        {
                            _calllogDataListener.calllogDataGenerated(_calllogData, _sessionEventIds, isFlushedFinal);
                        }
                    }
                });
                if (_calllogDataListener != null)
                    _calllogDataListener.calllogDataGenerated2(_calllogData, _sessionEventIds, isFlushedFinal);
            }
        }

        return !_config.newCop() ?
            new com.nuance.nmsp.client.sdk.components.resource.common.ManagerListener()
            {
                @Override
                public void connected(String sessionId, com.nuance.nmsp.client.sdk.components.resource.common.Resource resource)
                {
                    (new ConnectedRunnable(sessionId)).run();
                }

                @Override
                public void connectionFailed(com.nuance.nmsp.client.sdk.components.resource.common.Resource resource, short reasonCode)
                {
                    (new ConnectionFailedRunnable(reasonCode)).run();
                }

                @Override
                public void disconnected(com.nuance.nmsp.client.sdk.components.resource.common.Resource resource, short reasonCode)
                {
                    (new DisconnectedRunnable(reasonCode)).run();
                }

                @Override
                public void shutdownCompleted() {
                    (new ShutdownCompletedRunnable()).run();
                }

                @SuppressWarnings({ "rawtypes" })
                @Override
                public void callLogDataGenerated(byte[] callLogData, List sessionEventIds)
                {
                    (new CallLogDataGeneratedRunnable(callLogData, sessionEventIds)).run();
                }
            }
            :
            new com.nuance.nmsp.client2.sdk.components.resource.common.ManagerListener()
            {
                @Override
                public void connected(String sessionId, com.nuance.nmsp.client2.sdk.components.resource.common.Resource resource)
                {
                    (new ConnectedRunnable(sessionId)).run();
                }

                @Override
                public void connectionFailed(com.nuance.nmsp.client2.sdk.components.resource.common.Resource resource, short reasonCode)
                {
                    (new ConnectionFailedRunnable(reasonCode)).run();
                }

                @Override
                public void disconnected(com.nuance.nmsp.client2.sdk.components.resource.common.Resource resource, short reasonCode)
                {
                    (new DisconnectedRunnable(reasonCode)).run();
                }

                @Override
                public void shutdownCompleted() {
                    (new ShutdownCompletedRunnable()).run();
                }

                @SuppressWarnings({ "rawtypes" })
                @Override
                public void callLogDataGenerated(byte[] callLogData, List sessionEventIds)
                {
                    (new CallLogDataGeneratedRunnable(callLogData, sessionEventIds)).run();
                }
            };
    }

    private Object createNmspManager()
    {
        try
        {

            Vector<com.nuance.nmsp.client.sdk.components.general.Parameter>  params1 = null;
            Vector<com.nuance.nmsp.client2.sdk.components.general.Parameter> params2 = null;

            List<NMTConfig> configList = _config.configs();
            if (configList != null)
            {
                if (!_config.newCop())
                    params1 = new Vector<com.nuance.nmsp.client.sdk.components.general.Parameter>();
                else
                    params2 = new Vector<com.nuance.nmsp.client2.sdk.components.general.Parameter>();
                for (NMTConfig config : configList)
                {
                    Enumeration<String> keys = config.keys();
                    while (keys.hasMoreElements()) {
                        String key = keys.nextElement();

                        if (key != null) {
                            Object value = config.getValue(key);
                            if (value != null) {
                                if (!_config.newCop())
                                    params1.add(new com.nuance.nmsp.client.sdk.components.general.Parameter(key, value, com.nuance.nmsp.client.sdk.components.general.Parameter.Type.SDK));
                                else
                                    params2.add(new com.nuance.nmsp.client2.sdk.components.general.Parameter(key, value, com.nuance.nmsp.client2.sdk.components.general.Parameter.Type.SDK));
                            }

                            if (key.equals(NMSPDefines.NMSP_DEFINES_CALLLOG_DISABLE)) {
                                String szDiable = new String((byte[])config.getValue(key));
                                if (szDiable.equals("FALSE"))
                                    _isCalllogDisabled = false;
                            }
                        }
                    }
                }
            }

            return !_config.newCop() ?
                com.nuance.nmsp.client.sdk.components.resource.common.ManagerFactory.createManager(_config.gwHost(), _config.gwPort(), _config.nmaid(), _config.appKey(), _config.uid(), (com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec)_config.nmspRecorderCodec(), (com.nuance.nmsp.client.sdk.common.defines.NMSPDefines.Codec)_config.nmspPlayerCodec(), params1,
                    _mgrListener1)
                :
                com.nuance.nmsp.client2.sdk.components.resource.common.ManagerFactory.createManager(_config.gwHost(), _config.gwPort(), _config.nmaid(), (com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec)_config.nmspRecorderCodec(), (com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec)_config.nmspPlayerCodec(), params2,
                            _mgrListener2);
        } catch (Throwable tr)
        {
            Logger.error(this, "Unable to create NMSP manager", tr);
            return null;
        }
    }

    // only commandCreated and queryResult are issued from main thread here, others are issued internally with main thread
    private Object createNmasResourceListener()
    {
        return !_config.newCop() ?
            // SDK1
            new com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResourceListener()
            {
                @Override
                public void PDXCommandCreated(final String sessionId)
                {
                    if (_registeredRscListener1 != null)
                        _registeredRscListener1.PDXCommandCreated(sessionId);
                }

                @Override
                public void PDXCreateCommandFailed(short reasonCode)
                {
                    if (_registeredRscListener1 != null)
                        _registeredRscListener1.PDXCreateCommandFailed(reasonCode);
                }

                @SuppressWarnings("rawtypes")
                @Override
                public void getParameterCompleted(short arg0, Vector arg1, long arg2)
                {
                }

                @Override
                public void getParameterFailed(short arg0, short arg1, long arg2)
                {
                }

                @Override
                public void resourceUnloaded(short arg0)
                {
                }

                @SuppressWarnings("rawtypes")
                @Override
                public void setParameterCompleted(short arg0, Vector arg1, long arg2)
                {
                }

                @Override
                public void setParameterFailed(short arg0, short arg1, long arg2)
                {
                }
            }
            :
            // SDK2
            new com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResourceListener()
            {
                @Override
                public void PDXCommandCreated(String sessionId)
                {
                    if (_registeredRscListener2 != null)
                        _registeredRscListener2.PDXCommandCreated(sessionId);
                }

                @Override
                public void PDXCreateCommandFailed(short reasonCode)
                {
                    if (_registeredRscListener2 != null)
                        _registeredRscListener2.PDXCreateCommandFailed(reasonCode);
                }

                @SuppressWarnings("rawtypes")
                @Override
                public void getParameterCompleted(short arg0, Vector arg1, long arg2)
                {
                }

                @Override
                public void getParameterFailed(short arg0, short arg1, long arg2)
                {
                }

                @Override
                public void resourceUnloaded(short arg0)
                {
                }

                @SuppressWarnings("rawtypes")
                @Override
                public void setParameterCompleted(short arg0, Vector arg1, long arg2)
                {
                }

                @Override
                public void setParameterFailed(short arg0, short arg1, long arg2)
                {
                }

                @Override
                public void sendMdsMessageStatus(int statusCode, int transactionId) {
                    Checker.checkState(this, false, "NMASResourceListener.sendMdsMessageStatus() should NEVER be triggered here!!!");
                }

                @Override
                public boolean receiveMdsMessageStatus(int statusCode,
                        final int transactionId, final com.nuance.nmsp.client2.sdk.components.core.internal.pdx.PDXDictionary message) {
                    Logger.debug(CloudServicesImpl.this, "receiveMdsMessageStatus(" + statusCode + ", " + transactionId + ", " + message + ")");

                    // For phase one we only pass on successful MDS messages
                    if (statusCode != STATUS_CODE_MDS_MESSAGE_RECEIVED)
                        return true;

                    _mainThreadHandler.post(new Runnable()
                    {
                        public void run()
                        {
                            _mdsListener.onMessageReceived(transactionId, message == null ? null : PdxConverter.dictionaryFromPdx(message));
                        }
                    });
                    return true;
                }
            };
    }

    /**
     * Create the NMAS resource.
     */
    private void createNmasResource(ASSession sess)
    {
        if (!_config.newCop())
        {
            Vector<com.nuance.nmsp.client.sdk.components.general.Parameter> params = new Vector<com.nuance.nmsp.client.sdk.components.general.Parameter>();
            params.add(new com.nuance.nmsp.client.sdk.components.general.Parameter(NMSPDefines.NMSP_DEFINES_DISABLE_BLUETOOTH,
                                                                                   "TRUE".getBytes(),
                                                                                   com.nuance.nmsp.client.sdk.components.general.Parameter.Type.SDK));
            _rsc1 = com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResourceFactory.createNMASResource(_mgr1, _rscListener1, params, _config.applicationName());
        } else
        {
            Vector<com.nuance.nmsp.client2.sdk.components.general.Parameter> params = new Vector<com.nuance.nmsp.client2.sdk.components.general.Parameter>();
            params.add(new com.nuance.nmsp.client2.sdk.components.general.Parameter(NMSPDefines.NMSP_DEFINES_DISABLE_BLUETOOTH,
                                                                                    "TRUE".getBytes(),
                                                                                    com.nuance.nmsp.client2.sdk.components.general.Parameter.Type.SDK));
            _rsc2 = com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResourceFactory.createNMASResource(_mgr2, _rscListener2, params, _config.applicationName(), sess.NuanceUserId, ASSession.ACCESS_TOKEN_PREFIX + sess.AccessToken);
        }
    }

    @Override
    public void setCloudServicesListener(CloudServicesListener listener) {
        _csListener = listener;
    }
}
