package com.nuance.dragon.toolkit.cloudservices;

import com.nuance.dragon.toolkit.core.data.Data.Dictionary;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * A Cloud Simple Command executes a simple text command through a
 * {@link CloudServices}.
 */
public class CloudSimpleCommand
{
    /**
     * Listens for the result of the command transaction
     */
    public interface Listener
    {
        /**
         * The command succeeded.
         * @param result The transaction result object.
         */
        void onSuccess(TransactionResult result);

        /**
         * There was an error when transmitting the command.
         * @param error The transaction error object.
         */
        void onError(TransactionError error);
    }

    private final CloudServices _cloud;
    private final String _commandName;
    private final Dictionary _commandSettings;
    private final NMTHandler _mainThreadHandler;
    private Transaction _currentTransaction;

    /**
     * @param cloudServices The cloud services instance to use.
     * @param commandName The command name.
     * @param commandSettings The command settings. May be null.
     */
    public CloudSimpleCommand(CloudServices cloudServices, String commandName, Dictionary commandSettings)
    {
        Checker.checkArgForNull("cloudServices", cloudServices);
        Checker.checkStringArgForNullOrEmpty("commandName", commandName);
        _cloud = cloudServices;
        _commandName = commandName;
        _commandSettings = commandSettings;
        _mainThreadHandler = _cloud.getMainThreadHandler();
    }

    /**
     * Transmit the command.
     * @param resultListener The result listener.
     */
    public void send(Listener resultListener)
    {
        send(10000, resultListener);
    }

    /**
     * Send the command.
     * @param timeoutMs The command timeout in milliseconds.
     * @param resultListener The result listener.
     */
    public void send(final int timeoutMs, final Listener resultListener)
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                // Cancel any command in progress
                cancel();
        
                _currentTransaction = new Transaction(_commandName, _commandSettings, new Transaction.Listener()
                {
                    @Override
                    public void onTransactionStarted(Transaction t)
                    {
                        // Nothing to do
                    }
        
                    @Override
                    public void onTransactionProcessingStarted(Transaction t)
                    {
                        // Nothing to do
                    }
        
                    @Override
                    public void onTransactionResult(Transaction t, TransactionResult result, boolean complete)
                    {
                        if (t != _currentTransaction)
                            return;
        
                        _currentTransaction = null;
        
                        if (resultListener != null)
                            resultListener.onSuccess(result);
                    }
        
                    @Override
                    public void onTransactionError(Transaction t, TransactionError error)
                    {
                        if (t != _currentTransaction)
                            return;
        
                        _currentTransaction = null;
        
                        if (resultListener != null)
                            resultListener.onError(error);
                    }
        
                    @Override
                    public void onTransactionIdGenerated(String transactionId)
                    {
                        // TODO Auto-generated method stub
                    }
                }, timeoutMs);
        
                _cloud.addTransaction(_currentTransaction, Transaction.DefaultPriorities.MEDIUM);
                _currentTransaction.finish();
            }
        });
    }

    /**
     * Cancel the command. This is safe to call even if there is no active command.
     */
    public void cancel()
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                if (_currentTransaction != null)
                {
                    Transaction t = _currentTransaction;
                    _currentTransaction = null;
                    t.cancel();
                }
            }
        });
    }
}
