package com.nuance.dragon.toolkit.cloudservices;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.core.data.Data;

/**
 * The command context provides the methods that a transaction must use to send
 * a command to the NMAS server.
 */
interface CommandContext
{
    interface NMTAudioSink
    {
        void addAudioBuf(byte[] buffer, int offset, int length, boolean isLastBuf);
    }
    
    public boolean newCop();
    
    /**
     * Get a session from AccountManager.
     */
    public void getSession();

    /**
     * Initialize the command. This must be called prior to calling any other
     * methods.
     *
     * @param cmdName The name of the command to execute on the Nuance Mobile
     *                Application Server host (cannot be null)
     * @param cmdSettings the specific keys for this command (this will go
     *                    in the BCP begin) (maybe null)
     *
     *                    User may specify language code through cmdSetting parameter with key of 'dictation_language',
     *                    for example, user may set 'dictation_language' = 'eng-USA' to use 'US English' as language.
     *                    if 'dictation_language' is not set, NMT fetches default language{@link com.nuance.dragon.toolkit.language.Language} from locale settings.
     *
     * @param timeout The timeout (in ms) of the command. (should be greater than 0)
     *
     * @throws Exception (if the internal NMSP SDK could not initialize a command due to the invalid internal state)
     */
    public void initializeCommand(String cmdName, Data.Dictionary cmdSettings, int timeout) throws Exception;

    /**
     * Release the NMAS resource associated with the current command, if
     * it has been called once, the consequent calls will be ignored.
     */
    public void freeResource();

    /**
     * End the command. After calling this, no more parameters may be sent,
     * although more audio can be sent for an existing audio parameter via
     * {@link #sendAudioForParam}.
     *
     * @throws Exception (if the internal NMSP SDK could not perform this action in its current
     *                    state, could be network is down or the command already expired)
     */
    void endCommand() throws Exception;

    /**
     * Send a text parameter.
     *
     * @param name name of the String parameter expected by the command
     * @param text text data
     *
     * @throws Exception (if the internal NMSP SDK could not perform this action in its current
     *                    state, could be network is down or the command already expired)
     */
    void sendTextParam(String name, String text) throws Exception;

    /**
     * Send a choice parameter, which holds the name and value of a possible choice
     * according to the results returned by the command on the Nuance Mobile Application
     * Server host.
     *
     * @param name        name of the PDX choice parameter expected by the command
     * @param choicename  a value that represents the choice specified by the client
     *
     * @throws Exception (if the internal NMSP SDK could not perform this action in its current
     *                    state, could be network is down or the command already expired)
     */
    void sendChoiceParam(String name, String choicename) throws Exception;

    /**
     * Send binary data to the command.
     *
     * @param name  name of the PDX data parameter expected by the command
     * @param data  raw binary data
     *
     * @throws Exception (if the internal NMSP SDK could not perform this action in its current
     *                    state, could be network is down or the command already expired)
     */
    void sendDataParam(String name, byte[] data) throws Exception;

    /**
     * Send a sequence start parameter, which indicates
     * the start of a series of data chunks that are going to be sent.
     *
     * @param name  name of the PDX Sequence Start Parameter
     * @param dict  a PDXDictionary containing all the dictionary contents
     *
     * @throws Exception (if the internal NMSP SDK could not perform this action in its current
     *                    state, could be network is down or the command already expired)
     */
    void sendSeqStartParam(String name, Data.Dictionary dict) throws Exception;

    /**
     * Send a Sequence Chunk Parameter. PDX Sequence Chunk is used to send
     * large amounts of data in dictionary format (like a contact list in {@link Data.Dictionary}
     * format) in smaller pieces (chunks).
     *
     * @param name name of the Sequence Chunk Parameter
     * @param dict {@link Data.Dictionary} that contains all the dictionary contents
     *
     * @throws Exception (if the internal NMSP SDK could not perform this action in its current
     *                    state, could be network is down or the command already expired)
     */
    void sendSeqChunkParam(String name, Data.Dictionary dict) throws Exception;

    /**
     * Send a Sequence End Parameter, which indicates
     * the end of a series of data chunks that were sent.
     *
     * @param name  name of the Sequence End Parameter
     * @param dict  PDXDictionary containing all the dictionary contents
     *
     * @throws Exception (if the internal NMSP SDK could not perform this action in its current
     *                    state, could be network is down or the command already expired)
     */
    void sendSeqEndParam(String name, Data.Dictionary dict) throws Exception;

    /**
     * Send a parameter for the TTS text and pushes the generated audio
     * buffers to the audio sink object. You can hook the synthesized audio to the
     * Player object, which lets the Player play synthesized audio samples as they
     * are streamed to the mobile device.
     *
     * While it is receiving the synthesized audio buffers from the server, the library
     * streams the audio buffers to the associated audioSink object.
     *
     * @param name        name of the TTS parameter
     * @param tts_text    text to be synthesized
     * @param audioSink   NMSPAudioSink object to which the synthesized buffers are sent
     *
     * @throws Exception (if the internal NMSP SDK could not perform this action in its current
     *                    state, could be network is down or the command already expired)
     */
    void sendTTSParam(String name, Data.Dictionary dict, NMTAudioSink audioSink) throws Exception;

    /**
     * Send a dictionary parameter, which can contain other type of data.
     *
     * @param name parameter name required by the command defined on the Nuance Mobile
     *             Application Server
     * @param dict the associated data in {@link Data.Dictionary} format
     *
     * @throws Exception (if the internal NMSP SDK could not perform this action in its current
     *                    state, could be network is down or the command already expired)
     */
    void sendDictParam(String name, Data.Dictionary dict) throws Exception;

    /**
     * Send an audio parameter. Use {@link #sendAudioForParam} to send audio
     * for this parameter.
     *
     * @param name the name of the audio parameter required by the command
     *
     * @throws Exception (if the internal NMSP SDK could not perform this action in its current
     *                    state, could be network is down or the command already expired)
     */
    void sendAudioParam(String name) throws Exception;

    /**
     * Send audio for a parameter that was previously specified via
     * {@link #sendAudioParam}.
     *
     * @param audioParamName The name of the existing audio parameter.
     * @param audio The audio chunk to send (may be null to send no audio).
     * @param isFinal True if no more audio chunks will be added ('audio' may
     * be null).
     *
     * @throws Exception (if the internal NMSP SDK could not perform this action in its current
     *                    state, could be network is down or the command already expired)
     */
    void sendAudioForParam(String audioParamName, AudioChunk audio, boolean isFinal) throws Exception;
}
