package com.nuance.dragon.toolkit.cloudservices;

import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.Vector;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.cloudservices.util.PdxConverter;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.internal.SystemClock;
import com.nuance.dragon.toolkit.oem.impl.HandlerOem;

/**
 * Actual implementation of the CommandContext, this class represent a NMSP level
 * command (transaction), it is responsible to communicate with the speech server.
 */
class CommandContextImpl implements CommandContext
{
    private final CloudServices _cs;

    // SDK1
    private final com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResourceListener     _rscListener1;
    private final com.nuance.nmsp.client.sdk.components.resource.nmas.PDXCommandListener       _cmdListener1;
    private com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResource                   _rsc1;
    private com.nuance.nmsp.client.sdk.components.resource.nmas.Command                        _cmd1;
    private final Map<String, com.nuance.nmsp.client.sdk.components.resource.nmas.AudioParam>  _audioParams1;
    // SDK2
    private final com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResourceListener    _rscListener2;
    private final com.nuance.nmsp.client2.sdk.components.resource.nmas.PDXCommandListener      _cmdListener2;
    private com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource                  _rsc2;
    private com.nuance.nmsp.client2.sdk.components.resource.nmas.Command                       _cmd2;
    private final Map<String, com.nuance.nmsp.client2.sdk.components.resource.nmas.AudioParam> _audioParams2;

    private final CloudServicesConfig                                                          _config;
    private String                                                                             _cmdName;
    private final Transaction                                                                  _tran;
    private final AccountManager                                                               _accManager;
    private final NMTHandler                                                                   _mainThreadHandler;
    private ASSession                                                                          _sess;

    /**
     * @param mgr          the NMSP Manager which manages the connection between client and speech server
     * @param config       the cloud service configuration object including the basic settings for NMSP module
     * @param transaction  the could service transaction which will be associate with the current command
     */
    CommandContextImpl(CloudServices cs, Object mgr, CloudServicesConfig config, Transaction transaction)
    {
        _cs = cs;
        _config = config;
        _tran = transaction;
        _accManager = _cs.getAccountManager();
        _cmdName = "no_cmd_yet";
        _mainThreadHandler = new HandlerOem();

        if (!config.newCop())
        {
            // SDK1
            _rscListener1 = (com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResourceListener)createNmasResourceListener();
            _cmdListener1 = (com.nuance.nmsp.client.sdk.components.resource.nmas.PDXCommandListener)createCommandListener();
            _audioParams1 = new HashMap<String, com.nuance.nmsp.client.sdk.components.resource.nmas.AudioParam>();
            _rscListener2 = null;
            _cmdListener2 = null;
            _audioParams2 = null;

        } else
        {
            // SDK2
            _rscListener1 = null;
            _cmdListener1 = null;
            _audioParams1 = null;
            _rscListener2 = (com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResourceListener)createNmasResourceListener();
            _cmdListener2 = (com.nuance.nmsp.client2.sdk.components.resource.nmas.PDXCommandListener)createCommandListener();
            _audioParams2 = new HashMap<String, com.nuance.nmsp.client2.sdk.components.resource.nmas.AudioParam>();
        }

        _sess = null;
    }

    @Override
    public boolean newCop()
    {
        return _config.newCop();
    }

    @Override
    public void getSession()
    {
        _accManager.getSession(new ASSession.SessionListener()
        {
            @Override
            public void onGetSessionSucceeded(ASSession session) {
                _sess = session;
                _tran.currentState().getSessionSucceeded(session);
            }
            @Override
            public void onGetSessionFailed(ASError error) {
                _tran.currentState().getSessionFailed(error);
            }
        });
    }

    @Override
    public void initializeCommand(String cmdName, Data.Dictionary cmdSettings, int timeout) throws Exception
    {
        _cmdName = cmdName;
        createNmasResource();
        createNmasCommand(_cmdName, cmdSettings, timeout);
    }

    @Override
    public void endCommand() throws Exception
    {
        if (!_config.newCop())
        {
            // SDK1
            if (_cmd1 != null)
            {
                Logger.debug(this, "Ending command");
                _cmd1.end();
            }
        } else
        {
            // SDK2
            if (_cmd2 != null)
            {
                Logger.debug(this, "Ending command");
                _cmd2.end();
            }
        }
    }

    @Override
    public void sendAudioForParam(String audioParamName, AudioChunk audio, boolean isFinal) throws Exception
    {
        if ((!_config.newCop() ? _cmd1 : _cmd2) != null)
        {
            if (!_config.newCop())
            {
                // SDK1
                if (!_audioParams1.containsKey(audioParamName))
                {
                    sendAudioParam(audioParamName);
                }
            } else
            {
                // SDK2
                if (!_audioParams2.containsKey(audioParamName))
                {
                    sendAudioParam(audioParamName);
                }
            }

            com.nuance.nmsp.client.sdk.components.resource.nmas.AudioParam audioParam1  = !_config.newCop() ? _audioParams1.get(audioParamName) : null;
            com.nuance.nmsp.client2.sdk.components.resource.nmas.AudioParam audioParam2 = !_config.newCop() ? null : _audioParams2.get(audioParamName);
            if (audio != null)
            {
                if (audio.audioBytes != null)
                {
                    Logger.verbose(this, "[LATCHK] calling addAudioBuf() vocoded audio:" + audio);
                    if (!_config.newCop())
                        audioParam1.addAudioBuf(audio.audioBytes, 0, audio.audioBytes.length, isFinal);
                    else
                        audioParam2.addAudioBuf(audio.audioBytes, 0, audio.audioBytes.length, isFinal);
//                        Logger.debug(CommandContextImpl.this, "stream audio bytes [" + audio.audioBytes.length
//                                + "] last buffer [" + isFinal + "]");
                } else if (audio.audioShorts != null)
                {
                    byte[] buffer = new byte[audio.audioShorts.length * 2];
                    for (int i = 0; i < audio.audioShorts.length; i++)
                    {
                        buffer[i * 2] = (byte) audio.audioShorts[i];
                        buffer[i * 2 + 1] = (byte) (audio.audioShorts[i] >> 8);
                    }
                    Logger.verbose(this, "[LATCHK] calling addAudioBuf() pcm audio:" + audio);
                    if (!_config.newCop())
                        audioParam1.addAudioBuf(buffer, 0, buffer.length, isFinal);
                    else
                        audioParam2.addAudioBuf(buffer, 0, buffer.length, isFinal);
//                        Logger.debug(CommandContextImpl.this, "stream audio bytes [" + buffer.length
//                                + "] last buffer [" + isFinal + "]");
                }
            } else
            {
                assert (isFinal);
                Logger.debug(this, "[LATCHK] calling addAudioBuf() final");
                if (!_config.newCop())
                    audioParam1.addAudioBuf(null, 0, 0, isFinal);
                else
                    audioParam2.addAudioBuf(null, 0, 0, isFinal);
                Logger.debug(CommandContextImpl.this, "stream audio last buffer [" + isFinal + "]");
            }
        }
    }

    @Override
    public void sendAudioParam(String name) throws Exception
    {
        if ((!_config.newCop() ? _cmd1 : _cmd2) != null)
        {
            boolean containsKey = !_config.newCop() ? _audioParams1.containsKey(name) : _audioParams2.containsKey(name);
            if (!containsKey)
            {
                if (!_config.newCop())
                {
                    // SDK1
                    com.nuance.nmsp.client.sdk.components.resource.nmas.AudioParam audioParam = _rsc1
                            .createPDXAudioParam(name);
                    _audioParams1.put(name, audioParam);
                    sendParam(audioParam, name);
                } else
                {
                    // SDK2
                    com.nuance.nmsp.client2.sdk.components.resource.nmas.AudioParam audioParam = _rsc2
                            .createPDXAudioParam(name);
                    _audioParams2.put(name, audioParam);
                    sendParam(audioParam, name);
                }
            } else
            {
                Logger.warn(this, "This Audio Param [" + name + "] has already been sent, it will ignored");
            }
        }
    }

    @Override
    public void sendDictParam(String name, Data.Dictionary dict) throws Exception
    {
        if ((!_config.newCop() ? _cmd1 : _cmd2) != null)
        {
            if (!_config.newCop())
            {
                // SDK1
                com.nuance.nmsp.client.sdk.components.core.pdx.Dictionary nmspDict = PdxConverter.dictionaryToPdx(dict, _rsc1);
                com.nuance.nmsp.client.sdk.components.resource.nmas.Parameter dictParam = _rsc1.createPDXDictParam(name,
                        nmspDict);
                sendParam(dictParam, name);
            } else
            {
                // SDK2
                com.nuance.nmsp.client2.sdk.components.core.pdx.Dictionary nmspDict = PdxConverter.dictionaryToPdx(dict, _rsc2);
                com.nuance.nmsp.client2.sdk.components.resource.nmas.Parameter dictParam = _rsc2.createPDXDictParam(name,
                        nmspDict);
                sendParam(dictParam, name);
            }
        }
    }

    @Override
    public void sendTextParam(String name, String text) throws Exception
    {
        if ((!_config.newCop() ? _cmd1 : _cmd2) != null)
        {
            if (!_config.newCop())
            {
                // SDK1
                com.nuance.nmsp.client.sdk.components.resource.nmas.Parameter textParam = _rsc1.createPDXTextParam(name,
                        text);
                sendParam(textParam, name);
            } else
            {
                // SDK2
                com.nuance.nmsp.client2.sdk.components.resource.nmas.Parameter textParam = _rsc2.createPDXTextParam(name,
                        text);
                sendParam(textParam, name);
            }
        }
    }

    @Override
    public void sendChoiceParam(String name, String choicename) throws Exception
    {
        if ((!_config.newCop() ? _cmd1 : _cmd2) != null)
        {
            if (!_config.newCop())
            {
                // SDK1
                com.nuance.nmsp.client.sdk.components.resource.nmas.Parameter choiceParam = _rsc1.createPDXChoiceParam(name,
                        choicename);
                sendParam(choiceParam, name);
            } else
            {
                // SDK2
                com.nuance.nmsp.client2.sdk.components.resource.nmas.Parameter choiceParam = _rsc2.createPDXChoiceParam(name,
                        choicename);
                sendParam(choiceParam, name);
            }
        }
    }

    @Override
    public void sendDataParam(String name, byte[] data) throws Exception
    {
        if ((!_config.newCop() ? _cmd1 : _cmd2) != null)
        {
            if (!_config.newCop())
            {
                // SDK1
                com.nuance.nmsp.client.sdk.components.resource.nmas.Parameter dataParam = _rsc1.createPDXDataParam(name,
                        data);
                sendParam(dataParam, name);
            } else
            {
                // SDK2
                com.nuance.nmsp.client2.sdk.components.resource.nmas.Parameter dataParam = _rsc2.createPDXDataParam(name,
                        data);
                sendParam(dataParam, name);
            }
        }
    }

    @Override
    public void sendSeqStartParam(String name, Data.Dictionary dict) throws Exception
    {
        if ((!_config.newCop() ? _cmd1 : _cmd2) != null)
        {
            if (!_config.newCop())
            {
                // SDK1
                com.nuance.nmsp.client.sdk.components.core.pdx.Dictionary nmspDict = PdxConverter.dictionaryToPdx(dict, _rsc1);
                com.nuance.nmsp.client.sdk.components.resource.nmas.Parameter seqStartParam = _rsc1.createPDXSeqStartParam(
                        name, nmspDict);
                sendParam(seqStartParam, name);
            } else
            {
                // SDK2
                com.nuance.nmsp.client2.sdk.components.core.pdx.Dictionary nmspDict = PdxConverter.dictionaryToPdx(dict, _rsc2);
                com.nuance.nmsp.client2.sdk.components.resource.nmas.Parameter seqStartParam = _rsc2.createPDXSeqStartParam(
                        name, nmspDict);
                sendParam(seqStartParam, name);
            }
        }
    }

    @Override
    public void sendSeqChunkParam(String name, Data.Dictionary dict) throws Exception
    {
        if ((!_config.newCop() ? _cmd1 : _cmd2) != null)
        {
            if (!_config.newCop())
            {
                // SDK1
                com.nuance.nmsp.client.sdk.components.core.pdx.Dictionary nmspDict = PdxConverter.dictionaryToPdx(dict, _rsc1);
                com.nuance.nmsp.client.sdk.components.resource.nmas.Parameter seqChunkParam = _rsc1.createPDXSeqChunkParam(
                        name, nmspDict);
                sendParam(seqChunkParam, name);
            } else
            {
                // SDK2
                com.nuance.nmsp.client2.sdk.components.core.pdx.Dictionary nmspDict = PdxConverter.dictionaryToPdx(dict, _rsc2);
                com.nuance.nmsp.client2.sdk.components.resource.nmas.Parameter seqChunkParam = _rsc2.createPDXSeqChunkParam(
                        name, nmspDict);
                sendParam(seqChunkParam, name);
            }
        }
    }

    @Override
    public void sendSeqEndParam(String name, Data.Dictionary dict) throws Exception
    {
        if ((!_config.newCop() ? _cmd1 : _cmd2) != null)
        {
            if (!_config.newCop())
            {
                // SDK1
                com.nuance.nmsp.client.sdk.components.core.pdx.Dictionary nmspDict = PdxConverter.dictionaryToPdx(dict, _rsc1);
                com.nuance.nmsp.client.sdk.components.resource.nmas.Parameter seqEndParam = _rsc1.createPDXSeqEndParam(
                        name, nmspDict);
                sendParam(seqEndParam, name);
            } else
            {
                // SDK2
                com.nuance.nmsp.client2.sdk.components.core.pdx.Dictionary nmspDict = PdxConverter.dictionaryToPdx(dict, _rsc2);
                com.nuance.nmsp.client2.sdk.components.resource.nmas.Parameter seqEndParam = _rsc2.createPDXSeqEndParam(
                        name, nmspDict);
                sendParam(seqEndParam, name);
            }
        }
    }

    @Override
    public void sendTTSParam(String name, Data.Dictionary dict, final NMTAudioSink audioSink) throws Exception
    {
        if ((!_config.newCop() ? _cmd1 : _cmd2) != null)
        {
            if (!_config.newCop())
            {
                // SDK1
                com.nuance.nmsp.client.sdk.components.core.pdx.Dictionary nmspDict = PdxConverter.dictionaryToPdx(dict, _rsc1);
                com.nuance.nmsp.client.sdk.components.resource.nmas.Parameter ttsParam = _rsc1.createPDXTTSParam(name,
                        nmspDict, new com.nuance.nmsp.client.sdk.components.general.NMSPAudioSink()
                {
                    @Override
                    public void addAudioBuf(byte[] buffer, int offset, int length,
                            boolean isLastBuf) {
                        audioSink.addAudioBuf(buffer, offset, length, isLastBuf);
                    }
                });
                sendParam(ttsParam, name);
            } else
            {
                // SDK2
                com.nuance.nmsp.client2.sdk.components.core.pdx.Dictionary nmspDict = PdxConverter.dictionaryToPdx(dict, _rsc2);
                com.nuance.nmsp.client2.sdk.components.resource.nmas.Parameter ttsParam = _rsc2.createPDXTTSParam(name,
                        nmspDict, new com.nuance.nmsp.client2.sdk.components.general.NMSPAudioSink()
                        {
                            @Override
                            public void addAudioBuf(byte[] buffer, int offset, int length,
                                    boolean isLastBuf) {
                                audioSink.addAudioBuf(buffer, offset, length, isLastBuf);
                            }
                        });
                sendParam(ttsParam, name);
            }
        }
    }

    protected CloudServicesConfig getConfig()
    {
        return _config;
    }

    private void sendParam(Object param, String name) throws Exception
    {
        Logger.debug(this, "Sending param " + name);
        if (!_config.newCop())
            _cmd1.sendParam((com.nuance.nmsp.client.sdk.components.resource.nmas.Parameter)param);
        else
            _cmd2.sendParam((com.nuance.nmsp.client2.sdk.components.resource.nmas.Parameter)param);
    }

    /**
     * Create the NMAS resource.
     */
    private void createNmasResource()
    {
        if (!_config.newCop())
            _rsc1 = (com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResource)_cs.acquireNmspResource(_sess, _rscListener1);
        else
            _rsc2 = (com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource)_cs.acquireNmspResource(_sess, _rscListener2);
    }

    private void createNmasCommand(String command, Data.Dictionary cmdSettings, int timeout) throws com.nuance.nmsp.client.sdk.components.resource.common.ResourceException, com.nuance.nmsp.client2.sdk.components.resource.common.ResourceException
    {
        if ((!_config.newCop() ? _rsc1 : _rsc2) != null)
        {
            Logger.debug(this, "Creating command " + command);

            String carrier = _config.carrier();
            String model = _config.model();
            String os = _config.os();
            String osType = _config.osType();
            String osVersion = _config.osVersion();
            String locale = _config.locale();
            String networkType = _config.networkType();
            String client = _config.clientVersion();
            String appName = _config.applicationName();
            String nmaid = _config.nmaid();
            String language = _config.defaultLanguage();

            if (!_config.newCop())
            {
                // SDK1
                com.nuance.nmsp.client.sdk.components.core.pdx.Dictionary settings = null;
                if (cmdSettings != null)
                    settings = PdxConverter.dictionaryToPdx(cmdSettings, _rsc1);
                else
                    settings = _rsc1.createPDXDictionary();
                // give user a chance to enter their own device specific parameters
                if (settings.containsKey("dictation_language"))
                    language = settings.getUTF8String("dictation_language");
                if (!settings.containsKey("ui_language"))
                    settings.addUTF8String("ui_language", language.substring(0, 2).toLowerCase(Locale.getDefault()));
                if (!settings.containsKey("phone_submodel"))
                    settings.addUTF8String("phone_submodel", model);
                if (!settings.containsKey("phone_OS"))
                    settings.addUTF8String("phone_OS", os);
                if (!settings.containsKey("locale"))
                    settings.addUTF8String("locale", locale);
                if (!settings.containsKey("nmaid"))
                    settings.addUTF8String("nmaid", nmaid);
                if (!settings.containsKey("network_type"))
                    settings.addUTF8String("network_type", networkType);
                if (!settings.containsKey("application_name"))
                    settings.addUTF8String("application_name", appName);

                // TODO:
                // byte [] transactionId = new byte[100];
                // try {
                // transactionId = _config.message(transactionId, null);
                // } catch (Throwable tr)
                // {
                // }
                // dict.addByteString("app_transaction_id", transactionId);
                // for (String key : customKeys.keySet())
                // {
                // dict.addUTF8String(key, customKeys.);
                // }
                // Don't handle an exception here. Let the calling code handle it.
                _cmd1 = _rsc1.createCommand(_cmdListener1, command, client, language, carrier, model, settings, timeout);
            } else
            {
                // SDK2
                com.nuance.nmsp.client2.sdk.components.core.pdx.Dictionary settings = null;
                if (cmdSettings != null)
                    settings = PdxConverter.dictionaryToPdx(cmdSettings, _rsc2);
                else
                    settings = _rsc2.createPDXDictionary();
                // give user a chance to enter their own device specific parameters
                if (settings.containsKey("dictation_language"))
                    language = settings.getUTF8String("dictation_language");
                if (!settings.containsKey("ui_language"))
                    settings.addUTF8String("ui_language", language.substring(0, 2).toLowerCase(Locale.getDefault()));
                if (!settings.containsKey("phone_submodel"))
                    settings.addUTF8String("phone_submodel", model);
                if (!settings.containsKey("phone_OS"))
                    settings.addUTF8String("phone_OS", os);
                if (!settings.containsKey("locale"))
                    settings.addUTF8String("locale", locale);
                if (!settings.containsKey("nmaid"))
                    settings.addUTF8String("nmaid", nmaid);
                if (!settings.containsKey("network_type"))
                    settings.addUTF8String("network_type", networkType);
                if (!settings.containsKey("application_name"))
                    settings.addUTF8String("application_name", appName);

                // new connection info keys
                if (!settings.containsKey("nmt_connection_provider"))
                    settings.addUTF8String("nmt_connection_provider", carrier);
                if (!settings.containsKey("nmt_connection_type"))
                    settings.addUTF8String("nmt_connection_type", networkType);
                if (!settings.containsKey("nmt_version_build"))
                    settings.addUTF8String("nmt_version_build", client);
                if (!settings.containsKey("nmt_device_model"))
                    settings.addUTF8String("nmt_device_model", model);
                if (!settings.containsKey("nmt_os_type"))
                    settings.addUTF8String("nmt_os_type", osType);
                if (!settings.containsKey("nmt_os_version_build"))
                    settings.addUTF8String("nmt_os_version_build", osVersion);
                if (!settings.containsKey("nmt_time_3339"))
                    settings.addUTF8String("nmt_time_3339", SystemClock.get3339TimeString());

                // TODO:
                // byte [] transactionId = new byte[100];
                // try {
                // transactionId = _config.message(transactionId, null);
                // } catch (Throwable tr)
                // {
                // }
                // dict.addByteString("app_transaction_id", transactionId);
                // for (String key : customKeys.keySet())
                // {
                // dict.addUTF8String(key, customKeys.);
                // }
                // Don't handle an exception here. Let the calling code handle it.
                _cmd2 = _rsc2.createCommand(_cmdListener2, command, client, language, carrier, model, settings, timeout, _sess.NuanceUserId, ASSession.ACCESS_TOKEN_PREFIX + _sess.AccessToken);
                Logger.debug(this, "createNmasCommand with access token: " + ASSession.ACCESS_TOKEN_PREFIX + _sess.AccessToken + " NuanceUserId: " + _sess.NuanceUserId);
            }
        }
    }

    /**
     * Free the NMAS resource. Safe to call if the NMAS resource hasn't been
     * created, or has already been freed.
     */
    @Override
    public void freeResource()
    {
        if ((!_config.newCop() ? _rsc1 : _rsc2) != null)
        {
            if (!_config.newCop())
            {
                // SDK1
                _cs.releaseNmspResource();
                _rsc1 = null;
            } else
            {
                // SDK2
                _cs.releaseNmspResource();
                _rsc2 = null;
            }
        }
    }

    // only commandCreated and queryResult are issued from main thread here, others are issued internally with main thread
    private Object createNmasResourceListener()
    {
        class PDXCommandCreatedRunnable implements Runnable
        {
            private final String _sessionId;
            private PDXCommandCreatedRunnable(String sessionId)
            {
                _sessionId = sessionId;
            }
            @Override
            public void run() {
                Logger.debug(CommandContextImpl.this, "Command Created [" + _cmdName + "] session id [" + _sessionId + "]");
                _mainThreadHandler.post(new Runnable()
                {
                    public void run()
                    {
                        _tran.setSessionId(_sessionId);
                        _tran.currentState().commandCreated(_sessionId);
                    }
                });
            }
        };
        class PDXCreateCommandFailedRunnable implements Runnable
        {
            private final short _reasonCode;
            private PDXCreateCommandFailedRunnable(short reasonCode)
            {
                _reasonCode = reasonCode;
            }
            @Override
            public void run() {
                Logger.error(CommandContextImpl.this, "Command Creation Failed [" + _cmdName + "]");
                _tran.currentState().createCommandFailed(_reasonCode);
            }
        };

        return !_config.newCop() ?
            // SDK1
            new com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResourceListener()
            {
                @Override
                public void PDXCommandCreated(final String sessionId)
                {
                    (new PDXCommandCreatedRunnable(sessionId)).run();
                }

                @Override
                public void PDXCreateCommandFailed(short reasonCode)
                {
                    (new PDXCreateCommandFailedRunnable(reasonCode)).run();
                }

                @SuppressWarnings("rawtypes")
                @Override
                public void getParameterCompleted(short arg0, Vector arg1, long arg2)
                {
                }

                @Override
                public void getParameterFailed(short arg0, short arg1, long arg2)
                {
                }

                @Override
                public void resourceUnloaded(short arg0)
                {
                }

                @SuppressWarnings("rawtypes")
                @Override
                public void setParameterCompleted(short arg0, Vector arg1, long arg2)
                {
                }

                @Override
                public void setParameterFailed(short arg0, short arg1, long arg2)
                {
                }
            }
            :
            // SDK2
            new com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResourceListener()
            {
                @Override
                public void PDXCommandCreated(String sessionId)
                {
                    (new PDXCommandCreatedRunnable(sessionId)).run();
                }

                @Override
                public void PDXCreateCommandFailed(short reasonCode)
                {
                    (new PDXCreateCommandFailedRunnable(reasonCode)).run();
                }

                @SuppressWarnings("rawtypes")
                @Override
                public void getParameterCompleted(short arg0, Vector arg1, long arg2)
                {
                }

                @Override
                public void getParameterFailed(short arg0, short arg1, long arg2)
                {
                }

                @Override
                public void resourceUnloaded(short arg0)
                {
                }

                @SuppressWarnings("rawtypes")
                @Override
                public void setParameterCompleted(short arg0, Vector arg1, long arg2)
                {
                }

                @Override
                public void setParameterFailed(short arg0, short arg1, long arg2)
                {
                }

                @Override
                public void sendMdsMessageStatus(int statusCode, int transactionId) {
                    Checker.checkState(this, false, "NMASResourceListener.sendMdsMessageStatus() should NEVER be triggered here!!!");
                }

                @Override
                public boolean receiveMdsMessageStatus(int statusCode,
                        int transactionId, com.nuance.nmsp.client2.sdk.components.core.internal.pdx.PDXDictionary message) {
                    Checker.checkState(this, false, "NMASResourceListener.receiveMdsMessageStatus() should NEVER be triggered here!!!");
                    return true;
                }
            };
    }

    private Object createCommandListener()
    {
        class PDXCommandEventRunnable implements Runnable
        {
            private final short _event;
            private PDXCommandEventRunnable(short event)
            {
                _event = event;
            }
            @Override
            public void run() {
                Logger.debug(CommandContextImpl.this, "[LATCHK] Command Event: " + _event);
                _tran.currentState().commandEvent(_event);
            }
        }
        class PDXQueryErrorReturnedRunnable implements Runnable
        {
            private final Object _error;
            private PDXQueryErrorReturnedRunnable(Object error)
            {
                _error = error;
            }
            @Override
            public void run() {
                int error = !_config.newCop() ? ((com.nuance.nmsp.client.sdk.components.resource.nmas.QueryError)_error).getError() : ((com.nuance.nmsp.client2.sdk.components.resource.nmas.QueryError)_error).getError();
                String description = !_config.newCop() ? ((com.nuance.nmsp.client.sdk.components.resource.nmas.QueryError)_error).getDescription() : ((com.nuance.nmsp.client2.sdk.components.resource.nmas.QueryError)_error).getDescription();
                Logger.debug(CommandContextImpl.this, "[LATCHK] PDX Query Error Returned: " + error + "("
                        + description + ")");
                _tran.currentState().queryError(error, description);
            }
        }
        class PDXQueryResultReturnedRunnable implements Runnable
        {
            private final Object _result;
            private PDXQueryResultReturnedRunnable(Object result)
            {
                _result = result;
            }
            @Override
            public void run() {
                Logger.debug(CommandContextImpl.this, "[LATCHK] PDX Query Result Returned");
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (!_config.newCop())
                            _tran.currentState().queryResult(PdxConverter.dictionaryFromPdx((com.nuance.nmsp.client.sdk.components.resource.nmas.QueryResult)_result));
                        else
                            _tran.currentState().queryResult(PdxConverter.dictionaryFromPdx((com.nuance.nmsp.client2.sdk.components.resource.nmas.QueryResult)_result));
                    }
                });
            }
        }
        class PDXQueryRetryReturnedRunnable implements Runnable
        {
            private final Object _retry;
            private PDXQueryRetryReturnedRunnable(Object retry)
            {
                _retry = retry;
            }
            @Override
            public void run() {
                final int cause     = !_config.newCop() ? ((com.nuance.nmsp.client.sdk.components.resource.nmas.QueryRetry)_retry).getCause()  : ((com.nuance.nmsp.client2.sdk.components.resource.nmas.QueryRetry)_retry).getCause();
                final String prompt = !_config.newCop() ? ((com.nuance.nmsp.client.sdk.components.resource.nmas.QueryRetry)_retry).getPrompt() : ((com.nuance.nmsp.client2.sdk.components.resource.nmas.QueryRetry)_retry).getPrompt();
                final String name   = !_config.newCop() ? ((com.nuance.nmsp.client.sdk.components.resource.nmas.QueryRetry)_retry).getName()   : ((com.nuance.nmsp.client2.sdk.components.resource.nmas.QueryRetry)_retry).getName();
                Logger.debug(CommandContextImpl.this, "[LATCHK] PDX Query Retry Returned: " + cause + "("
                        + name + ")");
                _tran.currentState().queryRetry(prompt, name);
            }
        }

        return !_config.newCop() ?
            // SDK1
            new com.nuance.nmsp.client.sdk.components.resource.nmas.PDXCommandListener()
            {
                @Override
                public void PDXCommandEvent(short event)
                {
                    (new PDXCommandEventRunnable(event)).run();
                }

                @Override
                public void PDXQueryErrorReturned(com.nuance.nmsp.client.sdk.components.resource.nmas.QueryError error)
                {
                    (new PDXQueryErrorReturnedRunnable(error)).run();
                }

                @Override
                public void PDXQueryResultReturned(com.nuance.nmsp.client.sdk.components.resource.nmas.QueryResult result)
                {
                    (new PDXQueryResultReturnedRunnable(result)).run();
                }

                @Override
                public void PDXQueryRetryReturned(com.nuance.nmsp.client.sdk.components.resource.nmas.QueryRetry retry)
                {
                    (new PDXQueryRetryReturnedRunnable(retry)).run();
                }
            }
            :
            new com.nuance.nmsp.client2.sdk.components.resource.nmas.PDXCommandListener()
            {
                @Override
                public void PDXCommandEvent(short event)
                {
                    (new PDXCommandEventRunnable(event)).run();
                }

                @Override
                public void PDXQueryErrorReturned(com.nuance.nmsp.client2.sdk.components.resource.nmas.QueryError error)
                {
                    (new PDXQueryErrorReturnedRunnable(error)).run();
                }

                @Override
                public void PDXQueryResultReturned(com.nuance.nmsp.client2.sdk.components.resource.nmas.QueryResult result)
                {
                    (new PDXQueryResultReturnedRunnable(result)).run();
                }

                @Override
                public void PDXQueryRetryReturned(com.nuance.nmsp.client2.sdk.components.resource.nmas.QueryRetry retry)
                {
                    (new PDXQueryRetryReturnedRunnable(retry)).run();
                }
            };
    }
}
