package com.nuance.dragon.toolkit.cloudservices;

import java.util.Enumeration;
import java.util.Hashtable;

import android.content.Context;

import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.nmsp.client.sdk.common.defines.NMSPDefines;

/**
 * By default the system will not use any proxy configuration, the
 * existence of the ProxyConfig will tell NMT to pick up the system's
 * default proxy configuration from the device directly.
 *
 * There are also 2 different proxy connection options for the application
 * to decide what to do if the device system does not configure any proxy
 * settings.
 */
public class ProxyConfig implements NMTConfig
{
    /**
     * The proxy connection option:
     * <br><br>
     * USE_PROXY_ONLY: Always attempt to connect through a proxy,
     * fail if no proxy is configured
     * <br><br>
     * USE_PROXY_IF_AVAILABLE: Always attempt to connect through a proxy,
     * If no proxy is specified on the device, make a direct connection.
     * Otherwise: connect through the proxy. If the proxy socket fails or
     * proxy traffic is denied, the connection fails.
     */
    public static enum ProxyOption { USE_PROXY_ONLY, USE_PROXY_IF_AVAILABLE };

    private Hashtable<String, Object> _parameters = new Hashtable<String, Object>();
    private ProxyOption _proxyOption = ProxyOption.USE_PROXY_IF_AVAILABLE;

    /**
     * Make a ProxyConfig object with default proxy option: USE_PROXY_IF_AVAILABLE.
     * <br><br>
     * @param context      the android context object required for retrieving the
     *                     proxy setting from the device.
     **/
    public ProxyConfig(Context context)
    {
        this(context, ProxyOption.USE_PROXY_IF_AVAILABLE);
    }

    /**
     * Make a ProxyConfig object.
     *
     * @param context      the android context object required for retrieving the
     *                     proxy setting from the device.
     *
     * @param proxyOption  the desired proxy option chosen from the {@link ProxyOption}.
     */
    public ProxyConfig(Context context, ProxyOption proxyOption)
    {
        Checker.checkArgForNull("Context", context);

         _parameters.put(NMSPDefines.NMSP_DEFINES_ANDROID_CONTEXT, context);
         _proxyOption = proxyOption;

         if ( _proxyOption == ProxyOption.USE_PROXY_ONLY )
         {
             _parameters.put(NMSPDefines.NMSP_DEFINES_USING_DEVICE_PROXY_SETTING_ONLY, "TRUE".getBytes());
         }
         else if ( _proxyOption == ProxyOption.USE_PROXY_IF_AVAILABLE )
         {
             _parameters.put(NMSPDefines.NMSP_DEFINES_USING_DEVICE_PROXY_SETTING_IF_AVAILABLE, "TRUE".getBytes());
         }
    }

    /**
     * Retrieve the proxy option.
     *
     * @return the proxy option choice.
     */
    public ProxyOption getProxyOption()
    {
        return _proxyOption;
    }

    @Override
    public int size()
    {
        return _parameters.size();
    }

    @Override
    public Enumeration<String> keys()
    {
        return _parameters.keys();
    }

    @Override
    public Object getValue(String key)
    {
        return _parameters.get(key);
    }
}
