package com.nuance.dragon.toolkit.cloudservices;

import java.util.LinkedList;
import java.util.List;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.sources.SingleSinkSource;
import com.nuance.dragon.toolkit.cloudservices.CommandContext.NMTAudioSink;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * A TTS parameter to be sent via {@link Transaction#addParam}.
 * <p>
 * A TTS parameter will request audio to be streamed from the server. The
 * downloaded audio will be available via the {@link AudioSource} returned by
 * {@link #getAudioSource}.
 */
public class TTSParam extends Param
{
//TODO: (wlwl) TTSParam should extend a command AudioSourceBase class which
//             could be extended by RecorderSource as well
    private final AudioType _audioType;
    private final Data.Dictionary _dict;
    private TTSAudioSource _ttsAudioSource;

    /**
     * @param key The key for this parameter.
     * @param dict The dictionary containing the data for this parameter (e.g.,
     * the text to TTS).
     * @param audioType The type of audio.
     */
    public TTSParam(String key,
                    Data.Dictionary dict,
                    AudioType audioType)
    {
        super(TYPE_TTS, key);
        Checker.checkArgForNull("dict", dict);
        _dict = dict;
        _audioType = audioType;
        _ttsAudioSource = null;
    }

    @Override
    void send(CommandContext commandContext) throws Exception
    {
        if (_ttsAudioSource == null)
            _ttsAudioSource = new TTSAudioSource(_audioType, Factory.createNMTHandler());
        
        Checker.checkState(CloudServices.class, ((CommandContextImpl) commandContext).getConfig().playerCodec().equals(_ttsAudioSource.getAudioType()),
                "The audio type specified in the TTSSpec does not match the player codec specified in the CloudServices");
        commandContext.sendTTSParam(key, _dict, _ttsAudioSource);
    }

    /**
     * Get the audio source that will provide the TTS audio.
     * @return The audio source.
     */
    public AudioSource<AudioChunk> getAudioSource()
    {
        return  getAudioSource(null);
    }

    /**
     * Get the audio source that will provide the TTS audio.
     * @return The audio source.
     */
    public AudioSource<AudioChunk> getAudioSource(NMTHandler mainThreadHandler)
    {
        if (_ttsAudioSource == null)
            _ttsAudioSource = new TTSAudioSource(_audioType, mainThreadHandler);
        
        return _ttsAudioSource;
    }

    private final static class TTSAudioSource extends SingleSinkSource<AudioChunk> implements NMTAudioSink
    {
        private final AudioType _audioType;
        private final List<AudioChunk>     _buffer;
        private boolean                    _done;
        private boolean                    _lastBuffReceived; // touched from the main thread

        TTSAudioSource(AudioType audioType, NMTHandler mainThreadHandler)
        {
            super(mainThreadHandler);
            
            assert (audioType != null);

            _buffer = new LinkedList<AudioChunk>();
            _audioType = audioType;
            _done = false;
        }

        @Override
        public void addAudioBuf(byte[] buffer, int offset, int length, boolean isLastBuf)
        {
        	Logger.verbose(this, "[LATCHK] network TTS: addAudioBuf [" + isLastBuf + "]: packet size: " + (buffer != null ? buffer.length : "NULL") + " bytes. Offset: " + offset + ", length: " + length);
            if (_done)
            {
                return;
            } else if (isLastBuf)
            {
                _done = true;
            }

            if (buffer != null && length > 0)
            {
                if (_audioType.encoding == AudioType.Encoding.PCM_16)
                {
                    short[] shorts = new short[length / 2];
                    for (int i = 0; i < length / 2; i++)
                    {
                        shorts[i] = (short) (((short) buffer[offset + i * 2] & 0x00FF)
                                | ((short) buffer[offset + i * 2 + 1] << 8 & 0xFF00));
                    }
                    handleNewAudio(new AudioChunk(_audioType, shorts));
                } else if (_audioType.encoding == AudioType.Encoding.SPEEX)
                {
                    byte[] buf = new byte[length];
                    System.arraycopy(buffer, offset, buf, 0, length);
                    handleNewAudio(new AudioChunk(_audioType, buf, 0)); // TODO: can we get the duration?
                } else if (_audioType.encoding == AudioType.Encoding.OPUS)
                {
                    byte[] buf = new byte[length];
                    System.arraycopy(buffer, offset, buf, 0, length);
                    handleNewAudio(new AudioChunk(_audioType, buf, 0));
                } else if (_audioType.encoding == AudioType.Encoding.ULAW)
                {
                    byte[] buf = new byte[length];
                    System.arraycopy(buffer, offset, buf, 0, length);
                    handleNewAudio(new AudioChunk(_audioType, buf, 0));
                } else if (_audioType.encoding == AudioType.Encoding.GSM_FR)
                {
                    byte[] buf = new byte[length];
                    System.arraycopy(buffer, offset, buf, 0, length);
                    handleNewAudio(new AudioChunk(_audioType, buf, 0));
                } else if (_audioType.encoding == AudioType.Encoding.AMR0 ||
                           _audioType.encoding == AudioType.Encoding.AMR1 ||
                           _audioType.encoding == AudioType.Encoding.AMR2 ||
                           _audioType.encoding == AudioType.Encoding.AMR3 ||
                           _audioType.encoding == AudioType.Encoding.AMR4 ||
                           _audioType.encoding == AudioType.Encoding.AMR5 ||
                           _audioType.encoding == AudioType.Encoding.AMR6 ||
                           _audioType.encoding == AudioType.Encoding.AMR7)
                {
                    byte[] buf = new byte[length];
                    System.arraycopy(buffer, offset, buf, 0, length);
                    handleNewAudio(new AudioChunk(_audioType, buf, 0));
                } else if (_audioType.encoding == AudioType.Encoding.MP3_8KBPS ||
                           _audioType.encoding == AudioType.Encoding.MP3_16KBPS ||
                           _audioType.encoding == AudioType.Encoding.MP3_24KBPS ||
                           _audioType.encoding == AudioType.Encoding.MP3_32KBPS ||
                           _audioType.encoding == AudioType.Encoding.MP3_40KBPS ||
                           _audioType.encoding == AudioType.Encoding.MP3_48KBPS ||
                           _audioType.encoding == AudioType.Encoding.MP3_56KBPS ||
                           _audioType.encoding == AudioType.Encoding.MP3_64KBPS ||
                           _audioType.encoding == AudioType.Encoding.MP3_80KBPS ||
                           _audioType.encoding == AudioType.Encoding.MP3_96KBPS ||
                           _audioType.encoding == AudioType.Encoding.MP3_112KBPS ||
                           _audioType.encoding == AudioType.Encoding.MP3_128KBPS ||
                           _audioType.encoding == AudioType.Encoding.MP3_144KBPS ||
                           _audioType.encoding == AudioType.Encoding.MP3_160KBPS)
                {
                    byte[] buf = new byte[length];
                    System.arraycopy(buffer, offset, buf, 0, length);
                    handleNewAudio(new AudioChunk(_audioType, buf, 0));
                }
            }

            if (isLastBuf)
            {
                handleLastBuffReceived();
                handleSourceClosed();
            }
        }

        @Override
        protected AudioChunk getAudioChunk()
        {
        	Logger.verbose(this, "[LATCHK] network TTS: getAudioChunk: chunk count: " + _buffer.size());
            if (_buffer.isEmpty())
                return null;

            return _buffer.remove(0);
        }

        @Override
        public AudioType getAudioType()
        {
            return _audioType;
        }

        @Override
        public int getChunksAvailable()
        {
        	Logger.verbose(this, "[LATCHK] network TTS: getChunksAvailable: chunk count: " + _buffer.size());
            return _buffer.size();
        }

        @Override
        public boolean isActive()
        {
            if (_lastBuffReceived) {
                int chunksAvailable = getChunksAvailable();
                if (chunksAvailable == 0)
                    return false;
                else {
                    Logger.debug(this, "still active: " + chunksAvailable
                            + " available audio chunks left");
                }
            }
            return true;
        }

        private void handleNewAudio(final AudioChunk audio)
        {
            _mainThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    _buffer.add(audio);
                    notifyChunksAvailable();
                }
            });
        }

        private void handleSourceClosed()
        {
            _mainThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    notifySourceClosed();
                }
            });
        }

        private void handleLastBuffReceived()
        {
            _mainThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    _lastBuffReceived = true;
                }
            });
        }
    }
}
