package com.nuance.dragon.toolkit.cloudservices;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.WorkerThread;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.impl.HandlerOem;
import com.nuance.dragon.toolkit.oem.impl.WorkerThreadOem;

/**
 * Defines a basic transaction with an NMAS: <br>
 *  - Command name <br>
 *  - Parameters <br>
 *  - Audio sink(s), for accumulating TTS audio <br>
 *  - Audio source(s), for feeding the audio to be recognized
 */
public final class Transaction
{
    /**
     * Listens for when the transaction starts and finishes.
     */
    public interface Listener
    {
        /**
         * The transaction has started.
         *
         * @param t The transaction
         */
        public void onTransactionStarted(Transaction t);
        
        /**
         * The transaction is available for the server to process.
         * All the necessary information for the transaction to get started on the server side is being transmitted.
         *
         * @param t The transaction
         */
        public void onTransactionProcessingStarted(Transaction t);

        /**
         * A result was received from the server. The 'complete' parameter lets
         * the listener know if more results are coming from the server.
         *
         * @param t The transaction
         * @param result The result
         * @param complete True if there are no more results expected
         */
        public void onTransactionResult(Transaction t, TransactionResult result, boolean complete);

        /**
         * An error was received from the server and the transaction is
         * complete. This also is called if there is an error creating the
         * transaction internally.
         *
         * @param t The transaction
         * @param error The error
         */
        public void onTransactionError(Transaction t, TransactionError error);
        
		/**
		 * The Id of the transaction is generated. 
		 *
		 * @param transactionId   The Id of the transaction.
		 */ 
        public void onTransactionIdGenerated(String transactionId);
    }

    /**
     * Notifies the {@link CloudServices} module that the transaction is ready
     * to start or has finished.
     */
    interface InternalListener
    {
        /**
         * @param t The transaction
         */
        void onReadyToStart(Transaction t);

        /**
         * @param t The transaction
         */
        void onFinished(Transaction t);
    }

    /**
     * Some default transaction priorities used by DMT modules.
     */
    public static final class DefaultPriorities
    {
        /** Suggested low priority */
        public static final int LOW = 0;

        /** Suggested medium priority */
        public static final int MEDIUM = 5;

        /** Suggested high priority */
        public static final int HIGH = 10;

        /** Priority used for recognition transactions */
        public static final int RECOGNITION = 6;

        /** Priority used for TTS transactions */
        public static final int TEXT_TO_SPEECH = 5;

        /** Priority used for normal data upload transactions */
        public static final int DATA_UPLOAD = 1;

        /**
         * Priority used for initial data upload transactions (higher to ensure
         * that the initial upload occurs prior to other transactions which may
         * be dependent on the uploaded data).
         */
        public static final int FIRST_DATA_UPLOAD = 10;
        
        /** Priority used for data download transactions */
        public static final int DATA_DOWNLOAD = 1;
    }

    private boolean _readyToStart;
    private boolean _finished;
    private final List<Param> _params;
    private final String _name;
    private final Data.Dictionary _settings;
    private final Listener _listener;
    private final int _timeout;
    private InternalListener _internalListener;
    private TransactionState _currentState;
    private int _priority;
    private String _sessionId;

    private int _cadenceMs;
    private TransactionResult _lastCachedResult;
    private boolean _isFinal;
    private boolean _isCadenceStarted;
    private WorkerThread _resultThread;
    private volatile NMTHandler _mainThreadHandler;
    private final NMTHandler _resultThreadHandler;

    /**
     * Like {@link #Transaction(String, com.nuance.dragon.toolkit.core.data.Data.Dictionary, Listener, int, boolean)},
     * with the transaction starting 'enabled'.
     */
    public Transaction(String commandName, Data.Dictionary commandSettings, Listener listener, int timeout)
    {
        this(commandName, commandSettings, listener, timeout, true);
    }

    /**
     * Create a Transaction object.
     *
     * @param commandName      The name of the NMAS command.
     * @param commandSettings  The command settings, it could be NULL.
     * @param listener         The Transaction Listener.
     * @param timeout          The Transaction (command) time out value (in milliseconds).
     * @param startEnabled     False to delay the transaction executing until
     * {@link #enable} is called. A disabled transaction will block the
     * transaction queue until it is enabled or canceled.
     */
    public Transaction(String commandName, Data.Dictionary commandSettings, Listener listener, int timeout, boolean startEnabled)
    {
        this(commandName, commandSettings, listener, timeout, startEnabled, 0);
    }

    /**
     * Create a Transaction object.
     *
     * @param commandName      The name of the NMAS command.
     * @param commandSettings  The command settings, it could be NULL.
     * @param listener         The Transaction Listener.
     * @param timeout          The Transaction (command) time out value (in milliseconds).
     * @param startEnabled     False to delay the transaction executing until {@link #enable}
     *                         is called. A disabled transaction will block the transaction queue
     *                         until it is enabled or canceled.
     * @param cadenceMs        The cadence interval for word-streaming result
     */
    public Transaction(String commandName, Data.Dictionary commandSettings, Listener listener, int timeout, boolean startEnabled, int cadenceMs)
    {
        Checker.checkStringArgForNullOrEmpty("commandName", commandName);
        Checker.checkArgForNull("listener", listener);
        Checker.checkArgForCondition("timeout", "equal or greater than 0", (timeout >= 0));
        Checker.checkArgForCondition("cadenceMs", "equal or greater than 0", (cadenceMs >= 0));

        _params = new ArrayList<Param>();
        _name = commandName;
        _settings = commandSettings;
        _listener = listener;
        _timeout = timeout;
        _currentState = new TransactionStateIdle(this);
        _priority = 0;
        _readyToStart = startEnabled;

        _lastCachedResult = null;
        _cadenceMs = cadenceMs;
        _isCadenceStarted = false;
        _isFinal = false;

        if (cadenceMs > 0)
        {
            _resultThread = new WorkerThreadOem("com.nuance.dragon.toolkit.cloudservices.recognizer");
            _resultThread.start();

            _resultThreadHandler = _resultThread.getHandler();
        }
        else
        {
            _resultThread = null;
            _resultThreadHandler = null;
        }
        _mainThreadHandler = null;
    }

    /**
     * return the command name of this transaction
     *
     * @return The command name.
     */
    String getCommandName()
    {
        return _name;
    }

    /**
     * return the command settings
     *
     * @return The command settings.
     */
    Data.Dictionary getCommandSettings()
    {
        return _settings;
    }

    /**
     * return the command (transaction) timeout value
     *
     * @return The command timeout.
     */
    int getCommandTimeout()
    {
        return _timeout;
    }

    /**
     * set the priority of this transaction in the
     * cloud service transaction queue
     *
     * @param priority
     */
    void setPriority(int priority)
    {
        _priority = priority;
    }

    /**
     * get the priority of this transaction
     *
     * @return The priority.
     */
    int getPriority()
    {
        return _priority;
    }

    /**
     * Add a parameter to this transaction.
     *
     * @param p
     */
    public void addParam(final Param p)
    {
        Runnable r = new Runnable()
        {
            @Override
            public void run() {
                _params.add(p);
                _currentState.paramAdded(p);
            }
        };
        execute(r);
    }

    /**
     * Enable this transaction to be started. This has no effect if the
     * transaction was created enabled.
     */
    public void enable()
    {
        Runnable r = new Runnable()
        {
            @Override
            public void run() {
                readyToStart();
            }
        };
        execute(r);
    }

    /**
     * Ends the transaction. After this is called, no parameters can be added.
     * The transaction is not necessarily complete at this point since
     * parameters may be queued to be sent or we may be waiting on results, this
     * just means the client is finished constructing it.
     */
    public void finish()
    {
        Runnable r = new Runnable()
        {
            @Override
            public void run() {
                _finished = true;
                _currentState.transactionFinished();
            }
        };
        execute(r);
    }

    /**
     * Cancels this transaction. This message will be propagated all the way to
     * the server and the transaction will be removed from the Speech Core
     * queue.
     */
    public void cancel()
    {
        Runnable r = new Runnable()
        {
            @Override
            public void run() {
                _currentState.cancel();
            }
        };
        execute(r);
    }

    /**
     * If a transaction is active, it is not waiting in queue, it is either
     * currently running or waiting for a result.
     *
     * @return true if this transaction is active.
     */
    public boolean isActive()
    {
        return _currentState.isActive();
    }

    /**
     * If a transaction is finished, finish() has been called and no further
     * parameters can be added.
     *
     * @return true if this transaction has been finished.
     */
    public boolean isFinished()
    {
        return _finished;
    }

    /**
     * If a transaction is complete, it means that it is over. We have received
     * all results and no further callbacks will be executed. The transaction
     * can be discarded.
     *
     * @return true if this transaction has completed.
     */
    public boolean isComplete()
    {
        return _currentState.isComplete();
    }

    /**
     * Called when transaction has been added to CloudServices.
     */
    void transactionAdded()
    {
        _mainThreadHandler = new HandlerOem();
    }
    
    /**
     * Start the transaction (called by CloudServices)
     * @param commandContext
     */
    void start(CommandContext commandContext)
    {
        _currentState.start(commandContext);
    }

    /**
     * Monitor the transaction for when it is completed.
     * @param internalListener Listener to be notified if the transaction is
     * ready to start, or completed (or null to stop listening).
     */
    void monitorTransaction(InternalListener internalListener)
    {
        _internalListener = internalListener;
    }

    /**
     * Get all the parameters that have been added.
     * @return The parameters.
     */
    List<Param> getParams()
    {
        return _params;
    }

    /**
     * Get the transaction listener. This is the application's listener.
     * @return The listener.
     */
    Listener getListener()
    {
        return _listener;
    }

    /**
     * Get the internal listener. This is owned by the {@link CloudServices}
     * module, and is used to determine when a transaction is ready to start or
     * has finished.
     * @return The complete listener.
     */
    InternalListener getInternalListener()
    {
        return _internalListener;
    }

    /**
     * Set the cadence at which multiple results can be received.
     * <p>
     * This feature is relevant when results are streamed back from the server
     * (per-word recognition). Setting the cadence guarantees that all results
     * that are sent back will be aligned to the cadence period specified.
     * For example, if the cadence is set at 250ms, every 250ms the system
     * tries to obtain an updated result. If an updated result is found,
     * it is returned through the resultListener (see startRecognition).
     * If no updated result is available, the last result is sent back in this cycle.
     * The final result is always sent back as soon as it is available.
     * When the cadence is set to 0, updates are sent back through the Listener
     * as soon as they are available. The default cadence is 0.
     *
     * @param cadenceMs    The time between successive result updates.
     */
    public void setResultCadence(final int cadenceMs)
    {
        Runnable r = new Runnable()
        {
            @Override
            public void run() {
                _cadenceMs = cadenceMs;
            }
        };
        execute(r);
    }

    /**
     * Called by the Cloud Services module when a result is received.
     * @param result
     * @param complete
     */
    void handleResult(TransactionResult result, boolean complete)
    {
        Logger.debug(this, "[LATCHK] calling onTransactionResult()");

        _lastCachedResult = result;
        _isFinal = complete;

        if (_cadenceMs == 0 || !result.isCadenceRegulatable())
        {
            _listener.onTransactionResult(this, result, complete);
        }
        else if (!_isCadenceStarted)
        {
            if (!_isCadenceStarted)
                _isCadenceStarted = true;
            final Runnable cadenceJob = new Runnable()
            {
                @Override
                public void run()
                {
                    _mainThreadHandler.post(new Runnable() {
                        @Override
                        public void run()
                        {
                            Logger.debug(Transaction.this, "cadence result at: " + System.currentTimeMillis());
                            _listener.onTransactionResult(Transaction.this, _lastCachedResult, _isFinal);
                            if (_isFinal) {
                                _resultThread.stop();
                                _resultThread = null;
                                Logger.debug(Transaction.this, "cadence result thread stopped");
                            }
                        }
                    });
                    if ((_cadenceMs > 0 && !_isFinal))
                    {
                        _resultThreadHandler.postDelayed(this, _cadenceMs);
                    }
                }
            };
            _resultThreadHandler.postDelayed(cadenceJob, _cadenceMs);
        }
    }

    /**
     * Called by the Cloud Services module when an error is received.
     * @param error
     */
    void handleError(TransactionError error)
    {
        Logger.debug(this, "[LATCHK] calling handleError()");
        _listener.onTransactionError(this, error);
    }

    /**
     * Switch to another transaction state. This will call
     * {@link TransactionState#leave} on the old state, and
     * {@link TransactionState#enter} on the new state.
     * @param state
     */
    void switchToState(TransactionState state)
    {
        _currentState.leave();
        _currentState = state;
        _currentState.enter();
    }

    /**
     * Get the current transaction state.
     * @return The current state.
     */
    TransactionState currentState()
    {
        return _currentState;
    }

    /**
     * Set the session ID
     * @param sessionId
     */
    void setSessionId(String sessionId)
    {
        _sessionId = sessionId;
    }

    /**
     * @return the session ID previously set by {@link #setSessionId}.
     */
    String getSessionId()
    {
        return _sessionId;
    }

    /**
     * @return true if the transaction is ready to start
     */
    boolean isReadyToStart()
    {
        return _readyToStart;
    }

    private void readyToStart()
    {
        if (!_readyToStart)
        {
            _readyToStart = true;
            if (_internalListener != null)
            {
                _internalListener.onReadyToStart(this);
            }
        }
    }
    
    private void execute(Runnable r)
    {
        if (_mainThreadHandler == null)
            r.run();
        else
            _mainThreadHandler.postToLooper(r);
    }
}
