package com.nuance.dragon.toolkit.cloudservices;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.nmsp.client2.sdk.components.resource.nmas.PDXCommandListener;
import com.nuance.nmsp.client2.sdk.components.resource.nmas.QueryError;
import com.nuance.nmsp.client2.sdk.components.resource.nmas.QueryRetry;

/**
 * A transaction error.
 */
public class TransactionError implements JSONCompliant
{
    /**
     * Possible error types.
     */
    public static class Types
    {
        public static final int CANCELED = 0;
        public static final int CONNECTION_ERROR = 1;
        public static final int RETRY_ERROR = 2;
        public static final int QUERY_ERROR = 3;
        public static final int OTHER_ERROR = 4;
        public static final int SECURITY_ERROR = 5;
    }

    /**
     * Possible error codes when the type is {@link Types#CONNECTION_ERROR}.
     */
    public static class ConnectionErrorCodes
    {
        // For error codes defined in DMT, use a value too big for a short to ensure we don't clash with NMSP
        public static final int NETWORK_UNAVAILABLE             = 0x0fff0000;

        public static final int COMMAND_IDLE_FOR_TOO_LONG       = PDXCommandListener.COMMAND_IDLE_FOR_TOO_LONG;
        public static final int REMOTE_DISCONNECTION            = PDXCommandListener.REMOTE_DISCONNECTION;
        public static final int TIMED_OUT_WAITING_FOR_RESULT    = PDXCommandListener.TIMED_OUT_WAITING_FOR_RESULT;
        public static final int COMMAND_ENDED_UNEXPECTEDLY      = PDXCommandListener.COMMAND_ENDED;
    }

    private static final String KEY_TYPE = "type";
    private static final String KEY_TEXT = "text";
    private static final String KEY_PARAM = "param";
    private static final String KEY_PROMPT = "prompt";
    private static final String KEY_CODE = "code";
    private static final String KEY_SESSION = "session";

    private final int _type;
    private final String _text;
    private final String _param;
    private final String _prompt;
    private final int _code;
    private final String _sessionId;

    /**
     * Create a transaction error with type {@link Types#OTHER_ERROR} and default text.
     * @param transaction (cannot be null)
     */
    TransactionError(Transaction transaction)
    {
        this(transaction, Types.OTHER_ERROR, null, null, null, 0);
    }

    /**
     * Create a transaction error from a query retry.
     *
     * @param transaction (cannot be null)
     * @param retry (cannot be null)
     */
    TransactionError(Transaction transaction, QueryRetry retry)
    {
        this(transaction, Types.RETRY_ERROR, null, nullEmptyString(retry.getName()), nullEmptyString(retry.getPrompt()), 0);
    }

    /**
     * Create a transaction error from a query error.
     * @param transaction (cannot be null)
     * @param error (cannot be null)
     */
    TransactionError(Transaction transaction, QueryError error)
    {
        this(transaction, Types.QUERY_ERROR, nullEmptyString(error.getDescription()), null, null, error.getError());
    }

    /**
     * Create a transaction error with type {@link Types#OTHER_ERROR} and the given
     * text.
     * @param transaction
     * @param errorText
     */
    TransactionError(Transaction transaction, String errorText)
    {
        this(transaction, Types.OTHER_ERROR, errorText, null, null, 0);
    }

    /**
     * Create a transaction error with the given type and default text.
     * @param transaction
     * @param type
     */
    TransactionError(Transaction transaction, int type)
    {
        this(transaction, type, null, null, null, 0);
    }

    /**
     * Create a transaction error with the given type and supplemental code.
     * @param transaction
     * @param type
     * @param errorCode
     */
    TransactionError(Transaction transaction, int type, int errorCode)
    {
        this(transaction, type, null, null, null, errorCode);
    }

    /**
     * Create a transaction error with the given type, text, and supplemental code.
     * @param transaction
     * @param type
     * @param errorText
     * @param prompt
     * @param errorCode
     */
    TransactionError(Transaction transaction, int type, String errorText, String param, String prompt, int errorCode)
    {
        _type = type;
        _text = errorText;
        _param = param;
        _prompt = prompt;
        _code = errorCode;
        _sessionId = transaction.getSessionId();
    }

    /**
     * Create a transaction error from JSON.
     * @param json
     * @throws JSONException
     */
    private TransactionError(org.json.JSONObject json) throws JSONException
    {
        _type = json.getInt(KEY_TYPE);
        _text = json.optString(KEY_TEXT, null);
        _param = json.optString(KEY_PARAM, null);
        _prompt = json.optString(KEY_PROMPT, null);
        _code = json.optInt(KEY_CODE, 0);
        _sessionId = json.optString(KEY_SESSION, null);
    }

    /**
     * Get the session ID that this error came from.
     * @return The session ID, or null if the error occurred before a session
     * was established.
     */
    public String getSessionId()
    {
        return _sessionId;
    }

    /**
     * Get the type.
     * @return A type from {@link Types}
     */
    public int getType()
    {
        return _type;
    }

    /**
     * Get the text of the error.
     * @return May return null
     */
    public String getErrorText()
    {
        return _text;
    }

    /**
     * Get the parameter name associated with the error, if any.
     * @return May return null
     */
    public String getParameterWithError()
    {
        return _param;
    }

    /**
     * Get the suggestion prompt, if any.
     * @return May return null
     */
    public String getPrompt()
    {
        return _prompt;
    }

    /**
     * Get the extra error code, if applicable (i.e. for NMSP QueryError and
     * connection errors).
     * See {@link ConnectionErrorCodes} for error codes specific to connection
     * errors.
     * @return The error code
     */
    public int getErrorCode()
    {
        return _code;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        o.tryPut(KEY_TYPE, _type);
        o.tryPut(KEY_TEXT, _text);
        o.tryPut(KEY_PARAM, _param);
        o.tryPut(KEY_PROMPT, _prompt);
        o.tryPut(KEY_CODE, _code);
        return o;
    }

    public static TransactionError createFromJSON(org.json.JSONObject json) throws JSONException
    {
        return new TransactionError(json);
    }

    private static String nullEmptyString(String str)
    {
        if (str != null && str.length() == 0)
            return null;
        return str;
    }
}
