package com.nuance.dragon.toolkit.cloudservices;

import org.json.JSONException;

import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * A result from a Cloud transaction, received by {@link Transaction.Listener#onTransactionResult}.
 */
public class TransactionResult implements JSONCompliant
{
    protected static final String KEY_RESULT = "res";
    protected static final String KEY_SESSION = "session";
    private static final String KEY_RESULT_TYPE = "result_type";
    private static final String NO_RESULT_TYPE  = "no_result_type";
    private static final String KEY_RESULT_FORMAT = "result_format";
    private static final String NO_RESULT_FORMAT  = "no_result_format";
    private static final String KEY_FINAL_RESPONSE = "final_response";
    private static final String KEY_CADENCE_REGULATABLE = "cadence_regulatable_result";
    private static final String KEY_UTTERANCE_NUMBER = "utterance-number";
    private static final String VALUE_CADENCE_PARTIAL_RESULT = "partialRecognition";

    private Data.Dictionary _resultDict;
    private String _resultType;
    private String _resultFormat;
    private final String _sessionId;

    TransactionResult(Transaction transaction, Data.Dictionary result)
    {
        _resultDict = result;
        _resultType = getResultType(result);
        _resultFormat = getResultFormat(result);
        _sessionId = transaction.getSessionId();
    }

    /*TransactionResult(Transaction transaction, com.nuance.nmsp.client2.sdk.components.resource.nmas.QueryResult result)
    {
        _resultDict = PdxConverter.dictionaryFromPdx(result);
        _resultType = getResultType(_resultDict);
        _resultFormat = getResultFormat(_resultDict);
        _sessionId = transaction.getSessionId();
    }*/

    private TransactionResult(org.json.JSONObject json) throws JSONException
    {
        org.json.JSONObject result = json.optJSONObject(KEY_RESULT);
        if (result != null)
            _resultDict = Data.Dictionary.createFromJSON(result);
        else
            _resultDict = null;

        _resultType = getResultType(_resultDict);
        _resultFormat = getResultFormat(_resultDict);
        _sessionId = json.optString(KEY_SESSION, null);
    }

    private String getResultType(Data.Dictionary resultDict)
    {
        if (resultDict != null)
        {
            Data type = resultDict.get(KEY_RESULT_TYPE);
            if (type != null && type.getType() == Data.TYPE_STRING)
                return ((Data.String)type).value;
        }
        return NO_RESULT_TYPE;
    }

    private String getResultFormat(Data.Dictionary resultDict)
    {
        if (resultDict != null)
        {
            Data type = resultDict.get(KEY_RESULT_FORMAT);
            if (type != null && type.getType() == Data.TYPE_STRING)
                return ((Data.String)type).value;
        }
        return NO_RESULT_FORMAT;
    }

    /**
     * Get the session ID that this result came from.
     * @return The session ID. (never be null)
     */
    public String getSessionId()
    {
        return _sessionId;
    }

    /**
     * Get the result type, taken from the 'result_type' key of the result
     * dictionary.
     * @return The result type (never be null)
     */
    public String getResultType()
    {
        return _resultType;
    }

    /**
     * Get the result format, taken from the 'result_format' key of the result
     * dictionary.
     * @return The result format (never be null)
     */
    public String getResultFormat()
    {
        return _resultFormat;
    }

    /**
     * Get the contents of this result.
     * @return The contents (never be null)
     */
    public Data.Dictionary getContents()
    {
        return _resultDict;
    }

    /**
     * Is this the final result, or are we expecting more?
     * @return true if there will be no more results.
     */
    public boolean isFinal()
    {
        Data data = _resultDict.get(KEY_FINAL_RESPONSE);

        // quick fix for prior 3.9 servers
        // it will be ok for 3.9 server, the "final_response" will be only on the top level
        if (data == null)
        {
            Data.Dictionary appResultDict = _resultDict.getDictionary("appserver_results");
            if (appResultDict != null)
            {
                data = appResultDict.get(KEY_FINAL_RESPONSE);
            }
        }
        // quick fix, it will be ok for 3.9 server
        if (data != null && data instanceof Data.Integer)
        {
            return ((Data.Integer) data).value == 1;
        }

        return true;
    }

    /**
	 * Tells whether this cadence regulation applies to this result.
     * Not all results can be subject to cadence regulation. If this method returns
     * true, it indicates that cadence regulation can be used on this result.
     * Whether cadence regulation was indeed applied depends on whether it was
     * activated by the user or not.
	 * @return True if cadence regulation applies to this result, false otherwise.
	 */
    public boolean isCadenceRegulatable()
    {
        Data data = _resultDict.get(KEY_CADENCE_REGULATABLE);

        if (data != null && data instanceof Data.String)
        {
            return ((Data.String) data).value.equals(VALUE_CADENCE_PARTIAL_RESULT);
        }

        return false;
    }

    /**
     * Returns whether the utterance is complete or not. This method
     * should only be caleld if the intermediate response mode is enabled
     * and set to "UtteranceDetectionWithPartialRecognition".
     * @return true if the utterance is complete
     */
    public boolean isUtteranceComplete()
    {
    	Data data = _resultDict.get(KEY_UTTERANCE_NUMBER);

        // quick fix for prior 3.9 servers
        // it will be ok for 3.9 server, the "final_response" will be only on the top level
        if (data == null)
        {
            Data.Dictionary appResultDict = _resultDict.getDictionary("appserver_results");
            if (appResultDict != null)
            {
                data = appResultDict.get(KEY_UTTERANCE_NUMBER);
            }
        }
        // quick fix, it will be ok for 3.9 server
        if (data != null && data instanceof Data.Integer)
        {
            return ((Data.Integer) data).value >= 0;
        }

        return false;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        if (_resultDict != null)
        {
            o.tryPut(KEY_RESULT, _resultDict.toJSON());
        }
        o.tryPut(KEY_SESSION, _sessionId);
        return o;
    }

    public static TransactionResult createFromJSON(org.json.JSONObject json) throws JSONException
    {
        return new TransactionResult(json);
    }
}
