package com.nuance.dragon.toolkit.cloudservices;

import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.impl.HandlerOem;


abstract class TransactionState
{
    protected final CommandContext _context;
    protected final Transaction _transaction;
    private final boolean _isActive;
    private final boolean _isComplete;
    private boolean _isCurrent;
    protected NMTHandler _mainThreadHandler;

    TransactionState(CommandContext context, boolean isActive, boolean isComplete, Transaction transaction)
    {
        _transaction = transaction;
        _isActive = isActive;
        _isComplete = isComplete;
        _context = context;
    }

    /**
     * The transaction state has been entered.
     */
    void enter()
    {
        _mainThreadHandler = new HandlerOem();
        _isCurrent = true;
    }

    /**
     * The transaction state has been left.
     */
    void leave()
    {
        _isCurrent = false;
    }

    /**
     * @return Whether this is the current state of the transaction.
     */
    final boolean isCurrent()
    {
        return _isCurrent;
    }

    /**
     * Is this state an 'active' state?
     * @return true if this state is 'active'
     */
    final boolean isActive()
    {
        return _isActive;
    }

    /**
     * Is this state a 'complete' state?
     * @return true if this state is a terminal state.
     */
    final boolean isComplete()
    {
        return _isComplete;
    }

    /**
     * The transaction should start.
     *
     * @param commandContext
     */
    void start(CommandContext commandContext)
    {

    }

    /**
     * The transaction should be canceled.
     */
    void cancel()
    {
        switchToState(new TransactionStateFinished(_context, _transaction));
        _transaction.getListener().onTransactionError(_transaction, new TransactionError(_transaction, TransactionError.Types.CANCELED));
    }

    // Events
    /**
     * Retrieving a session succeeded.
     * @param param
     */
    void getSessionSucceeded(ASSession session)
    {

    }

    /**
     * Retrieving a session failed.
     * @param param
     */
    void getSessionFailed(ASError error)
    {
        handleError(new TransactionError(_transaction, error.toString()));
    }

    /**
     * A parameter has been added to the transaction.
     * @param param
     */
    void paramAdded(Param param)
    {

    }

    /**
     * The transaction has been finished (i.e. no more params will be added).
     */
    void transactionFinished()
    {

    }

    /**
     * The command has been created.
     * @param sessionId
     */
    void commandCreated(String sessionId)
    {
        _transaction.getListener().onTransactionIdGenerated(sessionId);
    }

    /**
     * The command has failed to be created.
     */
    void createCommandFailed(short reasonCode)
    {
        handleError(new TransactionError(_transaction, reasonCode == com.nuance.nmsp.client2.sdk.components.resource.common.ManagerListener.NMSP_REMOTE_DISCONNECT_SECURITY ? TransactionError.Types.SECURITY_ERROR : TransactionError.Types.CONNECTION_ERROR, TransactionError.ConnectionErrorCodes.NETWORK_UNAVAILABLE));
    }

    /**
     * A command event has occurred.
     * @param code
     */
    void commandEvent(short code)
    {
        handleError(new TransactionError(_transaction, TransactionError.Types.CONNECTION_ERROR, code));
    }

    /**
     * A command error has occurred.
     * @param code
     * @param description
     */
    void queryError(int code, String description)
    {
        handleError(new TransactionError(_transaction, TransactionError.Types.QUERY_ERROR, description, null, null, code));
    }

    /**
     * A query retry has occurred.
     * @param prompt
     * @param name
     */
    void queryRetry(String prompt, String name)
    {
        handleError(new TransactionError(_transaction, TransactionError.Types.RETRY_ERROR, null, name, prompt, 0));
    }

    /**
     * Handle any old error.
     * @param error
     */
    void handleError(final TransactionError error)
    {
        _mainThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (_isActive)
                {
                    switchToState(new TransactionStateFinished(_context, _transaction));
                    _transaction.getListener().onTransactionError(_transaction, error);
                }
            }
        });
    }

    /**
     * A command result has been received.
     * @param result
     */
    void queryResult(Data.Dictionary queryResult)
    {
        // Don't handle this by default
    }

    /**
     * Switches the transaction state
     *
     * @param newState
     */
    protected final void switchToState(TransactionState newState)
    {
        _transaction.switchToState(newState);
    }
}
