package com.nuance.dragon.toolkit.cloudservices;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.Logger;

/**
 * This state sends the transaction to the server. It initializes the command,
 * sends parameters, and waits for the transaction to be finished and/or for all
 * parameters to finish sending before transitioning to the
 * {@link TransactionStateFinished}.
 */
class TransactionStateSending extends TransactionState
{
    private final ArrayList<StreamParam>   _streamingParams;
    private boolean _completed;

    TransactionStateSending(CommandContext commandContext, Transaction transaction)
    {
        super(commandContext, true, false, transaction);
        _streamingParams = new ArrayList<StreamParam>();
        _completed = false;
    }

    @Override
    void enter()
    {
        super.enter();
        try
        {
            _context.initializeCommand(_transaction.getCommandName(), _transaction.getCommandSettings(), _transaction
                    .getCommandTimeout());
        } catch (Exception e)
        {
            Logger.error(this, "Failed to create command [" + _transaction.getCommandName() + "]", e);
            createCommandFailed(com.nuance.nmsp.client2.sdk.components.resource.common.ManagerListener.NMSP_REMOTE_DISCONNECT_UNKNOWN);
            return;
        }

        List<Param> params = _transaction.getParams();
        for (Param p : params)
        {
            try
            {
                p.send(_context);
            } catch (Exception e)
            {
                Logger.error(this, "Failed to send param [" + p.key + "] command [" + _transaction.getCommandName() + "]", e);
                handleError(new TransactionError(_transaction, TransactionError.Types.OTHER_ERROR, e.getMessage(), p.key, null, 0));
                return;
            }

            if (p.isStreamParam())
                _streamingParams.add((StreamParam)p);
        }

        if (_transaction.isFinished() || !_streamingParams.isEmpty()) {
            if (!_completed) {
                _completed = true;
                try {
                    _context.endCommand();
                    _transaction.getListener().onTransactionProcessingStarted(_transaction);
                }
                catch (Exception e) {
                    Logger.error(this, "Error ending command " + _transaction.getCommandName(), e);
                    handleError(new TransactionError(_transaction));
                }
            }

            if (!_streamingParams.isEmpty()) {
                switchToState(new TransactionStateStreaming(_context, _transaction, _streamingParams));
            }
        }

    }

    @Override
    public void paramAdded(Param param)
    {
        if (!_transaction.isFinished()) {
            if (param.isStreamParam()) {
                _streamingParams.add((StreamParam)param);
            }

            try {
                param.send(_context);
            }
            catch (Exception e) {
                Logger.error(this, "Failed to send param " + param.key, e);
                handleError(new TransactionError(_transaction, TransactionError.Types.OTHER_ERROR, null, param.key, null, 0));
            }
        } else {
            Logger.warn(this, "Transaction already finished, parameter " + param.key + " can not be sent");
        }
    }

    @Override
    void transactionFinished()
    {
        if (!_completed) {
            _completed = true;
            try {
                _context.endCommand();
                _transaction.getListener().onTransactionProcessingStarted(_transaction);
            }
            catch (Exception e) {
                Logger.error(this, "Error ending command " + _transaction.getCommandName(), e);
                handleError(new TransactionError(_transaction));
            }
        }

        if (!_streamingParams.isEmpty()) {
            switchToState(new TransactionStateStreaming(_context, _transaction, _streamingParams));
        }
    }

    @Override
    void queryResult(Data.Dictionary queryResult)
    {
        super.queryResult(queryResult);
        TransactionResult qResult = new TransactionResult(_transaction, queryResult);
        boolean isFinal = qResult.isFinal();
        if (isFinal)
        {
            switchToState(new TransactionStateFinished(_context, _transaction));
        }
        _transaction.handleResult(qResult, isFinal);
    }
}
