package com.nuance.dragon.toolkit.cloudservices.datadownloader;

import org.json.JSONException;

import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.DictionaryParam;
import com.nuance.dragon.toolkit.cloudservices.Transaction;
import com.nuance.dragon.toolkit.cloudservices.TransactionError;
import com.nuance.dragon.toolkit.cloudservices.TransactionResult;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.core.data.Data.Dictionary;
import com.nuance.dragon.toolkit.core.data.Data.Sequence;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.AsyncTaskHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * A Cloud Data Downloader downloads a server-side list through a
 * {@link CloudServices}.
 */
public class NcsCloudDataDownloader
{
    /**
     * Listens for the result of the download transaction
     */
    public interface DownloadListener
    {
        /**
         * The download succeeded
         * @param downloader The current downloader.
         * @param result The transaction result object.
         * @param resumePoint Where to resume the download. Could be null if no more items to download.
         */
        void onSuccess(NcsCloudDataDownloader downloader, TransactionResult result, ResumePoint resumePoint);

        /**
         * There was an error when downloading
         * @param downloader The current downloader.
         * @param error The transaction error object.
         */
        void onError(NcsCloudDataDownloader downloader, TransactionError error, ResumePoint resumePoint);
    }

    public static class ResumePoint implements JSONCompliant
    {
        private static final String KEY_REQUEST           = "rp_request";
        private static final String KEY_CURRENT_VERSION   = "rp_cur_version";
        private static final String KEY_EXPECTED_VERSION  = "rp_exp_version";
        private static final String KEY_CURRENT_NUM_ITEMS = "rp_cur_num_items";
        private static final String KEY_TOTAL_NUM_ITEMS   = "rp_tot_num_items";

        final NcsDownloadRequest request;
        final int currentVersion;
        final int expectedVersion;
        final int currentNumItems;
        final int totalNumItems;

        ResumePoint(NcsDownloadRequest request, int currentVersion, int expectedVersion, int currentNumItems, int totalNumItems)
        {
            this.request = request;
            this.currentVersion = currentVersion;
            this.expectedVersion = expectedVersion;
            this.currentNumItems = currentNumItems;
            this.totalNumItems = totalNumItems;
        }

        @Override
        public String toString()
        {
            return "[request:" + request + ", currentVersion:" + currentVersion + ", expectedVersion:" + expectedVersion + ", currentNumItems:" + currentNumItems + ", totalNumItems:" + totalNumItems + "]";
        }

        @Override
        public org.json.JSONObject toJSON()
        {
            JSONObject o = new JSONObject();
            o.tryPut(KEY_REQUEST, request);
            o.tryPut(KEY_CURRENT_VERSION, currentVersion);
            o.tryPut(KEY_EXPECTED_VERSION, expectedVersion);
            o.tryPut(KEY_CURRENT_NUM_ITEMS, currentNumItems);
            o.tryPut(KEY_TOTAL_NUM_ITEMS, totalNumItems);
            return o;
        }

        public static ResumePoint createFromJSON(org.json.JSONObject json) throws JSONException
        {
            NcsDownloadRequest request = NcsDownloadRequest.createFromJSON(json.getJSONObject(KEY_REQUEST));
            int currentVersion = json.getInt(KEY_CURRENT_VERSION);
            int expectedVersion = json.getInt(KEY_EXPECTED_VERSION);
            int currentNumItems = json.getInt(KEY_CURRENT_NUM_ITEMS);
            int totalNumItems = json.getInt(KEY_TOTAL_NUM_ITEMS);

            return new ResumePoint(request, currentVersion, expectedVersion, currentNumItems, totalNumItems);
        }
    }

    private static class Download
    {
        private Transaction transaction;
    }

    private static final String QUERY_PARAM_NAME = "REQUEST_INFO";

    private final CloudServices _cloud;
    private final NcsCloudDataDownloaderConfig _config;
    private final AsyncTaskHandler _taskHandler;
    private final NMTHandler _mainThreadHandler;
    private Download _download;

    /**
     * @param cloudServices The cloud services instance to use
     */
    public NcsCloudDataDownloader(CloudServices cloudServices)
    {
        this(cloudServices, new NcsCloudDataDownloaderConfig());
    }

    /**
     * @param cloudServices The cloud services instance to use
     * @param config The custom configuration, see {@link NcsCloudDataDownloaderConfig}
     */
    public NcsCloudDataDownloader(CloudServices cloudServices, NcsCloudDataDownloaderConfig config)
    {
        Checker.checkArgForNull("cloudServices", cloudServices);
        Checker.checkArgForNull("config", config);
        _cloud = cloudServices;
        _config = config;
        _taskHandler = new AsyncTaskHandler();
        _mainThreadHandler = _cloud.getMainThreadHandler();
    }

    public void download(NcsDownloadRequest request, DownloadListener downloadListener)
    {
        Checker.checkArgForNull("request", request);
        Checker.checkArgForNull("downloadListener", downloadListener);

        downloadInternal(new ResumePoint(request, -1, -1,  0, Integer.MAX_VALUE), Transaction.DefaultPriorities.DATA_DOWNLOAD, downloadListener);
    }

    /**
     * Resume the retrieval of the incomplete data returned by @see com.nuance.dragon.toolkit.cloudservices.datadownloader.CloudDataDownloader.DownloadListener#onSuccess(CloudDataDownloader, RetrievedData, int)
     * @param resumePoint Where to resume the download
     * @param listener The download listener
     */
    public void download(ResumePoint resumePoint,
                         DownloadListener listener)
    {
        Checker.checkArgForNull("previousResult", resumePoint);
        Checker.checkArgForNull("listener", listener);

        downloadInternal(resumePoint, Transaction.DefaultPriorities.DATA_DOWNLOAD, listener);
    }

    protected void downloadInternal(final ResumePoint rp,
                                    final int transactionPriority,
                                    final DownloadListener listener)
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                Logger.debug(this, "downloadInternal(" + rp + ", " + transactionPriority + ")");
        
                // Cancel any transaction in progress
                cancel();
        
                final Download download = new Download();
                _download = download;
        
                final Transaction t = new Transaction(_config.commandName, _config.commandSettings, new Transaction.Listener()
                {
                    @Override
                    public void onTransactionStarted(final Transaction t)
                    {
                        if (download != _download)
                            return;
        
                        final AsyncTaskHandler.TaskRunnable<DictionaryParam> queryParamTask
                              = new AsyncTaskHandler.TaskRunnable<DictionaryParam>()
                                {
                                    @Override
                                    public DictionaryParam run()
                                    {
                                        if (_download == download) {
                                            return new DictionaryParam(QUERY_PARAM_NAME, constructRequestInfo(rp.request.id, rp.request.type, rp.request.chunkSize, rp.currentNumItems, rp.currentVersion, rp.expectedVersion, rp.request.targetInfo.userId, rp.request.targetInfo.deviceId, rp.request.targetInfo.applicationId));
                                        }
                                        return null;
                                    }
        
                                    @Override
                                    public void onPostRun(DictionaryParam param)
                                    {
                                        if (_download == download) {
                                            t.addParam(param);
                                            t.finish();
                                        }
                                    }
                                };
        
                        _taskHandler.post(queryParamTask);
                    }
        
                    @Override
                    public void onTransactionResult(Transaction t, final TransactionResult result, boolean complete)
                    {
                        Logger.debug(this, "onTransactionResult(" + result + ", " + complete + ")");
        
                        if (download != _download)
                            return;
        
                        AsyncTaskHandler.TaskRunnable<ResumePoint> resultTask = new AsyncTaskHandler.TaskRunnable<ResumePoint>()
                        {
                            @Override
                            public ResumePoint run()
                            {
                                ResumePoint ret = null;
                                if (_download == download)
                                {
                                    try {
                                        Dictionary queryResult = result.getContents();
                                        int latestVersion = queryResult.getInt("latest_version").value;
                                        Sequence actions = queryResult.getSequence("actions");
                                        Sequence data = queryResult.getSequence("data");
        
                                        int currentVersion = rp.currentVersion;
                                        int expectedVersion = latestVersion;
        
                                        for (int i = 0; i < actions.size(); i++)
                                        {
                                            String action = actions.getString(i).value;
                                            if (action.equals("complete"))
                                            {
                                                Logger.debug(this, "onTransactionResult() \"complete\" recevied");
                                                currentVersion = latestVersion;
                                            }
                                        }
        
                                        int newCurrentNumItems = rp.currentNumItems + data.size();
        
                                        ret = new ResumePoint(rp.request, currentVersion, expectedVersion, newCurrentNumItems, /*TODO*/Integer.MAX_VALUE);
                                    }
                                    catch (Exception e) {
                                        Logger.error(this, "Error parsing the transaction result when downloading data", e);
                                    }
                                }
                                return ret;
                            }
        
                            @Override
                            public void onPostRun(ResumePoint rp)
                            {
                                Logger.debug(this, "onTransactionResult() rp:" + rp);
        
                                if (_download == download)
                                {
                                    _download = null;
        
                                    if (result == null)
                                    {
                                        listener.onError(NcsCloudDataDownloader.this, null, rp);
                                    }
                                    else
                                    {
                                        listener.onSuccess(NcsCloudDataDownloader.this,
                                                           result,
                                                           rp);
                                    }
                                }
                            }
                        };
        
                        _taskHandler.post(resultTask);
                    }
        
                    @Override
                    public void onTransactionError(Transaction t, TransactionError error)
                    {
                        Logger.error(this, "onTransactionError(): error code: " + (error != null ? error.getErrorCode() : "unknown"));
        
                        if (download == _download)
                            _download = null;
        
                        listener.onError(NcsCloudDataDownloader.this, error, rp);
                    }
        
                    @Override
                    public void onTransactionIdGenerated(String transactionId) {
                    }
        
                    @Override
                    public void onTransactionProcessingStarted(Transaction t)
                    {
                    }
        
                }, _config.commandTimeoutMs, true);
        
                _download.transaction = t;
        
                _cloud.addTransaction(t, transactionPriority);
            }
        });
    }

    /**
     * Cancel the current download, if any
     */
    public void cancel()
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                if (_download != null)
                {
                    if (_download.transaction != null)
                        _download.transaction.cancel();
                    _download = null;
                }
            }
        });
    }

    private static Dictionary constructRequestInfo(String id, String type, int maxNumItems, int currentNumItems, int currentVersion, int expectedVersion, String userId, String deviceId, String applicationId)
    {
        Data.Dictionary requestInfo = new Data.Dictionary();

        requestInfo.put("id", id);
        requestInfo.put("type", type);
        requestInfo.put("max_num_items", maxNumItems);
        requestInfo.put("current_num_items", currentNumItems);
        requestInfo.put("current_version", currentVersion);
        requestInfo.put("expected_version", expectedVersion);

        Data.Dictionary targetInfo = new Data.Dictionary();
        if (userId != null)
            targetInfo.put("user_id", userId);
        if (deviceId != null)
            targetInfo.put("device_id", deviceId);
        if (applicationId != null)
            targetInfo.put("application_id", applicationId);

        requestInfo.put("target_info", targetInfo);

        return requestInfo;
    }
}
