package com.nuance.dragon.toolkit.cloudservices.datadownloader;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

public class NcsDownloadRequest implements JSONCompliant
{
    /*
     * TargetInfo
     */
    public static class TargetInfo implements JSONCompliant
    {
        protected static final String KEY_USER_ID = "ti_uid";
        protected static final String KEY_DEVICE_ID = "ti_did";
        protected static final String KEY_APPLICATION_ID = "ti_aid";
        
        public final String userId;
        public final String deviceId;
        public final String applicationId;
        
        private TargetInfo(String uid, String did, String aid)
        {
            userId = uid;
            deviceId = did;
            applicationId = aid;
        }
        
        /*
         * TargetInfo Builder
         */
        public static class Builder
        {
            private String _userId        = "PUBLIC";
            private String _deviceId      = "PUBLIC";
            private String _applicationId = null;
            
            public Builder()
            {
            }
            
            public void setUserId(String userId)
            {
                _userId = userId;
            }
            
            public void setDeviceId(String deviceId)
            {
                _deviceId = deviceId;
            }
            
            public void setApplicationId(String applicationId)
            {
                _applicationId = applicationId;
            }
            
            public TargetInfo build()
            {
                return new TargetInfo(_userId, _deviceId, _applicationId);
            }
        }
        
        @Override
        public String toString()
        {
            return "[userId:" + userId + ", deviceId:" + deviceId + ", applicationId:" + applicationId + "]"; 
        }

        @Override
        public boolean equals(Object obj)
        {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            TargetInfo other = (TargetInfo)obj;
            
            if (userId == null || other.userId == null)
                return userId == other.userId;
            if (!userId.equals(other.userId))
                return false;
            
            if (deviceId == null || other.deviceId == null)
                return deviceId == other.deviceId;
            if (!deviceId.equals(other.deviceId))
                return false;
            
            if (applicationId == null || other.applicationId == null)
                return applicationId == other.applicationId;
            if (!applicationId.equals(other.applicationId))
                return false;
            
            return true;
        }
        
        @Override
        public int hashCode()
        {
            final int prime = 31;
            int result = 1;
            result = prime * result + ((userId == null) ? 0 : userId.hashCode());
            result = prime * result + ((deviceId == null) ? 0 : deviceId.hashCode());
            result = prime * result + ((applicationId == null) ? 0 : applicationId.hashCode());
            return result;
        }
        
        @Override
        public org.json.JSONObject toJSON()
        {
            JSONObject o = new JSONObject();
            o.tryPut(KEY_USER_ID, userId);
            o.tryPut(KEY_DEVICE_ID, deviceId);
            o.tryPut(KEY_APPLICATION_ID, applicationId);
            return o;
        }

        public static TargetInfo createFromJSON(org.json.JSONObject json) throws JSONException
        {
            String uid = json.has(KEY_USER_ID) ? json.getString(KEY_USER_ID) : null;
            String did = json.has(KEY_DEVICE_ID) ? json.getString(KEY_DEVICE_ID) : null;
            String aid = json.has(KEY_APPLICATION_ID) ? json.getString(KEY_APPLICATION_ID) : null;
            return new TargetInfo(uid, did, aid);
        }
    }

    private static final String KEY_ID          = "dr_id";
    private static final String KEY_TYPE        = "dr_type";
    private static final String KEY_TARGET_INFO = "dr_target_info";
    private static final String KEY_CHUNK_SIZE  = "dr_chunk_size";

    public final String id;
    public final String type;
    public final int chunkSize;
    public final TargetInfo targetInfo;

    private NcsDownloadRequest(String id, String type, int chunkSize, TargetInfo targetInfo)
    {
        this.id = id;
        this.type = type;
        this.chunkSize = chunkSize;
        this.targetInfo = targetInfo;
    }
    
    public static class Builder
    {
        private final String _id;
        private final String _type;
        private int _chunkSize = 1000;
        private TargetInfo _targetInfo = (new TargetInfo.Builder()).build();

        public Builder(String id, String type)
        {
            Checker.checkStringArgForNullOrEmpty("id", id);
            Checker.checkStringArgForNullOrEmpty("type", type);
            _id = id;
            _type = type;
        }
        
        public Builder setChunkSize(int chunkSize)
        {
            Checker.checkArgForCondition("commandSize", "stricly greater than 0", chunkSize > 0);
            _chunkSize = chunkSize;
            return this;
        }

        public Builder setTargetInfo(TargetInfo targetInfo)
        {
            Checker.checkArgForNull("targetInfo", targetInfo);
            _targetInfo = targetInfo;
            return this;
        }

        public NcsDownloadRequest build()
        {
            return new NcsDownloadRequest(_id, _type, _chunkSize, _targetInfo);
        }
    }
    
    @Override
    public String toString()
    {
        return "[id:" + id + ", type:" + type + ", chunkSize:" + chunkSize + ", targetInfo" + targetInfo + "]";
    }
    
    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        NcsDownloadRequest other = (NcsDownloadRequest)obj;
        if (!id.equals(other.id))
            return false;
        if (!type.equals(other.type))
            return false;
        if (chunkSize != other.chunkSize)
            return false;
        if (!targetInfo.equals(other.targetInfo))
            return false;
        return true;
    }
    
    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + id.hashCode();
        result = prime * result + type.hashCode();
        result = prime * result + chunkSize;
        result = prime * result + targetInfo.hashCode();
        return result;
    }

    @Override
    public org.json.JSONObject toJSON() {
        JSONObject o = new JSONObject();
        o.tryPut(KEY_ID, id);
        o.tryPut(KEY_TYPE, type);
        o.tryPut(KEY_CHUNK_SIZE, chunkSize);
        o.tryPut(KEY_TARGET_INFO, targetInfo);
        return o;
    }
    
    public static NcsDownloadRequest createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String id = json.getString(KEY_ID);
        String type = json.getString(KEY_TYPE);
        int chunkSize = json.getInt(KEY_CHUNK_SIZE);
        TargetInfo targetInfo = TargetInfo.createFromJSON(json.getJSONObject(KEY_TARGET_INFO));
        return new NcsDownloadRequest(id, type, chunkSize, targetInfo);
    }
}