package com.nuance.dragon.toolkit.cloudservices.datadownloader;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Hashtable;

import org.json.JSONArray;
import org.json.JSONException;

import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.TransactionError;
import com.nuance.dragon.toolkit.cloudservices.TransactionResult;
import com.nuance.dragon.toolkit.cloudservices.TransactionError.Types;
import com.nuance.dragon.toolkit.cloudservices.datadownloader.NcsCloudDataDownloader.DownloadListener;
import com.nuance.dragon.toolkit.cloudservices.datadownloader.NcsCloudDataDownloader.ResumePoint;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.core.data.Data.Dictionary;
import com.nuance.dragon.toolkit.core.data.Data.Sequence;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTContext;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.impl.FileManagerOem;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

public class NcsPersistentDownloader
{
    public static enum Status
    {
        COMPLETED, // Done downloading or up to date
        DOWNLOAD_IN_PROGRESS, // Self-explanatory
        //NEW_VERSION, // Hint the client to wipe his copy of the list before writing the new content to storage
        SOFT_ERROR, // Network error, basically something that is not fatal e.g. network connection and the client can eventually requeue the download
        HARD_ERROR, // Something bad happened (config problem), the client shall not requeue the download
    };

    public static class Downloadable
    {
        public final String id; // For convenience so client knows in which bucket to store the entities without doing more processing
        public final int chunkId; // Chunk UUID
        public final boolean newVersion; 
        public final Data.Sequence chunk; // The current RAM subset of the list to be stored by the client
        final ResumePoint resumePoint;
        
        private Downloadable(String id, int chunkId, boolean newVersion, Data.Sequence chunk, ResumePoint resumePoint)
        {
            this.id = id;
            this.chunkId = chunkId;
            this.newVersion = newVersion;
            this.chunk = chunk;
            this.resumePoint = resumePoint;
        }
    }

    public interface PersistentDownloadListener
    {
        void onDone(NcsPersistentDownloader downloader, Downloadable downloadable, float progress, Status status);
    }

    private final static String FILE_NAME = "rps.save";
    
    @SuppressWarnings("serial")
    private static class ResumePoints extends Hashtable<NcsDownloadRequest, ResumePoint> implements JSONCompliant
    {
        private static final String KEY_ARRAY = "rps_array";
        
        @Override
        public org.json.JSONObject toJSON()
        {
            JSONArray a = new JSONArray();
            for (ResumePoint rp : values())
            {
                a.put(rp.toJSON());
            }
            
            JSONObject o = new JSONObject();
            o.tryPut(KEY_ARRAY, a);
            return o;
        }
        
        public static ResumePoints createFromJSON(JSONObject o) throws JSONException
        {
            Checker.checkJsonForNull("o", o);
            ResumePoints ret = new ResumePoints();
            JSONArray a = o.getJSONArray(KEY_ARRAY);
            for (int i = 0; i < a.length(); i++)
            {
                ResumePoint rp = ResumePoint.createFromJSON(a.getJSONObject(i));
                ret.put(rp.request, rp);
            }
            return ret;
        }
    }
    
    private final NcsCloudDataDownloader _downloader;
    private ResumePoints _resumePoints = null;
    private FileManager _fileMgr;
    
    public NcsPersistentDownloader(CloudServices cloudServices)
    {
        _downloader = new NcsCloudDataDownloader(cloudServices);
        loadResumePoints(cloudServices.getContext().getAppContext());
    }

    public NcsPersistentDownloader(CloudServices cloudServices, NcsCloudDataDownloaderConfig config)
    {
        _downloader = new NcsCloudDataDownloader(cloudServices, config);
        loadResumePoints(cloudServices.getContext().getAppContext());
    }

    private void loadResumePoints(NMTContext context)
    {
        _fileMgr = new FileManagerOem(context, "persistent_downloader");
        if (_fileMgr.exists(FILE_NAME))
        {
            int fileSize = _fileMgr.getSize(FILE_NAME);
            FileInputStream fis = _fileMgr.openFileForReading(FILE_NAME);
            byte[] json = new byte[fileSize]; // TODO assuming file size is small
            try {
                fis.read(json); // TODO assuming whole file is read off in 1 shot
                _resumePoints = ResumePoints.createFromJSON(new JSONObject(new String(json, "UTF-8")));
            } catch (Exception e) {
                Logger.warn(this, "Exception thrown while accessing saved resume points:" + e);
            }
            
            try {
                fis.close();
            } catch (IOException e) {}
        }
        
        if (_resumePoints == null)
            _resumePoints = new ResumePoints();
        
        Logger.debug(this, "loadResumePoints() _resumePoints:" + _resumePoints);
    }
    
    private void saveResumePoints()
    {
        FileOutputStream fos = _fileMgr.openFileForWriting(FILE_NAME);
        try {
            fos.write(_resumePoints.toJSON().toString().getBytes()); // TODO assuming whole file is written off in 1 shot
            fos.close();
        } catch (IOException e) {
            try {
                fos.close();
            } catch (IOException ee) {}
            _fileMgr.delete(FILE_NAME);
            if (_fileMgr.exists(FILE_NAME))
                Logger.warn(this, "saveResumePoints(): cleaning up unsuccessful saved resume points failed!!! (" + FILE_NAME + ")");
        }
    }
    
    public void download(final NcsDownloadRequest request, final PersistentDownloadListener downloadListener)
    {
        Checker.checkArgForNull("request", request);
        Checker.checkArgForNull("downloadListener", downloadListener);
    
        DownloadListener listener = new DownloadListener()
        {
            @Override
            public void onSuccess(NcsCloudDataDownloader downloader,
                    TransactionResult result, ResumePoint rp) {
                Downloadable downloadable = null;
                float progress = -1;
                Status st = Status.HARD_ERROR;
                if (rp.expectedVersion >= 0)
                {
                    try {
                        Dictionary queryResult = result.getContents();
                        Sequence actions = queryResult.getSequence("actions");
                        Sequence data = queryResult.getSequence("data");
    
                        boolean newVersion = false;
                        for (int i = 0; i < actions.size(); i++)
                        {
                            String action = actions.getString(i).value;
                            if (action.equals("purge_all"))
                            {
                                newVersion = true;
                            }
                            else if (action.equals("request_more"))
                            {
                                st = Status.DOWNLOAD_IN_PROGRESS;
                            }
                            else if (action.equals("complete"))
                            {
                                st = Status.COMPLETED;
                            }
                        }
                        
                        progress = rp.currentNumItems / rp.totalNumItems;
                        downloadable = new Downloadable(request.id, java.util.Arrays.hashCode(new Object[] {rp.currentVersion, rp.currentNumItems, data}), newVersion, data, rp);
                    }
                    catch (Exception e) {
                        Logger.error(this, "Error parsing the transaction result when downloading data", e);
                        st = Status.HARD_ERROR;
                    }
                }
                
                downloadListener.onDone(NcsPersistentDownloader.this, downloadable, progress, st);
            }

            @Override
            public void onError(NcsCloudDataDownloader downloader,
                    TransactionError error, ResumePoint rp) {
                Status st;
                switch (error.getType())
                {
                    case Types.CONNECTION_ERROR:
                    case Types.RETRY_ERROR:
                        st = Status.SOFT_ERROR;
                        break;
                    case Types.CANCELED:
                    case Types.QUERY_ERROR:
                    case Types.OTHER_ERROR:
                    default:
                        st = Status.HARD_ERROR;
                        break;
                }
                downloadListener.onDone(NcsPersistentDownloader.this, null, (float)rp.currentNumItems / rp.totalNumItems, st);
            }
        };
        
        ResumePoint rp = _resumePoints.get(request);
        if (rp == null)
            _downloader.download(request, listener);
        else
            _downloader.download(rp, listener);
    }
    
    public void commit(Downloadable downloadable)
    {
        _resumePoints.put(downloadable.resumePoint.request, downloadable.resumePoint);
        saveResumePoints();
    }
    
    public void cancel()
    {
        _downloader.cancel();
    }
    
    public void clear(NcsDownloadRequest request)
    {
        _resumePoints.remove(request);
        saveResumePoints();
    }

    public void clearAll()
    {
        _resumePoints.clear();
        saveResumePoints();
    }
}
