package com.nuance.dragon.toolkit.cloudservices.datauploader;

import java.util.HashMap;
import java.util.Map;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.DictionaryParam;
import com.nuance.dragon.toolkit.cloudservices.SeqChunkParam;
import com.nuance.dragon.toolkit.cloudservices.SeqEndParam;
import com.nuance.dragon.toolkit.cloudservices.SeqStartParam;
import com.nuance.dragon.toolkit.cloudservices.Transaction;
import com.nuance.dragon.toolkit.cloudservices.TransactionError;
import com.nuance.dragon.toolkit.cloudservices.TransactionResult;
import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.core.WordList;
import com.nuance.dragon.toolkit.core.WordList.WordIterator;
import com.nuance.dragon.toolkit.core.data.Data.Dictionary;
import com.nuance.dragon.toolkit.core.data.Data.Sequence;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.AsyncTaskHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.util.internal.Pair;

/**
 * A Cloud Data Uploader uploads the contents of a {@link WordList} through a
 * {@link CloudServices}, for use in a later recognition.
 */
public class CloudDataUploader
{
    /**
     * Listens for the result of the upload transaction
     */
    public interface UploadListener
    {
        /**
         * The upload succeeded
         * @param uploader The current uploader.
         * @param result May be null if no upload occurred due to 0 effective
         * changes.
         * @param grammarId The grammar ID of the word list that was uploaded.
         * @param nothingToUpload True if there is nothing to upload.
         */
        void onSuccess(CloudDataUploader uploader,
                       TransactionResult result,
                       String grammarId,
                       boolean nothingToUpload);

        /**
         * There was an error when uploading
         * @param uploader The current uploader.
         * @param error The transaction error object.
         * @param grammarId The grammar ID of the word list that was uploaded.
         */
        void onError(CloudDataUploader uploader, TransactionError error,  String grammarId);
    }

    /**
     * Listens for the result of the delete transaction
     */
    public interface DeleteListener
    {
        /**
         * The deletion succeeded
         * @param uploader
         * @param result May be null if no upload occurred due to 0 effective
         * changes.
         */
        void onSuccess(CloudDataUploader uploader, TransactionResult result);

        /**
         * There was an error when deleting
         * @param uploader
         * @param error
         */
        void onError(CloudDataUploader uploader, TransactionError error);
    }

    private static class Upload
    {
        private Transaction transaction;
        private WordIterator iterator;
        private boolean isFullIterator;
        private boolean nothingToUpload;
        private boolean continueUpload;
        private int numDataBlocks;
        private int currentChecksum;
    }

    private static final String DATA_PARAM_NAME = "DATA_BLOCK";
    private static final String DONE_PARAM_NAME = "UPLOAD_DONE";

    private final CloudServices _cloud;
    private final CloudDataUploaderConfig _config;
    private final AsyncTaskHandler _taskHandler;
    private Upload _upload;

    /**
     * @param cloudServices The cloud services instance to use
     */
    public CloudDataUploader(CloudServices cloudServices)
    {
        this(cloudServices, new CloudDataUploaderConfig());
    }

    /**
     * @param cloudServices The cloud services instance to use
     * @param config The custom configuration, see {@link CloudDataUploaderConfig}
     */
    public CloudDataUploader(CloudServices cloudServices, CloudDataUploaderConfig config)
    {
        Checker.checkArgForNull("cloudServices", cloudServices);
        Checker.checkArgForNull("config", config);
        _cloud = cloudServices;
        _config = config;
        _taskHandler = new AsyncTaskHandler();
    }

    /**
     * Upload a word list to the server. This method will decide whether to
     * upload the full list, or to upload only recent modifications.
     * @param grammarId The grammar ID to attach to the word list
     * @param type The type of the content, e.g. "contacts" or "custom_words" or "structured_content"
     * @param category The category of the content. Null if type is not "structured_content"
     * @param list The list to upload
     * @param listener
     */
    public void uploadData(final String grammarId,
                           final String type,
                           final String category,
                           final WordList list,
                           final UploadListener listener)
    {
        uploadData(grammarId, type, category, list, false, listener);
    }

    /**
     * Upload a word list to the server. This method will decide whether to
     * upload the full list, or to upload only recent modifications.
     * @param grammarId The grammar ID to attach to the word list
     * @param type The type of the content, e.g. "contacts" or "custom_words" or "structured_content"
     * @param category The category of the content. Null if type is not "structured_content"
     * @param list The list to upload
     * @param forceFullUpload true to force the upload to clear and upload everything
     * @param listener
     */
    public void uploadData(final String grammarId,
                           final String type,
                           final String category,
                           final WordList list,
                           final boolean forceFullUpload,
                           final UploadListener listener)
    {
        Checker.checkStringArgForNullOrEmpty("grammarId", grammarId);
        Checker.checkStringArgForNullOrEmpty("type", type);
        Checker.checkArgForNull("list", list);
        Checker.checkArgForNull("listener", listener);

        uploadData(null, grammarId, type, category, list, forceFullUpload, Transaction.DefaultPriorities.FIRST_DATA_UPLOAD, listener);
    }

    private void uploadData(final WordIterator iterator,
                            final String grammarId,
                            final String type,
                            final String category,
                            final WordList list,
                            final boolean forceFullUpload,
                            final int transactionPriority,
                            final UploadListener listener)
    {
        // Cancel any transaction in progress
        cancel();

        final Upload upload = new Upload();
        _upload = upload;

        final Transaction t = new Transaction(_config.commandName, _config.commandSettings, new Transaction.Listener()
        {
            @Override
            public void onTransactionStarted(final Transaction t)
            {
                if (upload != _upload)
                    return;

                // The transaction shouldn't start until it has been enabled
                final WordIterator it = upload.iterator;
                final boolean clear = upload.isFullIterator;

                Assert.assertNotNull(it);

                final AsyncTaskHandler.TaskRunnable<DictionaryParam> uploadDoneTask
                      = new AsyncTaskHandler.TaskRunnable<DictionaryParam>()
                        {
                            @Override
                            public DictionaryParam run()
                            {
                                if (_upload == upload) {
                                    return new DictionaryParam(DONE_PARAM_NAME, constructUploadDone(it, grammarId, type, category, list, upload.numDataBlocks, upload.currentChecksum));
                                }
                                return null;
                            }

                            @Override
                            public void onPostRun(DictionaryParam param)
                            {
                                if (_upload == upload) {
                                    t.addParam(param);
                                    t.finish();
                                }
                            }
                        };

                AsyncTaskHandler.TaskRunnable<DictionaryParam> dataBlockTask
                      = new AsyncTaskHandler.TaskRunnable<DictionaryParam>()
                        {
                            @Override
                            public DictionaryParam run()
                            {
                                DictionaryParam param = null;

                                if (_upload == upload)
                                {
                                    Logger.debug(this, "Entering dataBlockTask, grammarId = " + grammarId + ", clear = " + (clear && !upload.continueUpload));
                                    final Dictionary dict = constructDataBlock(grammarId, type, category, constructActionsSequence(it, clear && !upload.continueUpload, type, _config.commandChunkSize));
                                    boolean hasMore = false;
                                    if (upload.iterator != null)
                                    {
                                        hasMore = upload.iterator.hasNext();
                                    }
                                    boolean keepCurrentCommand = _config.commandChunkSize * (upload.numDataBlocks + 1) < _config.commandTotalSize;

                                    if (upload.continueUpload)
                                    {
                                        if (hasMore && keepCurrentCommand)
                                        {
                                            param = new SeqChunkParam(DATA_PARAM_NAME, dict);
                                        }
                                        else
                                        {
                                            param = new SeqEndParam(DATA_PARAM_NAME, dict);
                                            upload.continueUpload = false;
                                        }
                                    }
                                    else
                                    {
                                        if (hasMore && keepCurrentCommand)
                                        {
                                            param = new SeqStartParam(DATA_PARAM_NAME, dict);
                                            upload.continueUpload = true;
                                        }
                                        else
                                        {
                                            param = new DictionaryParam(DATA_PARAM_NAME, dict);
                                        }
                                    }
                                       Logger.debug(this, "Exiting dataBlockTask, grammarId = " + grammarId);
                                }

                                return param;
                            }

                            @Override
                            public void onPostRun(DictionaryParam param)
                            {
                                if (_upload == upload) {
                                    upload.numDataBlocks++;
                                    t.addParam(param);

                                    if (upload.continueUpload)
                                    {
                                        _taskHandler.post(this);
                                    }
                                    else
                                    {
                                        _taskHandler.post(uploadDoneTask);
                                    }
                                }
                            }
                        };

                _taskHandler.post(dataBlockTask);
            }

            @Override
            public void onTransactionProcessingStarted(Transaction t)
            {
            }
            
            @Override
            public void onTransactionResult(Transaction t, final TransactionResult result, boolean complete)
            {
                if (upload != _upload)
                    return;

                boolean tempSuccess = false;
                try {
                    Dictionary dict = result.getContents().getSequence("result_list").getDictionary(0);
                    tempSuccess = dict.getString("status").value.equals("success");
                    if (!tempSuccess)
                        Logger.error(this, "onTransactionResult(): key \"status \" != \"success \"");
                }
                catch (Exception e) {
                    Logger.error(this, "Error parsing the transaction result when uploading grammar = " + grammarId, e);
                }

                final boolean success = tempSuccess;
                AsyncTaskHandler.TaskRunnable<Boolean> resultTask
                = new AsyncTaskHandler.TaskRunnable<Boolean>()
                  {
                      @Override
                      public Boolean run()
                      {
                          if (_upload == upload)
                          {
                              if (success) {
                                  boolean hasMore = false;
                                  if (upload.iterator != null) {
                                      if (upload.iterator.hasNext()) {
                                          hasMore = true;
                                      }
                                      else {
                                          upload.iterator.acceptChanges();
                                      }
                                  }
                                  return hasMore;
                              }
                              else
                              {
                                  if (upload.iterator != null)
                                      upload.iterator.discardChanges();
                              }
                          }
                          return false;
                      }

                      @Override
                      public void onPostRun(Boolean hasMore)
                      {
                          if (_upload == upload)
                          {
                              _upload = null;

                              if (success)
                              {
                                  if (hasMore) {
                                      Logger.debug(this, "More items, sending next command for grammarId = " + grammarId);
                                      uploadData(upload.iterator, grammarId, type, category, list, false, Transaction.DefaultPriorities.DATA_UPLOAD, listener);
                                  }
                                  else {
                                      listener.onSuccess(CloudDataUploader.this, result, grammarId, false);
                                  }
                              }
                              else
                              {
                                  Logger.error(this, "Server error while uploading grammar = " + grammarId);
                                  listener.onError(CloudDataUploader.this, null, grammarId);
                              }
                          }
                      }
                  };

                _taskHandler.post(resultTask);
            }

            @Override
            public void onTransactionError(Transaction t, TransactionError error)
            {
                /* If there was nothing to upload, we intentionally canceled the transaction.
                 * In that case, notify the user that the upload was successful (nothing to upload). */
                if (upload.nothingToUpload) {
                    _upload = null;
                    listener.onSuccess(CloudDataUploader.this, null, grammarId, true);
                } else {
                    Logger.error(this, "onTransactionError(): error code: " + (error != null ? error.getErrorCode() : "unknown"));

                    if (upload.iterator != null)
                        upload.iterator.discardChanges();

                    if (upload == _upload)
                        _upload = null;

                    listener.onError(CloudDataUploader.this, error, grammarId);
                }
            }

            @Override
            public void onTransactionIdGenerated(String transactionId)
            {
                // TODO Auto-generated method stub
            }
        }, _config.commandTimeoutMs, false); // Start disabled--don't start sending the transaction
                                             // until we're sure we need to actually run it.

        _upload.transaction = t;

        // Add the transaction right away, to avoid another transaction added
        // first in the queue. This is to support multiple CloudDataUploaders
        // being started in succession, where a subsequent recognition is queued
        // after them.
        _cloud.addTransaction(t, transactionPriority);

        // Run a task to initialize the WordIterator, and either enable the
        // transaction or cancel it (if the WordIterator is empty)
        _taskHandler.post(new AsyncTaskHandler.TaskRunnable<Pair<WordIterator, Boolean>>()
                          {
                            @Override
                            public Pair<WordIterator, Boolean> run()
                            {
                                boolean clear = (iterator == null) && (forceFullUpload || list.fullUpdateRequired());
                                WordIterator it = iterator;
                                if (it != null)  {
                                    upload.currentChecksum = it.getChecksum();
                                }
                                else {
                                    it = clear ? list.getFullIterator() : list.getModifiedIterator();
                                    upload.currentChecksum = list.getAcceptedChecksum();
                                }

                                if (!clear && !it.hasNext()) {
                                    it.discardChanges();
                                    return null; // No transaction is necessary
                                }

                                return new Pair<WordIterator, Boolean>(it, clear);
                            }

                            @Override
                            public void onPostRun(final Pair<WordIterator, Boolean> p)
                            {
                                if (upload != _upload)
                                    return;

                                if (p == null)
                                {
                                    // Upload isn't necessary
                                    _upload.nothingToUpload = true;
                                    // Cancel the transaction.  Wait for the transactions's onError
                                    // callback, and notify the caller from there.
                                    t.cancel();
                                }
                                else
                                {
                                    _upload.iterator = p.first;
                                    _upload.isFullIterator = p.second;

                                    _upload.transaction.enable();
                                }
                            }
                          });
    }

    /**
     * Delete everything from the server
     * @param listener
     */
    public void deleteAll(final DeleteListener listener)
    {
        // Cancel any transaction in progress
        cancel();

        final Upload upload = new Upload();
        _upload = upload;
        Transaction t = new Transaction(_config.commandName, _config.commandSettings, new Transaction.Listener()
        {
            @Override
            public void onTransactionStarted(Transaction t)
            {
            }

            @Override
            public void onTransactionProcessingStarted(Transaction t)
            {
            }

            @Override
            public void onTransactionResult(Transaction t, TransactionResult result, boolean complete)
            {
                if (upload != _upload)
                    return;

                _upload = null;

                listener.onSuccess(CloudDataUploader.this, result);
            }

            @Override
            public void onTransactionError(Transaction t, TransactionError error)
            {
                if (upload != _upload)
                    return;

                _upload = null;

                listener.onError(CloudDataUploader.this, error);
            }

            @Override
            public void onTransactionIdGenerated(String transactionId)
            {
                // TODO Auto-generated method stub
            }
        }, _config.commandTimeoutMs);
        upload.transaction = t;

        t.addParam(new DictionaryParam(DATA_PARAM_NAME, constructDeleteAllDataBlock()));
        t.addParam( new DictionaryParam(DONE_PARAM_NAME, constructDeleteAllUploadDone()));
        t.finish();

        _cloud.addTransaction(t, 0);
    }

    /**
     * Cancel the current upload or delete transaction, if any
     */
    public void cancel()
    {
        if (_upload != null)
        {
            if (_upload.transaction != null)
                _upload.transaction.cancel();
            _upload = null;
        }
    }

    /** May take a while, so call this from a worker/async thread */
    private static Sequence constructActionsSequence(WordIterator it, boolean clear, String type, int maxItems)
    {
        // actions sequence
        Sequence actions = new Sequence();

        if (clear)
        {
            Dictionary actionDict = new Dictionary();
            actionDict.put("action", "clear_all");
            actions.add(actionDict);
        }

        if (type.equals("structured_content"))
        {
            Map<Integer, WordAction> wordActions = new HashMap<Integer, WordAction>();
            while (it.hasNext() && maxItems-- > 0)
            {
                WordAction wordAction = it.next();
                Word word = wordAction.getWord();
                int contentId = word.getUserId();
                if (contentId >= 0)
                {
                    if (wordActions.containsKey(contentId))
                    {
                        WordAction existingWordAction = wordActions.get(contentId);
                        if (existingWordAction.getAction() == WordAction.ADD && wordAction.getAction() == WordAction.REMOVE)
                        {
                            wordActions.put(contentId, new WordAction(existingWordAction.getWord(), WordAction.UPDATE));
                        }
                        else if (existingWordAction.getAction() == WordAction.REMOVE && wordAction.getAction() == WordAction.ADD)
                        {
                            wordActions.put(contentId, new WordAction(word, WordAction.UPDATE));
                        }
                        else
                        {
                            Logger.error(CloudDataUploader.class, "Incorrect state. This situation should not be possible");
                        }
                    }
                    else
                    {
                        wordActions.put(contentId, wordAction);
                    }
                }
            }

            Sequence addListSeq = new Sequence();
            Sequence removeListSeq = new Sequence();
            Sequence updateListSeq = new Sequence();
            for (Map.Entry<Integer, WordAction> entry : wordActions.entrySet())
            {
                Dictionary dict = new Dictionary();
                WordAction wordAction = entry.getValue();
                Word word = wordAction.getWord();
                dict.put("content_id", word.getUserId());
                Dictionary content = word.getCustomForm();
                dict.put("content", content);
                switch (wordAction.getAction())
                {
                case WordAction.ADD:
                    addListSeq.add(dict);
                    break;
                case WordAction.UPDATE:
                    updateListSeq.add(dict);
                    break;
                case WordAction.REMOVE:
                    removeListSeq.add(dict);
                    break;
                }
            }
            if (addListSeq.size() > 0)
            {
                actions.add(constructActionDictionary(addListSeq, "add"));
            }
            if (updateListSeq.size() > 0)
            {
                actions.add(constructActionDictionary(updateListSeq, "update"));
            }
            if (removeListSeq.size() > 0)
            {
                actions.add(constructActionDictionary(removeListSeq, "remove"));
            }
        }
        else
        {
            boolean isAdding = true;
            Sequence listSeq = new Sequence();
            while (it.hasNext() && maxItems-- > 0)
            {
                WordAction wordAction = it.next();

                if (isAdding != wordAction.isAdded() && listSeq.size() > 0)
                {
                    // content dictionary
                    actions.add(constructActionDictionary(listSeq, isAdding ? "add" : "remove"));
                    listSeq = new Sequence();
                }

                Word word = wordAction.getWord();
                isAdding = wordAction.isAdded();
                if (type.equals("contacts"))
                {
                    // contact
                    Dictionary contact = word.getCustomForm();
                    if (contact == null)
                    {
                        contact = new Dictionary();
                        contact.put("full_name", word.getSurfaceForm());
                    }
                    listSeq.add(contact);
                }
                else
                {
                    String text = word.getSurfaceForm();
                    listSeq.add(text);
                }
            }

            if (listSeq.size() > 0)
            {
                // content dictionary
                actions.add(constructActionDictionary(listSeq, isAdding ? "add" : "remove"));
            }
        }

        return actions;
    }

    /** Lumps 'add' / 'remove' actions together in a single list */
    private static Dictionary constructActionDictionary(Sequence list, String action)
    {
        Dictionary contentDict = new Dictionary();
        contentDict.put("list", list);

        Dictionary actionDict = new Dictionary();
        actionDict.put("content", contentDict);
        actionDict.put("action", action);

        return actionDict;
    }

    private static Dictionary constructDataBlock(String grammarId, String type, String category, Sequence actionsSeq)
    {
        Dictionary dataBlock = new Dictionary();
        //contacts.put("delete_all", 1);

        // data list sequence
        Sequence dataListSequence = new Sequence();

        // data list entry dictionary
        Dictionary entry = new Dictionary();
        entry.put("id", grammarId);
        entry.put("type", type);
        if (category != null)
            entry.put("structured_content_category", category);

        entry.put("actions", actionsSeq);

        dataListSequence.add(entry);

        dataBlock.put("data_list", dataListSequence);

        return dataBlock;
    }

    private static Dictionary constructUploadDone(WordIterator it, String grammarId, String type,  String category, WordList list, int numDataBlocks, int currentChecksum)
    {
        Dictionary upload_done = new Dictionary();
        upload_done.put("num_data_blocks", numDataBlocks);
        Sequence checksums = new Sequence();
        Dictionary dict2 = new Dictionary();
        dict2.put("id", grammarId);
        dict2.put("type", type);
        if (category != null)
            dict2.put("structured_content_category", category);
        dict2.put("current_checksum", "" + currentChecksum);
        dict2.put("new_checksum", "" + it.getChecksum());
        dict2.put("algorithm_id", "MD5");
        checksums.add(dict2);
        upload_done.put("checksums", checksums);

        return upload_done;
    }

    private static Dictionary constructDeleteAllDataBlock()
    {
        Dictionary dataBlock = new Dictionary();
        dataBlock.put("delete_all", 1);
        dataBlock.put("data_list", new Sequence());
        return dataBlock;
    }

    private static Dictionary constructDeleteAllUploadDone()
    {
        Dictionary upload_done = new Dictionary();
        upload_done.put("num_data_blocks", 1);
        upload_done.put("checksums", new Sequence());

        return upload_done;
    }
}
