package com.nuance.dragon.toolkit.cloudservices.datauploader;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.DictionaryParam;
import com.nuance.dragon.toolkit.cloudservices.Param;
import com.nuance.dragon.toolkit.cloudservices.SeqChunkParam;
import com.nuance.dragon.toolkit.cloudservices.SeqEndParam;
import com.nuance.dragon.toolkit.cloudservices.SeqStartParam;
import com.nuance.dragon.toolkit.cloudservices.Transaction;
import com.nuance.dragon.toolkit.cloudservices.TransactionError;
import com.nuance.dragon.toolkit.cloudservices.TransactionResult;
import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.core.WordList;
import com.nuance.dragon.toolkit.core.WordList.WordIterator;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.core.data.Data.Dictionary;
import com.nuance.dragon.toolkit.core.data.Data.Sequence;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.AsyncTaskHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.util.internal.Pair;

/**
 * A Cloud Data Uploader uploads the contents of a {@link WordList} through a
 * {@link CloudServices}, for use in a later recognition.
 */
public class NcsCloudDataUploader
{
    /**
     *
     * The various upload outcomes.
     */
    public static enum Status
    {
        /** Upload (adding/deleting content) has succeeded. */
        SUCCESS,

        /** There is no content to upload (empty list or the system is in sync). */
        SUCCESS_NOTHING_TO_UPLOAD,

        /** Unable to add or delete content on the server. */
        ERROR_STORAGE,

        /** Unable to generate grammar for content on the server. */
        ERROR_GRAMMAR,

        /** Server busy, try again later. */
        ERROR_SERVER_BUSY,

        /** Network error (timeout, connection dropped, etc.). */
        ERROR_NETWORK
    };

    /**
     * Listens for the result of the upload transaction
     */
    public interface UploadListener
    {
        /**
         * The upload operation has completed
         * @param uploader The current uploader.
         * @param status The status from {@link Status}.
         * @param grammarId The grammar ID of the word list that was uploaded.
         * @param retryTimeMs When to retry to upload when the server is busy. 0 in all other cases.
         */
        void onDone(NcsCloudDataUploader uploader,
                    Status status,
                    String grammarId,
                    long retryTimeMs);

        /**
         * The server emitted a result. For debugging purposes mainly.
         * @param uploader The current uploader.
         * @param grammarId The grammar ID of the word list that was uploaded.
         * @param result The raw, remote result object.
         */
        void onServerResult(NcsCloudDataUploader uploader,
                            String grammarId,
                            TransactionResult result);

        /**
         * The server emitted an error. For debugging purposes mainly.
         * @param uploader The current uploader.
         * @param grammarId The grammar ID of the word list that was uploaded.
         * @param error The raw, remote error object.
         */
        void onServerError(NcsCloudDataUploader uploader,
                           String grammarId,
                           TransactionError error);
    }

    /**
     * Listens for the result of the delete transaction
     */
    public interface DeleteListener
    {
        /**
         * The deletion has completed
         * @param uploader
         * @param status The status from {@link Status}.
         */
        void onDone(NcsCloudDataUploader uploader, Status status);

        /**
         * The server emitted a result. For debugging purposes mainly.
         * @param uploader The current uploader.
         * @param result The raw, remote result object.
         */
        void onServerResult(NcsCloudDataUploader uploader,
                            TransactionResult result);

        /**
         * The server emitted an error. For debugging purposes mainly.
         * @param uploader The current uploader.
         * @param error The raw, remote error object.
         */
        void onServerError(NcsCloudDataUploader uploader,
                           TransactionError error);
    }

    private static final int STATUS_FILENAME_VERSION = 1;
    private static final String STATUS_FILENAME = "upload.stat";

    private static enum UploadMode
    {
        NONE,
        INCREMENTAL,
        FULL,
        BUSY
    };

    private static class Upload
    {
        private Transaction transaction;
        private WordIterator iterator;
        private UploadMode mode;
        private boolean continueUpload;
        private boolean hintFullUpload;
        private int numDataBlocks;
        private int currentVersion;
        private long currentRetryTimeMs;
    }

    private static final String VERSION_CMD_NAME = "NCS_GET_DATA_VERSION_CMD";
    private static final String UPLOAD_CMD_NAME = "NCS_UPLOAD_DATA_CMD";
    private static final String DELETE_CMD_NAME = "NCS_DELETE_DATA_CMD";

    private static final String HEADER_PARAM_NAME = "REQUEST_INFO";
    private static final String DATA_PARAM_NAME = "DATA_BLOCK";
    private static final String DONE_PARAM_NAME = "UPLOAD_INFO";

    private final CloudServices _cloud;
    private final NcsCloudDataUploaderConfig _config;
    private final FileManager _fileManager;
    private final AsyncTaskHandler _taskHandler;
    private final NMTHandler _mainThreadHandler;
    private Upload _upload;

    // All uploader share the same file
    private static final Object _fileSyncObj = new Object();

    /**
     * @param cloudServices The cloud services instance to use
     */
    public NcsCloudDataUploader(CloudServices cloudServices)
    {
        this(cloudServices, new NcsCloudDataUploaderConfig(), null);
    }

    /**
     * @param cloudServices The cloud services instance to use
     * @param config The custom configuration, see {@link NcsCloudDataUploaderConfig}
     * @param fileManager A platform file manager, if available. Can be null.
     */
    public NcsCloudDataUploader(CloudServices cloudServices, NcsCloudDataUploaderConfig config, FileManager fileManager)
    {
        Checker.checkArgForNull("cloudServices", cloudServices);
        Checker.checkArgForNull("config", config);
        _cloud = cloudServices;
        _config = config;
        _fileManager = fileManager;
        _taskHandler = new AsyncTaskHandler();
        _mainThreadHandler = _cloud.getMainThreadHandler();
    }

    /**
     * Upload a word list to the server. This method will decide whether to
     * upload the full list, or to upload only recent modifications.
     * @param grammarId The grammar ID to attach to the word list
     * @param type The type of the content, e.g. "contacts" or "custom_words" or "structured_content"
     * @param list The list to upload
     * @param forceFullUpload true to force the upload to clear and upload everything
     * @param listener
     */
    public void uploadData(final String grammarId,
                           final String type,
                           final String language,
                           final WordList list,
                           final boolean forceFullUpload,
                           final UploadListener listener)
    {
        Checker.checkStringArgForNullOrEmpty("grammarId", grammarId);
        Checker.checkStringArgForNullOrEmpty("type", type);
        Checker.checkStringArgForNullOrEmpty("language", language);
        Checker.checkArgForNull("list", list);
        Checker.checkArgForNull("listener", listener);

        uploadData(null, 0, grammarId, type, language, list, forceFullUpload, true, listener);
    }

    private void uploadData(final WordIterator currentIterator,
                            final int currentVersion,
                            final String grammarId,
                            final String type,
                            final String language,
                            final WordList list,
                            final boolean forceFullUpload,
                            final boolean initialUpload,
                            final UploadListener listener)
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                // Cancel any transaction in progress
                cancel();

                final Upload upload = new Upload();
                final int priority;

                if (currentIterator != null)
                {
                    // 2nd or N time we are calling the method, it is an ongoing upload
                    upload.transaction = getDataTransaction(upload, grammarId, type, language, list, listener);
                    priority = Transaction.DefaultPriorities.DATA_UPLOAD;
                }
                else
                {
                    // 1st time the method is called, check version first
                    // The transaction will not be enabled until the word list
                    // checksum has been computed.
                    upload.transaction = getVersionTransaction(upload, grammarId, type, language, list, listener);
                    priority = Transaction.DefaultPriorities.FIRST_DATA_UPLOAD;
                }

                // Add the transaction right away, to avoid another transaction added
                // first in the queue. This is to support multiple CloudDataUploaders
                // being started in succession, where a subsequent recognition is queued
                // after them.
                _cloud.addTransaction(upload.transaction, priority);
                _upload = upload;

                // Run a task to initialize the WordIterator, and either enable the
                // transaction or cancel it (if the WordIterator is empty)
                _taskHandler.post(new AsyncTaskHandler.TaskRunnable<Pair<WordIterator, UploadMode>>()
                                  {
                                    @Override
                                    public Pair<WordIterator, UploadMode> run()
                                    {
                                        boolean hintFullUpload = initialUpload && (forceFullUpload || list.fullUpdateRequired());
                                        WordIterator it = currentIterator;
                                        if (it != null)  {
                                            upload.currentVersion = currentVersion;
                                        }
                                        else {
                                            it = hintFullUpload ? list.getFullIterator() : list.getModifiedIterator();
                                            upload.currentVersion = list.getAcceptedChecksum();
                                        }
                                        upload.hintFullUpload = hintFullUpload;

                                        if (!hintFullUpload && !it.hasNext()) {
                                            it.discardChanges();
                                            return new Pair<WordIterator, UploadMode>(null, UploadMode.NONE); // No transaction is necessary
                                        }

                                        // If the server is still busy, cancel the pending transaction
                                        upload.currentRetryTimeMs = 0;
                                        long retryDateMs = readStatusFile();
                                        long currentRetryTimeMs = retryDateMs - System.currentTimeMillis();
                                        if (currentRetryTimeMs > 0) {
                                             upload.currentRetryTimeMs = currentRetryTimeMs;
                                             it.discardChanges();
                                             return new Pair<WordIterator, UploadMode>(null, UploadMode.BUSY);
                                        }

                                        return new Pair<WordIterator, UploadMode>(it, hintFullUpload ? UploadMode.FULL : UploadMode.INCREMENTAL);
                                    }

                                    @Override
                                    public void onPostRun(final Pair<WordIterator, UploadMode> p)
                                    {
                                        if (upload != _upload)
                                            return;

                                        WordIterator iterator = p.first;
                                        UploadMode mode = p.second;
                                        if (mode == UploadMode.NONE || mode == UploadMode.BUSY)
                                        {
                                            // Upload isn't necessary (nothing to upload or client thinks the server is still busy)
                                            _upload.mode = mode;
                                            // Cancel the transaction.  Wait for the transactions's onError
                                            // callback, and notify the caller from there.
                                            _upload.transaction.cancel();
                                        }
                                        else
                                        {
                                            _upload.iterator = iterator;

                                            // Kick-start whatever transaction is relevant at this point
                                            _upload.transaction.enable();
                                        }
                                    }
                                  });
            }
        });
    }

    /**
     * Delete everything from the server
     * @param listener
     */
    public void deleteAll(final DeleteListener listener)
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                // Cancel any transaction in progress
                cancel();

                final Upload upload = new Upload();
                _upload = upload;
                Transaction t = new Transaction(DELETE_CMD_NAME, null, new Transaction.Listener()
                {
                    @Override
                    public void onTransactionStarted(Transaction t)
                    {
                    }

                    @Override
                    public void onTransactionProcessingStarted(Transaction t)
                    {
                    }

                    @Override
                    public void onTransactionResult(Transaction t, TransactionResult result, boolean complete)
                    {
                        if (upload != _upload)
                            return;

                        _upload = null;

                        listener.onDone(NcsCloudDataUploader.this, Status.SUCCESS);
                    }

                    @Override
                    public void onTransactionError(Transaction t, TransactionError error)
                    {
                        if (upload != _upload)
                            return;

                        _upload = null;

                        listener.onDone(NcsCloudDataUploader.this, Status.ERROR_NETWORK);
                    }

                    @Override
                    public void onTransactionIdGenerated(String transactionId)
                    {
                    }
                }, _config.commandTimeoutMs);
                upload.transaction = t;

                // Since we don't share grammars across devices, we just send an empty dictionary here.
                t.addParam(new DictionaryParam(HEADER_PARAM_NAME, new Dictionary()));
                t.finish();

                _cloud.addTransaction(t, 0);
            }
        });
    }

    /**
     * Cancel the current upload or delete transaction, if any
     */
    public void cancel()
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                if (_upload != null)
                {
                    if (_upload.transaction != null)
                        _upload.transaction.cancel();
                    _upload = null;
                }
            }
        });
    }

    private Transaction getVersionTransaction(final Upload upload,
                                              final String grammarId,
                                              final String type,
                                              final String language,
                                              final WordList list,
                                              final UploadListener listener)
    {
        final Transaction t = new Transaction(VERSION_CMD_NAME, _config.commandSettings, new Transaction.Listener()
        {
            @Override
            public void onTransactionStarted(final Transaction t)
            {
                if (_upload != upload)
                    return;

                t.addParam(getGrammarInfo(grammarId, type));
                t.finish();
            }

            @Override
            public void onTransactionProcessingStarted(Transaction t)
            {
            }

            @Override
            public void onTransactionResult(Transaction t, final TransactionResult result, boolean complete)
            {
                if (_upload != upload)
                    return;

                listener.onServerResult(NcsCloudDataUploader.this, grammarId, result);

                boolean presentTemp = false;
                int serverVersionTemp = 0;
                long timeToUploadMsTemp = 0;
                if (result != null && result.getContents() != null)
                {
                    Data.Dictionary resultDict = result.getContents();
                    if (resultDict.getInt("version") != null)
                    {
                        presentTemp = true;
                        serverVersionTemp = resultDict.getInt("version").value;
                    }
                    timeToUploadMsTemp = getRetryTimeMs(resultDict.getInt("upload_in"));
                }

                final boolean present = presentTemp;
                final int serverVersion = serverVersionTemp;
                final long timeToUploadMs = timeToUploadMsTemp;
                _taskHandler.post(new AsyncTaskHandler.TaskRunnable<Boolean>()
                        {
                          @Override
                          public Boolean run()
                          {
                              writeStatusFile(timeToUploadMs);
                              return new Boolean(true);
                          }

                          @Override
                          public void onPostRun(Boolean success)
                          {
                              if (_upload != upload)
                                  return;

                              if (success)
                              {
                                  if (timeToUploadMs > 0)
                                  {
                                      // Server is busy
                                      listener.onDone(NcsCloudDataUploader.this, Status.ERROR_SERVER_BUSY, grammarId, timeToUploadMs);
                                  }
                                  else
                                  {
                                      final boolean mismatchBasedForceUpload;

                                      // Do some integrity check on the result
                                      if (present && upload.currentVersion == serverVersion)
                                      {
                                          mismatchBasedForceUpload = upload.hintFullUpload;
                                      }
                                      else
                                      {
                                          // Server or client can't be trusted, force an upload
                                          mismatchBasedForceUpload = true;
                                      }

                                      // Move on to the upload
                                      uploadData(upload.iterator,
                                                 upload.currentVersion,
                                                 grammarId, type, language, list,
                                                 mismatchBasedForceUpload, true,
                                                 listener);
                                  }
                              }
                              else
                              {
                                  listener.onDone(NcsCloudDataUploader.this, Status.ERROR_NETWORK, grammarId, 0);
                              }
                          }
                        });
            }

            @Override
            public void onTransactionError(Transaction t, TransactionError error)
            {
                if (_upload != upload)
                    return;

                listener.onServerError(NcsCloudDataUploader.this, grammarId, error);

                if (_upload.mode == UploadMode.NONE)
                {
                    // Nothing to upload
                    listener.onDone(NcsCloudDataUploader.this, Status.SUCCESS_NOTHING_TO_UPLOAD, grammarId, 0);
                }
                else if (_upload.mode == UploadMode.BUSY)
                {
                    listener.onDone(NcsCloudDataUploader.this, Status.ERROR_SERVER_BUSY, grammarId, _upload.currentRetryTimeMs);
                }
                else
                {
                    // Legit error
                    listener.onDone(NcsCloudDataUploader.this, Status.ERROR_NETWORK, grammarId, 0);
                }
            }

            @Override
            public void onTransactionIdGenerated(String transactionId)
            {
            }
        }, _config.commandTimeoutMs, false); // The caller starts the transaction when appropriate
        return t;
    }

    private Transaction getDataTransaction(final Upload upload,
                                           final String grammarId,
                                           final String type,
                                           final String language,
                                           final WordList list,
                                           final UploadListener listener)
    {
        Dictionary settings = _config.commandSettings;
        if (settings == null) {
            settings = new Data.Dictionary();
        }
        settings.put("dictation_language", language);
        settings.put("dictation_type", "websearch"); // Not relevant for upload but server needs it
        final Transaction t = new Transaction(UPLOAD_CMD_NAME, settings, new Transaction.Listener()
        {
            @Override
            public void onTransactionStarted(final Transaction t)
            {
                if (_upload != upload)
                    return;

                // The transaction shouldn't start until it has been enabled
                final WordIterator it = upload.iterator;

                Assert.assertNotNull(it);

                final AsyncTaskHandler.TaskRunnable<DictionaryParam> uploadDoneTask
                      = new AsyncTaskHandler.TaskRunnable<DictionaryParam>()
                        {
                            @Override
                            public DictionaryParam run()
                            {
                                if (_upload == upload) {
                                    boolean generateGrammar = true;
                                    if (upload.iterator != null) {
                                        // Generating the grammar is expensive, only do it for the last upload command.
                                        generateGrammar = !upload.iterator.hasNext();
                                    }
                                    return new DictionaryParam(DONE_PARAM_NAME, constructUploadInfo(grammarId, type, list, upload.currentVersion, it.getChecksum(), upload.hintFullUpload, generateGrammar));
                                }
                                return null;
                            }

                            @Override
                            public void onPostRun(DictionaryParam param)
                            {
                                if (_upload == upload) {
                                    t.addParam(param);
                                    t.finish();
                                }
                            }
                        };

                AsyncTaskHandler.TaskRunnable<DictionaryParam> dataBlockTask
                      = new AsyncTaskHandler.TaskRunnable<DictionaryParam>()
                        {
                            @Override
                            public DictionaryParam run()
                            {
                                DictionaryParam param = null;

                                if (_upload == upload)
                                {
                                    Logger.debug(this, "Entering dataBlockTask, grammarId = " + grammarId);
                                    final Dictionary dataBlockDict = constructDataBlock(it, _config.commandChunkSize);
                                    boolean hasMore = false;
                                    if (upload.iterator != null)
                                    {
                                        hasMore = upload.iterator.hasNext();
                                    }
                                    boolean keepCurrentCommand = _config.commandChunkSize * (upload.numDataBlocks + 1) < _config.commandTotalSize;

                                    if (upload.continueUpload)
                                    {
                                        if (hasMore && keepCurrentCommand)
                                        {
                                            param = new SeqChunkParam(DATA_PARAM_NAME, dataBlockDict);
                                        }
                                        else
                                        {
                                            param = new SeqEndParam(DATA_PARAM_NAME, dataBlockDict);
                                            upload.continueUpload = false;
                                        }
                                    }
                                    else
                                    {
                                        if (hasMore && keepCurrentCommand)
                                        {
                                            param = new SeqStartParam(DATA_PARAM_NAME, dataBlockDict);
                                            upload.continueUpload = true;
                                        }
                                        else
                                        {
                                            param = new DictionaryParam(DATA_PARAM_NAME, dataBlockDict);
                                        }
                                    }
                                       Logger.debug(this, "Exiting dataBlockTask, grammarId = " + grammarId);
                                }

                                return param;
                            }

                            @Override
                            public void onPostRun(DictionaryParam param)
                            {
                                if (_upload == upload) {
                                    upload.numDataBlocks++;
                                    t.addParam(param);

                                    if (upload.continueUpload)
                                    {
                                        _taskHandler.post(this);
                                    }
                                    else
                                    {
                                        _taskHandler.post(uploadDoneTask);
                                    }
                                }
                            }
                        };

                _taskHandler.post(dataBlockTask);
            }

            @Override
            public void onTransactionProcessingStarted(Transaction t)
            {
            }

            @Override
            public void onTransactionResult(Transaction t, final TransactionResult result, boolean complete)
            {
                if (_upload != upload)
                    return;

                listener.onServerResult(NcsCloudDataUploader.this, grammarId, result);
                Status tempStatus;
                long tempTimeToUploadMs = 0;
                try {
                    Data.Integer uploadIn = result.getContents().getDictionary("data_storage_result").getInt("upload_in");
                    String dataStorageResult = result.getContents().getDictionary("data_storage_result").getString("status").value;
                    String grammarGenerationResult = result.getContents().getDictionary("grammar_generation_result").getString("status").value;
                    if (dataStorageResult.equals("success") && (grammarGenerationResult.equals("success") || grammarGenerationResult.equals("not_applicable"))) {
                        tempStatus = Status.SUCCESS;
                    }
                    else if (dataStorageResult.equals("success") && !(grammarGenerationResult.equals("success") || grammarGenerationResult.equals("not_applicable"))) {
                        tempStatus = Status.ERROR_GRAMMAR;
                    }
                    else if (uploadIn != null) {
                        tempTimeToUploadMs = getRetryTimeMs(uploadIn);
                        tempStatus = Status.ERROR_SERVER_BUSY;
                    }
                    else {
                        tempStatus = Status.ERROR_STORAGE;
                    }
                }
                catch (Exception e) {
                    Logger.error(this, "Error parsing the transaction result when uploading grammar = " + grammarId, e);
                    tempStatus = Status.ERROR_NETWORK;
                }

                final Status status = tempStatus;
                final long timeToUploadMs = tempTimeToUploadMs;
                AsyncTaskHandler.TaskRunnable<Pair<Boolean, Integer>> resultTask
                = new AsyncTaskHandler.TaskRunnable<Pair<Boolean, Integer>>()
                  {
                      @Override
                      public Pair<Boolean, Integer> run()
                      {
                          if (_upload == upload)
                          {
                              if (status == Status.SUCCESS) {
                                  boolean hasMore = false;
                                  if (upload.iterator != null) {
                                      if (upload.iterator.hasNext()) {
                                          hasMore = true;
                                      }
                                      else {
                                          upload.iterator.acceptChanges();
                                      }
                                  }
                                  return new Pair<Boolean, Integer>(hasMore, upload.iterator.getChecksum());
                              }
                              else
                              {
                                  if (upload.iterator != null)
                                      upload.iterator.discardChanges();
                              }
                          }

                          writeStatusFile(timeToUploadMs);
                          return new Pair<Boolean, Integer>(false, 0);
                      }

                      @Override
                      public void onPostRun(Pair<Boolean, Integer> hasMoreAndCurrentVersion)
                      {
                          if (_upload == upload)
                          {
                              _upload = null;

                              if (status == Status.SUCCESS)
                              {
                                  if (hasMoreAndCurrentVersion.first) {
                                      Logger.debug(this, "More items, sending next command for grammarId = " + grammarId);
                                      uploadData(upload.iterator, hasMoreAndCurrentVersion.second, grammarId, type, language, list, false, false, listener);
                                  }
                                  else {
                                      listener.onDone(NcsCloudDataUploader.this, status, grammarId, 0);
                                  }
                              }
                              else
                              {
                                  Logger.error(this, "Server error while uploading grammar = " + grammarId);
                                  listener.onDone(NcsCloudDataUploader.this, status, grammarId, timeToUploadMs);
                              }
                          }
                      }
                  };

                _taskHandler.post(resultTask);
            }

            @Override
            public void onTransactionError(Transaction t, TransactionError error)
            {
                if (_upload != upload)
                    return;

                Logger.error(this, "onTransactionError(): error code: " + (error != null ? error.getErrorCode() : "unknown"));
                listener.onServerError(NcsCloudDataUploader.this, grammarId, error);

                if (upload.iterator != null)
                    upload.iterator.discardChanges();

                if (upload == _upload)
                    _upload = null;

                listener.onDone(NcsCloudDataUploader.this, Status.ERROR_NETWORK, grammarId, 0);
            }

            @Override
            public void onTransactionIdGenerated(String transactionId)
            {
            }
        }, _config.commandTimeoutMs, false); // The caller starts the transaction when appropriate
        return t;
    }

    private static Param getGrammarInfo(String id, String type)
    {
        Data.Dictionary requestInfo = new Data.Dictionary();
        requestInfo.put("id", id);
        requestInfo.put("type", type);
        return new DictionaryParam(HEADER_PARAM_NAME, requestInfo);
    }

    /** May take a while, so call this from a worker/async thread */
    private static Data.Dictionary constructDataBlock(WordIterator it, int maxItems)
    {
        Sequence putSequence = new Sequence();
        Sequence removeSequence = new Sequence();

        while (it.hasNext() && maxItems-- > 0)
        {
            WordAction wordAction = it.next();
            Word word = wordAction.getWord();
            Dictionary content = word.getCustomForm();
            if ((content == null || content.getEntries().isEmpty())
                && wordAction.getAction() != WordAction.REMOVE)
                continue;

            String id = String.valueOf(word.getUserId());
            switch (wordAction.getAction())
            {
            case WordAction.ADD:
            {
                Dictionary dict = new Dictionary();
                dict.put("id", id);
                dict.put("content", content);
                putSequence.add(dict);
                break;
            }
            case WordAction.UPDATE:
            {
                removeSequence.add(id);
                Dictionary dict = new Dictionary();
                dict.put("id", id);
                dict.put("content", content);
                putSequence.add(dict);
                break;
            }
            case WordAction.REMOVE:
            {
                removeSequence.add(id);
                break;
            }
            }
        }

        Dictionary retDataBlock = new Dictionary();
        retDataBlock.put("items_to_add", putSequence);
        retDataBlock.put("items_to_remove", removeSequence);
        return retDataBlock;
    }

    private static Dictionary constructUploadInfo(String grammarId, String type, WordList list, int currentVersion, int newVersion, boolean clear, boolean generateGrammar)
    {
        Dictionary upload_info = new Dictionary();
        upload_info.put("id", grammarId);
        upload_info.put("type", type);
        upload_info.put("version_current", currentVersion);
        upload_info.put("version_new", newVersion);
        upload_info.put("clear_all", clear ? 1 : 0);
        upload_info.put("generate_grammar", generateGrammar ? 1 : 0);
        return upload_info;
    }

    private static long getRetryTimeMs(Data.Integer uploadIn)
    {
        if (uploadIn != null)
            return uploadIn.value * 60 * 1000; // mn to ms conversion
        return 0;
    }

    // Helpers to write the status file.
    // Eventually those methods will be augmented to R/W more status information
    private void writeStatusFile(long retryTimeMs)
    {
        synchronized(_fileSyncObj)
        {
            FileOutputStream fos = null;
            ObjectOutputStream oos = null;
            try
            {
                if (_fileManager != null)
                {
                    fos = _fileManager.openFileForWriting(STATUS_FILENAME);
                    oos = new ObjectOutputStream(fos);
                    oos.writeInt(STATUS_FILENAME_VERSION);
                    oos.writeUTF(String.valueOf(System.currentTimeMillis() + retryTimeMs));
                }
            }
            catch (Exception e)
            {
                Logger.error(this, "Unable to write status file", e);
            }
            finally
            {
                if (oos != null)
                {
                    try {
                        oos.close();
                    } catch (IOException e) {}
                }
                if (fos != null)
                {
                    try {
                        fos.close();
                    } catch (IOException e) {}
                }
            }
        }
    }

    private long readStatusFile()
    {
        synchronized(_fileSyncObj)
        {
            long retryDateMs = System.currentTimeMillis();
            FileInputStream fis = null;
            ObjectInputStream ois = null;
            try
            {
                if (_fileManager != null)
                {
                    fis = _fileManager.openFileForReading(STATUS_FILENAME);
                    if (fis != null)
                    {
                        ois = new ObjectInputStream(fis);
                        if (STATUS_FILENAME_VERSION == ois.readInt())
                        {
                            retryDateMs = Long.valueOf(ois.readUTF());
                        }
                    }
                }
            }
            catch (Exception e)
            {
                Logger.error(this, "Unable to read status file", e);
            }
            finally
            {
                if (ois != null)
                {
                    try {
                        ois.close();
                    } catch (IOException e) {}
                }
                if (fis != null)
                {
                    try {
                        fis.close();
                    } catch (IOException e) {}
                }
            }
            return retryDateMs;
        }
    }
}
