package com.nuance.dragon.toolkit.cloudservices.recognizer;

import java.util.Set;

import org.json.JSONException;

import com.nuance.dragon.toolkit.core.WordList;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;

/**
 * A specification for a cloud recognition with asr command.
 */
public abstract class AsrSpec implements JSONCompliant
{
    private static final String KEY_ASR_TYPE        = "asrtype";
    
    public final static int ASR_TYPE_NVC = 0;
    public final static int ASR_TYPE_NCS = 1;
    
    /**
     * Get the type of grammar with id.
     *
     * GRAMMAR_STRUCTURED should NOT be mixed with
     * GRAMMAR_CONTACTS/GRAMMAR_CUSTOM_WORDS
     *
     * @return type of Grammar.
     */
    public abstract String getGrammarType(String id);

    /**
     * Get the category of grammar with id.
     *
     * @return category of Grammar.
     */
    public abstract String getGrammarCategory(String id);

    /**
     * Use this function to set the ID of the contact list that is needed by the associated Server Grammar.
     * Once the contactListId has been set, it cannot be changed.  Calling this function a second time will
     * have no effect.
     *
     * @param contactListId The ID of the contact list that is required by this object.
     * Cannot be null or empty.
     *
     */
    public abstract void setContactListId(String contactListId);

    /**
     * Use this function to determine the id of the contact list required by this object, if any.
     *
     * @return The contact list id that is required by this object.
     * Will be null if no contact list is required.
     */
    public abstract String getContactListId();

    /**
     * Use this function to add the IDs of Custom WordList that are expected by the associated Server Grammar.
     * This function can be called multiple times to add partial lists of IDs, but once an ID has been added,
     * it cannot be removed.
     * Note that the ID of the contact list does not need to be included in this list.
     *
     * @param wordListIds A Set object containing IDs of the custom word lists required by the Server Grammar.
     * Cannot be null or empty.
     */
    public abstract void addRequiredCustomWordListIds(Set<String> wordListIds);
    
    /**
     * Use this function to determine the IDs of the custom word lists required by this object, if any.
     *
     * @return The set of custom word list IDs that are required by the Server Grammar associated with
     * this object.
     * Will be empty (but never null) if no word lists are required.
     */
    public abstract Set<String> getRequiredCustomWordListIds();

    /**
     * Use this function to add the IDs of structured content
     * Note that we should not use 'structured content' mixed with 'contacts/custom_words'.
     *
     * @param id A string specifies content Id.
     * @param category A string specifies content category
     * Cannot be null or empty.
     */
    public abstract void addStructuredContent(String id, String category);

    /**
     * Returns the structured content IDs required by this object, if any.
     *
     * @return The set of structured content IDs associated with this object.
     * Will be empty (but never null) if no word lists are required.
     */
    public abstract Set<String> getStructuredContentIds();

    /**
     * This function is used to attach a {@link com.nuance.dragon.toolkit.core.WordList}
     * to this object to use as a contact list or custom word list.
     * Note that the ID of the WordList must match either the contact list, a custom word list or structured content list.
     * Also note that the attached WordLists are not included in the JSON representation of this object.
     * They must be attached each time the object is create from JSON.
     *
     * @param wordList The word list to attach.  Must not be null.
     * @param id The identifier for the word list.  Must not be null or empty.
     *
     * @return true if the list was attached, false if the list was not attached.
     */
    public abstract boolean attachWordList(WordList wordList, String id);

    /**
     * Add any extra settings desired for this command.
     *
     * @param settings The settings to use for this command.
     * Note that any previous settings that were attached will be replaced.
     * Use null to remove previously attached Settings.
     */
    public abstract void attachSettings(Data.Dictionary settings);

    /**
     * Set the dictation type that will be used by this scenario.
     *
     * @param dictationType The dictation type.
     */
    public abstract void setDictationType(String dictationType);

    /**
     * Create a {@link RecogSpec} that can be used to initialize a CloudRecognition.
     * Note that all required Contact Lists and Custom Word lists must be attached before
     * this method is called.
     *
     * @return A RecogSpec, or null if required parameters have not been defined.
     *
     */
    public abstract RecogSpec createRecogSpec();
    
    public static org.json.JSONObject toJSON(AsrSpec spec)
    {
        if (spec == null)
            return null;
        
        org.json.JSONObject o = spec.toJSON();
        try {
            if (spec instanceof NvcAsrSpec)
                o.put(KEY_ASR_TYPE, ASR_TYPE_NVC);
            else if (spec instanceof NcsAsrSpec)
                o.put(KEY_ASR_TYPE, ASR_TYPE_NCS);
            else
                throw new JSONException("AsrSpec class type not recognized!!!");
        } catch (JSONException e) {
            Logger.error("AsrSpec.toJSON()", "Error adding to JSON");
        }
        return o;
    }
    
    public static AsrSpec createFromJSON(org.json.JSONObject json) throws JSONException
    {
        int asrType = json.getInt(KEY_ASR_TYPE);
        if (asrType == ASR_TYPE_NVC)
            return NvcAsrSpec.createFromJSON(json);
        else if (asrType == ASR_TYPE_NCS)
            return NcsAsrSpec.createFromJSON(json);
        else
            throw new JSONException("ASR_TYPE_NVC not recognized!!!");
    }
}
