package com.nuance.dragon.toolkit.cloudservices.recognizer;

import java.util.List;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.pipes.SimplePipe;
import com.nuance.dragon.toolkit.cloudservices.AudioParam;
import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.DataParam;
import com.nuance.dragon.toolkit.cloudservices.Param;
import com.nuance.dragon.toolkit.cloudservices.Transaction;
import com.nuance.dragon.toolkit.cloudservices.TransactionError;
import com.nuance.dragon.toolkit.cloudservices.TransactionResult;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * A cloud recognizer performs a simple network recognition using a cloud
 * services transaction.
 */
public class CloudRecognizer
{
    /**
     * Listener interface for cloud recognition results.
     */
    public interface Listener
    {
        /**
         * A result was received.
         * @param result
         */
        void onResult(CloudRecognitionResult result);

        /**
         * An error occurred.
         * @param error
         */
        void onError(CloudRecognitionError error);
        
        /**
         * The transaction Id is generated.
		 *
		 * @param transactionId The Id of the transaction.
         */
        void onTransactionIdGenerated(String transactionId);
    }

    private static final class CloudRecognition
    {
        private Transaction _transaction;

        // We only want to finalize the transaction after the app has called
        // processResult(), and the transaction has started (which is where the
        // delayed params are added)
        private boolean _transactionStarted;
        private boolean _processing;
    }

    private final CloudServices _cloud;
    private final NMTHandler _mainThreadHandler;
    private final SimplePipe<AudioChunk> _audioPipe;
    private CloudRecognition _currentRecog;
    private String _sessionId;
    private String _transactionId;
    private int _cadenceMs;

    /**
     * @param cloudServices The cloud services instance to use
     */
    public CloudRecognizer(CloudServices cloudServices)
    {
        Checker.checkArgForNull("cloudServices", cloudServices);
        _cloud = cloudServices;
        _mainThreadHandler = _cloud.getMainThreadHandler();
        _audioPipe = new SimplePipe<AudioChunk>(cloudServices.getMainThreadHandler());
        _cadenceMs = 0;
    }


    /**
     * Start a recognition.
     * @param recogSpec The specification for the recognition.
     * @param audioSource The audio source. The recognizer will connect to the
     * audio source until {@link #processResult} or {@link #cancel} is called.
     * @param resultListener The result listener.
     */
    public void startRecognition(final RecogSpec recogSpec,
                                 final AudioSource<AudioChunk> audioSource,
                                 final Listener resultListener)
    {
        Checker.checkArgForNull("recogSpec", recogSpec);
        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForNull("resultListener", resultListener);

        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                // Cancel any recognition in progress
                cancel();
        
                final CloudRecognition recog = new CloudRecognition();
                _currentRecog = recog;
        
                recog._transaction = new Transaction(recogSpec.getCommand(), recogSpec.getSettings(), new Transaction.Listener()
                {
                    @Override
                    public void onTransactionStarted(Transaction t)
                    {
                        if (recog != _currentRecog)
                            return;
        
                        recog._transactionStarted = true;
        
                        // Send the delayed parameters
                        for (DataParam delayedParam : recogSpec.getDelayedParams())
                        {
                            recog._transaction.addParam(delayedParam);
                        }
        
                        if (recog._processing)
                            recog._transaction.finish();
                    }
        
                    @Override
                    public void onTransactionProcessingStarted(Transaction t)
                    {
                    }
        
                    @Override
                    public void onTransactionResult(Transaction t, TransactionResult result, boolean complete)
                    {
                        if (recog != _currentRecog)
                            return;
        
                        if (complete) {
                            _currentRecog = null;
                            //NPM: only disconnect audio source if this final result
                            _audioPipe.disconnectAudioSource();
                        }
                        setIds(result.getSessionId());
                        resultListener.onResult(new CloudRecognitionResult(result));
                    }
        
                    @Override
                    public void onTransactionError(Transaction t, TransactionError error)
                    {
                        if (recog == _currentRecog)
                            _currentRecog = null;
        
                        _audioPipe.disconnectAudioSource();
                        setIds(error.getSessionId());
        
                        resultListener.onError(new CloudRecognitionError(error));
                    }
        
                    @Override
                    public void onTransactionIdGenerated(String transactionId)
                    {
                        resultListener.onTransactionIdGenerated(transactionId);
                    }
                }, recogSpec.getTimeout(), true, _cadenceMs);
        
                // Hook up an audio pipe to enable us to disconnect in processResult()
                // to cause sourceClosed and finishing the command.
                _audioPipe.connectAudioSource(audioSource);
        
                List<DataParam> params = recogSpec.getParams();
                for (Param p : params)
                {
                    recog._transaction.addParam(p);
                }
        
                recog._transaction.addParam(new AudioParam(recogSpec.getAudioParam(), _audioPipe));
        
                _cloud.addTransaction(recog._transaction, Transaction.DefaultPriorities.RECOGNITION);
            }
        });
    }

    /**
     * End the recognition command, and wait for a result from the server. No
     * more audio or parameters can be processed after calling this.
     */
    public void processResult()
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                if (null != _currentRecog)
                {
                    if (_currentRecog._processing)
                    {
                        Logger.warn(this, "current recognition has been already processed");
                        return;
                    }
                    _currentRecog._processing = true;
        
                    if (_currentRecog._transactionStarted)
                        _currentRecog._transaction.finish();
                }
            }
        });
    }

    /**
     * Cancel the recognition. This is safe to call even if there is no active
     * recognition.
     */
    public void cancel()
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                _audioPipe.disconnectAudioSource();
                if (_currentRecog != null)
                {
                    Transaction t = _currentRecog._transaction;
                    _currentRecog = null;
                    t.cancel();
                }
            }
        });
    }

    /**
     * Send an extra parameter. This must be called after calling {@link #startRecognition},
     * but before calling {@link #processResult}.
     * @param param The parameter to add to the recognition command
     */
    public void sendParam(final Param param)
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                Checker.checkState(this, _currentRecog != null, "No recognition in progress");
                Checker.checkState(this, !_currentRecog._processing, "Can't add params after already processing");
        
                if (_currentRecog != null)
                {
                    _currentRecog._transaction.addParam(param);
                }
            }
        });
    }

    /**
     * Get the session ID.
     * @return The session ID.
     */
    public String getSessionId()
    {
        return _sessionId;
    }

    /**
     * Get the transaction ID.
     * @return The transaction ID.
     */
    public String getTransactionId()
    {
        return _transactionId;
    }

    /**
     * Set the cadence at which multiple results can be received.
     * <p>
     * This feature is relevant when results are streamed back from the server
     * (per-word recognition). Setting the cadence guarantees that all results
     * that are sent back will be aligned to the cadence period specified.
     * For example, if the cadence is set at 250ms, every 250ms the system
     * tries to obtain an updated result. If an updated result is found,
     * it is returned through the resultListener (see startRecognition).
     * If no updated result is available, the last result is sent back in this cycle.
     * The final result is always sent back as soon as it is available.
     * When the cadence is set to 0, updates are sent back through the Listener
     * as soon as they are available. The default cadence is 0.
     *
     * @param cadenceMs    The time between successive result updates.
     */
    public void setResultCadence(final int cadenceMs)
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                _cadenceMs = cadenceMs;
            }
        });
    }

    private void setIds(final String sessionId)
    {
        if (sessionId != null && sessionId.length() > 0)
        {
            int transIdIdx = sessionId.indexOf(":");
            if (transIdIdx != -1)
            {
                _sessionId = sessionId.substring(0, transIdIdx);
                _transactionId = sessionId.substring(transIdIdx + 1);
            }
        }
    }
}
