package com.nuance.dragon.toolkit.cloudservices.recognizer;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.json.JSONException;

import com.nuance.dragon.toolkit.cloudservices.DataParam;
import com.nuance.dragon.toolkit.cloudservices.DictionaryParam;
import com.nuance.dragon.toolkit.core.WordList;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.util.internal.ReadOnlyList;

/**
 * A specification for a cloud recognition with NVC_ASR_CMD.
 */
public class NvcAsrSpec extends AsrSpec
{
    private static class GrammarInfo implements JSONCompliant {
        private static final String KEY_TYPE                 = "type";
        private static final String KEY_CATEGORY             = "category";

        public final String  type;
        public final String  category;
        public WordList      wordList;

        public GrammarInfo(String inputType, String inputCategory) {
            type = inputType;
            category = inputCategory;
        }

        //wordList will not be added into Json
        @Override
        public org.json.JSONObject toJSON() {
            JSONObject o = new JSONObject();
            o.tryPut(KEY_TYPE, type);
            o.tryPut(KEY_CATEGORY, category);
            return o;
        }

        public static GrammarInfo createFromJSON(org.json.JSONObject json) throws JSONException
        {
            String type = json.getString(KEY_TYPE);

            String category = null;
            if (!json.isNull(KEY_CATEGORY)) {
                category = json.getString(KEY_CATEGORY);
            }
            GrammarInfo data = new GrammarInfo(type, category);
            return data;
        }

        public boolean setWordList(WordList wordList)
        {
            this.wordList = wordList;
            return true;
        }

        public WordList getWordList()
        {
            return wordList;
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result
                    + ((category == null) ? 0 : category.hashCode());
            result = prime * result + ((type == null) ? 0 : type.hashCode());
            // Commented intentionally
            //result = prime * result
            //        + ((wordList == null) ? 0 : wordList.hashCode());
            return result;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            GrammarInfo other = (GrammarInfo) obj;
            if (category == null) {
                if (other.category != null)
                    return false;
            } else if (!category.equals(other.category))
                return false;
            if (type == null) {
                if (other.type != null)
                    return false;
            } else if (!type.equals(other.type))
                return false;
            // Commented intentionally
            //if (wordList == null) {
            //    if (other.wordList != null)
            //        return false;
            //} else if (!wordList.equals(other.wordList))
            //    return false;
            return true;
        }
    }

    private static final String KEY_GRAMMAR_ID        = "grammarid";
    private static final String KEY_LANGUAGE          = "language";
    private static final String KEY_DICTATION_TYPE    = "dictationtype";
    private static final String KEY_CONTACTLIST_ID    = "contactlistid";
    private static final String KEY_GRAMMARLIST       = "grammarList";

    public static final String  GRAMMAR_CONTACTS       = "contacts";
    public static final String  GRAMMAR_CUSTOM_WORDS   = "custom_words";
    public static final String  GRAMMAR_STRUCTURED     = "structured_content";

    private final String                    _grammarId;
    private final String                    _language;
    private Data.Dictionary                 _settings;
    private String                          _contactListId;
    private String                          _dictationType;
    private HashMap<String, GrammarInfo>    _grammarList;

    /**
     * This object represents the information needed to do a Server based Constrained Recognition.
     */
    public NvcAsrSpec(String grammarId, String language)
    {
        _grammarId = grammarId;
        _language = language;
        _grammarList = new HashMap<String, GrammarInfo>();
        _contactListId = null;
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result
                + ((_contactListId == null) ? 0 : _contactListId.hashCode());
        result = prime * result
                + ((_dictationType == null) ? 0 : _dictationType.hashCode());
        result = prime * result
                + ((_grammarId == null) ? 0 : _grammarId.hashCode());
        result = prime * result
                + ((_grammarList == null) ? 0 : _grammarList.hashCode());
        result = prime * result
                + ((_language == null) ? 0 : _language.hashCode());
        result = prime * result
                + ((_settings == null) ? 0 : _settings.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        NvcAsrSpec other = (NvcAsrSpec) obj;
        if (_contactListId == null) {
            if (other._contactListId != null)
                return false;
        } else if (!_contactListId.equals(other._contactListId))
            return false;
        if (_dictationType == null) {
            if (other._dictationType != null)
                return false;
        } else if (!_dictationType.equals(other._dictationType))
            return false;
        if (_grammarId == null) {
            if (other._grammarId != null)
                return false;
        } else if (!_grammarId.equals(other._grammarId))
            return false;
        if (_grammarList == null) {
            if (other._grammarList != null)
                return false;
        } else if (!_grammarList.equals(other._grammarList))
            return false;
        if (_language == null) {
            if (other._language != null)
                return false;
        } else if (!_language.equals(other._language))
            return false;
        if (_settings == null) {
            if (other._settings != null)
                return false;
        } else if (!_settings.equals(other._settings))
            return false;
        return true;
    }

    /**
     * Get the type of grammar with id.
     *
     * GRAMMAR_STRUCTURED should NOT be mixed with
     * GRAMMAR_CONTACTS/GRAMMAR_CUSTOM_WORDS
     *
     * @return type of Grammar.
     */
    @Override
    public String getGrammarType(String id)
    {
        GrammarInfo info = _grammarList.get(id);
        if (info != null) {
            return info.type;
        }

        return "Invalid grammar ID";
    }

    /**
     * Get the category of grammar with id.
     *
     * @return category of Grammar.
     */
    @Override
    public String getGrammarCategory(String id)
    {
        GrammarInfo info = _grammarList.get(id);
        if (info != null) {
            return info.category;
        }

        return "Invalid grammar ID";
    }

    /**
     * Use this function to set the ID of the contact list that is needed by the associated Server Grammar.
     * Once the contactListId has been set, it cannot be changed.  Calling this function a second time will
     * have no effect.
     *
     * @param contactListId The ID of the contact list that is required by this object.
     * Cannot be null or empty.
     *
     */
    @Override
    public void setContactListId(String contactListId) {
        Checker.checkStringArgForNullOrEmpty("contactListId", contactListId);
        if (_contactListId == null) {
            _contactListId = contactListId;

            _grammarList.put(_contactListId, new GrammarInfo(GRAMMAR_CONTACTS, null));
        }
    }

    /**
     * Use this function to determine the id of the contact list required by this object, if any.
     *
     * @return The contact list id that is required by this object.
     * Will be null if no contact list is required.
     */
    @Override
    public String getContactListId() {
        return _contactListId;
    }

    /**
     * Use this function to add the IDs of Custom WordList that are expected by the associated Server Grammar.
     * This function can be called multiple times to add partial lists of IDs, but once an ID has been added,
     * it cannot be removed.
     * Note that the ID of the contact list does not need to be included in this list.
     *
     * @param wordListIds A Set object containing IDs of the custom word lists required by the Server Grammar.
     * Cannot be null or empty.
     */
    @Override
    public void addRequiredCustomWordListIds(Set<String> wordListIds) {
        Checker.checkArgForNull("wordListIds", wordListIds);
        Checker.checkArgForCondition("wordListIds", "not empty", !wordListIds.isEmpty());

        for (String id : wordListIds) {
            _grammarList.put(id, new GrammarInfo(GRAMMAR_CUSTOM_WORDS, null));
        }
    }

    /**
     * Use this function to determine the IDs of the custom word lists required by this object, if any.
     *
     * @return The set of custom word list IDs that are required by the Server Grammar associated with
     * this object.
     * Will be empty (but never null) if no word lists are required.
     */
    @Override
    public Set<String> getRequiredCustomWordListIds()
    {
        HashSet<String> requiredWordListIds = new HashSet<String>();
        for (Map.Entry<String, GrammarInfo> grammar :_grammarList.entrySet()) {
            GrammarInfo info = grammar.getValue();
            if (info.type.equals(GRAMMAR_CUSTOM_WORDS)) {
                requiredWordListIds.add(grammar.getKey());
            }
        }
        return requiredWordListIds;
    }

    /**
     * Use this function to add the IDs of structured content
     * Note that we should not use 'structured content' mixed with 'contacts/custom_words'.
     *
     * @param id A string specifies content Id.
     * @param category A string specifies content category
     * Cannot be null or empty.
     */
    @Override
    public void addStructuredContent(String id, String category) {
        Checker.checkArgForNull("id", id);
        Checker.checkArgForNull("category", category);

        _grammarList.put(id, new GrammarInfo(GRAMMAR_STRUCTURED, category));
    }

    /**
     * Returns the structured content IDs required by this object, if any.
     *
     * @return The set of structured content IDs associated with this object.
     * Will be empty (but never null) if no word lists are required.
     */
    @Override
    public Set<String> getStructuredContentIds() {
        HashSet<String> requiredWordListIds = new HashSet<String>();
        for (Map.Entry<String, GrammarInfo> grammar :_grammarList.entrySet()) {
            GrammarInfo info = grammar.getValue();
            if (info.type.equals(GRAMMAR_STRUCTURED)) {
                requiredWordListIds.add(grammar.getKey());
            }
        }
        return requiredWordListIds;
    }

    /**
     * This function is used to attach a {@link com.nuance.dragon.toolkit.core.WordList}
     * to this object to use as a contact list or custom word list.
     * Note that the ID of the WordList must match either the contact list, a custom word list or structured content list.
     * Also note that the attached WordLists are not included in the JSON representation of this object.
     * They must be attached each time the object is create from JSON.
     *
     * @param wordList The word list to attach.  Must not be null.
     * @param id The identifier for the word list.  Must not be null or empty.
     *
     * @return true if the list was attached, false if the list was not attached.
     */
    @Override
    public boolean attachWordList(WordList wordList, String id)
    {
        Checker.checkArgForNull("wordList", wordList);
        Checker.checkStringArgForNullOrEmpty("id", id);

        GrammarInfo info = _grammarList.get(id);
        if (info != null) {
            info.setWordList(wordList);
            return true;
        }

                return false;
            }

    private WordList getWordList(String id)
    {
        Checker.checkStringArgForNullOrEmpty("id", id);

        GrammarInfo info = _grammarList.get(id);
        if (info != null) {
            return info.getWordList();
        }

        return null;
    }

    /**
     * Add any extra settings desired for this command.
     *
     * @param settings The settings to use for this command.
     * Note that any previous settings that were attached will be replaced.
     * Use null to remove previously attached Settings.
     */
    @Override
    public void attachSettings(Data.Dictionary settings)
    {
        _settings = settings;
    }

    /**
     * Set the dictation type that will be used by this scenario.
     *
     * @param dictationType The dictation type.
     */
    @Override
    public void setDictationType(String dictationType)
    {
        _dictationType = dictationType;
    }

    /**
     * Create a {@link RecogSpec} that can be used to initialize a CloudRecognition.
     * Note that all required Contact Lists and Custom Word lists must be attached before
     * this method is called.
     *
     * @return A RecogSpec, or null if required parameters have not been defined.
     *
     */
    @Override
    public RecogSpec createRecogSpec()
    {
        // parameters:
        //     1. Contacts:
        //       WordList _contactList
        //       String type = "contacts"
        //       String grammarId = "[anyidstring]"
        //     2. server: String server
        //     3. recognitionGrxml
        //     4. interpretationGrxml
        //     5. Server Grammar
        //         id
        // _contactList (provided by application)

        // Create a sample recognition spec. based on the "NVC_ASR_CMD"
        // Optional Settings (set by app):
        // _settings.put("audio_source", "SpeakerAndMicrophone");
        // _settings.put("dictation_language", "spa-XLA");
        // _settings.put("application_name", "SampleApp");
        // _settings.put("field_id", "Recognition_Results");
        // _settings.put("location", "<+29.59300303, -98.37395602> +/- 2252.00m");
        // _settings.put("application_session_id", "1234567890");
        // _settings.put("utterance_number", "1");
        // _settings.put("application_state_id", "45");
        if (_settings == null) {
            _settings = new Data.Dictionary();
        }
        _settings.put("dictation_language", _language);
        _settings.put("dictation_type", _dictationType);
        RecogSpec retRecogSpec = new RecogSpec("NVC_ASR_CMD", _settings, "AUDIO_INFO")
        {
            @Override
            public List<DataParam> getDelayedParams()
            {
                final boolean isConstrainedRecognition = _dictationType.equalsIgnoreCase("oneshot");

                // The creation of the REQUEST_INFO param is delayed to ensure
                // that it references the up-to-date grammar checksums
                Data.Dictionary requestInfo = new Data.Dictionary();
                requestInfo.put("start", 0);
                requestInfo.put("end", 0);
                requestInfo.put("text", "");

                Data.Sequence grammarList = new Data.Sequence();
                Data.Sequence checksumList = new Data.Sequence();

                if (_contactListId != null) {
                    WordList contactList = getWordList(_contactListId);
                    if (contactList == null) {
                        Logger.error(this, "Missing required contact list: " + _contactListId);
                        return null;
                    }

                    if (!isConstrainedRecognition)
                    {
                        Data.Dictionary grammar = new Data.Dictionary();
                        grammar.put("id", _contactListId);
                        grammar.put("type", "contacts");
                        grammarList.add(grammar);
                    }

                    Data.Dictionary checksum = new Data.Dictionary();
                    checksum.put("id", _contactListId);
                    checksum.put("type", "contacts");
                    checksum.put("checksum", "" + contactList.getAcceptedChecksum());
                    checksumList.add(checksum);
                }

                for (Map.Entry<String, GrammarInfo> grammar :_grammarList.entrySet()) {
                    String id = grammar.getKey();
                    GrammarInfo info = grammar.getValue();

                    WordList wordList = getWordList(id);
                    if (wordList == null) {
                        Logger.error(this, "Missing required custom word list: " + id);
                        return null;
                    }

                    if (!isConstrainedRecognition)
                    {
                        Data.Dictionary grammarDict = new Data.Dictionary();
                        grammarDict.put("id", id);
                        grammarDict.put("type", info.type);
                        if (info.category != null) {
                            grammarDict.put("structured_content_category", info.category);
                        }
                        grammarList.add(grammarDict);
                    }

                    Data.Dictionary checksum = new Data.Dictionary();
                    checksum.put("id", id);
                    checksum.put("type", info.type);
                    if (info.category != null) {
                        checksum.put("structured_content_category", info.category);
                    }
                    checksum.put("checksum",  "" + wordList.getAcceptedChecksum());
                    checksumList.add(checksum);
                }

                if (isConstrainedRecognition) {
                    Data.Dictionary grammar = new Data.Dictionary();
                    grammar.put("id", _grammarId);
                    grammar.put("type", "predefined_static_grammars");
                    grammarList.add(grammar);
                }

                if (grammarList.size() > 0) {
                    requestInfo.put("grammar_list", grammarList);
                }
                if (checksumList.size() > 0) {
                    requestInfo.put("checksum_list", checksumList);
                }

                return new ReadOnlyList<DataParam>(new DictionaryParam("REQUEST_INFO", requestInfo));
            }
        };
        return retRecogSpec;
    }

    private void setGrammarList(HashMap<String, GrammarInfo> grammarList)
    {
        _grammarList = (HashMap<String, GrammarInfo>)grammarList;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        o.tryPut(KEY_GRAMMAR_ID, _grammarId);
        o.tryPut(KEY_LANGUAGE, _language);
        o.tryPut(KEY_DICTATION_TYPE, _dictationType);
        if (!_grammarList.isEmpty()) {
            try {
                org.json.JSONArray a = new org.json.JSONArray();
                for (Entry<String, GrammarInfo> e : _grammarList.entrySet())
                {
                    org.json.JSONObject g = new JSONObject();
                    g.put("id", e.getKey());
                    g.put("info", e.getValue().toJSON());
                    a.put(g);
                }
                o.put(KEY_GRAMMARLIST, a);
            } catch (JSONException excep) {}
        }
        if (_contactListId != null) {
            o.tryPut(KEY_CONTACTLIST_ID, _contactListId);
        }
        return o;
    }

    public static NvcAsrSpec createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String grammarId = json.getString(KEY_GRAMMAR_ID);
        String language = json.getString(KEY_LANGUAGE);
        String dictationType = json.getString(KEY_DICTATION_TYPE);
        HashMap<String, GrammarInfo> grammarList = new HashMap<String, GrammarInfo>();
        if (!json.isNull(KEY_GRAMMARLIST)) {
            org.json.JSONArray a = json.getJSONArray(KEY_GRAMMARLIST);
            for (int i = 0; i < a.length(); ++ i) {
                org.json.JSONObject g = a.getJSONObject(i);
                String id = g.getString("id");
                GrammarInfo info = (GrammarInfo)(GrammarInfo.createFromJSON(g.getJSONObject("info")));
                grammarList.put(id, info);
            }
        }
        NvcAsrSpec ret = new NvcAsrSpec(grammarId, language);
        ret.setDictationType(dictationType);
        if (!grammarList.isEmpty()) {
            ret.setGrammarList(grammarList);
        }

        if (!json.isNull(KEY_CONTACTLIST_ID)) {
            String contactListId = json.getString(KEY_CONTACTLIST_ID);
            if (contactListId != null && !contactListId.equals("")) {
                ret.setContactListId(contactListId);
            }
        }
        return ret;
    }
}
