package com.nuance.dragon.toolkit.cloudservices.recognizer;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.cloudservices.DataParam;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * A specification for a cloud recognition. Includes the command name,
 * the command settings, the audio parameter name and the spec for the
 * result structure.
 */
public class RecogSpec
{
    private String _command;
    private Data.Dictionary _settings;
    private String _audioParam;
    private final ArrayList<DataParam> _params;
    private final int _timeoutMs;

    /**
     * Create a recog spec with a default command timeout of 10 seconds.
     * @param command    The recognition command name, which has been deployed on the
     *                   Nuance Mobile Application Server. (cannot be null)
     * @param settings   The command settings (sent in the BCP begin, maybe null)
     * @param audioParam The audio param name. (cannot be null)
     */
    public RecogSpec(String command, Data.Dictionary settings, String audioParam)
    {
        Checker.checkArgForNull("command", command);
        Checker.checkArgForNull("audioParam", audioParam);

        _command = command;
        _settings = settings;
        _audioParam = audioParam;
        _params = new ArrayList<DataParam>();
        _timeoutMs = 10000;
    }

    /**
     * Create a recog spec.
     * @param command    The recognition command name, which has been deployed on the
     *                   Nuance Mobile Application Server. (cannot be null)
     * @param settings   The command settings (sent in the BCP begin, maybe null)
     * @param audioParam The audio param name. (cannot be null)
     * @param timeoutMs  The command timeout, in milliseconds
     */
    public RecogSpec(String command, Data.Dictionary settings, String audioParam, int timeoutMs)
    {
        Checker.checkArgForNull("command", command);
        Checker.checkArgForNull("audioParam", audioParam);
        Checker.checkArgForCondition("timeoutMs", "greater than 0", timeoutMs > 0);

        _command = command;
        _settings = settings;
        _audioParam = audioParam;
        _params = new ArrayList<DataParam>();
        _timeoutMs = timeoutMs;
    }

    /**
     * @return The recognition command name, which has been deployed on the
     *         Nuance Mobile Application Server. (never be null)
     */
    public final String getCommand()
    {
        return _command;
    }

    /**
     * @return The dictionary of command setting values. (maybe null)
     */
    public final Data.Dictionary getSettings()
    {
        return _settings;
    }

    /**
     * @return The command timeout, in ms
     */
    public final int getTimeout()
    {
        return _timeoutMs;
    }

    /**
     * Add a parameter to be sent during the recognition.
     * @param param Arbitrary parameters associated with the recognition specified by
     *              the current {@link RecogSpec}.
     */
    public final void addParam(DataParam param)
    {
        Checker.checkArgForNull("param", param);
        _params.add(param);
    }

    /**
     * Retrieves the parameters that have been added via {@link #addParam}.
     * @return never null, could be empty
     */
    public final List<DataParam> getParams()
    {
        return _params;
    }

    /**
     * @return The name of the audio parameter. (never be null)
     */
    public final String getAudioParam()
    {
        return _audioParam;
    }

    /**
     * Retrieves any parameters that need to be generated after the transaction
     * has started. This differs from {@link #getParams}, where the parameters
     * are generated ahead of time.
     * <p>
     * By default, this method will return an empty list. Override if necessary.
     * @return never null, could be empty
     */
    public List<DataParam> getDelayedParams()
    {
        return new ArrayList<DataParam>(0);
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result
                + ((_audioParam == null) ? 0 : _audioParam.hashCode());
        result = prime * result
                + ((_command == null) ? 0 : _command.hashCode());
        result = prime * result + ((_params == null) ? 0 : _params.hashCode());
        result = prime * result
                + ((_settings == null) ? 0 : _settings.hashCode());
        result = prime * result + _timeoutMs;
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        RecogSpec other = (RecogSpec) obj;
        if (_audioParam == null) {
            if (other._audioParam != null)
                return false;
        } else if (!_audioParam.equals(other._audioParam))
            return false;
        if (_command == null) {
            if (other._command != null)
                return false;
        } else if (!_command.equals(other._command))
            return false;
        if (_params == null) {
            if (other._params != null)
                return false;
        } else if (!_params.equals(other._params))
            return false;
        if (_settings == null) {
            if (other._settings != null)
                return false;
        } else if (!_settings.equals(other._settings))
            return false;
        if (_timeoutMs != other._timeoutMs)
            return false;
        return true;
    }
}
