package com.nuance.dragon.toolkit.cloudservices.util;

import java.util.Enumeration;
import java.util.Map;

import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.Logger;

/**
 * Converts between PDX and DMT dictionaries and sequences. This is separated
 * from the {@link Data} class so that it can be obfuscated away from the public
 * API.
 */
public final class PdxConverter
{
    // ================================ SDK1 ================================ 
    /**
     * Convert a DMT sequence to a PDX sequence.
     * @param sequence
     * @param resource
     * @return The PDX sequence
     */
    public static com.nuance.nmsp.client.sdk.components.core.pdx.Sequence sequenceToPdx(Data.Sequence sequence, com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResource resource)
    {
        com.nuance.nmsp.client.sdk.components.core.pdx.Sequence nmspSeq = resource.createPDXSequence();
        for (Data value : sequence.getValues())
        {
            switch(value.getType())
            {
            case Data.TYPE_DICT:
                nmspSeq.addDictionary(dictionaryToPdx((Data.Dictionary)value, resource));
                break;
            case Data.TYPE_SEQ:
                nmspSeq.addSequence(sequenceToPdx((Data.Sequence)value, resource));
                break;
            case Data.TYPE_INT:
                nmspSeq.addInteger(((Data.Integer) value).value);
                break;
            case Data.TYPE_BOOLEAN:
                nmspSeq.addBoolean(((Data.Boolean) value).value);
                break;
            case Data.TYPE_DOUBLE:
                nmspSeq.addDouble(((Data.Double) value).value);
                break;
            case Data.TYPE_BYTES:
                nmspSeq.addByteString(((Data.Bytes) value).value);
                break;
            case Data.TYPE_UTF8:
                nmspSeq.addUTF8String(((Data.String) value).value);
                break;
            case Data.TYPE_ASCII:
                nmspSeq.addAsciiString(((Data.String) value).value);
                break;
            case Data.TYPE_NULL:
                break;
            }
        }
        return nmspSeq;
    }

    /**
     * Convert a DMT dictionary to a PDX dictionary.
     * @param dictionary
     * @param resource
     * @return The PDX dictionary
     */
    public static com.nuance.nmsp.client.sdk.components.core.pdx.Dictionary dictionaryToPdx(Data.Dictionary dictionary, com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResource resource)
    {
        com.nuance.nmsp.client.sdk.components.core.pdx.Dictionary nmspDict = resource.createPDXDictionary();

        for (Map.Entry<String, Data> entry : dictionary.getEntries())
        {
            String key = entry.getKey();
            Data value = entry.getValue();
            switch(value.getType())
            {
            case Data.TYPE_DICT:

                nmspDict.addDictionary(key, dictionaryToPdx((Data.Dictionary)value, resource));
                break;
            case Data.TYPE_SEQ:
                nmspDict.addSequence(key, sequenceToPdx((Data.Sequence)value, resource));
                break;
            case Data.TYPE_INT:
                nmspDict.addInteger(key, ((Data.Integer) value).value);
                break;
            case Data.TYPE_BOOLEAN:
                nmspDict.addBoolean(key, ((Data.Boolean) value).value);
                break;
            case Data.TYPE_DOUBLE:
                nmspDict.addDouble(key, ((Data.Double) value).value);
                break;
            case Data.TYPE_BYTES:
                nmspDict.addByteString(key, ((Data.Bytes) value).value);
                break;
            case Data.TYPE_UTF8:
                nmspDict.addUTF8String(key, ((Data.String) value).value);
                break;
            case Data.TYPE_ASCII:
                nmspDict.addAsciiString(key, ((Data.String) value).value);
                break;
            case Data.TYPE_NULL:
                break;
            }
        }
        return nmspDict;
    }

    /**
     * Convert a PDX sequence to a DMT sequence.
     * @param pdxSequence
     * @return The DMT sequence
     */
    public static Data.Sequence sequenceFromPdx(com.nuance.nmsp.client.sdk.components.core.pdx.Sequence pdxSequence)
    {
        Data.Sequence ret = new Data.Sequence();
        for (int i = 0; i < pdxSequence.size(); i++)
        {
            addFromPdx(ret, pdxSequence, pdxSequence.getType(i), i);
        }
        return ret;
    }

    /**
     * Convert a PDX dictionary to a DMT dictionary.
     * @param pdxDictionary
     * @return The DMT dictionary
     */
    public static Data.Dictionary dictionaryFromPdx(com.nuance.nmsp.client.sdk.components.core.pdx.Dictionary pdxDictionary)
    {
        Data.Dictionary ret = new Data.Dictionary();
        @SuppressWarnings("unchecked")
        Enumeration<Object> e = pdxDictionary.keys();
        while (e.hasMoreElements())
        {
            String key = (String)e.nextElement();
            short type = pdxDictionary.getType(key);
            addFromPdx(ret, pdxDictionary, type, key);
        }
        return ret;
    }

    private static void addFromPdx(Data.Dictionary toDict,
                                   com.nuance.nmsp.client.sdk.components.core.pdx.Dictionary fromDict,
                                   short fromType,
                                   String key)
    {
        short type = fromType;
        switch (type)
        {
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_DICT:
            toDict.put(key, dictionaryFromPdx(fromDict.getDictionary(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_SEQ:
            toDict.put(key, sequenceFromPdx(fromDict.getSequence(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_INT:
            toDict.put(key, new Data.Integer(fromDict.getInteger(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_BOOLEAN:
            toDict.put(key, new Data.Boolean(fromDict.getBoolean(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_DOUBLE:
            toDict.put(key, new Data.Double(fromDict.getDouble(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_UTF8:
            toDict.put(key, new Data.String(fromDict.getUTF8String(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_ASCII:
            toDict.put(key, new Data.String(fromDict.getAsciiString(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_BYTES:
            toDict.put(key, new Data.Bytes(fromDict.getByteString(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_NULL:
            toDict.put(key, new Data.Null());
            break;
        default:
            Logger.warn(PdxConverter.class, "Data: Unsupported PDX type found in dictionary, skipping");
            break;
        }
    }

    private static void addFromPdx(Data.Sequence toSeq,
                                   com.nuance.nmsp.client.sdk.components.core.pdx.Sequence fromSeq,
                                   short fromType,
                                   int index)
    {
        short type = fromType;
        switch (type)
        {
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_DICT:
            toSeq.add(dictionaryFromPdx(fromSeq.getDictionary(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_SEQ:
            toSeq.add(sequenceFromPdx(fromSeq.getSequence(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_INT:
            toSeq.add(new Data.Integer(fromSeq.getInteger(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_BOOLEAN:
            toSeq.add(new Data.Boolean(fromSeq.getBoolean(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_DOUBLE:
            toSeq.add(new Data.Double(fromSeq.getDouble(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_UTF8:
            toSeq.add(new Data.String(fromSeq.getUTF8String(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_ASCII:
            toSeq.add(new Data.String(fromSeq.getAsciiString(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_BYTES:
            toSeq.add(new Data.Bytes(fromSeq.getByteString(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_NULL:
            toSeq.add(new Data.Null());
            break;
        default:
            Logger.warn(PdxConverter.class, "PdxValue Unsupported PDX type found in sequence, skipping");
            break;
        }
    }
    
    // ================================ SDK2 ================================ 
    /**
     * Convert a DMT sequence to a PDX sequence.
     * @param sequence
     * @param resource
     * @return The PDX sequence
     */
    public static com.nuance.nmsp.client2.sdk.components.core.pdx.Sequence sequenceToPdx(Data.Sequence sequence, com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource resource)
    {
        com.nuance.nmsp.client2.sdk.components.core.pdx.Sequence nmspSeq = resource.createPDXSequence();
        for (Data value : sequence.getValues())
        {
            switch(value.getType())
            {
            case Data.TYPE_DICT:
                nmspSeq.addDictionary(dictionaryToPdx((Data.Dictionary)value, resource));
                break;
            case Data.TYPE_SEQ:
                nmspSeq.addSequence(sequenceToPdx((Data.Sequence)value, resource));
                break;
            case Data.TYPE_INT:
                nmspSeq.addInteger(((Data.Integer) value).value);
                break;
            case Data.TYPE_BOOLEAN:
                nmspSeq.addBoolean(((Data.Boolean) value).value);
                break;
            case Data.TYPE_DOUBLE:
                nmspSeq.addDouble(((Data.Double) value).value);
                break;
            case Data.TYPE_BYTES:
                nmspSeq.addByteString(((Data.Bytes) value).value);
                break;
            case Data.TYPE_UTF8:
                nmspSeq.addUTF8String(((Data.String) value).value);
                break;
            case Data.TYPE_ASCII:
                nmspSeq.addAsciiString(((Data.String) value).value);
                break;
            case Data.TYPE_NULL:
                break;
            }
        }
        return nmspSeq;
    }

    /**
     * Convert a DMT dictionary to a PDX dictionary.
     * @param dictionary
     * @param resource
     * @return The PDX dictionary
     */
    public static com.nuance.nmsp.client2.sdk.components.core.pdx.Dictionary dictionaryToPdx(Data.Dictionary dictionary, com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource resource)
    {
        com.nuance.nmsp.client2.sdk.components.core.pdx.Dictionary nmspDict = resource.createPDXDictionary();

        for (Map.Entry<String, Data> entry : dictionary.getEntries())
        {
            String key = entry.getKey();
            Data value = entry.getValue();
            switch(value.getType())
            {
            case Data.TYPE_DICT:

                nmspDict.addDictionary(key, dictionaryToPdx((Data.Dictionary)value, resource));
                break;
            case Data.TYPE_SEQ:
                nmspDict.addSequence(key, sequenceToPdx((Data.Sequence)value, resource));
                break;
            case Data.TYPE_INT:
                nmspDict.addInteger(key, ((Data.Integer) value).value);
                break;
            case Data.TYPE_BOOLEAN:
                nmspDict.addBoolean(key, ((Data.Boolean) value).value);
                break;
            case Data.TYPE_DOUBLE:
                nmspDict.addDouble(key, ((Data.Double) value).value);
                break;
            case Data.TYPE_BYTES:
                nmspDict.addByteString(key, ((Data.Bytes) value).value);
                break;
            case Data.TYPE_UTF8:
                nmspDict.addUTF8String(key, ((Data.String) value).value);
                break;
            case Data.TYPE_ASCII:
                nmspDict.addAsciiString(key, ((Data.String) value).value);
                break;
            case Data.TYPE_NULL:
                break;
            }
        }
        return nmspDict;
    }

    /**
     * Convert a PDX sequence to a DMT sequence.
     * @param pdxSequence
     * @return The DMT sequence
     */
    public static Data.Sequence sequenceFromPdx(com.nuance.nmsp.client2.sdk.components.core.pdx.Sequence pdxSequence)
    {
        Data.Sequence ret = new Data.Sequence();
        for (int i = 0; i < pdxSequence.size(); i++)
        {
            addFromPdx(ret, pdxSequence, pdxSequence.getType(i), i);
        }
        return ret;
    }

    /**
     * Convert a PDX dictionary to a DMT dictionary.
     * @param pdxDictionary
     * @return The DMT dictionary
     */
    public static Data.Dictionary dictionaryFromPdx(com.nuance.nmsp.client2.sdk.components.core.pdx.Dictionary pdxDictionary)
    {
        Data.Dictionary ret = new Data.Dictionary();
        @SuppressWarnings("unchecked")
        Enumeration<Object> e = pdxDictionary.keys();
        while (e.hasMoreElements())
        {
            String key = (String)e.nextElement();
            short type = pdxDictionary.getType(key);
            addFromPdx(ret, pdxDictionary, type, key);
        }
        return ret;
    }

    private static void addFromPdx(Data.Dictionary toDict,
                                   com.nuance.nmsp.client2.sdk.components.core.pdx.Dictionary fromDict,
                                   short fromType,
                                   String key)
    {
        short type = fromType;
        switch (type)
        {
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_DICT:
            toDict.put(key, dictionaryFromPdx(fromDict.getDictionary(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_SEQ:
            toDict.put(key, sequenceFromPdx(fromDict.getSequence(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_INT:
            toDict.put(key, new Data.Integer(fromDict.getInteger(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_BOOLEAN:
            toDict.put(key, new Data.Boolean(fromDict.getBoolean(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_DOUBLE:
            toDict.put(key, new Data.Double(fromDict.getDouble(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_UTF8:
            toDict.put(key, new Data.String(fromDict.getUTF8String(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_ASCII:
            toDict.put(key, new Data.String(fromDict.getAsciiString(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_BYTES:
            toDict.put(key, new Data.Bytes(fromDict.getByteString(key)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_NULL:
            toDict.put(key, new Data.Null());
            break;
        default:
            Logger.warn(PdxConverter.class, "Data: Unsupported PDX type found in dictionary, skipping");
            break;
        }
    }

    private static void addFromPdx(Data.Sequence toSeq,
                                   com.nuance.nmsp.client2.sdk.components.core.pdx.Sequence fromSeq,
                                   short fromType,
                                   int index)
    {
        short type = fromType;
        switch (type)
        {
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_DICT:
            toSeq.add(dictionaryFromPdx(fromSeq.getDictionary(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_SEQ:
            toSeq.add(sequenceFromPdx(fromSeq.getSequence(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_INT:
            toSeq.add(new Data.Integer(fromSeq.getInteger(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_BOOLEAN:
            toSeq.add(new Data.Boolean(fromSeq.getBoolean(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_DOUBLE:
            toSeq.add(new Data.Double(fromSeq.getDouble(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_UTF8:
            toSeq.add(new Data.String(fromSeq.getUTF8String(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_ASCII:
            toSeq.add(new Data.String(fromSeq.getAsciiString(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_BYTES:
            toSeq.add(new Data.Bytes(fromSeq.getByteString(index)));
            break;
        case com.nuance.nmsp.client2.sdk.components.resource.nmas.NMASResource.PDX_DATA_TYPE_NULL:
            toSeq.add(new Data.Null());
            break;
        default:
            Logger.warn(PdxConverter.class, "PdxValue Unsupported PDX type found in sequence, skipping");
            break;
        }
    }
}
