package com.nuance.dragon.toolkit.cloudservices.vocalizer;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.pipes.SimplePipe;
import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.TTSParam;
import com.nuance.dragon.toolkit.cloudservices.Transaction;
import com.nuance.dragon.toolkit.cloudservices.TransactionError;
import com.nuance.dragon.toolkit.cloudservices.TransactionResult;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * A cloud vocalizer performs a network TTS using a cloud services transaction.
 * A single instance may be used for multiple TTS requests, but only one TTS
 * request may be active at a time.
 */
public class CloudVocalizer
{
    /**
     * Listens for the result of the TTS transaction. Note that this does NOT
     * reflect the state of playing audio.
     */
    public interface Listener
    {
        /**
         * The transaction is available for the server to process.
         * All the necessary information for the transaction to get started on the server side is being transmitted.
         */
        void onProcessingStarted();
        
        /**
         * The transaction completed successfully.
         */
        void onSuccess();

        /**
         * There was an error during the transaction.
         * @param error The error
         */
        void onError(TransactionError error);
    }

    private static final class CloudVocalization
    {
        private Transaction _transaction;
    }

    private final CloudServices _cloud;
    private final NMTHandler _mainThreadHandler;
    private SimplePipe<AudioChunk> _audioPipe;
    private CloudVocalization _currentVocalization;

    /**
     * @param cloudServices The cloud services instance to use
     */
    public CloudVocalizer(CloudServices cloudServices)
    {
        Checker.checkArgForNull("cloudServices", cloudServices);
        _cloud = cloudServices;
        _mainThreadHandler = _cloud.getMainThreadHandler();
    }

    /**
     * Generate TTS. If TTS generation is already in process, this will cancel
     * the old TTS request.
     * @param ttsSpec Describes the TTS to generate
     * @param listener Optional listener for the success / failure of the
     * transaction. May be null.
     * @return The audio source that will provide the generated TTS audio.
     */
    public AudioSource<AudioChunk> generateTts(final TtsSpec ttsSpec, final Listener listener)
    {
        Checker.checkArgForNull("ttsSpec", ttsSpec);
        
        final SimplePipe<AudioChunk> audioPipe = new SimplePipe<AudioChunk>(_mainThreadHandler);
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                // Cancel any transaction in progress
                cancel();
        
                final CloudVocalization vocalization = new CloudVocalization();
                _currentVocalization = vocalization;
        
                vocalization._transaction = new Transaction(ttsSpec.getCommand(), ttsSpec.getSettings(), new Transaction.Listener()
                {
                    @Override
                    public void onTransactionStarted(Transaction t)
                    {
                    }
        
                    @Override
                    public void onTransactionProcessingStarted(Transaction t)
                    {
                        if (listener != null)
                            listener.onProcessingStarted();
                    }
        
                    @Override
                    public void onTransactionResult(Transaction t, TransactionResult result, boolean complete)
                    {
                        if (vocalization == _currentVocalization && result.isFinal())
                            _currentVocalization = null;
        
                        if (listener != null)
                            listener.onSuccess();
                    }
        
                    @Override
                    public void onTransactionError(Transaction t, TransactionError error)
                    {
                        if (vocalization == _currentVocalization)
                            _currentVocalization = null;
        
                        if (listener != null)
                            listener.onError(error);
                    }
        
                    @Override
                    public void onTransactionIdGenerated(String transactionId)
                    {
                        // TODO Auto-generated method stub
                    }
                }, ttsSpec.getTimeout());
        
                TTSParam param = new TTSParam(ttsSpec.getTtsParamName(), ttsSpec.getTtsParamData(), ttsSpec.getAudioType());
                vocalization._transaction.addParam(param);
                vocalization._transaction.finish();
        
                // Hook up an audio pipe to enable us to disconnect in cancel().
                _audioPipe = audioPipe;
                _audioPipe.connectAudioSource(param.getAudioSource());
        
                _cloud.addTransaction(vocalization._transaction, Transaction.DefaultPriorities.TEXT_TO_SPEECH);
            }
        });
        return audioPipe;
    }

    /**
     * Cancel the current TTS request, if there is one.
     */
    public void cancel()
    {
        _mainThreadHandler.postToLooper(new Runnable()
        {
            @Override
            public void run() {
                if (_audioPipe != null)
                {
                    _audioPipe.disconnectAudioSource();
                    _audioPipe = null;
                }
        
                if (_currentVocalization != null)
                {
                    Transaction t = _currentVocalization._transaction;
                    _currentVocalization = null;
                    t.cancel();
                }
            }
        });
    }
}
