package com.nuance.dragon.toolkit.cloudservices.vocalizer;

import java.util.Map.Entry;

import org.json.JSONException;

import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * Defines a TTS request for {@link CloudVocalizer}.
 */
public class TtsSpec implements JSONCompliant
{
    private static final String KEY_COMMAND = "cmd";
    private static final String KEY_SETTINGS = "settings";
    private static final String KEY_TTS_PARAM_NAME = "pname";
    private static final String KEY_TTS_PARAM_DATA = "pdata";
    private static final String KEY_AUDIO_TYPE = "type";
    private static final String KEY_TIMEOUT = "timeout";

    private final String _command;
    private final Data.Dictionary _settings;
    private final String _ttsParamName;
    private final Data.Dictionary _ttsParamData;
    private final AudioType _audioType;
    private final int _timeout;

    /**
     * @param command The name of the TTS command, e.g. "NVC_TTS_COMMAND"
     * @param settings Custom command settings, if any. May be null.
     * @param ttsParamName The name of the TTS parameter.
     * @param ttsParamData The data to send in the TTS parameter.
     * @param audioType The type of the audio.
     */
    public TtsSpec(String command,
                   Data.Dictionary settings,
                   String ttsParamName,
                   Data.Dictionary ttsParamData,
                   AudioType audioType)
    {
        this(command, settings, ttsParamName, ttsParamData, audioType, 10000);
    }

    /**
     * @param command The name of the TTS command, e.g. "NVC_TTS_COMMAND"
     * @param settings Custom command settings, if any. May be null.
     * @param ttsParamName The name of the TTS parameter.
     * @param ttsParamData The data to send in the TTS parameter.
     * @param audioType The type of the audio.
     * @param timeout The timeout value in ms. Default is 10000.
     */
    public TtsSpec(String command,
                   Data.Dictionary settings,
                   String ttsParamName,
                   Data.Dictionary ttsParamData,
                   AudioType audioType,
                   int timeout)
    {
        Checker.checkStringArgForNullOrEmpty("command", command);
        Checker.checkStringArgForNullOrEmpty("ttsParamName", ttsParamName);
        Checker.checkArgForNull("ttsParamData", ttsParamData);
        Checker.checkArgForCondition("ttsParamData", "not empty", (ttsParamData.getEntries().size() > 0));
        Checker.checkArgForNull("audioType", audioType);
        Checker.checkArgForCondition("timeout", "greater than 0", timeout >= 0);

        _command = command;
        _settings = settings;
        _ttsParamName = ttsParamName;
        _ttsParamData = sanitizeTtsParamData(ttsParamData);
        _audioType = audioType;
        _timeout = timeout;
    }

    /**
     * @return The command name.
     */
    public String getCommand()
    {
        return _command;
    }

    /**
     * @return The dictionary of command setting values. May be null.
     */
    public Data.Dictionary getSettings()
    {
        return _settings;
    }

    /**
     * @return The name of the TTS parameter.
     */
    public String getTtsParamName()
    {
        return _ttsParamName;
    }

    /**
     * @return The data of the TTS parameter.
     */
    public Data.Dictionary getTtsParamData()
    {
        return _ttsParamData;
    }

    /**
     * @return The audio type.
     */
    public AudioType getAudioType()
    {
        return _audioType;
    }

    /**
     * @return The timeout value in ms.
     */
    public int getTimeout()
    {
        return _timeout;
    }

    // Only AudioType and _ttsParamData are relevant to declare equality
    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result
                + ((_audioType == null) ? 0 : _audioType.hashCode());
        result = prime * result
                + ((_ttsParamData == null) ? 0 : _ttsParamData.hashCode());
        return result;
    }

    // Only AudioType and _ttsParamData are relevant to declare equality
    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        TtsSpec other = (TtsSpec) obj;
        if (_audioType == null) {
            if (other._audioType != null)
                return false;
        } else if (!_audioType.equals(other._audioType))
            return false;
        if (_ttsParamData == null) {
            if (other._ttsParamData != null)
                return false;
        } else if (!_ttsParamData.equals(other._ttsParamData))
            return false;
        return true;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new  JSONObject();
        o.tryPut(KEY_COMMAND, _command);
        o.tryPut(KEY_SETTINGS, _settings.toJSON());
        o.tryPut(KEY_TTS_PARAM_NAME, _ttsParamName);
        o.tryPut(KEY_TTS_PARAM_DATA, _ttsParamData.toJSON());
        o.tryPut(KEY_AUDIO_TYPE, _audioType.toJSON());
        o.tryPut(KEY_TIMEOUT, _timeout);
        return o;
    }

    public static TtsSpec createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String command = json.getString(KEY_COMMAND);
        String ttsParamName = json.getString(KEY_TTS_PARAM_NAME);
        Data.Dictionary ttsParamData = Data.Dictionary.createFromJSON(json.getJSONObject(KEY_TTS_PARAM_DATA));
        Data.Dictionary settings = Data.Dictionary.createFromJSON(json.getJSONObject(KEY_SETTINGS));
        AudioType type = AudioType.fromJSON(json.getJSONObject(KEY_AUDIO_TYPE));
        int timeout = json.getInt(KEY_TIMEOUT);
        TtsSpec ret = new TtsSpec(command, settings, ttsParamName, ttsParamData, type, timeout);
        return ret;
    }

    // Control sequence support: replace default <ESC> character by custom <ESC> server character.
    // TODO: Kind of ugly, should we have NMT clients (or server) handle this instead?
    private static Data.Dictionary sanitizeTtsParamData(Data.Dictionary ttsParamData)
    {
        for (Entry<String, Data> entry : ttsParamData.getEntries()) {
            if (Data.TYPE_STRING == entry.getValue().getType()) {
                Data.String currentStr = (Data.String) entry.getValue();
                if (currentStr != null) {
                    String newStr = currentStr.value.replace(String.valueOf((char)0x1B), "\\!");
                    ttsParamData.put(entry.getKey(), newStr);
                }
            }
        }
        return ttsParamData;
    }
}
