package com.nuance.dragon.toolkit.core;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Map.Entry;

import org.json.JSONException;

import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.core.data.Data.Dictionary;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * Represents a custom word that may be recognized by a recognizer.
 */
public class Word implements JSONCompliant, Serializable
{
    private static final long serialVersionUID = 1L;

    private static final int INVALID_USER_ID = -1;

    private static final String KEY_CUSTOMFORM = "customform";
    private static final String KEY_USERID = "userid";
    private static final String KEY_SURFACE = "surface";
    private static final String KEY_SPOKEN = "spoken";

    // TODO: expose these?
    private static final int FIELD_MAX = 400;
    private static final int ADDITIONNAL_INFO_MAX = 2000;

    private final String _surfaceForm;
    private final String _spokenForm;
    private final Dictionary _customForm;
    private final int _userid;

    /**
     * @param surfaceForm The written form of the word.
     */
    public Word(String surfaceForm)
    {
        this(INVALID_USER_ID, surfaceForm, null);
    }

    /**
     * @param surfaceForm The written form of the word.
     * @param spokenForm May be null. This is the 'spoken' form of the word,
     * used for pronunciation.
     */
    public Word(String surfaceForm, String spokenForm)
    {
        this(INVALID_USER_ID, surfaceForm, spokenForm);
    }

    /**
     * @param userId Serial UID to identify the word.
     * @param surfaceForm The written form of the word.
     * @param spokenForm May be null. This is the 'spoken' form of the word,
     * used for pronunciation.
     */
    public Word(int userId, String surfaceForm, String spokenForm)
    {
        Checker.checkArgForNull("surfaceForm", surfaceForm);

        _surfaceForm = getSanitizedFieldForm(surfaceForm);
        _spokenForm = getSanitizedFieldForm(spokenForm);
        _customForm = null;
        _userid = userId;
    }

    /**
     * @param main The key/value pair describing the custom word content.
     * @param additional May be null. The key/value pair with additional custom
     * word information.
     */
    public Word(Dictionary main,
                Dictionary additional)
    {
        Checker.checkArgForNull("main", main);

        _surfaceForm = null;
        _spokenForm = null;
        _customForm = getSanitizedCustomForm(main, additional);
        _userid = INVALID_USER_ID;
    }

    /**
     * @param userId Serial UID to identify the word.
     * @param customForm The custom form of the word.
     */
    public Word(int userId, Dictionary customForm)
    {
        _surfaceForm = null;
        _spokenForm = null;
        _customForm = customForm;
        _userid = userId;
    }

    Word(ObjectInputStream inStream) throws IOException
    {
        boolean isCustomForm = inStream.readBoolean();
        _userid = inStream.readInt();
        if (isCustomForm)
        {
            try
            {
                _surfaceForm = null;
                _spokenForm = null;
                _customForm = (Dictionary)inStream.readObject();
            }
            catch (ClassNotFoundException e) {
                throw new IOException();
            }
        }
        else
        {
            _surfaceForm = getSanitizedFieldForm(inStream.readUTF());
            _spokenForm = inStream.readBoolean() ? getSanitizedFieldForm(inStream.readUTF()) : null;
            _customForm = null;
        }
    }

    /**
     * Get the surface form of the word. That is, the way the word appears in
     * written form.
     */
    public String getSurfaceForm()
    {
        return _surfaceForm;
    }

    /**
     * Get the spoken form of the word.  Note that most languages do not require
     * a spokenForm.  In that case it can be null.
     */
    public String getSpokenForm()
    {
        return _spokenForm;
    }

    /**
     * Get the custom word.
     * @return The custom word dictionary if any, else null.
     */
    public Data.Dictionary getCustomForm()
    {
        return _customForm;
    }

    /**
     * Get the custom word digest content.
     * @return The custom word digest content if any, else an empty string.
     */
    public String getCustomFormDigest()
    {
        String digest = "";

        if (_customForm != null)
        {
            digest = _customForm.toString();
        }

        return digest;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        Word other = (Word) obj;
        if (_customForm == null) {
            if (other._customForm != null)
                return false;
        } else if (!_customForm.equals(other._customForm))
            return false;
        if (_spokenForm == null) {
            if (other._spokenForm != null)
                return false;
        } else if (!_spokenForm.equals(other._spokenForm))
            return false;
        if (_surfaceForm == null) {
            if (other._surfaceForm != null)
                return false;
        } else if (!_surfaceForm.equals(other._surfaceForm))
            return false;
        if (_userid != other._userid)
            return false;
        return true;
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result
                + ((_customForm == null || _customForm.getEntries().isEmpty()) ? 0 : _customForm.hashCode());
        result = prime * result
                + ((_spokenForm == null) ? 0 : _spokenForm.hashCode());
        result = prime * result
                + ((_surfaceForm == null) ? 0 : _surfaceForm.hashCode());
        result = prime * result + _userid;
        return result;
    }

    @Override
    public String toString()
    {
        return "Word [_surfaceForm=" + _surfaceForm + ", _spokenForm="
                + _spokenForm + ", _customForm=" + _customForm + ", _userid="
                + _userid + "]";
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        o.tryPut(KEY_CUSTOMFORM, _customForm != null);
        o.tryPut(KEY_USERID, _userid);

        if (_customForm != null)
        {
            o.tryPut(KEY_CUSTOMFORM, _customForm.toJSON());
        }
        else {
            o.tryPut(KEY_SURFACE, _surfaceForm);
            o.tryPut(KEY_SPOKEN, _spokenForm);
        }
        return o;
    }

    public static Word createFromJSON(org.json.JSONObject json) throws JSONException
    {
        boolean isCustomForm = json.getBoolean(KEY_CUSTOMFORM);
        int userId = json.getInt(KEY_USERID);
        if (isCustomForm) {
            return new Word(userId, Dictionary.createFromJSON(json.getJSONObject(KEY_CUSTOMFORM)));
        }
        else {
            return new Word(userId, json.getString(KEY_SURFACE), json.optString(KEY_SPOKEN, null));
        }
    }

    /**
     * Write this object to a output stream.
     * @param outStream
     * @throws IOException
     */
    void writeTo(ObjectOutputStream outStream) throws IOException
    {
        outStream.writeBoolean(_customForm != null);
        outStream.writeInt(_userid);
        if (_customForm != null)
        {
            outStream.writeObject(_customForm);
        }
        else
        {
            outStream.writeUTF(_surfaceForm);
            outStream.writeBoolean(_spokenForm != null);
            if (_spokenForm != null)
                outStream.writeUTF(_spokenForm);
        }
    }

    private static String getSanitizedFieldForm(String field)
    {
        String sanitizedField = null;
        if (field != null) {
            sanitizedField = new String(field);
            if (sanitizedField.length() > FIELD_MAX) {
                Logger.warn(Word.class, "Trimming current field because it is too long: " +  sanitizedField);
                sanitizedField = sanitizedField.substring(0, FIELD_MAX - 1);
            }
        }
        return sanitizedField;
    }

    private static Dictionary getSanitizedCustomForm(Dictionary main, Dictionary additional)
    {
        Dictionary sanitizedDict = new Data.Dictionary();
        if (main != null) {
            for (Entry<String, Data> entry : main.getEntries()) {
                sanitizedDict.put(entry.getKey(), entry.getValue());
            }
        }

        if (additional != null) {
            String serializedAdditionnal = additional.toJSON().toString();
            boolean add = true;
            if (serializedAdditionnal != null && serializedAdditionnal.length() > ADDITIONNAL_INFO_MAX) {
                Logger.warn(Word.class, "Trimming current \"additional_info\" because it is too long.");
                add = false;
            }

            if (add)
                sanitizedDict.put("additional_info", additional);
        }
        return sanitizedDict;
    }

    /**
     * Get the user ID.
     * @return The user ID
     */
    public int getUserId()
    {
        return _userid;
    }
}
