package com.nuance.dragon.toolkit.core;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

import org.pcollections.TreePVector;

import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Specifies an addition or deletion of a word.
 */
public class WordAction implements Serializable
{
    private static final long serialVersionUID = 1L;

    public static final byte ADD     = 0x01;
    public static final byte REMOVE  = 0x02;
    public static final byte UPDATE  = 0x03;

    private final Word _word;
    private final byte _action;

    /**
     * @param word The word
     * @param added true if the word was added; false if the word was removed.
     */
    public WordAction(Word word, boolean added)
    {
        _word = word;

        //to keep back compatibility. Only ADD and REMOVE action.
        if (added) {
            _action = ADD;
        }
        else  {
            _action = REMOVE;
        }
    }

    /**
     * @param word The word
     * @param action The action assigned to word, 'ADD', 'REMOVE' or 'UPDATE'.
     */
    public WordAction(Word word, byte action)
    {
        Checker.checkState(this, action == ADD || action == REMOVE || action == UPDATE, "Invalid action");

        _word = word;
        _action = action;
    }

    WordAction(ObjectInputStream inStream) throws IOException
    {
        _action = inStream.readByte();
        _word = new Word(inStream);
    }

    /**
     * Get the word.
     * @return The word in this action.
     */
    public Word getWord()
    {
        return _word;
    }

    /**
     * Returns true if the word should be added, or false if the word should be
     * removed.
     * @return true if the word should be added, false if it should be removed.
     */
    public boolean isAdded()
    {
        return (_action == ADD);
    }

    /**
     * Returns action code
     * @return an integer of action type.
     */
    public int getAction()
    {
        return _action;
    }

    /**
     * Returns action in String
     * @return a string of action type.
     */
    public String getActionString()
    {
        if (_action == ADD) {
            return "add";
        }
        else if (_action == REMOVE) {
            return "remove";
        }
        else if (_action == UPDATE) {
            return "update";
        }
        else {
            Checker.checkState(this, false, "invalid action");
            return null;
        }
    }

    @Override
    public boolean equals(Object obj)
    {
        WordAction action = (WordAction)obj;
        return (_action == action._action) && _word.equals(action._word);
    }

    @Override
    public String toString()
    {
        return "[Word:" + _word + ", Action:" + _action + "]";
    }

    /**
     * Computes the hash code based on the word and add/remove action. Adding the hashcode
     * of an add action, and a remove action, for the same word, will equal 0.
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = _word.hashCode();

        return (prime * result + _action);
    }

    /**
     * Write this object to a output stream.
     * @param outStream
     * @throws IOException
     */
    void writeTo(ObjectOutputStream outStream) throws IOException
    {
        outStream.writeByte(_action);
        _word.writeTo(outStream);
    }

    /**
     * Write a list of word actions to an output stream.
     * @param actionList
     * @param outStream
     * @throws IOException
     */
    public static void writeListTo(List<WordAction> actionList, ObjectOutputStream outStream) throws IOException
    {
        outStream.writeInt(actionList.size());
        for (WordAction action : actionList)
        {
            action.writeTo(outStream);
        }
    }

    /**
     * Read a list of word actions from an input stream.
     * @param inStream
     * @return never null
     * @throws IOException
     */
    public static List<WordAction> readListFrom(ObjectInputStream inStream) throws IOException
    {
        int size = inStream.readInt();
        ArrayList<WordAction> actions = new ArrayList<WordAction>(size);
        while (size-- > 0)
        {
            actions.add(new WordAction(inStream));
        }
        return actions;
    }

    /**
     * Read a persistent list of word actions from an input stream.
     * @param inStream
     * @return never null
     * @throws IOException
     */
    public static TreePVector<WordAction> readPersistentListFrom(ObjectInputStream inStream) throws IOException
    {
        int size = inStream.readInt();
        TreePVector<WordAction> actions = TreePVector.empty();
        while (size-- > 0)
        {
            actions = actions.plus(new WordAction(inStream));
        }
        return actions;
    }
}
