package com.nuance.dragon.toolkit.core;

import java.util.Iterator;

/**
 * Manages a list of words actions that may be obtained from any source. Also
 * manages caching the words and monitoring for updates to the list. This class
 * would be extended to provide content from the phonebook, media library, etc.
 */
public abstract class WordList
{
    /**
     * Iterates through word actions.
     * <p>
     * Note: it is possible that while iterating through the list, the underlying
     * collection of words will have changed. This iterator reflects the state
     * of the word collection at the time {@link WordList#getFullIterator}
     * or {@link WordList#getModifiedIterator} was called. Likewise,
     * calling {@link #acceptChanges} will accept the actions specified by this
     * iterator, which may be different than the current state of the underlying
     * collection.
     */
    public interface WordIterator extends Iterator<WordAction>
    {
        /**
         * Accept the list of words that was traversed using the {@link #next}
         * method. After accepting, if the entire list was traversed, the list
         * of words returned by subsequent calls to
         * {@link WordList#getModifiedIterator} will only include the remaining
         * words in the list (if the whole list was not traversed) as well as
         * changes to this list.
         * <p>
         * After accepting or discarding, this iterator should must not be
         * accepted or discarded again.
         */
        void acceptChanges();

        /**
         * Discard the list represented by this iterator. It is not strictly
         * required to call this (i.e. you could just forget about the list
         * and never call {@link #acceptChanges}, but calling this method may
         * allow memory to be freed sooner.
         * <p>
         * After accepting or discarding, this iterator should must not be
         * accepted or discarded again.
         */
        void discardChanges();

        /**
         * @return The approximate number of word actions in the iterator.
         * If < 0, the number of word actions is unknown.
         */
        int getCount();

        /**
         * Get the checksum for the FULL list represented by this iterator.
         * Even though the iterator might only contain modifications, the checksum
         * reflects what the checksum of an equivalent full iterator.
         * @return The checksum
         */
        int getChecksum();
    }

    /**
     * Get the full list of word actions. Since this is the full list, all word
     * actions will be additions.
     * <p>
     * This method may block while the word list initializes!
     * @return Iterator to traverse all the word actions.
     */
    public abstract WordIterator getFullIterator();

    /**
     * Get the list of modifications since the last iterator that was accepted
     * via {@link WordIterator#acceptChanges}.
     * <p>
     * This method may block while the word list initializes!
     * @return Iterator to traverse the modified word actions.
     */
    public abstract WordIterator getModifiedIterator();

    /**
     * Returns true if the underlying word list has changed so drastically, that
     * the modifications couldn't be acquired. In this case, {@link #getModifiedIterator}
     * will not be sufficient, so {@link #getFullIterator} must be used.
     * <p>
     * This method may block while the word list initializes!
     * @return true if a full update is required.
     */
    public abstract boolean fullUpdateRequired();

    /**
     * This method may block while the word list initializes!
     * @return The checksum of the last accepted iterator.
     */
    public abstract int getAcceptedChecksum();
}
