package com.nuance.dragon.toolkit.core.calllog;

import java.util.List;

import com.nuance.dragon.toolkit.core.calllog.internal.Calllog;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Call log manager in NMT is to provide similar call logging APIs from NMSP SDK manager;
 * while it also keeps modularity. Without real call log implementation such as cloud services
 * attached, most of the APIs return null or have no effect.
 */
public class CalllogManager
{
    /**
     * Listens for call log data ready returned from NMSP SDK manager
     */
    public interface CalllogDataListener {
        /**
         * A chunk of call log data is available, application should use CalllogSender to send {@link CalllogSender}
         * or save it to local file.
         *
         * @param calllogData  The binary representation of the call logs recorded by the associated object,
         *                         could be NULL if call logging feature is disabled.
         * @param sessionEventIds  The pointer to an array of session event ids (as ASCII strings) packaged inside the
         *                         calllogData, could be NULL, if the buffer is NULL
         * @param isFlushed  To indicated if this output is the result of a flushCalllogData call.
         */
        public void calllogDataGenerated(byte[] calllogData, List<String> sessionEventIds, boolean isFlushed);

        /**
         * A chunk of call log data is available, application should use CalllogSender to send {@link CalllogSender}
         * or save it to local file.
         *
         * @param calllogData  The binary representation of the call logs recorded by the associated object,
         *                         could be NULL if call logging feature is disabled.
         * @param sessionEventIds  The pointer to an array of session event ids (as ASCII strings) packaged inside the
         *                         calllogData, could be NULL, if the buffer is NULL
         * @param isFlushed  To indicated if this output is the result of a flushCalllogData call.
         *
         *  @deprecated The only difference of this callback is it is called from internal worker thread other than
         *              caller thread
         */
        @Deprecated
        public void calllogDataGenerated2(byte[] calllogData, List<String> sessionEventIds, boolean isFlushed);
    }

    private Calllog _calllogImpl = null;

    /**
     * This is the event name the application needs to use to create the root node of the
     * transaction information log tree. Before starting each transaction, the application needs
     * to call logAppEvent function with this name, commit that SessionEventBuilder right away.
     *
     * And then get the ID for that committed call log event. Once the application is ready
     * to start a transaction (create a Transaction for a recognition, etc) The application
     * needs to pass the following key value pair {NMSP_DEFINES_CALLLOG_APP_TRANSACTION_REF_EVENT, ID}
     * as one element in the commandSettings dictionary.
     *
     * {@link com.nuance.dragon.toolkit.cloudservices.Transaction}
     */
    public static final String CALLLOG_APP_TRANSACTION_REF_EVENT = "APP_TRANSACTION_REF";

    /**
     * attach a Calllog implementation to call log manager and the call log starts
     * <p>
     * Note: Internal use only
     *
     * @param calllogImpl implementation of Calllog
     */
    public void attachCalllogImpl(Calllog calllogImpl)
    {
        Checker.checkArgForNull("calllogImpl", calllogImpl);

        _calllogImpl = calllogImpl;
    }

    /**
     * register {@link CalllogDataListener}
     *
     * @param calllogDataListener call log data listener to monitor the available log data
     */
    public void registerCalllogDataListener(CalllogDataListener calllogDataListener)
    {
        if (_calllogImpl != null)
        {
            _calllogImpl.registerCalllogDataListener(calllogDataListener);
        }
    }

    /**
     * check call log is available or not
     *
     * @return true call log is available false otherwise
     */
    public boolean isCalllogActive()
    {
        return (_calllogImpl != null);
    }

    /**
     * Retrieves the root parent Id of the current call log session, the root parent Id is
     * Unique per call log implementation such as cloud services.
     *
     * @return the root parent id as a String
     * @return null if the call logging feature is disabled
     *
     */
    public String getCalllogRootParentId()
    {
        if (_calllogImpl != null)
        {
            return _calllogImpl.getCalllogRootParentId();
        }
        return null;
    }

    /**
     * Processes any pending call logs to the callback {@link CalllogDataListener#calllogDataGenerated(byte[], List, boolean)}
     * This should be used whenever the client wants the logs to be flushed before the chunk
     * size limit has been reached
     */
    public void flushCalllogData()
    {
        if (_calllogImpl != null)
        {
            _calllogImpl.flushCallLogData();
        }
    }

    /**
     * Log an application event by creating a {@link SessionEventBuilder}. This session event
     * is a child of the current SDK root session event. The application can call this method
     * multiple times to retrieve session event builders at different time.
     *
     * For example, logging user behaviors for different recognition transactions.
     *
     * @param appEventName Application event name. Must not be NULL or empty.
     *
     * @param schemaVersion The call log schema version used by the client code. Could be Null
     *           or empty string if there is no schema version for the client to use.
     *
     *
     * @return {@link SessionEventBuilder} Application session event builder
     * @return null if the call logging feature is disabled
     */
    public SessionEventBuilder logAppEvent(String appEventName, String schemaVersion)
    {
        if (_calllogImpl != null)
        {
            return _calllogImpl.logAppEvent(appEventName, schemaVersion);
        }
        return null;
    }

    /**
     * create the CalllogSender object that can be used by application to send
     * log (see {@link CalllogSender}).
     *
     * @param senderListener Listener to monitor the call log submissions
     *
     * @return The CalllogSender object or null if the call logging feature is disabled
     */
    public CalllogSender createCalllogSender(final CalllogSender.SenderListener senderListener) {

        if (_calllogImpl != null)
        {
            return _calllogImpl.createCalllogSender(senderListener);
        }
        return null;
    }
}
