package com.nuance.dragon.toolkit.core.calllog;

/**
 * CalllogSender manages sending the serialized call logs to the server via its own channel.
 * The binary call logs data can be obtained from the listener  {@link SenderListener} that
 * manages the event notifications for the call log transmission over the network to the server.
 * In case of a failure, the remaining call log binary data will be returned to the user via
 * {@link SenderListener#failed(short, byte[])}. The user should manage to send the data at a
 * later time via {@link CalllogSender#send(byte[])}.
 */
public abstract class CalllogSender
{
    /** Indicates there is a network error */
    public static final short CALLLOG_NETWORK_ERROR = 0x02;

    /** Indicates the call log data to be sent is invalid */
    public static final short CALLLOG_INVALID_DATA  = 0x03;

    /** Indicates the connection attempt failed */
    public static final short CALLLOG_CONNECTION_FAILED = 0x04;

    /**
     * The Listener to monitor the call log submissions.
     */
    public static interface SenderListener
    {
        /**
         * The call log data associated with the current listener has been
         * sent to the server successfully. The application can drop the
         * associated binary data.
         */
        void succeeded(byte[] data);

        /**
         * The call log data associated with the current listener failed to
         * be sent to the server. The remaining binary data has been returned
         * and the application. Application should manage to send them again
         * via the method {@link CalllogSender#send(byte[])}.
         *
         * @param errorCode  Error code see {@link CalllogSender#CALLLOG_NETWORK_ERROR}
         * @param data       The remaining call log binary data which failed to be sent to
         *                   the server.
         */
        void failed(short errorCode, byte[] data);
    }

    /**
     * Sends the binary call log data to the server.
     *
     * @param data The binary call log data to be sent.
     */
    public abstract void send(byte[] data);
    
    /**
     * Shuts down the current call log sender and disconnects from gateway if the
     * current call log sender has a live connection with gateway.
     */
    public abstract void shutdown();
}
