package com.nuance.dragon.toolkit.core.data;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import junit.framework.Assert;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * Specifies the structure of Data (not content). When {@link #process} is called,
 * the structure specified here will be traversed, along with a {@link Data}
 * structure. Only the data specified in this data spec will be considered.
 * <p>
 * Data that is not flagged as 'keep' will be discarded, but sub-contents may
 * still be kept. Discarding data like this may have the result of 'flattening'
 * the data structure. Having keep==false only really makes sense for Dictionaries--
 * for everything else, it's the same as just not including them in the data spec
 * in the first place.
 * <p>
 * Example: <br>
 * The following code
 * <p>
 * <pre><code>
 *      DataSpec resultSpec = new DataSpec(Data.TYPE_DICT, null, true);
 *      resultSpec.addDictContent("prompt", Data.TYPE_STRING, "warning");
 *      resultSpec.addDictContent("transcriptions", Data.TYPE_SEQ, "choices")
 *                .setSequenceContent(Data.TYPE_STRING, "text");
 *      resultSpec.addDictContent("confidences", Data.TYPE_SEQ, "choices")
 *                .setSequenceContent(Data.TYPE_INT, "score");
 * </code></pre>
 * transforms
 * <pre><code>
 *     {
 *         "prompt" : "There is too much noise.",
 *         "transcriptions" : [ "This is a test.",
 *                              "This is a text." ],
 *         "confidences": : [ 800, 700 ]
 *     }
 * </code></pre>
 * into
 * <pre><code>
 *     {
 *         "warning" : "There is too much noise.",
 *         "choices" : [ {
 *                          "text" : "This is a test.",
 *                          "score" : 800
 *                       },
 *                       {
 *                          "text" : "This is a text.",
 *                          "score" : 700
 *                       } ]
 *     }
 * </code></pre>
 */
public class DataSpec implements JSONCompliant
{
    private static final String KEY_TYPE = "type";
    private static final String KEY_CONTENT = "content";
    private static final String KEY_ALIAS = "alias";
    private static final String KEY_DATA = "data";

    private final short _type;
    private final boolean _keep;
    private final String _alias;
    private DataSpec _sequenceContent;
    private Map<String, DataSpec> _dictContent;

    /**
     * @param type The data type. Use the values specified in {@link Data}.
     * @param alias The alias for this data. If the parent is a dictionary, this
     * data will be stored in key=alias instead of the dictionary's key. If the
     * parent is not a dictionary, this data will be stored in a NEW dictionary
     * at the key=alias.
     * @param keep Whether or not to keep this piece of data. Generally, if you
     * simply don't care about data, you shouldn't add it at all to the spec.
     * You would use keep==false if you wanted to 'flatten' the data structure--
     * i.e., you may have a dictionary with keep==false, because while you don't
     * want the dictionary itself in the processed result, you do want some of
     * its subcontents.
     */
    public DataSpec(short type, String alias, boolean keep)
    {
        _type = type;
        _alias = alias;
        _keep = keep;

        if (type == Data.TYPE_DICT)
        {
            _dictContent = new HashMap<String, DataSpec>();
        } else if (type == Data.TYPE_SEQ)
        {
            _sequenceContent = new DataSpec(Data.TYPE_NULL, null, false);
        }
    }

    /**
     * Get the type (values specified in {@link Data}).
     * @return The type
     */
    public short getType()
    {
        return _type;
    }

    /**
     * Set the content type for a sequence. Only call this method if the type of
     * this spec is {@link Data#TYPE_SEQ}. Note that this constrains every
     * element of the sequence to the same type.
     * @param type
     * @param alias If non-null, when data is processed, the returned data will
     * use this for the key, rather than the original key.
     * @return The spec of the type that was just added.
     */
    public DataSpec setSequenceContent(short type, String alias)
    {
        DataSpec sequenceContent = new DataSpec(type, alias, true);
        setSequenceContent(sequenceContent);
        return sequenceContent;
    }

    /**
     * Set the content for a sequence. Only call this method if the type of
     * this spec is {@link Data#TYPE_SEQ}. Note that this constrains every
     * element of the sequence to the same type.
     * @param content
     */
    public void setSequenceContent(DataSpec content)
    {
        Assert.assertEquals(_type, Data.TYPE_SEQ);
        _sequenceContent = content;
    }

    /**
     * Set the content type for a key of a dictionary. Only call this method if
     * the type of this spec is {@link Data#TYPE_DICT}.
     * @param key
     * @param type
     * @param alias If non-null, when data is processed, the returned data will
     * use this for the key, rather than the original key.
     * @return The spec of the type that was just added.
     */
    public DataSpec addDictContent(String key, short type, String alias)
    {
        DataSpec piece = new DataSpec(type, alias, true);
        addDictContent(key, piece);
        return piece;
    }

    /**
     * Set the content type for a key of a dictionary. Only call this method if
     * the type of this spec is {@link Data#TYPE_DICT}.
     * <p>
     * This will set the 'keep' flag of the new data to false.
     * @param key
     * @param type
     * @return The spec of the type that was just added.
     */
    public DataSpec addDictContentToDiscard(String key, short type)
    {
        DataSpec piece = new DataSpec(type, null, false);
        addDictContent(key, piece);
        return piece;
    }

    /**
     * Set the content type for a key of a dictionary. Only call this method if
     * the type of this spec is {@link Data#TYPE_DICT}.
     * @param key
     * @param content
     */
    public void addDictContent(String key, DataSpec content)
    {
        Assert.assertEquals(_type, Data.TYPE_DICT);
        _dictContent.put(key, content);
    }

    /**
     * Based on this spec, process the input.
     * @param input
     * @return The processed data. May return null if no matching data is found.
     */
    public Data process(Data input)
    {
        Data processed = processInternal(input);
        if (processed != null && _alias != null)
        {
            Data.Dictionary dict = new Data.Dictionary();
            dict.put(_alias, processed);
            processed = dict;
        }
        return processed;
    }

    private Data processInternal(Data input)
    {
        Data ret = null;
        if (input != null)
        {
            if (input.getType() == _type)
            {
                if (_type == Data.TYPE_DICT)
                {
                    Assert.assertNotNull(_dictContent);

                    Data.Dictionary retDict = null;

                    // If this is flagged as 'keep', we'll return this dictionary.
                    // Otherwise, we will return the first non-null subdata that
                    // we find.
                    if (_keep)
                    {
                        retDict = new Data.Dictionary();
                        ret = retDict;
                    }

                    Data.Dictionary dict = (Data.Dictionary)input;
                    Set<Map.Entry<String, DataSpec>> entries = _dictContent.entrySet();
                    for (Map.Entry<String, DataSpec> entry : entries)
                    {
                        String key = entry.getKey();
                        DataSpec data = entry.getValue();
                        Data val = dict.get(key);
                        if (val != null)
                        {
                            Data processed = data.processInternal(val);
                            if (processed != null)
                            {
                                // If we're keeping the parent dictionary, add
                                // this to it. Otherwise, return this, since it's
                                // the first matching data we've found.
                                if (retDict != null)
                                {
                                    String toKey;
                                    if (data._alias == null)
                                    {
                                        toKey = key;
                                    } else
                                    {
                                        toKey = data._alias;
                                    }

                                    Data existing = retDict.get(toKey);
                                    if (existing == null)
                                    {
                                        retDict.put(toKey, processed);
                                    } else
                                    {
                                        Logger.debug(this, "Attempting to merge into '" + toKey + "'");

                                        // For sequences, merge by combining the
                                        // dictionaries at corresponding indices.
                                        // For dictionaries, just combine them.
                                        if (processed.getType() == Data.TYPE_SEQ &&
                                            existing.getType() == Data.TYPE_SEQ)
                                        {
                                            Data.Sequence toSeq = (Data.Sequence)existing;
                                            Data.Sequence fromSeq = (Data.Sequence)processed;

                                            if (toSeq.size() == fromSeq.size())
                                            {
                                                final int len = toSeq.size();
                                                for (int i = 0; i < len; i++)
                                                {
                                                    Data fromEntry = fromSeq.get(i);
                                                    Data toEntry = toSeq.get(i);

                                                    if (fromEntry.getType() == Data.TYPE_DICT &&
                                                        toEntry.getType() == Data.TYPE_DICT)
                                                    {
                                                        Data.Dictionary fromDict = (Data.Dictionary)fromEntry;
                                                        Data.Dictionary toDict = (Data.Dictionary)toEntry;
                                                        toDict.putAll(fromDict);
                                                    } else
                                                    {
                                                        Logger.error(this, "Unable to merge sequence elements of '" + toKey + "' due to non-dictionary types");
                                                    }
                                                }
                                            }
                                        }
                                        else if (processed.getType() == Data.TYPE_DICT &&
                                                 existing.getType() == Data.TYPE_DICT)
                                        {
                                            Data.Dictionary fromDict = (Data.Dictionary)processed;
                                            Data.Dictionary toDict = (Data.Dictionary)existing;
                                            toDict.putAll(fromDict);
                                        }
                                        else
                                        {
                                            Logger.error(this, "Unable to merge multiple data with key '" + toKey + "' due to different types");
                                        }
                                    }
                                } else
                                {
                                    ret = processed;
                                    break;
                                }
                            }
                        }
                    }
                } else if (_type == Data.TYPE_SEQ)
                {
                    Assert.assertNotNull(_sequenceContent);

                    if (_keep)
                    {
                        Data.Sequence retSeq = new Data.Sequence();

                        Data.Sequence seq = (Data.Sequence)input;
                        List<Data> values = seq.getValues();
                        for (Data d : values)
                        {
                            Data processed = _sequenceContent.processInternal(d);
                            if (processed != null)
                            {
                                if (_sequenceContent._alias == null)
                                {
                                    retSeq.add(processed);
                                } else
                                {
                                    Data.Dictionary dict = new Data.Dictionary();
                                    dict.put(_sequenceContent._alias, processed);
                                    retSeq.add(dict);
                                }
                            }
                        }
                        ret = retSeq;
                    }
                } else
                {
                    if (_keep)
                    {
                        ret = input;
                    }
                }
            }
        }
        return ret;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        o.tryPut(KEY_TYPE, (int)_type);
        if (_alias != null)
            o.tryPut(KEY_ALIAS, _alias);

        if (_type == Data.TYPE_SEQ)
        {
            o.tryPut(KEY_CONTENT, _sequenceContent.toJSON());
        }
        else if (_type == Data.TYPE_DICT)
        {
            JSONObject content = new JSONObject();

            Set<Entry<String, DataSpec>> entries = _dictContent.entrySet();
            for (Entry<String, DataSpec> e : entries)
            {
                JSONObject entry = new JSONObject();

                DataSpec data = e.getValue();
                entry.tryPut(KEY_DATA, data.toJSON());

                content.tryPut(e.getKey(), entry);
            }
            o.tryPut(KEY_CONTENT, content);
        }
        return o;
    }

    @SuppressWarnings("unchecked")
    public static DataSpec createFromJSON(org.json.JSONObject json) throws JSONException
    {
        short type = (short)json.getInt(KEY_TYPE);
        String alias = json.optString(KEY_ALIAS);

        DataSpec ret = new DataSpec(type, alias, true);

        if (type == Data.TYPE_SEQ)
        {
            ret.setSequenceContent(DataSpec.createFromJSON(json.getJSONObject(KEY_CONTENT)));
        }
        else if (type == Data.TYPE_DICT)
        {
            org.json.JSONObject content = json.getJSONObject(KEY_CONTENT);
            Iterator<Object> keys = content.keys();

            while (keys.hasNext())
            {
                String key = keys.next().toString();
                org.json.JSONObject o = content.getJSONObject(key);

                DataSpec spec = DataSpec.createFromJSON(o.getJSONObject(KEY_DATA));
                ret.addDictContent(key, spec);
            }
        }
        return ret;
    }
}
