package com.nuance.dragon.toolkit.edr;

import com.nuance.dragon.toolkit.edr.internal.EdrConfigImpl;

/**
 * Used to set up parameters at initialization time.
 */
public abstract class EdrConfig {

    /**
     * The associated value should be a string representing a 6 letter language code (for example "eng-USA").
     */
    public static final String LANGUAGE_CODE_TO_LOAD = "recognizer.languageCode";

    /**
     * The associated value should be a string representing a system path to
     * an uncompressed version of the datapack to be loaded. The path should not point to the APK asset folder.
     */
    public static final String DATAPACK_FOLDER_PATH = "recognizer.datapackPath";

    /**
     * The associated value should be a string representing a system path to
     * a folder where the library has write access in order to store user profiles.
     * The folder must exists.
     */
    public static final String USER_FOLDER_PATH = "recognizer.profilesPath";

    /**
     * The associated value should be a string representing a system path to a folder
     * where the library has write access in order to create temporary files or folders.
     * The folder must exists.
     */
    public static final String TEMP_FOLDER_PATH = "tempPath";

    /**
     * The associated value should be a string containing either "true" or "false".
     * If set to "true", the recognizer will split the recognition session into multiple results based on the user speech pauses.
     * <p>
     * <b>Important caveat</b>: <br>
     * If the user stops talking while dictating a part of a token, the token output could be incorrect.
     * For example, if a user pauses while dictating a phone number. The phone number format won't be correctly displayed.
     *
     * <p>
     * Can be set at recognition time. Note that values set at recognition time are reseted to the default values once the recognition finishes.
     */
    public static final String UTTERANCE_DETECTION = "recognizer.utteranceDetection";

    /**
     * If "true", the recognizer will automatically set the left context between two utterances when utterance detection is activated. Default value is "true".
     * <p>
     * <b>Important caveat</b>: <br>
     * The same caveats described by {@link #UTTERANCE_DETECTION} applies. This flag doesn't enable utterance cross-formatting but influences the
     * utterance results.
     * The value of this parameter will influence cases like: "My name is Jon {speaker pauses during speech} and I like to play hockey.".
     * If this parameter is set to "true", the word "and" will start will a lower case since the left context is non empty (i.e. left context is set automatically to "My name is Jon").
     * If this parameter is set to "false", the word "And" will start with a capital letter since no left context would have been set (i.e. the recognizer will consider the second utterance as a new phrase).
     * <p>
     * Can be set at recognition time. Note that values set at recognition time are reseted to the default values once the recognition finishes.
     */
    public static final String UTTERANCE_DETECTION_AUTOMATIC_LEFT_CONTEXT = "recognizer.utteranceDetection.automaticLeftContext";

    /**
     * The associated value should be an integer greater or equal to 1.
     * Sets a limit on the number of choices proposed by the recognizer when reporting a recognition result.
     *
     * <p>
     * Can be set at recognition time. Note that values set at recognition time are reseted to the default values once the recognition finishes.
     */
    public static final String MAX_NBEST_LENGTH = "recognizer.maxNBestChoices";

    /**
     * If the string is non empty, it will be used as a left context for the first utterance detected during a recognition session. The default value is empty.
     *
     * <p>
     * It is highly recommended to set this parameter at recognition time. Note that values set at recognition time are reseted to the default values once the recognition finishes.
     */
    public static final String LEFT_CONTEXT = "recognition.leftContext";

    /**
     * Allows setting up a parameter used at initialization time.
     * @param param name of the parameter. See the various static strings of this class.
     * @param value value associated with the parameter. See the various static strings
     * of this class for the expected content.
     */
    public abstract void setParam(String param, String value);

    /**
     * Creates an EdrConfig compatible object.
     * @return an EdrConfig compatible object.
     */
    static public EdrConfig createEdrConfig() {
        return new EdrConfigImpl();
    }
}

