package com.nuance.dragon.toolkit.edr;

import java.util.List;

import android.content.Context;

import com.nuance.dragon.toolkit.edr.internal.EdrDatapackInstallerImpl;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Utility object used for the installation of datapacks.
 */
public abstract class EdrDatapackInstaller {

    /**
     * Interface used to track datapacks installation related requests.
     */
    public interface DatapackInstallerListener {
        /**
         * Called when the datapacks have been installed.
         * @param success true if all the datapacks could be installed. false if unsuccessful.
         */
        public void datapacksInstalled(boolean success);

        /**
         * Called when the datapacks have been removed.
         * @param succcess true if all the datapacks could be removed. false if unsuccessful.
         */
        public void datapacksRemoved(boolean succcess);

        /**
         * Called when datapacks that needs to be installed are available.
         * @param datapackNames the list of datapacks that should be installed. null if install path is invalid or context is invalid.
         */
        public void datapacksInstallNeeded(List<String> datapackNames);

        /**
         * Called when the datapacks from the asset folder has been synchronized with the install path.
         * @param succcess true if all the datapacks could be synced. false if unsuccessful.
         */
        public void datapacksSynced(boolean succcess);
    }

    /* Synchronous calls */
    /**
     * Returns a list of datapack names that are provided within the given context.
     * This method expects to find the APK's datapacks within the "edr/datapacks"
     * folder relative to the "assets" folder.
     * @return a list of available datapack names that can be installed. Returns null if the context is invalid.
     */
    public abstract List<String> getAssetDatapacks();

    /**
     * Returns a list of datapack names that are currently installed at the given install path.
     * @return a list of installed datapack names. Returns null if install path is invalid.
     */
    public abstract List<String> getInstalledDatapacks();

    /* Asynchronous calls */
    /**
     * Posts a request to retrieve a list of datapack names that should be installed since they
     * don't match the version provided by the assets.
     * This method expects to find the APK's datapacks within the "edr/datapacks"
     * folder relative to the "assets" folder.
     * @see DatapackInstallerListener#datapacksInstallNeeded(List)
     */
    public abstract void getDatapacksInstallNeeded();

    /**
     * Posts a request to install the given datapacks to the install folder path from the APK's assets folder.
     * This method expects to find the APK's datapacks within the "edr/datapacks"
     * folder relative to the "assets" folder.
     * This method will first copy the datapack to destination and remove the old version if it exists.
     * If any errors occurs while copying, the original datapack will be kept and the one being copied deleted.
     * <p>
     * Throws an exception if datapackNames is null.
     * @param datapackNames a list of datapack names to install.
     * @see DatapackInstallerListener#datapacksInstalled(boolean success)
     */
    public abstract void installDatapacks(List<String> datapackNames);

    /**
     * Posts a request to remove the given datapacks from the install path.
     * <p>
     * Throws an exception if datapackNames is null.
     * @param datapackNames a list of datapack names to remove.
     * @see DatapackInstallerListener#datapacksRemoved(boolean succcess)
     */
    public abstract void removeInstalledDatapacks(List<String> datapackNames);

    /**
     * Posts a request to synchronize all the datapacks provided from the APK's assets and the install path.
     * This method expects to find the APK's datapacks within the "edr/datapacks"
     * folder relative to the "assets" folder.
     * <p>
     * Rules: <br>
     * Datapacks that are not part of the assets but installed at the install path will be removed.
     * Datapacks that are part of the assets but not installed will be installed.
     * Datapacks that doesn't share the same versions will be replaced.
     * @see DatapackInstallerListener#datapacksSynced(boolean succcess)
     */
    public abstract void syncAssetsAndInstalledDatapacks();

    /**
     * Returns a DatapackInstaller compatible object.
     * @param context a valid Android context.
     * @param installPath an existing folder path were the library can create folders and write files.
     * @param listener a listener used to report datapack installation progress.
     * @return a DatapackInstaller compatible object.
     */
    public static EdrDatapackInstaller createDatapackInstaller(Context context, String installPath, DatapackInstallerListener listener) {
        Checker.checkArgForNull("context", context);
        Checker.checkArgForNull("installPath", installPath);
        Checker.checkArgForNull("listener", listener);

        return new EdrDatapackInstallerImpl(context, installPath, listener);
    }

}
