package com.nuance.dragon.toolkit.edr;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.SpeechDetectionListener;
import com.nuance.dragon.toolkit.edr.internal.EdrRecognizerImpl;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;

/**
 * Main class used to interact with the embedded dictation recognizer.
 */
public abstract class EdrRecognizer implements SpeechDetectionListener {

    /* Listeners */

    /* Initialization and release listeners */

    /**
     * Interface used to report an EdrRecognizer initialization progress.
     */
    public interface InitializeListener {
        /**
         * Called when an EdrRecognizer initialization has finished.
         * @param recognizer the object being initialized.
         * @param success returns true if the initialization has been successful.
         * @see EdrRecognizer#initialize(EdrConfig config, EdrRecognizer.InitializeListener listener)
         */
        void onLoaded(EdrRecognizer recognizer, boolean success);

        /**
         * Called when an EdrRecognizer cannot be initialized when a previous EdrRecognizer instance wasn't released.
         * @param recognizer the object that cannot be initialized.
         * @param usedInstance the instance that hasn't been released.
         * @see EdrRecognizer#initialize(EdrConfig config, EdrRecognizer.InitializeListener listener)
         */
        void onInstanceAlreadyLoaded(EdrRecognizer recognizer, EdrRecognizer usedInstance);
    };

    /**
     * Interface used to report the progress of an EdrRecognizer release.
     */
    public interface ReleaseListener {
        /**
         * Called when EdrRecognizer has been released.
         * @param recognizer the object being released.
         * @see EdrRecognizer#release(ReleaseListener listener)
         */
        void onRelease(EdrRecognizer recognizer);
    };

    /* Recognition related listener */

    /**
     * Interface used to report a recognition result and errors during recognition.
     */
    public interface RecognitionListener {
        /**
         * Called when a recognition result is available.
         * Note that the callback can be called multiple times during a recognition.
         * @param result result object associated with an utterance.
         * @see EdrResult
         */
        void onResult(EdrResult result);

        /**
         * Called when an error occurs during recognition.
         * @param err an error object.
         * @see EdrError
         */
        void onError(EdrError err);
    };

    /* User vocabularies related listener */

    /**
     * Interface used to report a user vocabulary loading progress.
     */
    public interface RebuildListener {
        /**
         * Called when a UserVocabulary object has been successfully loaded or cleared.
         * @param userVocab the loaded user vocabulary. param is null when a user vocabulary has been cleared.
         */
        void onComplete(UserVocabulary userVocab);

        /**
         * Called if an error occurred while loading a UserVocabulary object.
         * @param err an error object.
         */
        void onError(EdrError err);
    };

    /* Wake vs sleep listener */

    /**
     * Interface used to report the awaking state of the recognizer.
     */
    public interface AwakeStateListener {

        /**
         * Called once the recognizer is in the sleep state.
         */
        void onSleeping();

        /**
         * Called once the recognizer is in the wake state.
         */
        void onAwaken();

        /**
         * Called to report an error while processing a sleep or wake request.
         * @param err an error object.
         */
        void onError(EdrError err);
    };

    /**
     * Interface used to report changes to the user local profile.
     */
    public interface ProfileListener {

        /**
         * Called once the user speaker profile has been reset.
         */
        void onSpeakerProfileReset();

        /**
         * Called if there is any errors happening during a profile operation.
         */
        void onError(EdrError err);
    }

    /* Initialization and release requests */

    /**
     * Queues an initialization request.
     * Once the recognizer is initialized, a recognition session can be started.
     * @param config a object containing various parameters used to correctly initialize the recognizer.
     * @param listener an object used to report the initialization progress.
     * @see InitializeListener
     */
    public abstract void initialize(EdrConfig config, InitializeListener listener);

    /**
     * Queues a release request.
     * @param listener an object used to report the release progress.
     * @see ReleaseListener
     */
    public abstract void release(ReleaseListener listener);

    /* Recognition related requests */

    /**
     * Starts a recognition request.
     * <p>
     * <b>Important:</b> <br>
     * It is important to wait for {@link EdrResult#isAtEndOfRecognition()} to be true
     * before calling this method when a previous recognition request has been issued
     * except if a previous recognition has been cancelled.
     * Note that this call will block if there is other tasks that are being processed in the task queue.
     * @param audioSource an audio source. Cannot be null.
     * @param config EdrConfig object. Can be null. {@link EdrConfig} for a list of valid parameters.
     * @param listener an object used to report the recognition progress.
     * @throws UnsupportedAudioFormat
     * @see RecognitionListener
     */
    public abstract void startRecognition(
            AudioSource<AudioChunk> audioSource,
            EdrConfig config,
            RecognitionListener listener);

    /**
     * Stops the current recognition.
     * <p>
     * <b>Important:</b> <br>
     * Note that this call will block if there is other tasks that are being processed in the queue.
     * @see RecognitionListener
     */
    public abstract void stopListening();

    /**
     * Cancels the current recognition.
     * <p>
     * <b>Important:</b> <br>
     * Note that this call will block if there is other tasks that are being processed in the task queue.
     * @see RecognitionListener
     */
    public abstract void cancelRecognition();

    /* User vocabularies related requests */

    /**
     * Queues a load vocabulary request.
     * Allows the user to enrich the factory vocabulary with user added words.
     * Note that the recognizer returns a success even if he refuses some words
     * to be added if the recognizer can't guess a proper pronunciation.
     * <p>
     * <b>Important caveats:</b> <br>
     * The words are added only during the life span of the IEdrRecognizer object. Words added are lost once IEdrRecognizer is released.
     * Also note that depending on the number of out of vocabulary words added. This call can take several seconds/minutes to complete.
     *
     * @param userVocab a user vocabulary object. Can't be null.
     * @param listener an object used to report a user vocabulary loading progress.
     * @see RebuildListener
     */
    public abstract void loadUserVocabulary(UserVocabulary userVocab, RebuildListener listener);

    /**
     * Queues a clear vocabulary request.
     * This call clears all the words associated with the user vocabulary.
     * Any added words that were not part of the factory vocabulary will be removed (making them unrecognizable).
     * @param listener an object used to report a user vocabulary clear progress.
     * @see RebuildListener
     */
    public abstract void clearUserVocabulary(RebuildListener listener);

    /* Wake vs sleep requests */

    /**
     * Queues a sleep request.
     * When EdrRecognizer is sleeping, the process will consume less memory.
     * @param listener an object used to report a sleep progress.
     * @see AwakeStateListener
     */
    public abstract void sleep(AwakeStateListener listener);

    /**
     * Queues a wake request.
     * @param listener an object used to report a wake progress.
     * @see AwakeStateListener
     */
    public abstract void wake(AwakeStateListener listener);

    /* User profile related */

    /**
     * Queues a request to reset the speaker profile.
     * @param listener an object used to report a reset speaker profile progress.
     * @see ProfileListener
     */
    public abstract void resetSpeakerProfile(ProfileListener listener);

    /* Object creations */

    /**
     * Returns an EdrRecognizer compatible object.
     * <p>
     * <b>Important:</b> <br>
     * Make sure to wait for {@link ReleaseListener#onRelease(EdrRecognizer)} before calling this
     * method if you have called {@link EdrRecognizer#release(ReleaseListener)} on a previous
     * EdrRecognizer object instance.
     * @param fileManager for future use. Can be null.
     * @return an EdrRecognizer compatible object.
     */
    public static EdrRecognizer createEdrRecognizer(FileManager fileManager)
    {
        return new EdrRecognizerImpl(fileManager);
    }

    /**
     * Returns an EdrRecognizer compatible object.
     * <p>
     * <b>Important:</b> <br>
     * Make sure to wait for {@link ReleaseListener#onRelease(EdrRecognizer)} before calling this
     * method if you have called {@link EdrRecognizer#release(ReleaseListener)} on a previous
     * EdrRecognizer object instance.
     * @param fileManager for future use. Can be null.
     * @param workerThread a user defined thread from which consuming tasks are run.
     * @return an EdrRecognizer compatible object.
     */
    public static EdrRecognizer createEdrRecognizer(FileManager fileManager, NMTHandler workerThread)
    {
        return new EdrRecognizerImpl(fileManager, workerThread);
    }
}
