package com.nuance.dragon.toolkit.edr.internal;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

import android.content.Context;
import android.content.res.AssetManager;

import com.nuance.dragon.toolkit.edr.EdrDatapackInstaller;
import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.WorkerThread;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

public class EdrDatapackInstallerImpl extends EdrDatapackInstaller {

    private static final String DATAPACKS_ASSETS_RELATIVE_PATH = "edr/datapacks";
    private static final String VERSION_FILE_NAME = "version.txt";
    private static final int MAX_DELETION_TRY_COUNT = 3;

    private final Context _context;
    private AssetManager _assetManager = null;
    private String _installPath = null;
    private final DatapackInstallerListener _listener;

    private final NMTHandler _mainThreadHandler = Factory.createNMTHandler();
    private final WorkerThread _workerThread;
    private final NMTHandler _workerThreadHandler;

    public EdrDatapackInstallerImpl(Context context, String installPath, DatapackInstallerListener listener) {
        _context = context;
        _installPath = installPath;
        _listener = listener;

        if(_installPath != null) {
            _installPath = normalizePath(_installPath);

            File installPathFile = new File(_installPath);
            if(!installPathFile.exists() || !installPathFile.isDirectory() || !installPathFile.canWrite()) {
                Logger.error(this, "Invalid installation folder path for '" + _installPath + "'.");
                _installPath = null;
            }
        }

        if(_context != null) {
            _assetManager = _context.getAssets();

            String[] list = null;
            try {
                list = _assetManager.list(DATAPACKS_ASSETS_RELATIVE_PATH);
                if(list == null || list.length == 0) { // It's a file
                    Logger.error(this, "Expecting assets folder at '" + DATAPACKS_ASSETS_RELATIVE_PATH + "'.");
                    _assetManager = null;
                }
            } catch (IOException e) {
                _assetManager = null;
            }

            if(_assetManager == null)
                Logger.error(this, "The assets folder '" + DATAPACKS_ASSETS_RELATIVE_PATH + "' doesn't exists.");
        }

        _workerThread = Factory.createWorkerThread("com.nuance.dragon.toolkit.edr.internal.EdrDatapackInstallerImpl");
        _workerThread.start();
        _workerThreadHandler = _workerThread.getHandler();
    }

    protected void finalize() {
        _workerThread.stop();
    }

    @Override
    public List<String> getAssetDatapacks() {
        if(_assetManager == null)
            return null;

        String[] assetFolderList = null;
        try {
            assetFolderList = _assetManager.list(DATAPACKS_ASSETS_RELATIVE_PATH);
        } catch (IOException e) {}

        if(assetFolderList == null) {
            Logger.error(this, "Couldn't retrieve a file list from assets folder at path '" + DATAPACKS_ASSETS_RELATIVE_PATH + "'.");
            return null;
        }

        ArrayList<String> assetDatapacks = new ArrayList<String>();
        for(String assetName : assetFolderList) {
            String[] subFolderAssetList = null;
            try {
                subFolderAssetList = _assetManager.list(DATAPACKS_ASSETS_RELATIVE_PATH + "/" + assetName);
            } catch (IOException e) {
                continue;
            }

            if(subFolderAssetList.length > 0) { // Its a directory
                for(String subAssetName : subFolderAssetList) {
                    if(subAssetName.equals(VERSION_FILE_NAME)) {
                        assetDatapacks.add(assetName); // Add only to the list if the version file exists
                        break;
                    }
                }
            }
        }

        return assetDatapacks;
    }

    @Override
    public List<String> getInstalledDatapacks() {
        if(_installPath == null)
            return null;

        ArrayList<String> installedDatapacks = new ArrayList<String>();

        File installFolderFile = new File(_installPath);
        for (File f : installFolderFile.listFiles()) {
            for(File subFile : f.listFiles()) {
                if(subFile.getName().equals(VERSION_FILE_NAME))
                    installedDatapacks.add(f.getName());
            }
        }

        return installedDatapacks;
    }

    @Override
    public void getDatapacksInstallNeeded() {
        _workerThreadHandler.post(getDatapacksInstallNeededRunnable());
    }

    @Override
    public void installDatapacks(List<String> datapackNames) {
        Checker.checkArgForNull("datapackNames", datapackNames);
        _workerThreadHandler.post(getInstallDatapacksRunnable(datapackNames));
    }

    @Override
    public void removeInstalledDatapacks(List<String> datapackNames) {
        Checker.checkArgForNull("datapackNames", datapackNames);
        _workerThreadHandler.post(getRemoveInstalledDatapacksRunnable(datapackNames));
    }

    @Override
    public void syncAssetsAndInstalledDatapacks() {
        _workerThreadHandler.post(getsyncAssetsAndInstalledDatapacksRunnable());
    }

    private List<String> getDatapacksInstallNeededImpl() {
        if(_assetManager == null || _installPath == null)
            return null;

        List<String> assetDatapacks = getAssetDatapacks();
        if(assetDatapacks == null)
            return null;

        ArrayList<String> versionUnmatchedDatapacks = new ArrayList<String>();
        for(String assetName : assetDatapacks) {
            String installedFolderPath = _installPath + "/" + assetName;

            File installDatapackFile = new File(installedFolderPath);
            if(!installDatapackFile.exists())
                continue;

            String installedVersionContent = getInstalledDatapackVersion(installedFolderPath);
            String assetVersionContent = getAssetDatapackVersion(DATAPACKS_ASSETS_RELATIVE_PATH + "/" + assetName);

            if(assetVersionContent == null) {
                Logger.error(this, "Couldn't get the datapack version from '" + VERSION_FILE_NAME + "' for asset '" + assetName + "'. " + "Asset '" + assetName + "' will not be installed.");
                continue;
            }

            if(installedVersionContent == null || !installedVersionContent.equals(assetVersionContent))
                versionUnmatchedDatapacks.add(assetName);
        }

        return versionUnmatchedDatapacks;
    }

    private Runnable getDatapacksInstallNeededRunnable() {
        return new Runnable() {

            @Override
            public void run() {
                final List<String> list = getDatapacksInstallNeededImpl();

                if(_listener != null) {
                    _mainThreadHandler.post(new Runnable() {

                        @Override
                        public void run() {
                            _listener.datapacksInstallNeeded(list);
                        }
                    });
                }

            }
        };
    }

    private boolean installDatapacksImpl(List<String> datapackNames) {
        if(_assetManager == null || _installPath == null || datapackNames == null)
            return false;

        boolean successful = true;

        for(String datapackName : datapackNames) {
            String assetFolderPath = DATAPACKS_ASSETS_RELATIVE_PATH + "/" + datapackName;
            String destinationTempFolderPath = _installPath + "/" + datapackName + "_temp";
            String destinationOldFolderPath = _installPath + "/" + datapackName;

            // Copy the datapack to temporary folder
            File destinationTempFolderFile = new File(destinationTempFolderPath);
            if(destinationTempFolderFile.exists()) {
                if(!deleteFileOrDir(destinationTempFolderFile)) {
                    Logger.error(this, "Couldn't delete folder at path '" + destinationTempFolderFile + "'.");
                    successful = false;
                    continue;
                }
            }

            String[] assetFolderList = null;
            try {
                assetFolderList = _assetManager.list(assetFolderPath);
            } catch (IOException e) {}

            if(assetFolderList != null)
            {
                for(String assetName : assetFolderList) {
                    String assetPath = assetFolderPath + "/" + assetName;

                    if(!copyFileOrDir(_assetManager, assetFolderPath, assetPath, destinationTempFolderPath)) {
                        Logger.error(this, "Couldn't copy assets files from '" + assetFolderPath + "' to '"+ destinationTempFolderPath + "'.");
                        successful = false;
                        break;
                    }
                }
            }
            else
                successful = false;

            if(!successful) {
                deleteFileOrDir(destinationTempFolderFile);
                continue;
            }

            // Delete the old folder if it exists
            File destinationOldFolderFile = new File(destinationOldFolderPath);
            if(destinationOldFolderFile.exists()) {
                boolean deletionSuccess = false;
                if(deleteFileOrDir(destinationOldFolderFile))
                    deletionSuccess = true;

                if(!deletionSuccess) {
                    Logger.error(this, "Couldn't delete folder at path '" + destinationOldFolderPath + "'.");
                    successful = false;
                    continue;
                }
            }

            // Rename the temporary folder
            if(!destinationTempFolderFile.renameTo(destinationOldFolderFile)) {
                Logger.error(this, "Couldn't rename folder at path '" + destinationTempFolderFile.getAbsolutePath() + "' to '" + destinationOldFolderFile.getAbsolutePath() + "'.");
                successful = false;
            }
        }

        return successful;
    }

    private Runnable getInstallDatapacksRunnable(final List<String> datapackNames) {
        return new Runnable() {

            @Override
            public void run() {
                final boolean successful = installDatapacksImpl(datapackNames);

                if(_listener != null) {
                    _mainThreadHandler.post(new Runnable() {

                        @Override
                        public void run() {
                            _listener.datapacksInstalled(successful);
                        }
                    });
                }
            }

        };
    }

    private boolean removeInstalledDatapacksImpl(List<String> datapackNames) {
        if(_installPath == null || datapackNames == null)
            return false;

        boolean successful = true;
        for(String datapackName : datapackNames) {
            File installedFile = new File(_installPath + "/" + datapackName);
            if(!deleteFileOrDir(installedFile)) {
                Logger.error(this, "Couldn't delete folder at path '" + installedFile + "'.");
                successful = false;
            }
        }

        return successful;
    }

    private Runnable getRemoveInstalledDatapacksRunnable(final List<String> datapackNames) {
        return new Runnable() {

            @Override
            public void run() {
                final boolean successful = removeInstalledDatapacksImpl(datapackNames);

                if(_listener != null) {
                    _mainThreadHandler.post(new Runnable() {

                        @Override
                        public void run() {
                            _listener.datapacksRemoved(successful);
                        }
                    });
                }
            }

        };
    }

    private boolean syncAssetsAndInstalledDatapacksImpl() {
        boolean successful = true;

        List<String> assetsDatapacks = getAssetDatapacks();
        if(assetsDatapacks == null)
            return false;

        List<String> installedDatapacks = getInstalledDatapacks();
        if(installedDatapacks == null)
            return false;

        List<String> installNeededDatapacks = getDatapacksInstallNeededImpl();
        if(installNeededDatapacks == null)
            return false;

        ArrayList<String> toBeRemoved = new ArrayList<String>();
        for(String installedDatapack : installedDatapacks) {
            if(assetsDatapacks.contains(installedDatapack))
                continue;
            toBeRemoved.add(installedDatapack);
        }

        ArrayList<String> toBeInstalled = new ArrayList<String>();
        for(String assetsDatapack : assetsDatapacks) {
            if(installedDatapacks.contains(assetsDatapack))
                continue;
            toBeInstalled.add(assetsDatapack);
        }

        for(String installNeeded : installNeededDatapacks)
            toBeInstalled.add(installNeeded); //add those that don't have the same version as the assets folder

        if(!removeInstalledDatapacksImpl(toBeRemoved))
            successful = false;

        if(!installDatapacksImpl(toBeInstalled))
            successful = false;

        return successful;
    }

    private Runnable getsyncAssetsAndInstalledDatapacksRunnable() {
        return new Runnable() {

            @Override
            public void run() {
                final boolean successful = syncAssetsAndInstalledDatapacksImpl();

                if(_listener != null) {
                    _mainThreadHandler.post(new Runnable() {

                        @Override
                        public void run() {
                            _listener.datapacksSynced(successful);

                        }
                    });
                }
            }
        };
    }

    private String getAssetDatapackVersion(String path) {
        if(_assetManager == null)
            return null;

        String version = null;
        InputStream in = null;
        OutputStream out = null;

        try {
            in = _assetManager.open(path + "/" + VERSION_FILE_NAME);
            out = new ByteArrayOutputStream();

            byte[] buffer = new byte[1024];
            int read;
            while ((read = in.read(buffer)) != -1) {
                out.write(buffer, 0, read);
            }

            version = out.toString();

        } catch (IOException e) {
            return version;
        }
        finally {
            try {
                if(out != null)
                    out.close();
                if(in != null)
                    in.close();
            } catch (IOException e) {}
        }

        return version;
    }

    private String getInstalledDatapackVersion(String path) {
        if(_installPath == null)
            return null;

        String version = null;
        InputStream in = null;
        OutputStream out = null;

        try {
            in = new FileInputStream(path + "/" + VERSION_FILE_NAME);
            out = new ByteArrayOutputStream();

            byte[] buffer = new byte[1024];
            int read;
            while ((read = in.read(buffer)) != -1) {
                out.write(buffer, 0, read);
            }

            version = out.toString();

        } catch (FileNotFoundException e) {
            return version;
        } catch (IOException e) {
            return version;
        }
        finally {
            try {
                if(out != null)
                    out.close();
                if(in != null)
                    in.close();
            } catch (IOException e) {}
        }

        return version;
    }

    private static boolean deleteFileOrDir(File file) {
        if (file.isDirectory()) {
            for (File f : file.listFiles()) {
                int i = 1;
                while(i <= MAX_DELETION_TRY_COUNT) {
                    if(deleteFileOrDir(f))
                        break;
                    ++i;
                }

                if(i == MAX_DELETION_TRY_COUNT)
                    return false;
            }
        }

        return file.delete();
    }

    private static String normalizePath(String path) {
        if(path == null)
            return null;

        String newPath = path.trim();
        if(newPath.matches("^/+$"))
            return newPath;

        return newPath.replaceAll("/+$", "");
    }

    // The following functions all deal with file copying from assets to installation directory
    private static boolean copyFileOrDir(AssetManager assetManager, String srcRootFolderAssetPath, String srcAssetPath, String dstRootFolderPath) {
        String assets[] = null;

        // Note: Call below may have a 512 item limit, if we ever don't copy a file, we probably hit that.
        try {
            assets = assetManager.list(srcAssetPath);
        } catch (IOException e) {
            return false;
        }

        String dstItemPath = dstRootFolderPath + srcAssetPath.replaceAll("^" + srcRootFolderAssetPath, "");

        if (assets.length == 0) { // It's a file.
            if(!copyFile(assetManager, srcAssetPath, dstItemPath))
                return false;
        } else {
            String assetPath = normalizePath(srcAssetPath);

            File dir = new File(dstItemPath);
            if (!dir.exists() && !dir.mkdirs()) {
                Logger.error(EdrDatapackInstallerImpl.class, "Couldn't create folder at path '" + dstItemPath + "'.");
                return false;
            }

            for (int i = 0; i < assets.length; ++i) {
                if(!copyFileOrDir(assetManager, srcRootFolderAssetPath, assetPath + "/" + assets[i], dstRootFolderPath))
                    return false;
            }
        }

        return true;
    }

    private static boolean copyFile(AssetManager assetManager, String srcAssetPath, String dstFilePath) {
        boolean ret = true;

        InputStream in = null;
        OutputStream out = null;

        File dstFile = new File(dstFilePath);
        File dstFileParent = dstFile.getParentFile();
        if(dstFileParent != null && !dstFileParent.exists() && !dstFileParent.mkdirs()) {
        	Logger.error(EdrDatapackInstallerImpl.class, "Couldn't create folder at path '" + dstFileParent.getAbsolutePath() + "'.");
        	return false;
        }
		
        try {
            in = assetManager.open(srcAssetPath);
            out = new FileOutputStream(dstFilePath);

            byte[] buffer = new byte[1024];
            int read;
            while ((read = in.read(buffer)) != -1) {
                out.write(buffer, 0, read);
            }
        } catch (FileNotFoundException e) {
            Logger.error(EdrDatapackInstallerImpl.class, "Couldn't open file for writing at path '" + dstFilePath + "'." + " Reason: '" + e.getMessage() + "'");
            ret = false;
        } catch (IOException e) {
            Logger.error(EdrDatapackInstallerImpl.class, "Couldn't copy file from '" + srcAssetPath + "' to '" + dstFilePath + "'." + " Reason: '" + e.getMessage() + "'");
            ret = false;
        }
        finally {
            try {
                if(in != null)
                    in.close();
                if(out != null)
                    out.close();
            } catch (IOException e) {}
        }

        return ret;
    }


}
