package com.nuance.dragon.toolkit.edr.internal;

import java.util.List;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.edr.EdrConfig;
import com.nuance.dragon.toolkit.edr.EdrRecognizer;
import com.nuance.dragon.toolkit.edr.EdrResult;
import com.nuance.dragon.toolkit.edr.UserVocabulary;
import com.nuance.dragon.toolkit.edr.WordSet;
import com.nuance.dragon.toolkit.edr.internal.jni.DictationRecognizer;
import com.nuance.dragon.toolkit.edr.internal.jni.EDRMutableRecognizerEvent;
import com.nuance.dragon.toolkit.edr.internal.jni.EDRRecognizerEvent;
import com.nuance.dragon.toolkit.edr.internal.jni.EDRReturnCode;
import com.nuance.dragon.toolkit.edr.internal.jni.Event;
import com.nuance.dragon.toolkit.edr.internal.jni.EventResults;
import com.nuance.dragon.toolkit.edr.internal.jni.IDictationRecognizerListener;
import com.nuance.dragon.toolkit.edr.internal.jni.MutableInteger;
import com.nuance.dragon.toolkit.edr.internal.jni.StringArray;
import com.nuance.dragon.toolkit.edr.internal.jni.StringMap;
import com.nuance.dragon.toolkit.edr.internal.jni.edr;
import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.WorkerThread;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

public class EdrRecognizerImpl extends EdrRecognizer {

    private static EdrLogger _edrLogger = new EdrLogger();

    public static boolean LOAD_NATIVE_LIBRARY_SUCCESS = false;
    static {
        try {
            System.loadLibrary("mrec");
            System.loadLibrary("textproc");
            System.loadLibrary("server");
            System.loadLibrary("dmt_edr");
            LOAD_NATIVE_LIBRARY_SUCCESS = true;
        }
        catch (UnsatisfiedLinkError e) {
            Logger.error(EdrRecognizerImpl.class, "Failed to load native library.", e);
        }
    }

    private static class RecognitionInstance extends AudioSink<AudioChunk> implements IDictationRecognizerListener {

        final private EdrRecognizerImpl _recognizer;
        final private RecognitionListener _listener;

        public RecognitionInstance(EdrRecognizerImpl recognizer, RecognitionListener listener) {
            super(null);
            _recognizer = recognizer;
            _listener = listener;
        }

        @Override
        public void chunksAvailable(AudioSource<AudioChunk> source) {
            synchronized(_recognizer) {
                if (_recognizer._currentRecognitionInstance != this)
                {
                    Logger.warn(this, "Got audio for old recognition");
                    return;
                }
            }

            List<AudioChunk> chunks = source.getAllAudioChunksForSink(this);
            for (int i = 0; i < chunks.size(); i++) {
                AudioChunk chunk = chunks.get(i);

                short[] audioData = chunk.audioShorts;

                if(audioData != null && audioData.length > 0)
                    _recognizer._workerThreadHandler.post(_recognizer.getPushAudioRunnable(audioData));
            }
        }

        @Override
        public void framesDropped(AudioSource<AudioChunk> source) {
            _recognizer._workerThreadHandler.post(_recognizer.getStopListeningRunnable(false));
        }

        @Override
        public void sourceClosed(AudioSource<AudioChunk> source) {
            _recognizer._workerThreadHandler.post(_recognizer.getStopListeningRunnable(false));
        }

        @Override
        public void eventCallback(Event event) {
            EDRMutableRecognizerEvent type = new EDRMutableRecognizerEvent();
            event.getType(type);

            EDRRecognizerEvent eventType = type.toEDRRecognizerEvent();
            switch(eventType) {
            case EDRRECOGNIZEREVENT_AUDIO_SAMPLE_INFORMATION:
                break;
            case EDRRECOGNIZEREVENT_END_OF_RECOGNITION:
            {
                synchronized(_recognizer) {
                    _recognizer._currentRecognitionInstance = null;
                    EdrResultImpl edrResult = new EdrResultImpl(true);
                    _recognizer._mainThreadHandler.post(_recognizer.getOnResultRunnable(edrResult, _listener));
                }
                break;
            }
            case EDRRECOGNIZEREVENT_RESULTS:
            {
                EventResults resultsEvent = (EventResults)event;

                StringArray resultsArr = new StringArray();
                resultsEvent.getResults(resultsArr);
                EdrResultImpl edrResult = new EdrResultImpl(resultsArr);
                resultsArr.delete();

                _recognizer._mainThreadHandler.post(_recognizer.getOnResultRunnable(edrResult, _listener));
                break;
            }
            default:
                break;
            }
        }

        public void onError(EDRReturnCode code) {
            if(_listener != null && code != EDRReturnCode.EDRRETURNCODE_SUCCESS)
                _recognizer._mainThreadHandler.post(_recognizer.getRecognitionOnError(code, _listener));
        }
    }

    static private Lock _instanceUsingDictationRecognizerlock = new ReentrantLock();
    static private EdrRecognizerImpl _instanceUsingDictationRecognizer = null;

    private final NMTHandler _mainThreadHandler = Factory.createNMTHandler();
    //private final FileManager _fileManager;
    private final WorkerThread _workerThread;
    private final NMTHandler _workerThreadHandler;
    private RecognitionInstance _currentRecognitionInstance = null;
    private DictationRecognizer _dictationRecognizer = null;

    private final Lock _lock = new ReentrantLock();
    private final Condition _requestFinishedCond  = _lock.newCondition();
    private EDRReturnCode _lastEDRRequestReturnCode = EDRReturnCode.EDRRETURNCODE_FAILED;
    private boolean _requestFinished = false;
    private Runnable _waitingRunnable = null;

    public EdrRecognizerImpl(FileManager fileManager) {
        //Checker.checkArgForNull("fileManager", fileManager);

        //_fileManager = fileManager;

        _workerThread = Factory.createWorkerThread("com.nuance.dragon.toolkit.edr.internal.EdrRecognizerImpl");
        _workerThread.start();
        _workerThreadHandler = _workerThread.getHandler();
    }

    public EdrRecognizerImpl(FileManager fileManager, NMTHandler workerThread) {
        //Checker.checkArgForNull("fileManager", fileManager);
        Checker.checkArgForNull("workerThread", workerThread);

        //_fileManager = fileManager;
        _workerThread = null;
        _workerThreadHandler = workerThread;
    }

    @Override
    public void initialize(EdrConfig config, InitializeListener listener) {
        Checker.checkArgForNull("config", config);

        _instanceUsingDictationRecognizerlock.lock();
        EdrRecognizerImpl instanceUsingDictationRecognizer = _instanceUsingDictationRecognizer;
        _instanceUsingDictationRecognizerlock.unlock();

        if(instanceUsingDictationRecognizer != null) {
            // This check avoids a situation where a back to back call to the constructor is made without a proper release/onRelease call sequence.
            // We cannot release _dictationRecognizer at this moment since another thread/instance can still be using it.
            Logger.error(EdrRecognizerImpl.class, "A previous instance of the recognizer hasn't been released.");
            _mainThreadHandler.post(getOnInstanceAlreadyLoadedRunnable(this, instanceUsingDictationRecognizer, listener));
            return;
        }

        _instanceUsingDictationRecognizerlock.lock();
        _instanceUsingDictationRecognizer = this;
        _instanceUsingDictationRecognizerlock.unlock();

        EdrConfigImpl configBase = (EdrConfigImpl)config;
        if(!configBase.isParamDefined("log.level"))
            configBase.setParam("log.level", "5");

        _workerThreadHandler.post(getInitRunnable(config, listener));
    }

    @Override
    public void release(ReleaseListener listener) {
        _workerThreadHandler.post(getReleaseRunnable(listener));

        if(_workerThread != null)
            _workerThread.stop();
    }

    @Override
    public void startRecognition(
            AudioSource<AudioChunk> audioSource,
            EdrConfig params,
            RecognitionListener listener) {
        Checker.checkArgForNull("audioSource", audioSource);

        final RecognitionListener list = listener;

        // Validate audio source format
        AudioType audioType = audioSource.getAudioType();
        if(audioType.encoding != AudioType.Encoding.PCM_16 && audioType.frequency != AudioType.Frequency.FREQ_16KHZ) {
            Logger.error(this, "Supports only PCM 16 kHz with 16 bits per samples format.");
            if(list != null) {
                _mainThreadHandler.post(new Runnable() {

                    @Override
                    public void run() {
                        list.onError(new EdrErrorImpl(EDRReturnCode.EDRRETURNCODE_FAILED));
                    }
                });
            }

            return;
        }

        StringMap recogSettings = null;
        if(params != null) {
            EdrConfigImpl configBase = (EdrConfigImpl)params;
            recogSettings = configBase.getEDRSettingsMap();
        }

        waitEDRRequest(getStartRecognitionRunnable(audioSource, recogSettings, listener));

        synchronized(this) {
            if(_currentRecognitionInstance != null) {
                _currentRecognitionInstance.connectAudioSource(audioSource);
            }
        }
    }

    @Override
    public void stopListening() {
        synchronized(this) {
            if(_currentRecognitionInstance != null)
                _currentRecognitionInstance.disconnectAudioSource();
        }

        waitEDRRequest(getStopListeningRunnable(true));
    }

    @Override
    public void cancelRecognition() {
        synchronized(this) {
            if(_currentRecognitionInstance != null)
                _currentRecognitionInstance.disconnectAudioSource();
        }

        waitEDRRequest(getCancelRecognitionRunnable(true));
    }

    @Override
    public void loadUserVocabulary(UserVocabulary userVocab, RebuildListener listener) {
        Checker.checkArgForNull("userVocab", userVocab);
        _workerThreadHandler.post(getOnloadUserVocabularyRunnable(userVocab, listener));
    }

    @Override
    public void clearUserVocabulary(RebuildListener listener) {
        _workerThreadHandler.post(getClearUserVocabularyRunnable(listener));
    }

    @Override
    public void sleep(AwakeStateListener listener) {
        _workerThreadHandler.post(getSleepRunnable(listener));
    }

    @Override
    public void wake(AwakeStateListener listener) {
        _workerThreadHandler.post(getWakeRunnable(listener));
    }

    @Override
    public void resetSpeakerProfile(ProfileListener listener) {
        _workerThreadHandler.post(getResetSpeakerProfileRunnable(listener));
    }

    @Override
    public void onStartOfSpeech() {

    }

    @Override
    public void onEndOfSpeech() {
        _workerThreadHandler.post(getStopListeningRunnable(false));
    }

    private EDRReturnCode waitEDRRequest(Runnable runnable) {
        EDRReturnCode returnCode = EDRReturnCode.EDRRETURNCODE_FAILED;
        _lock.lock();
        try {
            _requestFinished = false;
            _waitingRunnable = runnable;

            _workerThreadHandler.post(runnable);
            while(!_requestFinished)
                _requestFinishedCond.await();

            returnCode = _lastEDRRequestReturnCode;
            _requestFinished = false;
            _waitingRunnable = null;

        } catch (InterruptedException e) {}
        finally {
            _lock.unlock();
        }

        return returnCode;
    }

    private void signalEDRRequest(EDRReturnCode returnCode, Runnable runnable) {
        _lock.lock();
        try {
            if(_waitingRunnable != runnable) {
                return;
            }

            _requestFinished = true;
            _lastEDRRequestReturnCode = returnCode;
            _requestFinishedCond.signal();
        }
        finally {
            _lock.unlock();
        }
    }

    private Runnable getOnLoadedRunnable(final EdrRecognizer recognizer, final boolean success, final InitializeListener listener) { // Callback that should be run on main thread.
        return new Runnable()
        {
            @Override
            public void run() {
                if(listener != null) {
                    listener.onLoaded(recognizer, success);
                }
            }
        };
    }

    private Runnable getOnInstanceAlreadyLoadedRunnable(final EdrRecognizer recognizer, final EdrRecognizer usedInstance, final InitializeListener listener) { // Callback that should be run on main thread.
        return new Runnable()
        {
            @Override
            public void run() {
                if(listener != null) {
                    listener.onInstanceAlreadyLoaded(recognizer, usedInstance);
                }
            }
        };
    }

    private Runnable getInitRunnable(final EdrConfig config, final InitializeListener listener) {
        final EdrRecognizerImpl recognizer = this;

        return new Runnable()
        {
            @Override
            public void run()
            {
                if(LOAD_NATIVE_LIBRARY_SUCCESS == false) {
                    _mainThreadHandler.post(getOnLoadedRunnable(recognizer, false, listener));
                    return;
                }

                _edrLogger.registerCallback();

                _dictationRecognizer = new DictationRecognizer();

                EdrConfigImpl configBase = (EdrConfigImpl)config;
                EDRReturnCode ret = _dictationRecognizer.init(configBase.getEDRSettingsMap());

                boolean success = false;
                if(ret == EDRReturnCode.EDRRETURNCODE_SUCCESS)
                    success = true;

                if(!success) {
                    _instanceUsingDictationRecognizerlock.lock();
                    _instanceUsingDictationRecognizer = null;
                    _instanceUsingDictationRecognizerlock.unlock();
                }

                if(listener != null) {
                    _mainThreadHandler.post(getOnLoadedRunnable(recognizer, success, listener));
                }
            }
        };
    }

    private Runnable getOnReleaseRunnable(final ReleaseListener listener) { // Callback that should be run on main thread.
        final EdrRecognizerImpl recognizer = this;

        return new Runnable()
        {
            @Override
            public void run()
            {
                if(listener != null)
                    listener.onRelease(recognizer);
            }
        };
    }

    private Runnable getReleaseRunnable(final ReleaseListener listener) {
        return new Runnable()
        {
            @Override
            public void run()
            {
                if(_dictationRecognizer != null) {
                    _dictationRecognizer.finish();
                    _dictationRecognizer.delete();
                    _dictationRecognizer = null;
                }

                synchronized(this) {
                    if(_currentRecognitionInstance != null) {
                        _currentRecognitionInstance = null;
                    }
                }

                _instanceUsingDictationRecognizerlock.lock();
                _instanceUsingDictationRecognizer = null;
                _instanceUsingDictationRecognizerlock.unlock();

                _edrLogger.unregisterCallback();

                if(listener != null)
                    _mainThreadHandler.post(getOnReleaseRunnable(listener));
            }
        };
    }

    private Runnable getRecognitionOnError(final EDRReturnCode code, final RecognitionListener listener) { // Callback that should be run on main thread.
        return new Runnable()
        {
            @Override
            public void run()
            {
                if(listener != null)
                    listener.onError(new EdrErrorImpl(code));
            }
        };
    }

    private synchronized Runnable getStartRecognitionRunnable(
            final AudioSource<AudioChunk> audioSource,
            final StringMap settings,
            final RecognitionListener listener) {

        final EdrRecognizerImpl recognizer = this;

        return new Runnable()
        {
            @Override
            public void run()
            {
                synchronized(recognizer) {
                    _currentRecognitionInstance = new RecognitionInstance(recognizer, listener);
                }

                EDRReturnCode ret = EDRReturnCode.EDRRETURNCODE_FAILED;
                if(settings != null) {
                    ret = _dictationRecognizer.startRecognition(settings, _currentRecognitionInstance);
                    settings.delete();
                }
                else
                    ret = _dictationRecognizer.startRecognition(_currentRecognitionInstance);

                synchronized(recognizer) {
                    if(_currentRecognitionInstance != null && ret != EDRReturnCode.EDRRETURNCODE_SUCCESS) {
                        _currentRecognitionInstance.onError(ret);
                        _currentRecognitionInstance = null;
                    }
                }

                signalEDRRequest(ret, this);
            }
        };
    }

    private Runnable getStopListeningRunnable(final boolean giveFeedback) {
        final EdrRecognizerImpl recognizer = this;

        return new Runnable()
        {
            @Override
            public void run()
            {
                EDRReturnCode ret = _dictationRecognizer.stopRecognition();

                synchronized(recognizer) {
                    if(giveFeedback && _currentRecognitionInstance != null && ret != EDRReturnCode.EDRRETURNCODE_SUCCESS)
                        _currentRecognitionInstance.onError(ret);
                }

                signalEDRRequest(ret, this);
            }
        };
    }

    private Runnable getCancelRecognitionRunnable(final boolean giveFeedback) {
        final EdrRecognizerImpl recognizer = this;

        return new Runnable()
        {
            @Override
            public void run()
            {
                EDRReturnCode ret = _dictationRecognizer.cancelRecognition();

                synchronized(recognizer) {
                    if(giveFeedback && _currentRecognitionInstance != null && ret != EDRReturnCode.EDRRETURNCODE_SUCCESS)
                        _currentRecognitionInstance.onError(ret);

                    _currentRecognitionInstance = null;
                }

                signalEDRRequest(ret, this);
            }
        };
    }

    private Runnable getPushAudioRunnable(final short[] bytes) {
        return new Runnable()
        {
            @Override
            public void run()
            {
                edr.JNIPushShortAudio(_dictationRecognizer, bytes);
            }
        };
    }

    private Runnable getOnResultRunnable(final EdrResult result, final RecognitionListener listener) { // Callback that should be run on main thread.
        return new Runnable()
        {
            @Override
            public void run()
            {
                if(listener != null)
                    listener.onResult(result);
            }
        };
    }

    private Runnable getUserVocabularyOnError(final EDRReturnCode code, final RebuildListener listener) { // Callback that should be run on main thread.
        return new Runnable()
        {
            @Override
            public void run()
            {
                if(listener != null)
                    listener.onError(new EdrErrorImpl(code));
            }
        };
    }

    private Runnable getUserVocabularyOnCompleteRunnable(final UserVocabulary userVocab, final RebuildListener listener) { // Callback that should be run on main thread.
        return new Runnable()
        {
            @Override
            public void run()
            {
                if(listener != null)
                    listener.onComplete(userVocab);
            }
        };
    }

    private Runnable getOnloadUserVocabularyRunnable(final UserVocabulary userVocab, final RebuildListener listener) {
        return new Runnable()
        {
            @Override
            public void run()
            {
                EDRReturnCode ret = unregisterEdrWordSets();
                if(ret != EDRReturnCode.EDRRETURNCODE_SUCCESS) {
                    _mainThreadHandler.post(getUserVocabularyOnError(ret, listener));
                    return;
                }

                // Register the new word sets
                UserVocabularyImpl userVocabImpl = (UserVocabularyImpl)userVocab;
                List<WordSet> wordSetlist = userVocabImpl.getWordSetList();

                StringArray registeredWordSets = new StringArray();
                registeredWordSets.clear();
                for(WordSet ws : wordSetlist) {
                    WordSetBase wsBase = (WordSetBase)ws;
                    com.nuance.dragon.toolkit.edr.internal.jni.WordSet edrWs = wsBase.getEDRWordSet();

                    EDRReturnCode currentRet = _dictationRecognizer.registerWordSet(edrWs);
                    edrWs.delete();

                    if(currentRet != EDRReturnCode.EDRRETURNCODE_SUCCESS)
                        ret = currentRet;

                    registeredWordSets.push(wsBase.getId());
                }

                if(ret != EDRReturnCode.EDRRETURNCODE_SUCCESS) {
                    _mainThreadHandler.post(getUserVocabularyOnError(ret, listener));
                    return;
                }

                // Load the word sets
                ret = _dictationRecognizer.buildWordSets(registeredWordSets);
                registeredWordSets.delete();

                if(ret != EDRReturnCode.EDRRETURNCODE_SUCCESS) {
                    _mainThreadHandler.post(getUserVocabularyOnError(ret, listener));
                    return;
                }

                if(listener != null)
                    _mainThreadHandler.post(getUserVocabularyOnCompleteRunnable(userVocab, listener));
            }
        };
    }

    private Runnable getClearUserVocabularyRunnable(final RebuildListener listener) {
        return new Runnable()
        {
            @Override
            public void run()
            {
                EDRReturnCode ret = unregisterEdrWordSets();
                if(ret != EDRReturnCode.EDRRETURNCODE_SUCCESS) {
                    _mainThreadHandler.post(getUserVocabularyOnError(ret, listener));
                    return;
                }

                ret = _dictationRecognizer.unbuildWordSets();
                if(ret != EDRReturnCode.EDRRETURNCODE_SUCCESS) {
                    _mainThreadHandler.post(getUserVocabularyOnError(ret, listener));
                    return;
                }

                if(listener != null)
                    _mainThreadHandler.post(getUserVocabularyOnCompleteRunnable(null, listener));
            }
        };
    }

    private Runnable getSleepRunnable(final AwakeStateListener listener) {
        return new Runnable()
        {
            @Override
            public void run()
            {
                final EDRReturnCode ret = _dictationRecognizer.sleep();
                if(listener != null) {
                    _mainThreadHandler.post(new Runnable() {

                        @Override
                        public void run() {
                            if(ret != EDRReturnCode.EDRRETURNCODE_SUCCESS)
                                listener.onError(new EdrErrorImpl(ret));
                            else
                                listener.onSleeping();
                        }
                    });
                }
            }
        };
    }

    private Runnable getWakeRunnable(final AwakeStateListener listener) {
        return new Runnable()
        {
            @Override
            public void run()
            {
                final EDRReturnCode ret = _dictationRecognizer.wake();
                if(listener != null) {
                    _mainThreadHandler.post(new Runnable() {

                        @Override
                        public void run() {
                            if(ret != EDRReturnCode.EDRRETURNCODE_SUCCESS)
                                listener.onError(new EdrErrorImpl(ret));
                            else
                                listener.onAwaken();
                        }
                    });
                }
            }
        };
    }

    private EDRReturnCode unregisterEdrWordSets() {
        StringArray registeredWordSets = new StringArray();
        _dictationRecognizer.getRegisteredWordSets(registeredWordSets);

        MutableInteger size = new MutableInteger();
        registeredWordSets.getSize(size);

        EDRReturnCode ret = EDRReturnCode.EDRRETURNCODE_SUCCESS;
        for(int i = 0; i < size.getValue(); ++i) {
            StringBuffer id = new StringBuffer();
            registeredWordSets.getStringAtIndex(i, id);

            EDRReturnCode currentRet = _dictationRecognizer.unregisterWordSet(id.toString());
            if(ret != EDRReturnCode.EDRRETURNCODE_SUCCESS)
                ret = currentRet;
        }

        registeredWordSets.delete();

        return ret;
    }

    private Runnable getResetSpeakerProfileRunnable(final ProfileListener listener) {
        return new Runnable() {

            @Override
            public void run() {
                final EDRReturnCode ret = _dictationRecognizer.resetSpeakerProfile();
                if(listener != null) {
                    _mainThreadHandler.post(new Runnable() {

                        @Override
                        public void run() {
                            if(ret == EDRReturnCode.EDRRETURNCODE_SUCCESS)
                                listener.onSpeakerProfileReset();
                            else
                                listener.onError(new EdrErrorImpl(ret));
                        }
                    });
                }
            }

        };
    }
}
