package com.nuance.dragon.toolkit.elvis;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Represents an Elvis constraint. A constraint defines a single Elvis grammar--
 * i.e., a series of word slot transitions that may be recognized by Elvis.
 */
public class Constraint implements JSONCompliant
{
    private static final String JSON_KEY_ID = "id";
    private static final String JSON_KEY_FROM = "from";
    private static final String JSON_KEY_TO = "to";
    private static final String JSON_KEY_PENALTY = "pen";

    /** Special word slot name that specifies the beginning of a constraint. */
    public static final String START = "_CONSTRAINT_CLASS_START";

    /** Special word slot name that specifies the end of a constraint. */
    public static final String END = "_CONSTRAINT_CLASS_END";

    private boolean _validated;
    private boolean _isValid;
    private int _classCount;
    private int _transitionCount;
    private boolean _hasPenalties;

    /**
     * Describes an Elvis constraint transition, from one word slot to another.
     */
    class Transition
    {
        /** The starting word slot, or null if the transition begins the constraint */
        final String from;
        /** The ending word slot, or null if the transition ends the constraint */
        final String to;
        /** The penalty for the constraint */
        final int penalty;

        private Transition(String from, String to, int penalty)
        {
            this.from = from.length() == 0 ? null : from;
            this.to = to.length() == 0 ? null : to;
            this.penalty = penalty;
        }

        boolean equals(Transition t) {
            if (this.from == null && t.from != null) {
                return false;
            }
            if (!this.from.equals(t.from)) {
                return false;
            }
            if (this.to == null && t.to != null) {
                return false;
            }
            if (!this.to.equals(t.to)) {
                return false;
            }
            return (this.penalty == t.penalty);
        }

        /* (non-Javadoc)
         * @see java.lang.Object#hashCode()
         */
        @Override
        public int hashCode()
        {
            final int prime = 31;
            int result = 1;
            result = prime * result + ((from == null) ? 0 : from.hashCode());
            result = prime * result + penalty;
            result = prime * result + ((to == null) ? 0 : to.hashCode());
            return result;
        }

        /* (non-Javadoc)
         * @see java.lang.Object#equals(java.lang.Object)
         */
        @Override
        public boolean equals(Object obj)
        {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (!(obj instanceof Transition))
                return false;
            Transition other = (Transition) obj;
            if (from == null)
            {
                if (other.from != null)
                    return false;
            } else if (!from.equals(other.from))
                return false;
            if (penalty != other.penalty)
                return false;
            if (to == null)
            {
                if (other.to != null)
                    return false;
            } else if (!to.equals(other.to))
                return false;
            return true;
        }
    }

    private final HashMap<String,Transition> _transitions;
    private final String _id;
    private boolean _alwaysRebuild;

    /**
     * Construct a new Constraint with the given Id.
     *
     * @param id A string that uniquely identifies this constraint.  Must not be null or empty.
     */
    public Constraint(String id)
    {
        Checker.checkStringArgsForAllNullOrEmpty("id", id);
        _transitions = new HashMap<String,Transition>();
        _id = id;
        _alwaysRebuild = false;
        _validated = false;
    }

    /**
     * Create a copy of the constraint.
     * @return A copy, never null.
     */
    public Constraint copy()
    {
        Constraint copy = new Constraint(_id);
        copy._transitions.putAll(_transitions);
        copy._validated = _validated;
        copy._isValid = _isValid;
        copy._classCount = _classCount;
        copy._transitionCount = _transitionCount;
        copy._hasPenalties = _hasPenalties;
        return copy;
    }

    /**
     * Get the ID of this constraint.
     *
     * @return The ID that was given to the constructor.  Will not be null or empty.
     */
    public String getId()
    {
        return _id;
    }

    /**
     * Set the Always Rebuild status of this constraint.
     */
    public void setAlwaysRebuild(boolean alwaysRebuild)
    {
        _alwaysRebuild = alwaysRebuild;
    }

    /**
     * Get the Always Rebuild status of this constraint.
     *
     * @return True if the constraint should always be rebuilt.
     *         False if the constraint should skipped if it already exists in the Elvis grammar.
     */
    public boolean getAlwaysRebuild()
    {
        return _alwaysRebuild;
    }

    /**
     * Add a transition to the constraint. <br>
     *
     * @param fromWordSlot The slot at the start of the transition.
     * Pass {@link #START} to add a root transition.
     * Cannot be null or empty string.  Cannot be (@link #END}.
     * @param toWordSlot The slot at the end of the transition.
     * Pass {@link #END} for a terminal transition.
     * Cannot be null or empty string.  Cannot be (@link #START}.
     * @param penalty The penalty for the transition. A penalty of 0 has no
     * effect. Positive penalties decrease the chance that Elvis will select the
     * transition, and negative penalties will increase the chance.
     */
    public void addTransition(String fromWordSlot, String toWordSlot, int penalty)
    {
        Checker.checkStringArgForNullOrEmpty("fromWordSlot", fromWordSlot);
        Checker.checkStringArgForNullOrEmpty("toWordSlot", toWordSlot);
        Checker.checkArgForCondition("fromWordSlot", "not equal to Constraint.END", !fromWordSlot.equals(Constraint.END));
        Checker.checkArgForCondition("toWordSlot", "not equal to Constraint.START", !toWordSlot.equals(Constraint.START));

        _transitions.put(fromWordSlot + ":" + toWordSlot, new Transition(fromWordSlot, toWordSlot, penalty));
        _validated = false;
    }

    @Override
    public JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        try
        {
            o.put(JSON_KEY_ID, _id);

            JSONArray from = new JSONArray();
            JSONArray to = new JSONArray();
            JSONArray penalty = new JSONArray();
            for (Transition t : _transitions.values())
            {
                from.put(t.from == null ? "" : t.from);
                to.put(t.to == null ? "" : t.to);
                penalty.put(t.penalty);
            }
            o.put(JSON_KEY_FROM, from);
            o.put(JSON_KEY_TO, to);
            o.put(JSON_KEY_PENALTY, penalty);
        }
        catch (JSONException e)
        {
            Logger.error(this, "Exception converting to JSON", e);
        }
        return o;
    }

    /**
     * Constructs a Constraint object from a JSON object.  Throws JSONException if a conversion error occurs.
     *
     * @return A constraint object.  Will never return null.
     */
    public static Constraint createFromJSON(JSONObject json) throws JSONException
    {
        String id = json.getString(JSON_KEY_ID);
        if (id == null || id.equals("")) {
            throw new JSONException("id not defined in JSON object.");
        }
        Constraint ret = new Constraint(id);

        JSONArray from = json.getJSONArray(JSON_KEY_FROM);
        JSONArray to = json.getJSONArray(JSON_KEY_TO);
        JSONArray penalty = json.getJSONArray(JSON_KEY_PENALTY);

        int size = from.length();
        if (size != from.length() || size != penalty.length())
        {
            throw new JSONException("The 'from', 'to', and 'penalty' arrays must have the same length.");
        }

        for (int i = 0; i < size; i++)
        {
            ret.addTransition(from.getString(i), to.getString(i), penalty.getInt(i));
        }
        return ret;
    }

    /**
     * Get the transitions (internal method).
     * @return The transitions. Will never be null.
     */
    List<Transition> getTransitions()
    {
        ArrayList<Transition> transitions = new ArrayList<Transition>(_transitions.size());
        transitions.addAll(_transitions.values());
        return transitions;
    }

    /**
     * Get the number of word classes used by this Constraint (internal method).
     * @return The word class count.
     */
    boolean isValid() {
        if (!_validated) {
            validate();
        }
        return _isValid;
    }

    /**
     * Get the number of word classes used by this Constraint (internal method).
     * @return The word class count.
     */
    int getTransitionCount() {
        if (!_validated) {
            validate();
        }
        return _transitionCount;
    }

    /**
     * Get the number of word classes used by this Constraint (internal method).
     * @return The word class count.
     */
    int getClassCount() {
        if (!_validated) {
            validate();
        }
        return _classCount;
    }

    /**
     * Determine whether any transitions in this Constraint have penalties (internal method).
     * @return true if there are penalties, or false otherwise.
     */
    boolean hasPenalties() {
        if (!_validated) {
            validate();
        }
        return _hasPenalties;
    }

    /**
     * Performs some basic validity checking for a constraint. (internal method).
     */
    private void validate() {

        int transitionCount = _transitions.size();
        boolean hasPenalties = false;
        boolean hasStart = false;
        boolean hasEnd = false;
        ArrayList<String> classes = new ArrayList<String>();
        int classCount;
        boolean isValid;

        for (Transition t : _transitions.values())
        {
            if (t.penalty != 0) {
                hasPenalties = true;
            }
            if (t.from.equals(START)) {
                hasStart = true;
            } else if (!classes.contains(t.from)) {
                classes.add(t.from);
            }
            if (t.to.equals(END)) {
                hasEnd = true;
            } else if (!classes.contains(t.to)) {
                classes.add(t.to);
            }
        }
        classCount = classes.size();
        classes.clear();
        classes = null;

        isValid = true;
        if (!hasStart) {
            Logger.error(this, "Constraint is missing transition from Constraint.START. [" + _id + "]");
            isValid = false;
        }
        if (!hasEnd) {
            Logger.error(this, "Constraint is missing transition to Constraint.END. [" + _id + "]");
            isValid = false;
        }
        if (transitionCount < 2) {
            Logger.error(this, "Constraint must contain at least 2 transitions. [" + _id + "]");
            isValid = false;
        }
        if (classCount < 1) {
            Logger.error(this, "Constraint must contain at least 1 word slot. [" + _id + "]");
            isValid = false;
        }

        _transitionCount = transitionCount;
        _classCount = classCount;
        _hasPenalties = hasPenalties;
        _isValid = isValid;
        _validated = true;
    }

    /* (non-Javadoc)
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + (_alwaysRebuild ? 1231 : 1237);
        result = prime * result + _classCount;
        result = prime * result + (_hasPenalties ? 1231 : 1237);
        result = prime * result + ((_id == null) ? 0 : _id.hashCode());
        result = prime * result + (_isValid ? 1231 : 1237);
        result = prime * result + _transitionCount;
        result = prime * result + ((_transitions == null) ? 0 : _transitions.hashCode());
        result = prime * result + (_validated ? 1231 : 1237);
        return result;
    }

    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (!(obj instanceof Constraint))
            return false;
        Constraint other = (Constraint) obj;
        if (_alwaysRebuild != other._alwaysRebuild)
            return false;
        if (_classCount != other._classCount)
            return false;
        if (_hasPenalties != other._hasPenalties)
            return false;
        if (_id == null)
        {
            if (other._id != null)
                return false;
        } else if (!_id.equals(other._id))
            return false;
        if (_isValid != other._isValid)
            return false;
        if (_transitionCount != other._transitionCount)
            return false;
        if (_transitions == null)
        {
            if (other._transitions != null)
                return false;
        } else if (!_transitions.equals(other._transitions))
            return false;
        if (_validated != other._validated)
            return false;
        return true;
    }
}
