package com.nuance.dragon.toolkit.elvis;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * Represents an Elvis configuration, which consists of one of the following:
 * <br> - A bin file
 * <br> - A language, a quality and a sampling rate
 * TODO Need more explanation
 */
public final class ElvisConfig implements JSONCompliant
{
    private static final String KEY_LANGUAGE = "language";
    private static final String KEY_QUALITY = "quality";
    private static final String KEY_SAMPLING_RATE = "sampling_rate";
    private static final String KEY_BINFILE = "binfile";

    final String language;
    final String quality;
    final String binfile;
    final int samplingRate;

    /**
     * @param binfile The bin file name. Must be a valid file name.
     */
    public ElvisConfig(String binfile)
    {
        Checker.checkStringArgForNullOrEmpty("binfile", binfile);
        language = null;
        quality = null;
        samplingRate = 0;
        this.binfile = binfile;
    }

    /**
     * @param samplingRate The sampling rate, e.g. 8000 or 16000.
     */
    public ElvisConfig(int samplingRate)
    {
        this(LangSpecs.UNSPECIFIED, ElvisQuality.UNSPECIFIED, samplingRate);
    }

    /**
     * @param language The language string, from {@link ElvisRecognizer.Languages}.
     * @param samplingRate The sampling rate, e.g. 8000 or 16000.
     */
    public ElvisConfig(ElvisLanguage language, int samplingRate)
    {
        this(language, ElvisQuality.UNSPECIFIED, samplingRate);
    }

    /**
     * @param language The language string, from {@link ElvisRecognizer.Languages}.
     * @param quality The quality string, from {@link ElvisQuality}.
     * @param samplingRate The sampling rate, e.g. 8000 or 16000.
     */
    public ElvisConfig(ElvisLanguage language, String quality, int samplingRate)
    {
        Checker.checkArgForNull("language", language);
        Checker.checkArgForNull("quality",  quality);
        this.language = language.name;
        this.quality = quality;
        this.samplingRate = samplingRate;
        binfile = null;
    }

    /**
     * @param language The language, from {@link com.nuance.dragon.toolkit.language.Languages}.
     * @param samplingRate The sampling rate, e.g. 8000 or 16000.
     */
    public ElvisConfig(com.nuance.dragon.toolkit.language.Language language, int samplingRate)
    {
        this(language, ElvisQuality.UNSPECIFIED, samplingRate);
    }

    /**
     * @param language The language, from {@link com.nuance.dragon.toolkit.language.Languages}.
     * @param quality The quality string, from {@link ElvisQuality}.
     * @param samplingRate The sampling rate, e.g. 8000 or 16000.
     */
    public ElvisConfig(com.nuance.dragon.toolkit.language.Language language, String quality,  int samplingRate)
    {
        Checker.checkArgForNull("language", language);
        Checker.checkArgForNull("quality", quality);
        ElvisLanguage elvisLanguage = language.getElvisLanguage();
        this.language = elvisLanguage.name;
        this.quality = quality;
        this.samplingRate = samplingRate;
        binfile = null;
    }

    private ElvisConfig(String language, String quality, int samplingRate)
    {
        this.language = language;
        this.samplingRate = samplingRate;
        this.quality = quality;
        this.binfile = null;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        if (language != null)
        {
            o.tryPut(KEY_LANGUAGE, language);
            o.tryPut(KEY_QUALITY, quality);
            o.tryPut(KEY_SAMPLING_RATE, samplingRate);
        } else
        {
            o.tryPut(KEY_BINFILE, binfile);
        }
        return o;
    }

    public static ElvisConfig createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String binfile = json.optString(KEY_BINFILE, null);
        if (binfile != null)
        {
            return new ElvisConfig(binfile);
        }
        return new ElvisConfig(json.getString(KEY_LANGUAGE), json.getString(KEY_QUALITY), json.getInt(KEY_SAMPLING_RATE));
    }
}
