package com.nuance.dragon.toolkit.elvis;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.util.internal.ObjectUtils;
import com.nuance.dragon.toolkit.util.internal.ReadOnlyList;

/**
 * An Elvis N-best list.
 * <p>
 * The N-best list contains an order list of possible Elvis results.
 */
public class ElvisNbestList extends ReadOnlyList<ElvisNbestList.Entry> implements JSONCompliant
{
    private static final String GENERIC_SPEECH_TEXT = "[?]";
    /**
     * A possible Elvis result.
     */
    public static class Entry implements JSONCompliant
    {
        /**
         * A 'word' in this Nbest entry.
         */
        public static class Word implements JSONCompliant
        {
            private static final int INVALID_PHRASE_ID = -1;

            private static final String KEY_WORD_SLOT = "slot";
            private static final String KEY_SURFACE_FORM = "surface";
            private static final String KEY_SPOKEN_FORM = "spoken";
            private static final String KEY_FULL_PHRASE = "phrase";
            private static final String KEY_WORD_IDS = "wordids";
            private static final String KEY_GENERIC = "generic";
            private static final String KEY_PHRASE_ID = "phraseid";

            /** The name of the word slot that this word belongs to */
            public final String wordSlot;
            /** The surface form of the word */
            public final String surfaceForm;
            /** The spoken form of the word, or null if there is no custom spoken form */
            public final String spokenForm;
            /** The full phrase text */
            public final String fullPhrase;
            /** The Elvis word ID(s) */
            public final int[] wordIds;
            /** True if this word is generic (detected, but non-recognized) speech. */
            public final boolean isGenericSpeech;
            /** The user-specified number */
            public final int phraseId;


            Word(String wordSlot,
                    String surfaceForm,
                    String spokenForm,
                    String fullPhrase,
                    int [] wordIds,
                    boolean isGenericSpeech)
            {
                this(wordSlot, surfaceForm, spokenForm, fullPhrase, wordIds, INVALID_PHRASE_ID, isGenericSpeech);
            }

            Word(String wordSlot,
                 String surfaceForm,
                 String spokenForm,
                 String fullPhrase,
                 int [] wordIds,
                 int phraseId,
                 boolean isGenericSpeech)
            {
                this.wordSlot = wordSlot;
                this.surfaceForm = surfaceForm;
                this.spokenForm = (spokenForm == null || spokenForm.length() == 0) ? null : spokenForm;
                this.fullPhrase = fullPhrase;
                this.wordIds = wordIds;
                this.phraseId = phraseId;
                this.isGenericSpeech = isGenericSpeech;
            }

            @Override
            public org.json.JSONObject toJSON()
            {
                JSONObject o = new JSONObject();

                o.tryPut(KEY_WORD_SLOT, wordSlot);
                o.tryPut(KEY_SURFACE_FORM, surfaceForm);
                o.tryPut(KEY_SPOKEN_FORM, spokenForm);
                o.tryPut(KEY_FULL_PHRASE, fullPhrase);
                o.tryPut(KEY_GENERIC, isGenericSpeech);
                o.tryPut(KEY_PHRASE_ID, phraseId);

                if (wordIds.length > 0)
                {
                    org.json.JSONArray wordIds = new org.json.JSONArray();
                    for (int id : this.wordIds)
                    {
                        wordIds.put(id);
                    }
                    o.tryPut(KEY_WORD_IDS, wordIds);
                }
                return o;
            }

            public static Word createFromJSON(org.json.JSONObject json) throws JSONException
            {
                String wordSlot = json.getString(KEY_WORD_SLOT);
                String surfaceForm = json.getString(KEY_SURFACE_FORM);
                String spokenForm = json.optString(KEY_SPOKEN_FORM, null);
                String fullPhrase = json.getString(KEY_FULL_PHRASE);
                boolean isGeneric = json.getBoolean(KEY_GENERIC);
                int phraseId = json.getInt(KEY_PHRASE_ID);
                int [] ids;

                org.json.JSONArray wordIds = json.optJSONArray(KEY_WORD_IDS);
                if (wordIds != null)
                {
                    ids = new int[wordIds.length()];
                    for (int i = 0; i < ids.length; i++)
                    {
                        ids[i] = wordIds.getInt(i);
                    }
                } else
                {
                    ids = new int[0];
                }
                return new Word(wordSlot, surfaceForm, spokenForm, fullPhrase, ids, phraseId, isGeneric);
            }

            @Override
            public int hashCode() {
                final int prime = 31;
                int result = 1;
                result = prime * result
                        + ((fullPhrase == null) ? 0 : fullPhrase.hashCode());
                result = prime * result + (isGenericSpeech ? 1231 : 1237);
                result = prime * result
                        + ((spokenForm == null) ? 0 : spokenForm.hashCode());
                result = prime * result
                        + ((surfaceForm == null) ? 0 : surfaceForm.hashCode());
                result = prime * result + Arrays.hashCode(wordIds);
                result = prime * result
                        + ((wordSlot == null) ? 0 : wordSlot.hashCode());
                result = prime *result + phraseId;
                return result;
            }

            @Override
            public boolean equals(Object o)
            {
                if (o instanceof Word)
                {
                    Word w = (Word) o;
                    if (!ObjectUtils.compareObjects(wordSlot, w.wordSlot))
                        return false;
                    // Note: If both words are generic speech, they match even
                    // if the other fields (surface form, wordIds, etc) do not match.
                    if (isGenericSpeech && w.isGenericSpeech)
                        return true;
                    if (isGenericSpeech != w.isGenericSpeech)
                        return false;
                    if (phraseId != w.phraseId)
                        return false;
                    if (!ObjectUtils.compareObjects(surfaceForm, w.surfaceForm))
                        return false;
                    if (!ObjectUtils.compareObjects(spokenForm, w.spokenForm))
                        return false;
                    if (!ObjectUtils.compareObjects(fullPhrase, w.fullPhrase))
                        return false;
                    if (!Arrays.equals(wordIds,w.wordIds))
                        return false;
                    return true;
                }
                return false;
            }

            @Override
            public String toString()
            {
                if (this.isGenericSpeech) {
                    return GENERIC_SPEECH_TEXT;
                }
                return this.fullPhrase;
            }
        }

        private static final String KEY_WORDS = "words";
        private static final String KEY_CONSTRAINT = "constraint";
        private static final String KEY_SCORE = "score";

        private final List<Word> _words;
        private final String _constraint;
        private final int _score;
        private boolean _hasGenericSpeech;

        /**
         * @param constraint The name of the constraint that was recognized
         * @param score Elvis's score for this result entry
         */
        Entry(String constraint, int score)
        {
            _constraint = constraint;
            _words = new ArrayList<Word>();
            _score = score;
        }

        /**
         * Get the name of this entry's constraint.
         * @return The name of this constraint.
         */
        public String getConstraint()
        {
            return _constraint;
        }

        /**
         * @return The entry's score as returned by Elvis.
         */
        public int getScore()
        {
            return _score;
        }

        /**
         * Determine whether any of the words in this list are generic speech.
         * @return True if any of the words are generic text
         */
        public boolean hasGenericSpeech()
        {
            return _hasGenericSpeech;
        }

        /**
         * Append a word to this entry.
         * @param word
         */
        void addWord(Word word)
        {
            _words.add(word);
            if (!_hasGenericSpeech && word.isGenericSpeech)
                _hasGenericSpeech = true;
        }

        /**
         * Append a word to this entry.
         * @param wordSlot
         * @param surfaceForm
         * @param spokenForm
         * @param fullPhrase
         * @param wordIds
         */
        void addWord(String wordSlot,
                     String surfaceForm,
                     String spokenForm,
                     String fullPhrase,
                     int [] wordIds)
        {
            addWord(new Word(wordSlot, surfaceForm, spokenForm, fullPhrase, wordIds, false));
        }

        /**
         * Set the word at the index to be generic
         * @param index
         */
        void setWordGeneric(int index)
         {
             Word w = _words.get(index);
             _words.set(index, new Word(w.wordSlot, GENERIC_SPEECH_TEXT, null, GENERIC_SPEECH_TEXT, w.wordIds, true));
             _hasGenericSpeech = true;
         }

        /**
         * Get a read-only copy of the words in this entry, in correct order.
         * @return The words in this entry.
         */
        public List<Word> getWords()
        {
            return new ReadOnlyList<Word>(_words);
        }

        /**
         * @return The transaction of the full entry
         */
        @Override
        public String toString()
        {
            StringBuilder builder = new StringBuilder();
            String separator = "";
            for (Word w : _words)
            {
                builder.append(separator);
                builder.append(w.toString());
                separator = " ";
            }
            return builder.toString();
        }

        @Override
        public org.json.JSONObject toJSON()
        {
            JSONObject o = new JSONObject();
            o.tryPut(KEY_CONSTRAINT, _constraint);
            o.tryPut(KEY_SCORE, _score);

            org.json.JSONArray words = new org.json.JSONArray();
            for (Word w : _words)
            {
                words.put(w.toJSON());
            }
            o.tryPut(KEY_WORDS, words);
            return o;
        }

        public static Entry createFromJSON(org.json.JSONObject json) throws JSONException
        {
            String constraint = json.getString(KEY_CONSTRAINT);
            int score = json.getInt(KEY_SCORE);
            Entry ret = new Entry(constraint, score);
            org.json.JSONArray words = json.getJSONArray(KEY_WORDS);
            for (int i = 0; i < words.length(); i++)
            {
                ret.addWord(Word.createFromJSON(words.getJSONObject(i)));
            }
            return ret;
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result + _score;
            result = prime * result
                    + ((_constraint == null) ? 0 : _constraint.hashCode());
            result = prime * result + (_hasGenericSpeech ? 1231 : 1237);
            result = prime * result
                    + ((_words == null) ? 0 : _words.hashCode());
            return result;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            Entry other = (Entry) obj;
            if (_score != other._score)
                return false;
            if (_constraint == null) {
                if (other._constraint != null)
                    return false;
            } else if (!_constraint.equals(other._constraint))
                return false;
            if (_hasGenericSpeech != other._hasGenericSpeech)
                return false;
            if (_words == null) {
                if (other._words != null)
                    return false;
            } else if (!_words.equals(other._words))
                return false;
            return true;
        }
    }

    private static final String KEY_ENTRIES = "entries";

    ElvisNbestList(List<Entry> entries)
    {
        super(entries);
    }

    ElvisNbestList(Entry ... entries)
    {
        super(entries);
    }

    private ElvisNbestList(int initialCapacity)
    {
        super(initialCapacity);
    }

    /**
     * Add an entry--add() doesn't work for read-only lists.
     * @param entry
     */
    void addEntry(Entry entry)
    {
        addInternal(entry);
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        org.json.JSONArray entries = new org.json.JSONArray();

        for (Entry e : this)
        {
            entries.put(e.toJSON());
        }
        o.tryPut(KEY_ENTRIES, entries);
        return o;
    }

    public static ElvisNbestList createFromJSON(org.json.JSONObject json) throws JSONException
    {
        org.json.JSONArray entries = json.getJSONArray(KEY_ENTRIES);
        ElvisNbestList ret = new ElvisNbestList(entries.length());

        for (int i = 0; i < entries.length(); i++)
        {
            ret.addEntry(Entry.createFromJSON(entries.getJSONObject(i)));
        }
        return ret;
    }
}
