package com.nuance.dragon.toolkit.elvis;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioEnergyListener;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.SpeechDetectionListener;
import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.language.LanguageEvent;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;

/**
 * The ElvisRecognizer class wraps the Elvis C library. The methods and callbacks
 * ('listeners') will deal with native Java objects, and convert as necessary
 * to C data types. This class does not expose all of Elvis's functionality--
 * only that which is intended to be used by this toolkit.
 * <p>
 * Creating: to create an ElvisRecognizer, call {@link #createElvisRecognizer}.
 * <p>
 * Initializing: this class is initialized by calling {@link #initialize}.
 * After initializing Elvis, a grammar should be specified via either
 * {@link #loadGrammar} or {@link #loadState}.
 * <p>
 * Recognizing: once Elvis has been initialized with the desired {@link Grammar}, begin
 * a recognition by calling {@link #startRecognition}. This method takes an audio
 * source, and listeners for speech detection (start, end) and speech recognition.
 * This class will internally manage consuming audio from the audio source,
 * feeding it to the Elvis engine, and providing updates to the listeners.
 */
public abstract class ElvisRecognizer extends LanguageEvent
{
    /** Not for public use. */
    public static boolean LOAD_NATIVE_LIBRARY_SUCCESS = NativeElvisImpl.LOAD_LIBRARY_SUCCESS;

    /** Constants to define the logging status. */
    public static final class Logging
    {
        /** Logging has not been enabled. */
        public static final int LOGGING_DISABLED = 0;

        /** Logging is enabled, but a recognition has not been started. */
        public static final int LOGGING_NOT_STARTED = 1;

        /** Logging is enabled, and a recognition is in progress. */
        public static final int LOGGING_IN_PROGRESS = 2;

        /** Recognition is complete and the log file was successfully written. */
        public static final int LOGGING_COMPLETE = 3;

        /** An error occurred during logging and no log file was written. */
        public static final int LOGGING_ERROR = 4;
    }

    /**
     * Wake-up modes supported by the engine.
     */
    public static final class EngineWakeupMode
    {
        /** Regular / default wake-up mode. */
        public static final int REGULAR = 0;

        /** Seamless / "one-shot" wake-up mode. */
        public static final int  SEAMLESS = 1;
    }

    /**
     * Listens for results or errors from an Elvis recognition.
     */
    public interface ResultListener
    {
        /**
         * Sent by Elvis when a result is available. The framework will take the
         * raw Elvis result and do some post-processing on it, including
         * removing semantic equivalents, and re-recognize a second time before
         * returning the result.
         *
         * @param result The Elvis result (never null)
         */
        public abstract void onResult(ElvisResult result);

        /**
         * A recognition error or internal error occurred. The Error object
         * contains details about what happened.
         *
         * @param error The error (never null)
         */
        public abstract void onError(ElvisError error);
    }

    /**
     * Listens for results or errors from an adaptation step.
     *
     * @deprecated use {@link AdaptationProgressListener} instead.
     */
    @Deprecated
    public interface AdaptationStepListener
    {
        /**
         * Sent by Elvis upon adaptation step completion
         *
         * @param success false if an error occurred
         */
        public abstract void onComplete(boolean success);
    }

    /**
     * Listens for results or errors from an adaptation step.
     */
    public interface AdaptationProgressListener extends AdaptationStepListener
    {
        /**
         * Sent by Elvis upon adaptation step completion
         *
         * @param result the adaptation result
         */
        void onComplete(ElvisAdaptationResult result);
    }

    /**
     * Listens for results or errors from an adaptation check.
     */
    public interface AdaptationCheckListener
    {
        /**
         * Sent by Elvis upon adaptation check completion
         *
         * @param adapted false if the system is not adapted or an error occurred
         */
        public abstract void onComplete(boolean adapted);
    }

    /**
     * Listens for results or errors from a grammar rebuild started by {@link
     * ElvisRecognizer#loadGrammar}.
     */
    public interface RebuildListener
    {
        /**
         * Represents a word that could not be added to the grammar during a rebuild.
         */
        public static class SkippedWord {

            private Word _word;
            private String _slotId;

            SkippedWord(Word word, String slotId)
            {
                _word = word;
                _slotId = slotId;
            }

            /**
             * @return The Id of the WordSlot that had a skipped word.
             */
            public String getSlotId() {
                return _slotId;
            }

            /**
             * @return The Word that could not be added.
             */
            public Word getWord() {
                return _word;
            }

        }

        /**
         * This method is called if a Grammar Rebuild complete successfully.
         *
         * @param grammar After a call to {@link ElvisRecognizer#loadGrammar} or
         * {@link ElvisRecognizer#loadStateAndGrammar}, a pointer to the same Grammar object will be returned.
         * After a call to {@link ElvisRecognizer#loadState} a new grammar representing the loaded grammar will be returned.
         * After a call to {@link ElvisRecognizer#clearGrammar}, a new empty grammar object will be returned.
         * @param skippedWords A list of words that could not be added to Elvis
         * because the pronunciation is not known. This object will never be
         * null, but may be empty if all words were added successfully.
         */
        void onComplete(Grammar grammar, List<SkippedWord> skippedWords);

        /**
         * A grammar format error or internal error occurred. The Error object
         * contains details about what happened.
         *
         * @param error The error (never null)
         */
        public abstract void onError(ElvisError error);
    }

    /**
     * Listens for the completion of {@link ElvisRecognizer#initialize}.
     */
    public interface InitializeListener
    {
        /**
         * The initialization completed
         * @param elvis The recognizer that was initialized.
         * @param success false if the initialization was unsuccessful
         */
        void onLoaded(ElvisRecognizer elvis, boolean success);
    }

    /**
     * Listens for the completion of {@link ElvisRecognizer#saveState}.
     */
    public interface SaveListener
    {
        /**
         * The save completed.
         * @param elvis The recognizer that was saved
         * @param success false if the save was unsuccessful
         */
        void onSaved(ElvisRecognizer elvis, boolean success);
    }

    /**
     * Listens for the completion of {@link ElvisRecognizer#release}.
     */
    public interface ReleaseListener
    {
        /**
         * The release completed
         * @param elvis The recognizer that was released
         */
        void onReleased(ElvisRecognizer elvis);
    }

    /**
     * Listens for the completion of {@link ElvisRecognizer#checkWakeupPhrases}.
     */
    public interface WakeupCheckListener
    {
        /**
         * Called when all phrases have been tested for validity.
         * If any phrase is too short, or does not have a defined pronunciation,
         * it will be included in the list of rejectedPhrases.  If the list is
         * empty, it means that all phrases were suitable.
         * @param rejectedPhrases A list containing the phrases that cannot
         *        be used for WakeupMode.  The phrases might be too short, or the
         *        pronunciation of the phrases cannot be determined.
         *        If all phrases are suitable, an empty list will be returned.
         *        This parameter will never be null.
         */
        void onComplete(List<String> rejectedPhrases);

        /*
         * Called if Elvis is in a state that does not allow checking of Wakeup Phrases.
         * @param error The error (never null)
         */
        void onError(ElvisError error);
    }

    /**
     * Listens for the completion of {@link ElvisRecognizer#createBlob}.
     */
    public interface BlobActionListener
    {
        /**
         * Called when createBlob has finished
         *
         * @param success false if the blob creation was unsuccessful
         * @param blob filled in with blob data when success is true
         */
        void onBlobCreate(boolean success, int [] blob);
        /**
         * Called when forgetBlob has finished
         *
         * @param success false if the blob destruction was unsuccessful
         */
        void onBlobForget(boolean success);

    }

    /**
     * Listens for the completion of {@link ElvisRecognizer#evaluateWakeupPhrase}.
     */
    public interface EvaluatePhraseListener
    {
        /**
         * Called when evaluateWakeupPhrase has finished
         *
         * @param success false if the blob evaluation was unsuccessful
         * @param evaluation phrase evaluation 0-100 when success is true
         */
        void onPhraseEvaluation(boolean success, int evaluation);

    }

    /**
     * Frequencies supported by Elvis.
     */
    public static final class Frequencies
    {
        public static final int FREQ_8KHZ  = 8000;
        public static final int FREQ_16KHZ = 16000;
    }

    /**
     * Languages supported by Elvis.
     */
    public static final class Languages extends LangSpecs
    {
    }

    /**
     * Get a list of available Elvis models. Each model consists of a language
     * and frequency.
     * <p>- This may be called from any thread, prior to initializing Elvis.
     * <br>- This method performs File I/O, so it may be preferable to call it
     * from a non-UI thread
     * <p>
     * TODO: This currently only looks at binfiles, not compiled languages.
     * @return The list of available models.
     */
    public abstract List<ElvisModelInfo> getAvailableModels();

    /**
     * Get a list of available Elvis models for a given language. Each model
     * consists of a language and frequency.
     * <p>- This may be called from any thread, prior to initializing Elvis.
     * <br>- This method performs File I/O, so it may be preferable to call it
     * from a non-UI thread
     * <p>
     * TODO: This currently only looks at binfiles, not compiled languages.
     * @param language The language to restrict results to (from {@link ElvisRecognizer.Languages}).
     * @return The list of available models.
     */
    public abstract List<ElvisModelInfo> getAvailableModels(ElvisLanguage language);

    /**
     * Get a list of available Elvis models for a given frequency. Each model
     * consists of a language and frequency.
     * <p>- This may be called from any thread, prior to initializing Elvis.
     * <br>- This method performs File I/O, so it may be preferable to call it
     * from a non-UI thread
     * <p>
     * TODO: This currently only looks at binfiles, not compiled languages.
     * @param freqHz The frequency to restrict results to (from {@link Frequencies}).
     * An invalid input will return an empty container.
     * @return The list of available models.
     */
    public abstract List<ElvisModelInfo> getAvailableModels(int freqHz);

    /**
     * Load Elvis. This must be called prior to doing anything with the Elvis
     * object. Typically you'd call this right after constructing the Elvis
     * object. Note that even though stateFileName is passed here, it will NOT
     * be loaded unless {@link #loadState} is called.
     * @param config The Elvis configuration
     * @param stateFileName The file where Elvis's state will be saved / loaded
     * by {@link #saveState} and {@link #loadState}.
     * @param listener Notified after the load completes
     */
    public abstract void initialize(ElvisConfig config, String stateFileName, InitializeListener listener);

    /**
     * Like {@link #initialize(ElvisConfig,String,InitializeListener)}, but no listener.
     * @param stateFileName The file where Elvis's state will be saved / loaded
     * by {@link #saveState} and {@link #loadState}.
     */
    public abstract void initialize(ElvisConfig config, String stateFileName);

    /**
     * Release all resources when done with the recognizer. The recognizer can
     * never be used after calling this. Not calling this may result in memory
     * leaks.
     * @param listener Notified after the release completes
     */
    public abstract void release(ReleaseListener listener);

    /**
     * Like {@link #release(ReleaseListener)}, but no listener.
     */
    public abstract void release();

    /**
     * Save current model set, grammar, and vocabulary to disk. <br>
     * Uses the file previously passed in {@link #initialize}.
     *
     * @param listener Notified after the save completes.
     */
    public abstract void saveState(SaveListener listener);

    /**
     * Like {@link #saveState(SaveListener)}, but no listener.
     */
    public abstract void saveState();

    /**
     * Load Elvis model set and stored vocabulary from disk.<br>
     * Uses the file previously passed in {@link #initialize}.
     *
     * @param rebuildListener Called when the Elvis state is loaded.
     */
    public abstract void loadState(RebuildListener rebuildListener);

    /**
     * Combined {@link #loadState} and {@link #loadGrammar}. If loading the state
     * fails, the grammar will be fully rebuild; otherwise, the grammar will be
     * loaded normally on top of the loaded state.
     * @param grammar The grammar to load. If  null, the behavior will be the
     * same as {@link #loadState}.
     * @param rebuildListener Will be notified when the state and/or grammar
     * has been loaded.
     */
    public abstract void loadStateAndGrammar(Grammar grammar,
                                             RebuildListener rebuildListener);

    /**
     * Load the given grammar into the Elvis engine. This process is
     * asynchronous. The listener will be notified when it completes. <p>
     * When the load succeeds, the Elvis state will automatically be saved.
     * @param grammar The grammar containing the word slots and constraints.
     * @param listener Will be notified when the rebuild has finished.
     */
    public abstract void loadGrammar(Grammar grammar,
                                     RebuildListener listener);

    /**
     * Clear all constraints and word slots from the Elvis engine.
     * @param listener Will be notified when grammar has been cleared.
     */
    public abstract void clearGrammar(RebuildListener listener);

    /**
     * Clear adaptation data from the Elvis engine.
     * <p><b>This method shall be called after {@link #initialize} but
     * before {@link #loadState}, {@link #loadGrammar} or {@link #loadStateAndGrammar}.</b>
     * @param listener Will be notified when adaptation data has been cleared.
     */
    public abstract void clearAdaptation(AdaptationStepListener listener);

    /**
     * Clear adaptation data from the Elvis engine.
     * <p><b>This method shall be called after {@link #initialize} but
     * before {@link #loadState}, {@link #loadGrammar} or {@link #loadStateAndGrammar}.</b>
     * @param all Will clear adaptation in every languages and frequencies. Default is false.
     * @param listener Will be notified when adaptation data has been cleared.
     */
    public abstract void clearAdaptation(boolean all, AdaptationStepListener listener);

    /**
     * Clear wake-up adaptation data from the Elvis engine.
     * <p><b>This method shall be called after {@link #initialize} but
     * before {@link #loadState}, {@link #loadGrammar} or {@link #loadStateAndGrammar}.</b>
     * @param listener Will be notified when adaptation data has been cleared.
     */
    public abstract void clearWuWAdaptation(AdaptationProgressListener listener);

    /**
     * Clear wake-up adaptation data from the Elvis engine.
     * <p><b>This method shall be called after {@link #initialize} but
     * before {@link #loadState}, {@link #loadGrammar} or {@link #loadStateAndGrammar}.</b>
     * @param all Will clear wake-up adaptation in every languages and frequencies. Default is false.
     * @param listener Will be notified when adaptation data has been cleared.
     */
    public abstract void clearWuWAdaptation(boolean all, AdaptationProgressListener listener);

    /**
     * Cancel a rebuild that is in progress. This may leave Elvis in an
     * indeterminate state, so be sure to load a grammar to completion before
     * using Elvis.
     */
    public abstract void cancelRebuild();

    /**
     * Find the constraint with the given name.
     * @param id The constraint ID to look for
     * @return The constraint with the given name, or null if none could be
     * found.
     */
    public abstract Constraint findConstraint(String id);

    /**
     * Find the word slot with the given name.
     * @param id The word slot ID to look for
     * @return The word slot with the given name, or null if none could be
     * found.
     */
    public abstract WordSlot findWordSlot(String id);

    /**
     * @deprecated Set a list of constraints to be active on the next recognition.
     *
     * @param activeConstraints The constraints to make active. The constraints
     * must have previously been added to the Elvis grammar using {@link #loadGrammar}.
     */
    @Deprecated
    public abstract void setActiveConstraints(List<Constraint> activeConstraints);

    /**
     * @deprecated Set a list of constraints to be active on the next recognition.
     *
     * @param activeConstraintNames The names of the constraints to make active.
     * The constraints must have previously been added to the Elvis grammar
     * using {@link #loadGrammar}.
     */
    @Deprecated
    public abstract void setActiveConstraintNames(List<String> activeConstraintNames);

    /**
     * Set custom pronunciations in the vocabulary.
     * In order to ensure that the custom pronunciations are added to the vocabulary
     * correctly, one must call {@link #clearGrammar(RebuildListener)} before
     * calling this method.
     *
     * @param words The words that need to have custom pronunciations.
     * Each word entry needs to have valid written and spoken forms.
     */
    public abstract void setCustomPronunciations(List<Word> words);

    /**
     * Check whether Elvis is currently adapted or not.
     *
     * @param adaptationCheckListener The listener that will receive a callback when the check is complete.
     */
    public abstract void checkAdaptationStatus(AdaptationCheckListener adaptationCheckListener);

    /**
     * Check whether a list of phrases can be used for WakeupMode.
     * If all phrases are suitable, wakeupCheckListener will be called with an empty list of phrases.
     * If any phrase is not suitable, wakeupCheckListener will be called with a list of phrases that cannot be used.
     *
     * @param wakeupPhrases The phrases to check for suitability.
     * @param wakeupCheckListener The listener that will receive a callback when the check is complete.
     */
    public abstract void checkWakeupPhrases(List<String> wakeupPhrases, WakeupCheckListener wakeupCheckListener);

    /**
     * @deprecated Set a list of wakeup phrases to be active the next time WakeupMode is started.
     *
     * @param wakeupPhrases The phrases to enable for {@link #startWakeupMode}.
     * These can't be too short, and they must be words that are in Elvis's
     * vocabulary and do not require a separate pronunciation to be specified.
     */
    @Deprecated
    public abstract void setWakeupPhrases(List<String> wakeupPhrases);

    /**
     * @deprecated Set the penalty for the next time WakeupMode is started.
     * This method is optional. When not called, the default penalty passed to the
     * engine is 0. Then, the native engine will consume the penalty parameter.
     *
     * @param penalty Penalty to allow trade-off between False Accept and False Reject.
     * A positive penalty will increase False Accept and lower False Reject.
     */
    @Deprecated
    public abstract void setWakeupPenalty(int penalty);

    /**
     * @deprecated Set a list of wakeup phrases to be active the next time WakeupMode is started.
     *
     * @param wakeupPhrases The phrases to enable for {@link #startWakeupMode}.
     * These can't be too short, and they must be words that are in Elvis's
     * vocabulary and do not require a separate pronunciation to be specified.
     */
    @Deprecated
    public final void setWakeupPhrases(String ... wakeupPhrases)
    {
        ArrayList<String> phrases = new ArrayList<String>(wakeupPhrases.length);
        for (String s : wakeupPhrases)
        {
            phrases.add(s);
        }
        setWakeupPhrases(phrases);
    }

    /**
     * @deprecated Call {@link #startRecognition(AudioSource,List,SpeechDetectionListener,ResultListener)} instead.
     * Begin performing a recognition using the connected audio source, based on
     * the current active constraint set. Prior to calling this, constraints
     * must have been made active by calling {@link #setActiveConstraints} or
     * {@link #setActiveConstraintNames}.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz)
     * @param speechDetectionListener Optional listener to be notified on start
     * of speech and end of speech. May be null.
     * @param resultListener Listener to be notified when recognition is
     * complete
     */
    @Deprecated
    public abstract void startRecognition(AudioSource<AudioChunk> audioSource,
                                          SpeechDetectionListener speechDetectionListener,
                                          ResultListener resultListener);

    /**
     * Begin performing a recognition, using the connected audio source, against
     * the passed active constraints.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz).
     * @param activeConstraintNames The names of the constraints to make active.
     * The constraints must have previously been added to the Elvis grammar
     * using {@link #loadGrammar}.
     * @param speechDetectionListener Optional listener to be notified on start
     * of speech and end of speech. May be null.
     * @param resultListener Listener to be notified when recognition is
     * complete.
     */
    public abstract void startRecognition(AudioSource<AudioChunk> audioSource,
                                          List<String> activeConstraintNames,
                                          SpeechDetectionListener speechDetectionListener,
                                          ResultListener resultListener);

    /**
     * Begin performing a recognition, using the connected audio source, against
     * the passed active constraints.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz).
     * @param activeConstraintNames The names of the constraints to make active.
     * The constraints must have previously been added to the Elvis grammar
     * using {@link #loadGrammar}.
     * @param speechDetectionListener Optional listener to be notified on start
     * of speech and end of speech. May be null.
     * @param audioEnergyListener Optional listener to be notified with audio
     * speech energy updates. May be null.
     * @param resultListener Listener to be notified when recognition is
     * complete.
     */
    public abstract void startRecognition(AudioSource<AudioChunk> audioSource,
                                          List<String> activeConstraintNames,
                                          SpeechDetectionListener speechDetectionListener,
                                          AudioEnergyListener audioEnergyListener,
                                          ResultListener resultListener);

    /**
     * @deprecated Call {@link #startWakeupMode(AudioSource,List,int,ResultListener)} instead.
     * Enter 'wakeup mode' using the connected audio source, based on the
     * current configured wakeup phrases. Prior to calling this, wakeup phrases
     * must have been made active by calling {@link #setWakeupPhrases}.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz)
     * @param resultListener Listener to be notified when a wakeup phrase is
     * detected
     */
    @Deprecated
    public abstract void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                         ResultListener resultListener);

    /**
     * @deprecated Call {@link #startWakeupMode(AudioSource,List,int,ResultListener)} instead.
     * Enter 'wakeup mode', using the connected audio source, against the passed
     * wake-up phrases.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz).
     * @param wakeupPhrases The active wake-up phrases. These can't be too short.
     * @param resultListener Listener to be notified when a wakeup phrase is
     * detected.
     */
    @Deprecated
    public abstract void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                         List<String> wakeupPhrases,
                                         ResultListener resultListener);

    /**
     * @deprecated Enter 'wakeup mode', using the connected audio source, against the passed
     * wake-up phrases.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz).
     * @param wakeupPhrases The active wake-up phrases. These can't be too short.
     * If the phrase pronunciations are not in the Elvis vocabulary, they can be added
     * calling {@link #setCustomPronunciations}.
     * @param penalty Penalty to allow trade-off between False Accept and False Reject.
     * A positive penalty will increase False Accept and lower False Reject.
     * @param resultListener Listener to be notified when a wakeup phrase is
     * detected.
     */
    @Deprecated
    public abstract void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                         List<String> wakeupPhrases,
                                         int penalty,
                                         ResultListener resultListener);

    /**
     * @deprecated Initiates a regular or seam-less wake-up word recognition.
     * @param audioSource The audio source to get audio from.
     * @param wakeupPhrases The active wake-up phrases. These can't be too short.
     * @param penalty Penalty to allow trade-off between False Accept and False Reject.
     * A positive penalty will increase False Accept and lower False Reject.
     * @param pauseDurationMs The expected pause duration (in ms) to differentiate
     * between a regular wake-up or a seam-less wake-up. Default is 1000 ms.
     * @param resultListener Listener to be notified when a wake-up phrase is
     * detected.
     */
    @Deprecated
    public abstract void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                         List<String> wakeupPhrases,
                                         int penalty, int pauseDurationMs,
                                         ResultListener resultListener);

    /**
     * @deprecated Initiates a regular or seam-less wake-up word recognition.
     * <b>This is mainly intented for internal customers and research purposes </b>.
     * @param audioSource The audio source to get audio from.
     * @param wakeupPhrases The active wake-up phrases. These can't be too short.
     * @param penalty Penalty to allow trade-off between False Accept and False Reject.
     * A positive penalty will increase False Accept and lower False Reject.
     * @param pauseDurationMs The expected pause duration (in ms) to differentiate
     * between a regular wake-up or a seam-less wake-up. Default is 1000 ms.
     * @param engineWakeupMode The engine wake-up mode override, see
     * {@link EngineWakeupMode}.
     * @param resultListener Listener to be notified when a wake-up phrase is
     * detected.
     */
    @Deprecated
    public abstract void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                         List<String> wakeupPhrases,
                                         int penalty, int pauseDurationMs, int engineWakeupMode,
                                         ResultListener resultListener);

    /**
     * @deprecated Initiates a seam-less wake-up word recognition.
     * @param audioSource The audio source to get audio from.
     * @param wakeupPhrases The active wake-up phrases. These can't be too short.
     * @param penalty Penalty to allow trade-off between False Accept and False Reject.
     * A positive penalty will increase False Accept and lower False Reject.
     * @param resultListener Listener to be notified when a wake-up phrase is
     * detected.
     */
    @Deprecated
    public abstract void startSeamlessWakeupMode(AudioSource<AudioChunk> audioSource,
                                                 List<String> wakeupPhrases,
                                                 int penalty,
                                                 ResultListener resultListener);

    /**
     * Initiates a regular or seam-less wake-up word recognition.
     * @param audioSource The audio source to get audio from.
     * @param wakeupParam The wake-up parameter, see {@link WakeupParam}.
     * @param audioEnergyListener Optional listener to be notified with audio
     * speech energy updates. May be null.
     * @param resultListener Listener to be notified when a wake-up phrase is
     * detected.
     */
    public abstract void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                         WakeupParam wakeupParam,
                                         AudioEnergyListener audioEnergyListener,
                                         ResultListener resultListener);

    /**
     * Begin performing an adaptation, using the connected audio source, against
     * the passed phrase.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz).
     * @param phrase The adaptation phrase to recognize against.
     * @param terminateSession True to indicate this is the last phrase to adapt.
     * This will instructs the engine to terminate the adaptation session and save
     * the adaptation data. When in a middle of an adaptation session, this parameter
     * shall be set to false.
     * @param speechDetectionListener Optional listener to be notified on start
     * of speech and end of speech. May be null.
     * @param stepListener Listener to be notified when adaptation step is
     * complete.
     */
    public abstract void startAdaptation(AudioSource<AudioChunk> audioSource,
                                         String phrase, boolean terminateSession,
                                         SpeechDetectionListener speechDetectionListener,
                                         AdaptationStepListener stepListener);

    /**
     * Begin performing an adaptation, using the connected audio source, against
     * the passed phrase.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz).
     * @param phrase The adaptation phrase to recognize against.
     * @param terminateSession True to indicate this is the last phrase to adapt.
     * This will instructs the engine to terminate the adaptation session and save
     * the adaptation data. When in a middle of an adaptation session, this parameter
     * shall be set to false.
     * @param speechDetectionListener Optional listener to be notified on start
     * of speech and end of speech. May be null.
     * @param audioEnergyListener Optional listener to be notified with audio
     * speech energy updates. May be null.
     * @param stepListener Listener to be notified when adaptation step is
     * complete.
     */
    public abstract void startAdaptation(AudioSource<AudioChunk> audioSource,
                                         String phrase, boolean terminateSession,
                                         SpeechDetectionListener speechDetectionListener,
                                         AudioEnergyListener audioEnergyListener,
                                         AdaptationStepListener stepListener);



    /**
     * Abort an eventual adaption session and do not cache the adaptation data accumulated
     * up to this point. This is different than {@link #cancelRecognition()} that cancels
     * the current adaptation stage but preserves the adaptation session in case same/other
     * phrases need to be added.
     * <p>
     * Note: This API is safe to call even when there is no adaptation session going on.
     */
    public abstract void abortAdaptation();

    /**
     * Stop listening to the audio, and process the result now. Applies during
     * a recognition, adaptation or wakeup mode.
     */
    public abstract void stopListening();

    /**
     * Cancels the current recognition and clears any audio in the queue. No
     * result is returned.
     * <p>
     * Note: use this method to cancel adaptation, speech detection and wakeup mode as well.
     */
    public abstract void cancelRecognition();

    /**
     * Prepare an Elvis instance to receive audio from a connected source and
     * simply detect start/end of speech. Does not perform a recognition.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz)
     * @param speechDetectionListener Listener to be notified on start of speech
     * and end of speech. May be null.
     */
    public abstract void startSpeechDetection(AudioSource<AudioChunk> audioSource,
                                              SpeechDetectionListener speechDetectionListener);


    /**
     * Prepare an Elvis instance to receive audio from a connected source and
     * simply detect start/end of speech. Does not perform a recognition.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz)
     * @param longSpeech True for longer sensitivity. False otherwise.
     * @param speechDetectionListener Listener to be notified on start of speech
     * and end of speech. May be null.
     */
    public abstract void startSpeechDetection(AudioSource<AudioChunk> audioSource, boolean longSpeech,
                                              SpeechDetectionListener speechDetectionListener);


    /**
     * @deprecated Prepare an Elvis instance to receive audio from a connected source and
     * simply detect start/end of speech. Does not perform a recognition.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz)
     * @param silenceDurationMs The silence duration (in ms) needed to declare end-of-speech.
     * @param speechDetectionListener Listener to be notified on start of speech
     * and end of speech. May be null.
     */
    @Deprecated
    public abstract void startSpeechDetection(AudioSource<AudioChunk> audioSource, int silenceDurationMs,
                                              SpeechDetectionListener speechDetectionListener);

    /**
     * @deprecated Prepare an Elvis instance to receive audio from a connected source and
     * simply detect start/end of speech. Does not perform a recognition.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz)
     * @param silenceDurationMs The silence duration (in ms) needed to declare end-of-speech.
     * @param speechDetectionListener Listener to be notified on start of speech
     * and end of speech. May be null.
     * @param audioEnergyListener Optional listener to be notified with audio
     * speech energy updates. May be null.
     */
    @Deprecated
    public abstract void startSpeechDetection(AudioSource<AudioChunk> audioSource, int silenceDurationMs,
                                              SpeechDetectionListener speechDetectionListener,
                                              AudioEnergyListener audioEnergyListener);

    /**
     * @deprecated Prepare an Elvis instance to receive audio from a connected source and
     * simply detect start/end of speech. Does not perform a recognition.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz)
     * @param silenceDurationMs The silence duration (in ms) needed to declare end-of-speech.
     * @param ignoreDurationMs The duration (in ms) to ignore before reporting speech notifications. Default: 0 ms.
     * @param resetChannel true to reset the Elvis front-end channel, else false. Default: true.
     * @param speechDetectionListener Listener to be notified on start of speech
     * and end of speech. May be null.
     * @param audioEnergyListener Optional listener to be notified with audio
     * speech energy updates. May be null.
     */
    @Deprecated
    public abstract void startSpeechDetection(AudioSource<AudioChunk> audioSource, int silenceDurationMs, int ignoreDurationMs, boolean resetChannel,
                                              SpeechDetectionListener speechDetectionListener,
                                              AudioEnergyListener audioEnergyListener);

    /**
     * Prepare an Elvis instance to receive audio from a connected source and
     * simply detect start/end of speech. Does not perform a recognition.
     *
     * @param audioSource The audio source to get audio from. The audio type
     * must match the Elvis model type (PCM 16Khz or 8Khz)
     * @param endpointingParam The end-pointing parameter, see {@link EndpointingParam}.
     * @param speechDetectionListener Listener to be notified on start of speech
     * and end of speech. May be null.
     * @param audioEnergyListener Optional listener to be notified with audio
     * speech energy updates. May be null.
     */
    public abstract void startSpeechDetection(AudioSource<AudioChunk> audioSource,
                                              EndpointingParam endpointingParam,
                                              SpeechDetectionListener speechDetectionListener,
                                              AudioEnergyListener audioEnergyListener);

    /**
     * This function controls the amount of information about the ElvisRecognizer
     * that will be written to the Android logs.
     *
     * Note: Do not use / disabled for your application release build.
     *
     * @param verbose True if detailed (verbose) status information for the ElvisRecognizer should be written
     *                to the Android logs.
     *                False if only minimal details and error information should be written to  the Android logs.
     */
    public abstract void enableVerboseAndroidLogging(boolean verbose);


    /**
     * This function enables logging of recognition data for a single recognition.
     * This function should be called after Elvis has been created and the grammar
     * has been loaded, but before calling:
     * {@link #startRecognition}, {@link #startWakeupMode}, or {@link #startSpeechDetection}
     *
     * Note: Do not use for your application release build.
     *
     * This function will have no effect if logging is already in progress.
     *
     * Use {@link #getRecognitionLoggingStatus()} after receiving any result or error callback to determine
     * the status of logging.
     *
     * @param destinationFolder The destination folder for log files.  A null value indicates logging should be disabled.
     * @param filenamePrefix An optional string to prepend to each logfile name.  May be null.
     * @param logMilliseconds The number of milliseconds of audio to include in the logs.
     *                        A value of 0 (or any negative value) indicates logging should be disabled.
     */
    public abstract void enableRecognitionLogging(String destinationFolder,
                                                  String filenamePrefix,
                                                  int logMilliseconds);

    /**
     * This function enables logging of recognition data for a single recognition.
     * This function should be called after Elvis has been created and the grammar
     * has been loaded, but before calling:
     * {@link #startRecognition}, {@link #startWakeupMode}, or {@link #startSpeechDetection}
     *
     * Note: Do not use for your application release build.
     *
     * This function will have no effect if logging is already in progress.
     *
     * Use {@link #getRecognitionLoggingStatus()} after receiving any result or error callback to determine
     * the status of logging.
     *
     * @param destinationFolder The destination folder for log files.  A null value indicates logging should be disabled.
     * @param filenamePrefix An optional string to prepend to each logfile name.  May be null.
     * @param logMilliseconds The number of milliseconds of audio to include in the logs.
     *                        A value of 0 (or any negative value) indicates logging should be disabled.
     * @param uuid An optional UUID to include in the recognition logs. Default value is null.
     */
    public abstract void enableRecognitionLogging(String destinationFolder,
                                                  String filenamePrefix,
                                                  int logMilliseconds,
                                                  String uuid);

    /**
     * Log misc data into Elvis engine, which will be included in the Elvis dump
     * eventually. Currently only a fixed set of tags are supported by the Elvis
     * engine. The application cannot pass arbitrary data.
     *
     * The supported tags are: "NVCSessionId", "NVCUniqueUserId". "NVCTransactionId"
     * will be supported in the future Elvis release.
     *
     * @param tag   the "name" of the data to be logged.
     * @param data  The "value" of the data to be logged.
     *
     * @return true for success or false for failure
     */
    public abstract boolean logMiscInfo(String tag, String data);

    /**
     * The specific NVC session id value.
     */
    public static final String TAG_NVC_SESSION_ID = "NVCSessionId";

    /**
     * This function checks the status of the most recently logged recognition.
     *
     * @return A value from {@link Logging}.
     */
    public abstract int getRecognitionLoggingStatus();

    /**
     * Create the blob file data to be saved and used for miniElvis
     *
     * @param blobActionListener Listener to be notified when action is finished {@link BlobActionListener#onBlobCreate}
     * @param phrases a list of phrases to include in the blob.
     * @param cpu optional short string describing processor CPU for the blob
     * @param numPhrases number of phrases in the list.
     * @param decoyPenalty the decoy penalty to be used when generating the blob data
     * @param isMagic boolean allow speech following the wakeup phrases.
     * @param maxDuration maximum speech duration in msec, should be set to -1.
     *
     */
    public abstract void createBlob(BlobActionListener blobActionListener, List<String> phrases,
                                    String cpu, int numPhrases, int decoyPenalty, boolean isMagic, int maxDuration);

    /**
     * Forget the current blob data.
     *
     * @param blobActionListener Listener to be notified when action is finished {@link BlobActionListener#onBlobForget}
     */
    public abstract void forgetBlob(BlobActionListener blobActionListener);

    /**
     * Evaluate wakeup phrase
     *
     * @param evaluationListener Listener to be notified when evaluation is done {@link EvaluatePhraseListener#onPhraseEvaluation}
     * @param phrase the phrase to evaluate
     */
    public abstract void evaluateWakeupPhrase(EvaluatePhraseListener evaluationListener, String phrase);

    /**
     * Create an {@link ElvisRecognizer}. The Elvis will use a dedicated worker thread. <p>
     * NOTE: Due to Elvis engine limitations, multiple Elvis instances must use
     * the same worker thread.
     * @param fileManager The file manager where Elvis should look for files.
     */
    public static ElvisRecognizer createElvisRecognizer(FileManager fileManager)
    {
        return new ElvisRecognizerImpl(fileManager);
    }

    /**
     * Create an {@link ElvisRecognizer}. The Elvis will use the given worker thread. <p>
     * NOTE: Due to Elvis engine limitations, multiple Elvis instances must use
     * the same worker thread.
     * @param fileManager The file manager where Elvis should look for files.
     * @param workerThread The worker thread to use. If null, a dedicated one
     * will be created.
     */
    public static ElvisRecognizer createElvisRecognizer(FileManager fileManager, NMTHandler workerThread)
    {
        return new ElvisRecognizerImpl(fileManager, workerThread);
    }

    /**
     * @return An Elvis recognizer 'impersonator'. The impersonator doesn't do
     * any real recognition--it is a very light 'fake' Elvis module that may be
     * useful for testing other modules.
     */
    public static ElvisRecognizer createElvisImpersonator()
    {
        return new ElvisRecognizerImpersonator();
    }
}
