package com.nuance.dragon.toolkit.elvis;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import junit.framework.Assert;
import android.os.SystemClock;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioEnergyListener;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.SpeechDetectionListener;
import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.elvis.ElvisRecognizer.RebuildListener.SkippedWord;
import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Base class for Elvis recognizer implementations.
 * <p>
 * There will probably be one 'real' subclass of this ({@link ElvisRecognizerImpl}),
 * along with variants used for unit testing (e.g. {@link ElvisRecognizerImpersonator}).
 */
@SuppressWarnings("deprecation")
abstract class ElvisRecognizerBase extends ElvisRecognizer
{
    /** The number of word classes to add at a time. Too few and things will be
     * slow; too many and there will be a delay when trying to interrupt a
     * rebuild. */
    private static final int WORDCLASS_BATCH_SIZE = 100;

    private enum RecognitionMode
    {
        RECOGNITION, WAKEUP, ADAPTATION, ENDPOINTING
    };

    /**
     * Internal class that keeps track of a single recognition.
     * TODO: move this to a separate file?
     */
    private class Recognition extends AudioSink<AudioChunk>
    {
        private final Grammar _recogGrammar;
        private ResultListener _resultListener;
        private AdaptationStepListener _adaptationStepListener;
        private SpeechDetectionListener _speechStartListener;
        private SpeechDetectionListener _speechEndListener;
        private AudioEnergyListener _audioEnergyListener;
        private boolean _started;
        private RecognitionMode _mode;
        private boolean _terminateAdaptationSession;
        private boolean _speechEnded; // Accessed from worker thread
        private short[] _prevAudio; // Place to save small buffers that aren't big enough to give to Elvis
        private long _prevAudioTimestamp; // The timestamp of _prevAudio
        private boolean _seamlessWakeup;
        private int _lastWakeupStatus;

        Recognition(Grammar grammar, ResultListener resultListener, SpeechDetectionListener speechDetectionListener, AudioEnergyListener audioEnergyListener, AdaptationStepListener adaptationStepListener)
        {
            super(null);
            _recogGrammar = grammar;
            _resultListener = resultListener;
            _speechStartListener = speechDetectionListener;
            _speechEndListener = speechDetectionListener;
            _audioEnergyListener = audioEnergyListener;
            _adaptationStepListener = adaptationStepListener;
        }

        /**
         * Start recognizing. This will connect the audio source, and initialize
         * the Elvis engine on the worker thread.
         * @param audioSource
         * @param activeConstraints
         */
        void startRecognizing(AudioSource<AudioChunk> audioSource,
                              final List<String> activeConstraints)
        {
            Assert.assertFalse(_started);

            if (!_started)
            {
                _started = true;
                _prevAudio = null;
                _workerThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        _mode = RecognitionMode.RECOGNITION;

                        boolean success = _nativeElvis.recognize(activeConstraints, true);
                        updateRecognitionLoggingStatus();
                        if (!success)
                        {
                            handleStartError();
                        }
                    }
                });
                connectAudioSource(audioSource);
            }
        }

        /**
         * Start WakeupMode. This will connect the audio source, and initialize
         * the Elvis engine on the worker thread.
         * @param audioSource
         * @param wakeupParam
         */
        void startWakeupMode(AudioSource<AudioChunk> audioSource, final WakeupParam wakeupParam)
        {
            Assert.assertFalse(_started);

            if (!_started)
            {
                _started = true;
                _prevAudio = null;
                _workerThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        _mode = RecognitionMode.WAKEUP;
                        int nativeMode;
                        switch (wakeupParam.getMode())
                        {
                        case WAKEUP_COMMAND:
                            nativeMode = NativeElvis.WAKEUP_COMMAND;
                            _seamlessWakeup = true;
                            break;
                        case WAKEUP_SHORT:
                            nativeMode = NativeElvis.WAKEUP_SHORT;
                            _seamlessWakeup = false;
                            break;
                        case WAKEUP_SNAP:
                            nativeMode = NativeElvis.WAKEUP_SNAP;
                            _seamlessWakeup = false;
                            break;
                        default:
                            nativeMode = NativeElvis.WAKEUP_LONG;
                            _seamlessWakeup = false;
                            break;
                        }

                        boolean success = _nativeElvis.wakeupMode(wakeupParam.getPhrases(),
                                                                  wakeupParam.getPenalty(),
                                                                  nativeMode,
                                                                  wakeupParam.getCommandStartBailoutMs(),
                                                                  wakeupParam.getCommandEndBailoutMs(),
                                                                  wakeupParam.getCommandMinSpeechMs(),
                                                                  wakeupParam.isAutoAdapt());
                        updateRecognitionLoggingStatus();
                        if (!success)
                        {
                            handleStartError();
                        }
                    }
                });
                connectAudioSource(audioSource);
            }
        }

        /**
         * Start adapting. This will connect the audio source, and initialize
         * the Elvis engine on the worker thread.
         * @param audioSource
         * @param phrase
         * @param terminateSession
         */
        void startAdapting(AudioSource<AudioChunk> audioSource,
                           final String phrase, final boolean terminateSession)
        {
            Assert.assertFalse(_started);

            if (!_started)
            {
                _started = true;
                _prevAudio = null;
                _workerThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        _mode = RecognitionMode.ADAPTATION;
                        _terminateAdaptationSession = terminateSession;

                        boolean success = _nativeElvis.adapt(phrase);
                        updateRecognitionLoggingStatus();
                        if (!success)
                        {
                            handleStartError();
                        }
                    }
                });
                connectAudioSource(audioSource);
            }
        }

        /**
         * Start endpointing. This will connect the audio source, and initialize
         * the Elvis engine on the worker thread.
         * @param audioSource
         * @param endpointingParam
         */
        void startEndpointing(AudioSource<AudioChunk> audioSource, final EndpointingParam endpointingParam)
        {
            Assert.assertFalse(_started);

            if (!_started)
            {
                _started = true;
                _prevAudio = null;
                final int silenceSampleCount = audioSource.getAudioType().getSampleCount(endpointingParam.getSilenceDurationMs());
                final int ignoreSampleCount = audioSource.getAudioType().getSampleCount(endpointingParam.getIgnoreDurationMs());
                _workerThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        _mode = RecognitionMode.ENDPOINTING;

                        boolean success = _nativeElvis.endpointingMode(silenceSampleCount, ignoreSampleCount, endpointingParam.getResetChannel(), _fxLoggingFolder);
                        updateRecognitionLoggingStatus();
                        if (!success)
                        {
                            handleStartError();
                        }
                    }
                });
                connectAudioSource(audioSource);
            }
        }

        /**
         * Stop recognizing or adapting. This will disconnect the audio source
         * and compute the Elvis result on the worker thread (if applicable). It
         * is safe to call this method if {@link #start} hasn't been called.
         * @param abort true if results should be discarded,
         *        false if results should be processed normally.
         */
        void stop(final boolean abort)
        {
            // Disable detection
            _speechStartListener = null;
            _speechEndListener = null;

            disconnectAudioSource();
            if (_started)
            {
                _started = false;
                _workerThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (_speechEnded)
                            return;

                        boolean success = true;
                        if (_mode == RecognitionMode.RECOGNITION || _mode == RecognitionMode.ADAPTATION) {
                            success = _nativeElvis.finishRecognizing(abort);
                        }
                        else {
                             if (_mode == RecognitionMode.ENDPOINTING) {
                                 success = _nativeElvis.finishEndpointing();
                             }
                        }

                        _speechEnded = true;

                        ElvisResult result = null;
                        ElvisAdaptationResult adaptResult = null;
                        if (!abort && success) {
                            /* Only process the result if recognition was NOT aborted. */
                            if (_mode == RecognitionMode.RECOGNITION) {
                                result = _nativeElvis.getResult(_recogGrammar);
                            }
                        }

                        if (_mode == RecognitionMode.WAKEUP) {
                            success = _nativeElvis.finishWakeupMode(_stateFileName);
                        } else {
                            if (!abort && success && _mode == RecognitionMode.ADAPTATION) {
                                adaptResult = _nativeElvis.processAdaptation();
                                if (adaptResult.getStatus() && _terminateAdaptationSession)
                                {
                                    if (!_nativeElvis.finishAdapting(_stateFileName, false))
                                        adaptResult = new ElvisAdaptationResult(false, adaptResult.getMiscLogging());
                                }
                            }
                            /* Release memory used by the completed recognition. */
                            _nativeElvis.releaseUnusedMemory();
                        }
                        updateRecognitionLoggingStatus();
                        if (abort) {
                            /* if recognition was aborted, we should not try to
                             * post the result or error back to the main thread. */
                            return;
                        }

                        final ElvisResult finalResult = result;
                        final ElvisAdaptationResult finalAdaptResult = adaptResult;
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                if(_mode == RecognitionMode.ADAPTATION)
                                {
                                   handleAdaptationStep(finalAdaptResult);
                                }
                                else
                                {
                                    if (finalResult == null)
                                        handleError(new ElvisError(ElvisError.Reason.NO_RESULT));
                                    else
                                        handleResult(finalResult);
                                }
                            }
                        });
                    }
                });
            }
        }

        /**
         * Cancels a recognition. This will stop the recognition process if
         * necessary, and report an error to the listener (if applicable).
         */
        void cancel()
        {
            stop(true);
            if(_mode == RecognitionMode.ADAPTATION)
            {
                handleAdaptationStep(new ElvisAdaptationResult(false, null));
            }
            else
            {
                handleError(new ElvisError(ElvisError.Reason.CANCELED));
            }
        }

        // TODO: Create a common utility for this method?
        private List<AudioChunk> resizeAudioChunks(List<AudioChunk> chunks, int maxDurationMs)
        {
            List<AudioChunk> retChunks = new ArrayList<AudioChunk>();
            for (AudioChunk chunk : chunks)
            {
                if (chunk.audioDuration <= maxDurationMs)
                {
                    // Add the chunk as is
                    retChunks.add(chunk);
                }
                else
                {
                    // Slice it in smaller chunks
                    int offset = 0;
                    int targetDurationSamples = chunk.audioType.getSampleCount(maxDurationMs);
                    while (offset < chunk.audioShorts.length)
                    {
                        short[] shorts;
                        int remainingSamples = chunk.audioShorts.length - offset;
                        if (remainingSamples >= targetDurationSamples)
                        {
                            shorts = new short[targetDurationSamples];
                        }
                        else
                        {
                            shorts = new short[remainingSamples];
                        }
                        System.arraycopy(chunk.audioShorts, offset, shorts, 0, shorts.length);
                        offset = offset + shorts.length;
                        retChunks.add(new AudioChunk(chunk.audioType, shorts));
                    }
                }
            }
            return retChunks;
        }

        // Get the sound level in RMS and supply it to the client in dB.
        private float getSoundLevel()
        {
            double level = 0;
            int rms = _nativeElvis.getSoundLevel();
            if (rms >= 1) {
                level = (20*Math.log10(rms));
                if (level > 90) {
                    level = 90; // Clamp at 90.
                }
            }
            return (float) level;
        }

        private void handleNewAudio(final AudioSource<AudioChunk> source)
        {
            final int frameDurationMs = 20; // Elvis needs 20ms of audio.
            List<AudioChunk> chunks = resizeAudioChunks(source.getAllAudioChunksForSink(this), frameDurationMs);
            for (int i = 0; i < chunks.size(); i++)
            {
                AudioChunk chunk = chunks.get(i);

                final short [] audio;
                final long audioTimestamp;

                if (_prevAudio != null)
                {
                    audio = new short[_prevAudio.length + chunk.audioShorts.length];
                    audioTimestamp = _prevAudioTimestamp;
                    System.arraycopy(_prevAudio, 0, audio, 0, _prevAudio.length);
                    System.arraycopy(chunk.audioShorts, 0, audio, _prevAudio.length, chunk.audioShorts.length);
                    _prevAudio = null;
                }
                else
                {
                    audio = chunk.audioShorts;
                    audioTimestamp = chunk.audioTimestamp;
                }

                if (audio.length < chunk.audioType.getSampleCount(frameDurationMs))
                {
                    _prevAudio = audio;
                    _prevAudioTimestamp = audioTimestamp;
                    continue;
                }

                _workerThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (_speechEnded)
                            return;

                        int tempSpeechType;
                        boolean wakeupEnded = false;
                        final ElvisResult result;
                        if (_mode == RecognitionMode.ENDPOINTING)
                        {
                            tempSpeechType = _nativeElvis.processEndpointingAudio(audio);
                            _speechEnded = tempSpeechType == NativeElvis.SPEECH_ENDED;
                            result = null;
                        }
                        else if (_mode == RecognitionMode.WAKEUP)
                        {
                            int wakeupStatus = _nativeElvis.processWakeupAudio(audio, audioTimestamp);
                            if (_lastWakeupStatus != wakeupStatus)
                            {
                                result = _nativeElvis.getWakeupResult(_seamlessWakeup);
                                switch (wakeupStatus)
                                {
                                    case NativeElvis.WUW_NO_WAKEUP:
                                    case NativeElvis.WUW_PENDING_WAKEUP:
                                        tempSpeechType = NativeElvis.SILENCE;
                                        break;

                                    case NativeElvis.WUW_WAKEUP:
                                        if (result == null)
                                        {
                                            Logger.error(this, "No wake-up result.");
                                            tempSpeechType = NativeElvis.SPEECH_ERROR;
                                            wakeupEnded = true;
                                        }
                                        else if (_seamlessWakeup)
                                        {
                                            tempSpeechType = NativeElvis.SILENCE;
                                            _mainThreadHandler.post(new Runnable()
                                            {
                                                @Override
                                                public void run()
                                                {
                                                    if (_currentRecognition != Recognition.this)
                                                    {
                                                        return;
                                                    }

                                                    if (_resultListener != null)
                                                    {
                                                        _resultListener.onResult(result);
                                                    }
                                                }
                                            });
                                        }
                                        else
                                        {
                                            tempSpeechType = NativeElvis.SPEECH_ENDED;
                                            wakeupEnded = true;
                                        }
                                        break;

                                    case NativeElvis.WUW_COMMAND_STARTED:
                                        tempSpeechType = NativeElvis.SPEECH;
                                        _mainThreadHandler.post(new Runnable()
                                        {
                                            @Override
                                            public void run()
                                            {
                                                if (_currentRecognition != Recognition.this)
                                                {
                                                    return;
                                                }

                                                if (_resultListener != null)
                                                {
                                                    _resultListener.onResult(result);
                                                }
                                            }
                                        });
                                        break;

                                    case NativeElvis.WUW_COMMAND_ENDED:
                                    case NativeElvis.WUW_COMMAND_TIMEOUT:
                                        tempSpeechType = NativeElvis.SPEECH_ENDED;
                                        wakeupEnded = true;
                                        break;

                                    case NativeElvis.WUW_ERROR:
                                    default:
                                        Logger.error(this, "Unexpected wake-up loop status: " + wakeupStatus);
                                        tempSpeechType = NativeElvis.SPEECH_ERROR;
                                        wakeupEnded = true;
                                        break;
                                }
                            }
                            else
                            {
                                // No wake-up state transition, keep looping
                                result = null;
                                tempSpeechType = wakeupStatus == NativeElvis.WUW_COMMAND_STARTED ? NativeElvis.SPEECH : NativeElvis.SILENCE;
                            }
                            _lastWakeupStatus = wakeupStatus;
                        }
                        else
                        {
                            tempSpeechType = _nativeElvis.processAudio(audio);
                            _speechEnded = tempSpeechType == NativeElvis.SPEECH_ENDED;
                            result = _speechEnded ? _nativeElvis.getResult(_recogGrammar) : null;
                        }

                        // Pass back the signal information for the current audio packet.
                        final float finalLevel = getSoundLevel();
                        final boolean finalSpeech = (tempSpeechType == NativeElvis.SPEECH || tempSpeechType == NativeElvis.SPEECH_ENDED);
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                if (_currentRecognition != Recognition.this)
                                {
                                    return;
                                }

                                if (_audioEnergyListener != null)
                                {
                                    _audioEnergyListener.onEnergyLevelAvailable(finalLevel, finalSpeech);
                                }
                            }
                        });

                        /* Note: for WakeupMode, we need to make sure to call finishWakeupMode after we get the
                         * result, or memory will not  be freed, and the next call to startWakeupMode or
                         * startRecognition will fail. */
                        if (_mode == RecognitionMode.WAKEUP && wakeupEnded) {
                            _nativeElvis.finishWakeupMode(_stateFileName);
                            tempSpeechType = NativeElvis.SPEECH_ENDED;
                            _speechEnded = true;
                        }
                        ElvisAdaptationResult adaptResult = null;
                        if (_mode == RecognitionMode.ADAPTATION && _speechEnded) {
                            adaptResult = _nativeElvis.processAdaptation();
                            if (adaptResult.getStatus() && _terminateAdaptationSession)
                            {
                                if (!_nativeElvis.finishAdapting(_stateFileName, false))
                                    adaptResult = new ElvisAdaptationResult(false, adaptResult.getMiscLogging());
                            }
                        }
                        updateRecognitionLoggingStatus();

                        final int speechType = tempSpeechType;
                        if (speechType != NativeElvis.SILENCE) {
                            final ElvisAdaptationResult finalAdaptResult = adaptResult;
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (_currentRecognition != Recognition.this)
                                    {
                                        return;
                                    }

                                    if (speechType == NativeElvis.SPEECH)
                                    {
                                        if (_speechStartListener != null)
                                        {
                                            Logger.debug(ElvisRecognizerBase.this, "Elvis detected start of speech");
                                            _speechStartListener.onStartOfSpeech();
                                            _speechStartListener = null;
                                        }
                                    }
                                    else /* SPEECH_ENDED or SPEECH_ERROR */
                                    {
                                        disconnectAudioSource();
                                        if (speechType == NativeElvis.SPEECH_ENDED) {
                                            Logger.debug(ElvisRecognizerBase.this, "Elvis detected end of speech.");
                                            if (_speechEndListener != null) {
                                                _speechEndListener.onEndOfSpeech();
                                                _speechEndListener = null;
                                            }
                                        } else {
                                            Logger.debug(ElvisRecognizerBase.this, "Elvis speech error.");
                                        }

                                        if(_mode == RecognitionMode.ADAPTATION)
                                        {
                                           handleAdaptationStep(finalAdaptResult);
                                        }
                                        else
                                        {
                                            if (result == null)
                                                handleError(new ElvisError(ElvisError.Reason.NO_RESULT));
                                            else
                                                handleResult(result);
                                        }

                                        // We're done, so clean things up
                                        // TODO: because of the thread switching,
                                        // there may be a few pending audio chunks
                                        // that get discarded. Ideally we can account
                                        // for these, so that remaining audio after
                                        // the Elvis recognition can be send somewhere
                                        // else.
                                        _started = false;
                                    }
                                }
                            });
                        }
                    }
                });
            }
        }

        /** Called from worker thread */
        private void handleStartError()
        {
            _mainThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    _started = false;
                    stop(true);
                    if (_mode == RecognitionMode.ADAPTATION)
                        handleAdaptationStep(new ElvisAdaptationResult(false, null));
                    else
                        handleError(new ElvisError(ElvisError.Reason.CANT_START));
                }
            });
        }

        /** Called from main thread */
        private void handleResult(ElvisResult result)
        {
            if (_currentRecognition == Recognition.this)
            {
                Logger.verbose(this, "[LATCHK] Engine combo wake-up latency: " + (SystemClock.uptimeMillis() - result.getWakeupEndTimestamp()) + " ms.");

                _currentRecognition = null;
                if (_resultListener != null)
                {
                    _resultListener.onResult(result);
                    _resultListener = null;
                }
            }
        }

        /** Called from main thread */
        private void handleError(ElvisError error)
        {
            if (_currentRecognition == Recognition.this)
            {
                _currentRecognition = null;
                if (_resultListener != null)
                {
                    _resultListener.onError(error);
                    _resultListener = null;
                }
            }
        }

        /** Called from main thread */
        private void handleAdaptationStep(ElvisAdaptationResult result)
        {
            if (_currentRecognition == Recognition.this)
            {
                _currentRecognition = null;
                if (_adaptationStepListener != null)
                {
                    if (_adaptationStepListener instanceof AdaptationProgressListener)
                        ((AdaptationProgressListener)_adaptationStepListener).onComplete(result);
                    else
                        _adaptationStepListener.onComplete(result.getStatus());
                    _adaptationStepListener = null;
                }
            }
        }

        @Override
        public void chunksAvailable(AudioSource<AudioChunk> source)
        {
            if (_currentRecognition != this) return;
            handleNewAudio(source);
            if (!source.isActive())
            {
                stop(false);
            }
        }

        @Override
        public void framesDropped(AudioSource<AudioChunk> source)
        {
            if (_currentRecognition != this) return;
            if(_mode == RecognitionMode.ADAPTATION)
            {
                handleAdaptationStep(new ElvisAdaptationResult(false, null));
            }
            else
            {
                handleError(new ElvisError(ElvisError.Reason.AUDIO_FRAMES_DROPPED));
            }
        }

        @Override
        public void sourceClosed(AudioSource<AudioChunk> source)
        {
            if (_currentRecognition != this) return;
            stop(false);
        }
    }

    /**
     * Internal class that keeps track of a single grammar rebuild.
     */
    private class Rebuild
    {
        private final Grammar _rebuildGrammar; // The grammar to build
        private final Grammar _newGrammar;     // The new grammar (starts as a copy of the old grammar)
        private final ArrayList<WordSlot.WordIterator> _savedWordSlots;
        private final RebuildListener _listener;
        private boolean _started;
        private final List<SkippedWord> _skippedWords;

        Rebuild(Grammar grammar, Grammar oldGrammar, RebuildListener listener)
        {
            _newGrammar = new Grammar();
            if (oldGrammar != null) {
                _newGrammar.addConstraints(oldGrammar.getConstraints(), true);
                _newGrammar.addWordSlots(oldGrammar.getSlots(), true);
            }
            _rebuildGrammar = grammar;
            _savedWordSlots = new ArrayList<WordSlot.WordIterator>();
            _listener = listener;
            _skippedWords = new ArrayList<SkippedWord>();
        }

        /**
         * Cancel a rebuild that is in progress. It may not stop immediately.
         */
        void cancel()
        {
            if (_started)
            {
                synchronized (this)
                {
                    _started = false;
                }
            }
        }

        /**
         * Starts the rebuild process.<p>
         * 1. Runs a job to add all constraints. This must occur before adding
         * word classes, because when an existing constraint is added, it is
         * actually deleted + re-added. This could result in automatically
         * deleting word classes that were just added. <br>
         * 2. Runs a number of jobs to add all word classes. <br>
         * 3. Runs a job to add all transitions. <br>
         * 4. Notifies of completion on the main thread.
         * @param forceFullRebuild true to force a full rebuild
         */
        void start(final boolean forceFullRebuild)
        {
            Assert.assertFalse(_started);

            synchronized (this)
            {
                _started = true;
            }

            final ArrayList<Constraint> transitionsToAdd = new ArrayList<Constraint>();
            // Transition Job is also to synchronize back with content list
            final Runnable transitionsJob = new Runnable()
            {
                @Override
                public void run()
                {
                    for (Constraint c : transitionsToAdd)
                    {
                        String constraint = c.getId();
                        for (Constraint.Transition t : c.getTransitions())
                        {
                            if (NativeElvis.TRANSITION_ADDED != _nativeElvis.addTransition(constraint, t.from, t.to, t.penalty))
                            {
                                Logger.error(ElvisRecognizerBase.this, "Error adding transition to native Elvis: " + constraint + " from: " +
                                                                       t.from + " to: " + t.to + " penalty: " + t.penalty);
                                _nativeElvis.deleteConstraint(constraint);
                                handleError(new ElvisError(ElvisError.Reason.TRANSITION_DEF_ERROR));
                                return;
                            }
                        }
                    }

                    // Finished, so save and notify the app
                    if (!_nativeElvis.saveGrammar(_stateFileName, _newGrammar))
                    {
                        Logger.error(ElvisRecognizerBase.this, "Error saving Elvis state to: " + _stateFileName);
                        handleError(new ElvisError(ElvisError.Reason.FILE_ERROR));
                        return;
                    }

                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            if (_started)
                            {
                                _started = false;

                                for (WordSlot.WordIterator it : _savedWordSlots)
                                    it.finished(true);

                                _savedWordSlots.clear();
                                _grammar.addConstraints(_newGrammar.getConstraints(), true);
                                _grammar.addWordSlots(_newGrammar.getSlots(), true);

                                if (_currentRebuild == Rebuild.this)
                                    _currentRebuild = null;

                                _listener.onComplete(_rebuildGrammar, _skippedWords);
                            }
                        }
                    });
                }
            };

            final Runnable wordsJob = new Runnable()
            {
                private Iterator<WordSlot> _slots = initSlots();
                private WordSlot.WordIterator _words;
                private String _slotName;

                private Iterator<WordSlot> initSlots()
                {
                    // Add the special endpointing wordslot
                    ArrayList<WordSlot> slots = new ArrayList<WordSlot>(_rebuildGrammar.getSlots().size());
                    slots.addAll(_rebuildGrammar.getSlots());
                    return slots.iterator();
                }

                @Override
                public void run()
                {
                    synchronized (Rebuild.this)
                    {
                        if (!_started)
                            return;
                    }
                    if (!_isElvisCreated) {
                        handleError(new ElvisError(ElvisError.Reason.NOT_INITIALIZED));
                        return;
                    }

                    if (_words == null || !_words.hasNext())
                    {
                        if (_slots.hasNext())
                        {
                            WordSlot nextSlot = _slots.next();
                            _slotName = nextSlot.getId();
                            WordSlot oldSlot = _newGrammar.getWordSlotByName(_slotName);

                            if (nextSlot.getWordListRequired() && !nextSlot.hasWordList()) {
                                Logger.error(ElvisRecognizerBase.this, "WordSlot " + nextSlot.getId() + " requires an attached wordlist.");
                                handleError(new ElvisError(ElvisError.Reason.MISSING_WORDLIST));
                                return;
                            }

                            if ((oldSlot == null) ||
                                (nextSlot.getRebuildType() == WordSlot.RebuildModes.WORDSLOT_REBUILD_FULL))
                            {
                                // This WordSlot does not exist, or should be over-written.
                                _newGrammar.addWordSlot(nextSlot.copy(), true);
                            } else if (oldSlot.getType() != nextSlot.getType())
                            {
                                if (nextSlot.getType() != WordSlot.Types.WORDSLOT_DEFAULT) {
                                    // Do no allow appending to a WordSlot of different type.
                                    // However, if the new WordSlot type is "WORDSLOT_DEFAULT",
                                    // we can append to any WordSlot type.
                                    Logger.error(ElvisRecognizerBase.this, "WordSlot types do not match:" + _slotName + ". Old: " +
                                            oldSlot.getType() + " New: " + nextSlot.getType());
                                    handleError(new ElvisError(ElvisError.Reason.WORDSLOT_DEF_ERROR));
                                    return;
                                }
                            }

                            // It should be safe to add existing classes
                            int stat = _nativeElvis.addWordClass(nextSlot);
                            if (stat == NativeElvis.WORDCLASS_ERROR)
                            {
                                handleError(new ElvisError(ElvisError.Reason.WORDSLOT_DEF_ERROR));
                                return;
                            }
                            else if (stat == NativeElvis.WORD_ADDED || forceFullRebuild)
                            {
                                // WORD_ADDED means that the word slot was newly added in Elvis,
                                // or that all existing members were deleted if the slot already existed.
                                // forceFullRebuild means we should always load all members.
                                // In all three case, we need to do a full rebuild.
                                _words = nextSlot.getWords(true);
                                _savedWordSlots.add(_words);
                            } else if (nextSlot.getRebuildType() == WordSlot.RebuildModes.WORDSLOT_REBUILD_UPDATE) {
                                // This is not a rebuild, and the WORDSLOT_REBUILD_UPDATE will add new words and remove
                                // deleted words
                                _words = nextSlot.getWords(false);
                                _savedWordSlots.add(_words);
                            }
                            // else {
                            //     If this is not a full rebuild, and the WordClass already exists,
                            //     and the rebuild type is WORDSLOT_REBUILD_SKIP_EXISTING, there is nothing to do.
                            // }

                            _workerThreadHandler.post(this);
                        } else
                        {
                            // Word slots are done, so start adding transitions
                            _workerThreadHandler.post(transitionsJob);
                        }
                    }
                    else
                    {
                        WordSlot slot = _newGrammar.getWordSlotByName(_slotName);
                        boolean addVariants = false;
                        if (slot != null && slot.useVariants())
                        {
                            addVariants = true;
                        }
                        for (int i = 0; i < WORDCLASS_BATCH_SIZE && _words.hasNext(); i++)
                        {
                            int stat = NativeElvis.WORD_ADDED;
                            WordAction wa = _words.getNext();
                            if (wa.isAdded())
                            {
                                stat = _nativeElvis.addWordClassMember(_slotName, wa.getWord(), addVariants);
                            }
                            else
                            {
                                if(_nativeElvis.removeWordClassMember(_slotName, wa.getWord()))
                                    Logger.debug(this, "Success removing word class member: " + wa.getWord().getSurfaceForm());
                                else
                                    Logger.error(this, "Failed removing word class member: " + wa.getWord().getSurfaceForm());
                            }
                            if (stat == NativeElvis.WORD_ERROR)
                            {
                                handleError(new ElvisError(ElvisError.Reason.WORDSLOT_DEF_ERROR));
                                return;
                            } else if (stat == NativeElvis.WORD_NO_PRONUNCIATION) {
                                _skippedWords.add(new SkippedWord(wa.getWord(), _slotName));
                            }
                        }
                        _workerThreadHandler.post(this);
                    }
                }
            };

            final Runnable constraintJob = new Runnable()
            {
                @Override
                public void run()
                {
                    synchronized (Rebuild.this)
                    {
                        if (!_started)
                            return;
                    }
                    ArrayList<Constraint> constraints = new ArrayList<Constraint>(_rebuildGrammar.getConstraints().size());
                    constraints.addAll(_rebuildGrammar.getConstraints());

                    for (Constraint c : constraints)
                    {
                        int constraintStat = NativeElvis.CONSTRAINT_ERROR;
                        String constraint = c.getId();
                        Constraint oldConstraint = _newGrammar.getConstraintByName(constraint);
                        if (!c.getAlwaysRebuild() && (oldConstraint != null))
                        {
                            /* Skip existing constraints that are not "always rebuild". */
                            continue;
                        }
                        if (c.isValid()) {
                            constraintStat = _nativeElvis.addConstraint(constraint, c.getClassCount(), c.getTransitionCount(), c.hasPenalties(), c.getAlwaysRebuild());
                        } else {
                            Logger.error(this, "Constraint not valid: " + c.getId() + ".  Skipping.");
                        }
                        if (constraintStat == NativeElvis.CONSTRAINT_ADDED)
                        {
                            // Delay adding the transitions until the word classes are added
                            transitionsToAdd.add(c);

                            // Add new Constraints to ElvisRecognizer grammar.
                            _newGrammar.addConstraint(c, true);
                        } else if (constraintStat == NativeElvis.CONSTRAINT_NOT_ADDED){
                            Logger.debug(ElvisRecognizerBase.this, "Constraint already exists in native Elvis: " + constraint);
                        } else {
                            Logger.error(ElvisRecognizerBase.this, "Error adding constraint to native Elvis: " + constraint);
                            handleError(new ElvisError(ElvisError.Reason.CONSTRAINT_DEF_ERROR));
                            return;
                        }
                    }

                    // Constraints are done, so start adding word classes
                    _workerThreadHandler.post(wordsJob);
                }
            };

            // Kick off rebuild by building jobs
            _workerThreadHandler.post(constraintJob);
        }

        /** Call this from the WORKER THREAD */
        private void handleError(final ElvisError error)
        {
            _mainThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    if (_started)
                    {
                        _started = false;

                        for (WordSlot.WordIterator it : _savedWordSlots)
                            it.finished(false);

                        _savedWordSlots.clear();
                        _listener.onError(error);
                        if (_currentRebuild == Rebuild.this) {
                            _currentRebuild = null;
                        }
                    }
                }
            });
        }
    }

    private final NativeElvis _nativeElvis;
    private Grammar _grammar;
    protected final NMTHandler _mainThreadHandler;
    private final NMTHandler _workerThreadHandler;
    private final ArrayList<String> _activeConstraints;
    private final List<String> _wakeupPhrases;
    private int _wakeUpPenalty;
    private int _recognitionLoggingStatus;
    private String _recognitionLoggingPrefix;
    private String _recognitionLoggingFolder;
    private String _fxLoggingFolder;
    private Recognition _currentRecognition;
    private Rebuild _currentRebuild;
    private String _stateFileName; // Touched from worker thread
    private boolean _isElvisCreated; // Should only be accessed from worker thread
    private boolean _isReleased;
    private boolean _isInitialized;

    /**
     * The ElvisRecognizer class encapsulates an instance of the Elvis
     * recognition engine.
     */
    ElvisRecognizerBase(NativeElvis nativeElvis, NMTHandler externalWorkerThreadHandler)
    {
        _nativeElvis = nativeElvis;
        _mainThreadHandler = Factory.createNMTHandler();
        _grammar = new Grammar();
        _activeConstraints = new ArrayList<String>();
        _wakeupPhrases =  new ArrayList<String>(0);
        _wakeUpPenalty = 0;

        if (externalWorkerThreadHandler == null)
            _workerThreadHandler = getWorkerThreadHandler();
        else
            _workerThreadHandler = externalWorkerThreadHandler;
    }

    @Override
    public List<ElvisModelInfo> getAvailableModels()
    {
        ArrayList<NativeModelInfo> nativeModels = new ArrayList<NativeModelInfo>();

        boolean success = _nativeElvis.listAvailableModels(nativeModels);
        if (success)
        {
            ArrayList<ElvisModelInfo> models = new ArrayList<ElvisModelInfo>(nativeModels.size());
            for (NativeModelInfo nativeModel : nativeModels)
            {
                int rateHz = nativeModel.rateKhz * 1000;
                String quality = nativeModel.quality;
                if (rateHz == ElvisRecognizer.Frequencies.FREQ_16KHZ ||
                    rateHz == ElvisRecognizer.Frequencies.FREQ_8KHZ)
                {
                    for (ElvisLanguage lang : ElvisRecognizer.Languages.ALL_LANGUAGES)
                    {
                        if (lang.name.equals(nativeModel.language))
                        {
                            models.add(new ElvisModelInfo(lang, quality, rateHz));
                        }
                    }
                }
            }
            return models;
        }
        Logger.error(this, "Unable to detect supported Elvis models");
        return new ArrayList<ElvisModelInfo>(0);
    }

    @Override
    public List<ElvisModelInfo> getAvailableModels(ElvisLanguage language)
    {
        List<ElvisModelInfo> models = getAvailableModels();
        for (int i = 0; i < models.size(); i++)
        {
            ElvisModelInfo info = models.get(i);
            if (!info.language.equals(language)) {
                models.remove(i);
                i--;
            }
        }
        return models;
    }

    @Override
    public List<ElvisModelInfo> getAvailableModels(int freqHz)
    {
        List<ElvisModelInfo> models = getAvailableModels();
        for (int i = 0; i < models.size(); i++)
        {
            ElvisModelInfo info = models.get(i);
            if (info.frequencyHz != freqHz) {
                models.remove(i);
                i--;
            }
        }
        return models;
    }

    @Override
    public void initialize(final ElvisConfig config, final String stateFileName, final InitializeListener listener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, !_isInitialized, "Already initialized");
        Checker.checkStringArgForNullOrEmpty("stateFileName", stateFileName);
        Checker.checkArgForNull("config", config);

        _isInitialized = true;
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _stateFileName = stateFileName;

                String language = config.language;
                int samplingRate = config.samplingRate;
                String binfile = config.binfile;

                // Find the best quality possible
                String quality = ElvisQuality.UNSPECIFIED.equals(config.quality) ? getQuality(language, samplingRate) : config.quality;

                if (binfile == null && language != null)
                    binfile = LangSpecs.getBinfileName(language, quality, samplingRate / 1000);

                final boolean success = _nativeElvis.createElvis(language, samplingRate, binfile);
                if (success) {
                    _nativeElvis.loadAdaptation(_stateFileName, ".adp", false);
                    _nativeElvis.loadAdaptation(_stateFileName, ".wuwadp", true);
                    _isElvisCreated = true;
                }
                _mainThreadHandler.post(new Runnable() {
                    @Override
                    public void run()
                    {
                        if (listener != null) {
                            listener.onLoaded(ElvisRecognizerBase.this, success);
                        }
                    }
                });
            }
        });
    }

    @Override
    public void initialize(ElvisConfig config, String stateFileName)
    {
        initialize(config, stateFileName, null);
    }

    @Override
    public void release(final ReleaseListener listener)
    {
        Checker.checkState(this, !_isReleased);
        if (!_isInitialized) {
            _isReleased = true;
            if (listener != null)
                listener.onReleased(ElvisRecognizerBase.this);
            return;
        }

        terminateCurrentRecognition();
        terminateAdaptation();
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (_isElvisCreated) {
                    _nativeElvis.destroyElvis();
                }
                _isElvisCreated = false;
                _isReleased = true;
                if (listener != null) {
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            listener.onReleased(ElvisRecognizerBase.this);
                        }
                    });
                }
            }
        });
    }

    @Override
    public void release()
    {
        release(null);
    }

    @Override
    public void saveState(final SaveListener listener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        final Grammar grammar = _grammar;

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final boolean success = _nativeElvis.saveGrammar(_stateFileName, grammar);
                if (listener != null)
                    _mainThreadHandler.post(new Runnable()
                                            {
                                                @Override
                                                public void run()
                                                {
                                                    listener.onSaved(ElvisRecognizerBase.this, success);
                                                }
                                            });
            }
        });
    }

    @Override
    public void saveState()
    {
        saveState(null);
    }

    @Override
    public void loadState(final RebuildListener rebuildListener)
    {
        loadStateAndGrammar(null, rebuildListener);
    }

    @Override
    public void loadStateAndGrammar(final Grammar grammar,
                                    final RebuildListener rebuildListener)
    {
        Checker.checkArgForNull("rebuildListener", rebuildListener);
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelRecognition();

        _grammar.clearConstraints();
        _grammar.clearWordSlots();

        // Create, but don't start
        final Rebuild rebuild;
        if (grammar != null)
            rebuild = new Rebuild(grammar, _grammar, rebuildListener);
        else
            rebuild = null;
        _currentRebuild = rebuild; // Set this now, so we can tell if it's canceled

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (!_isElvisCreated) {
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            if (rebuild == _currentRebuild)
                            {
                                Logger.error(ElvisRecognizerBase.this, "Cannot load Elvis State or Grammar.  Elvis is not initialized. ");
                                rebuildListener.onError(new ElvisError(ElvisError.Reason.NOT_INITIALIZED));
                            }
                        }
                    });
                    return;
                }
                // Elvis does not support loading the state file twice, or loading the state file
                // after constraints or word classes have been added.  Recreate Native Elvis
                // if necessary so that we can load the state file.
                final String stateFileName = _stateFileName;
                _nativeElvis.recreateElvis(false);
                // only loads adaptation psa once to the same elvis, it seems there is a bug on native elvis
                //_nativeElvis.loadAdaptation(_stateFileName, ".adp", false);
                //_nativeElvis.loadAdaptation(_stateFileName, ".wuwadp", true);
                final Grammar loadedGrammar = _nativeElvis.loadGrammar(_stateFileName);
                final List<SkippedWord> _skippedWords = new ArrayList<SkippedWord>();

                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (rebuild == _currentRebuild)
                        {
                            if (loadedGrammar == null && grammar == null)
                            {
                                Logger.error(ElvisRecognizerBase.this, "Error loading Elvis state from: " + stateFileName);
                                rebuildListener.onError(new ElvisError(ElvisError.Reason.FILE_ERROR));
                            } else if (grammar != null)
                            {
                                // refresh rebuild object
                                _currentRebuild = new Rebuild(grammar, loadedGrammar, rebuildListener);

                                // If no loaded grammar was loaded, make sure we do a full rebuild
                                if (loadedGrammar == null) {
                                    Logger.warn(ElvisRecognizerBase.this, "Unable to load Elvis state from: " + stateFileName + ", doing a full grammar rebuild");
                                    _currentRebuild.start(true);
                                } else
                                {
                                    Logger.debug(ElvisRecognizerBase.this, "Elvis state loaded, now building grammar");
                                    _currentRebuild.start(false);
                                }
                            } else
                            {
                                _grammar = loadedGrammar;
                                rebuildListener.onComplete(loadedGrammar, _skippedWords);
                            }
                        }
                    }
                });
            }
        });
    }

    @Override
    public void loadGrammar(Grammar grammar,
                            RebuildListener listener)
    {
        Checker.checkArgForNull("grammar", grammar);
        Checker.checkArgForNull("listener", listener);
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelRecognition();

        _currentRebuild = new Rebuild(grammar, _grammar, listener);
        _currentRebuild.start(false);
    }

    @Override
    public void clearGrammar(final RebuildListener listener)
    {
        cancelRebuild();
        cancelRecognition();
        final List<SkippedWord> _skippedWords = new ArrayList<SkippedWord>();

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (!_isElvisCreated){
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            if (listener != null) {
                                listener.onError(new ElvisError(ElvisError.Reason.NOT_INITIALIZED));
                            }
                        }
                    });
                    return;
                }

                final boolean success = _nativeElvis.clearGrammar();
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (listener != null) {
                            if (success) {
                                listener.onComplete(_grammar, _skippedWords);
                            } else {
                                listener.onError(new ElvisError(ElvisError.Reason.CONSTRAINT_DEF_ERROR));
                            }
                        }
                        _grammar = new Grammar();
                    }
                });
            }
        });
    }

    @Override
    public void clearAdaptation(final AdaptationStepListener listener)
    {
        clearAdaptation(false, listener);
    }

    @Override
    public void clearAdaptation(final boolean all, final AdaptationStepListener listener)
    {
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (!_isElvisCreated) {
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            if (listener != null) {
                                if (listener instanceof AdaptationProgressListener)
                                    ((AdaptationProgressListener)listener).onComplete(new ElvisAdaptationResult(false, null));
                                else
                                    listener.onComplete(false);
                            }
                        }
                    });
                    return;
                }

                final boolean success = _nativeElvis.clearAdaptation(_stateFileName, ".adp", all);
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (listener != null) {
                            if (listener instanceof AdaptationProgressListener)
                                ((AdaptationProgressListener)listener).onComplete(new ElvisAdaptationResult(success, null));
                            else
                                listener.onComplete(success);
                        }
                    }
                });
            }
        });
    }

    @Override
    public void clearWuWAdaptation(final AdaptationProgressListener listener)
    {
        clearWuWAdaptation(false, listener);
    }

    @Override
    public void clearWuWAdaptation(final boolean all, final AdaptationProgressListener listener)
    {
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (!_isElvisCreated) {
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            if (listener != null) {
                                listener.onComplete(new ElvisAdaptationResult(false, null));
                            }
                        }
                    });
                    return;
                }

                final boolean success = _nativeElvis.clearAdaptation(_stateFileName, ".wuwadp", all);
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (listener != null) {
                            listener.onComplete(new ElvisAdaptationResult(success, null));
                        }
                    }
                });
            }
        });
    }


    @Override
    public void cancelRebuild()
    {
        if (_currentRebuild != null)
        {
            _currentRebuild.cancel();
            _currentRebuild = null;
        }
    }

    @Override
    public Constraint findConstraint(String id)
    {
        return _grammar.getConstraintByName(id);
    }

    @Override
    public WordSlot findWordSlot(String id)
    {
        return _grammar.getWordSlotByName(id);
    }

    @Override
    public void setActiveConstraints(List<Constraint> activeConstraints)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        _activeConstraints.clear();
        if (activeConstraints != null)
        {
            _activeConstraints.ensureCapacity(activeConstraints.size());
            for (Constraint c : activeConstraints)
            {
                _activeConstraints.add(c.getId());
            }
        }
    }

    @Override
    public void setActiveConstraintNames(List<String> activeConstraintNames)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        _activeConstraints.clear();
        if (activeConstraintNames != null)
        {
            _activeConstraints.addAll(activeConstraintNames);
        }
    }

    @Override
    public void setCustomPronunciations(final List<Word> words)
    {
        Checker.checkArgForNull("words", words);
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        Checker.checkState(this, _currentRebuild == null, "A rebuild is in progress");

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (_isElvisCreated)
                {
                    for (Word word : words)
                    {
                        if(_nativeElvis.addExternalPronunciation(word) != NativeElvis.WORD_ADDED)
                            Logger.error(this, "Unable to set custom pronunciation for " + word.toString());
                    }
                }
                else
                {
                    Logger.error(this, "Unable to set custom pronunciations. Elvis is not created.");
                }
            }
        });
    }

    @Override
    public void setWakeupPhrases(List<String> wakeupPhrases)
    {
        _wakeupPhrases.clear();
        if (wakeupPhrases != null)
            _wakeupPhrases.addAll(wakeupPhrases);
    }

    @Override
    public void setWakeupPenalty(int penalty)
    {
        _wakeUpPenalty = penalty;
    }

    @Override
    public final void startRecognition(AudioSource<AudioChunk> audioSource,
                                       SpeechDetectionListener speechDetectionListener,
                                       ResultListener resultListener)
    {
        startRecognition(audioSource, new ArrayList<String>(_activeConstraints), speechDetectionListener, null, resultListener);
    }

    @Override
    public void startRecognition(AudioSource<AudioChunk> audioSource,
                                 List<String> activeConstraintNames,
                                 SpeechDetectionListener speechDetectionListener,
                                 ResultListener resultListener)
    {
        startRecognition(audioSource, activeConstraintNames, speechDetectionListener, null, resultListener);
    }

    @Override
    public void startRecognition(AudioSource<AudioChunk> audioSource,
                                 List<String> activeConstraintNames,
                                 SpeechDetectionListener speechDetectionListener,
                                 AudioEnergyListener audioEnergyListener,
                                 ResultListener resultListener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        Checker.checkState(this, _currentRebuild == null, "A rebuild is in progress");

        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForNull("activeConstraintNames", activeConstraintNames);
        Checker.checkArgForNull("resultListener", resultListener);

        if (_currentRebuild == null)
        {
            cancelRecognition();

            _currentRecognition = new Recognition(_grammar, resultListener, speechDetectionListener, audioEnergyListener, null);
            _currentRecognition.startRecognizing(audioSource, activeConstraintNames);
        }
    }

    @Override
    public void checkAdaptationStatus(final AdaptationCheckListener adaptationCheckListener)
    {
        Checker.checkArgForNull("adaptationCheckListener", adaptationCheckListener);
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        _workerThreadHandler.post(new Runnable() {

            @Override
            public void run()
            {
                if (!_isElvisCreated) {
                    _mainThreadHandler.post(new Runnable() {
                        @Override
                        public void run()
                        {
                            adaptationCheckListener.onComplete(false);
                        }});
                    return;
                }
                final boolean adapted =_nativeElvis.checkAdaptationStatus(_stateFileName);
                _mainThreadHandler.post(new Runnable() {
                    @Override
                    public void run()
                    {
                        adaptationCheckListener.onComplete(adapted);
                    }});
            }});
    }

    @Override
    public void checkWakeupPhrases(final List<String> wakeupPhrases, final WakeupCheckListener wakeupCheckListener)
    {
        Checker.checkArgForNull("wakeupPhrases", wakeupPhrases);
        Checker.checkArgForNull("wakeupCheckListener", wakeupCheckListener);
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        _workerThreadHandler.post(new Runnable() {

            @Override
            public void run()
            {
                if (!_isElvisCreated) {
                    _mainThreadHandler.post(new Runnable() {
                        @Override
                        public void run()
                        {
                            wakeupCheckListener.onError(new ElvisError(ElvisError.Reason.NOT_INITIALIZED));
                        }});
                    return;
                }
                final ArrayList<String> rejectedPhrases = new ArrayList<String>();
                for (String wakeupPhrase : wakeupPhrases) {
                    if (!_nativeElvis.checkWakeupPhrase(wakeupPhrase)) {
                        rejectedPhrases.add(wakeupPhrase);
                    }
                }
                _mainThreadHandler.post(new Runnable() {
                    @Override
                    public void run()
                    {
                        wakeupCheckListener.onComplete(rejectedPhrases);
                    }});
            }});
    }

    @Override
    public final void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                      ResultListener resultListener)
    {
        startWakeupMode(audioSource, new ArrayList<String>(_wakeupPhrases), _wakeUpPenalty,  0, EngineWakeupMode.REGULAR, resultListener);
    }

    @Override
    public void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                List<String> wakeupPhrases,
                                ResultListener resultListener)
    {
        startWakeupMode(audioSource, wakeupPhrases, 0, 0,  EngineWakeupMode.REGULAR, resultListener);
    }

    @Override
    public void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                List<String> wakeupPhrases,
                                int penalty,
                                ResultListener resultListener)
    {
        startWakeupMode(audioSource, wakeupPhrases, penalty, 0,  EngineWakeupMode.REGULAR, resultListener);
    }

    @Override
    public void startSeamlessWakeupMode(AudioSource<AudioChunk> audioSource,
                                        List<String> wakeupPhrases,
                                        int penalty,
                                        ResultListener resultListener)
    {
        startWakeupMode(audioSource, wakeupPhrases, penalty, 1000, EngineWakeupMode.SEAMLESS, resultListener);
    }

    @Override
    public void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                List<String> wakeupPhrases,
                                int penalty, int pauseDurationMs,
                                ResultListener resultListener)
    {
        startWakeupMode(audioSource, wakeupPhrases, penalty, pauseDurationMs, pauseDurationMs > 0 ? EngineWakeupMode.SEAMLESS : EngineWakeupMode.REGULAR, resultListener);
    }

    @Override
    public void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                List<String> wakeupPhrases,
                                int penalty, int pauseDurationMs, int engineWakeupMode,
                                ResultListener resultListener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        Checker.checkState(this, _currentRebuild == null, "A rebuild is in progress");

        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForNull("wakeupPhrases", wakeupPhrases);
        Checker.checkArgForCondition("pauseDurationMs", "greater than 0", pauseDurationMs >= 0);
        Checker.checkArgForCondition("engineWakeupMode", "REGULAR or SEAMLESS", engineWakeupMode == EngineWakeupMode.REGULAR ||
                                                                                engineWakeupMode == EngineWakeupMode.SEAMLESS);
        Checker.checkArgForNull("resultListener", resultListener);

        if (_currentRebuild == null)
        {
            cancelRecognition();

            _currentRecognition = new Recognition(_grammar, resultListener, null, null, null);
            _currentRecognition.startWakeupMode(audioSource, new WakeupParam.Builder(wakeupPhrases).setPenalty(penalty).setCommandStartBailoutDuration(pauseDurationMs).build());
        }
    }

    public void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                WakeupParam wakeupParam,
                                AudioEnergyListener audioEnergyListener,
                                ResultListener resultListener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        Checker.checkState(this, _currentRebuild == null, "A rebuild is in progress");

        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForNull("wakeupParam", wakeupParam);
        Checker.checkArgForNull("resultListener", resultListener);

        if (_currentRebuild == null)
        {
            cancelRecognition();

            _currentRecognition = new Recognition(_grammar, resultListener, null, audioEnergyListener, null);
            _currentRecognition.startWakeupMode(audioSource, wakeupParam);
        }
    }

    @Override
    public final void startAdaptation(AudioSource<AudioChunk> audioSource,
                                      String phrase,
                                      boolean terminateSession,
                                      SpeechDetectionListener speechDetectionListener,
                                      AdaptationStepListener stepListener)
    {
        startAdaptation(audioSource, phrase, terminateSession, speechDetectionListener, null, stepListener);
    }

    @Override
    public final void startAdaptation(AudioSource<AudioChunk> audioSource,
                                      String phrase,
                                      boolean terminateSession,
                                      SpeechDetectionListener speechDetectionListener,
                                      AudioEnergyListener audioEnergyListener,
                                      AdaptationStepListener stepListener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        Checker.checkState(this, _currentRebuild == null, "A rebuild is in progress");

        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkStringArgForNullOrEmpty("phrase", phrase);
        Checker.checkArgForNull("stepListener", stepListener);

        if (_currentRebuild == null)
        {
            cancelRecognition();

            _currentRecognition = new Recognition(_grammar, null, speechDetectionListener, audioEnergyListener, stepListener);
            _currentRecognition.startAdapting(audioSource, phrase, terminateSession);
        }
    }

    @Override
    public void abortAdaptation()
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        terminateAdaptation();
    }

    @Override
    public final void stopListening()
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        if (_currentRecognition != null)
        {
            _currentRecognition.stop(false);
        }
    }

    @Override
    public final void cancelRecognition()
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        terminateCurrentRecognition();
    }

    @Override
    public final void startSpeechDetection(AudioSource<AudioChunk> audioSource,
                                           SpeechDetectionListener speechDetectionListener)
    {
        // Duration comes from ELVIS team
        startSpeechDetection(audioSource, 1500, 0, true, speechDetectionListener, null);
    }

    @Override
    public final void startSpeechDetection(AudioSource<AudioChunk> audioSource, boolean longSpeech,
                                           SpeechDetectionListener speechDetectionListener)
    {
        // Durations come from ELVIS team
        startSpeechDetection(audioSource, longSpeech ? 1500 : 900, 0, true, speechDetectionListener, null);
    }

    @Override
    public final void startSpeechDetection(AudioSource<AudioChunk> audioSource, int silenceDurationMs,
                                           SpeechDetectionListener speechDetectionListener)
    {
        // Durations come from ELVIS team
        startSpeechDetection(audioSource, silenceDurationMs, 0, true, speechDetectionListener, null);
    }

    @Override
    public final void startSpeechDetection(AudioSource<AudioChunk> audioSource, int silenceDurationMs,
                                           SpeechDetectionListener speechDetectionListener,
                                           AudioEnergyListener audioEnergyListener)
    {
        startSpeechDetection(audioSource, silenceDurationMs, 0, true, speechDetectionListener, audioEnergyListener);
    }

    @Override
    public final void startSpeechDetection(AudioSource<AudioChunk> audioSource, int silenceDurationMs, int ignoreDurationMs, boolean resetChannel,
                                           SpeechDetectionListener speechDetectionListener,
                                           AudioEnergyListener audioEnergyListener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        Checker.checkState(this, _currentRebuild == null, "A rebuild is in progress");

        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForCondition("silenceDurationMs", "> 0", silenceDurationMs > 0);
        Checker.checkArgForCondition("ignoreDurationMs", ">= 0", ignoreDurationMs >= 0);
        Checker.checkArgForNull("speechDetectionListener", speechDetectionListener);

        if (_currentRebuild == null)
        {
            cancelRecognition();

            _currentRecognition = new Recognition(_grammar, null, speechDetectionListener, audioEnergyListener, null);
            _currentRecognition.startEndpointing(audioSource, new EndpointingParam.Builder()
                                                                  .setSilenceDuration(silenceDurationMs)
                                                                  .setIgnoreDuration(ignoreDurationMs)
                                                                  .setResetChannel(resetChannel).build());
        }
    }

    @Override
    public void startSpeechDetection(AudioSource<AudioChunk> audioSource,
                                     EndpointingParam endpointingParam,
                                     SpeechDetectionListener speechDetectionListener,
                                     AudioEnergyListener audioEnergyListener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        Checker.checkState(this, _currentRebuild == null, "A rebuild is in progress");

        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForNull("endpointingParam", endpointingParam);
        Checker.checkArgForNull("speechDetectionListener", speechDetectionListener);

        if (_currentRebuild == null)
        {
            cancelRecognition();

            _currentRecognition = new Recognition(_grammar, null, speechDetectionListener, audioEnergyListener, null);
            _currentRecognition.startEndpointing(audioSource, endpointingParam);
        }
    }

    @Override
    public void enableRecognitionLogging(final String destinationFolder, final String filenamePrefix, final int logMilliseconds)
    {
        enableRecognitionLogging(destinationFolder, filenamePrefix, logMilliseconds, null);
    }

    @Override
    public void enableRecognitionLogging(final String destinationFolder, final String filenamePrefix, final int logMilliseconds, final String uuid)
    {
        Checker.checkState(this, !_isReleased);
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (filenamePrefix == null || logMilliseconds <= 0)
                {
                    _nativeElvis.enableElvisRecognitionLogging(null, 0, null);
                    _fxLoggingFolder = null;
                    _recognitionLoggingFolder = null;
                    _recognitionLoggingPrefix = null;
                } else {
                    if (_nativeElvis.enableElvisRecognitionLogging(filenamePrefix, logMilliseconds, uuid))
                    {
                        _fxLoggingFolder = destinationFolder;
                        _recognitionLoggingFolder = destinationFolder;
                        _recognitionLoggingPrefix = filenamePrefix;
                    }
                }
                updateRecognitionLoggingStatus();
            }
        });
    }

    @Override
    public void createBlob(final BlobActionListener blobActionListener, final List<String> phrases,
                           final String cpu, final int numPhrases,
                           final int decoyPenalty, final boolean isMagic,
                           final int maxDuration)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        Checker.checkArgForNull("blobActionListener", blobActionListener);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (cpu != null)
                    _nativeElvis.setTargetCPU(cpu);

                final int [] blob = _nativeElvis.createBlob(phrases, numPhrases, decoyPenalty,
                                                           isMagic,
                                                           maxDuration);

                _mainThreadHandler.post(new Runnable() {
                    @Override
                    public void run()
                    {
                        if (blob == null)
                            blobActionListener.onBlobCreate(false, null);
                        else
                            blobActionListener.onBlobCreate(true, blob);
                    }});
            }
        });
    }

    @Override
    public void forgetBlob(final BlobActionListener blobActionListener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        Checker.checkArgForNull("blobActionListener", blobActionListener);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final int status = _nativeElvis.forgetBlob();

                _mainThreadHandler.post(new Runnable() {
                    @Override
                    public void run()
                    {
                        if (status == 0)
                            blobActionListener.onBlobForget(true);
                        else
                            blobActionListener.onBlobForget(false);
                    }});
            }
        });
    }

    @Override
    public void evaluateWakeupPhrase(final EvaluatePhraseListener evaluationListener, final String phrase)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        Checker.checkArgForNull("evaluationListener", evaluationListener);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final int status =
                    _nativeElvis.evaluateWakeupPhrase(phrase);
                _mainThreadHandler.post(new Runnable() {
                    @Override
                    public void run()
                    {
                        evaluationListener.onPhraseEvaluation(true, status);
                    }});
            }
        });
    }

    @Override
    public void enableVerboseAndroidLogging(boolean verbose)
    {
        _nativeElvis.enableElvisVerboseLogging(verbose);
    }

    @Override
    public int getRecognitionLoggingStatus()
    {
        return _recognitionLoggingStatus;
    }

    @Override
    public boolean logMiscInfo(String tag, String data)
    {
        return _nativeElvis.logMiscInfo(tag, data);
    }

    @Override
    public void onApplicationLanguageChange(Object unused, final Listener listener)
    {
        // clearGrammar() will also cancel the rebuild and the recognition
        clearGrammar(new RebuildListener()
        {
            @Override
            public void onComplete(Grammar grammar,
                    List<SkippedWord> skippedWords)
            {
                if (listener != null)
                    listener.onComplete(true);
            }

            @Override
            public void onError(ElvisError error)
            {
                if (listener != null)
                    listener.onComplete(false);
            }
        });
    }

    private void updateRecognitionLoggingStatus()
    {
        int oldStatus = _recognitionLoggingStatus;
        int newStatus = _nativeElvis.getElvisRecognitionLoggingStatus();
        if (oldStatus != newStatus) {
            if (newStatus == Logging.LOGGING_COMPLETE) {
                if (_recognitionLoggingFolder != null) {
                    if (_nativeElvis.copyElvisRecognitionLogs(_recognitionLoggingFolder, _recognitionLoggingPrefix)) {
                        _recognitionLoggingStatus = newStatus;
                        return;
                    }
                    _recognitionLoggingFolder = null;
                }
                _recognitionLoggingStatus = Logging.LOGGING_ERROR;
                return;
            }
            _recognitionLoggingStatus = newStatus;
        }
        return;
    }

    private void terminateCurrentRecognition()
    {
        if (_currentRecognition != null)
        {
            Recognition r = _currentRecognition;
            _currentRecognition.cancel();
            if (r == _currentRecognition)
                _currentRecognition = null;
        }
    }

    private void terminateAdaptation()
    {
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (_isElvisCreated)
                    _nativeElvis.finishAdapting(_stateFileName, true);
            }
        });
    }

    private String getQuality(String language, int samplingRate)
    {
        final String[] PRIORITIES = { ElvisQuality.ENDPOINTING, ElvisQuality.ULTRA_COMPACT, ElvisQuality.STANDARD, ElvisQuality.PREMIUM };
        int qualityIdx = 0;

        List<ElvisModelInfo> elvisModels = getAvailableModels();
        for (ElvisModelInfo model : elvisModels)
        {
            if (language.equals(model.language.name) && samplingRate == model.frequencyHz)
            {
                int availableQualityIdx = 0;
                for (int idx = 0; idx < PRIORITIES.length; idx++)
                {
                    if (PRIORITIES[idx].equals(model.quality))
                    {
                        availableQualityIdx = idx;
                    }
                }

                if (availableQualityIdx > qualityIdx)
                {
                    qualityIdx = availableQualityIdx;
                }
            }
        }

        return PRIORITIES[qualityIdx];
    }

    /**
     * Called on MAIN thread exactly once in the constructor to obtain the
     * worker thread handler, if the worker thread handler wasn't given in the
     * constructor.
     * @return The worker thread handler.
     */
    protected abstract NMTHandler getWorkerThreadHandler();
}
