package com.nuance.dragon.toolkit.elvis;

import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.WorkerThread;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Implementation of {@link ElvisRecognizer}.
 * <p>
 * Don't create this directly. Use {@link ElvisRecognizer#createElvisRecognizer}.
 */
final class ElvisRecognizerImpl extends ElvisRecognizerBase
{
    private final boolean _hasWorkerThread;

    // Only allow 1 worker thread and file factory for all Elvis instances.
    private static WorkerThread _workerThread;
    private static int _workerThreadCount;

    private boolean _released;

    /**
     * Create an Elvis Recognizer base on a compiled-in Elvis model. The Elvis
     * will use a dedicated worker thread. <p>
     * NOTE: Due to Elvis engine limitations, multiple Elvis instances must use
     * the same worker thread.
     * @param fileManager The file manager where Elvis should look for files.
     */
    public ElvisRecognizerImpl(FileManager fileManager)
    {
        this(fileManager, null);
    }

    /**
     * Create an Elvis Recognizer from a bin file. The Elvis will use the given
     * worker thread. <p>
     * NOTE: Due to Elvis engine limitations, multiple Elvis instances must use
     * the same worker thread.
     * @param fileManager The file manager where Elvis should look for files.
     * @param workerThread
     */
    public ElvisRecognizerImpl(FileManager fileManager, NMTHandler workerThread)
    {
        super(new NativeElvisImpl(fileManager), workerThread);
        Checker.checkArgForNull("fileManager", fileManager);
        _hasWorkerThread = workerThread == null;
    }

    @Override
    protected NMTHandler getWorkerThreadHandler()
    {
        if (_workerThread == null) {
            _workerThread = Factory.createWorkerThread("com.nuance.dragon.toolkit.elvis.ElvisRecognizerImpl");
            _workerThread.start();
        }
        _workerThreadCount++;
        return _workerThread.getHandler();
    }

    @Override
    public void release()
    {
        if (!_released)
        {
            _released = true;
            super.release();

            if (_hasWorkerThread)
            {
                _workerThreadCount--;
                if (_workerThreadCount == 0)
                {
                    _workerThread.stop();
                    _workerThread = null;
                }
            }
        }
    }
}
