package com.nuance.dragon.toolkit.elvis;

import junit.framework.Assert;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.recognition.LocalRecognitionResult;

/**
 * A raw Elvis result.
 */
public final class ElvisResult extends LocalRecognitionResult implements JSONCompliant
{
    /**
     * The wake-up status.
     */
    public class WakeupStatus
    {
        /** Wake-up word + command unconfirmed */
        public static final int MAYBE_COMMAND = 0;        

        /** Wake-up word only confirmed */
        public static final int NO_COMMAND = 1;

        /** Wake-up word + command start-of-speech confirmed */
        public static final int COMMAND_STARTED = 2;

        /** Wake-up word + command end-of-speech confirmed */
        public static final int COMMAND_ENDED = 3;
    }

    public static final int MIN_CONFIDENCE = 0;
    public static final int MAX_CONFIDENCE = 100;
    public static final int GATE_CONFIDENCE_UNAVAILABLE = -1;
    public static final int WAKEUP_TIMESTAMP_UNAVAILABLE = -1;

    private static final String KEY_NBEST = "nbest";
    private static final String KEY_MISCLOGGING = "misclogging";
    private static final String KEY_GATE_CONFIDENCE = "gate_confidence";
    private static final String KEY_WAKEUP_START_TIMESTAMP = "wakeup_start_timestamp";
    private static final String KEY_WAKEUP_END_TIMESTAMP = "wakeup_end_timestamp";
    private static final String KEY_WAKEUP_COMMAND_START_TIMESTAMP = "wakeup_cmd_start_timestamp";
    private static final String KEY_WAKEUP_COMMAND_END_TIMESTAMP = "wakeup_cmd_end_timestamp";
    private static final String KEY_WAKEUP_STATUS = "wakeup_status";
    private final ElvisNbestList _nbest;
    private final int _confidence;
    private final int _gateConfidence;
    private final long _wakeupStartTimestamp, _wakeupEndTimestamp, _wakeupCommandStartTimestamp, _wakeupCommandEndTimestamp;
    private final int _wakeupStatus;
    private final String _miscLogging;

    ElvisResult(ElvisNbestList nbest, int gateConfidence, long wakeupStartTimestamp, long wakeupEndTimestamp, long wakeupCommandStartTimestamp, long wakeupCommandEndTimestamp, int wakeupStatus, String miscLogging)
    {
        super(nbest != null ? nbest.toJSON() : null);

        Assert.assertFalse(nbest.isEmpty());
        _nbest = nbest;
        int confidence = MIN_CONFIDENCE;
        if (nbest.size() < 2) {
            confidence = MAX_CONFIDENCE;
        } else {
            confidence = nbest.get(1).getScore() - nbest.get(0).getScore();
            if (confidence < MIN_CONFIDENCE) {
                confidence = MIN_CONFIDENCE;
            }
            if (confidence > MAX_CONFIDENCE) {
                confidence = MAX_CONFIDENCE;
            }
        }
        _confidence = confidence;
        _gateConfidence = gateConfidence;
        _wakeupStartTimestamp = wakeupStartTimestamp;
        _wakeupEndTimestamp = wakeupEndTimestamp;
        _wakeupCommandStartTimestamp = wakeupCommandStartTimestamp;
        _wakeupCommandEndTimestamp = wakeupCommandEndTimestamp;
        _wakeupStatus = wakeupStatus;
        _miscLogging = miscLogging;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        o.tryPut(KEY_NBEST, _nbest.toJSON());
        o.tryPut(KEY_MISCLOGGING, _miscLogging);
        o.tryPut(KEY_GATE_CONFIDENCE, _gateConfidence);
        o.tryPut(KEY_WAKEUP_START_TIMESTAMP, _wakeupStartTimestamp);
        o.tryPut(KEY_WAKEUP_END_TIMESTAMP, _wakeupEndTimestamp);
        o.tryPut(KEY_WAKEUP_COMMAND_START_TIMESTAMP, _wakeupCommandStartTimestamp);
        o.tryPut(KEY_WAKEUP_COMMAND_END_TIMESTAMP, _wakeupCommandEndTimestamp);
        o.tryPut(KEY_WAKEUP_STATUS, _wakeupStatus);
        return o;
    }

    public static ElvisResult createFromJSON(org.json.JSONObject json) throws JSONException
    {
        int gateConfidence;
        long wakeupStartTimestamp, wakeupEndTimestamp, wakeupCommandStartTimestamp, wakeupCommandEndTimestamp;
        int wakeupStatus;
        String miscLogging;

        try {
            gateConfidence = json.getInt(KEY_GATE_CONFIDENCE);
            wakeupStartTimestamp = json.getLong(KEY_WAKEUP_START_TIMESTAMP);
            wakeupEndTimestamp = json.getLong(KEY_WAKEUP_END_TIMESTAMP);
            wakeupCommandStartTimestamp = json.getLong(KEY_WAKEUP_COMMAND_START_TIMESTAMP);
            wakeupCommandEndTimestamp = json.getLong(KEY_WAKEUP_COMMAND_END_TIMESTAMP);
            wakeupStatus = json.getInt(KEY_WAKEUP_STATUS);
            miscLogging = json.getString(KEY_MISCLOGGING);
        } catch(JSONException e) {
            gateConfidence = GATE_CONFIDENCE_UNAVAILABLE;
            wakeupStartTimestamp = WAKEUP_TIMESTAMP_UNAVAILABLE;
            wakeupEndTimestamp = WAKEUP_TIMESTAMP_UNAVAILABLE;
            wakeupCommandStartTimestamp = WAKEUP_TIMESTAMP_UNAVAILABLE;
            wakeupCommandEndTimestamp = WAKEUP_TIMESTAMP_UNAVAILABLE;
            wakeupStatus = WakeupStatus.NO_COMMAND;
            miscLogging = "unavailable";
        }

        return new ElvisResult(ElvisNbestList.createFromJSON(json.getJSONObject(KEY_NBEST)),
                               gateConfidence, wakeupStartTimestamp, wakeupEndTimestamp, wakeupCommandStartTimestamp, wakeupCommandEndTimestamp, wakeupStatus, miscLogging);
    }

    /**
     * Get the list of choices as an N-best list. The list will never be empty.
     * @return The N-best list.
     */
    public ElvisNbestList getChoiceList()
    {
        return _nbest;
    }

    /**
     * @return The number of choices in the N-best list. Always > 0. This is the
     * same as the size of the N-best list returned by {@link #getChoiceList}.
     */
    public int getChoiceCount()
    {
        return _nbest.size();
    }

    /**
     * Returns the calculated confidence that the top choice is correct.
     * 100 indicates highest confidence, 0 indicates lowest confidence. For low
     * confidence results, it may be a good idea to consider other options in
     * the N-best list.
     * @return The confidence
     */
    public int getConfidence()
    {
        return _confidence;
    }

    /**
     * Returns the confidence that the gate command is correct.
     * 100 indicates highest confidence, 0 indicates lowest confidence.
     * -1 indicates that the gate confidence is not available.
     * (Gate confidence is only available for premium models)
     * @return The confidence
     */
    public int getGateConfidence()
    {
        return _gateConfidence;
    }

    /**
     * Indicates the wake-up type (i.e. wake-up word +/- <open-speech>).
     * @return The wake-up type, see {@link WakeupStatus}.
     */
    public int getWakeupStatus()
    {
        return _wakeupStatus;
    }

    /**
     * Returns the time stamp of start of the wake-up utterance.
     * @return The start of utterance time stamp, in ms. -1 if not available.
     */
    public long getWakeupStartTimestamp()
    {
        return _wakeupStartTimestamp;
    }

    /**
     * Returns the time stamp of the end of wake-up utterance.
     * @return The start of utterance time stamp, in ms. -1 if not available.
     */
    public long getWakeupEndTimestamp()
    {
        return _wakeupEndTimestamp;
    }

    /**
     * Returns the time stamp of start of the wake-up command utterance.
     * @return The start of utterance time stamp, in ms. -1 if not available.
     */
    public long getWakeupCommandStartTimestamp()
    {
        return _wakeupCommandStartTimestamp;
    }

    /**
     * Returns the time stamp of the end of wake-up command utterance.
     * @return The start of utterance time stamp, in ms. -1 if not available.
     */
    public long getWakeupCommandEndTimestamp()
    {
        return _wakeupCommandEndTimestamp;
    }    

    /**
     * Returns the result log to be eventually sent to the server for data mining purposes.
     * @return The log buffer. null if not available.
     */
    public String getMiscLogging()
    {
        return _miscLogging;
    }    

    /**
     * @return The transcription of the top entry
     */
    @Override
    public String toString()
    {
        return _nbest.get(0).toString();
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + _confidence;
        result = prime * result + _gateConfidence;
        result = prime * result
                + ((_miscLogging == null) ? 0 : _miscLogging.hashCode());
        result = prime * result + ((_nbest == null) ? 0 : _nbest.hashCode());
        result = prime
                * result
                + (int) (_wakeupCommandEndTimestamp ^ (_wakeupCommandEndTimestamp >>> 32));
        result = prime
                * result
                + (int) (_wakeupCommandStartTimestamp ^ (_wakeupCommandStartTimestamp >>> 32));
        result = prime * result
                + (int) (_wakeupEndTimestamp ^ (_wakeupEndTimestamp >>> 32));
        result = prime
                * result
                + (int) (_wakeupStartTimestamp ^ (_wakeupStartTimestamp >>> 32));
        result = prime * result + _wakeupStatus;
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        ElvisResult other = (ElvisResult) obj;
        if (_confidence != other._confidence)
            return false;
        if (_gateConfidence != other._gateConfidence)
            return false;
        if (_miscLogging == null) {
            if (other._miscLogging != null)
                return false;
        } else if (!_miscLogging.equals(other._miscLogging))
            return false;
        if (_nbest == null) {
            if (other._nbest != null)
                return false;
        } else if (!_nbest.equals(other._nbest))
            return false;
        if (_wakeupCommandEndTimestamp != other._wakeupCommandEndTimestamp)
            return false;
        if (_wakeupCommandStartTimestamp != other._wakeupCommandStartTimestamp)
            return false;
        if (_wakeupEndTimestamp != other._wakeupEndTimestamp)
            return false;
        if (_wakeupStartTimestamp != other._wakeupStartTimestamp)
            return false;
        if (_wakeupStatus != other._wakeupStatus)
            return false;
        return true;
    }
}
