package com.nuance.dragon.toolkit.elvis;

import com.nuance.dragon.toolkit.oem.api.internal.Checker;


/**
 * A container for all the parameters consumed
 * by an Elvis end-pointing session.
 */
public class EndpointingParam
{
    /** Default parameters */
    public static final int DEFAULT_SILENCE_DURATION_MS = 900;
    public static final int DEFAULT_IGNORE_DURATION_MS = 0;
    public static final boolean DEFAULT_RESET_CHANNEL = true;

    /**
     * Builder for creating instances of {@code EndpointingParam}.
     */
    public static class Builder
    {
        private int silenceDurationMs;
        private int ignoreDurationMs;
        private boolean resetChannel;

        /**
         * Creates the builder.
         */
        public Builder()
        {
            this.silenceDurationMs = DEFAULT_SILENCE_DURATION_MS;
            this.ignoreDurationMs = DEFAULT_IGNORE_DURATION_MS;
            this.resetChannel = DEFAULT_RESET_CHANNEL;
        }

        /**
         * Set custom silence duration.
         * @param silenceDurationMs The silence duration (in ms) needed to declare end-of-speech.
         *                          Default is {@link #DEFAULT_SILENCE_DURATION_MS}.
         */
        public Builder setSilenceDuration(int silenceDurationMs)
        {
            this.silenceDurationMs = silenceDurationMs;
            return this;
        }

        /**
         * Set custom ignore duration, essentially for audio surgery purposes.
         * @param ignoreDurationMs The duration (in ms) to ignore before reporting speech notifications
         *                          Default is {@link #DEFAULT_IGNORE_DURATION_MS}.
         */
        public Builder setIgnoreDuration(int ignoreDurationMs)
        {
            this.ignoreDurationMs = ignoreDurationMs;
            return this;
        }

        /**
         * Set custom reset channel flag.
         *  @param resetChannel true to reset the Elvis front-end channel, else false
         *                           Default is {@link #DEFAULT_RESET_CHANNEL}.
         */
        public Builder setResetChannel(boolean resetChannel)
        {
            this.resetChannel = resetChannel;
            return this;
        }

        /**
         * Build the parameters.
         * @return The {@link EndpointingParam} instance.
         */
        public EndpointingParam build()
        {
            Checker.checkArgForCondition("silenceDurationMs", ">= 0", silenceDurationMs >= 0);
            Checker.checkArgForCondition("ignoreDurationMs", ">= 0", ignoreDurationMs >= 0);
            return new EndpointingParam(this);
        }
    }

    private final int _silenceDurationMs;
    private final int _ignoreDurationMs;
    private final boolean _resetChannel;

    private EndpointingParam(Builder builder)
    {
        _silenceDurationMs = builder.silenceDurationMs;
        _ignoreDurationMs = builder.ignoreDurationMs;
        _resetChannel = builder.resetChannel;
    }

    int getSilenceDurationMs()
    {
        return _silenceDurationMs;
    }

    int getIgnoreDurationMs()
    {
        return _ignoreDurationMs;
    }

    boolean getResetChannel()
    {
        return _resetChannel;
    }

    @Override
    public String toString() {
        return "EndpointingParam [_silenceDurationMs=" + _silenceDurationMs
                + ", _ignoreDurationMs=" + _ignoreDurationMs
                + ", _resetChannel=" + _resetChannel + "]";
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + _ignoreDurationMs;
        result = prime * result + (_resetChannel ? 1231 : 1237);
        result = prime * result + _silenceDurationMs;
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        EndpointingParam other = (EndpointingParam) obj;
        if (_ignoreDurationMs != other._ignoreDurationMs)
            return false;
        if (_resetChannel != other._resetChannel)
            return false;
        if (_silenceDurationMs != other._silenceDurationMs)
            return false;
        return true;
    }
}
