package com.nuance.dragon.toolkit.elvis;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.json.JSONException;

import com.nuance.dragon.toolkit.core.WordList;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * Represents the Elvis grammar. Contains populated word slots and constraints.
 */
public class Grammar implements JSONCompliant
{
    private static final String KEY_SLOTS = "slots";
    private static final String KEY_CONSTRAINTS = "constraints";

    private final HashMap<String,Constraint> _constraints;
    private final HashMap<String,WordSlot> _slots;

    /**
     * Create an empty grammar.
     */
    Grammar()
    {
        _slots = new HashMap<String,WordSlot>();
        _constraints = new HashMap<String,Constraint>();
    }

    /* (non-Javadoc)
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((_constraints == null) ? 0 : _constraints.hashCode());
        result = prime * result + ((_slots == null) ? 0 : _slots.hashCode());
        return result;
    }

    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (!(obj instanceof Grammar))
            return false;
        Grammar other = (Grammar) obj;
        if (_constraints == null)
        {
            if (other._constraints != null)
                return false;
        } else if (!_constraints.equals(other._constraints))
            return false;
        if (_slots == null)
        {
            if (other._slots != null)
                return false;
        } else if (!_slots.equals(other._slots))
            return false;
        return true;
    }

    /**
     * Construct a grammar with the specified WordSlot and Constraint lists.
     *
     * @param slots The list of WordSlots in this grammar.
     *              Must not be empty or null.
     *              The WordSlot list must not contain duplicate IDs.
     *
     * @param constraints The list of Constraints in this grammar.
     *                    Must not be empty or null.
     *                    The Constraint list must not contain duplicate IDs.
     *
     * @return a Grammar object if successful or null if the argument list is not valid.
     */
    public static Grammar createGrammar(List<WordSlot> slots, List<Constraint> constraints)
    {
        Checker.checkArgForCondition("constraints", "not null or empty", constraints != null && !constraints.isEmpty());
        Checker.checkArgForCondition("slots", "not null or empty", slots != null && !slots.isEmpty());

        Grammar grammar = new Grammar();
        for (WordSlot ws : slots) {
            if (grammar._slots.put(ws.getId(), ws.copy()) != null) {
                Logger.error(grammar, "Duplicate WordSlot ID in slots: " + ws.getId());
                return null;
            }
        }
        for (Constraint c : constraints) {
            if (grammar._constraints.put(c.getId(), c.copy()) != null) {
                Logger.error(grammar, "Duplicate Constraint ID in slots: " + c.getId());
                return null;
            }
        }
        return grammar;
    }

    /**
     * Use this function to determine the IDs of the word lists required by this Grammar, if any.
     *
     * @return A list of Word list IDs that are required by this object.
     * Will be an empty list (never null) if no word lists are required.
     */
    public Set<String> getRequiredWordListIds() {
        HashSet<String> requiredWordListIds = new HashSet<String>();
        for (Map.Entry<String, WordSlot> entry : _slots.entrySet()) {
            String name = entry.getKey();
            WordSlot ws = entry.getValue();
            if (ws.getWordListRequired()) {
                requiredWordListIds.add(name);
            }
        }
        return requiredWordListIds;
    }

    /**
     * Attach a word list to this Grammar.  Note that the name of the word list must
     * exactly match the name of a word slot defined in the grammar.
     *
     * @param wordList The word list to attach.  Must not be null.
     * @param id The name of the word slot where the list should be attached.  Must not be null or empty.
     *
     * @return true if the list was attached, false if the list was not attached.
     */
    public boolean attachWordList(WordList wordList, String id)
    {
        Checker.checkStringArgForEmpty("id", id);
        Checker.checkArgForNull("wordList", wordList);
        WordSlot ws = _slots.get(id);
        if (ws == null) {
            return false;
        }
        ws.attachWordList(wordList);
        return true;
    }

    /**
     * And the specified WordSlot to the grammar.
     *
     * @param slot a WordSlot to be added.  May be null.
     * @param true if existing WordSlot with the same name should be replaced.
     *        false if existing WordSlot with the same place should be skipped.
     */
    void addWordSlot(WordSlot slot, boolean overwrite)
    {
        if (slot == null) {
            return;
        }
        if (overwrite || !_slots.containsKey(slot.getId())) {
            _slots.put(slot.getId(), slot.copy());
        }
    }

    /**
     * And the specified WordSlots to the grammar.
     *
     * @param slots a list of WordSlots to be added.  May be null or empty.
     * @param true if existing WordSlots with the same name should be replaced.
     *        false if existing WordSlots with the same place should be skipped.
     */
    void addWordSlots(List<WordSlot> slots, boolean overwrite)
    {
        if (slots == null) {
            return;
        }
        for (WordSlot ws : slots) {
            if (overwrite || !_slots.containsKey(ws.getId())) {
                _slots.put(ws.getId(), ws.copy());
            }
        }
    }

    /**
     * Return a list containing all the WordSlot objects in this grammar.
     *
     * @return The word slots.  Will never be null, but may be an empty list.
     */
    public List<WordSlot> getSlots()
    {
        List<WordSlot> slots = new ArrayList<WordSlot>();
        for (WordSlot ws : _slots.values()) {
            slots.add(ws);
        }
        return slots;
    }

    /**
     * Get the WordSlot object with the given name.
     *
     * @return a WordSlot object, or null if the name is not defined.
     */
    public WordSlot getWordSlotByName(String name)
    {
        return _slots.get(name);
    }

    /**
     * Delete all WordSlots from the Grammar.
     */
    void clearWordSlots() {
        _slots.clear();
    }

    /**
     * And the specified Constraint to the grammar.
     *
     * @param constraint a list of Constraints to be added.  May be null.
     * @param true if existing Constraints with the same name should be replaced.
     *        false if existing Constraints with the same place should be skipped.
     */
    void addConstraint(Constraint constraint, boolean overwrite)
    {
        if (constraint == null) {
            return;
        }
        if (overwrite || !_constraints.containsKey(constraint.getId())) {
            _constraints.put(constraint.getId(), constraint.copy());
        }
    }

    /**
     * And the specified List of Constraints to the grammar.
     *
     * @param constraints a list of Constraints to be added.  May be null or empty.
     * @param true if existing Constraints with the same name should be replaced.
     *        false if existing Constraints with the same place should be skipped.
     */
    void addConstraints(List<Constraint> constraints, boolean overwrite)
    {
        if (constraints == null) {
            return;
        }
        for (Constraint c : constraints) {
            if (overwrite || !_constraints.containsKey(c.getId())) {
                _constraints.put(c.getId(), c.copy());
            }
        }
    }

    /**
     * Get the list of constraints.
     *
     * @return The constraints. Will never be null.
     */
    public List<Constraint> getConstraints()
    {
        List<Constraint> constraints = new ArrayList<Constraint>();
        for (Constraint c : _constraints.values()) {
            constraints.add(c);
        }
        return constraints;
    }

    /**
     * Get the Constraint object with the given name.
     *
     * @return a Constraint object, or null if the name is not defined.
     */
    public Constraint getConstraintByName(String name)
    {
        return _constraints.get(name);
    }

    /**
     * Delete all Constraints from the Grammar.
     */
    void clearConstraints() {
        _constraints.clear();
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();

        o.tryPutList(KEY_SLOTS, getSlots());
        o.tryPutList(KEY_CONSTRAINTS, getConstraints());

        return o;
    }

    /**
     * @param json A JSON object that represents a Grammar.
     *
     * @return A Grammar object if successful.  Throws JSONException if there is an error.
     */
    public static Grammar createFromJSON(org.json.JSONObject json) throws JSONException
    {

        ArrayList<WordSlot> w = new ArrayList<WordSlot>();
        ArrayList<Constraint> c = new ArrayList<Constraint>();

        org.json.JSONArray slots = json.getJSONArray(KEY_SLOTS);
        for (int i = 0; i < slots.length(); i++)
        {
            w.add(WordSlot.createFromJSON(slots.getJSONObject(i)));
        }

        org.json.JSONArray constraints = json.getJSONArray(KEY_CONSTRAINTS);
        for (int i = 0; i < constraints.length(); i++)
        {
            c.add(Constraint.createFromJSON(constraints.getJSONObject(i)));
        }
        Grammar grammar = Grammar.createGrammar(w, c);
        if (grammar == null) {
            throw new JSONException("Error is WordSlot or Constraint list.  Unable to create grammar.");
        }
        return grammar;
    }
}
