package com.nuance.dragon.toolkit.elvis;

import java.util.List;

import com.nuance.dragon.toolkit.core.Word;

/**
 * An interface that provides access to the Elvis engine. All methods here are
 * expected to be called from a single worker thread.
 */
interface NativeElvis
{
    // TODO: these probably need to be revised
    static final int SILENCE = 0;
    static final int SPEECH = 1;
    static final int SPEECH_ENDED = 2;
    static final int SPEECH_ERROR = 3;

    static final int CONSTRAINT_ADDED = 0;
    static final int CONSTRAINT_NOT_ADDED = 1;
    static final int CONSTRAINT_ERROR = 2;

    static final int TRANSITION_ADDED = 0;
    static final int TRANSITION_ERROR = 1;

    static final int WORDCLASS_ADDED = 0;
    static final int WORDCLASS_NOT_ADDED = 1;
    static final int WORDCLASS_ERROR = 2;

    static final int WORD_ADDED = 0;
    static final int WORD_NO_PRONUNCIATION = 1;
    static final int WORD_ERROR = 2;

    static final int WAKEUP_LONG = 0;
    static final int WAKEUP_SHORT = 1;
    static final int WAKEUP_ENROLL = 2;
    static final int WAKEUP_SEAMLESS = 3;
    static final int WAKEUP_WORDSPOT = 4;
    static final int WAKEUP_CHOPPED = 5;
    static final int WAKEUP_COMMAND = 6;
    static final int WAKEUP_SNAP = 7;

    static final int WUW_NO_WAKEUP = 0;
    static final int WUW_PENDING_WAKEUP = 1;
    static final int WUW_WAKEUP = 2;
    static final int WUW_COMMAND_STARTED = 3;
    static final int WUW_COMMAND_ENDED = 4;
    static final int WUW_COMMAND_TIMEOUT = 5;
    static final int WUW_ERROR = 6;

    /**
     * Called from ANY thread to synchronously request the supported models.
     * @param retModels Add the detected models here.
     * @return false if unsuccessful
     */
    boolean listAvailableModels(List<NativeModelInfo> retModels);

    /**
     * Called on worker thread. Language / samplingRate will be tried if present.
     * If they are absent or unsupported, binFile will be tried if present.
     * @param language
     * @param samplingRate
     * @param binFile
     * @return false if unsuccessful
     */
    boolean createElvis(String language, int samplingRate, String binFile);

    /**
     * If necessary, this function recreates Elvis using the parameters that were
     * originally passed to createElvis().
     * The recommended value for the force parameter is false.
     * Call this function before calling loadElvis to ensure Elvis is in a state
     * that supports loading.
     *
     * @param If true, delete and create Elvis even if no save file, constraints, or word classes have been loaded.
     *        If false, only recreate Elvis if necessary.
     * @return false if unsuccessful.
     */
    boolean recreateElvis(boolean force);

    /**
     * Called on worker thread.
     */
    void destroyElvis();

    /**
     * Load a saved Elvis grammar (called on worker thread).
     * The Native Elvis library only supports loading when Elvis is empty.
     * If a state has previously been loaded, or if any Constraints or WordClasses
     * have been added to Elvis, this function will fail.
     *
     * @param fileName The file to load from.
     * @return The restored grammar object, or null if an error occurred.
     */
    Grammar loadGrammar(String fileName);

    /**
     * Save the current Elvis grammar (called on worker thread).
     * @param fileName The file to save to.
     * @param grammar
     * @return false if unsuccessful
     */
    boolean saveGrammar(String fileName, Grammar grammar);

    /**
     * Add a constraint to Elvis (called on worker thread).
     * @param name Name of the constraint.
     * @param classCount The total number of word classes used in this constraint.
     * @param transitionCount The total number of transitions used in this constraint.
     * @param hasPenalty Must be true if at least one transition in this constraint has a non-zero penalty value.
     * @param reset If true, and the constraint already exists, the old constraint
     * will be replaced with this one.
     * @return CONSTRAINT_ADDED if the constraint was added successfully
     * CONSTRAINT_NOT_ADDED if constraint already existed and was not replaced and "reset" was false,
     * CONSTRAINT_ERROR if an error occurred, or if "reset" was true and replacing it failed.
     */
    int addConstraint(String name, int classCount, int transitionCount, boolean hasPenalties, boolean reset);

    /**
     * Delete the specified constraint from Elvis (called on worker thread).
     * @param name The name of the constraint to delete.
     * @return true for success or false for an error.
     */
    boolean deleteConstraint(String name);

    /**
     * Check whether a constraint already exists in Elvis (called on worker thread).
     * @param name Name of the constraint.
     * @return true if the constraint exists.
     */
    boolean constraintExists(String name);

    /**
     * Clear all constraints and WordClasses from Elvis (called on worker thread).
     * @return true for success or false for an error.
     */
    boolean clearGrammar();

    /**
     * Add a word class (called on worker thread).
     * @param slot a WordSlot to add to the grammar.
     * @return
     * WORDCLASS_ADDED if the word class was added successfully <br>
     * WORDCLASS_NOT_ADDED if the word class wasn't added because it already exists <br>
     * WORDCLASS_ERROR if an error occurred, or if rebuildType was FULL and deleting the old definition failed.
     */
    int addWordClass(WordSlot slot);

    /**
     * Add members to an existing word class (called on worker thread).
     * @param wordclass The name of the word class.
     * @param word The word to add to this class.
     * @param addVariants True if variants (shorter versions of a name, song title, etc) should be automatically added.
     * @return WORD_ADDED if the word was added successfully
     * WORD_NO_PRONUNCIATION if the word cannot be added because the pronunciation isn't known.
     * WORD_ERROR if an error occurred.
     */
    int addWordClassMember(String wordclass, Word word, boolean addVariants);

    /**
     * Remove members from an existing word class (called on worker thread).
     * @param wordclass The name of the word class.
     * @param word The word to remove from this class.
     * @return true for success or false for an error.
     */
    boolean removeWordClassMember(String wordclass, Word word);

    /**
     * Add am external pronunciation (called on worker thread).
     * @param word The word to add.
     * @return
     * @return WORD_ADDED if the word was added successfully
     * WORD_NO_PRONUNCIATION if the word cannot be added because the pronunciation isn't known.
     * WORD_ERROR if an error occurred.
     */
    int addExternalPronunciation(Word word);

    /**
     * Add a transition in an existing constraint (called on worker thread).
     * @param constraint The constraint to add the transition to.
     * @param fromWordClass The word class that is transitioned from.
     * @param toWordClass The word class that is transitioned to.
     * @param penalty The penalty.
     * @return TRANSITION_ADDED if the transition was added successfully
     * TRANSITION_ERROR if an error occurred.
     */
    int addTransition(String constraint, String fromWordClass, String toWordClass, int penalty);

    /**
     * Start an adaptation session (called from worker thread).
     * @param phrase
     * @return false if unsuccessful
     */
    boolean adapt(String phrase);

    /**
     * Process an adapted utterance (called from worker thread).
     * @return The result
     */
    ElvisAdaptationResult processAdaptation();

    /**
     * Load eventual adaptation data.
     * @param fileName The file to load from.
     * @return false if unsuccessful
     */
    boolean loadAdaptation(String fileName, String ext, boolean isUnsupervised);

    /**
     * End an adaptation session.
     * @param fileName The file to save to.
     * @param abort true if adaptation data should be discarded. False if adaptation data should be processed.
     * @return false if unsuccessful
     */
    boolean finishAdapting(String fileName, boolean abort);

    /**
     * Check if Elvis is adapted.
     * @param fileName The current adaptation file.
     * @return false if unsuccessful or not adapted
     */
    boolean checkAdaptationStatus(String fileName);

    /**
     * Clear adaptation data from Elvis (called on worker thread).
     * @param fileName The file to load from.
     * @param all Will clear adaptation in every languages and frequencies.
     * @return false if unsuccessful
     */
    boolean clearAdaptation(String fileName, String ext, boolean all);

    /**
     * Start a recognition (called from worker thread).
     * @param resetChannel true to reset the Elvis front-end channel, else false.
     * @return false if unsuccessful
     */
    boolean recognize(List<String> activeConstraints, boolean resetChannel);

    /**
     * Start wakeupMode (called from worker thread).
     * @return false if unsuccessful
     */
    boolean wakeupMode(List<String> wakeupPhrases, int penalty, int mode, int commandStartBailoutMs, int commandEndBailoutMs, int commandMinSpeechMs, boolean autoAdapt);

    /**
     * Start a end-of-speech detection (called from worker thread).
     * @param neededSamples The number of silence samples needed to declare end-of-speech.
     * @param ignoredSamples The number of samples to ignore.
     * @param resetChannel true to reset the Elvis front-end channel, else false.
     * @param loggingFolder The logging folder.
     * @return false if unsuccessful
     */
    boolean endpointingMode(int neededSamples, int ignoredSamples, boolean resetChannel, String loggingFolder);

    /**
     * Finish a recognition (called from worker thread). Call this to cause
     * Elvis to process a result prior to seeing end of speech.
     * @param abort true if results should be discarded.  false if results should be processed.
     * @return false if unsuccessful
     */
    boolean finishRecognizing(boolean abort);

    /**
     * Finish WakeupMode (called from worker thread). Call this to cause
     * Elvis to process a result prior to seeing wake up phrase.
     * @param fileName The file to save to.
     * @return false if unsuccessful
     */
    boolean finishWakeupMode(String fileName);

    /**
     * Finish a end-of-speech detection (called from worker thread). Call this to cause
     * Elvis to clean-up the endpointing session.
     * @return false if unsuccessful
     */
    boolean finishEndpointing();

    /**
     * Handles audio that has just been recorded (called from worker thread).
     * @param shorts
     *
     * @return The status of the recorded audio: {@link #SPEECH},
     *         {@link #SPEECH_ENDED}, {@link #SILENCE}, or {@link #SPEECH_ERROR}.
     */
    int processAudio(short[] shorts);

    /**
     * Handles audio that has just been recorded (called from worker thread).
     * @param shorts
     * @param timestamp
     *
     * @return The wake-up status.
     */
    int processWakeupAudio(short[] shorts, long timestamp);

    /**
     * Handles audio that has just been recorded (called from worker thread).
     * @param shorts
     *
     * @return The status of the recorded audio: {@link #SPEECH},
     *         {@link #SPEECH_ENDED}, {@link #SILENCE}, or {@link #SPEECH_ERROR}.
     */
    int processEndpointingAudio(short[] shorts);

    /**
     * Get the average sound level of the last pushed audio frame.
     * @return The sound level.
     */
    int getSoundLevel();

    /**
     * Obtain the result from the previous recognition (called from worker thread).
     * @param grammar
     * @return The result.
     */
    ElvisResult getResult(Grammar grammar);

    /**
     * Obtain the top result returned by WakeupMode (called from worker thread).
     * @param seamless True if in seamless mode, else false.
     * @return The result.
     */
    ElvisResult getWakeupResult(boolean seamless);

    /**
     * Check whether a phrase is suitable to use as a Wakeup phrase
     * @return true if the phrase can be used, false if it cannot
     */
    boolean checkWakeupPhrase(String wakeupPhrase);

    /**
     * After a recognition is complete and the result has been retrieved,
     * release memory that is no longer needed.
     */
    void releaseUnusedMemory();

    /**
     * Enable or disable verbose Elvis logging. (called from worker thread)
     *
     * @param verbose True if verbose (detailed) l
     *
     */
    void enableElvisVerboseLogging(boolean verbose);

    /**
     * Enable or disable Elvis logging. (called from worker thread)
     *
     * @param logPrefix A string to use for logging file names.
     * @param logMillis the number of milliseconds of audio to log.
     * @param uuid The optional UUID of the log.
     *
     * @return true for success or false for failure.
     *
     */
    boolean enableElvisRecognitionLogging(String logPrefix, int logMillis, String uuid);

    /**
     * Copies internal Elvis logging files to the requested location. (called from worker thread)
     *
     * @param destinationFolder the folder where the files should be copied.
     * @param destinationPrefix the prefix to use for the destination files.
     *
     * @return true for success or false for failure
     */
    boolean copyElvisRecognitionLogs(String destinationFolder, String destinationPrefix);

    /**
     * Returns the Elvis Recognition Logging status. (called from worker thread)
     */
    int getElvisRecognitionLoggingStatus();

    /**
     * Log misc data into Elvis engine, which will be included in the Elvis dump
     * eventually. Currently only a fixed set of tags are supported by the Elvis
     * engine. The application cannot pass arbitrary data.
     *
     * The supported tags are: "NVCSessionId", "NVCUniqueUserId". "NVCTransactionId"
     * will be supported in the future Elvis release.
     *
     * @param tag   the "name" of the data to be logged.
     * @param data  The "value" of the data to be logged.
     *
     * @return true for success or false for failure
     */
    boolean logMiscInfo(String tag, String data);

    /**
     * Return the blob file data to be saved and used for miniElvis
     *
     * @param ppPhrases a list of phrases to include in the blob.
     * @param numPhrases number of phrases in the list.
     * @param decoyPenalty the decoy penalty to be used when generating the blob data
     * @param bMagic boolean allow speech following the wakeup phrases.
     * @param maxDuration maximum speech duration in msec, should be set to -1.
     *
     * @return integer array containing blob file data.
     */
     int [] createBlob(List<String> phrases, int numPhrases, int decoyPenalty, boolean isMagic, int maxDuration);

    /**
     * Set the target CPU type for blob generation
     *
     * @param pCpu a string containing the cpu name ("x86", "CAPE", etc.)
     *
     * @return error status
     */
     int setTargetCPU(String cpu);

    /**
     * Forget the current blob data.
     *
     * @return error status
     */
    int forgetBlob();

    /**
     * Evaluate a phrase for wakeup
     *
     * @param phrase a string to evaluate
     *
     * @return phrase strength 0-100
     */
     int evaluateWakeupPhrase(String phrase);

}
