package com.nuance.dragon.toolkit.elvis;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.json.JSONObject;

import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.AudioType.Encoding;
import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.elvis.ElvisRecognizer.Logging;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.util.JSONUtils;

/**
 * Internal class that directly interfaces with the Elvis engine through JNI.
 * This is split from ElvisRecognizerImpl so that ElvisRecognizerImpl can be
 * obfuscated (this class doesn't need to be, since it isn't public).
 */
class NativeElvisImpl implements NativeElvis
{
    static boolean LOAD_LIBRARY_SUCCESS = true;
    static {
        try {
            System.loadLibrary("dmt_elvis");
        }
        catch (UnsatisfiedLinkError e) {
            Logger.error(NativeElvisImpl.class, "Failed to load native library.", e);
            LOAD_LIBRARY_SUCCESS = false;
        }
    }

    public native static int elvisGetParamsFromBinFile(FileManager fileManager, String binFile, BinFileParametersJni binFileParams);
    public native long elvisCreate(FileManager fileManager, String binfile, String language, int samplingRate);
    public native void elvisDestroy(long elvis);
    public native int elvisSave(long elvis, String fileName);
    public native int elvisLoad(long elvis, String vocFileName, String adaptFileName, boolean isUnsupervised);
    public native int elvisAddWordClass(long elvis, String name, int type, boolean force);
    public native int elvisAddClassMember(long elvis, String wordClass, String surfacePhrase, String spokenPhrase, int phraseID, boolean addVariants, int variantPenalty);
    public native boolean elvisRemoveClassMember(long elvis, String wordClass, String surfacePhrase);
    public native int elvisAddExternalPronunciation(long elvis, String surfacePhrase, String spokenPhrase);
    public native int elvisAddConstraint(long elvis, String constraint, int classCount, int transitionCount, boolean hasPenalties, boolean force);
    public native int elvisDeleteConstraint(long elvis, String constraint);
    public native boolean elvisConstraintExists(long elvis, String constraint);
    public native int elvisClearGrammar(long elvis);
    public native int elvisAddTransition(long elvis, String contraint, String fromClass, String toClass, int penalty);
    public native int elvisStartRecognizing(long elvis, String activeConstraints[], int numActiveConstraints, boolean resetChannel);
    public native int elvisStopRecognizing(long elvis, boolean abort);
    public native int elvisStartAdaptation(long elvis, String adaptationPhrase);
    public native int elvisProcessAdaptation(long elvis);
    public native int elvisEndAdaptationSession(long elvis, String fileName, boolean abort);
    public native boolean elvisCheckAdaptationStatus(long elvis);
    public native int elvisProcessAudio(long elvis, short data[], int numSamples);
    public native int elvisStartWakeupMode(long elvis, String wakeupPhrases[], int penalty, int mode, int commandStartBailoutMs, int commandEndBailoutMs, int commandMinSpeechMs, boolean autoAdapt);
    public native int elvisStopWakeupMode(long elvis, String fileName);
    public native int elvisProcessWakeupAudio(long elvis, short data[], int numSamples);
    public native boolean elvisCheckWakeupPhrase(long elvis, String wakeupPhrase);
    public native int elvisStartEndpointing(long elvis, int neededSamples, int ignoredSamples, boolean resetChannel, String loggingFolder);
    public native int elvisStopEndpointing(long elvis);
    public native int elvisProcessEndpointingAudio(long elvis, short data[], int numSamples);
    public native int elvisGetChoiceCount(long elvis);
    public native int elvisGetGateConfidence(long elvis);
    public native int elvisGetSoundLevel(long elvis);
    public native int elvisGetWakeupStartTime(long elvis, boolean commandPiece);
    public native int elvisGetWakeupEndTime(long elvis, boolean commandPiece);
    public native int elvisGetChoiceListEntry(long elvis, int choice, List<NbestEntryJni> entries);
    public native int elvisGetWakeupEntry(long elvis, List<NbestEntryJni> entries);
    public native void elvisEnableVerboseLogging(boolean verbose);
    public native int elvisEnableRecognitionLogging(long elvis, String logPrefix, int logMillis, String uuid);
    public native int elvisGetRecognitionLoggingStatus(long elvis);
    public native int elvisReleaseUnusedMemory(long elvis);
    public native int elvisGetMiscLoggingSize(long elvis);
    public native int elvisGetMiscLogging(long elvis, byte[] miscLogging);
    public native boolean elvisLogMiscInfo(long elvis, String tag, String data);
    public native int [] elvisCreateBlob(long elvis, String ppPhrases[], int numPhrases,
                                          int decoyPenalty, int bMagic, int maxDuration);
    public native int elvisSetTargetCPU(long elvis, String pCpu);
    public native int elvisForgetBlob(long elvis);
    public native int elvisEvaluateWakeupPhrase(long elvis, String phrase);

    private long _elvis;
    private boolean _released;
    private final FileManager _fileManager;
    private String _language;
    private int _samplingRate;
    private String _binFile;
    private boolean _dirty;

    // To manage the wake-up state
    private long _lastWakeupTimestamp;
    private int _lastWakeupStatus;
    private ElvisResult _lastWakeupResult;

    /* Note: The value of these constants must match the values in ngf_elvis.h */
    private static final int ELVIS_OK = 0;                     /*!< Normal return. */
    private static final int ELVIS_OOM = 1;                    /*!< Heap exhausted, or internal buffer full.*/
    private static final int ELVIS_INVALID_LANGUAGE = 2;       /*!< Invalid language selected.*/
    private static final int ELVIS_FILE_ERROR = 3;             /*!< Corrupt, unavailable, or non-existent file.*/
    private static final int ELVIS_ARGUMENT_ERROR = 4;         /*!< An invalid argument was passed to native Elvis. */
    private static final int ELVIS_UNSUPPORTED_OPERATION = 5;  /*!< Requested operation is not supported at this time. */
    private static final int ELVIS_NO_PRONUNCIATION = 6;       /*!< Unable to add word because Elvis has no pronunciation. */
    private static final int ELVIS_VOCABULARY_FULL = 7;        /*!< Unable to add word because Elvis's vocabulary is full. */
    private static final int ELVIS_OTHER_ERROR = 8;            /*!< An error occurred.*/

    private static final int ELVIS_NOSPEECH = 0;     /*!< FE has not yet detected speech */
    private static final int ELVIS_SPEECH = 1;       /*!< FE has detected onset of speech */
    private static final int ELVIS_DONE = 2;         /*!< FE thinks utt is done or wakeup phrase was detected. */
    private static final int ELVIS_SPEECH_ERROR = 3; /*!< An error condition occurred. */

    /**
     * Gets Language Code and Sample Rate from specified bin file
     * @param binFileName filename of bin file
     * @return BinFileParametersJni or null if bin file could not be found
     */
    static BinFileParametersJni getBinFileParameters(FileManager fileManager, String binFileName)
    {
        BinFileParametersJni ret = new BinFileParametersJni();

        if (elvisGetParamsFromBinFile(fileManager, binFileName, ret) != ELVIS_OK)
        {
            ret = null;
        }

        return ret;
    }

    NativeElvisImpl(FileManager fileManager)
    {
        _fileManager = fileManager;
    }

    @Override
    public boolean listAvailableModels(List<NativeModelInfo> retModels)
    {
        Pattern filePattern = Pattern.compile(LangSpecs.BINFILE_REGEX);

        String[] files = _fileManager.findFiles(filePattern);
        if (files != null && files.length > 0)
        {
            for (String file : files)
            {
                Matcher matcher = filePattern.matcher(file);
                if (matcher.matches())
                {
                    String lang    = LangSpecs.getMatchedBinfileLanguage(matcher);
                    String quality = LangSpecs.getMatchedBinfileQuality(matcher);
                    String rate    = LangSpecs.getMatchedBinfileRateKhz(matcher);

                    if (lang != null && rate != null)
                    {
                        retModels.add(new NativeModelInfo(lang, quality, Integer.parseInt(rate)));
                    }
                }
            }
        }

        return true;
    }

    @Override
    public boolean createElvis(String language, int samplingRate, String binFile)
    {
        if (_elvis != 0) {
            Logger.error(this, "createElvis: native Elvis already exists.");
            return false;
        }

        if (binFile == null) {
            /* User did not specify a binFile.  Try to load compiled in models. */
            if (language != null) {
                _elvis = elvisCreate(_fileManager, null, new String(language), samplingRate);
            } else {
                Logger.error(this, "No binfile or language specified.");
            }
        } else {
            /* If a binFile was specified, try to load the binFile. */
            BinFileParametersJni binInfo = getBinFileParameters(_fileManager, binFile);
            if (binInfo == null) {
                Logger.warn(this, "createElvis: Elvis binfile not found: " + binFile);
                if (language != null) {
                    /* The specified binFile does not exist, but a language was specified.  Try to load compiled in models. */
                    _elvis = elvisCreate(_fileManager, null, new String(language), samplingRate);
                }
            } else {
                /* If the user specified the language, both the language and sample rate must match the binFile. */
                if (language != null) {
                    if (!language.equals(binInfo._languageCode) || binInfo._sampleRate != samplingRate) {
                        Logger.error(this, "createElvis: wrong language / sampling rate detected for Elvis bin file");
                        Logger.error(this, "Found " + binInfo._languageCode + ":" + binInfo._sampleRate + ", expected " + language + ":" + samplingRate);
                        Logger.error(this, "Unable to create native Elvis with requested language and frequency.");
                        return false;
                    }
                } else {
                    language = binInfo._languageCode;
                    samplingRate = binInfo._sampleRate;
                }
                _elvis = elvisCreate(_fileManager, new String(binFile), null, 1234);
            }
        }

        if (_elvis != 0) {
            Logger.debug(this, "Created native Elvis");
            _language = language;
            _samplingRate = samplingRate;
            _binFile = binFile;
            _dirty = false;
            return true;
        } else {
            Logger.error(this, "Unable to create native Elvis");
            return false;
        }
    }

    @Override
    public boolean recreateElvis(boolean force)
    {
        if (_dirty || force) {
            if (_elvis != 0) {
                elvisDestroy(_elvis);
                _elvis = 0;
            }
            String language = _language;
            String binFile = _binFile;
            int samplingRate = _samplingRate;
            return createElvis(language, samplingRate, binFile);
        }
        return (_elvis != 0);
    }

    @Override
    public void destroyElvis()
    {
        if (_elvis == 0) {
            Logger.error(this, "destroyElvis: native Elvis does not exist!");
            return;
        }

        elvisDestroy(_elvis);
        _elvis = 0;
        Logger.debug(this, "Destroyed native Elvis");
    }

    @Override
    public Grammar loadGrammar(String fileName)
    {
        String psaFile = fileName + _language + _samplingRate + ".psa";
        String grmFile = fileName + _language + _samplingRate + ".grm";
        // TODO manage file names and extensions
        if (_elvis == 0) {
            Logger.error(this, "loadGrammar(" + fileName + "): native Elvis does not exist.");
            return null;
        }

        FileInputStream s = _fileManager.openFileForReading(grmFile);
        Grammar grammar = null;
        if (s != null)
        {
            try {
                JSONObject json = JSONUtils.readFromStream(s);
                s.close();
                grammar = Grammar.createFromJSON(json);
            } catch (Exception e) {
                Logger.error(this, "Error reading Elvis grm file", e);
            }
        }
        if (grammar == null) {
            return null;
        }

        int stat = elvisLoad(_elvis, psaFile, null, false);
        if (stat != ELVIS_OK) {
            Logger.error(this, "loadGrammar (" + psaFile + ") error loading Elvis: " + stat);
            return null;
        }
        _dirty = true;
        return grammar;
    }

    @Override
    public boolean saveGrammar(String fileName, Grammar grammar)
    {
        String psaFile = fileName + _language + _samplingRate + ".psa";
        String grmFile = fileName + _language + _samplingRate + ".grm";
        // TODO manage file names and extensions
        if (_elvis == 0) {
            Logger.error(this, "saveGrammar(" + fileName + "): native Elvis does not exist.");
            return false;
        }
        int stat = elvisSave(_elvis, psaFile);
        if (stat != ELVIS_OK) {
            Logger.error(this, "saveGrammar (" + psaFile + ") error saving native Elvis grammar: " + stat);
            return false;
        }

        FileOutputStream s = _fileManager.openFileForWriting(grmFile);
        if (s != null) {
            try {
                JSONUtils.saveToStream(grammar.toJSON(), s);
                s.close();
                return true;
            } catch (Exception e) {
                Logger.error(this, "Error saving Elvis grammar", e);
            }
        }
        return false;
    }

    @Override
    public int addConstraint(String name, int classCount, int transitionCount, boolean hasPenalties, boolean reset)
    {
        if (_elvis == 0) {
            Logger.error(this, "addConstraint (" + name + "): native Elvis does not exist!");
            return CONSTRAINT_ERROR;
        }

        _dirty = true;
        return elvisAddConstraint(_elvis, name, classCount, transitionCount, hasPenalties, reset);
    }

    @Override
    public boolean deleteConstraint(String name)
    {
        if (_elvis == 0) {
            Logger.error(this, "deleteConstraint (" + name + "): native Elvis does not exist!");
            return false;
        }

        return (ELVIS_OK == elvisDeleteConstraint(_elvis, name));
    }

    @Override
    public boolean constraintExists(String name) {
        if (_elvis == 0) {
            Logger.error(this, "constraintExists (" + name + "): native Elvis does not exist!");
            return false;
        }

        return elvisConstraintExists(_elvis, name);
    }

    @Override
    public boolean clearGrammar()
    {
        if (_elvis == 0) {
            Logger.error(this, "clearConstraints: native Elvis does not exist!");
            return false;
        }

        int status = elvisClearGrammar(_elvis);
        if (status != ELVIS_OK) {
            Logger.error(this, "Erroring Clearing Elvis constraints.");
            return false;
        }
        Logger.debug(this, "Cleared Elvis constraints.");
        return true;
    }

    @Override
    public int addWordClass(WordSlot slot)
    {
        if (slot == null) {
            Logger.error(this, "addWordClass (): invalid WordSlot!");
            return WORDCLASS_ERROR;
        }
        String name = slot.getId();
        if (_elvis == 0) {
            Logger.error(this, "addWordClass (" + name + "): native Elvis does not exist!");
            return WORDCLASS_ERROR;
        }
        boolean reset = (slot.getRebuildType() == WordSlot.RebuildModes.WORDSLOT_REBUILD_FULL ? true : false);
        int ret = elvisAddWordClass(_elvis, name, slot.getType(), reset);
        if (ret == WORDCLASS_ERROR)
            Logger.error(this, "Error adding word class to native Elvis: " + name + ".");
        _dirty = true;

        return ret;
    }

    @Override
    public int addWordClassMember(String wordclass, Word word, boolean addVariants)
    {
        if (_elvis == 0) {
            Logger.error(this, "addWordClassMembers (Class " + wordclass + "): native Elvis does not exist!");
            return WORD_ERROR;
        }
        if (word == null)
        {
            Logger.error(this, "addWordClassMembers (Class " + wordclass + "): word list is null!");
            return WORD_ERROR;
        }
        String surfaceForm = word.getSurfaceForm();
        String spokenForm = word.getSpokenForm();
        final int variantPenalty = 10;
        int status = elvisAddClassMember(_elvis, wordclass, surfaceForm, spokenForm, word.getUserId(), addVariants, variantPenalty);
        if (status == ELVIS_NO_PRONUNCIATION) {
            Logger.warn(this, "No pronunciation found.  Cannot add member to word class (" + wordclass + "): " + surfaceForm + "," + spokenForm);
            return WORD_NO_PRONUNCIATION;
        } else if (status != ELVIS_OK) {
            Logger.error(this, "Could not add member to word class (" + wordclass + "): " + surfaceForm + "," + spokenForm);
            return WORD_ERROR;
        }
        return WORD_ADDED;
    }

    @Override
    public boolean removeWordClassMember(String wordclass, Word word)
    {
        if (_elvis == 0) {
            Logger.error(this, "removeWordClassMember (Class " + wordclass + "): native Elvis does not exist!");
            return false;
        }
        if (word == null)
        {
            Logger.error(this, "removeWordClassMember (Class " + wordclass + "): word is null!");
            return false;
        }
        return elvisRemoveClassMember(_elvis, wordclass, word.getSurfaceForm());
    }

    @Override
    public int addExternalPronunciation(Word word)
    {
        if (word == null) {
            Logger.error(this, "addExternalPronunciation (): invalid word!");
            return WORD_ERROR;
        }

        if (_elvis == 0) {
            Logger.error(this, "addExternalPronunciation (" + word.toString() + "): native Elvis does not exist!");
            return WORD_ERROR;
        }

        _dirty = true;
        String surfaceForm = word.getSurfaceForm();
        String spokenForm = word.getSpokenForm();
        int status = elvisAddExternalPronunciation(_elvis, surfaceForm, spokenForm);
        if (status == ELVIS_NO_PRONUNCIATION) {
            Logger.warn(this, "addExternalPronunciation (" + word.toString() + "): no pronunciation found.");
            return WORD_NO_PRONUNCIATION;
        } else if (status != ELVIS_OK) {
            Logger.error(this, "addExternalPronunciation (" + word.toString() + "): native Elvis error.");
            return WORD_ERROR;
        }
        return WORD_ADDED;
    }

    @Override
    public int addTransition(String constraint, String fromWordClass, String toWordClass, int penalty)
    {
        if (_elvis == 0) {
            Logger.error(this, "Error adding transitions to constraint (" + constraint + "): " + fromWordClass + "," + toWordClass + ": native Elvis does not exist.");
            return TRANSITION_ERROR;
        }
        if (fromWordClass == null) {
            fromWordClass = Constraint.START;
        }
        if (toWordClass == null) {
            toWordClass = Constraint.END;
        }
        int status = elvisAddTransition(_elvis, constraint, fromWordClass, toWordClass, penalty);
        if (status != ELVIS_OK) {
            Logger.error(this, "Error adding transitions to constraint (" + constraint + "): " + fromWordClass + "," + toWordClass);
            return TRANSITION_ERROR;
        }
        return TRANSITION_ADDED;

    }

    @Override
    public boolean recognize(List<String> activeConstraints, boolean resetChannel)
    {
        if (_elvis == 0) {
            Logger.error(this, "Error starting recognize: native Elvis does not exist.");
            return false;
        }
        if (activeConstraints == null || activeConstraints.isEmpty()) {
            Logger.error(this, "Error starting recognize: empty active constraint list.");
            return false;
        }
        int constraintCount = activeConstraints.size();
        String constraints[] = new String[constraintCount];
        int count = 0;
        Logger.debug(this, "Starting Recognition with following constraints:");
        for (String c : activeConstraints)
        {
            constraints[count] = c;
            Logger.debug(this, "Constraint [" + count + "]: " + constraints[count]);
            count++;
        }
        return (ELVIS_OK == elvisStartRecognizing(_elvis, constraints, count, resetChannel));
    }

    @Override
    public boolean wakeupMode(List<String> wakeupPhrases, int penalty, int mode, int commandStartBailoutMs, int commandEndBailoutMs, int commandMinSpeechMs, boolean autoAdapt)
    {
        if (_elvis == 0) {
            Logger.error(this, "Error starting wakeup mode: native Elvis does not exist.");
            return false;
        }
        if (wakeupPhrases == null || wakeupPhrases.isEmpty()) {
            Logger.error(this, "Error starting wake-up: empty wake-up phrase list.");
            return false;
        }
        int wakeupPhraseCount = wakeupPhrases.size();
        String phrases[] = new String[wakeupPhraseCount];
        int count = 0;
        Logger.debug(this, "Starting Wakeup Mode with following phrases:");
        for (String s : wakeupPhrases)
        {
            phrases[count] = s;
            Logger.debug(this, "Phrase [" + count + "]: " + phrases[count]);
            count++;
        }
        _lastWakeupResult = null;
        return (ELVIS_OK == elvisStartWakeupMode(_elvis, phrases, penalty, mode, commandStartBailoutMs,commandEndBailoutMs, commandMinSpeechMs, autoAdapt));
    }

    @Override
    public boolean finishRecognizing(final boolean abort)
    {
        if (_elvis != 0)
        {
            return (ELVIS_OK == elvisStopRecognizing(_elvis, abort));
        }
        return false;
    }

    @Override
    public boolean finishWakeupMode(String fileName)
    {
        String adpFile = fileName + _language + _samplingRate + ".wuwadp";
        if (_elvis != 0)
        {
            return (ELVIS_SPEECH_ERROR != elvisStopWakeupMode(_elvis, adpFile));
        }
        return false;
    }

    @Override
    public boolean endpointingMode(int neededSamples, int ignoredSamples, boolean resetChannel, String loggingFolder)
    {
        if (_elvis == 0) {
            Logger.error(this, "Error starting endpointing mode: native Elvis does not exist.");
            return false;
        }
        Logger.debug(this, "Starting Endpointing Mode.");
        return (ELVIS_OK == elvisStartEndpointing(_elvis, neededSamples, ignoredSamples, resetChannel, loggingFolder));
    }

    @Override
    public int processEndpointingAudio(short[] shorts)
    {
        if (_elvis == 0) {
            Logger.error(this, "processEndpointingAudio: native Elvis does not exist.");
            return SPEECH_ERROR;
        }

        // Feed to Elvis
        int speech = SILENCE;
        if (shorts != null) {
            speech = elvisProcessEndpointingAudio(_elvis, shorts, shorts.length);
        }
        return speech;
    }

    @Override
    public boolean finishEndpointing()
    {
        if (_elvis != 0)
        {
            return (ELVIS_OK == elvisStopEndpointing(_elvis));
        }
        return false;
    }

    @Override
    public boolean adapt(String phrase)
    {
        if (_elvis == 0) {
            Logger.error(this, "Error starting adaptation: native Elvis does not exist.");
            return false;
        }
        if (phrase == null || phrase.length() == 0) {
            Logger.error(this, "Error starting adaptation: invalid phrase.");
            return false;
        }
        return (ELVIS_OK == elvisStartAdaptation(_elvis, phrase));
    }

    @Override
    public ElvisAdaptationResult processAdaptation()
    {
        boolean status = false;
        String miscLogging = null;
        if (_elvis != 0)
        {
            status = ELVIS_OK == elvisProcessAdaptation(_elvis);
            miscLogging = getMiscLogging();
        }
        return new ElvisAdaptationResult(status, miscLogging);
    }

    @Override
    public boolean loadAdaptation(String fileName, String ext, boolean isUnsupervised)
    {
        String adpFile = fileName + _language + _samplingRate + ext;
        // TODO manage file name and extension
        if (_elvis == 0) {
            Logger.error(this, "loadAdaptation(" + fileName + "): native Elvis does not exist.");
            return false;
        }

        if (_fileManager.exists(adpFile))
        {
            int stat = elvisLoad(_elvis, null, adpFile, isUnsupervised);
            if (stat != ELVIS_OK) {
                Logger.error(this, "loadAdaptation (" + adpFile + ") error loading Elvis with adaptation data: " + stat);
                return false;
            }
        }
        return true;
    }

    @Override
    public boolean finishAdapting(String fileName, boolean abort)
    {
        String adpFile = fileName + _language + _samplingRate + ".adp";
        if (_elvis != 0)
        {
            return (ELVIS_OK == elvisEndAdaptationSession(_elvis, adpFile, abort));
        }
        return false;
    }

    @Override
    public boolean checkAdaptationStatus(String fileName)
    {
        if (_elvis == 0) {
            Logger.error(this, "Error checking adaptation status: native Elvis does not exist.");
            return false;
        }
        if (_fileManager.exists(fileName + _language + _samplingRate + ".adp")) {
            return elvisCheckAdaptationStatus(_elvis);
        }
        return false;
    }

    @Override
    public boolean clearAdaptation(String fileName, String ext, boolean all)
    {
        if (_elvis == 0) {
            Logger.error(this, "clearAdaptation: native Elvis does not exist!");
            return false;
        }
        elvisDestroy(_elvis);
        _elvis = 0;

        if (all) {
            String[] files = _fileManager.findFiles(ext);
            if (files != null) {
                for (String f : files) {
                    _fileManager.delete(f);
                }
            }
        }
        else {
            _fileManager.delete(fileName + _language + _samplingRate + ext);
        }
        return createElvis(_language, _samplingRate, _binFile);
    }

    @Override
    public int processAudio(short[] shorts)
    {
        if (_elvis == 0) {
            Logger.error(this, "processAudio: native Elvis does not exist.");
            return SPEECH_ERROR;
        }

        // Feed to Elvis
        int speech = SILENCE;
        if (shorts != null) {
            speech = elvisProcessAudio(_elvis, shorts, shorts.length);
        }
        return speech;
    }

    @Override
    public int processWakeupAudio(short[] shorts, long timestamp)
    {
        if (_elvis == 0) {
            Logger.error(this, "processWakeupAudio: native Elvis does not exist.");
            return SPEECH_ERROR;
        }

        // Feed to Elvis
        int status = WUW_NO_WAKEUP;
        if (shorts != null) {
            status = elvisProcessWakeupAudio(_elvis, shorts, shorts.length);
        }
        _lastWakeupStatus = status;
        _lastWakeupTimestamp = timestamp;
        return status;
    }

    @Override
    public int getSoundLevel()
    {
        if (_elvis == 0) {
            Logger.error(this, "getSoundLevel: native Elvis does not exist.");
            return -1;
        }
        return elvisGetSoundLevel(_elvis);
    }

    @Override
    public ElvisResult getResult(Grammar grammar)
    {
        ElvisResult result = null;
        ElvisNbestList nbest = null;
        if (_elvis == 0) {
            Logger.error(this, "getResult: native Elvis does not exist.");
            return null;
        }
        int choices = elvisGetChoiceCount(_elvis);
        int gateConfidence = elvisGetGateConfidence(_elvis);

        if (choices < 0) {
            Logger.error(this, "getResult: error retrieving choices.");
        }
        else if (choices == 0) {
            Logger.debug(this, "getResult: no choices available.");
        }
        else if (choices > 0) {
            Logger.debug(this, "getResult: " + choices + " choices.");
            ArrayList<NbestEntryJni> jniEntries = new ArrayList<NbestEntryJni>(choices);
            for (int i = 0; i < choices; i ++) {
                int status = elvisGetChoiceListEntry(_elvis, i, jniEntries);
                if (status != ELVIS_OK) {
                    Logger.error(this, "elvis result error.");
                    break;
                }
            }

            ArrayList<ElvisNbestList.Entry> entries = new ArrayList<ElvisNbestList.Entry>(choices);

            for (int i = 0; i < jniEntries.size(); i++)
            {
                entries.add(jniEntries.get(i).createRealEntry(grammar));
            }

            nbest = new ElvisNbestList(entries);
        }
        if (nbest != null  && !nbest.isEmpty()) {
            result = new ElvisResult(nbest, gateConfidence, ElvisResult.WAKEUP_TIMESTAMP_UNAVAILABLE, ElvisResult.WAKEUP_TIMESTAMP_UNAVAILABLE, ElvisResult.WAKEUP_TIMESTAMP_UNAVAILABLE, ElvisResult.WAKEUP_TIMESTAMP_UNAVAILABLE, ElvisResult.WakeupStatus.NO_COMMAND, getMiscLogging());
            DefaultResultFilter drf = new DefaultResultFilter();
            result = drf.applyFilter(result);
        } else {
            Logger.warn(this, "No nbest results.");
        }
        return result;
    }

    @Override
    public ElvisResult getWakeupResult(boolean seamless)
    {
        ElvisResult result = null;
        ElvisNbestList nbest = null;
        if (_elvis == 0) {
            Logger.error(this, "getWakeupResult: native Elvis does not exist.");
            return null;
        }

        AudioType audioType = new AudioType(Encoding.PCM_16, _samplingRate);
        switch (_lastWakeupStatus)
        {
            case NativeElvis.WUW_WAKEUP:
                // Get the initial result
                int start = elvisGetWakeupStartTime(_elvis, false);
                int end = elvisGetWakeupEndTime(_elvis, false);

                ArrayList<NbestEntryJni> jniEntries = new ArrayList<NbestEntryJni>(1);
                int status = elvisGetWakeupEntry(_elvis, jniEntries);
                if (status != ELVIS_OK || jniEntries.isEmpty()) {
                    Logger.error(this, "elvis result error.");
                } else
                {
                    long wakeupStartTimestamp = ElvisResult.WAKEUP_TIMESTAMP_UNAVAILABLE;
                    if (start != -1) {
                        wakeupStartTimestamp = _lastWakeupTimestamp - audioType.getDuration(start);
                    }
                    else {
                        Logger.error(this, "getWakeupResult: error retrieving wake-up start time.");
                    }
                    long wakeupEndTimestamp = ElvisResult.WAKEUP_TIMESTAMP_UNAVAILABLE;
                    if (end != -1) {
                        wakeupEndTimestamp = _lastWakeupTimestamp - audioType.getDuration(end);
                    }
                    else {
                        Logger.error(this, "getWakeupResult: error retrieving wake-up end time.");
                    }

                    nbest = new ElvisNbestList(jniEntries.get(0).createRealEntry(null));
                    result = new ElvisResult(nbest, ElvisResult.GATE_CONFIDENCE_UNAVAILABLE,
                                             wakeupStartTimestamp,
                                             wakeupEndTimestamp,
                                             ElvisResult.WAKEUP_TIMESTAMP_UNAVAILABLE,
                                             ElvisResult.WAKEUP_TIMESTAMP_UNAVAILABLE,
                                             seamless ? ElvisResult.WakeupStatus.MAYBE_COMMAND : ElvisResult.WakeupStatus.NO_COMMAND,
                                             getMiscLogging());
                }
                break;

            case NativeElvis.WUW_COMMAND_STARTED:
                if (_lastWakeupResult != null)
                {
                    int commandStart = elvisGetWakeupStartTime(_elvis, true);

                    long commandStartTimestamp = ElvisResult.WAKEUP_TIMESTAMP_UNAVAILABLE;
                    if (commandStart != -1) {
                        commandStartTimestamp = _lastWakeupTimestamp - audioType.getDuration(commandStart);
                    }
                    else {
                        Logger.error(this, "getWakeupResult: error retrieving wake-up command start time.");
                    }

                    result = new ElvisResult(_lastWakeupResult.getChoiceList(),
                                             _lastWakeupResult.getGateConfidence(),
                                             _lastWakeupResult.getWakeupStartTimestamp(),
                                             _lastWakeupResult.getWakeupEndTimestamp(),
                                             commandStartTimestamp,
                                             ElvisResult.WAKEUP_TIMESTAMP_UNAVAILABLE,
                                             ElvisResult.WakeupStatus.COMMAND_STARTED,
                                             getMiscLogging());
                }
                break;

            case NativeElvis.WUW_COMMAND_ENDED:
                if (_lastWakeupResult != null)
                {
                    int commandEnd = elvisGetWakeupEndTime(_elvis, true);

                    long commandEndTimestamp = ElvisResult.WAKEUP_TIMESTAMP_UNAVAILABLE;
                    if (commandEnd != -1) {
                        commandEndTimestamp = _lastWakeupTimestamp - audioType.getDuration(commandEnd);
                    }
                    else {
                        Logger.error(this, "getWakeupResult: error retrieving wake-up command end time.");
                    }

                    result = new ElvisResult(_lastWakeupResult.getChoiceList(),
                                             _lastWakeupResult.getGateConfidence(),
                                             _lastWakeupResult.getWakeupStartTimestamp(),
                                             _lastWakeupResult.getWakeupEndTimestamp(),
                                             _lastWakeupResult.getWakeupCommandStartTimestamp(),
                                             commandEndTimestamp,
                                             ElvisResult.WakeupStatus.COMMAND_ENDED,
                                             getMiscLogging());
                }
                break;

            case NativeElvis.WUW_COMMAND_TIMEOUT:
                if (_lastWakeupResult != null)
                {
                    result = new ElvisResult(_lastWakeupResult.getChoiceList(),
                                             _lastWakeupResult.getGateConfidence(),
                                             _lastWakeupResult.getWakeupStartTimestamp(),
                                             _lastWakeupResult.getWakeupEndTimestamp(),
                                             _lastWakeupResult.getWakeupCommandStartTimestamp(),
                                             _lastWakeupResult.getWakeupCommandEndTimestamp(),
                                             ElvisResult.WakeupStatus.NO_COMMAND,
                                             getMiscLogging());
                }
                break;

            case NativeElvis.WUW_NO_WAKEUP:
            case NativeElvis.WUW_PENDING_WAKEUP:
            case NativeElvis.WUW_ERROR:
            default:
                // Nothing to do
                break;
        }

        _lastWakeupResult = result;
        return result;
    }

    @Override
    public boolean checkWakeupPhrase(String wakeupPhrase)
    {
        if (_elvis == 0) {
            Logger.error(this, "Error checking wakeup phrase: native Elvis does not exist.");
            return false;
        }
        return elvisCheckWakeupPhrase(_elvis, wakeupPhrase);
    }

    @Override
    public void releaseUnusedMemory()
    {
        if (_elvis == 0) {
            Logger.error(this, "Error releasing unused memory: native Elvis does not exist.");
            return;
        }
        elvisReleaseUnusedMemory(_elvis);
    }

    @Override
    public void enableElvisVerboseLogging(boolean verbose)
    {
        elvisEnableVerboseLogging(verbose);
    }

    @Override
    public boolean enableElvisRecognitionLogging(String logPrefix, int logMillis, String uuid)
    {
        if (_elvis == 0) {
            Logger.error(this, "Error enabling recognition logging: native Elvis does not exist.");
            return false;
        }
        return (elvisEnableRecognitionLogging(_elvis, logPrefix, logMillis, uuid) == 0);
    }

    @Override
    public int getElvisRecognitionLoggingStatus()
    {
        if (_elvis == 0) {
            Logger.error(this, "Error getting logging status: native Elvis does not exist.");
            return Logging.LOGGING_ERROR;
        }
        return elvisGetRecognitionLoggingStatus(_elvis);
    }

    @Override
    public boolean copyElvisRecognitionLogs(String destinationFolder, String destinationPrefix)
    {
        String sourcePsaFile = destinationPrefix + "elvisdump.psa";
        String absDestPsaFile = destinationFolder + File.separator + sourcePsaFile;
        String sourceRprFile = destinationPrefix + "elvisdump.rpr";
        String absDestRprFile = destinationFolder + File.separator + sourceRprFile;
        String sourceUttFile = destinationPrefix + "elvisdump.utt";
        String absDestUttFile = destinationFolder + File.separator + sourceUttFile;
        String sourceWavFile = destinationPrefix + "elvisdump.sph";
        String absDestWavFile = destinationFolder + File.separator + sourceWavFile;

        File folder = new File(destinationFolder);
        if (!folder.exists()) {
            if (!folder.mkdir()) {
                Logger.error(this, "Destination folder cannot be created.");
                return false;
            }
        } else if (!folder.isDirectory()) {
            Logger.error(this, "Destination folder is not a valid directory.");
            return false;
        }
        if (copyFileToAbsoluteLocation(sourcePsaFile, absDestPsaFile))
            if(copyFileToAbsoluteLocation(sourceRprFile, absDestRprFile))
                if (copyFileToAbsoluteLocation(sourceUttFile, absDestUttFile))
                    return copyFileToAbsoluteLocation(sourceWavFile, absDestWavFile);
        return false;
    }

    private boolean copyFileToAbsoluteLocation(String source, String destination)
    {
        FileInputStream fis = _fileManager.openFileForReading(source);
        if (fis == null) {
            Logger.error(this, "Unable to read file:" + source);
            return false;
        }
        FileOutputStream fos = null;

        File destFile = new File (destination);
        try {
            if (destFile.exists()) {
                if (!destFile.delete()) {
                    Logger.error(this, "Can't delete file: " + destination);
                    fis.close();
                    return false;
                }
            }
            if (!destFile.createNewFile()) {
                Logger.error(this, "Could not create log output file:" + destination);
                fis.close();
                return false;
            }
            fos = new FileOutputStream(destFile);
            byte[] buff = new byte[1024];
            int bytesRead = 0;
            while((bytesRead = fis.read(buff)) > -1 )
            {
                fos.write(buff, 0, bytesRead);
            }
            fis.close();
            fis = null;
            fos.close();
            fos = null;
        } catch (IOException e) {
            Logger.error(this, "Error copying file:" + e.toString());
            try {
                if (fis != null)
                    fis.close();
            } catch (IOException ioe) {
                Logger.error(this, "Error closing file:" + ioe.toString());
            }
            try {
                if (fos != null)
                    fos.close();
            } catch (IOException ioe) {
                Logger.error(this, "Error closing file:" + ioe.toString());
            }
            return false;
        }
        return true;
    }

    private String getMiscLogging()
    {
        int miscLoggingLen = elvisGetMiscLoggingSize(_elvis);
        if (miscLoggingLen > 0)
        {
            byte[] miscLogging = new byte[miscLoggingLen];
            Arrays.fill(miscLogging, (byte) 0x00);
            int ret = elvisGetMiscLogging(_elvis, miscLogging);
            if (ret == 0)
            {
                String miscLoggingStr = null;
                try {
                    int idx = 0;
                    for (idx = 0; idx < miscLogging.length && miscLogging[idx] != 0x00; idx++) {}
                    miscLoggingStr = new String(miscLogging, 0, idx, "UTF-8");
                } catch (UnsupportedEncodingException e) {
                    Logger.error(this, "Unable to get ELVIS misc. logging.", e);
                }
                return miscLoggingStr;
            }
        }
        return null;
    }

    @Override
    public boolean logMiscInfo(String tag, String data)
    {
        Checker.checkStringArgForNullOrEmpty("tag", tag);
        Checker.checkStringArgForNullOrEmpty("data", data);

        return elvisLogMiscInfo(_elvis, tag, data);
    }

    @Override
    public int [] createBlob(List<String> phrases, int numPhrases, int decoyPenalty, boolean isMagic, int maxDuration)
    {
        if (_elvis == 0) {
            Logger.error(this, "createBlob: native Elvis does not exist.");
            return null;
        }
        if (phrases == null || phrases.isEmpty()) {
            Logger.error(this, "createBlob: empty phrase list.");
            return null;
        }
        int wakeupPhraseCount = phrases.size();
        String phrasesAry[] = new String[wakeupPhraseCount];
        int count = 0;
        Logger.debug(this, "createBlob with following phrases:");
        for (String s : phrases)
        {
            phrasesAry[count] = s;
            Logger.debug(this, "Phrase [" + count + "]: " + phrasesAry[count]);
            count++;
        }
        return elvisCreateBlob(_elvis, phrasesAry, numPhrases, decoyPenalty,
                               isMagic ? 1 : 0, maxDuration);
    }


    @Override
    public int setTargetCPU(String cpu)
    {
        if (_elvis == 0) {
            Logger.error(this, "setTargetCPU: native Elvis does not exist.");
            return -1;
        }
        return elvisSetTargetCPU(_elvis, cpu);
    }

    @Override
    public int forgetBlob()
    {
        if (_elvis == 0) {
            Logger.error(this, "forgetBlob: native Elvis does not exist.");
            return -1;
        }
        return elvisForgetBlob(_elvis);
    }

    @Override
    public int evaluateWakeupPhrase(String phrase)
    {
        if (_elvis == 0) {
            Logger.error(this, "evaluateWakeupPhrase: native Elvis does not exist.");
            return -1;
        }
        return elvisEvaluateWakeupPhrase(_elvis, phrase);
    }

}
