package com.nuance.dragon.toolkit.elvis;

import java.util.List;

import com.nuance.dragon.toolkit.oem.api.internal.Checker;


/**
 * A container for all all the parameters consumed
 * by an Elvis wake-up session.
 */
public class WakeupParam
{
    /** Default parameters */
    public static final int DEFAULT_PENALTY = 0;
    public static final int DEFAULT_COMMAND_START_BAILOUT_MS = -1;
    public static final int DEFAULT_COMMAND_END_BAILOUT_MS = 900;
    public static final int DEFAULT_COMMAND_MIN_SPEECH_MS = 200;

    /**
     * Elvis Wakeup Mode
     */
    public static enum WakeupMode
    {
        /** The standard mode for wake-up, with very low FA */
        WAKEUP_LONG,
        /**
         * The mode is used where a low FR rate is needed, and a higher FA rate is acceptable
         */
        WAKEUP_SHORT,
        /**
         * The mode of wake-up is immediately followed by a command, wake-up detection is followed by
         * an endpointing session.
         */
        WAKEUP_COMMAND,
        /**
         * The mode is optimized for low latency, low FR, and a higher FA rate is acceptable.
         * It is highly recommended that penalty is set to around -30 for this mode.
         */
        WAKEUP_SNAP
    };

    /**
     * Builder for creating instances of {@code WakeupParam}.
     */
    public static class Builder
    {
        private List<String> phrases;
        private WakeupMode mode;
        private int penalty;
        private int commandStartBailoutMs;
        private int commandEndBailoutMs;
        private int commandMinSpeechMs;
        private boolean isAutoAdapt;

        /**
         * Creates the builder with mandatory parameter(s).
         * @param phrases The active wake-up phrases. These can't be too short.
         */
        public Builder(List<String> phrases)
        {
            this.phrases = phrases;
            this.mode = WakeupMode.WAKEUP_LONG;
            this.penalty = DEFAULT_PENALTY;
            this.commandStartBailoutMs = DEFAULT_COMMAND_START_BAILOUT_MS;
            this.commandEndBailoutMs = DEFAULT_COMMAND_END_BAILOUT_MS;
            this.commandMinSpeechMs = DEFAULT_COMMAND_MIN_SPEECH_MS;
            this.isAutoAdapt = false;
        }

        /**
         * Set wake-up mode.
         * @param mode A wake-up mode from {@link WakeupMode}. Default is {@link WakeupMode#WAKEUP_LONG}.
         *
         */
        public Builder setMode(WakeupMode mode)
        {
            this.mode = mode;
            return this;
        }

        /**
         * Set custom penalty.
         * @param penalty A positive penalty will increase False Accept and lower False Reject.
         *                Default is {@link #DEFAULT_PENALTY}.
         */
        public Builder setPenalty(int penalty)
        {
            this.penalty = penalty;
            return this;
        }

        /**
         * Set custom command start bailout duration.
         * @param commandStartBailoutMs The expected silence duration (in ms) to differentiate
         *                              between a regular wake-up or a seam-less wake-up.
         *                              If set to default(-1), a minimum silence length is set.
         *
         *                              Default is {@link #DEFAULT_COMMAND_START_BAILOUT_MS}.
         *
         * <p>
         * Note: we need to use {@link #setMode(WakeupParam.WakeupMode)} to turn on command wake-up mode explicitly.
         */
        public Builder setCommandStartBailoutDuration(int commandStartBailoutMs)
        {
            this.commandStartBailoutMs = commandStartBailoutMs;
            return this;
        }

        /**
         * Set custom command end bailout duration.
         * @param commandEndBailoutMs The expected silence duration (in ms) after a wake-up + command
         *                            has been uttered in order to declare command-ended / end-of-speech.
         *                            Default is {@link #DEFAULT_COMMAND_END_BAILOUT_MS}.
         */
        public Builder setCommandEndBailoutDuration(int commandEndBailoutMs)
        {
            this.commandEndBailoutMs = commandEndBailoutMs;
            return this;
        }

        /**
         * Set custom command minimum speech duration.
         * @param commandMinSpeechMs The minimum duration (in ms) of a command in order
         *                           to differentiate a command from noise /speech glitches
         *                           following the wake-up word.
         *                           Default is {@link #DEFAULT_COMMAND_MIN_SPEECH_MS}.
         */
        public Builder setCommandMinSpeechDuration(int commandMinSpeechMs)
        {
            this.commandMinSpeechMs = commandMinSpeechMs;
            return this;
        }

        /**
         * Set wake-up adaptation mode, which by default is disabled.
         */
        public Builder setAutoAdapt()
        {
            this.isAutoAdapt = true;
            return this;
        }

        /**
         * Build the parameters.
         * @return The {@link WakeupParam} instance.
         */
        public WakeupParam build()
        {
            Checker.checkArgForNull("phrases", phrases);
            Checker.checkArgForCondition("commandStartBailoutMs", ">= -1", commandStartBailoutMs >= -1);
            Checker.checkArgForCondition("commandEndBailoutMs", ">= 0", commandEndBailoutMs >= 0);
            Checker.checkArgForCondition("commandMinSpeechMs", ">= 0", commandMinSpeechMs >= 0);
            return new WakeupParam(this);
        }
    }

    private final List<String> _phrases;
    private final WakeupMode _mode;
    private final int _penalty;
    private final int _commandStartBailoutMs;
    private final int _commandEndBailoutMs;
    private final int _commandMinSpeechMs;
    private final boolean _isAutoAdapt;

    private WakeupParam(Builder builder)
    {
        _phrases = builder.phrases;
        _mode = builder.mode;
        _penalty = builder.penalty;
        _commandStartBailoutMs = builder.commandStartBailoutMs;
        _commandEndBailoutMs = builder.commandEndBailoutMs;
        _commandMinSpeechMs = builder.commandMinSpeechMs;
        _isAutoAdapt = builder.isAutoAdapt;
    }

    List<String> getPhrases()
    {
        return _phrases;
    }

    WakeupMode getMode()
    {
        return _mode;
    }

    int getPenalty()
    {
        return _penalty;
    }

    int getCommandStartBailoutMs()
    {
        return _commandStartBailoutMs;
    }

    int getCommandEndBailoutMs()
    {
        return _commandEndBailoutMs;
    }

    int getCommandMinSpeechMs()
    {
        return _commandMinSpeechMs;
    }

    boolean isAutoAdapt()
    {
        return _isAutoAdapt;
    }

    @Override
    public String toString() {
        return "WakeupParam [_phrases=" + _phrases + ", _mode" + _mode
                + ", _penalty=" + _penalty
                + ", _commandStartBailoutMs=" + _commandStartBailoutMs
                + ", _commandEndBailoutMs=" + _commandEndBailoutMs
                + ", _commandMinSpeechMs=" + _commandMinSpeechMs
                + ", _isAutoAdapt=" + _isAutoAdapt + "]";
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + _commandEndBailoutMs;
        result = prime * result + _commandMinSpeechMs;
        result = prime * result + _commandStartBailoutMs;
        result = prime * result + (_isAutoAdapt ? 1231 : 1237);
        result = prime * result + _penalty;
        result = prime * result + ((_mode == null) ? 0 : _mode.hashCode());
        result = prime * result
                + ((_phrases == null) ? 0 : _phrases.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        WakeupParam other = (WakeupParam) obj;
        if (_commandEndBailoutMs != other._commandEndBailoutMs)
            return false;
        if (_commandMinSpeechMs != other._commandMinSpeechMs)
            return false;
        if (_commandStartBailoutMs != other._commandStartBailoutMs)
            return false;
        if (_isAutoAdapt != other._isAutoAdapt)
            return false;
        if (_penalty != other._penalty)
            return false;
        if (_mode != other._mode)
            return false;
        if (_phrases == null) {
            if (other._phrases != null)
                return false;
        } else if (!_phrases.equals(other._phrases))
            return false;
        return true;
    }
}
