package com.nuance.dragon.toolkit.elvis;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import junit.framework.Assert;

import org.json.JSONException;

import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.core.WordList;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONArray;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * Represents an Elvis word class.
 */
public class WordSlot implements JSONCompliant
{
    private static final String JSON_KEY_ID = "id";
    private static final String JSON_KEY_WORDS = "words";
    private static final String JSON_KEY_SLOT_TYPE = "slot_type";
    private static final String JSON_KEY_REBUILD_MODE = "rebuild_mode";
    private static final String JSON_KEY_REQUIRES_WORDLIST = "requires_wordlist";

    private final List<Word> _words;
    private final List<WordList> _wordLists;
    private final String _id;
    private int _type;
    private final int _rebuildMode;
    private boolean _wordListRequired;

    /**
     * The word slot types supported by Elvis.
     * <p>
     * The type serves as a hint to the Elvis engine about how to tokenize text
     * and compute pronunciations.
     */
    public static final class Types
    {
        /** Regular words and phrases. Only exact matches will be recognized. */
        public static final int WORDSLOT_DEFAULT = 0;

        /** Names. Partial name results (e.g. first name) will be recognized. */
        public static final int WORDSLOT_NAME = 1;

        /** Music. Partial results (e.g. part of a title) will be recognized. */
        public static final int WORDSLOT_MUSIC = 2;

        /** Phone numbers. */
        public static final int WORDSLOT_PHONE_DIGIT = 3;

        /** Placeholder for text that will be recognized elsewhere, like on the
         *  server during a hybrid recognition (short end-of-speech timeout). */
        public static final int WORDSLOT_GENERIC_SHORT = 4;

        /** Placeholder for text that will be recognized elsewhere, like on the
         *  server during a hybrid recognition (long end-of-speech timeout). */
        public static final int WORDSLOT_GENERIC_LONG = 5;

        /** Names. variant name results (e.g. first name/last name/full name) will be recognized. */
        public static final int WORDSLOT_VARIANT_NAME = 6;
    }

    /**
     * The modes that should be applied when adding a word slot to Elvis during
     * a grammar rebuild.
     */
    public static final class RebuildModes
    {
        /**
         * When a grammar is loaded, the definition of this WordSlot will replace any
         * previous definition.  The old value of the WordSlot will be deleted before
         * the WordSlot is added.
         */
        public static final int WORDSLOT_REBUILD_FULL = 0;

        /**
         * If the WordSlot already exists, existing words will be updated (added or
         * deleted). Combine the previous definition and the new definition.
         */
        public static final int WORDSLOT_REBUILD_UPDATE = 1;

        /**
         * If the WordSlot already exists, it will be skipped during the rebuild.  Keep
         * the previously loaded value and ignore the new definition.
         */
        public static final int WORDSLOT_REBUILD_SKIP_EXISTING = 2;
    }

    /**
     * Iterates through all the words in this word slot.
     */
    public static class WordIterator
    {
        private final LinkedList<Word> _words;
        private final LinkedList<WordList.WordIterator> _wordListIterators;
        private final LinkedList<WordList.WordIterator> _completedWordListIterators;
        private WordAction _next;

        private WordIterator(List<Word> words, List<WordList> wordLists, boolean fullRebuild)
        {
            _words = new LinkedList<Word>();
            _wordListIterators = new LinkedList<WordList.WordIterator>();
            _completedWordListIterators = new LinkedList<WordList.WordIterator>();

            if (words != null)
                _words.addAll(words);

            if (wordLists != null)
            {
                for (WordList list : wordLists)
                {
                    WordList.WordIterator it;
                    if (fullRebuild)
                        it = list.getFullIterator();
                    else
                        it = list.getModifiedIterator();

                    _wordListIterators.add(it);
                }
            }

            _next = getNextWord();
        }

        /**
         * Get the next word action.
         * @return The next word action, or null when the end has been reached.
         */
        public WordAction getNext()
        {
            WordAction next = _next;
            if (_next != null)
            {
                _next = getNextWord();
            }
            return next;
        }

        /**
         * @return True if there is a next.
         */
        public boolean hasNext()
        {
            return _next != null;
        }

        /**
         * Call when all the words have been consumed, and they have been applied.
         * @param success Whether or not the words have been successfully applied.
         */
        public void finished(boolean success)
        {
            Assert.assertTrue(_wordListIterators.isEmpty());

            for (WordList.WordIterator it : _completedWordListIterators)
            {
                if (success)
                    it.acceptChanges();
                else
                    it.discardChanges();
            }
        }

        private WordAction getNextWord()
        {
            if (!_words.isEmpty())
            {
                WordAction wa = null;
                Word w = _words.remove();
                if (w != null)
                    wa = new WordAction (w, true);
                return wa;
            }

            while (!_wordListIterators.isEmpty())
            {
                WordAction action = null;
                if (_wordListIterators.getFirst().hasNext())
                {
                    action = _wordListIterators.getFirst().next();
                }

                if (action != null)
                {
                    return action;
                }

                _completedWordListIterators.add(_wordListIterators.remove());
            }

            // Signal that the list is done
            return null;
        }
    }

    /**
     * Uses default type of {@link Types#WORDSLOT_DEFAULT} and default rebuild mode
     * of {@link RebuildModes#WORDSLOT_REBUILD_SKIP_EXISTING}.
     * @param id A string that identifies the WordSlot.  Must not be null or empty string.
     */
    public WordSlot(String id)
    {
        this(id, Types.WORDSLOT_DEFAULT, RebuildModes.WORDSLOT_REBUILD_SKIP_EXISTING);
    }

    /**
     * Uses default rebuild mode of {@link RebuildModes#WORDSLOT_REBUILD_SKIP_EXISTING}.
     * @param id A string that identifies the WordSlot.  Must not be null or empty string.
     * @param type One of {@link Types}.
     */
    public WordSlot(String id, int type)
    {
        this(id, type, RebuildModes.WORDSLOT_REBUILD_SKIP_EXISTING);
    }

    /**
     * @param id A string that identifies the WordSlot.  Must not be null or empty string.
     * @param type One of {@link Types}.
     * @param rebuildMode One of {@link RebuildModes}.
     */
    public WordSlot(String id, int type, int rebuildMode)
    {
        Checker.checkStringArgsForAllNullOrEmpty("id", id);
        _id = id;
        _type = type;
        _words = new ArrayList<Word>();
        _wordLists = new ArrayList<WordList>();
        if (_type == Types.WORDSLOT_GENERIC_SHORT || _type == Types.WORDSLOT_GENERIC_LONG) {
            _rebuildMode = RebuildModes.WORDSLOT_REBUILD_SKIP_EXISTING;
            /* Add members for "generic" ("garbage") speech. */
            _words.add(new Word(":GARBAGE:", null));
            _words.add(new Word("0", null));
            _words.add(new Word("1", null));
            _words.add(new Word("2", null));
            _words.add(new Word("3", null));
            _words.add(new Word("4", null));
            _words.add(new Word("5", null));
            _words.add(new Word("6", null));
            _words.add(new Word("7", null));
            _words.add(new Word("8", null));
            _words.add(new Word("9", null));
        } else
        {
            _rebuildMode = rebuildMode;
        }
    }

    /**
     * Create a copy of the word slot.
     * @return A copy, never null.
     */
    public WordSlot copy()
    {
        WordSlot copy = new WordSlot(_id, _type, _rebuildMode);
        copy._words.addAll(_words);
        copy._wordLists.addAll(_wordLists);
        copy._wordListRequired = _wordListRequired;
        return copy;
    }

    /* (non-Javadoc)
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((_id == null) ? 0 : _id.hashCode());
        result = prime * result + _rebuildMode;
        result = prime * result + _type;
        result = prime * result + (_wordListRequired ? 1231 : 1237);
        result = prime * result + ((_wordLists == null) ? 0 : _wordLists.hashCode());
        result = prime * result + ((_words == null) ? 0 : _words.hashCode());
        return result;
    }

    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (!(obj instanceof WordSlot))
            return false;
        WordSlot other = (WordSlot) obj;
        if (_id == null)
        {
            if (other._id != null)
                return false;
        } else if (!_id.equals(other._id))
            return false;
        if (_rebuildMode != other._rebuildMode)
            return false;
        if (_type != other._type)
            return false;
        if (_wordListRequired != other._wordListRequired)
            return false;
        if (_wordLists == null)
        {
            if (other._wordLists != null)
                return false;
        } else if (!_wordLists.equals(other._wordLists))
            return false;
        if (_words == null)
        {
            if (other._words != null)
                return false;
        } else if (!_words.equals(other._words))
            return false;
        return true;
    }

    /**
     * Get the ID of the word slot.
     *
     * @return The word slot ID as passed to the constructor.
     */
    public String getId()
    {
        return _id;
    }

    /**
     * Get the type of the word slot.
     *
     * @return The word slot type as passed to the constructor.
     */
    public int getType()
    {
        return _type;
    }

    /**
     * A package private method that indicates whether Variants should be used for this WordSlot.
     *
     * @return True if variants should be used, false if they should not.
     */
    boolean useVariants()
    {
        // Currently, Names and Music types always use variants.
        // In the future, this option might be setable by the user.
        if (_type == Types.WORDSLOT_NAME || _type == Types.WORDSLOT_MUSIC) {
            return true;
        }
        return false;
    }

    /**
     * Is this a generic speech WordSlot?
     *
     * @return true if type is WORDSLOT_GENERIC_SHORT or WORDSLOT_GENERIC_LONG.
     */
    public boolean isGeneric()
    {
        return ((_type == Types.WORDSLOT_GENERIC_SHORT) || (_type == Types.WORDSLOT_GENERIC_LONG));
    }

    /**
     * Get the Always Rebuild status of this Word Slot.
     *
     * @return the Rebuild Type for this Word Slot.
     */
    public int getRebuildType()
    {
        if (_rebuildMode != RebuildModes.WORDSLOT_REBUILD_FULL)
        {
            for (WordList list : _wordLists)
            {
                if (list.fullUpdateRequired())
                    return RebuildModes.WORDSLOT_REBUILD_FULL;
            }
        }
        return _rebuildMode;
    }

    /**
     * Get an iterator for the words in this word slot. May be called on any thread.
     * @param fullRebuild true if the full word list should always be returned; false
     * to automatically determine which words to return.
     * @return The iterator, never null.
     */
    public WordIterator getWords(boolean fullRebuild)
    {
        return new WordIterator(_words, _wordLists, fullRebuild || getRebuildType() == RebuildModes.WORDSLOT_REBUILD_FULL); // TODO: inefficient?
    }

    /**
     * @param required true if this WordSlot requires that a WordList be attached.
     */
    public void setWordListRequired(boolean required)
    {
        _wordListRequired = required;
    }

    /**
     * @return true if this WordSlot requires that a WordList be attached.
     */
    public boolean getWordListRequired()
    {
        return _wordListRequired;
    }

    /**
     * @return true if there is at least one WordList attached to this WordSlot.
     */
    public boolean hasWordList() {
        return !_wordLists.isEmpty();
    }

    /**
     * Attach a word list, which will serve as a source of content for this word
     * slot.
     * TODO: is it ok that this info is not included in the JSON representation?
     * @param wordList
     */
    public void attachWordList(WordList wordList)
    {
        if (!_wordLists.contains(wordList))
            _wordLists.add(wordList);
    }

    /**
     * Add a word (or phrase) to the word slot.
     *
     * @param word The word to add.
     * <p>
     * If the word has a spoken form, it will only
     * be used for Japanese--the surface form will be the Kanji spelling, and
     * the spoken form will be the Kata-Kana spelling.
     */
    public void addWord(Word word)
    {
        Checker.checkArgForNull("word", word);
        _words.add(word);
    }

    /**
     * Add words to the word slot.
     *
     * @param words The words to add
     */
    public void addWords(Iterable<Word> words)
    {
        Checker.checkArgForNull("words", words);
        for (Word w : words)
            _words.add(w);
    }

    /**
     * Remove words from the word slot. This will only remove words that exactly
     * match both the surface and spoken form.
     *
     * @param words The words to remove.
     */
    public void removeWords(Iterable<Word> words)
    {
        Checker.checkArgForNull("words", words);

        for (Word w : words)
            _words.remove(w);
    }

    /**
     * Clear all words from the slot.
     */
    public void clear()
    {
        _words.clear();
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();

        o.tryPut(JSON_KEY_ID, _id);
        o.tryPut(JSON_KEY_SLOT_TYPE, _type);
        o.tryPut(JSON_KEY_REBUILD_MODE, _rebuildMode);
        o.tryPut(JSON_KEY_REQUIRES_WORDLIST, _wordListRequired);
        if (!this.isGeneric()) {
            /* Note: We do not need to store the surface forms or spoken forms
             *       for generic words.  The word type itself provides enough
             *       information to rebuild these classes correctly.
             */
            JSONArray words = new JSONArray();
            for (Word w : _words)
            {
                words.put(w.toJSON());
            }
            o.tryPut(JSON_KEY_WORDS, words);
        }
        return o;
    }

    public static WordSlot createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String id = json.getString(JSON_KEY_ID);
        int type = json.getInt(JSON_KEY_SLOT_TYPE);
        int rebuildMode = json.getInt(JSON_KEY_REBUILD_MODE);
        WordSlot ret = new WordSlot(id, type, rebuildMode);
        boolean wordListRequired = json.getBoolean(JSON_KEY_REQUIRES_WORDLIST);
        ret.setWordListRequired(wordListRequired);

        if (!ret.isGeneric()) {
            /* Note: We do not need to store the surface forms or spoken forms
             *       for generic words.  The word type itself provides enough
             *       information to rebuild these classes correctly.
             */
            org.json.JSONArray words = json.getJSONArray(JSON_KEY_WORDS);

            int size = words.length();
            for (int i = 0; i < size; i++)
            {
                ret.addWord(Word.createFromJSON(words.getJSONObject(i)));
            }
        }
        return ret;
    }
}
