package com.nuance.dragon.toolkit.grammar;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.json.JSONException;

import com.nuance.dragon.toolkit.Build;
import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.TransactionError;
import com.nuance.dragon.toolkit.cloudservices.TransactionResult;
import com.nuance.dragon.toolkit.cloudservices.datauploader.NcsCloudDataUploader;
import com.nuance.dragon.toolkit.cloudservices.datauploader.NcsCloudDataUploaderConfig;
import com.nuance.dragon.toolkit.core.WordList;
import com.nuance.dragon.toolkit.grammar.content.ContentManager;
import com.nuance.dragon.toolkit.language.Language;
import com.nuance.dragon.toolkit.language.LanguageEvent;
import com.nuance.dragon.toolkit.language.Languages;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONArray;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * A grammar depot contains all of the scenarios and grammars necessary for
 * Elvis, cloud, or hybrid recognition.
 */
public class NcsGrammarDepot extends LanguageEvent implements JSONCompliant
{
    /**
     * Defines a grammar for a specific language.
     */
    public static class GrammarEntry
    {
        private final Language _language;
        private final Map<String, NcsScenario> _scenarios;
        private final com.nuance.dragon.toolkit.elvis.Grammar _elvisGrammar;
        private final com.nuance.dragon.toolkit.vocon.Grammar _voconGrammar;

        public GrammarEntry(Language language, List<NcsScenario> scenarios,
                            com.nuance.dragon.toolkit.elvis.Grammar elvisGrammar,
                            com.nuance.dragon.toolkit.vocon.Grammar voconGrammar)
        {
            _language = language;
            _scenarios = new HashMap<String, NcsScenario>();
            _elvisGrammar = elvisGrammar;
            _voconGrammar = voconGrammar;

            for (NcsScenario s : scenarios)
            {
                _scenarios.put(s.getName(), s);
            }
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result
                    + ((_elvisGrammar == null) ? 0 : _elvisGrammar.hashCode());
            result = prime * result
                    + ((_language == null) ? 0 : _language.hashCode());
            result = prime * result
                    + ((_scenarios == null) ? 0 : _scenarios.hashCode());
            result = prime * result
                    + ((_voconGrammar == null) ? 0 : _voconGrammar.hashCode());
            return result;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            GrammarEntry other = (GrammarEntry) obj;
            if (_elvisGrammar == null) {
                if (other._elvisGrammar != null)
                    return false;
            } else if (!_elvisGrammar.equals(other._elvisGrammar))
                return false;
            if (_language == null) {
                if (other._language != null)
                    return false;
            } else if (!_language.equals(other._language))
                return false;
            if (_scenarios == null) {
                if (other._scenarios != null)
                    return false;
            } else if (!_scenarios.equals(other._scenarios))
                return false;
            if (_voconGrammar == null) {
                if (other._voconGrammar != null)
                    return false;
            } else if (!_voconGrammar.equals(other._voconGrammar))
                return false;
            return true;
        }
    }

    /*
     * This class associates a name with a WordList.  This object can be passed to the
     * GrammarDepot to identify the contact list for this grammar.
     */
    public static class ContentManagerEntry {
        private String _name;
        private ContentManager _contentManager;

        /*
         * Create a ContentManagerEntry.
         * @param name Must not be null or empty.
         * @param contentManager must not be null.
         */
        public ContentManagerEntry(String name, ContentManager contentManager) {
            Checker.checkStringArgForNullOrEmpty("name", name);
            Checker.checkArgForNull("contentManager", contentManager);
            _name = name;
            _contentManager = contentManager;
        }
    }

    private static class UploadWordListRequest
    {
        private String _wordListId;
        private CloudServices _cloudServices;
        private boolean _clearLists;
        private GrammarUploadListener _grammarUploadListener;

        public UploadWordListRequest(String wordListId, CloudServices cloudServices,
                                     boolean clearLists, GrammarUploadListener grammarUploadListener)
        {
            _wordListId = wordListId;
            _cloudServices = cloudServices;
            _clearLists = clearLists;
            _grammarUploadListener = grammarUploadListener;
        }
    }

    /** Constants to define the grammar upload status. */
    public static final class GrammarUploadStatus
    {
        /** Upload (adding/deleting content) has succeeded. */
        public static final int SUCCESS = 0;

        /** There is no content to upload. */
        public static final int NOTHING_TO_UPLOAD = 1;

        /** Invalid word list ID. */
        public static final int INVALID_WORD_LIST_ID = 2;

        /** Unable to add or delete content on the server. */
        public static final int UPLOAD_ERROR = 3;

        /** Upload is in progress. */
        public static final int UPLOAD_IN_PROGRESS = 4;

        /** Unknown error. */
        public static final int UNKNOWN_ERROR = 5;
    }

    /**
     * Listens for the completion of {@link NcsGrammarDepot#uploadServerWordLists}.
     */
    public interface GrammarUploadListener
    {
        /**
         * The upload completed.
         * @param grammarId The ID of the uploaded grammar if it's known, "unknown" otherwise.
         * @param status The status of the uploaded grammar (see #{@link GrammarUploadStatus}).
         * @param retryTimeMs When to retry to upload when the server is busy. 0 in all other cases.
         */
        void onComplete(String grammarId, int status, long retryTimeMs);
    }

    private static final String VERSION = "NMT version: " + Build.VERSION + " " + Build.CODENAME;
    private static final String UNKNOWN = "unknown";

    private static final String KEY_SCENARIOS = "scenarios";
    private static final String KEY_ELVIS_GRAMMARS = "elvis_grammar";
    private static final String KEY_VOCON_GRAMMARS = "vocon_grammar";
    private static final String KEY_LANGUAGE = "language";
    private static final String KEY_GRAMMARS = "grammars";

    private final Map<String, GrammarEntry> _grammars;
    private final List<GrammarEntry> _grammarsInOrder;  // Used to preserve order for JSON
    private GrammarEntry _currentGrammar;
    private final Map<String, ContentManager> _contentManagers;
    private final FileManager _fileManager;
    private final HashMap<String, WordList> _loadedElvisWordLists;
    private final HashMap<String, WordList> _loadedVoconWordLists;
    private final HashMap<String, WordList> _loadedServerWordLists;
    private HashMap<String, NcsCloudDataUploader> _uploads;
    private final ArrayList<UploadWordListRequest> _uploadWordListRequests;
    private int _contentManagerInitCount;
    private boolean _canUploadWordLists;
    private NcsCloudDataUploaderConfig _uploadConfig;

    /**
     * @param grammars The grammar entries, each specifying a language, Elvis
     * grammar, VoCon grammar, and scenario list. The first grammar entry will
     * be considered the 'default' language. Only use this constructor for Grammars
     * that do no contain any references to contact lists or custom word lists.
     */
    public NcsGrammarDepot(List<GrammarEntry> grammars)
    {
        Checker.checkArgForNull("grammars", grammars);
        Checker.checkArgForCondition("grammars", "non-empty", !grammars.isEmpty());
        _grammarsInOrder = new ArrayList<GrammarEntry>(grammars.size());
        _grammarsInOrder.addAll(grammars);
        _grammars = new HashMap<String, GrammarEntry>();
        for (GrammarEntry e : grammars)
        {
            _grammars.put(e._language.getGrammarLanguage(), e);
        }
        _currentGrammar = grammars.get(0);
        _contentManagers = null;
        _fileManager = null;
        _loadedElvisWordLists = null;
        _loadedVoconWordLists = null;
        _loadedServerWordLists = null;
        _uploadWordListRequests = new ArrayList<UploadWordListRequest>();
        _contentManagerInitCount = 0;
        _canUploadWordLists = false;
        _uploadConfig = new NcsCloudDataUploaderConfig();
    }

    /**
     * @param grammars The grammar entries, each specifying a language, Elvis
     * grammar, and scenario list. The first grammar entry will be considered
     * the 'default' language.
     * @param contentManagers A list of ContentManagers that can be used to populate contact lists
     * and custom word lists.
     * @param fileManager A file manager that can be used to load or save the state of the ContentManagers associated
     * with this grammar.
     */
    public NcsGrammarDepot(List<GrammarEntry> grammars, List<ContentManagerEntry> contentManagers, FileManager fileManager)
    {
        Checker.checkArgForNull("grammars", grammars);
        Checker.checkArgForCondition("grammars", "non-empty", !grammars.isEmpty());
        Checker.checkArgForNull("contentManagers", contentManagers);
        Checker.checkArgForCondition("contentManagers", "non-empty", !contentManagers.isEmpty());
        Checker.checkArgForNull("fileManager", fileManager);
        Logger.info(this, VERSION);
        _grammarsInOrder = new ArrayList<GrammarEntry>(grammars.size());
        _grammarsInOrder.addAll(grammars);
        _grammars = new HashMap<String, GrammarEntry>();
        for (GrammarEntry e : grammars)
        {
            _grammars.put(e._language.getGrammarLanguage(), e);
        }
        _currentGrammar = grammars.get(0);
        _contentManagers = new HashMap<String, ContentManager>();
        for (ContentManagerEntry cme : contentManagers) {
            if (cme._name == null || (cme._name.length() == 0) || cme._contentManager == null) {
                throw new IllegalArgumentException("ContentManagerEntry cannot have an empty or null name or ContentManager.");
            }
            if (_contentManagers.containsKey(cme._name)) {
                throw new IllegalArgumentException("contentManagers must not contain 2 ContentManagerEntry with the same name.");
            }
            final int contentManagerInitTotal = contentManagers.size();
            final ContentManager.SyncListener initListener = new ContentManager.SyncListener()
            {
                @Override
                public void onDone(ContentManager contentManager)
                {
                    _contentManagerInitCount++;
                    if(_contentManagerInitCount >= contentManagerInitTotal)
                    {
                        // All content managers are now ready to be used,
                        // we can upload lists.
                        _canUploadWordLists = true;

                        // Dequeue pending requests
                        for (UploadWordListRequest request : _uploadWordListRequests)
                        {
                            if (request._wordListId != null)
                            {
                                uploadServerWordList(request._wordListId,
                                                     request._cloudServices,
                                                     request._clearLists,
                                                     request._grammarUploadListener);
                            }
                            else
                            {
                                uploadServerWordLists(request._cloudServices,
                                                      request._clearLists,
                                                      request._grammarUploadListener);
                            }
                        }
                        _uploadWordListRequests.clear();
                    }
                }
            };
            cme._contentManager.setInitListener("GrammarDepot", initListener);
            _contentManagers.put(cme._name, cme._contentManager);
        }
        _fileManager = fileManager;
        _loadedElvisWordLists = new HashMap<String, WordList>();
        _loadedVoconWordLists = new HashMap<String, WordList>();
        _loadedServerWordLists = new HashMap<String, WordList>();
        _uploadWordListRequests = new ArrayList<UploadWordListRequest>();
        _contentManagerInitCount = 0;
        _canUploadWordLists = false;
        _uploadConfig = new NcsCloudDataUploaderConfig();
    }

    /**
     * Get the list of languages that are supported by this GrammarDepot.
     *
     * @return A list of available languages.
     */
    public List<Language> getAvailableLanguages() {
        ArrayList<Language> languages = new ArrayList<Language>(_grammarsInOrder.size());
        for (GrammarEntry ge : _grammarsInOrder) {
            languages.add(ge._language);
        }
        return languages;
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((_currentGrammar == null) ? 0 : _currentGrammar.hashCode());
        result = prime * result + ((_fileManager == null) ? 0 : _fileManager.hashCode());
        result = prime * result + ((_grammars == null) ? 0 : _grammars.hashCode());
        result = prime * result + ((_grammarsInOrder == null) ? 0 : _grammarsInOrder.hashCode());
        result = prime * result + ((_loadedElvisWordLists == null) ? 0 : _loadedElvisWordLists.hashCode());
        result = prime * result + ((_loadedVoconWordLists == null) ? 0 : _loadedVoconWordLists.hashCode());
        result = prime * result + ((_loadedServerWordLists == null) ? 0 : _loadedServerWordLists.hashCode());
        result = prime * result + ((_uploads == null) ? 0 : _uploads.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (!(obj instanceof NcsGrammarDepot))
            return false;
        NcsGrammarDepot other = (NcsGrammarDepot) obj;
        if (_currentGrammar == null)
        {
            if (other._currentGrammar != null)
                return false;
        } else if (!_currentGrammar.equals(other._currentGrammar))
            return false;
        if (_fileManager == null)
        {
            if (other._fileManager != null)
                return false;
        } else if (!_fileManager.equals(other._fileManager))
            return false;
        if (_grammars == null)
        {
            if (other._grammars != null)
                return false;
        } else if (!_grammars.equals(other._grammars))
            return false;
        if (_grammarsInOrder == null)
        {
            if (other._grammarsInOrder != null)
                return false;
        } else if (!_grammarsInOrder.equals(other._grammarsInOrder))
            return false;
        if (_loadedElvisWordLists == null)
        {
            if (other._loadedElvisWordLists != null)
                return false;
        } else if (!_loadedElvisWordLists.equals(other._loadedElvisWordLists))
            return false;
        if (_loadedVoconWordLists == null)
        {
            if (other._loadedVoconWordLists != null)
                return false;
        } else if (!_loadedVoconWordLists.equals(other._loadedVoconWordLists))
            return false;
        if (_loadedServerWordLists == null)
        {
            if (other._loadedServerWordLists != null)
                return false;
        } else if (!_loadedServerWordLists.equals(other._loadedServerWordLists))
            return false;
        if (_uploads == null)
        {
            if (other._uploads != null)
                return false;
        } else if (!_uploads.equals(other._uploads))
            return false;
        return true;
    }

    /**
     * Set the custom cloud data uploader configuration.
     * @param config The custom configuration, see {@link NcsCloudDataUploaderConfig}.
     */
    public void setCloudDataUploaderConfig(NcsCloudDataUploaderConfig config)
    {
        Checker.checkArgForNull("config", config);
        _uploadConfig = config;
    }

    /**
     * Set the current language. Subsequent calls to {@link #getElvisGrammar()}
     * and {@link #getScenario(String)} will use this language.
     * @param language A language from {@link com.nuance.dragon.toolkit.language.Languages}.
     */
    public void setLanguage(Language language)
    {
        GrammarEntry grammar = _grammars.get(language.getGrammarLanguage());
        Checker.checkArgForCondition("language", "a configured language", grammar != null);
        _currentGrammar = grammar;
    }

    /**
     * Get the current language.
     */
    public Language getLanguage()
    {
        return _currentGrammar._language;
    }

    /**
     * Mark the specified word lists (contact lists or custom word lists) as needing to be resent to
     * the server.  The next time the {@link #uploadServerWordLists} routine is run, the entire contents
     * of the flagged word lists will be resent to the server.
     *
     * Use this method to reset any Word Lists that are reported as being out of sync by either:
     * {@link com.nuance.dragon.toolkit.recognition.RecognitionInterpreter#getUpdateRequiredList} or
     * {@link com.nuance.dragon.toolkit.hybrid.HybridRecognitionResult#getSlotsToUpdate()}.
     * It is recommended to call {@link #uploadServerWordLists} after this method is called so that the
     * server will be updated.
     * @param wordLists The list of Word Lists to be updated.  May be null or empty if nothing needs to be updated.
     * @return A list of wordLists that could not be reset.  If all lists are successful, an empty list will be returned.
     * Word Lists will fail to be reset if the name does not exist in the GrammarDepot for the current language.
     * Failures can also occur if there are errors reading or writing the underlying files that represent the Word Lists.
     */
    public List<String> resetServerWordLists(List<String> wordLists)
    {
        ArrayList<String> failedWordListIds = new ArrayList<String>();
        if (wordLists == null || wordLists.isEmpty()) {
            return failedWordListIds;
        }
        HashSet<String> filteredWordList = new HashSet<String>(wordLists.size());
        filteredWordList.addAll(wordLists);

        HashSet<String> allWordListIds = new HashSet<String>();
        String contactListId = null;
        for (Map.Entry<String, NcsScenario> entry : _currentGrammar._scenarios.entrySet()) {
            NcsScenario scenario = entry.getValue();
            if (contactListId == null) {
                contactListId = scenario.getRequiredContactListId();
            }
            allWordListIds.addAll(scenario.getRequiredWordListIds());
        }

        for (String wordListId : filteredWordList)
        {
            if (allWordListIds.contains(wordListId)) {
                WordList wl = createOrGetServerWordList(wordListId, true);
                if (wl == null)
                {
                    Logger.error(this, "Unable to reset custom word list: " + wordListId);
                    failedWordListIds.add(wordListId);
                }
            } else {
                if (wordListId.equals("contacts") && (contactListId != null)) {
                    WordList wl = createOrGetServerWordList(contactListId, true);
                    if (wl == null)
                    {
                        Logger.error(this, "Unable to reset contact list: " + contactListId);
                        failedWordListIds.add(wordListId);
                    }
                } else {
                    Logger.error(this, "Word list not found in grammar depot: " + wordListId);
                    failedWordListIds.add(wordListId);
                }
            }
        }
        return failedWordListIds;
    }

    /**
     * Upload contact list or a custom Word List required by the server grammar for
     * the current language. Only uploads the changes since the last upload.
     * Large uploads will be split into multiple smaller transactions.
     * @param wordListId The word list ID to upload.
     * @param cloudServices A cloudServices object that can be used to send the data to the server.
     * @param clearList True to do a full reset and upload all content to the server.
     * False to only upload changes since the last upload.
     * will make the entire upload take longer, but will allow other
     * transactions (e.g. recognitions) to execute in the middle.
     * @param grammarUploadListener The onComplete method will be called to indicate success or failure.
     */
    public void uploadServerWordList(String wordListId, CloudServices cloudServices,
                                     boolean clearList,
                                     final GrammarUploadListener grammarUploadListener)
    {
        if (_uploads != null && _uploads.containsKey(wordListId)) {
            Logger.warn(this,"Upload (\"" + wordListId + "\") is already in progress. Not starting new upload.");

            if (grammarUploadListener != null)
                grammarUploadListener.onComplete(wordListId, GrammarUploadStatus.UPLOAD_IN_PROGRESS, 0);
            return;
        }

        if(_canUploadWordLists)
        {
            NcsCloudDataUploader.UploadListener uploadListener = new NcsCloudDataUploader.UploadListener()
            {
                @Override
                public void onDone(NcsCloudDataUploader uploader, NcsCloudDataUploader.Status status, String grammarId, long retryTimeMs)
                {
                    Logger.debug(this, "Upload done for " + grammarId + ": " + status.name());

                    if (_uploads == null)
                        return;

                    _uploads.remove(grammarId);
                    if (_uploads.isEmpty())
                        _uploads = null;

                    if (status == NcsCloudDataUploader.Status.SUCCESS_NOTHING_TO_UPLOAD)
                    {
                        if (grammarUploadListener != null)
                            grammarUploadListener.onComplete(grammarId, GrammarUploadStatus.NOTHING_TO_UPLOAD, retryTimeMs);
                    }
                    else if (status == NcsCloudDataUploader.Status.SUCCESS)
                    {
                        if (grammarUploadListener != null)
                            grammarUploadListener.onComplete(grammarId, GrammarUploadStatus.SUCCESS, retryTimeMs);
                    }
                    else
                    {
                        // TODO: Add more error granularity
                        if (grammarUploadListener != null)
                            grammarUploadListener.onComplete(grammarId, GrammarUploadStatus.UPLOAD_ERROR, retryTimeMs);
                    }
                }

                @Override
                public void onServerResult(NcsCloudDataUploader uploader, String grammarId, TransactionResult result)
                {
                    Logger.debug(this, "Server result for " + grammarId + ": " + result.getContents());
                }

                @Override
                public void onServerError(NcsCloudDataUploader uploader, String grammarId, TransactionError error)
                {
                    Logger.error(this, "Server error for " + grammarId + ": " + error.getErrorCode());
                }
            };

            if (_uploads == null) {
                _uploads =  new HashMap<String, NcsCloudDataUploader>();
            }

            // Figure out the list type
            String type = null;
            for (Map.Entry<String, NcsScenario> entry : _currentGrammar._scenarios.entrySet()) {
                NcsScenario scenario = entry.getValue();
                String contactListId = scenario.getRequiredContactListId();
                if (contactListId != null && contactListId.equals(wordListId)) {
                    type = "contacts";
                    break;
                }

                Set<String> ids = scenario.getRequiredWordListIds();
                for (String id : ids) {
                    if (id.equals(wordListId)) {
                        type = scenario.getWordListType(id);
                        break;
                    }
                }
            }

            // Upload right away. CloudServices will take care of
            // queuing multiple transactions.
            boolean uploaded = false;
            if (wordListId != null) {
                WordList wordList = createOrGetServerWordList(wordListId, false);
                if (wordList == null) {
                    Logger.error(this, "Upload failed. Cannot load word list");
                    cancelServerWordListsUpload();
                    if (grammarUploadListener != null)
                        grammarUploadListener.onComplete(wordListId, GrammarUploadStatus.INVALID_WORD_LIST_ID, 0);
                    return;
                }
                final NcsCloudDataUploader currentUploader = new NcsCloudDataUploader(cloudServices, _uploadConfig, _fileManager);
                _uploads.put(wordListId, currentUploader);
                currentUploader.uploadData(wordListId, type, _currentGrammar._language.getCloudLanguage(), wordList, clearList, uploadListener);
                uploaded = true;
            }

            // There was nothing to upload.  Call the onComplete callback and exit.
            if (!uploaded && grammarUploadListener != null) {
                grammarUploadListener.onComplete(UNKNOWN, GrammarUploadStatus.NOTHING_TO_UPLOAD, 0);
            }
        }
        else
        {
            // Upload the word list later
            _uploadWordListRequests.add(new UploadWordListRequest(wordListId, cloudServices,
                                                                  clearList, grammarUploadListener));
        }
    }

    /**
     * Upload all contact lists and custom Word Lists required by the server grammar for
     * the current language. Only uploads the changes since the last upload.
     * Large uploads will be split into multiple smaller transactions.
     * @param cloudServices A cloudServices object that can be used to send the data to the server.
     * @param grammarUploadListener The onComplete method will be called to indicate success or failure.
     */
    public void uploadServerWordLists(CloudServices cloudServices,
                                      final GrammarUploadListener grammarUploadListener)
    {
        uploadServerWordLists(cloudServices, false, grammarUploadListener);
    }

    /**
     * Upload all contact lists and custom Word Lists required by the server grammar for
     * the current language.
     * @param cloudServices A cloudServices object that can be used to send the data to the server.
     * @param clearLists True to do a full reset and upload all content to the server.
     * @param grammarUploadListener The onComplete method will be called to indicate success or failure.
     */
    public void uploadServerWordLists(CloudServices cloudServices,
                                      boolean clearLists,
                                      final GrammarUploadListener grammarUploadListener)
    {
        if (_uploads != null) {
            Logger.warn(this,"Upload is already in progress. Not starting new upload.");

            if (grammarUploadListener != null)
                grammarUploadListener.onComplete(UNKNOWN, GrammarUploadStatus.UPLOAD_IN_PROGRESS, 0);
            return;
        }

        if(_canUploadWordLists)
        {
            final HashMap<String, NcsCloudDataUploader> uploadList = new  HashMap<String, NcsCloudDataUploader>();
            NcsCloudDataUploader.UploadListener uploadListener = new NcsCloudDataUploader.UploadListener()
            {
                @Override
                public void onDone(NcsCloudDataUploader uploader, NcsCloudDataUploader.Status status, String grammarId, long retryTimeMs)
                {
                    Logger.debug(this, "Upload done for " + grammarId + ": " + status.name());

                    if (_uploads != uploadList)
                        return;

                    if (status == NcsCloudDataUploader.Status.SUCCESS ||
                        status == NcsCloudDataUploader.Status.SUCCESS_NOTHING_TO_UPLOAD)
                    {
                        uploadList.remove(grammarId);

                        if (uploadList.isEmpty())
                        {
                            _uploads = null;

                            if (grammarUploadListener != null)
                                grammarUploadListener.onComplete(UNKNOWN, GrammarUploadStatus.SUCCESS, retryTimeMs);
                        }
                    }
                    else
                    {
                        _uploads = null;

                        uploadList.remove(grammarId);
                        for (Map.Entry<String, NcsCloudDataUploader> entry : uploadList.entrySet())
                            entry.getValue().cancel();

                        if (grammarUploadListener != null)
                            grammarUploadListener.onComplete(UNKNOWN, GrammarUploadStatus.UPLOAD_ERROR, retryTimeMs);
                    }
                }

                @Override
                public void onServerResult(NcsCloudDataUploader uploader, String grammarId, TransactionResult result)
                {
                    Logger.debug(this, "Server result for " + grammarId + ": " + result.getContents());
                }

                @Override
                public void onServerError(NcsCloudDataUploader uploader, String grammarId, TransactionError error)
                {
                    Logger.error(this, "Server error for " + grammarId + ": " + error.getErrorCode());
                }
            };

            String contactListId = null;
            ArrayList<String> requiredWordListIds = new ArrayList<String>();
            ArrayList<String> requiredwordListTypes = new ArrayList<String>();
            for (Map.Entry<String, NcsScenario> entry : _currentGrammar._scenarios.entrySet()) {
                NcsScenario scenario = entry.getValue();
                if (contactListId == null) {
                    contactListId = scenario.getRequiredContactListId();
                } else {
                    String newContactListId = scenario.getRequiredContactListId();
                    if (newContactListId != null && !newContactListId.equals(contactListId)) {
                        // TODO handle error: multiple contact list IDs.
                    }
                }
                Set<String> ids = scenario.getRequiredWordListIds();
                for (String id : ids) {
                    if (!requiredWordListIds.contains(id))
                    {
                        requiredWordListIds.add(id);
                        requiredwordListTypes.add(scenario.getWordListType(id));
                    }
                }
            }

            _uploads = uploadList;

            // Upload everything right away. CloudServices will take care of
            // queuing multiple transactions.
            boolean uploaded = false;
            if (contactListId != null) {
                // Try to upload the Contact List if there is one.
                WordList contactList = createOrGetServerWordList(contactListId, false);
                if (contactList == null) {
                    Logger.error(this, "Upload failed.  Cannot load contact list");
                    cancelServerWordListsUpload();
                    if (grammarUploadListener != null)
                        grammarUploadListener.onComplete(contactListId, GrammarUploadStatus.INVALID_WORD_LIST_ID, 0);
                    return;
                }
                NcsCloudDataUploader uploader = new NcsCloudDataUploader(cloudServices, _uploadConfig, _fileManager);
                uploadList.put(contactListId, uploader);
                uploader.uploadData(contactListId, "contacts", _currentGrammar._language.getCloudLanguage(), contactList, clearLists, uploadListener);
                uploaded = true;
            }

            for (int i = 0; i < requiredWordListIds.size(); ++ i)
            {
                String wordListId = requiredWordListIds.get(i);
                String type = requiredwordListTypes.get(i);
                String language = _currentGrammar._language.getCloudLanguage();

                WordList wl = createOrGetServerWordList(wordListId, false);
                if (wl == null)
                {
                    Logger.error(this, "Upload failed.  Cannot load word list:" + wordListId);
                    cancelServerWordListsUpload();
                    if (grammarUploadListener != null)
                        grammarUploadListener.onComplete(wordListId, GrammarUploadStatus.INVALID_WORD_LIST_ID, 0);
                    return;
                }

                NcsCloudDataUploader uploader = new NcsCloudDataUploader(cloudServices, _uploadConfig, _fileManager);
                uploadList.put(wordListId, uploader);
                uploader.uploadData(wordListId, type, language, wl, clearLists, uploadListener);
                uploaded = true;
            }

            // There was nothing to upload.  Call the onComplete callback and exit.
            if (!uploaded && grammarUploadListener != null) {
                _uploads = null;
                grammarUploadListener.onComplete(UNKNOWN, GrammarUploadStatus.NOTHING_TO_UPLOAD, 0);
            }
        }
        else
        {
            // Upload all the word lists later
            _uploadWordListRequests.add(new UploadWordListRequest(null, cloudServices,
                                                                  clearLists,
                                                                  grammarUploadListener));
        }
    }

    /**
     * Delete all contact lists and custom Word Lists from the server.
     * @param cloudServices A cloudServices object that can be used to send the data to the server.
     * @param grammarUploadListener The onComplete method will be called to indicate success or failure.
     */
    public void clearServerWordLists(CloudServices cloudServices, final GrammarUploadListener grammarUploadListener)
    {
        if (_uploads != null) {
            Logger.warn(this,"Upload is already in progress.  Cannot clear Word List.");
            return;
        }

        final HashMap<String, NcsCloudDataUploader> uploads = new HashMap<String, NcsCloudDataUploader>(1);
        _uploads = uploads;

        NcsCloudDataUploader uploader = new NcsCloudDataUploader(cloudServices, _uploadConfig, _fileManager);
        uploader.deleteAll( new NcsCloudDataUploader.DeleteListener()
        {
            @Override
            public void onDone(NcsCloudDataUploader uploader, NcsCloudDataUploader.Status status)
            {
                if (_uploads != uploads)
                    return;

                boolean success = false;
                if (status == NcsCloudDataUploader.Status.SUCCESS)
                {
                    String contactListId = null;
                    HashSet<String> requiredWordListIds = new HashSet<String>();
                    for (Map.Entry<String, NcsScenario> entry : _currentGrammar._scenarios.entrySet()) {
                        NcsScenario scenario = entry.getValue();
                        if (contactListId == null) {
                            contactListId = scenario.getRequiredContactListId();
                        } else {
                            String newContactListId = scenario.getRequiredContactListId();
                            if (newContactListId != null && !newContactListId.equals(contactListId)) {
                                // TODO handle error: multiple contact list IDs.
                            }
                        }
                        requiredWordListIds.addAll(scenario.getRequiredWordListIds());
                    }
                    if (contactListId != null) {
                        createOrGetServerWordList(contactListId, true);
                    }
                    for (String wordListId : requiredWordListIds) {
                        createOrGetServerWordList(wordListId, true);
                    }
                    success = true;
                }

                _uploads = null;
                if (grammarUploadListener != null)
                    grammarUploadListener.onComplete(UNKNOWN, success ? GrammarUploadStatus.SUCCESS : GrammarUploadStatus.UPLOAD_ERROR, 0);
            }

            @Override
            public void onServerResult(NcsCloudDataUploader uploader, TransactionResult result)
            {
                Logger.debug(this, "Server result: " + result.getContents());
            }

            @Override
            public void onServerError(NcsCloudDataUploader uploader, TransactionError error)
            {
                Logger.error(this, "Server error: " + error.getErrorCode());
            }
        });
    }

    /**
     * Cancel a ServerWordList upload that is in progress.
     */
    public void cancelServerWordListsUpload()
    {
        if (_uploads != null)
        {
            Map<String, NcsCloudDataUploader> uploads = _uploads;
            _uploads = null;

            for (Map.Entry<String, NcsCloudDataUploader> entry : uploads.entrySet())
            {
                entry.getValue().cancel();
            }
        }
    }

    /**
     * Get a scenario for the current language
     * @param scenarioName
     * @return The recognition scenario with the given name, or null if it
     * doesn't exist
     */
    public NcsScenario getScenario(String scenarioName)
    {
        NcsScenario scenario = _currentGrammar._scenarios.get(scenarioName);
        String contactListId = scenario.getRequiredContactListId();
        if (contactListId != null) {
            WordList wl = createOrGetServerWordList(contactListId, false);
            if (wl == null) {
                Logger.error(this, "Missing required Server contact list: " + contactListId);
                return null;
            }
            if (!scenario.attachWordList(wl, contactListId)) {
                Logger.error(this, "Failed to attach required Server contact list: " + contactListId);
                return null;
            }
        }
        for (String wordListId : scenario.getRequiredWordListIds()) {
            WordList wl = createOrGetServerWordList(wordListId, false);
            if (wl == null) {
                Logger.error(this, "Missing required Server wordlist: " + wordListId);
                return null;
            }
            if (!scenario.attachWordList(wl, wordListId)) {
                Logger.error(this, "Failed to attach required Server wordlist: " + wordListId);
                return null;
            }
        }
        return scenario;
    }

    /**
     * Get the Elvis grammar for the current language
     * @return The full Elvis grammar
     */
    public com.nuance.dragon.toolkit.elvis.Grammar getElvisGrammar()
    {
        com.nuance.dragon.toolkit.elvis.Grammar elvisGrammar = _currentGrammar._elvisGrammar;
        for (String wordListId : elvisGrammar.getRequiredWordListIds()) {
            WordList wl = createOrGetElvisWordList(wordListId, false);
            if (wl == null) {
                Logger.error(this, "Missing required Elvis wordlist: " + wordListId);
                return null;
            }
            if (!elvisGrammar.attachWordList(wl, wordListId))
            {
                Logger.error(this, "Failed to attach required Elvis wordlist: " + wordListId);
                return null;
            }
        }
        return elvisGrammar;
    }

    /**
     * Get the Vocon grammar for the current language
     * @return The full Vocon grammar
     */
    public com.nuance.dragon.toolkit.vocon.Grammar getVoconGrammar()
    {
        com.nuance.dragon.toolkit.vocon.Grammar voconGrammar = _currentGrammar._voconGrammar;
        for (String wordListId : voconGrammar.getRequiredWordListIds()) {
            WordList wl = createOrGetVoconWordList(wordListId, false);
            if (wl == null) {
                Logger.error(this, "Missing required Vocon wordlist: " + wordListId);
                return null;
            }
            if (!voconGrammar.attachWordList(wl, wordListId))
            {
                Logger.error(this, "Failed to attach required Vocon wordlist: " + wordListId);
                return null;
            }
        }
        return voconGrammar;
    }

    /**
     * Get the accepted word list checksum
     * @param wordListId The word list ID
     * @return The accepted checksum
     */
    public int getAcceptedWordListChecksum(String wordListId)
    {
        return createOrGetServerWordList(wordListId, false).getAcceptedChecksum();
    }

    @Override
    public void onApplicationLanguageChange(Object cloudServices, final Listener listener)
    {
        Checker.checkArgForNull("cloudServices", cloudServices);
        Checker.checkArgForCondition("cloudServices", "instance of CloudServices", cloudServices instanceof CloudServices);

        cancelServerWordListsUpload();
        clearServerWordLists((CloudServices)cloudServices, new GrammarUploadListener()
        {
            @Override
            public void onComplete(String grammarId, int status, long retryTimeMs)
            {
               if (listener != null)
                  listener.onComplete(status == GrammarUploadStatus.SUCCESS);
            }
        });
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();

        // Use an array to preserve the order, since the default has to stay first
        JSONArray grammars = new JSONArray();

        for (GrammarEntry grammarEntry : _grammarsInOrder)
        {
            JSONObject grammarObject = new JSONObject();

            JSONObject scenariosObject = new JSONObject();
            for (NcsScenario scenario : grammarEntry._scenarios.values())
            {
                scenariosObject.tryPut(scenario.getName(), scenario);
            }
            grammarObject.tryPut(KEY_LANGUAGE, grammarEntry._language.getGrammarLanguage());
            grammarObject.tryPut(KEY_SCENARIOS, scenariosObject);
            grammarObject.tryPut(KEY_ELVIS_GRAMMARS, grammarEntry._elvisGrammar);
            grammarObject.tryPut(KEY_VOCON_GRAMMARS, grammarEntry._voconGrammar);

            grammars.tryPut(grammarObject);
        }
        o.tryPut(KEY_GRAMMARS, grammars);
        return o;
    }

    /**
     * Create A GrammarDepot from JSON without file manager or content managers
     *
     * <p>
     * Note: Only use this API for Grammars that do no contain any references to contact
     * lists or custom word lists.
     * </p>
     *
     * @param json A JSON object that represents a GrammarDepot
     */
    public static NcsGrammarDepot createFromJSON(org.json.JSONObject json) throws JSONException
    {
        List<ContentManagerEntry> contentManagers = null;

        return createFromJSON(json, null, contentManagers);
    }

    /**
     * Create A GrammarDepot from JSON.
     *
     * @param json A JSON object that represents a GrammarDepot
     * @param fileManager A FileManager pointing to a folder that should be used to save WordLists data related to
     * this GrammarDepot.
     * @param contentManagers A list of {@link com.nuance.dragon.toolkit.grammar.content.ContentManager}s that can be used
     * to retrieve contact lists and custom word lists required by the associated server grammars.
     */
    public static NcsGrammarDepot createFromJSON(org.json.JSONObject json, FileManager fileManager, ContentManagerEntry ... contentManagers) throws JSONException
    {
        ArrayList<ContentManagerEntry> contentManagerList = new ArrayList<ContentManagerEntry>(contentManagers.length);
        for (ContentManagerEntry e : contentManagers)
            contentManagerList.add(e);

        return createFromJSON(json, fileManager, contentManagerList);
    }

    /**
     * Create A GrammarDepot from JSON.
     *
     * @param json A JSON object that represents a GrammarDepot
     * @param fileManager A FileManager pointing to a folder that should be used to save WordLists data related to
     * this GrammarDepot.
     * @param contentManagers A list of {@link com.nuance.dragon.toolkit.grammar.content.ContentManager}s that can be used
     * to retrieve contact lists and custom word lists required by the associated server grammars.
     */
    @SuppressWarnings("rawtypes")
    public static NcsGrammarDepot createFromJSON(org.json.JSONObject json, FileManager fileManager, List<ContentManagerEntry> contentManagers) throws JSONException
    {
        ArrayList<GrammarEntry> grammarEntries = new ArrayList<GrammarEntry>();
        HashMap<String, ContentManager> managers = null;
        if (contentManagers != null) {
            managers = new HashMap<String, ContentManager>(contentManagers.size());
            for (ContentManagerEntry cme : contentManagers) {
                if (cme._name == null || (cme._name.length() == 0) || cme._contentManager == null) {
                    throw new IllegalArgumentException("ContentManagerEntry cannot have an empty or null name or ContentManager.");
                }
                if (managers.containsKey(cme._name)) {
                    throw new IllegalArgumentException("contentManagers must not contain 2 ContentManagerEntry with the same name.");
                }
                managers.put(cme._name, cme._contentManager);
            }
        } else {
            managers = new HashMap<String, ContentManager>();
        }

        org.json.JSONArray grammars = json.getJSONArray(KEY_GRAMMARS);
        for (int i = 0; i < grammars.length(); i++)
        {
            org.json.JSONObject grammarObject = grammars.getJSONObject(i);
            String grammarLanguage = grammarObject.getString(KEY_LANGUAGE);
            Language language = Languages.lookupFromGrammarLanguage(grammarLanguage);
            if (language == null) {
                throw new JSONException("Undefined grammar language: " + grammarLanguage);
            }
            org.json.JSONObject scenariosObject = grammarObject.getJSONObject(KEY_SCENARIOS);
            com.nuance.dragon.toolkit.elvis.Grammar elvisGrammar = com.nuance.dragon.toolkit.elvis.Grammar.createFromJSON(grammarObject.getJSONObject(KEY_ELVIS_GRAMMARS));
            Set<String> requiredWordListIds = elvisGrammar.getRequiredWordListIds();
            HashSet<String> allRequiredWordListIds = new HashSet<String>();
            if (requiredWordListIds != null && !requiredWordListIds.isEmpty()) {
                for (String id : requiredWordListIds) {
                    if (!managers.containsKey(id)) {
                        throw new IllegalArgumentException("contentManagers must contain a ContentManagerEntry named: " + id);
                    }
                }
                allRequiredWordListIds.addAll(requiredWordListIds);
            }

            com.nuance.dragon.toolkit.vocon.Grammar voconGrammar = com.nuance.dragon.toolkit.vocon.Grammar.createFromJSON(grammarObject.getJSONObject(KEY_VOCON_GRAMMARS));
            Set<String> requiredDSWordListIds = voconGrammar.getRequiredWordListIds();
            if (requiredDSWordListIds != null && !requiredDSWordListIds.isEmpty()) {
                for (String id : requiredDSWordListIds) {
                    if (!managers.containsKey(id)) {
                        throw new IllegalArgumentException("contentManagers must contain a ContentManagerEntry named: " + id);
                    }
                }
                allRequiredWordListIds.addAll(requiredWordListIds);
            }

            ArrayList<NcsScenario> scenarios = new ArrayList<NcsScenario>();
            Iterator scenIt = scenariosObject.keys();
            while (scenIt.hasNext())
            {
                String scenName = scenIt.next().toString();
                NcsScenario scenario = NcsScenario.createFromJSON(scenariosObject.getJSONObject(scenName));
                requiredWordListIds = scenario.getRequiredWordListIds();
                if (requiredWordListIds != null && !requiredWordListIds.isEmpty()) {
                    for (String id : requiredWordListIds) {
                        if (!managers.containsKey(id)) {
                            throw new IllegalArgumentException("contentManagers must contain a ContentManagerEntry named: " + id);
                        }
                    }
                    allRequiredWordListIds.addAll(requiredWordListIds);
                }
                scenarios.add(scenario);
            }
            // Make sure that WordLists in contentManagers and allRequiredWordListIds match exactly.
            String notUsed = "";
            for (Map.Entry<String, ContentManager> entry : managers.entrySet()) {
                String id = entry.getKey();
                if (!allRequiredWordListIds.contains(id)) {
                    if (notUsed.length() > 0) {
                        notUsed += ", ";
                    }
                    notUsed += id;
                }
            }
            if (notUsed.length() > 0) {
                throw new IllegalArgumentException("The following entries in contentManagers are not referenced by this GrammarDepot: " + notUsed);
            }
            grammarEntries.add(new GrammarEntry(language, scenarios, elvisGrammar, voconGrammar));
        }

        if (fileManager == null && contentManagers == null)
        {
            return new NcsGrammarDepot(grammarEntries);
        }

        return new NcsGrammarDepot(grammarEntries, contentManagers, fileManager);
    }

    private WordList createOrGetServerWordList(String wordListId, boolean reset) {
        String key = "server" + _currentGrammar._language + wordListId;
        if (_loadedServerWordLists.containsKey(key) && !reset) {
            return _loadedServerWordLists.get(key);
        } else {
            if (_contentManagers == null || _fileManager == null) {
                Logger.error(this, "Missing content manager or file manager.");
                return null;
            }
            ContentManager cm = _contentManagers.get(wordListId);
            if (cm == null) {
                Logger.error(this, "Cannot find content manager for word list: " + wordListId);
                return null;
            }
            WordList wl = cm.createWordList(_fileManager, key + ".lst", reset);
            _loadedServerWordLists.put(key, wl);
            return wl;
        }
    }

    private WordList createOrGetElvisWordList(String wordListId, boolean reset) {
        String key = "elvis" + _currentGrammar._language + wordListId;
        if (_loadedElvisWordLists.containsKey(key) && !reset) {
            return _loadedElvisWordLists.get(key);
        } else {
            if (_contentManagers == null || _fileManager == null) {
                return null;
            }
            ContentManager cm = _contentManagers.get(wordListId);
            if (cm == null) {
                return null;
            }
            WordList wl = cm.createWordList(_fileManager, key + ".lst", reset);
            _loadedElvisWordLists.put(key, wl);
            return wl;
        }
    }

    private WordList createOrGetVoconWordList(String wordListId, boolean reset) {
        String key = "vocon" + _currentGrammar._language + wordListId;
        if (_loadedVoconWordLists.containsKey(key) && !reset) {
            return _loadedVoconWordLists.get(key);
        } else {
            if (_contentManagers == null || _fileManager == null) {
                return null;
            }
            ContentManager cm = _contentManagers.get(wordListId);
            if (cm == null) {
                return null;
            }
            WordList wl = cm.createWordList(_fileManager, key + ".lst", reset);
            _loadedVoconWordLists.put(key, wl);
            return wl;
        }
    }
}
