package com.nuance.dragon.toolkit.grammar;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.json.JSONException;

import com.nuance.dragon.toolkit.cloudservices.recognizer.AsrSpec;
import com.nuance.dragon.toolkit.cloudservices.recognizer.RecogSpec;
import com.nuance.dragon.toolkit.core.WordList;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.util.internal.ReadOnlyList;
import com.nuance.dragon.toolkit.vocon.VoconContext;

/**
 * Defines a recognition scenario that uses {@link com.nuance.dragon.toolkit.elvis.ElvisRecognizer Elvis}
 * and {@link com.nuance.dragon.toolkit.cloudservices.CloudServices CloudServices}.
 */
public class Scenario implements JSONCompliant
{
    private static final String KEY_CONSTRAINTS = "constraints";
    private static final String KEY_CONTEXTS = "contexts";
    private static final String KEY_COMMAND = "cmd";
    private static final String KEY_NAME = "name";

    protected final ReadOnlyList<String> _elvisConstraints;
    protected final ReadOnlyList<VoconContext> _voconContexts;
    protected final AsrSpec _asrSpec;
    protected final String _name;

    /**
     * Each Scenario represents a collection of constraints (or commands) that can be recognized
     * at a particular time.  The scenario encapsulates all the information needed to start
     * either an Elvis recognition or a server based recognition (or both) with a particular
     * list of commands active.
     * Note that a Scenario corresponds to a grammar definition file that exists on a server
     * so they generally must be loaded from the GrammarDepot, and cannot reliably be constructed
     * at run-time without specific knowledge of what files and settings exist on the server.
     *
     * @param name The name of the scenario.
     * @param elvisConstraints The Elvis constraints that will be active
     * for this scenario. May be null to create a server-only recognition
     * scenario.
     * @param voconContexts The VoCon contexts that will be active
     * for this scenario. May be null to create a server-only recognition
     * scenario.
     * @param asrSpec The server recognition command for this scenario. May
     * be null to create an local-only recognition scenario.
     */
    public Scenario(String name,
                    List<String> elvisConstraints,
                    List<VoconContext> voconContexts,
                    AsrSpec asrSpec)
    {
        Checker.checkArgForNull("name", name);
        Checker.checkArgsForNull("elvisConstraints, voconContext and asrSpec", elvisConstraints, voconContexts, asrSpec);

        _name = name;

        if (elvisConstraints == null)
        {
            _elvisConstraints = new ReadOnlyList<String>();
        } else
        {
            _elvisConstraints = new ReadOnlyList<String>(elvisConstraints);
        }

        if (voconContexts == null)
        {
            _voconContexts = new ReadOnlyList<VoconContext>();
        } else
        {
            _voconContexts = new ReadOnlyList<VoconContext>(voconContexts);
        }

        _asrSpec = asrSpec;
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime
                * result
                + ((_elvisConstraints == null) ? 0 : _elvisConstraints
                        .hashCode());
        result = prime * result + ((_name == null) ? 0 : _name.hashCode());
        result = prime * result
                + ((_asrSpec == null) ? 0 : _asrSpec.hashCode());
        result = prime * result
                + ((_voconContexts == null) ? 0 : _voconContexts.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        Scenario other = (Scenario) obj;
        if (_elvisConstraints == null) {
            if (other._elvisConstraints != null)
                return false;
        } else if (!_elvisConstraints.equals(other._elvisConstraints))
            return false;
        if (_name == null) {
            if (other._name != null)
                return false;
        } else if (!_name.equals(other._name))
            return false;
        if (_asrSpec == null) {
            if (other._asrSpec != null)
                return false;
        } else if (!_asrSpec.equals(other._asrSpec))
            return false;
        if (_voconContexts == null) {
            if (other._voconContexts != null)
                return false;
        } else if (!_voconContexts.equals(other._voconContexts))
            return false;
        return true;
    }

    /**
     * The name of this scenario.
     * @return The name of the scenario
     */
    public String getName()
    {
        return _name;
    }

    /**
     * Get the Elvis constraints specified by this scenario.
     * @return The Elvis constraints. If empty, this scenario does not support
     * Elvis recognition.
     */
    public List<String> getElvisConstraints()
    {
        return _elvisConstraints;
    }

    /**
     * Get the VoCon context specified by this scenario. This is equivalent
     * to get the 1st entry returned by {@link #getVoconContexts}.
     * @return The VoCon context. May be null if this scenario does not
     * support server recognition.
     */
    public VoconContext getVoconContext()
    {
        VoconContext c = null;
        if (_voconContexts != null && !_voconContexts.isEmpty())
        {
            c = _voconContexts.get(0);
        }
        return c;
    }

    /**
     * Get the VoCon contexts specified by this scenario.
     * @return The VoCon contexts. If empty, this scenario does not support
     * VoCon recognition.
     */
    public List<VoconContext> getVoconContexts()
    {
        return _voconContexts;
    }

    /**
     * Get a cloud recog spec for this this scenario.
     * @return The cloud recog spec. May be null if this scenario does not
     * support server recognition.  Will also be null if this scenario requires
     * word lists or contact lists, and those lists have not been attached.
     */
    public RecogSpec getCloudRecogSpec()
    {
        if (_asrSpec != null) {
            return _asrSpec.createRecogSpec();
        }
        return null;
    }

    /**
     * Attach a word list to this scenario.  Note that the name of the word list must
     * exactly match the name of a word list defined in the associated server grammar.
     *
     * @param wordList The word list to attach.  Must not be null.
     * @param id The name of the word list to attach.  Must not be null or empty.
     *
     * @return true if the list was attached, false if the list was not attached.
     */
    public boolean attachWordList(WordList wordList, String id)
    {
        Checker.checkStringArgForEmpty("id", id);
        Checker.checkArgForNull("wordList", wordList);
        if (_asrSpec != null) {
            return _asrSpec.attachWordList(wordList, id);
        }
        return false;
    }

    /**
     * Use this function to determine the ID of the contact lists required by this object, if any.
     *
     * @return The contact list ID if required, or null if not required.
     */
    public String getRequiredContactListId() {
        if (_asrSpec != null) {
            return _asrSpec.getContactListId();
        }
        return null;
    }

    /**
     * Use this function to determine the IDs of the word lists required by this object, if any.
     *
     * @return A list of word list IDs that are required by this object.
     * Will be an empty list (never null) if no word lists are required.
     */
    public Set<String> getRequiredWordListIds() {
        if (_asrSpec != null) {
            Set<String> requiredWordListIds = _asrSpec.getRequiredCustomWordListIds();
            requiredWordListIds.addAll(_asrSpec.getStructuredContentIds());
            return requiredWordListIds;
        }
        return new HashSet<String>();
    }

    /**
     * Use this function to determine the type of a word list.
     *
     * @param id The word list ID.
     * @return The type of the word list.
     */
    public String getWordListType(String id) {
        if (_asrSpec != null)
            return _asrSpec.getGrammarType(id);
        return "Invalid NVC ASR Spec";
    }

    /**
     * Use this function to determine the category of a word list.
     *
     * @param id The word list ID.
     * @return The category of the word list.
     */
    public String getWordListCategory(String id) {
        if (_asrSpec != null)
            return _asrSpec.getGrammarCategory(id);
        return "Invalid NVC ASR Spec";
    }

    /**
     * Attach additional settings to this scenario.  Settings are optional.
     *
     * @param settings the settings to attach.  May be null to remove previously attached settings.
     */
    public void attachSettings(Data.Dictionary settings)
    {
        if (_asrSpec != null) {
            _asrSpec.attachSettings(settings);
        }
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();

        o.tryPut(KEY_NAME, _name);
        o.tryPut(KEY_COMMAND, AsrSpec.toJSON(_asrSpec));
        o.tryPutList(KEY_CONTEXTS, _voconContexts);
        o.tryPutStringList(KEY_CONSTRAINTS, _elvisConstraints);
        return o;
    }

    public static Scenario createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String name = json.getString(KEY_NAME);

        org.json.JSONObject commandObject = json.optJSONObject(KEY_COMMAND);
        AsrSpec as = commandObject == null ? null : AsrSpec.createFromJSON(commandObject);

        org.json.JSONArray contextArray = json.optJSONArray(KEY_CONTEXTS);
        ArrayList<VoconContext> contexts = null;
        if (contextArray != null)
        {
            contexts = new ArrayList<VoconContext>(contextArray.length());
            for (int i = 0; i < contextArray.length(); i++)
            {
                contexts.add(VoconContext.createFromJSON(contextArray.getJSONObject(i)));
            }
        }

        org.json.JSONArray constraintArray = json.optJSONArray(KEY_CONSTRAINTS);
        ArrayList<String> constraints = null;
        if (constraintArray != null)
        {
            constraints = new ArrayList<String>(constraintArray.length());
            for (int i = 0; i < constraintArray.length(); i++)
            {
                constraints.add(constraintArray.getString(i));
            }
        }

        if (as == null && contexts == null && constraints == null)
            throw new JSONException("Must have a " + KEY_COMMAND + " or " + KEY_CONTEXTS + " or " + KEY_CONSTRAINTS + " key");

        return new Scenario(name, constraints, contexts, as);
    }
}
