package com.nuance.dragon.toolkit.grammar.content;

import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.ApplicationInfo;
import android.content.pm.PackageManager;
import android.content.pm.ResolveInfo;

import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;

/**
 * A {@link ContentManager} that provides the full list of launchable
 * applications on the device.
 */
public class AppsManager extends ContentManager
{
    private final BroadcastReceiver _receiver;

    /**
     * Like {@link #AppsManager(String, FileManager, boolean, Context)} using auto
     * content sync-up without resetting the cache file.
     */
    public AppsManager(String fileName, FileManager fileManager, Context context)
    {
        this(fileName, fileManager, false, false, context);
    }

    /**
     * Like {@link #AppsManager(String, FileManager, boolean, Context)} without
     * resetting the cache file.
     */
    public AppsManager(String fileName, FileManager fileManager, boolean manualPlatformFullUpdate, Context context)
    {
        this(fileName, fileManager, false, manualPlatformFullUpdate, context);
    }

    /**
     * @param fileName The name of the file to store the full content list cache
     * @param fileManager The file manager used to store the content list cache
     * @param resetFile True if the cache file should be reset
     * @param manualPlatformFullUpdate True if the auto content sync-up with
     * platform is not required. Therefore, {@link #forceRefresh()} needs to be
     * triggered manually to force a sync-up
     * @param context The app context
     */
    public AppsManager(String fileName, FileManager fileManager, boolean resetFile, boolean manualPlatformFullUpdate, Context context)
    {
        super(fileName, fileManager, manualPlatformFullUpdate, context);
        _receiver = new BroadcastReceiver()
        {
            @Override
            public void onReceive(Context arg0, Intent arg1)
            {
                onContentUpdate();
            }
        };
        initialize(resetFile);
    }

    @Override
    protected void watchForContentUpdates(Context context)
    {
        IntentFilter f = new IntentFilter();
        f.addDataScheme("package");
        f.addAction(Intent.ACTION_PACKAGE_ADDED);
        f.addAction(Intent.ACTION_PACKAGE_REMOVED);
        f.addAction(Intent.ACTION_PACKAGE_CHANGED);
        f.addAction(Intent.ACTION_PACKAGE_REPLACED);
        context.registerReceiver(_receiver, f);
    }

    @Override
    protected void stopWatchingForContentUpdates(Context context)
    {
        context.unregisterReceiver(_receiver);
    }

    @Override
    protected PlatformFullIterator getPlatformFullIterator()
    {
        final PackageManager pm = _context.getPackageManager();
        final Set<String> results = new TreeSet<String>(new Comparator<String>()
        {
            @Override
            public int compare(String s1, String s2)
            {
                return s1.compareTo(s2);
            }
        });

        List<ApplicationInfo> installedApplications = null;
        try
        {
            installedApplications = pm.getInstalledApplications(0);
        }
        catch (Exception e)
        {
            Logger.error(this, "Unable to fetch installed application list from device.", e);
        }

        if (installedApplications != null)
        {
            for (int idx1 = 0; idx1 < installedApplications.size(); idx1++)
            {
                String packageName = installedApplications.get(idx1).packageName;
                if (packageName == null || packageName.length() == 0)
                    continue;

                Intent intent = new Intent(Intent.ACTION_MAIN);
                intent.addCategory(Intent.CATEGORY_LAUNCHER);
                intent.setPackage(packageName);

                try
                {
                    List<ResolveInfo> resolveInfos = pm.queryIntentActivities(intent, 0);
                    for (int idx2 = 0; idx2 < resolveInfos.size(); idx2++)
                    {
                        String appName = resolveInfos.get(idx2).loadLabel(pm).toString();
                        if (appName.length() > 0)
                        {
                            Logger.debug(this, "Indexing application: " + appName);
                            results.add(appName);
                        }
                    }
                }
                catch (Exception e)
                {
                    Logger.error(this, "Unable to fetch application name in " + packageName, e);
                    continue;
                }
            }
        }

        return new PlatformFullIterator()
        {
            private Iterator<String> it = results != null ? results.iterator() : null;

            @Override
            public boolean hasNext()
            {
                if (it != null)
                    return it.hasNext();
                return false;
            }

            @Override
            public WordAction next()
            {
                WordAction wordAction = null;
                if (it != null)
                {
                    String name = it.next();
                    wordAction = new WordAction(new Word(name, null), true);
                }
                return wordAction;
            }

            @Override
            public void remove()
            {
                // Not allowed
            }

            @Override
            public int getSize()
            {
                return results != null ? results.size() : 0;
            }
        };
    }
}
