package com.nuance.dragon.toolkit.grammar.content;

import android.content.Context;
import android.database.Cursor;
import android.net.Uri;
import android.provider.ContactsContract;
import android.provider.ContactsContract.CommonDataKinds;
import android.provider.ContactsContract.Data;

import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.oem.api.FileManager;

/**
 * A {@link ContentManager} that provides the full list of contacts in the
 * device's phonebook.
 */
public class ContactManager extends UriContentManager
{
    //use data table instead of contacts table, it joins many other tables' column
    private final static Uri URI = ContactsContract.Data.CONTENT_URI;
    //private final static Uri URI = ContactsContract.Contacts.CONTENT_URI;

    // join contacts table columns to query data table at the same time with mime type
    private final static String[] PROJECTION = {
        ContactsContract.Contacts.LOOKUP_KEY,
        ContactsContract.Contacts.DISPLAY_NAME,
        ContactsContract.Contacts.IN_VISIBLE_GROUP,
        ContactsContract.CommonDataKinds.StructuredName.PHONETIC_GIVEN_NAME,
        ContactsContract.CommonDataKinds.StructuredName.PHONETIC_MIDDLE_NAME,
        ContactsContract.CommonDataKinds.StructuredName.PHONETIC_FAMILY_NAME
    };

    /**
     * Like {@link #ContactManager(String, FileManager, boolean, Context)} using auto
     * content sync-up without resetting the cache file.
     */
    public ContactManager(String fileName, FileManager fileManager, Context context)
    {
        this(fileName, fileManager, false, false, context);
    }

    /**
     * Like {@link #ContactManager(String, FileManager, boolean, Context)} without
     * resetting the cache file.
     */
    public ContactManager(String fileName, FileManager fileManager, boolean manualPlatformFullUpdate, Context context)
    {
        this(fileName, fileManager, false, manualPlatformFullUpdate, context);
    }

    /**
     * @param fileName The name of the file to store the full content list cache
     * @param fileManager The file manager used to store the content list cache
     * @param resetFile True if the cache file should be reset
     * @param manualPlatformFullUpdate True if the auto content sync-up with
     * platform is not required. Therefore, {@link #forceRefresh()} needs to be
     * triggered manually to force a sync-up
     * @param context The app context
     */
    public ContactManager(String fileName, FileManager fileManager, boolean resetFile, boolean manualPlatformFullUpdate, Context context)
    {
        super(fileName, fileManager, URI, resetFile, manualPlatformFullUpdate, context);
    }

    @Override
    protected PlatformFullIterator getPlatformFullIterator()
    {
        return new UriPlatformFullIterator(_context,
                                            _uri,
                                            PROJECTION,
                                            Data.MIMETYPE + "='" + CommonDataKinds.StructuredName.CONTENT_ITEM_TYPE + "'",
                                            null,
                                            null)
        {
            private int colLookupKey;
            private int colName;
            private int colVisible;
            private int colGivenName;
            private int colMiddleName;
            private int colFamilyName;

            @Override
            protected void init(Cursor cursor)
            {
                colLookupKey  = cursor.getColumnIndex(PROJECTION[0]);
                colName       = cursor.getColumnIndex(PROJECTION[1]);
                colVisible    = cursor.getColumnIndex(PROJECTION[2]);
                colGivenName  = cursor.getColumnIndex(PROJECTION[3]);
                colMiddleName = cursor.getColumnIndex(PROJECTION[4]);
                colFamilyName = cursor.getColumnIndex(PROJECTION[5]);
            }

            @Override
            protected WordAction getNext(Cursor cursor)
            {
                try {
                    while (cursor.moveToNext()) {
                        if (cursor.getInt(colVisible) == 0)
                            continue;

                        String lookupKey = cursor.getString(colLookupKey);
                        String name = cursor.getString(colName);

                        if (name == null || name.length() == 0 || lookupKey == null || name.equals('\u0000'))
                            continue;

                        return new WordAction(new Word(name, getPhoneticName(cursor)), true);
                    }
                }
                catch (RuntimeException e) {
                }
                return null;
            }

            private String getPhoneticName(Cursor cursor) {
                String phoneticName = "";

                String phoneticFirstName = cursor.getString(colGivenName);
                String phoneticMiddleName = cursor.getString(colMiddleName);
                String phoneticFamilyName = cursor.getString(colFamilyName);

                if (phoneticFamilyName != null && phoneticFamilyName.length() != 0) {
                    phoneticName = phoneticFamilyName;
                }
                if (phoneticMiddleName != null && phoneticMiddleName.length() != 0) {
                    phoneticName += phoneticMiddleName;
                }
                if (phoneticFirstName != null && phoneticFirstName.length() != 0) {
                    phoneticName += phoneticFirstName;
                }

                phoneticName = (phoneticName.length() == 0 ? null : phoneticName);

                return /*removeInvalidData(*/phoneticName/*)*/;
            }

            /*private String removeInvalidData(String input)
            {
                int index=0, value;

                if(input==null||input.length()==0)
                 return "";

                char[] result = new char[input.length()];
                for (int i=0;i<input.length();i++) {
                    value = input.codePointAt(i);
                    if ((value>0x3040&&value<0x3130)||(value>0xff60&&value<0xffa0))  // Only Japanese character
                    {
                        result[index++]=input.charAt(i);
                    }
                }
                return new String(result,0,index);
            }*/

        };
    }


}
