package com.nuance.dragon.toolkit.grammar.content;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import junit.framework.Assert;

import org.pcollections.TreePVector;

import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.core.WordList;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.AsyncTaskHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Implementation of {@link WordList} that pulls from a {@link ContentManager}.
 * Keeps track of the items accepted by a wordlist iterator.
 */
class ContentList extends WordList
{
    private class WordIteratorBase implements WordList.WordIterator
    {
        private final List<WordAction> _changeList;
        private final TreePVector<WordAction> _oldAccepted;
        private TreePVector<WordAction> _newAccepted;
        private final Iterator<WordAction> _it;
        private final int _count;
        private int _numAccepted;
        private int _checksum;
        private boolean _finished;

        /**
         * @param acceptedList The last accepted list
         * @param changeList The changes between the full list, and the last accepted list
         */
        WordIteratorBase(TreePVector<WordAction> acceptedList, List<WordAction> changeList)
        {
            Assert.assertNotNull(changeList);
            Assert.assertNotNull(acceptedList);
            _changeList = changeList;
            _oldAccepted = acceptedList;
            _newAccepted = acceptedList;

            _numAccepted = 0;
            _it = _changeList.iterator();
            _count = _changeList.size();
        }

        private void finish(boolean accept)
        {
            Checker.checkState(this, !_finished);
            _finished = true;
            accept(this, _oldAccepted, accept ? _newAccepted : null, getChecksum());
        }

        @Override
        public void acceptChanges()
        {
            finish(_numAccepted > 0);
        }

        @Override
        public void discardChanges()
        {
            finish(false);
        }

        @Override
        public boolean hasNext()
        {
            return _it.hasNext();
        }

        @Override
        public WordAction next()
        {
            WordAction next = _it.next();
            _numAccepted++;

            if (next.getAction() == WordAction.ADD ||
                next.getAction() == WordAction.UPDATE)
            {
                WordAction toBeAdded = new WordAction(next.getWord(), true);
                _checksum += next.getWord().hashCode();
                _newAccepted = _newAccepted.plus(toBeAdded);
            }
            else {
                WordAction toBeDeleted = new WordAction(next.getWord(), true);
                if (_newAccepted.contains(toBeDeleted))
                {
                    _newAccepted = _newAccepted.minus(toBeDeleted);
                    _checksum -= next.getWord().hashCode();
                } else
                {
                    Assert.assertTrue(false);
                }
            }

            return next;
        }

        @Override
        public void remove()
        {
            //_it.remove();
        }

        @Override
        public int getCount()
        {
            return _count;
        }

        @Override
        public int getChecksum()
        {
            return _checksum;
        }
    }

    private static class AcceptedState
    {
        final TreePVector<WordAction> acceptedList;
        final int acceptedChecksum;
        final String fullListId;

        AcceptedState(TreePVector<WordAction> acceptedList, int acceptedChecksum, String fullListUuid)
        {
            this.acceptedList = acceptedList;
            this.acceptedChecksum = acceptedChecksum;
            this.fullListId = fullListUuid;
        }
    }

    private static int FILE_VERSION = 1;

    private final FileManager _fileManager;
    private final String _fileName;
    private final boolean _startFromScratch;
    private final AsyncTaskHandler _workerThreadHandler;
    private final Object _syncObj;

    private WordIteratorBase _curIt;

    private List<WordAction> _fullList;
    private TreePVector<WordAction> _acceptedList;  // The most recently accepted list
    private List<WordAction> _changeList;    // The changes since accepting
    private TreePVector<WordAction> _acceptedList4It; // Accepted list to give to new iterators
    private List<WordAction> _changeList4It; // Change list to give to new iterators
    private String _fullListId; // When this changes, it means the full list has been reset.
                                // So the accepted state here should also be reset.
                                // This should only be set upon initialization.
    private int _acceptedChecksum;
    private boolean _accepting;

    private final Object _initSync = new Object();
    private boolean _initializing;          // initialize() has been called, but the update has not completed
    private boolean _initializeComplete;    // The initial update has completed

    /**
     * Create from main thread. Make sure to call {@link #initialize} on the worker thread when it
     * is ready to be initialized!
     * @param manager
     * @param fileManager
     * @param fileName
     * @param startFromScratch
     * @param sync
     */
    ContentList(ContentManager manager,
                FileManager fileManager,
                String fileName,
                boolean startFromScratch,
                Object sync)
    {
        _fileManager = fileManager;
        _fileName = fileName;
        _startFromScratch = startFromScratch;
        _workerThreadHandler = new AsyncTaskHandler();
        _syncObj = sync;

        synchronized (_initSync)
        {
            _initializing = true;
        }
    }

    /**
     * Call from worker thread
     * @param fullList The full content list
     * @param fullListId The generated ID of the full list. This changes whenever
     * the ContentManager's full file is reset.
     * @param rebuild True to force the content list to start from scratch
     */
    void initialize(final List<WordAction> fullList, final String fullListId, boolean rebuild)
    {
        Assert.assertNotNull(fullListId);

        TreePVector<WordAction> acceptedList = null;
        int acceptedChecksum = 0;
        List<WordAction> changeList;

        _fullListId = fullListId;

        if (!rebuild && !_startFromScratch) {
            AcceptedState acceptedState = loadFile();
            if (acceptedState != null)
            {
                // Only use the saved state if the ID matches the one from the content manager
                if (fullListId.equals(acceptedState.fullListId))
                {
                    acceptedList = acceptedState.acceptedList;
                    acceptedChecksum = acceptedState.acceptedChecksum;
                }
            }
        }

        if (acceptedList == null)
        {
            changeList = fullList;
            acceptedList = TreePVector.empty();
            _fileManager.delete(_fileName);
        } else
        {
            changeList = findChangeList(fullList, acceptedList);
        }

        synchronized (_syncObj)
        {
            _acceptedChecksum = acceptedChecksum;
            _acceptedList = acceptedList;
            _fullList = fullList;
            _changeList = changeList;
        }

        completeInitialization();
    }

    /** Called from any thread */
    void updateFullList(final List<WordAction> fullList, final ContentManager manager)
    {
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                synchronized (_syncObj)
                {
                    _fullList = fullList;
                    _changeList = findChangeList(fullList, _acceptedList);
                    Logger.debug(this, "Word list sync real done");
                    manager.onSyncDone();
                }
            }
        });
    }

    @Override
    public boolean fullUpdateRequired()
    {
        waitForInitialization();

        synchronized(_syncObj) {
            if (_accepting)
            {
                Logger.debug(this, "Full update NOT required, the list is currently being accepted.");
                return false;
            }

            if (_acceptedList.isEmpty() && !_fullList.isEmpty())
            {
                Logger.debug(this, "Full update required, the accepted list is empty.");
                return true;
            }

            if (_changeList.size() > _fullList.size())
            {
                Logger.debug(this, "Full update required, the change list is bigger than the full list.");
                return true;
            }
        }
        Logger.debug(this, "Full update NOT required.");
        return false;
    }

    @Override
    public int getAcceptedChecksum()
    {
        waitForInitialization();

        synchronized (_syncObj)
        {
            return _acceptedChecksum;
        }
    }

    @Override
    public WordIterator getFullIterator()
    {
        return getIterator(true);
    }

    @Override
    public WordIterator getModifiedIterator()
    {
        return getIterator(false);
    }

    private WordIterator getIterator(boolean isFullList)
    {
        waitForInitialization();

        synchronized(_syncObj) {
            if (isFullList && !_fullList.isEmpty())
            {
                // An empty accepted list, and a full change list, make the iterator act like a full list
                TreePVector<WordAction> acceptedList = TreePVector.empty();
                _curIt = new WordIteratorBase(acceptedList, _fullList);
            }
            else if (_accepting)
                _curIt = new WordIteratorBase(_acceptedList4It, _changeList4It);
            else
                _curIt = new WordIteratorBase(_acceptedList, _changeList);
            return _curIt;
        }
    }

    private void waitForInitialization()
    {
        synchronized (_initSync)
        {
            while (!_initializeComplete)
            {
                Logger.debug(this, "Waiting for initialization to complete...");
                if (!_initializing)
                {
                    Logger.error(this, "Not initialized or initializing!");
                    break;
                }

                try {
                    _initSync.wait();
                }
                catch (InterruptedException e) {
                }
            }
        }
    }

    private void completeInitialization()
    {
        synchronized (_initSync)
        {
            if (!_initializeComplete)
            {
                _initializing = false;
                _initializeComplete = true;
                _initSync.notifyAll();
            }
        }
    }

    /** Called from any thread */
    private void accept(WordList.WordIterator iterator, final List<WordAction> initialList, final TreePVector<WordAction> acceptedList, final int checksum)
    {
        synchronized (_syncObj)
        {
            if (iterator != _curIt)
            {
                Logger.error(this, "Can't accept changes on old iterator");
                return;
            }

            _curIt = null;

            if (acceptedList != null)
            {
                Logger.debug(this, "Storing temporary list while accept finishes");
                _accepting = true;
                _acceptedList4It = acceptedList;
                _changeList4It = findChangeList(initialList, acceptedList);
                _acceptedChecksum = checksum;

                _workerThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        synchronized (_syncObj)
                        {
                            if (!_accepting)
                            {
                                Logger.error(ContentList.this, "Accepting unexpectedly interrupted");
                                return;
                            }

                            _acceptedChecksum = checksum;
                            _acceptedList = acceptedList;
                            _changeList = findChangeList(_fullList, acceptedList);
                            _acceptedList4It = null;
                            _changeList4It = null;
                            _accepting = false;

                            Logger.debug(this, "Updated accepted list");
                        }

                        saveFile(new AcceptedState(acceptedList, checksum, _fullListId));
                    }
                });
            }
        }
    }

   /**
    * helper method to find the change list from a full iterator and an accepted list.
    * @param fullIt
    * @param acceptedList
    * @return change list
    */
    private List<WordAction> findChangeList(List<WordAction> fullList, List<WordAction> accepted)
    {
        if (accepted.isEmpty())
            return fullList;

        List<WordAction> changeList = new ArrayList<WordAction>();

        // Make a copy to mess with, since we don't know where 'accepted' came from (could be
        // a blocking full-rebuild iterator, which doesn't support remove)
        List<WordAction> acceptedList = new ArrayList<WordAction>(accepted);

        for (WordAction action : fullList)
        {
            if (acceptedList.contains(action))
                acceptedList.remove(action);
            else
                changeList.add(action);
        }
        for (WordAction action : acceptedList)
        {
            changeList.add(new WordAction(action.getWord(), false));
        }
        return changeList;
    }

    private boolean saveFile(AcceptedState acceptedState)
    {
        synchronized(_syncObj)
        {
            Logger.debug(this, "Writing to file " + _fileName);
            boolean success = false;
            ObjectOutputStream oos = ContentManager.createWriteStream(_fileManager, _fileName);
            if (oos != null)
            {
                try {
                    oos.writeInt(FILE_VERSION);
                    oos.writeUTF(acceptedState.fullListId);
                    oos.writeInt(acceptedState.acceptedChecksum);
                    WordAction.writeListTo(acceptedState.acceptedList, oos);
                    Logger.debug(this, "Saved " + acceptedState.acceptedList.size() + " items");
                    success = true;
                }
                catch (IOException e) {
                    Logger.error(this, "Error reading to file " + _fileName, e);
                }

                try { oos.close(); } catch(IOException e) { }
            }
            Logger.debug(this, "Done writing to file");
            return success;
        }
    }

    private AcceptedState loadFile()
    {
        synchronized(_syncObj)
        {
            Logger.debug(this, "Reading from file " + _fileName);
            TreePVector<WordAction> acceptedList = null;
            int acceptedChecksum = 0;
            String fullListId = null;
            ObjectInputStream ois = ContentManager.createReadStream(_fileManager, _fileName);
            if (ois != null)
            {
                try {
                    int fileVersion = ois.readInt();
                    if (fileVersion == FILE_VERSION)
                    {
                        fullListId = ois.readUTF();
                        acceptedChecksum = ois.readInt();
                        acceptedList = WordAction.readPersistentListFrom(ois);
                        Logger.debug(this, "Loaded " + acceptedList.size() + " items");
                    }
                }
                catch (IOException e) {
                    Logger.error(this, "Error reading from file " + _fileName, e);
                }

                try { ois.close(); } catch(IOException e) { }
            }
            Logger.debug(this, "Done reading from file");
            return acceptedList == null ? null : new AcceptedState(acceptedList, acceptedChecksum, fullListId);
        }
    }
}
