package com.nuance.dragon.toolkit.grammar.content;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.UUID;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import junit.framework.Assert;
import android.content.Context;
import android.os.Handler;
import android.util.Pair;

import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.core.WordList;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.AsyncTaskHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.util.internal.BlockingIteratorList;

/**
 * Base class for content managers. The content manager keeps track of the full list of content.
 * Multiple {@link WordList} instances can be obtained via {@link #createWordList}, each
 * maintaining a separate persistent state.
 */
public abstract class ContentManager
{
    /**
     * Listens for the completion of the sync. operation.
     */
    public interface SyncListener
    {
        /**
         * The sync. operation has completed.
         * @param contentManager The current content manager
         */
        void onDone(ContentManager contentManager);
    }

    /**
     * Iterator returned by {@link ContentManager#getPlatformFullIterator}.
     * Must iterate through the full content list, returning {@link WordAction}s
     * signifying 'added' words.
     */
    protected interface PlatformFullIterator extends Iterator<WordAction>
    {
        /**
         * @return The estimated number of total items in the iterator. < 0 if
         * it can't be determined.
         */
        int getSize();
    }

    private static class PendingFullUpdate
    {
        interface FullUpdater
        {
            /**
             * Called when the timer has finally elapsed, and it's time to do the update.
             * @param pendingUpdate
             */
            void update(PendingFullUpdate pendingUpdate);
        }

        private final Runnable _runnable;
        private final Handler _handler;
        private boolean _reset;
        private boolean _canceled;

        PendingFullUpdate(final int waitTime, final int maxWaitTime, final FullUpdater updater)
        {
            _handler = new Handler();
            _runnable = new Runnable()
            {
                private int _waited;

                @Override
                public void run()
                {
                    if (_canceled)
                        return;

                    _waited += waitTime;
                    if (_reset && _waited < maxWaitTime)
                    {
                        _reset = false;
                        _handler.postDelayed(_runnable, waitTime);
                    } else
                    {
                        _canceled = true;
                        updater.update(PendingFullUpdate.this);
                    }
                }
            };
            _handler.postDelayed(_runnable, waitTime);
        }

        /** Reset the update timer */
        void reset()
        {
            _reset = true;
        }

        void cancel()
        {
            _canceled = true;
            _handler.removeCallbacks(_runnable);
        }
    }

    // Saved to files to prevent loading an old version. Increment when the file format changes.
    private static final int FILE_VERSION = 12346;

    // value constants
    private static final int SYNC_UP_DELAY = 10000;
    private static final int MAX_SYNC_UP_DELAY = 3600000;

    protected final Context _context;
    private final String _fileName;
    private final FileManager _fileManager;
    private final Handler _mainThreadHandler;
    private final AsyncTaskHandler _rebuildThreadHandler;    // Full rebuilds happen here, and can take a long time
    private final AsyncTaskHandler _workerThreadHandler;     // Most things happen here, necessary due to file IO, change-list detection
    private PendingFullUpdate _pendingFullUpdate;
    private final Object _syncObj = new Object();
    private boolean _initialized;           // Whether initialize() has been called
    private boolean _ready;                 // The content manager is ready to be used
    private boolean _initializeComplete;
    private boolean _manualPlatformFullUpdate;
    private int _limit;

    // lists
    private List<WordAction> _fullList; // The list of everything (accepted + change list)
    private String _fullListUuid;       // Updated every time the full list is reset from scratch.
                                        // Allows the ContentList to know whether it should
                                        // also start from scratch.

    // Content lists
    private final List<WeakReference<ContentList>> _contentLists;

    // Optional sync. listeners
    private SyncListener _initListener, _updateListener;

    /** Make sure to call {@link #initialize}! */
    public ContentManager(String fileName, FileManager fileManager, boolean manualPlatformFullUpdate, Context context)
    {
        Checker.checkStringArgForNullOrEmpty("fileName", fileName);
        Checker.checkArgForNull("fileManager", fileManager);
        Checker.checkArgForNull("context", context);

        _fileName = fileName;
        _fileManager = fileManager;
        _manualPlatformFullUpdate = manualPlatformFullUpdate;
        _context = context.getApplicationContext();
        _contentLists = new ArrayList<WeakReference<ContentList>>();

        _mainThreadHandler = new Handler();
        _rebuildThreadHandler = new AsyncTaskHandler();
        _workerThreadHandler = new AsyncTaskHandler();

        // Initialize list to empty
        _fullList = new ArrayList<WordAction>(0);

        // Set stock limit
        _limit = 10000;
    }

    /**
     * Initialize the word list. Call this exactly once in your subclass constructor.
     * @param resetFile true if the file should be reset
     */
    protected void initialize(final boolean resetFile)
    {
        Checker.checkState(this, !_initialized, "Already initialized");
        _initialized = true;

        watchForContentUpdates(_context);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                initialLoad(resetFile);
            }
        });
    }

    /**
     * Not for public use.
     * @param clazz class
     * @param listener listener
     */
    public void setInitListener(String clazz, SyncListener listener)
    {
        // "clazz" is just here to make sure that if external
        // consumers call this method, nothing will happen.
        // "clazz" being the class name of the caller.
        if (clazz != null && clazz.equals("GrammarDepot"))
        {
            // Fire listener immediately to make sure consumer
            // (i.e GrammarDepot) always get the notification
            // even if the content manager has been initialized
            // sometimes earlier.
            if (_ready)
            {
                if (listener != null)
                    listener.onDone(this);
            }
            else
            {
                // Set it for later
                _initListener = listener;
            }
        }
    }

    /**
     * Set a limit for the number of device's entries indexed by this content
     * manager. This is useful to comply with the limitations of our
     * recognition back-ends and hosting facilities.
     * @param limit The maximum number of entries for this content manager.
     */
    public final void setLimit(int limit)
    {
        _limit = limit;
    }

    /**
     * Set a listener to be notified when the device's content has been updated.
     * @param listener notification listener
     */
    public void setUpdateListener(SyncListener listener)
    {
        _updateListener = listener;
    }

    /**
     * Dispose the word list. Call this exactly once when finished.
     */
    public void release()
    {
        Checker.checkState(this, _initialized, "Not initialized");
        _initialized = false;

        stopWatchingForContentUpdates(_context);
        if (_pendingFullUpdate != null)
        {
            _pendingFullUpdate.cancel();
            _pendingFullUpdate = null;
        }

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _contentLists.clear();
                writeToFile(_fullList, _fullListUuid);
                _fullList = null;
                _rebuildThreadHandler.cancelAll();
                _workerThreadHandler.cancelAll();
            }
        });
    }

    /**
     * Get a word list. The status of the accepted items will be kept track of in the file.
     * <p>
     * For example, you may have two components in your application that rely on the contact
     * list, but main two different accepted states. You can use this method to obtain two
     * separate word lists without needing separate content managers.
     * @param fileManager The file manager to use for the word list's file.
     * @param fileName The file to save the word list's state.
     * @param startFromScratch If true, start fresh, i.e. the accepted list is empty, and the
     * modified list is the same as the full list.
     * @return A word list
     */
    public WordList createWordList(FileManager fileManager, String fileName, boolean startFromScratch)
    {
        Checker.checkState(this, _initialized, "Not initialized");
        Checker.checkArgForNull("fileManager", fileManager);
        Checker.checkStringArgForNullOrEmpty("fileName", fileName);
        Checker.checkArgForCondition(fileName, "different from the ContentManager's file name", !fileName.equals(_fileName));

        final ContentList list = new ContentList(this, fileManager, fileName, startFromScratch, _syncObj);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _contentLists.add(new WeakReference<ContentList>(list));

                if (_initializeComplete)
                    list.initialize(_fullList, _fullListUuid, false); // TODO: When should we pass true here?
            }
        });
        return list;
    }

    /**
     * Force the word list to update itself from its content provider.
     */
    public void forceRefresh()
    {
        Checker.checkState(this, _initialized, "Not initialized");
        goPlatformFullUpdate();
    }

    /** Perform the initial load on startup. Call on worker thread. */
    private void initialLoad(boolean resetFile)
    {
        List<WordAction> readList = null;
        String fullListUuid = null;
        if (resetFile)
        {
            _fileManager.delete(_fileName);
        } else
        {
            Pair<List<WordAction>, String> readListAndUuid = readFromFile();
            if (readListAndUuid != null)
            {
                readList = readListAndUuid.first;
                fullListUuid = readListAndUuid.second;
            }
        }

        boolean completedInitialization = false;
        if (readList != null)
        {
            synchronized (_syncObj)
            {
                _fullList = readList;
                _fullListUuid = fullListUuid;
            }
            completeInitialization(readList, fullListUuid, false);
            completedInitialization = true;
        }

        if (fullListUuid == null)
            fullListUuid = UUID.randomUUID().toString(); // We're generating from scratch with no file, so generate a new UUID

        if (_manualPlatformFullUpdate && readList == null)
        {
            synchronized (_syncObj)
            {
                _fullList = new ArrayList<WordAction>(0);
                _fullListUuid = fullListUuid;
            }
            completeInitialization(readList, fullListUuid, true);
            completedInitialization = true;
        }

        if (_manualPlatformFullUpdate)
            return;

        // Always do a full update
        final PlatformFullIterator newFullList = getPlatformFullIterator();
        if (newFullList == null || !newFullList.hasNext())
        {
            synchronized (_syncObj)
            {
                // The full list is already empty
                _fullListUuid = fullListUuid;
            }

            if (!completedInitialization)
                completeInitialization(new ArrayList<WordAction>(0), fullListUuid, false);
        } else
        {
            // If we're doing a full rebuild to start, we use a special RebuildingList
            // that returns a blocking iterator
            final BlockingIteratorList<WordAction> rebuildingList = new BlockingIteratorList<WordAction>(newFullList.getSize());
            if (readList == null)
            {
                synchronized (_syncObj)
                {
                    _fullList = rebuildingList;
                    _fullListUuid = fullListUuid;
                }
            }

            // Complete the initialization once we have the rebuilding list in place
            if (!completedInitialization)
                completeInitialization(rebuildingList, fullListUuid, true);

            final List<WordAction> fullList = _fullList;
            _rebuildThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    final List<WordAction> fastList = new ArrayList<WordAction>();
                    boolean hasNext = true;
                    int itemCount = 0;
                    do
                    {
                        WordAction next = newFullList.next();
                        hasNext = newFullList.hasNext() && itemCount < _limit;
                        rebuildingList.add(next, !hasNext);
                        fastList.add(next);
                        itemCount++;
                    } while (hasNext);

                    _workerThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            if (_fullList != fullList)
                            {
                                Logger.warn(ContentManager.this, "Rebuild finished, but appears to have been aborted");
                                return;
                            }

                            String uuid;
                            synchronized (_syncObj)
                            {
                                uuid = _fullListUuid;
                                // update ContentLists if readList is not null initially
                                if (rebuildingList != _fullList)
                                    updateContentListsFullLists(fastList, uuid);

                                _fullList = fastList;
                            }
                            // Now save
                            writeToFile(fastList, uuid);
                        }
                    });
                }
            });
        }
    }

    /** call from worker thread */
    private void completeInitialization(List<WordAction> fullList, String fullListUuid, boolean rebuild)
    {
        synchronized (_syncObj)
        {
            _initializeComplete = true;
            for (int i = 0; i < _contentLists.size(); i++)
            {
                WeakReference<ContentList> listRef = _contentLists.get(i);
                ContentList contentList = listRef.get();
                if (contentList != null)
                {
                    contentList.initialize(fullList, fullListUuid, rebuild);
                } else
                {
                    _contentLists.remove(i);
                    i--;
                }
            }
        }

        onSyncDone();
    }

    /**
     * Request a full sync up with the OS. It is delayed to avoid doing full
     * sync-up all the time. A non-delayed version is goPlatformFullUpdate().
     */
    private void requestPlatformFullUpdate()
    {
        if (_pendingFullUpdate != null)
        {
            _pendingFullUpdate.reset();
        } else
        {
            _pendingFullUpdate = new PendingFullUpdate(SYNC_UP_DELAY, MAX_SYNC_UP_DELAY,
                                                       new PendingFullUpdate.FullUpdater()
            {
                @Override
                public void update(PendingFullUpdate pendingUpdate)
                {
                    Assert.assertEquals(_pendingFullUpdate, pendingUpdate);
                    _pendingFullUpdate = null;
                    goPlatformFullUpdate();
                }
            });
        }
    }

    /**
     * Send a message to the worker thread to perform a full sync.
     */
    private void goPlatformFullUpdate()
    {
        Logger.debug(this, "goPlatformFullUpdate() called");
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                platformFullUpdate();
            }
        });
    }

    /** Perform the full sync. Call from worker thread. */
    private void platformFullUpdate()
    {
        // get the new full iterator from OS
        final PlatformFullIterator newFullList = getPlatformFullIterator();

        _rebuildThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                int size = newFullList.getSize();
                if (size < 0) size = 0;
                final List<WordAction> list = new ArrayList<WordAction>(size);

                int itemCount = 0;
                while (newFullList.hasNext() && itemCount < _limit)
                {
                    list.add(newFullList.next());
                    itemCount++;
                }

                _workerThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        // save the resulting lists
                        String uuid;
                        synchronized(_syncObj) {
                            if (_fullList == null) // Means we've been disposed
                                return;

                            _fullList = list;
                            uuid = _fullListUuid;
                            updateContentListsFullLists(list, uuid);
                        }

                        writeToFile(list, uuid);
                    }
                });
            }
        });
    }

    /** Call from any thread */
    void onSyncDone()
    {
        _mainThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _ready = true;

                // The initialization listener is one-shot only
                if (_initListener != null)
                {
                    _initListener.onDone(ContentManager.this);
                    _initListener = null;
                }

                if (_updateListener != null)
                    _updateListener.onDone(ContentManager.this);
            }
        });
    }

    /** Call from worker thread */
    private void updateContentListsFullLists(List<WordAction> fullList, String uuid)
    {
        for (int i = 0; i < _contentLists.size(); i++)
        {
            WeakReference<ContentList> listRef = _contentLists.get(i);
            ContentList contentList = listRef.get();
            if (contentList != null)
            {
                contentList.updateFullList(fullList, this);
            } else
            {
                _contentLists.remove(i);
                i--;
            }
        }
    }

    /** Call from worker thread */
    private Pair<List<WordAction>, String> readFromFile()
    {
        synchronized(_syncObj)
        {
            Logger.debug(this, "Reading from file " + _fileName);

            List<WordAction> fullList = null;
            String uuid = null;

            ObjectInputStream ois = createReadStream(_fileManager, _fileName);
            if (ois != null)
            {
                try {
                    int version = ois.readInt();
                    if (version == FILE_VERSION)
                    {
                        uuid = ois.readUTF();
                        fullList = WordAction.readListFrom(ois);
                        Logger.debug(this, "Loaded " + fullList.size() + " items");
                    }
                }
                catch (IOException e) {
                    Logger.error(this, "Error reading from file " + _fileName, e);
                }

                try { ois.close(); } catch(IOException e) { Logger.error(this, "Error closing file " + _fileName, e); }
            }

            Logger.debug(this, "Done reading from file");

            return fullList == null ? null : new Pair<List<WordAction>, String>(fullList, uuid);
        }
    }

    /** Call from worker thread */
    private void writeToFile(List<WordAction> fullList, String uuid)
    {
        synchronized(_syncObj)
        {
            Logger.debug(this, "Writing to file " + _fileName);
            ObjectOutputStream oos = createWriteStream(_fileManager, _fileName);
            if (oos != null)
            {
                try {
                    oos.writeInt(FILE_VERSION);
                    oos.writeUTF(uuid);
                    WordAction.writeListTo(fullList, oos);
                    Logger.debug(this, "Saved " + fullList.size() + " items");
                }
                catch (IOException e) {
                    Logger.error(this, "Error writing to file " + _fileName, e);
                }

                try { oos.close(); } catch(IOException e) { Logger.error(this, "Error saving file " + _fileName, e); }
            }
            Logger.debug(this, "Done writing to file");
        }
    }

    static ObjectInputStream createReadStream(FileManager fileManager, String fileName)
    {
        InputStream is = fileManager.openFileForReading(fileName);
        if (is == null) {
            Logger.warn(ContentManager.class, "Fail to open read-stream for file " + fileName);
            return null;
        }

        GZIPInputStream zis = null;
        ObjectInputStream ois = null;
        try
        {
            zis = new GZIPInputStream(is);
            ois = new ObjectInputStream(zis);

            return ois;
        } catch (Exception e)
        {
            Logger.error(ContentManager.class, "Error creating read-stream for file " + fileName, e);
        }

        if (zis != null)
            try { zis.close(); } catch (IOException e) {}
        if (is != null)
            try { is.close(); } catch (IOException e) {}
        return null;
    }

    static ObjectOutputStream createWriteStream(FileManager fileManager, String fileName)
    {
        OutputStream os = fileManager.openFileForWriting(fileName);
        if (os == null) {
            Logger.warn(ContentManager.class, "Fail to open write-stream for file " + fileName);
            return null;
        }

        GZIPOutputStream zos = null;
        ObjectOutputStream oos = null;
        try
        {
            zos = new GZIPOutputStream(os);
            oos = new ObjectOutputStream(zos);

            return oos;
        }
        catch (Exception e)
        {
            Logger.error(ContentManager.class, "Error creating write-stream for file " + fileName, e);
        }

        if (zos != null)
            try { zos.close(); } catch (IOException e) {}
        if (os != null)
            try { os.close(); } catch (IOException e) {}
        return null;
    }

    /**
     * Subclasses should call this whenever a content update is detected.
     */
    protected void onContentUpdate()
    {
        if (!_manualPlatformFullUpdate)
            requestPlatformFullUpdate();
    }

    /**
     * Start watching for content updates. Call {@link #onContentUpdate} whenever an update
     * is detected.
     * @param context
     */
    protected abstract void watchForContentUpdates(Context context);

    /**
     * Stop watching for content updates.
     * @param context
     */
    protected abstract void stopWatchingForContentUpdates(Context context);

    /**
     * To be implemented by subclass to provide the platform full iterator.
     * @return The full iterator.
     */
    protected abstract PlatformFullIterator getPlatformFullIterator();
}
