package com.nuance.dragon.toolkit.grammar.content;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import android.content.Context;

import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.oem.api.FileManager;

/**
 * A simple content manager that can be modified directly at runtime. Probably not too useful
 * in the real world, but nice for testing.
 */
public class SimpleContentManager extends ContentManager
{
    private static class FullIterator implements PlatformFullIterator
    {
        private final Iterator<WordAction> _it;
        private final int _size;

        FullIterator(List<WordAction> list)
        {
            _it = new ArrayList<WordAction>(list).iterator();
            _size = list.size();
        }

        @Override
        public boolean hasNext()
        {
            return _it.hasNext();
        }

        @Override
        public WordAction next()
        {
            return _it.next();
        }

        @Override
        public void remove()
        {
            _it.remove();
        }

        @Override
        public int getSize()
        {
            return _size;
        }
    }

    private boolean _updateToReport;
    private boolean _watchingForUpdates;
    private final ArrayList<WordAction> _fullList;

    /**
     * @param fileName The name of the file to store the full content list cache
     * @param fileManager The file manager used to store the content list cache
     * @param resetFile True if the cache file should be reset
     * @param manualPlatformFullUpdate True if the auto content sync-up with
     * platform is not required. Therefore, {@link #forceRefresh()} needs to be
     * triggered manually to force a sync-up
     * @param context The app context
     * @param initialContent The content to initialize with
     */
    public SimpleContentManager(String fileName,
                                FileManager fileManager,
                                boolean resetFile,
                                boolean manualPlatformFullUpdate,
                                Context context,
                                Word ... initialContent)
    {
        super(fileName, fileManager, manualPlatformFullUpdate, context);
        _fullList = new ArrayList<WordAction>(initialContent.length);
        for (Word w : initialContent)
        {
            _fullList.add(new WordAction(w, true));
        }
        initialize(resetFile);
    }

    /**
     * Add content to the content manager's list. This will trigger a full update.
     * @param words The words to add.
     */
    public void addContent(Collection<? extends Word> words)
    {
        _fullList.ensureCapacity(_fullList.size() + words.size());
        for (Word w : words)
        {
            _fullList.add(new WordAction(w, true));
        }
        handleUpdate();
    }

    /**
     * Add content to the content manager's list. This will trigger a full update.
     * @param words The words to add.
     */
    public void addContent(Word ... words)
    {
        _fullList.ensureCapacity(_fullList.size() + words.length);
        for (Word w : words)
        {
            _fullList.add(new WordAction(w, true));
        }
        handleUpdate();
    }

    /**
     * Remove content to the content manager's list. This will trigger a full update.
     * @param words The words to add.
     */
    public void removeContent(Collection<? extends Word> words)
    {
        for (Word w : words)
        {
            _fullList.remove(new WordAction(w, true));
        }
        handleUpdate();
    }

    /**
     * Remove content to the content manager's list. This will trigger a full update.
     * @param words The words to remove.
     */
    public void removeContent(Word ... words)
    {
        for (Word w : words)
        {
            _fullList.remove(new WordAction(w, true));
        }
        handleUpdate();
    }

    /**
     * Update content to the content manager's list. This will trigger a full update.
     * @param words The words to remove.
     */
    public void updateContent(Collection<? extends Word> words)
    {
        for (Word w : words)
        {
            for (WordAction wc : _fullList)
            {
                if (wc.getWord().getUserId() == w.getUserId())
                {
                    if (wc.getAction() ==  WordAction.UPDATE)
                        _fullList.remove(new WordAction(wc.getWord(), WordAction.UPDATE));
                    else
                        _fullList.remove(new WordAction(wc.getWord(), true));
                    break;
                }
            }
            _fullList.add(new WordAction(w, WordAction.UPDATE));
        }
        handleUpdate();
    }

    /**
     * Update content to the content manager's list. This will trigger a full update.
     * @param words The words to remove.
     */
    public void updateContent(Word ... words)
    {
        for (Word w : words)
        {
            for (WordAction wc : _fullList)
            {
                if (wc.getWord().getUserId() == w.getUserId())
                {
                    if (wc.getAction() ==  WordAction.UPDATE)
                        _fullList.remove(new WordAction(wc.getWord(), WordAction.UPDATE));
                    else
                        _fullList.remove(new WordAction(wc.getWord(), true));
                    break;
                }
            }
            _fullList.add(new WordAction(w, WordAction.UPDATE));
        }
        handleUpdate();
    }

    /**
     * Clear the content from the content manager's list. This will trigger a
     * full update.
     */
    public void clearContent()
    {
        _fullList.clear();
        handleUpdate();
    }

    @Override
    protected void watchForContentUpdates(Context context)
    {
        _watchingForUpdates = true;
        if (_updateToReport)
            handleUpdate();
    }

    @Override
    protected void stopWatchingForContentUpdates(Context context)
    {
        _watchingForUpdates = false;
    }

    @Override
    protected PlatformFullIterator getPlatformFullIterator()
    {
        return new FullIterator(_fullList);
    }

    private void handleUpdate()
    {
        if (_watchingForUpdates)
        {
            _updateToReport = false;
            onContentUpdate();
        } else
        {
            _updateToReport = true;
        }
    }
}
