package com.nuance.dragon.toolkit.grammar.content;

import android.content.Context;
import android.database.Cursor;
import android.provider.ContactsContract;
import android.provider.ContactsContract.CommonDataKinds;
import android.provider.ContactsContract.Data;

import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.core.data.Data.Dictionary;
import com.nuance.dragon.toolkit.oem.api.FileManager;

/**
 * A {@link ContentManager} that provides the full list of contacts in the
 * device's phonebook.
 */
public class StructuredContactManager extends ContactManager
{
    // join contacts table columns to query data table at the same time with mime type
    private final static String[] PROJECTION = {
        ContactsContract.Contacts.LOOKUP_KEY,
        ContactsContract.Contacts.DISPLAY_NAME,
        ContactsContract.Contacts.IN_VISIBLE_GROUP,
        ContactsContract.CommonDataKinds.StructuredName._ID,
        ContactsContract.CommonDataKinds.StructuredName.GIVEN_NAME,
        ContactsContract.CommonDataKinds.StructuredName.MIDDLE_NAME,
        ContactsContract.CommonDataKinds.StructuredName.FAMILY_NAME,
        ContactsContract.CommonDataKinds.StructuredName.PHONETIC_GIVEN_NAME,
        ContactsContract.CommonDataKinds.StructuredName.PHONETIC_MIDDLE_NAME,
        ContactsContract.CommonDataKinds.StructuredName.PHONETIC_FAMILY_NAME,
    };

    /**
     * Like {@link #StructuredContactManager(String, FileManager, boolean, Context)} using auto
     * content sync-up without resetting the cache file.
     */
    public StructuredContactManager(String fileName, FileManager fileManager, Context context)
    {
        this(fileName, fileManager, false, false, context);
    }

    /**
     * Like {@link #StructuredContactManager(String, FileManager, boolean, Context)} without
     * resetting the cache file.
     */
    public StructuredContactManager(String fileName, FileManager fileManager, boolean manualPlatformFullUpdate, Context context)
    {
        this(fileName, fileManager, false, manualPlatformFullUpdate, context);
    }

    /**
     * @param fileName The name of the file to store the full content list cache
     * @param fileManager The file manager used to store the content list cache
     * @param resetFile True if the cache file should be reset
     * @param manualPlatformFullUpdate True if the auto content sync-up with
     * platform is not required. Therefore, {@link #forceRefresh()} needs to be
     * triggered manually to force a sync-up
     * @param context The app context
     */
    public StructuredContactManager(String fileName, FileManager fileManager, boolean resetFile, boolean manualPlatformFullUpdate, Context context)
    {
        super(fileName, fileManager, resetFile, manualPlatformFullUpdate, context);
    }

    @Override
    protected PlatformFullIterator getPlatformFullIterator()
    {
        return new UriPlatformFullIterator(_context,
                                            _uri,
                                            PROJECTION,
                                            Data.MIMETYPE + "='" + CommonDataKinds.StructuredName.CONTENT_ITEM_TYPE + "'",
                                            null,
                                            null)
        {
            private int colLookupKey;
            private int colId;
            private int colName;
            private int colVisible;
            private int colGivenName;
            private int colMiddleName;
            private int colFamilyName;
            private int colPhoneticGivenName;
            private int colPhoneticMiddleName;
            private int colPhoneticFamilyName;

            @Override
            protected void init(Cursor cursor)
            {
                colLookupKey  = cursor.getColumnIndex(PROJECTION[0]);
                colName       = cursor.getColumnIndex(PROJECTION[1]);
                colVisible    = cursor.getColumnIndex(PROJECTION[2]);
                colId         = cursor.getColumnIndex(PROJECTION[3]);
                colGivenName  = cursor.getColumnIndex(PROJECTION[4]);
                colMiddleName = cursor.getColumnIndex(PROJECTION[5]);
                colFamilyName = cursor.getColumnIndex(PROJECTION[6]);
                colPhoneticGivenName  = cursor.getColumnIndex(PROJECTION[7]);
                colPhoneticMiddleName = cursor.getColumnIndex(PROJECTION[8]);
                colPhoneticFamilyName = cursor.getColumnIndex(PROJECTION[9]);
            }

            @Override
            protected WordAction getNext(Cursor cursor)
            {
                try {
                    while (cursor.moveToNext()) {
                        if (cursor.getInt(colVisible) == 0)
                            continue;

                        String lookupKey = cursor.getString(colLookupKey);
                        String name = cursor.getString(colName);

                        if (name == null || name.length() == 0 || lookupKey == null)
                            continue;

                        Dictionary content = new Dictionary();

                        int contentId      = cursor.getInt(colId);

                        String firstName = cursor.getString(colGivenName);
                        String middleName = cursor.getString(colMiddleName);
                        String familyName = cursor.getString(colFamilyName);

                        String phoneticFirstName = cursor.getString(colPhoneticGivenName);
                        if (phoneticFirstName == null || phoneticFirstName.length() == 0) {
                            phoneticFirstName = firstName;
                        }

                        String phoneticMiddleName = cursor.getString(colPhoneticMiddleName);
                        if (phoneticMiddleName == null || phoneticMiddleName.length() == 0) {
                            phoneticMiddleName = middleName;
                        }

                        String phoneticFamilyName = cursor.getString(colPhoneticFamilyName);
                        if (phoneticFamilyName == null || phoneticFamilyName.length() == 0) {
                            phoneticFamilyName = familyName;
                        }

//                        content.put("ad", getUniqueAddressTypes(addresses));
//                        content.put("c", contact.getCompany());

                        if (firstName != null && firstName.length() > 0) {
                            content.put("fn", firstName);
                        }

                        if (phoneticFirstName != null && phoneticFirstName.length() > 0) {
                            content.put("fnp", phoneticFirstName);
                        }

                        if (familyName != null && familyName.length() > 0) {
                            content.put("ln", familyName);
                        }

                        if (phoneticFamilyName != null && phoneticFamilyName.length() > 0) {
                            content.put("lnp", phoneticFamilyName);
                        }

                        if (middleName != null && middleName.length() > 0) {
                            content.put("mn", middleName);
                        }

                        if (phoneticMiddleName != null && phoneticMiddleName.length() > 0) {
                            content.put("mnp", phoneticMiddleName);
                        }
//                        content.put("nn", );
//                        content.put("nnp", );

                        return new WordAction(new Word(contentId, content), WordAction.ADD);
                    }
                }
                catch (RuntimeException e) {
                }
                return null;
            }
        };
    }


}
