package com.nuance.dragon.toolkit.grammar.content;

import android.content.Context;
import android.database.Cursor;

import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.core.data.Data.Dictionary;
import com.nuance.dragon.toolkit.oem.api.FileManager;

/**
 * A {@link ContentManager} that provides the full list of songs in the
 * device's music library.
 */
public class StructuredSongManager extends SongManager
{
    private static final String[] PROJECTION =  {
        android.provider.MediaStore.Audio.AudioColumns._ID,
        android.provider.MediaStore.MediaColumns.TITLE,
        android.provider.MediaStore.Audio.AudioColumns.IS_MUSIC,
        android.provider.MediaStore.Audio.AudioColumns.ARTIST,
        android.provider.MediaStore.Audio.AudioColumns.ALBUM,
    };

    /**
     * Like {@link #StructuredSongManager(String, FileManager, boolean, Context)} using auto
     * content sync-up without resetting the cache file.
     */
    public StructuredSongManager(String fileName, FileManager fileManager, Context context)
    {
        this(fileName, fileManager, false, false, context);
    }

    /**
     * Like {@link #StructuredSongManager(String, FileManager, boolean, Context)} without
     * resetting the cache file.
     */
    public StructuredSongManager(String fileName, FileManager fileManager, boolean manualPlatformFullUpdate, Context context)
    {
        this(fileName, fileManager, false, manualPlatformFullUpdate, context);
    }

    /**
     * @param fileName The name of the file to store the full content list cache
     * @param fileManager The file manager used to store the content list cache
     * @param resetFile True if the cache file should be reset
     * @param manualPlatformFullUpdate True if the auto content sync-up with
     * platform is not required. Therefore, {@link #forceRefresh()} needs to be
     * triggered manually to force a sync-up
     * @param context The app context
     */
    public StructuredSongManager(String fileName, FileManager fileManager, boolean resetFile, boolean manualPlatformFullUpdate, Context context)
    {
        super(fileName, fileManager, resetFile, manualPlatformFullUpdate, context);
    }

    @Override
    protected PlatformFullIterator getPlatformFullIterator()
    {
        return new UriPlatformFullIterator(_context, _uri, PROJECTION)
        {
            private int colId;
            private int colTitle;
            private int colIsMusic;
            private int colArtist;
            private int colAlbum;

            @Override
            protected void init(Cursor cursor)
            {
                colId      = cursor.getColumnIndex(PROJECTION[0]);
                colTitle   = cursor.getColumnIndex(PROJECTION[1]);
                colIsMusic = cursor.getColumnIndex(PROJECTION[2]);
                colArtist  = cursor.getColumnIndex(PROJECTION[3]);
                colAlbum   = cursor.getColumnIndex(PROJECTION[4]);
            }

            @Override
            protected WordAction getNext(Cursor cursor)
            {
                while (cursor.moveToNext()) {

                    int contentId      = cursor.getInt(colId);
                    String title = cursor.getString(colTitle);
                    int isMusic = cursor.getInt(colIsMusic);
                    String artist = cursor.getString(colArtist);
                    String album = cursor.getString(colAlbum);

                    if (title != null && title.length() == 0)
                        title = null;
                    if (title == null || isMusic == 0) continue;

                    Dictionary content = new Dictionary();
                    content.put("s", title);

                    if (album != null && title.length() > 0) {
                        content.put("al", album);
                    }

                    if (artist != null && artist.length() > 0) {
                        content.put("ar", artist);
                    }

//                    content.put("g", music.getGenre());
//                    content.put("r", music.getReleaseYear());

                    return new WordAction(new Word(contentId, content), WordAction.ADD);
                }
                return null;
            }
        };
    }
}
