package com.nuance.dragon.toolkit.grammar.content;

import android.content.Context;
import android.database.ContentObserver;
import android.database.Cursor;
import android.net.Uri;
import android.os.Handler;

import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;

/**
 * Base class for content managers that get their content from a URI.
 */
public abstract class UriContentManager extends ContentManager
{
    /**
     *  Content iterator backed by the URI's {@link Cursor}. It is alright for these
     *  methods to be (lightly) blocking, since they will be accessed from the
     *  content manager's background thread.
     */
    protected abstract static class UriPlatformFullIterator implements PlatformFullIterator
    {
        private Cursor _cursor;
        private WordAction _next = null;
        private final int _size;

        // TODO What are these?
        protected UriPlatformFullIterator(Context context, Uri uri, String[] projection)
        {
            try {
                _cursor = context.getContentResolver().query(
                        uri,
                        projection,
                        null, null, null);
            } catch (Exception ex)
            {
                Logger.error(this, "Error getting cursor for URI " + uri.toString());
            }
            _size = getCursorSize();
        }

        protected UriPlatformFullIterator(Context context, Uri uri, String[] projection, String selection, String[] selectionArgs, String sortOrder)
        {
            try {
                _cursor = context.getContentResolver().query(
                        uri,
                        projection,
                        selection, selectionArgs, sortOrder);
            } catch (Exception ex)
            {
                Logger.error(this, "Error getting cursor for URI " + uri.toString());
            }
            _size = getCursorSize();
        }

        // used by unittest
        protected UriPlatformFullIterator(Cursor cursor)
        {
            _cursor = cursor;
            _size = getCursorSize();
        }

        private int getCursorSize()
        {
            int size;
            if (_cursor != null)
            {
                init(_cursor);
                size = _cursor.getCount();
                if (_cursor.moveToFirst())
                {
                    _cursor.moveToPrevious();
                }
                else
                {
                    _cursor.close();
                    _cursor = null;
                }
            } else
            {
                size = 0;
            }
            return size;
        }

        @Override
        public final boolean hasNext()
        {
            if (_next != null)
                return true;

            _next = _cursor == null ? null : getNext(_cursor);
            if (_next == null && _cursor != null) {
                _cursor.close();
                _cursor = null;
            }
            return _next == null ? false : true;
        }

        @Override
        public final WordAction next()
        {
            WordAction next;
            if (_next != null)
            {
                next = _next;
                _next = null;
                return next;
            }

            next = (_cursor == null ? null : getNext(_cursor));
            if (_next == null && _cursor != null) {
                _cursor.close();
                _cursor = null;
            }
            return next;
        }

        @Override
        public final void remove()
        {
            // not allowed
        }

        @Override
        public final int getSize()
        {
            return _size;
        }

        /**
         * Called once as soon as the cursor is obtained. You might initialize your
         * column indices here.
         * @param cursor
         */
        protected abstract void init(Cursor cursor);

        /**
         * @param cursor The current cursor
         * @return Must return the next word action, or null if there are no more.
         */
        protected abstract WordAction getNext(Cursor cursor);
    }

    protected final Uri _uri;
    private ContentObserver _observer;

    /**
     * @param fileName The name of the file to store the full content list cache
     * @param fileManager The file manager used to store the content list cache
     * @param uri The URI to read content from
     * @param resetFile True if the cache file should be reset
     * @param manualPlatformFullUpdate True if the auto content sync-up with
     * platform is not required. Therefore, {@link #forceRefresh()} needs to be
     * triggered manually to force a sync-up
     * @param context The app context
     */
    protected UriContentManager(String fileName, FileManager fileManager, Uri uri, boolean resetFile, boolean manualPlatformFullUpdate, Context context)
    {
        super(fileName, fileManager, manualPlatformFullUpdate, context);
        _uri = uri;
        initialize(resetFile);
    }

    @Override
    protected void watchForContentUpdates(Context context)
    {
        _observer = new ContentObserver(new Handler()) {
            @Override
            public boolean deliverSelfNotifications() {
                return false;
            }

            @Override
            public void onChange(boolean selfChange) {
                if (!selfChange) {
                    onContentUpdate();
                }
            }
        };

        context.getContentResolver().registerContentObserver(_uri, true, _observer);
    }

    @Override
    protected void stopWatchingForContentUpdates(Context context)
    {
        if (_observer != null) {
            context.getContentResolver().unregisterContentObserver(_observer);
            _observer = null;
        }
    }
}
