package com.nuance.dragon.toolkit.grammar.content;

import java.util.ArrayList;
import java.util.HashSet;

import android.content.Context;
import android.database.ContentObserver;
import android.database.Cursor;
import android.net.Uri;
import android.os.Handler;
import android.provider.ContactsContract;
import android.provider.ContactsContract.CommonDataKinds;
import android.provider.ContactsContract.Data;

import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;

/**
 * A {@link ContentManager} that provides the full list of contacts in the
 * device's phonebook.
 * Compared to {@link ContactManager}, this manager generates additional name
 * variants. This is useful to support partial name recognition when using
 * a recognition back-end (ex: Vocon) that doesn't to it natively.
 */
public class VariantContactManager extends ContentManager
{
    /**
     * The name variants supported by Android.
     */
    public static final class NameVariants
    {
        /** The display name (e.g. default / full name). */
        public static final int DISPLAY_NAME = 0x40000000;

        /** The given name (e.g. first name). */
        public static final int GIVEN_NAME = 0x80000000;

        /** The family name (e.g. last name). */
        public static final int FAMILY_NAME = 0xC0000000;
    }

    private class NameEntry
    {
        final String surfaceForm;
        final String spokenForm;
        final int variant;

        NameEntry(String surfaceForm, String spokenForm, int variant)
        {
            this.surfaceForm = surfaceForm;
            this.spokenForm = spokenForm;
            this.variant = variant;
        }

        int getUniqueId()
        {
            short checksum = 0;
            if (spokenForm != null) {
               for (int idx = 0; idx < spokenForm.length(); idx++)
                   checksum += (idx + 1) * spokenForm.charAt(idx);
            }
            if (surfaceForm != null) {
               for (int idx = 0; idx < surfaceForm.length(); idx++)
                   checksum += (idx + 1) * surfaceForm.charAt(idx);
            }
            return variant | ((int)checksum);
        }
    }

    private final Context _context;
    protected final Uri _uri;
    private ContentObserver _observer;
    private Cursor _cursor;
    private WordAction _next;

    /**
     * Like {@link #VariantContactManager(String, FileManager, boolean, Context)} using auto
     * content sync-up without resetting the cache file.
     */
    public VariantContactManager(String fileName, FileManager fileManager, Context context)
    {
        this(fileName, fileManager, false, false, context);
    }

    /**
     * Like {@link #VariantContactManager(String, FileManager, boolean, Context)} without
     * resetting the cache file.
     */
    public VariantContactManager(String fileName, FileManager fileManager, boolean manualPlatformFullUpdate, Context context)
    {
        this(fileName, fileManager, false, manualPlatformFullUpdate, context);
    }

    /**
     * @param fileName The name of the file to store the full content list cache
     * @param fileManager The file manager used to store the content list cache
     * @param resetFile True if the cache file should be reset
     * @param manualPlatformFullUpdate True if the auto content sync-up with
     * platform is not required. Therefore, {@link #forceRefresh()} needs to be
     * triggered manually to force a sync-up
     * @param context The app context
     */
    public VariantContactManager(String fileName, FileManager fileManager, boolean resetFile, boolean manualPlatformFullUpdate, Context context)
    {
        super(fileName, fileManager, manualPlatformFullUpdate, context);
        _context = context.getApplicationContext();
        _uri = ContactsContract.Data.CONTENT_URI;
        initialize(resetFile);
    }

    /**
     * Convenience method to get the name variant from a word ID.
     * @param id The word ID of interest returned by the recognizer (Elvis phraseID or VoCon userID).
     * @return The name variant from {@link NameVariants}.
     */
    public static int getNameVariantFromId(int id)
    {
        return id & 0xC0000000;
    }

    @Override
    protected void watchForContentUpdates(Context context)
    {
        _observer = new ContentObserver(new Handler()) {
            @Override
            public boolean deliverSelfNotifications() {
                return false;
            }

            @Override
            public void onChange(boolean selfChange) {
                if (!selfChange) {
                    onContentUpdate();
                }
            }
        };

        context.getContentResolver().registerContentObserver(_uri, true, _observer);
    }

    @Override
    protected void stopWatchingForContentUpdates(Context context)
    {
        if (_observer != null) {
            context.getContentResolver().unregisterContentObserver(_observer);
            _observer = null;
        }
    }

    @Override
    protected PlatformFullIterator getPlatformFullIterator()
    {
        final ArrayList<NameEntry> variants = new ArrayList<NameEntry>(3);
        final HashSet<Integer> hashes = new HashSet<Integer>();
        _cursor = null;
        _next = null;

        final String colVisible = ContactsContract.Contacts.IN_VISIBLE_GROUP;
        final String colLookupKey = ContactsContract.Contacts.LOOKUP_KEY;
        final String colDisplayName = ContactsContract.Contacts.DISPLAY_NAME;
        final String colGivenName = ContactsContract.CommonDataKinds.StructuredName.GIVEN_NAME;
        final String colFamilyName = ContactsContract.CommonDataKinds.StructuredName.FAMILY_NAME;
        final String colPhoneticGivenName = ContactsContract.CommonDataKinds.StructuredName.PHONETIC_GIVEN_NAME;
        final String colPhoneticMiddleName = ContactsContract.CommonDataKinds.StructuredName.PHONETIC_MIDDLE_NAME;
        final String colPhoneticFamilyName = ContactsContract.CommonDataKinds.StructuredName.PHONETIC_FAMILY_NAME;

        try {
            _cursor = _context.getContentResolver().query(_uri,
                                                          new String[] { colVisible, colLookupKey, colDisplayName, colGivenName, colFamilyName, colPhoneticGivenName, colPhoneticMiddleName, colPhoneticFamilyName},
                                                          Data.MIMETYPE + "='" + CommonDataKinds.StructuredName.CONTENT_ITEM_TYPE + "'",
                                                          null, null);
        } catch (Exception ex)
        {
            Logger.error(this, "Error getting cursor for URI " + _uri.toString());
        }

        return new PlatformFullIterator()
        {
            @Override
            public final boolean hasNext()
            {
                if (_next != null)
                    return true;

                _next = _cursor == null ? null : getNext(_cursor);
                if (_next == null && _cursor != null) {
                    _cursor.close();
                    _cursor = null;
                }
                return _next == null ? false : true;
            }

            @Override
            public final WordAction next()
            {
                WordAction next;
                if (_next != null)
                {
                    next = _next;
                    _next = null;
                    return next;
                }

                next = (_cursor == null ? null : getNext(_cursor));
                if (_next == null && _cursor != null) {
                    _cursor.close();
                    _cursor = null;
                }
                return next;
            }

            private WordAction getNext(Cursor cursor)
            {
                // Try reading from cache
                NameEntry cached = getNextCachedEntry();
                if (cached != null)
                    return new WordAction(new Word(cached.getUniqueId(), cached.surfaceForm, cached.spokenForm), true);

                // Then, read from SQL DB
                try {
                    while (cursor.moveToNext())
                    {
                        if (cursor.getInt(_cursor.getColumnIndex(colVisible)) == 0 ||
                            cursor.getString(_cursor.getColumnIndex(colLookupKey)) == null)
                            continue;

                        String displayName = cursor.getString(cursor.getColumnIndex(colDisplayName));
                        String givenName = cursor.getString(cursor.getColumnIndex(colGivenName));
                        String familyName = cursor.getString(cursor.getColumnIndex(colFamilyName));

                        // Get pronunciations
                        String phoneticGivenName = cursor.getString(cursor.getColumnIndex(colPhoneticGivenName));
                        String phoneticMiddleName = cursor.getString(cursor.getColumnIndex(colPhoneticMiddleName));
                        String phoneticFamilyName = cursor.getString(cursor.getColumnIndex(colPhoneticFamilyName));
                        String phoneticDisplayName = getPhoneticDisplayName(phoneticGivenName, phoneticMiddleName, phoneticFamilyName);

                        // Add name variants
                        variants.clear();
                        if (displayName != null && displayName.length() > 0 && !hashes.contains(displayName.hashCode()))
                        {
                            hashes.add(displayName.hashCode());
                            variants.add(new NameEntry(displayName,
                                                       (phoneticDisplayName != null && phoneticDisplayName.length() > 0) ? phoneticDisplayName : null,
                                                       NameVariants.DISPLAY_NAME));
                        }
                        if (givenName != null && givenName.length() > 0 && !hashes.contains(givenName.hashCode()))
                        {
                            hashes.add(givenName.hashCode());
                            variants.add(new NameEntry(givenName,
                                                       (phoneticGivenName != null && phoneticGivenName.length() > 0) ? phoneticGivenName : null,
                                                       NameVariants.GIVEN_NAME));
                        }
                        if (familyName != null && familyName.length() > 0 && !hashes.contains(familyName.hashCode()))
                        {
                            hashes.add(familyName.hashCode());
                            variants.add(new NameEntry(familyName,
                                                       (phoneticFamilyName != null && phoneticFamilyName.length() > 0) ? phoneticFamilyName : null,
                                                       NameVariants.FAMILY_NAME));
                        }

                        cached = getNextCachedEntry();
                        if (cached != null)
                            return new WordAction(new Word(cached.getUniqueId(), cached.surfaceForm, cached.spokenForm), true);
                    }
                }
                catch (RuntimeException e) {
                }
                return null;
            }

            @Override
            public void remove()
            {
                // Not allowed
            }

            @Override
            public int getSize()
            {
                return -1;
            }

            private NameEntry getNextCachedEntry()
            {
                if (variants.isEmpty())
                    return null;
                return variants.remove(0);
            }

            private String getPhoneticDisplayName(String phoneticGivenName, String phoneticMiddleName, String phoneticFamilyName)
            {
                String phoneticDisplayName = "";

                if (phoneticFamilyName != null && phoneticFamilyName.length() != 0) {
                    phoneticDisplayName = phoneticFamilyName;
                }
                if (phoneticMiddleName != null && phoneticMiddleName.length() != 0) {
                    phoneticDisplayName += phoneticMiddleName;
                }
                if (phoneticGivenName != null && phoneticGivenName.length() != 0) {
                    phoneticDisplayName += phoneticGivenName;
                }

                return (phoneticDisplayName.length() == 0 ? null : phoneticDisplayName);
            }
        };
    }
}
