package com.nuance.dragon.toolkit.hybrid;

import java.util.ArrayList;
import java.util.List;

import android.util.Pair;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.SpeechDetectionListener;
import com.nuance.dragon.toolkit.audio.pipes.DuplicatorPipe;
import com.nuance.dragon.toolkit.audio.pipes.SpeexEncoderPipe;
import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.DataParam;
import com.nuance.dragon.toolkit.cloudservices.Param;
import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognitionError;
import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognitionResult;
import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognizer;
import com.nuance.dragon.toolkit.cloudservices.recognizer.RecogSpec;
import com.nuance.dragon.toolkit.elvis.ElvisError;
import com.nuance.dragon.toolkit.elvis.ElvisRecognizer;
import com.nuance.dragon.toolkit.elvis.ElvisResult;
import com.nuance.dragon.toolkit.elvis.EndpointingParam;
import com.nuance.dragon.toolkit.grammar.Scenario;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.AsyncTaskHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.recognition.InterpretException;
import com.nuance.dragon.toolkit.recognition.InterpretedRecognition;
import com.nuance.dragon.toolkit.recognition.InterpretedRecognitionCombiner;
import com.nuance.dragon.toolkit.recognition.LocalRecognitionResult;
import com.nuance.dragon.toolkit.recognition.RecognitionInterpreter;

/**
 * Implementation of {@link HybridRecognizer}.
 * <p>
 * Don't create this directly. Use {@link HybridRecognizer#createHybridRecognizer(ElvisRecognizer,CloudServices,RecognitionInterpreter,RecognitionInterpreter,InterpretedRecognitionCombiner)}.
 */
final class ElvisHybridRecognizerImpl extends HybridRecognizer
{
    private static final class HybridRecognition
    {
        private CloudRecognizer _cloudRecognizer;
        private InterpretedRecognition _cloudRecognitionResult;
        private boolean _elvisRecognizerActive;
        private InterpretedRecognition _elvisResult;
        private ElvisError _elvisError;
        private AsyncTaskHandler.TaskRunnable<Pair<HybridRecognitionResult, HybridRecognitionError>> _reportingTask;
        private List<String> _slotsToUpdate;
        private boolean _localFxOnly;
    }

    private DuplicatorPipe<AudioChunk> _duplicator;
    private SpeexEncoderPipe _speexEncoderPipe;
    private AudioSource<AudioChunk> _localAudioSource;
    private AudioSource<AudioChunk> _remoteAudioSource;
    private final ElvisRecognizer _elvisRecognizer;
    private final CloudServices _cloudServices;
    private RecognitionInterpreter<LocalRecognitionResult> _elvisInterpreter;
    private RecognitionInterpreter<CloudRecognitionResult> _cloudInterpreter;
    private InterpretedRecognitionCombiner _combiner;
    private final AsyncTaskHandler _workerThreadHandler;
    private HybridRecognition _currentRecognition;

    /**
     * @deprecated Call {@link #ElvisHybridRecognizerImpl(ElvisRecognizer,CloudServices)} instead.
     * Creates a new HybridRecognizer with references to a local listener to
     * receive Elvis results and a remote listener to receive cloud recognition
     * results.
     * @param elvisRecognizer A valid Elvis recognizer instance. Must not be null
     * @param cloudServices A valid Cloud Services instance. Must not be null
     * @param elvisInterpreter The Elvis recognition interpreter to use. Must not be null
     * @param cloudInterpreter The cloud recognition interpreter to use. Must not be null
     * @param hybridCombiner Used to combine the interpreted recognitions. Must not be null.
     * TODO: should we pass this in startRecognition() instead?
     */
    @Deprecated
    public ElvisHybridRecognizerImpl(ElvisRecognizer elvisRecognizer,
                                     CloudServices cloudServices,
                                     RecognitionInterpreter<LocalRecognitionResult> elvisInterpreter,
                                     RecognitionInterpreter<CloudRecognitionResult> cloudInterpreter,
                                     InterpretedRecognitionCombiner hybridCombiner)
    {
        Checker.checkArgForNull("elvisRecognizer", elvisRecognizer);
        Checker.checkArgForNull("cloudServices", cloudServices);
        Checker.checkArgForNull("elvisInterpreter", elvisInterpreter);
        Checker.checkArgForNull("cloudInterpreter", cloudInterpreter);
        Checker.checkArgForNull("hybridCombiner", hybridCombiner);

        _elvisRecognizer = elvisRecognizer;
        _cloudServices = cloudServices;
        _elvisInterpreter = elvisInterpreter;
        _cloudInterpreter = cloudInterpreter;
        _combiner = hybridCombiner;
        _workerThreadHandler = new AsyncTaskHandler();
    }

    /**
     * Creates a new HybridRecognizer with references to a local listener to
     * receive Elvis results and a remote listener to receive cloud recognition
     * results.
     * @param elvisRecognizer A valid Elvis recognizer instance. Must not be null
     * @param cloudServices A valid Cloud Services instance. Must not be null
     */
    public ElvisHybridRecognizerImpl(ElvisRecognizer elvisRecognizer,
                                     CloudServices cloudServices)
    {
        Checker.checkArgForNull("elvisRecognizer", elvisRecognizer);
        Checker.checkArgForNull("cloudServices", cloudServices);

        _elvisRecognizer = elvisRecognizer;
        _cloudServices = cloudServices;
        _workerThreadHandler = new AsyncTaskHandler();
    }

    private void reportError(HybridRecognition recognition,
                              HybridRecognitionError error,
                              Listener listener)
    {
        if (recognition == _currentRecognition)
            _currentRecognition = null;

        listener.onError(error);
    }

    private void reportResult(final HybridRecognition recognition,
                              final InterpretedRecognition elvisResult,
                              final InterpretedRecognition cloudResult,
                              final InterpretedRecognitionCombiner combiner,
                              final Listener listener)
    {
        if (recognition != _currentRecognition)
            return;

        // Interpreting the cloud result can be slow, so do it asynchronously
        recognition._reportingTask = new AsyncTaskHandler.TaskRunnable<Pair<HybridRecognitionResult, HybridRecognitionError>>()
        {
            @Override
            public Pair<HybridRecognitionResult, HybridRecognitionError> run()
            {
                try {
                    HybridRecognitionResult hybridResult = new HybridRecognitionResult(elvisResult, cloudResult, combiner, recognition._slotsToUpdate);
                    return new Pair<HybridRecognitionResult, HybridRecognitionError>(hybridResult, null);
                }
                catch (InterpretException e) {
                    HybridRecognitionError error = new HybridRecognitionError();
                    error.setCloudError(e);
                    error.setElvisError(e);
                    return new Pair<HybridRecognitionResult, HybridRecognitionError>(null, error);
                }
            }

            @Override
            public void onPostRun(Pair<HybridRecognitionResult, HybridRecognitionError> result)
            {
                if (recognition != _currentRecognition)
                    return;

                _currentRecognition = null;

                if (result.first != null)
                    listener.onResult(result.first);
                else
                    listener.onError(result.second);
            }
        };
        _workerThreadHandler.post(_currentRecognition._reportingTask);
    }

    @Override
    public void startRecognition(AudioSource<AudioChunk> audioSource,
                                 Scenario scenario,
                                 final Listener listener)
    {
        startRecognition(audioSource, scenario, _elvisInterpreter, _cloudInterpreter, _combiner, listener);
    }

    @Override
    public void startRecognition(AudioSource<AudioChunk> audioSource,
                                 Scenario scenario,
                                 RecognitionInterpreter<LocalRecognitionResult> localArbitrationInterpreter,
                                 Listener listener)
    {
        final String msg = "Method is not supported. Use VoconHybridRecognizer instead.";
        Logger.error(this, msg);
        throw new UnsupportedOperationException(msg);
    }

    @Override
    public void startRecognition(AudioSource<AudioChunk> audioSource,
                                 Scenario scenario,
                                 final RecognitionInterpreter<LocalRecognitionResult> localInterpreter,
                                 final RecognitionInterpreter<CloudRecognitionResult> cloudInterpreter,
                                 final InterpretedRecognitionCombiner hybridCombiner,
                                 final Listener listener)
    {
        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForNull("scenario", scenario);
        Checker.checkArgForNull("localInterpreter", localInterpreter);
        Checker.checkArgForNull("cloudInterpreter", cloudInterpreter);
        Checker.checkArgForNull("hybridCombiner", hybridCombiner);
        Checker.checkArgForNull("listener", listener);

        RecogSpec cloudRecogSpec = scenario.getCloudRecogSpec();
        List<String> elvisConstraints = scenario.getElvisConstraints();
        Checker.checkArgsForNull("Cloud recognition spec and Elvis constraints ", cloudRecogSpec, elvisConstraints);

        // Cancel any recognition in progress
        cancel();

        final HybridRecognition recog = new HybridRecognition();
        recog._localFxOnly = elvisConstraints == null || elvisConstraints.isEmpty();
        _currentRecognition = recog;

        // Set-up audio chain
        setupAudioChain(audioSource);

        // Start remote recognition
        if (cloudRecogSpec != null)
        {
            recog._cloudRecognizer = new CloudRecognizer(_cloudServices);
            recog._cloudRecognizer.startRecognition(cloudRecogSpec, _remoteAudioSource, new CloudRecognizer.Listener()
            {
                @Override
                public void onResult(CloudRecognitionResult result)
                {
                    Logger.debug(ElvisHybridRecognizerImpl.this, "Cloud result received");
                    if (recog._cloudRecognizer == null) {
                        Logger.error(ElvisHybridRecognizerImpl.this, "Got unexpected Cloud result");
                        return;
                    }

                    try {
                        recog._cloudRecognitionResult = cloudInterpreter.getInterpretedResult(result);
                        recog._slotsToUpdate = cloudInterpreter.getUpdateRequiredList(result);
                    } catch (InterpretException ex)
                    {
                        onError(null, ex);
                        return;
                    }

                    recog._cloudRecognizer = null;
                    if (!recog._elvisRecognizerActive || recog._localFxOnly) {
                        cleanupAudioChain();
                        reportResult(recog, recog._elvisResult, recog._cloudRecognitionResult, hybridCombiner, listener);
                    }
                }

                @Override
                public void onError(CloudRecognitionError error)
                {
                    Logger.debug(ElvisHybridRecognizerImpl.this, "Cloud error received");
                    onError(error, null);
                }

                private void onError(CloudRecognitionError cloudError, InterpretException otherError)
                {
                    if (recog._cloudRecognizer == null) {
                        Logger.error(ElvisHybridRecognizerImpl.this, "Got unexpected Cloud error");
                        return;
                    }

                    recog._cloudRecognizer = null;
                    if (!recog._elvisRecognizerActive || recog._localFxOnly) {
                        cleanupAudioChain();

                        if (recog._elvisResult != null) {
                            // A cloud error occurred, but we can still report the elvis result
                            reportResult(recog, recog._elvisResult, null, hybridCombiner, listener);
                        } else
                        {
                            HybridRecognitionError error = new HybridRecognitionError();
                            error.setElvisError(recog._elvisError);
                            error.setCloudError(cloudError);
                            error.setCloudError(otherError);
                            reportError(recog, error, listener);
                        }
                    }
                }

                @Override
                public void onTransactionIdGenerated(String transactionId)
                {
                    if ( recog._localFxOnly == false )
                    {
                        _elvisRecognizer.logMiscInfo(ElvisRecognizer.TAG_NVC_SESSION_ID, transactionId);
                    }
                }
            });
        }

        // If somehow the recognition already ended by now, don't bother with
        // Elvis
        if (recog != _currentRecognition)
            return;

        // Start local recognition
        if (recog._localFxOnly)
        {
            // Just use Elvis for the endpointing
            Logger.warn(this, "Elvis is only used for end-pointing in this recognition.");
            recog._elvisRecognizerActive = true;
            EndpointingParam.Builder builder = new EndpointingParam.Builder();
            _elvisRecognizer.startSpeechDetection(_localAudioSource,
                                                  builder.build(),
                                                  new SpeechDetectionListener()
            {
                @Override
                public void onStartOfSpeech()
                {
                    listener.onStartOfSpeech();
                }

                @Override
                public void onEndOfSpeech()
                {
                    if (!recog._elvisRecognizerActive) {
                        Logger.error(ElvisHybridRecognizerImpl.this, "Got unexpected Elvis end-of-speech");
                        return;
                    }

                    recog._elvisRecognizerActive = false;
                    _elvisRecognizer.stopListening();
                    cleanupAudioChain();
                    listener.onEndOfSpeech();

                    if (recog._cloudRecognizer != null)
                    {
                        recog._cloudRecognizer.processResult();
                    }
                }
            },
            null);
        } else
        {
            recog._elvisRecognizerActive = true;
            _elvisRecognizer.startRecognition(_localAudioSource, elvisConstraints,
                new SpeechDetectionListener() {
                    @Override
                    public void onStartOfSpeech()
                    {
                        listener.onStartOfSpeech();
                    }

                    @Override
                    public void onEndOfSpeech()
                    {
                        _elvisRecognizer.stopListening();
                        cleanupAudioChain();
                        listener.onEndOfSpeech();
                    }
                },
                null,
                new ElvisRecognizer.ResultListener()
                {
                    @Override
                    public void onResult(ElvisResult result) {
                        Logger.debug(ElvisHybridRecognizerImpl.this, "Elvis result received");
                        if (!recog._elvisRecognizerActive) {
                            Logger.error(ElvisHybridRecognizerImpl.this, "Got unexpected Elvis result");
                            return;
                        }

                        try {
                            recog._elvisResult = localInterpreter.getInterpretedResult(result);
                        }
                        catch (InterpretException e) {
                            onError(null, e);
                            return;
                        }

                        if (recog._cloudRecognizer != null)
                        {
                            // Send params to the cloud if necessary
                            List<DataParam> params = new ArrayList<DataParam>(3);
                            boolean keepWaiting = localInterpreter.processForCloud(result, params);
                            for (Param p : params)
                            {
                                recog._cloudRecognizer.sendParam(p);
                            }
                            recog._cloudRecognizer.processResult();

                            if (!keepWaiting)
                            {
                                // Let the command finish, but don't wait for it anymore.
                                recog._cloudRecognizer = null;
                            }
                        }
                        recog._elvisRecognizerActive = false;

                        if (recog._cloudRecognizer == null) {
                            cleanupAudioChain();
                            reportResult(recog, recog._elvisResult, recog._cloudRecognitionResult, hybridCombiner, listener);
                        }
                    }

                    @Override
                    public void onError(ElvisError error) {
                        Logger.debug(ElvisHybridRecognizerImpl.this, "Elvis error received");
                        onError(error, null);
                    }

                    private void onError(ElvisError error, InterpretException otherError) {
                        if (!recog._elvisRecognizerActive) {
                            Logger.error(ElvisHybridRecognizerImpl.this, "Got unexpected Elvis error");
                            return;
                        }

                        recog._elvisRecognizerActive = false;
                        cleanupAudioChain();
                        if (recog._cloudRecognizer == null) {
                            HybridRecognitionError hybridError = new HybridRecognitionError();
                            hybridError.setElvisError(error);
                            hybridError.setElvisError(otherError);

                            reportError(recog, hybridError, listener);
                        }
                        else {
                            recog._elvisError = error; // Save it for later TODO
                            recog._cloudRecognizer.processResult();
                        }
                    }
                });
        }
    }

    @Override
    public void stopRecognition()
    {
        if (_currentRecognition != null)
        {
            _elvisRecognizer.stopListening();
            cleanupAudioChain();

            if (_currentRecognition._localFxOnly)
            {
                _currentRecognition._elvisRecognizerActive = false;

                // In the case of using the engine FX only, we complete the command here
                CloudRecognizer cloudRecognizer = _currentRecognition._cloudRecognizer;
                if (cloudRecognizer != null)
                {
                    cloudRecognizer.processResult();
                }
            }
        }
    }

    @Override
    public void cancel()
    {
        if (_currentRecognition != null)
        {
            // When canceling, make sure that we don't report any results
            _currentRecognition._elvisResult = null;
            _currentRecognition._cloudRecognitionResult = null;

            CloudRecognizer cloudRecognizer = _currentRecognition._cloudRecognizer;
            _currentRecognition = null;

            _elvisRecognizer.cancelRecognition();
            if (cloudRecognizer != null)
            {
                cloudRecognizer.cancel();
            }
        }

        // The error callbacks should clean things up
    }

    private void setupAudioChain(AudioSource<AudioChunk> audioSource)
    {
        cleanupAudioChain();

        // Configure local audio source
        _duplicator = new DuplicatorPipe<AudioChunk>();
        _duplicator.connectAudioSource(audioSource);
        _localAudioSource = _duplicator;

        // Configure remote audio source
        _speexEncoderPipe = new SpeexEncoderPipe();
        _speexEncoderPipe.connectAudioSource(_duplicator);
        _remoteAudioSource = _speexEncoderPipe;
    }

    private void cleanupAudioChain()
    {
        if (_duplicator != null) {
            _duplicator.disconnectAudioSource();
            _duplicator = null;
        }

        if (_speexEncoderPipe != null) {
            _speexEncoderPipe.disconnectAudioSource();
            _speexEncoderPipe.release();
            _speexEncoderPipe = null;
        }
    }
}
