package com.nuance.dragon.toolkit.hybrid;

import org.json.JSONException;

import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognitionError;
import com.nuance.dragon.toolkit.elvis.ElvisError;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.recognition.InterpretException;
import com.nuance.dragon.toolkit.vocon.VoconError;

/**
 * A hybrid recognition error.
 */
public final class HybridRecognitionError implements JSONCompliant
{
    private static final String KEY_ELVIS_ERROR = "elvis";
    private static final String KEY_ELVIS_EX = "elvis_ex";
    private static final String KEY_VOCON_ERROR = "vocon";
    private static final String KEY_VOCON_EX = "vocon_ex";
    private static final String KEY_CLOUD_ERROR = "cloud";
    private static final String KEY_CLOUD_EX = "cloud_ex";

    private ElvisError _elvisError;
    private VoconError _voconError;
    private CloudRecognitionError _cloudError;
    private InterpretException _elvisException;
    private InterpretException _voconException;
    private InterpretException _cloudException;

    void setElvisError(ElvisError error)
    {
        _elvisError = error;
    }

    void setElvisError(InterpretException exception)
    {
        _elvisException = exception;
    }

    void setVoconError(VoconError error)
    {
        _voconError = error;
    }

    void setVoconError(InterpretException exception)
    {
        _voconException = exception;
    }

    void setCloudError(CloudRecognitionError error)
    {
        _cloudError = error;
    }

    void setCloudError(InterpretException exception)
    {
        _cloudException = exception;
    }

    /**
     * Get the Elvis error that occurred, if any
     * @return May return null
     */
    public ElvisError getElvisError()
    {
        return _elvisError;
    }

    /**
     * Get the Vocon error that occurred, if any
     * @return May return null
     */
    public VoconError getVoconError()
    {
        return _voconError;
    }

    /**
     * Get the error that occurred while interpreting the Elvis result, if any
     * @return May return null
     */
    public String getElvisInterpretationError()
    {
        return _elvisException == null ? null : _elvisException.getMessage();
    }

    /**
     * Get the error that occurred while interpreting the Vocon result, if any
     * @return May return null
     */
    public String getVoconInterpretationError()
    {
        return _voconException == null ? null : _voconException.getMessage();
    }

    /**
     * Get the Cloud error that occurred, if any
     * @return May return null
     */
    public CloudRecognitionError getCloudError()
    {
        return _cloudError;
    }

    /**
     * Get the error that occurred while interpreting the cloud result, if any
     * @return May return null
     */
    public String getCloudInterpretationError()
    {
        return _cloudException == null ? null : _cloudException.getMessage();
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject json = new JSONObject();
        json.tryPut(KEY_ELVIS_ERROR, _elvisError);
        json.tryPut(KEY_VOCON_ERROR, _voconError);
        json.tryPut(KEY_CLOUD_ERROR, _cloudError);

        if (_elvisException != null)
            json.tryPut(KEY_ELVIS_EX, _elvisException.getMessage());
        if (_voconException != null)
            json.tryPut(KEY_ELVIS_EX, _voconException.getMessage());
        if (_cloudException != null)
            json.tryPut(KEY_CLOUD_EX, _cloudException.getMessage());
        return json;
    }

    public static HybridRecognitionError createFromJSON(org.json.JSONObject json) throws JSONException
    {
        org.json.JSONObject elvis = json.optJSONObject(KEY_ELVIS_ERROR);
        org.json.JSONObject vocon = json.optJSONObject(KEY_VOCON_ERROR);
        org.json.JSONObject cloud = json.optJSONObject(KEY_CLOUD_ERROR);

        String elvisEx = json.optString(KEY_ELVIS_EX);
        String voconEx = json.optString(KEY_VOCON_EX);
        String cloudEx = json.optString(KEY_CLOUD_EX);

        ElvisError elvisError = elvis == null ? null : ElvisError.createFromJSON(elvis);
        VoconError voconError = vocon == null ? null : VoconError.createFromJSON(vocon);
        CloudRecognitionError cloudError = cloud == null ? null : CloudRecognitionError.createFromJSON(elvis);

        HybridRecognitionError error = new HybridRecognitionError();
        error.setElvisError(elvisError);
        error.setVoconError(voconError);
        error.setCloudError(cloudError);

        if (elvisEx != null)
            error.setElvisError(new InterpretException(elvisEx));
        if (voconEx != null)
            error.setVoconError(new InterpretException(voconEx));
        if (cloudEx != null)
            error.setCloudError(new InterpretException(cloudEx));

        return error;
    }
}
