package com.nuance.dragon.toolkit.hybrid;

import java.util.ArrayList;
import java.util.List;

import org.json.JSONException;

import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognitionError;
import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognitionResult;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.recognition.InterpretException;
import com.nuance.dragon.toolkit.recognition.InterpretedRecognition;
import com.nuance.dragon.toolkit.recognition.InterpretedRecognitionCombiner;
import com.nuance.dragon.toolkit.recognition.LocalRecognitionResult;

/**
 * A hybrid recognition result.
 */
public class HybridRecognitionResult implements JSONCompliant
{
    private static final String KEY_LOCAL_RESULT = "local";
    private static final String KEY_REMOTE_RESULT = "remote";
    private static final String KEY_RESULT = "result";
    private static final String KEY_SLOTS_TO_UPDATE = "slots_to_update";

    private final InterpretedRecognition _localResult;
    private final InterpretedRecognition _remoteResult;
    private final InterpretedRecognition _result;
    private final ArrayList<String>      _slotsToUpdate;

    HybridRecognitionResult(InterpretedRecognition localResult,
                            InterpretedRecognition remoteResult,
                            InterpretedRecognitionCombiner combiner,
                            List<String> slotsToUpdate) throws InterpretException
    {
        _localResult = localResult;
        _remoteResult = remoteResult;
        if (combiner != null) {
            _result = combiner.combine(localResult, remoteResult);
        }
        else {
            _result = new InterpretedRecognition();
        }
        _slotsToUpdate = new ArrayList<String>();
        if (slotsToUpdate != null) {
            _slotsToUpdate.addAll(slotsToUpdate);
        }
    }

    private HybridRecognitionResult(InterpretedRecognition localResult,
                                    InterpretedRecognition remoteResult,
                                    InterpretedRecognition result,
                                    List<String> slotsToUpdate)
    {
        _localResult = localResult;
        _remoteResult = remoteResult;
        _result = result;
        _slotsToUpdate = new ArrayList<String>();
        if (slotsToUpdate != null) {
            _slotsToUpdate.addAll(slotsToUpdate);
        }
    }

    /**
     * Get the hybrid recognition result.
     * @return The hybrid recognition result, based on the local and remote
     * results.
     */
    public InterpretedRecognition getResult()
    {
        return _result;
    }

    /**
     * Get the original local result.
     * @return The original local (Elvis or Vocon) result.
     */
    public InterpretedRecognition getLocalResult()
    {
        return _localResult;
    }

    /**
     * Get the original remote result.
     * @return The original remote (cloud) result.
     */
    public InterpretedRecognition getRemoteResult()
    {
        return _remoteResult;
    }

    /**
     * This method can be used to determine if any server word lists are out of sync between the client and the server.
     *
     * @return returns a list of slots that should be updated, or an empty list (not null) if all slots are up to date.
     */
    public List<String> getSlotsToUpdate()
    {
        ArrayList<String> slotsToUpdate = new ArrayList<String>(_slotsToUpdate.size());
        slotsToUpdate.addAll(_slotsToUpdate);
        return slotsToUpdate;
    }

    /**
     * Get the original local result.
     * @return The original local (Elvis or Vocon) raw result, if any.
     */
    public LocalRecognitionResult getRawLocalResult()
    {
        return null;
    }

    /**
     * Get the original remote result.
     * @return The original remote (cloud) raw results, if any.
     */
    public List<CloudRecognitionResult> getRawRemoteResults()
    {
        return null;
    }

    /**
     * Get the eventual remote error.
     * @return The remote (cloud) error, if any.
     */
    public CloudRecognitionError getEventualRemoteError()
    {
        return null;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        o.tryPut(KEY_RESULT, _result);
        if (_localResult != null)
            o.tryPut(KEY_LOCAL_RESULT, _localResult.toJSON());
        if (_remoteResult != null)
            o.tryPut(KEY_REMOTE_RESULT, _remoteResult.toJSON());
        if (_slotsToUpdate.size() > 0) {
            org.json.JSONArray jsonSlots = new org.json.JSONArray();
            for (int i = 0; i < _slotsToUpdate.size(); i++) {
                jsonSlots.put(_slotsToUpdate.get(i));
            }
            o.tryPut(KEY_SLOTS_TO_UPDATE, jsonSlots);
        }
        return o;
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result
                + ((_localResult == null) ? 0 : _localResult.hashCode());
        result = prime * result
                + ((_remoteResult == null) ? 0 : _remoteResult.hashCode());
        result = prime * result + ((_result == null) ? 0 : _result.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        HybridRecognitionResult other = (HybridRecognitionResult) obj;
        if (_localResult == null) {
            if (other._localResult != null)
                return false;
        } else if (!_localResult.equals(other._localResult))
            return false;
        if (_remoteResult == null) {
            if (other._remoteResult != null)
                return false;
        } else if (!_remoteResult.equals(other._remoteResult))
            return false;
        if (_result == null) {
            if (other._result != null)
                return false;
        } else if (!_result.equals(other._result))
            return false;
        if (_slotsToUpdate == null) {
            if (other._slotsToUpdate != null)
                return false;
        } else if (!_slotsToUpdate.equals(other._slotsToUpdate))
            return false;
        return true;
    }

    public static HybridRecognitionResult createFromJSON(org.json.JSONObject json) throws JSONException
    {
        // TODO: decide which of these can be null
        org.json.JSONObject resultObject = json.optJSONObject(KEY_RESULT);
        org.json.JSONObject localObject = json.optJSONObject(KEY_LOCAL_RESULT);
        org.json.JSONObject remoteObject = json.optJSONObject(KEY_REMOTE_RESULT);
        org.json.JSONArray jsonSlots = json.optJSONArray(KEY_SLOTS_TO_UPDATE);
        ArrayList<String> slotsToUpdate = null;
        if (jsonSlots != null) {
            slotsToUpdate = new ArrayList<String>(jsonSlots.length());
            for (int i = 0; i < jsonSlots.length(); i++) {
                slotsToUpdate.add(jsonSlots.getString(i));
            }
        }

        InterpretedRecognition localResult = localObject == null ? null : InterpretedRecognition.createFromJSON(localObject);
        InterpretedRecognition remoteResult = remoteObject == null ? null : InterpretedRecognition.createFromJSON(remoteObject);
        InterpretedRecognition result = resultObject == null ? null : InterpretedRecognition.createFromJSON(resultObject);
        return new HybridRecognitionResult(localResult, remoteResult, result, slotsToUpdate);
    }
}
