package com.nuance.dragon.toolkit.hybrid;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.SpeechDetectionListener;
import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognitionResult;
import com.nuance.dragon.toolkit.elvis.ElvisRecognizer;
import com.nuance.dragon.toolkit.grammar.Scenario;
import com.nuance.dragon.toolkit.recognition.InterpretedRecognitionCombiner;
import com.nuance.dragon.toolkit.recognition.LocalRecognitionResult;
import com.nuance.dragon.toolkit.recognition.RecognitionInterpreter;
import com.nuance.dragon.toolkit.vocon.VoconRecognizer;

/**
 * The Hybrid recognizer is a consolidated entry point for server and local
 * recognition. When recognize is called, the rules are evaluated as to whether
 * they can be locally recognized and whether they require server recognition.
 * Recognition is then started on one or both, the audio source is told to begin
 * recording, and audio state feedback is provided to the delegate. Once the
 * utterance is end-pointed, the audio source is stopped and a timer is started
 * if server recognition was required. When the interpretation of the local
 * result is provided it is logged to the server and it is presented as the
 * result if it is complete. Otherwise either the expiration of the timer or a
 * server result triggers the emission of a result.
 * <p>
 * To create a HybridRecognizer, call {@link #createHybridRecognizer(ElvisRecognizer,CloudServices)}
 * or {@link #createHybridRecognizer(VoconRecognizer,CloudServices)}.
 */
public abstract class HybridRecognizer
{
    /**
     * Listens for results or errors from a hybrid recognition.
     */
    public interface Listener extends SpeechDetectionListener
    {
        /**
         * Sent by the hybrid recognizer when a result is ready.
         * @param result The recognition result (never null)
         */
        public abstract void onResult(HybridRecognitionResult result);

        /**
         * A recognition error or internal error occurred. The Error object
         * contains details about what happened.
         * @param error The error (never null)
         */
        public abstract void onError(HybridRecognitionError error);
    }

    /**
     * @deprecated Call {@link #startRecognition(AudioSource,Scenario,RecognitionInterpreter,RecognitionInterpreter,InterpretedRecognitionCombiner,Listener)} instead.
     * Recognize a scenario. A scenario (see {@link Scenario}) specifies a server command and a set of
     * embedded constraints to be recognized in parallel.
     * @param audioSource The audio source to get audio from. The audio type
     * must both match the Elvis model type plus the Cloud audio codec type
     * (PCM 16Khz + Speex Wide-band or 8Khz + Speex Narrow-band). Must not be null.
     * @param scenario The recognition scenario for this recognition. Must not be null.
     * @param listener Listener to be notified when recognition is
     * complete in order to get results or failure indication. Must not be null.
     */
    public abstract void startRecognition(AudioSource<AudioChunk> audioSource,
                                          Scenario scenario,
                                          Listener listener);

    /**
     * Recognize a scenario. A scenario (see {@link Scenario}) specifies a server command and a set of
     * embedded constraints to be recognized in parallel.
     * @param audioSource The audio source to get audio from. The audio type
     * must both match the Elvis model type plus the Cloud audio codec type
     * (PCM 16Khz + Speex Wide-band or 8Khz + Speex Narrow-band). Must not be null.
     * @param scenario The recognition scenario for this recognition. Must not be null.
     * @param localArbitrationInterpreter The local recognition interpreter to use for arbitration. Must not be null.
     * @param listener Listener to be notified when recognition is
     * complete in order to get results or failure indication. Must not be null.
     */
    public abstract void startRecognition(AudioSource<AudioChunk> audioSource,
                                          Scenario scenario,
                                          RecognitionInterpreter<LocalRecognitionResult> localArbitrationInterpreter,
                                          Listener listener);

    /**
     * Recognize a scenario. A scenario (see {@link Scenario}) specifies a server command and a set of
     * embedded constraints to be recognized in parallel.
     * @param audioSource The audio source to get audio from. The audio type
     * must both match the Elvis model type plus the Cloud audio codec type
     * (PCM 16Khz + Speex Wide-band or 8Khz + Speex Narrow-band). Must not be null.
     * @param scenario The recognition scenario for this recognition. Must not be null.
     * @param localInterpreter The local recognition interpreter to use. Must not be null.
     * @param cloudInterpreter The cloud recognition interpreter to use. Must not be null.
     * @param hybridCombiner Used to combine the interpreted recognitions. Must not be null.
     * @param listener Listener to be notified when recognition is
     * complete in order to get results or failure indication. Must not be null.
     */
    public abstract void startRecognition(AudioSource<AudioChunk> audioSource,
                                          Scenario scenario,
                                          RecognitionInterpreter<LocalRecognitionResult> localInterpreter,
                                          RecognitionInterpreter<CloudRecognitionResult> cloudInterpreter,
                                          InterpretedRecognitionCombiner hybridCombiner,
                                          Listener listener);

    /**
     * Stop the current recognition and return a result.
     */
    public abstract void stopRecognition();

    /**
     * Cancel the current recognition and do not return a result.
     */
    public abstract void cancel();

    /**
     * @deprecated Call {@link #createHybridRecognizer(ElvisRecognizer,CloudServices)} instead.
     * Creates a new Elvis-based HybridRecognizer with references to a local listener to
     * receive Elvis results and a remote listener to receive cloud recognition
     * results.
     * @param elvisRecognizer Must not be null
     * @param cloudServices Must not be null
     * @param elvisInterpreter The Elvis recognition interpreter to use. Must not be null
     * @param cloudInterpreter The cloud recognition interpreter to use. Must not be null
     * @param hybridCombiner Used to combine the interpreted recognitions. Must not be null.
     */
    @Deprecated
    public static HybridRecognizer createHybridRecognizer(ElvisRecognizer elvisRecognizer,
                                                          CloudServices cloudServices,
                                                          RecognitionInterpreter<LocalRecognitionResult> elvisInterpreter,
                                                          RecognitionInterpreter<CloudRecognitionResult> cloudInterpreter,
                                                          InterpretedRecognitionCombiner hybridCombiner)
    {
        return new ElvisHybridRecognizerImpl(elvisRecognizer, cloudServices, elvisInterpreter, cloudInterpreter, hybridCombiner);
    }

    /**
     * Creates a new Elvis-based HybridRecognizer with references to a local listener to
     * receive Elvis results and a remote listener to receive cloud recognition
     * results.
     * @param elvisRecognizer Must not be null
     * @param cloudServices Must not be null
     */
    public static HybridRecognizer createHybridRecognizer(ElvisRecognizer elvisRecognizer,
                                                          CloudServices cloudServices)
    {
        return new ElvisHybridRecognizerImpl(elvisRecognizer, cloudServices);
    }

    /**
     * Creates a new Vocon-based HybridRecognizer with references to a local listener to
     * receive Vocon results and a remote listener to receive cloud recognition
     * results.
     * @param voconRecognizer Must not be null
     * @param cloudServices Must not be null
     */
    public static HybridRecognizer createHybridRecognizer(VoconRecognizer voconRecognizer,
                                                          CloudServices cloudServices)
    {
        return new VoconHybridRecognizerImpl(voconRecognizer, cloudServices);
    }
}
