package com.nuance.dragon.toolkit.hybrid;

import java.util.LinkedList;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioPipe;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.pipes.BufferingPipe;
import com.nuance.dragon.toolkit.audio.pipes.ResamplerPipe;
import com.nuance.dragon.toolkit.audio.pipes.SpeexDecoderPipe;
import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.TransactionError;
import com.nuance.dragon.toolkit.cloudservices.vocalizer.CloudVocalizer;
import com.nuance.dragon.toolkit.cloudservices.vocalizer.TtsSpec;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.vocalizer.Vocalizer;

/**
* The Hybrid vocalizer is a consolidated entry point for server or local TTS.
* The server is always favored as the TTS back-end for generating utterances.
* However, if there is a network error (connection problem, timeout), the utterance
* will eventually be generated by the embedded TTS engine as a fallback.
* <p>
* The Hybrid vocalizer flow can be specified via the the {@link HybridTtsSpec}. The generated
* audio source will always be PCM 16k (server audio being systematically decoded) so the
* callee's audio playback configuration doesn't have to change at run-time depending on which
* NMT TTS back-end generated the audio.
*/
public class HybridVocalizer
{
    /**
     * Listens for the generation success or error notifications of the hybrid TTS system.
     */
    public interface Listener
    {
        /**
         * The TTS generation succeeded.
         * @param fromCloud True if the TTS generation happened on the server, false if it happened locally.
         */
        public abstract void onSuccess(boolean fromCloud);

        /**
         * The TTS generation failed.
         * @param error The {@link HybridTtsError}
         */
        public abstract void onError(HybridTtsError error);
    }

    private static final class TtsRequest
    {
        final HybridTtsSpec hybridTtsSpec;
        final AudioPipe<AudioChunk, AudioChunk> pipe;
        final Listener listener;

        public TtsRequest(HybridTtsSpec hybridTtsSpec, AudioPipe<AudioChunk, AudioChunk> pipe, Listener listener)
        {
            this.hybridTtsSpec = hybridTtsSpec;
            this.pipe = pipe;
            this.listener = listener;
        }
    }

    private final Vocalizer _localVocalizer;
    private final CloudVocalizer _cloudVocalizer;
    private final AudioType _outAudioType;
    private final LinkedList<TtsRequest> _ttsQueue;
    private TtsRequest _currentTtsRequest;
    private SpeexDecoderPipe _decoder;
    private ResamplerPipe _resampler;

    /**
     * Creates a new HybridVocalizer.
     * @param localVocalizer Must not be null
     * @param cloudServices Must not be null
     */
    public HybridVocalizer(Vocalizer localVocalizer, CloudServices cloudServices)
    {
        Checker.checkArgForNull("localVocalizer", localVocalizer);
        Checker.checkArgForNull("cloudServices", cloudServices);

        _localVocalizer = localVocalizer;
        _cloudVocalizer = new CloudVocalizer(cloudServices);
        _ttsQueue = new LinkedList<TtsRequest>();
        _outAudioType = AudioType.PCM_16k; // TODO: Expose?
    }

    /**
     * Generate TTS. If TTS generation is already in process, this will cancel
     * the old TTS request.
     * @param hybridTtsSpec Describes the TTS to generate.
     * @param listener Optional listener for the success / failure of the generation. May be null.
     * @return The audio source that will provide the generated TTS audio.
     */
    public AudioSource<AudioChunk> generateTts(HybridTtsSpec hybridTtsSpec, Listener listener)
    {
        Checker.checkArgForNull("hybridTtsSpec", hybridTtsSpec);

        BufferingPipe<AudioChunk> pipe = new BufferingPipe<AudioChunk>(_outAudioType);
        _ttsQueue.add(new TtsRequest(hybridTtsSpec, pipe, listener));
        startNext();
        return pipe;
    }

    private void startNext()
    {
        if (_currentTtsRequest != null || _ttsQueue.isEmpty())
            return;

        final TtsRequest next = _ttsQueue.remove();
        final TtsSpec cloudSpec = next.hybridTtsSpec.getCloudSpec();
        final Data.Dictionary localSpec = next.hybridTtsSpec.getLocalSpec();
        Checker.checkArgsForNull("Cloud and local TTS specs ", cloudSpec, localSpec);

        final BufferingPipe<AudioChunk> engineBufferingPipe = new BufferingPipe<AudioChunk>(cloudSpec != null ? cloudSpec.getAudioType() : _localVocalizer.getAudioType());
        final Vocalizer.TtsListener localTtsListener = new Vocalizer.TtsListener()
        {
            @Override
            public void ttsGenerationStarted(String text, Object context, Vocalizer vocalizer)
            {
                // Nothing to do
                return;
            }

            @Override
            public void ttsGenerationFinished(String text, Object context, Vocalizer vocalizer, boolean success)
            {
                if (_currentTtsRequest == null)
                    return;

                Listener listener = _currentTtsRequest.listener;
                if (success)
                {
                    if (listener != null)
                        listener.onSuccess(false);
                }
                else
                {
                    _currentTtsRequest.pipe.disconnectAudioSource();
                    if (listener != null)
                        listener.onError(new HybridTtsError(null, true));
                }
                _currentTtsRequest = null;
                startNext();
            }

            @Override
            public void ttsStreamingStarted(String text, Object context,  Vocalizer vocalizer)
            {
                // Nothing to do
                return;
            }

            @Override
            public void ttsStreamingFinished(String text, Object context,  Vocalizer vocalizer)
            {
                // Nothing to do
                return;
            }
        };

        final CloudVocalizer.Listener cloudTtsListener = new CloudVocalizer.Listener()
        {
            @Override
            public void onSuccess()
            {
                if (_currentTtsRequest == null)
                    return;

                Listener listener = _currentTtsRequest.listener;
                cleanupAudioChain();
                _decoder = new SpeexDecoderPipe();
                _decoder.connectAudioSource(engineBufferingPipe);
                _currentTtsRequest.pipe.connectAudioSource(_decoder);
                _currentTtsRequest = null;

                if (listener != null)
                    listener.onSuccess(true);
                startNext();
            }

            @Override
            public void onError(TransactionError error)
            {
                if (_currentTtsRequest == null)
                    return;

                Listener listener = _currentTtsRequest.listener;
                if (localSpec != null)
                {
                    cleanupAudioChain();
                    _resampler = new ResamplerPipe(_outAudioType);
                    _resampler.connectAudioSource(_localVocalizer.generateTts(localSpec, localTtsListener, null));
                    _currentTtsRequest.pipe.connectAudioSource(_resampler);
                }
                else
                {
                    _currentTtsRequest.pipe.disconnectAudioSource();
                    _currentTtsRequest = null;
                    if (listener != null)
                        listener.onError(new HybridTtsError(error, false));
                    startNext();
                }
            }

            @Override
            public void onProcessingStarted()
            {
            }
        };

        _currentTtsRequest = next;

        if (cloudSpec != null)
        {
            Assert.assertTrue(engineBufferingPipe.getAudioType().encoding == AudioType.Encoding.SPEEX);
            engineBufferingPipe.connectAudioSource(_cloudVocalizer.generateTts(cloudSpec, cloudTtsListener));
        }
        else
        {
            // Emulate a network error to generate the utterance locally
            cloudTtsListener.onError(null);
        }
    }

    /**
     * Cancel any TTS that is in progress.
     */
    public void cancelTts()
    {
        cleanupAudioChain();
        _localVocalizer.cancelTts();
        _cloudVocalizer.cancel();

        // Clean-up here
        if (_currentTtsRequest != null)
        {
            _currentTtsRequest.pipe.disconnectAudioSource();
            if (_currentTtsRequest.listener != null)
                _currentTtsRequest.listener.onError(new HybridTtsError(null, true));
            _currentTtsRequest = null;
        }
        for (TtsRequest r : _ttsQueue)
        {
            r.pipe.disconnectAudioSource();
            if (r.listener != null)
                r.listener.onError(new HybridTtsError(null, true));
        }
        _ttsQueue.clear();
    }

    private void cleanupAudioChain()
    {
        if (_decoder != null)
        {
            _decoder.disconnectAudioSource();
            _decoder.release();
            _decoder = null;
        }

        if (_resampler != null)
        {
            _resampler.disconnectAudioSource();
            _resampler.release();
            _resampler = null;
        }
    }
}
