package com.nuance.dragon.toolkit.hybrid;

import java.util.ArrayList;
import java.util.List;

import android.util.Pair;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.SpeechDetectionListener;
import com.nuance.dragon.toolkit.audio.pipes.DuplicatorPipe;
import com.nuance.dragon.toolkit.audio.pipes.SpeexEncoderPipe;
import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.DataParam;
import com.nuance.dragon.toolkit.cloudservices.Param;
import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognitionError;
import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognitionResult;
import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognizer;
import com.nuance.dragon.toolkit.cloudservices.recognizer.RecogSpec;
import com.nuance.dragon.toolkit.grammar.Scenario;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.AsyncTaskHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.recognition.InterpretException;
import com.nuance.dragon.toolkit.recognition.InterpretedRecognition;
import com.nuance.dragon.toolkit.recognition.InterpretedRecognitionCombiner;
import com.nuance.dragon.toolkit.recognition.LocalRecognitionResult;
import com.nuance.dragon.toolkit.recognition.RecognitionInterpreter;
import com.nuance.dragon.toolkit.vocon.VoconContext;
import com.nuance.dragon.toolkit.vocon.VoconError;
import com.nuance.dragon.toolkit.vocon.VoconRecognizer;
import com.nuance.dragon.toolkit.vocon.VoconResult;

/**
 * Implementation of {@link HybridRecognizer}.
 * <p>
 * Don't create this directly. Use {@link HybridRecognizer#createHybridRecognizer(VoconRecognizer,CloudServices,RecognitionInterpreter,RecognitionInterpreter,InterpretedRecognitionCombiner)}.
 */
final class VoconHybridRecognizerImpl extends HybridRecognizer
{
    private static final class HybridRecognition
    {
        private CloudRecognizer _cloudRecognizer;
        private List<CloudRecognitionResult> _cloudRecognitionRawResults;
        private InterpretedRecognition _cloudRecognitionResult;
        private CloudRecognitionError _cloudError;
        private boolean _voconRecognizerActive;
        private LocalRecognitionResult _voconRawResult;
        private InterpretedRecognition _voconResult;
        private VoconError _voconError;
        private AsyncTaskHandler.TaskRunnable<Pair<HybridRecognitionResult, HybridRecognitionError>> _reportingTask;
        private List<String> _slotsToUpdate;
        private boolean _localFxOnly;
    }

    private DuplicatorPipe<AudioChunk> _duplicator;
    private SpeexEncoderPipe _speexEncoderPipe;
    private AudioSource<AudioChunk> _localAudioSource;
    private AudioSource<AudioChunk> _remoteAudioSource;
    private final VoconRecognizer _voconRecognizer;
    private final CloudServices _cloudServices;
    private final AsyncTaskHandler _workerThreadHandler;
    private HybridRecognition _currentRecognition;

    /**
     * Creates a new HybridRecognizer with references to a local listener to
     * receive Vocon results and a remote listener to receive cloud recognition
     * results.
     * @param voconRecognizer A valid Vocon recognizer instance. Must not be null
     * @param cloudServices A valid Cloud Services instance. Must not be null
     */
    public VoconHybridRecognizerImpl(VoconRecognizer voconRecognizer,
                                     CloudServices cloudServices)
    {
        Checker.checkArgForNull("voconRecognizer", voconRecognizer);
        Checker.checkArgForNull("cloudServices", cloudServices);

        _voconRecognizer = voconRecognizer;
        _cloudServices = cloudServices;
        _workerThreadHandler = new AsyncTaskHandler();
    }

    private void reportError(HybridRecognition recognition,
                             HybridRecognitionError error,
                             Listener listener)
    {
        if (recognition == _currentRecognition)
            _currentRecognition = null;

        listener.onError(error);
    }

    private void reportResult(final HybridRecognition recognition,
                              final LocalRecognitionResult voconRawResult,
                              final List<CloudRecognitionResult> cloudRawResults,
                              final CloudRecognitionError eventualCloudError,
                              final boolean isFinal,
                              final Listener listener)
    {
        if (recognition != _currentRecognition)
            return;

        recognition._reportingTask = new AsyncTaskHandler.TaskRunnable<Pair<HybridRecognitionResult, HybridRecognitionError>>()
        {
            @Override
            public Pair<HybridRecognitionResult, HybridRecognitionError> run()
            {
                try {
                    CompositeRecognitionResult compositeHybridResult = new CompositeRecognitionResult(voconRawResult, cloudRawResults, eventualCloudError, recognition._slotsToUpdate);
                    return new Pair<HybridRecognitionResult, HybridRecognitionError>(compositeHybridResult, null);
                }
                catch (InterpretException e) {
                    HybridRecognitionError error = new HybridRecognitionError();
                    error.setCloudError(e);
                    error.setVoconError(e);
                    return new Pair<HybridRecognitionResult, HybridRecognitionError>(null, error);
                }
            }

            @Override
            public void onPostRun(Pair<HybridRecognitionResult, HybridRecognitionError> result)
            {
                if (recognition != _currentRecognition)
                    return;

                if (isFinal || result.first == null)
                    _currentRecognition = null;

                if (result.first != null)
                    listener.onResult(result.first);
                else
                    listener.onError(result.second);
            }
        };
        _workerThreadHandler.post(_currentRecognition._reportingTask);
    }

    private void reportResult(final HybridRecognition recognition,
                              final InterpretedRecognition voconResult,
                              final InterpretedRecognition cloudResult,
                              final InterpretedRecognitionCombiner combiner,
                              final Listener listener)
    {
        if (recognition != _currentRecognition)
            return;

        // Interpreting the cloud result can be slow, so do it asynchronously
        recognition._reportingTask = new AsyncTaskHandler.TaskRunnable<Pair<HybridRecognitionResult, HybridRecognitionError>>()
        {
            @Override
            public Pair<HybridRecognitionResult, HybridRecognitionError> run()
            {
                try {
                    HybridRecognitionResult hybridResult = new HybridRecognitionResult(voconResult, cloudResult, combiner, recognition._slotsToUpdate);
                    return new Pair<HybridRecognitionResult, HybridRecognitionError>(hybridResult, null);
                }
                catch (InterpretException e) {
                    HybridRecognitionError error = new HybridRecognitionError();
                    error.setCloudError(e);
                    error.setVoconError(e);
                    return new Pair<HybridRecognitionResult, HybridRecognitionError>(null, error);
                }
            }

            @Override
            public void onPostRun(Pair<HybridRecognitionResult, HybridRecognitionError> result)
            {
                if (recognition != _currentRecognition)
                    return;

                _currentRecognition = null;

                if (result.first != null)
                    listener.onResult(result.first);
                else
                    listener.onError(result.second);
            }
        };
        _workerThreadHandler.post(_currentRecognition._reportingTask);
    }

    @Override
    public void startRecognition(AudioSource<AudioChunk> audioSource,
                                 Scenario scenario,
                                 final Listener listener)
    {
        final String msg = "Method is not supported. See other startRecognition() methods.";
        Logger.error(this, msg);
        throw new UnsupportedOperationException(msg);
    }

    @Override
    public void startRecognition(AudioSource<AudioChunk> audioSource,
                                 Scenario scenario,
                                 final RecognitionInterpreter<LocalRecognitionResult> localArbitrationInterpreter,
                                 final Listener listener)
    {
        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForNull("scenario", scenario);
        Checker.checkArgForNull("localArbitrationInterpreter", localArbitrationInterpreter);
        Checker.checkArgForNull("listener", listener);

        RecogSpec cloudRecogSpec = scenario.getCloudRecogSpec();
        List<VoconContext> voconContexts = scenario.getVoconContexts();
        Checker.checkArgsForNull("Cloud recognition spec and Vocon contexts", cloudRecogSpec, voconContexts);

        // Cancel any recognition in progress
        cancel();

        final HybridRecognition recog = new HybridRecognition();
        recog._cloudRecognitionRawResults = new ArrayList<CloudRecognitionResult>();
        recog._localFxOnly = voconContexts == null || voconContexts.isEmpty();
        _currentRecognition = recog;

        // Set-up audio chain
        setupAudioChain(audioSource);

        // Start remote recognition
        if (cloudRecogSpec != null)
        {
            recog._cloudRecognizer = new CloudRecognizer(_cloudServices);
            recog._cloudRecognizer.startRecognition(cloudRecogSpec, _remoteAudioSource, new CloudRecognizer.Listener()
            {
                @Override
                public void onResult(CloudRecognitionResult result)
                {
                    Logger.debug(VoconHybridRecognizerImpl.this, "Cloud result received");
                    if (recog._cloudRecognizer == null) {
                        Logger.error(VoconHybridRecognizerImpl.this, "Got unexpected Cloud result");
                        return;
                    }

                    recog._cloudRecognitionRawResults.add(result);
                    if (result.isFinal()) {
                        recog._cloudRecognizer = null;
                    }

                    if (!recog._voconRecognizerActive || recog._localFxOnly) {
                        cleanupAudioChain();
                        reportResult(recog, recog._voconRawResult, recog._cloudRecognitionRawResults, recog._cloudError, result.isFinal(), listener);
                    }
                }

                @Override
                public void onError(CloudRecognitionError error)
                {
                    Logger.debug(VoconHybridRecognizerImpl.this, "Cloud error received");

                    if (recog._cloudRecognizer == null) {
                        Logger.error(VoconHybridRecognizerImpl.this, "Got unexpected Cloud error");
                        return;
                    }

                    recog._cloudError = error;
                    recog._cloudRecognizer = null;
                    if (!recog._voconRecognizerActive || recog._localFxOnly) {
                        cleanupAudioChain();

                        if (recog._voconResult != null) {
                            reportResult(recog, recog._voconRawResult, recog._cloudRecognitionRawResults, recog._cloudError, true, listener);
                        } else
                        {
                            HybridRecognitionError hybridError = new HybridRecognitionError();
                            hybridError.setVoconError(recog._voconError);
                            hybridError.setCloudError(error);
                            reportError(recog, hybridError, listener);
                        }
                    }
                }

                @Override
                public void onTransactionIdGenerated(String transactionId)
                {
                    // TODO Auto-generated method stub

                }
            });
        }

        // If somehow the recognition already ended by now, don't bother with VoCon
        if (recog != _currentRecognition)
            return;

        // Start local recognition
        if (recog._localFxOnly)
        {
            // Just use Vocon for the endpointing
            Logger.warn(this, "VoCon is only used for end-pointing in this recognition.");
            recog._voconRecognizerActive = true;
            _voconRecognizer.startSpeechDetection(_localAudioSource,
                                                  new SpeechDetectionListener()
            {
                @Override
                public void onStartOfSpeech()
                {
                    listener.onStartOfSpeech();
                }

                @Override
                public void onEndOfSpeech()
                {
                    if (!recog._voconRecognizerActive) {
                        Logger.error(VoconHybridRecognizerImpl.this, "Got unexpected Vocon end-of-speech");
                        return;
                    }

                    recog._voconRecognizerActive = false;
                    _voconRecognizer.stopListening();
                    cleanupAudioChain();
                    listener.onEndOfSpeech();

                    if (recog._cloudRecognizer != null)
                    {
                        recog._cloudRecognizer.processResult();
                    }
                }
            });
        } else
        {
            // Multi-context recognition management (only applicable for custom hybrid recognition,
            // to recognize the main context and a OOG context for example).
            final int totalResultCount = voconContexts.size();
            final List<VoconResult> results = new ArrayList<VoconResult>();

            recog._voconRecognizerActive = true;
            _voconRecognizer.startRecognition(_localAudioSource, voconContexts,
                new SpeechDetectionListener() {
                    @Override
                    public void onStartOfSpeech()
                    {
                        listener.onStartOfSpeech();
                    }

                    @Override
                    public void onEndOfSpeech()
                    {
                        _voconRecognizer.stopListening();
                        cleanupAudioChain();
                        listener.onEndOfSpeech();
                    }
                },
                null, // TODO_ADR: Expose signal quality notifications?
                new VoconRecognizer.ResultListener()
                {
                    @Override
                    public void onResult(VoconResult result) {
                        Logger.debug(VoconHybridRecognizerImpl.this, "Vocon result received");
                        if (!recog._voconRecognizerActive) {
                            Logger.error(VoconHybridRecognizerImpl.this, "Got unexpected Vocon result");
                            return;
                        }

                        results.add(result);
                        if (results.size() >= totalResultCount)
                        {
                            // All results came back from VoCon, propagate the best one
                            VoconResult bestResult = results.remove(0);
                            for (VoconResult r : results)
                            {
                                int lastBestConfidence = bestResult.getConfidence();
                                int currentConfidence = r.getConfidence();
                                Logger.debug(this, "Current confidence: " + currentConfidence + " ,last best confidence: " + lastBestConfidence);

                                if(currentConfidence >= lastBestConfidence)
                                    bestResult = r;
                            }
                            recog._voconRawResult = bestResult;

                            if (recog._cloudRecognizer != null)
                            {
                                // Send params to the cloud if necessary
                                List<DataParam> params = new ArrayList<DataParam>(3);
                                boolean keepWaiting = localArbitrationInterpreter.processForCloud(bestResult, params);
                                for (Param p : params)
                                {
                                    recog._cloudRecognizer.sendParam(p);
                                }
                                recog._cloudRecognizer.processResult();

                                if (!keepWaiting)
                                {
                                    // Let the command finish, but don't wait for it anymore.
                                    recog._cloudRecognizer = null;
                                }
                            }
                            recog._voconRecognizerActive = false;

                            if (recog._cloudRecognizer == null) {
                                cleanupAudioChain();
                                reportResult(recog, recog._voconRawResult, recog._cloudRecognitionRawResults, recog._cloudError, true, listener);
                            }
                        }
                    }

                    @Override
                    public void onError(VoconError error) {
                        Logger.debug(VoconHybridRecognizerImpl.this, "Vocon error received");
                        onError(error, null);
                    }

                    private void onError(VoconError error, InterpretException otherError) {
                        if (!recog._voconRecognizerActive) {
                            Logger.error(VoconHybridRecognizerImpl.this, "Got unexpected Vocon error");
                            return;
                        }

                        recog._voconRecognizerActive = false;
                        cleanupAudioChain();
                        if (recog._cloudRecognizer == null) {
                            HybridRecognitionError hybridError = new HybridRecognitionError();
                            hybridError.setVoconError(error);
                            hybridError.setVoconError(otherError);

                            reportError(recog, hybridError, listener);
                        }
                        else {
                            recog._voconError = error; // Save it for later TODO
                            recog._cloudRecognizer.processResult();
                        }
                    }
                });
        }
    }

    @Override
    public void startRecognition(AudioSource<AudioChunk> audioSource,
                                 Scenario scenario,
                                 final RecognitionInterpreter<LocalRecognitionResult> localInterpreter,
                                 final RecognitionInterpreter<CloudRecognitionResult> cloudInterpreter,
                                 final InterpretedRecognitionCombiner hybridCombiner,
                                 final Listener listener)
    {
        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForNull("scenario", scenario);
        Checker.checkArgForNull("localInterpreter", localInterpreter);
        Checker.checkArgForNull("cloudInterpreter", cloudInterpreter);
        Checker.checkArgForNull("hybridCombiner", hybridCombiner);
        Checker.checkArgForNull("listener", listener);

        RecogSpec cloudRecogSpec = scenario.getCloudRecogSpec();
        List<VoconContext> voconContexts = scenario.getVoconContexts();
        Checker.checkArgsForNull("Cloud recognition spec and Vocon contexts", cloudRecogSpec, voconContexts);

        // Cancel any recognition in progress
        cancel();

        final HybridRecognition recog = new HybridRecognition();
        recog._localFxOnly = voconContexts == null || voconContexts.isEmpty();
        _currentRecognition = recog;

        // Set-up audio chain
        setupAudioChain(audioSource);

        // Start remote recognition
        if (cloudRecogSpec != null)
        {
            recog._cloudRecognizer = new CloudRecognizer(_cloudServices);
            recog._cloudRecognizer.startRecognition(cloudRecogSpec, _remoteAudioSource, new CloudRecognizer.Listener()
            {
                @Override
                public void onResult(CloudRecognitionResult result)
                {
                    Logger.debug(VoconHybridRecognizerImpl.this, "Cloud result received");
                    if (recog._cloudRecognizer == null) {
                        Logger.error(VoconHybridRecognizerImpl.this, "Got unexpected Cloud result");
                        return;
                    }

                    try {
                        recog._cloudRecognitionResult = cloudInterpreter.getInterpretedResult(result);
                        recog._slotsToUpdate = cloudInterpreter.getUpdateRequiredList(result);
                    } catch (InterpretException ex)
                    {
                        onError(null, ex);
                        return;
                    }

                    recog._cloudRecognizer = null;
                    if (!recog._voconRecognizerActive || recog._localFxOnly) {
                        cleanupAudioChain();
                        reportResult(recog, recog._voconResult, recog._cloudRecognitionResult, hybridCombiner, listener);
                    }
                }

                @Override
                public void onError(CloudRecognitionError error)
                {
                    Logger.debug(VoconHybridRecognizerImpl.this, "Cloud error received");
                    onError(error, null);
                }

                private void onError(CloudRecognitionError cloudError, InterpretException otherError)
                {
                    if (recog._cloudRecognizer == null) {
                        Logger.error(VoconHybridRecognizerImpl.this, "Got unexpected Cloud error");
                        return;
                    }

                    recog._cloudRecognizer = null;
                    if (!recog._voconRecognizerActive || recog._localFxOnly) {
                        cleanupAudioChain();

                        if (recog._voconResult != null) {
                            // A cloud error occurred, but we can still report the vocon result
                            reportResult(recog, recog._voconResult, null, hybridCombiner, listener);
                        } else
                        {
                            HybridRecognitionError error = new HybridRecognitionError();
                            error.setVoconError(recog._voconError);
                            error.setCloudError(cloudError);
                            error.setCloudError(otherError);
                            reportError(recog, error, listener);
                        }
                    }
                }

                @Override
                public void onTransactionIdGenerated(String transactionId)
                {
                    // TODO Auto-generated method stub

                }
            });
        }

        // If somehow the recognition already ended by now, don't bother with VoCon
        if (recog != _currentRecognition)
            return;

        // Start local recognition
        if (recog._localFxOnly)
        {
            // Just use Vocon for the endpointing
            Logger.warn(this, "VoCon is only used for end-pointing in this recognition.");
            recog._voconRecognizerActive = true;
            _voconRecognizer.startSpeechDetection(_localAudioSource,
                                                  new SpeechDetectionListener()
            {
                @Override
                public void onStartOfSpeech()
                {
                    listener.onStartOfSpeech();
                }

                @Override
                public void onEndOfSpeech()
                {
                    if (!recog._voconRecognizerActive) {
                        Logger.error(VoconHybridRecognizerImpl.this, "Got unexpected Vocon end-of-speech");
                        return;
                    }

                    recog._voconRecognizerActive = false;
                    _voconRecognizer.stopListening();
                    cleanupAudioChain();
                    listener.onEndOfSpeech();

                    if (recog._cloudRecognizer != null)
                    {
                        recog._cloudRecognizer.processResult();
                    }
                }
            });
        } else
        {
            recog._voconRecognizerActive = true;
            _voconRecognizer.startRecognition(_localAudioSource, voconContexts,
                new SpeechDetectionListener() {
                    @Override
                    public void onStartOfSpeech()
                    {
                        listener.onStartOfSpeech();
                    }

                    @Override
                    public void onEndOfSpeech()
                    {
                        _voconRecognizer.stopListening();
                        cleanupAudioChain();
                        listener.onEndOfSpeech();
                    }
                },
                null, // TODO_ADR: Expose signal quality notifications?
                new VoconRecognizer.ResultListener()
                {
                    @Override
                    public void onResult(VoconResult result) {
                        Logger.debug(VoconHybridRecognizerImpl.this, "Vocon result received");
                        if (!recog._voconRecognizerActive) {
                            Logger.error(VoconHybridRecognizerImpl.this, "Got unexpected Vocon result");
                            return;
                        }

                        try {
                            recog._voconResult = localInterpreter.getInterpretedResult(result);
                        }
                        catch (InterpretException e) {
                            onError(null, e);
                            return;
                        }

                        if (recog._cloudRecognizer != null)
                        {
                            // Send params to the cloud if necessary
                            List<DataParam> params = new ArrayList<DataParam>(3);
                            boolean keepWaiting = localInterpreter.processForCloud(result, params);
                            for (Param p : params)
                            {
                                recog._cloudRecognizer.sendParam(p);
                            }
                            recog._cloudRecognizer.processResult();

                            if (!keepWaiting)
                            {
                                // Let the command finish, but don't wait for it anymore.
                                recog._cloudRecognizer = null;
                            }
                        }
                        recog._voconRecognizerActive = false;

                        if (recog._cloudRecognizer == null) {
                            cleanupAudioChain();
                            reportResult(recog, recog._voconResult, recog._cloudRecognitionResult, hybridCombiner, listener);
                        }
                    }

                    @Override
                    public void onError(VoconError error) {
                        Logger.debug(VoconHybridRecognizerImpl.this, "Vocon error received");
                        onError(error, null);
                    }

                    private void onError(VoconError error, InterpretException otherError) {
                        if (!recog._voconRecognizerActive) {
                            Logger.error(VoconHybridRecognizerImpl.this, "Got unexpected Vocon error");
                            return;
                        }

                        recog._voconRecognizerActive = false;
                        cleanupAudioChain();
                        if (recog._cloudRecognizer == null) {
                            HybridRecognitionError hybridError = new HybridRecognitionError();
                            hybridError.setVoconError(error);
                            hybridError.setVoconError(otherError);

                            reportError(recog, hybridError, listener);
                        }
                        else {
                            recog._voconError = error; // Save it for later TODO
                            recog._cloudRecognizer.processResult();
                        }
                    }
                });
        }
    }

    @Override
    public void stopRecognition()
    {
        if (_currentRecognition != null)
        {
            _voconRecognizer.stopListening();
            cleanupAudioChain();

            if (_currentRecognition._localFxOnly)
            {
                _currentRecognition._voconRecognizerActive = false;

                // In the case of using the engine FX only, we complete the command here
                CloudRecognizer cloudRecognizer = _currentRecognition._cloudRecognizer;
                if (cloudRecognizer != null)
                {
                    cloudRecognizer.processResult();
                }
            }
        }
    }

    @Override
    public void cancel()
    {
        if (_currentRecognition != null)
        {
            // When canceling, make sure that we don't report any results
            _currentRecognition._voconResult = null;
            _currentRecognition._cloudRecognitionResult = null;

            CloudRecognizer cloudRecognizer = _currentRecognition._cloudRecognizer;
            _currentRecognition = null;

            _voconRecognizer.cancelRecognition();
            if (cloudRecognizer != null)
            {
                cloudRecognizer.cancel();
            }
        }

        // The error callbacks should clean things up
    }

    private void setupAudioChain(AudioSource<AudioChunk> audioSource)
    {
        cleanupAudioChain();

        // Configure local audio source
        _duplicator = new DuplicatorPipe<AudioChunk>();
        _duplicator.connectAudioSource(audioSource);
        _localAudioSource = _duplicator;

        // Configure remote audio source
        _speexEncoderPipe = new SpeexEncoderPipe();
        _speexEncoderPipe.connectAudioSource(_duplicator);
        _remoteAudioSource = _speexEncoderPipe;
    }

    private void cleanupAudioChain()
    {
        if (_duplicator != null) {
            _duplicator.disconnectAudioSource();
            _duplicator = null;
        }

        if (_speexEncoderPipe != null) {
            _speexEncoderPipe.disconnectAudioSource();
            _speexEncoderPipe.release();
            _speexEncoderPipe = null;
        }
    }
}
