package com.nuance.dragon.toolkit.language;

import java.util.Locale;

import com.nuance.dragon.toolkit.util.internal.ObjectUtils;

/**
 * A DMT language that provides references to the corresponding languages for
 * Vocalizer, ElvisRecognizer, and the GrammarDepot.
 */
public class Language
{
    private final com.nuance.dragon.toolkit.elvis.ElvisLanguage _elvisLanguage;
    private final com.nuance.dragon.toolkit.vocalizer.VocalizerLanguage _vocalizerLanguage;
    private final com.nuance.dragon.toolkit.vocon.VoconLanguage _voconLanguage;
    private final String _grammarLanguage;
    private final String _desc;

    Language(com.nuance.dragon.toolkit.elvis.ElvisLanguage elvisLanguage,
             com.nuance.dragon.toolkit.vocalizer.VocalizerLanguage vocalizerLanguage,
             com.nuance.dragon.toolkit.vocon.VoconLanguage voconLanguage,
             String grammarLanguage,
             String description)
    {
        _elvisLanguage = elvisLanguage;
        _vocalizerLanguage = vocalizerLanguage;
        _voconLanguage = voconLanguage;
        _grammarLanguage = grammarLanguage;
        _desc = description;
    }

    /**
     * Get the Elvis language, which can be passed to an Elvis recognizer.
     * @return The Elvis language, or null if this NMT language does not have
     * Elvis support.
     */
    public com.nuance.dragon.toolkit.elvis.ElvisLanguage getElvisLanguage()
    {
        return _elvisLanguage;
    }

    /**
     * Get the Vocalizer language, which can be passed to a Vocalizer.
     * @return The Vocalizer language, or null if this NMT language does not
     * have Vocalizer support.
     */
    public com.nuance.dragon.toolkit.vocalizer.VocalizerLanguage getVocalizerLanguage()
    {
        return _vocalizerLanguage;
    }

    /**
     * Get the Vocon language, which can be passed to a Vocon.
     * @return The Vocon language, or null if this NMT language does not
     * have Vocon support.
     */
    public com.nuance.dragon.toolkit.vocon.VoconLanguage getVoconLanguage()
    {
        return _voconLanguage;
    }

    /**
     * Get the Cloud language, which can be passed to cloud-related modules like
     * CloudRecognizer.
     * @return The Grammar language (never null).
     */
    public String getCloudLanguage()
    {
        return _grammarLanguage;
    }

    /**
     * Get the Grammar language, which can be passed to a Grammar Depot.
     * @return The Grammar language (never null).
     */
    public String getGrammarLanguage()
    {
        return _grammarLanguage;
    }

    /**
     * Get the Grammar language, which can be passed to a Grammar Depot, from a Locale object.
     * @param l The input Locale object.
     * @return The Grammar language.
     */
    public static String getGrammarLanguage(Locale l)
    {
        return GrammarLanguage.getGrammarLanguage(l);
    }

    @Override
    /**
     * @return The description
     */
    public String toString()
    {
        return _desc;
    }

    @Override
    public boolean equals(Object item)
    {
        if (item == this)
            return true;

        if (item == null)
            return false;

        if (!(item instanceof Language))
            return false;

        Language lang = (Language)item;
        return _grammarLanguage.equals(lang._grammarLanguage) &&
               ObjectUtils.compareObjects(_elvisLanguage, lang._elvisLanguage) &&
               ObjectUtils.compareObjects(_vocalizerLanguage, lang._vocalizerLanguage) &&
               ObjectUtils.compareObjects(_voconLanguage, lang._voconLanguage);
    }

    @Override
    public int hashCode()
    {
        return _grammarLanguage.hashCode() +
               (_elvisLanguage == null ? 0 : _elvisLanguage.hashCode()) +
               (_vocalizerLanguage == null ? 0 : _vocalizerLanguage.hashCode()) +
               (_voconLanguage == null ? 0 : _voconLanguage.hashCode());
    }
}
