package com.nuance.dragon.toolkit.nvsl;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.util.internal.ReadOnlyList;

public class EnrollParam
{
    /** Default value for degradation level */
    public static int DEFAULT_DEGRADATION_LEVEL = 0;

    /**
     * Builder for creating instances of {@link EnrollParam}.
     */
    public static class Builder
    {
        private List<String> _noisyFiles;
        private int _degradationLevel;

        /**
         * Creates the builder with mandatory parameter(s).
         */
        public Builder()
        {
            _noisyFiles = new ArrayList<String>();
            _degradationLevel = DEFAULT_DEGRADATION_LEVEL;
        }

        /**
         * Set noisy files for enrollment with summed noise. The sample rate of audio
         * in noisy file should be consistent with the engine configuration being used.
         * @param fileList A list of noisy file to create voiceprint with robustness
         */
        public Builder setNoisyFiles(List<String> fileList)
        {
            this._noisyFiles = fileList;
            return this;
        }

        /**
         * Set degradation level for noise addition
         * @param level The level of degradation
         *        0 = low
         *        1 = medium
         *        2 = high
         *        Default level is 0
         */
        public Builder setDegradationLevel(int level)
        {
            this._degradationLevel = level;
            return this;
        }

        /**
         * Build the parameters.
         * @return The {@link EnrollParam} instance.
         */
        public EnrollParam build()
        {
            Checker.checkArgForNull("_noisyFiles", _noisyFiles);
            Checker.checkArgForCondition("_degradationLevel", ">= 0 and <= 2", (_degradationLevel >= 0 && _degradationLevel <= 2));
            return new EnrollParam(this);
        }

    }

    private List<String> _noisyFiles;
    private int _degradationLevel;

    private EnrollParam(Builder builder)
    {
        this._noisyFiles = builder._noisyFiles;
        this._degradationLevel = builder._degradationLevel;
    }

    List<String> getNoisyFiles()
    {
        return new ReadOnlyList<String>(_noisyFiles);
    }

    int getDegradationLevel()
    {
        return _degradationLevel;
    }

    @Override
    public String toString()
    {
        return "EnrollParam [_noisyFiles=" + _noisyFiles
                + ", _degradationLevel=" + _degradationLevel + "]";
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((_noisyFiles == null) ? 0 : _noisyFiles.hashCode());
        result = prime * result + _degradationLevel;
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        EnrollParam other = (EnrollParam) obj;
        if (_degradationLevel != other._degradationLevel)
            return false;
        if (_noisyFiles == null) {
            if (other._noisyFiles != null)
                return false;
        } else if (!_noisyFiles.equals(other._noisyFiles))
            return false;
        return true;
    }

}
