package com.nuance.dragon.toolkit.nvsl;

import com.nuance.dragon.toolkit.oem.api.internal.Checker;


public class IdentifyParam
{
    /** Default parameters */
    public static final boolean DEFAULT_MULTIIDENT_OFF = false;
    public static final int DEFAULT_MAX_RESULTS = 3;
    public static final int DEFAULT_PARTIAL_RESULT_TIMER_MS = 1000;

    /**
     * Builder for creating instances of {@link IdentifyParam}.
     */
    public static class Builder
    {
        private boolean _isMultiIdent;
        private int _maxResults;
        private int _partialResultTimerMs;

        /**
         * Creates the builder with mandatory parameter(s).
         */
        public Builder()
        {
            this._isMultiIdent = DEFAULT_MULTIIDENT_OFF;
            this._maxResults = DEFAULT_MAX_RESULTS;
            this._partialResultTimerMs = DEFAULT_PARTIAL_RESULT_TIMER_MS;
        }

        /**
         * Set multi-identification on/off.
         * @param flag A switch to turn multi-identification on or off
         *             Default is {@link #DEFAULT_MULTIIDENT_OFF}.
         */
        public Builder setMultiIdent(boolean flag)
        {
            this._isMultiIdent = flag;
            return this;
        }

        /**
         * Set custom maximum number of n-best list of identification results.
         * This number doesn't impact multi-identification result.
         * @param maxResults The maximum number of n-best results of a single identification
         *                   Default is {@link #DEFAULT_MAX_RESULTS}.
         */
        public Builder setMaxResults(int maxResults)
        {
            this._maxResults = maxResults;
            return this;
        }

        /**
         * Set custom timer of multi-speaker identification partial result, minimum 1000ms.
         * This timer doesn't take effect if task is a single speaker verification/identification
         * @param timerMs The timer of streaming diarization in millisecond
         *                   Default is {@link #DEFAULT_PARTIAL_RESULT_TIMER_MS}.
         */
        public Builder setPartialResultTimer(int timerMs)
        {
            this._partialResultTimerMs = timerMs;
            return this;
        }

        /**
         * Build the parameters.
         * @return The {@link IdentifyParam} instance.
         */
        public IdentifyParam build()
        {
            Checker.checkArgForCondition("maxResults", ">= 1", _maxResults >= 1);
            Checker.checkArgForCondition("_partialResultTimerMs", ">= 1000", _partialResultTimerMs >= 1000);
            return new IdentifyParam(this);
        }
    }

    private boolean _isMultiIdent;
    private int _maxResults;
    private int _partialResultTimerMs;

    private IdentifyParam(Builder builder)
    {
        this._isMultiIdent = builder._isMultiIdent;
        this._maxResults = builder._maxResults;
        this._partialResultTimerMs = builder._partialResultTimerMs;
    }

    boolean isMultiIdentEnabled()
    {
        return _isMultiIdent;
    }

    int getMaxResults()
    {
        return _maxResults;
    }

    int getPartialResultTimerMs()
    {
        return _partialResultTimerMs;
    }

    @Override
    public String toString()
    {
        return "IdentifyParam [_isMultiIdent=" + _isMultiIdent
                + ", _maxResults=" + _maxResults
                + ", _partialResultTimerMs=" + _partialResultTimerMs + "]";
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + (_isMultiIdent ? 1231 : 1237);
        result = prime * result + _maxResults;
        result = prime * result + _partialResultTimerMs;
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        IdentifyParam other = (IdentifyParam) obj;
        if (_isMultiIdent != other._isMultiIdent)
            return false;
        if (_maxResults != other._maxResults)
            return false;
        if (_partialResultTimerMs != other._partialResultTimerMs)
            return false;
        return true;
    }
}
