package com.nuance.dragon.toolkit.nvsl;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * Contains configuration files used for initializing NVSL
 * global resource, engine, framework, etc.
 */
public class NVSLConfig implements JSONCompliant
{
    private static final String KEY_XMLCONFFILE = "xmlconffile";
    private static final String KEY_SPKIDFILE = "spkidfile";
    private static final String KEY_LANGIDFILE = "langidfile";
    private static final String KEY_FALSEMEAN = "falsemean";
    private static final String KEY_FALSESTD = "falsestd";

    final String xmlConfFile;
    final String speakerIdFile;
    final String languageIdFile;

    String calibrationFile;
    float falseMean =  0.0f;
    float falseStd = 1.0f;
    float minAudioLength = 1.0f;
    float maxAudioLength = 600.0f;
    float minSpeechLength = 0.0f;
    float maxSpeechLength = 500.0f;
    float audioNoisyThreshold = 10.0f;
    float audioSoftThreshold = 2.0f;
    float audioLoudThreshold = 600.0f;

    /**
     * @param speakerIdFile speaker identification and segmentation configuration {see @link NVSLAbility for possible values}.
     */
    public NVSLConfig(String speakerIdFile) {
        this("NVSLEngine.xml", speakerIdFile, null);
    }

    /**
     * @param speakerIdFile speaker identification and segmentation configuration {see @link NVSLAbility for possible values}.
     * @param languageIdFile language identification configuration {see @link NVSLAbility for possible values}.
     */
    public NVSLConfig(String speakerIdFile, String languageIdFile)
    {
        this("NVSLEngine.xml", speakerIdFile, languageIdFile);
    }

    /**
     * @param xmlConfFile NVSL XML configuration file.
     * @param speakerIdFile speaker identification and segmentation configuration.
     * @param languageIdFile language identification configuration.
     */
    public NVSLConfig(String xmlConfFile, String speakerIdFile, String languageIdFile)
    {
        Checker.checkStringArgForNullOrEmpty("xmlConfFile", xmlConfFile);
        Checker.checkStringArgForNullOrEmpty("speakerIdFile", speakerIdFile);
        if (languageIdFile != null)
            Checker.checkStringArgForEmpty("languageIdFile", languageIdFile);

        this.xmlConfFile = xmlConfFile;
        this.speakerIdFile = speakerIdFile;
        this.languageIdFile = languageIdFile;
        this.calibrationFile = null;
    }

    /**
     * Is the configuration in Text-Independent (TI) operation mode
     */
    public boolean isTIMode(){

        if (speakerIdFile.startsWith("TD"))
            return false;
        else
            return true;
    }

    /**
     * Is the configuration supporting language identification
     */
    public boolean isLangID(){
        return languageIdFile != null;
    }

    /**
     * In case of mapped score is clumped together, application could set
     * new impostor mean and standard deviation to achieve a better mapped score
     *
     * @param falseMean Impostor score average from development set.
     * @param falseStd Impostor score standard deviation estimate
     */
    public void setFalseMeanAndStd(float falseMean, float falseStd)
    {
        this.falseMean = falseMean;
        this.falseStd = falseStd;
    }

    /**
     * Gets the mean from the impostor distribution.
     * @return the mean
     * @see NVSLConfig#setFalseMeanAndStd(float, float)
     */
    public float getFalseMean()
    {
        return falseMean;
    }

    /**
     * Gets the standard deviation from the impostor distribution.
     * @return the standard deviation
     * @see NVSLConfig#setFalseMeanAndStd(float, float)
     */
    public float getFalseStdDeviation()
    {
        return falseStd;
    }

    /**
     * Sets speaker identification calibration file. Default calibration data of
     * framework configuration will be used when it is not specified.
     *
     * @param calibrationFile Calibration data file name
     */
    public void setCalibrationFile(String calibrationFile)
    {
        this.calibrationFile = calibrationFile;
    }

    /**
     * Returns the minimum duration in seconds allowed for the audio signal to be processed.
     *
     * @see NVSLConfig#setMinAudioLength(float)
     */
    public float getMinAudioLength()
    {
        return minAudioLength;
    }

    /**
     * Sets the minimum duration in seconds allowed for the audio signal to be processed,
     *
     * If audio length is below this limit {@link NVSLError.Reason#AUDIO_TOO_SHORT} will be thrown
     * when checking the audio quality during enrollment
     *
     * Default value: 1.0f
     *
     * @param minAudioLength the minimum audio length
     */
    public void setMinAudioLength(float minAudioLength)

    {
        this.minAudioLength = minAudioLength;
    }

    /**
     * Returns the maximum amount in seconds allowed for the audio signal to be processed.
     *
     * @see NVSLConfig#setMaxAudioLength(float)
     */
    public float getMaxAudioLength()
    {
        return maxAudioLength;
    }

    /**
     * Sets the maximum amount in seconds allowed for the audio signal to be processed,
     *
     * If audio length exceeds this limit {@link NVSLError.Reason#AUDIO_TOO_LONG} will be thrown
     * when checking the audio quality during enrollment
     *
     * Default value: 600.0f
     *
     * @param maxAudioLength the maximum allowed length for the audio signal
     */
    public void setMaxAudioLength(float maxAudioLength)
    {
        this.maxAudioLength = maxAudioLength;
    }

    /**
     * Returns the minimum duration in seconds allowed for the speech to be processed.
     *
     * @see NVSLConfig#setMinSpeechLength(float)
     */
    public float getMinSpeechLength()
    {
        return minSpeechLength;
    }

    /**
     * Sets the minimum duration in seconds allowed for the speech to be processed,
     *
     * If speech length is below this limit {@link NVSLError.Reason#SPEECH_TOO_SHORT} will be thrown
     * when checking the audio quality during enrollment
     *
     * Default value: 0.0f
     *
     * @param minSpeechLength the minimum speech length
     */
    public void setMinSpeechLength(float minSpeechLength)

    {
        this.minSpeechLength = minSpeechLength;
    }

    /**
     * Returns the maximum amount in seconds allowed for the speech to be processed.
     *
     * @see NVSLConfig#setMaxSpeechLength(float)
     */
    public float getMaxSpeechLength()
    {
        return maxSpeechLength;
    }

    /**
     * Sets the maximum amount in seconds allowed for the speech to be processed,
     *
     * If speech length exceeds this limit {@link NVSLError.Reason#SPEECH_TOO_LONG} will be thrown
     * when checking the audio quality during enrollment
     *
     * Default value: 500.0f
     *
     * @param maxSpeechLength the maximum allowed length for the processed speech
     */
    public void setMaxSpeechLength(float maxSpeechLength)
    {
        this.maxSpeechLength = maxSpeechLength;
    }

    /**
     * Returns the SNR lower threshold to accept an audio segment
     *
     * @return the SNR lower threshold
     * @see NVSLConfig#setAudioNoisyThreshold(float)
     */
    public float getAudioNoisyThreshold()
    {
        return audioNoisyThreshold;
    }

    /**
     * Sets the SNR lower threshold to accept an audio segment
     *
     * If audio SNR is below this limit {@link NVSLError.Reason#AUDIO_TOO_NOISY} will be thrown
     * when checking the audio quality during enrollment
     *
     * Default value: 10.0f
     *
     * @param audioNoisyThreshold
     */
    public void setAudioNoisyThreshold(float audioNoisyThreshold)
    {
        this.audioNoisyThreshold = audioNoisyThreshold;
    }

    /**
     * Returns the speech level lower threshold to accept an audio segment
     *
     * @return the speech level lower threshold
     * @see NVSLConfig#setAudioSoftThreshold(float)
     */
    public float getAudioSoftThreshold()
    {
        return audioSoftThreshold;
    }

    /**
     * Sets the speech level lower threshold to accept an audio segment
     *
     * If audio speech level is below this limit {@link NVSLError.Reason#AUDIO_TOO_SOFT} will be thrown
     * when checking the audio quality during enrollment
     *
     * Default value: 4.0f
     *
     * @param audioSoftThreshold
     */
    public void setAudioSoftThreshold(float audioSoftThreshold)
    {
        this.audioSoftThreshold = audioSoftThreshold;
    }

    /**
     * Gets the upper threshold to detect distorted (or "clipped") audio utterances.
     *
     * @return the upper audio distortion threshold
     * @see NVSLConfig#setAudioLoudThreshold(float)
     */
    public float getAudioLoudThreshold()
    {
        return audioLoudThreshold;
    }

    /**
     * Sets the upper threshold to detect distorted (or "clipped") audio utterances that might
     * degrade performance in Enrollment, Identification, etc.
     *
     * The unit is the ratio of saturated audio samples per minute.
     * (Known as "saturation level" in NVSL documentation)
     * <p>
     * If audio speech level exceeds this limit {@link NVSLError.Reason#AUDIO_TOO_LOUD} will be thrown
     * when checking the audio quality during enrollment
     *
     * Default value: 600.0f
     *
     * @param audioLoudThreshold
     */
    public void setAudioLoudThreshold(float audioLoudThreshold)
    {
        this.audioLoudThreshold = audioLoudThreshold;
    }

    String getConfSetName()
    {
        String confSetName = null;

        if (null != speakerIdFile) {
            confSetName = speakerIdFile.replace(".nsrpkg", "");
        }

        return confSetName;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();

        o.tryPut(KEY_XMLCONFFILE, xmlConfFile);
        o.tryPut(KEY_SPKIDFILE, speakerIdFile);
        o.tryPut(KEY_LANGIDFILE, languageIdFile);
        o.tryPut(KEY_FALSEMEAN, falseMean);
        o.tryPut(KEY_FALSESTD, falseStd);

        return o;
    }

    public static NVSLConfig createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String xmlConfFile = json.optString(KEY_XMLCONFFILE, null);
        String speakerIdFile = json.optString(KEY_SPKIDFILE, null);
        String languageIdFile = json.optString(KEY_LANGIDFILE, null);
        float falseMean = (float)json.getDouble(KEY_FALSEMEAN);
        float falseStd = (float)json.getDouble(KEY_FALSESTD);

        NVSLConfig config =  new NVSLConfig(xmlConfFile, speakerIdFile, languageIdFile);
        config.setFalseMeanAndStd(falseMean, falseStd);
        return config;
    }
}
