package com.nuance.dragon.toolkit.nvsl;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;


/**
 * A NVSL error.
 */
public class NVSLError implements JSONCompliant {

       /**
     * Possible reasons for NVSL errors.
     */
    public static class Reason
    {
        public static final int CANCELED              = 0;
        public static final int AUDIO_FRAMES_DROPPED  = 1;
        public static final int NO_RESULT             = 2;
        public static final int CANT_START            = 3;
        public static final int FILE_ERROR            = 4;
        public static final int LANGID_NOT_SUPPORTED  = 5;
        public static final int NOT_INITIALIZED       = 6;
        public static final int CONFIG_NOT_COMPATIBLE = 7;
        public static final int AUDIO_CHECK_ERROR     = 8;
        public static final int AUDIO_INCONSISTENT    = 9;
        public static final int MAX_AUDIO_NUM         = 10;
        public static final int AUDIO_TOO_SHORT       = 11;
        public static final int AUDIO_TOO_LONG        = 12;
        public static final int AUDIO_TOO_SOFT        = 13;
        public static final int AUDIO_TOO_LOUD        = 14;
        public static final int AUDIO_TOO_NOISY       = 15;
        public static final int SPEECH_TOO_SHORT      = 16;
        public static final int SPEECH_TOO_LONG       = 17;



        private static final String [] REASON_TEXTS = {"Identification canceled",
                                                       "Audio frames dropped",
                                                       "No identified result",
                                                       "Can't start",
                                                       "File error",
                                                       "Language detection not supported",
                                                       "NVSL is not initialized.",
                                                       "Voiceprint config is not compatible with engine",
                                                       "Audio check process error",
                                                       "Audios are not consistent",
                                                       "Max audio number is reached in voiceprint",
                                                       "Audio information is too short",
                                                       "Audio information is too long",
                                                       "Audio information is too soft",
                                                       "Audio information is too loud",
                                                       "Audio information is too noisy",
                                                       "Speech information is too short",
                                                       "Speech information is too long",
                                                       };
    }

    private static final String KEY_CODE = "code";

    private final int _code;
    private final String _reason;

    NVSLError(int code)
    {
        _code = code;
        _reason = Reason.REASON_TEXTS[code];
    }

    /**
     * Get the reason code, from {@link Reason}.
     * @return The reason code
     */
    public int getReasonCode()
    {
        return _code;
    }

    /**
     * @return A description corresponding with the reason code
     */
    public String getReason()
    {
        return _reason;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        o.tryPut(KEY_CODE, _code);
        return o;
    }

    // TODO never null?
    public static NVSLError createFromJSON(org.json.JSONObject json) throws JSONException
    {
        int code = json.getInt(KEY_CODE);
        if (code < 0 || code >= Reason.REASON_TEXTS.length)
            throw new JSONException("Invalid code");
        return new NVSLError(code);
    }

    @Override
    public String toString()
    {
        return _reason;
    }
}
