package com.nuance.dragon.toolkit.nvsl;

import java.util.ArrayList;
import java.util.List;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * A group contains list of speakers {@link NVSLSpeaker}. A directory with name as
 * group id corresponding to the group would be created.
 */
public class NVSLGroup implements JSONCompliant
{
    private static final String KEY_GROUP_ID  = "groupid";
    private static final String KEY_SPEAKERS   = "speakers";

    private final String _groupId;
    private final List<NVSLSpeaker> _speakers;

    NVSLGroup(String groupId)
    {
        Checker.checkStringArgForNullOrEmpty("groupId", groupId);
        _groupId = groupId;
        _speakers = new ArrayList<NVSLSpeaker>();
    }

    /**
     * Returns the unique string identifying this group.
     * @return the group id
     */
    public String getGroupId()
    {
        return _groupId;
    }

    /**
     * Creates a speaker and attaches it to this group.
     * If the speaker existed already, it returns the existing one.
     *
     * @param speakerId the unique string identifying the new speaker
     * @return the newly created {@link NVSLSpeaker}
     */
    public NVSLSpeaker createSpeaker(String speakerId)
    {
        NVSLSpeaker sp;
        if ((sp = getSpeaker(speakerId)) != null)
            return sp;

        sp = new NVSLSpeaker(speakerId);
        _speakers.add(sp);
        return sp;
    }

    /**
     * Returns the NVSLSpeaker object with the given id if available in this group.
     *
     * @param speakerId
     * @return the NVSLSpeaker object or null if it was not found in this group
     */
    public NVSLSpeaker getSpeaker(String speakerId)
    {
        Checker.checkStringArgForNullOrEmpty("speakerId", speakerId);
        //empty list
        if (_speakers.size() == 0) return null;

        for (NVSLSpeaker sp: _speakers)
        {
            if ( sp.getSpeakerId().equals(speakerId))
            {
                return sp;
            }
        }
        return null;
    }

    /**
     * Removes a speaker from this group.
     *
     * @param speakerId the unique string identifying the speaker
     * @param fileManager The file manager which helps to look for files.
     * @return true if removal was successful
     */
    public boolean removeSpeaker(FileManager fileManager, String speakerId)
    {
        //empty list
        if(_speakers.size() == 0) return false;

        NVSLSpeaker sp = getSpeaker(speakerId);
        if(sp == null) return false;

        sp.removeAllVoiceprints(fileManager);

        return _speakers.remove(sp);
    }

    /**
     * Removes all attached speakers from the group.
     *
     * @param fileManager The file manager which helps to look for files.
     */
    public void removeAllSpeakers(FileManager fileManager)
    {
        while (_speakers.size() > 0) {
            NVSLSpeaker sp = _speakers.get(0);
            removeSpeaker(fileManager, sp.getSpeakerId());
        }
    }

    /**
     * Returns the list of speakers attached to this group.
     * @return the list of speakers
     */
    public List<NVSLSpeaker> getSpeakers()
    {
        return _speakers;
    }

    private void addSpeaker(NVSLSpeaker speaker)
    {
        _speakers.add(speaker);
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result
                + ((_groupId == null) ? 0 : _groupId.hashCode());
        result = prime * result
                + ((_speakers == null) ? 0 : _speakers.hashCode());

        return result;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (!(obj instanceof NVSLGroup))
            return false;
        NVSLGroup other = (NVSLGroup) obj;
        if (_groupId == null) {
            if (other._groupId != null)
                return false;
        } else if (!_groupId.equals(other._groupId))
            return false;
        if (_speakers == null) {
            if (other._speakers != null)
                return false;
        } else if (!_speakers.equals(other._speakers))
            return false;

        return true;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();

        o.tryPut(KEY_GROUP_ID, _groupId);

        org.json.JSONArray speakers = new org.json.JSONArray();
        for (NVSLSpeaker sp : _speakers)
        {
            speakers.put(sp.toJSON());
        }
        o.tryPut(KEY_SPEAKERS, speakers);

        return o;
    }

    public static NVSLGroup createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String groupId = json.getString(KEY_GROUP_ID);
        NVSLGroup group = new NVSLGroup(groupId);

        org.json.JSONArray speakers = json.getJSONArray(KEY_SPEAKERS);
        for (int i = 0; i < speakers.length(); i++)
        {
            group.addSpeaker(NVSLSpeaker.createFromJSON(speakers.getJSONObject(i)));
        }

        return group;
    }


}
