package com.nuance.dragon.toolkit.nvsl;

import java.util.List;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;


public abstract class NVSLIdentifier {

    /** Not for public use. */
    public static boolean LOAD_NATIVE_LIBRARY_SUCCESS = NativeNVSLImpl.LOAD_LIBRARY_SUCCESS;

    /**
     * Listens for results or errors from a NVSL identifier.
     */
    public interface ResultListener
    {
        /**
         * Results available
         *
         * @param result The results received
         */
        void onResult(NVSLResult result);

        /**
         * Error occurred
         *
         * @param error The error description
         */
        void onError(NVSLError error);
    }

    /**
     * Listens for results or errors from an enrollment/adaptation step.
     */
    public interface EnrollmentListener
    {
        /**
         * Sent by NVSL upon enrollment/adaptation step completion
         */
        void onComplete();

        /**
         * Error occurred
         *
         * @param error The error description
         */
        void onError(NVSLError error);
    }

    /**
     * Listens for audio signal updates during a NVSL identifier.
     */
    public interface SignalListener
    {
        /**
         * Update available
         *
         * @param audioInfo {@link NVSLAudioInfo}
         */
        void onUpdate(NVSLAudioInfo audioInfo);
    }

    /**
     * Listens for the completion of {@link NVSLIdentifier#initialize}.
     */
    public interface InitializeListener
    {
        /**
         * The initialization completed
         *
         * @param nvsl The identifier that was initialized.
         * @param success false if the initialization was unsuccessful
         */
        void onLoaded(NVSLIdentifier nvsl, boolean success);
    }

    /**
     * Listens for the completion of {@link NVSLIdentifier#release}.
     */
    public interface ReleaseListener
    {
        /**
         * The release completed
         * @param nvsl The identifier that was released
         */
        void onReleased(NVSLIdentifier nvsl);
    }

    /**
     * Listens for the operations of {@link NVSLGroup}.
     */
    public interface GroupListener
    {
        /**
         * The operation completed.
         * @param group The group on which the operation occurred
         * @param success false if the group is not created
         */
        void onDone(NVSLGroup group,  boolean success);
    }

    /**
     * Load NVSL. This must be called prior to doing anything with the NVSLIdentifier object.
     * Typically you'd call this right after constructing the NVSLIdentifier object. Note that
     * any input filename indicated in NVSLConfig must be available through the FileManager passed
     * in the factory creators {@link NVSLIdentifier#createNVSLIdentifier}.
     *
     * @param config The NVSL configuration
     * @param listener Notified after the load completes
     */
    public abstract void initialize(NVSLConfig config, InitializeListener listener);

    /**
     * Load NVSL. This must be called prior to doing anything with the NVSLIdentifier object.
     * Typically you'd call this right after constructing the NVSLIdentifier object. Note that
     * any input filename indicated in NVSLConfig must be available through the FileManager passed
     * in the factory creators {@link NVSLIdentifier#createNVSLIdentifier}.
     *
     * @param config The NVSL configuration
     */
    public abstract void initialize(NVSLConfig config);

    /**
     * Releases the resources created by the identifier.
     * This method should be called when the identifier is no longer needed. It is
     * a non-blocking call.
     *
     * @param listener A listener that confirms when the identifier is released
     */
    public abstract void release(ReleaseListener listener);

    /**
     * Releases the resources created by the identifier.
     * This method should be called when the identifier is no longer needed. It is
     * a non-blocking call.
     */
    public abstract void release();

    /**
     * Terminates the current identification.
     * NVSL will try to enroll/adapt/identify all the audio received up until this method is called.
     * The resultListener will be used to deliver the results.
     */
    public abstract void stopListening();


    /**
     * Cancels the current identification/enrollment/adaptation.
     * NVSL will ignore all received audio, and no results will be sent back.
     */
    public abstract void cancel();


    /**
     * Initiates a enrollment.
     * <p>
     * In Text-Dependent mode, in order to get a relatively high and stable identified score,
     * application can issue several times of enrollment for the same speaker. A voiceprint
     * will be created accumulatively based on audio data of every time.
     *
     * @param audioSource The audioSource that provides the audio.
     * @param voiceprint NVSLVoiceprint includes voice tag to be registered as native voiceprint
     * @param signalListener A listener that provides the intermediate audio signal updates, or the warning description
     *           if a warning occurred.
     * @param enrollmentListener Listener to be notified when enrollment step is
     *           complete.
     */
    @Deprecated
    public abstract void startEnrollment(AudioSource<AudioChunk> audioSource,
                                         NVSLVoiceprint voiceprint,
                                         SignalListener signalListener,
                                         EnrollmentListener enrollmentListener);

    /**
     * Initiates a enrollment.
     * <p>
     * In Text-Dependent mode, in order to get a relatively high and stable identified score,
     * application can issue several times of enrollment for the same speaker. A voiceprint
     * will be created accumulatively based on audio data of every time.
     *
     * @param audioSource The audioSource that provides the audio.
     * @param voiceprint NVSLVoiceprint includes voice tag to be registered as native voiceprint
     * @param enrollParam
     * @param signalListener A listener that provides the intermediate audio signal updates, or the warning description
     *           if a warning occurred.
     * @param enrollmentListener Listener to be notified when enrollment step is
     *           complete.
     */
    public abstract void startEnrollment(AudioSource<AudioChunk> audioSource,
                                         NVSLVoiceprint voiceprint,
                                         EnrollParam enrollParam,
                                         SignalListener signalListener,
                                         EnrollmentListener enrollmentListener);


    /**
     * Initiates an adaptation on an existing voiceprint.
     * <p>
     * In order to get a relatively high and stable identified score, application can
     * initiate an adaptation for the same speaker.
     *
     * @param audioSource The audioSource that provides the audio.
     * @param voiceprint NVSLVoiceprint includes voice tag to be adapted as native voiceprint
     * @param signalListener A listener that provides the intermediate audio signal updates, or the warning description
     *           if a warning occurred.
     * @param adaptationListener Listener to be notified when adaptation step is
     *           complete.
     */
    public abstract void startAdaptation(AudioSource<AudioChunk> audioSource,
                                         NVSLVoiceprint voiceprint,
                                         SignalListener signalListener,
                                         EnrollmentListener adaptationListener);


    /**
     * @deprecated Initiates an identification.
     * @param audioSource The audioSource that provides the audio.
     * @param voiceprints The list of NVSLVoiceprint to be used in identification.
     * @param signalListener A listener that provides the intermediate audio signal updates, or the warning description
     *          if a warning occurred.
     *
     * @param resultListener A listener that provides the results, or the error description
     *          if an error occurred.
     */
    @Deprecated
    public abstract void startIdentification(AudioSource<AudioChunk> audioSource,
                                              List<NVSLVoiceprint> voiceprints,
                                              SignalListener signalListener,
                                              ResultListener resultListener);

    /**
     * @deprecated Initiates an identification.
     * @param audioSource The audioSource that provides the audio.
     * @param group The group against whose speakers' voiceprints, identification is performing
     * @param signalListener A listener that provides the intermediate audio signal updates, or the warning description
     *          if a warning occurred.
     *
     * @param resultListener A listener that provides the results, or the error description
     *          if an error occurred.
     */
    @Deprecated
    public abstract void startIdentification(AudioSource<AudioChunk> audioSource,
                                              NVSLGroup group,
                                              SignalListener signalListener,
                                              ResultListener resultListener);

    /**
     * Initiates an identification.
     * @param audioSource The audioSource that provides the audio.
     * @param voiceprints The list of NVSLVoiceprint to be used in identification.
     * @param identifyParam The identification parameter, see {@link IdentifyParam}
     * @param signalListener A listener that provides the intermediate audio signal updates, or the warning description
     *          if a warning occurred.
     *
     * @param resultListener A listener that provides the results, or the error description
     *          if an error occurred.
     */
    public abstract void startIdentification(AudioSource<AudioChunk> audioSource,
                                              List<NVSLVoiceprint> voiceprints,
                                              IdentifyParam identifyParam,
                                              SignalListener signalListener,
                                              ResultListener resultListener);

    /**
     * Initiates an identification.
     * @param audioSource The audioSource that provides the audio.
     * @param group The group against whose speakers' voiceprints, identification is performing
     * @param identifyParam The identification parameter, see {@link IdentifyParam}
     * @param signalListener A listener that provides the intermediate audio signal updates, or the warning description
     *          if a warning occurred.
     *
     * @param resultListener A listener that provides the results, or the error description
     *          if an error occurred.
     */
    public abstract void startIdentification(AudioSource<AudioChunk> audioSource,
                                              NVSLGroup group,
                                              IdentifyParam identifyParam,
                                              SignalListener signalListener,
                                              ResultListener resultListener);

    /**
     * Initiates a language detection.
     * <p>
     * The duration of speech in the audio is a primary factor in accuracy. Nuance recommends a minimum of 30 seconds.
     *
     * @param audioSource The audioSource that provides the audio.
     * @param signalListener A listener that provides the intermediate audio signal updates, or the warning description
     *          if a warning occurred.
     * @param resultListener A listener that provides the results, or the error description
     *          if an error occurred.
     */
    public abstract void startLanguageDetection(AudioSource<AudioChunk> audioSource,
                                                  SignalListener signalListener,
                                                  ResultListener resultListener);

    /**
     * Creates a group of speakers. If there is a saved metadata for the group Id,
     * load the data to create a group.
     *
     * @param groupId the unique identifier of the group.
     * @param listener Listener to be notified of the creation.
     */
    public abstract void createGroup(String groupId, GroupListener listener);

    /**
     * Deletes a group of speakers
     * @param group the group to be deleted
     * @param listener Listener to be notified of the deletion.
     */
    public abstract void deleteGroup(NVSLGroup group, GroupListener listener);

    /**
     * Saves a group of speakers
     * @param group the group to be saved
     * @param listener Listener to be notified of the save.
     */
    public abstract void saveGroup(NVSLGroup group, GroupListener listener);

    /**
     * This function controls the amount of information about the NVSLIdentifier
     * that will be written to the Android logs.
     *
     * Note: Do not use / disabled for your application release build.
     *
     * @param verbose True if detailed (verbose) status information for the NVSLIdentifier should be written
     *                to the Android logs.
     *                False if only minimal details and error information should be written to  the Android logs.
     */
    public abstract void enableVerboseAndroidLogging(boolean verbose);

    /**
     * Create a {@link NVSLIdentifier}. The NVSLIdentifier will use the given worker thread.
     *
     * @param fileManager The file manager where NVSLIdentifier should look for files.
     * @param workerThread The worker thread to use. If null, a dedicated one
     *          will be created.
     */
    public static NVSLIdentifier createNVSLIdentifier(FileManager fileManager, NMTHandler workerThread)
    {
        return new NVSLIdentifierImpl(fileManager, workerThread);
    }

    /**
     * Create a {@link NVSLIdentifier}. The NVSLIdentifier will create its own dedicated worker
     * thread.
     *
     * @param fileManager The file manager where NVSLIdentifier should look for files.
     */
    public static NVSLIdentifier createNVSLIdentifier(FileManager fileManager)
    {
        return new NVSLIdentifierImpl(fileManager);
    }
}
