package com.nuance.dragon.toolkit.nvsl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.regex.Pattern;

import org.json.JSONObject;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.sources.BurstFileRecorderSource;
import com.nuance.dragon.toolkit.nvsl.NativeNVSLImpl.TaskType;
import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.util.JSONUtils;
import com.nuance.dragon.toolkit.util.internal.Pair;
import com.nuance.dragon.toolkit.util.internal.ThreadUtils;

/**
 * Base class for NVSL implementations.
 * <p>
 * There will probably be one 'real' subclass of this ({@link NVSLIdentifierImpl}).
 */
abstract class NVSLIdentifierBase extends NVSLIdentifier
{
    private final NativeNVSL _nativeNVSL;
    private Identification _currentIdentification;
    private boolean _isInitialized;
    private boolean _isNVSLCreated;
    private boolean _isReleased;
    private String _confSetName;
    private boolean _isLangIDSupported;
    private EnrollmentMode _eMode;
    private final FileManager _fileManager;
    private NVSLGroup _currentGroup;
    private AudioBounds _audioCheckBounds;
    private final NMTHandler _workerThreadHandler;
    private final NMTHandler _mainThreadHandler;

    private interface NVSLInnerListener
    {
        void onAudioCheckComplete(NVSLVoiceprint voiceprint, Pair<String, String> audioFilePair);

        void onConsistencyCheckComplete(NVSLVoiceprint voiceprint, Pair<String, String> audioFilePair);
    }

    private enum EnrollmentMode
    {
        TI_MODE, TD_MODE
    };

    private class AudioBounds
    {
        final float minAudioLength;
        final float maxAudioLength;
        final float minSpeechLength;
        final float maxSpeechLength;
        final float audioNoisyThreshold;
        final float audioSoftThreshold;
        final float audioLoudThreshold;

        AudioBounds(float audioMin, float audioMax, float speechMin, float speechMax, float snr, float speech, float saturation)
        {
            minAudioLength = audioMin;
            maxAudioLength = audioMax;
            minSpeechLength = speechMin;
            maxSpeechLength = speechMax;
            audioNoisyThreshold = snr;
            audioSoftThreshold = speech;
            audioLoudThreshold = saturation;
        }
    }

    private class NoiseSumParam
    {
        final String noisyFilename;
        final int rand;
        final int degradationLevel;
        final String audioNoiseDumpFile;

        NoiseSumParam(String noisyFilename, int rand, int degradationLevel, String audioNoiseDumpFile)
        {
            this.noisyFilename = noisyFilename;
            this.rand = rand;
            this.degradationLevel = degradationLevel;
            this.audioNoiseDumpFile = audioNoiseDumpFile;
        }
    }

    private class Identification extends AudioSink<AudioChunk>
    {
        private ResultListener _resultListener;
        private final SignalListener _signalListener;
        private EnrollmentListener _enrollmentListener;
        private final LinkedBlockingQueue<AudioChunk> _audioQueue; // Touched from both threads
        private AudioChunk _currentAudioChunk;
        private int _currentAudioChunkOffset;
        private boolean _started; // Touched from both threads
        private boolean _isAdaptation = false;
        private BurstFileRecorderSource _fileAudioSource;
        //private String _tempAudioFileName;
        private int _errCode = -1;

        Identification(ResultListener resultListener, SignalListener signalListener, EnrollmentListener enrollmentListener)
        {
            super(null);
            _resultListener = resultListener;
            _signalListener = signalListener;
            _enrollmentListener = enrollmentListener;
            _audioQueue = new LinkedBlockingQueue<AudioChunk>();
        }

        void resetToStart()
        {
            _started = true;
            _audioQueue.clear();
            _currentAudioChunk = null;
            _currentAudioChunkOffset = 0;
        }

        void startIdentifying(final AudioSource<AudioChunk> audioSource, final List<NVSLVoiceprint> voiceprints, final IdentifyParam identifyParam, final boolean isLangID)
        {
            if (isLangID)
            {
                if(!_isLangIDSupported)
                {
                    handleError(new NVSLError(NVSLError.Reason.LANGID_NOT_SUPPORTED));
                    return;
                }
            }

            resetToStart();
            connectAudioSource(audioSource);

            _workerThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    NativeNVSL.NVSLAudioListener audioListener = new NativeNVSL.NVSLAudioListener()
                    {
                        @Override
                        public void onStartAudioPull(String audioFileName)
                        {
                        }

                        @Override
                        public void onStopAudioPull(String audioFileName)
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    disconnectAudioSource();
                                }
                            });
                        }

                        @Override
                        public void onSignalUpdate(AudioInfoJni audioInfo)
                        {
                            if (_signalListener != null) {
                                final NVSLAudioInfo userAudioInfo = audioInfo.createAudioInfo();
                                _mainThreadHandler.post(new Runnable()
                                {
                                    @Override
                                    public void run()
                                    {
                                        _signalListener.onUpdate(userAudioInfo);
                                    }
                                });
                            }
                        }

                        @Override
                        public AudioChunk onGetAudioChunk(int maxLen)
                        {
                            long frameShiftMs = audioSource.getAudioType().getDuration(maxLen);
                            if (poll(frameShiftMs))
                                return getNextAudioChunk(maxLen);
                            return null;
                        }

                        @Override
                        public void onCanceled()
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    disconnectAudioSource();
                                }
                            });
                        }
                    };

                    NativeNVSL.NVSLResultListener resultListener = new NativeNVSL.NVSLResultListener()
                    {

                        @Override
                        public void onIdentificationResult(ResultJni result) {
                            final NVSLResult finalResult = result.createNVSLResult(_confSetName);
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (finalResult != null)
                                        handleResult(finalResult);
                                    else
                                        handleError(new NVSLError(NVSLError.Reason.NO_RESULT));
                                }
                            });
                        }

                    };

                    if (isLangID) {
                        if(!_nativeNVSL.startDetectingLanguage(audioSource.getAudioType().frequency, audioListener, resultListener))
                        {
                            Logger.error(this, "Unable to start Detection.");
                            handleStartError();
                            return;
                        }

                    } else {
                        // collect voiceprint file paths
                        List<String> vpFilePaths = new ArrayList<String>();
                        if (!voiceprints.isEmpty())
                        {
                            for (NVSLVoiceprint voiceprint : voiceprints)
                            {
                                if (null == voiceprint.getConfigSetName() || voiceprint.getConfigSetName().equals(""))
                                    voiceprint.setConfigSetName(_confSetName);
                                if (!_confSetName.equals(voiceprint.getConfigSetName()))
                                    continue;

                                vpFilePaths.add(voiceprint.getVpFileName());
                            }
                        }
                        else
                        {
                            String[] vpFileNames = _fileManager.findFiles(Pattern.compile("(.*).svm"));
                            if (vpFileNames != null)
                            {
                                for (String fileName : vpFileNames)
                                {
                                    vpFilePaths.add(fileName);
                                }
                            }
                        }
                        if(!_nativeNVSL.startIdentifying(vpFilePaths, audioSource.getAudioType().frequency, identifyParam.getMaxResults(), identifyParam.isMultiIdentEnabled(), identifyParam.getPartialResultTimerMs(), audioListener, resultListener))
                        {
                            Logger.error(this, "Unable to start Identifier.");
                            handleStartError();
                            return;
                        }
                    }
                }
            });
        }

        NVSLInnerListener _innerListener = new NVSLInnerListener()
        {
            @Override
            public void onAudioCheckComplete(final NVSLVoiceprint voiceprint, Pair<String, String> audioFilePair)
            {
                if (Identification.this == _currentIdentification)
                {
                    if (_eMode == EnrollmentMode.TI_MODE)
                    {
                        // TI mode doesn't accumulate audio segs
                        voiceprint.clearAllAudioSegs(_fileManager);
                        voiceprint.addAudioSegFile(audioFilePair, voiceprint.audioType);
                        _currentIdentification.startEnrollingInternal(voiceprint, voiceprint.isAdaptation());
                    }
                    else
                    {
                        _currentIdentification.startConsistencyCheck(voiceprint, audioFilePair, _innerListener);
                    }
                }
                else if (_enrollmentListener != null)
                {
                    _enrollmentListener.onError(new NVSLError(NVSLError.Reason.AUDIO_CHECK_ERROR));
                }
            }

            @Override
            public void onConsistencyCheckComplete(NVSLVoiceprint voiceprint, Pair<String, String> audioFilePair)
            {
                voiceprint.addAudioSegFile(audioFilePair, voiceprint.audioType);
                Logger.debug(this, "audio consistency check ok, file: [" + audioFilePair + "] added to voiceprint" );
                if (Identification.this == _currentIdentification)
                    _currentIdentification.startEnrollingInternal(voiceprint, voiceprint.isAdaptation());
                else if (_enrollmentListener != null)
                {
                    _enrollmentListener.onError(new NVSLError(NVSLError.Reason.AUDIO_INCONSISTENT));
                }
            }
        };

        void startEnrolling(AudioSource<AudioChunk> audioSource, NVSLVoiceprint voiceprint, EnrollParam enrollParam)
        {
            _isAdaptation = true;
            if (!voiceprint.checkAudioConsistent(audioSource.getAudioType()))
            {
                handleError(new NVSLError(NVSLError.Reason.CANT_START));
                return;
            }
            voiceprint.setAdaptation(false);

            String groupDir = (null == _currentGroup) ? "" : (_currentGroup.getGroupId() + File.separator);
            Pair<String, String> audioDumpFilePair = voiceprint.generateVpAudioSegFileName();
            String audioDumpFile = groupDir + audioDumpFilePair.first;
            NoiseSumParam noiseSumParam = null;
            if (!enrollParam.getNoisyFiles().isEmpty())
            {
                int numOfFiles = enrollParam.getNoisyFiles().size();
                int rand = new Random().nextInt();
                String noiseInFile = enrollParam.getNoisyFiles().get(Math.abs(rand % numOfFiles));
                String audioNoiseDumpFile = groupDir + audioDumpFilePair.second;
                noiseSumParam =  new NoiseSumParam(noiseInFile, rand, enrollParam.getDegradationLevel(), audioNoiseDumpFile);
            }
            _currentIdentification.startAudioCheck(audioSource, voiceprint, audioDumpFile, TaskType.SPK_ENR_TASK, noiseSumParam, _innerListener);
        }

        void startAdaptation(AudioSource<AudioChunk> audioSource, NVSLVoiceprint voiceprint)
        {
            _isAdaptation = true;
            if (!voiceprint.checkAudioConsistent(audioSource.getAudioType()))
            {
                handleError(new NVSLError(NVSLError.Reason.CANT_START));
                return;
            }
            voiceprint.setAdaptation(true);

            String groupDir = (null == _currentGroup) ? "" : (_currentGroup.getGroupId() + File.separator);
            Pair<String, String> audioDumpFilePair = voiceprint.generateVpAudioSegFileName();
            String audioDumpFile = groupDir + audioDumpFilePair.first;
            _currentIdentification.startAudioCheck(audioSource, voiceprint, audioDumpFile, TaskType.SPK_ENR_TASK, null, _innerListener);
        }

        /*
         * start enrollment by using audio segment files associated with voiceprint
         */
        private void startEnrollingInternal(final NVSLVoiceprint voiceprint, final boolean isAdaption)
        {
            if ((voiceprint.getConfigSetName() != null && !voiceprint.getConfigSetName().equals(""))
                && !_confSetName.equals(voiceprint.getConfigSetName()))
            {
                handleError(new NVSLError(NVSLError.Reason.CONFIG_NOT_COMPATIBLE));
                return;
            }

            resetToStart();

            _workerThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    if (null == voiceprint.getConfigSetName() || voiceprint.getConfigSetName().equals(""))
                        voiceprint.setConfigSetName(_confSetName);
                    String vpFileName = voiceprint.generateVpFileName();
                    String groupDir = (null == _currentGroup) ? "" : (_currentGroup.getGroupId() + File.separator);
                    final String vpFilePath =  groupDir + vpFileName;
                    voiceprint.setVpFileName(vpFilePath);

                    NativeNVSL.NVSLAudioListener audioListener = new NativeNVSL.NVSLAudioListener()
                    {
                        @Override
                        public void onStartAudioPull(final String audioFileName)
                        {
                            // onStartAudioPull will be called once for each file source, so we need to reset for each new file
                            resetToStart();
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    //_tempAudioFileName = audioFileName;
                                    _fileAudioSource = new BurstFileRecorderSource(voiceprint.audioType, audioFileName, _fileManager, true);
                                    _fileAudioSource.startRecording();
                                    connectAudioSource(_fileAudioSource);
                                }
                            });
                        }

                        @Override
                        public void onStopAudioPull(final String audioFileName)
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    disconnectAudioSource();
                                    if (_fileAudioSource != null)
                                    {
                                        _fileAudioSource.stopRecording();
                                        _fileAudioSource = null;
                                    }
                                }
                            });
                        }

                        @Override
                        public void onSignalUpdate(AudioInfoJni audioInfo)
                        {
                            // do not call onUpdate callback, onUpdate is called during audio check
                        }

                        @Override
                        public AudioChunk onGetAudioChunk(int maxLen)
                        {
                            long frameShiftMs;
                            frameShiftMs = voiceprint.audioType.getDuration(maxLen);

                            if (poll(frameShiftMs))
                                return getNextAudioChunk(maxLen);

                            return null;
                        }

                        @Override
                        public void onCanceled()
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    disconnectAudioSource();
                                    if (_fileAudioSource != null)
                                    {
                                        _fileAudioSource.stopRecording();
                                        _fileAudioSource = null;
                                    }
                                }
                            });
                        }
                    };

                    boolean succeed = false;
                    if (isAdaption)
                    {
                        // Start Adaption.
                        succeed = _nativeNVSL.startAdapting(vpFilePath, voiceprint.getAudioSegFile(voiceprint.getNumOfAudioSegFiles()-1).first,
                                                            voiceprint.audioType.frequency, audioListener);
                    }
                    else
                    {
                        List<String> audioFilesToEnroll = new ArrayList<String>();
                        audioFilesToEnroll.addAll(voiceprint.getAudioSegFiles());
                        audioFilesToEnroll.addAll(voiceprint.getAudioNoiseFiles());
                        succeed = _nativeNVSL.startEnrolling(vpFilePath, audioFilesToEnroll, voiceprint.audioType.frequency,
                                                            audioListener);
                    }

                    if (!succeed)
                    {
                        Logger.error(this, "Unable to start enrollment/adaption.");
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                //voiceprint.removeAudioFile(_fileManager, _tempAudioFileName);
                                stop(true);
                                handleError(new NVSLError(NVSLError.Reason.CANT_START));
                            }
                        });
                        return;
                    }
                    else
                    {
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                handleAdaptationStep();
                            }
                        });
                    }
                }
            });
        }

        private void startAudioCheck(final AudioSource<AudioChunk> audioSource,
                                     final NVSLVoiceprint voiceprint,
                                     final String audioDumpFile,
                                     final int taskType,
                                     final NoiseSumParam noiseSumParam,
                                     final NVSLInnerListener innerListener)
        {
            final boolean isNoiseToAdd = (noiseSumParam == null) ? false : true;
            resetToStart();
            connectAudioSource(audioSource);

            _workerThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    NativeNVSL.NVSLAudioListener audioListener = new NativeNVSL.NVSLAudioListener()
                    {
                        @Override
                        public void onStartAudioPull(String audioFileName)
                        {
                        }

                        @Override
                        public void onStopAudioPull(String audioFileName)
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    disconnectAudioSource();
                                }
                            });
                        }

                        @Override
                        public void onSignalUpdate(AudioInfoJni audioInfo)
                        {
                            if (_signalListener != null) {
                                final NVSLAudioInfo userAudioInfo = audioInfo.createAudioInfo();
                                if (audioInfo.isAudioEnded)
                                {
                                    if (userAudioInfo.processedAudio < _audioCheckBounds.minAudioLength)
                                        _errCode = NVSLError.Reason.AUDIO_TOO_SHORT;
                                    else if (userAudioInfo.processedAudio > _audioCheckBounds.maxAudioLength)
                                        _errCode = NVSLError.Reason.AUDIO_TOO_LONG;
                                    else if (userAudioInfo.processedSpeech < _audioCheckBounds.minSpeechLength)
                                        _errCode = NVSLError.Reason.SPEECH_TOO_SHORT;
                                    else if (userAudioInfo.processedSpeech > _audioCheckBounds.maxSpeechLength)
                                        _errCode = NVSLError.Reason.SPEECH_TOO_LONG;
                                    else if (userAudioInfo.snr < _audioCheckBounds.audioNoisyThreshold)
                                        _errCode = NVSLError.Reason.AUDIO_TOO_NOISY;
                                    else if (userAudioInfo.speechLevel < _audioCheckBounds.audioSoftThreshold)
                                        _errCode = NVSLError.Reason.AUDIO_TOO_SOFT;
                                    else if (userAudioInfo.saturationLevel > _audioCheckBounds.audioLoudThreshold)
                                        _errCode = NVSLError.Reason.AUDIO_TOO_LOUD;
                                }
                                else
                                {
                                    _mainThreadHandler.post(new Runnable()
                                    {
                                        @Override
                                        public void run()
                                        {
                                            _signalListener.onUpdate(userAudioInfo);
                                        }
                                    });
                                }
                            }
                        }

                        @Override
                        public AudioChunk onGetAudioChunk(int maxLen)
                        {
                            long frameShiftMs = audioSource.getAudioType().getDuration(maxLen);
                            if (poll(frameShiftMs))
                                return getNextAudioChunk(maxLen);
                            return null;
                        }

                        @Override
                        public void onCanceled()
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    disconnectAudioSource();
                                }
                            });
                        }
                    };

                    boolean isSuccesful = isNoiseToAdd ? _nativeNVSL.startAudioCheck(audioSource.getAudioType().frequency, taskType, _fileManager.getAbsFilePathForWriting(audioDumpFile),
                                                                                     _fileManager.getAbsFilePathForReading(noiseSumParam.noisyFilename),
                                                                                     noiseSumParam.rand,
                                                                                     noiseSumParam.degradationLevel,
                                                                                     _fileManager.getAbsFilePathForWriting(noiseSumParam.audioNoiseDumpFile),
                                                                                     audioListener) :
                                                         _nativeNVSL.startAudioCheck(audioSource.getAudioType().frequency, taskType, _fileManager.getAbsFilePathForWriting(audioDumpFile), audioListener);

                    if(!isSuccesful || _errCode > 0)
                    {

                        _fileManager.delete(audioDumpFile);
                        if (isNoiseToAdd) _fileManager.delete(noiseSumParam.audioNoiseDumpFile);
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                if (_errCode > 0)
                                {
                                    NVSLError audioErr = new NVSLError(_errCode);
                                    Logger.error(this, "audio check error: " + audioErr.getReason());
                                    handleError(audioErr);
                                }
                                else
                                {
                                    Logger.error(this, "Unable to start audio check.");
                                    handleError(new NVSLError(NVSLError.Reason.AUDIO_CHECK_ERROR));
                                }

                            }
                        });
                        return;
                    }
                    else
                    {
                        if (voiceprint.getNumOfAudioSegFiles() >= voiceprint.getMaxAudioSegNum())
                        {
                            Logger.error(this, "Registered audio files in voiceprint: " + voiceprint.getVoiceTag() + " are full");
                            _fileManager.delete(audioDumpFile);
                            if (isNoiseToAdd) _fileManager.delete(noiseSumParam.audioNoiseDumpFile);
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    handleError(new NVSLError(NVSLError.Reason.MAX_AUDIO_NUM));
                                }
                            });
                        }
                        else
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (innerListener != null)
                                    {
                                        innerListener.onAudioCheckComplete(voiceprint, new Pair<String, String>(audioDumpFile, isNoiseToAdd? noiseSumParam.audioNoiseDumpFile : null));
                                    }
                                }
                            });
                        }
                    }
                }
            });
        }

        private void startConsistencyCheck(final NVSLVoiceprint voiceprint, final Pair<String, String> audioFilePair, final NVSLInnerListener innerListener)
        {
            resetToStart();
            _workerThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    NativeNVSL.NVSLAudioListener audioListener = new NativeNVSL.NVSLAudioListener()
                    {
                        @Override
                        public void onStartAudioPull(final String audioFileName)
                        {
                            // onStartAudioPull will be called once for each file source, so we need to reset for each new file
                            resetToStart();
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    _fileAudioSource = new BurstFileRecorderSource(voiceprint.audioType, audioFileName, _fileManager, true);
                                    _fileAudioSource.startRecording();
                                    connectAudioSource(_fileAudioSource);
                                }
                            });
                        }

                        @Override
                        public void onStopAudioPull(String audioFileName)
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (_fileAudioSource != null)
                                    {
                                        _fileAudioSource.stopRecording();
                                        _fileAudioSource = null;
                                    }
                                    disconnectAudioSource();
                                }
                            });
                        }

                        @Override
                        public void onSignalUpdate(AudioInfoJni audioInfo)
                        {
                            // do not call onUpdate callback in consistency check
                        }

                        @Override
                        public AudioChunk onGetAudioChunk(int maxLen)
                        {
                            long frameShiftMs = voiceprint.audioType.getDuration(maxLen);
                            if (poll(frameShiftMs))
                                return getNextAudioChunk(maxLen);
                            return null;
                        }

                        @Override
                        public void onCanceled()
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (_fileAudioSource != null)
                                    {
                                        _fileAudioSource.stopRecording();
                                        _fileAudioSource = null;
                                    }
                                    disconnectAudioSource();
                                }
                            });
                        }
                    };

                    // only check the last audio file in the list of voiceprint and current audio file
                    List<String> fileNames = new ArrayList<String>();
                    int nCount = voiceprint.getNumOfAudioSegFiles();
                    if (nCount > 0)
                    {
                        fileNames.add(voiceprint.getAudioSegFile(nCount - 1).first);
                    }
                    fileNames.add(audioFilePair.first);
                    if(!_nativeNVSL.startCheckConsistency(fileNames, voiceprint.audioType.frequency, audioListener))
                    {
                        Logger.error(this, "Unable to start audio consistency check.");
                        _fileManager.delete(audioFilePair.first);
                        if (audioFilePair.second != null) _fileManager.delete(audioFilePair.second);
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                handleError(new NVSLError(NVSLError.Reason.AUDIO_INCONSISTENT));
                            }
                        });
                        return;
                    }
                    else
                    {
                      //audio files in voiceprint and current file are consistent
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                if (innerListener != null)
                                {
                                    innerListener.onConsistencyCheckComplete(voiceprint, audioFilePair);
                                }
                            }
                        });
                    }
                }
            });
        }

        void stop(final boolean abort)
        {
            _started = false;
            if (abort)
                _nativeNVSL.cancel();
        }

        void cancel()
        {
            stop(true);
            handleError(new NVSLError(NVSLError.Reason.CANCELED));
        }

        /** Called from worker thread */
        private void handleStartError()
        {
            _mainThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    stop(true);
                    handleError(new NVSLError(NVSLError.Reason.CANT_START));
                }
            });
        }

        /** Called from main thread */
        private void handleResult(NVSLResult result)
        {
            if (_currentIdentification == Identification.this)
            {

                if (_resultListener != null)
                {
                    _resultListener.onResult(result);
                }
                if (result.isFinal())
                {
                    _resultListener = null;
                    _currentIdentification = null;
                }
            }
        }

        /** Called from main thread */
        private void handleError(NVSLError error)
        {
            if (_currentIdentification == Identification.this)
            {
                _currentIdentification = null;
                if (_isAdaptation)
                {
                    if (_enrollmentListener != null)
                    {
                        _enrollmentListener.onError(error);
                        _enrollmentListener = null;
                    }
                }
                else
                {
                    if (_resultListener != null)
                    {
                        _resultListener.onError(error);
                        _resultListener = null;
                    }
                }
            }
        }

        /** Called from main thread */
        private void handleAdaptationStep()
        {
            if (_currentIdentification == Identification.this)
            {
                _currentIdentification = null;
                if (_enrollmentListener != null)
                {
                    _enrollmentListener.onComplete();
                    _enrollmentListener = null;
                }
            }
        }


        /** Called from worker thread */
        private boolean poll(long periodMs)
        {
            // Poll until we see a chunk or the audio source disconnects
            //Logger.debug(this, "poll status: start: " + _started + ", current audio chunk=" + _currentAudioChunk + ", is audio queue empty: " + _audioQueue.isEmpty());
            while (_started &&
                   _currentAudioChunk == null &&
                   _audioQueue.isEmpty())
            {
                 ThreadUtils.sleep(periodMs);
            }

            // Return true if there is audio available, else false.
            return _currentAudioChunk != null || _audioQueue.size() > 0;
        }

        /** Called from worker thread */
        private AudioChunk getNextAudioChunk(final int requiredChunkLen)
        {
            if (_currentAudioChunk != null)
            {
                short[] frames = null;
                if (_currentAudioChunkOffset + requiredChunkLen < _currentAudioChunk.audioShorts.length)
                {
                    // Get a full sub-chunk out of the cached chunk
                    frames = new short[requiredChunkLen];
                    System.arraycopy(_currentAudioChunk.audioShorts, _currentAudioChunkOffset, frames, 0, frames.length);
                    _currentAudioChunkOffset = _currentAudioChunkOffset + frames.length;
                    return new AudioChunk(_currentAudioChunk.audioType, frames);
                }
                else
                {
                    if (_currentAudioChunkOffset < _currentAudioChunk.audioShorts.length)
                    {
                        // Get only what's left in the cached chunk
                        frames = new short[_currentAudioChunk.audioShorts.length - _currentAudioChunkOffset];
                        System.arraycopy(_currentAudioChunk.audioShorts, _currentAudioChunkOffset, frames, 0, frames.length);
                        AudioChunk leftover = new AudioChunk(_currentAudioChunk.audioType, frames);
                        _currentAudioChunk = null;
                        _currentAudioChunkOffset = 0;
                        return leftover;
                    }
                }
            }

            // Caching next chunk
            _currentAudioChunkOffset = 0;
            _currentAudioChunk = _audioQueue.peek();
            if (_currentAudioChunk != null)
            {
                _audioQueue.remove();
                return getNextAudioChunk(requiredChunkLen);
            }
            return null;
        }

        private void handleNewAudio(AudioSource<AudioChunk> source)
        {
            if (this != _currentIdentification)
            {
                Logger.warn(this, "Got audio for old identification/enrollment/adaption");
                return;
            }

            _audioQueue.addAll(source.getAllAudioChunksForSink(this));
        }

        @Override
        public void chunksAvailable(AudioSource<AudioChunk> source)
        {
            handleNewAudio(source);
        }

        @Override
        public void framesDropped(AudioSource<AudioChunk> source)
        {
            if (this != _currentIdentification)
                return;

            handleError(new NVSLError(NVSLError.Reason.AUDIO_FRAMES_DROPPED));
        }

        @Override
        public void sourceClosed(AudioSource<AudioChunk> source)
        {
            if (this != _currentIdentification)
                return;

            stop(false);
        }
    }

    NVSLIdentifierBase(FileManager fileManager, NMTHandler externalWorkerThreadHandler)
    {
        _fileManager = fileManager;
        _nativeNVSL = new NativeNVSLImpl(_fileManager);

        _mainThreadHandler = Factory.createNMTHandler();
        if (externalWorkerThreadHandler == null)
            _workerThreadHandler = getWorkerThreadHandler();
        else
            _workerThreadHandler = externalWorkerThreadHandler;

    }

    @Override
    public void enableVerboseAndroidLogging(boolean verbose) {
        _nativeNVSL.enableVoconVerboseLogging(verbose);
    }

    @Override
    public void initialize(final NVSLConfig config, final InitializeListener listener) {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, !_isInitialized, "Already initialized");
        Checker.checkArgForNull("config", config);

        _isInitialized = true;
        _isLangIDSupported = config.isLangID();
        _eMode = config.isTIMode() ? EnrollmentMode.TI_MODE : EnrollmentMode.TD_MODE;
        _confSetName = config.getConfSetName();
        _audioCheckBounds = new AudioBounds(config.getMinAudioLength(),
                                            config.getMaxAudioLength(),
                                            config.getMinSpeechLength(),
                                            config.getMaxSpeechLength(),
                                            config.getAudioNoisyThreshold(),
                                            config.getAudioSoftThreshold(),
                                            config.getAudioLoudThreshold());
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final boolean success = _nativeNVSL.createNVSL(config);
                if (success)
                {
                    _isNVSLCreated = true;
                }

                if (listener != null) {

                    _mainThreadHandler.post(new Runnable() {
                        @Override
                        public void run()
                        {
                            listener.onLoaded(NVSLIdentifierBase.this, success);
                        }
                    });
                }
            }
        });
    }

    @Override
    public void initialize(final NVSLConfig config) {
        initialize(config, null);
    }

    @Override
    public void release(final ReleaseListener listener) {
        Checker.checkState(this, !_isReleased);
        if (!_isInitialized) {
            _isReleased = true;
            if (listener != null)
                listener.onReleased(this);
            return;
        }

        //terminateCurrentRecognition(); ?

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                // Destroy NVSL engine
                if (_isNVSLCreated) {
                    _nativeNVSL.destroyNVSL();
                }
                _isNVSLCreated = false;
                _isReleased = true;
                if (listener != null) {
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            listener.onReleased(NVSLIdentifierBase.this);
                        }
                    });
                }
            }
        });
    }

    @Override
    public void release() {
        release(null);
    }

    @Override
    public void stopListening()
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        if (_currentIdentification != null)
        {
            _currentIdentification.stop(false);
        }
    }

    @Override
    public void cancel()
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        terminateCurrentIdentification();
    }

    @Override
    public void startEnrollment(AudioSource<AudioChunk> audioSource,
                                NVSLVoiceprint voiceprint,
                                SignalListener signalListener,
                                EnrollmentListener enrollmentListener)
    {
        startEnrollment(audioSource, voiceprint, new EnrollParam.Builder().build(), signalListener, enrollmentListener);
    }

    @Override
    public void startEnrollment(AudioSource<AudioChunk> audioSource,
                                NVSLVoiceprint voiceprint,
                                EnrollParam enrollParam,
                                SignalListener signalListener,
                                EnrollmentListener enrollmentListener)
    {
        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForCondition("audioSource", "a supported audio type", audioSource.getAudioType().equals(AudioType.PCM_8k) || audioSource.getAudioType().equals(AudioType.PCM_16k));
        Checker.checkArgForNull("voiceprint", voiceprint);
        Checker.checkArgForNull("enrollParam", enrollParam);

        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancel();
        _currentIdentification = new Identification(null, signalListener, enrollmentListener);
        _currentIdentification.startEnrolling(audioSource, voiceprint, enrollParam);
    }

    @Override
    public void startAdaptation(AudioSource<AudioChunk> audioSource,
                                NVSLVoiceprint voiceprint,
                                SignalListener signalListener,
                                EnrollmentListener adaptationListener)
    {
        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForCondition("audioSource", "a supported audio type", audioSource.getAudioType().equals(AudioType.PCM_8k) || audioSource.getAudioType().equals(AudioType.PCM_16k));
        Checker.checkArgForNull("voiceprint", voiceprint);

        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);


        cancel();
        _currentIdentification = new Identification(null, signalListener, adaptationListener);
        _currentIdentification.startAdaptation(audioSource, voiceprint);
    }

    @Override
    public void startIdentification(AudioSource<AudioChunk> audioSource,
                                    List<NVSLVoiceprint> voiceprints,
                                     SignalListener signalListener,
                                     ResultListener resultListener)
    {
        startIdentification(audioSource, voiceprints, new IdentifyParam.Builder().build(), signalListener, resultListener);
    }

    @Override
    public void startIdentification(AudioSource<AudioChunk> audioSource,
                                    NVSLGroup group,
                                    SignalListener signalListener,
                                    ResultListener resultListener)
    {
        startIdentification(audioSource, group, new IdentifyParam.Builder().build(), signalListener, resultListener);
    }

    public void startIdentification(AudioSource<AudioChunk> audioSource,
                                    List<NVSLVoiceprint> voiceprints,
                                    IdentifyParam identifyParam,
                                    SignalListener signalListener,
                                    ResultListener resultListener)
    {
        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForCondition("audioSource", "a supported audio type", audioSource.getAudioType().equals(AudioType.PCM_8k) || audioSource.getAudioType().equals(AudioType.PCM_16k));
        Checker.checkArgForNull("voiceprints", voiceprints);
        Checker.checkArgForNull("identifyParam", identifyParam);
        Checker.checkArgForNull("resultListener", resultListener);

        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);


        cancel();
        _currentIdentification = new Identification(resultListener, signalListener, null);
        _currentIdentification.startIdentifying(audioSource, voiceprints, identifyParam, false);
    }

    public void startIdentification(AudioSource<AudioChunk> audioSource,
                                    NVSLGroup group,
                                    IdentifyParam identifyParam,
                                    SignalListener signalListener,
                                    ResultListener resultListener)
    {
        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForCondition("audioSource", "a supported audio type", audioSource.getAudioType().equals(AudioType.PCM_8k) || audioSource.getAudioType().equals(AudioType.PCM_16k));
        Checker.checkArgForNull("group", group);
        Checker.checkArgForNull("identifyParam", identifyParam);
        Checker.checkArgForNull("resultListener", resultListener);

        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);


        cancel();
        _currentIdentification = new Identification(resultListener, signalListener, null);

        List<NVSLVoiceprint> voiceprints = new ArrayList<NVSLVoiceprint>();
        for (NVSLSpeaker speaker : group.getSpeakers())
        {
            for (NVSLVoiceprint voiceprint : speaker.getVoiceprints())
            {
                voiceprints.add(voiceprint);
            }
        }
        _currentGroup = group;
        _currentIdentification.startIdentifying(audioSource, voiceprints, identifyParam, false);
    }

    @Override
    public void startLanguageDetection(AudioSource<AudioChunk> audioSource,
                                       SignalListener signalListener,
                                       ResultListener resultListener)
    {
        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForCondition("audioSource", "a supported audio type", audioSource.getAudioType().equals(AudioType.PCM_8k) || audioSource.getAudioType().equals(AudioType.PCM_16k));
        Checker.checkArgForNull("resultListener", resultListener);

        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancel();
        _currentIdentification = new Identification(resultListener, signalListener, null);
        _currentIdentification.startIdentifying(audioSource, null, null, true);
    }

    private void terminateCurrentIdentification()
    {
        if (_currentIdentification != null)
        {
            Identification r = _currentIdentification;
            _currentIdentification.cancel();
            if (r == _currentIdentification)
                _currentIdentification = null;
        }
    }

    @Override
    public void createGroup(final String groupId, final GroupListener listener)
    {
        Checker.checkStringArgForNullOrEmpty("groupId", groupId);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                String groupDir = groupId + File.separator;
                String metaFile =  groupDir + groupId + ".json";
                NVSLGroup group = null;
                if (_fileManager.exists(metaFile))
                {
                    FileInputStream s = _fileManager.openFileForReading(metaFile);
                    if (s != null)
                    {
                        try {
                            JSONObject json = JSONUtils.readFromStream(s);
                            s.close();
                            group = NVSLGroup.createFromJSON(json);
                        } catch (Exception e) {
                            Logger.error(this, "Error reading NVSL group meta file", e);
                        }
                    }
                }
                else
                {
                    group = new NVSLGroup(groupId);
                    _fileManager.createDir(group.getGroupId());
                }

                if (group != null)
                {
                    _currentGroup = group;
                    if (listener != null)
                    {
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                listener.onDone(_currentGroup, true);
                            }
                        });
                    }
                }
                else
                {
                    final NVSLGroup fakeGroup = new NVSLGroup(groupId);
                    if (listener != null)
                    {
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                listener.onDone(fakeGroup, false);
                            }
                        });
                    }
                }
            }
        });
    }

    @Override
    public void deleteGroup(final NVSLGroup group, final GroupListener listener)
    {
        Checker.checkArgForNull("group", group);

        group.removeAllSpeakers(_fileManager);
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                String groupDir = group.getGroupId() + File.separator;
                String metaFile =  groupDir + group.getGroupId() + ".json";
                if (_fileManager.exists(metaFile))
                    _fileManager.delete(metaFile);

                if (_fileManager.deleteDir(group.getGroupId()))
                {
                    if (listener != null)
                    {
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                if (_currentGroup == group)
                                    _currentGroup = null;
                                listener.onDone(group, true);
                            }
                        });
                    }
                }
                else
                {
                    if (listener != null)
                    {
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                listener.onDone(group, false);
                            }
                        });
                    }
                }
            }
        });
    }

    @Override
    public void saveGroup(final NVSLGroup group, final GroupListener listener)
    {
        Checker.checkArgForNull("group", group);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                String groupDir = group.getGroupId() + File.separator;
                String metaFile =  groupDir + group.getGroupId() + ".json";

                boolean success = false;
                FileOutputStream s = _fileManager.openFileForWriting(metaFile);
                if (s != null)
                {
                    try {
                        JSONUtils.saveToStream(group.toJSON(), s);
                        s.close();
                        success = true;
                    } catch (Exception e) {
                        Logger.error(this, "Error saving NVSL group", e);
                        success = false;
                    }
                }

                if (success)
                {
                    if (listener != null)
                    {
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                listener.onDone(group, true);
                            }
                        });
                    }
                }
                else
                {
                    if (listener != null) {
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                listener.onDone(group, false);
                            }
                        });
                    }
                }
            }
        });
    }

    /**
     * Called on MAIN thread exactly once in the constructor to obtain the
     * worker thread handler, if the worker thread handler wasn't given in the
     * constructor.
     * @return The worker thread handler.
     */
    protected abstract NMTHandler getWorkerThreadHandler();
}
