package com.nuance.dragon.toolkit.nvsl;

import java.util.ArrayList;
import java.util.List;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.util.internal.ReadOnlyList;

/**
 * A NVSL identification result
 * <p>
 * The result may contain several kinds of NVSL result:
 * speaker identification result, gender result, language detection result, etc.
 */
public final class NVSLResult implements JSONCompliant
{
    /**
     * Result entry for identification
     */
    public static class IdentEntry implements JSONCompliant
    {
        private static final String KEY_VOICEID = "voiceid";
        private static final String KEY_RAWSCORE = "rawscore";
        private static final String KEY_MAPPEDSCORE = "mappedscore";

        /** The voice ID that has been identified during the identification session.
         *  The string format is [speakerId]_[voiceTag] or [voiceTag] if speakerId is null in {@link NVSLVoiceprint}
         */
        public String voiceId;
        /**
         * The speaker identification raw score. The score value ranges from negative infinity
         * value to positive infinity value.
         */
        public int rawScore;
        /**
         * The consistent score mapping a raw score onto a scale of -10000 to +10000.
         * <p>
         * When properly calibrated, the returned score value can provide some intuitive value.
         * For example, positive score values correspond to the true user while negative score
         * values correspond to impostors. Also, as the magnitude of the score increases,
         * so does the confidence. So, for example, a mapped score of -9000 would have a high confidence
         * of being an impostor whereas a score of +1500 would have a low confidence of being the true user.
         */
        public int mappedScore;

        IdentEntry(String voiceId, int rawScore, int mappedScore)
        {
            this.voiceId = voiceId;
            this.rawScore = rawScore;
            this.mappedScore = mappedScore;
        }

        @Override
        public org.json.JSONObject toJSON()
        {
            JSONObject o = new JSONObject();
            o.tryPut(KEY_VOICEID, voiceId);
            o.tryPut(KEY_RAWSCORE, rawScore);
            o.tryPut(KEY_MAPPEDSCORE, mappedScore);

            return o;
        }

        public static IdentEntry createFromJSON(org.json.JSONObject json) throws JSONException
        {
            String voiceId = json.getString(KEY_VOICEID);
            int rawScore = json.getInt(KEY_RAWSCORE);
            int mappedScore = json.getInt(KEY_MAPPEDSCORE);

            return new IdentEntry(voiceId, rawScore, mappedScore);
        }

        @Override
        public int hashCode()
        {
            final int prime = 31;
            int result = 1;
            result = prime * result + rawScore;
            result = prime * result + mappedScore;
            result = prime * result
                    + ((voiceId == null) ? 0 : voiceId.hashCode());

            return result;
        }

        @Override
        public boolean equals(Object obj)
        {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            IdentEntry other = (IdentEntry) obj;
            if (rawScore != other.rawScore)
                return false;
            if (mappedScore != other.mappedScore)
                return false;
            if (voiceId == null) {
                if (other.voiceId != null)
                    return false;
            } else if (!voiceId.equals(other.voiceId))
                return false;

            return true;
        }
    }

    public static class Interval implements JSONCompliant
    {
        private static final String KEY_START = "start";
        private static final String KEY_END = "end";

        public float startInterval;

        public float endInterval;

        Interval(float startInterval, float endInterval)
        {
            this.startInterval = startInterval;
            this.endInterval = endInterval;
        }

        @Override
        public org.json.JSONObject toJSON()
        {
            JSONObject o = new JSONObject();
            o.tryPut(KEY_START, startInterval);
            o.tryPut(KEY_END, endInterval);

            return o;
        }

        public static Interval createFromJSON(org.json.JSONObject json) throws JSONException
        {
            float startInterval = (float)json.getDouble(KEY_START);
            float endInterval = (float)json.getDouble(KEY_END);

            return new Interval(startInterval, endInterval);
        }

        @Override
        public int hashCode()
        {
            final int prime = 31;
            int result = 1;
            result = prime * result + Float.floatToIntBits(startInterval);
            result = prime * result + Float.floatToIntBits(endInterval);

            return result;
        }

        @Override
        public boolean equals(Object obj)
        {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            Interval other = (Interval) obj;
            if (startInterval != other.startInterval)
                return false;
            if (endInterval != other.endInterval)
                return false;

            return true;
        }
    }

    public static class Speaker implements JSONCompliant
    {
        private static final String KEY_ENTRIES = "entries";
        private static final String KEY_INTERVALS = "intervals";
        private static final String KEY_GENDERCONF = "genderconf";
        private static final String KEY_GENDER = "gender";

        private final List<IdentEntry> _entries;
        private final List<Interval> _intervals;
        private int _genderConfidence;
        private char _genderChar;

        Speaker()
        {
            _entries = new ArrayList<IdentEntry>();
            _intervals = new ArrayList<Interval>();
        }

        void addIdentEntry(IdentEntry entry)
        {
            _entries.add(entry);
        }

        void addInterval(Interval interval)
        {
            _intervals.add(interval);
        }

        void setGenderResult(char cGender, int genderConf)
        {
            _genderChar = cGender;
            _genderConfidence = genderConf;
        }

        /**
         * Get a read-only copy of the identified entry list, in correct order.
         * @return The entries in this result, could be empty.
         */
        public List<IdentEntry> getIdentEntries()
        {
            return new ReadOnlyList<IdentEntry>(_entries);
        }

        List<IdentEntry> getOrgIdentEntries()
        {
            return _entries;
        }

        /**
         * Get a read-only copy of the interval list, which is associated to this speaker.
         * The list is empty, when multi-speaker identification is not performed.
         * @return The intervals in this result, could be empty.
         */
        public List<Interval> getIntervals()
        {
            return new ReadOnlyList<Interval>(_intervals);
        }

        /**
         * Get the gender char.
         * @return The gender char 'F': female, 'M': male.
         */
        public char getGenderChar()
        {
            return _genderChar;
        }

        /**
         * Returns how confident the system is that the analyzed
         * speaker is male or female. The confidence values range from 50 to 100
         *
         * @return Gender confidence value or -1 if not available
         */
        public int getGenderConfidence()
        {
            return _genderConfidence;
        }

        @Override
        public org.json.JSONObject toJSON()
        {
            JSONObject o = new JSONObject();
            o.tryPut(KEY_GENDERCONF, _genderConfidence);
            o.tryPut(KEY_GENDER, _genderChar);

            org.json.JSONArray entries = new org.json.JSONArray();
            for (IdentEntry entry : _entries)
            {
                entries.put(entry.toJSON());
            }
            o.tryPut(KEY_ENTRIES, entries);

            org.json.JSONArray intervals = new org.json.JSONArray();
            for (Interval interval : _intervals)
            {
                intervals.put(interval.toJSON());
            }
            o.tryPut(KEY_INTERVALS, intervals);
            return o;
        }

        public static Speaker createFromJSON(org.json.JSONObject json) throws JSONException
        {
            char genderChar = (char)json.getInt(KEY_GENDER);
            int genderConf = json.getInt(KEY_GENDERCONF);

            Speaker speaker =  new Speaker();
            speaker.setGenderResult(genderChar, genderConf);

            org.json.JSONArray entries = json.getJSONArray(KEY_ENTRIES);
            for (int i = 0; i < entries.length(); i++)
            {
                speaker.addIdentEntry(IdentEntry.createFromJSON(entries.getJSONObject(i)));
            }

            org.json.JSONArray intervals = json.getJSONArray(KEY_INTERVALS);
            for (int i = 0; i < intervals.length(); i++)
            {
                speaker.addInterval(Interval.createFromJSON(intervals.getJSONObject(i)));
            }
            return speaker;
        }

        @Override
        public int hashCode()
        {
            final int prime = 31;
            int result = 1;
            result = prime * result
                    + ((_entries == null) ? 0 : _entries.hashCode());
            result = prime * result
                    + ((_intervals == null) ? 0 : _intervals.hashCode());
            result = prime * result + _genderConfidence;
            result = prime * result + _genderChar;

            return result;
        }

        @Override
        public boolean equals(Object obj)
        {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (!(obj instanceof Speaker))
                return false;
            Speaker other = (Speaker) obj;
            if (_genderConfidence != other._genderConfidence)
                return false;
            if (_genderChar != other._genderChar)
                return false;
            if (_entries == null) {
                if (other._entries != null)
                    return false;
            } else if (!_entries.equals(other._entries))
                return false;
            if (_intervals == null) {
                if (other._intervals != null)
                    return false;
            } else if (!_intervals.equals(other._intervals))
                return false;

            return true;
        }
    }

    /**
     * Language detection result optional
     */
    public static class Language implements JSONCompliant {
        private static final String KEY_LANGCODE = "langCode";
        private static final String KEY_LANGDESCR = "langDescr";
        private static final String KEY_SCORE = "score";

        /** Language Code string */
        public final String langCode;
        /** Language description string. */
        public final String langDescr;
        /**
         * Language identification score, a language is correctly identified
         * with score > 0. The score value ranges from 0 to 1000
         */
        public final int score;

        Language(String langCode, String langDescr, int score) {
            this.langCode = langCode;
            this.langDescr = langDescr;
            this.score = score;
        }

        @Override
        public org.json.JSONObject toJSON()
        {
            JSONObject o = new JSONObject();
            o.tryPut(KEY_LANGCODE, langCode);
            o.tryPut(KEY_LANGDESCR, langDescr);
            o.tryPut(KEY_SCORE, score);

            return o;
        }

        public static Language createFromJSON(org.json.JSONObject json) throws JSONException
        {
            String langCode = json.getString(KEY_LANGCODE);
            String langDescr = json.getString(KEY_LANGDESCR);
            int score = json.getInt(KEY_SCORE);

            return new Language(langCode, langDescr, score);
        }

        @Override
        public int hashCode()
        {
            final int prime = 31;
            int result = 1;
            result = prime * result + score;
            result = prime * result
                    + ((langCode == null) ? 0 : langCode.hashCode());
            result = prime * result
                    + ((langDescr == null) ? 0 : langDescr.hashCode());

            return result;
        }

        @Override
        public boolean equals(Object obj)
        {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            Language other = (Language) obj;
            if (score != other.score)
                return false;
            if (langCode == null) {
                if (other.langCode != null)
                    return false;
            } else if (!langCode.equals(other.langCode))
                return false;
            if (langDescr == null) {
                if (other.langDescr != null)
                    return false;
            } else if (!langDescr.equals(other.langDescr))
                return false;

            return true;
        }
    }

    private static final String KEY_SPEAKERS = "speakers";
    private static final String KEY_LANGUAGES = "languages";
    private static final String KEY_ISFINAL = "final";

    private final List<Speaker> _speakers;
    private final List<Language> _languages;
    private boolean _isFinal;

    public NVSLResult(boolean isFinal) {
        _speakers = new ArrayList<Speaker>();
        _languages = new ArrayList<Language>();
        _isFinal = isFinal;
    }

    public void addSpeaker(Speaker speaker)
    {
        _speakers.add(speaker);
    }

    void addLanguage(Language language)
    {
        _languages.add(language);
    }

    /**
     * Get a read-only copy of the speakers in this result.
     * @return The speakers in this result.
     */
    public List<Speaker> getSpeakers()
    {
        return new ReadOnlyList<Speaker>(_speakers);
    }

    /**
     * Get a read-only copy of the language results in this result, in correct order.
     * @return The languages in this result.
     */
    public List<Language> getLanguages()
    {
        return new ReadOnlyList<Language>(_languages);
    }

    /**
     * Return true if it is final result, otherwise false
     * @return final flag
     */
    public boolean isFinal()
    {
        return _isFinal;
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + (_isFinal ? 1231 : 1237);
        result = prime * result
                + ((_speakers == null) ? 0 : _speakers.hashCode());
        result = prime * result
                + ((_languages == null) ? 0 : _languages.hashCode());

        return result;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (!(obj instanceof NVSLResult))
            return false;
        NVSLResult other = (NVSLResult) obj;
        if (_isFinal != other._isFinal)
            return false;
        if (_speakers == null) {
            if (other._speakers != null)
                return false;
        } else if (!_speakers.equals(other._speakers))
            return false;
        if (_languages == null) {
            if (other._languages != null)
                return false;
        } else if (!_languages.equals(other._languages))
            return false;

        return true;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();

        o.tryPut(KEY_ISFINAL, _isFinal);

        org.json.JSONArray speakers = new org.json.JSONArray();
        for (Speaker speaker : _speakers)
        {
            speakers.put(speaker.toJSON());
        }
        o.tryPut(KEY_SPEAKERS, speakers);

        org.json.JSONArray languages = new org.json.JSONArray();
        for (Language lang : _languages)
        {
            languages.put(lang.toJSON());
        }
        o.tryPut(KEY_LANGUAGES, languages);
        return o;
    }

    public static NVSLResult createFromJSON(org.json.JSONObject json) throws JSONException
    {
        boolean isFinal = json.getBoolean(KEY_ISFINAL);

        NVSLResult res =  new NVSLResult(isFinal);

        org.json.JSONArray speakers = json.getJSONArray(KEY_SPEAKERS);
        for (int i = 0; i < speakers.length(); i++)
        {
            res.addSpeaker(Speaker.createFromJSON(speakers.getJSONObject(i)));
        }

        org.json.JSONArray languages = json.getJSONArray(KEY_LANGUAGES);
        for (int i = 0; i < languages.length(); i++)
        {
            res.addLanguage(Language.createFromJSON(languages.getJSONObject(i)));
        }
        return res;
    }
}
