package com.nuance.dragon.toolkit.nvsl;

import java.util.ArrayList;
import java.util.List;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * A speaker can have several different voiceprints {@link NVSLVoiceprint}.
 * A speaker should be associated to a group.
 */
public class NVSLSpeaker implements JSONCompliant
{
    private static final String KEY_SPEAKER_ID  = "speakerid";
    private static final String KEY_VOICEPRINTS   = "voiceprints";

    private final String _speakerId;
    private final List<NVSLVoiceprint> _voiceprints;

    NVSLSpeaker(String speakerId)
    {
        Checker.checkStringArgForNullOrEmpty("speakerId", speakerId);
        _speakerId = speakerId;
        _voiceprints = new ArrayList<NVSLVoiceprint>();
    }

    /**
     * Returns the speaker unique identifier.
     * @return the speaker unique identifier
     */
    public String getSpeakerId()
    {
        return _speakerId;
    }

    /**
     * Creates a voiceprint and attaches it to this speaker.
     * If the voiceprint existed already, it returns the existing one.
     *
     * @param vpTag the string identifying the new voiceprint
     * @return the newly created {@link NVSLVoiceprint}
     */
    public NVSLVoiceprint createVoiceprint(String vpTag)
    {
        Checker.checkStringArgForNullOrEmpty("vpTag", vpTag);

        NVSLVoiceprint vp;
        //if the voiceprint existed already, return it
        if((vp = getVoiceprint(vpTag)) != null)
        {
            return vp;
        }

        //if the voiceprint doesn't exist, create a new one
        vp = new NVSLVoiceprint(_speakerId, vpTag);
        _voiceprints.add(vp);

        return vp;
    }

    /**
     * Returns a voiceprint with the given id tag, if found attached to this speaker.
     *
     * @param vpTag the voiceprint tag
     * @return the voiceprint with the given tag or null if none was found.
     */
    public NVSLVoiceprint getVoiceprint(String vpTag)
    {
        Checker.checkStringArgForNullOrEmpty("vpTag", vpTag);
        //empty list
        if (_voiceprints.size() == 0) return null;

        for (NVSLVoiceprint vp: _voiceprints)
        {
            if ( vp.getVoiceTag().equals(vpTag))
            {
                return vp;
            }
        }
        return null;
    }

    /**
     * Removes from this speaker the voiceprint with the specified identification tag.
     * Removal corresponds to deletion of all the voiceprint data from the persistent storage.
     *
     * @param vpTag the voiceprint tag
     * @param fileManager The file manager which helps to look for files.
     * @return true if removal was successful. False if a voiceprint with that tag was not found.
     */
    public boolean removeVoiceprint(FileManager fileManager, String vpTag)
    {
        //empty list
        if(_voiceprints.size() == 0) return false;

        NVSLVoiceprint vp = getVoiceprint(vpTag);
        if(vp == null) return false;

        vp.removeAllAudioSegs(fileManager);
        vp.removeVpFile(fileManager);

        return _voiceprints.remove(vp);
    }

    /**
     * Removes all the voiceprints from this speaker.
     *
     * @param fileManager The file manager which helps to look for files.
     */
    public void removeAllVoiceprints(FileManager fileManager)
    {
        while (_voiceprints.size() > 0)
        {
            NVSLVoiceprint vp = _voiceprints.get(0);
            removeVoiceprint(fileManager, vp.getVoiceTag());
        }
    }

    /**
     * Returns the list of voiceprints belonging to this speaker.
     * @return the list of voiceprints
     */
    public List<NVSLVoiceprint> getVoiceprints()
    {
        return _voiceprints;
    }

    private void addVoiceprint(NVSLVoiceprint voiceprint)
    {
        _voiceprints.add(voiceprint);
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result
                + ((_speakerId == null) ? 0 : _speakerId.hashCode());
        result = prime * result
                + ((_voiceprints == null) ? 0 : _voiceprints.hashCode());

        return result;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (!(obj instanceof NVSLSpeaker))
            return false;
        NVSLSpeaker other = (NVSLSpeaker) obj;
        if (_speakerId == null) {
            if (other._speakerId != null)
                return false;
        } else if (!_speakerId.equals(other._speakerId))
            return false;
        if (_voiceprints == null) {
            if (other._voiceprints != null)
                return false;
        } else if (!_voiceprints.equals(other._voiceprints))
            return false;

        return true;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();

        o.tryPut(KEY_SPEAKER_ID, _speakerId);

        org.json.JSONArray voiceprints = new org.json.JSONArray();
        for (NVSLVoiceprint vp : _voiceprints)
        {
            voiceprints.put(vp.toJSON());
        }
        o.tryPut(KEY_VOICEPRINTS, voiceprints);

        return o;
    }

    public static NVSLSpeaker createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String speakerId = json.getString(KEY_SPEAKER_ID);
        NVSLSpeaker speaker = new NVSLSpeaker(speakerId);

        org.json.JSONArray voiceprints = json.getJSONArray(KEY_VOICEPRINTS);
        for (int i = 0; i < voiceprints.length(); i++)
        {
            speaker.addVoiceprint(NVSLVoiceprint.createFromJSON(voiceprints.getJSONObject(i)));
        }

        return speaker;
    }
}
