package com.nuance.dragon.toolkit.nvsl;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

import org.json.JSONException;

import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.AsyncTaskHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.util.internal.Pair;
import com.nuance.dragon.toolkit.util.internal.ReadOnlyList;

/**
 * A class representing voiceprints. A voiceprint is a logical entity that is
 * associated to audio files ("enrollment" process) from which the engine extracts
 * the voice biometric features that results into a binary file.
 * <p>
 * A voiceprint is uniquely (relative to its speaker) identified by a "voiceprint tag"
 * and keeps memory of the engine configuration it was created with by storing a name
 * to the engine configuration.
 * <p>
 * NVSLVoiceprint objects can be created stand-alone or by speaker
 */
public class NVSLVoiceprint implements JSONCompliant
{
    private static final String KEY_VOICE_TAG  = "voicetag";
    private static final String KEY_SPEAKER_ID   = "speakerid";
    private static final String KEY_CONFSET_NAME = "confsetname";
    private static final String KEY_AUDIOSEGFILES   = "audiosegfiles";
    private static final String KEY_VPFILENAME   = "vpfilename";
    private static final String KEY_AUDIOTYPE   = "audiotype";

    /**
     * The maximum number of audio segments allowed for a voiceprint.
     */
    private final String _voiceTag;
    private final String _speakerId;
    private String _configSetName;
    // saved audio segment files and audio with noise (if applicable) associated with this voiceprint
    private final List<Pair<String, String>> _audioSegFiles;
    private String _vpFileName;
    private boolean _isAdaptation;
    private final AsyncTaskHandler _workerThreadHandler;
    private final int _maxAudioSeg;

    AudioType audioType;

    /**
     * Create NVSLVoiceprint.
     * @param voicetag The voiceprint's description
     */
    public NVSLVoiceprint(String voicetag)
    {
        this(null, voicetag, 30);
    }

    /**
     * Create NVSLVoiceprint.
     *
     * @param speakerId voice speaker owner Id
     * @param voicetag The voiceprint's description
     */
    NVSLVoiceprint(String speakerId, String voicetag)
    {
        this(speakerId, voicetag, 30);
    }

    /**
     * Create NVSLVoiceprint.
     *
     * @param speakerId Voice speaker owner Id
     * @param voicetag The voiceprint's description
     * @param maxAudioSeg The maximum number of audio segments a voiceprint can hold.
     *                    By default it is 30.
     */
    NVSLVoiceprint(String speakerId, String voicetag, int maxAudioSeg)
    {
        Checker.checkStringArgForNullOrEmpty("voicetag", voicetag);

        _voiceTag = voicetag;
        _speakerId = speakerId;
        _audioSegFiles = new ArrayList<Pair<String, String>>();
        audioType = AudioType.UNKNOWN;
        _configSetName = "";
        _vpFileName = "";
        _workerThreadHandler = new AsyncTaskHandler();
        _maxAudioSeg = maxAudioSeg;
    }

    /**
     * Attach a unique configuration label to this voiceprint
     *
     * @param configSetName The engine configuration combined name (e.g.
     *            LVSLv3_TI_COMPACT_v1.0)
     */
    void setConfigSetName(String configSetName)
    {
        _configSetName = configSetName;
    }

    /**
     * Get a configuration name attached to this voiceprint
     *
     * @return configuration name
     */
    public String getConfigSetName()
    {
        return _configSetName;
    }

    /**
     * Get voice tag for this voiceprint
     */
    public String getVoiceTag()
    {
        return _voiceTag;
    }

    /**
     * Get speaker Id for this voiceprint
     */
    public String getSpeakerId()
    {
        return _speakerId;
    }

    /**
     * Returns the number of audio segment files belonging to this voiceprint.
     *
     * @return the number of audio segment files
     */
    public int getNumOfAudioSegFiles()
    {
        return (null == _audioSegFiles) ? 0 : _audioSegFiles.size();
    }

    /**
     * Returns a read-only copy of audio segment file names belonging
     * to this voiceprint.
     *
     * @return the list of audio segment file names
     */
    public List<String> getAudioSegFiles()
    {
        // TODO(LXU): tempList may not be GCed
        List<String> tempList = new ArrayList<String>();
        for (Pair<String, String> pair : _audioSegFiles)
            tempList.add(pair.first);
        return new ReadOnlyList<String>(tempList);
    }

    /**
     * Returns the {@link AudioType} of audio segments associated to this voiceprint
     *
     * @return the AudioType
     */
    public AudioType getAudioType()
    {
        return audioType;
    }

    /**
     * Returns the maximum number the voiceprint can hold.
     */
    public int getMaxAudioSegNum()
    {
        return _maxAudioSeg;
    }

    List<String> getAudioNoiseFiles()
    {
        // TODO(LXU): tempList may not be GCed
        List<String> tempList = new ArrayList<String>();
        for (Pair<String, String> pair : _audioSegFiles)
        {
            if (pair.second != null)
                tempList.add(pair.second);
        }
        return new ReadOnlyList<String>(tempList);
    }

    /**
     * Removes all audio files owned by this voiceprint.
     *
     * @param fileManager The file manager which helps to look for files.
     */
    void removeAllAudioSegs(FileManager fileManager)
    {
        while (_audioSegFiles.size() > 0)
        {
            removeAudioFile(fileManager, 0);
        }
    }

    /**
     * Clears all audio files owned by this voiceprint, without worker thread
     *
     * @param fileManager The file manager which helps to look for files.
     */
    void clearAllAudioSegs(FileManager fileManager)
    {
        for (Pair<String, String> pair : _audioSegFiles)
        {
            if (fileManager.exists(pair.first))
            {
                fileManager.delete(pair.first);
            }
            if (pair.second != null && fileManager.exists(pair.second))
            {
                fileManager.delete(pair.second);
            }
        }
        _audioSegFiles.clear();
    }

    /**
     * Removes voiceprint data file
     *
     * @param fileManager The file manager which helps to look for files.
     */
    void removeVpFile(FileManager fileManager)
    {
        if (null != _vpFileName && fileManager.exists(_vpFileName))
        {
            fileManager.delete(_vpFileName);
            _vpFileName = "";
        }
    }

    Pair<String, String> generateVpAudioSegFileName()
    {
        Calendar cal = Calendar.getInstance();
        String affix = new SimpleDateFormat("yyyyMMddHHmmss", java.util.Locale.getDefault()).format(cal.getTime());

        String prefix = "";
        if (_speakerId != null)
            prefix = _speakerId;

        String audioSegFileNmae = new String(prefix + affix + ".pcm");
        String audioNoiseSegFileName = new String(prefix + affix + "_Noise"+ ".pcm");

        return new Pair<String, String>(audioSegFileNmae, audioNoiseSegFileName);
    }

    /**
     * Voiceprint file name are composed as:
     * [speakerId]_[voiceTag]_[configSetName].svm
     */
    String generateVpFileName()
    {
        StringBuilder fileNameBuilder = new StringBuilder();
        if (_speakerId != null)
            fileNameBuilder.append(_speakerId + "_");
        fileNameBuilder.append(_voiceTag);
        if (_configSetName != null)
            fileNameBuilder.append("_" + _configSetName + ".svm");

        return fileNameBuilder.toString();
    }

    void setVpFileName(String vpFileName)
    {
        _vpFileName = vpFileName;
    }

    /**
     * Get the voiceprint file name.
     *
     * @return Voiceprint file name are composed as:
     *         [speakerId]_[voiceTag]_[configSetName].svm
     */
    public String getVpFileName()
    {
        return _vpFileName;
    }

    void addAudioSegFile(Pair<String, String> audioFilePair, AudioType audioType)
    {
        if (!checkAudioConsistent(audioType))
        {
            Logger.warn(this, "Unable to add audio segment file to voiceprint: frequency inconsistent");
        }

        _audioSegFiles.add(audioFilePair);
    }

    Pair<String, String> getAudioSegFile(int index)
    {
        if (null == _audioSegFiles || _audioSegFiles.size() == 0 ||
            index < 0 || index >= _audioSegFiles.size())
        {
            return null;
        }
        return _audioSegFiles.get(index);
    }

    boolean isAdaptation()
    {
        return _isAdaptation;
    }

    void setAdaptation(boolean isAdaptation)
    {
        this._isAdaptation = isAdaptation;
    }

    boolean checkAudioConsistent(AudioType audioType)
    {
        if (this.audioType == AudioType.UNKNOWN)
        {
            this.audioType = audioType;
            return true;
        }
        else if (!this.audioType.equals(audioType))
        {
            Logger.warn(this, "Audio segment is inconsistent to voiceprint.");
            return false;
        }
        return true;
    }

    private boolean removeAudioFile(final FileManager fileManager, int index)
    {
        if (null == _audioSegFiles || _audioSegFiles.size() == 0 ||
            index < 0 || index >= _audioSegFiles.size())
        {
            return false;
        }

        final Pair<String, String> audioFilePair = _audioSegFiles.get(index);

        Runnable deleteTask = new Runnable()
        {
            @Override
            public void run()
            {
                if (fileManager.exists(audioFilePair.first))
                {
                    fileManager.delete(audioFilePair.first);
                }

                if (audioFilePair.second != null && fileManager.exists(audioFilePair.second))
                {
                    fileManager.delete(audioFilePair.second);
                }
            }

        };
        _workerThreadHandler.post(deleteTask);
        return _audioSegFiles.remove(index) != null;
    }

    boolean removeAudioFile(final FileManager fileManager, final Pair<String, String> audioFilePair)
    {
        if (null == _audioSegFiles || _audioSegFiles.size() == 0 || !_audioSegFiles.contains(audioFilePair))
        {
            return false;
        }

        Runnable deleteTask = new Runnable()
        {
            @Override
            public void run()
            {
                if (fileManager.exists(audioFilePair.first))
                {
                    fileManager.delete(audioFilePair.first);
                }

                if (audioFilePair.second != null && fileManager.exists(audioFilePair.second))
                {
                    fileManager.delete(audioFilePair.second);
                }
            }

        };
        _workerThreadHandler.post(deleteTask);

        return _audioSegFiles.remove(audioFilePair);
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result
                + ((_voiceTag == null) ? 0 : _voiceTag.hashCode());
        result = prime * result
                + ((_speakerId == null) ? 0 : _speakerId.hashCode());
        result = prime * result
                + ((_configSetName == null) ? 0 : _configSetName.hashCode());
        result = prime * result
                + ((_audioSegFiles == null) ? 0 : _audioSegFiles.hashCode());
        result = prime * result
                + ((_vpFileName == null) ? 0 : _vpFileName.hashCode());
        result = prime * result
                + ((audioType == null) ? 0 : audioType.hashCode());

        return result;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (!(obj instanceof NVSLVoiceprint))
            return false;
        NVSLVoiceprint other = (NVSLVoiceprint) obj;

        if (_voiceTag == null) {
            if (other._voiceTag != null)
                return false;
        } else if (!_voiceTag.equals(other._voiceTag))
            return false;
        if (_speakerId == null) {
            if (other._speakerId != null)
                return false;
        } else if (!_speakerId.equals(other._speakerId))
            return false;
        if (_configSetName == null) {
            if (other._configSetName != null)
                return false;
        } else if (!_configSetName.equals(other._configSetName))
            return false;
        if (_vpFileName == null) {
            if (other._vpFileName != null)
                return false;
        } else if (!_vpFileName.equals(other._vpFileName))
            return false;
        if (audioType == null) {
            if (other.audioType != null)
                return false;
        } else if (!audioType.equals(other.audioType))
            return false;
        if (_audioSegFiles == null) {
            if (other._audioSegFiles != null)
                return false;
        } else if (!_audioSegFiles.equals(other._audioSegFiles))
            return false;

        return true;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();

        o.tryPut(KEY_VOICE_TAG, _voiceTag);
        o.tryPut(KEY_SPEAKER_ID, _speakerId);
        o.tryPut(KEY_CONFSET_NAME, _configSetName);
        o.tryPut(KEY_VPFILENAME, _vpFileName);
        o.tryPut(KEY_AUDIOTYPE, audioType);
        org.json.JSONArray audioSegFiles = new org.json.JSONArray();
        for (Pair<String, String> pair : _audioSegFiles)
        {
            JSONObject pairO = new JSONObject();
            pairO.tryPut("first", pair.first);
            pairO.tryPut("second", pair.second);
            audioSegFiles.put(pairO);
        }
        o.tryPut(KEY_AUDIOSEGFILES, audioSegFiles);

        return o;
    }


    /**
     * @param json A JSON object that represents a NVSLVoiceprint.
     *
     * @return A NVSLVoiceprint object
     */
    public static NVSLVoiceprint createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String voicetag = json.getString(KEY_VOICE_TAG);
        String speakerId = json.getString(KEY_SPEAKER_ID);
        NVSLVoiceprint voiceprint = new NVSLVoiceprint(speakerId, voicetag);

        String confSetName = json.getString(KEY_CONFSET_NAME);
        voiceprint.setConfigSetName(confSetName);

        String vpFileName = json.getString(KEY_VPFILENAME);
        voiceprint.setVpFileName(vpFileName);

        voiceprint.audioType = AudioType.fromJSON(json.getJSONObject(KEY_AUDIOTYPE));

        org.json.JSONArray audioSegFiles = json.getJSONArray(KEY_AUDIOSEGFILES);
        for (int i = 0; i < audioSegFiles.length(); i++)
        {
            org.json.JSONObject pairO = audioSegFiles.getJSONObject(i);
            String first = pairO.optString("first", null);
            String second = pairO.optString("second", null);
            voiceprint.addAudioSegFile(new Pair<String, String>(first, second), voiceprint.audioType);
        }

        return voiceprint;
    }
}
