package com.nuance.dragon.toolkit.nvsl;


import java.util.List;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;

class NativeNVSLImpl implements NativeNVSL
{
    static boolean LOAD_LIBRARY_SUCCESS = true;
    static
    {
        try
        {
            System.loadLibrary("dmt_nvsl");
        }
        catch (UnsatisfiedLinkError e)
        {
            Logger.error(NativeNVSLImpl.class, "Failed to load native library.", e);
            LOAD_LIBRARY_SUCCESS = false;
        }
    }

    private native int nvslCreate(long[] nvsl, FileManager fileManager, NVSLConfig config, InputDeviceListener listener);
    private native void nvslDestroy(long nvsl);
    private native int nvslStartEnrolling(long nvsl, String vpFilePath, String[] audioFiles, int freqHz);
    private native int nvslStartAdapting(long nvsl, String vpFilePath, String audioFile, int freqHz);
    private native int nvslStartIdentifying(long nvsl, String[] vpFilePaths, int freqHz, int maxResults, boolean isMultiIdent, int partialResultTimerMs);
    private native int nvslStartDetectingLanuage(long nvsl, int freqHz);
    private native void nvslCancel(long nvsl);
    private native int nvslStartCheckConsistency(long nvsl, String[] audioFiles, int freqHz);
    private native int nvslStartAudioCheck(long nvsl, int taskType, int freqHz, String audioDumpFilePath, String noiseInFilePath, int rand, int degradationLevel, String audioPlusNoiseFilePath);
    private native void nvslEnableVerboseLogging(boolean verbose);

    public interface InputDeviceListener
    {
        void onStartAudio(String audioFileName);

        void onStopAudio(String audioFileName);

        void onProcessedAudio(AudioInfoJni audioInfo);

        int onGetAudio(short[] audio);

        void onIdentificationResult(ResultJni result);

        void onCanceled();
    }

    public class TaskType
    {
        public static final int SPK_ID_TASK = 0;                    /* Speaker identification task                 */
        public static final int SPK_ENR_TASK = 1;                   /* Speaker enrollment task                     */
        public static final int SPK_ID_CALIBRATION_TASK = 2;        /* Speaker identification calibration task     */
        public static final int SPK_SEGME_TASK = 3;                 /* Speaker Segmentation task                   */
        public static final int MULTI_SPK_ID_TASK = 4;              /* Multi-Speaker Identification task           */
        public static final int LANGUAGE_ID_TASK = 5;               /* Language identification task                */
        public static final int LANGUAGE_ID_PREPROCESSING_TASK = 6; /* Language identification preprocessing task  */
    }

    /* Note: The value of these constants must match the values in ngf_nvsl.h */
    public class Status
    {
        public static final int OK = 0;                      /*!< Normal return. */
        public static final int ARGUMENT_ERROR = 1;         /*!< Function parameter error */
        public static final int MEMORY_ERROR = 2 ;            /*!< Memory allocation failure*/
        public static final int FILE_ERROR = 3;              /*!< Filesystem I/O error*/
        public static final int CALLFLOW_ERROR = 4;            /*!< Wrong call flow */
        public static final int SYSTEMCALL_ERROR = 5;       /*!< System call failure */
        public static final int HANDLE_ERROR = 6;              /*!<  Invalid NVSL handle */
        public static final int INTERNAL_ERROR = 7;         /*!< Unexpected internal error */
        public static final int AUDIO_ERROR = 8;               /*!< Audio source error */
        public static final int CONF_ERROR = 9;               /*!< Configuration error */
        public static final int NOISEADD_ERROR = 10;          //*!< Noise addition error */
        public static final int OTHER_ERROR = 11;           /*!< An error occurred.*/
    }

    private FileManager _fileManager;
    private long nvslInstance;

    private NVSLAudioListener _audioListener;
    private NVSLResultListener _resultListener;

    public NativeNVSLImpl(FileManager fileManager)
    {
        _fileManager = fileManager;
    }

    @Override
    public boolean createNVSL(NVSLConfig config)
    {
        boolean success = false;

        //destroy nvsl, if it already exists
        destroyNVSL();

        if (config != null)
        {

            // START JNI_CODE
            long[] nvslTemp = new long[1];
            int status = nvslCreate(nvslTemp, _fileManager, config, new InputDeviceListener()
            {
                @Override
                public void onStartAudio(String audioFileName)
                {
                    Logger.debug(this, "In onStartAudio().");

                    if (_audioListener != null)
                        _audioListener.onStartAudioPull(audioFileName);
                }

                @Override
                public void onStopAudio(String audioFileName)
                {
                    Logger.debug(this, "In onStopAudio().");

                    if (_audioListener != null)
                        _audioListener.onStopAudioPull(audioFileName);
                }

                @Override
                public void onProcessedAudio(AudioInfoJni audioInfo)
                {
                    Logger.debug(this, "In onProcessedAudio().");

                    if (_audioListener != null)
                        _audioListener.onSignalUpdate(audioInfo);
                }


                @Override
                public int onGetAudio(short[] audio)
                {
                    AudioChunk chunk;
                    int filled = 0;

                    if (_audioListener != null)
                    {
                        chunk = _audioListener.onGetAudioChunk(audio.length);
                        if (chunk != null)
                        {
                            filled = chunk.audioShorts.length;
                            System.arraycopy(chunk.audioShorts, 0, audio, 0, filled);
                        }
                    }

                    return filled;
                }

                @Override
                public void onIdentificationResult(ResultJni result)
                {
                    Logger.debug(this, "In onIdentificationResult().");

                    if (_resultListener != null)
                        _resultListener.onIdentificationResult(result);
                }

                @Override
                public void onCanceled()
                {
                    Logger.debug(this, "In onCanceled().");

                    if (_audioListener != null)
                        _audioListener.onCanceled();
                }
            });

            if (status == Status.OK && nvslTemp[0] != 0)
            {
                Logger.debug(this, "Created native NVSL");
                nvslInstance = nvslTemp[0];
                success = true;
            }
            else
            {
                Logger.error(this, "Unable to create native NVSL (error status = " + status + ").");
            }
        }
        return success;
    }

    @Override
    public void destroyNVSL()
    {
        if (nvslInstance == 0)
        {
            Logger.warn(this, "Error destroy NVSL: native NVSL does not exist!");
            return;
        }

        nvslDestroy(nvslInstance);
        nvslInstance = 0;
        Logger.debug(this, "Destroyed native NVSL");
    }

    @Override
    public boolean startEnrolling(String vpFilePath, List<String> audioFiles, int freqHz, NVSLAudioListener audioListener)
    {
        if (nvslInstance == 0)
        {
            Logger.error(this, "Error starting nvsl enrollment: native NVSL does not exist.");
            return false;
        }

        assert(audioFiles != null);
        assert(!audioFiles.isEmpty());

        String[] audioFileArray = new String[audioFiles.size()];
        audioFileArray = audioFiles.toArray(audioFileArray);

        int i = 1;
        Logger.debug(this, "To enroll voiceprint based on the following audio files:");
        for (String file : audioFileArray)
        {
            Logger.debug(this, "audio file [" + i + "]: " + file);
            i++;
        }

        _audioListener = audioListener;
        _resultListener = null;
        return (Status.OK == nvslStartEnrolling(nvslInstance, vpFilePath, audioFileArray, freqHz));
    }

    @Override
    public boolean startAdapting(String vpFilePath, String audioFile, int freqHz, NVSLAudioListener audioListener)
    {
        if (nvslInstance == 0)
        {
            Logger.error(this, "Error starting nvsl enrollment: native NVSL does not exist.");
            return false;
        }

        _audioListener = audioListener;
        _resultListener = null;
        return (Status.OK == nvslStartAdapting(nvslInstance, vpFilePath, audioFile, freqHz));

    }

    @Override
    public boolean startIdentifying(List<String> vpFilePaths, int freqHz, int maxResults, boolean isMultiIdent, int partialResultTimerMs, NVSLAudioListener audioListener, NVSLResultListener resultListener)
    {
        if (nvslInstance == 0)
        {
            Logger.error(this, "Error starting nvsl enrollment: native NVSL does not exist.");
            return false;
        }

        assert(vpFilePaths != null);

        String vpFilePathArray[];
        if (vpFilePaths.isEmpty())
        {
            Logger.warn(this, "To start identification against no voiceprint");
            return false;
        }
        else
        {
            vpFilePathArray = new String[vpFilePaths.size()];
            vpFilePathArray = vpFilePaths.toArray(vpFilePathArray);

            Logger.debug(this, "Starting identification with following voiceprint tags:");
            int i = 1;
            for (String vpFilePath : vpFilePathArray)
            {
                Logger.debug(this, "Voiceprint [" + i + "]: " + vpFilePath);
                i++;
            }
        }

        _audioListener = audioListener;
        _resultListener = resultListener;
        return (Status.OK == nvslStartIdentifying(nvslInstance, vpFilePathArray, freqHz, maxResults, isMultiIdent, partialResultTimerMs));
    }

    @Override
    public boolean startDetectingLanguage(int freqHz, NVSLAudioListener audioListener, NVSLResultListener resultListener)
    {
        if (nvslInstance == 0)
        {
            Logger.error(this, "Error starting nvsl language detection: native NVSL does not exist.");
            return false;
        }

        _audioListener = audioListener;
        _resultListener = resultListener;
        return (Status.OK == nvslStartDetectingLanuage(nvslInstance, freqHz));
    }

    @Override
    public void enableVoconVerboseLogging(boolean verbose)
    {
        nvslEnableVerboseLogging(verbose);
    }

    @Override
    public void cancel()
    {

        if (nvslInstance != 0) {
            nvslCancel(nvslInstance);
        }
    }

    @Override
    public boolean startCheckConsistency(List<String> audioFiles, int freqHz, NVSLAudioListener audioListener)
    {
        if (nvslInstance == 0)
        {
            Logger.error(this, "Error starting nvsl consistency check: native NVSL does not exist.");
            return false;
        }
        _audioListener = audioListener;
        _resultListener = null;

        assert(audioFiles != null);
        assert(!audioFiles.isEmpty());

        String audioFileArray[];
        if (audioFiles.size() == 1)
        {
            Logger.warn(this, "To check consistency based on only one file");
            return true;
        }
        else
        {
            audioFileArray = new String[audioFiles.size()];
            audioFileArray = audioFiles.toArray(audioFileArray);

            int i = 0;
            Logger.debug(this, "Starting consistency check with following audio files:");
            for (String name : audioFileArray)
            {
                Logger.debug(this, "audio file [" + i + "]: " + name);
                i++;
            }
        }
        return (Status.OK == nvslStartCheckConsistency(nvslInstance, audioFileArray, freqHz));
    }

    @Override
    public boolean startAudioCheck(int freqHz, int taskType, String audioDumpFilePath, NVSLAudioListener audioListener)
    {
        if (nvslInstance == 0)
        {
            Logger.error(this, "Error starting nvsl audio check: native NVSL does not exist.");
            return false;
        }

        _audioListener = audioListener;
        _resultListener = null;
        return (Status.OK == nvslStartAudioCheck(nvslInstance, taskType, freqHz, audioDumpFilePath, null, 0, 0, null));
    }

    @Override
    public boolean startAudioCheck(int freqHz, int taskType, String audioDumpFilePath, String noiseInFilePath, int rand, int degradationLevel, String audioPlusNoiseFilePath, NVSLAudioListener audioListener)
    {
        if (nvslInstance == 0)
        {
            Logger.error(this, "Error starting nvsl audio check: native NVSL does not exist.");
            return false;
        }

        _audioListener = audioListener;
        _resultListener = null;
        return (Status.OK == nvslStartAudioCheck(nvslInstance, taskType, freqHz, audioDumpFilePath, noiseInFilePath, rand, degradationLevel, audioPlusNoiseFilePath));
    }
}
