package com.nuance.dragon.toolkit.oem.api;

import com.nuance.dragon.toolkit.oem.impl.ContextOem;
import com.nuance.dragon.toolkit.oem.impl.DeviceInfoOem;
import com.nuance.dragon.toolkit.oem.impl.FileManagerOem;
import com.nuance.dragon.toolkit.oem.impl.HandlerOem;
import com.nuance.dragon.toolkit.oem.impl.HttpUtilOem;
import com.nuance.dragon.toolkit.oem.impl.WorkerThreadOem;

public class Factory {

    /**
     * Creates a NMT handler allows you to send and process Runnable objects associated with a
     * thread's message queue. On Android this method returns a handler with the looper for
     * the current thread; On J2SE this method returns a handler associated with current thread
     * if it's a {@link WorkerThread} otherwise there is no thread associated with the handler.
     */
    public static NMTHandler createNMTHandler()
    {
        return new HandlerOem();
    }

    /**
     * Creates a NMT handler allows you to send and process Runnable objects associated with a
     * thread's message queue.
     *
     * @param obj On Android it is android.os.Handler; on J2SE it is {@link WorkerThread}
     */
    public static NMTHandler createNMTHandler(Object obj)
    {
        return new HandlerOem(obj);
    }

    /**
     * Creates a NMT context, which is interface to global information about an application environment.
     *
     * @param obj On Android it is android.content.Context; on J2SE it could be null.
     */
    public static NMTContext createNMTContext(Object obj)
    {
        return new ContextOem(obj);
    }

    /**
     * Creates a simple worker thread wrapper. This class starts a dedicated thread with
     * a {@link NMTHandler} attached for posting messages.
     */
    public static WorkerThread createWorkerThread()
    {
        return new WorkerThreadOem();
    }

    /**
     * Creates a simple worker thread wrapper. This class starts a dedicated thread with
     * a {@link NMTHandler} attached for posting messages.
     *
     * @param threadName The name of the new thread
     */
    public static WorkerThread createWorkerThread(String threadName)
    {
        return new WorkerThreadOem(threadName);
    }

    public static HttpUtil createHttpUtil(boolean enableSelfSignedCert, String certSummary, String certData, NMTContext context)
    {
        return new HttpUtilOem(enableSelfSignedCert, certSummary, certData, context);
    }

    public static DeviceInfo getDeviceInfo(NMTContext context)
    {
        return DeviceInfoOem.getDeviceInfoOem(context);
    }

    /**
     * Creates a file manager that only looks in the app's built-in asset directory.
     * Android automatically compresses asset files. This causes issues when reading and using them.
     * To avoid this, we use special extensions to prevent automatic compression
     *
     * @param context The NMT context which wraps application context on Android.
     *
     * @param assetExt The special extension to automatically add when looking for asset
     * files. For example, you might pass ".jpg" if assets are given a .jpg
     * extension to avoid automatic Android asset compression. Null means don't look for assets with special extension
     *
     * @param assetPath The relative path within the assets (i.e., "docs/home.html") or expansion files to search for files in the assets.
     * If null, asset files will not be supported. Asset files are read-only.
     */
    public static FileManager createFileManager(NMTContext context, String assetExt, String assetPath)
    {
        return new FileManagerOem(context, assetExt, assetPath);
    }

    /**
     * Creates a file manager that only looks in the device's build-in directory.
     *
     * @param context The NMT context which wraps application context on Android.
     *
     * @param internalFilePath The relative path within application's default data directory
     * to search for files in the file system. If null, internal files will not be supported.
     * These files can be read or written.
     */
    public static FileManager createFileManager(NMTContext context, String internalFilePath)
    {
        return new FileManagerOem(context, internalFilePath);
    }

    /**
     * Creates a file manager that can look in the app's built-in asset directory and the device's built-in directory
     *
     * @param context The NMT context which wraps application context on Android.
     *
     * @param assetExt The special extension to automatically add when looking for asset
     * files. For example, you might pass ".jpg" if assets are given a .jpg
     * extension to avoid automatic Android asset compression. Null means don't look for assets with special extension
     *
     * @param assetPath The relative path with the assets(i.e., "docs/home.html") to search for files in the assets. If
     * null, asset files will not be supported. Asset files are read-only.
     *
     * @param internalFilePath The relative path within application's default data directory
     * to search for files in the file system. If null, internal files will not be supported.
     * These files can be read or written.
     */
    public static FileManager createFileManager(NMTContext context, String assetExt, String assetPath, String internalFilePath)
    {
        return new FileManagerOem(context, assetExt, assetPath, internalFilePath);
    }

    /**
     * Creates a file manager that can look in the app's built-in asset directory and the device's built-in directory,
     * and the device's external disk (e.g. SD card)
     *
     * @param context The NMT context which wraps application context on Android.
     *
     * @param assetExt The special extension to automatically add when looking for asset
     * files. For example, you might pass ".jpg" if assets are given a .jpg
     * extension to avoid automatic Android asset compression. Null means don't look for assets with special extension
     *
     * @param assetPath The relative path within the assets (i.e., "docs/home.html") or expansion files to search for files in the assets. If
     * null, asset files will not be supported. Asset files are read-only.
     *
     * @param internalFilePath The relative path within application's default data directory
     * to search for files in the file system. If null, internal files will not be supported.
     * These files can be read or written.
     *
     * @param externalFilePath An absolute path to a read-only external directory.
     * If blank, on Android defaults to Environment#getExternalStorageDirectory; on J2SE defaults to user.home directory
     * If null, external files will not be supported.
     */
    public static FileManager createFileManager(NMTContext context, String assetExt, String assetPath, String internalFilePath, String externalFilePath)
    {
        return new FileManagerOem(context, assetExt, assetPath, internalFilePath, externalFilePath);
    }
}
