package com.nuance.dragon.toolkit.oem.api;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Used to find files and create file handlers. Supported 4 forms of directories:
 * <p>Assets: A file system that is built into an Android application (e.g. an APK).
 * <br>Expansion files: Google Play file system for APKs > 50 MB.
 * <br>Internal: The device's built in-disk.
 * <br>External: The device's external disk (e.g. an SD card)
 */
public abstract class FileManager
{
    /**
     * Status codes related to expansion file mounting.
     */
    public static enum ExpansionStatus
    {
        OK, FILE_NOT_FOUND, DIR_NOT_FOUND, FILE_CORRUPTED, SDCARD_ERROR, UNPACKING_ERROR, UNKNOWN_ERROR
    }

    public static final int MODE_READ = 0;
    public static final int MODE_WRITE = 1;

    protected final String _assetExt;
    private final String _assetPath;
    private final TreeSet<String> _assets;
    protected final String _internalFilePath;
    protected final String _externalFilePath;
    private final NMTContext _context;

    protected FileManager(NMTContext context, String assetExt, String assetPath, String internalFilePath, String externalFilePath)
    {
        Checker.checkArgForNull("context", context);
        Checker.checkArgsForNull("assetExt, assetPath, internalFilePath and externalFilePath",
                                 assetExt, assetPath, internalFilePath, externalFilePath);

        _context = context.getAppContext();
        _assetExt = assetExt == null ? "" : assetExt;
        _assetPath = assetPath;
        _assets = new TreeSet<String>();

        if (internalFilePath == null)
            _internalFilePath = null;
        else if (internalFilePath.length() == 0)
            _internalFilePath = _context.getFilesDir().toString();
        else
            _internalFilePath = _context.getFilesDir() + File.separator + internalFilePath;

        if (externalFilePath == null)
            _externalFilePath = null;
        else if (externalFilePath.length() == 0)
            _externalFilePath = _context.getExternalStorageDirectory().toString();
        else
            _externalFilePath = externalFilePath;

        if (assetPath != null)
        {
            String [] assets = _context.listAssets(assetPath);
            for (String a : assets)
            {
                if (a.endsWith(_assetExt))
                    _assets.add(a);
            }
        }
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result
                + ((_assetExt == null) ? 0 : _assetExt.hashCode());
        result = prime * result
                + ((_assetPath == null) ? 0 : _assetPath.hashCode());
        result = prime * result + ((_assets == null) ? 0 : _assets.hashCode());
        // Commented intentionally
        //result = prime * result
        //        + ((_context == null) ? 0 : _context.hashCode());
        //result = prime * result
        //        + ((_expansions == null) ? 0 : _expansions.hashCode());
        result = prime
                * result
                + ((_externalFilePath == null) ? 0 : _externalFilePath
                        .hashCode());
        result = prime
                * result
                + ((_internalFilePath == null) ? 0 : _internalFilePath
                        .hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        FileManager other = (FileManager) obj;
        if (_assetExt == null) {
            if (other._assetExt != null)
                return false;
        } else if (!_assetExt.equals(other._assetExt))
            return false;
        if (_assetPath == null) {
            if (other._assetPath != null)
                return false;
        } else if (!_assetPath.equals(other._assetPath))
            return false;
        if (_assets == null) {
            if (other._assets != null)
                return false;
        } else if (!_assets.equals(other._assets))
            return false;
        // Commented intentionally
        //if (_context == null) {
        //    if (other._context != null)
        //        return false;
        //} else if (!_context.equals(other._context))
        //    return false;
        //if (_expansions == null) {
        //    if (other._expansions != null)
        //        return false;
        //} else if (!_expansions.equals(other._expansions))
        //    return false;
        if (_externalFilePath == null) {
            if (other._externalFilePath != null)
                return false;
        } else if (!_externalFilePath.equals(other._externalFilePath))
            return false;
        if (_internalFilePath == null) {
            if (other._internalFilePath != null)
                return false;
        } else if (!_internalFilePath.equals(other._internalFilePath))
            return false;
        return true;
    }

    /**
     * Get the expansion file status.
     * @param fileName The absolute expansion file name (ex: "/sdcard/Android/obb/com.nuance.balerion/main.1.com.nuance.balerion.obb").
     * @return A status code from {@link ExpansionStatus}.
     */
    public abstract ExpansionStatus getExpansionFileStatus(String fileName);

    /**
     * This will mount data carried in the expansion file.
     * @param fileName The absolute expansion file name (ex: "/sdcard/Android/obb/com.nuance.balerion/main.1.com.nuance.balerion.obb").
     * @return A status code from {@link ExpansionStatus}.
     */
    public abstract ExpansionStatus mountExpansionFile(String fileName);

    /**
     * @deprecated Call {@link #mountExpansionFile(String,String)} instead.
     * This will mount data carried in the expansion file.
     * @param path The absolute expansion file path (ex: "/sdcard/Android/obb/com.nuance.balerion/").
     * @param fileName The relative expansion file name (ex: "main.1.com.nuance.balerion.obb").
     * @param subPath The path within the expansion file.
     * @return A status code from {@link ExpansionStatus}.
     */
    @Deprecated
    public abstract ExpansionStatus mountExpansionFile(String path, String fileName, String subPath);

    /**
     * This will mount data carried in the expansion file.
     * @param fileName The absolute expansion file name (ex: "/sdcard/Android/obb/com.nuance.balerion/main.1.com.nuance.balerion.obb").
     * @param subPath The path within the expansion file.
     * @return A status code from {@link ExpansionStatus}.
     */
    public abstract ExpansionStatus mountExpansionFile(String fileName, String subPath);

    /**
     * @deprecated Call {@link #unmountExpansionFile(String)} instead.
     * This will unmount data carried in the expansion file.
     * @param path The absolute expansion file path (ex: "/sdcard/Android/obb/com.nuance.balerion/").
     * @param fileName The expansion file name (ex: "main.1.com.nuance.balerion.obb").
     */
    @Deprecated
    public abstract void unmountExpansionFile(String path, String fileName);

    /**
     * This will unmount data carried in the expansion file.
     * @param fileName The absolute expansion file name (ex: "/sdcard/Android/obb/com.nuance.balerion/main.1.com.nuance.balerion.obb").
     */
    public abstract void unmountExpansionFile(String fileName);

    /**
     * Get a file output stream for writing to this file. Will automatically
     * create any parent directories that don't yet exist. Only applicable
     * when the internal directory is supported.
     *
     * @param fileName The file name to search for
     * @return The stream, or null if the stream couldn't be opened.
     */
    public FileOutputStream openFileForWriting(String fileName)
    {
        Checker.checkState(this, _internalFilePath != null, "File manager isn't configured with an internal file path");
        Checker.checkArgForNull("fileName", fileName);

        File f = new File(_internalFilePath, fileName);

        boolean dirExists = ensureDirectory(f.getParentFile());
        if (dirExists)
        {
            try {
                return new FileOutputStream(f);
            }
            catch (Exception e) {
                Logger.error(this, "Unable to open output stream for file " + fileName);
            }
        }
        return null;
    }

    /**
     * Get a file output stream for writing characters to this file. Will automatically
     * create any parent directories that don't yet exist. Only applicable
     * when the internal directory is supported.
     *
     * @param fileName The file name to search for
     * @return The stream, or null if the stream couldn't be opened.
     */
    public FileWriter openFileForWritingChar(String fileName)
    {
        Checker.checkState(this, _internalFilePath != null, "File manager isn't configured with an internal file path");
        Checker.checkArgForNull("fileName", fileName);

        File f = new File(_internalFilePath, fileName);

        boolean dirExists = ensureDirectory(f.getParentFile());
        if (dirExists)
        {
            try {
                return new FileWriter(f);
            }
            catch (Exception e) {
                Logger.error(this, "Unable to open output stream for file " + fileName);
            }
        }
        return null;
    }

    /**
     * Get the absolute path of a file for writing to this file. Will automatically
     * create any parent directories that don't yet exist. Only applicable
     * when the internal directory is supported.
     *
     * @param fileName The file name to search for
     * @return The absolute path
     */
    public String getAbsFilePathForWriting(String fileName)
    {
        Checker.checkState(this, _internalFilePath != null, "File manager isn't configured with an internal file path");
        Checker.checkArgForNull("fileName", fileName);

        File f = new File(_internalFilePath, fileName);

        boolean dirExists = ensureDirectory(f.getParentFile());
        if (dirExists)
        {
            try {
                return (f.createNewFile() ? f.getAbsolutePath() : null);
            }
            catch (Exception e) {
                Logger.error(this, "Unable to create file: " + fileName + " for writing");
            }
        }
        return null;
    }

    /**
     * Get a file input stream for reading from this file.
     * @param fileName The file name to search for
     * @return The stream, or null if the stream couldn't be opened
     */
    public FileInputStream openFileForReading(String fileName)
    {
        Checker.checkArgForNull("fileName", fileName);

        File f = null;
        NMTAssetFileDescriptor af = null;
        if (_externalFilePath != null)
        {
            f = new File(_externalFilePath, fileName);
            if (!f.exists())
                f = null;
        }

        if (_internalFilePath != null && f == null)
        {
            f = new File(_internalFilePath, fileName);
            if (!f.exists())
                f = null;
        }

        if (f == null) {
            af = getAssetFile(fileName, _assetExt);
        }

        if (f != null || af != null)
        {
            try {
                return (f != null) ? new FileInputStream(f) : af.createInputStream();
            }
            catch (Exception e) {
                Logger.error(this, "Unable to open input stream for file " + fileName);
            }
        }
        return null;
    }

    /**
     * Get the absolute path of a file for reading from this file.
     * Only applicable when the external or internal directory is supported.
     *
     * @param fileName The file name to search for
     * @return The absolute path
     */
    public String getAbsFilePathForReading(String fileName)
    {
        Checker.checkArgForNull("fileName", fileName);

        File f = null;
        if (_externalFilePath != null)
        {
            f = new File(_externalFilePath, fileName);
            if (!f.exists())
                f = null;
        }

        if (_internalFilePath != null && f == null)
        {
            f = new File(_internalFilePath, fileName);
            if (!f.exists())
                f = null;
        }

        return (f != null ? f.getAbsolutePath() : null);
    }

    /**
     * Returns an array of filenames containing the passed pattern. Searches in
     * all supported directories
     * @param pattern the pattern to check if it's contained in a file name
     * @return the array of filenames if there are matches. Otherwise, returns null.
     */
    public String[] findFiles(final String pattern)
    {
        return findFiles( new FilenameFilter()
        {
            @Override
            public boolean accept(File dir, String filename)
            {
                return filename.contains(pattern);
            }

            @Override
            public String toString()
            {
                return pattern;
            }
        });
    }

    /**
     * Returns an array of filenames exactly matching the passed regular
     * expression pattern. Searches in all supported directories
     * @param pattern the regular expression pattern to match against the file
     * name.
     * @return the array of filenames if there are matches. Otherwise, returns null.
     */
    public String[] findFiles(final Pattern pattern)
    {
        return findFiles(new FilenameFilter()
        {
            @Override
            public boolean accept(File dir, String filename)
            {
                Matcher patternMatcher = pattern.matcher(filename);
                return patternMatcher.matches();
            }

            @Override
            public String toString()
            {
                return pattern.pattern();
            }
        });
    }

    /**
     * Returns an array of filenames containing the passed pattern. Searches in
     * all supported directories
     * @param filter The filename filter
     * @return the array of filenames if there are matches. Otherwise, returns null.
     */
    protected String[] findFiles(FilenameFilter filter)
    {
        String retMatches[] = null;

        try {
            ArrayList<String> matches = new ArrayList<String>();

            if (_assetPath != null) {
                for (String fileName : _assets) {
                    String nameNoAssetExt = fileName.substring(0, fileName.endsWith(_assetExt) ? fileName.length()-_assetExt.length()
                                                                                               : fileName.length());
                    if (filter.accept(null, nameNoAssetExt)) {
                        matches.add(nameNoAssetExt);
                    }
                }
            }

            if (_internalFilePath != null) {
                String filenames[] = new File(_internalFilePath).list(filter);
                if (filenames != null) {
                    for (String fileName : filenames)
                        matches.add(fileName);
                }
            }

            if (_externalFilePath != null) {
                String filenames[] = new File(_externalFilePath).list(filter);
                if (filenames != null) {
                    for (String fileName : filenames)
                        matches.add(fileName);
                }
            }

            if (matches.size() > 0) {
                retMatches = new String[matches.size()];
                matches.toArray(retMatches);
            }
            else {
                Logger.debug(this, "No file matches \"" + filter.toString() + "\" pattern.");
                retMatches = null;
            }
        }
        catch (Exception e) {
            Logger.error(this, "Error while searching for files with a pattern.", e);
            retMatches = null;
        }

        return retMatches;
    }

    /**
     * Returns true if the file exists. Searches in all supported directories.
     * @param fileName The filename to search for. It will be concatenated with
     *                 assetExt to be searched in built-in assets directory.
     * @return true if the file exists.
     */
    public boolean exists(String fileName)
    {
        if (_assetPath != null && _assets.contains(fileName + _assetExt))
            return true;

        if (_internalFilePath != null && new File(_internalFilePath, fileName).exists())
            return true;

        if (_externalFilePath != null && new File(_externalFilePath, fileName).exists())
            return true;

        return false;
    }

    /**
     * Renames a file. Only applicable when the internal directory is supported.
     * @param from The name of the file to rename
     * @param to The new filename
     * @return true if the file was renamed.
     */
    public boolean rename(String from, String to)
    {
        if (_internalFilePath != null)
        {
            File old = new File(_internalFilePath, from);
            File dest = new File(_internalFilePath, to);
            return old.renameTo(dest);
        }
        Logger.warn(this, "Unable to rename file " + from + " to " + to);
        return false;
    }

    /**
     * Delete a file. Only applicable when the internal directory is supported.
     * @param file of the file to be deleted
     * @return true if the file was deleted.
     *
     * @throws NullPointerException if file is null
     */
    public boolean delete(String file)
    {
        if (_internalFilePath != null)
        {
            return new File(_internalFilePath, file).delete();
        }
        Logger.warn(this, "Unable to delete file " + file);
        return false;
    }

    /**
     * Create a directory. Only applicable when the internal directory is supported.
     *
     * @param dir to be created
     * @return true if the directory was created.
     *
     * @throws NullPointerException if dir is null
     */
    public boolean createDir(String dir)
    {
        if (_internalFilePath != null)
        {
            File directory = new File(_internalFilePath, dir);
            return ensureDirectory(directory);
        }
        Logger.warn(this, "Unable to create directory " + dir);
        return false;
    }

    /**
     * Delete a directory. Only applicable when the internal directory is supported.
     *
     * @param dir to be deleted
     * @return true if the directory was deleted.
     *
     * @throws NullPointerException if dir is null
     */
    public boolean deleteDir(String dir) {
        if (_internalFilePath != null)
        {
            File f = new File(_internalFilePath, dir);
            return deleteDir(f);
        }
        Logger.warn(this, "Unable to delete directory " + dir);
        return false;
    }

    private boolean deleteDir(File dir)
    {
        if (!dir.exists()) return false;

        if (dir.isDirectory())
        {
            for (File item : dir.listFiles())
            {
                if (item.isDirectory())
                {
                    if (!deleteDir(item))
                        return false;
                }
                else
                {
                    if (!item.delete())
                        return false;
                }
            }
        }
        return dir.delete();
    }

    /**
     * Get the file size in all supported directories
     *
     * @param fileName Name of file to analyze. It will be concatenated with assetExt as a full file
     *                 name for using in assets directory
     * @return -1 if the file doesn't exist.
     */
    public int getSize(String fileName)
    {
        if (_externalFilePath != null)
        {
            File file = new File(_externalFilePath, fileName);
            if (file.exists()) return (int)file.length();
        }

        if (_internalFilePath != null)
        {
            File file = new File(_internalFilePath, fileName);
            if (file.exists()) return (int)file.length();
        }

        NMTAssetFileDescriptor afd = getAssetFile(fileName, _assetExt);
        if (afd != null)
        {
            int len = (int)afd.getDeclaredLength();
            return len;
        }

        return -1;
    }

    /**
     * Internal method for getting an asset file. Only called if the mode is
     * {@link #MODE_READ}. Returns null if no asset file is found.
     * @param fileName name of file to get.
     * @param ext extension of file to get.
     * @return The asset file, or null if it couldn't be found.
     */
    protected NMTAssetFileDescriptor getAssetFile(String fileName, String ext)
    {
        if (_assetPath != null)
        {
            String fileNameWithExt = fileName + ext;

            if (_assets.contains(fileNameWithExt))
            {
                String assetName = _assetPath.length() == 0 ? fileNameWithExt : _assetPath + File.separator + fileNameWithExt;
                NMTAssetFileDescriptor afd = _context.openFd(assetName);
                if (afd != null)
                {
                    Logger.debug(this, "Found asset file " + fileNameWithExt);
                    return afd;
                } else
                {
                    Logger.error(this, "Unexpected error opening asset file " + fileNameWithExt);
                }
            } else
            {
                Logger.debug(this, "Asset file " + fileName + " not found");
            }
        }
        return null;
    }

    /**
     * Internal method for getting a real file (either on device's internal or external storage).
     * @param directory Directory to look for file. Null is equivalent to a directory that doesn't exist
     * @param fileName name of file to find
     * @param mode The read write mode. Should either be {@link #MODE_READ} or
     * {@link #MODE_WRITE}. Anything else would result in unexpected behavior
     * @return The file, or null if it couldn't be found.
     */
    protected File getRealFile(String directory, String fileName, int mode)
    {
        if (directory != null)
        {
            File file = new File(directory, fileName);
            if (!file.exists()) {
                if (mode == MODE_WRITE) {
                    /* In MODE_WRITE attempt to create the file if it does not exist. */
                    try {
                        if (!ensureDirectory(new File(directory))) {
                            Logger.warn(this, "Target folder: " + directory + " is not a directory.  Cannot create " + fileName);
                            return null;
                        }
                        file.createNewFile();
                    } catch (IOException e) {
                        Logger.error(this, "File " + directory + File.separator + fileName + " could not be created: " + e);
                        return null;
                    }
                } else {
                    /* In MODE_READ, return an error if the file does not exist. */
                    Logger.debug(this, "File " + fileName + " not found");
                    return null;
                }
            }
            Logger.debug(this, "Found file " + directory + File.separator + fileName);
            return file;
        }
        return null;
    }

    /**
     * Determine if dir is a directory. If not, attempts to create the directory
     * (and all missing parents)
     * @param dir
     * @return true if the directory exists or was created
     */
    protected boolean ensureDirectory(File dir)
    {
        if (dir == null)
            return false;

        if (dir.isDirectory())
            return true;

        File parentDir = dir.getParentFile();
        if (parentDir == null)
            return false;

        if (!parentDir.isDirectory())
        {
            boolean success = ensureDirectory(parentDir);
            if (!success)
                return false;
        }
        return dir.mkdir();
    }
}
