package com.nuance.dragon.toolkit.oem.api.internal;

import java.util.LinkedList;

import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.impl.AsyncTaskOem;

/**
 * A simplified version of an Android handler that executes Runnables using
 * AsyncTasks, to avoid the need for a dedicated worker thread.
 * <p>
 * This is very light-weight, as it does nothing when it is sitting idle.
 */
public final class AsyncTaskHandler
{
    /**
     * A runnable that performs a task on the async task thread, and then
     * notifies of the result on the calling thread.
     * @param <ResultType> The type of the object to pass in the calling thread
     * notification
     */
    public static abstract class TaskRunnable<ResultType extends Object>
    {
        /**
         * Runs the task on the async task thread.
         * @return The result of the task
         */
        public abstract ResultType run();

        /**
         * Called on the calling thread after {@link #run} has been run
         * @param result The result obtained from {@link #run}
         */
        public abstract void onPostRun(ResultType result);

        private final void doRun(NMTHandler mainHandler)
        {
            final ResultType result = run();
            mainHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    onPostRun(result);
                }
            });
        }
    }

    private class Task extends AsyncTaskOem<Void, Void, Void>
    {
        @Override
        protected Void doInBackground(Void... params)
        {
            while (true)
            {
                Runnable r;
                synchronized (_taskSync)
                {
                    if (_currentTask != this)
                    {
                        Logger.warn(AsyncTaskHandler.this, "Task is not current, aborting");
                        break;
                    }

                    if (_tasks.isEmpty())
                    {
                        _currentTask = null;
                        break;
                    }

                    r = _tasks.removeFirst();
                }

                if (isCancelled())
                {
                    Logger.warn(AsyncTaskHandler.this, "Task is canceled, aborting");
                    break;
                }

                synchronized (_runnableSync)
                {
                    r.run();
                }
            }
            return null;
        }
    }

    private final LinkedList<Runnable> _tasks;
    private final Object _taskSync = new Object();
    private final Object _runnableSync = new Object();
    private Task _currentTask;

    public AsyncTaskHandler()
    {
        _tasks = new LinkedList<Runnable>();
    }

    /**
     * Post a new runnable. It is guaranteed that runnables will be executed in
     * the order that they are added, and that only one runnable will be
     * executing at once.
     * <p>
     * Note: it is possible that different threads will be used to run different
     * runnables; however, the runnables are all synchronized with each other,
     * so it is almost as though they are run on the same thread
     * @param runnable
     */
    public void post(Runnable runnable)
    {
        synchronized (_taskSync)
        {
            _tasks.add(runnable);

            if (_currentTask == null)
            {
                _currentTask = new Task();
                _currentTask.execute((Void)null);
            }
        }
    }

    /**
     * Like {@link #post(TaskRunnable)}, but posts a task runnable, which
     * additionally notifies the calling thread with a result on completion.
     * @param runnable
     */
    public void post(final TaskRunnable<? extends Object> runnable)
    {
        final NMTHandler h = Factory.createNMTHandler();
        synchronized (_taskSync)
        {

            _tasks.add(new Runnable()
            {
                @Override
                public void run()
                {
                    runnable.doRun(h);
                }
            });

            if (_currentTask == null)
            {
                _currentTask = new Task();
                _currentTask.execute((Void)null);
            }
        }
    }

    /**
     * Cancel any runnables that are waiting to execute.
     */
    public void cancelAll()
    {
        synchronized (_taskSync)
        {
            _tasks.clear();
            if (_currentTask != null)
            {
                _currentTask.cancel(false);
                _currentTask = null;
            }
        }
    }
}
