package com.nuance.dragon.toolkit.oem.impl;

import java.io.File;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;

import android.content.res.AssetFileDescriptor;
import android.os.Environment;

import com.nuance.dragon.toolkit.file.FileJni;
import com.nuance.dragon.toolkit.file.FileZip;
import com.nuance.dragon.toolkit.file.FileZip.ZipEntryRO;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTAssetFileDescriptor;
import com.nuance.dragon.toolkit.oem.api.NMTContext;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.util.internal.Pair;

public class FileManagerOem extends FileManager {

    private class ExpansionFile
    {
        final FileZip fileZip;
        final String subPath;

        ExpansionFile(FileZip fileZip, String subPath)
        {
            this.fileZip = fileZip;
            this.subPath = subPath;
        }
    }

    private final LinkedHashMap<String, ExpansionFile> _expansions;

    public FileManagerOem(NMTContext context, String assetExt, String assetPath)
    {
        this(context, assetExt, assetPath, null, null);
    }

    public FileManagerOem(NMTContext context, String internalFilePath)
    {
        this(context, null, null, internalFilePath, null);
    }

    public FileManagerOem(NMTContext context, String assetExt, String assetPath, String internalFilePath)
    {
        this(context, assetExt, assetPath, internalFilePath, null);
    }

    public FileManagerOem(NMTContext context, String assetExt, String assetPath, String internalFilePath, String externalFilePath)
    {
        super(context, assetExt, assetPath, internalFilePath, externalFilePath);
        _expansions = new LinkedHashMap<String, ExpansionFile>();
    }

    /**
     * Get the specified file (for JNI). <p>
     * Search with the following priority: <br>
     *  - External files <br>
     *  - Internal files <br>
     *  - Asset files <p>
     * TODO: it seems like there can be problems with the file descriptor getting
     * closed prematurely when using this to read/write on the Java side (maybe
     * C too?).
     * @param fileName The file name to search for in all supported directories
     * @param mode The read write mode. Should either be {@link #MODE_READ} or
     * {@link #MODE_WRITE}. Anything else would result in unexpected behavior
     * @return The file, or null if it couldn't be found or opened
     *
     */
    FileJni getFileJni(String fileName, int mode)
    {
        if (mode == MODE_READ)
        {
            File file = getRealFile(_externalFilePath, fileName, MODE_READ);
            if (file != null)
                return new FileJni(file);
        }

        File file  = getRealFile(_internalFilePath, fileName, mode);
        if (file != null)
            return new FileJni(file);

        AssetFileDescriptor afd = null;
        if (mode == MODE_READ)
        {
        	NMTAssetFileDescriptor nmtAfd = getAssetFile(fileName, _assetExt);
        	if (nmtAfd != null)
        		afd = (AssetFileDescriptor)nmtAfd.getNativeDescriptor();
        }
        return afd == null ? null : new FileJni(afd);
    }

    @Override
    public ExpansionStatus getExpansionFileStatus(String fileName)
    {
        Checker.checkStringArgForNullOrEmpty("fileName", fileName);

        ExpansionStatus status = checkExpansionStorage(fileName);
        if (status == ExpansionStatus.OK)
        {
            status = mountExpansionFileInternal(fileName).second;
        }

        return status;
    }

    @Override
    public ExpansionStatus mountExpansionFile(String fileName)
    {
        return mountExpansionFile(fileName, "");
    }

    @Override
    public ExpansionStatus mountExpansionFile(String path, String fileName, String subPath)
    {
        Checker.checkArgForNull("path", path);
        Checker.checkStringArgForNullOrEmpty("fileName", fileName);

        return mountExpansionFile(path + File.separator + fileName, subPath);
    }

    @Override
    public ExpansionStatus mountExpansionFile(String fileName, String subPath)
    {
        Checker.checkStringArgForNullOrEmpty("fileName", fileName);
        Checker.checkArgForNull("subPath", subPath);

        ExpansionStatus status = ExpansionStatus.OK;
        FileZip fileZip = null;

        status = checkExpansionStorage(fileName);
        if (status == ExpansionStatus.OK)
        {
            Pair<FileZip, ExpansionStatus> p = mountExpansionFileInternal(fileName);
            fileZip = p.first;
            status = p.second;
        }

        if (status == ExpansionStatus.OK)
            _expansions.put(fileName, new ExpansionFile(fileZip, subPath));

        return status;
    }

    @Override
    public void unmountExpansionFile(String path, String fileName)
    {
        Checker.checkArgForNull("path", path);
        Checker.checkStringArgForNullOrEmpty("fileName", fileName);

        unmountExpansionFile(path + File.separator + fileName);
    }

    @Override
    public void unmountExpansionFile(String fileName) {
        Checker.checkStringArgForNullOrEmpty("fileName", fileName);

        if (_expansions.remove(fileName) == null)
            Logger.warn(this, "Unable to unmount " + fileName);
    }

    @Override
    protected String[] findFiles(FilenameFilter filter)
    {
        String retMatches[] = null;

        ArrayList<String> matches = new ArrayList<String>();

        ArrayList<ExpansionFile> tempExpansions = new ArrayList<ExpansionFile>();
        for (String key : _expansions.keySet())
        {
            tempExpansions.add(_expansions.get(key));
        }

        for (int idx = tempExpansions.size() - 1; idx >= 0; idx--)
        {
            if (tempExpansions.get(idx).fileZip != null)
            {
                String subPath = tempExpansions.get(idx).subPath.length() > 0 ? tempExpansions.get(idx).subPath + File.separator : "";
                ZipEntryRO entries[] = tempExpansions.get(idx).fileZip.getEntriesAt(subPath);
                if (entries != null) {
                    for (ZipEntryRO entry : entries) {
                        String sanitizedFileName = entry.mFileName.replace(subPath, "");
                        if (filter.accept(null, sanitizedFileName)) {
                            matches.add(sanitizedFileName);
                        }
                    }
                }
            }
        }

        String[] filesMatched = super.findFiles(filter);
        if (filesMatched != null)
            matches.addAll(Arrays.asList(super.findFiles(filter)));

        if (matches.size() > 0) {
            retMatches = new String[matches.size()];
            matches.toArray(retMatches);
        }
        else {
            Logger.debug(this, "No file matches \"" + filter.toString() + "\" pattern.");
            retMatches = null;
        }

        return retMatches;
    }

    @Override
    public boolean exists(String fileName) {
        ArrayList<ExpansionFile> tempExpansions = new ArrayList<ExpansionFile>();
        for (String key : _expansions.keySet())
        {
            tempExpansions.add(_expansions.get(key));
        }

        for (int idx = tempExpansions.size() - 1; idx >= 0; idx--)
        {
            if (tempExpansions.get(idx).fileZip != null)
            {
                String subPath = tempExpansions.get(idx).subPath.length() > 0 ? tempExpansions.get(idx).subPath + File.separator : "";
                for (ZipEntryRO entry : tempExpansions.get(idx).fileZip.getEntriesAt(subPath)) {
                    if (entry.mFileName.contains(fileName))
                        return true;
                }
            }
        }
        return super.exists(fileName);
    }

    @Override
    protected NMTAssetFileDescriptor getAssetFile(String fileName, String ext)
    {
        ArrayList<ExpansionFile> tempExpansions = new ArrayList<ExpansionFile>();
        for (String key : _expansions.keySet())
        {
            tempExpansions.add(_expansions.get(key));
        }
        for (int idx = tempExpansions.size() - 1; idx >= 0; idx--)
        {
            if (tempExpansions.get(idx).fileZip != null)
            {
                String strippedFileName = fileName.replace(tempExpansions.get(idx).subPath + File.separator, "");
                AssetFileDescriptor afd = tempExpansions.get(idx).fileZip.getAssetFileDescriptor(tempExpansions.get(idx).subPath + File.separator + strippedFileName);
                if (afd != null)
                {
                    Logger.debug(this, "Found asset file (in expansion file) " + fileName);
                    return new AssetFileDescriptorOem(afd);
                }
            }
        }
        return super.getAssetFile(fileName, ext);
    }

    private static ExpansionStatus checkExpansionStorage(String fileName)
    {
        ExpansionStatus status = ExpansionStatus.UNKNOWN_ERROR;

        boolean isSdCardPath = fileName.contains(Environment.getExternalStorageDirectory().getPath());

        if (!isSdCardPath || Environment.getExternalStorageState().equals(Environment.MEDIA_MOUNTED))
        {
            // Build the full path to the app's expansion files
            File expPath = new File(fileName).getParentFile();

            // Check that expansion file path exists
            if (expPath != null && expPath.exists())
            {
                status = ExpansionStatus.OK;
            }
            else
            {
                status = ExpansionStatus.DIR_NOT_FOUND;
            }
        }
        else
        {
            status = ExpansionStatus.SDCARD_ERROR;
        }

        return status;
    }

    private static Pair<FileZip, ExpansionStatus> mountExpansionFileInternal(final String fileName)
    {
        FileZip expansionZip = null;
        ExpansionStatus status = ExpansionStatus.UNKNOWN_ERROR;

        try
        {
            File expansion = new File(fileName);
            if (expansion.isFile())
            {
                expansionZip = new FileZip(fileName);
                status = ExpansionStatus.OK;
            }
            else {
                status = ExpansionStatus.FILE_NOT_FOUND;
            }
        }
        catch (Exception e)
        {
            Logger.warn(FileManager.class, "Unable to parse expansion files.");
            status = ExpansionStatus.FILE_CORRUPTED;
        }

        return new Pair<FileZip, ExpansionStatus>(expansionZip, status);
    }
}
