package com.nuance.dragon.toolkit.recognition;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.cloudservices.DataParam;
import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognitionResult;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.core.data.DataSpec;

/**
 * Interprets simple Cloud recognizer results.
 */
public class CloudInterpreter implements RecognitionInterpreter<CloudRecognitionResult>
{
    public static final String KEY_CHOICES = "choices";
    public static final String KEY_SCORE = "score";
    public static final String KEY_TRANSCRIPTION = "trans";
    public static final String KEY_PIECES = "pieces";

    private final DataSpec _resultSpec;

    /**
     * @param resultSpec A data-spec that transforms the NMAS result into the following
     * structure:
     * <p>
     * <pre>
     * { "choices" : [ { "trans" : "The transcription",
     *                   "score" : (int),               // Optional score
     *                   "pieces" : [                   // Optional recognition pieces
     *                               { "trans" : "The",
     *                                 "score" : (int) },   // Optional score
     *                               { "trans" : "transcription",
     *                                 "score" : (int) }
     *                              ]
     *                 } .. ]
     * </pre>
     */
    public CloudInterpreter(DataSpec resultSpec)
    {
        _resultSpec = resultSpec;
    }

    @Override
    public InterpretedRecognition getInterpretedResult(CloudRecognitionResult cloudResult) throws InterpretException
    {
        InterpretedRecognition recognition = new InterpretedRecognition();
        Data.Dictionary dict = cloudResult.getDictionary();
        if (dict == null)
            throw new InterpretException("No result dictionary found");

        if (_resultSpec != null)
        {
            Data processed = _resultSpec.process(dict);
            dict = null;
            if (processed != null && processed.getType() == Data.TYPE_DICT)
            {
                dict = (Data.Dictionary)processed;
            }
        }

        if (dict == null)
            throw new InterpretException("Unable to process result dictionary according to data spec");

        Data.Sequence choices = dict.getSequence(KEY_CHOICES);
        if (choices == null)
            throw new InterpretException("No choices found");

        for (int i = 0; i < choices.size(); i++)
        {
            Data.Dictionary choiceDict = choices.getDictionary(i);
            if (choiceDict != null)
            {
                Data.String trans = choiceDict.getString(KEY_TRANSCRIPTION);
                Data.Integer transScore = choiceDict.getInt(KEY_SCORE);

                if (trans == null)
                    throw new InterpretException("No transcription found");

                List<InterpretedRecognition.Phrase> phrases = new ArrayList<InterpretedRecognition.Phrase>();
                List<InterpretedRecognition.Word> words = new ArrayList<InterpretedRecognition.Word>();
                Data.Sequence pieces = choiceDict.getSequence(KEY_PIECES);
                if (pieces == null)
                {
                    words.add(new InterpretedRecognition.Word(trans.value, null, null, InterpretedRecognition.Word.TIMING_NOT_SUPPORTED, InterpretedRecognition.Word.TIMING_NOT_SUPPORTED));
                    phrases.add(new InterpretedRecognition.Phrase(null, trans.value, words, null, true));
                } else
                {
                    for (int j = 0; j < pieces.size(); j++)
                    {
                        Data.Dictionary piece = pieces.getDictionary(j);
                        Data.String pieceTrans = piece.getString(KEY_TRANSCRIPTION);
                        //Data.Integer pieceTransScore = piece.getInt(KEY_SCORE);
                        words.add(new InterpretedRecognition.Word(pieceTrans.value, null, null, InterpretedRecognition.Word.TIMING_NOT_SUPPORTED, InterpretedRecognition.Word.TIMING_NOT_SUPPORTED));
                    }
                    phrases.add(new InterpretedRecognition.Phrase(null, trans.value, words, null, true));
                }
                recognition.addChoice(trans.value, "CLOUD_RECOGNITION", transScore == null ? 0 : transScore.value, phrases, null);
            }
        }
        return recognition;
    }

    @Override
    public boolean processForCloud(CloudRecognitionResult recognitionResult, List<DataParam> params)
    {
        return true;
    }

    @Override
    public List<String> getUpdateRequiredList(CloudRecognitionResult recognitionResult)
    {
        return new ArrayList<String>(0);
    }
}
