package com.nuance.dragon.toolkit.recognition;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.cloudservices.DataParam;
import com.nuance.dragon.toolkit.elvis.ElvisNbestList;
import com.nuance.dragon.toolkit.elvis.ElvisResult;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.util.internal.ObjectUtils;
import com.nuance.dragon.toolkit.util.internal.StringUtils;


/**
 * Interprets raw Elvis results.
 * <p>
 * The purpose of the Elvis interpreter is to construct an 'interpreted' Elvis
 * result object from the raw Elvis result structure. The interpreted result
 * will be similar to the results returned by other speech recognition engines
 * (e.g. Vocon or NMAS), to enable easier merging of results from multiple sources.
 */
public class ElvisInterpreter implements RecognitionInterpreter<LocalRecognitionResult>
{
    private static final char SLOT_GROUPING_TOKEN = '~';

    @Override
    public InterpretedRecognition getInterpretedResult(LocalRecognitionResult elvisResult) throws InterpretException
    {
        Checker.checkArgForCondition("elvisResult", "instanceof ElvisResult", elvisResult instanceof ElvisResult);

        ElvisResult result = (ElvisResult) elvisResult;
        InterpretedRecognition recognition = new InterpretedRecognition(result.getGateConfidence());
        ElvisNbestList nbest = result.getChoiceList();

        for (ElvisNbestList.Entry entry : nbest)
        {
            List<InterpretedRecognition.Phrase> iPhrases = new ArrayList<InterpretedRecognition.Phrase>();
            List<ElvisNbestList.Entry.Word> words = entry.getWords();
            ArrayList<InterpretedRecognition.Word> iWords = new ArrayList<InterpretedRecognition.Word>(words.size());
            if (words.isEmpty())
                throw new InterpretException("Elvis Nbest list can't be empty");

            ElvisNbestList.Entry.Word nextWord = null;
            String nextSlot = null;
            boolean nextIsGrouped = false;

            String wordToAdd = null;
            String pronToAdd = null;

            int size = words.size();
            for (int i = -1; i < size; i++) // First iteration is just to initialize nextWord
            {
                ElvisNbestList.Entry.Word curWord = nextWord;
                String curSlot = nextSlot;
                boolean curIsGrouped = nextIsGrouped;

                nextWord = i+1 < size ? words.get(i+1) : null;
                if (nextWord != null) {
                    int splitIndex = nextWord.wordSlot == null ? -1 : nextWord.wordSlot.indexOf(SLOT_GROUPING_TOKEN);
                    nextIsGrouped = splitIndex >= 0;
                    nextSlot = nextIsGrouped ? nextWord.wordSlot.substring(0, splitIndex) : nextWord.wordSlot;
                }

                if (curWord != null)
                {
                    boolean lastInSlot = (nextWord == null ||
                                          curWord.isGenericSpeech ||
                                          nextWord.isGenericSpeech ||
                                          !ObjectUtils.compareObjects(curSlot, nextSlot));

                    wordToAdd = StringUtils.appendWithNulls(wordToAdd, curWord.fullPhrase);
                    pronToAdd = StringUtils.appendWithNulls(pronToAdd, curWord.spokenForm);

                    if (!curIsGrouped || lastInSlot)
                    {
                        iWords.add(new InterpretedRecognition.Word(wordToAdd, null, pronToAdd, InterpretedRecognition.Word.TIMING_NOT_SUPPORTED, InterpretedRecognition.Word.TIMING_NOT_SUPPORTED));
                        wordToAdd = pronToAdd = null;
                    }

                    if (lastInSlot)
                    {
                        iPhrases.add(new InterpretedRecognition.Phrase(curSlot, StringUtils.joinStrings(iWords, " "), iWords, null, !curWord.isGenericSpeech));
                        iWords.clear();
                    }
                }
            }
            recognition.addChoice(entry.toString(), entry.getConstraint(), entry.getScore(), iPhrases, null);
        }
        return recognition;
    }

    @Override
    public boolean processForCloud(LocalRecognitionResult elvisResult, List<DataParam> params)
    {
        Checker.checkArgForCondition("elvisResult", "instanceof ElvisResult", elvisResult instanceof ElvisResult);

        // TODO: should we send params to the cloud?

        // Only wait if the top choice has generic speech
        return ((ElvisResult) elvisResult).getChoiceList().get(0).hasGenericSpeech();
    }

    @Override
    public List<String> getUpdateRequiredList(LocalRecognitionResult recognitionResult)
    {
        return new ArrayList<String>(0);
    }
}
