package com.nuance.dragon.toolkit.recognition;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import junit.framework.Assert;

import org.json.JSONException;

import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.json.JSONArray;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.recognition.dictation.DictationResult;
import com.nuance.dragon.toolkit.recognition.dictation.internal.DictationResultImpl;
import com.nuance.dragon.toolkit.util.internal.ObjectUtils;
import com.nuance.dragon.toolkit.util.internal.ReadOnlyList;
import com.nuance.dragon.toolkit.util.internal.StringUtils;

/**
 * A generic interpreted recognition object. This can be used to represent a
 * number of different types of recognition results (e.g. Elvis, Cloud, Vocon).
 */
public final class InterpretedRecognition implements JSONCompliant
{
    /**
     * The smallest unit in a recognition--an individual word. The word may have
     * timing and/or pronunciation information.
     */
    public static final class Word implements JSONCompliant
    {
        /** The value of {@link #getStartTime} and {@link #getEndTime} if timing isn't supported. */
        public static final int TIMING_NOT_SUPPORTED = 0;

        private static final String KEY_TEXT = "text";
        private static final String KEY_FULLTEXT = "fulltext";
        private static final String KEY_PRON = "pron";
        private static final String KEY_START = "start";
        private static final String KEY_END = "end";

        private final String _text;
        private final String _fullText;
        private final String _pron;
        private final long _start;
        private final long _end;

        /**
         * @param text Can't be null
         * @param fullText May be null
         * @param pronunciation May be null
         * @param start Start time in ms, or {@link #TIMING_NOT_SUPPORTED} if unsupported.
         * @param end Start time in ms, or {@link #TIMING_NOT_SUPPORTED} if unsupported.
         */
        public Word(String text, String fullText, String pronunciation, long start, long end)
        {
            _text = text;
            _fullText = fullText;
            _pron = pronunciation;
            _start = start;
            _end = end;
        }

        /**
         * Get the text that was recognized.
         * @return The recognized text
         */
        public String getText()
        {
            return _text;
        }

        /**
         * Get the full form of the text that was recognized. This may differ
         * from {@link #getText} in some cases where recognized text is not the
         * same as the full form of the word. For example, "Andy" may be
         * recognized, but the full word may be "Andrew".
         * @return The full form of the recognized text
         */
        public String getFullText()
        {
            return _fullText == null ? _text : _fullText;
        }

        /**
         * Get the custom pronunciation of the recognized text, if it is
         * different than the recognized text itself.
         * @return The pronunciation, or null if there is not a custom pronunciation.
         */
        public String getPronunciation()
        {
            return _pron;
        }

        /**
         * Get the start time of this word, in ms.
         * <p>If timings are unsupported, both {@link #getStartTime} and
         * {@link #getEndTime} will return {@link #TIMING_NOT_SUPPORTED}.
         * @return The start time, or {@link #TIMING_NOT_SUPPORTED}
         */
        public long getStartTime()
        {
            return _start;
        }

        /**
         * Get the end time of this word, in ms.
         * <p>If timings are unsupported, both {@link #getStartTime} and
         * {@link #getEndTime} will return {@link #TIMING_NOT_SUPPORTED}.
         * @return The end time, or {@link #TIMING_NOT_SUPPORTED}
         */
        public long getEndTime()
        {
            return _end;
        }

        /**
         * Determine whether the start and end time of this word are valid.
         * @return True if the timings are valid
         */
        public boolean isTimingSupported()
        {
            return _start != TIMING_NOT_SUPPORTED || _end != TIMING_NOT_SUPPORTED;
        }

        @Override
        public org.json.JSONObject toJSON()
        {
            JSONObject json = new JSONObject();
            json.tryPut(KEY_TEXT, _text);
            json.tryPut(KEY_FULLTEXT, _fullText);
            json.tryPut(KEY_PRON, _pron);
            json.tryPut(KEY_START, _start);
            json.tryPut(KEY_END, _end);
            return json;
        }

        public static Word createFromJSON(org.json.JSONObject json) throws JSONException
        {
            String text = json.getString(KEY_TEXT);
            String fullText = json.optString(KEY_FULLTEXT, null);
            String pron = json.optString(KEY_PRON, null);
            int start = json.getInt(KEY_START);
            int end = json.getInt(KEY_END);
            return new Word(text, fullText, pron, start, end);
        }

        /**
         * Returns the text of the word.
         * @see java.lang.Object#toString()
         */
        @Override
        public String toString()
        {
            return _text;
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result + (int) (_end ^ (_end >>> 32));
            result = prime * result
                    + ((_fullText == null) ? 0 : _fullText.hashCode());
            result = prime * result + ((_pron == null) ? 0 : _pron.hashCode());
            result = prime * result + (int) (_start ^ (_start >>> 32));
            result = prime * result + ((_text == null) ? 0 : _text.hashCode());
            return result;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            Word other = (Word) obj;
            if (_end != other._end)
                return false;
            if (_fullText == null) {
                if (other._fullText != null)
                    return false;
            } else if (!_fullText.equals(other._fullText))
                return false;
            if (_pron == null) {
                if (other._pron != null)
                    return false;
            } else if (!_pron.equals(other._pron))
                return false;
            if (_start != other._start)
                return false;
            if (_text == null) {
                if (other._text != null)
                    return false;
            } else if (!_text.equals(other._text))
                return false;
            return true;
        }
    }

    /**
     * A recognition 'phrase', which represents a chunk of text from a recognition.
     * The phrase may have an associated key, and start/end timing information.
     */
    public static final class Phrase implements JSONCompliant
    {
        private static final String KEY_KEY = "key";
        private static final String KEY_TRANSCRIPTION = "trans";
        private static final String KEY_WORDS = "words";
        private static final String KEY_EXTRA = "extra";
        private static final String KEY_ISRECOGNIZED = "isrecognized";
        private static final String KEY_DICTATIONRESULT = "dictresult";

        private final String _key;
        private final String _transcription;
        private final ReadOnlyList<Word> _words;
        private final Data.Dictionary _extra;
        private final DictationResult _dictationResult;
        private final boolean _isRecognized;

        /**
         * @param key May be null
         * @param transcription Full transcription text. Can't be null
         * @param words Can't be null
         * @param extra Optional extra information
         * @param isRecognized True if the phrase was successfully recognized.
         * False for unrecognized text--i.e., we know something was spoken, but
         * we dont' know what it was.
         */
        public Phrase(String key, String transcription, List<Word> words, Data.Dictionary extra, boolean isRecognized)
        {
            this(key, transcription, words, null, extra, isRecognized);
        }

        /**
         * @param key May be null
         * @param dictationResult Can't be null
         * @param extra Optional extra information
         */
        public Phrase(String key, DictationResult dictationResult, Data.Dictionary extra)
        {
            this(key, dictationResult.toString(), null, dictationResult, extra, true);
        }

        private Phrase(String key, String transcription, List<Word> words, DictationResult dictationResult, Data.Dictionary extra, boolean isRecognized)
        {
            Assert.assertTrue(!isRecognized || (dictationResult != null || (words != null && !words.isEmpty())));
            _key = key;
            _transcription = transcription;
            _words = words == null ? new ReadOnlyList<Word>() : new ReadOnlyList<Word>(words);
            _extra = extra;
            _dictationResult = dictationResult;
            _isRecognized = isRecognized;
        }

        /**
         * Get the key for this phrase.
         * @return The key (may be null).
         */
        public String getKey()
        {
            return _key;
        }

        /**
         * Determine whether this phrase contains recognized text. Some phrases
         * might represent text that was detected by the engine, but not
         * recognized.
         * @return true if the phrase contains recognized text.
         */
        public boolean isRecognized()
        {
            return _isRecognized;
        }

        /**
         * Get the words in this phrase. Will be empty in the case of a dictation
         * result, or non-recognized text.
         * @return List of words.
         */
        public List<Word> getWords()
        {
            return _words;
        }

        /**
         * Get the start time of this phrase (in milliseconds).
         * @return The start time.
         */
        public long getStartTime()
        {
            if (!_words.isEmpty())
            {
                return _words.get(0)._start;
            }
            return 0;
        }

        /**
         * Get the end time of this phrase (in milliseconds).
         * @return The end time.
         */
        public long getEndTime()
        {
            if (!_words.isEmpty())
            {
                return _words.get(_words.size()-1)._end;
            }
            return 0;
        }

        /**
         * Get the extra data for this phrase.
         * @return May return null.
         */
        public Data.Dictionary getExtra()
        {
            return _extra;
        }

        /**
         * Get the dictation result that comes with this phrase.
         * @return May return null if this phrase does not have a dictation
         * result.
         */
        public DictationResult getDictationResult()
        {
            return _dictationResult;
        }

        /**
         * Returns the text of the phrase.
         * @see java.lang.Object#toString()
         */
        @Override
        public java.lang.String toString()
        {
            return _transcription;
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime
                    * result
                    + ((_dictationResult == null) ? 0 : _dictationResult
                            .hashCode());
            result = prime * result
                    + ((_extra == null) ? 0 : _extra.hashCode());
            result = prime * result + (_isRecognized ? 1231 : 1237);
            result = prime * result + ((_key == null) ? 0 : _key.hashCode());
            result = prime
                    * result
                    + ((_transcription == null) ? 0 : _transcription.hashCode());
            result = prime * result
                    + ((_words == null) ? 0 : _words.hashCode());
            return result;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            Phrase other = (Phrase) obj;
            if (_dictationResult == null) {
                if (other._dictationResult != null)
                    return false;
            } else if (!_dictationResult.equals(other._dictationResult))
                return false;
            if (_extra == null) {
                if (other._extra != null)
                    return false;
            } else if (!_extra.equals(other._extra))
                return false;
            if (_isRecognized != other._isRecognized)
                return false;
            if (_key == null) {
                if (other._key != null)
                    return false;
            } else if (!_key.equals(other._key))
                return false;
            if (_transcription == null) {
                if (other._transcription != null)
                    return false;
            } else if (!_transcription.equals(other._transcription))
                return false;
            if (_words == null) {
                if (other._words != null)
                    return false;
            } else if (!_words.equals(other._words))
                return false;
            return true;
        }

        @Override
        public org.json.JSONObject toJSON()
        {
            JSONObject json = new JSONObject();
            json.tryPut(KEY_KEY, _key);
            json.tryPut(KEY_TRANSCRIPTION, _transcription);
            if (_extra != null)
                json.tryPut(KEY_EXTRA, _extra.toJSON());
            json.tryPutList(KEY_WORDS, _words);
            json.tryPut(KEY_ISRECOGNIZED, _isRecognized);
            if (_dictationResult != null)
                json.tryPut(KEY_DICTATIONRESULT, _dictationResult.toJSON());

            return json;
        }

        public static Phrase createFromJSON(org.json.JSONObject json) throws JSONException
        {
            String key = json.optString(KEY_KEY);
            String transcription = json.getString(KEY_TRANSCRIPTION);
            boolean isRecognized = json.getBoolean(KEY_ISRECOGNIZED);

            org.json.JSONArray a = json.getJSONArray(KEY_WORDS);
            ArrayList<Word> words = new ArrayList<Word>(a.length());
            for (int i = 0; i < a.length(); i++)
            {
                words.add(Word.createFromJSON(a.getJSONObject(i)));
            }

            Data.Dictionary extra = null;
            org.json.JSONObject o = json.optJSONObject(KEY_EXTRA);
            if (o != null)
                extra = Data.Dictionary.createFromJSON(o);

            DictationResult dictationResult = null;
            o = json.optJSONObject(KEY_DICTATIONRESULT);
            if (o != null)
                dictationResult = DictationResultImpl.createFromJSON(o);

            return new Phrase(key, transcription, words, dictationResult, extra, isRecognized);
        }
    }

    /**
     * A recognition choice. A recognition may consist of multiple alternative
     * choices.
     */
    public static final class Choice implements JSONCompliant
    {
        private static final String KEY_TRANSCRIPTION = "trans";
        private static final String KEY_RECOGNITION_TYPE = "type";
        private static final String KEY_SCORE = "score";
        private static final String KEY_PHRASES = "phrases";
        private static final String KEY_EXTRA = "extra";

        private final String _transcription;
        private final String _recognitionType;
        private final Map<String, Phrase> _phraseMap;
        private final List<Phrase> _phrases;
        private final int _score;
        private final Data.Dictionary _extra;
        private final boolean _isAllRecognized;

        private Choice(String transcription,
                       String recognitionType,
                       int score,
                       List<Phrase> phrases,
                       Data.Dictionary extra)
        {
            _recognitionType = recognitionType;
            _phraseMap = new HashMap<String, Phrase>();
            _phrases = new ArrayList<Phrase>(phrases.size());
            _transcription = transcription;
            _score = score;
            _extra = extra;

            boolean allRecognized = true;
            _phrases.addAll(phrases);
            for (Phrase p : phrases)
            {
                _phraseMap.put(p.getKey(), p);
                if (!p.isRecognized())
                    allRecognized = false;
            }
            _isAllRecognized = allRecognized;
        }

        /**
         * Get the type of the recognition (e.g. "call_cmd")
         * @return The type.
         */
        public String getRecognitionType()
        {
            return _recognitionType;
        }

        /**
         * Determine whether all phrases in the choice are fully recognized.
         * @see Phrase#isRecognized
         * @return True if all phrases in the choice are fully recognized
         */
        public boolean isAllRecognized()
        {
            return _isAllRecognized;
        }

        /**
         * Determine whether this choice and another choice match.
         * @param otherChoice The other choice
         * @return true if this choice and the other choice have the same
         * recognition type and exactly the same phrase keys.
         */
        public boolean matches(Choice otherChoice)
        {
            if (!isAllRecognized())
                return true;

            if (!ObjectUtils.compareObjects(otherChoice._recognitionType, _recognitionType))
                return false;

            int len = _phrases.size();

            List<Phrase> otherPhrases = otherChoice._phrases;
            if (otherPhrases.size() != len)
                return false;

            for (int i = 0; i < len; i++)
            {
                Phrase thisPhrase = _phrases.get(i);
                Phrase otherPhrase = otherPhrases.get(i);

                if (!ObjectUtils.compareObjects(thisPhrase.getKey(), otherPhrase.getKey()))
                {
                    return false;
                }
            }
            return true;
        }

        /**
         * Return a new choice, with unrecognized phrases resolved using another
         * interpreted recognition.
         * @param otherRecog The other recognition. This method will search
         * through all its choices until it finds one with a recognized phrase
         * matching the key of the unrecognized phrase in this choice.
         * @param allowUnrecognized True if the returned choice may have
         * remaining unrecognized phrases. False if not--null will be returned
         * if any unrecognized phrases would remain.
         * @return The new choice with as many unrecognized phrases resolved as
         * possible, or null if some phrases couldn't be resolved.
         */
        public Choice resolveUnrecognizedPhrases(InterpretedRecognition otherRecog, boolean allowUnrecognized)
        {
            if (isAllRecognized())
                return this;

            Data.Dictionary extra = null;
            ArrayList<Phrase> newPhrases = new ArrayList<Phrase>(_phrases.size());
            int phraseIndex = -1;
            for (Phrase phrase : _phrases)
            {
                phraseIndex += 1;
                Phrase newPhrase = null;
                if (!phrase._isRecognized)
                {
                    String key = phrase.getKey();
                    if (key != null)
                    {
                        for (Choice otherChoice : otherRecog._choices)
                        {
                            if (matches(otherChoice))
                            {
                                Phrase otherPhrase = otherChoice.getPhrase(key);
                                if (otherPhrase == null && phraseIndex < otherChoice.getPhrases().size())
                                    otherPhrase = otherChoice.getPhrases().get(phraseIndex);
                                if (otherPhrase != null && otherPhrase._isRecognized)
                                {
                                    newPhrase = otherPhrase;
                                    if (extra == null)
                                        extra = otherChoice._extra;
                                    else if (otherChoice._extra != null)
                                        extra.putAll(otherChoice._extra);
                                    break;
                                }
                            }
                        }
                    }

                    if (newPhrase == null)
                    {
                        if (!allowUnrecognized) { // Fail
                            Logger.warn(this, "Unable to resolve unrecognized phrases in " + _transcription);
                            return null;
                        }

                        newPhrase = phrase; // Stick with the old one
                    }
                } else
                {
                    newPhrase = phrase;
                }
                newPhrases.add(newPhrase);
            }

            return new Choice(StringUtils.joinStrings(newPhrases, " "), _recognitionType, _score, newPhrases, extra);
        }

        /**
         * Get the score for this choice. TODO: what does a score mean?
         * @return The score.
         */
        public int getScore()
        {
            return _score;
        }

        /**
         * Get the phrases, in order of their timings, from this choice.
         * @return The phrases.
         */
        public List<Phrase> getPhrases()
        {
            return _phrases;
        }

        /**
         * Get the phrase with the given key.
         * @param key
         * @return The phrase, or null if no phrase exists with the key.
         */
        public Phrase getPhrase(String key)
        {
            return _phraseMap.get(key);
        }

        /**
         * Get the extra data for this choice.
         * @return May return null. Could be used for NLU results.
         */
        public Data.Dictionary getExtra()
        {
            return _extra;
        }

        /**
         * Get the full text transcription of this choice. May not be formatted
         * correctly for display, e.g. there may be unexpected spaces.
         * @see java.lang.Object#toString()
         */
        @Override
        public String toString()
        {
            return _transcription;
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result
                    + ((_extra == null) ? 0 : _extra.hashCode());
            result = prime * result + (_isAllRecognized ? 1231 : 1237);
            result = prime * result
                    + ((_phraseMap == null) ? 0 : _phraseMap.hashCode());
            result = prime * result
                    + ((_phrases == null) ? 0 : _phrases.hashCode());
            result = prime
                    * result
                    + ((_recognitionType == null) ? 0 : _recognitionType
                            .hashCode());
            result = prime * result + _score;
            result = prime
                    * result
                    + ((_transcription == null) ? 0 : _transcription.hashCode());
            return result;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            Choice other = (Choice) obj;
            if (_extra == null) {
                if (other._extra != null)
                    return false;
            } else if (!_extra.equals(other._extra))
                return false;
            if (_isAllRecognized != other._isAllRecognized)
                return false;
            if (_phraseMap == null) {
                if (other._phraseMap != null)
                    return false;
            } else if (!_phraseMap.equals(other._phraseMap))
                return false;
            if (_phrases == null) {
                if (other._phrases != null)
                    return false;
            } else if (!_phrases.equals(other._phrases))
                return false;
            if (_recognitionType == null) {
                if (other._recognitionType != null)
                    return false;
            } else if (!_recognitionType.equals(other._recognitionType))
                return false;
            if (_score != other._score)
                return false;
            if (_transcription == null) {
                if (other._transcription != null)
                    return false;
            } else if (!_transcription.equals(other._transcription))
                return false;
            return true;
        }

        @Override
        public org.json.JSONObject toJSON()
        {
            JSONObject json = new JSONObject();
            json.tryPut(KEY_TRANSCRIPTION, _transcription);
            json.tryPut(KEY_RECOGNITION_TYPE, _recognitionType);
            json.tryPut(KEY_SCORE,  _score);

            if (_extra != null)
                json.tryPut(KEY_EXTRA, _extra);

            JSONArray a = new JSONArray();
            for (Phrase p : _phrases)
            {
                a.tryPut(p.toJSON());
            }
            json.tryPut(KEY_PHRASES, a);

            return json;
        }

        public static Choice createFromJSON(org.json.JSONObject json) throws JSONException
        {
            String trans = json.getString(KEY_TRANSCRIPTION);
            String type = json.getString(KEY_RECOGNITION_TYPE);
            int s = json.getInt(KEY_SCORE);

            org.json.JSONArray a = json.getJSONArray(KEY_PHRASES);
            ArrayList<Phrase> phrases = new ArrayList<Phrase>(a.length());
            for (int i = 0; i < a.length(); i++)
            {
                phrases.add(Phrase.createFromJSON(a.getJSONObject(i)));
            }

            Data.Dictionary extra = null;
            org.json.JSONObject o = json.optJSONObject(KEY_EXTRA);
            if (o != null)
                extra = Data.Dictionary.createFromJSON(o);

            return new Choice(trans, type, s, phrases, extra);
        }
    }

    public static final int GATE_CONFIDENCE_UNAVAILABLE = -1;

    private static final String KEY_CHOICES = "choices";
    private static final String KEY_PROMPT = "prompt";
    private static final String KEY_GATE_CONFIDENCE = "gate_confidence";

    private final List<Choice> _choices;
    private final String _prompt;
    private final int _gateConfidence;

    public InterpretedRecognition()
    {
        this(null);
    }

    public InterpretedRecognition(int gateConfidence)
    {
        _choices = new ArrayList<Choice>();
        _prompt = null;
        _gateConfidence = gateConfidence;
    }

    public InterpretedRecognition(String prompt)
    {
        _choices = new ArrayList<Choice>();
        _prompt = prompt;
        _gateConfidence = GATE_CONFIDENCE_UNAVAILABLE;
    }

    InterpretedRecognition(List<Choice> choices, String prompt, int gateConfidence)
    {
        _choices = new ArrayList<Choice>(choices.size());
        _choices.addAll(choices);
        _prompt = prompt;
        _gateConfidence = gateConfidence;
    }

    /**
     * Get the prompt that accompanies the result.
     * @return May return null
     */
    public String getPrompt()
    {
        return _prompt;
    }

    /**
     * Equivalent to calling {@link Choice#getScore} of the 0th choice.
     * @return The score of the best choice.
     */
    public int getScore()
    {
        if (!_choices.isEmpty())
        {
            return _choices.get(0).getScore();
        }
        return 0;
    }

    /**
     * Returns the confidence that the gate command is correct.
     * 100 indicates highest confidence, 0 indicates lowest confidence.
     * -1 indicates that the gate confidence is not available.
     * (Gate confidence is only available for Elvis premium models)
     * @return The confidence
     */
    public int getGateConfidence()
    {
        return _gateConfidence;
    }

    /**
     * Equivalent to calling {@link Choice#getPhrases}.
     * @return The phrases.
     */
    public List<Phrase> getPhrases()
    {
        if (!_choices.isEmpty())
        {
            return _choices.get(0).getPhrases();
        }
        return new ArrayList<Phrase>(0);
    }

    /**
     * Equivalent to calling {@link Choice#getPhrase}.
     * @param key
     * @return The phrase, or null if no phrase exists with the key.
     */
    public Phrase getPhrase(String key)
    {
        if (!_choices.isEmpty())
        {
            return _choices.get(0).getPhrase(key);
        }
        return null;
    }

    /**
     * Add a choice.
     * @param transcription
     * @param recognitionType e.g. "call_cmd"
     * @param score
     * @param phrases
     * @param extra May be null.
     */
    public void addChoice(String transcription,
                          String recognitionType,
                          int score,
                          List<Phrase> phrases,
                          Data.Dictionary extra)
    {
        Choice c = new Choice(transcription, recognitionType, score, phrases, extra);
        _choices.add(c);
    }

    /**
     * Get the choice count.
     * @return The number of choices.
     */
    public int getChoiceCount()
    {
        return _choices.size();
    }

    /**
     * Get the first choice.
     * @return The first choice (the same as calling {@link #getChoice getChoice(0)}).
     */
    public Choice getFirstChoice()
    {
        return _choices.get(0);
    }

    /**
     * @param index
     * @return The choice at the given index
     */
    public Choice getChoice(int index)
    {
        return _choices.get(index);
    }

    /**
     * Returns the text of the first choice.
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString()
    {
        if (!_choices.isEmpty())
        {
            return _choices.get(0).toString();
        }
        return "[no choices]";
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result
                + ((_choices == null) ? 0 : _choices.hashCode());
        result = prime * result + _gateConfidence;
        result = prime * result + ((_prompt == null) ? 0 : _prompt.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        InterpretedRecognition other = (InterpretedRecognition) obj;
        if (_choices == null) {
            if (other._choices != null)
                return false;
        } else if (!_choices.equals(other._choices))
            return false;
        if (_gateConfidence != other._gateConfidence)
            return false;
        if (_prompt == null) {
            if (other._prompt != null)
                return false;
        } else if (!_prompt.equals(other._prompt))
            return false;
        return true;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject json = new JSONObject();
        JSONArray a = new JSONArray();
        for (Choice c : _choices)
        {
            a.tryPut(c.toJSON());
        }
        json.tryPut(KEY_CHOICES, a);
        json.tryPut(KEY_PROMPT, _prompt);
        json.tryPut(KEY_GATE_CONFIDENCE, _gateConfidence);
        return json;
    }

    public static InterpretedRecognition createFromJSON(org.json.JSONObject json) throws JSONException
    {
        org.json.JSONArray c = json.getJSONArray(KEY_CHOICES);
        ArrayList<Choice> choices = new ArrayList<Choice>(c.length());
        for (int i = 0; i < c.length(); i++)
        {
            choices.add(Choice.createFromJSON(c.getJSONObject(i)));
        }
        return new InterpretedRecognition(choices, json.optString(KEY_PROMPT, null), json.optInt(KEY_GATE_CONFIDENCE, GATE_CONFIDENCE_UNAVAILABLE));
    }
}
