package com.nuance.dragon.toolkit.recognition;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.cloudservices.DataParam;
import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognitionResult;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.recognition.dictation.DictationResult;
import com.nuance.dragon.toolkit.recognition.dictation.DictationResultManager;
import com.nuance.dragon.toolkit.recognition.interpretation.Interpretation;
import com.nuance.dragon.toolkit.recognition.interpretation.InterpretationResult;

/**
 * Interpreter for the NVC_ASR_CMD
 */
public class NvcAsrCloudInterpreter implements RecognitionInterpreter<CloudRecognitionResult>
{
    @Override
    public InterpretedRecognition getInterpretedResult(CloudRecognitionResult recognitionResult) throws InterpretException
    {
        InterpretedRecognition interpretedRecognition = null;

        Data resultData = recognitionResult.getDictionary();
        if (resultData == null || resultData.getType() != Data.TYPE_DICT)
            throw new InterpretException("Couldn't find top-level result dictionary");

        Logger.debug(this, "Trying to parse result: " + resultData.toString());
        Data.Dictionary resultDict = (Data.Dictionary)resultData;

        Data.String prompt = resultDict.getString("prompt");
        Data.String resultType = resultDict.getString("result_format");

        if (prompt != null)
            interpretedRecognition = new InterpretedRecognition(prompt.value);
        else
            interpretedRecognition = new InterpretedRecognition();

        if (resultType == null || !resultType.value.equals("interp_xml_results"))
            throw new InterpretException("Expected result_format == interp_xml_results");

        Data.String interpXmlResult = resultDict.getString("xml_results");
        if (interpXmlResult == null)
            throw new InterpretException("Expected xml_results");

        InterpretationResult ir = DictationResultManager.createInterpretationResult(interpXmlResult.value.getBytes());
        if (ir == null) {
            throw new InterpretException("No interpretation result!");
        }
        List<Interpretation> interpretations = ir.getInterpretations();
        if (interpretations.isEmpty())
            throw new InterpretException("No interpretations found");

        for (Interpretation interp : interpretations)
        {
            List<InterpretedRecognition.Phrase> phrases = new ArrayList<InterpretedRecognition.Phrase>();

            List<Interpretation.Slot> slots = interp.getSlots();
            for (Interpretation.Slot slot : slots)
            {
                // We don't care about this special slot. TODO: is this the right place to filter it out?
                if ("fill".equals(slot.name))
                    continue;

                List<InterpretedRecognition.Word> words = new ArrayList<InterpretedRecognition.Word>();

                words.add(new InterpretedRecognition.Word(slot.text, null, null, InterpretedRecognition.Word.TIMING_NOT_SUPPORTED, InterpretedRecognition.Word.TIMING_NOT_SUPPORTED));
                DictationResult dictationResult = slot.dictationResult;

                InterpretedRecognition.Phrase phrase;
                if (dictationResult != null)
                    phrase = new InterpretedRecognition.Phrase(slot.name, dictationResult, null);
                else
                    // TODO: split words from slot.text?
                    phrase = new InterpretedRecognition.Phrase(slot.name, slot.text, words, null, true);

                phrases.add(phrase);
            }

            // TODO: score?
            interpretedRecognition.addChoice(interp.toString(), interp.getRecogType(), 0, phrases, null);
        }
        return interpretedRecognition;
    }

    @Override
    public boolean processForCloud(CloudRecognitionResult recognitionResult, List<DataParam> params)
    {
        return true;
    }

    @Override
    public List<String> getUpdateRequiredList(CloudRecognitionResult recognitionResult)
    {
        Data resultData = recognitionResult.getDictionary();
        if (resultData == null || resultData.getType() != Data.TYPE_DICT)
            return new ArrayList<String>(0);

        Data.Dictionary resultDict = (Data.Dictionary)resultData;

        Data uploadData = resultDict.get("force_upload_flags");
        if (uploadData != null && uploadData.getType() == Data.TYPE_SEQ) {
            Data.Sequence uploadFlags = (Data.Sequence)uploadData;
            ArrayList<String> updateList = new ArrayList<String>(uploadFlags.size());
            for (int i = 0; i < uploadFlags.size(); i++) {
                Data uploadItem = uploadFlags.get(i);
                if (uploadItem.getType() == Data.TYPE_DICT) {
                    Data.Dictionary uploadItemDict = (Data.Dictionary)uploadItem;
                    Data.Integer force_upload = uploadItemDict.getInt("force_upload");
                    if (force_upload != null && (force_upload.value == 1)) {
                        Data.String id = uploadItemDict.getString("id");
                        if (!updateList.contains(id.value)) {
                            updateList.add(id.value);
                        }
                    }
                }
            }
            return updateList;
        }

        return new ArrayList<String>(0);
    }
}
