package com.nuance.dragon.toolkit.recognition;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.cloudservices.DataParam;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.vocon.VoconResult;

/**
 * Interprets raw VoCon results.
 * <p>
 * The purpose of the VoCon arbitrator is to determine if VoCOn result
 * is in grammar and doesn't contain open-speech. This conditions if the system
 * needs the server assistance for retrieving the rest of the results. Since the
 * server results can be NLU typed, there is no need to try to interpret or format
 * the VoconResult. The arbitration rules can vary from one application to another.
 * Therefore, this class can be overridden by customers.
 */
public class VoconArbitrationInterpreter implements RecognitionInterpreter<LocalRecognitionResult>
{
    /**
     * Indicates if a decoy grammar has been recognized.
     */
    protected boolean _isDecoyGrammar;

    /**
     * The arbitration decision.
     */
    public static enum Decision
    {
        LOCAL, OPEN_SPEECH, OUT_OF_GRAMMAR, DECOY_GRAMMAR, NOISE
    }

    @Override
    public InterpretedRecognition getInterpretedResult(LocalRecognitionResult voconResult) throws InterpretException
    {
        Checker.checkArgForCondition("voconResult", "instanceof VoconResult", voconResult instanceof VoconResult);

        // TODO: Here, an application might want to differentiate between
        // being OOG by low-confidence or by decoy grammar recognition.
        _isDecoyGrammar = false;

        return null;
    }

    @Override
    public boolean processForCloud(LocalRecognitionResult voconResult, List<DataParam> params)
    {
        Checker.checkArgForCondition("voconResult", "instanceof VoconResult", voconResult instanceof VoconResult);

        String decision = "local";
        boolean keepWaiting = false;

        // We want the cloud help if (order matters):
        // - OOG (poor confidence, optional decoy grammar recognized with the OOG context parameter maxed out).
        // - Unidentified speech/noise (speech detected but no NBest result).
        // - Open-speech ("send text to John Smith were are you?").
        VoconResult res = (VoconResult)voconResult;
        if (res.isOutOfGrammar())
        {
            decision = _isDecoyGrammar ? "decoy-grammar": "out-of-grammar";
            keepWaiting = true;
        }
        else if (res.isSpeech() && res.getChoiceCount() == 0)
        {
            decision = "noise";
            keepWaiting = true;
        }
        else
        {
            if (res.hasGenericSpeech(0))
            {
                decision = "open-speech";
                keepWaiting = true;
            }
        }

        Logger.debug(this, "\"vocon_decision\" = \"" + decision + "\"");

        // TODO: should we send params to the cloud?

        return keepWaiting;
    }

    @Override
    public List<String> getUpdateRequiredList(LocalRecognitionResult recognitionResult)
    {
        return new ArrayList<String>(0);
    }
}
