package com.nuance.dragon.toolkit.recognition;

import java.util.ArrayList;
import java.util.List;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import com.nuance.dragon.toolkit.cloudservices.DataParam;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.recognition.InterpretedRecognition.Word;
import com.nuance.dragon.toolkit.vocon.VoconResult;


/**
 * Interprets raw VoCon results.
 * <p>
 * The purpose of the VoCon interpreter is to construct an 'interpreted' VoCon
 * result object from the raw VoCon result structure. The interpreted result
 * will be similar to the results returned by other speech recognition engines
 * (e.g. VoCon or NMAS), to enable easier merging of results from multiple sources.
 */
public class VoconInterpreter implements RecognitionInterpreter<LocalRecognitionResult>
{
    @Override
    public InterpretedRecognition getInterpretedResult(LocalRecognitionResult voconResult) throws InterpretException
    {
        Checker.checkArgForCondition("voconResult", "instanceof VoconResult", voconResult instanceof VoconResult);

        VoconResult res = (VoconResult)voconResult;
        InterpretedRecognition recognition = new InterpretedRecognition();

        JSONArray jsonNbest = res.getChoiceList();
        boolean isOutOfGrammar = res.isOutOfGrammar();
        for (int idx1 = 0; idx1 < jsonNbest.length(); idx1++)
        {
            try
            {
                org.json.JSONObject choice = jsonNbest.getJSONObject(idx1);
                int score = choice.getInt("_score");
                String fullRule= choice.getString("_startRule");
                String rule = fullRule.substring(fullRule.indexOf('#') + 1);
                String transcription = "";

                List<InterpretedRecognition.Phrase> phrases = new ArrayList<InterpretedRecognition.Phrase>();
                List<InterpretedRecognition.Word> words = new ArrayList<InterpretedRecognition.Word>();

                JSONArray items = choice.getJSONArray("_items");
                for (int idx2 = 0; idx2 < items.length(); idx2++)
                {
                    String key = null;
                    String subTranscription = "";
                    boolean isRecognized = true;

                    JSONObject currentItem = items.getJSONObject(idx2);
                    if(currentItem.getString("_type").equals("terminal"))
                    {
                        // Node is not a tag
                        String ortography = currentItem.getString("_orthography");
                        isRecognized = (!ortography.equals("<...>") || currentItem.getInt("_conf") == 0) && !isOutOfGrammar;

                        Word iw = new Word(ortography, ortography, null,
                                           currentItem.getInt("_beginTimeMs"),
                                           currentItem.getInt("_endTimeMs"));
                        transcription += ortography + " ";
                        subTranscription += ortography + " ";
                        words.add(iw);
                    }
                    else
                    {
                        // Node is a tag
                        String fullKey = currentItem.getString("_name");
                        key = fullKey.substring(fullKey.indexOf('#') + 1);
                        JSONArray subItems = currentItem.getJSONArray("_items");
                        for (int idx3 = 0; idx3 < subItems.length(); idx3++)
                        {
                            JSONObject currentSubItem = subItems.getJSONObject(idx3);
                            String ortography = currentSubItem.getString("_orthography");
                            isRecognized = (!ortography.equals("<...>") || currentSubItem.getInt("_conf") == 0) && !isOutOfGrammar;

                            Word iw = new Word(ortography, ortography, null,
                                               currentSubItem.getInt("_beginTimeMs"),
                                               currentSubItem.getInt("_endTimeMs"));
                            transcription += ortography + " ";
                            subTranscription += ortography + " ";
                            words.add(iw);
                        }
                    }
                    phrases.add(new InterpretedRecognition.Phrase(key, subTranscription.trim(), words, null, isRecognized));
                    words.clear();
                }
                recognition.addChoice(transcription.trim(), rule, score, phrases, null);
            }
            catch (JSONException e)
            {
                final String msg = "Failing to parse VoCon result";
                Logger.error(this, msg, e);
                throw new InterpretException(msg);
            }
        }
        return recognition;
    }

    @Override
    public boolean processForCloud(LocalRecognitionResult voconResult, List<DataParam> params)
    {
        Checker.checkArgForCondition("voconResult", "instanceof VoconResult", voconResult instanceof VoconResult);

        // TODO: should we send params to the cloud?

        return ((VoconResult)voconResult).hasGenericSpeech(0) ||
                ((VoconResult)voconResult).isOutOfGrammar();
    }

    @Override
    public List<String> getUpdateRequiredList(LocalRecognitionResult recognitionResult)
    {
        return new ArrayList<String>(0);
    }
}
