package com.nuance.dragon.toolkit.recognition.dictation;

/**
 * The aggregate of each <code>AlterChoice</code> that is returned when
 * {@link DictationEditBuffer#getChoices(long, long)} is called.
 * It can also be used to replace the selected text in the edit buffer with one
 * of the choices returned.
 * @see DictationEditBuffer#getChoices(long, long)
 */
public interface AlterChoices {

    /**
     * Gets the total number of alternatives available.
     *
     * An <code>AlterChoices</code> object holds many alternatives to
     * choose from.
     * This method returns the total number of alternatives that are
     * available.
     *
     * @return the number of alternatives
     */
    public int size();

    /**
     * Gets the alternative at a certain position.
     *
     * An <code>AlterChoices</code> object holds many alternatives to choose
     * from. This method returns the alternative at a specified index.
     *
     * @param index Index of the requested AlterChoice
     * @return The alternative at position, or null if position is invalid.
     */
    public AlterChoice getChoiceAt(int index);

    /**
     * Selects the alternative to be used
     *
     * This method updates the <code>DictationEditBuffer</code> to which this
     * <code>AlterChoices</code> belongs with the choice at index. It replaces
     * the alternative currently used in <code>DictationEditBuffer</code> with
     * the one at index.
     * If provided index is invalid, an IndexOutOfBound exception is thrown.
     *
     * For this method to successfully complete, the <code>DictationEditBuffer</code>
     * this <code>AlterChoices</code> object belongs should not have been modified
     * before calling choose. If a modification to <code>DicationEditBuffer</code>
     * has made this <code>AlterChoices</code> object out-of-date with respect to
     * <code>DicationEditBuffer</code> an IllegalStateException will be thrown.
     *
     * @param index Index of the chosen AlterChoice
     */
    public void choose(int index) throws IndexOutOfBoundsException, IllegalStateException;

    /**
     * Returns all alternative choices as an array of strings.
     *
     * This method provides a convenient way of obtaining a list of alternatives
     * as strings; where, the first string in the array represents the first
     * alternative, the second represents the second alternative, etc...
     * Typically, this method facilitates obtaining and displaying the alternatives
     * so that an end user can make a selection of the alternative that should be
     * used.
     *
     * @return alternative choices in an array of strings
     */
    public String[] toStringArray();


}
