package com.nuance.dragon.toolkit.recognition.dictation;

/**
 * The result of a dictation, as received from the recognition server, contains
 * the most likely result along with alternative but less likely results. The
 * <code>DicationEditBuffer</code> facilitates retrieving the main result and the
 * alternative results of a dictation. In addition, it supports several manual
 * edits on the result.
 * Once created, the <code>DictationEditBuffer</code> represents the most likely
 * result by default.
 * @see DictationResultManager#createDictationEditBuffer(byte[], long)
 * @see DictationResultManager#createDictationEditBuffer(DictationResult, long)
 */
public interface DictationEditBuffer
{
    /**
     * Retrieve length of edit buffer
     *
     * This method returns the current total number of characters this
     * <code>DictationEditBuffer</code> contains.
     *
     * @return length of edit buffer
     */
    public long length();

    /**
     * Retrieve a list of all possible alternatives for text range of [start, end).
     *
     * By default the <code>DicationEditBuffer</code> represents the most likely
     * recognition result. To obtain alternative recognition results for part or
     * all of the current result, this method can be used.
     * This method will return all the possible alternatives within the specified range.
     * The returned results will be ordered from most likely to least likely, and the
     * currently used alternative will be included in the list.
     *
     * @param start The start position of text range
     * @param end The position one past the last character of the text range
     * @return <code>AlterChoices</code> object that aggregates the alternatives
     */
    public AlterChoices getChoices(long start, long end) throws IndexOutOfBoundsException, IllegalArgumentException;

    /**
     * Insert a char
     *
     * This method modifies the current <code>DictationEditBuffer</code> by adding
     * a character at the indicated position
     *
     * @param pos position to insert char
     * @param c char to be inserted
     */
    public void insert(long pos, char c) throws IndexOutOfBoundsException;

    /**
     * Insert a string
     *
     * This method modifies the current <code>DictationEditBuffer</code> by
     * inserting a string at the indicated position
     *
     * @param pos position to insert string
     * @param text string to be inserted
     * @throws IndexOutOfBoundsException
     */
    public void insert(long pos, String text) throws IndexOutOfBoundsException;

    /**
     * Delete a char
     *
     * This method removes a char at the indicated position.
     *
     * @param pos position of char to be deleted
     * @throws IndexOutOfBoundsException
     */
    public void delete(long pos) throws IndexOutOfBoundsException;

    /**
     * Delete text in range [start, end)
     *
     * This method removes the text contained in the indicated range.
     * The start position is included in the range, the end position is excluded
     * from the range. I.e., the character at end will not be deleted.
     *
     * @param start The start position of text to be deleted
     * @param end The position one past the last character to be deleted
     * @throws IndexOutOfBoundsException
     */
    public void delete(long start, long end) throws IndexOutOfBoundsException, IllegalArgumentException;

    /**
     * Replace editor text in range [start, end) with a string
     *
     * This method replaces the text contained in the indicated range with the
     * specified string.
     * The start position is included in the range, the end position is excluded
     * from the range. I.e., the character at end will not be deleted.
     *
     * @param start The start position of text to be replaced
     * @param end The position one past the last character to be replaced
     * @param text new text string to replace editor text.
     * @throws IndexOutOfBoundsException
     */
    public void replace(long start, long end, String text) throws IndexOutOfBoundsException;

    /**
     * Update whole editor buffer with new text
     *
     * This method will update the entire <code>DictationEditBuffer</code> with
     * the specified string. This method will compare the original text to the
     * updated text, and detect and apply the changes.
     *
     * @param text new text string to update editor buffer
     * @throws throws IllegalArgumentException
     */
    public void update(String text) throws IllegalArgumentException;

    /**
     * Returns the string representation of the current editor text
     *
     * This method returns the current <code>DictationEditBuffer</code> as a string.
     *
     * @return string representation of the current editor text
     */
    @Override
    public String toString();

    /**
     * Retrieve the list of <code>EditorItem</code> objects
     *
     * The <code>DictationEditBuffer</code> aggregates a list of <code>EditorItems</code>.
     * This method allows the user to obtain the list of <code>EditorItems</code> that
     * represent this <code>DictationEditBuffer</code>.
     *
     * @return Array of <code>EditorItem</code>
     */
    public EditorItem[] getItems();

    /**
     * Merge with a new dictation result buffer from dictation recognition server
     * and replace text within the range of [start, end)
     * @param buffer the binary data received from dictation recognition server
     * to be merged
     * @param start The start position of text to be replaced
     * @param end The position one past the last character to be replaced
     */
    public void merge(byte [] buffer, long start, long end) throws IllegalArgumentException;

    /**
     * Append a new dictation result buffer at the end of current buffer
     * @param buffer the binary data received from dictation recognition server
     * to be merged
     * @throws IllegalArgumentException
     */
    public void append(byte [] buffer) throws IllegalArgumentException;
}
