/**
 *  Copyright 2006-2012, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.dragon.toolkit.recognition.dictation;

import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.recognition.dictation.internal.DictationEditBufferImpl;
import com.nuance.dragon.toolkit.recognition.dictation.internal.DictationResultImpl;
import com.nuance.dragon.toolkit.recognition.dictation.internal.Util;
import com.nuance.dragon.toolkit.recognition.dictation.parser.DnsBinV1ResultParser;
import com.nuance.dragon.toolkit.recognition.dictation.parser.NLSMLSaxParser;
import com.nuance.dragon.toolkit.recognition.dictation.parser.ResultParser;
import com.nuance.dragon.toolkit.recognition.dictation.parser.XmlResultParser;
import com.nuance.dragon.toolkit.recognition.interpretation.InterpretationResult;

/**
 * Hides the results
 * type from the users and creates the unified representation of a dictation
 * result. The unified representation is presented to the mobile device through the interface
 * {@link DictationResult}.
 * <br>
 * <br>
 * Supports Dragon Naturally Speaking  (V.1 and V.2) binary results
 * and XML results.
 *
 * @see DictationResult
 *
 * @author Nuance Communication Inc.
 * @since 1.0
 */
public class DictationResultManager
{
    /**
     * The factory methods that identify the results format and return
     * the unified <code>DictationResult</code> data structure.
     *
     * @param buffer the binary data received from dictation recognition server
     * @return the unified data structure <code>DictationResult</code>
     * @throws IllegalArgumentException thrown if the parsers are unable to parse the data
     */
    public static DictationResult createDictationResult(byte[] buffer) throws IllegalArgumentException
    {
        return createDictationResultImpl(buffer);
    }

    private static DictationResult createDictationResultImpl(byte[] buffer) throws IllegalArgumentException
    {
        Logger.debug(DictationResultManager.class, "DictationResultImpl(buffer [size: " + buffer.length + "] )");

        if(buffer == null || buffer.length < 4)
        {
            String msg = "Cannot parse dictation results: The buffer length is too small to be containing any results.";

            Logger.error(DictationResultManager.class, msg);

            throw new IllegalArgumentException(msg);
        }

        DictationResult result = null;
        try {
            ResultParser parser = getParserBasedOnHeader(buffer);
            result = parser.parse();
        }catch (Exception e) {
            throw new IllegalArgumentException("Cannot parse dictation results: illegal format buffer.");
        }
        return result;
    }

    /**
     * Create dictation edit buffer from DictationResult object
     * @param result the DictationResult object containing the results received from the dictation recognition server
     * @param maxSize maximum buffer size.
     * @return  the unified data structure <code>DictationEditBuffer</code>
     */
    public static DictationEditBuffer createDictationEditBuffer(DictationResult result, long maxSize)
    {
        return new DictationEditBufferImpl(maxSize, (DictationResultImpl)result);
    }

    /**
     * Create dictation edit buffer from binary data
     * @param buffer the binary data received from dictation recognition server
     * @param maxSize maximum buffer size.
     * @return  the unified data structure <code>DictationEditBuffer</code>
     * @throws IllegalArgumentException
     */
    public static DictationEditBuffer createDictationEditBuffer(byte[] buffer, long maxSize) throws IllegalArgumentException
    {
        DictationResultImpl result = (DictationResultImpl)createDictationResultImpl(buffer);

        return new DictationEditBufferImpl(maxSize, result);
    }

    /**
     * Create an interpretation result from a byte buffer in nlsml format
     *
     * @param buffer the binary data received from dictation recognition server
     * @return the unified data structure <code>InterpretationResult</code>
     * @throws IllegalArgumentException
     */
    public static InterpretationResult createInterpretationResult(byte[] buffer) throws IllegalArgumentException
    {
        Logger.debug(DictationResultManager.class, "createInterpretationResult(buffer [size: " + buffer.length + "] )");

        if(buffer == null || buffer.length < 4)
        {
            String msg = "Cannot parse interpretation results: The buffer length is too small to be containing any results.";

            Logger.error(DictationResultManager.class, msg);

            throw new IllegalArgumentException(msg);
        }

        InterpretationResult result = null;
        try {
            NLSMLSaxParser parser = new NLSMLSaxParser(buffer);
            parser.parse();
            result = parser.getInterpretationResult();
        } catch (Exception e) {
            Logger.error(DictationResultManager.class, "Error parsing interpretation result", e);
            throw new IllegalArgumentException("Cannot parse interpretation results: illegal format buffer.");
        }
        return result;
    }

    private static ResultParser getParserBasedOnHeader(byte[] buffer)
    {
        if (buffer.length > 20) {
            String firstChars = Util.constructByteEncodedString(buffer, 0, 20, "ISO-8859-1");

            if (firstChars.startsWith("<?xml")
                    || (firstChars.indexOf("<?") > -1 && firstChars
                            .indexOf("xml") > -1))
            {
                Logger.debug(DictationResultManager.class, "Detected xml results. Using xml parser.");
                return new XmlResultParser(buffer);
            }
        }

        Logger.debug(DictationResultManager.class, "Detected binary results. Using binary parser.");
        return new DnsBinV1ResultParser(buffer);
    }
}
