package com.nuance.dragon.toolkit.recognition.dictation.internal;

import java.util.ArrayList;
import java.util.List;

import com.nuance.dragon.toolkit.recognition.dictation.AlterChoices;
import com.nuance.dragon.toolkit.recognition.dictation.DictationEditBuffer;
import com.nuance.dragon.toolkit.recognition.dictation.DictationResult;
import com.nuance.dragon.toolkit.recognition.dictation.DictationResultManager;
import com.nuance.dragon.toolkit.recognition.dictation.EditorItem;
import com.nuance.dragon.toolkit.recognition.dictation.EditorItem.Type;
import com.nuance.dragon.toolkit.recognition.dictation.Token;

public class DictationEditBufferImpl implements DictationEditBuffer
{
    public class TimeRange {
        public long start;
        public long end;

        public TimeRange(long startTime, long endTime) {
            start = startTime;
            end = endTime;
        }
    }

    public class RelativePosition {
        public int   index;
        public int   offset;

        public RelativePosition() {
            index = 0;
            offset = 0;
        }
    }

    public class RelativeRange {
        public RelativePosition   start;
        public RelativePosition   end;

        public RelativeRange() {
            start = new RelativePosition();
            end = new RelativePosition();
        }
    }


    private ArrayList<EditorItem> _items;
    private DictationResult _dictResult;
    private final long _maxSize;

    private long _stamp = 0;

    public DictationEditBufferImpl(long maxSize, final DictationResultImpl dictResult)
    {
        _maxSize = maxSize;
        _items = new ArrayList<EditorItem>();
        _dictResult = dictResult;

        if (_dictResult.size() > 0) {
            SentenceImpl sentence = dictResult.getSentenceAt(0);

            appendTokens(sentence.getTokens(), false);
        }
    }

    @Override
    public long length() {
        return this.toString().length();
    }

    @Override
    public AlterChoices getChoices(long start, long end) throws IndexOutOfBoundsException, IllegalArgumentException {
        if (start >= end)
            throw new IllegalArgumentException("The indicated range is invalid. The range is treated as [start, end), therefore start cannot be greater or equal to end.");

        if (start < 0 || start > length() || end < 0 || end > length())
            throw new IndexOutOfBoundsException("Indicated position is out of bounds.");

        return buildAlterChoices(start, end);
    }

    @Override
    public void insert(long pos, char c) throws IndexOutOfBoundsException {
        if (pos < 0 || pos > length())
            throw new IndexOutOfBoundsException("Indicated position is out of bounds.");

        insert(pos, Character.toString(c));
    }

    @Override
    public void insert(long pos, String text) throws IndexOutOfBoundsException
    {
        if (text == null || text.length() == 0) {
            return;
        }

        updateStamp();

        RelativePosition position = null;
        position = getRelativePostion(pos);

        String newText = null;
        int index = position.index;

        if (position.offset == 0) {
            newText = text;
        }
        else {
            EditorItem item = _items.get(position.index);

            String oldText = item.toString();
            newText = oldText.substring(0, position.offset) + text + oldText.substring(position.offset);
        }

        int spacePos= newText.indexOf(" ");
        int lastSpace = 0;
        while (spacePos != -1) {

            if (spacePos == lastSpace) {
                _items.add(index ++, new Word(" ", Type.WHITE_SPACE));
            }
            else if (spacePos > lastSpace) {
                _items.add(index ++, new Word(newText.substring(lastSpace, spacePos), Type.WORD));
                _items.add(index ++, new Word(" ", Type.WHITE_SPACE));
            }

            lastSpace = spacePos + 1;
            spacePos= newText.indexOf(" ", lastSpace);
        }

        if (lastSpace < newText.length()) {
            _items.add(index ++, new Word(newText.substring(lastSpace), Type.WORD));
        }

    }

    @Override
    public void delete(long pos) throws IndexOutOfBoundsException
    {
        delete(pos, pos+1);
    }

    @Override
    public void delete(long start, long end) throws IndexOutOfBoundsException, IllegalArgumentException
    {
        if (start >= end)
             throw new IllegalArgumentException("The indicated range is invalid. The range is treated as [start, end), therefore start cannot be greater or equal to end.");

        updateStamp();
        RelativeRange range = null;
        range = getRelativeRange(start, end);

        if (range.start.index == _items.size()) {
            throw new IndexOutOfBoundsException("Index out of bounds!");
        }

        if (range.end.index == range.start.index) {
            EditorItem item = _items.get(range.end.index);
            String text = item.toString();
            _items.set(range.end.index, new Word(text.substring(0, range.start.offset) + text.substring(range.end.offset), Type.WORD));
        }
        else {
            int startIndex = range.start.index;
            int endIndex = range.end.index;

            EditorItem item = _items.get(range.start.index);
            if (range.start.offset != 0){
                String text = item.toString();
                _items.set(range.start.index, new Word(text.substring(0, range.start.offset), Type.WORD));
                startIndex ++;
            }

            if (range.end.index != _items.size()) {
                item = _items.get(range.end.index);
                if (range.end.offset != 0) {
                    String text = item.toString();

                    _items.set(range.end.index, new Word(text.substring(range.end.offset), Type.WORD));
                }
            }

            if (startIndex < endIndex) {
                _items.subList(startIndex, endIndex).clear();
            }
        }
    }

    @Override
    public void replace(long start, long end, String text) throws IndexOutOfBoundsException {
        long lengthL = length();
        if (start < 0 || start > lengthL || end < 0 || end > lengthL)
            throw new IndexOutOfBoundsException("Indicated position is out of bounds.");

        updateStamp();
        delete(start, end);
        insert(start, text);
    }

    @Override
    public void update(String text) throws IllegalArgumentException
    {
        if (text == null) {
            throw new IllegalArgumentException("text cannot be null.");
        }

        updateStamp();
        String bufferString = this.toString();
        CharSequence s1 = bufferString.subSequence(0, bufferString.length());
        CharSequence s2 = text.subSequence(0, text.length());

        // Find the min and max sizes of the strings.
        //--------------------------------------------
        int s1_len = s1.length();
        int s2_len = s2.length();

        int min_size;
        if (s1_len > s2_len) {
            min_size = s2_len;
        }
        else {
            min_size = s1_len;
        }

        // Here we start searching for the region that changed.
        // b1, b2 are the starting points of changes in s1, s2
        // Compute b1 and b2 (which are always the same)
        //-----------------------------------------------------------
        int i;
        for (i = 0; i < min_size; ++ i) {
            if (s1.charAt(i) != s2.charAt(i)) {
                break;
            }
        }

        int b1 = i;
        int b2 = i;

        // e1, e2 are the ending points of the changes in s1, s2
        // To compute e1 and e2, we want to position the
        // pointers at the end of the strings.
        //--------------------------------------------------------
        for ( i = 0; i < min_size; ++ i) {
            if (s1.charAt(s1_len - i - 1) != s2.charAt(s2_len - i - 1)) {
                break;
            }
        }

        // Next compute e1 and e2 based on i
        //-----------------------------------
        int e1 = s1_len - i - 1;
        int e2 = s2_len - i - 1;

        // Now we can compute the things that changed:
        //---------------------------------------------
        int del_count = e1 - b1 + 1;
        if (del_count < 0) {
            del_count = 0;
        }

        int insert_count = e2 - b2 + 1;
        if (insert_count < 0) {
            insert_count = 0;
        }

        // A weakness of the algorithm ocurrs on the strings
        // like "This cat cat" and "This cat", where we get
        // del_count = insert_count = 0.
        // In these cases, we make sure that:
        // s2_len = s1_len - del_count + insert_count
        // If not, we use the difference is strlens to
        // append from the smaller string to the larger string.
        //----------------------------------------------------------
        if (s2_len != (s1_len - del_count + insert_count)) {
            if (s1_len > s2_len) {
                del_count = s1_len - s2_len;
            }
            else {
                insert_count = s2_len - s1_len;
            }
        }

        //Logger.debug(this, "**************" + toString().substring(b1, b1 + del_count) + " replaced to " + text.substring(b2, b2 + insert_count));
        if (del_count == 0 && insert_count == 0) {
            return;
        }

        if (insert_count == 0) {
            delete(b1, b1 + del_count);
        }
        else if (del_count == 0) {
            insert(b1, text.substring(b2, b2 + insert_count));
        }
        else {
            replace(b1, b1 + del_count, text.substring(b2, b2 + insert_count));
        }
    }

    @Override
    public String toString()
    {
        if (_items.size() == 0)
            return "";

        String text = new String();
        for (int index = 0; index < _items.size(); ++ index) {
            text += _items.get(index).toString();
        }

        return text;
    }

    @Override
    public EditorItem[] getItems() {
        return (EditorItem[]) _items.toArray(new EditorItem[_items.size()]);
    }

    public ArrayList<EditorItem> getItems(long start, long end)
    {
        if (_items.size() == 0)
            return null;

        ArrayList<EditorItem> items = new ArrayList<EditorItem>();
        long pos = 0;
        for (int index = 0; index < _items.size(); ++ index) {

            EditorItem item = (EditorItem) _items.get(index);

            if(pos > end) {
                break;
            }
            else if (pos + item.length() > start) {
                items.add(item);
            }

            pos += item.length();
        }

        return items;
    }

    @Override
    public void merge(byte[] buffer, long start, long end) throws IllegalArgumentException
    {
        if (start < 0 || end > length() || start >= end) {
            throw new IllegalArgumentException("Invalid start or end position");
        }

        updateStamp();

        TimeRange timeRange = getTimeRange(start, end);

        //removed items in (start, end)
        delete(start, end);

        //insert Tokens from new result
        DictationResult result = DictationResultManager.createDictationResult(buffer);
        if (result.size() > 0) {

            long newResultStart = timeRange.start;
            long newResultEnd = timeRange.start + result.sentenceAt(0).getEndTime();

            _dictResult.merge(result, timeRange.start, timeRange.end);

            SentenceImpl sentence = (SentenceImpl)_dictResult.sentenceAt(0);

            if (sentence.size() > 0) {

                ArrayList<Token> tokens = sentence.getTokens(newResultStart, newResultEnd);
                int index = getRelativePostion(start).index;
                if (index > 0) {

                    EditorItem last = _items.get(index -1);
                    if (last.getType() == Type.WHITE_SPACE) {
                        insertTokens(index, tokens, false);
                    }
                    else if(last.getType() == Type.TOKEN) {
                        insertTokens(index, tokens, !((Token)last).hasNoSpaceAfterDirective());
                    }
                    else {
                        insertTokens(index, tokens, true);
                    }
                }
                else {
                    insertTokens(index, tokens, false);
                }
            }

            if (timeRange.start == -1) {
                timeRange.start = _dictResult.sentenceAt(0).getStartTime();
            }

            if (timeRange.end == -1) {
                timeRange.end = _dictResult.sentenceAt(0).getEndTime();
            }
        }
        else {
            timeRange.start = 0;
            timeRange.end = 0;
        }

    }

    @Override
    public void append(byte[] buffer) throws IllegalArgumentException
    {
        updateStamp();

        DictationResult result = DictationResultManager.createDictationResult(buffer);
        if (result.size() > 0) {
            long newResultStart = 0;
            if (_dictResult.size() > 0) {
                newResultStart = _dictResult.sentenceAt(0).getEndTime();
            }

            _dictResult.append(result);
            SentenceImpl sentence = (SentenceImpl)_dictResult.sentenceAt(0);
            if (sentence.size() > 0) {

                ArrayList<Token> tokens = sentence.getTokens(newResultStart, sentence.getEndTime());
                if (_items.size() > 0) {
                    EditorItem last = _items.get(_items.size() -1);
                    if (last.getType() == Type.WHITE_SPACE) {
                        appendTokens(tokens, false);
                    }
                    else if(last.getType() == Type.TOKEN) {
                        appendTokens(tokens, !((Token)last).hasNoSpaceAfterDirective());
                    }
                    else {
                        appendTokens(tokens, true);
                    }
                }
                else {
                    //no item in current editor buffer
                    appendTokens(tokens, false);
                }
            }
        }

    }

    private RelativePosition getRelativePostion(long pos) throws IndexOutOfBoundsException
    {
        if (pos < 0) {
            throw new IndexOutOfBoundsException("Position is less than 0!");
        }

        RelativePosition position = new RelativePosition();

        long trace = 0;
        for (int index = 0; index < _items.size(); ++index)
        {
            EditorItem item = _items.get(index);
            if (pos >= trace && pos < trace + item.length()) {
                position.index = index;
                position.offset = (int)(pos - trace);
                return position;
            }

            trace += item.length();
        }

        if (pos == trace) {
            position.index = _items.size();
            position.offset = 0;
            return position;
        }
        else {
            throw new IndexOutOfBoundsException("Position is greater than buffer length!");
        }
    }

    private RelativeRange getRelativeRange(long start, long end)  throws IndexOutOfBoundsException
    {
        if (start < 0 || end < start) {
            throw new IndexOutOfBoundsException("start is less than 0 or end is less than start!");
        }

        RelativeRange range = new RelativeRange();

        long trace = 0;
        for (int index = 0; index < _items.size(); ++index)
        {
            EditorItem item = _items.get(index);
            if (start >= trace && start < trace + item.length()) {
                range.start.index = index;
                range.start.offset = (int)(start - trace);
            }

            if (end >= trace && end < trace + item.length()) {
                range.end.index = index;
                range.end.offset = (int)(end - trace);
                return range;
            }

            trace += item.length();
        }

        if (start == trace) {
            range.start.index = _items.size();
            range.start.offset = 0;
        }

        if (end == trace) {
            range.end.index = _items.size();
            range.end.offset = 0;
        }

        if (end > trace) {
            throw new IndexOutOfBoundsException("end is greater than buffer length!");
        }

        return range;
    }

    private TimeRange getTimeRange(long start, long end)
    {
        if (start < 0 || end < start) {
            throw new IndexOutOfBoundsException("start is less than 0 or end is less than start!");
        }

        TimeRange timeRange = new TimeRange(-1, -1);

        Token first = null;
        Token last = null;
        Token before = null;
        Token after = null;
        long trace = 0;
        for (int index = 0; index < _items.size(); ++index)
        {
            EditorItem item = _items.get(index);

            if (item.getType() == Type.TOKEN) {
                if (trace < start) {
                    before = (Token)item;
                }

                if (trace >= start && first == null) {
                    first = (Token)item;
                }

                if (trace < end ) {
                    last = (Token)item;
                }

                if (trace >= end) {
                    after = (Token)item;
                    break;
                }
            }

            trace += item.length();
        }

        if (end > trace) {
            throw new IndexOutOfBoundsException("end is greater than buffer length!");
        }

        if (first != null) {
            timeRange.start = first.getStartTime();
        }
        else if (before != null){
            timeRange.start = before.getEndTime();
        }

        if (last != null) {
            timeRange.end = last.getEndTime();
        }
        else if (after != null){
            timeRange.end = after.getStartTime();
        }

        return timeRange;
    }

    private void appendTokens(List<? extends Token> tokens, boolean add_leading_space)
    {
        int index = _items.size();
        insertTokens(index, tokens, add_leading_space);
    }

    private void insertTokens(int index, List<? extends Token> tokens, boolean add_leading_space)
    {
        long size = 0;
        boolean no_space_before = false;
        boolean no_space_after = !add_leading_space;

        if (tokens != null) {
            for (int i = 0; i < tokens.size(); ++ i) {

                TokenImpl token = (TokenImpl)tokens.get(i);
                no_space_before = token.hasNoSpaceBeforeDirective();

                if (size + token.length() > _maxSize)
                    break;

                if (!no_space_before && !no_space_after) {
                    _items.add(index ++, new Word(" ", Type.WHITE_SPACE));
                }
                _items.add(index ++, token);

                no_space_after = token.hasNoSpaceAfterDirective();
            }
        }

    }

//    private int getIndexOfNextItemOfType(int currentIndex, Type type)
//    {
//        for (int i = currentIndex + 1; i < _items.size(); i++)
//        {
//            EditorItem current = _items.get(i);
//            if (current.getType() == type) {
//                return i;
//            }
//        }
//        return -1;
//    }
//
//    private int getIndexOfPrevioustItemOfType(int currentIndex, Type type)
//    {
//        for (int i = currentIndex - 1; i >=0; i--)
//        {
//            EditorItem current = _items.get(i);
//            if (current.getType() == type) {
//                return i;
//            }
//        }
//        return -1;
//    }

    private void insertItemsAtIndex(int index, ArrayList<EditorItem> newItems)
    {
        _items.addAll(index, newItems);
    }

    public void replaceItems(long start, long end, ArrayList<EditorItem> newItems)
    {
        RelativePosition firstPos = getRelativePostion(start);
        delete(start, end);
        if(firstPos.offset == 0)
            insertItemsAtIndex(firstPos.index, newItems);
        else
            insertItemsAtIndex(firstPos.index + 1, newItems);
    }

    private void updateStamp()
    {
        _stamp++;
    }

    public boolean compareStamp(long stamp)
    {
        if (stamp == _stamp)
            return true;

        return false;
    }

    private AlterChoices buildAlterChoices(long start, long end)
    {
        RelativePosition firstPos = getRelativePostion(start);
        RelativePosition lastPos= getRelativePostion(end);

        AlterChoicesImpl ret = new AlterChoicesImpl(this, _dictResult.size(), start, end, _stamp);

        int firstIdx;
        int lastIdx;

        if (firstPos.offset != 0) {
            firstIdx = firstPos.index + 1;
            String subText = _items.get(firstPos.index).toString().substring(firstPos.offset);
            ret.addEntry(new Word(subText, Type.WORD));
        }
        else {
            firstIdx = firstPos.index;
        }

        // Inserting the partial text of the last item selected is handled at the end
        lastIdx = lastPos.index -1;

        for (int current = firstIdx; current <= lastIdx; ++ current)
        {
            //Get the type of the current EditItem
            Type type = _items.get(current).getType();

            switch(type)
            {
            // If the current EditItem is a WORD, add it as is to the AlterChoices,
            // since it has no alternatives
            case WORD:
                ret.addEntry(_items.get(current));
                break;
                // If the current EditItem is WHITE_SPACE, add it as is to the AlterChoices,
                // since it has no alternatives.
                // Note: White spaces that are in between TOKEN items will not and should not
                // be handled in here. They will be automatically assimilated into the Alternatives
                // and Alternative objects.
            case WHITE_SPACE:
                ret.addEntry(_items.get(current));
                break;
                // If the current EditItem is a TOKEN, move forward to cover all subsequent TOKENs,
                // and get the Alternatives for all the consecutive tokens and add them to AlterChoices.
            case TOKEN:
            {
//                // Find next Item of type WORD
//                int temp = getIndexOfNextItemOfType(current, Type.WORD);
//                int lastConsecutiveToken;
//
//                // If next Item of type WORD exceeds lastIdx, or if there is no
//                // next Item of type WORD, set temp to lastIdx
//                if (temp == -1 || temp > lastIdx) {
//                    temp = lastIdx;
//                }
//
//                // Get the previous TOKEN from temp, or temp if is a TOKEN.
//                // using (temp + 1) will guarantee we receive temp itself if
//                // it is a TOKEN
//                lastConsecutiveToken = getIndexOfPrevioustItemOfType(temp+1, Type.TOKEN);
//
//                // Get the alternatives spanning start to end
//                Token startToken = (Token)_items.get(current);
//                Token endToken   = (Token)_items.get(lastConsecutiveToken);
//
//                ArrayList< ArrayList<Token> > altsToAdd = new ArrayList< ArrayList<Token> >(_dictResult.size());
//                for (int sentIdx = 0; sentIdx < _dictResult.size(); sentIdx++)
//                {
//                ArrayList<Token> tokens = ((SentenceImpl)_dictResult.sentenceAt(sentIdx)).getTokens(startToken.getStartTime(), endToken.getEndTime());
//                altsToAdd.add(sentIdx, tokens);
//                }
//                ret.addEntry(altsToAdd);
//
//                // Move start accordingly
//                current = lastConsecutiveToken + 1;

                    Token token = (Token)_items.get(current);
                    for (int sentIdx = 0; sentIdx < _dictResult.size(); sentIdx++)
                    {
                        ArrayList<Token> tokens = ((SentenceImpl)_dictResult.sentenceAt(sentIdx)).getTokens(token.getStartTime(), token.getEndTime());
                        if (!tokens.isEmpty()) {
                            ret.addTokens(sentIdx, tokens);
                        }
                    }
                }
            }
        }

        if (lastPos.offset != 0) {
            String subText = _items.get(lastPos.index).toString().substring(0, lastPos.offset);
            ret.addEntry(new Word(subText, Type.WORD));
        }

        ret.removeDuplicates();

        return ret;
    }
}
