/**
 *  Copyright 2006-2011, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.dragon.toolkit.recognition.dictation.internal;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.json.JSONArray;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.recognition.dictation.Alternative;
import com.nuance.dragon.toolkit.recognition.dictation.Alternatives;
import com.nuance.dragon.toolkit.recognition.dictation.DictationResult;
import com.nuance.dragon.toolkit.recognition.dictation.Sentence;
import com.nuance.dragon.toolkit.recognition.dictation.Token;

public class DictationResultImpl extends DictationResult
{
    private ArrayList<SentenceImpl> sentences;
    private Map<String, String> extraInformation;

    private static final String KEY_SENTENCES = "sentences";
    private static final String KEY_EXTRA = "extra";

    public DictationResultImpl(ArrayList<SentenceImpl> nbestList, Map<String, String> extraInformation)
    {
        Logger.debug(this, "Received " + nbestList.size() +
            " sentences constituting the dictation result.");

        this.sentences = nbestList;
        this.extraInformation = extraInformation;

        for(int i = 0; i < sentences.size(); ++i)
        {
            SentenceImpl sentence = nbestList.get(i);
            sentence.replaceAllNewLineBySystemNewLine();
        }
    }

    @Override
    public Sentence sentenceAt(int index)
    {
        return getSentenceAt(index);
    }

    SentenceImpl getSentenceAt(int index)
    {
        if (index < 0 || index >= sentences.size())
            return null;

        return sentences.get(index);
    }

    @Override
    public int size()
    {
        return sentences.size();
    }

    @Override
    public Map<String, String> getExtraInformation()
    {
        return this.extraInformation;
    }

    @Override
    public String toString()
    {
        if(sentences.size() > 0)
            return sentences.get(0).toString();
        else
            return "";
    }

    private void updateExtraInformation(Map<String, String> newExtraInfo)
    {
        //update Extra information
        //utterance-length
        String utteranceLength = this.extraInformation.get("utterance-length");
        String utteranceLengthToUpdate = newExtraInfo.get("utterance-length");
        if (utteranceLength != null && utteranceLengthToUpdate != null) {
            this.extraInformation.put("utterance-length", (Integer.valueOf(Integer.parseInt(utteranceLength) + Integer.parseInt(utteranceLengthToUpdate))).toString());
        }

        //processed-audio-frame-count
        String processedAudioFrameCount = this.extraInformation.get("processed-audio-frame-count");
        String processedAudioFrameCountToUpdate = newExtraInfo.get("processed-audio-frame-count");
        if (processedAudioFrameCount != null && processedAudioFrameCountToUpdate != null) {
            this.extraInformation.put("processed-audio-frame-count", (Integer.valueOf(Integer.parseInt(processedAudioFrameCount) + Integer.parseInt(processedAudioFrameCountToUpdate))).toString());
        }

        //InputAudioLength
        String inputAudioLengthToUpdate = (String)newExtraInfo.get("InputAudioLength");
        if (inputAudioLengthToUpdate != null) {
            this.extraInformation.put("InputAudioLength", inputAudioLengthToUpdate);
        }
    }

     @Override
    public void merge(DictationResult newDR, long timingBegin, long timingEnd)
     {
        if ( timingBegin < 0 )
            throw new IllegalArgumentException("Timings cannot be negative");
        if ( timingBegin > timingEnd )
            throw new IllegalArgumentException("The timings are corrupted, the timingBegin is greater than the timingEnd");

        long longstTime = 0;
        for(int i=0; i < size(); ++i) {
            SentenceImpl sentence = getSentenceAt(i);
            long time = sentence.getEndTime();
            if (time > longstTime) {
                longstTime = time;
            }
        }

        if( timingEnd > longstTime)
            throw new IllegalArgumentException("timingEnd is greater than maximum time");

        long shiftTime = 0;
        for(int i=0; i<newDR.size(); ++i) {
            SentenceImpl sentence = (SentenceImpl)newDR.sentenceAt(i);
            long time = sentence.getEndTime();
            if (time > shiftTime) {
                shiftTime = time;
            }
        }

        for(int i=0; i<this.size(); ++i) {
            SentenceImpl sentenceToInsert = (i < newDR.size() ? ((DictationResultImpl)newDR).getSentenceAt(i) : null);
            SentenceImpl sentenceToModify = getSentenceAt(i);
            sentenceToModify.insertSentenceWithinTimings(sentenceToInsert, timingBegin, timingEnd, shiftTime);
        }

        for(int i=this.size(); i<newDR.size(); ++i) {
            SentenceImpl sentence = (SentenceImpl)newDR.sentenceAt(i);

            //copy the sentence because we should not shift timing of original sentence
            SentenceImpl sentenceToInsert = (SentenceImpl) sentence.clone();

            sentenceToInsert.shiftTimingsStartingAtToken(timingBegin, 0);
            sentences.add(sentenceToInsert);
        }

        updateExtraInformation(newDR.getExtraInformation());
    }

    @Override
    public void append(DictationResult newDR)
    {
        if (newDR == null) {
            throw new IllegalArgumentException("newDR cannot be null!");
        }

        long timingBegin = 0;
        for(int i=0; i<this.size(); ++i) {
            SentenceImpl sentence = getSentenceAt(i);
            long time = sentence.getEndTime();
            if (time > timingBegin) {
                timingBegin = time;
            }
        }

        for(int i=0; i<this.size(); ++i) {
            SentenceImpl sentenceToInsert = (i < newDR.size() ? (SentenceImpl)newDR.sentenceAt(i) : null);
            SentenceImpl sentenceToModify = getSentenceAt(i);

            sentenceToModify.insertSentenceWithinTimings(sentenceToInsert, timingBegin, timingBegin, 0);
        }

        for(int i=this.size(); i<newDR.size(); ++i) {
            SentenceImpl sentence = (SentenceImpl)newDR.sentenceAt(i);

            //copy the sentence because we should not shift timing of original sentence
            SentenceImpl sentenceToInsert = (SentenceImpl) sentence.clone();

            sentenceToInsert.shiftTimingsStartingAtToken(timingBegin, 0);
            sentences.add(sentenceToInsert);
        }

        updateExtraInformation(newDR.getExtraInformation());
    }

    @Override
    public Alternatives getAlternatives(long startTime, long endTime)
    {
        Logger.debug(this, "Getting alternatives of at times [" + startTime + ", " + endTime + "]");

        if (endTime < 0 || endTime < startTime)
            return null;

        long longstTime = 0;
        for(int i=0; i < size(); ++i) {
            SentenceImpl sentence = getSentenceAt(i);
            long time = sentence.getEndTime();
            if (time > longstTime) {
                longstTime = time;
            }
        }

        if (startTime >= longstTime) {
            return null;
        }

        ArrayList<Alternative> alters = new ArrayList<Alternative>();
        ArrayList<String> alterStrings = new ArrayList<String>();
        for(int i = 0; i < sentences.size(); ++ i)
        {
            SentenceImpl sentence = sentences.get(i);

            ArrayList<Token> alternativeTokens = sentence.getTokens(startTime, endTime);
            AlternativeImpl alternative = new AlternativeImpl(alternativeTokens);

            Logger.debug(this, "Got alternative [" + alternative + "] for sentence at index " + i);

            if(alternative.size() == 0) {
                Logger.debug(this, "Got no alternative for sentence at index " + i);
                continue;
            }

            String alterString = alternative.toString();
            if(alterStrings.contains(alterString))  {
                Logger.debug(this, "That alternative has already been " +
                        "given by another sentence.");
                continue;
            }

            alters.add(alternative);
            alterStrings.add(alterString);
        }

        if (alters.size() == 0) {
            return null;
        }

        return new AlternativesImpl(alters);
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        JSONArray sentenceArray = new JSONArray();
        for (SentenceImpl sentence : sentences)
        {
            sentenceArray.tryPut(sentence.toJSON());
        }
        o.tryPut(KEY_SENTENCES, sentenceArray);

        if (extraInformation != null)
        {
            JSONObject extraObj = new JSONObject();

            for (Map.Entry<String, String> entry : extraInformation.entrySet())
            {
                extraObj.tryPut(entry.getKey(), entry.getValue());
            }
            o.tryPut(KEY_EXTRA, extraObj);
        }
        return o;
    }

    public static DictationResultImpl createFromJSON(org.json.JSONObject json) throws JSONException
    {
        org.json.JSONArray sentenceArray = json.getJSONArray(KEY_SENTENCES);
        ArrayList<SentenceImpl> sentences = new ArrayList<SentenceImpl>(sentenceArray.length());
        for (int i = 0; i < sentenceArray.length(); i++)
        {
            sentences.add(SentenceImpl.createFromJSON(sentenceArray.getJSONObject(i)));
        }

        org.json.JSONObject extrasObject = json.optJSONObject(KEY_EXTRA);
        HashMap<String, String> extras;
        if (extrasObject != null)
        {
            extras = new HashMap<String, String>();

            org.json.JSONArray names = extrasObject.names();
            for (int i = 0; i < names.length(); i++)
            {
                String name = names.getString(i);
                extras.put(name, extrasObject.getString(name));
            }
        } else
        {
            extras = null;
        }
        return new DictationResultImpl(sentences, extras);
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((extraInformation == null) ? 0 : extraInformation.hashCode());
        result = prime * result + ((sentences == null) ? 0 : sentences.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        DictationResultImpl other = (DictationResultImpl) obj;
        if (extraInformation == null)
        {
            if (other.extraInformation != null)
                return false;
        }
        else if (!extraInformation.equals(other.extraInformation))
            return false;
        if (sentences == null)
        {
            if (other.sentences != null)
                return false;
        }
        else if (!sentences.equals(other.sentences))
            return false;
        return true;
    }
}
