/**
 *  Copyright 2006-2011, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.dragon.toolkit.recognition.dictation.internal;

import java.util.ArrayList;
import java.util.List;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.json.JSONArray;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.recognition.dictation.Sentence;
import com.nuance.dragon.toolkit.recognition.dictation.Token;

public class SentenceImpl implements Sentence, JSONCompliant
{
    private static final String KEY_TOKENS = "tokens";
    private static final String KEY_CONFIDENCE = "conf";

    private final ArrayList<TokenImpl> tokens;
    private double confidence;

    public SentenceImpl()
    {
        tokens = new ArrayList<TokenImpl>();
    }

    private SentenceImpl(ArrayList<TokenImpl> tokens, double confidence)
    {
        this.confidence = confidence;
        this.tokens = tokens;
    }

    void insertSentenceWithinTimings(SentenceImpl sentenceToInsert, long timingBegin, long timingEnd, long timeShiftThisSentence)
    {
        if(timingBegin < 0 || timingEnd < 0)
            throw new IllegalArgumentException("Timings cannot be negative");
        if(timingBegin > timingEnd)
            throw new IllegalArgumentException("The timings are corrupted, the timingBegin is greater than the timingEnd");

        SentenceImpl copiedSentenceToInsert = null;
        if (sentenceToInsert != null) {
            copiedSentenceToInsert = (SentenceImpl)sentenceToInsert.clone();
        }

        if(this.tokens.size() == 0)
        {
            if(copiedSentenceToInsert != null) {
                this.tokens.addAll(copiedSentenceToInsert.tokens);
                this.confidence = copiedSentenceToInsert.confidence;
            }
            return;
        }

        ArrayList<Token> tokensToRemove = getTokens(timingBegin, timingEnd);
        if(tokensToRemove.size() != 0)
        {
            for(int i=0;i<tokensToRemove.size();i++)
                this.tokens.remove(tokensToRemove.get(i));
        }

        //In order to keep alternatives, shift timings even sentenceToInsert is null
        this.shiftTimingsStartingAtTiming(timingBegin, timeShiftThisSentence);

        int indexWhereToInsert = findTokenStartTimeThatFollowThisTiming(timingBegin);
        if(indexWhereToInsert == -1)
            indexWhereToInsert = this.tokens.size();

        if(copiedSentenceToInsert != null && copiedSentenceToInsert.tokens.size() != 0)
        {
            copiedSentenceToInsert.shiftTimingsStartingAtToken(timingBegin, 0);
            insertSentenceAtIndex(indexWhereToInsert, copiedSentenceToInsert);
        }

    }

    void shiftTimingsStartingAtToken(long timeShift, int tokenIndex)
    {
        for(int i=tokenIndex;i<this.tokens.size();i++)
        {
            TokenImpl token = (TokenImpl)tokenAt(i);
            token.setStartTime(token.getStartTime()+timeShift);
            token.setEndTime(token.getEndTime()+timeShift);
        }
    }


    private void shiftTimingsStartingAtTiming(long timeWhereToStart, long timeShift)
    {
        int indexWhereToShift = findTokenStartTimeThatFollowThisTiming(timeWhereToStart);
        if(indexWhereToShift == -1)
            return;

        this.shiftTimingsStartingAtToken(timeShift, indexWhereToShift);
    }

    private int findTokenStartTimeThatFollowThisTiming(long timingBegin)
    {
        int index = -1;
        for(int i=0;i<this.tokens.size();i++)
        {
            TokenImpl token = (TokenImpl)tokenAt(i);
            if(token.getStartTime() >= timingBegin)
            {
                index = i;
                break;
            }
        }

        return index;
    }

    private void insertSentenceAtIndex(int index, SentenceImpl sentenceToInsert)
    {
        for(int j=sentenceToInsert.tokens.size()-1;j>=0;j--)
            this.tokens.add(index, sentenceToInsert.tokens.get(j));
    }

    public void appendTokenToSentence(TokenImpl token)
    {
        tokens.add(token);
    }

    List<TokenImpl> getTokens()
    {
        return tokens;
    }

    /**
     * Picks up the tokens that touches the provided timings:
     *
     * <pre>
     *   |  abc    |
     * de|fgh      |
     *   |       ij|klm
     * no|pqrstuvwx|yz
     * 12|         |
     *   |         |345
     * </pre>
     *
     * The content of the return list will be:
     * <pre>
     * abc
     * defgh
     * ijklm
     * nopqrstuvwxyz
     * </pre>
     *
     * 12 and 345 are only touching the boundary so they do not get picked up.
     *
     * @param startTime
     * @param endTime
     * @return list of tokens falling into range of (startTime, endTime)
     */
    public ArrayList<Token> getTokens(long startTime, long endTime)
    {
        ArrayList<Token> tokensTouchingTimeRegion = new ArrayList<Token>();

        for (int i=0;i<tokens.size();i++)
        {
            TokenImpl token = (TokenImpl)this.tokenAt(i);
            long tokenStartTime = token.getStartTime();
            long tokenEndTime = token.getEndTime();

            if (tokenStartTime < endTime && tokenEndTime > startTime)
                tokensTouchingTimeRegion.add(token);
            else if(tokenStartTime > endTime)
                break;
        }

        return tokensTouchingTimeRegion;
    }

    public void setConfidenceScore(double confidence)
    {
        this.confidence = confidence;
    }

    public double getConfidenceScore()
    {
        return this.confidence;
    }

    protected Sentence clone()
    {
        SentenceImpl copy = new SentenceImpl();
        copy.confidence = this.confidence;

        for(int i = 0; i < tokens.size(); ++ i) {
            copy.tokens.add(((TokenImpl)this.tokenAt(i)).clone());
        }
        return copy;
    }


    public String toString()
    {
        if(tokens.size() == 0)
            return "";

        StringBuffer ret = new StringBuffer();
        boolean no_space_before = false;
        boolean no_space_after = true;
        for (int i = 0; i < tokens.size(); ++ i) {

            TokenImpl token = (TokenImpl)(tokens.get(i));
            no_space_before = token.hasNoSpaceBeforeDirective();

            if ( !no_space_before && !no_space_after) {
                ret.append(' ');
            }

            ret.append(this.tokenAt(i).toString());
            no_space_after = token.hasNoSpaceAfterDirective();
        }

        return ret.toString();
    }

    public int size() {
        return this.tokens.size();
    }

    public Token tokenAt(int index) {
        TokenImpl token = (TokenImpl)this.tokens.get(index);
        return (Token)token;
    }

    void replaceAllNewLineBySystemNewLine()
    {
        String systemNewLine = System.getProperty("line.separator");
        // ---------------------------------------
        // Fix when line separator returns null
        // ---------------------------------------
        if (systemNewLine == null)
        {
            systemNewLine = "\n";
        }

        for(int i=0;i<this.tokens.size();i++)
        {
            TokenImpl token = (TokenImpl)tokenAt(i);

            StringBuffer word = new StringBuffer(token.getWord());
            for(int j=0;j<word.length();)
            {
                int index = word.toString().indexOf("\r\n", j);
                if(index == -1)
                    break;

                word.deleteCharAt(index);
                word.deleteCharAt(index);
                word.insert(index, '\n');

                j = index + 1;
            }

            for(int j=0;j<word.length();)
            {
                int index = word.toString().indexOf("\n", j);
                if(index == -1)
                    break;

                word.deleteCharAt(index);
                word.insert(index, systemNewLine);

                j = index + systemNewLine.length();
            }

            token.setWord(word.toString());
        }
    }

    @Override
    public long getStartTime()
    {
        if (tokens.isEmpty()) {
            return 0;
        }

        //start time may not be zero
        Token t = (Token)tokens.get(0);
        return t.getStartTime();
    }

    @Override
    public long getEndTime()
    {
        if (tokens.isEmpty()) {
            return 0;
        }

        int lastIndex = tokens.size() - 1;
        Token t = (Token)tokens.get(lastIndex);
        return t.getEndTime();
    }

    @Override
    public int hashCode()
    {
        return tokens.hashCode() + (int)(confidence * 1000);
    }

    @Override
    public boolean equals(Object other)
    {
        if (other == this)
            return true;
        if (other == null)
            return false;
        if (!(other instanceof SentenceImpl))
            return false;

        SentenceImpl otherSentence = (SentenceImpl)other;
        return confidence == otherSentence.confidence &&
               tokens.equals(otherSentence.tokens);
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject json = new JSONObject();
        json.tryPut(KEY_CONFIDENCE, confidence);

        JSONArray array = new JSONArray();
        for (TokenImpl t : tokens)
        {
            array.tryPut(t.toJSON());
        }
        json.tryPut(KEY_TOKENS, array);
        return json;
    }

    static SentenceImpl createFromJSON(org.json.JSONObject json) throws JSONException
    {
        double confidence = json.getDouble(KEY_CONFIDENCE);
        org.json.JSONArray array = json.getJSONArray(KEY_TOKENS);

        ArrayList<TokenImpl> tokens = new ArrayList<TokenImpl>(array.length());
        for (int i = 0; i < array.length(); i++)
        {
            tokens.add(TokenImpl.createFromJSON(array.getJSONObject(i)));
        }
        return new SentenceImpl(tokens, confidence);
    }
}
