/**
 *  Copyright 2006-2011, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.dragon.toolkit.recognition.dictation.internal;

import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.recognition.dictation.Token;

public class TokenImpl implements Token, JSONCompliant
{
    private static final String KEY_WORD = "word";
    private static final String KEY_START = "start";
    private static final String KEY_END = "end";
    private static final String KEY_CONFIDENCE = "conf";

    private String word;
    private List<String> homophones;
    private Hashtable<String, String> spokenForms;
    private long startTime;
    private long endTime;
    private double confidence;
    private boolean noSpaceBefore = false;
    private boolean noSpaceAfter = false;
    private static final String NO_SPACE_BEFORE = "\\*no-space-before";
    private static final String NO_SPACE_AFTER = "\\*no-space-after";

    public TokenImpl(String word, List<String> homophones, long startTime, long endTime, double confidence)
    {
        //this.word = new String(word);
        this.word = removeSpaceMarkers(word);

        this.homophones = new LinkedList<String>();
        if(homophones != null){
            this.homophones.addAll(homophones);
        }
        this.startTime = startTime;
        this.endTime = endTime;
        this.confidence = confidence;

    }

    public TokenImpl(String word, long startTime, long endTime, double confidence)
    {
        this(word, null, startTime, endTime, confidence);
    }

    public String getWord()
    {
        return word;
    }

    public void setWord(String word)
    {
        this.word = word;
    }

    @Override
    public long getStartTime()
    {
        return startTime;
    }

    public void setStartTime(long startTime)
    {
        this.startTime = startTime;
    }

    @Override
    public long getEndTime()
    {
        return endTime;
    }

    public void setEndTime(long endTime)
    {
        this.endTime = endTime;
    }

    @Override
    public double getConfidenceScore()
    {
        return confidence;
    }

    @Override
    public String toString()
    {
        return new String(this.word);
    }

    @Override
    public TokenImpl clone()
    {
        TokenImpl copy = new TokenImpl(this.word, this.homophones, this.startTime, this.endTime, this.confidence);
        copy.noSpaceAfter = this.noSpaceAfter;
        copy.noSpaceBefore = this.noSpaceBefore;
        return copy;
    }

    public void append(String word) {
        this.word += word;

    }

    @Override
    public boolean hasNoSpaceBeforeDirective() {
        return this.noSpaceBefore;
    }

    @Override
    public boolean hasNoSpaceAfterDirective() {
        return this.noSpaceAfter;
    }

    @Override
    public Type getType() {
        return Type.TOKEN;
    }

    @Override
    public int length() {
        return word.length();
    }

    @Override
    public int hashCode()
    {
        return word.hashCode() + (int)startTime + (int)endTime + (int)(confidence * 1000);
    }

    @Override
    public boolean equals(Object other)
    {
        if (other == this)
            return true;
        if (other == null)
            return false;
        if (!(other instanceof TokenImpl))
            return false;

        TokenImpl otherToken = (TokenImpl)other;
        return word.equals(otherToken.word) &&
               startTime == otherToken.startTime &&
               endTime == otherToken.endTime &&
               confidence == otherToken.confidence;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject json = new JSONObject();
        json.tryPut(KEY_WORD, word);
        json.tryPut(KEY_START, startTime);
        json.tryPut(KEY_END, endTime);
        json.tryPut(KEY_CONFIDENCE, confidence);
        return json;
    }

    static TokenImpl createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String word = json.getString(KEY_WORD);
        long start = json.getLong(KEY_START);
        long end = json.getLong(KEY_END);
        double conf = json.getDouble(KEY_CONFIDENCE);

        // TODO : Check who uses this JSON format, they will need to update the format
        return new TokenImpl(word, null, start, end, conf);
    }

    @Override
    public String getWrittenForm()
    {
        return this.word;
    }

    @Override
    public Hashtable<String, String> getSpokenForms()
    {
        return this.spokenForms;
    }

//    @Override
//    public List<String> getHomophones()
//    {
//        return this.homophones;
//    }

    public void addHomophone(String homophone)
    {
        this.homophones.add(removeSpaceMarkers(homophone));
    }

    public void addSpokenForm(String key, String value)
    {
        if(spokenForms == null){
            spokenForms = new Hashtable<String, String>();
        }

        spokenForms.put(key, removeSpaceMarkers(value));
    }

    private String removeSpaceMarkers(String pWord)
    {
        String newWord = pWord;
        if (newWord.indexOf(NO_SPACE_BEFORE) != -1)
        {
            noSpaceBefore = true;
            int indexOfNoSpaceBefore = newWord.indexOf(NO_SPACE_BEFORE);
            if (indexOfNoSpaceBefore + NO_SPACE_BEFORE.length() == newWord.length())
                newWord = newWord.substring(0, indexOfNoSpaceBefore);
            else
                newWord = newWord.substring(0, indexOfNoSpaceBefore) + newWord.substring(indexOfNoSpaceBefore + NO_SPACE_BEFORE.length());

        }

        if (newWord.indexOf(NO_SPACE_AFTER) != -1)
        {
            noSpaceAfter = true;
            int indexOfNoSpaceAfter = newWord.indexOf(NO_SPACE_AFTER);
            if (indexOfNoSpaceAfter + NO_SPACE_AFTER.length() == newWord.length())
                newWord = newWord.substring(0, indexOfNoSpaceAfter);
            else
                newWord = newWord.substring(0, indexOfNoSpaceAfter) + newWord.substring(indexOfNoSpaceAfter + NO_SPACE_AFTER.length());

        }
        return newWord;
    }

}
