/**
 *  Copyright 2006-2012, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.dragon.toolkit.recognition.dictation.parser;

import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.recognition.dictation.internal.DictationResultImpl;
import com.nuance.dragon.toolkit.recognition.dictation.internal.SentenceImpl;
import com.nuance.dragon.toolkit.recognition.dictation.internal.TokenImpl;
import com.nuance.dragon.toolkit.recognition.dictation.internal.Util;

// TODO need comments
public class DnsBinV1ResultParser implements ResultParser
{
    private static final int WORD_TIMINGS = 1;
    private static final int SENTENCES_CONFIDENCES = 3;
    private static final int ENCODING_EXTENSION_ID = 4;

    private static final int SECTION_HOMOPHONES = 6;
    private static final int SECTION_SEVEN = 7;
    private static final int SECTION_EIGHT = 8;
    private static final int SECTION_NINE = 9;

    private static final int EXTENSION_NOT_FOUND_OFFSET_VALUE = -1;

    private static final String DEFAULT_ENCODING = "Windows-1252";
    private static final String JAVA_NAMING_DEFAULT_ENCODING = "Cp1252";

    private final static int SHORT_SIZE = 2;
    private final static int INT_SIZE = 4;

    private byte[] buffer;
    private String encoding = JAVA_NAMING_DEFAULT_ENCODING;
    private BinaryFormatExtensionInfo binaryExtensionInfo = null;

    private class BinaryFormatExtensionInfo
    {
        public int wordTimingsExtensionOffset = EXTENSION_NOT_FOUND_OFFSET_VALUE;
        public int confidenceExtensionOffset = EXTENSION_NOT_FOUND_OFFSET_VALUE;
        public int encodingExtensionOffset = EXTENSION_NOT_FOUND_OFFSET_VALUE;
        public int homophonesOffset = EXTENSION_NOT_FOUND_OFFSET_VALUE;
        public int sectionSevenOffset = EXTENSION_NOT_FOUND_OFFSET_VALUE;
        public int sectionEightOffset = EXTENSION_NOT_FOUND_OFFSET_VALUE;
        public int sectionNineOffset = EXTENSION_NOT_FOUND_OFFSET_VALUE;
        public int pairKeyValueOffset = -1;
        public int sentencesOffset = -1;
    }

    private class ExtractTokenResult
    {
        public int newOffset = 0;
        public int newConfidenceOffset = 0;
        public TokenImpl token = null;
    }

    private class ExtractSentenceResult
    {
        public SentenceImpl sentence = new SentenceImpl();
        public int newOffset = 0;
        public int newConfidenceOffset = 0;
    }

    public String getEncoding()
    {
        if(encoding.equals(JAVA_NAMING_DEFAULT_ENCODING))
            return DEFAULT_ENCODING;

        return encoding;
    }

    public DnsBinV1ResultParser(byte[] buffer)
    {
        this.buffer = buffer;
        binaryExtensionInfo = parseExtensionOffset(buffer);
        encoding = findEncoding(buffer, binaryExtensionInfo);
    }

    private BinaryFormatExtensionInfo parseExtensionOffset(byte[] buffer)
    {
        BinaryFormatExtensionInfo info = new BinaryFormatExtensionInfo();

        int offset = 0;

        int numWords = getNumberOfWordsInWordList(offset);
        offset += SHORT_SIZE;
        offset = walkWordsList(offset, numWords);

        info.sentencesOffset = offset;

        int numSentences = getNumberOfSentences(offset);
        offset += INT_SIZE;
        offset = walkSentences(offset, numSentences);

        info.pairKeyValueOffset = offset;

        int numKeyValuePairs = getNumberOfKeyValuePairs(offset);
        offset += SHORT_SIZE;
        offset = walkWordsList(offset, numKeyValuePairs);

        info.confidenceExtensionOffset = getSectionExtensionOffset(SENTENCES_CONFIDENCES, offset, buffer.length);
        info.encodingExtensionOffset = getSectionExtensionOffset(ENCODING_EXTENSION_ID, offset, buffer.length);
        info.wordTimingsExtensionOffset = getSectionExtensionOffset(WORD_TIMINGS, offset, buffer.length);
        info.homophonesOffset = getSectionExtensionOffset(SECTION_HOMOPHONES, offset, buffer.length);
        info.sectionSevenOffset = getSectionExtensionOffset(SECTION_SEVEN, offset, buffer.length);
        info.sectionEightOffset = getSectionExtensionOffset(SECTION_EIGHT, offset, buffer.length);
        info.sectionNineOffset = getSectionExtensionOffset(SECTION_NINE, offset, buffer.length);

        return info;
    }

    private int walkWordsList(int offset, int nbWords)
    {
        int wordFound = 0;

        while(offset < buffer.length && wordFound < nbWords)
        {
            if(buffer[offset] == 0)
                wordFound++;

            offset++;
        }
        return offset;
    }

    private int walkSentences(int offset, int numSentences)
    {
        for(int i = 0; i < numSentences; i++)
        {
            int numWordsInThisSentence = getNumberOfWordsInTheSentence(offset);
            offset += INT_SIZE;

            /*for(int j = 0; j < numWordsInThisSentence; j++)
            {
                offset += 3 * SHORT_SIZE;
            }*/
            offset += (3 * SHORT_SIZE * numWordsInThisSentence);
        }
        return offset;
    }

    private String findEncoding(byte[] buffer, BinaryFormatExtensionInfo info)
    {
        if(info.encodingExtensionOffset == EXTENSION_NOT_FOUND_OFFSET_VALUE)
            return JAVA_NAMING_DEFAULT_ENCODING;
        else
            return parseEncodingExtension(info.encodingExtensionOffset);
    }

    private String convertECNEncodingNameToJava(String encoding)
    {
        if(encoding == DEFAULT_ENCODING)
            return JAVA_NAMING_DEFAULT_ENCODING;

        return encoding;
    }

    private String parseEncodingExtension(int offset)
    {
        offset += INT_SIZE;
        // we know the encoding extension is a NUL terminated string
        // and the Java String Constructor does not take NUL
        int strlen = (int)Util.bytesToUInt(buffer, offset) - 1;
        offset += INT_SIZE;
        String str = Util.constructByteEncodedString(buffer, offset, strlen, JAVA_NAMING_DEFAULT_ENCODING);

        if(!isEncodingValid(convertECNEncodingNameToJava(str)))
            str = JAVA_NAMING_DEFAULT_ENCODING;

        return str;
    }

    private boolean isEncodingValid(String encoding)
    {
        try
        {
            new String(new byte[]{(byte)20}, 0, 1, encoding);
            return true;
        }
        catch (UnsupportedEncodingException ex)
        {
            //#ifdef NMSP_LOG
//@                        if(log.isWarnEnabled())
//@                            log.warn(encoding + " character encoding is not available in your VM. Using the default one.");
            //#endif
        }

        return false;
    }

    public DictationResultImpl parse()
    {
        Logger.debug(this, "Unpacking DNS binary version 3.2 results.");

        ArrayList<String> wordList = extractWordList(0);

        Logger.debug(this, "Found " + wordList.size() + " in word list");
        for(int i=0;i<wordList.size();i++)
            Logger.debug(this, wordList.get(i).toString());

        ArrayList<SentenceImpl> sentences = extractSentences(binaryExtensionInfo,
                                            wordList);


        Logger.debug(this, "Found " + sentences.size() + " in n-best list");
        for(int i=0;i<sentences.size();i++) {
            Logger.debug(this, sentences.get(i).toString());
        }

        extractWordAlternatives(binaryExtensionInfo, sentences, wordList);


        Map<String, String> keyPairResult = extractKeyValuePairs(binaryExtensionInfo.pairKeyValueOffset);

        setSentenceConfidence(keyPairResult, sentences);

        normalizeKeyValues(keyPairResult);

        return new DictationResultImpl(sentences, keyPairResult);
    }

    private int getSectionExtensionOffset(int sectionExtension, int offset, int maxLength)
    {
        while(offset < maxLength)
        {
            if(getExtensionId(offset) == sectionExtension)
                return offset;

            offset += INT_SIZE;
            offset += (int)getPayloadSize(offset);
            offset += INT_SIZE;
        }

        return EXTENSION_NOT_FOUND_OFFSET_VALUE;
    }

    private double getConfidenceValue(int offset)
    {
        return Util.bytesToUShort(this.buffer, offset);
    }

    private int getNumberOfWordsInWordList(int offset)
    {
        int numWords = Util.bytesToUShort(this.buffer, offset);

        Logger.debug(this, "Number of words: " + numWords);

        return numWords;
    }

    private ArrayList<String> extractWordList(int offset)
    {
        int numWords = getNumberOfWordsInWordList(offset);
        offset += SHORT_SIZE;

        return extractWordList(numWords, offset);
    }

    private ArrayList<String> extractWordList(int numWords, int offset)
    {
        ArrayList<String> wordList = new ArrayList<String>();

        for(int i = 0;i < numWords;i++)
        {
            int strlen = Util.computeStrLen(buffer, offset);
            String str = Util.constructByteEncodedString(buffer, offset, strlen, getEncoding());
            wordList.add(str);

            Logger.debug(this, "Added a word to the list: [" + str +
                    "] offset [" + offset + "] len [" +
                    strlen + "]");

            offset += strlen + 1;
        }

        return wordList;
    }

    private int getNumberOfSentences(int offset)
    {
        int numSentences = Util.bytesToInt(buffer, offset);

        Logger.debug(this, "Number of sentences: " + numSentences);

        return numSentences;
    }

    private ArrayList<SentenceImpl> extractSentences(BinaryFormatExtensionInfo info, ArrayList<String> wordList)
    {
        boolean contains32BitsTimings = info.wordTimingsExtensionOffset != EXTENSION_NOT_FOUND_OFFSET_VALUE;
        int offset = contains32BitsTimings ? info.wordTimingsExtensionOffset : info.sentencesOffset;

        boolean containsConfidence = info.confidenceExtensionOffset != EXTENSION_NOT_FOUND_OFFSET_VALUE;

        if(contains32BitsTimings)
        {
            // move up extension header
            offset += INT_SIZE;
            // move up extension size
            offset += INT_SIZE;
        }

        int numSentences = getNumberOfSentences(offset);
        offset += INT_SIZE;

        int confidenceOffset = info.confidenceExtensionOffset;

        if(containsConfidence)
        {
            // move up extension id header
            confidenceOffset += INT_SIZE;
            // move up extension size
            confidenceOffset += INT_SIZE;
            // move up nb sentences
            confidenceOffset += INT_SIZE;
            // move up nb words
            confidenceOffset += INT_SIZE;
        }


        ArrayList<SentenceImpl> sentences = new ArrayList<SentenceImpl>();

        for(int i = 0; i < numSentences; i++)
        {
            ExtractSentenceResult result = null;

            if(contains32BitsTimings)
                result = extractSentence32Bits(offset, confidenceOffset, wordList);
            else
                result = extractSentence(offset, confidenceOffset, wordList);

            Logger.debug(this, "Number of words in Sentence " + i
                + ": [" + result.sentence.size() + "]");

            sentences.add(result.sentence);
            offset = result.newOffset;
            confidenceOffset = result.newConfidenceOffset;

            Logger.debug(this, "Extracted sentence: [" + result.sentence + "]");
        }

        return sentences;
    }

    private ExtractTokenResult parseToken(int offset, int confidenceOffset, int wordTimingSize, ArrayList<String> wordList)
    {

        ExtractTokenResult result = new ExtractTokenResult();

        int wordId = extractWordId(offset);
        offset += SHORT_SIZE;
        String word = (String)wordList.get(wordId);

        long startTime = 0;
        long endTime = 0;

        if(wordTimingSize == INT_SIZE)
        {
            startTime = extractWordTime32Bits(offset);
            offset += wordTimingSize;

            endTime = extractWordTime32Bits(offset);
            offset += wordTimingSize;
        }
        else
        {
            startTime = extractWordTime(offset);
            offset += wordTimingSize;

            endTime = extractWordTime(offset);
            offset += wordTimingSize;
        }

        double confidence = 0.0f;

        if(confidenceOffset != EXTENSION_NOT_FOUND_OFFSET_VALUE)
        {
            confidence = getConfidenceValue(confidenceOffset);
            confidenceOffset += SHORT_SIZE;
        }

        TokenImpl token = new TokenImpl(word, startTime, endTime, confidence);

        result.token = token;
        result.newOffset = offset;
        result.newConfidenceOffset = confidenceOffset;

        return result;
    }

    private ExtractTokenResult parseToken32Bits(int offset, int confidenceOffset, ArrayList<String> wordList)
    {
        return parseToken(offset, confidenceOffset, INT_SIZE, wordList);
    }

    private ExtractTokenResult parseToken16Bits(int offset, int confidenceOffset, ArrayList<String> wordList)
    {

        return parseToken(offset, confidenceOffset, SHORT_SIZE, wordList);
    }

    private int getNumberOfWordsInTheSentence(int offset)
    {
        int numWordsInThisSentence = Util.bytesToInt(buffer, offset);
        Logger.debug(this, "Number of words in a sentence: " + numWordsInThisSentence);
        return numWordsInThisSentence;
    }

    private ExtractSentenceResult extractSentence(int offset, int confidenceOffset, ArrayList<String> wordList)
    {
        int numWordsInThisSentence = getNumberOfWordsInTheSentence(offset);
        offset += INT_SIZE;

        SentenceImpl sentence = new SentenceImpl();

        ExtractSentenceResult result = new ExtractSentenceResult();
        for(int i = 0;i < numWordsInThisSentence;i++)
        {
            ExtractTokenResult tokenResult = parseToken16Bits(offset, confidenceOffset, wordList);
            sentence.appendTokenToSentence(tokenResult.token);

            offset = tokenResult.newOffset;
            confidenceOffset = tokenResult.newConfidenceOffset;

            TokenImpl token = (TokenImpl) tokenResult.token;
            Logger.debug(this, "Extracted word: startTime [" + token.getStartTime()
                    + "] endTime [" + token.getEndTime() + "] content [" + token.getWord() + "]");
        }

        result.newOffset = offset;
        result.newConfidenceOffset = confidenceOffset;
        result.sentence = sentence;

        return result;
    }

    private ExtractSentenceResult extractSentence32Bits(int offset, int confidenceOffset, ArrayList<String> wordList)
    {
        int numWordsInThisSentence = getNumberOfWordsInTheSentence(offset);
        offset += INT_SIZE;

        ExtractSentenceResult result = new ExtractSentenceResult();

        for(int i = 0;i < numWordsInThisSentence;i++)
        {
            ExtractTokenResult tokenResult = parseToken32Bits(offset, confidenceOffset, wordList);

            result.sentence.appendTokenToSentence(tokenResult.token);
            offset = tokenResult.newOffset;
            confidenceOffset = tokenResult.newConfidenceOffset;

            TokenImpl token = (TokenImpl) tokenResult.token;
            Logger.debug(this, "Extracted word: startTime [" + token.getStartTime()
                    + "] endTime [" + token.getEndTime() + "] content [" + token.getWord() + "]");
        }

        result.newOffset = offset;
        result.newConfidenceOffset = confidenceOffset;

        return result;
    }

    private int extractWordId(int offset)
    {
        int wordId = Util.bytesToUShort(buffer, offset);
        wordId -= 1; /* the word Id starts from 1 to n */
        return wordId;
    }

    private int extractWordTime(int offset)
    {
        int endTime = Util.bytesToUShort(buffer, offset);
        return endTime;
    }

    private long extractWordTime32Bits(int offset)
    {
        long time = Util.bytesToUInt(buffer, offset);
        return time;
    }

    private int getNumberOfKeyValuePairs(int offset)
    {
        int numKeyValue = Util.bytesToUShort(this.buffer, offset);

        Logger.debug(this, "Number of key-value pairs: " + numKeyValue);

        return numKeyValue;
    }

    private Map<String, String> extractKeyValuePairs(int offset)
    {
        int numKeyValuePairs = getNumberOfKeyValuePairs(offset);
        offset += SHORT_SIZE;

        HashMap<String, String> keyPairResult = new HashMap<String, String>();

        ArrayList<String> wordList = extractWordList(numKeyValuePairs, offset);

        Logger.debug(this, "Extracted " + wordList.size() + " words from the set of key-value pairs.");

        for(int i=0; i < wordList.size(); i++)
        {
            String word = (String)wordList.get(i);
            int indexOfEqual = word.indexOf('=');
            if(indexOfEqual == -1)
            {
                Logger.error(this, "Received an invalid key-value pair: " + (String)wordList.get(i));
                continue;
            }

            keyPairResult.put(word.substring(0, indexOfEqual), word.substring(indexOfEqual+1, word.length()));
        }


        return keyPairResult;
    }

    private void setSentenceConfidence(Map<String, String> keyValue, List<SentenceImpl> nbestList)
    {
        for(int i=0;i<nbestList.size();i++)
        {
            String confidenceStr = (String)keyValue.get("CFD" + i);
            if(confidenceStr == null)
                continue;

            double confidence;
            try
            {
                confidence = Double.parseDouble(confidenceStr);
            } catch(NumberFormatException e)
            {
                continue;
            }

            SentenceImpl sentence = (SentenceImpl)nbestList.get(i);
            sentence.setConfidenceScore(confidence);
        }
    }

    private void normalizeKeyValues(Map<String, String> keyValue)
    {
        for(int i=0;;i++)
        {
            if(keyValue.remove("CFD" + i) == null)
                break;
        }

        if(keyValue.containsKey("IAL"))
            keyValue.put("InputAudioLength", keyValue.get("IAL"));
    }

    private long getExtensionId(int offset)
    {
        long extensionId = Util.bytesToUInt(buffer, offset);

        Logger.debug(this, "Received extension id=" + extensionId);

        return extensionId;
    }

    private long getPayloadSize(int offset)
    {
        long payloadSize = Util.bytesToUInt(buffer, offset);

        Logger.debug(this, "Received payload size: " + payloadSize);

        return payloadSize;
    }


    @SuppressWarnings("unused")
    private void extractBinarySection(List<SentenceImpl> sentences, List<String> wordList, int offset)
    {
        int sectionId = 0;
        int sectionLength = 0;
        int nbSets = 0;
        int i = 0;

        int sentenceId = 0;
        int wordPosition = 0;
        int nbAlternatives = 0;
        int altIndex = 0;

        int altWordId = 0;
        long sentenceCount = 0;

        SentenceImpl curSentence = null;
        TokenImpl curToken = null;

        sentenceCount = sentences.size();

        Util.bytesToInt(buffer, offset);
        sectionId = Util.bytesToInt(buffer, offset);
        offset += INT_SIZE;
        sectionLength = Util.bytesToInt(buffer, offset);
        offset += INT_SIZE;
        nbSets = Util.bytesToInt(buffer, offset);
        offset += INT_SIZE;

        for (i = 0 ; i < nbSets; ++i){

            sentenceId = Util.bytesToInt(buffer, offset);
            offset += INT_SIZE;
            wordPosition = Util.bytesToInt(buffer, offset);
            offset += INT_SIZE;
            nbAlternatives = Util.bytesToInt(buffer, offset);
            offset += INT_SIZE;

            if(sentenceId >= sentenceCount){
                return;
            }

            for(altIndex = 0; altIndex < nbAlternatives; ++altIndex){

                altWordId = Util.bytesToUShort(buffer, offset);
                offset += SHORT_SIZE;
                if(altWordId > wordList.size()){
                    return;
                }

                curSentence = (SentenceImpl) sentences.get(sentenceId);
                curToken = (TokenImpl) curSentence.tokenAt(wordPosition);
                // WordIds are 1 based
                addWordAlternativeToToken(curToken, (String) wordList.get(altWordId - 1), sectionId);
            }
        }
    }

    private void addWordAlternativeToToken(TokenImpl curToken, String word, int sectionId)
    {
        // TODO : The binary format specification does not include a section name. We have to hardcode it here.
        switch(sectionId){
        case SECTION_HOMOPHONES:
            curToken.addHomophone(word);
            break;
        case SECTION_SEVEN:
            curToken.addSpokenForm("katakana", word);
            break;
        case SECTION_EIGHT:
            curToken.addSpokenForm("hiragana", word);
            break;
        case SECTION_NINE:
            curToken.addSpokenForm("romaji", word);
            break;
        default:
            break;
        };
    }

    private void extractWordAlternatives(BinaryFormatExtensionInfo exInfo, List<SentenceImpl> sentences, List<String> wordList)
    {
        if(exInfo.homophonesOffset != EXTENSION_NOT_FOUND_OFFSET_VALUE){
            extractBinarySection(sentences, wordList, exInfo.homophonesOffset);
        }
        if(exInfo.sectionSevenOffset != EXTENSION_NOT_FOUND_OFFSET_VALUE){
            extractBinarySection(sentences, wordList, exInfo.sectionSevenOffset);
        }
        if(exInfo.sectionEightOffset != EXTENSION_NOT_FOUND_OFFSET_VALUE){
            extractBinarySection(sentences, wordList, exInfo.sectionEightOffset);
        }
        if(exInfo.sectionNineOffset != EXTENSION_NOT_FOUND_OFFSET_VALUE){
            extractBinarySection(sentences, wordList, exInfo.sectionNineOffset);
        }
    }

}

