/**
 *  Copyright 2006-2012, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.dragon.toolkit.recognition.dictation.parser;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.recognition.dictation.DictationResult;
import com.nuance.dragon.toolkit.recognition.dictation.Sentence;
import com.nuance.dragon.toolkit.recognition.dictation.Token;
import com.nuance.dragon.toolkit.recognition.dictation.internal.DictationResultImpl;
import com.nuance.dragon.toolkit.recognition.dictation.internal.SentenceImpl;
import com.nuance.dragon.toolkit.recognition.dictation.internal.TokenImpl;
import com.nuance.dragon.toolkit.recognition.dictation.parser.XMLSaxParser.ResultsHandler;
import com.nuance.dragon.toolkit.recognition.interpretation.Interpretation.RawPiece;
import com.nuance.dragon.toolkit.recognition.interpretation.Interpretation.RawPiece.Type;
import com.nuance.dragon.toolkit.recognition.interpretation.InterpretationImpl;
import com.nuance.dragon.toolkit.recognition.interpretation.InterpretationResult;

class NLSMLResultsHandler extends DefaultHandler implements ResultsHandler
{

    public final static String                   ELEM_RESULT                 = "result";
    public final static String                   ELEM_AUDIO_STATIS           = "NSS_Audio_Statistics";
    public final static String                   ELEM_INTERPRETATION         = "interpretation";
    public final static String                   ELEM_SWI_LITERAL            = "SWI_literalTimings";
    public final static String                   ELEM_ALTERNATIVES           = "alternatives";
    public final static String                   ELEM_SWI_MEANING_FILLERS    = "SWI_meaning_fillers";
    public final static String                   ELEM_SLOT                   = "slot";
    public final static String                   ELEM_SLOT_TYPE              = "SLOT_TYPE";
    public final static String                   ELEM_DETAILS                = "details";
    public final static String                   ELEM_FILL                   = "fill";
    public final static String                   ELEM_INPUT                  = "input";
    public final static String                   ELEM_INPUT_POSTITN          = "input_postItn";
    public final static String                   ELEM_INPUT_PREITN           = "input_preItn";
    public final static String                   ELEM_WORD                   = "word";
    public final static String                   ELEM_WORD_ALTERNATIVES      = "word_alternatives";
    public final static String                   ELEM_WORD_ALTERNATIVE       = "word_alternative";

    public final static String                   WORD_ALTERNATIVE_TYPE_HOMOPHONE = "homophone";

    public final static String                   ATTR_INPUTAUDIOLENGTH       = "InputAudioLength";
    public final static String                   ATTR_CONFIDENCE             = "confidence";
    public final static String                   ATTR_START                  = "start";
    public final static String                   ATTR_END                    = "end";
    public final static String                   ATTR_TYPE                   = "type";
    public final static String                   ATTR_POST_ITN               = "post-itn";
    public final static String                   ATTR_WORD_ID                = "wordId";
    private final static String                  ATTR_NAME                   = "name";


    public final static int                      PARSING_NONE                = 0x00;
    public final static int                      PARSING_RESULT              = 0x01;
    public final static int                      PARSING_EXTRA_INFO          = 0x02;
    public final static int                      PARSING_INTERPRETATION      = 0x03;
    public final static int                      PARSING_SWI_LITERAL         = 0x04;
    public final static int                      PARSING_ALTERNATIVES        = 0x04;
    public final static int                      PARSING_INPUT_POSTITN       = 0x05;
    public final static int                      PARSING_INPUT_PREITN        = 0x06;
    public final static int                      PARSING_SWI_WORD            = 0x07;
    public final static int                      PARSING_POSTITN_WORD        = 0x08;
    public final static int                      PARSING_PREITN_WORD         = 0x09;
    private final static int                     PARSING_SWI_MEANING_FILLERS = 0x10;
    private final static int                     PARSING_SLOT                = 0x11;
    private final static int                     PARSING_SLOT_WORD           = 0x12;
    private final static int                     PARSING_SLOT_TYPE           = 0x13;
    private final static int                     PARSING_DETAILS             = 0x14;
    private final static int                     PARSING_DETAILS_FILL        = 0x15;
    private final static int                     PARSING_WORD_ALTERNATIVES   = 0x16;

    private Stack<String>                        tokenStack                  = new Stack<String>();

    private ArrayList<SentenceImpl>              sentences                   = new ArrayList<SentenceImpl>();
    private Map<String, String>                  extraInformation            = new HashMap<String, String>();
    private SentenceImpl                         SWISentence                 = null;
    private SentenceImpl                         postItnSentence             = null;
    private SentenceImpl                         preItnSentence              = null;
    private double                               sentenceConfidence          = 0.0;

    // For Interpretation Result
    private HashMap<String, ArrayList<SentenceImpl>> _fillerSentences        = new HashMap<String, ArrayList<SentenceImpl>>();
    private List<InterpretationImpl>             _interpretations            = new ArrayList<InterpretationImpl>();
    private InterpretationImpl                   _interpretation             = null;
    private SentenceImpl                         _slotSentence               = null;
    private String                               _slotKey                    = null;
    private String                               _details_fill_str           = null;
    private String                               _currentSlotName            = null;
    private String                               _currentSlotValue           = null;
    private double                               _currentSlotConfidence      = 0.0F;

    // To handle alternatives
    // This is a little convoluted, but this is a hashtable grouping words by id and the alternate forms for that word
    // Inside the alternate forms, we have a table holding the list of alternate words, grouped by the alternate word type
    private Hashtable<Integer, Hashtable<String, List<String>>> _wordAlternativeTable = new Hashtable<Integer, Hashtable<String, List<String>>>();
    private int                                  _wordAlternativeWordId      = 0;
    private String                               _wordAlternativeType        = null;
    private String                               _wordAlternativeValue       = null;


    private void startElementResult() throws SAXException
    {
        if (tokenStack.size() != 0)
            throw new SAXException("StartElement> Found " + ELEM_RESULT + " but it is not the first token.");
    }

    private void setSentenceConfidence(Attributes attributes) throws SAXException
    {
        if (attributes.getLength() > 0 && attributes.getValue(ATTR_CONFIDENCE) != null)
        {
            sentenceConfidence = Double.parseDouble(attributes.getValue(ATTR_CONFIDENCE));
        }
    }

    private void createWord(Sentence sentence, Attributes attributes) throws SAXException
    {
        if (attributes.getLength() < 3)
        {
            throw new SAXException("sentences is empty or attributes is missing in createWord!");
        }

        long startTime = Long.parseLong(attributes.getValue(ATTR_START));
        long endTime = Long.parseLong(attributes.getValue(ATTR_END));
        double confidence = Double.parseDouble(attributes.getValue(ATTR_CONFIDENCE));

        TokenImpl word = new TokenImpl("", startTime, endTime, confidence);
        ((SentenceImpl) sentence).appendTokenToSentence(word);
    }

    private void setWordText(Sentence sentence, String content) throws SAXException
    {
        if (sentence.size() == 0)
        {
            throw new SAXException("sentence is empty in setWord!");
        }

        int last = sentence.size() - 1;
        TokenImpl word = (TokenImpl) sentence.tokenAt(last);
        word.setWord(content);
    }

    private void pickSentence()
    {
        SentenceImpl sentence = null;

        if (postItnSentence != null)
        {
            sentence = postItnSentence;
        }
        else if (SWISentence != null)
        {
            sentence = SWISentence;
        }
        else if (preItnSentence != null)
        {
            sentence = preItnSentence;
        }

        _interpretation.setInnerSentence(sentence);
        sentences.add(sentence);

        SWISentence = null;
        postItnSentence = null;
        preItnSentence = null;
        sentenceConfidence = 0.0;
    }

    private void pickFiller()
    {
        if (_slotKey != null && _slotSentence != null)
        {
            ArrayList<SentenceImpl> _tmpList;
            if (!_fillerSentences.containsKey(_slotKey))
            {
                _tmpList = new ArrayList<SentenceImpl>();
                _fillerSentences.put(_slotKey, _tmpList);
            }
            else
            {
                _tmpList = _fillerSentences.get(_slotKey);
            }
            _tmpList.add(_slotSentence);
        }
        _slotKey = null;
        _slotSentence = null;
        sentenceConfidence = 0.0;
    }

    private class TokenTemp
    {
        Token _token;

        TokenTemp(Token token, String str)
        {
            _token = token;
            // TODO: what do we do with str?
        }
    }

    private void pickInterp()
    {
        // fill: "search [searcharg] * * * [/searcharg]"
        // will be transcripted into: 0, 1searcharg, 1searcharg, 1searcharg
        // and then we revisit it to construct the recognition pieces list
        if (_details_fill_str != null)
        {
            // Prepare the no-space-token list from the inner sentence of the interpretation object
            Sentence innerSentence = _interpretation.getInnerSentence();
            ArrayList<TokenTemp> noSpaceTokenList = new ArrayList<TokenTemp>();
            for (int i = 0; i < innerSentence.size(); i++)
            {
                Token token = innerSentence.tokenAt(i);
                boolean firstToken = true;
                for (String str : token.toString().split(" "))
                {
                    noSpaceTokenList.add(new TokenTemp(firstToken? token : null, str));
                    firstToken = false;
                }
            }

            // go through the filler's pattern and try to add tokens and dictation results into the
            // interpretation object
            // the no-space-token list should have the same number of token as the _details_fill_str
            String currentSlotName = null;
            int tokenIndex = 0;
            for (String token : _details_fill_str.split(" "))
            {
                if (token.startsWith("[") && token.endsWith("]") && !token.startsWith("[/"))
                {
                    // [searcharg] opening slot
                    currentSlotName = token.substring(1, token.length() - 1);
                }
                else if (token.startsWith("[/") && token.endsWith("]"))
                {
                    // [/searcharg] closing slot
                    DictationResult dictResult = new DictationResultImpl(_fillerSentences.get(currentSlotName), null);
                    RawPiece dictPiece = new RawPiece(Type.DICTATION_RESULT, currentSlotName,
                            dictResult);
                    _interpretation.addPiece(dictPiece);
                    _interpretation.addSlot(currentSlotName, dictResult);

                    currentSlotName = null;
                }
                else if (token.equals("*"))
                {
                    // * * *
                    tokenIndex++;
                }
                else
                {
                    // search
                    // [DMTADRSDK-163] TODO: Come with a more elegant fix to skip
                    // this whole blob when patterns doesn't match
                    // "search [searcharg] * * * [/searcharg]".
                    Token realToken = null;
                    if (tokenIndex < noSpaceTokenList.size())
                    {
                        realToken = noSpaceTokenList.get(tokenIndex)._token;
                    }

                    if (realToken != null)
                    {
                        RawPiece tokenPiece = new RawPiece(Type.TOKEN, "", realToken);
                        _interpretation.addPiece(tokenPiece);
                    }
                    tokenIndex++;
                }
            }
        }
        else
        {
            Sentence sentence = _interpretation.getInnerSentence();
            if (sentence != null)
            {
                for (int i = 0; i < sentence.size(); i++)
                {
                    Token vToken = sentence.tokenAt(i);
                    _interpretation.addPiece(new RawPiece(Type.TOKEN, "", vToken));
                }
            }
        }

        if (_interpretation.getInnerSentence() != null)
            _interpretations.add(_interpretation);
        _interpretation = null;
    }

    private int getParsingState()
    {
        int state = PARSING_NONE;
        for (int i = 0; i < tokenStack.size(); ++i)
        {
            if (tokenStack.get(i).equals(ELEM_RESULT))
            {
                state = PARSING_RESULT;
            }
            else if (state == PARSING_RESULT && tokenStack.get(i).equals(ELEM_AUDIO_STATIS))
            {
                // when ELEM_AUDIO_STATIS tag is second last token
                if (i == tokenStack.size() - 2)
                {
                    state = PARSING_EXTRA_INFO;
                    break;
                }
            }
            else if (state == PARSING_RESULT && tokenStack.get(i).equals(ELEM_INTERPRETATION))
            {
                state = PARSING_INTERPRETATION;
            }
            else if (state == PARSING_INTERPRETATION && tokenStack.get(i).equals(ELEM_SWI_LITERAL))
            {
                state = PARSING_SWI_LITERAL;
            }
            else if (state == PARSING_INTERPRETATION && tokenStack.get(i).equals(ELEM_ALTERNATIVES))
            {
                state = PARSING_ALTERNATIVES;
            }
            else if (state == PARSING_INTERPRETATION && tokenStack.get(i).equals(ELEM_SWI_MEANING_FILLERS))
            {
                state = PARSING_SWI_MEANING_FILLERS;
            }
            else if (state == PARSING_SWI_MEANING_FILLERS && tokenStack.get(i).equals(ELEM_SLOT))
            {
                state = PARSING_SLOT;
            }
            else if (state == PARSING_SWI_LITERAL && tokenStack.get(i).equals(ELEM_WORD))
            {
                state = PARSING_SWI_WORD;
            }
            else if ((state == PARSING_ALTERNATIVES || state == PARSING_SWI_MEANING_FILLERS)
                    && tokenStack.get(i).equals(ELEM_INPUT_POSTITN))
            {
                state = PARSING_INPUT_POSTITN;
            }
            else if (state == PARSING_INPUT_POSTITN && tokenStack.get(i).equals(ELEM_WORD))
            {
                state = PARSING_POSTITN_WORD;
            }
            else if (state == PARSING_INPUT_POSTITN && tokenStack.get(i).equals(ELEM_WORD_ALTERNATIVES))
            {
                state = PARSING_WORD_ALTERNATIVES;
            }
            else if ((state == PARSING_ALTERNATIVES || state == PARSING_SWI_MEANING_FILLERS)
                    && tokenStack.get(i).equals(ELEM_INPUT_PREITN))
            {
                state = PARSING_INPUT_PREITN;
            }
            else if (state == PARSING_INPUT_PREITN && tokenStack.get(i).equals(ELEM_WORD))
            {
                state = PARSING_PREITN_WORD;
            }
            else if (state == PARSING_SLOT && tokenStack.get(i).equals(ELEM_WORD))
            {
                state = PARSING_SLOT_WORD;
            }
            else if (state == PARSING_INTERPRETATION && tokenStack.get(i).equals(ELEM_SLOT_TYPE))
            {
                state = PARSING_SLOT_TYPE;
            }
            else if (state == PARSING_INTERPRETATION && tokenStack.get(i).equals(ELEM_DETAILS))
            {
                state = PARSING_DETAILS;
            }
            else if (state == PARSING_DETAILS && tokenStack.get(i).equals(ELEM_FILL))
            {
                state = PARSING_DETAILS_FILL;
            }
        }

        return state;
    }

    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        Logger.debug(this, "Received startElement " + localName);

        String token = localName;
        if (localName.equals(ELEM_RESULT))
        {
            startElementResult();
        }
        else if (localName.equals(ELEM_INTERPRETATION))
        {
            setSentenceConfidence(attributes);
            _interpretation = new InterpretationImpl(sentenceConfidence);
        }
        else if (localName.equals(ELEM_SWI_LITERAL))
        {
            SWISentence = new SentenceImpl();
            SWISentence.setConfidenceScore(sentenceConfidence);
        }
        else if(localName.equals(ELEM_WORD_ALTERNATIVE))
        {
            if(PARSING_WORD_ALTERNATIVES == getParsingState())
            {
                startElementWordAlternative(attributes);
            }
        }
        else if (localName.equals(ELEM_INPUT))
        {
            if (attributes.getLength() > 0 && attributes.getValue(ATTR_TYPE) != null)
            {
                if (attributes.getValue(ATTR_TYPE).equals(ATTR_POST_ITN))
                {
                    token = ELEM_INPUT_POSTITN;
                    postItnSentence = new SentenceImpl();
                    postItnSentence.setConfidenceScore(sentenceConfidence);
                }
                else if (attributes.getValue(ATTR_TYPE).equals(ATTR_POST_ITN))
                {
                    token = ELEM_INPUT_POSTITN;
                    preItnSentence = new SentenceImpl();
                    preItnSentence.setConfidenceScore(sentenceConfidence);

                }

                if (_slotSentence != null)
                {
                    double score = Double.parseDouble(attributes.getValue(ATTR_CONFIDENCE));
                    _slotSentence.setConfidenceScore(score);
                }
            }

        }
        else if (localName.equals(ELEM_SLOT))
        {
            int length = attributes.getLength();
            String attrValue = attributes.getValue(ATTR_NAME);
            if (length > 0 && attrValue != null && !attrValue.equals(""))
            {
                _slotKey = attrValue;
                _slotSentence = new SentenceImpl();
            }
        }
        else if (localName.equals(ELEM_WORD))
        {
            if (getParsingState() == PARSING_SWI_LITERAL)
            {
                createWord(SWISentence, attributes);
            }
            else if (getParsingState() == PARSING_INPUT_POSTITN)
            {
                createWord(postItnSentence, attributes);
            }
            else if (getParsingState() == PARSING_INPUT_PREITN)
            {
                createWord(preItnSentence, attributes);
            }
            else if (getParsingState() == PARSING_SLOT)
            {
                createWord(_slotSentence, attributes);
            }
        }

        if (getParsingState() == PARSING_DETAILS)
        {
            _currentSlotName = localName;
            if (attributes.getLength() > 0 && attributes.getValue(ATTR_CONFIDENCE) != null)
            {
                _currentSlotConfidence = Double.parseDouble(attributes.getValue(ATTR_CONFIDENCE));
            }
        }

        tokenStack.push(token);
    }

    public void endElement(String namespaceURI, String localName, String rawName) throws SAXException
    {
        Logger.debug(this, "Received endElement " + localName);

        String token = tokenStack.pop();
        if (token.equals(ELEM_INPUT_POSTITN))
        {
            token = ELEM_INPUT;
        }
        else if (token.equals(ELEM_INPUT_PREITN))
        {
            token = ELEM_INPUT;
        }

        if (!token.equals(localName))
        {
            // parsing error
            sentences = null;
            extraInformation = null;
            return;
        }

        if (localName.equals(ELEM_SLOT))
        {
            pickFiller();
        }

        if (localName.equals(ELEM_INTERPRETATION))
        {
            pickInterp();
        }

        if (localName.equals(ELEM_ALTERNATIVES))
        {
            mergeWordAlternatives();
            pickSentence();
        }
        // not sure we still need this
        // since we pick the filler from the slot element already
//        else if (localName.equals(ELEM_SWI_MEANING_FILLERS))
//        {
//            pickFiller();
//        }

        if (getParsingState() == PARSING_DETAILS)
        {
            _interpretation.addSlot(_currentSlotName, _currentSlotValue, _currentSlotConfidence);
        }
        if(localName.equals(ELEM_WORD_ALTERNATIVE))
        {
            endElementWordAlternative();
        }
    }

    public void characters(char[] ch, int offset, int length) throws SAXException
    {
        Logger.debug(this, "Received characters: " + new String(ch, offset, length));

        StringBuffer value = new StringBuffer(length);
        value.append(ch, offset, length);

        int state = getParsingState();
        if (state == PARSING_EXTRA_INFO)
        {

            String key = tokenStack.peek();

            if (key.equals(ATTR_INPUTAUDIOLENGTH))
            {
                extraInformation.put("IAL", new String(value));
            }

            extraInformation.put(key, new String(value));
        }
        else if (state == PARSING_SWI_WORD)
        {
            setWordText(SWISentence, new String(value));
        }
        else if (state == PARSING_POSTITN_WORD)
        {
            setWordText(postItnSentence, new String(value));
        }
        else if (state == PARSING_PREITN_WORD)
        {
            setWordText(preItnSentence, new String(value));
        }
        else if (state == PARSING_SLOT_WORD)
        {
            if (_slotSentence != null)
                setWordText(_slotSentence, new String(value));
        }
        else if (state == PARSING_SLOT_TYPE)
        {
            _interpretation.setSlotType(new String(value));
        }
        else if (state == PARSING_DETAILS_FILL)
        {
            _details_fill_str = new String(value);
            // <fill> is inside the <details> so it should do the
            // same thing as in the <details>
            _currentSlotValue = new String(value);
        }
        else if (state == PARSING_DETAILS)
        {
            _currentSlotValue = new String(value);
        }
        else if (state == PARSING_WORD_ALTERNATIVES)
        {
            _wordAlternativeValue = new String(value);
        }
    }

    private void mergeWordAlternatives() throws SAXException
    {
        if(_wordAlternativeTable.size() == 0){
            return;
        }

        if (postItnSentence == null)
        {
            throw new SAXException("There is no Post-ITN sentence in current context for word alternatives");
        }
        for(Map.Entry<Integer, Hashtable<String, List<String>>> wordAlternatives : _wordAlternativeTable.entrySet())
        {
            TokenImpl token = (TokenImpl) postItnSentence.tokenAt(wordAlternatives.getKey());
            for(Map.Entry<String, List<String>> altType : wordAlternatives.getValue().entrySet())
            {
                String altTypeName = altType.getKey();
                for(String word : altType.getValue())
                {
                    if(altTypeName.equals(WORD_ALTERNATIVE_TYPE_HOMOPHONE))
                    {
                        token.addHomophone(word);
                    }
                    else
                    {
                        token.addSpokenForm(altTypeName, word);
                    }
                }
            }
        }
        _wordAlternativeTable.clear();
    }

    private void startElementWordAlternative(Attributes attributes) throws SAXException
    {
        if(tokenStack.isEmpty() || !tokenStack.peek().equals(ELEM_WORD_ALTERNATIVES))
            throw new SAXException("StartElement> The " + ELEM_WORD_ALTERNATIVE +
                    " token was found without a " + ELEM_WORD_ALTERNATIVES +
                    " being opened before.");
        tokenStack.push(ELEM_WORD_ALTERNATIVE);

        String wordId = attributes.getValue(ATTR_WORD_ID);
        String type = attributes.getValue(ATTR_TYPE);

        if(wordId == null)
            throw new SAXException("StartElement> There is no wordId associated with this word_alternative.");
        if(type == null)
            throw new SAXException("StartElement> There is no type associated with this word_alternative.");


        try
        {
            _wordAlternativeWordId = Integer.parseInt(wordId);
            _wordAlternativeType = type;
        } catch(NumberFormatException e)
        {
            throw new SAXException("StartElement> Could not parse the wordId : " + wordId, e);
        }

    }

    private void endElementWordAlternative() throws SAXException
    {
        if(tokenStack.isEmpty() || tokenStack.peek() != ELEM_WORD_ALTERNATIVE)
        {
            throw new SAXException("End Element> The top of the stack does not contain the token " + ELEM_WORD_ALTERNATIVE);
        }
        if(_wordAlternativeValue == null)
        {
            throw new SAXException("Did not get any character from " + ELEM_WORD);
        }

        Hashtable<String, List<String>> altTable = _wordAlternativeTable.get(_wordAlternativeWordId);
        if(altTable == null){
            altTable = new Hashtable<String, List<String>>();
            _wordAlternativeTable.put(_wordAlternativeWordId, altTable);
        }
        List<String> altWords = altTable.get(_wordAlternativeType);
        if(altWords == null){
            altWords = new LinkedList<String>();
            altTable.put(_wordAlternativeType, altWords);
        }
        altWords.add(_wordAlternativeValue);

        _wordAlternativeWordId = 0;
        _wordAlternativeType = null;
        _wordAlternativeValue = null;

        // Remove the current ELEM_WORD_ALTERNATIVE from the stack
        tokenStack.pop();

    }

    @Override
    public DictationResultImpl getDictationResult()
    {
        return new DictationResultImpl(sentences, extraInformation);
    }

    @Override
    public InterpretationResult getInterpretationResult()
    {
        if (_interpretations != null && _interpretations.size() > 0)
            return new InterpretationResult(_interpretations, extraInformation);
        else
            return null;
    }
}
