/**
 *  Copyright 2006-2012, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.dragon.toolkit.recognition.dictation.parser;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Stack;
import java.util.StringTokenizer;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.recognition.dictation.internal.DictationResultImpl;
import com.nuance.dragon.toolkit.recognition.dictation.internal.SentenceImpl;
import com.nuance.dragon.toolkit.recognition.dictation.internal.TokenImpl;
import com.nuance.dragon.toolkit.recognition.dictation.parser.XMLSaxParser.ResultsHandler;
import com.nuance.dragon.toolkit.recognition.interpretation.InterpretationResult;

class XMLResultsHandler extends DefaultHandler implements ResultsHandler
{

    public final static String ELEM_RESULT = "result";
    public final static String ELEM_INPUT = "input";
    public final static String ELEM_INTERPRETATION = "interpretation";
    public final static String ELEM_AUDIO_STATS = "NSS_Audio_Statistics";
    public final static String ATTR_GRAMMAR = "grammar";
    public final static String ATTR_TIMING = "timing";
    public final static String ATTR_CONFIDENCE = "confidence";
    public final static String ATTR_WORDS_CONFIDENCE = "word_confidence";
    public final static String ATTR_MODE = "mode";
    public final static String SEP_COMMA = ",";
    public final static String SEP_SPACE = " ";
    public final static String SEP_HYPHEN = "-";

    private Stack<String> tokenStack = new Stack<String>();

    private StringBuffer inputBuf = null;
    private StringBuffer audioStatValue = null;

    private String timings = null;
    private String wordconfidences = null;
    private double sentenceConf = 0.0;

    private ArrayList<SentenceImpl> nbestList = new ArrayList<SentenceImpl>();
    private Map<String, String> extraInformation = new HashMap<String, String>();

    public void startElement(String uri, String localName, String qName,
            Attributes attributes) throws SAXException
    {
        Logger.debug(this, "Received startElement " + localName);

        if(localName.equals(ELEM_RESULT))
            startElementResult();
        else if(localName.equals(ELEM_INTERPRETATION))
            startElementInterpretation(attributes);
        else if(localName.equals(ELEM_INPUT))
            startElementInput();
        else if(localName.equals(ELEM_AUDIO_STATS))
            startElementAudioStats();
    }

    private void startElementResult() throws SAXException
    {
        if(tokenStack.size() != 0)
            throw new SAXException("StartElement> Found " + ELEM_RESULT +
                    " but it is not the first token.");
        tokenStack.push(ELEM_RESULT);
    }

    private void startElementInterpretation(Attributes attributes)
            throws SAXException
    {
        if(tokenStack.isEmpty() || tokenStack.peek() != ELEM_RESULT)
            throw new SAXException("StartElement> The " + ELEM_INTERPRETATION +
                    " token was found without a " + ELEM_RESULT +
                    " being opened before.");
        tokenStack.push(ELEM_INTERPRETATION);

        String timing = attributes.getValue(ATTR_TIMING);
        String confidence = attributes.getValue(ATTR_CONFIDENCE);
        String wordconfidence = attributes.getValue(ATTR_WORDS_CONFIDENCE);


        if(timing == null)
            throw new SAXException("StartElement> There are no timings associated with this interpretation.");
        if(confidence == null)
            throw new SAXException("StartElement> There is no confidence associated with this interpretation.");

        this.timings = timing;

        this.wordconfidences = wordconfidence;

        try
        {
            sentenceConf = Double.parseDouble(confidence);
        } catch(NumberFormatException e)
        {
            throw new SAXException("StartElement> Could not parse the confidence: " + confidence, e);
        }
    }

    private void startElementInput() throws SAXException
    {
        if(tokenStack.isEmpty() || tokenStack.peek() != ELEM_INTERPRETATION)
            throw new SAXException("StartElement> The " + ELEM_INPUT +
                    " token was found without a " + ELEM_INTERPRETATION +
                    " being opened before.");
        tokenStack.push(ELEM_INPUT);
    }

    private void startElementAudioStats() throws SAXException
    {
        if(tokenStack.isEmpty() || tokenStack.peek() != ELEM_RESULT)
            throw new SAXException("StartElement> The " + ELEM_AUDIO_STATS +
                    " token was found without a " + ELEM_RESULT +
                    " being opened before.");
        tokenStack.push(ELEM_AUDIO_STATS);
    }

    public void endElement(String namespaceURI, String localName, String rawName)
            throws SAXException
    {
        Logger.debug(this, "Received endElement " + localName);

        if(localName.equals(ELEM_RESULT))
        {
            if(tokenStack.isEmpty() || tokenStack.peek() != ELEM_RESULT)
                throw new SAXException("End Element> The top of the stack does not contain the token " + ELEM_RESULT);
            tokenStack.pop();
        }
        else if(localName.equals(ELEM_INTERPRETATION))
        {
            if(tokenStack.isEmpty() || tokenStack.peek() != ELEM_INTERPRETATION)
                throw new SAXException("End Element> The top of the stack does not contain the token " + ELEM_INTERPRETATION);
            tokenStack.pop();
        }
        else if(localName.equals(ELEM_INPUT))
        {
            endElementInput();
            tokenStack.pop();
        }
        else if(localName.equals(ELEM_AUDIO_STATS))
        {
            if(tokenStack.isEmpty() || tokenStack.peek() != ELEM_AUDIO_STATS)
                throw new SAXException("End Element> The top of the stack does not contain the token " + ELEM_AUDIO_STATS);
            tokenStack.pop();
        }
        else if(!tokenStack.isEmpty() && tokenStack.peek() == ELEM_AUDIO_STATS)
        {
            String value = (audioStatValue == null ? "" : audioStatValue.toString().trim());

            if(localName.equals("InputAudioLength"))
                extraInformation.put("IAL", value);

            extraInformation.put(localName, value);
            audioStatValue = null;
        }
    }

    private void endElementInput() throws SAXException
    {
        if(tokenStack.isEmpty() || tokenStack.peek() != ELEM_INPUT)
            throw new SAXException("End Element> The top of the stack does not contain the token " + ELEM_INTERPRETATION);

        if(timings == null)
            throw new SAXException("Did not get any timings from " + ELEM_INPUT);
        if(inputBuf == null)
            throw new SAXException("Did not get any character from " + ELEM_INPUT);

        String input = inputBuf.toString();
        StringTokenizer stWords = new StringTokenizer(input, SEP_SPACE);
        StringTokenizer stTiming = new StringTokenizer(timings, SEP_COMMA);
        StringTokenizer stWordConfidence = null;

        if(wordconfidences != null)
            stWordConfidence = new StringTokenizer(wordconfidences, SEP_COMMA);

        inputBuf = null;
        timings = null;
        wordconfidences = null;

        if(stWords.countTokens() != stTiming.countTokens())
        {
            throw new SAXException(
                    "timing("
                            + stTiming.countTokens()
                            + ") and inputs("
                            + stWords.countTokens()
                            + ") information does not have the same number of items");
        }



        SentenceImpl sentence = new SentenceImpl();
        sentence.setConfidenceScore(sentenceConf);
        while(stWords.hasMoreTokens())
        {
            TokenImpl token = parseWord(stWords, stTiming, stWordConfidence);
            sentence.appendTokenToSentence(token);
        }

        addToNBestList(sentence);
    }

    private TokenImpl parseWord(StringTokenizer stWords,
            StringTokenizer stTiming, StringTokenizer stWordConfidence)
    {
        String tokenString = stWords.nextToken();
        String timingString = stTiming.nextToken();

        double confidence = 0.0;
        if (stWordConfidence != null){
            String confidenceString = stWordConfidence.nextToken();
            confidence = Double.parseDouble(confidenceString);
        }

        long startTime = Long.parseLong(timingString.substring(0,
                timingString.indexOf(SEP_HYPHEN)));
        long endTime = Long.parseLong(timingString.substring(timingString
                .indexOf(SEP_HYPHEN) + 1));

        return new TokenImpl(tokenString, startTime, endTime, confidence);
    }

    private void addToNBestList(SentenceImpl sentence)
    {
        if(!nbestList.isEmpty())
        {
            for(int i=0;i<nbestList.size();i++)
            {
                if( ((SentenceImpl)nbestList.get(i)).getConfidenceScore() < sentence.getConfidenceScore())
                {
                    nbestList.add(i, sentence);
                    return;
                }
            }
        }

        nbestList.add(sentence);
    }

    public void characters(char[] ch, int offset, int length) throws SAXException
    {
        Logger.debug(this, "Received characters: " + new String(ch, offset, length));

        if(!tokenStack.isEmpty() && tokenStack.peek() == ELEM_INPUT)
        {
            if(inputBuf == null)
                inputBuf = new StringBuffer(length);
            inputBuf.append(ch, offset, length);
        }
        else if(!tokenStack.isEmpty() && tokenStack.peek() == ELEM_AUDIO_STATS)
        {
            if(audioStatValue == null)
                audioStatValue = new StringBuffer(length);
            audioStatValue.append(ch, offset, length);
        }
    }

    @Override
    public DictationResultImpl getDictationResult()
    {
        return new DictationResultImpl(nbestList, extraInformation);
    }

    @Override
    public InterpretationResult getInterpretationResult()
    {
        return null;
    }
}
