package com.nuance.dragon.toolkit.recognition.interpretation;

import java.util.List;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.recognition.dictation.DictationResult;
import com.nuance.dragon.toolkit.recognition.dictation.Token;

/**
 * An interpretation of recognized text. The contents of the interpretation depend on
 * the recognition / interpretation grammar that was used to process the speech.
 */
public interface Interpretation
{
    /**
     * An interpreted recognition slot. It represents a named chunk of the recognition--
     * for example, name could be "song_title" and text could be "Twinkle Twinkle Little Star".
     */
    public static final class Slot
    {
        /** The name of the slot, as defined in the interpretation grammar. Never null. */
        public final String name;
        /** The contents of the slot. Never null. */
        public final String text;
        /** The dictation result. May be null if this slot does not represent a full dictation. */
        public final DictationResult dictationResult;
        /** The confidence score. TODO: what does it mean? */
        public final double confidenceScore;

        public Slot(String name, String text, double confidenceScore)
        {
            this(name, text, null, confidenceScore);
        }

        public Slot(String name, DictationResult dictationResult)
        {
            this(name, dictationResult.toString(), dictationResult, dictationResult.sentenceAt(0).getConfidenceScore());
        }

        private Slot(String name, String text, DictationResult dictationResult, double confidenceScore)
        {
            Assert.assertNotNull(name);
            Assert.assertTrue(name.length() > 0);
            Assert.assertNotNull(text);

            this.name = name;
            this.text = text;
            this.dictationResult = dictationResult;
            this.confidenceScore = confidenceScore;
        }
    }

    /**
     * A raw recognition piece. It represents a piece of text PRIOR to interpretation.
     */
    public static final class RawPiece
    {
        /**
         * The type of the piece: either a simple token, or a dictation result.
         */
        public enum Type
        {
            TOKEN, DICTATION_RESULT
        };

        /** The name of the piece. TODO: how to explain this? */
        public final String          name;
        /** The type of the piece */
        public final Type            type;
        /** The token of this piece. Will be null if the type is not {@link Type#TOKEN} */
        public final Token           token;
        /** The dictation result of this piece. Will be null if the type is not {@link Type#DICTATION_RESULT} */
        public final DictationResult dictResult;
        /** The confidence score. TODO: what does it mean? */
        public final double          confidenceScore;

        public RawPiece(Type type, String name, Token token)
        {
            this(type, name, token, null, token.getConfidenceScore());
        }

        public RawPiece(Type type, String name, DictationResult dictResult)
        {
            this(type, name, null, dictResult, dictResult.sentenceAt(0).getConfidenceScore());
        }

        private RawPiece(Type type, String name, Token token, DictationResult dictResult, double confidenceScore)
        {
            this.type = type;
            this.name = name;
            this.token = token;
            this.dictResult = dictResult;
            this.confidenceScore = confidenceScore;
        }
    }

    /**
     * @return The raw recognition pieces.
     */
    List<RawPiece> getRecognitionPieces();

    /**
     * @return The interpreted slots.
     */
    List<Slot> getSlots();

    /**
     * Get the interpreted slot with the given name.
     * @param name The slot name to look for
     * @return The interpreted slot, or null if it couldn't be found.
     */
    Slot getSlot(String name);

    /**
     * Get the type of the recognition. Maps to SLOT_TYPE of the NVC_ASR_CMD. E.g. "call" or "play".
     * @return The recognition type, never null.
     */
    String getRecogType();

    /**
     * @return The overall confidence score of the interpretation. TODO: what does it mean?
     */
    double getConfidenceScore();

    /**
     * Generates the string representation of the interpretation.
     *
     * @return the interpretation as a string
     */
    String toString();
}
