package com.nuance.dragon.toolkit.remoting;

import java.util.HashMap;

import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * Represents a connection through which other services can be communicated with.
 */
public abstract class Connection
{
    private final String _name;
    private final HashMap<String, Connection> _connections;
    private final HashMap<String, Connection> _services;

    public Connection(String name)
    {
        _name = name;
        _connections = new HashMap<String, Connection>();
        _services = new HashMap<String, Connection>();
    }

    /**
     * Get the name of this connection.
     * @return The name.
     */
    public String getName()
    {
        return _name;
    }

    /**
     * Add another connection. Connections that have been added will be notified
     * when services are connected or disconnected.
     * @param connection
     */
    public void addConnection(Connection connection)
    {
        _connections.put(connection.getName(), connection);
    }

    /**
     * Remove a connection.
     * @param connection
     */
    public void removeConnection(Connection connection)
    {
        _connections.remove(connection.getName());
    }

    /**
     * A service has been connected. This connection will keep track of the
     * service name, and the connection from which the notification came, in
     * order to send messages to this service later. This method will notify
     * other connections of the new service.
     * @param serviceName The name of the service that was connected.
     * @param connection The connection from which the service notification came.
     */
    public void onServiceConnected(String serviceName, Connection connection)
    {
        _services.put(serviceName, connection);
        for (Connection c : _connections.values())
        {
            if (c == connection) continue;

            c.onServiceConnected(serviceName, this);
        }
    }

    /**
     * A service has been disconnected.
     * @param serviceName The name of the service that was disconnected.
     * @param connection The connection from which the service notification came.
     */
    public void onServiceDisconnected(String serviceName, Connection connection)
    {
        _services.remove(serviceName);
        for (Connection c : _connections.values())
        {
            if (c == connection) continue;

            c.onServiceDisconnected(serviceName, this);
        }
    }

    /**
     * Send a message through this connection.
     * @param to The service to send the message to.
     * @param from The service that is sending the message (may be used by the
     * destination service to send replies).
     * @param messageId The ID of the message to send.
     * @param message The contents of the message.
     * @return true if a receiver was found for the message; false otherwise.
     */
    public boolean sendMessage(String to, String from, String messageId, JSONObject message)
    {
        for (Connection c : _connections.values())
        {
            if (c.sendMessage(to, from, messageId, message))
                return true;
        }
        return false;
    }
}
