package com.nuance.dragon.toolkit.remoting;

import org.json.JSONException;

import android.os.Handler;

import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * Base class for a connection that connects to another remote connection.
 * Messages sent to this connection will all be routed to the remote connection.
 * Messages received from the remote connection will be sent to the attached
 * connections.
 * <p>
 * In general, remote connections should come in pairs, which should know how to
 * connect to each other.
 */
public abstract class RemoteConnection extends Connection
{
    private static final String JSON_KEY_TO = "to";
    private static final String JSON_KEY_FROM = "from";
    private static final String JSON_KEY_ID = "id";
    private static final String JSON_KEY_MSG = "msg";

    private final Handler _mainThreadHandler;

    public RemoteConnection(String name)
    {
        super(name);
        _mainThreadHandler = new Handler();
    }

    /**
     * Connect to the remote connection on the other side, and begin listening
     * for messages from the other side and sending messages to the other side.
     */
    public abstract void start();

    /**
     * Disconnect, and stop listening for messages and sending messages.
     */
    public abstract void stop();

    @Override
    public boolean sendMessage(String to, String from, String messageId, JSONObject message)
    {
        // Assume that the service is on the other end of the remote connection
        JSONObject messageWrapper = new JSONObject();
        messageWrapper.tryPut(JSON_KEY_TO, to);
        messageWrapper.tryPut(JSON_KEY_FROM, from);
        messageWrapper.tryPut(JSON_KEY_ID, messageId);
        messageWrapper.tryPut(JSON_KEY_MSG, message);
        sendMessage(messageWrapper.toString());
        return true;
    }

    private void onMessageReceived(JSONObject messageWrapper)
    {
        String to = messageWrapper.optString(JSON_KEY_TO);
        String from = messageWrapper.optString(JSON_KEY_TO);
        String id = messageWrapper.optString(JSON_KEY_TO);
        JSONObject message = new JSONObject(messageWrapper.optJSONObject(JSON_KEY_MSG));

        // TODO: kind of ugly, reverting to default behavior here of sending this
        // message to other connections
        super.sendMessage(to, from, id, message);
    }

    /**
     * Call this method to notify that there is a new message available. This
     * method may be called from any thread.
     */
    protected final void onMessageAvailable(final String message)
    {
        _mainThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                try
                {
                    onMessageReceived(new JSONObject(message));
                }
                catch (JSONException e)
                {
                    Logger.error(RemoteConnection.this, "Error parsing JSON", e);
                }
            }
        });
    }

    /**
     * This method sends a message to the remote connection on the other side.
     * This will be called from the main application thread.
     * @param message
     */
    protected abstract void sendMessage(String message);
}
