package com.nuance.dragon.toolkit.speechkit;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioEnergyListener;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.SpeechDetectionListener;
import com.nuance.dragon.toolkit.audio.pipes.AudioEnergyCalculatorPipe;
import com.nuance.dragon.toolkit.audio.pipes.EndPointerPipe;
import com.nuance.dragon.toolkit.audio.pipes.SpeexEncoderPipe;
import com.nuance.dragon.toolkit.audio.sources.MicrophoneRecorderSource;
import com.nuance.dragon.toolkit.audio.sources.RecorderSource;
import com.nuance.dragon.toolkit.audio.sources.StreamingFileRecorderSource;
import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognitionError;
import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognitionResult;
import com.nuance.dragon.toolkit.cloudservices.recognizer.CloudRecognizer;
import com.nuance.dragon.toolkit.cloudservices.recognizer.RecogSpec;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * A simple cloud recognizer that handles all the prompting, recording,
 * end-pointing, and audio routing for you.
 */
public final class SKCloudRecognizer
{
    public static final class PromptTypes
    {
        public static final int START_LISTENING = 0;
        public static final int STOP_LISTENING = 1;
        public static final int RESULT = 2;
        public static final int ERROR = 3;
    }

    public static final class EndPointerTypes
    {
        public static final int NONE = 0;
        public static final int LONG = 1;
        // TODO: short
    }

    /**
     * Listens for recognition progress.
     */
    public interface Listener
    {
        /**
         * Listening has begun.
         * @param recognizer
         */
        void onListeningStart(SKCloudRecognizer recognizer);

        /**
         * Listening has stopped.
         * @param recognizer
         */
        void onListeningStop(SKCloudRecognizer recognizer);

        /**
         * An error occurred.
         * @param recognizer
         * @param error
         */
        void onError(SKCloudRecognizer recognizer, CloudRecognitionError error);

        /**
         * A result was received.
         * @param recognizer
         * @param result
         */
        void onResult(SKCloudRecognizer recognizer, CloudRecognitionResult result);
    }

    private final CloudServices _cloud;
    private final CloudRecognizer _recognizer;
    private final RecorderSource<AudioChunk> _recorder;
    private final SpeexEncoderPipe _speexEncoder;
    private final RecogSpec _recogSpec;
    private final int _endPointerType;
    private final Listener _listener;

    private final AudioEnergyCalculatorPipe _energyCalculator;
    private float _audioEnergyLevel;

    private EndPointerPipe _endPointer;
    private boolean _started;
    private boolean _stopped;
    private boolean _done;

    private SKPrompt _startListeningPrompt;
    private SKPrompt _stopListeningPrompt;
    private SKPrompt _resultPrompt;
    private SKPrompt _errorPrompt;

    private SKPrompt _currentPrompt;

    SKCloudRecognizer(CloudServices cloud,
                      RecogSpec recogSpec,
                      int endPointerType,
                      String fileName,
                      boolean fileIsLittleEndian,
                      Listener listener)
    {
        _cloud = cloud;
        _recognizer = new CloudRecognizer(_cloud);
        if (fileName == null)
            _recorder = new MicrophoneRecorderSource(AudioType.PCM_16k);
        else
            _recorder = new StreamingFileRecorderSource(AudioType.PCM_16k, fileName, fileIsLittleEndian);
        _energyCalculator = new AudioEnergyCalculatorPipe(new AudioEnergyListener()
        {
            @Override
            public void onEnergyLevelAvailable(float energyLevel, boolean speech)
            {
                _audioEnergyLevel = energyLevel;
            }
        });
        _speexEncoder = new SpeexEncoderPipe();
        _recogSpec = recogSpec;
        _endPointerType = endPointerType;
        _listener = listener;
    }

    /**
     * Set a prompt to play during the recognition.
     * @param promptType
     * @param prompt
     */
    public void setPrompt(int promptType, SKPrompt prompt)
    {
        switch(promptType)
        {
        case PromptTypes.START_LISTENING:
            _startListeningPrompt = prompt;
            break;
        case PromptTypes.STOP_LISTENING:
            _stopListeningPrompt = prompt;
            break;
        case PromptTypes.RESULT:
            _resultPrompt = prompt;
            break;
        case PromptTypes.ERROR:
            _errorPrompt = prompt;
            break;
        }
    }

    /**
     * Returns a numeric representation of the audio energy of the last audio buffer
     * processed, The audio level is within the range from 0 to 90, the value is
     * proportional to the loudness of the audio. Based on the experiments, only the
     * audio with the energy value greater than 40 will be heard by normal human ears.
     *
     * @return audio energy level
     */
    public float getAudioLevel()
    {
        return _audioEnergyLevel;
    }

    /**
     * Start the recognition.
     */
    public void start()
    {
        Checker.checkState(this, !_started, "Already started");
        Checker.checkState(this, !_done, "Already finished");

        _started = true;
        _stopped = false;

        // Start the recorder, and wait for it to start
        _recorder.startRecording(new RecorderSource.Listener<AudioChunk>()
        {
            @Override
            public void onStarted(RecorderSource<AudioChunk> recorder)
            {
                if (_done)
                    return;

                if (_startListeningPrompt != null)
                {
                    Logger.debug(SKCloudRecognizer.this, "Recorder started, playing prompt");
                    _currentPrompt = _startListeningPrompt;
                    _startListeningPrompt = null;
                    _currentPrompt.start(new SKPrompt.Listener()
                    {
                        @Override
                        public void onFinished(SKPrompt prompt)
                        {
                            if (_done)
                                return;

                            Logger.debug(SKCloudRecognizer.this, "Prompt finished, listening");
                            _currentPrompt = null;
                            startListening();
                        }
                    });
                } else
                {
                    Logger.debug(SKCloudRecognizer.this, "Recorder started, listening");
                    startListening();
                }
            }

            @Override
            public void onStopped(RecorderSource<AudioChunk> recorder)
            {
            }

            @Override
            public void onError(RecorderSource<AudioChunk> recorder)
            {
            }
        });

        if (_endPointerType != EndPointerTypes.NONE)
        {
            _endPointer = new EndPointerPipe(new SpeechDetectionListener()
            {
                @Override
                public void onStartOfSpeech()
                {
                    if (_done)
                        return;
                }

                @Override
                public void onEndOfSpeech()
                {
                    if (_done)
                        return;

                    Logger.debug(SKCloudRecognizer.this, "End of speech detected, stopping recorder");

                    // Cleaning-up Speex audio chain
                    if (_endPointer != null)
                        _endPointer.disconnectAudioSource();
                    _speexEncoder.disconnectAudioSource();
                    _energyCalculator.disconnectAudioSource();

                    stopListening();
                }
            });
            _endPointer.connectAudioSource(_speexEncoder);
        }

        // Start the recognition, but don't connect the speex encoder to the
        // recorder until the recorder has started and the prompt (if any) has
        // finished.
        _recognizer.startRecognition(_recogSpec,
                                     _endPointer != null ? _endPointer : _speexEncoder,
                                     new CloudRecognizer.Listener()
        {
            @Override
            public void onResult(CloudRecognitionResult result)
            {
                if (_done)
                    return;

                Logger.debug(SKCloudRecognizer.this, "Result received");
                if (_resultPrompt != null)
                {
                    _resultPrompt.start(null);
                    _resultPrompt = null;
                }

                cleanup();

                _listener.onResult(SKCloudRecognizer.this, result);
            }

            @Override
            public void onError(CloudRecognitionError error)
            {
                Logger.debug(SKCloudRecognizer.this, "Error received");
                // Don't play the error prompt if this was canceled
                if (!_done && _errorPrompt != null)
                {
                    _errorPrompt.start(null);
                    _errorPrompt = null;
                }

                cleanup();

                _listener.onError(SKCloudRecognizer.this, error);
            }

            @Override
            public void onTransactionIdGenerated(String transactionId) {
                // TODO Auto-generated method stub
                
            }
        });
    }

    /**
     * Stop listening.
     */
    public void stopListening()
    {
        Checker.checkState(this, _started, "Not started yet");

        if (_stopped)
            return;

        Logger.debug(this, "Stopping recorder");
        _stopped = true;

        if (!_done)
        {
            _recognizer.processResult();
            _recorder.stopRecording();
        }
        else
        {
            Logger.warn(this, "current recognition has been already processed");
        }

        _listener.onListeningStop(this);

        if (_stopListeningPrompt != null)
        {
            if (_currentPrompt != null)
            {
                _currentPrompt.cancel();
            }
            _currentPrompt = _stopListeningPrompt;
            _stopListeningPrompt = null;
            _currentPrompt.start(new SKPrompt.Listener()
            {
                @Override
                public void onFinished(SKPrompt prompt)
                {
                    if (_done)
                        return;

                    if (prompt == _currentPrompt)
                        _currentPrompt = null;
                }
            });
        }
    }

    /**
     * Cancel the recognition.
     */
    public void cancel()
    {
        if (_done)
            return;

        Logger.debug(this, "Canceling");
        cleanup();
    }

    private void startListening()
    {
        // Now connect the speex encoder.
        _energyCalculator.connectAudioSource(_recorder);
        _speexEncoder.connectAudioSource(_energyCalculator);
        _listener.onListeningStart(this);
    }

    private void cleanup()
    {
        if (!_done)
        {
            _done = true;

            if (_currentPrompt != null)
            {
                _currentPrompt.cancel();
                _currentPrompt = null;
            }

            _startListeningPrompt = null;
            _stopListeningPrompt = null;
            _resultPrompt = null;
            _errorPrompt = null;

            _recorder.stopRecording();
            _recognizer.cancel();

            _speexEncoder.release();
        }
    }
}
