package com.nuance.dragon.toolkit.speechkit;

import java.util.ArrayList;
import java.util.LinkedList;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.pipes.SpeexDecoderPipe;
import com.nuance.dragon.toolkit.audio.sinks.PlayerSink;
import com.nuance.dragon.toolkit.audio.sinks.SpeakerPlayerSink;
import com.nuance.dragon.toolkit.cloudservices.CloudConfig;
import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.TransactionError;
import com.nuance.dragon.toolkit.cloudservices.vocalizer.CloudVocalizer;
import com.nuance.dragon.toolkit.cloudservices.vocalizer.TtsSpec;
import com.nuance.dragon.toolkit.core.data.Data;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
/**
 * A simple cloud vocalizer that handles queuing and playing multiple network
 * TTS requests for you.
 */
public class SKCloudVocalizer
{
    /**
     * Listen for TTS progress.
     */
    public interface Listener
    {
        /**
         * Speaking has begun for a phrase.
         * @param vocalizer
         * @param context The context that was passed to {@link SKCloudVocalizer#speakString}.
         */
        void onSpeakingBegin(SKCloudVocalizer vocalizer,
                             Object context);

        /**
         * Speaking has finished for a phrase.
         * @param vocalizer
         * @param moreToPlay true if there is more queued to play.
         * @param error non-null if there was an error.
         * @param context The context that was passed to {@link SKCloudVocalizer#speakString}.
         */
        void onSpeakingDone(SKCloudVocalizer vocalizer,
                            boolean moreToPlay,
                            TransactionError error,
                            Object context);
    }

    /**
     * A language.
     */
    public static final class Language
    {
        private final String _language;

        /**
         * @param language The language to use (passed in the 'tts_language'
         * key to the server)
         */
        public Language(String language)
        {
            _language = language;
        }
    }

    /**
     * A voice.
     */
    public static final class Voice
    {
        private final String _voice;

        /**
         * @param voice The voice to use (passed in the 'tts_voice' key to the
         * server)
         */
        public Voice(String voice)
        {
            _voice = voice;
        }
    }

    private static final class TtsRequest
    {
        final TtsSpec spec;
        final Object context;
        public TtsRequest(TtsSpec spec, Object context)
        {
            this.spec = spec;
            this.context = context;
        }
    }

    private final CloudServices _cloud;
    private final CloudConfig _config;
    private final Listener _listener;
    private final LinkedList<TtsRequest> _ttsQueue;
    private CloudVocalizer _vocalizer;
    private SpeexDecoderPipe _speexDecoder;
    private PlayerSink _player;

    SKCloudVocalizer(CloudServices cloud,
                     CloudConfig config,
                     Listener listener)
    {
        _cloud = cloud;
        _config = config;
        _listener = listener;
        _ttsQueue = new LinkedList<TtsRequest>();
    }

    /**
     * Speak some text using the specified language. Uses the NVC_TTS_CMD.
     * @param text The text to speak.
     * @param isSsml Whether the text is SSML or not.
     * @param language The language.
     * @param context The context to pass back in the {@link Listener} methods.
     */
    public void speakString(String text, boolean isSsml, Language language, Object context)
    {
        Checker.checkStringArgForNullOrEmpty("language", language._language);
        speakString(text, isSsml, language._language, null, context);
    }

    /**
     * Speak some text using the specified voice. Uses the NVC_TTS_CMD.
     * @param text The text to speak.
     * @param isSsml Whether the text is SSML or not.
     * @param voice The voice.
     * @param context The context to pass back in the {@link Listener} methods.
     */
    public void speakString(String text, boolean isSsml, Voice voice, Object context)
    {
        Checker.checkStringArgForNullOrEmpty("voice", voice._voice);
        speakString(text, isSsml, null, voice._voice, context);
    }

    private void speakString(String text, boolean isSsml, String language, String voice, Object context)
    {
        Checker.checkStringArgForNullOrEmpty("text", text);
        Data.Dictionary ttsParamData = new Data.Dictionary();
        ttsParamData.put("tts_input", text);
        ttsParamData.put("tts_type", isSsml ? "ssml" : "text");

        Data.Dictionary settings = new Data.Dictionary();
        if (voice != null)
        {
            settings.put("tts_voice", voice);
        } else
        {
            settings.put("tts_language", language);
        }

        TtsSpec spec = new TtsSpec("NVC_TTS_CMD", settings, "TEXT_TO_READ", ttsParamData, AudioType.SPEEX_WB);
        speakString(spec, context);
    }

    /**
     * Perform a TTS command using a custom TTS spec.
     * @param ttsSpec
     * @param context
     */
    public void speakString(TtsSpec ttsSpec, Object context)
    {
        Checker.checkArgForNull("ttsSpec", ttsSpec);
        Checker.checkArgForCondition("ttsSpec", "have the same audio output type of the SpeechKit object", ttsSpec.getAudioType().equals(_config.playerCodec));

        _ttsQueue.add(new TtsRequest(ttsSpec, context));

        if (_vocalizer == null)
        {
            startNext();
        }
    }

    /**
     * Cancel all active and queued TTS requests.
     */
    public void cancel()
    {
        cleanup();
    }

    private void cleanup()
    {
        if (_vocalizer != null)
        {
            _vocalizer.cancel();
             // Let the error callback set _vocalizer = null
        }

        if (_player != null)
        {
            _player.disconnectAudioSource();
            _player.stopPlaying();
            _player = null;
        }

        if (_speexDecoder != null)
        {
            _speexDecoder.disconnectAudioSource();
            _speexDecoder.release();
            _speexDecoder = null;
        }
    }

    private void startNext()
    {
        if (!_ttsQueue.isEmpty())
        {
            _vocalizer = new CloudVocalizer(_cloud);
            final CloudVocalizer vocalizer = _vocalizer;
            final TtsRequest next = _ttsQueue.remove();

            AudioSource<AudioChunk> ttsSource = _vocalizer.generateTts(next.spec, new CloudVocalizer.Listener()
            {
                @Override
                public void onProcessingStarted()
                {
                }
                
                @Override
                public void onSuccess()
                {
                }

                @Override
                public void onError(TransactionError error)
                {
                    if (_vocalizer != vocalizer)
                        return;

                    _vocalizer = null;

                    int numQueued = _ttsQueue.size();
                    ArrayList<TtsRequest> requests = new ArrayList<TtsRequest>(numQueued);
                    _ttsQueue.clear();

                    cleanup();

                    if (_listener != null)
                    {
                        _listener.onSpeakingDone(SKCloudVocalizer.this, numQueued-- > 0, error, next.context);
                        for (TtsRequest r : requests)
                        {
                            _listener.onSpeakingDone(SKCloudVocalizer.this, numQueued-- > 0, error, r.context);
                        }
                    }
                }
            });

            if (next.spec.getAudioType().encoding == AudioType.Encoding.SPEEX)
            {
                _speexDecoder = new SpeexDecoderPipe();
                _speexDecoder.connectAudioSource(ttsSource);
                _player = new SpeakerPlayerSink(_speexDecoder.getAudioType());
                _player.connectAudioSource(_speexDecoder);
            } else
            {
                _player = new SpeakerPlayerSink(ttsSource.getAudioType());
                _player.connectAudioSource(ttsSource);
            }

            _player.startPlaying(new PlayerSink.Listener()
            {
                @Override
                public void onStarted(PlayerSink player)
                {
                    if (vocalizer != _vocalizer)
                        return;

                    if (_listener != null)
                        _listener.onSpeakingBegin(SKCloudVocalizer.this, next.context);
                }

                @Override
                public void onStopped(PlayerSink player)
                {
                    if (vocalizer != _vocalizer)
                        return;

                    _vocalizer = null;

                    cleanup();

                    if (_listener != null)
                    {
                        _listener.onSpeakingDone(SKCloudVocalizer.this, !_ttsQueue.isEmpty(), null, next.context);
                    }
                    startNext();
                }
            });
        }
    }
}
