package com.nuance.dragon.toolkit.speechkit;

import java.util.LinkedList;

import junit.framework.Assert;
import android.content.Context;
import android.content.res.AssetFileDescriptor;
import android.os.Handler;
import android.os.Vibrator;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.sinks.PlayerSink;
import com.nuance.dragon.toolkit.audio.sinks.SpeakerPlayerSink;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.vocalizer.Vocalizer;

/**
 * A vibration or audio prompt used during a recognition.
 */
public abstract class SKPrompt
{
    /**
     * A composite prompt, made of of multiple prompts.
     */
    public static class Composite extends SKPrompt
    {
        private final LinkedList<SKPrompt> _prompts;
        private SKPrompt _current;

        public Composite()
        {
            _prompts = new LinkedList<SKPrompt>();
        }

        /**
         * Add a prompt to this composite prompt.
         * @param prompt
         */
        public void addPrompt(SKPrompt prompt)
        {
            _prompts.add(prompt);
        }

        @Override
        protected void onStart()
        {
            if (_prompts.isEmpty())
            {
                Logger.warn(this, "Composite prompt is empty");
                done();
            } else
            {
                startNext();
            }
        }

        @Override
        protected void onCancel()
        {
            _prompts.clear();
            if (_current != null)
            {
                _current.cancel();
                _current = null;
            }
        }

        private void startNext()
        {
            _current = _prompts.remove();
            _current.start(new Listener()
            {
                @Override
                public void onFinished(SKPrompt prompt)
                {
                    _current = null;
                    if (_prompts.isEmpty())
                    {
                        done();
                    } else
                    {
                        startNext();
                    }
                }
            });
        }
    }

    /**
     * A TTS prompt. Not public, because it can only be created by {@link SpeechKit#createVocalizerPrompt}.
     */
    static class VocalizerAudio extends SKPrompt
    {
        private final Vocalizer _vocalizer;
        private final String _text;
        private PlayerSink _player;

        /**
         * @param vocalizer
         * @param text
         */
        VocalizerAudio(Vocalizer vocalizer, String text)
        {
            _vocalizer = vocalizer;
            _text = text;
        }

        @Override
        protected void onStart()
        {
            AudioSource<AudioChunk> ttsSource = _vocalizer.generateTts(_text, null, null);
            _player = new SpeakerPlayerSink(ttsSource.getAudioType());
            _player.connectAudioSource(ttsSource);
            _player.startPlaying(new PlayerSink.Listener()
                                {
                                    @Override
                                    public void onStarted(PlayerSink player)
                                    {
                                    }

                                    @Override
                                    public void onStopped(PlayerSink player)
                                    {
                                        _player = null;
                                        done();
                                    }
                                });
        }

        @Override
        protected void onCancel()
        {
            if (_player != null)
            {
                _player.disconnectAudioSource();
                _player = null;
            }
        }
    }

    /**
     * An audio prompt from an audio source, such as TTS.
     */
    public static class SourceAudio extends SKPrompt
    {
        private final AudioSource<AudioChunk> _source;
        private PlayerSink _player;

        /**
         * @param source The audio source to play
         */
        public SourceAudio(AudioSource<AudioChunk> source)
        {
            _source = source;
            _player.connectAudioSource(_source);
        }

        @Override
        protected void onStart()
        {
            _player = new SpeakerPlayerSink(_source.getAudioType());
            _player.startPlaying(new PlayerSink.Listener()
                                {
                                    @Override
                                    public void onStarted(PlayerSink player)
                                    {
                                    }

                                    @Override
                                    public void onStopped(PlayerSink player)
                                    {
                                        _player = null;
                                        done();
                                    }
                                });
        }

        @Override
        protected void onCancel()
        {
            if (_player != null)
            {
                _player.disconnectAudioSource();
                _player = null;
            }
        }
    }

    /**
     * An audio prompt from a media file, such as a beep.
     */
    public static class MediaAudio extends SKPrompt
    {
        /**
         * @param audioFile The audio file to play
         */
        public MediaAudio(AssetFileDescriptor audioFile)
        {
        }

        @Override
        protected void onStart()
        {
            // TODO Auto-generated method stub
            Logger.error(this, "Audio prompts not implemented");
            done();
        }

        @Override
        protected void onCancel()
        {
            // TODO Auto-generated method stub
        }
    }

    /**
     * A vibration prompt.
     */
    public static class Vibrate extends SKPrompt
    {
        private Vibrator _vibrateService;
        private Runnable _doneRunnable;
        private Handler _doneHandler;
        private final int _duration;

        /**
         * @param context
         * @param duration The duration of the vibration in ms.
         */
        public Vibrate(Context context, int duration)
        {
            _vibrateService = (Vibrator)context.getSystemService(Context.VIBRATOR_SERVICE);
            _duration = duration;
        }

        @Override
        protected void onStart()
        {
            if (_vibrateService == null)
            {
                Logger.error(this, "Unable to get vibrator service");
                done();
                return;
            }

            _vibrateService.vibrate(_duration);

            _doneHandler = new Handler();
            _doneRunnable = new Runnable()
            {
                @Override
                public void run()
                {
                    _doneHandler = null;
                    _doneRunnable = null;
                    done();
                }
            };
            _doneHandler.postDelayed(_doneRunnable, _duration);
        }

        @Override
        protected void onCancel()
        {
            if (_vibrateService != null)
            {
                _vibrateService.cancel();
                _vibrateService = null;
            }

            if (_doneHandler != null)
            {
                _doneHandler.removeCallbacks(_doneRunnable);
                _doneHandler = null;
                _doneRunnable = null;
            }
        }
    }

    /**
     * Notifies when the prompt finishes, so the recognitino can continue.
     */
    interface Listener
    {
        /**
         * The prompt finished.
         * @param prompt
         */
        void onFinished(SKPrompt prompt);
    }

    private Listener _listener;

    SKPrompt()
    {
    }

    /**
     * @param listener May be null
     */
    final void start(Listener listener)
    {
        Assert.assertNull(_listener);
        _listener = listener;

        onStart();
    }

    /**
     * Cancel the prompt early. The listener will not be called.
     */
    final void cancel()
    {
        _listener = null;

        onCancel();
    }

    /**
     * Override to start the prompt. Make sure to call {@link #done} when the
     * prompt finishes.
     */
    protected abstract void onStart();

    /**
     * Override to stop the prompt early if possible.
     */
    protected abstract void onCancel();

    /**
     * Call this when the prompt finishes.
     */
    protected final void done()
    {
        if (_listener != null)
        {
            _listener.onFinished(this);
            _listener = null;
        }
    }
}
