package com.nuance.dragon.toolkit.speechkit;

import java.util.LinkedList;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.sinks.PlayerSink;
import com.nuance.dragon.toolkit.audio.sinks.SpeakerPlayerSink;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.vocalizer.Vocalizer;
/**
 * A simple vocalizer that handles queuing and playing multiple embedded
 * TTS requests for you.
 */
public class SKVocalizer
{
    /**
     * Listen for TTS progress.
     */
    public interface Listener
    {
        /**
         * Speaking has begun for a phrase.
         * @param vocalizer
         * @param context The context that was passed to {@link SKVocalizer#speakString}.
         */
        void onSpeakingBegin(SKVocalizer vocalizer,
                             Object context);

        /**
         * Speaking has finished for a phrase.
         * @param vocalizer
         * @param moreToPlay true if there is more queued to play.
         * @param error true if there was an error.
         * @param context The context that was passed to {@link SKVocalizer#speakString}.
         */
        void onSpeakingDone(SKVocalizer vocalizer,
                            boolean moreToPlay,
                            boolean error,
                            Object context);
    }

    private static final class TtsRequest
    {
        final String text;
        final Object context;
        public TtsRequest(String text, Object context)
        {
            this.text = text;
            this.context = context;
        }
    }

    private final Vocalizer _vocalizer;
    private final Listener _listener;
    private final LinkedList<TtsRequest> _ttsQueue;
    private PlayerSink _player;

    SKVocalizer(Vocalizer vocalizer,
                Listener listener)
    {
        _vocalizer = vocalizer;
        _listener = listener;
        _ttsQueue = new LinkedList<TtsRequest>();
    }

    /**
     * Speak some text.
     * @param text The text to speak.
     * @param context The context to pass back in the {@link Listener} methods.
     */
    public void speakString(String text, Object context)
    {
        Checker.checkArgForNull("text", text);

        _ttsQueue.add(new TtsRequest(text, context));
        startNext();
    }

    /**
     * Cancel all active and queued TTS requests.
     */
    public void cancel()
    {
        cleanup();
    }

    private void cleanup()
    {
        if (_vocalizer != null)
        {
            _vocalizer.cancelTts();
        }

        if (_player != null)
        {
            _player.disconnectAudioSource();
            _player.stopPlaying();
            _player = null;
        }

        _ttsQueue.clear();
    }

    private void startNext()
    {
        if (!_ttsQueue.isEmpty())
        {
            final Vocalizer vocalizer = _vocalizer;
            final TtsRequest next = _ttsQueue.remove();

            AudioSource<AudioChunk> ttsSource = _vocalizer.generateTts(next.text, new Vocalizer.TtsListener()
            {

                @Override
                public void ttsGenerationStarted(String text, Object context, Vocalizer vocalizer)
                {
                    if (_vocalizer != vocalizer)
                        return;
                }

                @Override
                public void ttsGenerationFinished(String text, Object context, Vocalizer vocalizer, boolean success)
                {
                    if (_vocalizer != vocalizer)
                        return;

                    if(!success) {
                        cleanup();

                        if (_listener != null)
                        {
                            _listener.onSpeakingDone(SKVocalizer.this, false, true, next.context);
                            for (TtsRequest r : _ttsQueue)
                            {
                                _listener.onSpeakingDone(SKVocalizer.this, false, true, r.context);
                            }
                        }
                        _ttsQueue.clear();
                    }
                }

                @Override
                public void ttsStreamingStarted(String text, Object context, Vocalizer vocalizer)
                {
                    if (_vocalizer != vocalizer)
                        return;
                }

                @Override
                public void ttsStreamingFinished(String text, Object context, Vocalizer vocalizer)
                {
                    if (_vocalizer != vocalizer)
                        return;
                }
            }, next.context);

            _player = new SpeakerPlayerSink(ttsSource.getAudioType());
            _player.connectAudioSource(ttsSource);
            _player.startPlaying(new PlayerSink.Listener()
            {
                @Override
                public void onStarted(PlayerSink player)
                {
                    if (vocalizer != _vocalizer)
                        return;

                    if (_listener != null)
                        _listener.onSpeakingBegin(SKVocalizer.this, next.context);
                }

                @Override
                public void onStopped(PlayerSink player)
                {
                    if (vocalizer != _vocalizer)
                        return;

                    cleanup();

                    if (_listener != null)
                    {
                        _listener.onSpeakingDone(SKVocalizer.this, !_ttsQueue.isEmpty(), false, next.context);
                    }
                    startNext();
                }
            });
        }
    }
}
