package com.nuance.dragon.toolkit.speechkit;

import com.nuance.dragon.toolkit.cloudservices.CloudConfig;
import com.nuance.dragon.toolkit.cloudservices.CloudServices;
import com.nuance.dragon.toolkit.cloudservices.recognizer.RecogSpec;
import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.NMTContext;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.vocalizer.Vocalizer;
import com.nuance.dragon.toolkit.vocalizer.Vocalizer.Engine;
import com.nuance.dragon.toolkit.vocalizer.VocalizerConfig;

/**
 * Collection of modules that support very simple recognition / TTS scenarios.
 */
public final class SpeechKit
{
    private boolean _released;
    private final CloudServices _cloud;
    private final Vocalizer _vocalizer;
    private final CloudConfig _config;

    /**
     * Create a SpeechKit. Make sure to call {@link #release} when you're done.
     * @param cloudConfig The cloudservices config, or null for no cloud support
     * @param vocalizerConfig The vocalizer config, or null for no vocalizer support
     * @param context
     */
    public SpeechKit(CloudConfig cloudConfig, VocalizerConfig vocalizerConfig, NMTContext context)
    {
        Checker.checkArgForNull("context", context);
        Checker.checkArgsForNull("cloudConfig and vocalizerConfig", cloudConfig, vocalizerConfig);

        context = context.getAppContext();
        if (cloudConfig != null) {
            _config = cloudConfig;
            _cloud = CloudServices.createCloudServices(context, cloudConfig);
        }
        else {
            _config = null;
            _cloud = null;
        }

        if (vocalizerConfig != null) {
            _vocalizer = Vocalizer.createVocalizer(Engine.VOCALIZER_EXPRESSIVE, Factory.createFileManager(context, ".jpg", "vexpressive", "vexpressive"));
            _vocalizer.load(vocalizerConfig);
        }
        else {
            _vocalizer = null;
        }
    }

    /**
     * Create a cloud recognizer. All audio recording and endpointing will be
     * handled internally using a microphone recorder.
     * @param recogSpec
     * @param endPointerType
     * @param listener
     * @return The cloud recognizer.
     */
    public SKCloudRecognizer createCloudRecognizer(RecogSpec recogSpec,
                                                   int endPointerType,
                                                   SKCloudRecognizer.Listener listener)
    {
        Checker.checkState(this, !_released);
        Checker.checkState(this, _cloud != null);
        Checker.checkArgForNull("recogSpec", recogSpec);
        Checker.checkArgForNull("listener", listener);

        SKCloudRecognizer r = new SKCloudRecognizer(_cloud, recogSpec, endPointerType, null, false, listener);
        return r;
    }

    /**
     * Create a cloud recognizer. All audio recording and endpointing will be
     * handled internally using a FILE recorder. This is primarily useful for
     * testing. Use {@link #createCloudRecognizer(RecogSpec, int, SKCloudRecognizer.Listener)}
     * for standard recognition using the device microphone.
     * @param recogSpec
     * @param endPointerType
     * @param fileName The file to read the audio from
     * @param fileIsLittleEndian True if the file audio is stored in little-endian
     * format.
     * @param listener
     * @return The cloud recognizer.
     */
    public SKCloudRecognizer createCloudRecognizer(RecogSpec recogSpec,
                                                   int endPointerType,
                                                   String fileName,
                                                   boolean fileIsLittleEndian,
                                                   SKCloudRecognizer.Listener listener)
    {
        Checker.checkState(this, !_released);
        Checker.checkState(this, _cloud != null);
        Checker.checkArgForNull("recogSpec", recogSpec);
        Checker.checkArgForNull("fileName", fileName);
        Checker.checkArgForNull("listener", listener);

        SKCloudRecognizer r = new SKCloudRecognizer(_cloud, recogSpec, endPointerType, fileName, fileIsLittleEndian, listener);
        return r;
    }

    /**
     * Create a cloud vocalizer. All audio playback will be handled internally.
     * @param listener May be null
     * @return The cloud vocalizer.
     */
    public SKCloudVocalizer createCloudVocalizer(SKCloudVocalizer.Listener listener)
    {
        Checker.checkState(this, !_released);
        Checker.checkState(this, _cloud != null);

        SKCloudVocalizer v = new SKCloudVocalizer(_cloud, _config, listener);
        return v;
    }

    /**
     * Create an embedded vocalizer. All audio playback will be handled internally.
     * @param listener May be null
     * @return The embedded vocalizer.
     */
    public SKVocalizer createVocalizer(SKVocalizer.Listener listener)
    {
        Checker.checkState(this, !_released);
        Checker.checkState(this, _vocalizer != null);

        SKVocalizer v = new SKVocalizer(_vocalizer, listener);
        return v;
    }

    /**
     * Create a TTS prompt to be used with a {@link SKCloudRecognizer}.
     * @param text
     * @return The prompt.
     */
    public SKPrompt createVocalizerPrompt(String text)
    {
        Checker.checkState(this, !_released);
        Checker.checkState(this, _vocalizer != null);
        Checker.checkArgForNull("text", text);

        return new SKPrompt.VocalizerAudio(_vocalizer, text);
    }

    /**
     * Get the last established session ID
     * @return The session ID, or null if no session has been established yet
     */
    public String getSessionID()
    {
        String ret = null;

        if (_cloud != null)
        {
            return _cloud.getSessionID();
        }

        return ret;
    }

    /**
     * Release the resources used by the SpeechKit.
     */
    public void release()
    {
        Checker.checkState(this, !_released);

        if (_cloud != null)
            _cloud.release();

        if (_vocalizer != null)
            _vocalizer.release();

        _released = true;
    }
}
