package com.nuance.dragon.toolkit.sse;

import java.util.Arrays;

import junit.framework.Assert;
import android.content.Context;

import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.util.internal.NativeUtils;

class NativeSseImpl implements NativeSse
{
    static boolean LOAD_LIBRARY_SUCCESS = true;

    static
    {
        boolean armV7         = NativeUtils.isArmV7();
        boolean neonSupported = NativeUtils.isNeonSupported();
        boolean x86           = NativeUtils.isX86();
        Logger.debug(NativeSseImpl.class, "armV7:" + armV7 + ", neonSupported:" + neonSupported + ", x86:" + x86);

        try {
            if (armV7 && neonSupported)
            {
                System.loadLibrary("dmt_sse_neon");
            }
            else {

                System.loadLibrary("dmt_sse");
            }
        } catch (Exception e) {
            Logger.error(NativeSseImpl.class, "Failed to load native library.", e);
        }
    }

    private native static long sseCreate();
    private native static void sseDestroy(long sse);
    private native static int sseInitialize(long sse, int sampleRate, int micNum,
                                            boolean useEchoCancel, boolean useNoiseSuppression, boolean useBeamForming, boolean usePic, byte[] rawConfigData);
    private native static int sseProcessWithRef(long sse, short[] microphoneIn, short[] referenceIn, short[] microphoneOut);
    private native static int sseProcessWithoutRef(long sse, short[] microphoneIn, short[] microphoneOut);
    private native static int sseProcessWithRecv(long sse, short[] microphoneIn, short[] receiveIn, short[] microphoneOut, short[] receiveOut);
    private native static int sseProcessPic(long sse, short[][] microphoneIns, short[] microphoneOut);
    private native static int sseGetVoconConfigurationSize(long sse);
    private native static int sseGetVoconConfiguration(long sse, byte[] configuration);
    private native static String sseGetVersion();

    private long _sse;

    public NativeSseImpl(Context context)
    {
    }

    @Override
    public boolean createSse()
    {
        if (_sse != 0)
            sseDestroy(_sse);

        _sse = sseCreate();
        return _sse != 0;
    }

    @Override
    public void destroySse()
    {
        if (_sse != 0)
        {
            sseDestroy(_sse);
            _sse = 0;
        }
    }

    @Override
    public boolean initialize(int sampleRate,
                              int micNum,
                              boolean useEchoCancel,
                              boolean useNoiseSuppression,
                              boolean useBeamForming,
                              boolean usePic,
                              byte[] rawConfigData)
    {
        if (_sse != 0)
        {
            int ret = sseInitialize(_sse, sampleRate, micNum, useEchoCancel, useNoiseSuppression, useBeamForming, usePic, rawConfigData);
            return ret == 0;
        }
        return false;
    }

    @Override
    public short[][] processRecv(short[] microphoneIn, short[] receiveIn)
    {
        if (receiveIn != null)
            Assert.assertEquals(microphoneIn.length, receiveIn.length);

        if (_sse != 0)
        {
            short[] microphoneOut = new short[microphoneIn.length];
            short[] receiveOut    = new short[microphoneIn.length];
            int ret = sseProcessWithRecv(_sse, microphoneIn, receiveIn, microphoneOut, receiveOut);
            if (ret == 0)
                return new short[][] {microphoneOut, receiveOut};
        }
        return null;
    }

    @Override
    public short[] processRef(short[] microphoneIn, short[] referenceIn)
    {
        // TODO: note that this is modding the length (SAMPLES) with the frame size
        // (MS). A better check would be samples to samples or ms to ms, but this is
        // easier since we don't know the sample rate here.
        // NPM: This doesn't catch all cases, but it's better than nothing
        Assert.assertTrue(microphoneIn.length % Sse.getRequiredFrameSize() == 0);
        if (referenceIn != null)
            Assert.assertEquals(microphoneIn.length, referenceIn.length);

        if (_sse != 0)
        {
            short[] microphoneOut = new short[microphoneIn.length];
            int ret;
            if (referenceIn != null)
                ret = sseProcessWithRef(_sse, microphoneIn, referenceIn, microphoneOut);
            else
                ret = sseProcessWithoutRef(_sse, microphoneIn, microphoneOut);

            if (ret == 0)
                return microphoneOut;
        }
        return null;
    }

    @Override
    public short[] processPic(short[][] microphoneIns)
    {
        // TODO: note that this is modding the length (SAMPLES) with the frame size
        // (MS). A better check would be samples to samples or ms to ms, but this is
        // easier since we don't know the sample rate here.
        // NPM: This doesn't catch all cases, but it's better than nothing
        Assert.assertTrue(microphoneIns[0].length % Sse.getRequiredFrameSize() == 0);

        if (_sse != 0)
        {
            short[] microphoneOut = new short[microphoneIns[0].length];
            int ret = sseProcessPic(_sse, microphoneIns, microphoneOut);

            if (ret == 0)
                return microphoneOut;
        }
        return null;
    }

    @Override
    public byte[] getVoconConfiguration()
    {
        if (_sse != 0)
        {
            int configLen = sseGetVoconConfigurationSize(_sse);
            if (configLen > 0)
            {
                byte[] config = new byte[configLen];
                Arrays.fill(config, (byte) 0x00);
                int ret = sseGetVoconConfiguration(_sse, config);
                if (ret == 0)
                    return config;
            }
        }
        return null;
    }

    @Override
    public String getVersion()
    {
        return sseGetVersion();
    }
}