package com.nuance.dragon.toolkit.sse;

import java.io.FileInputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.LinkedList;

import junit.framework.Assert;
import android.content.Context;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.MultiChannelAudioChunk;
import com.nuance.dragon.toolkit.audio.pipes.SseInitParam;
import com.nuance.dragon.toolkit.audio.pipes.SseTuningType;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * Performs echo cancellation using SSE
 */
@SuppressWarnings("deprecation")
public class Sse
{
    private final NativeSse _nativeSse;
    private boolean _useEchoCancel = false;
    private boolean _useNoiseSuppression = false;
    private boolean _useBeamForming = false;
    private byte[] _voconConfiguration = null;

    // The following list and counter is only used with 22050Hz audio
    private final LinkedList<short[]> _storedMicAudio;
    private final LinkedList<short[]> _storedRefAudio;
    private int _storedAudioLength;
    private static final int FRAME_SHIFT_22K = 220; // This value is parachuted from the SSE implementation

    public Sse(Context context)
    {
        _nativeSse = new NativeSseImpl(context);

        _storedMicAudio = new LinkedList<short[]>();
        _storedRefAudio = new LinkedList<short[]>();
        _storedAudioLength = 0;
    }

    /**
     * @return The frame size (in ms) that each chunk MUST have
     */
    public static int getRequiredFrameSize()
    {
        return 20;
    }

    /**
     * @deprecated Use {@link SseInitParam} instead
     * Sets whether to use echo cancellation or not
     * @param useEchoCancel
     */
    @Deprecated
    public void setUseEchoCancel(boolean useEchoCancel) {
        _useEchoCancel = useEchoCancel;
    }

    /**
     * @deprecated Use {@link SseInitParam} instead
     * Sets whether to use noise suppression or not
     * @param useNoiseSuppression
     */
    @Deprecated
    public void setUseNoiseSuppression(boolean useNoiseSuppression) {
        _useNoiseSuppression = useNoiseSuppression;
    }

    /**
     * @deprecated Use {@link SseInitParam} instead
     * Sets whether to use beam forming or not
     * @param useBeamForming
     */
    @Deprecated
    public void setUseBeamForming(boolean useBeamForming) {
        _useBeamForming = useBeamForming;
    }

    @Deprecated
    public boolean initialize(AudioType audioType, FileManager configDataFileManager, String configDataFileName, SseTuningType tuningType, boolean useVocon)
    {
        byte[] rawConfigData = null;
        if (configDataFileManager != null && configDataFileName != null)
        {
            FileInputStream fis = configDataFileManager.openFileForReading(configDataFileName);
            if (fis != null)
            {
                int size = configDataFileManager.getSize(configDataFileName);
                if (size > 0)
                {
                    rawConfigData = new byte[size];
                    try
                    {
                        int numOfBytesRead = fis.read(rawConfigData);
                        fis.close();
                        if (numOfBytesRead != size)
                            throw new IOException(".bsd file not read fully!");
                    } catch (IOException e)
                    {
                        Logger.error(this, ".bsd file:" + configDataFileName + " could not be read. e:" + e);
                        rawConfigData = null;
                    }
                }
            }
        }

        switch (tuningType) {
        case NONE :
            _useEchoCancel = false;
            _useNoiseSuppression = false;
            _useBeamForming = false;
            break;
        case ECHO_CANCEL:
            _useEchoCancel = true;
            break;
        case NOISE_SUPPRESS:
            _useNoiseSuppression = true;
            break;
        case BEAM_FORMING:
            _useBeamForming = true;
            break;
        case ECHO_AND_NOISE:
            _useEchoCancel = true;
            _useNoiseSuppression = true;
            break;
        case ECHO_AND_BEAM:
            _useEchoCancel = true;
            _useBeamForming = true;
            break;
        case NOISE_AND_BEAM:
            _useNoiseSuppression = true;
            _useBeamForming = true;
            break;
        case ECHO_NOISE_BEAM:
            _useEchoCancel = true;
            _useNoiseSuppression = true;
            _useBeamForming = true;
            break;
        default:
            _useEchoCancel = true;
            _useNoiseSuppression = true;
        }

        boolean success = _nativeSse.createSse() &&
                          _nativeSse.initialize(audioType.frequency, 1, _useEchoCancel, _useNoiseSuppression, _useBeamForming, false, rawConfigData);

        if (success && useVocon)
            _voconConfiguration = _nativeSse.getVoconConfiguration();
        return success;
    }

    /**
     * Initialize SSE engine
     * @param audioType It includes the sample rate to initialize SSE
     * @param initParam The parameters have to be set prior to SSE initialization
     * @param configDataFileManager Optional the file manager to find SSE initial configuration file
     * @param configDataFileName Optional the SSE initial configuration file. It will override any parameters
     *                           set in initParam if applicable
     * @return true successful, otherwise false
     */
    public boolean initialize(AudioType audioType, SseInitParam initParam, FileManager configDataFileManager, String configDataFileName)
    {
        Checker.checkArgForNull("initParam", initParam);

        byte[] rawConfigData = null;
        if (configDataFileManager != null && configDataFileName != null)
        {
            FileInputStream fis = configDataFileManager.openFileForReading(configDataFileName);
            if (fis != null)
            {
                int size = configDataFileManager.getSize(configDataFileName);
                if (size > 0)
                {
                    rawConfigData = new byte[size];
                    try
                    {
                        int numOfBytesRead = fis.read(rawConfigData);
                        fis.close();
                        if (numOfBytesRead != size)
                            throw new IOException(".bsd file not read fully!");
                    } catch (IOException e)
                    {
                        Logger.error(this, ".bsd file:" + configDataFileName + " could not be read. e:" + e);
                        rawConfigData = null;
                    }
                }
            }
        }

        boolean success = _nativeSse.createSse() &&
                          _nativeSse.initialize(audioType.frequency, initParam.getMicNum(), initParam.isEchoCancelOn(), initParam.isNrOn(),
                                  initParam.isBeamFormingOn(), initParam.isPicOn(), rawConfigData);

        if (success && initParam.isForVoconASR())
            _voconConfiguration = _nativeSse.getVoconConfiguration();
        return success;
    }

    public void release()
    {
        _nativeSse.destroySse();
    }

    /**
     * @param input Since input is a regular audioChunk one channel is assumed.
     * Hence only performing noise suppression.
     * @return The processed audio, or null if an error occurred
     */
    public AudioChunk process(AudioChunk input)
    {
        short[] mic;
        short[] ref = null;

        mic = new short[input.audioShorts.length];

        System.arraycopy(input.audioShorts, 0, mic, 0, mic.length);

        short [] output = _nativeSse.processRef(mic, ref);

        if (output != null)
        {
            return new AudioChunk(input.audioType, output, input.audioTimestamp);
        }
        return null;
    }


    /**
     * @param input Channel 1 must be the microphone, channel 2 must be the
     * receive
     * @return The processed audio, or null if an error occurred
     */
    public AudioChunk process(MultiChannelAudioChunk input)
    {
        Assert.assertTrue(input.channelCount == 1 || input.channelCount == 2);

        // NPM: This is a hack for 22050Hz audio. We receive the audio
        // as multiples of 441samples. 441 samples corresponds to 20ms.
        // However SSE expects 22050Hz audio to be a multiple of
        // 220 samples.
        if(input.audioType == AudioType.PCM_22k)
        {
            return process22050(input);
        }

        short[] mic;
        short[] ref;

        if (input.channelCount == 1)
        {
            mic = new short[input.audioShorts.length];
            ref = null;
        } else
        {
            mic = new short[input.audioShorts.length / 2];
            ref = new short[input.audioShorts.length / 2];
            System.arraycopy(input.audioShorts, mic.length, ref, 0, ref.length);
        }
        System.arraycopy(input.audioShorts, 0, mic, 0, mic.length);

        short [] output = _nativeSse.processRef(mic, ref);
        //short[] output = mic;
        if (output != null)
        {
            return new AudioChunk(input.audioType, output, input.audioTimestamp);
        }
        return null;
    }

    // NPM: This is a hack for 22050Hz audio. We receive the audio
    // as multiples of 441samples. 441 samples corresponds to 20ms.
    // However SSE expects 22050Hz audio to be a multiple of
    // 220 samples.
    private AudioChunk process22050(MultiChannelAudioChunk input)
    {

        short[] mic;
        short[] ref;

        // If 1 channel set reference to silence, this is the same as
        // using 1 channel. Setting the silence is necessary to make sure
        // that the amount of buffered MicAudio is the same as the
        // amount of buffered reference audio
        if (input.channelCount == 1)
        {
            mic = new short[input.audioShorts.length];
            ref = new short[input.audioShorts.length];
            // TODO: is this necessary?
            Arrays.fill(ref, (short) 0);

            Logger.warn(this, "Single channel obtained, reference set to silence");
        } else
        {
            mic = new short[input.audioShorts.length / 2];
            ref = new short[input.audioShorts.length / 2];
            System.arraycopy(input.audioShorts, mic.length, ref, 0, ref.length);
        }
        System.arraycopy(input.audioShorts, 0, mic, 0, mic.length);

        _storedMicAudio.add(mic);
        _storedRefAudio.add(ref);
        _storedAudioLength += mic.length;

        int audioToSendToSSE = _storedAudioLength - (_storedAudioLength % FRAME_SHIFT_22K);

        if( audioToSendToSSE > 0)
        {
            short[] micResized = removeAudio(_storedMicAudio, audioToSendToSSE);
            short[] refResized = removeAudio(_storedRefAudio, audioToSendToSSE);
            _storedAudioLength -= audioToSendToSSE;


            short [] output = _nativeSse.processRef(micResized, refResized);

            if (output != null)
            {
                // unknown time-stamp...
                return new AudioChunk(input.audioType, output);
            }
        }

        return null;
    }

    /**
     * To process passenger interference cancellation
     * @param input Channel 1 must be the desire microphone, channel 2 must be "noise reference"
     * @return The processed audio, or null if an error occurred
     */
    public AudioChunk processPic(MultiChannelAudioChunk input)
    {
        Assert.assertTrue(input.channelCount == 1 || input.channelCount == 2);

        // NPM: This is a hack for 22050Hz audio. We receive the audio
        // as multiples of 441samples. 441 samples corresponds to 20ms.
        // However SSE expects 22050Hz audio to be a multiple of
        // 220 samples.
        if(input.audioType == AudioType.PCM_22k)
        {
            return processPic22050(input);
        }

        short[][] microphoneIns = new short[2][];

        if (input.channelCount == 1)
        {
            microphoneIns[0] = new short[input.audioShorts.length];
            microphoneIns[1] = new short[input.audioShorts.length];
            // TODO: is this necessary?
            Arrays.fill(microphoneIns[1], (short) 0);

            Logger.warn(this, "Single channel obtained, reference set to silence");
        } else
        {
            microphoneIns = new short[2][];
            microphoneIns[0] = new short[input.audioShorts.length / 2];
            microphoneIns[1] = new short[input.audioShorts.length / 2];
            System.arraycopy(input.audioShorts, microphoneIns[0].length, microphoneIns[1], 0, microphoneIns[1].length);
        }
        System.arraycopy(input.audioShorts, 0, microphoneIns[0], 0, microphoneIns[0].length);

        short [] output = _nativeSse.processPic(microphoneIns);
        //short[] output = mic;
        if (output != null)
        {
            return new AudioChunk(input.audioType, output, input.audioTimestamp);
        }
        return null;
    }

    // NPM: This is a hack for 22050Hz audio. We receive the audio
    // as multiples of 441samples. 441 samples corresponds to 20ms.
    // However SSE expects 22050Hz audio to be a multiple of
    // 220 samples.
    private AudioChunk processPic22050(MultiChannelAudioChunk input)
    {

        short[][] microphoneIns = new short[2][];

        // If 1 channel set reference to silence, this is the same as
        // using 1 channel. Setting the silence is necessary to make sure
        // that the amount of buffered MicAudio is the same as the
        // amount of buffered reference audio
        if (input.channelCount == 1)
        {
            microphoneIns[0] = new short[input.audioShorts.length];
            microphoneIns[1] = new short[input.audioShorts.length];
            // TODO: is this necessary?
            Arrays.fill(microphoneIns[1], (short) 0);

            Logger.warn(this, "Single channel obtained, reference set to silence");
        } else
        {
            microphoneIns[0] = new short[input.audioShorts.length / 2];
            microphoneIns[1] = new short[input.audioShorts.length / 2];
            System.arraycopy(input.audioShorts, microphoneIns[0].length, microphoneIns[1], 0, microphoneIns[1].length);
        }
        System.arraycopy(input.audioShorts, 0, microphoneIns[0], 0, microphoneIns[0].length);

        _storedMicAudio.add(microphoneIns[0]);
        _storedRefAudio.add(microphoneIns[1]);
        _storedAudioLength += microphoneIns[0].length;

        int audioToSendToSSE = _storedAudioLength - (_storedAudioLength % FRAME_SHIFT_22K);

        if( audioToSendToSSE > 0)
        {
            short[][] micResized = new short[2][];
            micResized[0] = removeAudio(_storedMicAudio, audioToSendToSSE);
            micResized[1] = removeAudio(_storedRefAudio, audioToSendToSSE);
            _storedAudioLength -= audioToSendToSSE;


            short [] output = _nativeSse.processPic(micResized);

            if (output != null)
            {
                // unknown time-stamp...
                return new AudioChunk(input.audioType, output);
            }
        }

        return null;
    }

    /**
     * Get VoCon configuration data (only call when doing ASR / WUW / EP using VoCon)
     * @return blob of data only meaningful to VoCon, or null if not applicable.
     */
    public byte[] getVoconConfiguration()
    {
        return _voconConfiguration;
    }

    /**
     * Get embedded SSE library version
     * @return SSE library version string
     */
    public String getVersion()
    {
        return _nativeSse.getVersion();
    }

    /**
     * Remove a frame, and optionally return the removed audio
     * @param chunks
     * @param audioType The audio type of the audio
     * @param numSamples
     * @param newTimestamp The timestamp for the new frame
     * @return The audio
     */
    private static short[] removeAudio(LinkedList<short[]> chunks, final int numSamples)
    {
        short [] newAudio = new short[numSamples];
        int offset = 0;
        while (!chunks.isEmpty() && offset < numSamples)
        {
            int reqLen = numSamples - offset;

            short [] audio = chunks.remove();

            if (audio.length <= reqLen)
            {
                System.arraycopy(audio, 0, newAudio, offset, audio.length);
                offset += audio.length;
            }
            else
            {
                System.arraycopy(audio, 0, newAudio, offset, reqLen);
                offset = numSamples;

                short [] leftover = new short[audio.length - reqLen];
                System.arraycopy(audio, reqLen, leftover, 0, audio.length - reqLen);

                chunks.addFirst(leftover);
            }
        }
        // if offset is less than numSamples, it's OK, leave trailing zeros

        return newAudio;
    }
}
