package com.nuance.dragon.toolkit.util.internal;

import java.util.ArrayList;

import android.os.Handler;
import android.os.Message;

import com.nuance.dragon.toolkit.oem.api.Logger;

/**
 * Wrapper around {@link android.os.Handler} that queues messages until
 * {@link #attachToCurrentThread} is called.
 * <p>
 * Note: {@link Handler#removeCallbacks} doesn't work!
 */
public class AttachableHandler extends Handler
{
    private class QueuedMessage
    {
        final Message _msg;
        final long _uptimeMillis;
        QueuedMessage(Message msg, long uptimeMillis)
        {
            _msg = msg;
            _uptimeMillis = uptimeMillis;
        }
    }

    private Handler _realHandler;
    private final ArrayList<QueuedMessage> _queuedMessages;

    public AttachableHandler()
    {
        _queuedMessages = new ArrayList<QueuedMessage>();
    }

    /**
     * Attach this handler to the current thread.
     */
    public synchronized void attachToCurrentThread()
    {
        _realHandler = new Handler();
        if (_queuedMessages.size() > 0)
        {
            Logger.debug(this, "Attached handler to current thread, sending " + _queuedMessages.size() + " queued messages");
            for (QueuedMessage m : _queuedMessages)
            {
                _realHandler.sendMessageAtTime(m._msg, m._uptimeMillis);
            }
            _queuedMessages.clear();
        }
    }

    @Override
    public synchronized boolean sendMessageAtTime(Message msg, long uptimeMillis)
    {
        if (_realHandler == null)
        {
            _queuedMessages.add(new QueuedMessage(msg, uptimeMillis));
            return true;
        } else
        {
            if (_realHandler.getLooper().getThread().isAlive())
            {
                return _realHandler.sendMessageAtTime(msg, uptimeMillis);
            } else
            {
                Logger.debug(this, "this msg [" + msg + "] will be ignore since the Looper Thread has been terminated already!");
                return true;
            }
        }
    }
}
