package com.nuance.dragon.toolkit.util.internal;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.NoSuchElementException;

/**
 * A synchronized list that returns iterators which block as long as there is more
 * to be added to the underlying list.
 * @param <T>
 */
public class BlockingIteratorList<T extends Object> extends AbstractList<T>
{
    private class BlockingIterator implements ListIterator<T>
    {
        private int _nextIndex;

        BlockingIterator(int startIndex)
        {
            _nextIndex = startIndex;
        }

        @Override
        public boolean hasNext()
        {
            synchronized (_sync)
            {
                return !_finished || _nextIndex < _list.size();
            }
        }

        @Override
        public T next()
        {
            synchronized (_sync)
            {
                while (!_finished)
                {
                    if (_nextIndex < _list.size())
                        return _list.get(_nextIndex++);

                    try {
                        _sync.wait();
                    }
                    catch (InterruptedException e) {
                    }
                }
                return _list.get(_nextIndex++);
            }
        }

        @Override
        public boolean hasPrevious()
        {
            return _nextIndex > 0;
        }

        @Override
        public int nextIndex()
        {
            return _nextIndex;
        }

        @Override
        public T previous()
        {
            if (_nextIndex <= 0) throw new NoSuchElementException();

            synchronized (_sync)
            {
                return _list.get(--_nextIndex);
            }
        }

        @Override
        public int previousIndex()
        {
            if (_nextIndex <= 0) throw new NoSuchElementException();

            return _nextIndex-1;
        }

        @Override
        public void remove()
        {
            throw new UnsupportedOperationException("Blocking iterators don't support remove");
        }

        @Override
        public void add(T object)
        {
            throw new UnsupportedOperationException("Blocking iterators don't support add");
        }

        @Override
        public void set(T object)
        {
            throw new UnsupportedOperationException("Blocking iterators don't support set");
        }
    }

    private final ArrayList<T> _list;
    private final Object _sync;
    private final int _fullSize;
    private boolean _finished;

    /**
     * @param fullSize The expected size, or -1 if it is unknown.
     */
    public BlockingIteratorList(int fullSize)
    {
        _list = new ArrayList<T>();
        _sync = new Object();
        _fullSize = fullSize;
    }

    @Override
    public Iterator<T> iterator()
    {
        return new BlockingIterator(0);
    }

    @Override
    public ListIterator<T> listIterator()
    {
        return new BlockingIterator(0);
    }

    @Override
    public ListIterator<T> listIterator(int location)
    {
        synchronized (_sync)
        {
            if (location < 0 || location >= _list.size()) throw new IndexOutOfBoundsException();
        }
        return new BlockingIterator(location);
    }

    /**
     * Add an item, and indicate whether it is the last one. You MUST finish the list eventually,
     * or other threads may block indefinitely waiting for it.
     * @param item The item
     * @param last Whether or not this is the last one
     */
    public void add(T item, boolean last)
    {
        synchronized (_sync)
        {
            _list.add(item);
            _finished = last;

            _sync.notifyAll();
        }
    }

    @Override
    public T get(int location)
    {
        synchronized (_sync)
        {
            return _list.get(location);
        }
    }

    /**
     * The number of items that are either in the list, or will be in the list eventually.
     * <b>MAY RETURN -1 IF THE EVENTUAL SIZE IS UNKNOWN.</b>
     * @see java.util.AbstractCollection#size()
     */
    @Override
    public int size()
    {
        synchronized (_sync)
        {
            if (_finished)
            {
                return _list.size();
            }
            return _fullSize;
        }
    }
}
