package com.nuance.dragon.toolkit.util.internal;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

import android.content.ContentValues;
import android.content.Context;
import android.database.Cursor;
import android.database.sqlite.SQLiteDatabase;
import android.util.Log;

/**
 * A list implementation that is backed by the persistent device's storage.
 * Slower than a regular list (each operation is atomic) but possibly more memory efficient.
 *
 * <p><b>Since there are HW R/W operations going on in this implementation,
 * make sure to call all methods on a worker thread.</p></b>
 *
 * @param <T> Object type, must be serializable.
 */
public class PersistentStorageList<T extends Object> implements List<T>
{
    private final static String TAG = "PersistentStorageList";
    private final static String DATABASE_TABLE = "list";

    private final static String KEY_ID   = "_id";
    private final static String KEY_HASH = "_hash";
    private final static String KEY_BLOB = "_blob";

    private static final String DATABASE_CREATE =
            "CREATE Table " + DATABASE_TABLE + " ( "
            + KEY_ID + " integer PRIMARY KEY Autoincrement, "
            + KEY_HASH + " integer, "
            + KEY_BLOB + " BLOB );";

    private final Object _syncObj;
    private final String _name;
    private final Context _context;

    /**
     * Create the persistent list.
     * @param name The name of the list. Can be anything as far as it is unique.
     * @param context The application context.
     */
    public PersistentStorageList(String name, Context context)
    {
        _syncObj = new Object();
        _name = name;
        _context = context;

        synchronized (_syncObj)
        {
            SQLiteDatabase db = open(name, _context);
            exec(db, "DROP TABLE IF EXISTS " +  DATABASE_TABLE);
            exec(db, DATABASE_CREATE);
            close(db);
        }
    }

    @Override
    public T get(int location)
    {
        T object = null;

        synchronized (_syncObj)
        {
            SQLiteDatabase db = open(_name, _context);
            Cursor cursor = query(db, new String[] { KEY_BLOB }, "_id=" + (location + 1), null);
            if (cursor != null & cursor.moveToFirst())
            {
                object = fromByteArray(cursor.getBlob(cursor.getColumnIndex(KEY_BLOB)));
                cursor.close();
            }
            close(db);
        }
        return object;
    }

    @Override
    public T set(int location, T object)
    {
        T prevObject = get(location);
        if (object != null)
        {
            ContentValues values = new ContentValues();
            values.put(KEY_HASH, object.hashCode());
            values.put(KEY_BLOB, toByteArray(object));

            synchronized (_syncObj)
            {
                SQLiteDatabase db = open(_name, _context);
                update(db, values, "_id=" + (location + 1));
                close(db);
            }
        }
        return prevObject;
    }

    @Override
    public boolean add(T object)
    {
        if (object != null)
        {
            ContentValues values = new ContentValues();
            values.put(KEY_HASH, object.hashCode());
            values.put(KEY_BLOB, toByteArray(object));

            synchronized (_syncObj)
            {
                SQLiteDatabase db = open(_name, _context);
                insert(db, values);
                close(db);
            }
        }
        return true;
    }

    @Override
    public boolean addAll(Collection<? extends T> objects)
    {
        if (objects != null)
        {
            for (T object : objects)
                add(object);
        }
        return true;
    }

    @Override
    public void add(int location, T object)
    {
        throw new UnsupportedOperationException("PersistentStorageList doesn't support insert!");
    }

    @Override
    public boolean addAll(int location, Collection<? extends T> objects)
    {
        throw new UnsupportedOperationException("PersistentStorageList doesn't support insert!");
    }

    @Override
    public boolean contains(Object object)
    {
        boolean contains = false;

        if (object != null)
        {
            synchronized (_syncObj)
            {
                SQLiteDatabase db = open(_name, _context);
                Cursor cursor = query(db, new String[] { KEY_HASH }, "_hash=" + object.hashCode(), null);
                if (cursor != null & cursor.moveToFirst())
                {
                    contains = true;
                    cursor.close();
                }
                close(db);
            }
        }
        return contains;
    }

    @Override
    public boolean containsAll(Collection<?> objects)
    {
        boolean contains = false;

        if (objects != null)
        {
            final int totalCount = objects.size();
            int matchCount = 0;
            for (Object object : objects)
            {
                if(contains(object))
                    matchCount++;
            }
            contains = matchCount == totalCount;
        }
        return contains;
    }

    @Override
    public int indexOf(Object object)
    {
        int index = -1;

        if (object != null)
        {
            synchronized (_syncObj)
            {
                SQLiteDatabase db = open(_name, _context);
                Cursor cursor = query(db, new String[] { KEY_ID }, "_hash=" + object.hashCode(), "ORDER BY _id");
                if (cursor != null & cursor.moveToFirst())
                {
                    index = cursor.getInt(cursor.getColumnIndex(KEY_ID));
                    cursor.close();
                }
                close(db);
            }
        }
        return index;
    }

    @Override
    public int lastIndexOf(Object object)
    {
        int index = -1;

        if (object != null)
        {
            synchronized (_syncObj)
            {
                SQLiteDatabase db = open(_name, _context);
                Cursor cursor = query(db, new String[] { KEY_ID }, "_hash=" + object.hashCode(), "ORDER BY _id");
                if (cursor != null & cursor.moveToLast())
                {
                    index = cursor.getInt(cursor.getColumnIndex(KEY_ID));
                    cursor.close();
                }
                close(db);
            }
        }
        return index;
    }

    @Override
    public T remove(int location)
    {
        T object = get(location);

        synchronized (_syncObj)
        {
            SQLiteDatabase db = open(_name, _context);
            delete(db, "_id=" + (location + 1));
            close(db);
        }
        return object;
    }

    @Override
    public boolean remove(Object object)
    {
        boolean removed = false;
        if (object != null)
        {
            int rowDeleted = 0;
            synchronized (_syncObj)
            {
                SQLiteDatabase db = open(_name, _context);
                rowDeleted = delete(db, "_hash=" + object.hashCode());
                close(db);
            }
            removed = rowDeleted >= 1;
        }
        return removed;
    }

    @Override
    public boolean removeAll(Collection<?> objects)
    {
        boolean removed = false;
        if (objects != null)
        {
            for (Object object : objects)
            {
                if (remove(object) && !removed)
                    removed = true;
            }
        }
        return removed;
    }

    @Override
    public void clear()
    {
        synchronized (_syncObj)
        {
            SQLiteDatabase db = open(_name, _context);
            delete(db, null);
            close(db);
        }
    }

    @Override
    public int size()
    {
        int size = 0;

        synchronized (_syncObj)
        {
            SQLiteDatabase db = open(_name, _context);
            Cursor cursor = query(db, new String[] { KEY_BLOB }, null, null);
            if (cursor != null)
            {
                size = cursor.getCount();
                cursor.close();
            }
            close(db);
        }
        return size;
    }

    @Override
    public boolean isEmpty()
    {
        return size() == 0;
    }

    @Override
    public Iterator<T> iterator()
    {
        return new Iterator<T>()
        {
            private T _next = null;
            private int _idx = 0;

            @Override
            public boolean hasNext()
            {
                _next = get(_idx++);
                return _next != null;
            }

            @Override
            public T next()
            {
                return _next;
            }

            @Override
            public void remove()
            {
                throw new UnsupportedOperationException("PersistentStorageList.Iterator doesn't support remove!");
            }
        };
    }

    @Override
    public ListIterator<T> listIterator()
    {
        throw new UnsupportedOperationException("PersistentStorageList doesn't support ListIterator!");
    }

    @Override
    public ListIterator<T> listIterator(int location)
    {
        throw new UnsupportedOperationException("PersistentStorageList doesn't support ListIterator!");
    }

    @Override
    public boolean retainAll(Collection<?> objects)
    {
        throw new UnsupportedOperationException("PersistentStorageList doesn't support retainAll!");
    }

    @Override
    public List<T> subList(int start, int end)
    {
        // If we implement this method, it kind of defeat the point of this memory-friendly list implementation.
        throw new UnsupportedOperationException("PersistentStorageList doesn't support subList!");
    }

    @Override
    public Object[] toArray()
    {
        // If we implement this method, it kind of defeat the point of this memory-friendly list implementation.
        throw new UnsupportedOperationException("PersistentStorageList doesn't support toArray!");
    }

    @SuppressWarnings("hiding")
    @Override
    public <T> T[] toArray(T[] array)
    {
        // If we implement this method, it kind of defeat the point of this memory-friendly list implementation.
        throw new UnsupportedOperationException("PersistentStorageList doesn't support toArray!");
    }

    private byte[] toByteArray(T object)
    {
        byte[] bytes = null;
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        ObjectOutput out = null;
        try {
            out = new ObjectOutputStream(bos);
            out.writeObject(object);
            bytes = bos.toByteArray();
        } catch (IOException e) {
            Log.e(TAG, "Unable to serialize byte array", e);
        }
        finally
        {
            try {
                if (out != null)
                    out.close();
                if (bos != null)
                    bos.close();
            } catch (IOException e) {
                Log.e(TAG, "Unable to close ouput stream", e);
            }
        }
        return bytes;
    }

    @SuppressWarnings("unchecked")
    private T fromByteArray(byte[] byteArray)
    {
        T object = null;
        ByteArrayInputStream bis = null;
        ObjectInput in = null;
        try {
            bis = new ByteArrayInputStream(byteArray);
            in = new ObjectInputStream(bis);
            object = (T) in.readObject();
        }
        catch (Exception e) {
            Log.e(TAG, "Unable to deserialize byte array", e);
        }
        finally
        {
            try {
                if (bis != null)
                    bis.close();
                if (in != null)
                    in.close();
            } catch (IOException e) {
                Log.e(TAG, "Unable to close input stream", e);
            }
        }
        return object;
    }

    // Helpers to manipulate the SQL database without risks of crashing
    private static SQLiteDatabase open(String name, Context context)
    {
        SQLiteDatabase db = null;
        try {
            db = context.openOrCreateDatabase(name, Context.MODE_PRIVATE, null);
        }
        catch(Exception e) {
            Log.e(TAG, "Unable to create database " + name, e);
        }
        return db;
    }

    private static void close(SQLiteDatabase db)
    {
        if (db != null)
        {
            try {
                db.close();
            }
            catch(Exception e) {
                Log.e(TAG, "Unable to close database", e);
            }
        }
    }

    private static void exec(SQLiteDatabase db, String sql)
    {
        if (db != null)
        {
            try {
                db.execSQL(sql);
            }
            catch(Exception e) {
                Log.e(TAG, "Unable to execute SQL " + sql, e);
            }
        }
    }

    private static Cursor query(SQLiteDatabase db, String[] columns, String selection, String order)
    {
        Cursor cursor = null;
        if (db != null)
        {
            try
            {
                 cursor = db.query(true, DATABASE_TABLE, columns, selection, null, null, null, order, null);
            }
            catch(Exception e) {
                Log.e(TAG, "Unable to query database colums " + columns + "and selection "+ selection, e);
            }
        }
        return cursor;
    }

    private static void update(SQLiteDatabase db, ContentValues values, String where)
    {
        if (db != null)
        {
            try
            {
                 db.update(DATABASE_TABLE, values, where, null);
            }
            catch(Exception e) {
                Log.e(TAG, "Unable to update database with values " + values + "and where "+ where, e);
            }
        }
    }

    private static void insert(SQLiteDatabase db, ContentValues values)
    {
        if (db != null)
        {
            try
            {
                db.insert(DATABASE_TABLE, null, values);
            }
            catch(Exception e) {
                Log.e(TAG, "Unable to insert database with values " + values, e);
            }
        }
    }

    private static int delete(SQLiteDatabase db, String where)
    {
        int rowDeleted = 0;
        if (db != null)
        {
            try
            {
                rowDeleted = db.delete(DATABASE_TABLE, where, null);
            }
            catch(Exception e) {
                Log.e(TAG, "Unable to insert database where " + where, e);
            }
        }
        return rowDeleted;
    }
}
