package com.nuance.dragon.toolkit.vocalizer;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.sources.QueuingAudioSource;
import com.nuance.dragon.toolkit.audio.sources.SingleSinkSource;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * This audio source generates earcon-enabled speech from text encoded with earcon markup.
 * In order for the process to work, all expected earcons must have an Audio Source associated
 * to them.
 */
public class EarconTtsAudioSource extends SingleSinkSource<AudioChunk>
{
    public interface Listener
    {
        /**
         * The source has started generating audio.
         * @param source
         */
        void onStarted(EarconTtsAudioSource source);

        /**
         * The source has stopped generating audio.
         * @param source
         */
        void onStopped(EarconTtsAudioSource source);

        /**
         * The source has encountered an error while generating audio.
         * @param source
         */
        void onError(EarconTtsAudioSource source);
    }

    private class EarconTtsPart
    {
        EarconTtsPart(String text, AudioSource<AudioChunk> audioSource)
        {
            _text = text;
            _audioSource = audioSource;
        }
        // If specified, this part is a TTS to generate using Vocalizer.
        String _text;
        // If specified, this part is a Audio Source to be used.
        AudioSource<AudioChunk> _audioSource;
    }

    private EarconTtsSession _currentEarconTtsSession;
    private boolean _closed;
    private AudioSink<AudioChunk> _connectedSink;
    private final AudioType _audioType;
    private final AudioSink<AudioChunk> _internalSink;

    private class EarconTtsSession
    {
        Listener _listener;
        QueuingAudioSource<AudioChunk> _audioSourceQueue;
        boolean _isError;
    }

    /**
     * Creates an earcon-enabled TTS audio source.
     * @param audioType The audio type that will be output and also what is
     * expected from the provided Vocalizer and Audio Sources.
     */
    public EarconTtsAudioSource(AudioType audioType)
    {
        super(null);
        
        Checker.checkArgForNull("audioType", audioType);

        _closed = false;
        _audioType = audioType;
        _internalSink = new AudioSink<AudioChunk>(_mainThreadHandler) {

            @Override
            public void chunksAvailable(AudioSource<AudioChunk> source)
            {
                if (_connectedSink != null && _currentEarconTtsSession._audioSourceQueue == source)
                {
                    _connectedSink.chunksAvailable(EarconTtsAudioSource.this);
                }
            }

            @Override
            public void framesDropped(AudioSource<AudioChunk> source)
            {
                if (_connectedSink != null && _currentEarconTtsSession._audioSourceQueue == source)
                {
                    _connectedSink.framesDropped(EarconTtsAudioSource.this);
                }
            }

            @Override
            public void sourceClosed(AudioSource<AudioChunk> source)
            {
                if (_connectedSink != null && _currentEarconTtsSession._audioSourceQueue == source)
                {
                    _closed = true;
                    _connectedSink.sourceClosed(EarconTtsAudioSource.this);
                }
            }
        };
    }

    /**
     * Converts the earcon-enabled text to speech using the provided Vocalizer instance
     * and Audio Sources that match the expected earcon names.
     * @param vocalizer Vocalizer instance that will be used for TTS.
     * @param text Earcon-enabled text to be converted to speech.
     * @param earconSourceMap Association between earcon name and the audio to play.
     * A single Recorder Source instance cannot be used for multiple earcon names.
     * @param listener Instance that will be called for notifications.
     */
    public void start(Vocalizer vocalizer, String text, Map<String, AudioSource<AudioChunk>> earconSourceMap, Listener listener)
    {
        Checker.checkState(this, null == _currentEarconTtsSession, "There is an ongoing session.");
        Checker.checkArgForNull("vocalizer", vocalizer);
        Checker.checkArgForNull("text", text);
        Checker.checkArgForNull("earconSourceMap", earconSourceMap);

        List<EarconTtsPart> earconTtsPartList = buildEarconTtsParts(text, earconSourceMap);
        validateAudioTypeConsistency(vocalizer, earconTtsPartList);

        // At this point, all possible validation/preparation was done. Time to enqueue the sources.
        QueuingAudioSource<AudioChunk> audioSourceQueue = new QueuingAudioSource<AudioChunk>(_audioType, true);
        _currentEarconTtsSession = new EarconTtsSession();
        _currentEarconTtsSession._listener = listener;
        _currentEarconTtsSession._audioSourceQueue = audioSourceQueue;
        _currentEarconTtsSession._isError = false;
        _internalSink.connectAudioSource(audioSourceQueue);

          // A new session is about to start. Reset the closed flag.
        _closed = false;

        if (listener != null)
        {
            listener.onStarted(this);
        }

        for (EarconTtsPart earconTtsPart : earconTtsPartList)
        {
            AudioSource<AudioChunk> audioSource = null;
            if (null != earconTtsPart._text) {
                audioSource = vocalizer.generateTts(earconTtsPart._text, new Vocalizer.TtsListener()
                {
                    @Override
                    public void ttsStreamingStarted(String text, Object context,
                            Vocalizer vocalizer)
                    {
                        // Nothing to do here.
                    }

                    @Override
                    public void ttsStreamingFinished(String text, Object context,
                            Vocalizer vocalizer)
                    {
                        // Nothing to do here.
                    }

                    @Override
                    public void ttsGenerationStarted(String text, Object context,
                            Vocalizer vocalizer)
                    {

                        // Nothing to do here.
                    }

                    @Override
                    public void ttsGenerationFinished(String text, Object context,
                            Vocalizer vocalizer, boolean success)
                    {
                        if (!success && _currentEarconTtsSession != null)
                        {
                            _currentEarconTtsSession._isError = true;
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run() {
                                    stop();
                                }
                            });
                        }
                    }
                }, null);

            } else if (null != earconTtsPart._audioSource)
            {
                audioSource = earconTtsPart._audioSource;
            }

            // Enqueue the audio source for playback.
            if (audioSource != null)
            {
                audioSourceQueue.queue(audioSource);
            }
        }
    }

    /**
     * Abruptly stops the earcon-enabled TTS playback, if any ongoing.
     */
    public void stop()
    {
       _internalSink.disconnectAudioSource();

       if (_currentEarconTtsSession != null)
       {
           // A session was ongoing, the event can be raised.
           if (_currentEarconTtsSession._listener != null)
           {
               if (_currentEarconTtsSession._isError)
               {
                   _currentEarconTtsSession._listener.onError(this);
               } else
               {
                   _currentEarconTtsSession._listener.onStopped(this);
               }
           }
           _currentEarconTtsSession = null;
       }

        if (_connectedSink != null && _closed == false)
        {
            _connectedSink.sourceClosed(this);
            _closed = true;
        }
    }

    @Override
    protected AudioChunk getAudioChunk()
    {
        if (_currentEarconTtsSession != null)
        {
            return _currentEarconTtsSession._audioSourceQueue.getAudioChunkForSink(_internalSink);
        }
        return null;
    }

    @Override
    protected void onSinkConnected(AudioSink<AudioChunk> sink)
    {
        _connectedSink = sink;
    }

    @Override
    protected void onSinkDisconnected(AudioSink<AudioChunk> sink)
    {
        if (_connectedSink == sink)
        {
            _connectedSink = null;
            stop();
        }
    }

    @Override
    public AudioType getAudioType()
    {
        return _audioType;
    }

    @Override
    public boolean isActive()
    {
        return !_closed;
    }

    @Override
    public int getChunksAvailable()
    {
        if (_currentEarconTtsSession != null)
        {
            int available = _currentEarconTtsSession._audioSourceQueue.getChunksAvailable();
            return available;
        }
        return 0;
    }

    private List<EarconTtsPart> buildEarconTtsParts(String text, Map<String, AudioSource<AudioChunk>> earconSourceMap)
    {
        List<EarconTtsPart> earconTtsPartList = new ArrayList<EarconTtsPart>();

        // NOTE: Earcon format is "\u001b\\earcon=<name>\\"
        // For future use: Pattern earconPattern = Pattern.compile("\u001b\\\\([^=]+)=([^\\\\]*)\\\\");
        Pattern earconPattern = Pattern.compile("\u001b\\\\(earcon)=([^\\\\]*)\\\\");
        Matcher matcher = earconPattern.matcher(text);

        int textPos = 0;
        while (matcher.find()) {
            if (matcher.start() > textPos)
            {
                String rawText = text.substring(textPos, matcher.start());
                if (rawText.length() > 0)
                {
                    earconTtsPartList.add(new EarconTtsPart(rawText, null));
                }
            }

            // Uncomment the regex above and the line below to support more keys other than 'earcon'.
            // String earconKey = matcher.group(1);
            String earconValue = matcher.group(2);

            AudioSource<AudioChunk> earconSource = earconSourceMap.get(earconValue);
            if (null == earconSource)
            {
                throw new IllegalArgumentException(String.format("No Audio Source provided for Earcon Name [%s].", earconValue));
            }
            earconTtsPartList.add(new EarconTtsPart(null, earconSource));

            textPos = matcher.end();
        }

        if (text.length() > textPos)
        {
            String rawText = text.substring(textPos);
            if (rawText.length() > 0)
            {
                earconTtsPartList.add(new EarconTtsPart(rawText, null));
            }
        }

        return earconTtsPartList;
    }

    private void validateAudioTypeConsistency(Vocalizer vocalizer, List<EarconTtsPart> earconTtsPartList)
    {
        Checker.checkArgForCondition("Vocalizer", "a supported audio type", _audioType.equals(vocalizer.getAudioType()));

        for (EarconTtsPart earconTtsPart : earconTtsPartList)
        {
            if (null != earconTtsPart._audioSource)
            {
                Checker.checkArgForCondition("Recorder Source", "a supported audio type", _audioType.equals(earconTtsPart._audioSource.getAudioType()));
            }
        }
    }
}
