package com.nuance.dragon.toolkit.vocalizer;

import java.util.List;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;

/**
 * An interface that provides access to the Vocalizer engine. All methods here
 * are expected to be called from a single worker thread unless otherwise specified.
 */
interface NativeVocalizer
{
    interface SpeakListener
    {
        /**
         * Called from the worker thread when the TTS engine has generated new audio.
         * @param chunk A chunk of generated audio. Can be null.
         * @param done True when the engine is done generating, else false.
         */
        void onNewChunk(AudioChunk chunk, boolean done);
    }

    /**
     * Called from ANY thread to synchronously request the supported models.
     * @param retModels Add the detected models here.
     * @return false if unsuccessful
     */
    boolean listAvailableModels(List<VocalizerModelJni> retModels);

    /**
     * Called from the worker thread to initialize vocalizer with a given language
     * plus/minus a voice.
     * @param language
     * @param voice
     * @param freqKhz
     * @param gender
     * @return The type of the audio that will be generated, or null if failed
     */
    AudioType init(String language, String voice, int freqKhz, int gender);

    /**
     * Called from the worker thread to destroy vocalizer.
     */
    void destroy();

    /**
     * Called from the worker thread to execute a TTS request.
     * @param text The text to TTS
     * @param ssml True for SSML input, false for plain text (including control sequences)
     * @param speed The TTS speed
     * @param pitch The TTS pitch
     * @param volume The TTS volume
     * @param msPerChunk The length of a TTS chunk in milliseconds
     * @param listener The listener to receive the generated audio.
     * @return false if unsuccessful
     */
    boolean speak(String text, boolean ssml, int speed, int pitch, int volume, int msPerChunk, SpeakListener listener);

    /**
     * Called from the main thread to cancel the current TTS session.
     */
    void cancel();

    /**
     * Called from the worker thread to load a resource (user dictionary, rule
     * set, active prompt).
     * @param fileName The resource file name
     * @param mimeType The resource MIME type
     * (please refer to the Vocalizer Engine programmer's guide for more detail)
     * @return false if unsuccessful
     */
    boolean loadResource(String fileName, String mimeType);

    /**
     * Called from the worker thread to release a resource.
     * @param id The resource file name
     */
    void releaseResource(String fileName);

    /**
     * Enable or disable verbose Vocalizer logging. (called from worker thread)
     *
     * @param verbose True if verbose (detailed) l
     *
     */
    void enableVocalizerVerboseLogging(boolean verbose);
}
