package com.nuance.dragon.toolkit.vocalizer;

import java.util.HashMap;
import java.util.List;
import java.util.Map.Entry;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.AudioType.Encoding;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;

/**
 * Internal class that directly interfaces with the "Vocalizer For Automotive" engine through JNI.
 * This is split from VocalizerImpl so that VocalizerImpl can be obfuscated
 * (this class doesn't need to be, since it isn't public).
 */
class NativeVocalizerAutomotiveImpl implements NativeVocalizer
{
    static boolean LOAD_LIBRARY_SUCCESS = true;
    static {
        try {
            System.loadLibrary("dmt_vfa");
        }
        catch (UnsatisfiedLinkError e) {
            Logger.debug(NativeVocalizerAutomotiveImpl.class, "Failed to load native VFA library.");
            LOAD_LIBRARY_SUCCESS = false;
        }
    }

    public interface OutputDeviceListener
    {
        /**
         * Called from the worker thread when the TTS engine has generated new audio.
         * @param audio A frame of generated audio.
         * @param done True when the engine is done generating, else false.
         */
        void notify(short[] audio, boolean done);
    }

    public class Status {
        /*  Error codes mirrored from the C side */
        public static final int OK = 0;
        public static final int OOM = 1;
        public static final int INVALID_MODEL = 2;
        public static final int FILE_ERROR = 3;
        public static final int OTHER_ERROR = 4;
    }


    public native int vocalizerCreate(long[] vocalizer, FileManager fileManager, String language, String voice, int freqKhz, OutputDeviceListener listener);
    public native void vocalizerDestroy(long vocalizer);
    public native int getVocalizerModels(FileManager fileManager, List<VocalizerModelJni> models);
    public native int vocalizerSpeak(long vocalizer, String text, int speed, int volume, int msPerBuffer);
    public native void vocalizerCancel(long vocalizer);
    public native int vocalizerGetAudioFrequency(long vocalizer);
    public native int vocalizerLoadResource(long vocalizer, String fileName, String mimeType, long[] resource);
    public native void vocalizerReleaseResource(long vocalizer, long resource);
    public native void vocalizerEnableVerboseLogging(boolean verbose);
    
    private final FileManager _fileManager;
    private long _vocalizer;
    private AudioType _audioType;
    private final HashMap<String, Long> _resources;
    private SpeakListener _listener;
    private short[] _outputBuffer;
    private int _outputBufferFilledLen, _outputBufferTargetLen;

    NativeVocalizerAutomotiveImpl(FileManager fileManager)
    {
        _fileManager = fileManager;
        _resources = new HashMap<String, Long>();
    }

    @Override
    public boolean listAvailableModels(List<VocalizerModelJni> retModels)
    {
        int ret = getVocalizerModels(_fileManager, retModels);
        return ret == 0;
    }

    @Override
    public AudioType init(String language, String voice, int freqKhz, int gender)
    {
        // Destroy Vocalizer if it already exists
        destroy();

        _audioType = null;

        if ((language != null && language.length() > 0) ||
            (voice != null && voice.length() > 0) ) {

            // START JNI_CODE
            long[] vocalizerTemp = new long[1];
            int status = vocalizerCreate(vocalizerTemp, _fileManager, language, voice, freqKhz, new OutputDeviceListener()
            {
                @Override
                public void notify(short[] audio, boolean done)
                {
                    NativeVocalizerAutomotiveImpl.this.notify(audio, done);
                }
            });

            if (status == Status.OK && vocalizerTemp[0] != 0)
            {
                // Note that this array operation is a JNI-specific workaround
                // to simulate a pass by reference to the C side.
                _vocalizer = vocalizerTemp[0];

                switch(vocalizerGetAudioFrequency(_vocalizer)) {
                    case Vocalizer.Frequencies.FREQ_8KHZ:
                        _audioType = new AudioType(Encoding.PCM_16, AudioType.Frequency.FREQ_8KHZ);
                        break;
                    case Vocalizer.Frequencies.FREQ_11KHZ:
                        _audioType = new AudioType(Encoding.PCM_16, AudioType.Frequency.FREQ_11KHZ);
                        break;
                    case Vocalizer.Frequencies.FREQ_16KHZ:
                        _audioType = new AudioType(Encoding.PCM_16, AudioType.Frequency.FREQ_16KHZ);
                        break;
                    case Vocalizer.Frequencies.FREQ_22KHZ:
                        _audioType = new AudioType(Encoding.PCM_16, AudioType.Frequency.FREQ_22KHZ);
                        break;
                    default:
                        Logger.error(this, "Unable to create native Vocalizer (invalid model frequency).");
                        vocalizerDestroy(_vocalizer);
                        _vocalizer = 0;
                        break;
                }
            }
            else {
                Logger.error(this, "Unable to create native Vocalizer (error status = " + status + ").");
            }
        }

        return _audioType;
    }

    @Override
    public boolean speak(String text, boolean ssml, int speed, int pitch, int volume, int msPerChunk, SpeakListener listener)
    {
        // Note: SSML and pitch are not exposed in NMT VfA C layer (deprecated).
        if (_vocalizer != 0)
        {
            _listener = listener;
            _outputBufferFilledLen = 0;
            _outputBufferTargetLen = _audioType.getSampleCount(msPerChunk);
            _outputBuffer = new short[_outputBufferTargetLen];

            Logger.debug(this, "Output buffer target length = " + _outputBufferTargetLen);

            if (vocalizerSpeak(_vocalizer, text, speed, volume, msPerChunk) == Status.OK) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void cancel()
    {
        if (_vocalizer != 0) {
            vocalizerCancel(_vocalizer);
        }
    }

    @Override
    public void destroy()
    {
        if (_vocalizer != 0) {
            // Release loaded resources (if any)
            for (Entry<String,Long> resource : _resources.entrySet())
                vocalizerReleaseResource(_vocalizer, resource.getValue());
            _resources.clear();

            vocalizerDestroy(_vocalizer);
            _vocalizer = 0;
        }
    }

    @Override
    public boolean loadResource(String fileName, String mimeType)
    {
        boolean success = false;

        if (!_resources.containsKey(fileName))
        {
            // START JNI_CODE
            if (_vocalizer != 0) {
                // Again, this array operation is a JNI-specific workaround
                // to simulate a pass by reference to the C side.
                long[] resourceTemp = new long[1];
                int status = vocalizerLoadResource(_vocalizer, fileName, mimeType, resourceTemp);
                if (status == Status.OK && resourceTemp[0] != 0)
                {
                    _resources.put(fileName, resourceTemp[0]);
                    success = true;
                }
                else
                {
                    Logger.warn(this, "Unable to load resource (error status = " + status + ").");
                }
            }
            // END JNI_CODE
        }
        else
        {
            Logger.warn(this, "This resource is already loaded.");
        }

        return success;
    }

    @Override
    public void releaseResource(String fileName)
    {
        if (_vocalizer != 0) {
            Long resource = _resources.remove(fileName);
            if (resource != null)
            {
                vocalizerReleaseResource(_vocalizer, resource);
            }
            else
            {
                Logger.warn(this, "The vocalizer resource was never loaded or has already been released: " + fileName);
            }
        }
    }

    @Override
    public void enableVocalizerVerboseLogging(boolean verbose)
    {
        vocalizerEnableVerboseLogging(verbose);
    }

    private void notify(short[] audio, boolean done)
    {
        int filledIdx = 0;
        boolean filled = false;
        if (audio != null)
        {
            for (int idx = 0; idx < audio.length && !filled; idx++)
            {
                if (_outputBufferFilledLen < _outputBuffer.length)
                {
                    _outputBuffer[_outputBufferFilledLen] = audio[idx];
                    _outputBufferFilledLen++;
                }
                else
                {
                    filled = true;
                    filledIdx = idx;
                }
            }
        }

        if (filled || done)
        {
            AudioChunk chunk = getAudioChunk(_audioType, _outputBuffer, _outputBufferFilledLen);
            _outputBufferFilledLen = 0;

            // Add left-over (if any)
            if (audio != null)
            {
                for (int idx = filledIdx; idx < audio.length; idx++)
                {
                    _outputBuffer[_outputBufferFilledLen] = audio[idx];
                    _outputBufferFilledLen++;
                }
            }

            if(_listener != null)
                _listener.onNewChunk(chunk, done);
        }
    }

    // Helper that populates the audio chunk from the audio frame.
    private static AudioChunk getAudioChunk(AudioType audioType, short[] buffer, int len)
    {
        Logger.verbose(NativeVocalizerAutomotiveImpl.class, "getAudioChunk() length = " + len);
        short[] copy = new short[len];
        System.arraycopy(buffer, 0, copy, 0, copy.length);
        return new AudioChunk(audioType, copy);
    }
}
