package com.nuance.dragon.toolkit.vocalizer;

import java.util.LinkedList;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.sources.SingleSinkSource;
import com.nuance.dragon.toolkit.vocalizer.Vocalizer.TtsListener;

/**
 * An audio source that contains the audio generated by a single Vocalizer TTS
 * request.
 */
public class TtsAudioSource extends SingleSinkSource<AudioChunk>
{
    /**
     * Notifies of a TTS job completion
     */
    interface JobCompleteListener
    {
        /**
         * Called when this TTS audio source has been fully consumed.
         * @param job TTSAudioSource instance that completed
         */
        void onComplete(TtsAudioSource job);
    }

    private final Vocalizer _vocalizer;
    final String _text;
    final boolean _ssml;
    final int _msPerChunk;
    final int _samplesPerChunk;
    final int _bufferCount;
    final int _speed;
    final int _pitch;
    final int _volume;
    final Object _context;
    private final TtsListener _listener;
    private final JobCompleteListener _completeListener;
    private boolean _streamingStarted;
    private boolean _preloading;
    private boolean _completed; // Touched from both threads
    private boolean _active;
    //private int _duration;

    private final LinkedList<AudioChunk> _chunks;

    /**
     * @param vocalizer
     * @param text
     * @param ssml
     * @param speed
     * @param pitch
     * @param volume
     * @param msPerChunk Milliseconds per audio chunk.
     * @param samplesPerChunk
     * @param bufferChunkCount The number of chunks to attempt to keep buffered.
     * We won't report chunks available until we have this many chunks.
     * @param sink
     * @param completeListener
     * @param listener
     * @param context
     */
    TtsAudioSource(Vocalizer vocalizer,
                   String text,
                   boolean ssml,
                   int speed,
                   int pitch,
                   int volume,
                   int msPerChunk,
                   int samplesPerChunk,
                   int bufferChunkCount,
                   JobCompleteListener completeListener,
                   TtsListener listener,
                   Object context)
    {
        super(null);
        
        _vocalizer = vocalizer;
        _text = text;
        _ssml = ssml;
        _speed = speed;
        _pitch = pitch;
        _volume = volume;
        _msPerChunk = msPerChunk;
        _samplesPerChunk = samplesPerChunk;
        _bufferCount = bufferChunkCount;
        _completeListener = completeListener;
        _listener = listener;
        _context = context;

        _chunks = new LinkedList<AudioChunk>();
    }

    /** Cancel this TTS request. */
    synchronized public void cancel()
    {
        complete(false);
    }

    /**
     * Completes the TTS request, calling {@link #finish} if necessary and immediately
     * allowing the next request to be processed.
     * @param success
     */
    synchronized void complete(boolean success)
    {
        if (!_completed)
        {
            _completed = true;
            finish(success);
            _completeListener.onComplete(this);
        }
    }

    synchronized boolean isComplete()
    {
        return _completed;
    }

    /**
     * Start the TTS request:<br>
     *  - Notifies the listener that generation started<br>
     *  - Connects the sink to this audio source
     */
    void start()
    {
        Assert.assertFalse(_active);

        if (_listener != null)
            _listener.ttsGenerationStarted(_text, _context, _vocalizer);

        _active = true;

        if (_bufferCount > 0)
            _preloading = true;
    }

    /**
     * Add a generated audio chunk.
     * @param chunk
     */
    void addChunk(AudioChunk chunk)
    {
        _chunks.add(chunk);
        //_duration += chunk.audioDuration;

        if (_preloading && _chunks.size() >= _bufferCount)
            _preloading = false;

        if (!_preloading)
            notifyChunksAvailable();
    }

    /**
     * Finish the TTS request:<br>
     *  - Notify the listener that generation finished<br>
     *  - If we were preloading, report chunks available<br>
     *  - Report that this source is closed<br>
     *  This TtsAudioSource will continue to block vocalizer from generating more audio
     *  until it is canceled or disconnected from.
     * @param success
     */
    void finish(boolean success)
    {
        if (_active)
        {
            if (_listener != null)
                _listener.ttsGenerationFinished(_text, _context, _vocalizer, success);

            _active = false;
            if (_preloading && !_chunks.isEmpty())
                notifyChunksAvailable();

            notifySourceClosed();
        }
    }

    @Override
    public AudioType getAudioType()
    {
        Assert.assertNotNull(_vocalizer);

        return _vocalizer.getAudioType();
    }

    @Override
    public int getChunksAvailable()
    {
        return _chunks.size();
    }

    @Override
    public boolean isActive()
    {
        return _active;
    }

    @Override
    public AudioChunk getAudioChunk()
    {
        if (_chunks.isEmpty()) return null;

        if (!_streamingStarted)
        {
            _streamingStarted = true;
            if (_listener != null)
                _listener.ttsStreamingStarted(_text, _context, _vocalizer);
        }

        return _chunks.removeFirst();
    }

    @Override
    protected void onSinkDisconnected(AudioSink<AudioChunk> sink)
    {
        super.onSinkDisconnected(sink);

        if (_streamingStarted)
        {
            _streamingStarted = false;
            if (_listener != null)
                _listener.ttsStreamingFinished(_text, _context, _vocalizer);
        }

        cancel();
    }
}
