package com.nuance.dragon.toolkit.vocalizer;

import java.util.List;

import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.core.data.Data.Dictionary;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;

/**
 * The Vocalizer object wraps the C Vocalizer library. The methods and callbacks
 * ('listeners') will deal with native Java objects, and convert as necessary
 * to C data types. This class does not expose all of Vocalizer's functionality--
 * only that which is intended to be used by this toolkit.
 * <p>
 * To create a Vocalizer, call {@link #createVocalizer}.
 */
public abstract class Vocalizer
{
    /** Not for public use. */
    public static boolean LOAD_NATIVE_VFA_LIBRARY_SUCCESS = NativeVocalizerAutomotiveImpl.LOAD_LIBRARY_SUCCESS;

    /** Not for public use. */
    public static boolean LOAD_NATIVE_VEX_LIBRARY_SUCCESS = NativeVocalizerExpressiveImpl.LOAD_LIBRARY_SUCCESS;

    /**
     * Provides callbacks for for certain Text-to-speech generation events.
     */
    public interface TtsListener
    {
        /**
         * TTS has started generating.
         * @param text The text that is being generated.
         * @param context The context that was originally passed to {@link
         * Vocalizer#generateTts}. Can be null.
         * @param vocalizer The associated vocalizer.
         */
        public abstract void ttsGenerationStarted(String text, Object context, Vocalizer vocalizer);

        /**
         * TTS has finished generating.
         * @param text The text that was generated.
         * @param context The context that was originally passed to {@link
         * Vocalizer#generateTts}. Can be null.
         * @param vocalizer The associated vocalizer.
         * @param success True if the generation finished successfully, false otherwise.
         */
        public abstract void ttsGenerationFinished(String text, Object context, Vocalizer vocalizer, boolean success);

        /**
         * TTS has started streaming--i.e., the attached audio sink has
         * begun reading audio from the Vocalizer TTS source.
         * @param text The text that was generated.
         * @param context The context that was originally passed to {@link
         * Vocalizer#generateTts}. Can be null.
         * @param vocalizer The vocalizer.
         */
        public abstract void ttsStreamingStarted(String text, Object context, Vocalizer vocalizer);

        /**
         * TTS has finished streaming--i.e., the attached audio sink has
         * finally disconnected from the Vocalizer TTS source.
         * @param text The text that was generated.
         * @param context The context that was originally passed to {@link
         * Vocalizer#generateTts}. Can be null.
         * @param vocalizer The vocalizer.
         */
        public abstract void ttsStreamingFinished(String text, Object context, Vocalizer vocalizer);
    }

    /**
     * Provides callbacks for for certain events pertaining to loading the vocalizer
     */
    public interface LoadListener
    {
        /**
         * An attempt to load the vocalizer has been made
         * @param vocalizer The vocalizer that has been loaded
         * @param success Is false if an error was encountered
         */
        public abstract void onLoaded(Vocalizer vocalizer, boolean success);
    }

    /**
     * Provides callbacks for for certain events pertaining to releasing the vocalizer
     */
    public interface ReleaseListener
    {
        /**
         * The vocalizer has been released
         * @param vocalizer The vocalizer that has been released
         */
        public abstract void onReleased(Vocalizer vocalizer);
    }

    /**
     * Provides callbacks for for certain events pertaining to loading the vocalizer resources.
     */
    public interface ResourceLoadListener
    {
        /**
         * An attempt to load the resource has been made
         * @param vocalizer The vocalizer.
         * @param success Is false if an error was encountered
         */
        public abstract void onLoaded(Vocalizer vocalizer, boolean success);
    }

    /**
     * Provides callbacks for for certain events pertaining to releasing the vocalizer resources.
     */
    public interface ResourceReleaseListener
    {
        /**
         * An attempt to release the resource has been made
         * @param vocalizer The vocalizer owning the resource that has been released
         */
        public abstract void onReleased(Vocalizer vocalizer);
    }

    /**
     * Engines supported by Vocalizer.
     */
    public static enum Engine
    {
       VOCALIZER_EXPRESSIVE,  @Deprecated VOCALIZER_FOR_AUTOMOTIVE
    }

    /**
     * Frequencies supported by Vocalizer.
     */
    public static final class Frequencies
    {
        public static final int UNKNOWN    = 0;
        public static final int FREQ_8KHZ  = 8;
        public static final int FREQ_11KHZ = 11;
        public static final int FREQ_16KHZ = 16;
        public static final int FREQ_22KHZ = 22;
    }

    /**
     * Genders supported by Vocalizer.
     */
    public static final class Genders
    {
        public static final int UNKNOWN = 0;
        public static final int MALE    = 1;
        public static final int FEMALE  = 2;
    }

    /**
     * Languages supported by Vocalizer.
     */
    public static final class Languages extends LangSpecs
    {
    }

    /**
     * Voices supported by Vocalizer.
     */
    public static final class Voices extends VoiceSpecs
    {
    }

    /**
     * Get a list of available Vocalizer models. Each model consists of a voice
     * (with language), frequency, and model type.
     * <p>- This may be called from any thread, prior to loading Vocalizer.
     * <br>- This method performs File I/O, so it may be preferable to call it
     * from a non-UI thread.
     * @return The list of available models.
     */
    public abstract List<VocalizerModelInfo> getAvailableModels();

    /**
     * Get a list of available Vocalizer models for the given frequency. Each model
     * consists of a voice (with language), frequency, and model type.
     * <p>- This may be called from any thread, prior to loading Vocalizer.
     * <br>- This method performs File I/O, so it may be preferable to call it
     * from a non-UI thread.
     * @param freqKhz The frequency to restrict results to (from {@link Frequencies}).
     * An invalid input will return an empty container.
     * @return The list of available models.
     */
    public abstract List<VocalizerModelInfo> getAvailableModels(int freqKhz);

    /**
     * Get a list of available voices for the given frequency.
     * <p>- This may be called from any thread, prior to loading Vocalizer.
     * <br>- This method performs File I/O, so it may be preferable to call it
     * from a non-UI thread.
     * @param freqKhz The frequency to restrict results to (from {@link Frequencies}).
     * An invalid input will return an empty container.
     * @return The list of available voices.
     */
    public abstract List<VocalizerVoice> getAvailableVoices(int freqKhz);

    /**
     * Get a list of available voices for the given frequency.
     * <p>- This may be called from any thread, prior to loading Vocalizer.
     * <br>- This method performs File I/O, so it may be preferable to call it from
     * @param lang the associated language to restrict results to. Null means ignore this param
     * @param freqKhz The frequency to restrict results to (from {@link Frequencies}).
     *  An invalid input will return an empty container.
     * @return The list of available voices.
     */
    public abstract List<VocalizerVoice> getAvailableVoices(int freqKhz, VocalizerLanguage lang);

    /**
     * Get a list of available languages for the given frequency.
     * <p>- This may be called from any thread, prior to loading Vocalizer.
     * <br>- This method performs File I/O, so it may be preferable to call it from
     * @param freqKhz The frequency to restrict results to (from {@link Frequencies}).
     * An invalid input will return an empty container.
     * @return The list of available languages.
     */
    public abstract List<VocalizerLanguage> getAvailableLanguages(int freqKhz);

    /**
     * Release all resources when done with the vocalizer. The vocalizer can
     * never be used after calling this. Not calling this may result in memory
     * leaks.
     * @param listener Notified after the release completes. Null means that no listener will be notified
     */
    public abstract void release(ReleaseListener listener);

    /**
     * Equivalent to {@link #release(ReleaseListener)} with a null listener.
     */
    public abstract void release();

    /**
     * Load the specified vocalizer configuration. This method must be called
     * prior to generating TTS.
     * @param config The configuration parameters (e.g. Language/Voice + audio
     * frequency) to  load
     * @param listener Notified after the load completes. Null means that no
     * listener will be notified
     */
    public abstract void load(VocalizerConfig config, LoadListener listener);

    /**
     * Like {@link #load(VocalizerConfig, LoadListener)}, but no listener.
     * @param config The configuration parameters (e.g. Language/Voice) to  load
     */
    public abstract void load(VocalizerConfig config);

    /**
     * Get the TTS audio type (i.e. Audio Frequency). It is useful for configuring
     * the audio sink prior to start speech synthesis.
     * @return The TTS audio type. Will never be null. In case of an error, it will either:<br>
     * - returns the audio type that matches the frequency passed via {@link VocalizerConfig}.<br>
     * - if no frequency has been explicitly specified through {@link VocalizerConfig}, it will
     *   return {@link com.nuance.dragon.toolkit.audio.AudioType.Frequency#FREQ_16KHZ 16KHZ} as a default audio type.
     */
    public abstract AudioType getAudioType();

    /**
     * Set the TTS speed.
     * @param speed An integer percentage of normal speed:
     *              50 is 2x slower than normal, 100 is normal, and 400 is 4x
     *              faster than normal. Any value between 50 and 400 is ok.
     *              Values outside that range are clamped.<br>
     *              Default is 100.
     */
    public abstract void setTtsSpeed(int speed);

    /**
     * Set the TTS pitch.
     * @param pitch An integer percentage of normal pitch:
     *              50 is one octave lower than normal, 100 is normal, and 200 is
     *              one octave higher than normal. Any value between 50 and 200 is ok.
     *              Values outside that range are clamped.<br>
     *              Default is 100.
     */
    public abstract void setTtsPitch(int pitch);

    /**
     * Set the TTS volume.
     * @param volume An integer value between 0 (soft) to 100 (loud).
     *               For each 10 points, the volume changes by 3 dB.
     *               Values outside that range are clamped.<br>
     *               Default is 80.
     */
    public abstract void setTtsVolume(int volume);

    /**
     * Set the size of the audio chunks that will be generated and made
     * available via the {@link TtsAudioSource}.
     * @param sizeInMs The size of the chunks, in milliseconds. There is no
     *                 maximum value for this integer parameter. However,
     *                 minimum value has to be greater or equal to 20.
     *                 Values below 20 are clamped.<br>
     *                 Default is 200.
     */
    public abstract void setAudioChunkSize(int sizeInMs);

    /**
     * Set the number of audio chunks to attempt to keep available. This would
     * typically be used to generate TTS as it is played.
     * @param numChunks The number of chunks to keep available. If negative, Vocalizer
     * will generate TTS as fast as possible. Otherwise:<br>
     *  - Vocalizer will not report chunks available until this many chunks have
     *    been prebuffered, or TTS generation has completed.<br>
     *  - Vocalizer will only generate new chunks when its internal buffer dips
     *    below numChunks. <br>
     * Default is 5.
     */
    public abstract void setAudioChunkBufferCount(int numChunks);

    /**
     * Generate TTS audio. The generated audio will be available from the
     * returned audio source. If TTS audio is already being generated, this
     * request will be queued and will automatically begin when the previous
     * request has finished.
     *
     * @param text The text to speak with or without include control sequences.
                   For more information on control sequences, please read the
                   the Vocalizer User's Guide.
     * @param listener Optional listener for TTS generation status updates.
     * @param context Optional context object that will be passed back via the
     * listener.
     *
     * @return The audio source. This audio source will block subsequent TTS
     * from being generated until it is fully consumed or canceled.
     */
    public abstract TtsAudioSource generateTts(String text,
                                               TtsListener listener,
                                               Object context);

    /**
     * Generate TTS audio. The generated audio will be available from the
     * returned audio source. If TTS audio is already being generated, this
     * request will be queued and will automatically begin when the previous
     * request has finished.
     *
     * @param param The parameter to speak. For example:
     * <br>Data.Dictionary ttsParamData = new Data.Dictionary();
     * <br>ttsParamData.put("tts_input", "Hello, I'm Dragon.");
     * <br>ttsParamData.put("tts_type", "text");
     * @param listener Optional listener for TTS generation status updates.
     * @param context Optional context object that will be passed back via the
     * listener.
     *
     * @return The audio source. This audio source will block subsequent TTS
     * from being generated until it is fully consumed or canceled.
     */
    public abstract TtsAudioSource generateTts(Dictionary param,
                                               TtsListener listener,
                                               Object context);

    /**
     * Cancel any TTS that is in progress.
     */
    public abstract void cancelTts();

    /**
     * Load a resource (user dictionary, rule set or active prompt).
     * @param fileName The resource file name
     * @param mimeType The resource MIME type
     * (please refer to the Vocalizer Engine programmer's guide for more detail)
     * @param listener Optional listener for resource loading status updates
     */
    public abstract void loadResource(String fileName, String mimeType, ResourceLoadListener listener);

    /**
     * Release a resource that was previously loaded using
     * {@link #loadResource(String fileName, String mimeType, ResourceLoadListener listener)}.
     * @param fileName The resource file name
     * @param listener Optional listener for resource releasing status updates
     */
    public abstract void releaseResource(String fileName, ResourceReleaseListener listener);

    /**
     * This function controls the amount of information about the Vocalizer
     * that will be written to the Android logs.
     *
     * Note: Do not use / disabled for your application release build.
     *
     * @param verbose True if detailed (verbose) status information for the Vocalizer should be written
     *                to the Android logs.
     *                False if only minimal details and error information should be written to  the Android logs.
     */
    public abstract void enableVerboseAndroidLogging(boolean verbose);

    /**
     * Create a Vocalizer that uses a dedicated worker thread.
     *
     * @param fileManager The file manager where Vocalizer should look for files.
     */
    public static Vocalizer createVocalizer(FileManager fileManager)
    {
        return new VocalizerImpl(Engine.VOCALIZER_EXPRESSIVE, fileManager);
    }

    /**
     * Create a Vocalizer that uses the given worker thread handler.
     *
     * @param fileManager The file manager where Vocalizer should look for files.
     * @param workerThread The worker thread to use. If null, a dedicated one
     * will be created.
     */
    public static Vocalizer createVocalizer(FileManager fileManager, NMTHandler workerThread)
    {
        return new VocalizerImpl(Engine.VOCALIZER_EXPRESSIVE, fileManager, workerThread);
    }

    /**
     * Create a Vocalizer that uses a dedicated worker thread.
     *
     * @param engine The engine used for TTS generation, see {@link Engine}.
     * @param fileManager The file manager where Vocalizer should look for files.
     */
    public static Vocalizer createVocalizer(Engine engine, FileManager fileManager)
    {
        return new VocalizerImpl(engine, fileManager);
    }

    /**
     * Create a Vocalizer that uses the given worker thread handler.
     *
     * @param engine The engine used for TTS generation, see {@link Engine}.
     * @param fileManager The file manager where Vocalizer should look for files.
     * @param workerThread The worker thread to use. If null, a dedicated one
     * will be created.
     */
    public static Vocalizer createVocalizer(Engine engine, FileManager fileManager, NMTHandler workerThread)
    {
        return new VocalizerImpl(engine, fileManager, workerThread);
    }
}
