package com.nuance.dragon.toolkit.vocalizer;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import junit.framework.Assert;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.AudioType.Encoding;
import com.nuance.dragon.toolkit.core.data.Data.Dictionary;
import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.util.internal.ThreadUtils;
import com.nuance.dragon.toolkit.vocalizer.NativeVocalizer.SpeakListener;

/**
 * Base class for Vocalizer implementations.
 * <p>
 * There will probably be one 'real' subclass of this ({@link VocalizerImpl}),
 * along with variants used for unit testing (e.g. {@link VocalizerImpersonator}).
 */
abstract class VocalizerBase extends Vocalizer
{
    private final NativeVocalizer _nativeVocalizer;
    private final LinkedList<TtsAudioSource> _ttsSourcesToGenerate;
    private TtsAudioSource _currentTtsSource;
    private int _ttsSpeed;
    private int _ttsPitch;
    private int _ttsVolume;
    private int _ttsBufferCount;
    private int _ttsChunkSizeMs;
    private AudioType _audioType;
    private boolean _isLoaded;
    private boolean _isReleased;
    private final NMTHandler _mainThreadHandler;
    private final NMTHandler _workerThreadHandler;

    /**
     * @param nativeVocalizer
     * @param externalWorkerThreadHandler The external worker thread handler.
     * If this is null, {@link #getWorkerThreadHandler} will be called;
     * otherwise, we'll simply use the given handler.
     */
    VocalizerBase(NativeVocalizer nativeVocalizer, NMTHandler externalWorkerThreadHandler)
    {
        _nativeVocalizer = nativeVocalizer;
        _isReleased = false;
        _ttsSpeed = 100;
        _ttsPitch = 100;
        _ttsVolume = 80;
        _ttsBufferCount = 5;
        _ttsChunkSizeMs = 200;
        _audioType = AudioType.UNKNOWN;
        _ttsSourcesToGenerate = new LinkedList<TtsAudioSource>();
        _mainThreadHandler = Factory.createNMTHandler();
        if (externalWorkerThreadHandler == null)
            _workerThreadHandler = getWorkerThreadHandler();
        else
            _workerThreadHandler = externalWorkerThreadHandler;
    }

    @Override
    public List<VocalizerModelInfo> getAvailableModels()
    {
        return getAvailableModels(Frequencies.UNKNOWN);
    }


    @Override
    public final List<VocalizerModelInfo> getAvailableModels(int freqKhz)
    {
        ArrayList<VocalizerModelJni> nativeList = new ArrayList<VocalizerModelJni>();
        boolean success = _nativeVocalizer.listAvailableModels(nativeList);
        if (success)
        {
            ArrayList<VocalizerModelInfo> infos = new ArrayList<VocalizerModelInfo>(nativeList.size());
            for (VocalizerModelJni model : nativeList)
            {
                if (freqKhz == Frequencies.UNKNOWN || freqKhz == model.freqKhz)
                {
                    VocalizerVoice v = VocalizerVoice.findByName(model.voice);
                    if (v != null)
                    {
                        if (v.language.name.equalsIgnoreCase(model.language))
                        {
                            infos.add(new VocalizerModelInfo(v, model.freqKhz, model.model));
                        } else
                        {
                            Logger.warn(this, "Language name " + model.language + " doesn't match " + v.name + "'s language " + v.language.name);
                        }
                    } else
                    {
                        Logger.warn(this, "Unable to find Vocalizer voice " + model.voice);
                    }
                }
            }
            return infos;
        }
        Logger.error(this, "Unable to detect supported Vocalizer models");
        return new ArrayList<VocalizerModelInfo>();
    }

    @Override
    public final List<VocalizerVoice> getAvailableVoices(int freqKhz)
    {
        return getAvailableVoices(freqKhz, null);
    }

    @Override
    public final List<VocalizerVoice> getAvailableVoices(int freqKhz, VocalizerLanguage lang)
    {
        // get voices associated with this language
        VocalizerVoice[] languageVoices = null;
        if (lang != null)
        {
            languageVoices = lang.voices;
        }

        // get voices on phone
        List<VocalizerModelInfo> models = getAvailableModels(freqKhz);
        ArrayList<VocalizerVoice> voices = new ArrayList<VocalizerVoice>(models.size());
        for (VocalizerModelInfo model : models)
        {
            VocalizerVoice thisVoice = model.voice;

            if (!voices.contains(thisVoice))
            {
                boolean addVoiceToOutput = false;

                // if we are restricting output to voices supported by specified make sure it is in the voices array
                if (languageVoices != null)
                {
                    for (VocalizerVoice languageVoice : languageVoices)
                    {
                        if (languageVoice != null && languageVoice.equals(thisVoice))
                        {
                            addVoiceToOutput = true;
                            break;
                        }
                    }
                }

                // if we are not restricting output, just add
                else
                {
                    addVoiceToOutput = true;
                }

                if (addVoiceToOutput)
                {
                    voices.add(thisVoice);
                }
            }
        }
        return voices;
    }

    @Override
    public final List<VocalizerLanguage> getAvailableLanguages(int freqKhz)
    {
        List<VocalizerModelInfo> models = getAvailableModels(freqKhz);
        ArrayList<VocalizerLanguage> languages = new ArrayList<VocalizerLanguage>(models.size());
        for (VocalizerModelInfo model : models)
        {
            if (!languages.contains(model.voice.language))
                languages.add(model.voice.language);
        }
        return languages;
    }

    @Override
    public void release(final ReleaseListener listener)
    {
        Checker.checkState(this, !_isReleased);

        cancelTts();
        _isReleased = true;
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _nativeVocalizer.destroy();
                if (listener != null)
                {
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            listener.onReleased(VocalizerBase.this);
                        }
                    });
                }
            }
        });
    }

    @Override
    public void release()
    {
        Checker.checkState(this, !_isReleased);

        release(null);
    }

    @Override
    public void load(VocalizerConfig config, LoadListener listener)
    {
        Checker.checkArgForNull("config", config);

        loadInternal(config, listener);
    }

    @Override
    public void load(VocalizerConfig config)
    {
        load(config, (LoadListener)null);
    }

    private void loadInternal(final VocalizerConfig config, final LoadListener listener)
    {
        Checker.checkState(this, !_isReleased);

        _isLoaded = true;
        // Initialize audio type here to the passed audio frequency
        _audioType = new AudioType(Encoding.PCM_16, getFrequencyInHz(config.freqKhz));

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final AudioType audioType = _nativeVocalizer.init(config.language, config.voice, config.freqKhz, config.gender);
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (audioType != null)
                            _audioType = audioType;
                           if (listener != null)
                            listener.onLoaded(VocalizerBase.this, audioType != null);
                    }
                });
            }
        });
    }

    @Override
    public void setTtsSpeed(int speed)
    {
        Checker.checkState(this, !_isReleased);

        if (speed < 50) speed = 50;
        else if (speed > 400) speed = 400;
        _ttsSpeed = speed;
    }

    @Override
    public void setTtsPitch(int pitch)
    {
        Checker.checkState(this, !_isReleased);

        if (pitch < 50) pitch = 50;
        else if (pitch > 200) pitch = 200;
        _ttsPitch = pitch;
    }

    @Override
    public void setTtsVolume(int volume)
    {
        Checker.checkState(this, !_isReleased);

        if (volume < 0) volume = 0;
        else if (volume > 100) volume = 100;
        _ttsVolume = volume;
    }

    @Override
    public void setAudioChunkSize(int sizeInMs)
    {
        Checker.checkState(this, !_isReleased);

        if (sizeInMs < 20) sizeInMs = 20;
        _ttsChunkSizeMs = sizeInMs;
    }

    @Override
    public void setAudioChunkBufferCount(int numChunks)
    {
        Checker.checkState(this, !_isReleased);

        if (numChunks == 0) numChunks = 1;
        if (numChunks < -1) numChunks = -1;
        _ttsBufferCount = numChunks;
    }

    @Override
    public TtsAudioSource generateTts(String text,
                                      TtsListener listener,
                                      Object context)
    {
        Checker.checkArgForNull("text", text);

        Dictionary param = new Dictionary();
        param.put("tts_input", text);
        param.put("tts_type", "text");
        return generateTts(param, listener, context);
    }

    @Override
    public TtsAudioSource generateTts(Dictionary param,
                                      TtsListener listener,
                                      Object context)
    {
        Checker.checkArgForNull("param", param);
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isLoaded);

        Assert.assertNotNull(_audioType);

        TtsAudioSource job = new TtsAudioSource(this, param.getString("tts_input").value, param.getString("tts_type").value.equals("ssml"), _ttsSpeed, _ttsPitch, _ttsVolume, _ttsChunkSizeMs, _audioType.getSampleCount(_ttsChunkSizeMs), _ttsBufferCount, new TtsAudioSource.JobCompleteListener()
        {
            @Override
            public void onComplete(TtsAudioSource job)
            {
                if (_currentTtsSource == job)
                {
                    _currentTtsSource = null;

                    _nativeVocalizer.cancel();
                    startNextJob();
                } else
                {
                    _ttsSourcesToGenerate.remove(job);
                }
            }
        }, listener, context);
        _ttsSourcesToGenerate.add(job);
        if (_currentTtsSource == null)
        {
            startNextJob();
        }
        return job;
    }

    @Override
    public void cancelTts()
    {
        Checker.checkState(this, !_isReleased);

        Logger.debug(this, "Canceling " + (_ttsSourcesToGenerate.size() + (_currentTtsSource == null ? 0 : 1)) + " TTS jobs");

        ArrayList<TtsAudioSource> toCancel = new ArrayList<TtsAudioSource>(_ttsSourcesToGenerate.size() + 1);

        if (_currentTtsSource != null)
        {
            // Don't cancel the job until we've cleared out _ttsJobs, since
            // canceling will trigger the next job to start
            toCancel.add(_currentTtsSource);
            _currentTtsSource = null;

            _nativeVocalizer.cancel();
        }

        toCancel.addAll(_ttsSourcesToGenerate);
        _ttsSourcesToGenerate.clear();

        // Now that Vocalizer is cleaned up, cancel all the jobs
        for (TtsAudioSource job : toCancel)
        {
            job.cancel();
        }
    }

    @Override
    public void loadResource(final String fileName, final String mimeType, final ResourceLoadListener listener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkStringArgForNullOrEmpty("fileName", fileName);
        Checker.checkStringArgForNullOrEmpty("mimeType", mimeType);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final boolean success = _nativeVocalizer.loadResource(fileName, mimeType);

                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (listener != null)
                            listener.onLoaded(VocalizerBase.this, success);
                    }
                });
            }
        });
    }

    @Override
    public void releaseResource(final String fileName, final ResourceReleaseListener listener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkStringArgForNullOrEmpty("fileName", fileName);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _nativeVocalizer.releaseResource(fileName);

                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (listener != null)
                            listener.onReleased(VocalizerBase.this);
                    }
                });
            }
        });
    }

    @Override
    public void enableVerboseAndroidLogging(boolean verbose)
    {
        _nativeVocalizer.enableVocalizerVerboseLogging(verbose);
    }

    @Override
    public AudioType getAudioType()
    {
        return _audioType;
    }

    /** Called from main thread to start the next TTS job in the queue */
    private void startNextJob()
    {
        if (!_ttsSourcesToGenerate.isEmpty())
        {
            final TtsAudioSource job = _ttsSourcesToGenerate.removeFirst();
            _currentTtsSource = job;
            Logger.debug(this, "Generating TTS for '" + job._text + "'");

            job.start();
            commenceGenerating(job);
        }
    }

    /** Called from main thread to begin audio generation for the source */
    private void commenceGenerating(final TtsAudioSource ttsSource)
    {
        Assert.assertEquals(ttsSource, _currentTtsSource);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                boolean success = _nativeVocalizer.speak(ttsSource._text, ttsSource._ssml, ttsSource._speed, ttsSource._pitch, ttsSource._volume, ttsSource._msPerChunk, new SpeakListener()
                {
                    @Override
                    public void onNewChunk(final AudioChunk chunk, final boolean done)
                    {
                        if (ttsSource._bufferCount >= 0)
                        {
                            final int deltaBufferedDuration = (ttsSource.getChunksAvailable() - ttsSource._bufferCount) * ttsSource._msPerChunk;
                            if (deltaBufferedDuration > 0)
                            {
                                // Too much audio has been accumulated and not being consumed yet, wait a little
                                Logger.debug(this, "Wait " + deltaBufferedDuration + "ms before generating more audio...");
                                ThreadUtils.sleep(deltaBufferedDuration);
                            }
                        }

                        // Now handle the audio from the main thread
                        _mainThreadHandler.post(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                if (ttsSource != _currentTtsSource)
                                    return;

                                if (ttsSource.isComplete())
                                {
                                    Logger.error(VocalizerBase.this, "TTS job was completed, but still in queue!");
                                    Assert.assertTrue(false);
                                    _currentTtsSource = null;
                                    return;
                                }

                                if (chunk != null)
                                {
                                    ttsSource.addChunk(chunk);
                                }

                                if (done)
                                {
                                    Logger.debug(VocalizerBase.this, "Finished generating TTS for '" + ttsSource._text + "'");
                                    ttsSource.complete(true);
                                }
                            }
                        });
                    }
                });

                if (!success)
                {
                    // TODO: report error better?
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            ttsSource.cancel();
                        }
                    });
                }
            }
        });
    }

    /**
     * Helper that converts a frequency in KHz to the exact matching
     * frequency in Hz.
     * @param freqKHz The frequency in KHz
     * @return The frequency in Hz
     */
    private int getFrequencyInHz(int freqKHz)
    {
        int freqHz = 0;

        switch(freqKHz) {
            case Vocalizer.Frequencies.FREQ_8KHZ:
                freqHz = AudioType.Frequency.FREQ_8KHZ;
                break;
            case Vocalizer.Frequencies.FREQ_11KHZ:
                freqHz = AudioType.Frequency.FREQ_11KHZ;
                break;
            case Vocalizer.Frequencies.FREQ_16KHZ:
                freqHz = AudioType.Frequency.FREQ_16KHZ;
                break;
            case Vocalizer.Frequencies.FREQ_22KHZ:
                freqHz = AudioType.Frequency.FREQ_22KHZ;
                break;
            case Vocalizer.Frequencies.UNKNOWN:
            default:
                // Defaulting to 22k audio
                freqHz = AudioType.Frequency.FREQ_22KHZ;
                break;
        }
        return freqHz;
    }

    /**
     * Called on MAIN thread exactly once in the constructor to obtain the
     * worker thread handler, if the worker thread handler wasn't given in the
     * constructor.
     * @return The worker thread handler.
     */
    protected abstract NMTHandler getWorkerThreadHandler();
}
