package com.nuance.dragon.toolkit.vocalizer;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.vocalizer.Vocalizer.Frequencies;

/**
 * Represents a Vocalizer configuration, which consists of one of the following:
 *  <br> - A language string
 *  <br> - A voice string
 *  <br> - A language and voice string
 */
public final class VocalizerConfig implements JSONCompliant
{
    private static final String KEY_LANGUAGE = "language";
    private static final String KEY_VOICE = "voice";
    private static final String KEY_FREQUENCY = "frequency";
    private static final String KEY_GENDER = "gender";
    
    final String language;
    final String voice;
    final int freqKhz;
    final int gender;

    /**
     * @param language See {@link Vocalizer.Languages} for supported languages.
     * @throws IllegalArgumentException if language is null
     */
    public VocalizerConfig(VocalizerLanguage language)
    {
        Checker.checkArgForNull("language", language);
        this.language = language.name;
        this.voice = null;
        this.freqKhz = Vocalizer.Frequencies.UNKNOWN;
        this.gender = Vocalizer.Genders.UNKNOWN;
    }

    /**
     * Sets voice to input and language to language associated with voice
     * @param voice See {@link Vocalizer.Voices} for supported languages.
     * @throws IllegalArgumentException if voice is null
     */
    public VocalizerConfig(VocalizerVoice voice)
    {
        Checker.checkArgForNull("voice", voice);
        this.language = voice.language.name;
        this.voice = voice.name;
        this.freqKhz = Vocalizer.Frequencies.UNKNOWN;
        this.gender = Vocalizer.Genders.UNKNOWN;
    }

    /**
     * Sets language to input and sets voice to null.
     * @param language A raw language string. Should come from a {@link VocalizerLanguage#name}.
     * @throws IllegalArgumentException if language is null or empty
     */
    public VocalizerConfig(String language)
    {
        Checker.checkStringArgForNullOrEmpty("language", language);
        this.language = language;
        this.voice = null;
        this.freqKhz = Vocalizer.Frequencies.UNKNOWN;
        this.gender = Vocalizer.Genders.UNKNOWN;
    }

    /**
     * Sets language and voice to inputs.
     * @param language A raw language string. Should come from a {@link VocalizerLanguage#name}.
     * May be null if voice is non-null.
     * @param voice A raw voice string. Should come from a {@link VocalizerVoice#name}. May be
     * null if language is non-null.
     * @throws IllegalArgumentException if none of the strings is non-null and non-empty
     */
    public VocalizerConfig(String language, String voice)
    {
        Checker.checkStringArgsForAllNullOrEmpty("language and voice", language, voice);
        this.language = language;
        this.voice = voice;
        this.freqKhz = Vocalizer.Frequencies.UNKNOWN;
        this.gender = Vocalizer.Genders.UNKNOWN;
    }

    /**
     * Sets language and frequency to inputs.
     * @param freqKhz See {@link Frequencies} for allowed values.
     * @throws IllegalArgumentException if language is null or empty
     * or the passed frequency is not supported.
     */
    public VocalizerConfig(String language, int freqKhz)
    {
        Checker.checkStringArgForNullOrEmpty("language", language);
        Checker.checkArgForCondition("freqKhz", "equals to one of the values from \"Vocalizer.Frequencies\" class",
                                     freqKhz == Vocalizer.Frequencies.UNKNOWN ||
                                     freqKhz == Vocalizer.Frequencies.FREQ_8KHZ ||
                                     freqKhz == Vocalizer.Frequencies.FREQ_11KHZ ||
                                     freqKhz == Vocalizer.Frequencies.FREQ_16KHZ ||
                                     freqKhz == Vocalizer.Frequencies.FREQ_22KHZ);
        this.language = language;
        this.voice = null;
        this.freqKhz = freqKhz;
        this.gender = Vocalizer.Genders.UNKNOWN;
    }

    /**
     * Sets language, voice and frequency to inputs.
     * @param language A raw language string. Should come from a {@link VocalizerLanguage#name}.
     * May be null if voice is non-null.
     * @param voice A raw voice string. Should come from a {@link VocalizerVoice#name}. May be
     * null if language is non-null.
     * @param freqKhz See {@link Vocalizer.Frequencies} for allowed values.
     * @throws IllegalArgumentException if none of the strings is non-null and non-empty
     * or the passed frequency is not supported.
     */
    public VocalizerConfig(String language, String voice, int freqKhz)
    {
        Checker.checkStringArgsForAllNullOrEmpty("language and voice", language, voice);
        Checker.checkArgForCondition("freqKhz", "equals to one of the values from \"Vocalizer.Frequencies\" class",
                                     freqKhz == Vocalizer.Frequencies.UNKNOWN ||
                                     freqKhz == Vocalizer.Frequencies.FREQ_8KHZ ||
                                     freqKhz == Vocalizer.Frequencies.FREQ_11KHZ ||
                                     freqKhz == Vocalizer.Frequencies.FREQ_16KHZ ||
                                     freqKhz == Vocalizer.Frequencies.FREQ_22KHZ);
        this.language = language;
        this.voice = voice;
        this.freqKhz = freqKhz;
        this.gender = Vocalizer.Genders.UNKNOWN;
    }

    /**
     * Sets language, voice and frequency to inputs.
     * @param language A raw language string. Should come from a {@link VocalizerLanguage#name}.
     * May be null if voice is non-null.
     * @param voice A raw voice string. Should come from a {@link VocalizerVoice#name}. May be
     * null if language is non-null.
     * @param freqKhz See {@link Vocalizer.Frequencies} for allowed values.
     * @param gender See {@link Vocalizer.Genders} for allowed values.
     * @throws IllegalArgumentException if none of the strings is non-null and non-empty
     * or the passed frequency is not supported.
     */
    public VocalizerConfig(String language, String voice, int freqKhz, int gender)
    {
        Checker.checkStringArgsForAllNullOrEmpty("language and voice", language, voice);
        Checker.checkArgForCondition("freqKhz", "equals to one of the values from \"Vocalizer.Frequencies\" class",
                                     freqKhz == Vocalizer.Frequencies.UNKNOWN ||
                                     freqKhz == Vocalizer.Frequencies.FREQ_8KHZ ||
                                     freqKhz == Vocalizer.Frequencies.FREQ_11KHZ ||
                                     freqKhz == Vocalizer.Frequencies.FREQ_16KHZ ||
                                     freqKhz == Vocalizer.Frequencies.FREQ_22KHZ);
        Checker.checkArgForCondition("gender", "equals to one of the values from \"Vocalizer.Genders\" class",
                gender == Vocalizer.Genders.UNKNOWN ||
                gender == Vocalizer.Genders.MALE ||
                gender == Vocalizer.Genders.FEMALE);
        this.language = language;
        this.voice = voice;
        this.gender = gender;
        this.freqKhz = freqKhz;
    }    
    
    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        o.tryPut(KEY_LANGUAGE, language);
        o.tryPut(KEY_VOICE, voice);
        o.tryPut(KEY_FREQUENCY, freqKhz);
        o.tryPut(KEY_GENDER, gender);
        return o;
    }

    public static VocalizerConfig createFromJSON(org.json.JSONObject json) throws JSONException
    {
        int freqKhz = json.optInt(KEY_FREQUENCY, Vocalizer.Frequencies.UNKNOWN);
        int gender = json.optInt(KEY_GENDER, Vocalizer.Genders.UNKNOWN);
        String language = json.optString(KEY_LANGUAGE, null);
        String voice = null;
        if (language == null)
            voice = json.getString(KEY_VOICE);
        else
            voice = json.optString(KEY_VOICE, null);
        return new VocalizerConfig(language, voice, freqKhz, gender);
    }
}
