package com.nuance.dragon.toolkit.vocalizer;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * A Vocalizer language, with the voices belonging to the language. Note that not
 * all languages are necessarily supported--supported languages and voices are
 * dependent on the original build configuration of the Vocalizer engine, and the
 * model files present on the device.
 */
public final class VocalizerLanguage implements JSONCompliant
{
    private static final String KEY_NAME = "name";

    /** The language name. This is the string passed to the Vocalizer engine. Must be non-null. */
    public final String name;
    /** A description of the language. Must be non-null */
    public final String description;
    /** The voices belonging to this language .Must be non-null */
    public final VocalizerVoice[] voices;

    VocalizerLanguage(String language, String desc, String[] voiceNames, String[] voiceDescriptions)
    {
        this.name = language;
        this.description = desc;
        this.voices = createVoices(voiceNames, voiceDescriptions, this);
    }

    @Override
    /**
     * @return The description
     */
    public String toString()
    {
        return description;
    }

    @Override
    /**
     * Returns true if the input Object is an instance of language and the name is deep equal to this name
     * @param item The item to compare. If null, returns false
     */
    public boolean equals(Object item)
    {
        if (item == null)
            return false;

        if (!(item instanceof VocalizerLanguage))
            return false;

        VocalizerLanguage lang = (VocalizerLanguage)item;
        return name.equals(lang.name);
    }

    @Override
    public int hashCode()
    {
        return name.hashCode();
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        // Only store 'name'
        JSONObject o = new JSONObject();
        o.tryPut(KEY_NAME, name);
        return o;
    }

    /**
     *
     * @param json The json object to create a language from
     * @return The newly created language
     * @throws JSONException if language name is unknown
     */
    public static VocalizerLanguage createFromJSON(org.json.JSONObject json) throws JSONException
    {
        // Don't create a new Language; rather, always make sure we return one of the valid ones
        String name = json.getString(KEY_NAME);
        for (VocalizerLanguage lang : LangSpecs.ALL_LANGUAGES)
        {
            if (lang.name.equals(name))
                return lang;
        }
        throw new JSONException("Invalid language name: " + name);
    }

    private static VocalizerVoice[] createVoices(String [] names, String[] descriptions, VocalizerLanguage language)
    {
        VocalizerVoice[] voices = new VocalizerVoice[names.length];
        for (int i = 0; i < names.length; i++)
        {
            voices[i] = new VocalizerVoice(names[i], descriptions[i], language);
        }
        return voices;
    }
}
