package com.nuance.dragon.toolkit.vocon;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import junit.framework.Assert;

import org.json.JSONException;

import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.core.WordList;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONArray;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * Represents a Vocon dynamic slot.
 */
public class DynamicSlot implements JSONCompliant
{
    private static final String JSON_KEY_ID = "id";
    private static final String JSON_KEY_WORDS = "words";
    private static final String JSON_KEY_DOMAIN = "domain";
    private static final String JSON_KEY_REBUILD_MODE = "rebuildmode";

    private final String _id;
    private final List<Word> _words;
    private final List<WordList> _wordLists;
    private final String _domain;
    private final int _rebuildMode;

    /**
     * The domains supported by Vocon.
     * <p>
     * The domain serves as a hint to the Vocon engine about how to tokenize text
     * and compute pronunciations using the passed CLC file.
     */
    public static final class Domains
    {
        /** Default */
        public static final String DYNAMICSLOT_NORMAL = "normal";

        /** Partial name matching */
        public static final String DYNAMICSLOT_NAME = "name";

        /** Music (artist) matching */
        public static final String DYNAMICSLOT_ARTIST = "artist";

        /** Music (album) matching */
        public static final String DYNAMICSLOT_ALBUM = "album";

        /** Music (title) matching */
        public static final String DYNAMICSLOT_TITLE = "title";
    }

    /**
     * The modes that should be applied when adding a dynamic slot to Vocon during
     * a grammar rebuild.
     */
    public static final class RebuildModes
    {
        /**
         * When a grammar is loaded, the definition of this DynamicSlot will replace any
         * previous definition.  The old value of the DynamicSlot will be deleted before
         * the DynamicSlot is added.
         */
        public static final int DYNAMICSLOT_REBUILD_FULL = 0;

        /**
         * If the DynamicSlot already exists, existing words will be updated (added or
         * deleted). Combine the previous definition and the new definition.
         */
        public static final int DYNAMICSLOT_REBUILD_UPDATE = 1;

        /**
         * If the DynamicSlot already exists, it will be skipped during the rebuild.  Keep
         * the previously loaded value and ignore the new definition.
         */
        public static final int DYNAMICSLOT_REBUILD_SKIP_EXISTING = 2;
    }

    /**
     * Iterates through all the words in this dynamic slot.
     */
    public static class WordIterator
    {
        private final LinkedList<Word> _words;
        private final LinkedList<WordList.WordIterator> _wordListIterators;
        private final LinkedList<WordList.WordIterator> _completedWordListIterators;
        private WordAction _next;

        private WordIterator(List<Word> words, List<WordList> wordLists, boolean fullRebuild)
        {
            _words = new LinkedList<Word>();
            _wordListIterators = new LinkedList<WordList.WordIterator>();
            _completedWordListIterators = new LinkedList<WordList.WordIterator>();

            if (words != null)
                _words.addAll(words);

            if (wordLists != null)
            {
                for (WordList list : wordLists)
                {
                    WordList.WordIterator it;
                    if (fullRebuild)
                        it = list.getFullIterator();
                    else
                        it = list.getModifiedIterator();

                    _wordListIterators.add(it);
                }
            }

            _next = getNextWord();
        }

        /**
         * Get the next word action.
         * @return The next word action, or null when the end has been reached.
         */
        public WordAction getNext()
        {
            WordAction next = _next;
            if (_next != null)
            {
                _next = getNextWord();
            }
            return next;
        }

        /**
         * @return True if there is a next.
         */
        public boolean hasNext()
        {
            return _next != null;
        }

        /**
         * Call when all the words have been consumed, and they have been applied.
         * @param success Whether or not the words have been successfully applied.
         */
        public void finished(boolean success)
        {
            Assert.assertTrue(_wordListIterators.isEmpty());

            for (WordList.WordIterator it : _completedWordListIterators)
            {
                if (success)
                    it.acceptChanges();
                else
                    it.discardChanges();
            }
        }

        private WordAction getNextWord()
        {
            if (!_words.isEmpty())
            {
                WordAction wa = null;
                Word w = _words.remove();
                if (w != null)
                    wa = new WordAction (w, true);
                return wa;
            }

            while (!_wordListIterators.isEmpty())
            {
                WordAction action = null;
                if (_wordListIterators.getFirst().hasNext())
                {
                    action = _wordListIterators.getFirst().next();
                }

                if (action != null)
                {
                    return action;
                }

                _completedWordListIterators.add(_wordListIterators.remove());
            }

            // Signal that the list is done
            return null;
        }
    }

    /**
     * @param id A string that identifies the DynamicSlot..
     */
    public DynamicSlot(String id)
    {
        this(id, Domains.DYNAMICSLOT_NORMAL, RebuildModes.DYNAMICSLOT_REBUILD_SKIP_EXISTING);
    }

    /**
     * @param id A string that identifies the DynamicSlot.
     * @param domain One of {@link Domains}.
     */
    public DynamicSlot(String id, String domain)
    {
        this(id, domain, RebuildModes.DYNAMICSLOT_REBUILD_SKIP_EXISTING);
    }

    /**
     * @param id A string that identifies the DynamicSlot.
     * @param domain One of {@link Domains}.
     * @param rebuildMode One of {@link RebuildModes}.
     */
    public DynamicSlot(String id, String domain, int rebuildMode)
    {
        Checker.checkStringArgsForAllNullOrEmpty("id", id);
        Checker.checkStringArgsForAllNullOrEmpty("domain", domain);

        _id = id;
        _words = new ArrayList<Word>();
        _wordLists = new ArrayList<WordList>();
        _domain = domain;
        _rebuildMode = rebuildMode;
    }

    /**
     * Create a copy of the dynamic slot.
     * @return A copy, never null.
     */
    public DynamicSlot copy()
    {
        DynamicSlot copy = new DynamicSlot(_id, _domain, _rebuildMode);
        copy._words.addAll(_words);
        copy._wordLists.addAll(_wordLists);
        return copy;
    }

    /**
     * Get the ID of the dynamic slot.
     *
     * @return The dynamic slot ID as passed to the constructor.
     */
    public String getId()
    {
        return _id;
    }

    /**
     * Get the domain of the dynamic slot.
     *
     * @return The dynamic domain as passed to the constructor.
     */
    public String getDomain()
    {
        return _domain;
    }

    /* (non-Javadoc)
     * @see java.lang.Object#hashCode()
     */
    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((_domain == null) ? 0 : _domain.hashCode());
        result = prime * result + ((_id == null) ? 0 : _id.hashCode());
        result = prime * result + _rebuildMode;
        result = prime * result
                + ((_wordLists == null) ? 0 : _wordLists.hashCode());
        result = prime * result + ((_words == null) ? 0 : _words.hashCode());
        return result;
    }

    /* (non-Javadoc)
     * @see java.lang.Object#hashCode()
     */
    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        DynamicSlot other = (DynamicSlot) obj;
        if (_domain == null) {
            if (other._domain != null)
                return false;
        } else if (!_domain.equals(other._domain))
            return false;
        if (_id == null) {
            if (other._id != null)
                return false;
        } else if (!_id.equals(other._id))
            return false;
        if (_rebuildMode != other._rebuildMode)
            return false;
        if (_wordLists == null) {
            if (other._wordLists != null)
                return false;
        } else if (!_wordLists.equals(other._wordLists))
            return false;
        if (_words == null) {
            if (other._words != null)
                return false;
        } else if (!_words.equals(other._words))
            return false;
        return true;
    }

    /**
     * Get the Always Rebuild status of this Dynamic Slot.
     *
     * @return the Rebuild Type for this Dynamic Slot.
     */
    public int getRebuildType()
    {
        if (_rebuildMode != RebuildModes.DYNAMICSLOT_REBUILD_FULL)
        {
            for (WordList list : _wordLists)
            {
                if (list.fullUpdateRequired())
                    return RebuildModes.DYNAMICSLOT_REBUILD_FULL;
            }
        }
        return _rebuildMode;
    }

    /**
     * Get an iterator for the words in this dynamic slot. May be called on any thread.
     * @param fullRebuild true if the full word list should always be returned; false
     * to automatically determine which words to return.
     * @return The iterator, never null.
     */
    public WordIterator getWords(boolean fullRebuild)
    {
        return new WordIterator(_words, _wordLists, fullRebuild || getRebuildType() == RebuildModes.DYNAMICSLOT_REBUILD_FULL); // TODO: inefficient?
    }

    /**
     * @return true if there is at least one WordList attached to this DynamicSlot.
     */
    public boolean hasWordList() {
        return !_wordLists.isEmpty();
    }

    /**
     * Attach a word list, which will serve as a source of content for this word
     * slot.
     * TODO: is it ok that this info is not included in the JSON representation?
     * @param wordList
     */
    public void attachWordList(WordList wordList)
    {
        if (!_wordLists.contains(wordList))
            _wordLists.add(wordList);
    }

    /**
     * Add a word (or phrase) to the dynamic slot.
     *
     * @param word The word to add.
     * <p>
     * If the word has a spoken form, it will only
     * be used for Japanese--the surface form will be the Kanji spelling, and
     * the spoken form will be the Kata-Kana spelling.
     */
    public void addWord(Word word)
    {
        Checker.checkArgForNull("word", word);
        _words.add(word);
    }

    /**
     * Add words to the dynamic slot.
     *
     * @param words The words to add
     */
    public void addWords(Iterable<Word> words)
    {
        Checker.checkArgForNull("words", words);
        for (Word w : words)
            _words.add(w);
    }

    /**
     * Remove words from the dynamic slot. This will only remove words that exactly
     * match both the surface and spoken form.
     *
     * @param words The words to remove.
     */
    public void removeWords(Iterable<Word> words)
    {
        Checker.checkArgForNull("words", words);

        for (Word w : words)
            _words.remove(w);
    }

    /**
     * Clear all words from the slot.
     */
    public void clear()
    {
        _words.clear();
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();

        o.tryPut(JSON_KEY_ID, _id);
        o.tryPut(JSON_KEY_DOMAIN, _domain);
        o.tryPut(JSON_KEY_REBUILD_MODE, _rebuildMode);
        JSONArray words = new JSONArray();
        for (Word w : _words)
        {
            words.put(w.toJSON());
        }
        o.tryPut(JSON_KEY_WORDS, words);
        return o;
    }

    public static DynamicSlot createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String id = json.getString(JSON_KEY_ID);
        String domain = json.getString(JSON_KEY_DOMAIN);
        int rebuildMode = json.getInt(JSON_KEY_REBUILD_MODE);
        DynamicSlot ret = new DynamicSlot(id, domain, rebuildMode);
        org.json.JSONArray words = json.getJSONArray(JSON_KEY_WORDS);
        int size = words.length();
        for (int i = 0; i < size; i++)
        {
            ret.addWord(Word.createFromJSON(words.getJSONObject(i)));
        }
        return ret;
    }
}
