package com.nuance.dragon.toolkit.vocon;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.json.JSONException;

import com.nuance.dragon.toolkit.core.WordList;
import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * Represents the Vocon grammar. Contains populated {@link VoconContext} objects.
 */
public class Grammar implements JSONCompliant
{
    private static final String KEY_CONTEXTS  = "contexts";
    private static final String KEY_SLOTS = "dynmslot";

    private final HashMap<String, VoconContext> _contexts;
    private final HashMap<String, DynamicSlot> _dynamicSlots;

    /**
     * Create an empty grammar.
     */
    public Grammar()
    {
        _contexts = new HashMap<String, VoconContext>();
        _dynamicSlots = new HashMap<String, DynamicSlot>();
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result
                + ((_contexts == null) ? 0 : _contexts.hashCode());
        result = prime * result
                + ((_dynamicSlots == null) ? 0 : _dynamicSlots.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        Grammar other = (Grammar) obj;
        if (_contexts == null) {
            if (other._contexts != null)
                return false;
        } else if (!_contexts.equals(other._contexts))
            return false;
        if (_dynamicSlots == null) {
            if (other._dynamicSlots != null)
                return false;
        } else if (!_dynamicSlots.equals(other._dynamicSlots))
            return false;
        return true;
    }

    /**
     * Construct a grammar with the specified DynamicSlot and VoconContext lists.
     *
     * @param contexts The list of VoconContext in this grammar.
     *                 Must not be empty or null.
     *                 The context list must not contain duplicate IDs.
     *
     *  @param slots The list of DynamicSlot in this grammar.
     *               DynamicSlot could be null and empty
     *               The DynamicSlot list must not contain duplicate IDs.
     *
     * @return a Grammar object if successful or null if the argument list is not valid.
     */
    public static Grammar createGrammar(List<VoconContext> contexts, List<DynamicSlot> slots)
    {
        Checker.checkArgForCondition("contexts", "not null or empty", contexts != null && !contexts.isEmpty());

        Grammar grammar = new Grammar();

        for (VoconContext c : contexts) {
            if (grammar._contexts.put(c.getId(), c.copy()) != null) {
                Logger.error(grammar, "Duplicate VoconContext ID in slots: " + c.getId());
                return null;
            }
        }

        if (slots != null) {
            for (DynamicSlot ds : slots) {
                if (grammar._dynamicSlots.put(ds.getId(), ds.copy()) != null) {
                    Logger.error(grammar, "Duplicate DynamicSlot ID in slots: " + ds.getId());
                    return null;
                }
            }
        }
        return grammar;
    }

    /**
     * Get the list of Vocon contexts.
     *
     * @return The contexts. Will never be null.
     */
    public List<VoconContext> getContexts()
    {
        List<VoconContext> contexts = new ArrayList<VoconContext>();
        for (VoconContext context : _contexts.values()) {
            contexts.add(context);
        }
        return contexts;
    }

    /**
     * Get a Vocon context.
     * @param name The context file name.
     *
     * @return a VoconContext object, or null if the name is not defined.
     */
    public VoconContext getContextByName(String name)
    {
        return _contexts.get(name);
    }

    /**
     * And the specified context to the grammar.
     *
     * @param context vocon context file to add.  May be null.
     * @param true if existing contexts with the same name should be replaced.
     *        false if existing contexts with the same place should be skipped.
     */
    void addContext(VoconContext context, boolean overwrite)
    {
        if (context == null) {
            return;
        }
        if (overwrite || !_contexts.containsKey(context.getId())) {
            _contexts.put(context.getId(), context.copy());
        }
    }

    /**
     * And the specified list of context to the grammar.
     *
     * @param constraints a list of vocon contexts to be added.  May be null or empty.
     * @param true if existing contexts with the same name should be replaced.
     *        false if existing contexts with the same place should be skipped.
     */
    void addContexts(List<VoconContext> contexts, boolean overwrite)
    {
        if (contexts == null) {
            return;
        }
        for (VoconContext c : contexts) {
            if (overwrite || !_contexts.containsKey(c.getId())) {
                _contexts.put(c.getId(), c.copy());
            }
        }
    }

    /**
     * Delete all Contexts from the Grammar.
     */
    void clearContexts() {
        _contexts.clear();
    }

    /**
     * Use this function to determine the IDs of the word lists required by this VoconContext, if any.
     *
     * @return A list of Word list IDs that are required by this object.
     * Will be an empty list (never null) if no word lists are required.
     */
    public Set<String> getRequiredWordListIds() {
        HashSet<String> requiredWordListIds = new HashSet<String>();
        for (String name : _dynamicSlots.keySet()) {
            requiredWordListIds.add(name);
        }
        return requiredWordListIds;
    }

    /**
     * Attach a word list to this Grammar.  Note that the name of the word list must
     * exactly match the name of a word slot defined in the grammar.
     *
     * @param wordList The word list to attach.  Must not be null.
     * @param id The name of the word slot where the list should be attached.  Must not be null or empty.
     * @return true if the list was attached, false if the list was not attached.
     */
    public boolean attachWordList(WordList wordList, String id)
    {
        return attachWordList(wordList, id, DynamicSlot.Domains.DYNAMICSLOT_NORMAL);
    }

    /**
     * Attach a word list to this Grammar.  Note that the name of the word list must
     * exactly match the name of a word slot defined in the grammar.
     *
     * @param wordList The word list to attach.  Must not be null.
     * @param id The name of the word slot where the list should be attached.  Must not be null or empty.
     * @param domain The CLC domain.  Must not be null or empty.
     * @return true if the list was attached, false if the list was not attached.
     */
    public boolean attachWordList(WordList wordList, String id, String domain)
    {
        Checker.checkArgForNull("wordList", wordList);
        Checker.checkStringArgForEmpty("id", id);
        Checker.checkStringArgForEmpty("domain", domain);

        DynamicSlot ds = _dynamicSlots.get(id);


        if (ds == null) {
            ds = new DynamicSlot(id, domain);
            _dynamicSlots.put(id, ds);
        }

        ds.attachWordList(wordList);
        return true;
    }

    /**
     * And the specified DynamicSlot to the context.
     *
     * @param slot a DynamicSlot to be added.  May be null.
     * @param true if existing DynamicSlot with the same name should be replaced.
     *        false if existing DynamicSlot with the same place should be skipped.
     */
    void addWordSlot(DynamicSlot slot, boolean overwrite)
    {
        if (slot == null) {
            return;
        }
        if (overwrite || !_dynamicSlots.containsKey(slot.getId())) {
            _dynamicSlots.put(slot.getId(), slot.copy());
        }
    }

    /**
     * And the specified DynamicSlots to the grammar.
     *
     * @param slots a list of DynamicSlots to be added.  May be null or empty.
     * @param true if existing DynamicSlots with the same name should be replaced.
     *        false if existing DynamicSlots with the same place should be skipped.
     */
    void addWordSlots(List<DynamicSlot> slots, boolean overwrite)
    {
        if (slots == null) {
            return;
        }
        for (DynamicSlot ds : slots) {
            if (overwrite || !_dynamicSlots.containsKey(ds.getId())) {
                _dynamicSlots.put(ds.getId(), ds.copy());
            }
        }
    }


    /**
     * Return a list containing all the DynamicSlot objects in this grammar.
     *
     * @return The vocon dynamic slots.  Will never be null, but may be an empty list.
     */
    public List<DynamicSlot> getDynamicSlots()
    {
        List<DynamicSlot> slots = new ArrayList<DynamicSlot>();
        for (DynamicSlot ds : _dynamicSlots.values()) {
            slots.add(ds);
        }
        return slots;
    }

    /**
     * Get the DynamicSlot object with the given name.
     *
     * @return a DynamicSlot object, or null if the name is not defined.
     */
    public DynamicSlot getDynamicSlotByName(String name)
    {
        return _dynamicSlots.get(name);
    }

    /**
     * Delete all dynamic slots from the Context.
     */
    void clearDynamicSlots() {
        _dynamicSlots.clear();
    }

    @Override
    public org.json.JSONObject toJSON() {
        JSONObject o = new JSONObject();
        o.tryPutList(KEY_CONTEXTS, getContexts());
        o.tryPutList(KEY_SLOTS, getDynamicSlots());
        return o;
    }

    /**
     * @param json A JSON object that represents a Grammar.
     *
     * @return A Grammar object if successful.  Throws JSONException if there is an error.
     */
    public static Grammar createFromJSON(org.json.JSONObject json) throws JSONException
    {
        org.json.JSONArray c = json.getJSONArray(KEY_CONTEXTS);

        List<VoconContext> contexts = new ArrayList<VoconContext>(c.length());
        for (int i = 0; i < c.length(); ++ i) {
            contexts.add(VoconContext.createFromJSON(c.getJSONObject(i)));
        }

        org.json.JSONArray s = json.getJSONArray(KEY_SLOTS);
        List<DynamicSlot> slots = new ArrayList<DynamicSlot>(s.length());
        for (int i = 0; i < s.length(); ++ i) {
            slots.add(DynamicSlot.createFromJSON(s.getJSONObject(i)));
        }

        Grammar grammar = Grammar.createGrammar(contexts, slots);
        if (grammar == null) {
            throw new JSONException("Error is DynamicSlot or VoconContext list.  Unable to create grammar.");
        }
        return grammar;
    }
}
