package com.nuance.dragon.toolkit.vocon;

import java.util.List;
import java.util.Map;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;

/**
 * An interface that provides access to the Vocon engine. All methods here
 * are expected to be called from a single worker thread unless otherwise specified.
 * TODO: Since this is not used directly by customers, add JAVADOC for all methods
 * (like NativeVocalizer.java) ?
 */
interface NativeVocon
{
   class WakeupMode
    {
       static final int DEFAULT = 0;

       static final int SEAMLESS = 1;
    }

    enum ContextError
    {
        OK, WARNING, FATAL
    }

    interface RecognitionListener
    {
        void onStartAudioPull();

        void onStopAudioPull();

        boolean onHasAudioChunk();

        AudioChunk onGetAudioChunk(int maxLen);

        void onCanceled();

        void onStartOfSpeech();

        void onEndOfSpeech();

        void onRecognitionResult(String json);

        void onSignalUpdate(int snr, int energy);

        void onSignalWarning(int code);
    }

    boolean listAvailableModels(List<NativeModelInfo> retModels);

    boolean listAvailableClcs(List<NativeClcInfo> retClcs);

    boolean createVocon(String acModFileName, String clcFileName);

    void destroyVocon();

    boolean startRecognizing(Map<VoconParam, Integer> fxParams, List<VoconContext> contexts, AudioType audioType, boolean resetChannel, RecognitionListener listener);

    boolean startEndpointing(Map<VoconParam, Integer> fxParams, AudioType audioType, boolean resetChannel, RecognitionListener listener);

    void cancelAsr();

    void resetCancelAsr();

    ContextError createAsrContext(VoconContext context);

    ContextError createWuwContext(VoconContext context, long[] quality);

    void destroyContext(VoconContext context);

    boolean hasWordClass(String wordClass);

    boolean createWordClass(String wordClass);

    void destroyWordClass(String wordClass);

    boolean addWordClassMember(String wordClass, int userID, String word, String domain);

    boolean removeWordClassMember(String wordClass, int userID, String word, int matchType);

    boolean addWordClassBatchMembers(String wordClass, int[] userIDs, String[] words, String domain, String delimiters);

    boolean removeWordClassMemeberInUTF8(String wordClass, int userID, String word, int matchType, String delimiters);

    boolean commitWordClass(String wordClassName);

    boolean createFuzzyMatchTable(String fileName, String uid);

    int createFuzzyMatchColumn(String tableUid, String columnName, int attributes);

    boolean addIntegerToColumn(String tableUid, int columnId, int value);

    boolean addStringToColumn(String tableUid, int columnId, String value);

    boolean addSpeakableToColumn(String tableUid, int columnId, String value);

    boolean commitFuzzyMatchRow(String tableUid);

    boolean destroyFuzzyMatchTable(String tableUid);

    void deleteCompiledContext(String fileName);

    String getMatchedFileName(String regex);

    boolean updateComputedParam(VoconContext context, long paramId);

    boolean compileWordsToContext(List<String> addedWords, String wordClass);

    boolean compileFileToContext(String fileName);

    void destroyDynamicWordContext(String wordClass);

    boolean stopRecognitionResults();

    /**
     * Load a saved Vocon grammar (called on worker thread).
     * The Native Vocon library only supports loading when Vocon is empty.
     * If a state has previously been loaded, or if grammar has been loaded,
     * this function will fail.
     *
     * @param fileName The file to load from.
     * @return The restored grammar object, or null if an error occurred.
     */
    Grammar loadVocon(String fileName);

    /**
     * Save the current Vocon grammar (called on worker thread).
     * @param fileName The file to save to.
     * @param grammar
     * @return true if successful else return flase
     */
    boolean saveVocon(String fileName, Grammar grammar);

    boolean loadResource(String fileName);

    void releaseResource(String fileName);

    /**
     * Enable or disable verbose Vocon logging. (called from worker thread)
     *
     * @param verbose True if verbose (detailed) l
     *
     */
    void enableVoconVerboseLogging(boolean verbose);

    boolean loadAcModAdaptData(String fileName);

    boolean saveAcModAdaptData(String fileName);

    boolean loadFxAdaptData(String fileName);

    boolean saveFxAdaptData(String fileName);

    boolean loadSpeakerProfile(String fileName);

    boolean saveSpeakerProfile(String fileName);
}
