package com.nuance.dragon.toolkit.vocon;

import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.json.JSONObject;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.util.JSONUtils;
import com.nuance.dragon.toolkit.util.internal.NativeUtils;

class NativeVoconImpl implements NativeVocon {

    static boolean LOAD_LIBRARY_SUCCESS = true;
    static {
        boolean armV7         = NativeUtils.isArmV7();
        boolean neonSupported = NativeUtils.isNeonSupported();
        boolean x86           = NativeUtils.isX86();
        Logger.debug(NativeVoconImpl.class, "armV7:" + armV7 + ", neonSupported:" + neonSupported + ", x86:" + x86);

        try {
            if (armV7 && neonSupported)
                System.loadLibrary("dmt_vocon_neon");
            else
                System.loadLibrary("dmt_vocon");
        }
        catch (UnsatisfiedLinkError e) {
            Logger.error(NativeVoconImpl.class, "Failed to load native library.", e);
            LOAD_LIBRARY_SUCCESS = false;
        }
    }

    private native int voconCreate(long[] vocon, FileManager fileManager, String acModFileName, String clcFileName, InputDeviceListener listener);
    private native void voconDestroy(long vocon);
    private native int voconStartRecognizing(long vocon, long[] contexts, int freqHz, byte[] sse, boolean resetChannel);
    private native int voconStartEndpointing(long vocon, int freqHz, byte[] sse, boolean resetChannel);
    private native void voconCancelAsr(long vocon);
    private native void voconResetCancelAsr(long vocon);
    private native int voconSetFxParam(long vocon, int key, int value, String name);
    private native int voconActivateRule(long context, String rule);
    private native int voconDeactivateRules(long context);
    private native int voconSetTableQuery(long context, String query, int activationValue);

    private native long voconCreateFuzzyMatchTable(long vocon, String fileName);
    private native int voconCreateFuzzyMatchColumn(long table, String columnName, int attributes);
    private native int voconAddIntegerToColumn(long table, int columnId, int value);
    private native int voconAddStringsToColumn(long table, int columnId, String[] values);
    private native int voconAddSpeakablesToColumn(long vocon, long table, int columnId, String[] values);
    private native int voconCommitFuzzyMatchRow(long table);
    private native int voconDestroyFuzzyMatchTable(long table);

    private native int voconLoadAcModAdaptData(long vocon, String fileName);
    private native int voconSaveAcModAdaptData(long vocon, String fileName);

    private native int voconLoadFxAdaptData(long vocon, String fileName);
    private native int voconSaveFxAdaptData(long vocon, String fileName);

    private native int voconLoadSpeakerProfile(long vocon, String fileName);
    private native int voconSaveSpeakerProfile(long vocon, String fileName);

    private native long voconCreateAsrContext(long vocon, String fileName);
    private native long voconCreateWuwContext(long vocon, int mode, String grammarID, String phrase, long[] valid);
    private native String voconGetGrammarName(long voconContext);
    private native int voconSetCtxParam(long hostContext, int key, int value, String name);
    private native int voconDestroyContext(long vocon, long hostContext);
    private native long voconCreateWordClass(long vocon);
    private native int voconDestroyWordClass(long wordClass);
    private native int voconAddWord(long vocon, long wordClass, int userID, String word, String domain);
    private native int voconRemoveWord(long wordClass, int userID, String word, int matchType);
    private native int voconAddWords(long vocon, long wordClass, int[] userIDs, byte[][] words, String domain, String delimiters);
    private native int voconRemoveWordInUTF8(long wordClass, int userID, byte[] word, int matchType, String delimiters);
    private native int voconCommitWordClass(long wordClass);
    private native int voconAttachWordClass(long hostContext, String slotName, long wordClass);
    private native int voconDetachWordClass(long hostContext, String slotName, long wordClass);
    private native long voconLoadWordClass(long vocon, String fileName);
    private native int voconSaveWordClass(long vocon, long wordClass, String fileName);
    private native int voconAttachSubContext(long hostContext, String slotAssignments, long subContext);
    private native int voconDetachSubContext(long hostContext, String slotName);
    private native int voconLoadResource(long vocon, String fileName, long[] resource);
    private native void voconReleaseResource(long vocon, long resource);
    private native void voconEnableVerboseLogging(boolean verbose);
    private native int voconUpdateComputedParam(long context, long paramId);
    private native int voconGram2CompileToXlf(long vocon, String fileName);
    private native int voconGram2CompileXlfToCtx(long vocon, String fileName);
    private native int voconStopRecognitionResults(long vocon);

    public interface InputDeviceListener
    {
        void onStartAudio();

        void onStopAudio();

        void onProcessedAudio(int snr, int energy);

        void onAbnormalAudio(int code);

        boolean onHasAudio();

        int onGetAudio(short[] audio);

        void onCanceled();

        void onStartOfSpeech();

        void onEndOfSpeech();

        void onRecognitionResult(String json);
    }

    /* Note: The value of these constants must match the values in ngf_vocon.h */
    public class Status {
        public static final int OK = 0;                      /*!< Normal return. */
        public static final int OOM = 1;                     /*!< Heap exhausted, or internal buffer full.*/
        public static final int STATE_ERROR = 2 ;            /*!< The requested operation cannot be performed because the object is not in the proper state.*/
        public static final int ARGUMENT_ERROR = 3;          /*!< An invalid argument was passed to native Vocon. */
        public static final int INVALID_DATA = 4;            /*!< A stream contains invalid data. */
        public static final int VERSION_ERROR = 5;           /*!< A function has received the correct type of data, but cannot handle the version.*/
        public static final int CALLBACK_ERROR = 6;          /*!< A user callback has returned an error. */
        public static final int UNSUPPORTED_OPERATION = 7;     /*!< A requested feature is not supported. */
        public static final int OTHER_ERROR = 8;               /*!< An error occurred.*/
    }

    // TODO: split on a Pattern for more efficiency?
    private static final String FUZZY_MATCH_SPLIT_REGEX = ";";

    private FileManager _fileManager;
    private long voconInstance;
    private final HashMap<String, Long> _wordClasses;
    private final HashMap<String, Long> _resources;
    private final HashMap<String, Long> _fuzzyMatchTables;
    private final HashMap<String, Long> _wordContexts;
    private RecognitionListener _listener;

    public NativeVoconImpl(FileManager fileManager)
    {
        _fileManager = fileManager;
        _wordClasses = new HashMap<String, Long>();
        _resources = new HashMap<String, Long>();
        _fuzzyMatchTables = new HashMap<String, Long>();
        _wordContexts = new HashMap<String, Long>();
    }

    @Override
    public boolean listAvailableModels(List<NativeModelInfo> retModels) {
        Pattern filePattern = Pattern.compile(LangSpecs.ACMODFILE_REGEX);

        String[] files = _fileManager.findFiles(filePattern);
        if (files != null)
        {
            for (String file : files)
            {
                Matcher matcher = filePattern.matcher(file);
                if (matcher.matches())
                {
                    String lang = LangSpecs.getMatchedAcmodFileLanguage(matcher);
                    String size = LangSpecs.getMatchedAcmodFileSizeKb(matcher);

                    if (lang != null && size != null)
                    {
                        retModels.add(new NativeModelInfo(lang, Integer.parseInt(size)));
                    }
                }
            }
        }

        return true;
    }

    @Override
    public boolean listAvailableClcs(List<NativeClcInfo> retClcs) {
        Pattern filePattern = Pattern.compile(LangSpecs.CLCFILE_REGEX);

        String[] files = _fileManager.findFiles(filePattern);
        if (files != null)
        {
            for (String file : files)
            {
                Matcher matcher = filePattern.matcher(file);
                if (matcher.matches())
                {
                    String lang = LangSpecs.getMatchedClcFileLanguage(matcher);
                    String domain = LangSpecs.getMatchedClcFileDomain(matcher);

                    if (lang != null && domain != null)
                    {
                        retClcs.add(new NativeClcInfo(lang, domain));
                    }
                }
            }
        }

        return true;
    }

    @Override
    public boolean createVocon(String acModFileName, String clcFileName)
    {
        boolean success = false;

        // Destroy Vocon if it already exists
        destroyVocon();

        if (acModFileName != null && clcFileName != null)
        {
            // START JNI_CODE
            long[] voconTemp = new long[1];
            int status = voconCreate(voconTemp, _fileManager, acModFileName, clcFileName, new InputDeviceListener()
            {
                @Override
                public void onStartAudio()
                {
                    Logger.debug(this, "In onStartAudio().");

                    if (_listener != null)
                        _listener.onStartAudioPull();
                }

                @Override
                public void onStopAudio()
                {
                    Logger.debug(this, "In onStopAudio().");

                    if (_listener != null)
                        _listener.onStopAudioPull();
                }

                @Override
                public void onProcessedAudio(int snr, int energy)
                {
                    Logger.debug(this, "In onProcessedAudio().");

                    if (_listener != null)
                        _listener.onSignalUpdate(snr, energy);
                }

                @Override
                public void onAbnormalAudio(int code)
                {
                    Logger.debug(this, "In onAbnormalAudio().");

                    if (_listener != null)
                        _listener.onSignalWarning(code);
                }

                @Override
                public boolean onHasAudio()
                {
                    boolean hasAudio = false;

                    if (_listener != null)
                        hasAudio = _listener.onHasAudioChunk();

                    return hasAudio;
                }

                @Override
                public int onGetAudio(short[] audio)
                {
                    AudioChunk chunk;
                    int filled = 0;

                    if (_listener != null)
                    {
                        chunk = _listener.onGetAudioChunk(audio.length);
                        if (chunk != null)
                        {
                            filled = chunk.audioShorts.length;
                            System.arraycopy(chunk.audioShorts, 0, audio, 0, filled);
                        }
                    }

                    return filled;
                }

                @Override
                public void onCanceled()
                {
                    Logger.debug(this, "In onCanceled().");

                    if (_listener != null)
                        _listener.onCanceled();
                }

                @Override
                public void onStartOfSpeech()
                {
                    Logger.debug(this, "In onStartOfSpeech().");

                    if (_listener != null)
                        _listener.onStartOfSpeech();
                }

                @Override
                public void onEndOfSpeech()
                {
                    Logger.debug(this, "In onEndOfSpeech().");

                    if (_listener != null)
                        _listener.onEndOfSpeech();
                }

                @Override
                public void onRecognitionResult(String json)
                {
                    Logger.debug(this, "In onRecognitionResult().");

                    if (_listener != null)
                        _listener.onRecognitionResult(json);
                }
            });

            if (status == Status.OK && voconTemp[0] != 0)
            {
                Logger.debug(this, "Created native Vocon");
                voconInstance = voconTemp[0];
                success = true;

            }
            else
            {
                Logger.error(this, "Unable to create native Vocon (error status = " + status + ").");
            }
        }
        return success;
    }

    @Override
    public void destroyVocon() {
        if (voconInstance == 0) {
            Logger.warn(this, "Error destroy Vocon: native Vocon does not exist!");
            return;
        }

        // Release loaded resources (if any)
        for (Entry<String,Long> resource : _resources.entrySet())
            voconReleaseResource(voconInstance, resource.getValue());
        _resources.clear();

//        voconUnloadGrammar(voconInstance);

        voconDestroy(voconInstance);
        voconInstance = 0;
        Logger.debug(this, "Destroyed native Vocon");

    }

    @Override
    public boolean startRecognizing(Map<VoconParam, Integer> fxParams, List<VoconContext> contexts, AudioType audioType, boolean resetChannel, RecognitionListener listener)
    {
        if (voconInstance == 0)
        {
            Logger.error(this, "Error starting vocon recognition: native Vocon does not exist.");
            return false;
        }

        if (contexts == null || contexts.isEmpty())
        {
            Logger.error(this, "Error starting vocon recognition: no recognition context available.");
            return false;
        }

        long[] activeContexts = new long[contexts.size()];
        for (int idx = 0; idx < contexts.size(); idx++)
        {
            Long context = contexts.get(idx).getJniHandle();
            if (context == null) {
                Logger.error(this, "Error starting vocon recognition: context does not exist.");
                return false;
            }
            activeContexts[idx] = context;
        }

        if (fxParams != null && fxParams.size() > 0)
        {
            for (Entry<VoconParam, Integer> param: fxParams.entrySet())
            {
                VoconParam key = param.getKey();
                Integer value = fxParams.get(key);
                if (key.type == VoconParam.Type.FRONT_END && voconSetFxParam(voconInstance, key.id, value, key.name) != Status.OK) {
                    Logger.error(this, "Error starting vocon recognition: unable to set parameter (" + Integer.toHexString(key.id) + ", " + value + ") ");
                    return false;
                }
            }
        }

        _listener = listener;
        return (Status.OK == voconStartRecognizing(voconInstance, activeContexts, audioType.frequency, audioType.sse, resetChannel));
    }

    @Override
    public boolean startEndpointing(Map<VoconParam, Integer> fxParams, AudioType audioType, boolean resetChannel, RecognitionListener listener)
    {
        if (voconInstance == 0)
        {
            Logger.error(this, "Error starting vocon recognition: native Vocon does not exist.");
            return false;
        }

        if (fxParams != null && fxParams.size() > 0)
        {
            for (Entry<VoconParam, Integer> param: fxParams.entrySet())
            {
                VoconParam key = param.getKey();
                Integer value = fxParams.get(key);
                if (key.type == VoconParam.Type.FRONT_END && voconSetFxParam(voconInstance, key.id, value, key.name) != Status.OK) {
                    Logger.error(this, "Error starting vocon recognition: unable to set parameter (" + Integer.toHexString(key.id) + ", " + value + ") ");
                    return false;
                }
            }
        }

        _listener = listener;
        return (Status.OK == voconStartEndpointing(voconInstance, audioType.frequency, audioType.sse, resetChannel));
    }

    @Override
    public void cancelAsr()
    {
        if (voconInstance != 0)
        {
            voconCancelAsr(voconInstance);
        }
    }

    @Override
    public void resetCancelAsr()
    {
        if (voconInstance != 0)
        {
            voconResetCancelAsr(voconInstance);
        }
    }

    @Override
    public ContextError createAsrContext(VoconContext context)
    {
        if (context != null && voconInstance != 0)
        {
            String id = context.getId();

            // Create native context
            long contextHandle = voconCreateAsrContext(voconInstance, id);
            if (contextHandle != 0)
            {
                context.setJniHandle(contextHandle);
                String grammarName = voconGetGrammarName(contextHandle);
                if (grammarName == null)
                    grammarName = id;
                context.setGrammarName(grammarName);

                boolean stepError = false;

                //Attach eventual word slots and online compiled dynamic words as context
                List<String> slotIds = context.getSlotIds();
                if (slotIds != null && slotIds.size() > 0)
                {
                    for (String slotId : slotIds)
                    {
                        // word slot
                        String fullSlotName = getFullSlotName(grammarName, slotId);
                        if (Status.OK != voconAttachWordClass(contextHandle, fullSlotName, _wordClasses.get(slotId)))
                        {
                            Logger.error(this, "Create ASR context: Unable to attach slot " + slotId);
                            stepError = true;
                            break;
                        }

                        // word context
                        Long wordContextObj = _wordContexts.get(slotId);
                        if (wordContextObj != null)
                        {
                            if (Status.OK != voconAttachSubContext(context.getJniHandle(), fullSlotName, wordContextObj))
                            {
                                Logger.error(this, "Create ASR context: Unable to attach word context to" + slotId);
                                stepError = true;
                                break;
                            }
                        }
                    }
                }

                // Step-check / clean-up
                if (stepError)
                {
                    destroyContext(context);
                    return ContextError.FATAL;
                }

                // recursively add subcontexts
                stepError = !addSubContext(context);

                // Step-check / clean-up
                if (stepError)
                {
                    destroyContext(context);
                    return ContextError.FATAL;
                }

                // Activate eventual rules (constrained recognition only)
                List<String> rules = context.getRules();
                if (rules != null && rules.size() > 0)
                {
                    if (Status.OK != voconDeactivateRules(contextHandle))
                    {
                        Logger.error(this, "Create ASR context: Unable to deactivate rules");
                        stepError = true;
                    }
                    else
                    {
                        for (String rule : rules)
                        {
                            String fullRuleName = getFullSlotName(grammarName, rule);
                            if (Status.OK != voconActivateRule(contextHandle, fullRuleName))
                            {
                                Logger.error(this, "Create ASR context: Unable to activate rule " + rule);
                                stepError = true;
                                break;
                            }
                        }
                    }
                }

                // Step-check / clean-up
                if (stepError)
                {
                    destroyContext(context);
                    return ContextError.FATAL;
                }

                // Set table query (fuzzy-match recognition only)
                List<VoconTableQuery> tableQueries = context.getTableQueries();
                if (tableQueries != null && tableQueries.size() > 0)
                {
                    for (VoconTableQuery tableQuery : tableQueries)
                    {
                        if (Status.OK != voconSetTableQuery(contextHandle, tableQuery.query, tableQuery.activationValue))
                        {
                            Logger.error(this, "Create ASR context: Unable to set table query " + tableQuery);
                            stepError = true;
                            break;
                        }
                    }
                }

                // Step-check / clean-up
                if (stepError)
                {
                    destroyContext(context);
                    return ContextError.FATAL;
                }

                // Set parameters
                Map<VoconParam, Integer> params = context.getParams();
                if (params != null && params.size() > 0)
                {
                    for (Entry<VoconParam, Integer> param: params.entrySet())
                    {
                        VoconParam key = param.getKey();
                        Integer value = params.get(key);
                        if (key.type == VoconParam.Type.CONTEXT && voconSetCtxParam(contextHandle, key.id, value, key.name) != Status.OK)
                        {
                            Logger.error(this, "Create ASR context: Unable to set parameter " + param);
                            stepError = true;
                            break;
                        }
                    }
                }

                // Step-check / clean-up
                if (stepError)
                {
                    destroyContext(context);
                    return ContextError.FATAL;
                }
                return ContextError.OK;
            }
            Logger.error(this, "Create ASR context: Unable to create native context.");
            return ContextError.FATAL;
        }
        Logger.error(this, "Create ASR context: Invalid state.");
        return ContextError.FATAL;
    }

    @Override
    public ContextError createWuwContext(VoconContext context, long[] quality)
    {
        if (voconInstance != 0 && context != null) {
            long contextHandle = voconCreateWuwContext(voconInstance, context.getWakeupMode(), context.getId(), context.getWakeupPhrase(), quality);
            if (contextHandle != 0)
            {
                context.setJniHandle(contextHandle);
                boolean stepError = false;

                // Set parameters
                Map<VoconParam, Integer> params = context.getParams();
                if (params != null && params.size() > 0)
                {
                    for (Entry<VoconParam, Integer> param: params.entrySet())
                    {
                        VoconParam key = param.getKey();
                        Integer value = params.get(key);
                        if (key.type == VoconParam.Type.CONTEXT && voconSetCtxParam(contextHandle, key.id, value, key.name) != Status.OK)
                        {
                            Logger.error(this, "Create ASR context: Unable to set parameter " + param);
                            stepError = true;
                            break;
                        }
                    }
                }

                // Step-check / clean-up
                if (stepError)
                {
                    destroyContext(context);
                    return ContextError.FATAL;
                }
                return quality[0] == 1 ? ContextError.OK : ContextError.WARNING;
            }
        }
        return ContextError.FATAL;
    }

    @Override
    public void destroyContext(VoconContext context)
    {
        if (voconInstance != 0 && context != null)
        {
            // Detach previously attached slots and dynamic word contexts, if any
            List<String> slotIds = context.getSlotIds();
            if (slotIds != null && slotIds.size() > 0)
            {
                for (String slotId : slotIds)
                {
                    // word slot
                    long slotHandle = _wordClasses.get(slotId);
                    String slotName = getFullSlotName(context.getGrammarName(), slotId);
                    voconDetachWordClass(context.getJniHandle(), slotName, slotHandle);

                    // word context
                    Long wordContextHandle = _wordContexts.get(slotId);
                    if (wordContextHandle != null)
                    {
                        if (Status.OK != voconDetachSubContext(context.getJniHandle(), slotName))
                        {
                            Logger.warn(this, "Destroy ASR context: Unable to detach word context from" + slotId);
                        }
                    }
                }
            }

            // recursively remove subcontexts
            removeSubContext(context);

            // Release native context
            voconDestroyContext(voconInstance, context.getJniHandle());
            context.setJniHandle(0);
        }
    }

    @Override
    public boolean createWordClass(String wordClass)
    {
        if (voconInstance != 0) {
            long obj = voconCreateWordClass(voconInstance);
            if (obj != 0) {
                _wordClasses.put(wordClass, Long.valueOf(obj));
                return true;
            }
        }

        return false;
    }

    @Override
    public void destroyWordClass(String wordClass)
    {
        Long obj = _wordClasses.get(wordClass);
        if (obj != null) {
            _wordClasses.remove(wordClass);
            voconDestroyWordClass(obj.longValue());
        }
    }

    @Override
    public void destroyDynamicWordContext(String wordClass) {
        Long wordContextobj = _wordContexts.get(wordClass);
        if (wordContextobj != null) {
            _wordContexts.remove(wordClass);
            voconDestroyContext(voconInstance, wordContextobj);
        }
    }

    @Override
    public boolean hasWordClass(String wordClass)
    {
        if (voconInstance != 0) {
            return _wordClasses.containsKey(wordClass);
        }
        return false;
    }

    @Override
    public boolean addWordClassMember(String wordClass, int userID, String word, String domain)
    {
        if (voconInstance != 0) {
            Long obj = _wordClasses.get(wordClass);
            if (obj != null)
            {
                return (Status.OK == voconAddWord(voconInstance, obj.longValue(), userID, word, domain));
            }
        }
        return false;
    }

    @Override
    public boolean removeWordClassMember(String wordClass, int userID, String word, int matchType)
    {
        Long obj = _wordClasses.get(wordClass);
        if (obj != null)
        {
            if (Status.OK != voconRemoveWord(obj.longValue(), userID, word, matchType))
            {
                Logger.error(this, "removeWordClassMember: voconAttachWordClass error!");
                return false;
            }
            return true;
        }
        return false;
    }

    @Override
    public boolean addWordClassBatchMembers(String wordClass, int[] userIDs, String[] words, String domain, String delimiters)
    {
        if (voconInstance != 0) {
            Long obj = _wordClasses.get(wordClass);
            if (obj != null)
            {
                byte[][] utf8Bytes = new byte[words.length][];
                for (int i = 0; i < words.length; i++)
                {
                    try
                    {
                        utf8Bytes[i] = words[i].getBytes("UTF-8");
                    } catch (UnsupportedEncodingException e)
                    {
                        Logger.error(this, "addWordClassBatchMembers: ", e);
                        return false;
                    }
                }
                return (Status.OK == voconAddWords(voconInstance, obj.longValue(), userIDs, utf8Bytes, domain, delimiters));
            }
        }
        return false;
    }

    @Override
    public boolean removeWordClassMemeberInUTF8(String wordClass, int userID, String word, int matchType, String delimiters)
    {
        if (voconInstance != 0) {
            Long obj = _wordClasses.get(wordClass);
            if (obj != null)
            {
                byte[] utf8Byte = null;
                try
                {
                    utf8Byte = word.getBytes("UTF-8");
                } catch (UnsupportedEncodingException e)
                {
                    Logger.error(this, "removeWordClassMemeberInUTF8: ", e);
                    return false;
                }
                if (Status.OK != voconRemoveWordInUTF8(obj.longValue(), userID, utf8Byte, matchType, delimiters))
                {
                    Logger.error(this, "removeWordClassMemeberInUTF8: voconAttachWordClass error!");
                    return false;
                }
                return true;
            }
        }
        return false;
    }

    @Override
    public boolean commitWordClass(String wordClass)
    {
        Long obj = _wordClasses.get(wordClass);
        if (obj != null) {
            return (Status.OK == voconCommitWordClass(obj.longValue()));
        }
        return false;
    }

    @Override
    public boolean compileWordsToContext(List<String> addedWords, String wordClass)
    {
        int status = Status.OK;

        if (voconInstance == 0)
        {
            Logger.error(this, "Error online compiling dyanmic words: native Vocon does not exist.");
            return false;
        }

        String fileName = getWordGrammarFileName(wordClass);
        if (fileName == null) return false;

        // destroy previous compiled context
        Long wordContextHandle = _wordContexts.get(wordClass);
        if (wordContextHandle != null)
        {
            _wordContexts.remove(wordClass);
            voconDestroyContext(voconInstance, wordContextHandle);
        }

        Logger.debug(this, "Writing to file " + fileName + ".bnf");
        FileWriter os = _fileManager.openFileForWritingChar(fileName + ".bnf");
        if (os != null)
        {
            // use BufferedWriter to limit I/O operations
            BufferedWriter bf = new BufferedWriter(os);
            if (bf != null)
            {
                try
                {
                    bf.write("#BNF+EMV2.1;");
                    bf.newLine();
                    bf.write("!grammar " + fileName + ";");
                    bf.newLine();
                    bf.write("!start <" + wordClass + ">;");
                    bf.newLine();
                    bf.newLine();
                    bf.write("<" + wordClass + ">:");
                    bf.newLine();
                    int i = 0;
                    for (String word : addedWords)
                    {
                        i++;
                        bf.write(word + ((i == addedWords.size()) ? ";" : "|"));
                        bf.newLine();
                    }
                    bf.newLine();
                }
                catch (IOException e)
                {
                    Logger.error(this, "Error writing to file " + fileName + ".bnf", e);
                    status = Status.OTHER_ERROR;
                }
                try { bf.close(); } catch (IOException e) { status = Status.OTHER_ERROR; }
            }

            try { os.close(); } catch (IOException e) { status = Status.OTHER_ERROR; }
        }
        else
            status =  Status.OTHER_ERROR;

        if (status == Status.OK)
        {
            status = voconGram2CompileToXlf(voconInstance, fileName);

            if (status == Status.OK)
            {
                status = voconGram2CompileXlfToCtx(voconInstance, fileName);
                if (status == Status.OK)
                {
                    // We are sure that .fcf is created
                    long newWordContextHandle = voconCreateAsrContext(voconInstance, fileName + ".fcf");
                    if (newWordContextHandle != 0) {
                        _wordContexts.put(wordClass, Long.valueOf(newWordContextHandle));
                    }
                    else
                        status = Status.OTHER_ERROR;
                }
            }
        }

        // whether context object is created or not, we delete the related files
        _fileManager.delete(fileName + ".bnf");
        _fileManager.delete(fileName + ".xlf");
        _fileManager.delete(fileName + ".fcf");

        return (status == Status.OK ? true : false);
    }

    @Override
    public boolean compileFileToContext(String fileName)
    {
        if (voconInstance == 0)
        {
            Logger.error(this, "Error online compiling file: native Vocon does not exist.");
            return false;
        }

        int status = voconGram2CompileToXlf(voconInstance, fileName);

        if (status == Status.OK)
        {
            status = voconGram2CompileXlfToCtx(voconInstance, fileName);
        }

        return (status == Status.OK ? true : false);
    }

    @Override
    public boolean createFuzzyMatchTable(String fileName, String tableUid)
    {
        if (voconInstance != 0) {
            long table = voconCreateFuzzyMatchTable(voconInstance, fileName);
            if (table != 0) {
                _fuzzyMatchTables.put(tableUid, table);
                return true;
            }
        }
        return false;
    }

    @Override
    public int createFuzzyMatchColumn(String tableUid, String columnName, int attributes)
    {
        Long table = _fuzzyMatchTables.get(tableUid);
        if (table != null)
            return voconCreateFuzzyMatchColumn(table, columnName, attributes);
        return -1;
    }

    @Override
    public boolean addIntegerToColumn(String tableUid, int columnId, int value)
    {
        boolean success = false;
        Long table = _fuzzyMatchTables.get(tableUid);
        if (table != null && columnId != -1)
            success = (Status.OK == voconAddIntegerToColumn(table, columnId, value));
        return success;
    }

    @Override
    public boolean addStringToColumn(String tableUid, int columnId, String value)
    {
        boolean success = false;
        Long table = _fuzzyMatchTables.get(tableUid);
        if (table != null && columnId != -1)
            success = (Status.OK == voconAddStringsToColumn(table, columnId, value.split(FUZZY_MATCH_SPLIT_REGEX)));
        return success;
    }

    @Override
    public boolean addSpeakableToColumn(String tableUid, int columnId, String value)
    {
        boolean success = false;
        Long table = _fuzzyMatchTables.get(tableUid);
        if (voconInstance != 0 && table != null && columnId != -1)
            success = (Status.OK == voconAddSpeakablesToColumn(voconInstance, table, columnId, value.split(FUZZY_MATCH_SPLIT_REGEX)));
        return success;
    }

    @Override
    public boolean commitFuzzyMatchRow(String tableUid)
    {
        boolean success = false;
        Long table = _fuzzyMatchTables.get(tableUid);
        if (table != null)
            success = (Status.OK == voconCommitFuzzyMatchRow(table));
        return success;
    }

    @Override
    public boolean destroyFuzzyMatchTable(String tableUid)
    {
        boolean success = false;
        Long table = _fuzzyMatchTables.get(tableUid);
        if (table != null)
            success = (Status.OK == voconDestroyFuzzyMatchTable(table));
        _fuzzyMatchTables.remove(tableUid);
        return success;
    }

    @Override
    public void deleteCompiledContext(String fileName)
    {
        if (fileName != null)
            _fileManager.delete(fileName);
    }

    @Override
    public String getMatchedFileName(String regex)
    {
        Pattern filePattern = Pattern.compile(regex);

        String[] files = _fileManager.findFiles(filePattern);
        if (files != null)
        {
            for (String file : files)
            {
                return file;
            }
        }
        return null;
    }

    @Override
    public Grammar loadVocon(String fileName)
    {
        // vocon's own saved file name is handled in getWordClassFileName
        String statusFile = fileName + "_status" + ".grm";

        if (voconInstance != 0) {
            FileInputStream s = _fileManager.openFileForReading(statusFile);
            Grammar grammar = null;
            if (s != null)
            {
                try {
                    JSONObject json = JSONUtils.readFromStream(s);
                    s.close();
                    grammar = Grammar.createFromJSON(json);
                } catch (Exception e) {
                    Logger.error(this, "loadVocon: reading Vocon from file error", e);
                }
            }
            if (grammar == null) {
                return null;
            }

            for (VoconContext context : grammar.getContexts()) {
                List<String> slotIds = context.getSlotIds();
                for (String wordClass : slotIds) {
                    if (!hasWordClass(wordClass)) {
                        String wordClassFile = getWordClassFileName(fileName, wordClass);
                        long obj = voconLoadWordClass(voconInstance, wordClassFile);
                        if (obj == 0) {
                            Logger.error(this, "loadVocon (" + wordClassFile + ") native loading vocon error");
                            return null;
                        }

                        _wordClasses.put(wordClass, Long.valueOf(obj));
                    }
                }
            }

            return grammar;
        }

        return null;
    }

    @Override
    public boolean saveVocon(String fileName, Grammar grammar)
    {
        if (fileName == null || grammar == null) {
            return false;
        }

        // vocon's own saved file name is handled in getWordClassFileName
        String statusFile = fileName + "_status" + ".grm";

        if (voconInstance != 0) {

            FileOutputStream s = _fileManager.openFileForWriting(statusFile);
            if (s != null) {
                try {
                    JSONUtils.saveToStream(grammar.toJSON(), s);
                    s.close();
                } catch (Exception e) {
                    Logger.error(this, "Error saving Vocon grammar", e);
                    return false;
                }
            }

            //Save all word class
            for (Map.Entry<String, Long> entry : _wordClasses.entrySet()) {
                String wordClass = entry.getKey();
                Long obj = entry.getValue();
                if (obj != null) {
                    String wordClassfile = getWordClassFileName(fileName, wordClass);
                    int stat = voconSaveWordClass(voconInstance, obj, wordClassfile);
                    if (Status.OK != stat) {
                        Logger.error(this, "saveVocon (" + wordClassfile + ") native saving vocon error:" + stat + "with word slot: " + wordClass);
                        return false;
                    }
                }
            }

            return true;
        }
        return false;
    }

    @Override
    public boolean loadResource(String fileName)
    {
        boolean success = false;

        if (!_resources.containsKey(fileName))
        {
            // START JNI_CODE
            if (voconInstance != 0) {
                // Again, this array operation is a JNI-specific workaround
                // to simulate a pass by reference to the C side.
                long[] resourceTemp = new long[1];
                int status = voconLoadResource(voconInstance, fileName, resourceTemp);
                if (status == Status.OK && resourceTemp[0] != 0)
                {
                    _resources.put(fileName, resourceTemp[0]);
                    success = true;
                }
                else
                {
                    Logger.warn(this, "Unable to load resource (error status = " + status + ").");
                }
            }
            // END JNI_CODE
        }
        else
        {
            Logger.warn(this, "This resource is already loaded.");
        }

        return success;
    }

    @Override
    public void releaseResource(String fileName)
    {
        if (voconInstance != 0) {
            Long resource = _resources.remove(fileName);
            if (resource != null)
            {
                voconReleaseResource(voconInstance, resource);
            }
            else
            {
                Logger.warn(this, "The vocon resource was never loaded or has already been released: " + fileName);
            }
        }
    }

    @Override
    public void enableVoconVerboseLogging(boolean verbose)
    {
        voconEnableVerboseLogging(verbose);
    }

    @Override
    public boolean updateComputedParam(VoconContext context, long paramId)
    {
        return (Status.OK == voconUpdateComputedParam(context.getJniHandle(), paramId));
    }

    @Override
    public boolean stopRecognitionResults()
    {
        return (Status.OK == voconStopRecognitionResults(voconInstance));
    }

    public boolean loadAcModAdaptData(String fileName)
    {
        return (Status.OK == voconLoadAcModAdaptData(voconInstance, fileName));
    }

    public boolean saveAcModAdaptData(String fileName)
    {
        return (Status.OK == voconSaveAcModAdaptData(voconInstance, fileName));
    }

    public boolean loadFxAdaptData(String fileName)
    {
        return (Status.OK == voconLoadFxAdaptData(voconInstance, fileName));
    }

    public boolean saveFxAdaptData(String fileName)
    {
        return (Status.OK == voconSaveFxAdaptData(voconInstance, fileName));
    }

    public boolean loadSpeakerProfile(String fileName)
    {
        return (Status.OK == voconLoadSpeakerProfile(voconInstance, fileName));
    }

    public boolean saveSpeakerProfile(String fileName)
    {
        return (Status.OK == voconSaveSpeakerProfile(voconInstance, fileName));
    }

    private boolean addSubContext(VoconContext context)
    {

        boolean ret = true;
        //Attach eventual subContexts.
        Map<String, VoconContext> voconContexts = context.getSubContexts();
        if (voconContexts != null && voconContexts.size() > 0)
        {
            for (Entry<String, VoconContext> entry : voconContexts.entrySet())
            {
                // create subcontext
                if (ContextError.OK != createAsrContext(entry.getValue()) ) {
                    Logger.error(this, "Failure while creating subcontext");
                    ret = false;
                    break;
                }

                String slotAssignments = entry.getKey();
                if (!isSlotAssignments(slotAssignments))
                {
                    String grammarName = context.getGrammarName();
                    slotAssignments = getFullSlotName(grammarName, slotAssignments);
                }
                if (Status.OK != voconAttachSubContext(context.getJniHandle(), slotAssignments, entry.getValue().getJniHandle()))
                {
                    Logger.error(this, "Create ASR context: Unable to attach context to" + entry.getKey());
                    ret = false;
                    break;
                }
            }
        }
        return ret;
    }

    private void removeSubContext(VoconContext context)
    {
        //Detach eventual subContexts.
        Map<String, VoconContext> voconContexts = context.getSubContexts();
        if (voconContexts != null && voconContexts.size() > 0)
        {
            for (Entry<String, VoconContext> entry : voconContexts.entrySet())
            {
                String slotName = entry.getKey();
                if (isSlotAssignments(slotName))
                {
                    slotName = "";
                }
                else
                {
                    String grammarName = context.getGrammarName();
                    slotName = getFullSlotName(grammarName, slotName);
                }
                if (Status.OK != voconDetachSubContext(context.getJniHandle(), slotName))
                {
                    Logger.warn(this, "Destroy ASR context: Unable to detach context from" + entry.getKey());
                }

                destroyContext(entry.getValue());
            }
        }
    }

    private static String getWordGrammarFileName(String slotName)
    {
        if (slotName == null)
            return null;

        Calendar cal = Calendar.getInstance();
        String affix = new SimpleDateFormat("yyyyMMddHHmmss", java.util.Locale.getDefault()).format(cal.getTime());

        return (new String(slotName + "_" + affix));
    }

    private static String getWordClassFileName(String grammarID, String slotName)
    {
        if (slotName == null) {
            return null;
        }

        String filename = null;
        if (grammarID == null) {
            filename = "slot";
        }
        else {
            int pos = grammarID.indexOf('.');
            if (pos == -1) {
                filename = grammarID;
            }
            else {
                filename = grammarID.substring(0, pos);
            }
        }

        int pos = slotName.indexOf('#');
        if (pos == -1) {
            filename += slotName;
        }
        else {
            filename += slotName.substring(pos + 1);
        }

        return filename;
    }

    private static boolean isSlotAssignments(String input)
    {
        return input.indexOf(':') != -1;
    }

    private static String getFullSlotName(String grammarID, String slotName)
    {
        if (slotName.contains("#"))
            // Slot name is fully qualified
            return slotName;

        return grammarID + "#" + slotName;
    }
}
