package com.nuance.dragon.toolkit.vocon;


/**
 * The recognition tuning parameters being passed to the VoCon engine.
 */
public final class ParamSpecs
{
    /**
     * VoCon booleans.
     */
    public static class Boolean
    {
        /** Indicate a <b>false</b> value. */
        public final static int FALSE = 0;

        /** Indicate a <b>true</b> value. */
        public final static int TRUE = 1;
    }

    /**
     * Enumerates the speaker adaptation modes.
     */
    public static class SpeakerAdaptMode
    {
        /** During recognition the speaker profile is continuously adapted and used. This is the default mode. */
        public final static int AUTOMATIC = 0;

        /** During recognition, the speaker profile is not further adapted but the current profile is used. */
        public final static int APPLY = 1;

        /** During recognition the speaker profile is adapted but not used. UNUSED BY NMT */
        //public final static int TRAIN = 2;

        /** During recognition the speaker profile is not adapted and no profile is used. */
        public final static int DISABLE = 3;
    }

    /**
     * VoCon Front-End parameters.
     * These parameters maps to the LH_FX_PARAM_* VoCon C SDK parameters.
     */
    public static class Fx
    {
        /** Used to enable or disable the Voice Activity Detector, VAD (sometimes also called Begin Of Speech (BOS) detection).
         * If enabled, the feature extractor will try to detect the beginning of speech, and the audio samples before the
         * speech detection will not be send to the recognizer.
         *    <p> Range: LH_TRUE or LH_FALSE
         *    <p> Default: LH_FALSE
         */
        public final static VoconParam START_ENABLE = new VoconParam(0x2b7cebf1, VoconParam.Type.FRONT_END, "LH_FX_PARAM_START_ENABLE");

        /** Sensitivity of the voice activity detection system (VAD), if enabled.
         *  The higher this value, the easier begin of speech will be detected.
         *    <p> Range : 0 -> 100
         *    <p> Default: 50
         */
        public final static VoconParam SENSITIVITY = new VoconParam(0x2b7cf34a, VoconParam.Type.FRONT_END, "LH_FX_PARAM_SENSITIVITY");

        /** Absolute speech energy threshold for VAD. If the voice activity detector is enabled, the energy
         * of the input signal must come above this threshold to trigger the VAD.
         *     <p> Range: -7200 (-72 dB) -> 1800 (18 dB)
         *     <p> Default: -6000 (-60 dB)
         */
        public final static VoconParam ABSOLUTE_THRESHOLD = new VoconParam(0x2b7cfbb1, VoconParam.Type.FRONT_END, "LH_FX_PARAM_ABSOLUTE_THRESHOLD");

        /** The minimal duration, in milliseconds, of the speech before the VAD system actually detects beginning of speech.
         *     <p> Range: 10 ms -> 400 ms
         *     <p> Default: 60 ms
         */
        public final static VoconParam MINSPEECH = new VoconParam(0x2b7cfd00, VoconParam.Type.FRONT_END, "LH_FX_PARAM_MINSPEECH");

        /** Informs the AGC about whether a far-talk or a close-talk microphone is used.
         *     <p> Range: LH_TRUE or LH_FALSE
         *     <p> Default: LH_TRUE
         */
        public final static VoconParam FARTALK = new VoconParam(0x2b7cffc2, VoconParam.Type.FRONT_END, "LH_FX_PARAM_FARTALK");

        /** Used to enable or disable the Automatic Gain Control (AGC).
         * The value for the LH_FX_PARAM_AGC parameter must be a value between 0 and 65535 and
         * it specifies the current gain (0 means that the AGC will be disabled). A subsequent
         * request to modify the gain (through the LH_AUDIOCHAIN_EVENT_FX_GAINREQUEST event) will use this
         * value as the reference level to calculate the new requested gain. With this, setting and
         * getting of the volume or gain on the sound card is left to the application. For more
         * information on how to set/get the volume on your sound card refer to your platform
         * documentation.
         *     <p> Range: 0 - 65535
         *     <p> Default: 0, meaning AGC disabled
         */
        public final static VoconParam AGC = new VoconParam(0x2b7d045f, VoconParam.Type.FRONT_END, "LH_FX_PARAM_AGC");

        /** Leading silence timeout in milliseconds.
         * After each  lh_AudioSourceStart, processing of input data will stop after
         * LH_FX_PARAM_TIMEOUT_LSILENCE milliseconds of input audio have been processed and no begin of speech
         * has been detected.
         * A result is generated, but it will contain no hypothesis, since no audio has been sent to the
         * recognizer.
         * When all processing is done,  lh_AudioSourceAdvance will return with eFinished set to
         * LH_AUDIOCHAIN_FINISHED_TIMEOUT.
         * <p>  LH_FX_PARAM_START_ENABLE must be set to LH_TRUE when using this parameter,
         * otherwise it will have no effect.
         * <p>
         *
         * With the function  lh_AudioSourceControlTimeOut the application can still change the behavior
         * of the timeout counters while the engine is processing audio. This can be useful to take into
         * account external events, like e.g. the play back of an information message by the system.
         *
         * See also  LH_AUDIOSOURCE_PARAM_TIMEOUT,  LH_FX_PARAM_TIMEOUT_SPEECH,
         *  LH_FX_PARAM_TSILENCE,  lh_AudioSourceControlTimeOut.
         *
         *     <p> Range: 0, 100 ms -> 10000 ms
         *     <p> Default: 0 (no timeout on leading silence duration)
         */
        public final static VoconParam TIMEOUT_LSILENCE = new VoconParam(0x2b7d4f1a, VoconParam.Type.FRONT_END, "LH_FX_PARAM_TIMEOUT_LSILENCE");

        /** Speech duration timeout in milliseconds.
         * After each begin of speech detection, processing of input data will stop after
         * LH_FX_PARAM_TIMEOUT_SPEECH milliseconds of input audio have been processed.
         * A result is generated. When all processing
         * is done,  lh_AudioSourceAdvance will return with eFinished set to
         * LH_AUDIOCHAIN_FINISHED_TIMEOUT.
         * <p>
         * If  LH_FX_PARAM_START_ENABLE is set to LH_FALSE,
         * this parameter has the same behavior as  LH_AUDIOSOURCE_PARAM_TIMEOUT.
         * <p>
         * See also  LH_AUDIOSOURCE_PARAM_TIMEOUT,  LH_FX_PARAM_TIMEOUT_LSILENCE,
         *  LH_FX_PARAM_TSILENCE.
         *     <p> Range: 0, 100 ms -> 30000 ms
         *     <p> Default: 0 (no timeout on speech duration)
         */
        public final static VoconParam TIMEOUT_SPEECH = new VoconParam(0x2b7d5004, VoconParam.Type.FRONT_END, "LH_FX_PARAM_TIMEOUT_SPEECH");

        /** Trailing silence duration in milliseconds.
         * Minimum amount of trailing silence, in milliseconds, to be observed by the voice activity detector (VAD)
         * before the end of an utterance is detected.
         * A result is generated when trailing silence is detected. When all processing
         * is done,  lh_AudioSourceAdvance will return with eFinished set to
         *  LH_AUDIOCHAIN_FINISHED_TS.
         * <p>
         * If no trailing silence detection is used (this parameter is set to 0), then you must
         * call  lh_AudioSourceNoMoreData to stop the audio processing and get a recognition result.
         * Similarly, when using a SingleThreadRec object, the  LH_AUDIOIN_INTERFACE.pffetchdata() callback
         * routine must in that case return NULL, 0 to signal the end of the audio input and get a
         * recognition result.
         * <p>
         * See also  LH_AUDIOSOURCE_PARAM_TIMEOUT,  LH_FX_PARAM_TIMEOUT_LSILENCE,
         *  LH_FX_PARAM_TIMEOUT_SPEECH,  LH_CTX_PARAM_TSILENCE_FX,  LH_CTX_PARAM_TSILENCE.
         *     <p> Range: 0, 100 ms - 10000 ms
         *     <p> Default: 0 (no trailing silence detection)
         */
        public final static VoconParam TSILENCE = new VoconParam(0x2b7d5131, VoconParam.Type.FRONT_END, "LH_FX_PARAM_TSILENCE");

        /** One of the front end adaptations is an adaptation based on the gender of the
         * speaker. Actually, it is based on the speaker's vocal tract length, which is strongly correlated
         * to the gender, but also to the age of the speaker. Based on the estimated vocal tract length,
         * a different warp factor is applied to the signal, hence the name of this parameter.
         * This parameter influences this vocal tract length estimation and the resulting applied warp factor.
         *
         * The default setting of this parameter is  LH_FX_AUTOSELECT_MF_WARPFACTOR, meaning that the front end
         * automatically selects the most optimal warp factor, but excluding the warp factor that corresponds to the
         * vocal tract length of a child. This setting is recommended when the ASR system will mainly be used by grown-ups,
         * for instance in a car.
         *
         * When setting this parameter to  LH_FX_AUTOSELECT_MFC_WARPFACTOR, the front end also considers the child
         * warping factor in its automatic selection. This setting is recommended when the ASR system will also be used
         * by children, for instance in living room applications.
         *
         * Other allowed values are  LH_FX_NEUTRAL_WARPFACTOR,  LH_FX_MALE_WARPFACTOR,
         *  LH_FX_FEMALE_WARPFACTOR and  LH_FX_CHILD_WARPFACTOR. If setting one of those, no automatic selection happens but the
         * corresponding warp factor is always used (neutral warp factor lies in between male and female warp factor).
         * Unless you have a very specific reason, you should not fix the warp factor to one of those values, even not if you
         * know the true gender and age of the speaker. Because although there is a strong correlation between gender/age
         * and the optimal warp factor, there are speakers for which this does not hold.
         *
         * Possible values are:  LH_FX_AUTOSELECT_MF_WARPFACTOR,  LH_FX_AUTOSELECT_MFC_WARPFACTOR,
         *  LH_FX_NEUTRAL_WARPFACTOR,  LH_FX_MALE_WARPFACTOR,  LH_FX_FEMALE_WARPFACTOR,  LH_FX_CHILD_WARPFACTOR.
         *  <p> Default:  LH_FX_AUTOSELECT_MF_WARPFACTOR.
         *
         *  <p><b> Refer to Vocon documentation value for value range. </b>TODO: Expose possible values?
         */
        public final static VoconParam FIX_WARPFACTOR = new VoconParam(0x2b7d51f4, VoconParam.Type.FRONT_END, "LH_FX_PARAM_FIX_WARPFACTOR");

        /**
         * Adaptation speed selection.
         * This parameter allows to switch between two different adaptation settings.
         * When selecting the LH_FX_FAST_ADAPTATION mode, the front end will adapt faster
         * to a new situation after a speaker reset, resulting in a lower error rate on the
         * first few utterances after a speaker reset. However, it comes at the expense of an increased
         * latency on the first few utterances, because more calculations are required.
         * Once convergence has been reached, there is no increased latency anymore.
         * <p> When selecting the LH_FX_STANDARD_ADAPTATION mode, the adaptation is slower, but
         * there is no increased latency on the first few utterances compared to the later
         * utterances.
         * <p> Furthermore, when LH_FX_FAST_ADAPTATION is
         * selected, the features send from the front end to the recognizer are larger in size,
         * i.e. more memory is needed to store them. This means that the shared memory used in the
         * multithread buffer reader and writer objects ( lh_CreateMThreadBufReader) must be larger to hold the same amount
         * of signal and the memory needs for the utterance object returned in the  lh_SingleThreadRecProcessEx
         * will also be larger.
         * <p> It is recommended to use the LH_FX_FAST_ADAPTATION setting, unless this
         * increases the latency too much.
         * <p> Possible values are:  LH_FX_STANDARD_ADAPTATION,  LH_FX_FAST_ADAPTATION.
         * <p> Default:  LH_FX_FAST_ADAPTATION
         *
         *  <p><b> Refer to VoCon documentation value for value range. </b>TODO: Expose possible values?
         */
        public final static VoconParam ADAPTATION = new VoconParam(0x33acf7f6, VoconParam.Type.FRONT_END, "LH_FX_PARAM_ADAPTATION");

        /** Time interval, in milliseconds, after which a  LH_AUDIOCHAIN_EVENT_FX_TIMER event must be generated. <p>
         * When setting this parameter, the feature extractor will generate a  LH_AUDIOCHAIN_EVENT_FX_TIMER
         * event each time it has processed the specified amount of input signal. This is useful in
         * combination with the  lh_FxGetSignalLevels call, because it allows to get the
         * signal energy level at regular intervals.
         * <p> Default: 0, which disables the generation of  LH_AUDIOCHAIN_EVENT_FX_TIMER events.
         */
        public final static VoconParam EVENT_TIMER = new VoconParam(0x2b7d531f, VoconParam.Type.FRONT_END, "LH_FX_PARAM_EVENT_TIMER");

        /** If the application knows when speaker changes occur, it should set this parameter to LH_TRUE and call
         *  lh_EnvAdaptReset on the Fx object each time a speaker change occurs. If it knows which speaker is speaking and if it has
         * a previously archived session data for that speaker, the application should reload that session data with
         * the  lh_EnvAdaptReset call.
         * <p> If the application does not know for sure when speaker changes occur, it should set this parameter
         * to LH_FALSE. It can still call the  lh_FxPossiblyNewSpeaker function at times when it thinks that a
         * speaker change may occur.
         * <p> This parameter influences how much past information is used to estimate the current speaker and
         * environment settings. When setting this parameter to LH_TRUE, a long history is used; when setting it to
         * LH_FALSE, a shorter history is used. The latter means that the feature extractor will adapt faster to a new
         * situation, but also that it will be more sensitive to noise.
         * <p> Both functions  lh_EnvAdaptReset on the Fx object and  lh_FxPossiblyNewSpeaker can be called regardless the setting of this
         * parameter. The function  lh_EnvAdaptReset returns the feature extractor to the exact same state it was in at the time
         * the session data was archived, or to the startup state if no session data is provided. The function
         *  lh_FxPossiblyNewSpeaker keeps the current adaptation state, but temporarily shortens the history, such that the
         * feature extractor will quickly adapt to a new situation.
         * <p> Default: LH_TRUE.
         */
        public final static VoconParam KNOWN_SPEAKER_CHANGES = new VoconParam(0x2b7d5638, VoconParam.Type.FRONT_END, "LH_FX_PARAM_KNOWN_SPEAKER_CHANGES");

        /** If this parameter is set to  LH_TRUE, the function  lh_FxFetchPCMAudio can be used.
         *
         * When setting this parameter to  LH_TRUE, the FX object will store the input audio in a buffer,
         * which can be obtained by calling  lh_FxFetchPCMAudio.
         *
         * Buffering audio samples requires a considerable amount of memory. Therefore, you should
         * enable this parameter only if you really plan to retrieve the buffered samples with  lh_FxFetchPCMAudio.
         *
         * Default:  LH_FALSE
         *
         * <b> Note: this also copies the audio data to the SD card. Therefore, one should never enable this
         * parameter in an application release build. </b>
         */
        public final static VoconParam ENABLEFETCHPCMAUDIO = new VoconParam(0x6b239240, VoconParam.Type.FRONT_END, "LH_FX_PARAM_ENABLEFETCHPCMAUDIO");

        /**
         * Speaker adaptation mode
         *
         * Set the speaker adaptation mode that the recognizer should use.
         * If the recognizer is already started, this mode will only be applied on the next start.
         *
         * <p> Possible values values: See {@link SpeakerAdaptMode} for possible values and their meaning.
         *
         * Default: {@link SpeakerAdaptMode#AUTOMATIC}
         */
        public final static VoconParam SPEAKER_ADAPTATION = new VoconParam(0x32d9b8fb, VoconParam.Type.FRONT_END, "LH_FX_PARAM_SPEAKER_ADAPTATION");


        /** If this parameter is set to TRUE, the recognition will automatically restart at the end of each utterance.
         *
         * NOTE: use in conjunction with AUTOMATIC_RESTART
         *
         * A perfect use case for this parameter is open microphone recognition, where you have to continue listening to the audio until a
         * particular word is recognized.
         *
         * The automatic restart on the FX will also restart the recognition object. This means that all loaded contexts are reset to
         * their start state, they do not continue to listen where they ended in the previous utterance. For instance, if you have a
         * grammar which allows to say 'call John' or 'call John at home' and the FX detects trailing after 'call John', then the automatic
         * restart will again listen for 'call John' or 'call John at home', it CANNOT return just 'at home'. It is not possible to
         * load another context during the automatic restart.
         *
         * Default: FALSE
         */
        public final static VoconParam AUTOMATIC_RESTART = new VoconParam(0x6e648860, VoconParam.Type.FRONT_END, "LH_FX_PARAM_AUTOMATIC_RESTART");

        /** If this parameter is set to TRUE, the recognition will automatically restart at the end of each utterance.
         *
         * NOTE: use in conjunction with AUTOMATIC_RESTART
         *
         * A perfect use case for this parameter is open microphone recognition, where you have to continue listening to the audio until a
         * particular word is recognized.
         *
         * The automatic restart on the FX will also restart the recognition object. This means that all loaded contexts are reset to
         * their start state, they do not continue to listen where they ended in the previous utterance. For instance, if you have a
         * grammar which allows to say 'call John' or 'call John at home' and the FX detects trailing after 'call John', then the automatic
         * restart will again listen for 'call John' or 'call John at home', it CANNOT return just 'at home'. It is not possible to
         * load another context during the automatic restart.
         *
         * Default: FALSE
         */
        public final static VoconParam THREAD_AUTOMATIC_RESTART = new VoconParam(0x77b692bf, VoconParam.Type.FRONT_END, "LH_MTHREADBUFREADER_PARAM_AUTOMATIC_RESTART");
    }

    /**
     * VoCon recognition context parameters.
     * These parameters maps to the LH_CTX_PARAM_* VoCon C SDK parameters.
     */
    public static class Ctx
    {
        /** Trade-off between CPU-load, memory requirements and the obtained accuracy
         * of the search. With a lower value, less CPU-time and memory are needed,
         * but the accuracy also degrades. With a higher value, more CPU-time and
         * memory are needed, but the performance increases.
         *
         * <b> Note: see also the  LH_CTX_PARAM_MAXNBEST_SECONDPASS parameter. Just like
         * the  LH_CTX_PARAM_ACCURACY parameter it controls the trade-off between recognition
         * accuracy and CPU-load. Therefore, the recognition performance may improve if
         *  LH_CTX_PARAM_MAXNBEST_SECONDPASS is increased too. The difference is that
         *  LH_CTX_PARAM_ACCURACY influences the CPU load during the first search pass,
         * whereas  LH_CTX_PARAM_MAXNBEST_SECONDPASS only influences the CPU load during
         * the second search pass.
         *
         *   <p>  default: 10000 (You will always choose a lower value in practice),
         *   <p>  range: 100 -> 50000
         *   <p> Can be used on all DP types.
         *
         * <b> Important: With the functions  lh_ConfigUpdateComputedParam or  lh_ConfigUpdateAllComputedParams
         * you can let the DP compute a good estimate for this parameter.
         */
        public final static VoconParam ACCURACY = new VoconParam(0x154d5340, VoconParam.Type.CONTEXT, "LH_CTX_PARAM_ACCURACY");

        /** Maximum number of hypotheses returned in a result.
         *
         *   <p>  default: 3,
         *   <p>  range: 1 -> 1000
         *   <p> Can be used on all, except spelling, DP types.
         */
        public final static VoconParam MAXNBEST = new VoconParam(0x154d53ee, VoconParam.Type.CONTEXT, "LH_CTX_PARAM_MAXNBEST");

        /** This parameter controls the maximum number of hypotheses that
         * will be considered during the second pass of a search
         * algorithm. The hypotheses that are considered during the
         * second pass are re-scored with more precise acoustic models.
         * One can use this parameter to find a trade-off between the time
         * that the second pass takes, and the recognition accuracy.
         *
         * The higher this parameter, the more time is taken by the
         * second pass, but this usually has a positive impact on the
         * recognition accuracy.
         * If you set this parameter to a value less than  LH_CTX_PARAM_MAXNBEST
         * then you effectively also limit the number of hypotheses in the result
         * to  LH_CTX_PARAM_MAXNBEST_SECONDPASS.
         *
         * It is recommended to set  LH_CTX_PARAM_MAXNBEST to the number of
         * hypotheses that the application will really use and to set
         *  LH_PARAM_MAXNBEST_SECONDPASS to at least the same value,
         * but with a minimum of 5. Only in case this leads to an
         * unacceptably large second pass time,  LH_CTX_PARAM_MAXNBEST_SECONDPASS
         * should be set to a lower value than this recommendation.
         * If the recognition accuracy is not satisfactory with a value of 5, try
         * increasing this parameter.
         *
         * For Spelling DP the value of this parameter is fixed and and it can not be modified.
         *
         *    <p>  default: 5,
         *    <p>  range: 1 -> 1000
         *   <p> Can be used on all DP types, except spelling and slm.
         */
        public final static VoconParam MAXNBEST_SECONDPASS = new VoconParam(0x154d5443, VoconParam.Type.CONTEXT, "LH_CTX_PARAM_MAXNBEST_SECONDPASS");

        /** Minimum amount of trailing silence, in milliseconds, to be observed by the voice activity detector (VAD)
         * before the end of an utterance is detected. This parameter is similar to  LH_FX_PARAM_TSILENCE
         * but the trailing silence decision is
         * taken in the search algorithm instead of in the FX.
         * This allows different TSILENCE_FX values for different contexts.
         * Setting this value to 0 disables this feature.
         *
         *   <p>  default: 0 ms (disabled)
         *   <p>  range: 0 ms, 100 ms -> 5000 ms
         *   <p> Can be used on all DP types
         *
         * For an isolated word context, a value of 300 ms is recommended,
         * while for a continuous speech context a value of 800 ms is recommended.
         */
        public final static VoconParam TSILENCE_FX = new VoconParam(0x214c4d1a, VoconParam.Type.CONTEXT, "LH_CTX_PARAM_TSILENCE_FX");

        /** Minimum amount of trailing silence, in milliseconds, as observed by the search engine,
         * before the end of an utterance is detected. This parameter affects low-level behavior
         * of the algorithm and it should be used with care. Quite often, the optimal value of
         *  LH_CTX_PARAM_TSILENCE is twice as high as the value of  LH_CTX_PARAM_TSILENCE_FX.
         * The optimal value is context-dependent however. If you have a sufficient amount of
         * representative speech data, and the response time of the engine is too long, you may want to
         * experiment with different values.
         *
         *   <p>  default: 0 ms (disabled)
         *   <p>  range: 0 ms, 100ms -> 10000 ms
         *   <p> Can be used on all DP types.
         */
        public final static VoconParam TSILENCE = new VoconParam(0x16ef5416, VoconParam.Type.CONTEXT, "LH_CTX_PARAM_TSILENCE");

        /** This parameter affects low-level behavior of the algorithm. It is solely intended
         * for tuning tools. DO not touch, unless you know what you are doing.
         *  <p>  default: 2500
         *  <p>  range:   0 -> 10000
         *  <p> Can be used on all DP types.
         */
        public final static VoconParam INITBEAMWIDTH = new VoconParam(0x17df0963, VoconParam.Type.CONTEXT, "LH_CTX_PARAM_INITBEAMWIDTH");

        /** Word Penalty. This parameter can only be set (and get) when a context has been loaded and before starting the search engine.
         * This parameter affects low-level behavior of the algorithm. It is solely intended for tuning tools. DO not touch, unless you know what you are doing.
         * <p>  default: Depends on the acoustic models. Also depends on whether the context is a !spelling context or not.
         * <p>  range: 0->3000
         * <p> Not supported for contexts of type TREEDP (compiled with include contextcpl.html and the  treedp DP type).
         * <p> Not supported for contexts of type FSTDP  (compiled with include fieldcontextcpl.html, include listcontextcpl.html, include grmcpl.html, or include contextcpl.html  fstdphost DP type).
         * <p> For Fst contexts the word penalty can only be set at compilation time, changing this parameter at run time is not supported.
         *
         * <b> Important: With the functions  lh_ConfigUpdateComputedParam or  lh_ConfigUpdateAllComputedParams
         * you can let the context compute a good estimate for this parameter.
         */
        public final static VoconParam WORDPENALTY = new VoconParam(0x17dfe9b0, VoconParam.Type.CONTEXT, "LH_CTX_PARAM_WORDPENALTY");

        /** Decoding strategy.
         *
         * The decoding strategy is a read-only parameter set by the context creation tool.
         * It is only supported by contexts implementing the  VoCon_Interfaces_HostCtx and  VoCon_Interfaces_FieldCtx.
         * Use  lh_ConfigHasParam to check whether a context supports this parameter.
         *
         * See the help in the include grmcpl.html and include fieldcontextcpl.html compilation tools for more details on the meaning
         * of the decoding strategy and when to use which decoding strategy.
         *
         * When embedding a guest context in a host context, (see  lh_HostCtxSetCtx), both contexts must use the
         * same decoding strategy.
         *
         * Possible values:
         * <p>  LH_CTX_DECODING_STRATEGY_FORWARD, which is the default decoding strategy.
         * <p>  LH_CTX_DECODING_STRATEGY_BACKWARD.
         *
         *  <p><b> Refer to Vocon documentation value for value range. </b>TODO: Expose possible values?
         */
        public final static VoconParam DECODING_STRATEGY = new VoconParam(0x404d0c53, VoconParam.Type.CONTEXT, "LH_CTX_PARAM_DECODING_STRATEGY");

        /** Maximum amount of confidence level that indicates that a spoken utterance is out of grammar.
         * If confidence level is equal or below this parameter value, it means that the spoken utterance is out of grammar.
         * If confidence level is above value of this parameter, the spoken utterance can be probably in grammar.
         * Use  lh_NBestResultIsInGrammar to check whether the spoken utterance is in grammar.
         *   <p>  default: 5000
         *   <p>  range: 0 -> 10000
         *   <p> Putting the  LH_CTX_PARAM_IG_LOWCONF above the  LH_CTX_PARAM_IG_HIGHCONF has no effect.
         */
        public final static VoconParam IG_LOWCONF = new VoconParam(0x5237a110, VoconParam.Type.CONTEXT, "LH_CTX_PARAM_IG_LOWCONF");

        /** Minimum amount of confidence level that indicates that a spoken utterance is in grammar.
         * If confidence level is equal or above this parameter value, it means that the spoken utterance is in grammar
         * (even through it is the same as  LH_CTX_PARAM_IG_LOWCONF value).
         * If confidence level is below this parameter value, the spoken utterance can be probably in grammar or out of grammar.
         * Use  lh_NBestResultIsInGrammar to check whether the spoken utterance is in grammar.
         *
         *   <p>  default: 5000
         *   <p>  range: 0 -> 10000
         *   <p> Putting the  LH_CTX_PARAM_IG_HIGHCONF below the  LH_CTX_PARAM_IG_LOWCONF has no effect.
         */
        public final static VoconParam IG_HIGHCONF = new VoconParam(0x5237a0a4, VoconParam.Type.CONTEXT, "LH_CTX_PARAM_IG_HIGHCONF");

        /** The maximum number of DB rows that will be taken into account by an interpretation.
         *
         * <p><b>This parameter is only applicable for a fuzzy match context.</b>
         */
        public final static VoconParam MAX_COLLAPSED_LISTSIZE = new VoconParam(0x75e9470c, VoconParam.Type.CONTEXT, "LH_TABLECTX_PARAM_MAX_COLLAPSED_LISTSIZE");

        /** If the number rows in the interpretation is smaller or equal to MAX_NBR_UNCOLLAPSED_ENTRIES,
         * the collapsed rows will also be added as uncollapsed interpretations.
         * If the number of rows exceeds MAX_NBR_UNCOLLAPSED_ENTRIES no additional uncollapsed interpretations will be added.
         *
         * <p><b>This parameter is only applicable for a fuzzy match context.</b>
         */
        public final static VoconParam MAX_NBR_UNCOLLAPSED_ENTRIES = new VoconParam(0x76d465de, VoconParam.Type.CONTEXT, "LH_TABLECTX_PARAM_MAX_NBR_UNCOLLAPSED_ENTRIES");

        public final static VoconParam INVALID = new VoconParam(0x7fffffff, VoconParam.Type.CONTEXT, "LH_CTX_PARAM_INVALID");
    }
}
