package com.nuance.dragon.toolkit.vocon;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

public class VoconConfig implements JSONCompliant {

    private static final String KEY_ACMODFILE = "acmodfile";
    private static final String KEY_CLCFILE = "clcfile";
    private static final String KEY_LANGUAGE = "language";
    private static final String KEY_SIZE = "size";
    private static final String KEY_DOMAIN = "domain";

    final String acmodfile;
    final String clcfile;
    final String language;
    final int size;
    final String domain;

    /**
     * @param acmodfile The Acoustic Model file name. Must be a valid file name.
     * @param clcfile The Common Language Component file name. Must be a valid file name.
     */
    public VoconConfig(String acmodfile, String clcfile)
    {
        Checker.checkStringArgForNullOrEmpty("acmodfile", acmodfile);
        Checker.checkStringArgForNullOrEmpty("clcfile", clcfile);
        this.acmodfile = acmodfile;
        this.clcfile = clcfile;
        this.language = null;
        this.size = 0;
        this.domain = null;
    }

    /**
     * @param language The language string, from {@link VoconRecognizer.Languages}.
     */
    public VoconConfig(VoconLanguage language)
    {
        this(language, -1, VoconDomain.UNSPECIFIED);
    }

    /**
     * @param language The language, from {@link com.nuance.dragon.toolkit.language.Languages}.
     */
    public VoconConfig(com.nuance.dragon.toolkit.language.Language language)
    {
        this(language, -1, VoconDomain.UNSPECIFIED);
    }

    /**
     * @deprecated
     * @param language The language string, from {@link VoconRecognizer.Languages}.
     * @param size The approximate acoustic model size in KB, e.g. 900, 1300, 2500, 2700, etc. or -1 to load the biggest model available.
     * @param domain The domain string, from {@link VoconDomain}.
     */
    @Deprecated
    public VoconConfig(VoconLanguage language, int size, String domain)
    {
        Checker.checkArgForNull("language", language);
        Checker.checkStringArgForNullOrEmpty("domain",  domain);
        this.acmodfile = null;
        this.clcfile = null;
        this.language = language.name;
        this.size = size;
        this.domain = domain;
    }

    /**
     * @deprecated
     * @param language The language, from {@link com.nuance.dragon.toolkit.language.Languages}.
     * @param size The approximate acoustic model size in KB, e.g. 900, 1300, 2500, 2700, etc. or -1 to load the biggest model available.
     * @param domain The domain string, from {@link VoconDomain}.
     */
    @Deprecated
    public VoconConfig(com.nuance.dragon.toolkit.language.Language language, int size, String domain)
    {
        Checker.checkArgForNull("language", language);
        Checker.checkStringArgForNullOrEmpty("domain",  domain);
        this.acmodfile = null;
        this.clcfile = null;
        this.language = language.getVoconLanguage().name;
        this.size = size;
        this.domain = domain;
    }

    private VoconConfig(String language, int size, String domain)
    {
        this.acmodfile = null;
        this.clcfile = null;
        this.language = language;
        this.size = size;
        this.domain = domain;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        if (language != null)
        {
            o.tryPut(KEY_LANGUAGE, language);
            o.tryPut(KEY_SIZE, size);
            o.tryPut(KEY_DOMAIN, domain);
        } else
        {
            o.tryPut(KEY_ACMODFILE, acmodfile);
            o.tryPut(KEY_CLCFILE, clcfile);
        }
        return o;
    }

    public static VoconConfig createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String acmodfile = json.optString(KEY_ACMODFILE, null);
        String clcfile = json.optString(KEY_CLCFILE, null);
        if (acmodfile != null && clcfile != null)
        {
            return new VoconConfig(acmodfile, clcfile);
        }
        return new VoconConfig(json.getString(KEY_LANGUAGE), json.getInt(KEY_SIZE), json.getString(KEY_DOMAIN));
    }
}
