package com.nuance.dragon.toolkit.vocon;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * Represents a Vocon / "Host" recognition context.
 */
public class VoconContext implements JSONCompliant
{
    private static final String KEY_ID = "id";
    private static final String KEY_SLOT_IDS = "slotids";

    private final String _id;
    private String _grammarName;
    private final String _wakeupPhrase;
    private final int _wakeupMode;
    private final List<String> _slotIds; // only keep slot id here
    private List<String> _rules;
    private List<VoconTableQuery> _tableQueries;
    private Map<String, VoconContext> _subContexts;
    private Map<VoconParam, Integer> _params;
    private boolean _preloaded;
    private long _jniHandle; // only visible by the native layer

    /**
     * Create VoconContext.
     * @param id The context ID (can be the file name or a unique ID for example).
     */
    public VoconContext(String id)
    {
        this(id, null, -1);
    }

    /**
     * Create VoconContext.
     * @param id The context ID (can be the file name or a unique ID for example).
     * @param wakeupPhrase The wake-up phrase. Can be null (ASR context).
     * @param wakeupMode The wake-up mode.
     */
    public VoconContext(String id, String wakeupPhrase, int wakeupMode)
    {
        Checker.checkArgForNull("id", id);

        _id = id;
        _wakeupPhrase = wakeupPhrase;
        _wakeupMode = wakeupMode;
        _slotIds = new ArrayList<String>();
        _rules = new ArrayList<String>();
        _tableQueries = new ArrayList<VoconTableQuery>();
        _params = new HashMap<VoconParam, Integer>();
        _subContexts = new HashMap<String, VoconContext>();
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((_id == null) ? 0 : _id.hashCode());
        result = prime * result + (int) (_jniHandle ^ (_jniHandle >>> 32));
        result = prime * result + ((_params == null) ? 0 : _params.hashCode());
        result = prime * result + ((_rules == null) ? 0 : _rules.hashCode());
        result = prime * result
                + ((_slotIds == null) ? 0 : _slotIds.hashCode());
        result = prime * result
                + ((_tableQueries == null) ? 0 : _tableQueries.hashCode());
        result = prime * result + _wakeupMode;
        result = prime * result
                + ((_wakeupPhrase == null) ? 0 : _wakeupPhrase.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        VoconContext other = (VoconContext) obj;
        if (_id == null) {
            if (other._id != null)
                return false;
        } else if (!_id.equals(other._id))
            return false;
        if (_jniHandle != other._jniHandle)
            return false;
        if (_params == null) {
            if (other._params != null)
                return false;
        } else if (!_params.equals(other._params))
            return false;
        if (_rules == null) {
            if (other._rules != null)
                return false;
        } else if (!_rules.equals(other._rules))
            return false;
        if (_slotIds == null) {
            if (other._slotIds != null)
                return false;
        } else if (!_slotIds.equals(other._slotIds))
            return false;
        if (_tableQueries == null) {
            if (other._tableQueries != null)
                return false;
        } else if (!_tableQueries.equals(other._tableQueries))
            return false;
        if (_wakeupMode != other._wakeupMode)
            return false;
        if (_wakeupPhrase == null) {
            if (other._wakeupPhrase != null)
                return false;
        } else if (!_wakeupPhrase.equals(other._wakeupPhrase))
            return false;
        return true;
    }

    /**
     * Construct a VoconContext with the specified Vocon context information.
     *
     * @param fileName The context file name.
     * @param slotIds  The optional list of dynamic slot id specified in this Vocon context.
     *
     * @return a VoconContext object if successful or null if the argument list is not valid.
     */
    public static VoconContext createVoconContext(String fileName, List<String> slotIds)
    {
        Checker.checkStringArgForNullOrEmpty("fileName", fileName);
        Checker.checkArgsForNull("slotIds", slotIds);


        VoconContext voconContext = new VoconContext(fileName);

        voconContext._slotIds.addAll(slotIds);

        return voconContext;
    }

    /**
     * Create a copy of the VoconContext.
     * @return A copy, never null.
     */
    public VoconContext copy()
    {
        VoconContext copy = new VoconContext(_id);
        copy._slotIds.addAll(_slotIds);

        return copy;
    }

    /**
     * Return a list containing all the sub-Contexts in this VoconContext.
     *
     * @return The sub-Contexts.  Will never be null, but may be empty.
     */
    public Map<String, VoconContext> getSubContexts()
    {
        return _subContexts;
    }

    /**
     * Add a sub-context to this context
     *
     * @param subContext the sub-context to be added
     * @param slotAssignments Can be either a slot id or a slot assignments expression.
     * @return true if the sub-context is successfully added, otherwise false
     */
    public boolean addSubContext(VoconContext subContext, String slotAssignments)
    {
        if (subContext != null && subContext!= this && !_subContexts.containsKey(slotAssignments)) {
            _subContexts.put(slotAssignments, subContext);
            return true;
        }
        return false;
    }
    
    /**
     * Remove all sub-contexts from this context
     *
     * @return true if operation is successful, otherwise false
     */
    public boolean clearSubContexts()
    {
        if (_subContexts != null) {
            _subContexts.clear();
            return true;
        }
        return false;
    }

    /**
     * Return a list containing all the slot ids in this VoconContext.
     *
     * @return The slot Ids.  Will never be null, but may be an empty list.
     */
    public List<String> getSlotIds()
    {
        return _slotIds;
    }

    /**
     * Add a dynamic slot id to this context
     *
     * @param slotId the slot id to be added
     *
     * @return true if the slot id is successfully added, otherwise false
     */
    public boolean addSlotId(String slotId) {
        if (!_slotIds.contains(slotId)) {
            _slotIds.add(slotId);
            return true;
        }
        return false;
    }

    /**
     * Get the ID.
     *
     * @return The context ID.
     */
    public String getId()
    {
        return _id;
    }

    /**
     * Get the wake-up phrase.
     *
     * @return The wake-up phrase.
     */
    public String getWakeupPhrase()
    {
        return _wakeupPhrase;
    }

    /**
     * Get the wake-up mode.
     *
     * @return The wake-up mode.
     */
    public int getWakeupMode()
    {
        return _wakeupMode;
    }

    /**
     * Set the rules.
     * @param rules The rules.
     */
    public void setRules(List<String> rules)
    {
        Checker.checkArgForNull("rules", rules);

        _rules = rules;
    }

    /**
     * Get the rules.
     * @return The rules.
     */
    public List<String> getRules()
    {
        return _rules;
    }

    /**
     * Get the grammarName.
     * @return The rules.
     */
    String getGrammarName() {
        return _grammarName;
    }

    /**
     * Set the grammarName.
     * @return The rules.
     */
    void setGrammarName(String grammarName) {
        this._grammarName = grammarName;
    }

    /**
     * Set the table query (only applicable for a fuzzy-match context)
     * @param queries The table queries
     */
    public void setTableQueries(List<VoconTableQuery> queries)
    {
        Checker.checkArgForNull("queries", queries);
        Checker.checkArgForCondition("VoconContext", "fuzzy-match (.zcf) context", _id.endsWith(".zcf"));

        _tableQueries = queries;
    }

    /**
     * Get the table queries.
     * @return The table queries.
     */
    public List<VoconTableQuery> getTableQueries()
    {
        return _tableQueries;
    }

    /**
     * Set the params.
     * @param params The params. See {@link ParamSpecs.Ctx} for possible values.
     */
    public void setParams(Map<VoconParam, Integer> params)
    {
        Checker.checkArgForNull("params", params);

        _params.clear();
        _params.putAll(params);
    }

    /**
     * Get the params.
     * @return The params.
     */
    public Map<VoconParam, Integer> getParams()
    {
        return _params;
    }
    
    void setPreloaded(boolean preloaded)
    {
        _preloaded = preloaded;
    }
    
    boolean getPreloaded()
    {
        return _preloaded;
    }

    void setJniHandle(long handle)
    {
        _jniHandle = handle;
    }

    long getJniHandle()
    {
        return _jniHandle;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();

        o.tryPut(KEY_ID, _id);
        o.tryPutStringList(KEY_SLOT_IDS, getSlotIds());

        return o;
    }

    /**
     * @param json A JSON object that represents a VoconContext.
     *
     * @return A VoconContext object if successful. Throws JSONException if there is an error.
     */
    public static VoconContext createFromJSON(org.json.JSONObject json) throws JSONException
    {
        String id = json.getString(KEY_ID);

        org.json.JSONArray s = json.getJSONArray(KEY_SLOT_IDS);
        List<String> slots = new ArrayList<String>(s.length());
        for (int i = 0; i < s.length(); ++ i) {
            slots.add(s.getString(i));
        }

        VoconContext context = createVoconContext(id, slots);
        if (context == null) {
            throw new JSONException("Unable to create VoconContext.");
        }
        return context;
    }
}
