package com.nuance.dragon.toolkit.vocon;

import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;

/**
 * A Vocon error.
 */
public class VoconError implements JSONCompliant
{
    /**
     * Possible reasons for Vocon errors.
     */
    public static class Reason
    {
        public static final int CANCELED              = 0;
        public static final int AUDIO_FRAMES_DROPPED  = 1;
        public static final int NO_RESULT             = 2;
        public static final int CANT_START            = 3;
        public static final int CONTEXT_DEF_ERROR     = 4;
        public static final int WORDSLOT_DEF_ERROR    = 5;
        public static final int FILE_ERROR            = 6;
        public static final int MISSING_WORDLIST      = 7;
        public static final int NOT_INITIALIZED       = 8;

        public static final int LOAD_GRAMMAR_ERROR    = 9;
        public static final int CREATE_WORDCLASS      = 10;
        public static final int ADD_WORD_ERROR        = 11;
        public static final int REMOVE_WORD_ERROR       = 12;
        public static final int ATTACH_WORD_CLASS_ERROR = 13;
        public static final int CLEAR_GRAMMAR_ERROR     = 14;

        public static final int CONTEXT_COMPILATION_ERROR = 15;

        public static final int FILE_SAVE_ERROR = 16;
        public static final int FILE_LOAD_ERROR = 17;

        private static final String [] REASON_TEXTS = {"Recognition canceled",
                                                       "Audio frames dropped",
                                                       "No recognition result",
                                                       "Can't start",
                                                       "Context definition error",
                                                       "WordSlot definition error",
                                                       "File error",
                                                       "Missing wordlist",
                                                       "Vocon is not initialized.",
                                                       "Vocon load grammar file error",
                                                       "Vocon create word class error",
                                                       "Vocon add word error",
                                                       "Vocon remove word error",
                                                       "Vocon attach word class error",
                                                       "Vocon clear grammar error",
                                                       "Vocon context compilation error",
                                                       "Vocon save file error",
                                                       "Vocon load File error"};
    }

    private static final String KEY_CODE = "code";

    private final int _code;
    private final String _reason;

    VoconError(int code)
    {
        _code = code;
        _reason = Reason.REASON_TEXTS[code];
    }

    /**
     * Get the reason code, from {@link Reason}.
     * @return The reason code
     */
    public int getReasonCode()
    {
        return _code;
    }

    /**
     * @return A description corresponding with the reason code
     */
    public String getReason()
    {
        return _reason;
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        o.tryPut(KEY_CODE, _code);
        return o;
    }

    // TODO never null?
    public static VoconError createFromJSON(org.json.JSONObject json) throws JSONException
    {
        int code = json.getInt(KEY_CODE);
        if (code < 0 || code >= Reason.REASON_TEXTS.length)
            throw new JSONException("Invalid code");
        return new VoconError(code);
    }

    @Override
    public String toString()
    {
        return _reason;
    }
}
