package com.nuance.dragon.toolkit.vocon;

import java.util.List;
import java.util.Map;

import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.SpeechDetectionListener;
import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.language.LanguageEvent;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;

public abstract class VoconRecognizer extends LanguageEvent
{
    /** Not for public use. */
    public static boolean LOAD_NATIVE_LIBRARY_SUCCESS = NativeVoconImpl.LOAD_LIBRARY_SUCCESS;

    /**
     * Listens for results or errors from a recognition.
     */
    public interface ResultListener
    {
        /**
         * Results available
         * @param result The results received
         */
        void onResult(VoconResult result);

        /**
         * Error occurred
         * @param error The error description
         */
        void onError(VoconError error);
    }

    /**
     * Listens for audio signal updates during a recognition.
     */
    public interface SignalListener
    {
        /**
         * Update available
         * @param voconAudioInfo The current audio info as detected by Vocon.
         */
        void onUpdate(VoconAudioInfo voconAudioInfo);

        /**
         * Warning occurred
         * @param warning The warning description
         */
        void onWarning(VoconWarning warning);
    }

    /**
     * Listens for the completion of {@link VoconRecognizer#checkWakeupPhrases}.
     */
    public interface WakeupCheckListener
    {

        /**
         * Called when all the phrases have been tested for validity.
         * The phrase char* contained in the pWakeupCheckResult are the exact ones
         * passed to CheckWakeUpPhrases methods.
         *
         * @param wakeupcheckresult The checked phrase result containing quality of each phrases.
         */
        void onComplete(WakeupCheckResult wakeupcheckresult);

        /**
         * Called if VoCon is in a state that does not allow checking of Wakeup Phrases.
         * @param error The error (never null)
         */
        void onError(VoconError error);
    }

    /**
     * Listens for results or errors from a grammar rebuild started by {@link
     * VoconRecognizer#loadGrammar}.
     */
    public interface RebuildListener
    {
        /**
         * Represents a word that could not be added to the grammar during a rebuild.
         */
        public static class SkippedWord {

            private Word _word;
            private String _slotId;

            SkippedWord(Word word, String slotId)
            {
                _word = word;
                _slotId = slotId;
            }

            /**
             * @return The Id of the WordSlot that had a skipped word.
             */
            public String getSlotId() {
                return _slotId;
            }

            /**
             * @return The Word that could not be added.
             */
            public Word getWord() {
                return _word;
            }

        }

        /**
         * This method is called if a Grammar Rebuild complete successfully.
         *
         * @param grammar After a call to {@link VoconRecognizer#loadGrammar} or
         * {@link VoconRecognizer#loadStateAndGrammar}, a pointer to the same Grammar object will be returned.
         * After a call to {@link VoconRecognizer#loadState} a new grammar representing the loaded grammar will be returned.
         * After a call to {@link VoconRecognizer#clearGrammar}, a new empty grammar object will be returned.
         * @param skippedWords A list of words that could not be added to Vocon
         * because the pronunciation is not known. This object will never be
         * null, but may be empty if all words were added successfully.
         */
        void onComplete(Grammar grammar, List<SkippedWord> skippedWords);

        /**
         * A grammar format error or internal error occurred. The Error object
         * contains details about what happened.
         *
         * @param error The error (never null)
         */
        public abstract void onError(VoconError error);
    }

    /**
     * Listens for results or errors from a compilation started by {@link
     * VoconRecognizer#compileContext}.
     */
    public interface CompileListener
    {
        /**
         * The compilation succeeded.
         * @param fileName An ASR context file name, compiled from online application data,
         * which can be used by one of the start recognition methods.
         */
        void onComplete(String fileName);

        /**
         * A compilation error occurred. The Error object
         * contains details about what happened.
         *
         * @param error The error (never null)
         */
        void onError(VoconError error);
    }

    /**
     * Listens for a cleaning event.
     */
    public interface CleanListener
    {
        /**
         * The cleaning finished.
         */
        void onCleaned();
    }

    /**
     * Listens for the completion of {@link VoconRecognizer#initialize}.
     */
    public interface InitializeListener
    {
        /**
         * The initialization completed
         * @param vocon The recognizer that was initialized.
         * @param success false if the initialization was unsuccessful
         */
        void onLoaded(VoconRecognizer vocon, boolean success);
    }

    /**
     * Listens for the completion of {@link VoconRecognizer#release}.
     */
    public interface ReleaseListener
    {
        /**
         * The release completed
         * @param vocon The recognizer that was released
         */
        void onReleased(VoconRecognizer vocon);
    }

    /**
     * Provides callbacks for for certain events pertaining to loading the recognizer resources.
     */
    public interface ResourceLoadListener
    {
        /**
         * An attempt to load the resource has been made
         * @param vocon The recognizer.
         * @param success Is false if an error was encountered
         */
        public abstract void onLoaded(VoconRecognizer vocon, boolean success);
    }

    /**
     * Listens for the completion of {@link VoconRecognizer#saveState}.
     */
    public interface SaveListener
    {
        /**
         * The save completed.
         * @param vocon The recognizer that was saved
         * @param success false if the save was unsuccessful
         */
        void onSaved(VoconRecognizer vocon, boolean success);
    }

    /**
     * Listens for the completion of a kind of file saving operation
     */
    public interface FileSaveListener
    {
        /**
         * The save operation completed.
         * @param vocon The recognizer
         * @param result The result of the saving operation. Indicates success or failure.
         */
        void onFileSaved(VoconRecognizer vocon, FileOperationResult result);
    }

    /**
     * Listens for the completion of a kind of file loading operation
     */
    public interface FileLoadListener
    {
        /**
         * The load operation completed.
         * @param vocon The recognizer
         * @param result The result of the load operation. Indicates success or failure.
         */
        void onFileLoaded(VoconRecognizer vocon, FileOperationResult result);
    }

    /**
     * Provides callbacks for for certain events pertaining to releasing the vocalizer resources.
     */
    public interface ResourceReleaseListener
    {
        /**
         * An attempt to release the resource has been made
         * @param vocon The recognizer owning the resource that has been released
         */
        public abstract void onReleased(VoconRecognizer vocon);
    }

    /**
     * Listens for the completion of {@link VoconRecognizer#preloadContext}.
     */
    public interface PreloadContextListener
    {
        /**
         * The context preload completed.
         * @param context The VoCon context that has been preloaded
         * @param success Is false if an error was encountered
         */
        public abstract void onPreloaded(VoconContext context, boolean success);
    }

    /**
     * Languages supported by Vocon.
     */
    public static final class Languages extends LangSpecs
    {
    }

    /**
     * This function controls the amount of information about the VoconRecognizer
     * that will be written to the Android logs.
     *
     * Note: Do not use / disabled for your application release build.
     *
     * @param verbose True if detailed (verbose) status information for the VoconRecognizer should be written
     *                to the Android logs.
     *                False if only minimal details and error information should be written to  the Android logs.
     */
    public abstract void enableVerboseAndroidLogging(boolean verbose);


    /**
     * Create a {@link VoconRecognizer}. The VoconRecognizer will use the given worker thread. <p>
     * @param fileManager The file manager where VoconRecognizer should look for files.
     * @param workerThread The worker thread to use. If null, a dedicated one
     * will be created.
     */
    public static VoconRecognizer createVoconRecognizer(FileManager fileManager, NMTHandler workerThread)
    {
        return new VoconRecognizerImpl(fileManager, workerThread);
    }

    /**
     * Create a {@link VoconRecognizer}. The VoconRecognizer will create its own dedicated worker
     * thread. <p>
     * @param fileManager The file manager where VoconRecognizer should look for files.
     */
    public static VoconRecognizer createVoconRecognizer(FileManager fileManager)
    {
        return new VoconRecognizerImpl(fileManager);
    }

    /**
     * Returns the list of available language models. Each model consists of
     * a language and frequency.
     * <p>- This may be called from any thread, prior to initializing engine.
     * <br>- This method performs File I/O, so it may be preferable to call it
     * from a non-UI thread
     *
     * @return Supported language models.
     */
    public abstract List<VoconModelInfo> getAvailableModels();

    /**
     * Get a list of available language models for a given language. Each model
     * consists of a language and frequency.
     * <p>- This may be called from any thread, prior to initializing engine.
     * <br>- This method performs File I/O, so it may be preferable to call it
     * from a non-UI thread
     *
     * @param language The language to restrict results to (from {@link VoconLanguage}).
     * @return The list of available models.
     */
    public abstract List<VoconModelInfo> getAvailableModels(VoconLanguage language);

    /**
     * Load Vocon. This must be called prior to doing anything with the VoconRecognizer
     * object. Typically you'd call this right after constructing the VoconRecognizer
     * object. Note that any input filename indicated in VoconConfig must be available
     * through the FileManager passed in the constructor.
     * @param config The Vocon configuration
     * @param stateFileName The file where Vocon's state will be saved / loaded
     * @param listener Notified after the load completes
     */
    public abstract void initialize(VoconConfig config, String stateFileName, InitializeListener listener);

    /**
     * Load Vocon. This must be called prior to doing anything with the VoconRecognizer
     * object. Typically you'd call this right after constructing the VoconRecognizer
     * object. Note that any input filename indicated in VoconConfig must be available
     * through the FileManager passed in the constructor.
     * @param config The Vocon configuration
     * @param stateFileName The file where Vocon's state will be saved / loaded
     */
    public abstract void initialize(VoconConfig config, String stateFileName);

    /**
     * Set the VoCon front-end parameters prior to a recognition, needs to be used
     * along with {@link #startRecognition(AudioSource, List, SpeechDetectionListener, SignalListener, ResultListener)}.
     * @param params The front-end parameters. See {@link ParamSpecs.Fx} for possible values.
     */
    public abstract void setFxParams(Map<VoconParam, Integer> params);

    /**
    * Set the VoCon context parameters prior to a recognition, needs to be used
    * along with {@link #startRecognition(AudioSource, List, SpeechDetectionListener, SignalListener, ResultListener)}.
    * @param params The context parameters. See {@link ParamSpecs.Ctx} for possible values.
    */
   public abstract void setCtxParams(Map<VoconParam, Integer> params);

    /**
     * Initiates a recognition (the front-end parameters can be provided independently using {@link #setFxParams(Map)}).
     * @param audioSource The audioSource that provides the audio.
     * @param activeContexts The list of VoCon contexts to be used in recognition.
     * @param speechDetectionListener A listener that end-points speech.
     * @param signalListener A listener that provides the intermediate audio signal updates, or the warning description
     * if a warning occurred.
     * @param resultListener A listener that provides the results, or the error description
     * if an error occurred.
     */
    public abstract void startRecognition(AudioSource<AudioChunk> audioSource,
                                          List<VoconContext> activeContexts,
                                          SpeechDetectionListener speechDetectionListener,
                                          SignalListener signalListener,
                                          ResultListener resultListener);

    /**
     * Initiates a recognition.
     * @param audioSource The audioSource that provides the audio.
     * @param fxParams The front-end parameters. See {@link ParamSpecs.Fx} for possible values.
     * @param activeContexts The list of VoCon contexts to be used in recognition.
     * @param speechDetectionListener A listener that end-points speech.
     * @param signalListener A listener that provides the intermediate audio signal updates, or the warning description
     * if a warning occurred.
     * @param resultListener A listener that provides the results, or the error description
     * if an error occurred.
     */
    public abstract void startRecognition(AudioSource<AudioChunk> audioSource,
                                          Map<VoconParam, Integer> fxParams,
                                          List<VoconContext> activeContexts,
                                          SpeechDetectionListener speechDetectionListener,
                                          SignalListener signalListener,
                                          ResultListener resultListener);

    /**
     * Initiates a recognition.
     * @param audioSource The audioSource that provides the audio.
     * @param activeContexts The list of VoCon contexts to be used in recognition.
     * @param recogParam The recognition parameters {@link RecognitionParam}
     * @param speechDetectionListener A listener that end-points speech.
     * @param signalListener A listener that provides the intermediate audio signal updates, or the warning description
     * if a warning occurred.
     * @param resultListener A listener that provides the results, or the error description
     * if an error occurred.
     */
    public abstract void startRecognition(AudioSource<AudioChunk> audioSource,
                                          List<VoconContext> activeContexts,
                                          RecognitionParam recogParam,
                                          SpeechDetectionListener speechDetectionListener,
                                          SignalListener signalListener,
                                          ResultListener resultListener);

    /**
     * Prepare a VoCon instance to receive audio from a connected source and
     * simply detect start/end of speech.
     *
     * <p>Since an actual recognition is performed by NMT for this operation
     * (but the result is not returned), the application developer shall create
     * a minimal context .bnf file, using VoCon grammar compiler. The resulting
     * .fcf file shall be copied in the same location than the other VoCon lingware files.
     * The name of the file must be "eos_~language~.fcf" where "~language~" is the 3-letter
     * VoCon language code ("enu", "frf", etc.).
     *
     * @param audioSource The audioSource that provides the audio.
     * @param speechDetectionListener Listener to be notified on start of speech
     * and end of speech.
     */
    public abstract void startSpeechDetection(AudioSource<AudioChunk> audioSource,
                                              SpeechDetectionListener speechDetectionListener);

    /**
     * Prepare a VoCon instance to receive audio from a connected source and
     * simply detect start/end of speech.
     *
     * <p>Since an actual recognition is performed by NMT for this operation
     * (but the result is not returned), the application developer shall create
     * a minimal context .bnf file, using VoCon grammar compiler. The resulting
     * .fcf file shall be copied in the same location than the other VoCon lingware files.
     * The name of the file must be "eos_~language~.fcf" where "~language~" is the 3-letter
     * VoCon language code ("enu", "frf", etc.).
     *
     * @param audioSource The audio source to get audio from.
     * @param longSpeech True for longer sensitivity. False otherwise.
     * @param speechDetectionListener Listener to be notified on start of speech
     * and end of speech.
     */
    public abstract void startSpeechDetection(AudioSource<AudioChunk> audioSource, boolean longSpeech,
                                              SpeechDetectionListener speechDetectionListener);

    /**
     * Prepare a VoCon instance to receive audio from a connected source and
     * simply detect start/end of speech. Does not perform a recognition.
     *
     * @param audioSource The audio source to get audio from.
     * @param silenceDurationMs The silence duration (in ms) needed to declare end-of-speech.
     * @param speechDetectionListener Listener to be notified on start of speech
     * and end of speech.
     */
    public abstract void startSpeechDetection(AudioSource<AudioChunk> audioSource, int silenceDurationMs,
                                              SpeechDetectionListener speechDetectionListener);

    /**
     * Prepare a VoCon instance to receive audio from a connected source and
     * simply detect start/end of speech. Does not perform a recognition.
     *
     * @param audioSource The audio source to get audio from.
     * @param silenceDurationMs The silence duration (in ms) needed to declare end-of-speech.
     * @param speechDetectionListener Listener to be notified on start of speech
     * and end of speech.
     * @param signalListener Listener to be notified of audio signal updates (SNR, quality warnings).
     * and end of speech. May be null.
     */
    public abstract void startSpeechDetection(AudioSource<AudioChunk> audioSource, int silenceDurationMs,
                                              SpeechDetectionListener speechDetectionListener,
                                              SignalListener signalListener);

    /**
     * Prepare a VoCon instance to receive audio from a connected source and
     * simply detect start/end of speech. Does not perform a recognition.
     *
     * @param audioSource The audio source to get audio from.
     * @param silenceDurationMs The silence duration (in ms) needed to declare end-of-speech.
     * @param resetChannel true to reset the VoCon front-end channel, else false. Default: true.
     * @param speechDetectionListener Listener to be notified on start of speech
     * and end of speech.
     * @param signalListener Listener to be notified of audio signal updates (SNR, quality warnings).
     * and end of speech. May be null.
     */
    public abstract void startSpeechDetection(AudioSource<AudioChunk> audioSource, int silenceDurationMs, boolean resetChannel,
                                              SpeechDetectionListener speechDetectionListener,
                                              SignalListener signalListener);

    /**
     * Initiates a constrained, timed recognition.
     * <p>This is a just convenience method which wraps {@link #startRecognition(AudioSource, Map, List, SpeechDetectionListener, SignalListener, ResultListener)}
     * with the proper {@link VoconParam} recognition parameters.
     * @param audioSource The audioSource that provides the audio.
     * @param activeContext The HostContext to be used in recognition.
     * @param speechDetectionListener A listener that end-points speech.
     * @param signalListener A listener that provides the intermediate audio signal updates, or the warning description
     * if a warning occurred.
     * @param resultListener A listener that provides the results, or the error description
     * if an error occurred.
     */
    public abstract void startTimedRecognitionMode(AudioSource<AudioChunk> audioSource,
                                                   VoconContext activeContext,
                                                   SpeechDetectionListener speechDetectionListener,
                                                   SignalListener signalListener,
                                                   ResultListener resultListener);

    /**
     * Initiates a fuzzy match recognition.
     * <p>
     * This is a just convenience method which wraps {@link #startRecognition(AudioSource, Map, List, SpeechDetectionListener, SignalListener, ResultListener)}
     * with the proper {@link VoconParam} recognition parameters for fuzzy matching.
     * <p>
     * The Fuzzy match result will be truncated after the 10th entry and expanded if the number of rows of an interpretation is smaller or equal to 10.
     * @param audioSource The audioSource that provides the audio.
     * @param activeContext The HostContext to be used in recognition.
     * @param speechDetectionListener A listener that end-points speech.
     * @param signalListener A listener that provides the intermediate audio signal updates, or the warning description
     * if a warning occurred.
     * @param resultListener A listener that provides the results, or the error description
     * if an error occurred.
     */
    public abstract void startFuzzyMatchRecognition(AudioSource<AudioChunk> audioSource,
                                                    VoconContext activeContext,
                                                    SpeechDetectionListener speechDetectionListener,
                                                    SignalListener signalListener,
                                                    ResultListener resultListener);

    /**
     * Initiates a wake-up word recognition.
     * @param audioSource The audio source to get audio from.
     * @param wakeupPhrases The active wake-up phrases. These can't be too short.
     * @param penalty Penalty to allow trade-off between False Accept and False Reject.
     * A positive penalty will increase False Accept and lower False Reject.
     * @param resultListener Listener to be notified when a wake-up phrase is
     * detected.
     */
    public abstract void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                         List<String> wakeupPhrases,
                                         int penalty,
                                         ResultListener resultListener);

    /**
     * Initiates a wake-up word recognition.
     * @param audioSource The audio source to get audio from.
     * @param fxParams The front-end parameters. See {@link ParamSpecs.Fx} for possible values.
     * @param wakeupPhrases The active wake-up phrases. These can't be too short.
     * @param penalty Penalty to allow trade-off between False Accept and False Reject.
     * A positive penalty will increase False Accept and lower False Reject.
     * @param resultListener Listener to be notified when a wake-up phrase is
     * detected.
     */
    public abstract void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                         Map<VoconParam, Integer> fxParams,
                                         List<String> wakeupPhrases,
                                         int penalty,
                                         ResultListener resultListener);

    /**
     * Initiates a seam-less wake-up word recognition.
     * @param audioSource The audio source to get audio from.
     * @param wakeupPhrases The active wake-up phrases. These can't be too short.
     * @param penalty Penalty to allow trade-off between False Accept and False Reject.
     * A positive penalty will increase False Accept and lower False Reject.
     * @param pauseDurationMs The expected pause duration (in ms) to differentiate
     * between a regular wake-up or a seam-less wake-up. Passing 0 is equivalent to
     * call {@link #startWakeupMode(AudioSource, List, int, ResultListener)}.
     * @param resultListener Listener to be notified when a wake-up phrase is
     * detected.
     */
    public abstract void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                         List<String> wakeupPhrases,
                                         int penalty, int pauseDurationMs,
                                         ResultListener resultListener);

    /**
     * Initiates a seam-less wake-up word recognition.
     * @param audioSource The audio source to get audio from.
     * @param fxParams The front-end parameters. See {@link ParamSpecs.Fx} for possible values.
     * @param wakeupPhrases The active wake-up phrases. These can't be too short.
     * @param penalty Penalty to allow trade-off between False Accept and False Reject.
     * A positive penalty will increase False Accept and lower False Reject.
     * @param pauseDurationMs The expected pause duration (in ms) to differentiate
     * between a regular wake-up or a seam-less wake-up. Passing 0 is equivalent to
     * call {@link #startWakeupMode(AudioSource, List, int, ResultListener)}.
     * @param resultListener Listener to be notified when a wake-up phrase is
     * detected.
     */
    public abstract void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                         Map<VoconParam, Integer> fxParams,
                                         List<String> wakeupPhrases,
                                         int penalty, int pauseDurationMs,
                                         ResultListener resultListener);

    /**
     * Initiates a seam-less wake-up word recognition.
     * @param audioSource The audio source to get audio from.
     * @param wakeupPhrases The active wake-up phrases. These can't be too short.
     * @param penalty Penalty to allow trade-off between False Accept and False Reject.
     * A positive penalty will increase False Accept and lower False Reject.
     * @param resultListener Listener to be notified when a wake-up phrase is
     * detected.
     */
    public abstract void startSeamlessWakeupMode(AudioSource<AudioChunk> audioSource,
                                                 List<String> wakeupPhrases,
                                                 int penalty,
                                                 ResultListener resultListener);

    /**
     * Initiates a seam-less wake-up word recognition.
     * @param audioSource The audio source to get audio from.
     * @param fxParams The front-end parameters. See {@link ParamSpecs.Fx} for possible values.
     * @param wakeupPhrases The active wake-up phrases. These can't be too short.
     * @param penalty Penalty to allow trade-off between False Accept and False Reject.
     * A positive penalty will increase False Accept and lower False Reject.
     * @param resultListener Listener to be notified when a wake-up phrase is
     * detected.
     */
    public abstract void startSeamlessWakeupMode(AudioSource<AudioChunk> audioSource,
                                                 Map<VoconParam, Integer> fxParams,
                                                 List<String> wakeupPhrases,
                                                 int penalty,
                                                 ResultListener resultListener);

    /**
     * Terminates the current recognition.
     * VoCon will try to recognize all the audio received up until this method is called.
     * The resultListener will be used to deliver the resulting recognition results.
     */
    public abstract void stopListening();

    /**
     * Cancels the current recognition.
     * VoCon will ignore all received audio, and no results will be sent back.
     */
    public abstract void cancelRecognition();

    /**
     * Releases the resources created by the recognizer.
     * This method should be called when the recognizer is no longer needed. It is
     * a non-blocking call.
     * @param listener A listener that confirms when the recognizer is released
     */
    public abstract void release(ReleaseListener listener);

    /**
     * Releases the resources created by the recognizer.
     * This method should be called when the recognizer is no longer needed. It is
     * a non-blocking call.
     */
    public abstract void release();

    /**
     * Check whether a list of phrases can be used for WakeupMode.
     * If all phrases are suitable, wakeupCheckListener will be called with an empty list of phrases.
     * If any phrase is not suitable, wakeupCheckListener will be called with a list of phrases that cannot be used.
     *
     * @param wakeupPhrases The phrases to check for suitability.
     * @param wakeupCheckListener The listener that will receive a callback when the check is complete.
     */
    public abstract void checkWakeupPhrases(List<String> wakeupPhrases, WakeupCheckListener wakeupCheckListener);

    /**
     * Load the given VoCon grammar into the VoCon engine. This process is
     * asynchronous. The listener will be notified when it completes. <p>
     * @param grammar The grammar containing the word slots and constraints.
     * @param listener Will be notified when the rebuild has finished.
     */
    public abstract void loadGrammar(Grammar grammar, RebuildListener listener);

    /**
     * Load the given VoCon grammar into the VoCon engine. This process is
     * asynchronous. The listener will be notified when it completes. <p>
     * @param grammar The grammar containing the word slots and constraints.
     * @param onlineCompile The flag indicating dynamic words will be compiled online.
     * By default, dynamic words will be added to word slot one by one.
     * <p>
     * Note: This option is only applicable to domain "name" {@link DynamicSlot.Domains}
     * </p>
     * @param listener Will be notified when the rebuild has finished.
     *
     * @deprecated use {@link #loadGrammar(Grammar, GrammarLoadParam, RebuildListener)} instead.
     */
    @Deprecated
    public abstract void loadGrammar(Grammar grammar, boolean onlineCompile, RebuildListener listener);


    /**
     * Load the given VoCon grammar into the VoCon engine. This process is
     * asynchronous. The listener will be notified when it completes. <p>
     * @param grammar The grammar containing the word slots and constraints.
     * @param loadingParam The configuration for loading grammar {@link GrammarLoadParam}
     * @param listener Will be notified when the rebuild has finished.
     */
    public abstract void loadGrammar(Grammar grammar, GrammarLoadParam loadingParam, RebuildListener listener);

    /**
     * Clear grammar from the VoCon engine.
     * @param listener Will be notified when grammar has been cleared.
     */
    public abstract void clearGrammar(RebuildListener listener);

    /**
     * Cancel a rebuild that is in progress.
     */
    public abstract void cancelRebuild();

    /**
     * Load a resource (user dictionary or rule set).
     * Should be called prior to {@link #loadGrammar(Grammar grammar, RebuildListener listener)} to take effect.
     * @param fileName The resource file name (must be ".dcb" for a user dictionary, ".rst" for a rule set).
     * @param listener Optional listener for resource loading status updates
     */
    public abstract void loadResource(String fileName, ResourceLoadListener listener);

    /**
     * Release a resource that was previously loaded using
     * {@link #loadResource(String fileName, ResourceLoadListener listener)}.
     * @param fileName The resource file name (must be ".dcb" for a user dictionary, ".rst" for a rule set).
     * @param listener Optional listener for resource releasing status updates
     */
    public abstract void releaseResource(String fileName, ResourceReleaseListener listener);

    /**
     * Compile a fuzzy-match context.
     * @param entries The list of {@link FuzzyMatchData} to compile: <p><b>The list must be composed of homogeneous
     * entries with properly sanitized fields to ensure successful compilation and subsequent recognition.</b></p>
     * @param fileName The compiled context file name (ex: "fuzzysearch-08192013-121503.zcf").
     * @param listener Will be notified when context has been compiled.
     *
     * @deprecated use {@link #compileFuzzyMatchContext} instead
     */
    @Deprecated
    public abstract void compileContext(List<FuzzyMatchData> entries, String fileName, CompileListener listener);

    /**
     * Cancel the fuzzy-match context compilation.
     *
     * @deprecated use {@link #cancelFuzzyMatchCompilation} instead
     */
    @Deprecated
    public abstract void cancelCompilation();

    /**
     * Compile a fuzzy-match context.
     * @param entries The list of {@link FuzzyMatchData} to compile: <p><b>The list must be composed of homogeneous
     * entries with properly sanitized fields to ensure successful compilation and subsequent recognition.</b></p>
     * @param fileName The compiled context file name (ex: "fuzzysearch-08192013-121503.zcf").
     * @param listener Will be notified when context has been compiled.
     */
    public abstract void compileFuzzyMatchContext(List<FuzzyMatchData> entries, String fileName, CompileListener listener);

    /**
     * Cancel the fuzzy-match context compilation.
     */
    public abstract void cancelFuzzyMatchCompilation();

    /**
     * Compile a BNF context.
     * @param fileName The compiled context file name (ex: "sample.bnf").
     * @param listener Will be notified when context has been compiled.
     */
    public abstract void compileBnfContext(String fileName, CompileListener listener);

    /**
     * Clear a compiled (regular or fuzzy-match) context.
     * @param fileName The compiled context file name.
     * @param listener Will be notified when context has been cleaned.
     */
    public abstract void clean(String fileName, CleanListener listener);

    /**
     * Clear a list of compiled (regular or fuzzy-match) contexts.
     * @param fileNames The compiled context file names.
     * @param listener Will be notified when contexts has been cleaned.
     */
    public abstract void clean(List<String> fileNames, CleanListener listener);

    /**
     * Save current model set, grammar, and vocabulary to disk. <br>
     *
     * @param listener Notified after the save completes.
     */
    public abstract void saveState(SaveListener listener);

    /**
     * Like {@link #saveState(SaveListener)}, but no listener.
     */
    public abstract void saveState();

    /**
     * Load VoCon model set and stored vocabulary from disk.<br>
     *
     * @param rebuildListener Called when the VoCon state is loaded.
     */
    public abstract void loadState(RebuildListener rebuildListener);

    /**
     * Combined {@link #loadState} and {@link #loadGrammar}. If loading the state
     * fails, the grammar will be fully rebuild; otherwise, the grammar will be
     * loaded normally on top of the loaded state.
     * @param grammar The grammar to load. If  null, the behavior will be the
     * same as {@link #loadState}.
     * @param rebuildListener Will be notified when the state and/or grammar
     * has been loaded.
     */
    public abstract void loadStateAndGrammar(Grammar grammar,
                                             RebuildListener rebuildListener);

    /**
     * Combined {@link #loadState} and {@link #loadGrammar}. If loading the state
     * fails, the grammar will be fully rebuild; otherwise, the grammar will be
     * loaded normally on top of the loaded state.
     * @param grammar The grammar to load. If  null, the behavior will be the
     * same as {@link #loadState}.
     * @param onlineCompile The flag indicating dynamic words will be compiled online.
     * By default, dynamic words will be added to word slot one by one.
     * <p>
     * Note: This option is only applicable to domain "name" {@link DynamicSlot.Domains}
     * </p>
     * @param rebuildListener Will be notified when the state and/or grammar
     * has been loaded.
     *
     * @deprecated use {@link #loadStateAndGrammar(Grammar, GrammarLoadParam, RebuildListener)} instead.
     */
    @Deprecated
    public abstract void loadStateAndGrammar(Grammar grammar,
                                             boolean onlineCompile,
                                             RebuildListener rebuildListener);

    /**
     * Combined {@link #loadState} and {@link #loadGrammar}. If loading the state
     * fails, the grammar will be fully rebuild; otherwise, the grammar will be
     * loaded normally on top of the loaded state.
     * @param grammar The grammar to load. If  null, the behavior will be the
     * same as {@link #loadState}.
     * @param loadingParam The configuration for loading grammar {@link GrammarLoadParam}
     * @param rebuildListener Will be notified when the state and/or grammar
     * has been loaded.
     */
    public abstract void loadStateAndGrammar(Grammar grammar, GrammarLoadParam loadingParam, RebuildListener rebuildListener);

    /**
     * Preload a VoCon context.
     *
     * @param context The context to preload.
     * @param listener Called when the context is preloaded.
     */
    public abstract void preloadContext(VoconContext context, PreloadContextListener listener);

    /**
     * Unload a preloaded VoCon context.
     *
     * @param context The context to unload.
     */
    public abstract void unloadContext(VoconContext context);

    /**
     * Stops VoCon from giving the following results if using multiple contexts.
     *
     * <p>
     * Note:
     * #1 The {@link RecognitionParam.Builder#setBlockingIssueResults()} must be set and passed in {@link #startRecognition(AudioSource, List, RecognitionParam, SpeechDetectionListener, SignalListener, ResultListener)}
     * #2 It must be called from within {@link ResultListener#onResult(VoconResult)} so that the effect
     * is taken into account for the next following VoCon result.
     * Otherwise the effect is undefined.
     * </p>
     */
    public abstract void stopRecognitionResults();

    /**
     * Saves the current acoustic model adaptation to file.
     *
     * WARNING: This method will preemptively cancel any active recognition or rebuild operation.
     *
     * @param fileName The file name where the adaptation data will be saved.
     * @param saveListener Listener that will indicate the success or failure of this operation.
     */
    public abstract void saveAcousticModelAdaptation(String fileName, FileSaveListener saveListener);

    /**
     * Loads an acoustic model adaptation from file.
     *
     * Passing a null fileName resets the Acoustic Model Adaptation to factory state.
     *
     * WARNING: This method will preemptively cancel any active recognition or rebuild operation.
     *
     * @param fileName The full file name from where the adaptation data will be loaded.
     * @param loadListener Listener that will indicate the success or failure of this operation.
     */
    public abstract void loadAcousticModelAdaptation(String fileName, FileLoadListener loadListener);

    /**
     * Saves the current front end adaptation to file.
     *
     * WARNING: This method will preemptively cancel any active recognition or rebuild operation.
     *
     * @param fileName The file name where the adaptation data will be saved.
     * @param saveListener Listener that will indicate the success or failure of this operation.
     */
    public abstract void saveFxAdaptation(String fileName, FileSaveListener saveListener);

    /**
     * Loads a front end adaptation from file.
     *
     * Passing a null fileName resets the Acoustic Model Adaptation to factory state.
     *
     * WARNING: This method will preemptively cancel any active recognition or rebuild operation.
     *
     * @param fileName The full file name from where the adaptation data will be loaded.
     * @param loadListener Listener that will indicate the success or failure of this operation.
     */
    public abstract void loadFxAdaptation(String fileName, FileLoadListener loadListener);

     /**
     * Saves the current speaker profile to file.
     *
     * The use of this method is only relevant when using the SPEAKER_ADAPTATION VoconParam.
     *
     * WARNING: This method will preemptively cancel any active recognition or rebuild operation.
     *
     * @param fileName The file name where the user profile will be saved.
     * @param saveListener Listener that will indicate the success or failure of this operation.
     */
    public abstract void saveSpeakerProfile(String fileName, FileSaveListener saveListener);

    /**
     * Loads a speaker profile from file.
     *
     * Passing a null fileName will reset the speaker profile
     *
     * WARNING: This method will preemptively cancel any active recognition or rebuild operation.
     *
     * @param fileName The file name from where the user profile data will be loaded.
     * @param loadListener Listener that will indicate the success or failure of this operation.
     */
    public abstract void loadSpeakerProfile(String fileName, FileLoadListener loadListener);
}
