package com.nuance.dragon.toolkit.vocon;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.UUID;
import java.util.concurrent.LinkedBlockingQueue;

import junit.framework.Assert;

import org.json.JSONArray;
import org.json.JSONException;

import android.os.SystemClock;

import com.nuance.dragon.toolkit.Build;
import com.nuance.dragon.toolkit.audio.AudioChunk;
import com.nuance.dragon.toolkit.audio.AudioSink;
import com.nuance.dragon.toolkit.audio.AudioSource;
import com.nuance.dragon.toolkit.audio.AudioType;
import com.nuance.dragon.toolkit.audio.SpeechDetectionListener;
import com.nuance.dragon.toolkit.core.Word;
import com.nuance.dragon.toolkit.core.WordAction;
import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.util.internal.ThreadUtils;
import com.nuance.dragon.toolkit.vocon.FileOperationResult.OperationResult;
import com.nuance.dragon.toolkit.vocon.NativeVocon.ContextError;
import com.nuance.dragon.toolkit.vocon.NativeVocon.WakeupMode;
import com.nuance.dragon.toolkit.vocon.ParamSpecs.Ctx;

/**
 * Base class for VoCon implementations.
 * <p>
 * There will probably be one 'real' subclass of this ({@link VoconImpl}).
 */
abstract class VoconRecognizerBase extends VoconRecognizer
{
    /** The number of word classes to add at a time. Too few and things will be
     * slow; too many and there will be a delay when trying to interrupt a
     * rebuild. */
    private static final int WORDCLASS_BATCH_SIZE = 500;

    /** The wake-up default confidence threshold.
     */
    private static final int WAKEUP_CONFIDENCE_THRESHOLD = 4700;

    private final NativeVocon _nativeVocon;
    private Recognition _currentRecognition;
    private boolean _isInitialized;
    private boolean _isVoconCreated;
    private boolean _isReleased;
    private final NMTHandler _workerThreadHandler;
    private final NMTHandler _mainThreadHandler;
    private String _stateFileName; // Touched from worker thread
    private Grammar _grammar;
    private Rebuild _currentRebuild;
    private Compilation _currentCompilation;
    private Map<VoconParam, Integer> _fxParams, _ctxParams;

    private class Recognition extends AudioSink<AudioChunk>
    {
        private ResultListener _resultListener;
        private SpeechDetectionListener _speechStartListener;
        private SpeechDetectionListener _speechEndListener;
        private final SignalListener _signalListener;
        private final LinkedBlockingQueue<AudioChunk> _audioQueue; // Touched from both threads
        private AudioChunk _currentAudioChunk;
        private int _currentAudioChunkOffset;
        private boolean _started; // Touched from both threads
        private boolean _audioDone;
        private long _refWakeupTimestamp;
        private final Object _sync;
        private boolean isResultsBlocked; // Touched from both threads
        private boolean _isNoMoreResults;

        Recognition(ResultListener resultListener, SpeechDetectionListener speechDetectionListener, SignalListener signalListener)
        {
            super(null);

            _resultListener = resultListener;
            _speechStartListener = speechDetectionListener;
            _speechEndListener = speechDetectionListener;
            _signalListener = signalListener;
            _audioQueue = new LinkedBlockingQueue<AudioChunk>();
            _sync = new Object();
            isResultsBlocked = false;
            _isNoMoreResults = false;
        }

        void startRecognizing(final AudioSource<AudioChunk> audioSource, final List<VoconContext> contexts, final RecognitionParam recogParam)
        {
            final AudioType audioType = audioSource.getAudioType();
            _started = true;
            _nativeVocon.resetCancelAsr();
            _workerThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    // Create the contexts
                    for (int idx = 0; idx < contexts.size(); idx++)
                    {
                        if (contexts.get(idx).getPreloaded())
                            continue;

                        // Load context from file
                        if (ContextError.FATAL == _nativeVocon.createAsrContext(contexts.get(idx)))
                        {
                            Logger.error(this, "Unable to create context: " + contexts.get(idx).getId());
                            for (VoconContext context : contexts)
                                _nativeVocon.destroyContext(context);
                            handleStartError();
                            return;
                        }

                        updateComputedParam(contexts.get(idx), Ctx.ACCURACY.id);

                    }

                    // Start recognition.
                    boolean success = _nativeVocon.startRecognizing(recogParam.getFxParams(), contexts, audioType, false, new NativeVocon.RecognitionListener()
                    {
                        private int _currentResultCount = 0;
                        private final int _totalResultCount = contexts.size();

                        @Override
                        public void onStartAudioPull()
                        {
                            // Nothing do to, audio is already connected at this point.
                        }

                        @Override
                        public void onStopAudioPull()
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    disconnectAudioSource();
                                }
                            });
                        }

                        @Override
                        public void onSignalUpdate(final int snr, final int energy)
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (_signalListener != null) {
                                        VoconAudioInfo voconAudioInfo = new VoconAudioInfo();
                                        voconAudioInfo.snrDb = snr * 0.01f;
                                        voconAudioInfo.energyDb = energy * 0.01f;
                                        _signalListener.onUpdate(voconAudioInfo);
                                    }

                                }
                            });
                        }

                        @Override
                        public void onSignalWarning(final int code)
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (_signalListener != null)
                                        _signalListener.onWarning(new VoconWarning(code));
                                }
                            });
                        }

                        @Override
                        public boolean onHasAudioChunk()
                        {
                            return _currentAudioChunk != null;
                        }

                        @Override
                        public AudioChunk onGetAudioChunk(int maxLen)
                        {
                            if (poll())
                                return getNextAudioChunk(maxLen);
                            return null;
                        }

                        @Override
                        public void onCanceled()
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    disconnectAudioSource();
                                    handleError(new VoconError(VoconError.Reason.CANCELED));
                                }
                            });
                        }

                        @Override
                        public void onStartOfSpeech()
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (_speechStartListener != null)
                                    {
                                        _speechStartListener.onStartOfSpeech();
                                        _speechStartListener = null;
                                    }
                                }
                            });
                        }

                        @Override
                        public void onEndOfSpeech()
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (_speechEndListener != null)
                                    {
                                        _speechEndListener.onEndOfSpeech();
                                        _speechEndListener = null;
                                    }
                                }
                            });
                        }

                        @Override
                        public void onRecognitionResult(String json)
                        {
                            isResultsBlocked = recogParam.isBlockingIssueResults();
                            final VoconResult result = new VoconResult(json, VoconResult.REF_WAKEUP_TIMESTAMP_UNAVAILABLE);
                            final Runnable resultTask = new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    _currentResultCount++;
                                    if (_currentRecognition == Recognition.this)
                                    {
                                        final ResultListener resultListener = _resultListener;

                                        if (resultListener != null)
                                        {
                                            resultListener.onResult(result);
                                        }

                                        if (_currentResultCount >= _totalResultCount || _isNoMoreResults)
                                        {
                                            _started = false;
                                            disconnectAudioSource();

                                            _currentRecognition = null;
                                            _resultListener = null;
                                        }
                                    }
                                    if (recogParam.isBlockingIssueResults())
                                    {
                                        synchronized(_sync)
                                        {
                                            isResultsBlocked = false;
                                            _sync.notifyAll();
                                        }
                                    }
                                }
                            };
                            if (recogParam.isBlockingIssueResults())
                            {
                                synchronized(_sync)
                                {
                                    _mainThreadHandler.post(resultTask);

                                    while (isResultsBlocked)
                                    {
                                        try {
                                            _sync.wait();
                                        } catch (InterruptedException e) {
                                            Logger.error(Recognition.this, "Cannot block on workerthread: " + e);
                                        }
                                    }
                                }
                            }
                            else
                            {
                                _mainThreadHandler.post(resultTask);
                            }
                        }
                    });

                    // Clean-up current recognition contexts
                    for (VoconContext context : contexts)
                    {
                        if (context.getPreloaded())
                            continue;

                        _nativeVocon.destroyContext(context);
                    }

                    if (!success) {
                        Logger.error(this, "Unable to start recognizer.");
                        handleStartError();
                    }
                }
            });
            connectAudioSource(audioSource);
        }

        void startEndpointing(final AudioSource<AudioChunk> audioSource, final Map<VoconParam, Integer> fxParams, final boolean resetChannel)
        {
            final AudioType audioType = audioSource.getAudioType();
            _started = true;
            _nativeVocon.resetCancelAsr();
            _workerThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    // Start recognition.
                    if(!_nativeVocon.startEndpointing(fxParams, audioType, resetChannel, new NativeVocon.RecognitionListener()
                    {
                        @Override
                        public void onStartAudioPull()
                        {
                            // Nothing do to, audio is already connected at this point.
                        }

                        @Override
                        public void onStopAudioPull()
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (_currentRecognition == Recognition.this)
                                    {
                                        _started = false;
                                        disconnectAudioSource();
                                        _currentRecognition = null;
                                    }
                                }
                            });
                        }

                        @Override
                        public void onSignalUpdate(final int snr, final int energy)
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (_signalListener != null) {
                                        VoconAudioInfo voconAudioInfo = new VoconAudioInfo();
                                        voconAudioInfo.snrDb = snr * 0.01f;
                                        voconAudioInfo.energyDb = energy * 0.01f;
                                        _signalListener.onUpdate(voconAudioInfo);
                                    }
                                }
                            });
                        }

                        @Override
                        public void onSignalWarning(final int code)
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (_signalListener != null)
                                        _signalListener.onWarning(new VoconWarning(code));
                                }
                            });
                        }

                        @Override
                        public boolean onHasAudioChunk()
                        {
                            return _currentAudioChunk != null;
                        }

                        @Override
                        public AudioChunk onGetAudioChunk(int maxLen)
                        {
                            if (poll())
                                return getNextAudioChunk(maxLen);
                            return null;
                        }

                        @Override
                        public void onCanceled()
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    disconnectAudioSource();
                                    handleError(new VoconError(VoconError.Reason.CANCELED));
                                }
                            });
                        }

                        @Override
                        public void onStartOfSpeech()
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (_speechStartListener != null)
                                    {
                                        _speechStartListener.onStartOfSpeech();
                                        _speechStartListener = null;
                                    }
                                }
                            });
                        }

                        @Override
                        public void onEndOfSpeech()
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (_speechEndListener != null)
                                    {
                                        _speechEndListener.onEndOfSpeech();
                                        _speechEndListener = null;
                                    }
                                }
                            });
                        }

                        @Override
                        public void onRecognitionResult(String json)
                        {
                            // Nothing do to, event is ignored in end-pointing mode.
                        }
                    })) {
                        Logger.error(this, "Unable to start recognizer.");
                        handleStartError();
                        return;
                    }
                }
            });
            connectAudioSource(audioSource);
        }

        void startWakeupMode(final AudioSource<AudioChunk> audioSource,
                             final Map<VoconParam, Integer> fxParams,
                             final List<String> wakeupPhrases,
                             final int penalty,
                             final int pauseDurationMs)
        {
            final int totalEventCount = (pauseDurationMs > 0 ? 2 : 1) * wakeupPhrases.size(); // each phrase is tied to 1 default WuW context + 1 seamless WuW
            final List<VoconResult> results = new ArrayList<VoconResult>(totalEventCount);
            _started = true;
            _nativeVocon.resetCancelAsr();
            startWakeupModeInternal(audioSource, fxParams, wakeupPhrases, pauseDurationMs, true, new ResultListener()
            {
                private VoconResult _intermediateBestResult = null;
                private VoconResult _finalBestResult = null;
                private final int _acceptConfidence = WAKEUP_CONFIDENCE_THRESHOLD - penalty;
                private boolean _matchFoundAndCommunicated = false;

                @Override
                public void onResult(final VoconResult result)
                {
                    if (Recognition.this != _currentRecognition)
                        return;

                    if (_started)
                    {
                        results.add(result);

                        //Seamless mode
                        if (pauseDurationMs > 0)
                        {
                            // Partial notification
                            if (results.size() == (totalEventCount / 2))
                            {
                                VoconResult partialBestResult = null;
                                for (VoconResult r : results)
                                {
                                    // Speech filter is only applicable for regular wake-up word contexts
                                    if (r.isSpeech())
                                    {
                                        int lastBestConfidence = partialBestResult != null ? partialBestResult.getConfidence() : 0;
                                        int currentConfidence = r.getConfidence();
                                        Logger.debug(this, "Current partial confidence: " + currentConfidence + ", last best confidence: " + lastBestConfidence);

                                        if(currentConfidence >= lastBestConfidence)
                                        {
                                            partialBestResult = r;
                                            partialBestResult.setRecognizedWakeupPhrase(wakeupPhrases);  // Wake-up type remains unknown
                                        }
                                    }
                                }

                                // Decide to propagate the intermediate result or not
                                if (partialBestResult != null && partialBestResult.getConfidence() >= _acceptConfidence)
                                {
                                    _intermediateBestResult = partialBestResult;

                                    // Fire partial notification to indicate that a wake-up occurred.
                                    if (_resultListener != null)
                                    {
                                        _resultListener.onResult(_intermediateBestResult);
                                    }
                                }
                            }

                            if (results.size() >= totalEventCount)
                            {
                                // Check GC best one
                                VoconResult bestGCResult = null;
                                for (int idx = totalEventCount / 2; idx < totalEventCount; idx++)
                                {
                                    VoconResult r = results.get(idx);
                                    int lastBestGCConfidence = bestGCResult != null ? bestGCResult.getConfidence() : 0;
                                    int currentConfidence = r.getConfidence();
                                    Logger.debug(this, "Current GC confidence: " + currentConfidence + ", last best GC confidence: " + lastBestGCConfidence);

                                    if (validateGCResult(r, pauseDurationMs) && currentConfidence >= lastBestGCConfidence)
                                    {
                                        bestGCResult = r;
                                        bestGCResult.setRecognizedWakeupPhrase(wakeupPhrases);
                                        bestGCResult.setRecognizedWakeupType();
                                    }
                                }
                                if (bestGCResult != null && bestGCResult.getConfidence() > _acceptConfidence)
                                    _finalBestResult = bestGCResult;

                                Logger.debug(this, "seamless GC final result: " + (_finalBestResult == null ? null : _finalBestResult.toJSON()));

                                if (_finalBestResult == null && _intermediateBestResult != null)
                                {
                                    _finalBestResult = _intermediateBestResult;
                                    _finalBestResult.setRecognizedWakeupType();
                                }
                                results.clear();

                                handleFinalResult();
                            }
                        }
                        else
                        {
                            // Default mode
                            if (results.size() >= totalEventCount)
                            {
                                // All results came back from VoCon, find the best one
                                _finalBestResult = _intermediateBestResult;
                                for (VoconResult r : results)
                                {
                                    int lastBestConfidence = _finalBestResult != null ? _finalBestResult.getConfidence() : 0;
                                    int currentConfidence = r.getConfidence();
                                    Logger.debug(this, "Current confidence: " + currentConfidence + ", last best confidence: " + lastBestConfidence);

                                    if (currentConfidence >= lastBestConfidence)
                                    {
                                        _finalBestResult = r;
                                        _finalBestResult.setRecognizedWakeupPhrase(wakeupPhrases);
                                        _finalBestResult.setRecognizedWakeupType();
                                    }
                                }
                                results.clear();

                                handleFinalResult();
                            }
                        }
                    }
                    else
                    {
                        handleError(new VoconError(VoconError.Reason.NO_RESULT));
                    }
                }

                @Override
                public void onError(final VoconError error)
                {
                    if (Recognition.this != _currentRecognition)
                        return;

                    // MINOR HACK: if error is "CANCELED", and result has been sent, do nothing. In automatic restart
                    // mode, we have to cancel the recognition ourselves once WakeUp match is found, this is not an error.
                    if(error.getReasonCode() == VoconError.Reason.CANCELED && _matchFoundAndCommunicated == true)
                    {
                        Logger.debug(this, "Internally triggered cancellation, since WuW is detected.");
                    }
                    else if (_started)
                    {
                        _started = false;
                        disconnectAudioSource();
                        handleError(error);
                    }
                    else
                    {
                        handleError(new VoconError(VoconError.Reason.NO_RESULT));
                    }
                }

                private void handleFinalResult()
                {
                    Logger.debug(this, "final final result: " + (_finalBestResult == null ? null : _finalBestResult.toJSON()));
                    // Decide to propagate the result or keep going
                    if (_finalBestResult != null && _finalBestResult.getConfidence() >= _acceptConfidence)
                    {
                        stop(true);
                        disconnectAudioSource();
                        handleResult(_finalBestResult);
                        _matchFoundAndCommunicated = true;
                    }
                    else if (!_audioDone)
                    {
                        Logger.debug(this, "Wake-up mode: low confidence, ignoring result.");

                        // Restart wake-up mode recognition
                        _intermediateBestResult = null;
                    }
                    else
                    {
                        // in case result confidence score doesn't pass threshold and audio is stopped, cancel the recognition because
                        // it is on continuous mode
                        stop(true);
                        disconnectAudioSource();
                        handleError(new VoconError(VoconError.Reason.NO_RESULT));
                    }
                }

                private boolean validateGCResult(VoconResult result, int silenceDurationMs)
                {
                    boolean valid = false;
                    JSONArray nbest = result.getChoiceList();

                    if (nbest != null && nbest.length() > 0)
                    {
                        try {
                            JSONArray items = nbest.getJSONObject(0).getJSONArray("_items");
                            org.json.JSONObject lastItem = items.getJSONObject(items.length() - 1);
                            if (lastItem.getString("_orthography").equals("<...>"))
                            {
                                // Specified by VoCon research to avoid seamless FAs.
                                int openSpeechDuration = lastItem.getInt("_endTimeMs") - lastItem.getInt("_beginTimeMs");
                                Logger.debug(this, "Open speech duration following wake-up word: " + openSpeechDuration + " ms.");

                                if (openSpeechDuration - silenceDurationMs > 200)
                                {
                                    // Seamless wake-up with long enough open speech
                                    valid = true;
                                }
                            }
                        } catch (JSONException e) {
                            Logger.error(this, "Unable to compute wake-up result validity.", e);
                        }
                    }
                    return valid;
                }
            });
            connectAudioSource(audioSource);
        }

        private void startWakeupModeInternal(final AudioSource<AudioChunk> audioSource,
                                             final Map<VoconParam, Integer> fxParams,
                                             final List<String> wakeupPhrases,
                                             final int pauseDurationMs, final boolean resetChannel,
                                             final ResultListener innerResultListener)
        {
            final AudioType audioType = audioSource.getAudioType();
            final List<VoconContext> contexts = new ArrayList<VoconContext>(wakeupPhrases.size());

            final Map<VoconParam, Integer> defaultCtxParams = new HashMap<VoconParam, Integer>();
            defaultCtxParams.put(ParamSpecs.Ctx.TSILENCE_FX, 300);
            defaultCtxParams.put(ParamSpecs.Ctx.TSILENCE, 600);

            // Seamless EOS parameters are clamped to slightly bigger values
            // than the default wake-up parameters: this will guarantee that
            // the regular results will be emitted before the seamless results.
            final Map<VoconParam, Integer> seamlessCtxParams = new HashMap<VoconParam, Integer>();
            if (pauseDurationMs <= 320)
            {
                seamlessCtxParams.put(ParamSpecs.Ctx.TSILENCE_FX, 320);
                seamlessCtxParams.put(ParamSpecs.Ctx.TSILENCE, 640);
            }
            else
            {
                seamlessCtxParams.put(ParamSpecs.Ctx.TSILENCE_FX, pauseDurationMs);
                seamlessCtxParams.put(ParamSpecs.Ctx.TSILENCE, pauseDurationMs * 2);
            }

            _workerThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    _refWakeupTimestamp = VoconResult.REF_WAKEUP_TIMESTAMP_UNAVAILABLE;

                    // Set the WUW contexts
                    for (String wakeupPhrase : wakeupPhrases)
                    {
                        // Create the WuW isolated then the seamless context
                        for (int idx = 0 ; idx < (pauseDurationMs > 0 ? 2 : 1); idx++)
                        {
                            String modeStr = "default";
                            int mode = WakeupMode.DEFAULT;
                            Map<VoconParam, Integer> ctxParams = defaultCtxParams;
                            if (idx > 0)
                            {
                                modeStr = "seamless";
                                mode = WakeupMode.SEAMLESS;
                                ctxParams = seamlessCtxParams;
                            }

                            final VoconContext wakeupContext = new VoconContext(wakeupPhrase.hashCode() + "#" + modeStr, wakeupPhrase, mode);
                            wakeupContext.setParams(ctxParams);
                            if (ContextError.FATAL != _nativeVocon.createWuwContext(wakeupContext, new long[1]))
                            {
                                contexts.add(wakeupContext);
                            }
                            else
                            {
                                Logger.error(this, "Unable to create WuW context.");
                                for (VoconContext context : contexts)
                                    _nativeVocon.destroyContext(context);
                                handleStartError();
                                return;
                            }
                        }
                    }

                    // Start recognition.
                    if(!_nativeVocon.startRecognizing(fxParams, contexts, audioType, resetChannel, new NativeVocon.RecognitionListener()
                    {
                        private int _currentResultCount = 0;

                        @Override
                        public void onStartAudioPull()
                        {
                            logBufferedAudio("onStartAudioPull");
                        }

                        @Override
                        public void onStopAudioPull()
                        {
                            logBufferedAudio("onStopAudioPull");
                        }

                        @Override
                        public void onSignalUpdate(final int snr, final int energy)
                        {
                            // Nothing do to, event is ignored in WuW mode.
                        }

                        @Override
                        public void onSignalWarning(final int code)
                        {
                            // Nothing do to, event is ignored in WuW mode.
                        }

                        @Override
                        public void onCanceled()
                        {
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    innerResultListener.onError(new VoconError(VoconError.Reason.CANCELED));
                                }
                            });
                        }

                        @Override
                        public void onStartOfSpeech()
                        {
                            logBufferedAudio("onStartOfSpeech");
                        }

                        @Override
                        public void onEndOfSpeech()
                        {
                            logBufferedAudio("onEndOfSpeech");
                        }

                        @Override
                        public boolean onHasAudioChunk()
                        {
                            return _currentAudioChunk != null;
                        }


                        @Override
                        public AudioChunk onGetAudioChunk(int maxLen)
                        {
                            if (poll()) {
                                AudioChunk next = getNextAudioChunk(maxLen);
                                if (next != null && _refWakeupTimestamp == VoconResult.REF_WAKEUP_TIMESTAMP_UNAVAILABLE)
                                    _refWakeupTimestamp = next.audioTimestamp;
                                return next;

                            }

                            // Given that wakeup is in continuous mode, once no more audio found
                            // and no match found, cancel recognition
                            if(_currentResultCount == 0){
                                stop(true);
                            }

                            return null;
                        }

                        @Override
                        public void onRecognitionResult(String json)
                        {
                            logBufferedAudio("onRecognitionResult");
                            _currentResultCount++;

                            final VoconResult result = new VoconResult(json, _refWakeupTimestamp);
                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    innerResultListener.onResult(result);
                                }
                            });
                        }
                    })) {
                        Logger.error(this, "Unable to start recognizer.");
                        for (VoconContext context : contexts)
                            _nativeVocon.destroyContext(context);
                        handleStartError();
                        return;
                    }

                    // Clean-up contexts
                    for (VoconContext context : contexts)
                        _nativeVocon.destroyContext(context);
                }
            });
        }

        void stop(final boolean abort)
        {
            if (abort) {
                // Disable detection
                _speechStartListener = null;
                _speechEndListener = null;

                _nativeVocon.cancelAsr();
                _started = false;
            }

            // Setting this flag will have this module drain the queued audio before stopping.
            _audioDone = true;
        }

        void cancel()
        {
            stop(true);
            handleError(new VoconError(VoconError.Reason.CANCELED));
        }

        private void handleNewAudio(AudioSource<AudioChunk> source)
        {
            _audioQueue.addAll(source.getAllAudioChunksForSink(this));
            logBufferedAudio("handleNewAudio");
        }

        @Override
        public void chunksAvailable(AudioSource<AudioChunk> source)
        {
            if (this != _currentRecognition)
                return;

            // NOTE: To prevent fetching audio after the audioDone flag is set (e.g. stopListening),
            // only queue new audio when not set.
            if (!_audioDone) {
                handleNewAudio(source);
            }
        }

        @Override
        public void framesDropped(AudioSource<AudioChunk> source)
        {
            if (this != _currentRecognition)
                return;

            handleError(new VoconError(VoconError.Reason.AUDIO_FRAMES_DROPPED));
        }

        @Override
        public void sourceClosed(AudioSource<AudioChunk> source)
        {
            if (this != _currentRecognition)
                return;

            stop(false);
        }

        /** Called from worker thread */
        private void handleStartError()
        {
            _mainThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    stop(true);
                    handleError(new VoconError(VoconError.Reason.CANT_START));
                }
            });
        }

        /** Called from main thread */
        private void handleResult(VoconResult result)
        {
            if (_currentRecognition == Recognition.this)
            {
                if (result.getWakeupType() != VoconResult.WakeupType.UNKNOWN)
                    Logger.verbose(this, "[LATCHK] Engine combo wake-up latency: " + (SystemClock.uptimeMillis() - result.getWakeupEndTimestamp()) + " ms.");

                _currentRecognition = null;
                if (_resultListener != null)
                {
                    _resultListener.onResult(result);
                    _resultListener = null;
                }
            }
        }

        /** Called from main thread */
        private void handleError(VoconError error)
        {
            if (_currentRecognition == Recognition.this)
            {
                _currentRecognition = null;
                if (_resultListener != null)
                {
                    _resultListener.onError(error);
                    _resultListener = null;
                }
            }
        }

        /** Called from worker thread */
        private boolean poll()
        {
            // Poll until we see a chunk or the audio source disconnects
            while (_started &&
                   !_audioDone &&
                   _currentAudioChunk == null &&
                   _audioQueue.isEmpty())
            {
                 // We poll at 10 ms which is twice as fast as the smallest captured audio chunk
                 ThreadUtils.sleep(10);
            }

            // Return true if there is audio available, else false.
            return _started && (_currentAudioChunk != null || _audioQueue.size() > 0);
        }

        /** Called from worker thread */
        private AudioChunk getNextAudioChunk(final int requiredChunkLen)
        {
            if (_currentAudioChunk != null)
            {
                short[] frames = null;
                if (_currentAudioChunkOffset + requiredChunkLen < _currentAudioChunk.audioShorts.length)
                {
                    // Get a full sub-chunk out of the cached chunk
                    frames = new short[requiredChunkLen];
                    System.arraycopy(_currentAudioChunk.audioShorts, _currentAudioChunkOffset, frames, 0, frames.length);
                    _currentAudioChunkOffset = _currentAudioChunkOffset + frames.length;
                    return new AudioChunk(_currentAudioChunk.audioType, frames);
                }
                else
                {
                    if (_currentAudioChunkOffset < _currentAudioChunk.audioShorts.length)
                    {
                        // Get only what's left in the cached chunk
                        frames = new short[_currentAudioChunk.audioShorts.length - _currentAudioChunkOffset];
                        System.arraycopy(_currentAudioChunk.audioShorts, _currentAudioChunkOffset, frames, 0, frames.length);
                        AudioChunk leftover = new AudioChunk(_currentAudioChunk.audioType, frames);
                        _currentAudioChunk = null;
                        _currentAudioChunkOffset = 0;
                        return leftover;
                    }
                }
            }

            // Caching next chunk
            _currentAudioChunkOffset = 0;
            _currentAudioChunk = _audioQueue.peek();
            if (_currentAudioChunk != null)
            {
                _audioQueue.remove();
                return getNextAudioChunk(requiredChunkLen);
            }
            return null;
        }

        private void logBufferedAudio(String tag)
        {
            if (Build.DEBUG) {
                int duration = 0;
                for (AudioChunk c : _audioQueue) {
                    duration += c.audioDuration;
                }
                Logger.verbose(this, "Buffered audio [" + tag + "] : " + duration + " ms.");
            }
        }
    }

    /**
     * Internal class that keeps track of a single grammar rebuild.
     */
    private class Rebuild
    {
        private final Grammar _currentGrammar; // The grammar to build
        private final Grammar _oldGrammar;     // The new grammar (starts as a copy of the old grammar)
        private final boolean _onlineCompile;
        private final String _delimiters;
        private final RebuildListener _listener;
        private boolean _started;
        private final ArrayList<DynamicSlot.WordIterator> _savedWordSlots;

        Rebuild(Grammar grammar, Grammar oldGrammar, GrammarLoadParam loadingParam, RebuildListener listener)
        {
            _currentGrammar = grammar;
            _oldGrammar = new Grammar();
            if (oldGrammar != null) {
                _oldGrammar.addContexts(oldGrammar.getContexts(), true);
                _oldGrammar.addWordSlots(oldGrammar.getDynamicSlots(), true);
            }
            _onlineCompile = loadingParam.isOnlineCompile();
            _delimiters = loadingParam.getDelimiters();
            _listener = listener;
            _savedWordSlots = new ArrayList<DynamicSlot.WordIterator>();
        }

        /**
         * Cancel a rebuild that is in progress. It may not stop immediately.
         */
        void cancel()
        {
            if (_started)
            {
                synchronized (this)
                {
                    _started = false;
                }
            }
        }

        /**
         * Starts the rebuild process.<p>
         * 1. Run a job to synchronize all vocon contexts in grammar. Because contexts are
         * added to vocon object when recognizing, so nothing to add for native vocon here<br>
         * 2. Runs a number of jobs to add all word classes, which is to synchronize both context
         * and native vocon<br>
         * 3. Notifies of completion on the main thread.
         * @param forceFullRebuild true to force a full rebuild
         */
        void start(final boolean forceFullRebuild)
        {
            Checker.checkState(this, !_started);

            synchronized (this)
            {
                _started = true;
            }

            // to synchronize slots and words
            final Runnable wordsJob = new Runnable()
            {
                private Iterator<DynamicSlot> _slots = initSlots();
                private DynamicSlot.WordIterator _words;
                private String _slotName;
                private boolean isUpdated = false;
                List<String> _addedWords = new ArrayList<String>();
                List<Integer> _userIDs = new ArrayList<Integer>();

                private Iterator<DynamicSlot> initSlots()
                {
                    ArrayList<DynamicSlot> slots = new ArrayList<DynamicSlot>(_currentGrammar.getDynamicSlots().size());
                    slots.addAll(_currentGrammar.getDynamicSlots());
                    return slots.iterator();
                }

                @Override
                public void run()
                {
                    synchronized (Rebuild.this)
                    {
                        if (!_started)
                            return;
                    }

                    if (!_isVoconCreated) {
                        handleError(new VoconError(VoconError.Reason.NOT_INITIALIZED));
                        return;
                    }

                    if (_currentGrammar == null) {
                        handleError(new VoconError(VoconError.Reason.LOAD_GRAMMAR_ERROR));
                        return;
                    }

                    if (_words == null || !_words.hasNext())
                    {
                        if (_slotName != null) {
                            if (isUpdated) {
                                DynamicSlot slot = _oldGrammar.getDynamicSlotByName(_slotName);
                                if (_onlineCompile && slot.getDomain().equals(DynamicSlot.Domains.DYNAMICSLOT_NAME))
                                {
                                    if (!_nativeVocon.compileWordsToContext(_addedWords, _slotName))
                                    {
                                        Logger.error(this, "Failed online compiling dynamic words: " + _slotName);
                                        handleError(new VoconError(VoconError.Reason.LOAD_GRAMMAR_ERROR));
                                        return;
                                    }
                                }
                                if (!_nativeVocon.commitWordClass(_slotName)) {
                                    handleError(new VoconError(VoconError.Reason.CREATE_WORDCLASS));
                                    return;
                                }
                            }
                        }
                        if (_slots.hasNext())
                        {
                            isUpdated = false;
                            DynamicSlot nextSlot = _slots.next();
                            _slotName = nextSlot.getId();
                            DynamicSlot oldSlot = _oldGrammar.getDynamicSlotByName(_slotName);

                            if (!nextSlot.hasWordList()) {
                                Logger.error(VoconRecognizerBase.this, "DynamicSlot " + nextSlot.getId() + " requires an attached wordlist.");
                                handleError(new VoconError(VoconError.Reason.MISSING_WORDLIST));
                                return;
                            }

                            int rebuildType = nextSlot.getRebuildType();
                            if ((oldSlot == null) ||
                                (rebuildType == DynamicSlot.RebuildModes.DYNAMICSLOT_REBUILD_FULL))
                            {
                                // This DynamicSlot does not exist, or should be over-written.
                                _oldGrammar.addWordSlot(nextSlot.copy(), true);
                            }
                            //TODO: else if (oldSlot.domain != nextSlot.domain)

                            // the slot is new to vocon, need to rebuild
                            String wordClassName = nextSlot.getId();
                            if (!_nativeVocon.hasWordClass(wordClassName)) {
                                boolean isAdded = _nativeVocon.createWordClass(wordClassName);
                                if (!isAdded)
                                {
                                    handleError(new VoconError(VoconError.Reason.WORDSLOT_DEF_ERROR));
                                    return;
                                }
                                _words = nextSlot.getWords(true);
                                _savedWordSlots.add(_words);
                                isUpdated = true;

                            }
                            // forceFullRebuild means we need to rebuild all word list
                            else if (forceFullRebuild || rebuildType ==  DynamicSlot.RebuildModes.DYNAMICSLOT_REBUILD_FULL
                                    || _onlineCompile) {
                                _words = nextSlot.getWords(true);
                                _savedWordSlots.add(_words);

                            } else if (rebuildType == DynamicSlot.RebuildModes.DYNAMICSLOT_REBUILD_UPDATE) {
                                // This is not a rebuild, and the DYNAMICSLOT_REBUILD_UPDATE will add new words and remove
                                // deleted words
                                _words = nextSlot.getWords(false);
                                _savedWordSlots.add(_words);
                            }
                            // else {
                            //     If this is not a full rebuild, and the DynamicSlot already exists,
                            //     and the rebuild type is DYNAMICSLOT_REBUILD_SKIP_EXISTING, there is nothing to do.
                            // }

                            _workerThreadHandler.post(this);
                        } else
                        {
                            // Word slots are done, so start to sync with contentlist
                            // Finished, so save and notify the app
                            if (!_nativeVocon.saveVocon(_stateFileName, _oldGrammar))
                            {
                                Logger.error(VoconRecognizerBase.this, "Error saving Vocon state to: " + _stateFileName);
                                handleError(new VoconError(VoconError.Reason.FILE_ERROR));
                                return;
                            }

                            _mainThreadHandler.post(new Runnable()
                            {
                                @Override
                                public void run()
                                {
                                    if (_started)
                                    {
                                        _started = false;

                                        for (DynamicSlot.WordIterator it : _savedWordSlots)
                                            it.finished(true);

                                        _savedWordSlots.clear();
                                        _grammar.addContexts(_oldGrammar.getContexts(), true);
                                        _grammar.addWordSlots(_oldGrammar.getDynamicSlots(), true);

                                        if (_currentRebuild == Rebuild.this)
                                            _currentRebuild = null;

                                        _listener.onComplete(_currentGrammar, null);
                                    }
                                }
                            });

                        }
                    }
                    else
                    {
                        //Logger.warn(Rebuild.this, "(LXU): Begin to do vocon word job with slot: " + _slotName);
                        DynamicSlot slot = _oldGrammar.getDynamicSlotByName(_slotName);
                        boolean stat = true;
                        int count = 0;

                        for (int i = 0; i < WORDCLASS_BATCH_SIZE && _words.hasNext(); i++)
                        {
                            WordAction wa = _words.getNext();
                            if (wa.isAdded())
                            {
                                count++;
                                Word word = wa.getWord();
                                String wordString = word.getSurfaceForm();
                                if (wordString != null) {
                                    _addedWords.add(wordString);
                                }
                                _userIDs.add(word.getUserId());
                                isUpdated = true;
                            }
                            else
                            {
                                Word word = wa.getWord();
                                String wordString = word.getSurfaceForm();
                                if (wordString != null) {

                                    stat = _nativeVocon.removeWordClassMemeberInUTF8(_slotName, word.getUserId(), wordString, 1, _delimiters);
                                }

                                if (!stat)
                                {
                                    Logger.error(this, "Failed removing word class member: " + wa.getWord().getSurfaceForm());
                                    handleError(new VoconError(VoconError.Reason.REMOVE_WORD_ERROR));
                                    return;
                                }
                                isUpdated = true;
                            }
                        }
                        if (!_addedWords.isEmpty() &&
                            (!_onlineCompile || !slot.getDomain().equals(DynamicSlot.Domains.DYNAMICSLOT_NAME)))
                        {
                            int[] userIDArray = new int[count];
                            String[] wordArray = new String[count];
                            for (int i = 0; i < _userIDs.size(); i++)
                            {
                                userIDArray[i] = _userIDs.get(i);
                            }
                            stat = _nativeVocon.addWordClassBatchMembers(_slotName, userIDArray, _addedWords.toArray(wordArray), slot.getDomain(), _delimiters);
                            if (!stat)
                            {
                                handleError(new VoconError(VoconError.Reason.ADD_WORD_ERROR));
                                return;
                            }
                            _addedWords.clear();
                            _userIDs.clear();
                        }
                        //Logger.warn(Rebuild.this, "(LXU): End to do vocon word job with slot: " + _slotName + " " + count + " words added");
                        _workerThreadHandler.post(this);
                    }
                }
            };

            // to synchronize contexts in grammar and slots in every context,
            // not unloaded existing contexts and slots
            final Runnable contextJob = new Runnable()
            {

                @Override
                public void run()
                {
                    synchronized (Rebuild.this)
                    {
                        if (!_started)
                            return;
                    }

                    ArrayList<VoconContext> contexts =
                            new ArrayList<VoconContext>(_currentGrammar.getContexts().size());

                    contexts.addAll(_currentGrammar.getContexts());

                    for (VoconContext ctx : contexts) {
                        String ctxName = ctx.getId();
                        VoconContext oldCtx = _oldGrammar.getContextByName(ctxName);

                        // no context in old grammar, add it
                        if (oldCtx == null) {
                            _oldGrammar.addContext(ctx, true);
                        }
                        //there is old context, need to synchronize the word slots
                        else {
                            for (String id : ctx.getSlotIds()) {
                                boolean isAdded = oldCtx.addSlotId(id);
                                if (!isAdded)
                                    Logger.debug(VoconRecognizerBase.this, "slot:" + id + " already exists in context: " + oldCtx.getId());
                            }
                        }
                    }

                    // Contexts are done, so start adding word classes
                    _workerThreadHandler.post(wordsJob);
                }

            };

            // Kick off rebuild by building jobs
            _workerThreadHandler.post(contextJob);

        }


        /** Call this from the WORKER THREAD */
        private void handleError(final VoconError error)
        {
            _mainThreadHandler.post(new Runnable()
            {
                @Override
                public void run()
                {
                    if (_started)
                    {
                        _started = false;

//                        for (WordSlot.WordIterator it : _savedWordSlots)
//                            it.finished(false);
//
//                        _savedWordSlots.clear();
                        _listener.onError(error);
                        if (_currentRebuild == Rebuild.this) {
                            _currentRebuild = null;
                        }
                    }
                }
            });
        }
    }

    /**
     * Internal class that keeps track of a context compilation.
     */
    private class Compilation
    {
        private final CompileListener _listener;
        private boolean _started;

        Compilation(CompileListener listener)
        {
            _listener = listener;
        }

        /**
         * Cancel a compilation that is in progress. It may not stop immediately.
         */
        void cancel()
        {
            if (_started)
            {
                synchronized (this)
                {
                    _started = false;
                }
            }
        }

        /**
         * Starts the compilation process.
         * @param entries The fuzzy-match entries.
         * @param fileName The output file name.
         */
        void start(final List<FuzzyMatchData> entries, final String fileName)
        {
            Checker.checkState(this, !_started);

            synchronized (this)
            {
                _started = true;
            }

            final String tableUid = UUID.randomUUID().toString();
            final HashMap<String, Integer> columns = new HashMap<String, Integer>();
            final List<FuzzyMatchData> allEntries = new ArrayList<FuzzyMatchData>(entries.size());
            allEntries.addAll(entries);

            final Runnable completionRunnable = new Runnable()
            {
                @Override
                public void run()
                {
                    final boolean success = _nativeVocon.destroyFuzzyMatchTable(tableUid);
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            if (_started && success)
                            {
                                _started = false;
                                _currentCompilation = null;
                                _listener.onComplete(fileName);
                            }
                            else
                            {
                                // Delete the generated context for convenience
                                clean(fileName, new CleanListener()
                                {
                                    @Override
                                    public void onCleaned()
                                    {
                                        _started = false;
                                        _currentCompilation = null;
                                        _listener.onError(new VoconError(VoconError.Reason.CONTEXT_COMPILATION_ERROR));
                                    }
                                });
                            }
                        }
                    });
                }
            };

            final Runnable addRunnable = new Runnable()
            {
                @Override
                public void run()
                {
                    synchronized (Compilation.this)
                    {
                        if (!_started)
                        {
                            _workerThreadHandler.post(completionRunnable);
                            return;
                        }
                    }

                    boolean success = true;
                    FuzzyMatchData row = allEntries.remove(0);

                    for (Entry<String, FuzzyMatchData.Value> e : row.data.entrySet())
                    {
                        String columnName = e.getKey();
                        FuzzyMatchData.Value fuzzyMatchValue = e.getValue();

                        if (fuzzyMatchValue.type == FuzzyMatchData.Type.INTEGER)
                        {
                            success = _nativeVocon.addIntegerToColumn(tableUid, columns.get(columnName), (Integer)fuzzyMatchValue.object);
                        }
                        else if (fuzzyMatchValue.type == FuzzyMatchData.Type.STRING)
                        {
                            success = _nativeVocon.addStringToColumn(tableUid, columns.get(columnName), (String)fuzzyMatchValue.object);
                        }
                        else if (fuzzyMatchValue.type == FuzzyMatchData.Type.SPEAKABLE)
                        {
                            success = _nativeVocon.addSpeakableToColumn(tableUid, columns.get(columnName), (String)fuzzyMatchValue.object);
                        }
                        else
                        {
                            Assert.assertTrue("Invalid FuzzyMatchData type.", false);
                        }

                        if (!success)
                            break;
                    }

                    if (success)
                    {
                        success = _nativeVocon.commitFuzzyMatchRow(tableUid);
                        if (allEntries.size() > 0 && success)
                        {
                            _workerThreadHandler.post(this);
                        }
                        else
                            _workerThreadHandler.post(completionRunnable);
                    }
                    else
                        _workerThreadHandler.post(completionRunnable);
                }
            };

            final Runnable createRunnable = new Runnable()
            {
                @Override
                public void run()
                {
                    synchronized (Compilation.this)
                    {
                        if (!_started)
                        {
                            _workerThreadHandler.post(completionRunnable);
                            return;
                        }
                    }

                    boolean success = _nativeVocon.createFuzzyMatchTable(fileName, tableUid);
                    // create row for column names
                    if (success)
                    {
                        for (FuzzyMatchData row : allEntries)
                        {
                            for (Entry<String, FuzzyMatchData.Value> e : row.data.entrySet())
                            {
                                String columnName = e.getKey();
                                FuzzyMatchData.Value fuzzyMatchValue = e.getValue();
                                if (!columns.containsKey(columnName))
                                {
                                    final int id = _nativeVocon.createFuzzyMatchColumn(tableUid, columnName, fuzzyMatchValue.flags);
                                    if (id != -1) {
                                        columns.put(columnName, id);
                                    }
                                    else {
                                        success = false;
                                        break;
                                    }
                                }
                            }
                            if (!success) break;
                        }
                    }
                    if (success)
                    {
                        success = _nativeVocon.commitFuzzyMatchRow(tableUid);
                    }
                    _workerThreadHandler.post(success ? addRunnable : completionRunnable);
                }

            };

            // Kick-start compilation
            _workerThreadHandler.post(createRunnable);
        }
    }

    VoconRecognizerBase(NativeVocon nativeVocon, NMTHandler externalWorkerThreadHandler)
    {
        _nativeVocon = nativeVocon;
        _grammar = new Grammar();

        _mainThreadHandler = Factory.createNMTHandler();
        if (externalWorkerThreadHandler == null)
            _workerThreadHandler = getWorkerThreadHandler();
        else
            _workerThreadHandler = externalWorkerThreadHandler;

    }

    @Override
    public void enableVerboseAndroidLogging(boolean verbose) {
        _nativeVocon.enableVoconVerboseLogging(verbose);
    }

    @Override
    public List<VoconModelInfo> getAvailableModels()
    {
        ArrayList<NativeModelInfo> nativeModels = new ArrayList<NativeModelInfo>();

        boolean success = _nativeVocon.listAvailableModels(nativeModels);
        if (success)
        {
            ArrayList<VoconModelInfo> models = new ArrayList<VoconModelInfo>(nativeModels.size());
            for (NativeModelInfo nativeModel : nativeModels)
            {
                int sizeKb = nativeModel.sizeKb;
                for (VoconLanguage lang : LangSpecs.ALL_LANGUAGES)
                {
                    if (lang.name.equals(nativeModel.language))
                    {
                        models.add(new VoconModelInfo(lang, sizeKb));
                    }
                }
            }
            return models;
        }
        Logger.error(this, "Unable to detect supported VoCon models");
        return new ArrayList<VoconModelInfo>(0);
    }

    @Override
    public List<VoconModelInfo> getAvailableModels(VoconLanguage language)
    {
        List<VoconModelInfo> models = getAvailableModels();
        for (int i = 0; i < models.size(); i++)
        {
            VoconModelInfo info = models.get(i);
            if (!info.language.equals(language)) {
                models.remove(i);
                i--;
            }
        }
        return models;
    }

    @Override
    public void initialize(final VoconConfig config, final String stateFileName, final InitializeListener listener) {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, !_isInitialized, "Already initialized");
        Checker.checkArgForNull("config", config);

        _isInitialized = true;
        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _stateFileName = stateFileName;
                String language = config.language;
                String acModFile = config.acmodfile;
                String clcFile = config.clcfile;

                // Find the biggest Acoustic Model possible
                int size = config.size == -1 ? getSize(language) : config.size;
                if (acModFile == null && language != null)
                {
                    String regex = LangSpecs.getAcmodFileNameRegEx(language, size);
                    acModFile = _nativeVocon.getMatchedFileName(regex);
                }

                // Find the best CLC possible
                String domain = VoconDomain.UNSPECIFIED.equals(config.domain) ? getDomain(language) : config.domain;
                if (clcFile == null && language != null)
                {
                    String regex = LangSpecs.getClcFileNameRegEx(language, domain);
                    clcFile = _nativeVocon.getMatchedFileName(regex);
                }

                final boolean success = _nativeVocon.createVocon(acModFile, clcFile);
                if (success)
                {
                    _isVoconCreated = true;
                }

                _mainThreadHandler.post(new Runnable() {
                    @Override
                    public void run()
                    {
                        if (listener != null) {
                            listener.onLoaded(VoconRecognizerBase.this, success);
                        }
                    }
                });
            }
        });
    }

    @Override
    public void initialize(VoconConfig config, String stateFileName) {
        initialize(config, stateFileName, null);
    }

    @Override
    public void release(final ReleaseListener listener) {

        Checker.checkState(this, !_isReleased);
        if (!_isInitialized) {
            _isReleased = true;
            if (listener != null)
                listener.onReleased(this);
            return;
        }

        terminateCurrentRecognition();
        clearGrammar(null);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                // Destroy recognition engine
                if (_isVoconCreated) {
                    _nativeVocon.destroyVocon();
                }
                _isVoconCreated = false;
                _isReleased = true;
                if (listener != null) {
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            listener.onReleased(VoconRecognizerBase.this);
                        }
                    });
                }
            }
        });

    }

    @Override
    public void checkWakeupPhrases(final List<String> wakeupPhrases, final WakeupCheckListener wakeupCheckListener)
    {
        Checker.checkArgForNull("wakeupPhrases", wakeupPhrases);
        Checker.checkArgForNull("wakeupCheckListener", wakeupCheckListener);
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        _workerThreadHandler.post(new Runnable() {

            @Override
            public void run()
            {
                if (!_isVoconCreated) {
                    _mainThreadHandler.post(new Runnable() {
                        @Override
                        public void run()
                        {
                            wakeupCheckListener.onError(new VoconError(VoconError.Reason.NOT_INITIALIZED));
                        }});
                    return;
                }
                final WakeupCheckResult wakeupCheckResult = new WakeupCheckResult();
                long[] aQuality = new long[1];
                for (String wakeupPhrase : wakeupPhrases)
                {
                    VoconContext context = new VoconContext("wakeupcheck", wakeupPhrase, WakeupMode.DEFAULT);
                    _nativeVocon.createWuwContext(context, aQuality);
                    wakeupCheckResult.AddPhraseQuality(wakeupPhrase, (int) aQuality[0]);
                    _nativeVocon.destroyContext(context);
                }
                _mainThreadHandler.post(new Runnable() {
                    @Override
                    public void run()
                    {
                        wakeupCheckListener.onComplete(wakeupCheckResult);
                    }});
            }});
    }

    @Override
    public void release() {
        release(null);
    }

    @Override
    public void setFxParams(Map<VoconParam, Integer> params)
    {
        _fxParams = params;
    }

    @Override
    public void setCtxParams(Map<VoconParam, Integer> params)
    {
        _ctxParams = params;
    }

    @Override
    public void startSpeechDetection(AudioSource<AudioChunk> audioSource,
                                     SpeechDetectionListener speechDetectionListener)
    {
        startSpeechDetection(audioSource, 1500, true, speechDetectionListener, null);
    }

    @Override
    public void startSpeechDetection(AudioSource<AudioChunk> audioSource, boolean longSpeech,
                                     SpeechDetectionListener speechDetectionListener)
    {
        startSpeechDetection(audioSource, longSpeech ? 1500 : 900, true, speechDetectionListener, null);
    }

    @Override
    public void startSpeechDetection(AudioSource<AudioChunk> audioSource, int silenceDurationMs,
                                     SpeechDetectionListener speechDetectionListener)
    {
        startSpeechDetection(audioSource, silenceDurationMs, true, speechDetectionListener, null);
    }

    @Override
    public void startSpeechDetection(AudioSource<AudioChunk> audioSource, int silenceDurationMs,
                                     SpeechDetectionListener speechDetectionListener,
                                     SignalListener signalListener)
    {
        startSpeechDetection(audioSource, silenceDurationMs, true, speechDetectionListener, signalListener);
    }

    @Override
    public void startSpeechDetection(AudioSource<AudioChunk> audioSource, int silenceDurationMs, boolean resetChannel,
                                     SpeechDetectionListener speechDetectionListener,
                                     SignalListener signalListener)
    {
        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForCondition("silenceDurationMs", "> 0", silenceDurationMs > 0);
        Checker.checkArgForNull("speechDetectionListener", speechDetectionListener);

        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        Checker.checkState(this, _currentRebuild == null, "A rebuild is in progress");

        Map<VoconParam, Integer> fxParams = new HashMap<VoconParam, Integer>();
        fxParams.put(ParamSpecs.Fx.START_ENABLE, ParamSpecs.Boolean.TRUE);
        fxParams.put(ParamSpecs.Fx.MINSPEECH, 60);
        fxParams.put(ParamSpecs.Fx.TIMEOUT_LSILENCE, 0);
        fxParams.put(ParamSpecs.Fx.TIMEOUT_SPEECH, 10000);
        fxParams.put(ParamSpecs.Fx.TSILENCE, silenceDurationMs);
        fxParams.put(ParamSpecs.Fx.EVENT_TIMER, 100);
        fxParams.put(ParamSpecs.Fx.KNOWN_SPEAKER_CHANGES, ParamSpecs.Boolean.FALSE);
        fxParams.put(ParamSpecs.Fx.FARTALK, ParamSpecs.Boolean.TRUE);
        fxParams.put(ParamSpecs.Fx.SPEAKER_ADAPTATION, ParamSpecs.SpeakerAdaptMode.DISABLE);

        cancelRecognition();
        _currentRecognition = new Recognition(null, speechDetectionListener, signalListener);
        _currentRecognition.startEndpointing(audioSource, fxParams, resetChannel);
    }

    @Override
    public void startTimedRecognitionMode(AudioSource<AudioChunk> audioSource,
                                          VoconContext activeContext,
                                          SpeechDetectionListener speechDetectionListener,
                                          SignalListener signalListener,
                                          ResultListener resultListener)
    {
        Checker.checkArgForNull("activeContext", activeContext);

        Map<VoconParam, Integer> fxParams = new HashMap<VoconParam, Integer>();
        fxParams.put(ParamSpecs.Fx.START_ENABLE, ParamSpecs.Boolean.TRUE);
        fxParams.put(ParamSpecs.Fx.MINSPEECH, 60);
        fxParams.put(ParamSpecs.Fx.TIMEOUT_LSILENCE, 10000);
        fxParams.put(ParamSpecs.Fx.TIMEOUT_SPEECH, 10000);
        fxParams.put(ParamSpecs.Fx.EVENT_TIMER, 100);
        fxParams.put(ParamSpecs.Fx.KNOWN_SPEAKER_CHANGES, ParamSpecs.Boolean.FALSE);
        fxParams.put(ParamSpecs.Fx.FARTALK, ParamSpecs.Boolean.FALSE);

        Map<VoconParam, Integer> ctxParams = new HashMap<VoconParam, Integer>();
        ctxParams.put(ParamSpecs.Ctx.TSILENCE, 600);
        ctxParams.put(ParamSpecs.Ctx.TSILENCE_FX, 300);

        ArrayList<VoconContext> activeContexts = new ArrayList<VoconContext>(1);
        activeContext.setParams(ctxParams);
        activeContexts.add(activeContext);

        startRecognition(audioSource, fxParams, activeContexts, speechDetectionListener, signalListener, resultListener);
    }

    @Override
    public void startFuzzyMatchRecognition(AudioSource<AudioChunk> audioSource,
                                          VoconContext activeContext,
                                          SpeechDetectionListener speechDetectionListener,
                                          SignalListener signalListener,
                                          ResultListener resultListener)
    {
        Checker.checkArgForNull("activeContext", activeContext);

        Map<VoconParam, Integer> fxParams = new HashMap<VoconParam, Integer>();
        fxParams.put(ParamSpecs.Fx.START_ENABLE, ParamSpecs.Boolean.TRUE);
        fxParams.put(ParamSpecs.Fx.MINSPEECH, 60);
        fxParams.put(ParamSpecs.Fx.TIMEOUT_LSILENCE, 10000);
        fxParams.put(ParamSpecs.Fx.TIMEOUT_SPEECH, 10000);
        fxParams.put(ParamSpecs.Fx.EVENT_TIMER, 100);
        fxParams.put(ParamSpecs.Fx.KNOWN_SPEAKER_CHANGES, ParamSpecs.Boolean.FALSE);
        fxParams.put(ParamSpecs.Fx.FARTALK, ParamSpecs.Boolean.FALSE);

        Map<VoconParam, Integer> ctxParams = new HashMap<VoconParam, Integer>();
        ctxParams.put(ParamSpecs.Ctx.TSILENCE, 0);
        ctxParams.put(ParamSpecs.Ctx.TSILENCE_FX, 600);
        ctxParams.put(ParamSpecs.Ctx.ACCURACY, 5000);
        ctxParams.put(ParamSpecs.Ctx.MAX_COLLAPSED_LISTSIZE, 10);
        ctxParams.put(ParamSpecs.Ctx.MAX_NBR_UNCOLLAPSED_ENTRIES, 10);

        ArrayList<VoconContext> activeContexts = new ArrayList<VoconContext>(1);
        activeContext.setParams(ctxParams);
        activeContexts.add(activeContext);

        startRecognition(audioSource, fxParams, activeContexts, speechDetectionListener, signalListener, resultListener);
    }

    @Override
    public void startRecognition(AudioSource<AudioChunk> audioSource,
                                 List<VoconContext> activeContexts,
                                 SpeechDetectionListener speechDetectionListener,
                                 SignalListener signalListener,
                                 ResultListener resultListener)
    {
        Checker.checkArgForNull("activeContexts", activeContexts);

        if (_ctxParams != null && _ctxParams.size() > 0) {
            for (VoconContext context : activeContexts) {
                context.setParams(_ctxParams);
            }
        }

        startRecognition(audioSource, _fxParams, activeContexts, speechDetectionListener, signalListener, resultListener);
    }

    @Override
    public void startRecognition(AudioSource<AudioChunk> audioSource,
                                 Map<VoconParam, Integer> fxParams,
                                 List<VoconContext> activeContexts,
                                 SpeechDetectionListener speechDetectionListener,
                                 SignalListener signalListener,
                                 ResultListener resultListener)
    {
        startRecognition(audioSource, activeContexts, new RecognitionParam.Builder(fxParams).build(), speechDetectionListener, signalListener, resultListener);
    }

    @Override
    public void startRecognition(AudioSource<AudioChunk> audioSource,
                                 List<VoconContext> activeContexts,
                                 RecognitionParam recogParam,
                                 SpeechDetectionListener speechDetectionListener,
                                 SignalListener signalListener,
                                 ResultListener resultListener)
    {
        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForNull("activeContexts", activeContexts);
        Checker.checkArgForNull("resultListener", resultListener);

        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        Checker.checkState(this, _currentRebuild == null, "A rebuild is in progress");

        cancelRecognition();
        _currentRecognition = new Recognition(resultListener, speechDetectionListener, signalListener);
        _currentRecognition.startRecognizing(audioSource, activeContexts, recogParam);
    }

    @Override
    public void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                List<String> wakeupPhrases,
                                int penalty,
                                ResultListener resultListener)
    {
        startWakeupMode(audioSource, wakeupPhrases, penalty, 0, resultListener);
    }

    @Override
    public void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                Map<VoconParam, Integer> fxParams,
                                List<String> wakeupPhrases,
                                int penalty,
                                ResultListener resultListener)
    {
        startWakeupMode(audioSource, fxParams, wakeupPhrases, penalty, 0, resultListener);
    }

    @Override
    public void startSeamlessWakeupMode(AudioSource<AudioChunk> audioSource,
                                        List<String> wakeupPhrases,
                                        int penalty,
                                        ResultListener resultListener)
    {
        startWakeupMode(audioSource, wakeupPhrases, penalty, 1000, resultListener);
    }

    @Override
    public void startSeamlessWakeupMode(AudioSource<AudioChunk> audioSource,
                                        Map<VoconParam, Integer> fxParams,
                                        List<String> wakeupPhrases,
                                        int penalty,
                                        ResultListener resultListener)
    {
        startWakeupMode(audioSource, fxParams, wakeupPhrases, penalty, 1000, resultListener);
    }

    @Override
    public void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                List<String> wakeupPhrases,
                                int penalty,
                                int pauseDurationMs,
                                ResultListener resultListener)
    {
        Map<VoconParam, Integer> fxParams = new HashMap<VoconParam, Integer>();
        fxParams.put(ParamSpecs.Fx.START_ENABLE, ParamSpecs.Boolean.TRUE);
        fxParams.put(ParamSpecs.Fx.MINSPEECH, 150);
        fxParams.put(ParamSpecs.Fx.TIMEOUT_LSILENCE, 0);
        fxParams.put(ParamSpecs.Fx.TIMEOUT_SPEECH, 3000);
        fxParams.put(ParamSpecs.Fx.EVENT_TIMER, 100);
        fxParams.put(ParamSpecs.Fx.KNOWN_SPEAKER_CHANGES, ParamSpecs.Boolean.FALSE);
        fxParams.put(ParamSpecs.Fx.FARTALK, ParamSpecs.Boolean.TRUE);
        fxParams.put(ParamSpecs.Fx.SPEAKER_ADAPTATION, ParamSpecs.SpeakerAdaptMode.DISABLE);

        startWakeupMode(audioSource, fxParams, wakeupPhrases, penalty, pauseDurationMs, resultListener);
    }

    @Override
    public void startWakeupMode(AudioSource<AudioChunk> audioSource,
                                Map<VoconParam, Integer> fxParams,
                                List<String> wakeupPhrases,
                                int penalty,
                                int pauseDurationMs,
                                ResultListener resultListener)
    {
        Checker.checkArgForNull("audioSource", audioSource);
        Checker.checkArgForNull("fxParams", fxParams);
        Checker.checkArgForCondition("fxParams", "not empty", fxParams.size() > 0);
        Checker.checkArgForNull("wakeupPhrases", wakeupPhrases);
        Checker.checkArgForCondition("wakeupPhrases", "not empty", wakeupPhrases.size() > 0);
        Checker.checkArgForCondition("pauseDurationMs", "greater than 0", pauseDurationMs >= 0);
        Checker.checkArgForNull("resultListener", resultListener);

        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);
        Checker.checkState(this, _currentRebuild == null, "A rebuild is in progress");

        // These params must be present when using wake-up mode, they will be forced
        fxParams.put(ParamSpecs.Fx.AUTOMATIC_RESTART, ParamSpecs.Boolean.TRUE);
        fxParams.put(ParamSpecs.Fx.THREAD_AUTOMATIC_RESTART, ParamSpecs.Boolean.TRUE);

        cancelRecognition();
        _currentRecognition = new Recognition(resultListener, null, null);
        _currentRecognition.startWakeupMode(audioSource, fxParams, wakeupPhrases, penalty, pauseDurationMs);
    }

    @Override
    public void stopListening()
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        if (_currentRecognition != null)
        {
            _currentRecognition.stop(false);
        }
    }

    @Override
    public void cancelRecognition()
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        terminateCurrentRecognition();
    }

    @Override
    public void loadGrammar(Grammar grammar, RebuildListener listener)
    {
        loadGrammar(grammar, new GrammarLoadParam.Builder().build(), listener);
    }

    @Override
    public void loadGrammar(Grammar grammar, boolean onlineCompile, RebuildListener listener)
    {
        loadGrammar(grammar, new GrammarLoadParam.Builder().setOnlineCompile(onlineCompile).build(), listener);
    }

    @Override
    public void loadGrammar(Grammar grammar, GrammarLoadParam loadingParam, RebuildListener listener)
    {
        Checker.checkArgForNull("grammar", grammar);
        Checker.checkArgForNull("loadingParam", loadingParam);
        Checker.checkArgForNull("listener", listener);
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelRecognition();

        _currentRebuild = new Rebuild(grammar, _grammar, loadingParam, listener);
        _currentRebuild.start(false);
    }

    @Override
    public void clearGrammar(final RebuildListener listener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelRecognition();

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                boolean tempSuccess = false;

                // Destroy grammar objects
                if (_grammar != null && _isVoconCreated)
                {
                    for (VoconContext context: _grammar.getContexts())
                    {
                        for (String wordClass : context.getSlotIds())
                        {
                            if (_nativeVocon.hasWordClass(wordClass))
                            {
                                _nativeVocon.destroyWordClass(wordClass);
                            }

                            _nativeVocon.destroyDynamicWordContext(wordClass);
                        }
                    }
                    tempSuccess = true;
                }

                final boolean success = tempSuccess;
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (listener != null) {
                            if (success) {
                                listener.onComplete(_grammar, null);
                            } else {
                                listener.onError(new VoconError(VoconError.Reason.CLEAR_GRAMMAR_ERROR));
                            }
                        }
                        _grammar = new Grammar();
                    }
                });
            }
        });
    }

    @Override
    public void cancelRebuild()
    {
        if (_currentRebuild != null)
        {
            _currentRebuild.cancel();
            _currentRebuild = null;
        }
    }

    @Override
    public void saveState(final SaveListener listener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelRecognition();

        final Grammar grammar = _grammar;

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (_grammar == null) {
                    Logger.error(this, "Error: grammar is not loaded.");
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            if (listener != null)
                                listener.onSaved(VoconRecognizerBase.this, false);
                        }
                    });
                }
                else
                {
                    final boolean success = _nativeVocon.saveVocon(_stateFileName, grammar);
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            if (listener != null)
                                listener.onSaved(VoconRecognizerBase.this, success);
                        }
                    });
                }
            }
        });

    }

    @Override
    public void saveState() {
        saveState(null);
    }

    @Override
    public void loadState(final RebuildListener rebuildListener) {
        loadStateAndGrammar(null, rebuildListener);
    }

    @Override
    public void loadStateAndGrammar(final Grammar grammar,
                                    final RebuildListener rebuildListener)
    {
        loadStateAndGrammar(grammar, new GrammarLoadParam.Builder().build(), rebuildListener);
    }

    @Override
    public void loadStateAndGrammar(final Grammar grammar,
                                    final boolean onlineCompile,
                                    final RebuildListener rebuildListener)
    {
        loadStateAndGrammar(grammar, new GrammarLoadParam.Builder().setOnlineCompile(onlineCompile).build(), rebuildListener);
    }

    @Override
    public void loadStateAndGrammar(final Grammar grammar,
                                    final GrammarLoadParam loadingParam,
                                    final RebuildListener rebuildListener)
    {
        Checker.checkArgForNull("loadingParam", loadingParam);
        Checker.checkArgForNull("rebuildListener", rebuildListener);
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelRecognition();

        _grammar.clearContexts();
        _grammar.clearDynamicSlots();

        // Create, but don't start
        final Rebuild rebuild;
        if (grammar != null)
            rebuild = new Rebuild(grammar, _grammar, loadingParam, rebuildListener);
        else
            rebuild = null;
        _currentRebuild = rebuild; // Set this now, so we can tell if it's canceled

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (!_isVoconCreated) {
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            if (rebuild == _currentRebuild)
                            {
                                Logger.error(VoconRecognizerBase.this, "Cannot load VoCon State or Grammar.  VoCon is not initialized. ");
                                rebuildListener.onError(new VoconError(VoconError.Reason.NOT_INITIALIZED));
                            }
                        }
                    });
                    return;
                }

                final String stateFileName = _stateFileName;
                final Grammar loadedGrammar = _nativeVocon.loadVocon(_stateFileName);

                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (rebuild == _currentRebuild)
                        {
                            if (loadedGrammar == null && grammar == null)
                            {
                                Logger.error(VoconRecognizerBase.this, "VoCon loading VoCon state from: " + stateFileName);
                                rebuildListener.onError(new VoconError(VoconError.Reason.FILE_ERROR));
                            } else if (grammar != null)
                            {
                                // refresh rebuild object
                                _currentRebuild = new Rebuild(grammar, loadedGrammar, loadingParam, rebuildListener);

                                // If no loaded grammar was loaded, make sure we do a full rebuild
                                if (loadedGrammar == null) {
                                    Logger.warn(VoconRecognizerBase.this, "Unable to load VoCon state from: " + stateFileName + ", doing a full grammar rebuild");
                                    _currentRebuild.start(true);
                                } else
                                {
                                    Logger.debug(VoconRecognizerBase.this, "VoCon state loaded, now building grammar");
                                    _currentRebuild.start(false);
                                }
                            }
                            else
                            {
                                Logger.debug(VoconRecognizerBase.this, "VoCon grammar loaded, now building grammar");

                                _grammar = loadedGrammar;
                                rebuildListener.onComplete(loadedGrammar, null);
                            }
                        }
                    }
                });
            }
        });
    }

    @Override
    public void loadResource(final String fileName, final ResourceLoadListener listener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkStringArgForNullOrEmpty("fileName", fileName);
        Checker.checkState(this, _currentRebuild == null, "A rebuild is in progress");

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final boolean success = _nativeVocon.loadResource(fileName);

                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (listener != null)
                            listener.onLoaded(VoconRecognizerBase.this, success);
                    }
                });
            }
        });
    }

    @Override
    public void releaseResource(final String fileName, final ResourceReleaseListener listener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkStringArgForNullOrEmpty("fileName", fileName);
        Checker.checkState(this, _currentRebuild == null, "A rebuild is in progress");

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                _nativeVocon.releaseResource(fileName);

                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (listener != null)
                            listener.onReleased(VoconRecognizerBase.this);
                    }
                });
            }
        });
    }

    @Override
    public void preloadContext(final VoconContext context, final PreloadContextListener listener)
    {
        Checker.checkArgForNull("context", context);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                boolean success;
                if (context.getPreloaded())
                {
                    success = true;
                }
                else
                {
                    // Load context from file
                    if (ContextError.FATAL == _nativeVocon.createAsrContext(context))
                    {
                        Logger.error(this, "Unable to create context: " + context.getId());
                        _nativeVocon.destroyContext(context);
                        context.setPreloaded(false);
                        success = false;
                    }
                    else
                    {
                        updateComputedParam(context, Ctx.ACCURACY.id);
                        context.setPreloaded(true);
                        success = true;
                    }
                }

                final boolean successFinal = success;
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        listener.onPreloaded(context, successFinal);
                    }
                });
            }
        });
    }

    @Override
    public void unloadContext(final VoconContext context)
    {
        Checker.checkArgForNull("context", context);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                if (context.getPreloaded())
                {
                    _nativeVocon.destroyContext(context);
                    context.setPreloaded(false);
                }
            }
        });
    }

    @Override
    public void compileContext(List<FuzzyMatchData> entries, String fileName, CompileListener listener)
    {
        Checker.checkArgForNull("entries", entries);
        Checker.checkArgForCondition("entries", "not empty", entries.size() > 0);
        Checker.checkStringArgForEmpty("fileName", fileName);
        Checker.checkArgForNull("listener", listener);
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelCompilation();
        cancelRecognition();

        _currentCompilation = new Compilation(listener);
        _currentCompilation.start(entries, fileName);
    }

    @Override
    public void compileFuzzyMatchContext(List<FuzzyMatchData> entries, String fileName, CompileListener listener)
    {
        Checker.checkArgForNull("entries", entries);
        Checker.checkArgForCondition("entries", "not empty", entries.size() > 0);
        Checker.checkStringArgForEmpty("fileName", fileName);
        Checker.checkArgForNull("listener", listener);
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelCompilation();
        cancelRecognition();

        _currentCompilation = new Compilation(listener);
        _currentCompilation.start(entries, fileName);
    }

    @Override
    public void clean(final String fileName, final CleanListener listener)
    {
        Checker.checkStringArgForNullOrEmpty("fileName", fileName);

        clean(new ArrayList<String>()
        {
            private static final long serialVersionUID = 1L;
            {
                 add(fileName);
            }
        }, listener);
    }

    @Override
    public void clean(final List<String> fileNames, final CleanListener listener)
    {
        Checker.checkArgForNull("fileNames", fileNames);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                for (String fileName : fileNames)
                {
                    _nativeVocon.deleteCompiledContext(fileName);
                    _mainThreadHandler.post(new Runnable()
                    {
                        @Override
                        public void run()
                        {
                            if (listener != null)
                                listener.onCleaned();
                        }
                    });
                }
            }
        });
    }

    @Override
    public void cancelCompilation()
    {
        if (_currentCompilation != null)
        {
            _currentCompilation.cancel();
            _currentCompilation = null;
        }
    }

    @Override
    public void cancelFuzzyMatchCompilation()
    {
        if (_currentCompilation != null)
        {
            _currentCompilation.cancel();
            _currentCompilation = null;
        }
    }

    @Override
    public void compileBnfContext(final String fileName, final CompileListener listener)
    {
        Checker.checkArgForNull("fileName", fileName);

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final boolean result = _nativeVocon.compileFileToContext(fileName);
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        if (listener != null)
                        {
                            if (result)
                            {
                                listener.onComplete(fileName);
                            }
                            else
                            {
                                // Delete the generated context for convenience
                                clean(fileName, new CleanListener()
                                {
                                    @Override
                                    public void onCleaned()
                                    {
                                        listener.onError(new VoconError(VoconError.Reason.CONTEXT_COMPILATION_ERROR));
                                    }
                                });
                            }
                        }
                    }
                });
            }
        });
    }

    @Override
    public void onApplicationLanguageChange(Object unused, final Listener listener)
    {
        // clearGrammar() will also cancel the rebuild and the recognition
        clearGrammar(new RebuildListener()
        {
            @Override
            public void onComplete(Grammar grammar,
                    List<SkippedWord> skippedWords)
            {
                if (listener != null)
                    listener.onComplete(true);
            }

            @Override
            public void onError(VoconError error)
            {
                if (listener != null)
                    listener.onComplete(false);
            }
        });
    }

    @Override
    public void stopRecognitionResults()
    {
        if (_currentRecognition != null)
        {
            _currentRecognition._isNoMoreResults = true;
            _nativeVocon.stopRecognitionResults();
        }
    }

    private void updateComputedParam(VoconContext context, long paramId) {
        Map<String, VoconContext> subContexts = context.getSubContexts();
        if (subContexts.size() == 0)
        {
            if (context.getSlotIds().size() != 0)
            {
                _nativeVocon.updateComputedParam(context, paramId);
            }
        }
        else
        {
            for (String key: subContexts.keySet())
            {
                updateComputedParam(subContexts.get(key), paramId);
            }
        }
    }

    private int getSize(String language)
    {
        int bestSizeKb = 0;
        List<VoconModelInfo> models = getAvailableModels();
        for (VoconModelInfo model : models)
        {
            if (language.equals(model.language.name) && model.sizeKb > bestSizeKb)
                bestSizeKb = model.sizeKb;
        }
        return bestSizeKb;
    }

    private String getDomain(String language)
    {
        final String[] PRIORITIES = { VoconDomain.TRAINING, VoconDomain.NORMAL, VoconDomain.NAME, VoconDomain.MUSIC, VoconDomain.SUPER_CLC, VoconDomain.SUPER_CLC_ML };
        int domainIdx = 0;

        ArrayList<NativeClcInfo> nativeClcs = new ArrayList<NativeClcInfo>();

        boolean success = _nativeVocon.listAvailableClcs(nativeClcs);
        if (success)
        {
            for (NativeClcInfo clc : nativeClcs)
            {
                if (language.equals(clc.language))
                {
                    int availableDomainIdx = 0;
                    for (int idx = 0; idx < PRIORITIES.length; idx++)
                    {
                        if (PRIORITIES[idx].equals(clc.domain))
                        {
                            availableDomainIdx = idx;
                        }
                    }

                    if (availableDomainIdx > domainIdx)
                    {
                        domainIdx = availableDomainIdx;
                    }
                }
            }
        }

        return PRIORITIES[domainIdx];
    }

    private void terminateCurrentRecognition()
    {
        if (_currentRecognition != null)
        {
            Recognition r = _currentRecognition;
            _currentRecognition.cancel();
            if (r == _currentRecognition)
                _currentRecognition = null;
        }
    }

    @Override
    public void saveAcousticModelAdaptation(final String fileName, final FileSaveListener saveListener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelRecognition();

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final boolean result = _nativeVocon.saveAcModAdaptData(fileName);
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        OperationResult enumResult = result ? OperationResult.SUCCESS : OperationResult.FAILURE;
                        if (saveListener != null)
                            saveListener.onFileSaved(VoconRecognizerBase.this, new FileOperationResult(fileName, enumResult));
                    }
                });
            }
        });
    }

    @Override
    public void loadAcousticModelAdaptation(final String fileName, final FileLoadListener loadListener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelRecognition();

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final boolean result = _nativeVocon.loadAcModAdaptData(fileName);
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                       OperationResult enumResult = result ? OperationResult.SUCCESS : OperationResult.FAILURE;
                        if (loadListener != null)
                            loadListener.onFileLoaded(VoconRecognizerBase.this, new FileOperationResult(fileName, enumResult));
                    }
                });
            }
        });
    }

    @Override
    public void saveFxAdaptation(final String fileName, final FileSaveListener saveListener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelRecognition();

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final boolean result = _nativeVocon.saveFxAdaptData(fileName);
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        OperationResult enumResult = result ? OperationResult.SUCCESS : OperationResult.FAILURE;
                        if (saveListener != null)
                            saveListener.onFileSaved(VoconRecognizerBase.this, new FileOperationResult(fileName, enumResult));
                    }
                });
            }
        });

    }

    @Override
    public void loadFxAdaptation(final String fileName, final FileLoadListener loadListener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelRecognition();

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final boolean result = _nativeVocon.loadFxAdaptData(fileName);
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        OperationResult enumResult = result ? OperationResult.SUCCESS : OperationResult.FAILURE;
                        if (loadListener != null)
                            loadListener.onFileLoaded(VoconRecognizerBase.this, new FileOperationResult(fileName, enumResult));
                    }
                });
            }
        });
    }

    @Override
    public void saveSpeakerProfile(final String fileName, final FileSaveListener saveListener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelRecognition();

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final boolean result = _nativeVocon.saveSpeakerProfile(fileName);
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        OperationResult enumResult = result ? OperationResult.SUCCESS : OperationResult.FAILURE;
                        if (saveListener != null)
                            saveListener.onFileSaved(VoconRecognizerBase.this, new FileOperationResult(fileName, enumResult));
                    }
                });
            }
        });
    }

    @Override
    public void loadSpeakerProfile(final String fileName, final FileLoadListener loadListener)
    {
        Checker.checkState(this, !_isReleased);
        Checker.checkState(this, _isInitialized);

        cancelRebuild();
        cancelRecognition();

        _workerThreadHandler.post(new Runnable()
        {
            @Override
            public void run()
            {
                final boolean result = _nativeVocon.loadSpeakerProfile(fileName);
                _mainThreadHandler.post(new Runnable()
                {
                    @Override
                    public void run()
                    {
                        OperationResult enumResult = result ? OperationResult.SUCCESS : OperationResult.FAILURE;
                        if (loadListener != null)
                            loadListener.onFileLoaded(VoconRecognizerBase.this, new FileOperationResult(fileName, enumResult));
                    }
                });
            }
        });
    }
    /**
     * Called on MAIN thread exactly once in the constructor to obtain the
     * worker thread handler, if the worker thread handler wasn't given in the
     * constructor.
     * @return The worker thread handler.
     */
    protected abstract NMTHandler getWorkerThreadHandler();
}
