package com.nuance.dragon.toolkit.vocon;

import com.nuance.dragon.toolkit.oem.api.Factory;
import com.nuance.dragon.toolkit.oem.api.FileManager;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.NMTHandler;
import com.nuance.dragon.toolkit.oem.api.WorkerThread;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;

final class VoconRecognizerImpl extends VoconRecognizerBase {

    private final boolean _hasWorkerThread;

    // Only allow 1 worker thread and file factory for all VoCon instances.
    private static WorkerThread _workerThread;
    private static int _workerThreadCount;

    private boolean _released;

    /**
     * Create a VoCon Recognizer. The VoCon will use a dedicated worker thread. <p>
     * NOTE: Due to VoCon engine limitations, multiple VoCon instances must use
     * the same worker thread.
     * @param fileManager The file manager where VoCon should look for files.
     */
    public VoconRecognizerImpl(FileManager fileManager) {
        this(fileManager, null);
    }

    /**
     * Create a VoCon Recognizer. The VoCon will use the given worker thread. <p>
     * NOTE: Due to VoCon engine limitations, multiple VoCon instances must use
     * the same worker thread. <p>
     * NOTE #2: Due to the audio pull model of VoCon engine, audio capture
     * must run on a different thread than the VoCon worker thread.
     * @param fileManager The file manager where VoCon should look for files.
     * @param workerThread
     */
    public VoconRecognizerImpl(FileManager fileManager, NMTHandler workerThread) {
        super(new NativeVoconImpl(fileManager), workerThread);
        Checker.checkArgForNull("fileManager", fileManager);
        _hasWorkerThread = workerThread == null;
    }

    @Override
    protected NMTHandler getWorkerThreadHandler() {
        if (_workerThread == null) {
            _workerThread = Factory.createWorkerThread("com.nuance.dragon.toolkit.vocon.VoconRecognizerImpl");
            _workerThread.start();
        }
        _workerThreadCount++;
        return _workerThread.getHandler();
    }

    @Override
    public void release()
    {
        if (!_released)
        {
            _released = true;
            super.release();

            if (_hasWorkerThread)
            {
                _workerThreadCount--;
                if (_workerThreadCount == 0)
                {
                    _workerThread.stop();
                    _workerThread = null;
                    Logger.debug(this, "VoCon worker thread stops and nulls out");
                }
            }
        }
    }
}
