package com.nuance.dragon.toolkit.vocon;

import java.util.List;

import org.json.JSONArray;
import org.json.JSONException;

import com.nuance.dragon.toolkit.oem.api.JSONCompliant;
import com.nuance.dragon.toolkit.oem.api.Logger;
import com.nuance.dragon.toolkit.oem.api.internal.Checker;
import com.nuance.dragon.toolkit.oem.api.json.JSONObject;
import com.nuance.dragon.toolkit.recognition.LocalRecognitionResult;

/**
 * A raw Vocon result.
 */
public final class VoconResult extends LocalRecognitionResult implements JSONCompliant
{
    /**
     * The wake-up type.
     */
    public class WakeupType
    {
        /** Unknown */
        public static final int UNKNOWN = 0;

        /** Wake-up word only */
        public static final int REGULAR = 1;

        /** Wake-up word + <open-speech> */
        public static final int SEAMLESS = 2;
    }

    /**
     * The result type.
     */
    public static enum ResultType
    {
        NBEST, TABLE, UNKNOWN
    }

    public static final int REF_WAKEUP_TIMESTAMP_UNAVAILABLE = -1;
    private static final String KEY_RAW_RESULT = "rawres";

    private ResultType _resultType;
    private boolean _isSpeech;
    private boolean _isOutOfGrammar;
    private JSONArray _nbest;
    private JSONArray _interpretations;
    private int _confidence;
    private int _userID;
    private long _refWakeupTimestamp;
    private String _wakeupPhrase;
    private int _wakeupType;

    VoconResult(String json, long refWakeupTimestamp)
    {
        super(json);

        _resultType = ResultType.UNKNOWN;
        _isSpeech = false;
        _isOutOfGrammar = false;
        _nbest = null;
        _interpretations = null;
        _confidence = -1;
        _userID = 0;
        _refWakeupTimestamp = refWakeupTimestamp;
        _wakeupPhrase = null;
        _wakeupType = WakeupType.UNKNOWN;

        try {
            String resultType = _rawResult.getString("_resultType");
            if (resultType.equals("NBest"))
            {
                _resultType      = ResultType.NBEST;
                _isSpeech        = _rawResult.getString("_isSpeech").equals("yes");
                _isOutOfGrammar  = _rawResult.getString("_isInGrammar").equals("no");
                _nbest           = _rawResult.getJSONArray("_hypotheses");
                if (_nbest.length() > 0)
                {
                    _confidence = _nbest.getJSONObject(0).getInt("_conf");
                    _userID = getUserID(_nbest);
                }
            }
            else if (resultType.equals("Table"))
            {
                _resultType = ResultType.TABLE;
                _interpretations = _rawResult.getJSONArray("_interpretations");
            }
            else
            {
                // Should never happen unless Vocon is not setup correctly
                Logger.error(this, "VoCon result type not supported.");
            }
        } catch (JSONException e)
        {
            Logger.error(this, "Unable to parse VoCon results.", e);
        }
    }

    /**
     * Get the result type. This is necessary to parse the result object correctly.
     * @return The result type. See {@link ResultType} for possible values.
     */
    public ResultType getResultType()
    {
        return _resultType;
    }

    /**
     * Indicate if speech was detected or not in the input utterance.
     * Only applicable for {@link ResultType#NBEST} result type.
     * @return True if speech was detected. Else false.
     */
    public boolean isSpeech()
    {
        Checker.checkState("ResultType", _resultType == ResultType.NBEST);
        return _isSpeech;
    }

    /**
     * Indicate if the utterance was in/out of grammar.
     * Only applicable for {@link ResultType#NBEST} result type.
     * @return True if out-of-grammar was detected. Else false.
     */
    public boolean isOutOfGrammar()
    {
        Checker.checkState("ResultType", _resultType == ResultType.NBEST);
        return _isOutOfGrammar;
    }

    /**
     * Indicate if the utterance contains generic speech.
     * Only applicable for {@link ResultType#NBEST} result type.
     * @param choiceIdx The choice index in the NBest list.
     * @return True if generic speech was detected. Else false.
     */
    public boolean hasGenericSpeech(int choiceIdx)
    {
        Checker.checkState("ResultType", _resultType == ResultType.NBEST);
        boolean isOpenSpeech = false;
        if (choiceIdx < _nbest.length())
        {
            try {
                JSONArray items = _nbest.getJSONObject(choiceIdx).getJSONArray("_items");
                for (int idx1 = 0; idx1 < items.length(); idx1++)
                {
                    org.json.JSONObject currentItem = items.getJSONObject(idx1);
                    if(currentItem.getString("_type").equals("terminal"))
                    {
                        if (currentItem.getString("_orthography").equals("<...>") &&
                            currentItem.getInt("_conf") > 0)
                        {
                            isOpenSpeech = true;
                            break;
                        }
                    }
                    else
                    {
                        JSONArray subItems = currentItem.getJSONArray("_items");
                        for (int idx2 = 0; idx2 < subItems.length(); idx2++)
                        {
                            org.json.JSONObject currentSubItem = subItems.getJSONObject(idx2);
                            if(currentSubItem.getString("_type").equals("terminal"))
                            {
                                if (currentSubItem.getString("_orthography").equals("<...>") &&
                                    currentSubItem.getInt("_conf") > 0)
                                {
                                    isOpenSpeech = true;
                                    break;
                                }
                            }
                        }
                    }
                }
            } catch (JSONException e)
            {
                Logger.error(this, "Unable to detect generic speech data", e);
            }
        }
        return isOpenSpeech;
    }

    /**
     * Returns the calculated confidence that the top choice is correct.
     * 10000 indicates highest confidence, 0 indicates lowest confidence. For low
     * confidence results, it may be a good idea to consider other options in
     * the N-best list.
     * Only applicable for {@link ResultType#NBEST} result type.
     * @return The confidence.
     */
    public int getConfidence()
    {
        Checker.checkState("ResultType", _resultType == ResultType.NBEST);
        return _confidence;
    }

    /**
     * Returns the user ID of the first tag encountered in the top choice.
     * user ID == 0 or user ID == 1 will be ignored since these are reserved
     * values for invalid or unused user IDs.
     * This is is a convenience method to help support partial name matching.
     * Only applicable for {@link ResultType#NBEST} result type.
     * @return The user ID.
     */
    public int getUserID()
    {
        Checker.checkState("ResultType", _resultType == ResultType.NBEST);
        return _userID;
    }

    void setRecognizedWakeupPhrase(List<String> wakeupPhrases)
    {
         if (_nbest != null && _nbest.length() > 0)
         {
             try
             {
                 final String startRule = _nbest.getJSONObject(0).optString("_startRule");
                 if (startRule != null && startRule.length() > 0)
                 {
                     int idx = startRule.indexOf("#");
                     if (idx > 0)
                     {
                         Integer hash = Integer.valueOf(startRule.substring(0, idx));
                         for (String p : wakeupPhrases)
                         {
                             if (p.hashCode() == hash)
                             {
                                 // Phrase has been found
                                 _wakeupPhrase = p;
                                 return;
                             }
                         }
                     }
                 }
             }
             catch (Exception e)
             {
                 Logger.error(this, "Error setting recognized wake-up phrase", e);
             }
         }
    }

    void setRecognizedWakeupType()
    {
         if (_nbest != null && _nbest.length() > 0)
         {
             try
             {
                 final String startRule = _nbest.getJSONObject(0).optString("_startRule");
                 if (startRule != null && startRule.length() > 0)
                 {
                     _wakeupType = startRule.contains("seamless") ? WakeupType.SEAMLESS : WakeupType.REGULAR;
                 }
             }
             catch (Exception e)
             {
                 Logger.error(this, "Error setting recognized wake-up phrase", e);
             }
         }
    }

    /**
     * Returns the wake-up phrase that has been recognized during the
     * wake-up session.
     * @return The recognized wake-up phrase, or null if not in wake-up mode.
     */
    public String getRecognizedWakeupPhrase()
    {
        Checker.checkState("ResultType", _resultType == ResultType.NBEST);
        return _wakeupPhrase;
    }

    /**
     * Indicates the wake-up type (i.e. wake-up word +/- <open-speech>).
     * @return The wake-up type, see {@link WakeupType}.
     */
    public int getWakeupType()
    {
        Checker.checkState("ResultType", _resultType == ResultType.NBEST);
        return _wakeupType;
    }

    /**
     * Returns the time stamp of start of the wake-up utterance.
     * @return The start of utterance time stamp, in ms. -1 if not available.
     */
    public long getWakeupStartTimestamp()
    {
        Checker.checkState("ResultType", _resultType == ResultType.NBEST);
        long timestamp = -1;

        if (_refWakeupTimestamp != REF_WAKEUP_TIMESTAMP_UNAVAILABLE)
        {
            try {
                timestamp = _refWakeupTimestamp + _nbest.getJSONObject(0).optInt("_beginTimeMs");
            } catch (JSONException e) {
                Logger.error(this, "Unable to parse wake-up start timestamp.", e);
            }
        }
        return timestamp;
    }

    /**
     * Returns the time stamp of the end of wake-up utterance.
     * @return The start of utterance time stamp, in ms. -1 if not available.
     */
    public long getWakeupEndTimestamp()
    {
        Checker.checkState("ResultType", _resultType == ResultType.NBEST);
        long timestamp = -1;

        if (_refWakeupTimestamp != REF_WAKEUP_TIMESTAMP_UNAVAILABLE)
        {
            try {
                JSONArray items = _nbest.getJSONObject(0).getJSONArray("_items");
                for (int idx = 0; idx < items.length(); idx++)
                {
                    org.json.JSONObject currentItem = items.getJSONObject(idx);
                    if (currentItem.getString("_orthography").equals("wuw_X"))
                    {
                        // This is the new end of the wake-up phrase
                        timestamp = _refWakeupTimestamp + currentItem.getInt("_endTimeMs");
                    }
                }
            } catch (JSONException e) {
                Logger.error(this, "Unable to parse wake-up end timestamp.", e);
            }
        }
        return timestamp;
    }

    /**
     * Get the list of choices as an N-best list.
     * Only applicable for {@link ResultType#NBEST} result type.
     * @return The N-best list.
     */
    public JSONArray getChoiceList()
    {
        Checker.checkState("ResultType", _resultType == ResultType.NBEST);
        return _nbest;
    }

    /**
     * Only applicable for {@link ResultType#NBEST} result type.
     * @return The number of choices in the N-best list. This is the
     * same as the size of the N-best list returned by {@link #getChoiceList}.
     */
    public int getChoiceCount()
    {
        Checker.checkState("ResultType", _resultType == ResultType.NBEST);
        return _nbest.length();
    }

    /**
     * Get the list of interpretation as fuzzy match result.
     * Only applicable for {@link ResultType#TABLE} result type.
     * @return The fuzzy match result list.
     */
    public JSONArray getInterpretationList()
    {
        Checker.checkState("ResultType", _resultType == ResultType.TABLE);
        return _interpretations;
    }

    /**
     * Only applicable for {@link ResultType#TABLE} result type.
     * @return The number of interpretations in fuzzy match result list. This is the
     * same as the size of fuzzy match result list returned by {@link #getInterpretationList}.
     */
    public int getInterpretationCount()
    {
        Checker.checkState("ResultType", _resultType == ResultType.TABLE);
        return _interpretations.length();
    }

    @Override
    public org.json.JSONObject toJSON()
    {
        JSONObject o = new JSONObject();
        o.tryPut(KEY_RAW_RESULT, _rawResult);
        return o;
    }

    /**
     * @return The transcription of the top entry.
     */
    public String toString()
    {
        // TODO
        if (_resultType != ResultType.NBEST)
            return super.toString();
            
        String transcription = "";

        try {
            org.json.JSONObject topEntry = _nbest.getJSONObject(0);
            JSONArray items = topEntry.getJSONArray("_items");
            for (int idx1 = 0; idx1 < items.length(); idx1++)
            {
                org.json.JSONObject currentItem = items.getJSONObject(idx1);
                if(currentItem.getString("_type").equals("terminal"))
                {
                    transcription += currentItem.getString("_orthography") + " ";
                }
                else
                {
                    JSONArray subItems = currentItem.getJSONArray("_items");
                    for (int idx2 = 0; idx2 < subItems.length(); idx2++)
                    {
                        transcription += subItems.getJSONObject(idx2).getString("_orthography") + " ";
                    }
                }
            }
        } catch (JSONException e) {
            Logger.error(this, "Failed to parse top-entry.", e);
        }
        return transcription;
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + _confidence;
        result = prime
                * result
                + ((_interpretations == null) ? 0 : _interpretations.hashCode());
        result = prime * result + (_isOutOfGrammar ? 1231 : 1237);
        result = prime * result + (_isSpeech ? 1231 : 1237);
        result = prime * result + ((_nbest == null) ? 0 : _nbest.hashCode());
        result = prime * result
                + (int) (_refWakeupTimestamp ^ (_refWakeupTimestamp >>> 32));
        result = prime * result
                + ((_resultType == null) ? 0 : _resultType.hashCode());
        result = prime * result + _userID;
        result = prime * result
                + ((_wakeupPhrase == null) ? 0 : _wakeupPhrase.hashCode());
        result = prime * result + _wakeupType;
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        VoconResult other = (VoconResult) obj;
        if (_confidence != other._confidence)
            return false;
        if (_interpretations == null) {
            if (other._interpretations != null)
                return false;
        } else if (!_interpretations.equals(other._interpretations))
            return false;
        if (_isOutOfGrammar != other._isOutOfGrammar)
            return false;
        if (_isSpeech != other._isSpeech)
            return false;
        if (_nbest == null) {
            if (other._nbest != null)
                return false;
        } else if (!_nbest.equals(other._nbest))
            return false;
        if (_refWakeupTimestamp != other._refWakeupTimestamp)
            return false;
        if (_resultType != other._resultType)
            return false;
        if (_userID != other._userID)
            return false;
        if (_wakeupPhrase == null) {
            if (other._wakeupPhrase != null)
                return false;
        } else if (!_wakeupPhrase.equals(other._wakeupPhrase))
            return false;
        if (_wakeupType != other._wakeupType)
            return false;
        return true;
    }

    private static int getUserID(JSONArray nbest) throws JSONException
    {
        final String KEY = "_userID.lo32";
        int userID = 0;

        JSONArray items = nbest.getJSONObject(0).getJSONArray("_items");
        for (int idx1 = 0; idx1 < items.length(); idx1++)
        {
            org.json.JSONObject currentItem = items.getJSONObject(idx1);
            if(currentItem.getString("_type").equals("tag"))
            {
                JSONArray subItems = currentItem.getJSONArray("_items");
                for (int idx2 = 0; idx2 < subItems.length(); idx2++)
                {
                    org.json.JSONObject currentSubItem = subItems.getJSONObject(idx2);
                    if (currentSubItem.has(KEY))
                    {
                        userID = currentSubItem.getInt(KEY);
                        if (userID > 0)
                            break;
                    }
                }
            }
        }
        return userID;
    }
}
