package com.nuance.nmsp.client.sdk.common.util;

//#ifdef NMSP_LOG
import com.nuance.nmsp.client.sdk.common.oem.api.LogFactory;
import com.nuance.nmsp.client.sdk.common.oem.api.LogFactory.Log;
//#endif

/**
 * Utility methods that check conditions in API calls, and throw exceptions when
 * necessary.
 */
public final class Checker
{
    //#ifdef NMSP_LOG
    private static final Log log = LogFactory.getLog(Checker.class);
    //#endif
    
    /**
     * Throw an exception if stateCheck is false.
     * @param module The module of the check
     * @param stateCheck If false, an exception will be thrown
     */
    public static void checkState(Object module, boolean stateCheck)
    {
        checkState(module, stateCheck, null);
    }
    
    /**
     * Throw an exception if stateCheck is false.
     * @param module The module of the check
     * @param stateCheck If false, an exception will be thrown
     * @param message The additional message to add to the exception.
     */
    public static void checkState(Object module, boolean stateCheck, String message)
    {
        if (!stateCheck)
        {
            String name = null;
            if (module instanceof String)
                name = (String)module;
            else if (module != null)
                name = ClassUtils.getShortClassName(module.getClass());
            
            IllegalStateException ex;
            if (name == null)
            {
                if (message == null)
                    ex = new IllegalStateException();
                else
                    ex = new IllegalStateException(message);
            } else
            {
                if (message == null)
                    ex = new IllegalStateException(name + " is in an invalid state");
                else
                    ex = new IllegalStateException(name + " is in an invalid state: " + message);
            }
            //#ifdef NMSP_LOG
            log.error("State check failed", ex);
            //#endif
            throw ex;
        }
    }
    
    /**
     * Throw an exception if the arg is null.
     * @param argName
     * @param arg
     * 
     * @return the original argument
     */
    public static Object checkArgForNull(String argName, Object arg)
    {
        if (arg == null) {
            throwArgException(argName + " must not be null");
        }
        return arg;
    }
    
    /**
     * Throw an exception if none of the objects is non-null.
     * @param argNames
     * @param args
     */
    public static void checkArgsForNull(String argNames, Object ... args)
    {
        for (Object arg : args)
        {
            if (arg != null)
                return;
        }
        throwArgException("At least one of " + argNames + " must not be null or empty");
    }

    /**
     * Throw an exception if the string arg is null or empty.
     * @param argName
     * @param arg
     * 
     * @return the original String argument
     */
    public static String checkStringArgForNullOrEmpty(String argName, String arg)
    {
        if (arg == null || arg.length() == 0) {
            throwArgException(argName + " must not be null or empty");
        }
        
        return arg;
    }
    
    /**
     * Throw an exception if the string arg is non-null and empty
     * 
     * 
     */
    public static String checkStringArgForEmpty(String argName, String arg)
    {
        if (arg != null && arg.length() == 0)
        {
            throwArgException(argName + " could be null but must not be empty if non-null");
        }
        
        return arg;
    }
    
    /**
     * Throw an exception if none of the strings is non-null and non-empty.
     * @param argNames
     * @param args
     */
    public static void checkStringArgsForAllNullOrEmpty(String argNames, String ... args)
    {
        for (String arg : args)
        {
            if (arg != null && arg.length() > 0)
                return;
        }
        throwArgException("At least one of " + argNames + " must not be null or empty");
    }
    
    /**
     * Throw an exception if the argument does not comply with the value check 
     * 
     * @param argName
     * @param mustBe
     * @param valueCheck
     */
    public static void checkArgForCondition(String argName, String mustBe, boolean valueCheck)
    {
    	if (!valueCheck)
    	{
    		throw new IllegalArgumentException(argName + " must be " + mustBe);
    	}
    }
    
    private static void throwArgException(String msg)
    {
        IllegalArgumentException ex = new IllegalArgumentException(msg);
        //#ifdef NMSP_LOG
        log.error(msg);
        //#endif
        throw ex;
    }
}
