package com.nuance.nmsp.client.sdk.components.core.internal.calllog;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Vector;

//#ifdef NMSP_LOG
import com.nuance.nmsp.client.sdk.common.oem.api.LogFactory;
import com.nuance.nmsp.client.sdk.common.oem.api.LogFactory.Log;
//#endif
import com.nuance.nmsp.client.sdk.common.oem.api.MessageSystem;
import com.nuance.nmsp.client.sdk.common.oem.api.MessageSystem.MessageData;
import com.nuance.nmsp.client.sdk.common.oem.api.MessageSystem.MessageHandler;
import com.nuance.nmsp.client.sdk.common.oem.api.MiscSystem;
import com.nuance.nmsp.client.sdk.common.protocols.ProtocolBuilder;
import com.nuance.nmsp.client.sdk.common.protocols.ProtocolDefines;
import com.nuance.nmsp.client.sdk.common.protocols.XModeMsgHeader;
import com.nuance.nmsp.client.sdk.common.util.ByteConversion;
import com.nuance.nmsp.client.sdk.components.core.XMode;
import com.nuance.nmsp.client.sdk.components.core.XMode.XModeListener;
import com.nuance.nmsp.client.sdk.components.core.calllog.CalllogSender;
import com.nuance.nmsp.client.sdk.components.core.internal.calllog.CalllogImpl.SessionEventData;
import com.nuance.nmsp.client.sdk.components.general.Parameter;
import com.nuance.nmsp.client.sdk.oem.MessageSystemOEM;


public class CalllogSenderImpl implements CalllogSender, XModeListener, MessageHandler {

    //#ifdef NMSP_LOG
    private static final Log log = LogFactory.getLog(CalllogSenderImpl.class);
    //#endif

    private static final byte CMD_SEND_LOGS               = 0x01;
    private static final byte CMD_BCP_LOG                 = 0x02;
    private static final byte CMD_BCP_LOG_RESPONSE        = 0x03;
    private static final byte CMD_DISCONNECT_AND_SHUTDOWN = 0x04;
    
    private static final byte STATE_DISCONNECTED        = 0x01;
    private static final byte STATE_CONNECTING          = 0x02;
    private static final byte STATE_CONNECTED           = 0x03;
    
    // This is used to process the message queue
    private final MessageSystem  _messageSystem         = new MessageSystemOEM();
    // This is the object with the callbacks we'll be using to notify send failure/success
    private SenderListener       _closeListener         = null;
    // We need to keep a reference to the serialized data so we can return in to the called if the send fails
    private Queue                _sendingQueue          = new LinkedList();
    private XMode                _xmode                 = null;
    // Once we unserialize the events this will hold them
    private final List           _events                = new LinkedList();
    // Not sure why session id is not used
    private static byte[]        _emptySessionId        = new byte[16];
    private int                  _requestIdCounter      = 1;    
    
    // These are the gateway parameters retrieved from the serialized data
    private String               _gatewayIP             = null;
    private short                _gatewayPort           = 0;
    private String               _applicationId         = null;
    private byte[]               _appKey                = null; 
    private Vector               _xmodeParams           = null;
    private Vector                _overrideParameters	= null;
    private String               _uid                   = null;
    
    private int                  _xmodeState            = STATE_DISCONNECTED;
    private int                  _outstandingEvents     = 0;
    
    // This keeps track of the total size actually sent to the socket. If this size would exeed
    // CALLLOG_CHUNK_SIZE_MAXIMUM, the data is not sent and the client is notified with the 
    // SenderListener.failed() callback
    private int                  _sizeSent              = 0;
    private boolean shutdownUponDisconnect = false;
    
    public CalllogSenderImpl(SenderListener listener, Vector overrideParameters)
    {
        //#ifdef NMSP_LOG
        if (log.isDebugEnabled())
        {
            log.debug("CalllogSenderImpl() " + this + " with listener (" + listener + ")");
        }
        //#endif

        _closeListener = listener;        
        _overrideParameters = overrideParameters == null ? new Vector() : overrideParameters;
        for( int i = 0; i < _overrideParameters.size(); ++i){
            if(!(_overrideParameters.get(i) instanceof Parameter)){
                //#ifdef NMSP_LOG
                if (log.isErrorEnabled())
                {
                    log.debug("Trying to add connection parameters of type other than Parameter, skipping override list");
                }
                //#endif
                _overrideParameters = new Vector();
                break;
            }
        }
    }
    
    public void addConnectionParameters(Vector extraParameters)
    {
    
    }
    
    // Application thread
    // Override
    public void shutdown()
    {
        //#ifdef NMSP_LOG
        if (log.isDebugEnabled())
        {
            log.debug("shutdown the current call log sender: " + this);
        }
        //#endif
    	
        _messageSystem.send(new MessageData(CMD_DISCONNECT_AND_SHUTDOWN, null ), this, Thread.currentThread(), _messageSystem.getVRAddr()[0]);
    }
    
    // Application thread
    // Override
    public void send(byte[] data)
    {
        _requestIdCounter = 1;
        _messageSystem.send(new MessageData(CMD_SEND_LOGS, new Object[]{ data } ), this, Thread.currentThread(), _messageSystem.getVRAddr()[0]);
    }

      
    /** XMode Listener methods **/
    // sdk thread
    // Override
    public void xmodeMsgCallback(XModeMsgHeader header, byte[] payloadBuf)
    {
        //#ifdef NMSP_LOG
        if(log.isDebugEnabled())
            log.debug("xmodeMsgCallback, protocol: " + header.protocol + ", command: " + header.cmd);
        //#endif
        switch (header.protocol)
        {
        case ProtocolDefines.XMODE_PROTOCOL_COP:
        	switch (header.cmd)
        	{
        	case ProtocolDefines.XMODE_COP_COMMAND_CONNECTED:
                //#ifdef NMSP_LOG
                if(log.isDebugEnabled())
                    log.debug("connected(" + Log.FormatUuid(_xmode.sessionId) + ") for Call Log Sender " + this);
                //#endif
                
            }
        }
    }

    // sdk thread
    // Override
    public void xmodeMsgSent(String context, Object data)
    {
        if (!context.startsWith(XMode.NET_CONTEXT_SEND_BCP_LOG)){
            return;
        }

        //#ifdef NMSP_LOG
        if (log.isDebugEnabled())
            log.debug("onMsgSent(" + context + ")");
        //#endif
        
        _outstandingEvents--;
        
        if(_outstandingEvents == 0)
        {
            _closeListener.succeeded((byte[])_sendingQueue.remove());
            trySendingLogs();
        }
    }

    // sdk thread
    // Override
    public void xmodeMsgNotSent(String context, Object data)
    {
        if (!context.startsWith(XMode.NET_CONTEXT_SEND_BCP_LOG)){
            return;
        }
        //#ifdef NMSP_LOG
        if (log.isDebugEnabled())
            log.debug("onMsgNotSent(" + context + ")");
        //#endif
        if(_outstandingEvents > 0)
        {
            notifyError(CalllogSender.CALLLOG_NETWORK_ERROR);
        }

    }

    // sdk thread
    // Override
    public void socketClosed(short reason, short subReason)
    {
        //#ifdef NMSP_LOG
        if ( log.isDebugEnabled() )
        {
            log.debug("socketClosed() reason [" + reason + "] subReason [" + subReason + "]");
        }
        //#endif
        _xmodeState = STATE_DISCONNECTED;
        
        if ( reason == XMode.XMODE_SOCKET_CLOSED_OPEN_SOCKET_FAILED ||
             reason == XMode.XMODE_SOCKET_CLOSED_OPEN_SOCKET_FAILED_SECURITY )
        {
            notifyError(CalllogSender.CALLLOG_NETWORK_ERROR);
        }
        
        if ( reason == XMode.XMODE_SOCKET_CLOSED_DISCONNECT && shutdownUponDisconnect )
        {
            _messageSystem.stop();
            //_closeListener.shutdownCompleted();
        }
    }
    
    private void notifyError(short errorCode)
    {
        // In error state, we cannot continue sending the rest of the data in
        // the sending queue, so we have to report CALLLOG_NETWORK_ERROR for
        // each of them.
        while ( _sendingQueue.size() > 0 )
        {
            byte[] data = (byte[]) _sendingQueue.remove();
            _closeListener.failed(errorCode, data);
        }
        
        _outstandingEvents = 0;
        _events.clear();
        _sendingQueue.clear();
        if(_xmode != null){
        _xmode.disconnect();
        _xmode = null;
        }

    }
    
    // sdk thread
    // Override
    public void socketOpened()
    {
        //#ifdef NMSP_LOG
        if (log.isDebugEnabled())
            log.debug("socketOpened()");
        //#endif
        
    }
    
    // sdk thread
    // Override
    public void copConnected()
    {
        _xmodeState = STATE_CONNECTED;
        trySendingLogs();
    }

    private void trySendingLogs()
    {
        if(_outstandingEvents == 0 && _sendingQueue.size() > 0)
        {
        	List event2Send = unserializeEvents((byte[])_sendingQueue.peek());
        	if (event2Send != null)
        	{
        		_events.addAll(event2Send);
        		_outstandingEvents = _events.size();
        	}
        }
        //#ifdef NMSP_LOG
        if (log.isDebugEnabled())
            log.debug("trySendingLogs() , _events.size()" + _events.size());
        //#endif
        
        while (_events.size() != 0)
        {
            bcpLog((SessionEventData)_events.remove(0));
        }
    }
    
    /******** BCP LOG ********/
    private void bcpLog(SessionEventData sessionEventData)
    {
        //#ifdef NMSP_LOG
     if (log.isDebugEnabled())
        log.debug("bcpLog, packet len [" + sessionEventData.data.length + "]");
        //#endif

        _messageSystem.send(new MessageData(CMD_BCP_LOG, new Object[]{ sessionEventData } ), this, Thread.currentThread(), _messageSystem.getVRAddr()[0]);
    }
    
    private int getRequestId()
    {
        return _requestIdCounter ++;  
    }
    
    // handlers
    private void handleBcpLog(SessionEventData eventLog)
    {
    	// This can happen if we previously called notifyError but other logs were in the sdk thread queue
    	// Then xmode will be disconnected and null so we don't bother creating the output buffer.
    	// Call notifyError again to make sure _sendingQueue is properly emptied
    	if(_xmode == null){
            notifyError(CalllogSender.CALLLOG_NETWORK_ERROR);
    		return;
    	}
    	eventLog.requestId = getRequestId();
        
        int payloadSize = ProtocolDefines.XMODE_SERVER_SESSION_UUID_SIZE +
                          ProtocolDefines.XMODE_BCP_TRANSACTION_ID_SIZE +
                          ProtocolDefines.XMODE_BCP_REQUEST_ID_SIZE +
                          ProtocolDefines.XMODE_PAYLOAD_LENGTH_SIZE + 
                          eventLog.data.length;
        int offset = 0;
        byte[] payload = new byte[payloadSize];
        
        //transaction id is not used.
        payload[offset] = 0;
        offset += ProtocolDefines.XMODE_BCP_TRANSACTION_ID_SIZE;
        
        ByteConversion.intToBytes(eventLog.requestId, payload, offset);
        offset += ProtocolDefines.XMODE_BCP_REQUEST_ID_SIZE;
        
        payload = ProtocolBuilder.appendBCPSessionUUID(payload, eventLog.sessionEventId.getBytes());
        offset += ProtocolDefines.XMODE_SERVER_SESSION_UUID_SIZE;

        ByteConversion.intToBytes(eventLog.data.length, payload, offset);
        offset += ProtocolDefines.XMODE_XML_PAYLOAD_SIZE;

        System.arraycopy(eventLog.data, 0, payload, offset, eventLog.data.length);
        
        byte[] buffer = ProtocolBuilder.buildXModeBuf(
                ProtocolDefines.XMODE_PROTOCOL_BCP,
                ProtocolDefines.XMODE_VERSION_BCP,
                ProtocolDefines.XMODE_BCP_COMMAND_BCP_LOG, 
                payload);
        
        if (_sizeSent + buffer.length > CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM){
            //#ifdef NMSP_LOG
            if (log.isErrorEnabled())
                log.error("handleBcpLog() : Calllog reached maximum size [" + _sizeSent + buffer.length + "] Maximum is : [" + CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM + "]");
            //#endif
            notifyError(CalllogSender.CALLLOG_INVALID_DATA);
            return;
        }
        _sizeSent += buffer.length;
        
        _xmode.sendXModeMsg(buffer, XMode.SEND_MSG_PRIORITY_HIGH, XMode.NET_CONTEXT_SEND_BCP_LOG);
    }

    private int decodeSize(byte[] sizeData)
    {
        int size;
        if (sizeData.length == 1) {
            size = (sizeData[0] & 0xFF);
        } else if (sizeData.length == 2) {
            size = (((sizeData[1] & 0xFF) << 8) + (sizeData[0] & 0xFF));
        } else if (sizeData.length == 3) {
            size = (((sizeData[2] & 0xFF) << 16) + ((sizeData[1] & 0xFF) << 8) + (sizeData[0] & 0xFF));
        } else {
            size = (((sizeData[3] & 0xFF) << 24) + ((sizeData[2] & 0xFF) << 16)
                    + ((sizeData[1] & 0xFF) << 8) + (sizeData[0] & 0xFF));
        }
        
        return size;
    }
    
    private class DataOffset
    {
        public int offset = 0;
    }
    
    private byte[] unserializeData(byte[] serializedData, DataOffset offset)
    {
        // Get the real size value
        byte[] sizeData = new byte[4];
        System.arraycopy(serializedData, offset.offset, sizeData, 0, sizeData.length);
        offset.offset += sizeData.length;
        // Get the data 
        int decodedDataSize = decodeSize(sizeData);
        byte[] data = new byte[decodedDataSize];
        System.arraycopy(serializedData, offset.offset, data, 0, data.length);
        offset.offset += data.length;
        
        return data;
    }
    
    private List unserializeEvents(byte[] serializedData)
    {
        List events = new LinkedList();
        DataOffset off = new DataOffset();
        byte[] sslParamsCount = new byte[1];
        byte[] serializedHash = new byte[20];
        _xmodeParams = new Vector();
        byte[] fullSize;
        int totalSize = 0;
        
        fullSize = unserializeData(serializedData, off);
        totalSize = ByteConversion.bytesToInt(fullSize, 0);
        
        System.arraycopy(serializedData, 4 + 4, serializedHash, 0, serializedHash.length);
        byte[] expectedHash = MiscSystem.HMAC_SHA1(CalllogImpl.INTERNAL_SHA1_KEY.getBytes(), serializedData, 28);
        off.offset += 20;
        
        for(int i = 0 ; i < expectedHash.length; ++i){
            if(expectedHash[i] != serializedHash[i]){
                //#ifdef NMSP_LOG
                if (log.isErrorEnabled())
                    log.error("unserializeEvents() hashes do not match");
                //#endif
                return null;
            }
        }
       
        try
        {
            _gatewayIP = new String(unserializeData(serializedData, off));
            byte[] portAsByte = unserializeData(serializedData, off);
            // Convert port short to byte array
            _gatewayPort = ByteBuffer.wrap(portAsByte).order(ByteOrder.LITTLE_ENDIAN).getShort();;
            _applicationId = new String(unserializeData(serializedData, off));
            _appKey = unserializeData(serializedData, off);
            _uid = new String(unserializeData(serializedData, off));
            sslParamsCount = unserializeData(serializedData, off);
            
            for(int i = 0 ; i < sslParamsCount[0]; i++)
            {
                Parameter p = null;
                
                // We need to unserialize everything anyway since it increases the data offset
                String paramName = new String(unserializeData(serializedData, off));
            	byte[] paramTypeAsBytes = unserializeData(serializedData, off);
                Parameter.Type type = Parameter.Type.SDK;
                byte[] paramValue = unserializeData(serializedData, off);
                
                // Try to find the current parameter in the override parameter list. If we find it then we'll pass it to xmode
                // Otherwise we'll use what is serialized in data. This is for parameters that cannot be serialized such as
                // an Android Context. This also allows the CalllogSender to use different connection parameters than those
                // which were used during the log creation.
                for (int overrideParamIndex = 0; overrideParamIndex < _overrideParameters.size(); ++overrideParamIndex){
                    Parameter currentParam = (Parameter)_overrideParameters.get(overrideParamIndex);
                    String currentParamName = currentParam.getName(); 
                    if (paramName.equals(currentParamName)){
                        p = currentParam;
                        break;
                    }
                }
                if(p == null){
                    p = new Parameter(paramName, paramValue, type);
                }
                _xmodeParams.add(p);
            }
            
            while(off.offset < totalSize)
            {
                SessionEventData sessionEventData = new SessionEventData();
                sessionEventData.data = unserializeData(serializedData, off);
                sessionEventData.sessionEventId = new String(unserializeData(serializedData, off));
    
                events.add(sessionEventData);
            }
        }
        catch(Exception e)
        {
            //#ifdef NMSP_LOG
            if (log.isErrorEnabled())
                log.error("unserializeEvents() failed!!! : " + e.getLocalizedMessage());
            //#endif
            return null;
        }
        return events;
    }
    
    private void handleSendLogs(byte[] data)
    {
        if (data == null)
        {
            _closeListener.failed(CALLLOG_INVALID_DATA, data);
            return;
        }
        _sendingQueue.add(data);
        
        //#ifdef NMSP_LOG
        if (log.isDebugEnabled())
        {
            if (data != null)
            {
                log.debug("handleSendLogs() data " + data + " length " + data.length + ")");
            }
        }
        //#endif

        //#ifdef NMSP_LOG
        if (log.isDebugEnabled())
        {
            log.debug("handleSendLogs() xmode state " + _xmodeState + ")");
        }
        //#endif
        
        if (_xmodeState == STATE_DISCONNECTED)
        {
            //#ifdef NMSP_LOG
            if (log.isDebugEnabled())
            {
                log.debug("handleSendLogs() connect");
            }
            //#endif
            _xmodeState = STATE_CONNECTING;
            unserializeEvents(data);
            _xmode = new XMode(_gatewayIP, _gatewayPort, _applicationId, _appKey, _uid, this, _xmodeParams, _messageSystem);
            _xmode.setForLogOnly();
            _xmode.connect((short)0, (short)0, null);
        }
        else if (_xmodeState == STATE_CONNECTED)
        {
            //#ifdef NMSP_LOG
            if (log.isDebugEnabled())
            {
                log.debug("handleSendLogs() trying to send log in connected state.");
            }
            //#endif

            trySendingLogs();
        }
    }
    
    private void handleDisconnectAndShutdown()
    {
        shutdownUponDisconnect = true;
        if(_xmode != null){
        _xmode.disconnect();
        }
    }
    
    public void handleMessage(Object msgData, Object msgSender)
    {
        MessageData messageData = (MessageData) msgData;
        switch (messageData.command)
        {
            case CMD_SEND_LOGS:
                Object[] data = (Object[]) messageData.data;
                handleSendLogs((byte[])data[0]);
                break;
            case CMD_BCP_LOG:
                Object[] currentEvent = (Object[]) messageData.data;
                handleBcpLog((SessionEventData) currentEvent[0]);
                break;
            case CMD_BCP_LOG_RESPONSE:
                break;
            case CMD_DISCONNECT_AND_SHUTDOWN:
                handleDisconnectAndShutdown();
                break;
            default:
//                super.handleMessage(msgData, msgSender);
        }
    }
    
}
