/**
 *  Copyright 2006-2011, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.nmsp.client.sdk.components.general;

import com.nuance.nmsp.client.sdk.common.util.ShortConstant;

/**
 * Represents a parameter. Includes methods for retrieving
 * the parameter type, name, and value.
 */
public class Parameter {

    private String name;
    private Object value;
    private Parameter.Type type;

    /**
     * Parameter type. Currently, the supported parameter types are: APP, NSS, 
	 * SDK, GWLOG, NSSLOG, and SLOG. The other types are reserved for future use. 
	 * For more information, see the <em>Programmer's Guide</em> section on session-related 
	 * parameters (in the chapter on functionality common across SDK modules).
     *
     */
    public static class Type extends ShortConstant {
        private Type(short value) {
            super(value);
        }

        public static final Type SDK    = new Type((short) 0x0000);
        public static final Type NMSP   = new Type((short) 0x0001);
        public static final Type APP    = new Type((short) 0x0002);
        public static final Type NSS    = new Type((short) 0x0003);
        public static final Type SLOG   = new Type((short) 0x0004);
        public static final Type NSSLOG = new Type((short) 0x0005);
        public static final Type GWLOG  = new Type((short) 0x0006);
        public static final Type SVSP   = new Type((short) 0x0007);
        public static final Type SIP    = new Type((short) 0x0008);
        public static final Type SDP    = new Type((short) 0x0009);

        public String toString() {
            if (equals(Type.NMSP)) {
                return "nmsp";
            } else if (equals(Type.APP)) {
                return "app";
            } else if (equals(Type.NSS)) {
                return "nss";
            } else if (equals(Type.SLOG)) {
                return "slog";
            } else if (equals(Type.NSSLOG)) {
                return "nsslog";
            } else if (equals(Type.GWLOG)) {
                return "gwlog";
            } else if (equals(Type.SVSP)) {
                return "svsp";
            } else if (equals(Type.SIP)) {
                return "sip";
            } else if (equals(Type.SDP)) {
                return "sdp";
            } else if (equals(Type.SDK)) {
                return "sdk";
            } else {
                return null;
            }
        }
    }

    /**
     * Constructs a new parameter object to set a parameter.
     * 
     * @param name parameter name
     * @param value parameter value
     * @param type parameter type
     */
    public Parameter(String name, Object value, Parameter.Type type) {
        this.name = name;
        this.value = value;
        this.type = type;
    }

    /**
     * Constructs a new parameter object to get a parameter.
     * 
     * @param name parameter name
     * @param type parameter type
     */
    public Parameter(String name, Parameter.Type type) {
        this.name = name;
        this.value = new byte[0];
        this.type = type;
    }

    /**
     * Returns the name of the parameter.
     * 
     * @return parameter name
     */
    public String getName() {
        return name;
    }

    /**
     * Returns the value of the parameter.
     * 
     * @return parameter value
     */
    public byte[] getValue() {
        return value instanceof byte[] ? (byte[])value : "THIS IS NOT A STRING PARAMETER!!!".getBytes();
    }
    
    /**
     * Returns the raw value of the parameter.
     * 
     * @return parameter value
     */
    public Object getValueRaw() {
        return value;
    }

    /**
     * Returns the type of the parameter.
     * 
     * @return parameter type
     */
    public Parameter.Type getType() {
        return type;
    }
    
    /**
     * Creates and returns a copy of this object.
     * 
     * @return a clone of this parameter instance.
     */
    public Parameter clone() {
        Object value_copy = value;
        if (value instanceof byte[]) {
            byte[] value_origin = (byte[])value;
            value_copy = new byte[value_origin.length];
            System.arraycopy(value_origin, 0, value_copy, 0, value_origin.length);
        }
        return new Parameter(name, value_copy, type);
    }
}
