/**
 *  Copyright 2006-2011, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.nmsp.client.sdk.components.resource.internal.nmas;

//#ifdef NMSP_LOG
import com.nuance.nmsp.client.sdk.common.oem.api.LogFactory;
import com.nuance.nmsp.client.sdk.common.oem.api.LogFactory.Log;
//#endif
import java.io.ByteArrayOutputStream;

import com.nuance.nmsp.client.sdk.components.general.TransactionProcessingException;
import com.nuance.nmsp.client.sdk.components.resource.internal.common.NMSPSession;
import com.nuance.nmsp.client.sdk.components.resource.nmas.AudioParam;
import com.nuance.nmsp.client.sdk.common.oem.api.MessageSystem;
import com.nuance.nmsp.client.sdk.common.oem.api.MessageSystem.MessageData;
import com.nuance.nmsp.client.sdk.common.util.Util;
//#ifdef NMSP_CPL_ENABLED
//@import java.util.Vector;
//@import com.nuance.nmsp.client.sdk.common.oem.api.log.CPLRecord;
//#endif

public class PDXAudioParam extends PDXParam implements AudioParam, MessageSystem.MessageHandler {
    //#ifdef NMSP_LOG
    private static final Log log = LogFactory.getLog(PDXAudioParam.class);
    //#endif

    private int bufferId;
    private NMSPSession session;
    private boolean lastAudioBufAdded = false;
    private MessageSystem msgSys;
    
    private static final byte CMD_AUDIOPARAM_SEND_VAPBEGIN = 0x01;
    private static final byte CMD_AUDIOPARAM_ADD_AUDIO_BUFF = 0x02;
    
    private final static String MESG_AUDIO_OFFSET_ILLEGAL = "the offset of the \"buffer\" is less than 0";
    private final static String MESG_AUDIO_LENGTH_ILLEGAL = "the indicated length of the \"buffer\" is less than 1 byte";

    //#ifdef NMSP_CPL_ENABLED
//@    private boolean firstAudioBufAdded = false;
//@    public long recordStartTime;
//@    public long recordStopTime;
//@    public long streamStartTime;
//@    public long streamEndTime;
    //#endif

    public PDXAudioParam(String name, NMSPSession session, MessageSystem msgSys) {
        super(name, PDXParam.AUDIO_PARAMETER);
        //#ifdef NMSP_LOG2
//@        log.debug("PDXAudioParam() " + name);
        //#endif
        
        this.session = session;
        this.msgSys = msgSys;
        this.bufferId = session.getNewAudioId();
        //#ifdef NMSP_CPL_ENABLED
//@        this.recordStartTime = 0;
//@        this.recordStopTime = 0;
//@        this.streamStartTime = 0;
//@        this.streamEndTime = 0;
        //#endif
        
        msgSys.send(new MessageData(CMD_AUDIOPARAM_SEND_VAPBEGIN, null), this, Thread.currentThread(), msgSys.getVRAddr()[0]);
    }
    
    //#ifdef NMSP_CPL_ENABLED
//@    public void addAudioBuf(byte[] buffer, boolean isLastBuf, long recordStartTime, long recordStopTime) throws TransactionProcessingException {
    //#else
    public void addAudioBuf(byte[] buffer, int offset, int length, boolean isLastBuf) throws TransactionProcessingException {
    //#endif
        //#ifdef NMSP_LOG2
//@        log.debug("PDXAudioParam.addAudioBuf() " + isLastBuf);
        //#endif
        if (buffer == null && !isLastBuf) {
            //#ifdef NMSP_LOG
            log.error("PDXAudioParam.addAudioBuf() " + MESG_AUDIO_NULL);
            //#endif
            throw new NullPointerException(MESG_AUDIO_NULL);
        }
        
        if (buffer != null && offset < 0) {
            //#ifdef NMSP_LOG
            log.error("PDXAudioParam.addAudioBuf() " + MESG_AUDIO_OFFSET_ILLEGAL);
            //#endif
            throw new IllegalArgumentException(MESG_AUDIO_OFFSET_ILLEGAL);
        }
        if (buffer != null && length <= 0) {
            //#ifdef NMSP_LOG
            log.error("PDXAudioParam.addAudioBuf() " + MESG_AUDIO_LENGTH_ILLEGAL);
            //#endif
        	throw new IllegalArgumentException(MESG_AUDIO_LENGTH_ILLEGAL);
        }

        if (lastAudioBufAdded) {
            //#ifdef NMSP_LOG
            log.error("PDXAudioParam.addAudioBuf() " + MESG_TRANSACTION_PROCESSING_ERROR);
            //#endif
            throw new TransactionProcessingException(MESG_TRANSACTION_PROCESSING_ERROR);
        }
        
        if (isLastBuf) {
            lastAudioBufAdded = true;
        }
        
        byte[] audioBuffer = null;
        if (buffer != null) {
            audioBuffer = new byte[length];
            System.arraycopy(buffer, offset, audioBuffer, 0, length);
        }

        Object array[] = new Object[2];
        array[0] = audioBuffer;
        array[1] = new Boolean(isLastBuf);
        msgSys.send(new MessageData(CMD_AUDIOPARAM_ADD_AUDIO_BUFF, array), this, Thread.currentThread(), msgSys.getVRAddr()[0]);
    }

    protected int getBufferId() {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXAudioParam.getBufferId() " + this.bufferId);
        //#endif
        return this.bufferId;
    }

    private final static String MESG_AUDIO_NULL                   = "in (NMSPAudioSink)PDXAudioParam.addAudioBuf(), the param \"buffer\" is null.";
    private final static String MESG_TRANSACTION_PROCESSING_ERROR = "last audio buffer already added!";
    
    public void handleMessage(Object msgData, Object msgSender) {
        MessageData tranMsgData = (MessageData) msgData;
        Object array[];
        switch (tranMsgData.command) {
            case CMD_AUDIOPARAM_SEND_VAPBEGIN:
                handleSendVapBegin();
                break;
            case CMD_AUDIOPARAM_ADD_AUDIO_BUFF:
                array = (Object[]) tranMsgData.data;
                handleAddAudioBuff((byte[]) array[0], ((Boolean)array[1]).booleanValue());
                break;
        }
    }
    
    private void handleSendVapBegin() {
        session.vapRecordBegin(bufferId);
    }
    
    private void handleAddAudioBuff(byte[] buffer, boolean isLastBuf) {
        if (buffer != null) {
        	//TODO check with server guys what is the real protocol for buffer length
            if(Util.isSpeexCodec(session.getInputCodec()) ||
               Util.isOpusCodec(session.getInputCodec())) {
                 ByteArrayOutputStream byteArray = new ByteArrayOutputStream();
                 
                 if(buffer.length < 0x80) {
                     byteArray.write(buffer.length & 0x7F);
                 }
                 else if(buffer.length < 0x4000) {
                     byteArray.write(((buffer.length >> 7) & 0x7F) | 0x80);
                     byteArray.write(buffer.length & 0x7F);
                 }
                 else if(buffer.length < 0x200000) {
                     byteArray.write(((buffer.length >> 14) & 0x7F) | 0x80);
                     byteArray.write(((buffer.length >> 7) & 0x7F) | 0x80);
                     byteArray.write(buffer.length & 0x7F);
                 }
                 else if(buffer.length < 0x10000000) {
                     byteArray.write(((buffer.length >> 21) & 0x7F) | 0x80);
                     byteArray.write(((buffer.length >> 14) & 0x7F) | 0x80);
                     byteArray.write(((buffer.length >> 7) & 0x7F) | 0x80);
                     byteArray.write(buffer.length & 0x7F);
                 }
                 else {
                     //#ifdef NMSP_LOG
                         log.error("buffer size is too big!");
                     //#endif
                     return;
                 }
                 byteArray.write(buffer, 0, buffer.length);
                 session.vapRecord(byteArray.toByteArray(), bufferId);
            }
            else {
            	session.vapRecord(buffer, bufferId);
            }
        }

        //#ifdef NMSP_CPL_ENABLED
//@        if(!firstAudioBufAdded) {
//@            streamStartTime = System.currentTimeMillis();
//@            this.recordStartTime = recordStartTime;
//@            firstAudioBufAdded = true;
//@        }
        //#endif
        
        if (isLastBuf) {
            //#ifdef NMSP_CPL_ENABLED
//@            streamEndTime = System.currentTimeMillis();
//@            this.recordStopTime = recordStopTime;
            //#endif
            session.vapRecordEnd(bufferId);
        }
    }
}
