/**
 *  Copyright 2006-2011, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.nmsp.client.sdk.components.resource.nmas;

import com.nuance.nmsp.client.sdk.components.general.TransactionAlreadyFinishedException;
import com.nuance.nmsp.client.sdk.components.general.TransactionExpiredException;

/**
 * Represents the transaction between the client application
 * and the associated command on the Nuance Mobile Application Server.
 * To create an object that implements the Command interface, use {@link com.nuance.nmsp.client.sdk.components.resource.nmas.NMASResource createCommand()}.
 * 
 * @see NMASResource
 * 
 * @author Nuance Communication Inc.
 */
public interface Command {
    /**
     * Sends a generic parameter (a <code>QueryParam</code> message) to the Nuance Mobile 
     * Application Server.
     * 
     * @param param a PDX parameter
     * 
     * @throws TransactionAlreadyFinishedException  when the transaction is in the <em>FINISHED</em> state
     * @throws TransactionExpiredException  when the transaction is in the <em>EXPIRED</em> state
     */
    public void sendParam(Parameter param) throws TransactionAlreadyFinishedException, TransactionExpiredException;

    /**
     * Sends enrollment audio to the Nuance Mobile Application Server.
     * 
     * For example, in a class implementing NMSPAudioSink:
     * <pre>
     * <code>
     * public void addAudioBuf(byte[] buffer, boolean isLastBuf) throws TransactionProcessingException {
     *       byte[] tempBuf = new byte[buffer.length];
     *       System.arraycopy(buffer, 0, tempBuf, 0, buffer.length);
     *       try {
     *           getPdxCommand().sendEnrollmentAudio(tempBuf);
     *       } catch (Exception e) {
     *         e.printStackTrace();
     *       }
     * } 
     * </code>
     * </pre>
     * @param audio    audio buffer
     * 
     * @throws TransactionAlreadyFinishedException  when the transaction is in the <em> FINISHED</em> state
     * @throws TransactionExpiredException  when the transaction is in the <em> EXPIRED</em> state
     */
    public void sendEnrollmentAudio(byte[] audio) throws TransactionAlreadyFinishedException, TransactionExpiredException;

    /**
     * Ends the transaction (by sending a <code>QueryEnd</code> message), which notifies the 
     * Nuance Mobile Application Server that the client application has finished sending
     * the required parameters and is ready to receive results. 
     * <br /><br />
     * <b><em>Note: </em></b>The Nuance Mobile Application Server does not return a result to the client application
     * until this function is called. (However, it can return <code>QueryError</code> messages without calling this function.)
     * It is a good practice to call this function as soon as the
     * application knows that the last {@link #sendParam(Parameter) sendParam()} call was made. In the case of a simple single
     * audio parameter, the application can do the following:<br />
     * <pre>
     * <code>
     *    command.sendParam(myAudioParam)
     *    command.end()
     * </code>
     * </pre>
     * 
     * @throws TransactionAlreadyFinishedException  indicates that the transaction is in the <em>FINISHED</em> state
     * @throws TransactionExpiredException  indicates that the transaction is in the <em>EXPIRED</em> state
     * 
     */
    public void end() throws TransactionAlreadyFinishedException, TransactionExpiredException;
    
    /**
     * Indicates if there is latency in the network or if
     * there is no connection to the network or the NMSP Gateway.
     * Can be used to indicate this information on the mobile device, 
     * for example, with an icon or a message.
     * <br /><br />
	 * <b><em>Note: </em></b>Do not use on Android. It has not been implemented and will always return true.<br />
     * @return true if the network is connected
     */
    public boolean isNetworkHealthy();
    
    /**
     * The completion cause type, which indicates how the
     * <code>Command</code> terminates.
     */
    public static class CompletionCause {
        private String log;
        private CompletionCause(String log) {
            this.log = log;
        }
        public String toString() {
            return log;
        }
    }

    /**
     * The user aborts the application by pressing the END key to cancel processing
     * (for example, BlackBerry red key).
     */
    public static final CompletionCause ABORT_KEY_END = new CompletionCause("ABORT_END");
    
    /**
     * The user aborts the application by pressing the BACK key to cancel processing
     * (for example, BlackBerry back key or Close in the default menu).
     */
    public static final CompletionCause ABORT_KEY_BACK = new CompletionCause("ABORT_BACK");
    
    /**
     * The user aborted the current transaction to perform a NEW command.
     */
    public static final CompletionCause ABORT_NEW = new CompletionCause("ABORT_NEW");
    
    /**
     * The user aborted the application by pressing the side key. 
     */
    public static final CompletionCause STOPPED_TOO_SOON = new CompletionCause("STOPPED_TOO_SOON");
    
    /**
     * The application was forced to go into the background
     * (due to incoming call, incoming SMS, etc.)
     */
    public static final CompletionCause PREEMPTED = new CompletionCause("PREEMPTED");

    /**
     * When a <code>Command</code> ends, you can set the proper completion
     * cause by calling this function. The SDK will collect these logs for improvement. 
     * 
     * @param cplCause completion cause of the current <code>Command</code> 
     */
    public void setLog(CompletionCause compCause);
}
