/**
 *  Copyright 2006-2011, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.nmsp.client.sdk.components.resource.nmas;

import com.nuance.nmsp.client.sdk.components.core.pdx.Dictionary;
import com.nuance.nmsp.client.sdk.components.core.pdx.Sequence;
import com.nuance.nmsp.client.sdk.components.general.NMSPAudioSink;
import com.nuance.nmsp.client.sdk.components.resource.common.Resource;
import com.nuance.nmsp.client.sdk.components.resource.common.ResourceException;

/**
 * Accesses the Nuance Mobile Application 
 * Server resource through the NMSP 3.0 SDK.
 * 
 * @see NMASResourceListener
 * 
 * @author Nuance Communication Inc.
 */
public interface NMASResource extends Resource {
    
    public static final short PDX_DATA_TYPE_INT   = (short) (0xC0 & 0xFF); //192
    public static final short PDX_DATA_TYPE_BYTES = 0x04;                  //004
    public static final short PDX_DATA_TYPE_UTF8  = (short) (0xFF & 0xC1); //193
    public static final short PDX_DATA_TYPE_ASCII = 0x16;                  //022
    public static final short PDX_DATA_TYPE_NULL  = 0x05;                  //005
    public static final short PDX_DATA_TYPE_DICT  = (short) (0xFF & 0xE0); //224
    public static final short PDX_DATA_TYPE_SEQ   = 0x10;                  //016
    public static final short PDX_DATA_TYPE_DOUBLE = 0x06;				   //006
    public static final short PDX_DATA_TYPE_BOOLEAN = 0x07;                //007
    
    /**
     * Creates a Command object to hold the transaction parameter values   
     * that the gate command expects to receive. Most of these parameters
     * contain connection information. The PDXCommandListener object is  
     * attached to the newly created Command object. 
     *  
     * @param commandListener    a PDXCommandListener object that receives the events of
     *                           PDX transactions <br /> <br />
     * @param commandName        command name (for example, GetWeather) to execute on the
     *                           Nuance Mobile Application Server host  <br /><br />
     * @param applicationVersion version of the application running on the mobile device; used to coordinate over-the-air updates <br /><br />
     * @param dictationLanguage  language used for dictation on the mobile device;<br /> <br />
     * @param carrier            mobile service provider  <br /><br />
     * @param phoneModel         mobile device model <br /><br />
     * @param optionalKeys       dictionary of key/values; values cannot be of type SEQUENCE or DICTIONARY <br /><br />    
     * @param commandTimeout     length of time, in milliseconds, that the client application waits for a response from
     *                           Nuance Mobile Application Server before timing out; must be greater than 0
     *                           <br /><br />
     *         
     * The following three parameters are used by Nuance Mobile Application Server to select 
     * the appropriate set of actions and parameters for each command. Using the Command Action 
	 * Definition Language (CADL), these parameters must map to the gate commands running on the 
	 * Nuance Mobile Application Server host
	 .
     *                           <br /> <br />
     * @param commandName        command name (for example, GetWeather) to execute on the
     *                           Nuance Mobile Application Server host  <br /><br />
     * @param carrier            mobile service provider <br /><br />
     * @param phoneModel         mobile device model
     * 
     * @return Command
     * @throws ResourceException 
     * 
     * @throws IllegalArgumentException        if transaction parameter value is
     *                                         incorrect     
     */
     public Command createCommand(PDXCommandListener commandListener, String commandName, String applicationVersion,
    		String dictationLanguage, String carrier, String phoneModel, Dictionary optionalKeys,
            long commandTimeout) throws ResourceException;
    
    /**
     * Creates a PDX Audio Parameter. The returned <code>AudioParam</code> is an {@link NMSPAudioSink} and should be
     * used to pass the audio parameter to the Nuance Mobile Application Server resource and stream the audio buffers to 
     * the NMSP Gateway.
     * 
     * @param name    name of the PDX Audio Parameter
     * 
     * @return AudioParam
     */
    public AudioParam createPDXAudioParam(String name);
    
    /**
     * Creates a parameter to hold the name and value of a possible choice according to the 
     * results returned by the command on the Nuance Mobile Application Server host. 
     * 
     * @param name        name of the PDX choice parameter expected by the command
     * @param choicename  a value that represents the choice specified by the client 
     * 
     * @return Parameter
     */
    public Parameter createPDXChoiceParam(String name, String choicename);
    
    /**
     * Creates a parameter to send binary data to the command.  
     * 
     * @param name  name of the PDX data parameter expected by the command
     * @param data  raw binary data 
     * 
     * @return Parameter
     */
    public Parameter createPDXDataParam(String name, byte[] data);
    
    /**
     * Creates a parameter to send text data to the Nuance Mobile  
     * Application Server network command.
     * 
     * @param name  name of the String parameter expected by the command
     * @param text  text data 
     * 
     * @return Parameter
     */
    public Parameter createPDXTextParam(String name, String text);
    
    /**
     * Creates a parameter to send PDX Dictionary data to the Nuance Mobile  
     * Application Server network command. The PDX Dictionary contains one or 
     * more key-value pairs.
     * 
     * @param name  name of the PDX Dictionary Parameter expected by the command
     * @param dict  PDX Dictionary
     * 
     * @return Parameter
     */
    public Parameter createPDXDictParam(String name, Dictionary dict);
    
    /**
     * Creates a PDX Sequence Start Parameter, which indicates 
     * the start of a series of data chunks that are going to be sent.
     * 
     * @param name  name of the PDX Sequence Start Parameter
     * @param dict  a PDXDictionary containing all the dictionary contents
     * 
     * @return Parameter
     */
    public Parameter createPDXSeqStartParam(String name, Dictionary dict);
    
    /**
     * Creates a PDX Sequence Chunk Parameter. PDX Sequence Chunk is used to send
     * large amounts of data in dictionary format (like a contact list in PDXDictionary
     * format) in smaller pieces (chunks).
     * 
     * @param name name of the PDX Sequence Chunk Parameter
     * @param dict PDXDictionary that contains all the dictionary contents
     * 
     * @return Parameter
     */
    public Parameter createPDXSeqChunkParam(String name, Dictionary dict);
    
    /**
     * Creates a PDX Sequence End Parameter, which indicates 
     * the end of a series of data chunks that were sent.
     * 
     * @param name  name of the PDX Sequence End Parameter
     * @param dict  PDXDictionary containing all the dictionary contents
     * 
     * @return Parameter
     */
    public Parameter createPDXSeqEndParam(String name, Dictionary dict);
    
    /**
     * Creates a parameter for the TTS text and pushes the generated audio
     * buffers to the audio sink object. You can hook the synthesized audio to the 
     * Player object, which lets the Player play synthesized audio samples as they 
     * are streamed to the mobile device. To
     * hook the synthesized audio to the Player object, pass the Player object as the 
     * {@link NMSPAudioSink}.
     * 
     * For example: 
     * <pre><code>
     *     Player player = mgr.createPlayer(...);
     *     Parameter ttsparam = NMSPPDXManager.createPDXTTSParam(<i>name, tts_dict, player</i>);
     *     cmd.sendParam(ttsparam);
     *     cmd.end;
     * </code></pre>
     * While it is receiving the synthesized audio buffers from the server, the library streams the audio buffers 
     * to the associated Player for playback.
     * 
     * For how to configure the TTS dictionary, please consult the NMAS (Nuance Mobile Application
     * Server) document. One example is shown below the the "text" type TTS request.
     * 
     * <pre><code>
     * {
     *     audio_id  : 123,
     *     tts_type  : "text", "uri", "ssml" or html
     *     tts_input : "HELLO, My name is Angelina Jolie"(Text content)
     *                 or "\<ssml\> ... \</ssml\>" (SSML Content)
     *                 or "\<html\> ... \</html\>" (HTML Content)
     * }
     * </code></pre>
     * 
     * @param name        name of the PDX TTS parameter
     * @param tts_dict    dictionary contains the tts configuration and content to be synthesized
     * @param audioSink   NMSPAudioSink object to which the synthesized buffers are sent
     * 
     * @return TTSParam
     */
    public Parameter createPDXTTSParam(String name, Dictionary tts_dict, NMSPAudioSink audioSink);
    
    /**
     * Creates a PDX Dictionary, which can contain six types of data:
     * integer, byte, UTF8, ASCII, Dictionary, or Sequence.
     * 
     * @return Dictionary
     */
    public Dictionary createPDXDictionary();
    
    /**
     * Creates a PDX Sequence, which contains a list of elements. Can  
     * can be any of these six data types:
     * integer, byte string, UTF8 string, ASCII string, Dictionary, or Sequence.
     * 
     * @return Sequence
     */
    public Sequence createPDXSequence();
}
