package com.nuance.nmsp.client.sdk.oem.bluetooth;

import android.bluetooth.BluetoothDevice;
import android.content.Context;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;

/**
 * Reflection-based wrapper for the BluetoothHeadset API.
 */
public class BluetoothHeadset {

    public interface ServiceListener {
        public void onServiceConnected();
        public void onServiceDisconnected();
    }

    private static final Constructor<?> sConstructor;
    private static final Method sGetState;
    private static final Method sStartVoiceRecognition;
    private static final Method sStopVoiceRecognition;
    private static final Method sGetCurrentHeadset;
    private static final Method sDisconnectHeadset;
    private static final Method sConnectHeadset;
    private static final Method sClose;
    private static final Class<?> sServiceListenerClass;
    private static final boolean sGetStateRequiresDevice;
    private static final Method sGetPriority;
    private static final Method sSetPriority;

    private static final String ON_SERVICE_CONNECTED = "onServiceConnected";
    private static final String ON_SERVICE_DISCONNECTED = "onServiceDisconnected";

    public static final String ACTION_STATE_CHANGED;
    public static final String ACTION_AUDIO_STATE_CHANGED;
    public static final String EXTRA_STATE;
    public static final String EXTRA_AUDIO_STATE;

    /** There was an error trying to obtain the state */
    public static final int STATE_ERROR;
    /** No headset currently connected */
    public static final int STATE_DISCONNECTED;
    /** Connection attempt in progress */
    public static final int STATE_CONNECTING;
    /** A headset is currently connected */
    public static final int STATE_CONNECTED;

    /** A SCO audio channel is not established */
    public static final int AUDIO_STATE_DISCONNECTED;
    /** A SCO audio channel is established */
    public static final int AUDIO_STATE_CONNECTED;

    /** Default priority for headsets that  for which we will accept
     * inconing connections and auto-connect */
    //public static final int PRIORITY_AUTO_CONNECT;  // 1000
    /** Default priority for headsets that  for which we will accept
     * inconing connections but not auto-connect */
    //public static final int PRIORITY_ON;            // 100
    /** Default priority for headsets that should not be auto-connected
     * and not allow incoming connections. */
    public static final int PRIORITY_OFF;           // 0
    /** Default priority when not set or when the device is unpaired */
    public static final int PRIORITY_UNDEFINED;     // -1

    static {
        Reflection r = new Reflection();
        Class<?> cls = r.getClassForName("android.bluetooth.BluetoothHeadset");
        Class<?> clsListener = r.getClassForName("android.bluetooth.BluetoothHeadset$ServiceListener");

        Method tmpGetState;
        Method tmpDisconnect;
        boolean tmpGetStateRequiresDevice = false;

        sConstructor = r.getConstructor(cls, Context.class, clsListener);

        // since sGetState is 'final', below finds valid getState method in local
        //  and then assigns it to sGetState.
        tmpGetState = r.getMethodOrNull(cls, "getState");
        if(tmpGetState == null) {
            tmpGetState = r.getMethod(cls, "getState", BluetoothDevice.class);
            tmpGetStateRequiresDevice = true;
        }
        sGetState = tmpGetState;
        sGetStateRequiresDevice = tmpGetStateRequiresDevice;

        sGetCurrentHeadset = r.getMethod(cls, "getCurrentHeadset");
        sConnectHeadset = r.getMethod(cls, "connectHeadset", BluetoothDevice.class);

        tmpDisconnect = r.getMethodOrNull(cls, "disconnectHeadset");
        if(tmpDisconnect == null) {
            tmpDisconnect = r.getMethod(cls, "disconnectHeadset", BluetoothDevice.class);
        }
        sDisconnectHeadset = tmpDisconnect;

        sStartVoiceRecognition = r.getMethod(cls, "startVoiceRecognition");
        sStopVoiceRecognition = r.getMethod(cls, "stopVoiceRecognition");
        sClose = r.getMethod(cls, "close");

        sGetPriority = r.getMethod(cls, "getPriority", BluetoothDevice.class);
        sSetPriority = r.getMethod(cls, "setPriority", BluetoothDevice.class, int.class);

        sServiceListenerClass = clsListener;

        ACTION_STATE_CHANGED = (String) r.getFieldValue(cls, "ACTION_STATE_CHANGED");
        ACTION_AUDIO_STATE_CHANGED = (String) r.getFieldValue(cls, "ACTION_AUDIO_STATE_CHANGED");
        EXTRA_STATE = (String) r.getFieldValue(cls, "EXTRA_STATE");
        EXTRA_AUDIO_STATE = (String) r.getFieldValue(cls, "EXTRA_AUDIO_STATE");

        STATE_ERROR = (Integer) r.getFieldValue(cls, "STATE_ERROR");
        STATE_DISCONNECTED = (Integer) r.getFieldValue(cls, "STATE_DISCONNECTED");
        STATE_CONNECTING = (Integer) r.getFieldValue(cls, "STATE_CONNECTING");
        STATE_CONNECTED = (Integer) r.getFieldValue(cls, "STATE_CONNECTED");

        AUDIO_STATE_DISCONNECTED = (Integer) r.getFieldValue(cls, "AUDIO_STATE_DISCONNECTED");
        AUDIO_STATE_CONNECTED = (Integer) r.getFieldValue(cls, "AUDIO_STATE_CONNECTED");      

        // Commented out PRIORITY_AUTO_CONNECT and PRIORITY_ON since they don't exist in Android 2.1
        //PRIORITY_AUTO_CONNECT = (Integer)r.getFieldValue(cls, "PRIORITY_AUTO_CONNECT", 1000);
        //PRIORITY_ON           = (Integer)r.getFieldValue(cls, "PRIORITY_ON", 100);
        PRIORITY_OFF          = (Integer)r.getFieldValue(cls, "PRIORITY_OFF");
        PRIORITY_UNDEFINED    = (Integer)r.getFieldValue(cls, "PRIORITY_UNDEFINED", -1);
    }

    private Object mBluetoothHeadset;
    private ServiceListener mServiceListener;

    public BluetoothHeadset(Context context, ServiceListener listener) {
        mServiceListener = listener;
        Object serviceListenerProxy = Proxy.newProxyInstance(this.getClass().getClassLoader(),
                new Class<?>[] {
                    sServiceListenerClass
                }, new InvocationHandler() {
                    public Object invoke(Object proxy, Method method, Object[] args)
                            throws Throwable {
                        if (method.getName().equals(ON_SERVICE_CONNECTED)) {
                            if (mServiceListener != null)
                                mServiceListener.onServiceConnected();
                        } else if (method.getName().equals(ON_SERVICE_DISCONNECTED)) {
                            if (mServiceListener != null)
                                mServiceListener.onServiceDisconnected();
                        }
                        return null;
                    }
                });

        try {
            mBluetoothHeadset = sConstructor.newInstance(context, serviceListenerProxy);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public boolean connectHeadset(BluetoothDevice device) {
        boolean status = false;

        try {
            if(sConnectHeadset != null) {
                status = (Boolean)sConnectHeadset.invoke(mBluetoothHeadset, device);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }

        return status;
    }

    public void disconnectHeadset() {
        try {
            if(sDisconnectHeadset != null) {
                sDisconnectHeadset.invoke(mBluetoothHeadset);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }
    }

    public void disconnectHeadset(BluetoothDevice device) {
        try {
            if(sDisconnectHeadset != null) {
                sDisconnectHeadset.invoke(mBluetoothHeadset, device);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }
    }

    public BluetoothDevice getConnectedDevice() {
        BluetoothDevice device = null;

        try {
            if(sGetCurrentHeadset != null) {
                device = (BluetoothDevice)sGetCurrentHeadset.invoke(mBluetoothHeadset);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }

        return device;
    }

    public int getState() {
        try {
            if(sGetStateRequiresDevice) {
                Object current = sGetCurrentHeadset.invoke(mBluetoothHeadset);
                return(Integer) sGetState.invoke(mBluetoothHeadset, current);
            } else {
                return(Integer) sGetState.invoke(mBluetoothHeadset);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }

        return 0;
    }

    public boolean startVoiceRecognition() {
        try {
            return(Boolean) sStartVoiceRecognition.invoke(mBluetoothHeadset);
        } catch(Exception e) {
            e.printStackTrace();
        }
        return false;
    }

    public boolean stopVoiceRecognition() {
        try {
            return(Boolean) sStopVoiceRecognition.invoke(mBluetoothHeadset);
        } catch(Exception e) {
            e.printStackTrace();
        }
        return false;
    }

    public void close() {
        try {
            sClose.invoke(mBluetoothHeadset);
        } catch(Exception e) {
            e.printStackTrace();
        }
    }

    public int getPriority(BluetoothDevice device) {
        int priority = PRIORITY_UNDEFINED;

        try {
            if(sGetPriority != null) {
                priority = (Integer)sGetPriority.invoke(mBluetoothHeadset, device);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }

        return priority;
    }

    public boolean setPriority(BluetoothDevice device, int priority) {
        boolean status = false;
        try {
            if(sSetPriority != null) {
                status = (Boolean)sSetPriority.invoke(mBluetoothHeadset, device, priority);
            }
        } catch(Exception e) {
            e.printStackTrace();
        }

        return status;
    }

}
