package com.nuance.nmsp.client2.sdk.components.core.calllog;

/**
 * CalllogSender manages sending the serialized call logs to the server via its own Channel.
 * The binary call logs data can be obtained from the 
 * {@link com.nuance.nmsp.client.sdk.components.resource.common.ManagerListener#callLogDataGenerated(byte[])}
 * method.
 * 
 * There is a listener {@link SenderListener} manages the event notifications for the
 * call log transmission over the network to the server. Once the submission is done, no matter success
 * or failure, the SenderListener will be called with the status code. The application can manage which
 * data buffer has been sent.
 * 
 * @see {@link com.nuance.nmsp.client.sdk.components.resource.common.ManagerListener#callLogDataGenerated(byte[])}
 */
public interface CalllogSender
{
	/*!< Indicates there is a network error */
	public static final short CALLLOG_NETWORK_ERROR = 0x02;

	/*!< Indicates the call log data to be sent is invalid */
	public static final short CALLLOG_INVALID_DATA  = 0x03;
	
	/*!< Indicates the connection attempt failed */
	public static final short CALLLOG_CONNECTION_FAILED = 0x04;

    /**
     * The Listener to monitor the call log submissions.
     */
    public static interface SenderListener
    {
        /**
         * The call log data associated with the current listener has been
         * sent to the server successfully. The application can drop the
         * associated binary data.
         * 
         * @param data   The call log data being sent to the server successfully.
         *               Must not be null.
         */
        void succeeded(byte[] data);

        /**
         * The call log data associated with the current listener failed to
         * be sent to the server.The remaining binary data has been returned
         * and the application. Application should manage to send them again
         * via the method {@link CalllogSender#send(byte[])}.
         * 
         * @param errorCode  Error code.
         * @param data       The remaining call log binary data which failed
         *                   to be sent to the server. Must not be null.
         */
        void failed(short errorCode, byte[] data);
    }
    
    /**
     * Send the binary call log data to the server. 
     * 
     * This is used to send the binary call log data obtained from the
     * {@link com.nuance.nmsp.client.sdk.components.resource.common.ManagerListener#callLogDataGenerated(byte[])}.
     * The data has to be exactly the same as when it is returned by the SDK and can not be modified.
     * 
     * @param data      The binary call log data to be sent.
     * 
     * @see {@link com.nuance.nmsp.client.sdk.components.resource.common.ManagerListener#shutdownCompleted(byte[])}
     */
    public void send(byte[] data, String uid, String accessToken);
    
    /**
     * Shuts down the current call log sender and disconnect from gateway if the
     * current call log sender has a live connection with gateway.
     */
    public void shutdown();
}
