package com.nuance.nmsp.client2.sdk.components.core.internal.calllog;

import java.nio.ByteBuffer;
import java.util.LinkedList;
import java.util.List;
import java.util.UUID;
import java.util.Vector;

//#ifdef NMSP_LOG
import com.nuance.nmsp.client2.sdk.common.oem.api.LogFactory;
import com.nuance.nmsp.client2.sdk.common.oem.api.LogFactory.Log;
//#endif

import com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines;
import com.nuance.nmsp.client2.sdk.common.oem.api.MessageSystem;
import com.nuance.nmsp.client2.sdk.common.oem.api.MiscSystem;
import com.nuance.nmsp.client2.sdk.common.oem.api.MessageSystem.MessageData;
import com.nuance.nmsp.client2.sdk.common.oem.api.MessageSystem.MessageHandler;
import com.nuance.nmsp.client2.sdk.common.util.ByteConversion;
import com.nuance.nmsp.client2.sdk.common.util.Checker;
import com.nuance.nmsp.client2.sdk.components.core.calllog.SessionEvent;
import com.nuance.nmsp.client2.sdk.components.core.calllog.SessionEventAlreadyCommittedException;
import com.nuance.nmsp.client2.sdk.components.core.internal.pdx.PDXDictionary;
import com.nuance.nmsp.client2.sdk.components.core.internal.pdx.PDXSequence;
import com.nuance.nmsp.client2.sdk.components.core.pdx.Dictionary;
import com.nuance.nmsp.client2.sdk.components.core.pdx.Sequence;
import com.nuance.nmsp.client2.sdk.components.general.Parameter;
import com.nuance.nmsp.client2.sdk.components.general.ResourceUtil;

public class CalllogImpl implements MessageHandler
{
    //#ifdef NMSP_LOG
    private static final Log log = LogFactory.getLog(CalllogImpl.class);
    //#endif

    public static class SessionEventData {
        public int           requestId;
        public String        sessionEventId; 
        public byte[]        data;
    };
    
    private static final String NMSP_CALLLOG_SCHEMA_VERSION       = "2.0";

    // default data retention days should be -1, to keep the data on the server forever
    public  static final int  CALLLOG_RETENTION_DAYS_DEFAULT  = -1;
    
    public  static final int  CALLLOG_CHUNK_SIZE_DEFAULT  = 1000 * 100;
    public  static final int  CALLLOG_CHUNK_SIZE_MINIMUM  = 1000 * 20;
    public  static final int  CALLLOG_CHUNK_SIZE_MAXIMUM  = 1024 * 1024 * 2;
    
    private static final byte CMD_ADD_SESSION_EVENT    = 0x10; // be different from super class commands
    //private static final byte CMD_LOAD_FILE            = 0x11;
    //private static final byte CMD_BCP_LOG              = 0x12;
    private static final byte CMD_BCP_LOG_RESPONSE     = 0x13;
    private static final byte CMD_FLUSH_LOGS           = 0x15;
    private static final byte CMD_STOP_LOGS            = 0x16;
    
    
    public static final String 	INTERNAL_SHA1_KEY      = "b84798d890d847063ac0145b627a1ecfe4538006";
    //public static final int	 	INTERNAL_MAGIC_NUMBER  = 0x4b934b1;
    
    private final List        _events                  = new LinkedList();

    private long              _totalSize               = 0;
    private SessionEvent      _rootSessionEvent;
    private boolean           _disposed                = false;
    private MessageSystem     _messageSystem;
    
    private byte[]            _serverConfig            = null;
    private int               _chunkSize               = 0;
    private CalllogListener   _listener                = null;
    private List 			  _sessionEventDictList    = new LinkedList();
    
    public CalllogImpl(String gatewayIP, 
            short gatewayPort, 
            String applicationId, 
            int chunkSize,
            int retentionDays, 
            String rootId, 
            Vector params,
            MessageSystem messageSystem,
            CalllogListener listener)
    {
        
        serializeServerConfig(gatewayIP, gatewayPort, applicationId, params);
        
        _chunkSize = chunkSize;
        _messageSystem = messageSystem;
        _listener = listener;
        
        if(rootId == null){
            rootId = UUID.randomUUID().toString();
        }

        try {
            _rootSessionEvent = SessionEventBuilderImpl.createRootSessionEventBuilder(
                                                                    rootId, 
                                                                    SessionEvent.NMSP_CALLLOG_NMSPSDK_EVENT,
                                                                    NMSP_CALLLOG_SCHEMA_VERSION,
                                                                    retentionDays,
                                                                    this, 
                                                                    SessionEvent.NMSP_CALLLOG_NMSPSDK_EVENT)
                                                                    .putString("Nmaid", applicationId)
                                                                    .putString("client_hardware_model", ResourceUtil.getDeviceModel())
                                                                    .putString("client_hardware_submodel", ResourceUtil.getDeviceSubModel())
                                                                    .putString("client_os_type", ResourceUtil.getOSType())
                                                                    .putString("client_os_version", ResourceUtil.getOSVersion())
                                                                    .putString("client_sdk_release", NMSPDefines.VERSION)
                                                                    .putString("Origin", SessionEvent.NMSP_CALLLOG_ORIGIN_CLIENT)
                                                                    .commit();
        } catch (SessionEventAlreadyCommittedException e) {
            
        }
        

    }
    
    public SessionEvent getRootSessionEvent()
    {
        return _rootSessionEvent;
    }

    // sdk thread
    public void dispose()
    {
        //#ifdef NMSP_LOG
        if (log.isDebugEnabled())
            log.debug("onSessionDisconnected()");
        //#endif
        
        _disposed = true;
    }

    
    void addSessionEvent(SessionEventImpl sessionEvent)
    {
        if (_disposed)
            return;
        _messageSystem.send(new MessageData(CMD_ADD_SESSION_EVENT, new Object[] {sessionEvent}), this,
            Thread.currentThread(), _messageSystem.getVRAddr()[0]);
    }
    
    private void serializeServerConfig(String   server, 
                                      short     port, 
                                      String    appId, 
                                      List      sslParams)
    {
        int dataSize = 0;
        // Calculate the byte array size needed
        byte[] gatewayIpAsBytes     = server.getBytes();
        byte[] gatewayPortAsBytes     = new byte[2];
        byte[] appIdAsBytes         = appId.getBytes();
        byte[] sslParamCount        = new byte[1];
        // Convert port short to byte array
        gatewayPortAsBytes[0] = (byte)(port & 0xff);
        gatewayPortAsBytes[1] = (byte)((port >> 8) & 0xff);
        
        // set the number of ssl parameter we want to serialize
        if(sslParams == null){
        	sslParamCount[0] = 0;
        }else{
        	sslParamCount[0] = (byte) sslParams.size();
        }
        dataSize += 4; // sslParamCount size
        dataSize += 1; // sslParamCount value
                
        dataSize += gatewayIpAsBytes.length;
        dataSize += 4;
        dataSize += gatewayPortAsBytes.length;
        dataSize += 4;
        dataSize += appIdAsBytes.length;
        dataSize += 4;
       
        for (int i = 0; i < sslParamCount[0]; i++)
        {
            if(!(sslParams.get(i) instanceof Parameter)){
                return;
            }
            Parameter p = (Parameter) sslParams.get(i);
            
            dataSize += 4; // name length 
            dataSize += p.getName().length();
            dataSize += 4; // Parameter type
            dataSize += 2; // Parameter value
            dataSize += 4; // value length 
            dataSize += p.getValue().length;
        }
        
        ByteBuffer returnArray = ByteBuffer.allocate(dataSize); 
        
        // This serializes the gateway properties
        try{
            serializeByteArray(gatewayIpAsBytes, returnArray);
            serializeByteArray(gatewayPortAsBytes, returnArray);
            serializeByteArray(appIdAsBytes, returnArray);
        }catch(Exception e)
        {
            returnArray.clear();
            //#ifdef NMSP_LOG
                if (log.isErrorEnabled())
                    log.error("serializeEvents() failed : " + e.getMessage());
            //#endif
        }   

        // This serializes the SSL parameters
        serializeByteArray(sslParamCount, returnArray);
        for(int i = 0; i < sslParamCount[0]; i++)
        {
            Parameter p = (Parameter) sslParams.get(i);
            try
            {
                byte[] typeAsByte = new byte[2];
                typeAsByte[0] = (byte)(p.getType().getValue() & 0xff);
                typeAsByte[1] = (byte)((p.getType().getValue() >> 8) & 0xff);
                
                // This serializes the name of the parameter
                serializeByteArray(p.getName().getBytes(), returnArray);
                serializeByteArray(typeAsByte, returnArray);
                serializeByteArray(p.getValue(), returnArray);
            }
            catch(Exception e)
            {
                returnArray.clear();
                //#ifdef NMSP_LOG
                    if (log.isErrorEnabled())
                        log.error("serializeEvents() failed : " + e.getMessage());
                //#endif
            }            
        }
        _serverConfig = returnArray.array();
    }

    private byte[] serializeEvents(SessionEventData data)
    {

        byte[] bufferSize            = new byte[4];
        
        int dataSize = 0;
        
        dataSize += 4; // Size of the whole buffer
        dataSize += 4; // Content of the size of the whole buffer
        dataSize += 20; // SHA1 hash length
        dataSize += _serverConfig.length;

        dataSize += 4; // data.length size
        dataSize += data.data.length;
        dataSize += 4; // Session Id size
        dataSize += data.sessionEventId.getBytes().length;
        
        ByteBuffer returnArray = ByteBuffer.allocate(dataSize); 
        
        ByteConversion.intToBytes(dataSize,bufferSize, 0);
        serializeByteArray(bufferSize, returnArray);
        // Make room for the hash
        returnArray.put(new byte[20]);
        //byte[] magicNumber = new byte[4];
        //ByteConversion.intToBytes(INTERNAL_MAGIC_NUMBER, magicNumber, 0);
        //returnArray.put(magicNumber);
        // Special method because we do not want to encode the _serverConfig size
        serializeByteArray(_serverConfig, returnArray, false);

        if (data.data.length == 0){
            return null;
        }

        try
        {
            serializeByteArray(data.data, returnArray);
            serializeByteArray(data.sessionEventId.getBytes(), returnArray);                
        }
        catch(Exception e)
        {
            returnArray.clear();
            //#ifdef NMSP_LOG
            if (log.isErrorEnabled())
                log.error("serializeEvents() failed : " + e.getMessage());
            //#endif
        }

        
        // Compute the hash from that data and copy it back to the whole buffer
        byte[] hashedSessionId = MiscSystem.HMAC_SHA1(INTERNAL_SHA1_KEY.getBytes(), returnArray.array(), 28);
        byte[] finalArray = returnArray.array();
        System.arraycopy(hashedSessionId, 0, finalArray, 8, hashedSessionId.length);
        return finalArray;
    }
    
    private void serializeByteArray(byte[] bytes, ByteBuffer buffer)
    {
        serializeByteArray(bytes, buffer, true);
    }
    
    private void serializeByteArray(byte[] bytes, ByteBuffer buffer, boolean encodeSize)
    {
        if(encodeSize){
            byte[][] bytesSize = encodeSize(bytes.length);
            //buffer.put(bytesSize[0], 0, bytesSize[0].length);
            buffer.put(bytesSize[1], 0, bytesSize[1].length);
        }
        buffer.put(bytes, 0, bytes.length);
    }

    private byte[][] encodeSize(int size)
    {
        byte[][] ret = new byte[2][];
        byte[] sizeOfSize = new byte[1];
        byte[] v;

        sizeOfSize[0] = 4;
        v = new byte[4];
        v[0] = (byte) (size & 0xFF);
        v[1] = (byte) ((size >> 8) & 0xFF);
        v[2] = (byte) ((size >> 16) & 0xFF);
        v[3] = (byte) ((size >> 24) & 0xFF);
        
        ret[0] = sizeOfSize;
        ret[1] = v;
        return ret;
    }
    
    private void handleAddSessionEvent(SessionEventImpl sessionEvent)
    {
        SessionEventData eventData = new SessionEventData();
        
        sessionEvent.generateSeqIds();
        
        Dictionary sessionEventDict = sessionEvent.getDictionaryFormat();
        _sessionEventDictList.add(sessionEventDict);
        
        
        eventData.data = sessionEvent.getBinaryFormat();
        eventData.sessionEventId = sessionEvent._rootParent.getRootParentId() + sessionEvent._sequenceId;
        
        //#ifdef NMSP_LOG
    	    if (eventData.data.length > CALLLOG_CHUNK_SIZE_MAXIMUM){
    	        log.error("Trying to add new event larger than : " + CALLLOG_CHUNK_SIZE_MAXIMUM);
                return;
            }
    	//#endif
        
        if (4 +    // Size of the whole buffer
                _totalSize + 
                4 + // Size of the _serverConfig length 
                _serverConfig.length +
                4 + // Size of the data.length
                eventData.data.length +  
                4 + // Size of the sessionId length
                eventData.sessionEventId.getBytes().length
                > _chunkSize) {
                handleFlushLogs(false);
        }
        //save sessionEventId
        _totalSize += eventData.data.length;
        _totalSize += 4; // data.length size
        _totalSize += eventData.sessionEventId.getBytes().length;
        _totalSize += 4; // sessionEventId.getBytes().length size
    }

    public void genSeqId(SessionEventImpl sessionEvent) 
    {
        sessionEvent.genSeqId();
    }
    
    public void handleMessage(Object msgData, Object msgSender)
    {
        MessageData messageData = (MessageData) msgData;
        Object[] params;
        
        switch (messageData.command)
        {
            case CMD_ADD_SESSION_EVENT:
                params = (Object[]) messageData.data;
                handleAddSessionEvent((SessionEventImpl) params[0]);
                break;
            case CMD_BCP_LOG_RESPONSE:
                break;
            case CMD_FLUSH_LOGS:
            	params = (Object[]) messageData.data;
                handleFlushLogs(((Boolean) params[0]).booleanValue());
                break;
            case CMD_STOP_LOGS:
            	handleStopLogs();
            	break;
            default:
//                super.handleMessage(msgData, msgSender);
        }
    }
    
    private void handleFlushLogs(boolean isFromClient)
    {
        // if we are trying to flush logs but there are no logs to flush
        if(_sessionEventDictList.size() == 0){
        	if (isFromClient)
        	{
        		// output null to indicate the client that there is
        		// no data to flush
        		_listener.calllogDataGenerated(null, null);
        	}
            return;
        }
        
        // This is to handle multiple logs in a single event
    	PDXDictionary batchDict = new PDXDictionary();
    	Sequence batchSeq = new PDXSequence();
    	List sessionEventIds = new LinkedList();
    	int i = 0;
        while (_sessionEventDictList.size() > 0)
        {
        	Dictionary dict = (Dictionary) _sessionEventDictList.remove(0);
        	batchSeq.addDictionary(dict);
        	
        	Sequence metaSeq = dict.getSequence("meta");
        	String rpi = "";
        	int seqId = 0;
        	
        	String sessionEventName = "";
        	
        	for ( int j = 0; j < metaSeq.size() ; ++j )
        	{
        		Dictionary metaDict = metaSeq.getDictionary(j);
        		String key = metaDict.getAsciiString("k");
        		
        		if (key.equals(SessionEventImpl.NMSP_CALLLOG_META_ROOT_PARENT_ID))
        		{
        			rpi = metaDict.getUTF8String("v");
        		}
        		else if (key.equals(SessionEventImpl.NMSP_CALLLOG_META_SEQUENCE_ID))
        		{
        			seqId = metaDict.getInteger("v");
        		}
                else if (key.equals(SessionEventImpl.NMSP_CALLLOG_META_NAME))
                {
                    sessionEventName = metaDict.getUTF8String("v"); 
                }
        	}
        	
        	if ( isFromClient == true && i == 0 )
        	{
        	    sessionEventIds.add("###flushed###:" + sessionEventName + ":" + rpi + "." + seqId);
        	}
        	else
        	{
        	    sessionEventIds.add(sessionEventName + ":" + rpi + "." + seqId);
        	}
        	
        	i++;
        }
        batchDict.addSequence("events", batchSeq);
        
    	SessionEventData data = new SessionEventData();
    	data.sessionEventId = _rootSessionEvent.getId();
    	data.data = batchDict.toByteArray();
    	
        byte[] bytesToSend = null;
        bytesToSend = serializeEvents(data);
        //#ifdef NMSP_LOG
                if (bytesToSend == null)
                    log.error("CalllogImpl.close serialized data is null");
        //#endif
        
        _totalSize = 0;
        _listener.calllogDataGenerated(bytesToSend, sessionEventIds);        
    }

    public void flushLogsToListener(boolean isFromClient)
    {
        _messageSystem.send(new MessageData(CMD_FLUSH_LOGS, new Object[] {new Boolean(isFromClient)}), this,
                Thread.currentThread(), _messageSystem.getVRAddr()[0]);
    }
    
    private void handleStopLogs()
    {
        _listener.calllogDataGenerated(null, null);   
    }
    
    public void stopCalllog()
    {
        _messageSystem.send(new MessageData(CMD_STOP_LOGS, null), this,
                Thread.currentThread(), _messageSystem.getVRAddr()[0]);    	
    }

    public String getRootParentId()
    {
        return ((RootSessionEventImpl)_rootSessionEvent).getRootParentId();
    }
}
