package com.nuance.nmsp.client2.sdk.components.core.internal.calllog;

import java.util.Hashtable;

import com.nuance.nmsp.client2.sdk.components.core.calllog.SessionEvent;
import com.nuance.nmsp.client2.sdk.components.core.calllog.SessionEventAlreadyCommittedException;
import com.nuance.nmsp.client2.sdk.components.core.calllog.SessionEventBuilder;


public class SessionEventBuilderImpl implements SessionEventBuilder
{
    private final SessionEventImpl _sessionEvent;
    private final Hashtable _table = new Hashtable();
    private boolean _committed = false;
    private final int APPROX_OVERHEAD_PER_PUT = 16;
    // This is to make sure that the resulting SessionEvent will not exceed the maximum
    // BCP_LOG message size defined in CalllogImpl. This is an approximation since the
    // exact calculation would require a lot of code duplication and is not needed here.
    private int _approxDataSize  = 0;

    static SessionEventBuilderImpl createRootSessionEventBuilder(String rootParentId, String applicationName, String schemaVersion, int retentionDays, CalllogImpl calllog, String eventName)
    {
        if (eventName == null) {
            throw new IllegalArgumentException("eventName is null.");
        }
        SessionEventImpl sessionEvent = new RootSessionEventImpl(rootParentId, applicationName, schemaVersion, retentionDays, calllog, eventName);
        return new SessionEventBuilderImpl(sessionEvent);
    }

    static SessionEventBuilderImpl createRegularSessionEventBuilder(SessionEventImpl parent, String eventName)
    {
		return createRegularSessionEventBuilder(parent, eventName, null);
    }
    
    static SessionEventBuilderImpl createRegularSessionEventBuilder(SessionEventImpl parent, String eventName, String schemaVersion)
    {
        if (eventName == null)
        {
            throw new IllegalArgumentException("eventName is null.");
        }
        
        SessionEventImpl sessionEvent = new RegularSessionEventImpl(parent, eventName, schemaVersion);
        return new SessionEventBuilderImpl(sessionEvent);
    }

    private SessionEventBuilderImpl(SessionEventImpl sessionEvent)
    {
        _sessionEvent = sessionEvent;
        
        CalllogImpl calllog = (CalllogImpl) _sessionEvent.getCalllog();
        calllog.genSeqId(sessionEvent);
    }
    
    //Override
    public SessionEventBuilder putBinary(String key, byte[] value) throws SessionEventAlreadyCommittedException
    {
        if (_committed) {
            throw new SessionEventAlreadyCommittedException("SessionEvent is already committed.");
        }
        
        int keyLength = key.getBytes().length;
        
        if(value.length + keyLength + _approxDataSize > CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM * 0.9){
        	throw new IllegalArgumentException("Trying to add too much data in putBinary. Maximum is : " + CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM * 0.9);
        }
        _approxDataSize += value.length + keyLength + APPROX_OVERHEAD_PER_PUT;

        _table.put(key, value);
        return this;
    }

    //Override
    public SessionEventBuilder putBoolean(String key, boolean value) throws SessionEventAlreadyCommittedException
    {
        if (_committed) {
            throw new SessionEventAlreadyCommittedException("SessionEvent is already committed.");
        }
        
        int keyLength = key.getBytes().length;
        int valueLength = Boolean.valueOf(value).toString().getBytes().length;
        
        if(valueLength + keyLength + _approxDataSize > CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM * 0.9){
        	throw new IllegalArgumentException("Trying to add too much data in putBinary. Maximum is : " + CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM * 0.9);
        }
        _approxDataSize += valueLength + keyLength + APPROX_OVERHEAD_PER_PUT;

    	_table.put(key, new Boolean(value));
        return this;
    }

    //Override
    public SessionEventBuilder putFloat(String key, double value) throws SessionEventAlreadyCommittedException
    {
        if (_committed) {
            throw new SessionEventAlreadyCommittedException("SessionEvent is already committed.");
        }
        
        int keyLength = key.getBytes().length;
        int valueLength = Double.valueOf(value).toString().getBytes().length;
        
        if(valueLength + keyLength + _approxDataSize > CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM * 0.9){
        	throw new IllegalArgumentException("Trying to add too much data in putBinary. Maximum is : " + CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM * 0.9);
        }
        _approxDataSize += valueLength + keyLength + APPROX_OVERHEAD_PER_PUT;

    	_table.put(key, new Double(value));
        return this;
    }

    //Override
    public SessionEventBuilder putInteger(String key, int value) throws SessionEventAlreadyCommittedException
    {
        if (_committed) {
            throw new SessionEventAlreadyCommittedException("SessionEvent is already committed.");
        }
        
        int keyLength = key.getBytes().length;
        int valueLength = 4;
        
        if(valueLength + keyLength + _approxDataSize > CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM * 0.9){
        	throw new IllegalArgumentException("Trying to add too much data in putBinary. Maximum is : " + CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM * 0.9);
        }
        _approxDataSize += valueLength + keyLength + APPROX_OVERHEAD_PER_PUT;

        _table.put(key, new Integer(value));
        return this;
    }

    //Override
    public SessionEventBuilder putLong(String key, long value) throws SessionEventAlreadyCommittedException
    {
        if (_committed) {
            throw new SessionEventAlreadyCommittedException("SessionEvent is already committed.");
        }
        
        int keyLength = key.getBytes().length;
        int valueLength = 4;
        
        if(valueLength + keyLength + _approxDataSize > CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM * 0.9){
        	throw new IllegalArgumentException("Trying to add too much data in putBinary. Maximum is : " + CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM * 0.9);
        }
        _approxDataSize += valueLength + keyLength + APPROX_OVERHEAD_PER_PUT;

    	_table.put(key, new Long(value));
        return this;
    }
    
    //Override
    public SessionEventBuilder putString(String key, String value) throws SessionEventAlreadyCommittedException
    {
        if (_committed) {
            throw new SessionEventAlreadyCommittedException("SessionEvent is already committed.");
        }

        int keyLength = key.getBytes().length;
        int valueLength = value.getBytes().length;
        
        if(valueLength + keyLength + _approxDataSize > CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM * 0.9){
        	throw new IllegalArgumentException("Trying to add too much data in putBinary. Maximum is : " + CalllogImpl.CALLLOG_CHUNK_SIZE_MAXIMUM * 0.9);
        }
        _approxDataSize += valueLength + keyLength + APPROX_OVERHEAD_PER_PUT;

    	_table.put(key, value);
        return this;
    }
    
    //Override
    public SessionEventBuilder createRemoteEvent(String eventName) throws SessionEventAlreadyCommittedException
    {
        if (eventName == null) {
            throw new IllegalArgumentException("eventName is null.");
        }
    	_sessionEvent.addRemoteEventRefid(eventName);
    	return this;
    }

    //Override
    public SessionEvent commit()
    {
        if (!_committed)
        {
            _sessionEvent.done(_table);
            _committed = true;
            return _sessionEvent;
        }
        else
        {
            return null;
        }
    }

    public SessionEvent cancel()
    {
        if (_committed == false)
        {
            _sessionEvent.done(_table);
            _committed = true;
        }
        return _sessionEvent;
    }
    
    // for cancel events in order to construct a "CancelLogEvent"
    public void setEventToCancel(String eventIdToCancel, String eventNameToCancel)
    {
        _sessionEvent._eventIdToCancel   = eventIdToCancel;
        _sessionEvent._eventNameToCancel = eventNameToCancel;
    }
}
