/**
 *  Copyright 2006-2011, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.nmsp.client2.sdk.components.core.internal.pdx;

import java.io.ByteArrayOutputStream;

//#ifdef NMSP_LOG2
//@import com.nuance.nmsp.client.sdk.common.oem.api.LogFactory;
//@import com.nuance.nmsp.client.sdk.common.oem.api.LogFactory.Log;
//#endif

//PDX Encoding Rules
//LOGICAL STRUCTURE  {
//      CLASS;
//      LENGTH;
//      unsigned char content[];
//    };
//
//For datum between 0 and 127 bytes long the packet should look like this.
//
//struct  {
//  unsigned byte class;
//  unsigned byte length_bits; // dec: 0 - 127, hex: 0 - 0x7F
//  unsigned char[] content
//};
//
//
//For datum between 128 and 255 bytes long the packet should look like this.
//
//struct {
//  unsigned byte class;
//  unsigned byte length_bits; // = 0x81
//  unsigned byte short_length; // dec: 0-255, 0 - 0xFF
//  unsigned char[] content;
//};
//
//
//For datum between 256 and 65535 bytes long the packet should look like this.
//
//struct {
//  unsigned byte class;
//  unsigned byte length_bits; // = 0x82
//  unsigned short long_length;
//  unsigned char[] content;
//};
//
//
//For datum between 65536 and 4294967295 bytes long the packet should look like this.
//
//struct {
//  unsigned byte class;
//  unsigned byte length_bits; // = 0x84
//  unsigned long very_long_length;
//  unsigned char[] content;
//};
//
//
//Name             CLASS Signature  Valid Lengths     Notes
//----------------------------------------------------
//Little Endian                                  We do not use natural BER integers (Class = 0x02) because they are encoded big-endian and we want to  
//INTEGER         0xC0              1, 2, 4         optimize for the most common case - little-endian server communicating with a little-endian handset.
//BYTE STRING     0x04              0-n             Arbitrary byte content
//UTF8-STRING     0xC1              0-n             UTF-8 encoded string
//ASCII-STRING     0x16              0-n             ASCII string
//NULL             0x05              0    
//SEQUENCE       0x10              0-n             Given that the set of objects may not be homogeneous, when decoding, the number of objects will not be known until you have decoded subsequent packets whose total size equals length.
//DICTIONARY     0xE0                             Contains alternating key value objects.  By convention, the Key component of the dictionary MUST be an ASCII-String.

public class PDXClass {
    //#ifdef NMSP_LOG2
//@    private static final Log log = LogFactory.getLog(PDXClass.class);
    //#endif

    private short CLASS;

    protected static final short BYTES = 0x04;                 //004
    protected static final short NULL = 0x05;                  //005
    protected static final short DOUBLE = 0x06;				   //006
    protected static final short BOOLEAN = 0x07;               //007
    protected static final short SEQ = 0x10;                   //016
    protected static final short ASCII = 0x16;                 //022
    protected static final short INT = (short) (0xC0 & 0xFF);  //192
    public static final short UTF8 = (short) (0xFF & 0xC1);    //193    
    protected static final short DICT = (short) (0xFF & 0xE0); //224
    
    public PDXClass(short c) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXClass() " + c);
        //#endif
        this.CLASS = c;
    }

    public short getType() {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXClass.getType() " + this.CLASS);
        //#endif
        return this.CLASS;
    }

    public int getLength(byte[] byteClass, int offset) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXClass.getLength()");
        //#endif
        int length_bits = (byteClass[offset] & 0xFF);
        if (length_bits == 0x81) {
            return (byteClass[offset + 1] & 0xFF);
        } else if (length_bits == 0x82) {
            return (((byteClass[offset + 1] & 0xFF) << 8)
                    + (byteClass[offset + 2] & 0xFF));
        } else if (length_bits == 0x84) {
            return (((byteClass[offset + 1] & 0xFF) << 24)
                    + ((byteClass[offset + 2] & 0xFF) << 16)
                    + ((byteClass[offset + 3] & 0xFF) << 8)
                    + (byteClass[offset + 4] & 0xFF));
        } else {
            return length_bits;
        }
    }

    public int getLengthSize(int length) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXClass.getLengthSize()");
        //#endif
        if (length <= 127) {
            return 1;
        } else if (length <= 255) {
            return 2;
        } else if (length <= 65535) {
            return 3;
        } else {
            return 5;
        }
    }

    public byte[] toByteArray(byte[] content) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXClass.toByteArray()");
        //#endif
        ByteArrayOutputStream pdxer = new ByteArrayOutputStream();
        pdxer.write(this.CLASS & 0xFF);
        if (content.length > 65535) {
            pdxer.write(0x84);
            pdxer.write((content.length >> 24) & 0xFF);
            pdxer.write((content.length >> 16) & 0xFF);
            pdxer.write((content.length >> 8) & 0xFF);
            pdxer.write(content.length & 0xFF);
        } else if (content.length > 255) {
            pdxer.write(0x82);
            pdxer.write((content.length >> 8) & 0xFF);
            pdxer.write(content.length & 0xFF);
        } else if (content.length > 127) {
            pdxer.write(0x81);
            pdxer.write(content.length);
        } else if (content.length >= 0) {
            pdxer.write(content.length);
        }

        pdxer.write(content, 0, content.length);
        return pdxer.toByteArray();
    }
}
