/**
 *  Copyright 2006-2011, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.nmsp.client2.sdk.components.core.internal.pdx;

import java.io.ByteArrayOutputStream;
import java.util.Enumeration;
import java.util.Vector;

//#ifdef NMSP_LOG
import com.nuance.nmsp.client2.sdk.common.oem.api.LogFactory;
import com.nuance.nmsp.client2.sdk.common.oem.api.LogFactory.Log;
//#endif
import com.nuance.nmsp.client2.sdk.components.core.pdx.Dictionary;
import com.nuance.nmsp.client2.sdk.components.core.pdx.Sequence;

/**
 * PDX sequence representation
 * 
 */
public class PDXSequence extends PDXClass implements Sequence {
  //#ifdef NMSP_LOG
    private static final Log log = LogFactory.getLog(PDXSequence.class);
    //#endif

    private Vector sequence;

    public PDXSequence() {
        super(PDXClass.SEQ);
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence()");
        //#endif
        sequence = new Vector();
    }

    public PDXSequence(byte[] byteSequence) {
        super(PDXClass.SEQ);
        //#ifdef NMSP_LOG3
//@        String inputHex = "";
//@        for(int i = 0; i < byteSequence.length; i++) {
//@            inputHex += Integer.toString( ( byteSequence[i] & 0xff ) + 0x100, 16).substring( 1 );
//@        }
//@        log.debug("PDXSequence() " + inputHex);
        //#endif
        sequence = new Vector();
        int index = 0;
        int classType;
        int length;
        byte[] classContent;

        while (index < byteSequence.length) {
            // Get the value class type
            classType = byteSequence[index++] & 0xFF;
            length = getLength(byteSequence, index);
            index += getLengthSize(length);

            classContent = new byte[length];
            System.arraycopy(byteSequence, index, classContent, 0, classContent.length);
            index += length;

            // Build the sequence by adding values
            switch (classType) {
            case PDXClass.BOOLEAN:
            	sequence.addElement(new PDXBoolean(classContent));
            	break;
            case PDXClass.DOUBLE:
            	sequence.addElement(new PDXDouble(classContent));
            	break;
            case PDXClass.INT:
                sequence.addElement(new PDXInteger(classContent));
                break;
            case PDXClass.BYTES:
                sequence.addElement(new PDXByteString(classContent));
                break;
            case PDXClass.UTF8:
                sequence.addElement(new PDXUTF8String(classContent));
                break;
            case PDXClass.ASCII:
                sequence.addElement(new PDXAsciiString(classContent));
                break;
            case PDXClass.NULL:
                sequence.addElement(new PDXNull());
                break;
            case PDXClass.DICT:
                sequence.addElement(new PDXDictionary(classContent));
                break;
            case PDXClass.SEQ:
                sequence.addElement(new PDXSequence(classContent));
                break;
            default:
                //#ifdef NMSP_LOG
                log.error("PDXSequence() Unknown PDXClass type: " + classType + ". ");
                //#endif
                break;
            }
        }
    }

    /**
     * Retrieves the number of elements in a sequence.
     * 
     * @return int
     */
    public int size() {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.size()");
        //#endif
        return sequence.size();
    }

    /**
     * Retrieves an Enumeration of elements in a sequence.
     * 
     * @return Enumeration
     */
    // public Enumeration elements()
    // {
    // return sequence.elements();
    // }
    /**
     * Adds a sequence element.
     * 
     * @param pdxClass
     *            element
     * 
     * @return
     */
    // public void add(PDXClass pdx)
    // {
    // sequence.addElement(pdx);
    // }
    
    public short getType(int index) {
        if (index >= sequence.size()) {
            //#ifdef NMSP_LOG
            log.error("PDXSequence.getType() index " + index + " is out of range. ");
            //#endif
            throw new IndexOutOfBoundsException();
        }
        
        PDXClass pdxClass = (PDXClass) sequence.elementAt(index);
        return pdxClass.getType();
    }
    
    /**
     * Retrieves a sequence boolean element.
     * 
     * @param index
     *            element index
     * 
     * @throws IndexOutOfBoundsException, RuntimeException
     * 
     * @return boolean
     */
    public boolean getBoolean(int index) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.getBoolean() " + index);
        //#endif
        if (index >= sequence.size()) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getBoolean() index " + index + " is out of range. ");
          //#endif
          throw new IndexOutOfBoundsException();
        }

        PDXClass pdxClass = (PDXClass) sequence.elementAt(index);
        if (pdxClass.getType() != PDXClass.BOOLEAN) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getBoolean() index " + index + " is not a PDXBoolean. ");
          //#endif
          throw new RuntimeException(Sequence.INDEX_IS_OF_WRONG_TYPE);
        }

        return ((PDXBoolean) pdxClass).getValue();
    }
    
    /**
     * Retrieves a sequence double element.
     * 
     * @param index
     *            element index
     * 
     * @throws IndexOutOfBoundsException, RuntimeException
     * 
     * @return double
     */
    public double getDouble(int index) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.getDouble() " + index);
        //#endif
        if (index >= sequence.size()) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getDouble() index " + index + " is out of range. ");
          //#endif
          throw new IndexOutOfBoundsException();
        }

        PDXClass pdxClass = (PDXClass) sequence.elementAt(index);
        if (pdxClass.getType() != PDXClass.DOUBLE) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getDouble() index " + index + " is not a PDXDouble. ");
          //#endif
          throw new RuntimeException(Sequence.INDEX_IS_OF_WRONG_TYPE);
        }

        return ((PDXDouble) pdxClass).getValue();
    }
    
    /**
     * Retrieves a sequence integer element.
     * 
     * @param index
     *            element index
     * 
     * @throws IndexOutOfBoundsException, RuntimeException
     * 
     * @return int
     */
    public int getInteger(int index) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.getInteger() " + index);
        //#endif
        if (index >= sequence.size()) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getInteger() index " + index + " is out of range. ");
          //#endif
          throw new IndexOutOfBoundsException();
        }

        PDXClass pdxClass = (PDXClass) sequence.elementAt(index);
        if (pdxClass.getType() != PDXClass.INT) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getInteger() index " + index + " is not a PDXInteger. ");
          //#endif
          throw new RuntimeException(Sequence.INDEX_IS_OF_WRONG_TYPE);
        }

        return ((PDXInteger) pdxClass).getValue();
    }

    /**
     * Retrieves a sequence byte string element.
     * 
     * @param index
     *            element index
     * 
     * @throws IndexOutOfBoundsException, RuntimeException
     * 
     * @return byte[]
     */
    public byte[] getByteString(int index) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.getByteString() " + index);
        //#endif
        if (index >= sequence.size()) {
            //#ifdef NMSP_LOG
            log.error("PDXSequence.getByteString() index " + index + " is out of range. ");
            //#endif
            throw new IndexOutOfBoundsException();
        }

        PDXClass pdxClass = (PDXClass) sequence.elementAt(index);
        if (pdxClass.getType() != PDXClass.BYTES) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getByteString() index " + index + " is not a PDXByteString. ");
          //#endif
          throw new RuntimeException(Sequence.INDEX_IS_OF_WRONG_TYPE);
        }

        return ((PDXByteString) pdxClass).getValue();
    }

    /**
     * Retrieves a sequence UTF8 string element.
     * 
     * @param index
     *            element index
     * 
     * @throws IndexOutOfBoundsException, RuntimeException
     * 
     * @return String
     */
    public String getUTF8String(int index) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.getUTF8String() " + index);
        //#endif
        if (index >= sequence.size()) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getUTF8String() index " + index + " is out of range. ");
          //#endif
          throw new IndexOutOfBoundsException();
        }

        PDXClass pdxClass = (PDXClass) sequence.elementAt(index);
        if (pdxClass.getType() != PDXClass.UTF8) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getUTF8String() index " + index + " is not a PDXUTF8String. ");
          //#endif
          throw new RuntimeException(Sequence.INDEX_IS_OF_WRONG_TYPE);
        }

        return ((PDXUTF8String) pdxClass).getValue();
    }

    /**
     * Retrieves a sequence ASCII string element.
     * 
     * @param index
     *            element index
     * 
     * @throws IndexOutOfBoundsException, RuntimeException
     * 
     * @return String
     */
    public String getAsciiString(int index) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.getAsciiString() " + index);
        //#endif
        if (index >= sequence.size()) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getAsciiString() index " + index + " is out of range. ");
          //#endif
          throw new IndexOutOfBoundsException();
        }

        PDXClass pdxClass = (PDXClass) sequence.elementAt(index);
        if (pdxClass.getType() != PDXClass.ASCII) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getAsciiString() index " + index + " is not a PDXAsciiString. ");
          //#endif
          throw new RuntimeException(Sequence.INDEX_IS_OF_WRONG_TYPE);
        }

        return ((PDXAsciiString) pdxClass).getValue();
    }

    /**
     * Retrieves a sequence Dictionary element.
     * 
     * @param index
     *            element index
     * 
     * @throws IndexOutOfBoundsException, RuntimeException
     * 
     * @return Dictionary
     */
    public Dictionary getDictionary(int index) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.getDictionary() " + index);
        //#endif
        if (index >= sequence.size()) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getDictionary() index " + index + " is out of range. ");
          //#endif
          throw new IndexOutOfBoundsException();
        }

        PDXClass pdxClass = (PDXClass) sequence.elementAt(index);
        if (pdxClass.getType() != PDXClass.DICT) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getDictionary() index " + index + " is not a PDXDictionary. ");
          //#endif
          throw new RuntimeException(Sequence.INDEX_IS_OF_WRONG_TYPE);
        }

        return ((PDXDictionary) pdxClass);
    }

    /**
     * Retrieves a sequence Sequence element.
     * 
     * @param index
     *            element index
     * 
     * @throws IndexOutOfBoundsException, RuntimeException
     * 
     * @return Sequence
     */
    public Sequence getSequence(int index) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.getSequence() " + index);
        //#endif
        if (index >= sequence.size()) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getSequence() index " + index + " is out of range. ");
          //#endif
          throw new IndexOutOfBoundsException();
        }

        PDXClass pdxClass = (PDXClass) sequence.elementAt(index);
        if (pdxClass.getType() != PDXClass.SEQ) {
          //#ifdef NMSP_LOG
          log.error("PDXSequence.getSequence() index " + index + " is not a PDXSequence. ");
          //#endif
          throw new RuntimeException(Sequence.INDEX_IS_OF_WRONG_TYPE);
        }

        return ((PDXSequence) pdxClass);
    }

    public void addBoolean(boolean value){
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.addDouble() " + value);
        //#endif
        PDXBoolean pdxBoolean = new PDXBoolean(value);
        sequence.addElement(pdxBoolean);		
    }
    
    public void addDouble(double value){
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.addDouble() " + value);
        //#endif
        PDXDouble pdxDouble = new PDXDouble(value);
        sequence.addElement(pdxDouble);	
    }
    
    public void addAsciiString(String value) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.addAsciiString() " + value);
        //#endif
    	if(value == null) throw new IllegalArgumentException("value is null.");
        PDXAsciiString string = new PDXAsciiString(value);
        sequence.addElement(string);
    }

    public void addByteString(byte[] value) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.addByteString() " + value);
        //#endif
        if(value == null) throw new IllegalArgumentException("value is null.");
        PDXByteString string = new PDXByteString(value);
        sequence.addElement(string);
    }

    public void addDictionary(Dictionary value) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.addDictionary() " + value);
        //#endif
        if(value == null) throw new IllegalArgumentException("value is null.");
        if (((PDXClass) value).getType() != PDXClass.DICT) {
            //#ifdef NMSP_LOG
            log.error("PDXSequence.addDictionary() value is not a valid dictionary.");
            //#endif
            throw new IllegalArgumentException("value is not a valid dictionary. ");
        }
        sequence.addElement(value);
    }

    public void addInteger(int value) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.addInteger() " + value);
        //#endif
        PDXInteger integer = new PDXInteger(value);
        sequence.addElement(integer);
    }

    public void addSequence(Sequence value) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.addSequence() " + value);
        //#endif
        if(value == null) throw new IllegalArgumentException("value is null.");
        if (((PDXClass) value).getType() != PDXClass.SEQ) {
            //#ifdef NMSP_LOG
            log.error("PDXSequence.addSequence() value is not a valid sequence.");
            //#endif
            throw new IllegalArgumentException("value is not a valid sequence. ");
        }
        sequence.addElement(value);
    }

    public void addUTF8String(String value) {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.addUTF8String() " + value);
        //#endif
        if(value == null) throw new IllegalArgumentException("value is null.");
        PDXUTF8String string = new PDXUTF8String(value);
        sequence.addElement(string);
    }

    public byte[] toByteArray() {
        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.toByteArray()");
        //#endif
        ByteArrayOutputStream output = new ByteArrayOutputStream();
        Enumeration elements = sequence.elements();
        while (elements.hasMoreElements()) {
            PDXClass pdxElement = (PDXClass) elements.nextElement();
            try {
                switch (pdxElement.getType()) {
                case PDXClass.BOOLEAN:
                	output.write(((PDXBoolean) pdxElement).toByteArray());
                	break;
                case PDXClass.DOUBLE:
                	output.write(((PDXDouble) pdxElement).toByteArray());
                	break;
                case PDXClass.INT:
                    output.write(((PDXInteger) pdxElement).toByteArray());
                    break;
                case PDXClass.BYTES:
                    output.write(((PDXByteString) pdxElement).toByteArray());
                    break;
                case PDXClass.UTF8:
                    output.write(((PDXUTF8String) pdxElement).toByteArray());
                    break;
                case PDXClass.ASCII:
                    output.write(((PDXAsciiString) pdxElement).toByteArray());
                    break;
                case PDXClass.NULL:
                    output.write(((PDXNull) pdxElement).toByteArray());
                    break;
                case PDXClass.DICT:
                    output.write(((PDXDictionary) pdxElement).toByteArray());
                    break;
                case PDXClass.SEQ:
                    output.write(((PDXSequence) pdxElement).toByteArray());
                    break;
                }
            } catch (java.io.IOException ioe) {
                //#ifdef NMSP_LOG
                log.error("PDXSequence.toByteArray() " + ioe.toString() + ". ");
                //#endif
            }
        }
        return super.toByteArray(output.toByteArray());
    }
    
    public String toString()
    {
        return this.toString(0);
    }
    
    protected String toString(int indentLevel)
    {
        String stringSequence = "";
        String baseIndent = "";
        String indent = "";
        for(int i = 0; i < indentLevel - 1; i++) {
            baseIndent += "    ";
        }
        if(indentLevel > 0) {
            indent = baseIndent + "    ";
        }
        
        Enumeration elements = sequence.elements();
        stringSequence +=  "[ \n";
        while(elements.hasMoreElements()) {
            PDXClass value = (PDXClass)elements.nextElement();
            switch (value.getType()) {
            case PDXClass.BOOLEAN:
            	stringSequence += indent + "<BOOLEAN> " + ((PDXBoolean)value).getValue();
            	break;
            case PDXClass.DOUBLE:
            	stringSequence += indent + "<DOUBLE> " + ((PDXDouble)value).getValue();
            	break;
            case PDXClass.INT:
                stringSequence += indent + "<INT> " + ((PDXInteger)value).getValue();
                break;
            case PDXClass.BYTES:
                stringSequence += indent + "<BYTES> \"" + ((PDXByteString)value).getValue() + "\" ";
                break;
            case PDXClass.UTF8:
                stringSequence += indent + "<UTF8> \"" + ((PDXUTF8String)value).getValue() + "\" ";
                break;
            case PDXClass.ASCII:
                stringSequence += indent + "<ASCII> \"" + ((PDXAsciiString)value).getValue() + "\" ";
                break;
            case PDXClass.NULL:
                stringSequence += indent + "<NULL> ";
                break;
            case PDXClass.DICT:
                stringSequence += indent + ((PDXDictionary)value).toString(indentLevel + 1);
                break;
            case PDXClass.SEQ:
                stringSequence += indent + ((PDXSequence)value).toString(indentLevel + 1);
                break;
            }
            if(elements.hasMoreElements()) {
                stringSequence += ",";
            }
            stringSequence += "\n";
        }
        stringSequence += baseIndent + "] ";

        //#ifdef NMSP_LOG2
//@        log.debug("PDXSequence.toString()\n" + stringSequence);
        //#endif
        return stringSequence;
    }
}
