/**
 *  Copyright 2006-2011, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.nmsp.client2.sdk.components.general;

import com.nuance.nmsp.client2.sdk.components.general.TransactionProcessingException;

/**
 * Interface to either capture the audio buffers streamed 
 * by the text-to-speech (TTS) resource or to send an incoming voice stream.
 * 
 * <p>
 * Implement an <code>NMSPAudioSink</code> interface when you want your
 * application to:
 * <ul>
 * <li>Save the recorded audio samples, for example, to a file.</li>
 * <li>Play the TTS audio samples with the application's own audio player.</li>
 * </ul>
 * Otherwise, the application simply needs to call the
 * {@link #addAudioBuf(byte[], int, int, boolean)} function when:
 * <ul>
 * <li>It has its own recorder. In this case, call the
 * <code>addAudioBuf(buf, int, int, false)</code> method every time the recorder gives you
 * an audio sample.</li>
 * <li>It retrieves audio from a file.</li>
 * </ul>
 * <p>
 * When the last buffer is sent, call <code>addAudioBuf(buf, 0, length, true)</code>.
 * <p>
 * <b><em>Tip:</em></b> For better performance, call
 * {@link #addAudioBuf(byte[], int, int, boolean)} multiple times with small buffers, not
 * one time with a very large buffer. The {@link #addAudioBuf(byte[], int, int, boolean)}
 * function works best when the audio buffer contains 200 to 500 milliseconds of
 * audio. This amount of audio represents a different number of bytes depending
 * on the codec you use. If you have to use one big buffer, your application must
 * ensure that the buffer size does not exceed the server maximum buffer size.
 * To change the default value of the server maximum buffer size, edit the value
 * assigned to the <code>XmodeInitialBufferSize</code> parameter in the
 * <code>NMSPGW.cfg</code> configuration file on the NMSP Gateway host.
 * 
 */
public interface NMSPAudioSink {
	/**
	 * 
	 * Call this method to send an incoming voice stream.
	 * 
	 * Alternatively, implement this method to capture the audio buffer streamed by the NMSP TTS resource.
	 * <p>
	 * <b><em>Note:</em></b> You will know that you
	 * have received the last audio sample when the <code>isLastBuf</code>
	 * argument is set to true. The last audio buffer can also be null.
	 * 
	 * 
	 * @param buffer
	 *            buffer that contains an audio sample; for better performance, the 
	 * sample should contain at least 250 ms of audio; for Speex, a buffer should correspond 
	 * to a <a href="http://speex.org/docs/manual/speex-manual/node7.html#SECTION007100000000000000000">packet</a> 
	 *(several frames) ended with a terminator code 
	 *
	 * @param offset
	 *            the start offset of the data
	 * @param length
	 *            the maximum number of bytes read
	 * @param isLastBuf
	 *            indicates the last audio buffer; set to true when the last audio buffer is sent or received
	 * 
	 * @throws TransactionProcessingException
	 *             thrown when the <code>NMSPAudioSink</code> encounters
	 *             problems processing the audio
	 */
	// #ifdef NMSP_CPL_ENABLED
//@    public void addAudioBuf(byte[] buffer, int offset, int length, boolean isLastBuf, long
//@            recordingStartTime, long recordingEndTime) throws
//@            TransactionProcessingException;
	// #else
	public void addAudioBuf(byte[] buffer, int offset, int length, boolean isLastBuf)
			throws TransactionProcessingException;
	// #endif
}
