/**
 *  Copyright 2006-2011, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.nmsp.client2.sdk.components.resource.common;

import java.util.Vector;

import com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines;
import com.nuance.nmsp.client2.sdk.components.general.Parameter;
import com.nuance.nmsp.client2.sdk.components.resource.internal.common.ManagerImpl;
import com.nuance.nmsp.client2.sdk.oem.MessageSystemOEM;
import com.nuance.nmsp.client2.sdk.oem.NetworkSystemOEM;
import com.nuance.nmsp.client2.sdk.oem.TimerSystemOEM;
//#ifdef NMSP_LOG
import com.nuance.nmsp.client2.sdk.common.oem.api.LogFactory;
import com.nuance.nmsp.client2.sdk.common.oem.api.LogFactory.Log;
//#endif

/**
 * This class has a static method for creating a
 * {@link Manager} interface.
 * 
 * @see ManagerListener
 */
public class ManagerFactory {
    //#ifdef NMSP_LOG
    private static final Log log = LogFactory.getLog(ManagerFactory.class);
    //#endif
    
    private ManagerFactory() {
    }

    /**
     * Creates a {@link Manager} interface.
     * 
     * @param gatewayIP NMSP Gateway IP address or hostname
     * @param gatewayPort NMSP Gateway port
     * @param applicationId application ID provided by Nuance Communications, Inc.
     * @param appKey application private key provided by Nuance Communications, Inc.
     * @param uid the unique id
     * @param inputCodec input codec; see {@link com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec}
     * @param outputCodec output codec; see {@link com.nuance.nmsp.client2.sdk.common.defines.NMSPDefines.Codec}
     * @param parameters vector of the class {@link Parameter}. This is where you can specify 
     * session-related parameters (a manager object represents a 
     * session with NMSP Gateway). Supported parameters types are APP, NMSP, and SDK. 
     * For more information, see the <em>Programmer's Guide</em> section on 
     * session-related parameters.
     * @param managerListener listener object to receive notifications from the Manager
     * 
     * @return Manager
     */
    public static Manager createManager(String gatewayIP, short gatewayPort, String nmaid,
                                        NMSPDefines.Codec inputCodec, NMSPDefines.Codec outputCodec, Vector parameters, 
                                        ManagerListener managerListener) {
        //#ifdef NMSP_LOG
        if(log.isDebugEnabled())
            log.debug("createManager");
        //#endif
        if (gatewayIP == null || gatewayIP.length() == 0) {
            //#ifdef NMSP_LOG
            log.error("NullPointerException gatewayIP is NULL. ");
            //#endif
            throw new NullPointerException("gatewayIP must be provided!");
        }
        if (gatewayPort <= 0) {
            //#ifdef NMSP_LOG
            log.error("IllegalArgumentException gatewayPort is invalid. ");
            //#endif
            throw new IllegalArgumentException("gatewayPort invalid value!");
        }
        if (nmaid == null) {
            //#ifdef NMSP_LOG
            log.error("NullPointerException applicationId is NULL. ");
            //#endif
            throw new NullPointerException("Application id can not be null!");
        }
        if (inputCodec == null) {
            //#ifdef NMSP_LOG
            log.error("NullPointerException inputCodec is NULL. ");
            //#endif
            throw new NullPointerException("inputCodec can not be null!");
        }
        if (outputCodec == null) {
            //#ifdef NMSP_LOG
            log.error("NullPointerException outputCodec is NULL. ");
            //#endif
            throw new NullPointerException("outputCodec can not be null!");
        }
        if (managerListener == null) {
            //#ifdef NMSP_LOG
            log.error("NullPointerException managerListener is NULL. ");
            //#endif
            throw new NullPointerException("managerListener can not be null!");
        }
        if (parameters != null) {
            for (int i = 0; i < parameters.size(); i++) {
                Parameter parameter = (Parameter) parameters.elementAt(i);
                if(parameter.getType() == Parameter.Type.SLOG ||
                   parameter.getType() ==  Parameter.Type.NSSLOG ||
                   parameter.getType() ==  Parameter.Type.GWLOG ||
                   parameter.getType() ==  Parameter.Type.SVSP ||
                   parameter.getType() ==  Parameter.Type.NSS ||
                   parameter.getType() ==  Parameter.Type.SIP ||
                   parameter.getType() ==  Parameter.Type.SDP) {
                    //#ifdef NMSP_LOG
                    log.error("IllegalArgumentException Parameter type: " + parameter.getType() + " not allowed. ");
                    //#endif
                    throw new IllegalArgumentException("Parameter type: " + parameter.getType() + " not allowed. ");
                }
            }
        }
        // TODO
        return new ManagerImpl(gatewayIP, gatewayPort, nmaid, inputCodec, outputCodec, parameters, managerListener);
    }
    
    public static void terminate() {
    	NetworkSystemOEM.terminate();
    	TimerSystemOEM.terminate();
    	MessageSystemOEM.terminate();
    }
}
