/**
 *  Copyright 2006-2011, Nuance Communications Inc. All rights reserved.
 *
 *  Nuance Communications, Inc. provides this document without representation
 *  or warranty of any kind. The information in this document is subject to
 *  change without notice and does not represent a commitment by Nuance
 *  Communications, Inc. The software and/or databases described in this
 *  document are furnished under a license agreement and may be used or
 *  copied only in accordance with the terms of such license agreement.
 *  Without limiting the rights under copyright reserved herein, and except
 *  as permitted by such license agreement, no part of this document may be
 *  reproduced or transmitted in any form or by any means, including, without
 *  limitation, electronic, mechanical, photocopying, recording, or otherwise,
 *  or transferred to information storage and retrieval systems, without the
 *  prior written permission of Nuance Communications, Inc.
 *
 *  Nuance, the Nuance logo, Nuance Recognizer, and Nuance Vocalizer are
 *  trademarks or registered trademarks of Nuance Communications, Inc. or its
 *  affiliates in the United States and/or other countries. All other
 *  trademarks referenced herein are the property of their respective owners.
 **/

package com.nuance.nmsp.client2.sdk.oem;

import java.net.SocketAddress;
import java.net.InetSocketAddress;
import java.util.UUID;

import android.content.Context;
import android.content.SharedPreferences;
import android.os.Build;
import android.telephony.TelephonyManager;

import com.nuance.nmsp.client2.sdk.common.oem.api.DeviceInfo;
import com.nuance.nmsp.client2.sdk.common.oem.api.LogFactory;
import com.nuance.nmsp.client2.sdk.common.oem.api.LogFactory.Log;

public class DeviceInfoOEM implements DeviceInfo  { 
	
	// Note: Permission in manifest <uses-permission android:name="android.permission.READ_PHONE_STATE" />
	private static final String DEVICEINFO = "nuance_sdk_pref";
	
	//#ifdef NMSP_LOG
 private final static Log log = LogFactory.getLog(DeviceInfoOEM.class);
	//#endif
	
	private TelephonyManager mTelephonyMgr;
	private Context _context = null;
	
	// Ctor for those properties does need the Context
	private DeviceInfoOEM(Context context)
	{
		_context = context;
		mTelephonyMgr =  (TelephonyManager)context.getSystemService(Context.TELEPHONY_SERVICE);
	}
	
	// Ctor for those properties does not need the Context
	private DeviceInfoOEM()
	{
	}

	/**
	 * @return DeviceInfo
	 */
	public static synchronized DeviceInfo getInstance(Context context)
	{
		return new DeviceInfoOEM(context);
	}
	
	public static synchronized DeviceInfo getInstance()
	{
		return new DeviceInfoOEM();
	}

	public String getDeviceModel() {		
        return Build.MODEL;
	}

    public String getDeviceSubModel() {
        return Build.DEVICE;
    }

	public String getOperatorName() {
		
		String opName = mTelephonyMgr.getNetworkOperatorName(); 
		return opName == null ? "" : opName;
	}

	public String getPhoneNumber() {
		
		String tel = mTelephonyMgr.getLine1Number(); 
		return tel == null ? "" : tel;	
	}

	public String getUniqueID() {		
		//String Android_ID = System.getString(this.getContentResolver(), System.ANDROID_ID);  
		/* Device ID is not used for uniqued ID any more
		String imei = mTelephonyMgr.getDeviceId();
		return imei == null ? "" : imei ;
		*/
		
		SharedPreferences sdkPref = _context.getSharedPreferences(DEVICEINFO, Context.MODE_PRIVATE);
		String imei = sdkPref.getString("NUANCE_NMSP_UID", "");
		
		if (imei.length() == 0) {
			imei = UUID.randomUUID().toString();
			
			if (imei != null) {
				imei = imei.replaceAll("-", "");
			}
			
			SharedPreferences.Editor editor = sdkPref.edit();
			editor.putString("NUANCE_NMSP_UID", imei);
			
			if (!editor.commit()) {
				//#ifdef NMSP_LOG
				log.error("Storing nuance sdk uid has failed");
				//#endif
			}
		}
		
		imei =  (imei == null) ? "" : imei;
		
		return imei;
		
	}

    public String getOSType() {
        return "Android";
    }

	public String getOSVersion() {
		return Build.VERSION.RELEASE;
	}
	
	public SocketAddress getProxyAddress()
	{
		//#ifdef NMSP_LOG
		if ( log.isDebugEnabled() )
		{
			log.debug("getProxyAddress Build.VERSION.SDK_INT: " + Build.VERSION.SDK_INT);
		}
		//#endif
		
		if ( Build.VERSION.SDK_INT >= 11 ) // Build.VERSION_CODES.VHONEYCOMB
		{
			//#ifdef NMSP_LOG
			if ( log.isDebugEnabled() )
			{
				log.debug("getProxyAddress SDK_INT >= 11");
			}
			//#endif
			
			return getProxyAddressHoneycomb();
		}
		
		if ( _context == null )
		{
			//#ifdef NMSP_LOG
			if ( log.isErrorEnabled() )
			{
				log.error("the android context is needed to retrieve the proxy address from the device.");
			}
			//#endif
			throw new IllegalStateException("the android context is needed to retrieve the proxy address from the device.");
		}

		return getProxyAddressLegacy();
	}

	private SocketAddress makeAddress(String address, int port)
	{
		//#ifdef NMSP_LOG
		if ( log.isDebugEnabled() )
		{
			log.debug("makeAddress -- address: " + address + ", port: " + port);
		}
		//#endif
		
		if (null == address || "".equals(address) || 1 > port )
		{
			//#ifdef NMSP_LOG
			if ( log.isDebugEnabled() )
			{
				log.debug("makeAddress, invalid parameters, return null.");
			}
			//#endif
			
			return null;
		}

		return new InetSocketAddress(address, port);
	}
	
	private SocketAddress getProxyAddressHoneycomb()
	{
		//#ifdef NMSP_LOG
		if ( log.isDebugEnabled() )
		{
			log.debug("getProxyAddressHoneycomb");
		}
		//#endif
		
		final String proxyAddress = System.getProperty("http.proxyHost");
		final String portStr = System.getProperty("http.proxyPort");

		//#ifdef NMSP_LOG
		if ( log.isDebugEnabled() )
		{
			log.debug("retrieved address: " + proxyAddress + ", port: " + portStr);
		}
		//#endif
		
		final int proxyPort = portStr != null ? Integer.parseInt(portStr) : -1;
		
		return makeAddress(proxyAddress, proxyPort);
	}
	
	private SocketAddress getProxyAddressLegacy()
	{
		//#ifdef NMSP_LOG
		if ( log.isDebugEnabled() )
		{
			log.debug("getProxyAddressLegacy");
		}
		//#endif

		final String proxyAddress = android.net.Proxy.getHost(_context);
		final int proxyPort = android.net.Proxy.getPort(_context);
		
		//#ifdef NMSP_LOG
		if ( log.isDebugEnabled() )
		{
			log.debug("getProxyAddressLegacy proxyAddress[" + proxyAddress + "] port[" + proxyPort + "]");
		}
		//#endif

		return makeAddress(proxyAddress, proxyPort);
	}
}
