package com.nuance.nmsp.client2.sdk.oem.socket;

import java.io.BufferedReader;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.UnknownHostException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
//#ifdef NMSP_LOG
import com.nuance.nmsp.client2.sdk.common.oem.api.LogFactory;
import com.nuance.nmsp.client2.sdk.common.oem.api.LogFactory.Log;
//#endif
import com.nuance.nmsp.client2.sdk.oem.DeviceInfoOEM;
import com.nuance.nmsp.client2.sdk.oem.socket.SocketSetting.ProxyOption;
import com.nuance.nmsp.client2.sdk.oem.socket.ssl.NmspSSLSocketFactory;

public class NmspSocketFactory
{
	//#ifdef NMSP_LOG
	private final static Log log = LogFactory.getLog(NmspSocketFactory.class);
	//#endif

	private static class ProxyConnectRunnable implements Runnable
	{
		private final String hostName;
		private final int port;
		private final OutputStream outputStream;

		ProxyConnectRunnable(String hostName, int port, OutputStream outputStream)
		{
			this.hostName = hostName;
			this.port = port;
			this.outputStream = outputStream;
		}

		//Override
		public void run()
		{
			final String proxyRequest = String.format("CONNECT %s:%d HTTP/1.0\r\n\r\n", hostName, port);
			try
			{
				outputStream.write(proxyRequest.getBytes("UTF-8"));
			}
			catch (IOException e)
			{
				//#ifdef NMSP_LOG
				if ( log.isErrorEnabled() )
				{
					log.error("proxy request write error: [" + e.getClass().getName() + "] Message - [" + e.getMessage() + "]");
				}
				//#endif
				try 
				{
					// just close the stream to cause the read
					// to fail as well
					outputStream.close();
				} catch (IOException ioe) {}
			}
		}
	}

	private static Socket handleProxyRequest(SocketAddress proxyAddress, String hostName, int port) throws IOException
	{
		final Socket proxySocket = new Socket();
		try
		{
			proxySocket.connect(proxyAddress, 2000); // 2 seconds as the connection timeout for connecting to the proxy server 
		} catch (Exception e)
		{
			//#ifdef NMSP_LOG
			if ( log.isDebugEnabled() )
			{
				log.warn("proxy socket connecting failed!", e);
			}
			//#endif
			return null;
		}
		
		final Thread t = new Thread(new ProxyConnectRunnable(hostName, port, proxySocket.getOutputStream()));
		t.start();

		final InputStreamReader isr = new InputStreamReader(proxySocket.getInputStream());
		final BufferedReader br = new BufferedReader(isr);

		// RFC2626, Section 6 specifies an HTTP response...
		// The first line is the Status-Line:
		//    Status-Line = HTTP-Version SP Status-Code SP Reason-Phrase CRLF
		// The Status-Code element is a 3-digit integer result code of the attempt to
		// understand and satisfy the request.
		//    HTTP-Version = "HTTP" "/" 1*DIGIT "." 1*DIGIT

		final String statusLine = br.readLine();

		//#ifdef NMSP_LOG
		if ( log.isDebugEnabled() )
		{
			log.debug("proxy status: [" + statusLine + "]");
		}
		//#endif

		boolean proxyConnectSuccess = false;
		// readLine consumed the CRLF
		final Pattern statusLinePattern = Pattern.compile("^HTTP/\\d+\\.\\d+ (\\d\\d\\d) .*");
		final Matcher statusLineMatcher = statusLinePattern.matcher(statusLine);
		if (statusLineMatcher.matches())
		{
			final String statusCode = statusLineMatcher.group(1);
			if (null != statusCode && 0 < statusCode.length() && '2' == statusCode.charAt(0))
			{
				proxyConnectSuccess = true;
			}
		}

		// Consume rest of proxy response
		String line;
		while ( "".equals((line = br.readLine())) == false )
		{
			//#ifdef NMSP_LOG
			if ( log.isDebugEnabled() )
			{
				log.debug("proxy: [" + line + "]");
			}
			//#endif
		}

		if ( proxyConnectSuccess == false )
		{
			throw new IOException("Proxy connect error");
		}

		return proxySocket;
	}

	public static Socket createSocket(SocketSetting setting, String hostName, int port) throws UnknownHostException, IOException, SecurityException
	{
		//#ifdef NMSP_LOG
		if ( log.isDebugEnabled() )
		{
			log.debug("creating socket hostName [" + hostName + "] port [" + port
					+ "] useProxy [" + setting.proxyOption + "] isSSL [" + setting.isSSL + "]");
		}
		//#endif

		if ( setting.proxyOption == ProxyOption.PROXY_ONLY || setting.proxyOption == ProxyOption.PROXY_IF_AVAIL )
		{
			//#ifdef NMSP_LOG
			if ( log.isDebugEnabled() )
			{
				log.debug("checking proxy");
			}
			//#endif

			SocketAddress proxyAddress = null;
			try
			{
				proxyAddress = DeviceInfoOEM.getInstance(setting.context).getProxyAddress();
			}
			catch (RuntimeException e)
			{
				//#ifdef NMSP_LOG
				if ( log.isDebugEnabled() )
				{
					log.debug("Encountered error while trying to get the system proxy address", e);
				}
				//#endif
			}

			//#ifdef NMSP_LOG
			if ( log.isDebugEnabled() )
			{
				log.debug("System proxy address: [" + proxyAddress + "]");
			}
			//#endif

			//
			// the proxyAddress will be null if there
			// is no proxy set up on the device
			//
			if ( null != proxyAddress )
			{
				final Socket proxySocket = handleProxyRequest(proxyAddress, hostName, port);

				if ( proxySocket != null )
				{
					if ( setting.isSSL == true )
					{
						return NmspSSLSocketFactory.createSocket(proxySocket, hostName, port, setting.sslSetting);
					}
					else
					{
						return proxySocket;
					}
				}
				else
				{
					// the proxySocket will be null if the proxy server does not function well
					return null;
				}
			}
			
			if ( setting.proxyOption == ProxyOption.PROXY_ONLY )
			{
				return null;
			}
			// if setting.proxyOption is ProxyOption.PROXY_AVAIL, we will continue to create regular socket
		}
		
		if ( setting.isSSL == true )
		{
			return NmspSSLSocketFactory.createSocket(hostName, port, setting.sslSetting);
		}
		else
		{
			return new Socket(hostName, port);
		}
	}
}
