package com.nuance.dragon.toolkit.nvsl;

/**
 * An enumeration to indicates predefined configuration types for SpeakerID and LanguageID engines.
 * This includes all the configurations currently released, even though some of them might not be available
 * in the specific product version or application installer.
 * A configuration type contains the following information:
 * <p>
 * <br> - The name of engine configuration
 * <br> - It is a language or speaker identification engine
 *
 */
public final class NVSLAbility
{
    // For Speaker identification
    /**
     * This is the most accurate and default engine configuration for
     * text independent speaker recognition applications.
     */
    public final static String NVSL_TI_COMPACT = "NVSLv4_TI_COMPACT_v1.0.nsrpkg";

    /**
     * This is the compact version of the text independent configuration.
     */
    public final static String LVSL_TI_COMPACT = "LVSLv3_TI_COMPACT_v1.0.nsrpkg";

    /**
     * This is Engine for text-independent speaker recognition applications.
     * It exploits the diagonal, approximated, full posterior scoring. Possibly the most accurate
     * text-independent system for short duration utterances. It is NOT compatible with {@link NVSLAbility#NVSL_TI_COMPACT}.
     */
    public final static String NVSLv4_TI_FPD_COMPACT = "NVSLv4_TI_FPD_COMPACT_v1.1.nsrpkg";

    /**
     * This is a widely used text dependent configuration which is very fast
     * and usually provides sufficient accuracy.
     */
    public final static String TD_HMGM = "TD_HMGM_v1.1.nsrpkg";

    /**
     * This is a double technology configuration, combining iVector and HMGM models.
     */
    public final static String TD_IVEC_COMPACT_HMGM = "TD_IVEC-COMPACT_HMGM_v1.1.nsrpkg";

    /**
     * This is 8 KHz NVSL engine configuration suited for mobility variable password scenario.
     * It internally exploits two HMGM subsystems. Very reduced memory footprint.
     */
    public final static String TD_MOBILITY_8KHz_2xHMGM = "TD_MOBILITY_8KHz_2xHMGM_v1.0.nsrpkg";

    /**
     * This is 8 KHz NVSL engine configuration suited for mobility variable password scenario.
     * It is based on the combination of a HMGM and a Joint Factor Analysis subsystem.
     * Reduced memory footprint.
     */
    public final static String TD_MOBILITY_8KHz_JFA_HMGM = "TD_MOBILITY_8KHz_JFA_HMGM_v1.0.nsrpkg";

    /**
     * This is 16 KHz NVSL engine configuration suited for mobility variable password scenario.
     * It internally exploits two HMGM subsystems. Very reduced memory footprint.
     */
    public final static String TD_MOBILITY_16KHz_2xHMGM = "TD_MOBILITY_16KHz_2xHMGM_v1.0.nsrpkg";

    /**
     * This is 16 KHz NVSL engine configuration suited for mobility variable password scenario.
     * It is based on the combination of a HMGM and a Joint Factor Analysis subsystem.
     * Reduced memory footprint.
     */
    public final static String TD_MOBILITY_16KHz_JFA_HMGM = "TD_MOBILITY_16KHz_JFA_HMGM_v1.0.nsrpkg";

    /**
     * It is very effective for liveness detection and short duration utterances.
     * This engine exploits full-posterior i-vector scoring. It requires slightly
     * more CPU than {@link NVSLAbility#TD_TP_IVEC_COMPACT}, and is NOT compatible with it.
     */
    public final static String TD_TP_IVEC_FPD_COMPACT = "TD_TP_IVEC_FPD-COMPACT_v1.1.nsrpkg";

    /**
     * This configuration is used for liveness detection when the highest accuracy is needed.
     * This engine is very expensive in terms of CPU and voiceprint footprints. It exploits
     * the exact algorithm of the full-posterior i-vector scoring. It is NOT compatible with
     * {@link NVSLAbility#TD_TP_IVEC_COMPACT} or {@link NVSLAbility#TD_TP_IVEC_FPD_COMPACT} and lacks a default calibration set
     */
    public final static String TD_TP_IVEC_FPF_COMPACT = "TD_TP_IVEC_FPF-COMPACT_v1.1.nsrpkg";

    /**.
     * This is the compact version of the text dependent and text-prompted configuration.
     */
    public final static String TD_TP_IVEC_COMPACT = "TD_TP_IVEC-COMPACT_v1.1.nsrpkg";

    /**
     * This is a basic text prompted algorithm that is usually used as a basis
     * for fusion with other algorithms.
     */
    public final static String TP_HMM = "TP_HMM_v1.1.nsrpkg";

    // For Language identification
    /**
     * This is the default language identification engine configuration.
     */
    public final static String LVSL_22LANG = "LVSLv3_22Lang_v1.0.nlrpkg";

    /**
     * This is An alternative language identification engine configuration
     * based on a combination of acoustic support vector machine and phonetic
     * neural-network FST N-Gram models.
     */
    public final static String LVSL_WESTERN = "LVSLv3_Western_v1.0.nlrpkg";

    /**
     * This is an alternative language identification engine configuration
     * based on an acoustic support vector machine model.
     */
    public final static String LVSL_WESTERN_FAST = "LVSLv3_WesternFast_v1.0.nlrpkg";
}
