package com.nuance.dragon.toolkit.vocon;

import java.util.HashMap;

import com.nuance.dragon.toolkit.oem.api.internal.Checker;

/**
 * A fuzzy-match data for runtime compilation.
 * <p>
 * There must be a column named 'entry_id' and {@link Flags} must be set to {@link Flags#INTEGER}
 */
public class FuzzyMatchData
{
    /**
     * The optional column flags (mirrored from VoCon Hybrid native API).
     */
    public class Flags
    {
        /** This attribute specifies that the column is speakable. It cannot be combined with Flags.INTEGER. */
        public static final int SPEAKABLE = 0x01;

        /** This attribute specifies that the column can be used in the query string for filtering. */
        public static final int ACTIVATABLE = 0x02;

        /** This attribute specifies that the column contains integers. If not set then the column contains strings. */
        public static final int INTEGER = 0x04;

        /** This attribute specifies that the column contains an array of string or integers. If not set then column has single values. */
        public static final int ARRAY = 0x08;

        /** When this attribute is set then collapsed results will present all the values of this column in an array. */
        public static final int RESULTACCUMULATE = 0x10;
    }

    enum Type
    {
        INTEGER, STRING, SPEAKABLE
    }

    class Value
    {
        final Type type;
        final Object object;
        final int flags;

        Value(Type type, Object object, int flags)
        {
            this.type = type;
            this.object = object;
            this.flags = flags;
        }
    }

    final HashMap<String, Value> data;

    /**
     * FuzzyMatchData constructor.
     */
    public FuzzyMatchData()
    {
        data = new HashMap<String, Value>();
    }

    /**
     * Add an integer value.
     * @param key The key.
     * @param value The value.
     */
    public void addInteger(String key, int value)
    {
        Checker.checkStringArgForEmpty("key", key);
        data.put(key, new Value(Type.INTEGER, value, Flags.ACTIVATABLE | Flags.INTEGER));
    }

    /**
     * Add an integer value.
     * @param key The key.
     * @param value The value.
     * @param flags The optional {@link Flags} combination. Can be 0.
     */
    public void addInteger(String key, int value, int flags)
    {
        Checker.checkStringArgForEmpty("key", key);
        data.put(key, new Value(Type.INTEGER, value, flags));
    }

    /**
     * Add a string value.
     * @param key The key.
     * @param value The value.
     */
    public void addString(String key, String value)
    {
        Checker.checkStringArgForEmpty("key", key);
        Checker.checkStringArgForEmpty("value", value);
        data.put(key, new Value(Type.STRING, value, FuzzyMatchData.Flags.ACTIVATABLE | FuzzyMatchData.Flags.ARRAY));
    }

    /**
     * Add a string value.
     * @param key The key.
     * @param value The value.
     * @param flags The optional {@link Flags} combination. Can be 0.
     */
    public void addString(String key, String value, int flags)
    {
        Checker.checkStringArgForEmpty("key", key);
        Checker.checkStringArgForEmpty("value", value);
        data.put(key, new Value(Type.STRING, value, flags));
    }

    /**
     * Add a a speakable string value.
     * @param key The key.
     * @param value The value.
     */
    public void addSpeakableString(String key, String value)
    {
        Checker.checkStringArgForEmpty("key", key);
        Checker.checkStringArgForEmpty("value", value);
        data.put(key, new Value(Type.SPEAKABLE, value, FuzzyMatchData.Flags.ACTIVATABLE | FuzzyMatchData.Flags.ARRAY | FuzzyMatchData.Flags.SPEAKABLE));
    }

    /**
     * Add a a speakable string value.
     * @param key The key.
     * @param value The value.
     * @param flags The optional {@link Flags} combination. Can be 0.
     */
    public void addSpeakableString(String key, String value, int flags)
    {
        Checker.checkStringArgForEmpty("key", key);
        Checker.checkStringArgForEmpty("value", value);
        data.put(key, new Value(Type.SPEAKABLE, value, flags));
    }
}
